/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

/*
 * Admin Superuser username/password get/set.
 *
 * Get Operation
 * in:  op=get
 * out: admpw.uid=<userid from config/admpw file>
 *
 * Set Operation
 * in:op=set&admpw.uid=<newuid>&admpw.pw=<new password>
 * If admpw.uid is found in cgi parameters, update uid in the config/admpw file
 * If admpw.pw is found in cgi parameters, update password in:
 *   1) "userPassword" attribute in the SIE on the DS
 *   2) siepid: parameter in config/adm.conf
 *   3) config/admpw file. The password is stored in encrypted form
 */

#include "libadmin/libadmin.h"
#include <stdio.h>
#include <stdarg.h>  /* for vsprintf */
#include <stdlib.h>
#include <string.h>

#include "libadminutil/resource.h"
#include "libadminutil/admutil.h"
#include "libadminutil/distadm.h"
#include "libadmsslutil/psetcssl.h"
#include "libadmsslutil/admsslutil.h"

static void output_admuid(char *file);
static void update_uidpwd();
static void update_admpwd(char *newuid, char *newpw);
static void update_adm_conf(char *newpw);
static void update_ds(char *newpw);

/*
 * Logging function
 */
static FILE * logfp;
static int log_enabled = 0;
static void logMsg(char *format, ...) {
	char logfile[512];

	if (!log_enabled) return;

	if (logfp==NULL) {
		sprintf(logfile, "%s/../logs/admpw.dbg", getenv("ADMSERV_ROOT"));
		logfp = fopen(logfile, "w");
	}

	if (logfp != NULL) {
		va_list ap;
		va_start(ap,format);
		vfprintf(logfp,format,ap);
		va_end(ap);
		fflush(logfp);
	}
}

/*
 * i18n conversions defines and function;
 * properties file  = "admpw.properties"
 */
#define RESOURCE_FILE "admpw"
#define resource_key(a,b)   a b

#define DBT_NO_METHOD 			resource_key(RESOURCE_FILE, "1")
#define DBT_UNKNOWN_METHOD 		resource_key(RESOURCE_FILE, "2")
#define DBT_NO_QUERY_STRING 		resource_key(RESOURCE_FILE, "3")
#define DBT_NO_OPERATION 		resource_key(RESOURCE_FILE, "4")
#define DBT_UNKNOWN_OPERATION		resource_key(RESOURCE_FILE, "5")
#define DBT_OPEN_ADMPW       		resource_key(RESOURCE_FILE, "6")
#define DBT_READ_ADMPW            	resource_key(RESOURCE_FILE, "7")
#define DBT_ADMPW_CORRUPTED		resource_key(RESOURCE_FILE, "8")
#define DBT_NO_PARAM			resource_key(RESOURCE_FILE, "9")
#define DBT_OPEN_ADMPW_WR		resource_key(RESOURCE_FILE, "10")
#define DBT_WR_ADMPW			resource_key(RESOURCE_FILE, "11")
#define DBT_OPEN_ADMCONF_RD		resource_key(RESOURCE_FILE, "12")
#define DBT_NO_SIEPID            	resource_key(RESOURCE_FILE, "13")
#define DBT_OPEN_ADMCONF_WR		resource_key(RESOURCE_FILE, "14")
#define DBT_NO_USERNAME			resource_key(RESOURCE_FILE, "15")
#define DBT_NO_USERDN			resource_key(RESOURCE_FILE, "16")
#define DBT_SECURITY_INIT		resource_key(RESOURCE_FILE, "17")
#define DBT_PSET_CREATE			resource_key(RESOURCE_FILE, "18")
#define DBT_PSET_ERROR			resource_key(RESOURCE_FILE, "19")
#define DBT_PSET_SET			resource_key(RESOURCE_FILE, "20")

static  char *acceptLanguage = (char*)"en";
static  Resource *i18nResource = NULL;

static void i18nInit() {
	char resPath[256], *execPath;
	execPath = getcwd(resPath, 256);
	if (execPath) {
		strcpy(resPath, execPath);
		strcat(resPath, "/property");
		i18nResource = res_init_resource(resPath, RESOURCE_FILE);
	}
	if (getenv("HTTP_ACCEPT_LANGUAGE")) {
		acceptLanguage = getenv("HTTP_ACCEPT_LANGUAGE");
	}
}
static void i18nCleanup() {
	if (i18nResource != NULL) {
	res_destroy_resource(i18nResource);
	}
}	
static const char *i18nMsg(char *msgid, char *defaultMsg) {
	const char *msg=NULL;

	if (i18nResource) {
            msg = res_getstring(i18nResource, msgid, acceptLanguage);
      }
	if (msg == NULL) {
		msg = (const char*)defaultMsg;
	}
	return msg;
}

/*
 *  Main
 */
int main(int argc, char *argv[])
{
    int _ai=ADMUTIL_Init();

    char *method = getenv("REQUEST_METHOD");
    char *root = getenv("NETSITE_ROOT");
    char *qs = 0, *op=0;
    char error_info[128];
    char filename[BIG_LINE];
	    
    logMsg(" In %s\n", argv[0]);

    i18nInit();	
    sprintf(filename, "%s/admpw", getenv("ADMSERV_ROOT"));

	/* GET or POST method */
    if (!method || !*method) {
        /* non UI CGI */
        rpt_err(SYSTEM_ERROR, i18nMsg(DBT_NO_METHOD,"No method is specified"), NULL, NULL);
    }
    else if (!strcmp(method, "GET")) {
        /* UI CGI - either upgrade or normal admin CGI */
        qs = getenv("QUERY_STRING");
	
        if (!qs || !*qs) {
            rpt_err(INCORRECT_USAGE, i18nMsg(DBT_NO_QUERY_STRING, "NO QUERY_STRING DATA"), NULL, NULL);
        }
        else {
            get_begin(qs);
        }
    }
    else if (!strcmp(method, "POST")) {
        post_begin(stdin);
    }
    else {
        sprintf(error_info, i18nMsg(DBT_UNKNOWN_METHOD, "Unknown Method (%s)"), method);
        rpt_err(INCORRECT_USAGE, error_info, NULL, NULL);
    }

	logMsg("method=%s\n", method);
	
    op=get_cgi_var("op", NULL, NULL);
	
	logMsg("op=%s\n", op);

	if (op == NULL) {
        rpt_err(INCORRECT_USAGE, i18nMsg(DBT_NO_OPERATION, "No operation defined"), NULL, NULL);
    }
	else if (strcmp(op,"get")==0) {
		output_admuid(filename);
	}
	else if (strcmp(op,"set")==0) {		
		update_uidpwd();		
	}
	else {
        sprintf(error_info, i18nMsg(DBT_UNKNOWN_OPERATION, "Unknown Operation (%s)"), op);
        rpt_err(INCORRECT_USAGE, error_info, NULL, NULL);
    }
	
   exit(0);
   return 0;
}



/*
 * Output user name from the admpwd. The password does not need to be sent
 */

static void output_admuid(char *filename) {
	FILE *f = fopen(filename, "r");
	char *uid=NULL ,*pw=NULL, *col=NULL;
	static char inbuf[BIG_LINE];

	if (f==NULL) {
		fclose(f);
		rpt_err(SYSTEM_ERROR, i18nMsg(DBT_OPEN_ADMPW, "Can not open admpw file"), NULL, NULL);
	}
	if (fgets(inbuf, sizeof(inbuf), f) == NULL) {
		fclose(f);
		rpt_err(SYSTEM_ERROR, i18nMsg(DBT_READ_ADMPW, "Can not read admpw file"), NULL, NULL);
	}
	logMsg("line=%s@\n", inbuf);
	fclose(f);

	col = strchr(inbuf, ':');
	if (col == NULL) {
		rpt_err(SYSTEM_ERROR, i18nMsg(DBT_ADMPW_CORRUPTED, "admpw file corrupted"), NULL, NULL);
	}
	
	uid = inbuf; *col=0; pw=col+1;
	logMsg("uid=%s, pw=%s\n", uid, pw);

	fprintf(stdout, "Content-type: text/html\n\n");
    fprintf(stdout, "admpw.uid: %s\n", uid);
}

/*
 * Modify local superuser username and password
 */
static void update_uidpwd() {

	FILE *f;
	char *uid=NULL ,*pw=NULL, *col=NULL;
	char *newuid=NULL, *newpw=NULL;
	static char filename[BIG_LINE];
	static char inbuf[BIG_LINE];

	sprintf(filename, "%s/admpw", getenv("ADMSERV_ROOT"));

	newuid = get_cgi_var("admpw.uid",NULL, NULL);
	newpw  = get_cgi_var("admpw.pw",NULL, NULL);
	if (newuid==NULL && newpw==NULL) {
		rpt_err(INCORRECT_USAGE, i18nMsg(DBT_NO_PARAM, "No parameters to set"), NULL, NULL);
	}

	logMsg("newuid=%s, newpw=%s\n", newuid ? newuid:"", newpw ? newpw : "");

	if ((f=fopen(filename, "r"))==NULL) {
		fclose(f);
		rpt_err(SYSTEM_ERROR, i18nMsg(DBT_OPEN_ADMPW,"Can not open admpw file"), NULL, NULL);
	}
	if (fgets(inbuf, sizeof(inbuf), f) == NULL) {
		fclose(f);
		rpt_err(SYSTEM_ERROR, i18nMsg(DBT_READ_ADMPW, "Can not read admpw file"), NULL, NULL);
	}
	logMsg("line=%s@\n", inbuf);
	fclose(f);

	col = strchr(inbuf, ':');
	if (col == NULL) {
		rpt_err(SYSTEM_ERROR, i18nMsg(DBT_ADMPW_CORRUPTED,"admpw file corrupted"), NULL, NULL);
	}
	
	uid = inbuf; *col=0; pw=col+1;
	logMsg("uid=%s, pw=%s\n", uid, pw);
	
	if (newuid == NULL) {
		newuid = uid;
	}

	if (newpw != NULL) {
		update_ds(newpw);
		update_adm_conf(newpw);
		update_admpwd(newuid, admin40_pw_enc(newpw));
	}
	else {
		update_admpwd(newuid, pw);
	}

	rpt_success(NULL);
}

/*
 * Modify admpw.
 */
static void update_admpwd(char *newuid, char *newpw) {

	FILE *f;
	int cnt;
	static char filename[BIG_LINE];
	static char inbuf[BIG_LINE];
	static char outbuf[BIG_LINE];

	sprintf(filename, "%s/admpw", getenv("ADMSERV_ROOT"));

	f = fopen(filename, "w");
	if (f==NULL) {
		fclose(f);
		rpt_err(SYSTEM_ERROR, i18nMsg(DBT_OPEN_ADMPW_WR, "Can not open admpw file for writing"), NULL, NULL);
	}

	sprintf(outbuf,"%s:%s", newuid, newpw);
	logMsg("output=%s\n", outbuf);

	cnt = fprintf(f,"%s", outbuf);

	if (cnt != (int)strlen(outbuf)) {
		rpt_err(SYSTEM_ERROR, i18nMsg(DBT_WR_ADMPW, "Failed to write to admpw"), NULL, NULL);
	}
	
	fclose(f);
}

/*
 * Modify adm.conf
 */
static void update_adm_conf(char *newpw) {

	FILE *f;
	int i, modified=0;
	static char filename[BIG_LINE];
	char buf[80];
	int linecnt=0;
	char *lines[50];
	static char inbuf[BIG_LINE];


	sprintf(filename, "%s/adm.conf", getenv("ADMSERV_ROOT"));

	f = fopen(filename, "r");
	if (f==NULL) {
		fclose(f);
		rpt_err(SYSTEM_ERROR, i18nMsg(DBT_OPEN_ADMCONF_RD, "Can not open adm.conf for reading"), NULL, NULL);
	}

	while(fgets(inbuf, sizeof(inbuf), f) != NULL) {
		if (strstr(inbuf,"siepid:") == inbuf) { /* Line starts with "Password:" */
			sprintf(buf, "siepid: %s\n", newpw);
			lines[linecnt++] = strdup(buf);
			modified=1;
		}
		else {
			lines[linecnt++] = strdup(inbuf);
		}
	}	
	fclose(f);

	if (!modified) {
		rpt_err(SYSTEM_ERROR, i18nMsg(DBT_NO_SIEPID, "Parameter siepid: not found in adm.conf"), NULL, NULL);
	}

	f = fopen(filename, "w");
	if (f==NULL) {
		fclose(f);
		rpt_err(SYSTEM_ERROR, i18nMsg(DBT_OPEN_ADMCONF_WR, "Can not open adm.conf for writing"), NULL, NULL);
	}

	for (i=0; i < linecnt; i++) {
		fprintf(f, "%s", lines[i]);
	}

	fclose(f);

}

/*
 * Modify userpassword in the DS
 */
static void update_ds(char *pwd) {
  
  int   err, rv, errorCode;
  PsetHndl pset;
  char *username = 0;
  char *localAdmin = 0;
  char *binddn = 0;
  char *bindpw = 0;
  char  error_info[128];

  /* Get UserDN and User Password  */

  rv = ADM_GetUserDNString(&err, &binddn);
  if (rv < 0 || !binddn || !*binddn) {
    rv = ADM_GetCurrentUsername(&err, &username);
    if (rv < 0 || !username || !*username) {
      rpt_err(ELEM_MISSING, i18nMsg(DBT_NO_USERNAME, "No User Name"), NULL, NULL);
    }
    else {
      /* No DN, maybe it is local super */
      localAdmin = admGetLocalAdmin(NULL, &rv);
      if (localAdmin) {
	if (strcmp(username, localAdmin)) {
	  rpt_err(ELEM_MISSING, i18nMsg(DBT_NO_USERDN, "No User DN"), NULL, NULL);
	}
	else {
	  binddn = NULL;
	  bindpw = NULL;
	}
      }
      else {
	rpt_err(ELEM_MISSING, i18nMsg(DBT_NO_USERDN, "No User DN"), NULL, NULL);
      }
    }
  }

  if (binddn) rv = ADM_GetCurrentPassword(&err, &bindpw);

  /* Initialize the pset  */

  errorCode = ADMSSL_InitSimple();
  if (errorCode) {
    rpt_err(APP_ERROR, i18nMsg(DBT_SECURITY_INIT, "Security Initialization Failed"), NULL, NULL);
  }

  pset = psetCreateSSL("admin-serv", 
		       /* configRoot */ NULL,
		       /* userDN */ binddn, 
		       /* passwd */ bindpw,
		       /* errorcode */ &rv);

  if (!pset) {
    sprintf(error_info, i18nMsg(DBT_PSET_ERROR,"PSETERROR: %d"), rv);
    rpt_err(APP_ERROR, i18nMsg(DBT_PSET_CREATE, "PSET Creation Failed"), NULL, error_info);
  }

  errorCode = psetSetSingleValueAttr(pset, "userpassword", pwd);

  if (errorCode) {
    logMsg("psetErr=%d\n", errorCode);
    psetDelete(pset);
    sprintf(error_info, i18nMsg(DBT_PSET_ERROR,"PSETERROR: %d"), errorCode);
    rpt_err(APP_ERROR, "PSET SET Failed", NULL, error_info);
  }
}
