/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * dsconfig.c -- CGI configuration update handler -- directory gateway
 *
 * Copyright (c) 1996 Netscape Communications Corp.
 * All rights reserved.
 */

 /*
 * Default Directory Server setting get/set. Read/modify DS ldap url from dbswitch.conf
 *
 * Get Operation
 * in:  op=getconfig
 * out: dsconfig.host=<host from DS ldap url>
 * out: dsconfig.port=<port from DS ldap url>
 * out: dsconfig.basedn=<dn from DS ldap url>
 * out: dsconfig.sll= 'true' if ldaps used, 'false' if ldap used
 *
 * Set Operation
 * in:op=setconfig&dsconfig.host=<host>&dsconfig.port=<port>&
 * dsconfig.basedn=<basedn>&dsconfig.ssl=<true|false>
 *   1) modify DS ldap url in dbswitch.conf with supplied parameters
 *   2) modify ldapHost: and ldapPort: in adm.conf
 */

#include "libadmin/libadmin.h"
#include "string.h"

#include "ldaputil/errors.h"
#include "ldaputil/ldaputil.h"
#include "ldaputil/dbconf.h"

#include "libadminutil/resource.h"
#include "libadminutil/admutil.h"
#include "libadminutil/distadm.h"

#include "libadmsslutil/psetcssl.h"
#include "libadmsslutil/admsslutil.h"
#include "libadmsslutil/certmgt.h"
#include "cert.h"

/*#ifdef XP_WIN32
#define strcasecmp stricmp
#define strncasecmp _strnicmp
#endif*/

static void handle_getconfig();
static void handle_setconfig();
static char *create_new_ldapurl(char *new_host, int new_port, char *new_basedn, int new_ssl);
static void update_adm_conf(char *ldapHost, int ldapPort, char *sieDN, char *isieDN);
static void update_dbswitch_conf(char *ldapURL);
static char *nonull_value(char *str);


/*
 * Logging function
 */
static FILE * logfp;
static int log_enabled = 0;
static void logMsg(char *format, ...) {
  char logfile[512];

  if (!log_enabled) return;

  if (logfp==NULL) {
    sprintf(logfile, "%s/../logs/dsconfig.dbg", getenv("ADMSERV_ROOT"));
    logfp = fopen(logfile, "w");
  }

  if (logfp != NULL) {
    va_list ap;
    va_start(ap,format);
    vfprintf(logfp,format,ap);
    va_end(ap);
    fflush(logfp);
  }
}

/*
 * i18n conversions defines and function;
 * properties file  = "dsconfig.properties"
 */
#define RESOURCE_FILE "dsconfig"
#define resource_key(a,b)   a b

#define DBT_NO_METHOD 			resource_key(RESOURCE_FILE, "1")
#define DBT_UNKNOWN_METHOD 		resource_key(RESOURCE_FILE, "2")
#define DBT_NO_QUERY_STRING 		resource_key(RESOURCE_FILE, "3")
#define DBT_NO_OPERATION 		resource_key(RESOURCE_FILE, "4")
#define DBT_UNKNOWN_OPERATION		resource_key(RESOURCE_FILE, "5")
#define DBT_RD_DBSWITCH			resource_key(RESOURCE_FILE, "6")
#define DBT_BAD_DBSWITCH		resource_key(RESOURCE_FILE, "7")
#define DBT_BAD_LDAPURL			resource_key(RESOURCE_FILE, "8")
#define DBT_BAD_PORT			resource_key(RESOURCE_FILE, "9")
#define DBT_NO_CERTDB	       	resource_key(RESOURCE_FILE, "10")
#define DBT_OPEN_CERTDB	       	resource_key(RESOURCE_FILE, "11")
#define DBT_BAD_SSL       		resource_key(RESOURCE_FILE, "12")
#define DBT_OPEN_DBSWITCH_RD 		resource_key(RESOURCE_FILE, "13")
#define DBT_NO_DBHANDLE      		resource_key(RESOURCE_FILE, "14")
#define DBT_OPEN_DBSWITCH_WR 		resource_key(RESOURCE_FILE, "15")
#define DBT_OPEN_ADM_RD     		resource_key(RESOURCE_FILE, "16")
#define DBT_NO_LDAPHOST  		resource_key(RESOURCE_FILE, "17")
#define DBT_NO_LDAPPORT			resource_key(RESOURCE_FILE, "18")
#define DBT_OPEN_ADM_WR     		resource_key(RESOURCE_FILE, "19")
#define DBT_NO_SIE     		        resource_key(RESOURCE_FILE, "20")
#define DBT_NO_ISIE     		resource_key(RESOURCE_FILE, "21")


static  char *acceptLanguage = "en";
static  Resource *i18nResource = NULL;

static void i18nInit() {
  char resPath[256], *execPath;
  execPath = getcwd(resPath, 256);
  if (execPath) {
    strcpy(resPath, execPath);
    strcat(resPath, "/property");
    i18nResource = res_init_resource(resPath, RESOURCE_FILE);
  }
  if (getenv("HTTP_ACCEPT_LANGUAGE")) {
    acceptLanguage = getenv("HTTP_ACCEPT_LANGUAGE");
  }
}
static void i18nCleanup() {
  if (i18nResource != NULL) {
    res_destroy_resource(i18nResource);
  }
}	
static const char *i18nMsg(char *msgid, char *defaultMsg) {
  const char *msg=NULL;

  if (i18nResource) {
    msg = res_getstring(i18nResource, msgid, acceptLanguage);
  }
  if (msg == NULL) {
    msg = (const char*)defaultMsg;
  }
  return msg;
}

static char * nonull_value(char *str) {
  return (str!=NULL) ? str : (char *)"";
}

/*
 *  Main
 */
int main(int argc, char *argv[])
{
  int _ai=ADMUTIL_Init();

  char *method = getenv("REQUEST_METHOD");
  char *root = getenv("NETSITE_ROOT");
  char *qs = 0, *op=0;
  char error_info[128];

  logMsg(" In %s\n", argv[0]);

  i18nInit();

  /* GET or POST method */
  if (!method || !*method) {
    /* non UI CGI */
    rpt_err(SYSTEM_ERROR, i18nMsg(DBT_NO_METHOD,"No method is specified"), NULL, NULL);
  }
  else if (!strcmp(method, "GET")) {
    /* UI CGI - either upgrade or normal admin CGI */
    qs = getenv("QUERY_STRING");

    if (!qs || !*qs) {
      rpt_err(INCORRECT_USAGE, i18nMsg(DBT_NO_QUERY_STRING ,"NO QUERY_STRING DATA"), NULL, NULL);
    }
    else {
      get_begin(qs);
    }
  }
  else if (!strcmp(method, "POST")) {
    post_begin(stdin);
  }
  else {
    sprintf(error_info, i18nMsg(DBT_UNKNOWN_METHOD,"Unknown Method (%s)"), method);
    rpt_err(INCORRECT_USAGE, error_info, NULL, NULL);
  }

  logMsg("method=%s\n", method);

  op=get_cgi_var("op", NULL, NULL);

  logMsg("op=%s\n", op);

  if (op == NULL) {
    rpt_err(INCORRECT_USAGE, i18nMsg(DBT_NO_OPERATION,"No operation is defined"), NULL, NULL);
  }
  if (strcmp(op,"getconfig") == 0) {
    handle_getconfig();
  }
  else if (strcmp(op,"setconfig") == 0) {
    handle_setconfig();
  }
  else {
    sprintf(error_info, i18nMsg(DBT_UNKNOWN_OPERATION,"Unknown Operation (%s)"), op);
    rpt_err(INCORRECT_USAGE, error_info, NULL, NULL);
  }

  return 0;
}


/*
 * Return current ldap url setting
 */
static void handle_getconfig()
{
  DBConfDBInfo_t *db_info;
  int		rc;
  LDAPURLDesc	*ludp;
  char dbswitch_file[BIG_LINE];

  logMsg("In handle_getconfig\n");

  sprintf(dbswitch_file, "%s/shared/config/dbswitch.conf", getenv("NETSITE_ROOT"));

  logMsg("dbswitch=%s\n", dbswitch_file);

  if (( rc = dbconf_read_default_dbinfo(dbswitch_file, &db_info ))!= LDAPU_SUCCESS ) {
    logMsg("dbconf_read_default_dbinfo(%s), rc=%d\n", dbswitch_file, rc);
    rpt_err(SYSTEM_ERROR, i18nMsg(DBT_RD_DBSWITCH, "Failed to read dbswitch.conf"), NULL, NULL);
  }
  if ( db_info == NULL ) {
    logMsg("dbconf_read_default_dbinfo(), returned db_info=NULL\n");
    rpt_err(SYSTEM_ERROR, i18nMsg(DBT_RD_DBSWITCH, "Failed to read dbswitch.conf"), NULL, NULL);
  }

  if ( strcasecmp( db_info->dbname, DBCONF_DEFAULT_DBNAME ) != 0 ) {
    logMsg("db_info->dbname is '%s', expecting '%s'\n", db_info->dbname, DBCONF_DEFAULT_DBNAME);
    rpt_err(SYSTEM_ERROR, i18nMsg(DBT_BAD_DBSWITCH, "Bad content/format of dbswitch.conf"), NULL, NULL);
  }
  logMsg("baseurl=%s\n", db_info->url);

  if (( rc = ldap_url_parse( db_info->url, &ludp )) != 0 ) {
    logMsg("ldap_url_parse(%s) failed, rc=%d\n", db_info->url, rc);
    rpt_err(SYSTEM_ERROR, i18nMsg(DBT_BAD_LDAPURL,"Bad ldap url in dbswitch.conf"), NULL, NULL);
  }
  else {
    int ssl;

    fprintf(stdout, "Content-type: text/html\n\n");

    fprintf(stdout, "dsconfig.host:%s\n", nonull_value(ludp->lud_host));
    logMsg("dsconfig.host:%s\n", nonull_value(ludp->lud_host));

    fprintf(stdout, "dsconfig.port:%d\n", ludp->lud_port);
    logMsg("dsconfig.port:%d\n", ludp->lud_port);

    ssl = (ludp->lud_options & LDAP_URL_OPT_SECURE);
    fprintf(stdout, "dsconfig.ssl:%s\n", ssl ? "true" : "false");
    logMsg("dsconfig.ssl:%s\n", ssl ? "true" : "false");


    fprintf(stdout, "dsconfig.basedn:%s\n", nonull_value(ludp->lud_dn));
    logMsg("dsconfig.basedn:%s\n", nonull_value(ludp->lud_dn));

    fprintf(stdout, "NMC_Status: 0\n");

  }
}

/*
 * Modify ldap url setting
 */
static void handle_setconfig() {
  char *sieDN=NULL, *isieDN=NULL, *host=NULL, *basedn=NULL, *ldapURL=NULL, *port_s=NULL, *ssl_s=NULL, *alias=NULL;
  int ssl=-1, port=-1;
  char keyfile[512], certfile[512];


  /* basedn can not be changed any more  Miodrag 08/21/98
     basedn = get_cgi_var( "dsconfig.basedn", NULL, NULL );
  */
  sieDN  = get_cgi_var( "dsconfig.sieDN", NULL, NULL );
  isieDN = get_cgi_var( "dsconfig.isieDN", NULL, NULL );
  host   = get_cgi_var( "dsconfig.host", NULL, NULL );
  port_s = get_cgi_var( "dsconfig.port", NULL, NULL );
  ssl_s  = get_cgi_var( "dsconfig.ssl", NULL, NULL );
  alias  = get_cgi_var( "dsconfig.alias", NULL, NULL );

  if (port_s != NULL) {
    port   = atoi( port_s);
    if (port < 1 || port > 65535) {
      rpt_err(INCORRECT_USAGE, i18nMsg(DBT_BAD_PORT,"Bad value for dsconfig.port"), NULL, NULL);
    }
  }
  if (ssl_s != NULL) {
    if (strcasecmp(ssl_s, "true") == 0) {
      if(!alias)
	rpt_err(INCORRECT_USAGE, i18nMsg(DBT_NO_CERTDB,"No certificate database specified for LDAP/SSL usage!"), NULL, NULL);
      ssl = 1;
      sprintf(keyfile, "alias%c%s-key3.db",
	      FILE_PATHSEP,
	      alias);
      sprintf(certfile, "alias%c%s-cert8.db",
	      FILE_PATHSEP,
	      alias);
      set_security(NULL, getenv("NETSITE_ROOT"), NULL, certfile, keyfile);
    }
    else if (strcasecmp(ssl_s, "false") == 0) {
      ssl = 0;
    }
    else {
      rpt_err(INCORRECT_USAGE, i18nMsg(DBT_BAD_SSL,"Bad value for dsconfig.ssl"), NULL, NULL);
    }
  }

  ldapURL = create_new_ldapurl(host, port, basedn, ssl);

  logMsg("new ldap url = %s", nonull_value(ldapURL));

  if (ldapURL != NULL) {
    update_dbswitch_conf(ldapURL);
    update_adm_conf(host, port, sieDN, isieDN);

    fprintf(stdout, "Content-type: text/html\n\n");
    fprintf(stdout, "NMC_Status: 0\n");
  }
}


/*
 * Create new ldap url
 */
static char* create_new_ldapurl(char *new_host, int new_port, char *new_basedn, int new_ssl)
{
  DBConfDBInfo_t *db_info;
  int		rc;
  LDAPURLDesc	*ludp;
  int sslflag;
  char url[BIG_LINE];
  char *host, *basedn, *ssl;
  int port;
  char dbswitch_file[BIG_LINE];

  sprintf(dbswitch_file, "%s/shared/config/dbswitch.conf", getenv("NETSITE_ROOT"));

  if (( rc = dbconf_read_default_dbinfo(dbswitch_file, &db_info ))!= LDAPU_SUCCESS ) {
    logMsg("dbconf_read_default_dbinfo(%s), rc=%d\n", dbswitch_file, rc);
    rpt_err(SYSTEM_ERROR, i18nMsg(DBT_RD_DBSWITCH,"Failed to read dbswitch.conf"), NULL, NULL);
  }
  if ( db_info == NULL ) {
    logMsg("dbconf_read_default_dbinfo(), returned db_info=NULL\n");
    rpt_err(SYSTEM_ERROR, i18nMsg(DBT_RD_DBSWITCH,"Failed to read dbswitch.conf"), NULL, NULL);
  }

  if ( strcasecmp( db_info->dbname, DBCONF_DEFAULT_DBNAME ) != 0 ) {
    logMsg("db_info->dbname is '%s', expecting '%s'\n", db_info->dbname, DBCONF_DEFAULT_DBNAME);
    rpt_err(SYSTEM_ERROR, i18nMsg(DBT_BAD_DBSWITCH,"Bad content/format of dbswitch.conf"), NULL, NULL);
  }
  logMsg("baseurl=%s\n", db_info->url);

  if (( rc = ldap_url_parse( db_info->url, &ludp )) != 0 ) {
    logMsg("ldap_url_parse(%s) failed, rc=%d\n", db_info->url, rc);
    rpt_err(SYSTEM_ERROR, i18nMsg(DBT_BAD_LDAPURL,"Bad ldap url in dbswitch.conf"), NULL, NULL);
  }

  else {

    host = (new_host != NULL) ? new_host : ludp->lud_host;
    port = (new_port != -1) ? new_port : ludp->lud_port;
    basedn = (new_basedn != NULL) ? new_basedn : ludp->lud_dn;
    sslflag = (new_ssl != -1) ? new_ssl : (ludp->lud_options & LDAP_URL_OPT_SECURE);
    ssl = (sslflag) ? (char *)"s" : (char *)"";

    sprintf(url, "ldap%s://%s:%d/%s", ssl, host, port, basedn);
  }
  return strdup(url);
}

static char inbuf[BIG_LINE];

/*
 * Modify dbswitch.conf
 */
static void update_dbswitch_conf(char *ldapURL) {

  FILE *f;
  int i, modified=0;
  static char filename[BIG_LINE];
  char buf[80];
  int linecnt=0;
  char *lines[100];


  sprintf(filename, "%s/shared/config/dbswitch.conf", getenv("NETSITE_ROOT"));

  f = fopen(filename, "r");
  if (f==NULL) {
    rpt_err(SYSTEM_ERROR, i18nMsg(DBT_OPEN_DBSWITCH_RD,"Can not open dbswitch.conf for reading"), NULL, NULL);
  }

  while(fgets(inbuf, sizeof(inbuf), f) != NULL) {
    if (strncasecmp(inbuf,"directory default ", strlen("directory default ")) == 0) {
      sprintf(buf, "directory default %s\n", ldapURL);
      lines[linecnt++] = strdup(buf);
      modified=1;
    }
    else {
      lines[linecnt++] = strdup(inbuf);
    }

  }
  fclose(f);

  if (!modified) {
    rpt_err(SYSTEM_ERROR, i18nMsg(DBT_NO_DBHANDLE,"Dbhandle 'default' not found in dbswitch.conf"), NULL, NULL);
  }


  f = fopen(filename, "w");
  if (f==NULL) {
    rpt_err(SYSTEM_ERROR, i18nMsg(DBT_OPEN_DBSWITCH_WR,"Can not open dbswitch.conf for writing"), NULL, NULL);
    return;
  }

  for (i=0; i < linecnt; i++) {
    fprintf(f, "%s", lines[i]);
  }

  fclose(f);

}

/*
 * Modify adm.conf
 */
static void update_adm_conf(char *ldapHost, int ldapPort, char *sieDN, char *isieDN) {

  FILE *f;
  int i, modifiedHost=0, modifiedPort=0, modifiedSIE=0, modifiedISIE=0;
  static char filename[BIG_LINE];
  char buf[500];
  int linecnt=0;
  char *lines[50];

  sprintf(filename, "%s/adm.conf", getenv("ADMSERV_ROOT"));

  f = fopen(filename, "r");
  if (f==NULL) {
    rpt_err(SYSTEM_ERROR, i18nMsg(DBT_OPEN_ADM_RD, "Can not open adm.conf for reading"), NULL, NULL);
  }

  while(fgets(inbuf, sizeof(inbuf), f) != NULL) {
    if (strstr(inbuf,"ldapHost:") == inbuf) { /* Line starts with "ldapHost:" */
      sprintf(buf, "ldapHost: %s\n", ldapHost);
      lines[linecnt++] = strdup(buf);
      modifiedHost=1;
    }
    else if (strstr(inbuf,"ldapPort:") == inbuf) { /* Line starts with "ldapPort:" */
      sprintf(buf, "ldapPort: %d\n", ldapPort);
      lines[linecnt++] = strdup(buf);
      modifiedPort=1;
    }
    else if ((strncasecmp(inbuf, "sie:", strlen("sie:")) == 0) &&
	     (sieDN)) {                            /* Line starts with "sie:" */
      sprintf(buf, "sie: %s\n", sieDN);
      lines[linecnt++] = strdup(buf);
      modifiedSIE=1;
    }
    else if ((strncasecmp(inbuf, "isie:", strlen("isie:")) == 0) &&
	     (isieDN)) {                           /* Line starts with "isie:" */
      sprintf(buf, "isie: %s\n", isieDN);
      lines[linecnt++] = strdup(buf);
      modifiedISIE=1;
    }
    else {
      lines[linecnt++] = strdup(inbuf);
    }

  }
  fclose(f);

  if (!modifiedHost) {
    rpt_err(SYSTEM_ERROR, i18nMsg(DBT_NO_LDAPHOST, "Parameter ldapHost: not found in adm.conf"), NULL, NULL);
  }

  if (!modifiedPort) {
    rpt_err(SYSTEM_ERROR, i18nMsg(DBT_NO_LDAPPORT, "Parameter ldapPort: not found in adm.conf"), NULL, NULL);
  }

  if ((sieDN) && (!modifiedSIE)) {
    rpt_err(SYSTEM_ERROR, i18nMsg(DBT_NO_SIE, "Parameter sie: not found in adm.conf"), NULL, NULL);
  }

  if ((isieDN) && (!modifiedISIE)) {
    rpt_err(SYSTEM_ERROR, i18nMsg(DBT_NO_ISIE, "Parameter isie: not found in adm.conf"), NULL, NULL);
  }

  f = fopen(filename, "w");
  if (f==NULL) {
    rpt_err(SYSTEM_ERROR, i18nMsg(DBT_OPEN_ADM_WR, "Can not open adm.conf for writing"), NULL, NULL);
  }

  for (i=0; i < linecnt; i++) {
    fprintf(f, "%s", lines[i]);
  }

  fclose(f);
}
