/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * mergeConfig.cpp
 *
 * Merges all SIE entry configuration info to a target Directory Server
 * All blame to Adam Prishtina (adam@netscape.com)
 */

#include <stdlib.h>
#include <stdio.h>
#include "seccomon.h"
#include "secport.h"
#include "libadminutil/distadm.h"
#include "libadminutil/admutil.h"
#include "libadminutil/resource.h"
#include "libadmsslutil/psetcssl.h"
#include "libadmsslutil/admsslutil.h"
#include "libadminutil/resource.h"
#include "ldapu.h"
#include <ldap_ssl.h>


/*
   properties file name = "mergeConfig.properties"
*/
#define RESOURCE_FILE "mergeConfig"

#define resource_key(a,b)   a b

/* define strings here */
#define DBT_INVALID_METHOD               resource_key(RESOURCE_FILE, "1")
#define DBT_INVALID_SOURCE_SR            resource_key(RESOURCE_FILE, "2")
#define DBT_ADMSSL_INITFAIL              resource_key(RESOURCE_FILE, "3")
#define DBT_NO_DEST_HOST                 resource_key(RESOURCE_FILE, "4")
#define DBT_NO_DEST_PORT                 resource_key(RESOURCE_FILE, "5")
#define DBT_NO_DEST_SECURITY             resource_key(RESOURCE_FILE, "6")
#define DBT_NO_DEST_BINDDN               resource_key(RESOURCE_FILE, "7")
#define DBT_NO_DEST_BINDPW               resource_key(RESOURCE_FILE, "8")
#define DBT_NO_SOURCE_GROUPDN            resource_key(RESOURCE_FILE, "9")
#define DBT_SOURCE_LDAPAUTH_FAIL         resource_key(RESOURCE_FILE, "10")
#define DBT_DEST_LDAPAUTH_FAIL           resource_key(RESOURCE_FILE, "11")
#define DBT_SIE_DATA_FAIL                resource_key(RESOURCE_FILE, "12")
#define DBT_SIE_CREATE_FAIL              resource_key(RESOURCE_FILE, "13")
#define DBT_SOURCE_PSET_FAIL             resource_key(RESOURCE_FILE, "14")
#define DBT_DEST_PSET_FAIL               resource_key(RESOURCE_FILE, "15")
#define DBT_REPLICATE_SIE_FAIL           resource_key(RESOURCE_FILE, "16")
#define DBT_MISSING_USERNAME             resource_key(RESOURCE_FILE, "17")
#define DBT_MISSING_USERDN               resource_key(RESOURCE_FILE, "18")
#define DBT_NO_DEST_DOMAIN               resource_key(RESOURCE_FILE, "19")
#define DBT_GLOBAL_PREF_FAIL             resource_key(RESOURCE_FILE, "20")


Resource       *i18nResource;
char           *acceptLanguage;

char*
getResourceString(char *key) {
  return (char *)(res_getstring(i18nResource, key, acceptLanguage));
}


struct isie_data {

  char *source_isieDN;
  char *dest_isieDN;
  struct sie_data *sie_list;
  struct isie_data *next;

} isie_data;

struct sie_data {

  char *sieDN;
  char *hostname;
  char *server_root;
  struct sie_data *next;

} sie_data;


#define FREESTRING(x) if(x) free(x);

/*
 * Delete the isie and sie data structs.
 * If delete_from_ds is true, an error occurred in mergeConfig and
 * we delete the new entries from the dest DS.
 */
void delete_isie_data(struct isie_data *isie_list, Ldap *ldap_dest, int delete_from_ds) {

  struct isie_data *current_isie;
  struct isie_data *delete_isie;

  struct sie_data *current_sie;
  struct sie_data *delete_sie;

  current_isie = isie_list;
  while(current_isie != NULL) {

    if(delete_from_ds && current_isie->dest_isieDN)
      removeISIE(ldap_dest, current_isie->dest_isieDN);

    FREESTRING(current_isie->source_isieDN);
    FREESTRING(current_isie->dest_isieDN);
    
    current_sie = current_isie->sie_list;
    while(current_sie != NULL) {

      FREESTRING(current_sie->sieDN);
      FREESTRING(current_sie->hostname);
      FREESTRING(current_sie->server_root);

      delete_sie = current_sie;
      current_sie = current_sie->next;
      free(delete_sie);
    }

    delete_isie = current_isie;
    current_isie = current_isie->next;
    free(delete_isie);
  }
}


int get_isie_data(char *host,
		  int port, 
		  int security, 
		  char *servergroup_dn,
		  char *binddn, 
		  char *bindpw, 
		  struct isie_data **isie_list) {


  int ver = LDAP_VERSION3;
  LDAP *server;
  int ldapError;
  LDAPMessage *isie_entry, *sie_entry;
  LDAPMessage *result, *result2;

  struct isie_data *current_isie;
  struct sie_data *current_sie;

  char *isie_dn, *sie_dn;
  char **vals;

  *isie_list = NULL;

  if(security) {
    if(!(server = ldapssl_init(host, port, 1)))
      return -1;
  }
  else {
    if(!(server = ldap_init(host, port)))
      return -1;
  }

  int returnval = ldap_set_option(server, LDAP_OPT_PROTOCOL_VERSION, &ver);
  if(returnval != LDAP_SUCCESS)
    return -1;

  if ((ldapError = ldap_simple_bind_s(server, binddn, bindpw))
      != LDAP_SUCCESS ) {
    switch (ldapError) {
    case LDAP_INAPPROPRIATE_AUTH:
    case LDAP_INVALID_CREDENTIALS:
    case LDAP_INSUFFICIENT_ACCESS:
      /* authenticate failed: Should not continue */
#ifdef LDAP_DEBUG
      ldap_perror( ld, "ldap_simple_bind_s" );
#endif
      return -1;
    case LDAP_NO_SUCH_OBJECT:
    case LDAP_ALIAS_PROBLEM:
    case LDAP_INVALID_DN_SYNTAX:
      /* Not a good DN */
      return -1;
    default:
      ldap_unbind(server);
      return -1;
    }
  }


  if((ldapError = ldap_search_s(server, servergroup_dn, LDAP_SCOPE_ONELEVEL,
				"(objectclass=*)", NULL, 0, &result)) != LDAP_SUCCESS)
    return -1;
  
  for(isie_entry = ldap_first_entry(server, result);
      isie_entry != NULL;
      isie_entry = ldap_next_entry(server, isie_entry)) {

    if(!(*isie_list)) {
      *isie_list = (struct isie_data *)malloc(sizeof(struct isie_data));
      current_isie = *isie_list;
      current_isie->next = NULL;
    }
    else {
      current_isie->next = (struct isie_data *)malloc(sizeof(struct isie_data));
      current_isie = current_isie->next;
      current_isie->next = NULL;
    }    
    
    isie_dn = ldap_get_dn(server, isie_entry);
    current_isie->source_isieDN = strdup(isie_dn);
    ldap_memfree(isie_dn);

    current_isie->dest_isieDN = NULL;
    current_isie->sie_list = NULL;

    if((ldapError = ldap_search_s(server, current_isie->source_isieDN, LDAP_SCOPE_ONELEVEL,
				  "(objectclass=netscapeServer)", NULL, 0, &result2)) != LDAP_SUCCESS)
      return -1;
    
    for(sie_entry = ldap_first_entry(server, result2);
	sie_entry != NULL;
	sie_entry = ldap_next_entry(server, sie_entry)) {
      
      if(!current_isie->sie_list) {
	current_isie->sie_list = (struct sie_data *)malloc(sizeof(struct sie_data));
	current_sie = current_isie->sie_list;
	current_sie->next = NULL;
      }
      else {
	current_sie->next = (struct sie_data *)malloc(sizeof(struct sie_data));
	current_sie = current_sie->next;
	current_sie->next = NULL;
      }
      
      sie_dn = ldap_get_dn(server, sie_entry);      
      current_sie->sieDN = strdup(sie_dn);
      ldap_memfree(sie_dn);
      
      if((vals = ldap_get_values(server, sie_entry, "serverhostname")) != NULL) {
	current_sie->hostname = strdup(vals[0]);
	ldap_value_free(vals);
      }
      else
	return -1;
      
      if((vals = ldap_get_values(server, sie_entry, "serverroot")) != NULL) {
	current_sie->server_root = strdup(vals[0]);
	ldap_value_free(vals);
      }
      else
	return -1;
    } /* SIE processing */

  } /* ISIE processing */

  return 0;
}


/* 
 * Returns 1 if at least one value of an attributed was replaced with new group dn,
 * 0 if nothing was changed.
 */

int replace_dn(char **attribute_list, char *source_groupDN, char *dest_groupDN) {

  char replace_string[BIG_LINE];
  char *temp_string;
  char *start_ptr, *end_ptr;

  int i, replaced=0;

  if((!attribute_list) ||
     (!source_groupDN) ||
     (!dest_groupDN))
     return 0;

  for(i=0; attribute_list[i]; i++) {

    /* no matches */
    if(!strstr(attribute_list[i], source_groupDN))
      continue;

    replaced=1;

    /* exact match */
    if(!strcmp(attribute_list[i], source_groupDN)) {
      free(attribute_list[i]);
      attribute_list[i] = strdup(dest_groupDN);
      continue;
    }

    /* at least one substring */
    strcpy(replace_string, "");
    temp_string = strdup(attribute_list[i]);
    start_ptr = temp_string;
    end_ptr = strstr(temp_string, source_groupDN);

    while((start_ptr && *start_ptr != '\0') && 
	  (end_ptr && *end_ptr != '\0')) {
      end_ptr[0] = '\0';
      strcat(replace_string, start_ptr);
      strcat(replace_string, dest_groupDN);
      start_ptr = end_ptr + strlen(source_groupDN);
      if(start_ptr && *start_ptr != '\0')
	end_ptr = strstr(start_ptr, source_groupDN);
    }
    if(start_ptr && *start_ptr!= '\0')
      strcat(replace_string, start_ptr);

    free(temp_string);
    free(attribute_list[i]);
    attribute_list[i] = strdup(replace_string);
  }

  return replaced;
}


#define KEEP_SOURCE 0
#define KEEP_DEST 1
#define MULTI_MERGE 2
#define STRING_MERGE 3
#define ADD_IF_EMPTY 4
#define NO_VALID_RULE 5

int find_merge_rule(char **merge_rules, char *attrName) {

  // merge_rules is an array of the form:
  // <attr name>: [KEEP_SOURCE | KEEP_DEST | MULTI_MERGE | STRING_MERGE | ADD_IF_EMPTY]
  
  int count = 0;
  char temp[BIG_LINE];

  if(!strchr(merge_rules[0], '{')) {

     // node rule found - it overrides attribute rules
     if(strstr(merge_rules[0], "KEEP_SOURCE"))
       return KEEP_SOURCE;
     else if(strstr(merge_rules[0], "KEEP_DEST"))
       return KEEP_DEST;
     else if(strstr(merge_rules[0], "MULTI_MERGE"))
       return MULTI_MERGE;
     else if(strstr(merge_rules[0], "STRING_MERGE"))
       return STRING_MERGE;
     else if(strstr(merge_rules[0], "ADD_IF_EMPTY"))
       return ADD_IF_EMPTY;
     else
       // invalid merge rule
       return NO_VALID_RULE;
  }
     
  while(merge_rules[count]) {
    sprintf(temp, "{%s}", attrName);
    if(strstr(merge_rules[count], temp))
      break;
    count++;
  }

  if(merge_rules[count]) {
    // merge rule found
    if(strstr(merge_rules[count], "KEEP_SOURCE"))
      return KEEP_SOURCE;
    else if(strstr(merge_rules[count], "KEEP_DEST"))
      return KEEP_DEST;
    else if(strstr(merge_rules[count], "MULTI_MERGE"))
      return MULTI_MERGE;
    else if(strstr(merge_rules[count], "STRING_MERGE"))
      return STRING_MERGE;
    else if(strstr(merge_rules[count], "ADD_IF_EMPTY"))
      return ADD_IF_EMPTY;
    else
      // invalid merge rule
      return NO_VALID_RULE;
  }
   
  // no merge rule found for this attr
  return NO_VALID_RULE;
}


char**
admutil_strsdup(char** orig) {
  int  cnt = 0, i = 0;
  char **dest, **tmpptr;
  
  if (!orig) return NULL;

  tmpptr = orig;
  while (*tmpptr++ != NULL) cnt++;

  dest = (char**)malloc((cnt+1)*sizeof(char*));

  for (i=0; i < cnt; i++) dest[i] = strdup(orig[i]);

  dest[cnt] = NULL;

  return dest;
}


LDAPMod*
create_mod( char* namep, char **values, int mode)
{
 
  LDAPMod       *newMod;
  
  newMod = (LDAPMod*)malloc(sizeof(LDAPMod));
  newMod->mod_op = mode;
  newMod->mod_type = strdup(namep);
  if (values) newMod->mod_values = admutil_strsdup(values);
  else newMod->mod_values = NULL;

  return newMod;
}


int recursive_mergeGlobalPrefs(LDAP *source_server, 
			       char *source_pref_dn,
			       LDAP *dest_server,
			       char *dest_pref_dn) {

  LDAPMessage *source_entry;
  LDAPMessage *source_result, *dest_result;

  int ldapError;
  char **merge_rules;
  char *attrName;

  char **source_values, **dest_values;
  char **merge_values;

  LDAPMod **merged_mods;
  int attr_count, count, count2;

  BerElement *ber;

  char *next_source_dn, next_dest_dn[BIG_LINE];
  char temp[BIG_LINE], *temp2;

  int ldap_action;

  if((ldapError = ldap_search_s(source_server, source_pref_dn, LDAP_SCOPE_BASE,
				"(objectclass=*)", NULL, 0, &source_result)) != LDAP_SUCCESS)
    return -1;

  ldapError = ldap_search_s(dest_server, dest_pref_dn, LDAP_SCOPE_BASE,
				"(objectclass=*)", NULL, 0, &dest_result);
  if(ldapError == LDAP_SUCCESS)
    // destination node exists; we will be modifying it
    ldap_action = LDAP_MOD_REPLACE;
  else if(ldapError == LDAP_NO_SUCH_OBJECT) {
    // destination node does not exist; we will be adding it
    ldap_action = LDAP_MOD_ADD;
    attr_count=0;
    merged_mods = NULL;
  }
  else
    // something serious happened
    return -1;

  // PROCESS CURRENT NODE
  
  merge_rules = ldap_get_values(source_server, source_result, "nsmerge");
  if(merge_rules) {
    // at least one merge rule found for this node; process it

    for (attrName = ldap_first_attribute(source_server, source_result, &ber);
	 attrName != NULL;
	 attrName = ldap_next_attribute(source_server, source_result, ber)) {
      
      // PROCESS ALL ATTRIBUTES IN THE NODE

      if(ldap_action == LDAP_MOD_REPLACE) {
	
	if (strcmp( attrName, "objectclass") &&
	    strcmp( attrName, "cn") &&
	    strcmp( attrName, "aci") &&
	    strcmp( attrName, "creatorsname") &&
	    strcmp( attrName, "createtimestamp") &&
	    strcmp( attrName, "modifytimestamp") &&
	    strcmp( attrName, "modifiersname" )) {
	  
	  // get source and dest values
	  source_values = ldap_get_values(source_server, source_result, attrName);
	  dest_values = ldap_get_values(dest_server, dest_result, attrName);
	  
	  if(!source_values)
	    continue;
	  
	  // destination node exists
	  switch(find_merge_rule(merge_rules, attrName)) {
          case KEEP_SOURCE : {
	    merged_mods = (LDAPMod **)malloc(2*sizeof(LDAPMod *));
	    merged_mods[0] = create_mod(attrName, source_values, ldap_action);
	    merged_mods[1] = NULL;
	    
	    if(((ldapError = ldap_modify_s(dest_server, dest_pref_dn, merged_mods)) != LDAP_SUCCESS) &&
	       (ldapError != LDAP_TYPE_OR_VALUE_EXISTS))
	      return -1;
	    break;
	  }
	  case MULTI_MERGE : {
	    // append dest multi-valued attribute values to source multi-valued attribute values
	    count=0;
	    while(source_values[count]) {
	      if(count==0)
		merge_values = (char **)malloc(2*sizeof(char *));
	      else
		merge_values = (char **)realloc(merge_values, (count+2)*sizeof(char *));
	      merge_values[count] = strdup(source_values[count]);
	      count++;
	    }
	    count2=0;
	    while(dest_values && dest_values[count2]) {
	      if(count==0)
		merge_values = (char **)malloc(2*sizeof(char *));
	      else
		merge_values = (char **)realloc(merge_values, (count+2)*sizeof(char *));
	      merge_values[count++] = strdup(dest_values[count2++]);
	    }
	    merge_values[count] = NULL;

	    merged_mods = (LDAPMod **)malloc(2*sizeof(LDAPMod *));
	    merged_mods[0] = create_mod(attrName, merge_values, ldap_action);
	    merged_mods[1] = NULL;

	    if(((ldapError = ldap_modify_s(dest_server, dest_pref_dn, merged_mods)) != LDAP_SUCCESS) &&
	       (ldapError != LDAP_TYPE_OR_VALUE_EXISTS))
	      return -1;
	    break;
	  }
	  case STRING_MERGE : {
	    // string append dest attribute to source attribute (single-valued attributes only)
	    merge_values = (char **)malloc(2*sizeof(char *));
	    merge_values[0] = (char *)malloc(strlen(source_values[0]) + (dest_values && strcmp(dest_values[0], "") ? strlen(dest_values[0]) : 0) + 2);
	    strcpy(merge_values[0], source_values[0]);
	    if(dest_values && strcmp(dest_values[0], "")) {
	      strcat(merge_values[0], ",");
	      strcat(merge_values[0], dest_values[0]);
	    }
	    merge_values[1] = NULL;

	    merged_mods = (LDAPMod **)malloc(2*sizeof(LDAPMod *));
	    merged_mods[0] = create_mod(attrName, merge_values, ldap_action);
	    merged_mods[1] = NULL;

	    if(((ldapError = ldap_modify_s(dest_server, dest_pref_dn, merged_mods)) != LDAP_SUCCESS) &&
	       (ldapError != LDAP_TYPE_OR_VALUE_EXISTS))
	      return -1;
	    break;
	  }
	  case ADD_IF_EMPTY : {
	    // KEEP_SOURCE if no destination value exists
	    if(!dest_values) {

	      merged_mods = (LDAPMod **)malloc(2*sizeof(LDAPMod *));
	      merged_mods[0] = create_mod(attrName, source_values, ldap_action);
	      merged_mods[1] = NULL;

	      if(((ldapError = ldap_modify_s(dest_server, dest_pref_dn, merged_mods)) != LDAP_SUCCESS) &&
		 (ldapError != LDAP_TYPE_OR_VALUE_EXISTS))
		return -1;
	    }
	    break;
	  }
	  default : {
	    break;
	    // NO_VALID_RULE or KEEP_DEST
	    // do nothing, i.e. keep the destination value (even if it doesn't exist)
	  }
	  } // switch statement
	}
      }
      else {
	// dest node doesn't exist

	// get source values
	source_values = ldap_get_values(source_server, source_result, attrName);
	
	if(!source_values)
	  continue;

	if(strcmp( attrName, "aci") &&
	   strcmp( attrName, "creatorsname") &&
	   strcmp( attrName, "createtimestamp") &&
	   strcmp( attrName, "modifytimestamp") &&
	   strcmp( attrName, "modifiersname" )) {
	  
	  switch(find_merge_rule(merge_rules, attrName)) {
	  case KEEP_DEST : {
	    // this makes no sense since no dest node exists - ignore this node completely
	    goto done;
	  }
	  default : {
	    // in all other cases, any type of merge means KEEP_SOURCE
	    // NO_VALID_RULE causes a merge as well, since all other attributes need to merge (at least so far)
	    if(!merged_mods) {
	      merged_mods = (LDAPMod **)malloc(2*sizeof(LDAPMod *));
	      merged_mods[1] = NULL;
	    }
	    else {
	      merged_mods = (LDAPMod **)realloc(merged_mods, (attr_count+2)*sizeof(LDAPMod *));
	      merged_mods[attr_count+1] = NULL;
	    }
	    merged_mods[attr_count] = create_mod(attrName, source_values, ldap_action);
	    attr_count++;
	    break;
	  }
	  } // switch statement

	} // valid attribute to merge

      } // node exists

    } // process each attribute

    // if the destination node didn't exist, add the complete node now
    if(ldap_action == LDAP_MOD_ADD) {
      if((ldapError = ldap_add_s(dest_server, dest_pref_dn, merged_mods)) != LDAP_SUCCESS)
	return -1;
    }

  } // if merge rules exist

  // now process the children
done:
  if((ldapError = ldap_search_s(source_server, source_pref_dn, LDAP_SCOPE_ONELEVEL,
				"(objectclass=*)", NULL, 0, &source_result)) != LDAP_SUCCESS)
    return -1;

  for(source_entry = ldap_first_entry(source_server, source_result);
      source_entry != NULL;
      source_entry = ldap_next_entry(source_server, source_entry)) {
    
    next_source_dn = ldap_get_dn(source_server, source_entry);
    strcpy(temp, next_source_dn);
    temp2 = strchr(temp, ',');
    temp2[2] = '\0'; // include the space
    strcpy(next_dest_dn, temp);
    strcat(next_dest_dn, dest_pref_dn);
    
    if(recursive_mergeGlobalPrefs(source_server,
				  next_source_dn,
				  dest_server,
				  next_dest_dn) == -1)
      return -1;
    
  } // processing all children

  return 0;
}


/*
 * mergeGlobalPrefs()
 * Merge the "Global Preferences" section of the tree
 */


int mergeGlobalPrefs(char *source_host,
		     int source_port,
		     int source_security,
		     char *source_binddn,
		     char *source_bindpw,
		     char *source_pref_dn,
		     char *dest_host,
		     int dest_port,
		     int dest_security,
		     char *dest_binddn,
		     char *dest_bindpw,
		     char *dest_pref_dn) {

  int ver = LDAP_VERSION3;
  LDAP *source_server, *dest_server;
  int ldapError;

  int returnval;

  // INIT SOURCE DS

  if(source_security) {
    if(!(source_server = ldapssl_init(source_host, source_port, 1)))
      return -1;
  }
  else {
    if(!(source_server = ldap_init(source_host, source_port)))
      return -1;
  }

  returnval = ldap_set_option(source_server, LDAP_OPT_PROTOCOL_VERSION, &ver);
  if(returnval != LDAP_SUCCESS)
    return -1;

  if ((ldapError = ldap_simple_bind_s(source_server, source_binddn, source_bindpw))
      != LDAP_SUCCESS ) {
    switch (ldapError) {
    case LDAP_INAPPROPRIATE_AUTH:
    case LDAP_INVALID_CREDENTIALS:
    case LDAP_INSUFFICIENT_ACCESS:
      /* authenticate failed: Should not continue */
#ifdef LDAP_DEBUG
      ldap_perror( ld, "ldap_simple_bind_s" );
#endif
      return -1;
    case LDAP_NO_SUCH_OBJECT:
    case LDAP_ALIAS_PROBLEM:
    case LDAP_INVALID_DN_SYNTAX:
      /* Not a good DN */
      return -1;
    default:
      ldap_unbind(source_server);
      return -1;
    }
  }

  // INIT DEST DS

  if(dest_security) {
    if(!(dest_server = ldapssl_init(dest_host, dest_port, 1)))
      return -1;
  }
  else {
    if(!(dest_server = ldap_init(dest_host, dest_port)))
      return -1;
  }

  returnval = ldap_set_option(dest_server, LDAP_OPT_PROTOCOL_VERSION, &ver);
  if(returnval != LDAP_SUCCESS)
    return -1;

  if ((ldapError = ldap_simple_bind_s(dest_server, dest_binddn, dest_bindpw))
      != LDAP_SUCCESS ) {
    switch (ldapError) {
    case LDAP_INAPPROPRIATE_AUTH:
    case LDAP_INVALID_CREDENTIALS:
    case LDAP_INSUFFICIENT_ACCESS:
      /* authenticate failed: Should not continue */
#ifdef LDAP_DEBUG
      ldap_perror( ld, "ldap_simple_bind_s" );
#endif
      return -1;
    case LDAP_NO_SUCH_OBJECT:
    case LDAP_ALIAS_PROBLEM:
    case LDAP_INVALID_DN_SYNTAX:
      /* Not a good DN */
      return -1;
    default:
      ldap_unbind(dest_server);
      return -1;
    }
  }

  return(recursive_mergeGlobalPrefs(source_server, source_pref_dn, dest_server, dest_pref_dn));
}


int main(int argc, char *argv[])
{
  int            _ai = ADMUTIL_Init();
  PsetHndl       pset_source, pset_dest;
  char           *method;
  int            rv;

  char admroot[BIG_LINE];
  char *server_root = NULL;
  char error_info[BIG_LINE];
  AdmldapInfo ldapInfo = NULL;
  char line[BIG_LINE];

  struct isie_data *source_isie_list, *source_isie_ptr;
  struct sie_data *source_sie_ptr;

  int            err;
  int            *errp = &err;
  char *source_binddn = 0;
  char *source_bindpw = 0;
  char *username = 0;
  char *localAdmin = 0;

  char *source_groupdn;
  char *dest_domain;
  char *dest_host;
  char *dest_port;
  char *dest_secure;
  char *dest_binddn;
  char *dest_bindpw;
  char source_url[BIG_LINE], dest_url[BIG_LINE];
  char *dest_basedn;
  char *dest_group_dn;

  char response[4096], sie_response[4096], isie_response[4096];
  char *temp, *temp2;  

  method = getenv("REQUEST_METHOD");

  if(!method)
    return 1;
  if(!strcmp(method, "GET")) {

    server_root = getenv("NETSITE_ROOT");
    
    /* ---------------------- INIT i18n ---------------------------------------- */

    char *lang=getenv("HTTP_ACCEPT_LANGUAGE");
    sprintf(line, "%s%cbin/admin/admin/bin/property", server_root, FILE_PATHSEP);
    i18nResource = res_init_resource(line, RESOURCE_FILE);
    acceptLanguage = "en";
    if (lang) acceptLanguage = strdup(lang);

    /* ---------------------- INIT SOURCE DS ---------------------------------- */

    /* Get bind DN and bind PW */
    
    rv = ADM_GetUserDNString(&err, &source_binddn);
    if (rv < 0 || !source_binddn || !*source_binddn) {
      rv = ADM_GetCurrentUsername(&err, &username);
      if (rv < 0 || !username || !*username)
	rpt_err(ELEM_MISSING, getResourceString(DBT_MISSING_USERNAME), NULL, NULL);
      else {
	/* No DN, maybe it is local super */
	localAdmin = admGetLocalAdmin(NULL, &rv);
	if (localAdmin) {
	  if (PORT_Strcmp(username, localAdmin))
	    rpt_err(ELEM_MISSING, getResourceString(DBT_MISSING_USERDN), NULL, NULL);
	  else {
	    source_binddn = NULL;
	    source_bindpw = NULL;
	  }
	}
	else
	  rpt_err(ELEM_MISSING, getResourceString(DBT_MISSING_USERDN), NULL, NULL);
      }
    }
    
    if (source_binddn) rv = ADM_GetCurrentPassword(&err, &source_bindpw);

    sprintf(admroot, "%s%cadmin-serv%cconfig", 
	    server_root,
	    FILE_PATHSEP,
	    FILE_PATHSEP);
    
    ldapInfo = admldapBuildInfo(admroot, &rv);
    
    if (!ldapInfo) {
      rpt_err(INCORRECT_USAGE, getResourceString(DBT_INVALID_SOURCE_SR), NULL, NULL);
    }
    
    rv = ADMSSL_InitSimple();
    if (rv) {
      sprintf(error_info, getResourceString(DBT_ADMSSL_INITFAIL), rv);
      rpt_err(APP_ERROR, error_info, NULL, NULL);
    }
    
    /* ---------------------- READ TARGET DS INPUT ---------------------------- */
    
    char *qs = getenv("QUERY_STRING");
    get_begin(qs);

    dest_domain = get_cgi_var("dest_domain", "", "");
    if(!dest_domain)
      rpt_err(ELEM_MISSING, getResourceString(DBT_NO_DEST_DOMAIN), NULL, NULL);

    dest_host = get_cgi_var("dest_host", "", "");
    if(!dest_host)
      rpt_err(ELEM_MISSING, getResourceString(DBT_NO_DEST_HOST), NULL, NULL);
    
    dest_port = get_cgi_var("dest_port", "", "");
    if(!dest_port)
      rpt_err(ELEM_MISSING, getResourceString(DBT_NO_DEST_PORT), NULL, NULL);
    
    dest_secure = get_cgi_var("dest_secure", "", "");
    if(!dest_secure)
      rpt_err(ELEM_MISSING, getResourceString(DBT_NO_DEST_SECURITY), NULL, NULL);
    
    dest_binddn = get_cgi_var("dest_binddn", "", "");
    if(!dest_binddn)
      rpt_err(ELEM_MISSING, getResourceString(DBT_NO_DEST_BINDDN), NULL, NULL);
    
    dest_bindpw = get_cgi_var("dest_bindpw", "", "");
    if(!dest_bindpw)
      rpt_err(ELEM_MISSING, getResourceString(DBT_NO_DEST_BINDPW), NULL, NULL);

    source_groupdn = get_cgi_var("source_groupdn", "", "");
    if(!source_groupdn)
      rpt_err(ELEM_MISSING, getResourceString(DBT_NO_SOURCE_GROUPDN), NULL, NULL);


    char *dest_basedn = (char *)malloc((strlen(dest_domain) + 20) * sizeof(char));
    sprintf(dest_basedn, "ou=%s, o=NetscapeRoot", dest_domain);

    temp = strdup(source_groupdn);
    char *source_basedn = strtok(temp, ",");
    source_basedn = strtok(NULL, ",");
    source_basedn = strtok(NULL, "\0");

    if(source_basedn[0] == ' ')
      source_basedn = &(source_basedn[1]);

    sprintf(dest_url, "ldap%s://%s:%s/%s",
	    (!strcmp(dest_secure, "1") ? "s" : ""),
	    dest_host,
	    dest_port,
	    dest_basedn);

    sprintf(source_url, "ldap%s://%s:%d/%s",
	    ((admldapGetSecurity(ldapInfo)) ? "s" : ""),
	    admldapGetHost(ldapInfo),
	    admldapGetPort(ldapInfo),
	    source_basedn);


    /* -------------------- CREATE DEST SIE ENTRIES ---------------------- */
    
    LdapError ldapError;
    
    Ldap ldap_source(ldapError, 
		     source_url,
		     source_binddn,
		     source_bindpw,
		     NULL,
		     NULL);
    
    if(ldapError != OKAY)
      rpt_err(APP_ERROR, getResourceString(DBT_SOURCE_LDAPAUTH_FAIL), NULL, NULL);
    
    Ldap ldap_dest(ldapError, 
		   dest_url,
		   dest_binddn,
		   dest_bindpw,
		   NULL,
		   NULL);
    
    if(ldapError != OKAY)
      rpt_err(APP_ERROR, getResourceString(DBT_DEST_LDAPAUTH_FAIL), NULL, NULL);
    
    
    rv = get_isie_data(admldapGetHost(ldapInfo),
		       admldapGetPort(ldapInfo),
		       admldapGetSecurity(ldapInfo),
		       source_groupdn,
		       source_binddn,
		       source_bindpw,
		       &source_isie_list);
    
    if(rv) {
      delete_isie_data(source_isie_list, NULL, 0);
      rpt_err(APP_ERROR, getResourceString(DBT_SIE_DATA_FAIL), NULL, NULL);
    }
    

    strcpy(sie_response, "");
    strcpy(isie_response, "");

    for(source_isie_ptr = source_isie_list; 
	source_isie_ptr != NULL; 
	source_isie_ptr = source_isie_ptr->next) {
      
      /* LOOP THROUGH EACH ISIE IN THE SOURCE GROUP DN */

      int isie_created = 0;
      LdapEntry *appEntry = new LdapEntry(&ldap_source);
      
      if((ldapError = appEntry->retrieve(source_isie_ptr->source_isieDN)) != OKAY) {
	delete_isie_data(source_isie_list, &ldap_dest, 1);
	rpt_err(APP_ERROR, getResourceString(DBT_SIE_DATA_FAIL), NULL, NULL);
      }
      
      /* change ldap ptr from source to dest */
      appEntry->change_ldap(&ldap_dest);

      /* Remove ACI's and uniquemember attributes... */
      appEntry->deleteAttribute("aci");
      appEntry->deleteAttribute("uniquemember");

      for(source_sie_ptr = source_isie_ptr->sie_list;
	  source_sie_ptr != NULL;
	  source_sie_ptr = source_sie_ptr->next) {

	/* LOOP THROUGH EACH SIE IN THE CURRENT ISIE */

	LdapEntry *sieEntry = new LdapEntry(&ldap_source);
	
	if((ldapError = sieEntry->retrieve(source_sie_ptr->sieDN)) != OKAY) {
	  delete_isie_data(source_isie_list, &ldap_dest, 1);
	  rpt_err(APP_ERROR, getResourceString(DBT_SIE_DATA_FAIL), NULL, NULL);
	}
	
	/* change ldap ptr from source to dest */
	sieEntry->change_ldap(&ldap_dest);
	
	/* Remove ACI's and uniquemember attributes... */
	sieEntry->deleteAttribute("aci");
	sieEntry->deleteAttribute("uniquemember");

	/* call createSIE to find a new program group */
	ldapError = createSIE(sieEntry, appEntry, source_sie_ptr->hostname,
			      source_sie_ptr->server_root,
			      dest_basedn);

	/* record the dest ISIE DN in case the CGI fails and needs to undo changes */
	if(!source_isie_ptr->dest_isieDN && appEntry->entryDN())
	  source_isie_ptr->dest_isieDN = strdup(appEntry->entryDN());

	if(ldapError != OKAY) {
	  delete_isie_data(source_isie_list, &ldap_dest, 1);
	  rpt_err(APP_ERROR, getResourceString(DBT_SIE_CREATE_FAIL), NULL, NULL);    
	}


	/* 
	 * Get dest group DN.
	 * For replacing destination attribute values.
	 */
	
	temp2 = strdup(sieEntry->entryDN());
	dest_group_dn = strtok(temp2, ",");
	dest_group_dn = strtok(NULL, ",");
	dest_group_dn = strtok(NULL, "\0");
	
	if(dest_group_dn[0] == ' ')
	  dest_group_dn = &(dest_group_dn[1]);
	
	/* 
	 * now that we know the destination server group, change all dest
	 * sie and isie attributes that contain the old server group info.
	 */

	char **source_sie_attrnames = sieEntry->getAttributeNames();
	char **source_sie_attrvals;
	
	int i;
	
	/* Replace all source program group DN's with dest group DN's. */
	for(i=0; source_sie_attrnames[i]; i++) {
	  source_sie_attrvals = sieEntry->getAttributes(source_sie_attrnames[i]);
	  if(replace_dn(source_sie_attrvals, source_groupdn, dest_group_dn)) {
	    sieEntry->deleteAttribute(source_sie_attrnames[i]);
	    sieEntry->addAttributes(source_sie_attrnames[i], (const char **)source_sie_attrvals);
	  }
	}
	sieEntry->replace(sieEntry->entryDN());

	if(!isie_created) {
	  /* do this only once per ISIE creation */
	  char **source_isie_attrnames = appEntry->getAttributeNames();
	  char **source_isie_attrvals;
	  
	  /* Replace all source program group DN's with dest group DN's. */      
	  for(i=0; source_isie_attrnames[i]; i++) {
	    source_isie_attrvals = appEntry->getAttributes(source_isie_attrnames[i]);
	    if(replace_dn(source_isie_attrvals, source_groupdn, dest_group_dn)) {
	      appEntry->deleteAttribute(source_isie_attrnames[i]);
	      appEntry->addAttributes(source_isie_attrnames[i], (const char **)source_isie_attrvals);
	    }
	  }
	  appEntry->replace(appEntry->entryDN());
	  
	  isie_created = 1;
	}
	
	strcat(sie_response, "SIE:");
	strcat(sie_response, sieEntry->entryDN());
	strcat(sie_response, "%0A");

      } /* creating SIE's for current ISIE */
	
      strcat(isie_response, "ISIE:");
      strcat(isie_response, appEntry->entryDN());
      strcat(isie_response, "%0A");


      /* ------------------------- MERGE ALL ENTRIES UNDER CURRENT ISIE --------------------------- */
      
      /* 
       * Initialize the source and dest pset.
       * Use ISIE DN's to make sure all entries are merged.
       */
	
      pset_source = psetRealCreateSSL(admldapGetHost(ldapInfo), 
				      admldapGetPort(ldapInfo),
				      admldapGetSecurity(ldapInfo),
				      source_isie_ptr->source_isieDN,
				      source_binddn,
				      source_bindpw,
				      NULL,
				      &rv);
	
      if (!pset_source) {
	delete_isie_data(source_isie_list, &ldap_dest, 1);
	rpt_err(APP_ERROR, getResourceString(DBT_SOURCE_PSET_FAIL), NULL, NULL);
      }

      pset_dest = psetRealCreateSSL(dest_host, 
				    atoi(dest_port),
				    atoi(dest_secure),
				    (char *)appEntry->entryDN(),  /* where createSIE created the ISIE entry on dest DS */
				    dest_binddn,
				    dest_bindpw,
				    NULL,
				    &rv);
	
      if (!pset_dest) {
	psetDelete(pset_source);
	delete_isie_data(source_isie_list, &ldap_dest, 1);
	rpt_err(APP_ERROR, getResourceString(DBT_DEST_PSET_FAIL), NULL, NULL);
      }
	
	
      rv = psetReplicateSIE(pset_source, 
			    pset_dest, 
			    source_groupdn,
			    dest_group_dn);
      
      psetDelete(pset_source);
      psetDelete(pset_dest);
      
      free(temp2);
      
      if(rv) {
	delete_isie_data(source_isie_list, &ldap_dest, 1);
	rpt_err(APP_ERROR, getResourceString(DBT_REPLICATE_SIE_FAIL), NULL, NULL);
      }

    } /* processing each ISIE */

    // All ISIE's processed, now merge the Global Preferences section

    {
      char source_pref_dn[BIG_LINE];
      char dest_pref_dn[BIG_LINE];

      strcpy(source_pref_dn, "ou=Global Preferences, ");
      strcat(source_pref_dn, source_basedn);
      strcpy(dest_pref_dn, "ou=Global Preferences, ");
      strcat(dest_pref_dn, dest_basedn);

      if(mergeGlobalPrefs(admldapGetHost(ldapInfo),
			  admldapGetPort(ldapInfo),
			  admldapGetSecurity(ldapInfo),
			  source_binddn,
			  source_bindpw,
			  source_pref_dn,
			  dest_host,
			  atoi(dest_port),
			  atoi(dest_secure),
			  dest_binddn,
			  dest_bindpw,
			  dest_pref_dn) == -1) {
	delete_isie_data(source_isie_list, NULL, 0);
	rpt_err(APP_ERROR, getResourceString(DBT_GLOBAL_PREF_FAIL), NULL, NULL);
      }
    }

    delete_isie_data(source_isie_list, NULL, 0);

  } /* GET */

  sprintf(response,
	  "%s%s",
	  sie_response,
	  isie_response);

  rpt_success(response);
  return 0;
}
