/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 *
 *	This module contains routines used by the config CGI to set 
 *	attributes in server.xml (a iWS6 config file).
 *
 *  We are not using any XML parsing package for modifying XML files.
 *  Instead, a set of simple custom routines is provided here. Those
 *  routines are design to work in accord with the server.xml template
 *  in ns/netsite/admserv/newinst/common/config directory.
 *  This is not a robust solution, so if the template file is to be
 *  changed, care shoud be taken that this code is not broken.
 *
 */
#include <stdio.h>
#include <string.h>

#define BIG_BUF 4096

/*
 * A structure for describing changes in a xml file
 */
#define XMLMOD_MOD_ATTR 0
#define XMLMOD_ADD_LINE 1
#define XMLMOD_DEL_LINE 2

typedef struct {
int   modop;    /* One of XMLMOD_ defines */ 
char *section;  /* The line identifier, a string that the line starts with */
char *attr;
char *val;
} xml_file_mod_t;


/*
 * Check if the line starts with the section id string. The id string may
 * be prefixed with white spaces and '<'.
 * Return  1 (true), 0 (false)
 */
static int match_line(char *line, char *section) {

    char *c, *section_ref = strstr(line, section);

    if (section_ref == NULL) {
        return 0;
    }
    if (section_ref != line) {
        for (c=line; c != section_ref; c++) {
			if (*c != ' ' && *c != '\t' && *c!= '<') {
                return 0;
            }
        }
    }

    return 1;
}
    
/*
 * Find the a name value pair in the form attr="val" and replace the val
 * with the new one.
 * Return number of changes or -1 if the line is malformed.
 */
static int set_xml_attr(char *line, char *attr, char *val) {
    char *c, *attr_ref;
    char out[BIG_BUF];

    /* Find the attribute */
    attr_ref = strstr (line, attr);
    if (attr_ref == NULL) {
       return 0;
    }

    /* Replace the attribute value */
    memset(out, 0, BIG_BUF); 
    memcpy(out, line, attr_ref-line);

    /* Adance to the first '"', the start of the value */
    for (c=attr_ref; c && *c != '"'; c++) {
       ; /* do nothing */ 
    }
    if (c == NULL) {
        return -1; /* malformed line */
    }
    else {
        c++;
    }

    /* Skip the current value */
    while ( c && *c != '"') {
        c++;
    }
    if (c == NULL) {
        return -1; /* malformed line */
    }
    sprintf(out,"%s%s=\"%s%s", out, attr, val, c);
    strcpy(line, out);
    return 1;
}
    
/*
 * Modify file according to the list of xml_file_mod_t descriptors.
 * First rename the file, then copy it to the original name, one
 * line at a time, applying the changes using set_xml_atrr function. 
 * Return 0 for success, -1 for error
 */
static int modify_xml_file(char *filename, xml_file_mod_t **mods) {
    char filename0[BIG_BUF];
    char line[BIG_BUF];
    FILE *f0, *f;
    int i;

    /* Create temporary file with the current filename value */
    sprintf(filename0, "%s.0", filename);
    if (rename(filename, filename0) != 0) {
        /*fatal_error("Can't rename %s", filename, 0);*/
        return -1;
    }

   f0 = fopen (filename0, "r");
   if (f0 == NULL) {
       /*fatal_error("Can not open %s", filename0, 0);*/
       return -1;
   }
    
   /* The output file */
   f = fopen (filename, "w");
   if (f == NULL) {
       /*fatal_error("Can not open %s for writing", filename, 0);*/
       return -1;
   }

   while (fgets(line, BIG_BUF, f0)) {
       int save_line = 1;
       char * new_line = 0;
       for (i=0; mods[i]; i++) {
           if (! match_line(line, mods[i]->section)) {
               continue;
           } 
           if (mods[i]->modop == XMLMOD_DEL_LINE) {
               save_line = 0;
               break;
           }
           if (mods[i]->modop == XMLMOD_ADD_LINE) {
               /* A new line, to be added after the current one */
               new_line = mods[i]->val;
               continue;
           }

           if (mods[i]->modop == XMLMOD_MOD_ATTR) {
                if (set_xml_attr(line, mods[i]->attr, mods[i]->val) < 0) {

                    /* Fatal error, file has invalid format; */
                    /* Restore previous content              */
                    fclose(f);
                    fclose(f0);
                    remove(filename);
                    rename(filename0, filename);
                    return -1;
                }
           }
       }
       if (save_line) {
           fprintf(f, "%s", line);
       }
       if (new_line) {
           fprintf(f, "%s\n", new_line);
       }
   }

   fclose(f);
  
   /* Remove temporary file */
   fclose(f0);
   remove(filename0);

   return 0;
}

/*
 * Set port in server.xml
 */
int modxml_server_port(char *sroot,  char *port) {
    char xmlfile[BIG_BUF];

    xml_file_mod_t port_mod =
        {XMLMOD_MOD_ATTR, "LS id=\"ls1\"", "port", NULL};
    xml_file_mod_t *mods[2] = { NULL, NULL};

    port_mod.val = port;
    mods[0] = &port_mod;

    sprintf(xmlfile, "%s/admin-serv/config/server.xml", sroot);
    return modify_xml_file(xmlfile, mods);
}

/*
 * Set listening ip in server.xml
 */
int modxml_server_ip(char *sroot, char* ip) {
    char xmlfile[BIG_BUF];

    xml_file_mod_t ip_mod =
        {XMLMOD_MOD_ATTR, "LS id=\"ls1\"", "ip", NULL};
    xml_file_mod_t *mods[2] = { NULL, NULL};

    ip_mod.val = ip;
    mods[0] = &ip_mod;

    sprintf(xmlfile, "%s/admin-serv/config/server.xml", sroot);
    return modify_xml_file(xmlfile, mods);
}

/*
 * Set accesslog in server.xml
 */
int modxml_server_accesslog(char *sroot, char *accesslog) {
    char xmlfile[BIG_BUF];

    xml_file_mod_t accesslog_mod =
    {XMLMOD_MOD_ATTR, "VARS", "accesslog", NULL };
    xml_file_mod_t *mods[2] = {NULL, NULL};

    accesslog_mod.val = accesslog;
    mods[0] = &accesslog_mod;

    sprintf(xmlfile, "%s/admin-serv/config/server.xml", sroot);
    return modify_xml_file(xmlfile, mods);
}


/*
 * Turn off the security. Must also delete SSLPARAMS section as 
 * iWS will complain about it and will not start.
 */
int modxml_security_off(char *sroot) {
    char xmlfile[BIG_BUF];

    xml_file_mod_t sec_off =
        {XMLMOD_MOD_ATTR, "LS id=\"ls1\"", "security", "off"};
    xml_file_mod_t del_ssl =
        {XMLMOD_DEL_LINE, "SSLPARAMS", NULL, NULL};
    xml_file_mod_t *mods[3] = { NULL,  NULL, NULL};

    mods[0] = &sec_off;
    mods[1] = &del_ssl;

    sprintf(xmlfile, "%s/admin-serv/config/server.xml", sroot);
    return modify_xml_file(xmlfile, mods);
}


/*
 * Turn on the security and set ssl attributes.
 */
int modxml_security_on(char *sroot, char *certnickname,
                       char *ssl2,  char *ssl2ciphers,
                       char *ssl3,  char *ssl3tlsciphers,
                       char *tls,   char *tlsrollback,
                       char *clientauth) {

    char xmlfile[BIG_BUF];
    char attr[BIG_BUF];

    xml_file_mod_t sec_on =
        {XMLMOD_MOD_ATTR, "LS id=\"ls1\"", "security", "on"};
    xml_file_mod_t del_old_ssl =
        {XMLMOD_DEL_LINE, "SSLPARAMS", NULL, NULL};
    xml_file_mod_t add_ssl =
        {XMLMOD_ADD_LINE, "CONNECTIONGROUP", NULL, NULL};
    xml_file_mod_t *mods[4];

    sprintf(attr, "    <SSLPARAMS servercertnickname=\"%s\"", certnickname);
    sprintf(attr, "%s ssl2=\"%s\" ssl2ciphers=\"%s\"", attr, ssl2, ssl2ciphers);
    sprintf(attr, "%s ssl3=\"%s\" ssl3tlsciphers=\"%s\"", attr, ssl3, ssl3tlsciphers);
    sprintf(attr, "%s tls=\"%s\" tlsrollback=\"%s\"", attr, tls, tlsrollback);
    sprintf(attr, "%s clientauth=\"%s\" />", attr, clientauth);

    add_ssl.val = attr;
    mods[0] = &sec_on;
    mods[1] = &del_old_ssl;
    mods[2] = &add_ssl;
    mods[3] = NULL;

    sprintf(xmlfile, "%s/admin-serv/config/server.xml", sroot);
    return modify_xml_file(xmlfile, mods);
}
