/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
#include "libadmin/libadmin.h"
#include <stdio.h>
#include <stdarg.h>  /* for vsprintf */
#include <stdlib.h>
#include <string.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include "libadminutil/resource.h"
#include "libadminutil/admutil.h"
#include "libadminutil/distadm.h"
#include "libadmin/sconf.h"

#define INTsystem_free free
#define INTsystem_malloc malloc
#define INTsystem_realloc realloc

#define MAGT_CONF "/plugins/snmp/magt/CONFIG"
#define DEFLT_TRAP_PORT_STR "162"
#define  SEPERATOR 0x1F

void output_community_list(char *configfile);
void update_community_list(char *community, char *configfile);
community_list * add_community_entry(community_list *list, char* name, char *op);

void output_manager_list(char *configfile);
void update_manager_list(char *manager, char *configfile);
manager_list * add_manager_entry(manager_list *list, char* name, char *port, char*community);
static void verify_manager(char *host);

#define COMMUNITY_PROPERTY "snmp.communities"
#define MANAGER_PROPERTY "snmp.managers"

static char output_buf[4096];

/*
 * Logging function for debugging
 * Logging is enabled by creating file <server-root>/admin-serv/logs/<cgi-name>.dbg
 */
static FILE * logfp=NULL;
static int enable_verified=0;
static void logMsg(char *format, ...) {
	char logfile[512];

	if (!logfp && enable_verified) return;

	/* Automatically enable logging if <cgi-name>.dbg file exists in the logs directory */
	if (!logfp && !enable_verified) {
		enable_verified = 1;
		sprintf(logfile, "%s/../logs/snmpconf.dbg", getenv("ADMSERV_ROOT"));
		logfp = fopen(logfile, "r");
		if (logfp == NULL) {
			return;
		}
		fclose(logfp);
		logfp = fopen(logfile, "w");
	}

	if (logfp != NULL) {
		va_list ap;
		va_start(ap,format);
		vfprintf(logfp,format,ap);
		va_end(ap);
		fflush(logfp);
	}
}
/*
 * i18n conversions defines and function;
 * properties file  = "snmpconf.properties"
 */
#define RESOURCE_FILE "snmpconf"
#define resource_key(a,b)   a b

#define DBT_NO_METHOD 			resource_key(RESOURCE_FILE, "1")
#define DBT_UNKNOWN_METHOD 		resource_key(RESOURCE_FILE, "2")
#define DBT_NO_QUERY_STRING 		resource_key(RESOURCE_FILE, "3")
#define DBT_NO_OPERATION 		resource_key(RESOURCE_FILE, "4")
#define DBT_UNKNOWN_OPERATION		resource_key(RESOURCE_FILE, "5")
#define DBT_NO_COMMUNITIES		resource_key(RESOURCE_FILE, "6")
#define DBT_NO_MANAGERS			resource_key(RESOURCE_FILE, "7")
#define DBT_NO_TAB			resource_key(RESOURCE_FILE, "8")
#define DBT_OPEN_FILE			resource_key(RESOURCE_FILE, "9")
#define DBT_OPEN_TEMP			resource_key(RESOURCE_FILE, "10")
#define DBT_OPEN_RD			resource_key(RESOURCE_FILE, "11")
#define DBT_OPEN_WR			resource_key(RESOURCE_FILE, "12")
#define DBT_BAD_FILE_FORMAT		resource_key(RESOURCE_FILE, "13")
#define DBT_BAD_MANAGER    		resource_key(RESOURCE_FILE, "14")

static  char *acceptLanguage = (char*)"en";
static  Resource *i18nResource = NULL;

static void i18nInit() {
	char resPath[256], *execPath;
	execPath = getcwd(resPath, 256);
	if (execPath) {
		strcpy(resPath, execPath);
		strcat(resPath, "/property");
		i18nResource = res_init_resource(resPath, RESOURCE_FILE);
	}
	if (getenv("HTTP_ACCEPT_LANGUAGE")) {
		acceptLanguage = getenv("HTTP_ACCEPT_LANGUAGE");
	}
}
static void i18nCleanup() {
	if (i18nResource != NULL) {
	res_destroy_resource(i18nResource);
	}
}	
static const char *i18nMsg(char *msgid, char *defaultMsg) {
	const char *msg=NULL;

	if (i18nResource) {
            msg = res_getstring(i18nResource, msgid, acceptLanguage);
      }
	if (msg == NULL) {
		msg = (const char*)defaultMsg;
	}
	return msg;
}

/*
 *  Main
 */
int main(int argc, char *argv[])
{
    int _ai=ADMUTIL_Init();

    char *method = getenv("REQUEST_METHOD");
    char *root = getenv("NETSITE_ROOT");
    char *qs = 0, *op=0;
    char error_info[128];
    char filename[BIG_LINE];
    char *community, *manager;
	    
    logMsg(" In %s\n", argv[0]);

    i18nInit();

    sprintf(filename, "%s%s", root, MAGT_CONF);

	/* GET or POST method */
    if (!method || !*method) {
        /* non UI CGI */
        rpt_err(SYSTEM_ERROR, i18nMsg(DBT_NO_METHOD, "No method is specified"), NULL, NULL);
    }
    else if (!strcmp(method, "GET")) {
        /* UI CGI - either upgrade or normal admin CGI */
        qs = getenv("QUERY_STRING");
	
        if (!qs || !*qs) {
            rpt_err(INCORRECT_USAGE, i18nMsg(DBT_NO_QUERY_STRING ,"NO QUERY_STRING DATA"), NULL, NULL);
        }
        else {
            get_begin(qs);
        }
    }
    else if (!strcmp(method, "POST")) {
        post_begin(stdin);
    }
    else {
        sprintf(error_info, i18nMsg(DBT_UNKNOWN_METHOD,"Unknown Method (%s)"), method);
        rpt_err(INCORRECT_USAGE, error_info, NULL, NULL);
    }

	logMsg("method=%s\n", method);
	
    op=get_cgi_var("op", NULL, NULL);
	
	logMsg("op=%s\n", op);

	if (op == NULL) {
        rpt_err(INCORRECT_USAGE, i18nMsg(DBT_NO_OPERATION, "No operation defined"), NULL, NULL);
    }
	else if (strcmp(op,"getconfig")==0) {
		logMsg("Get List:\n");
		
		fprintf(stdout, "Content-type: text/html\n\n");		
		output_community_list(filename);
		output_manager_list(filename);
		
		exit(0);		
	}
	else if (strcmp(op,"setconfig")==0) {
		
		logMsg("Set List:\n");
		community=get_cgi_var(COMMUNITY_PROPERTY, NULL, NULL);
		if (community==NULL) {
			rpt_err(INCORRECT_USAGE, i18nMsg(DBT_NO_COMMUNITIES, "No snmp.communities in setconfig"), NULL, NULL);
		}
		update_community_list(community, filename);

		manager=get_cgi_var(MANAGER_PROPERTY, NULL, NULL);
		if (manager==NULL) {
			rpt_err(INCORRECT_USAGE, i18nMsg(DBT_NO_MANAGERS, "No snmp.managers in setconfig"), NULL, NULL);
		}
		update_manager_list(manager, filename);

		rpt_success(NULL);
		exit(0);
	}
	else {
        sprintf(error_info, i18nMsg(DBT_UNKNOWN_OPERATION, "Unknown Operation (%s)"), op);
        rpt_err(INCORRECT_USAGE, error_info, NULL, NULL);
    }
	
   return 0;
}


/**************************************
 *          Community List functions
 **************************************/

void output_community_list( char *configfile) {
	community_list *list = NULL;
	
	list = get_community_list(configfile);
	output_buf[0]=0;
	
	while(list && list->community)	{
		
		char *community=list->community->name;
		char *operation=list->community->operation;

		logMsg("IN >%s=%s\n", community, operation);
		
		if (strstr(operation, "SET OP") != NULL) {
			operation= (char*)"SET";
		}
		else if (strstr(operation, "GET OP") != NULL) {
			operation= (char*)"GET";
		}
		else {
			operation= (char*)"ALL";
		}		
		
		if (strlen(output_buf) == 0) {
			sprintf(output_buf, "%s\t%s",community, operation);
		}
		else {
			sprintf(output_buf, "%s,%s\t%s", output_buf, community, operation);
		}

		list=list->next;
	}
	
	if (output_buf[0] == 0) { /* empty list, use dash as a special value*/
		sprintf(output_buf, "-");
	}

	fprintf(stdout,"%s:%s\n", COMMUNITY_PROPERTY, output_buf);
	logMsg("OUT>%s:%s\n", COMMUNITY_PROPERTY, output_buf);

}

void update_community_list(char *community, char *configfile) {
	char *comma, *tab, *entry, *name, *operation;
	char op_buf[128];
	
	community_list * list = NULL;

	logMsg("Update Community <%s>\n", community);
		
	/* dash character is a special value used to denote an empty list */
	if (strcmp(community,"-")==0) {
		logMsg("Comunity List is empty\n");
		write_community_list(NULL, configfile);
		return;
	}

	while (community != NULL) {
		
		if ((comma=strchr(community, ',')) != NULL) {
			*comma = 0;
		}
		
		if ((tab=strchr(community,'\t')) == NULL) {
			logMsg("Error no tab in <%s>\n", community);
			rpt_err(INCORRECT_USAGE, i18nMsg(DBT_NO_TAB, "No <TAB> separator in input data"), NULL, NULL);
		}
		*tab=0;
		
		name = community;
		operation = tab+1;

		logMsg("%s=%s\n", name, operation);
		sprintf(op_buf, "ALLOW %s OPERATIONS", operation);
		list = add_community_entry(list, name, op_buf);
		
		if (comma != NULL) { /* go to next entry */
			community = comma + 1;
		}
		else {
			community = NULL; /* done */
		}
	}

	
	{
		community_list *l = list;
		while (l != NULL) {
			logMsg("list.name=<%s> list.op=<%s>\n",
				l->community->name, l->community->operation);
			l=l->next;
		}
	}

	write_community_list(list, configfile);
}


community_list * add_community_entry(community_list *list, char* name, char *op) {
	
	community_list *entry;
	entry = (community_list*)MALLOC(sizeof(struct community_list));
	entry->community = (community_entry *)MALLOC(sizeof(struct community_entry));
	entry->next = NULL;
	entry->community->name = (char *)MALLOC(strlen(name)+1);
	entry->community->operation = (char *)MALLOC(strlen(op)+1);

	strcpy(entry->community->name, name);
	strcpy(entry->community->operation, op);

	if (list != NULL) {
		community_list *last_entry = list;
		while (last_entry->next != NULL) { /* Go to the last entry */
			last_entry = last_entry->next;
		}
		last_entry->next = entry;
		return list;
	}
	else {
		return entry;
	}
}
		

/************************************
 *          Manager List functions
 ************************************/

void output_manager_list( char *configfile) {
	manager_list *list = NULL;
	
	list = get_manager_list(configfile);
	output_buf[0]=0;
	
	while(list && list->manager)	{
		
		char *manager=list->manager->name;
		char *trap_port=list->manager->trap_port;
		char *community=list->manager->community;

		logMsg("IN ><%s><%s><%s>\n", manager, trap_port, community);
		
		if (strlen(output_buf) == 0) {
			sprintf(output_buf, "%s\t%s\t%s",manager, trap_port, community);
		}
		else {
			sprintf(output_buf, "%s,%s\t%s\t%s", output_buf, manager, trap_port, community);
		}

		list=list->next;
	}
	
	if (output_buf[0] == 0) { /* empty list, use dash as a special value*/
		sprintf(output_buf, "-");
	}

	fprintf(stdout,"%s:%s\n", MANAGER_PROPERTY, output_buf);
	logMsg("OUT>%s:%s\n", MANAGER_PROPERTY, output_buf);

}

void update_manager_list(char *manager, char *configfile) {
	char *comma, *tab1, *tab2, *entry, *name, *trap_port, *community;
	char op_buf[128];
	
	manager_list * list = NULL;

	logMsg("Update manager<%s>\n", manager);
		
	/* dash character is a special value used to denote an empty list */
	if (strcmp(manager,"-")==0) {
		logMsg("Manager List is empty\n");
		write_manager_list(NULL, configfile);
		return;
	}

	while (manager!= NULL) {
		
		if ((comma=strchr(manager, ',')) != NULL) {
			*comma = 0;
		}
		
		if ((tab1=strchr(manager,'\t')) == NULL) {
			logMsg("Error no first tab in <%s>\n", manager);
			rpt_err(INCORRECT_USAGE, i18nMsg(DBT_NO_TAB, "No <TAB> separator in input data"), NULL, NULL);
		}
		if ((tab2=strchr(tab1+1,'\t')) == NULL) {
			logMsg("Error no second tab in <%s>\n", manager);
			rpt_err(INCORRECT_USAGE, i18nMsg(DBT_NO_TAB, "No <TAB> separator in input data"), NULL, NULL);
		}
		*tab1=0; *tab2=0;
		
		name = manager;
		trap_port = tab1+1;
		community = tab2+1;

		logMsg("<%s><%s><%s>\n", name, trap_port, community);
		verify_manager(name);
		list = add_manager_entry(list, name, trap_port, community);
		
		if (comma != NULL) { /* go to next entry */
			manager= comma + 1;
		}
		else {
			manager= NULL; /* done */
		}
	}

	
	{
		manager_list *l = list;
		while (l != NULL) {
			logMsg("list.name=<%s> list.port=<%s> list.community=<%s>\n",
				l->manager->name, l->manager->trap_port, l->manager->community);
			l=l->next;
		}
	}

	write_manager_list(list, configfile);
}

/**
 * Verify id host name/IP address is correct. Master agent will not start
 * if an incorrect host name is found in the configuration file
 */
static void verify_manager(char *name) {
	struct hostent *hostent = gethostbyname(name);

	logMsg ("verify %s\n", name);

	if (hostent == NULL) {
		char msg[256];
		sprintf(msg, i18nMsg(DBT_BAD_MANAGER, "Incorrect manager specified (%s): DNS lookup failed, or bad IP address format"), name);
		rpt_err(SYSTEM_ERROR, msg, NULL, NULL);
	}
}


manager_list * add_manager_entry(manager_list *list, char* name, char *port, char*community) {
	
	manager_list *entry;
	entry = (manager_list*)MALLOC(sizeof(struct manager_list));
	entry->manager= (manager_entry *)MALLOC(sizeof(struct manager_entry));
	entry->next = NULL;
	entry->manager->name = (char *)MALLOC(strlen(name)+1);
	entry->manager->trap_port = (char *)MALLOC(strlen(port)+1);
	entry->manager->community = (char *)MALLOC(strlen(community)+1);

	strcpy(entry->manager->name, name);
	strcpy(entry->manager->trap_port, port);
	strcpy(entry->manager->community, community);

	if (list != NULL) {
		manager_list *last_entry = list;
		while (last_entry->next != NULL) { /* Go to the last entry */
			last_entry = last_entry->next;
		}
		last_entry->next = entry;
		return list;
	}
	else {
		return entry;
	}
}

/******************************************
 *   Functions from snmpcomm.c
 ******************************************/

/* following rountins should be moved to some lib Gdong */

community_list * get_community_list(char * file_name)
{
  community_list * ret_list_head=NULL;
  community_list * ret_list, *ret_tmp;
  FILE *fp;
  char line[BIG_LINE];
  char * ptr;

  if ( (fp = fopen(file_name, "r")) == NULL ) {
    sprintf(line, i18nMsg(DBT_OPEN_FILE, "Can't open config file (%s)"), file_name);
    rpt_err(SYSTEM_ERROR, line, NULL, NULL);
  } else {
    ret_list_head = (community_list *)MALLOC(sizeof(struct community_list));
    ret_list = ret_list_head;
    ret_list->community = NULL;
    ret_list->next = NULL;
    ret_tmp=ret_list;
    
    while (fgets(line, BIG_LINE, fp)){
      if (strncmp(line, "COMMUNITY",9) == 0){
	if(ret_list==NULL){
	  ret_list = (community_list *)MALLOC(sizeof(struct community_list));
	  ret_list->community = NULL;
	  ret_list->next = NULL;
	  ret_tmp->next = ret_list;
	  ret_tmp = ret_list;
	}
	ret_list->community = (community_entry *)MALLOC(sizeof(struct community_entry));
	ptr = strchr(line, '\t'); 
	if(ptr != NULL)
	  ptr++;
	else {
	  ptr = strchr(line, ' ');
	  if (ptr != NULL)
	    ptr++;
	}

	ret_list->community->name = (char *)MALLOC(strlen(ptr)+4);
	strcpy(ret_list->community->name, ptr);
       	ptr = strchr(ret_list->community->name, '\n');
	if(ptr != NULL)
	  *ptr='\0';

	if (fgets(line, BIG_LINE, fp) ==0)
	  return NULL;

	ptr = strchr(line, '\t'); 
	if(ptr != NULL)
	  ptr++;
	else {
	  ptr = strchr(line, ' ');
	  if (ptr != NULL)
	    ptr++;
	  else
	    ptr = line;
	}
	/*	ptr = strchr(line, 'A');  */

	ret_list->community->operation = (char *)MALLOC(strlen(ptr)+4);
	strcpy(ret_list->community->operation, ptr);
	ptr = strchr(ret_list->community->operation, '\n');
	if(ptr != NULL)
	  *ptr='\0';

	ret_list = NULL;
      }
    }    
    fclose(fp);
  }
  return (community_list *)ret_list_head;
}

void free_community_all_list(community_list * list)
{
  community_entry * entry;
  community_list  * list_tmp = list;
  community_list  * ptr;

  if(list)
    return;

  if (list_tmp->community)
    free_community_entry(list->community);
  
  list_tmp = list_tmp->next;

  while(list_tmp){
    entry = list_tmp->next->community;
    if (entry)
      free_community_entry(entry);
  
    ptr = list_tmp->next;
    free_community_list(list_tmp);
    list_tmp = ptr;
  }

  free_community_list(list);
}

void free_community_entry(community_entry * entry)
{
  if(entry)
    return;

  if(entry->name)
    FREE(entry->name);
  if(entry->operation)
    FREE(entry->operation);

  FREE(entry);
}
  
void free_community_list(community_list * list)
{
  if(list)
    return;

  if (list->community)
    FREE(list->community);
  
  FREE(list);
}

int write_community_list(community_list * list, char * file_name)
{
  FILE * fp, *tmp_fp;
  community_list * list_tmp = list;
  community_entry * entry;
  char * ptr;
  char  line[BIG_LINE];

  if( (tmp_fp=tmpfile()) == NULL )
    rpt_err(SYSTEM_ERROR, i18nMsg(DBT_OPEN_TEMP, "Can not open temp file"), NULL, NULL);

  if( (fp=fopen(file_name,"r")) == NULL ) {
    sprintf(line, i18nMsg(DBT_OPEN_RD,"Can not open config (%s) for reading"), file_name);
    rpt_err(SYSTEM_ERROR, line, NULL, NULL);
  }

  while(fgets(line, BIG_LINE, fp)){
    if (strncmp(line, "COMMUNITY", 9) == 0)
      {
	fgets(line, BIG_LINE, fp);
        continue;
      }
 
    fputs(line, tmp_fp);
  }

  fclose(fp);
  
  if ((fp = fopen(file_name, "w")) == NULL) {
    sprintf(line, i18nMsg(DBT_OPEN_WR, "Can not open config (%s) for writing"), file_name);
    rpt_err(SYSTEM_ERROR, line, NULL, NULL);
  }

  rewind(tmp_fp);
 
  while(fgets(line, BIG_LINE, tmp_fp))
    fputs(line, fp);

  while(list_tmp){
    entry = list_tmp->community;
    fputs("COMMUNITY\t", fp);
    fputs(entry->name, fp);
    ptr = strchr(entry->name, '\n');
    if(ptr == NULL)
      fputs("\n", fp);

    fputs("\t", fp);
    fputs(entry->operation, fp);
    ptr = strchr(entry->operation, '\n');
    if(ptr == NULL)
      fputs("\n", fp);

    list_tmp = list_tmp->next;
  }
  fclose(fp);

  return 0;

}

void modify_community(char * community, char * operation, char * file_name, int flag)
{
  community_list * list;
  community_list * list_tmp, *ptr;

  if (flag == 0){
    /* remove this entry */
    list = get_community_list(file_name);
    
    list_tmp = list;
    if (list_tmp->community){
      if(strcmp(list_tmp->community->name, community) == 0){
	ptr = list_tmp->next;
	free_community_entry(list_tmp->community);
	free_community_list(list_tmp);
	
	write_community_list(ptr, file_name);
	return;
      }
    } else {
      return;
    }

    ptr = list_tmp;
    list_tmp = list_tmp->next;

    while(list_tmp){
      if (list_tmp->community){
	if(strcmp(list_tmp->community->name, community) == 0){
	  ptr->next = list_tmp->next;
	  free_community_entry(list_tmp->community);
	  free_community_list(list_tmp);
	
	  write_community_list(list, file_name);
	  return;
	}
      }
      ptr = list_tmp;
      list_tmp = list_tmp->next;
    }
  } else {
    list = get_community_list(file_name);
    
    list_tmp = list;
    ptr = list_tmp;

    if (list_tmp->community){
      if(strcmp(list_tmp->community->name, community) == 0){
	list_tmp->community->name = (char *)REALLOC(list_tmp->community->name, strlen(community));
	strcpy(list_tmp->community->name, community);
	list_tmp->community->operation = (char *)REALLOC(list_tmp->community->operation, strlen(operation));
	strcpy(list_tmp->community->operation, operation);

	write_community_list(list, file_name);
	return;
      } else {
	list_tmp = list_tmp->next;
      }
    } 

    while(list_tmp){
      if (list_tmp->community){
	if(strcmp(list_tmp->community->name, community) == 0){
	  list_tmp->community->name = (char *)REALLOC(list_tmp->community->name, strlen(community));
	  strcpy(list_tmp->community->name, community);
	  list_tmp->community->operation = (char *)REALLOC(list_tmp->community->operation, strlen(operation));
	  strcpy(list_tmp->community->operation, operation);
	
	  write_community_list(list, file_name);
	  return;
	}
      }
      ptr = list_tmp;
      list_tmp = list_tmp->next;
    }
    
    if(ptr->community){
      ptr->next = (community_list *)MALLOC(sizeof(struct community_list));

      list_tmp = ptr->next;
    } else {
      list_tmp = ptr;
    }
    
    list_tmp->community = (community_entry *)MALLOC(sizeof(struct community_entry));
    list_tmp->next = NULL;

    list_tmp->community->name = (char *)MALLOC(strlen(community)+4);
    strcpy(list_tmp->community->name, community);
    list_tmp->community->operation = (char *)MALLOC(strlen(operation)+4);
    strcpy(list_tmp->community->operation, operation);

    write_community_list(list, file_name);
  }
}
		
/******************************************
 *   Functions from snmptrpc.c
 ******************************************/

/* following rountins should be moved to some lib Gdong */

manager_list * get_manager_list(char * file_name)
{
  manager_list * ret_list_head=NULL;
  manager_list * ret_list, *ret_tmp;
  FILE *fp;
  char line[BIG_LINE];
  char * ptr;

  if ( (fp = fopen(file_name, "r")) == NULL ) {
     sprintf(line, i18nMsg(DBT_OPEN_RD,"Can not open CONFIG file (%s) fir reading"), file_name);
     rpt_err(SYSTEM_ERROR, line, NULL, NULL);
  } else {
    ret_list_head = (manager_list *)MALLOC(sizeof(struct manager_list));
    ret_list = ret_list_head;
    ret_list->manager = NULL;
    ret_list->next = NULL;
    ret_tmp=ret_list;
    
    while (fgets(line, BIG_LINE, fp)){
      if (strncmp(line, "MANAGER",7) == 0){
	if(ret_list==NULL){
	  ret_list = (manager_list *)MALLOC(sizeof(struct manager_list));
	  ret_list->manager = NULL;
	  ret_list->next = NULL;
	  ret_tmp->next = ret_list;
	  ret_tmp = ret_list;
	}
	ret_list->manager = (manager_entry *)MALLOC(sizeof(struct manager_entry));
	ptr = strchr(line, '\t'); 
	if(ptr != NULL)
	  ptr++;
	else {
	  ptr = strchr(line, ' ');
	  if (ptr != NULL)
	    ptr++;
	}

	ret_list->manager->name = (char *)MALLOC(strlen(ptr)+4);
	strcpy(ret_list->manager->name, ptr);
       	ptr = strchr(ret_list->manager->name, '\n');
	if(ptr != NULL)
	  *ptr='\0';

	if (fgets(line, BIG_LINE, fp) ==0)
	  rpt_err(SYSTEM_ERROR, i18nMsg(DBT_BAD_FILE_FORMAT,"Bad config file format"), NULL, NULL);

	ptr = strrchr(line, 'T'); /* SEND ALL TRAPS TO PORT */
	if(ptr != NULL)
	  { ptr++;ptr++; }
	else 
	  rpt_err(SYSTEM_ERROR, i18nMsg(DBT_BAD_FILE_FORMAT,"Bad config file format"), NULL, NULL);

	ret_list->manager->trap_port = (char *)MALLOC(strlen(ptr)+4);
	strcpy(ret_list->manager->trap_port, ptr);
	ptr = strchr(ret_list->manager->trap_port, '\n');
	if(ptr != NULL)
	  *ptr='\0';

	if (fgets(line, BIG_LINE, fp) ==0)
	  rpt_err(SYSTEM_ERROR, i18nMsg(DBT_BAD_FILE_FORMAT,"Bad config file format"), NULL, NULL);

	ptr = strrchr(line, 'Y'); /* WITH COMMUNITY */
	if(ptr != NULL)
	  { ptr++; ptr++; }
	else 
	  rpt_err(SYSTEM_ERROR, i18nMsg(DBT_BAD_FILE_FORMAT,"Bad config file format"), NULL, NULL);

	ret_list->manager->community = (char *)MALLOC(strlen(ptr)+4);
	strcpy(ret_list->manager->community, ptr);
	ptr = strchr(ret_list->manager->community, '\n');
	if(ptr != NULL)
	  *ptr='\0';

	ret_list = NULL;
      }
    }    
    fclose(fp);
  }
  return (manager_list *)ret_list_head;
}

void free_manager_all_list(manager_list * list)
{
  manager_entry * entry;
  manager_list  * list_tmp = list;
  manager_list  * ptr;

  if(list)
    return;

  if (list_tmp->manager)
    free_manager_entry(list->manager);
  
  list_tmp = list_tmp->next;

  while(list_tmp){
    entry = list_tmp->next->manager;
    if (entry)
      free_manager_entry(entry);
  
    ptr = list_tmp->next;
    free_manager_list(list_tmp);
    list_tmp = ptr;
  }

  free_manager_list(list);
}

void free_manager_entry(manager_entry * entry)
{
  if(entry)
    return;

  if(entry->name)
    FREE(entry->name);
  if(entry->trap_port)
    FREE(entry->trap_port);
  if(entry->community)
    FREE(entry->community);

  FREE(entry);
}
  
void free_manager_list(manager_list * list)
{
  if(list)
    return;

  if (list->manager)
    FREE(list->manager);
  
  FREE(list);
}

int write_manager_list(manager_list * list, char * file_name)
{
  FILE * fp, *tmp_fp;
  manager_list * list_tmp = list;
  manager_entry * entry;
  char * ptr;
  char  line[BIG_LINE];

  if( (tmp_fp=tmpfile()) == NULL )
    rpt_err(SYSTEM_ERROR, i18nMsg(DBT_OPEN_TEMP,"Can not open a temp file"), NULL, NULL);

  if( (fp=fopen(file_name,"r")) == NULL ) {
    sprintf(line, i18nMsg(DBT_OPEN_RD,"Can not open CONFIG file (%s) for reading"), file_name);
    rpt_err(SYSTEM_ERROR, line, NULL, NULL);
  }

  while(fgets(line, BIG_LINE, fp)){
    if (strncmp(line, "MANAGER", 7) == 0)
      {
	fgets(line, BIG_LINE, fp);
	fgets(line, BIG_LINE, fp);
        continue;
      }
 
    fputs(line, tmp_fp);
  }

  fclose(fp);
  
  if ((fp = fopen(file_name, "w")) == NULL) {
    sprintf(line, i18nMsg(DBT_OPEN_WR,"Can not open CONFIG file (%s) for writing"), file_name);
    rpt_err(SYSTEM_ERROR, line, NULL, NULL);
  }
  rewind(tmp_fp);
 
  while(fgets(line, BIG_LINE, tmp_fp))
    fputs(line, fp);

  while(list_tmp){
    entry = list_tmp->manager;
    fputs("MANAGER\t", fp);
    fputs(entry->name, fp);
    ptr = strchr(entry->name, '\n');
    if(ptr == NULL)
      fputs("\n", fp);

    fputs("\t", fp);
    fputs("SEND ALL TRAPS TO PORT ", fp);
    fputs(entry->trap_port, fp);
    ptr = strchr(entry->trap_port, '\n');
    if(ptr == NULL)
      fputs("\n", fp);

    fputs("\t", fp);
    fputs("WITH COMMUNITY ", fp);
    fputs(entry->community, fp);
    ptr = strchr(entry->community, '\n');
    if(ptr == NULL)
      fputs("\n", fp);

    list_tmp = list_tmp->next;
  }
  fclose(fp);

  return 0;

}

void modify_manager(char * manager, char * trap_port, char * community, char * file_name, int flag)
{
  manager_list * list;
  manager_list * list_tmp, *ptr;

  if (flag == 0){
    /* remove this entry */
    list = get_manager_list(file_name);
    
    list_tmp = list;
    if (list_tmp->manager){
      if(strcmp(list_tmp->manager->name, manager) == 0){
	ptr = list_tmp->next;
	free_manager_entry(list_tmp->manager);
	free_manager_list(list_tmp);
	
	write_manager_list(ptr, file_name);
	return;
      }
    } else {
      return;
    }

    ptr = list_tmp;
    list_tmp = list_tmp->next;

    while(list_tmp){
      if (list_tmp->manager){
	if(strcmp(list_tmp->manager->name, manager) == 0){
	  ptr->next = list_tmp->next;
	  free_manager_entry(list_tmp->manager);
	  free_manager_list(list_tmp);
	
	  write_manager_list(list, file_name);
	  return;
	}
      }
      ptr = list_tmp;
      list_tmp = list_tmp->next;
    }
  } else {
    list = get_manager_list(file_name);
    
    list_tmp = list;
    ptr = list_tmp;

    if (list_tmp->manager){
      if(strcmp(list_tmp->manager->name, manager) == 0){
	list_tmp->manager->name = (char *)REALLOC(list_tmp->manager->name, strlen(manager));
	strcpy(list_tmp->manager->name, manager);
	list_tmp->manager->trap_port = (char *)REALLOC(list_tmp->manager->trap_port, strlen(trap_port));
	strcpy(list_tmp->manager->trap_port, trap_port);
	list_tmp->manager->community = (char *)REALLOC(list_tmp->manager->community, strlen(community));
	strcpy(list_tmp->manager->community, community);
	write_manager_list(list, file_name);
	return;
      } else {
	list_tmp = list_tmp->next;
      }
    } 

    while(list_tmp){
      if (list_tmp->manager){
	if(strcmp(list_tmp->manager->name, manager) == 0){
	  list_tmp->manager->name = (char *)REALLOC(list_tmp->manager->name, strlen(manager));
	  strcpy(list_tmp->manager->name, manager);
	  list_tmp->manager->trap_port = (char *)REALLOC(list_tmp->manager->trap_port, strlen(trap_port));
	  strcpy(list_tmp->manager->trap_port, trap_port);
	  list_tmp->manager->community = (char *)REALLOC(list_tmp->manager->community, strlen(community));
	  strcpy(list_tmp->manager->community, community);
	
	  write_manager_list(list, file_name);
	  return;
	}
      }
      ptr = list_tmp;
      list_tmp = list_tmp->next;
    }
    
    if(ptr->manager){
      ptr->next = (manager_list *)MALLOC(sizeof(struct manager_list));

      list_tmp = ptr->next;
    } else {
      list_tmp = ptr;
    }
    
    list_tmp->manager = (manager_entry *)MALLOC(sizeof(struct manager_entry));
    list_tmp->next = NULL;

    list_tmp->manager->name = (char *)MALLOC(strlen(manager)+4);
    strcpy(list_tmp->manager->name, manager);
    list_tmp->manager->trap_port = (char *)MALLOC(strlen(trap_port)+4);
    strcpy(list_tmp->manager->trap_port, trap_port);
    list_tmp->manager->community = (char *)MALLOC(strlen(community)+4);
    strcpy(list_tmp->manager->community, community);

    write_manager_list(list, file_name);
  }
}
