/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * snmpmctl.c - start/stop/restart SNMP Master agent
 *
 * Status Operation
 * in:  OPERATION=STATUS
 * out: status=ON|OFF
 *
 * Stop Operation
 * in:  OPERATION=STOP
 * out: 
 *
 * Restart Operation
 * in:  OPERATION=RESTART
 * out: 
 *
 */

#include "libadmin/libadmin.h"
#include "libadminutil/resource.h"
#include "libadminutil/distadm.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>

#include "libadminutil/admutil.h"
#include "libadmin/sconf.h"

#define MASTERAGT_PID   "pid_masteragt"
#define MAGT_CONF "/plugins/snmp/magt/CONFIG"
#define SEPERATOR 0x1F

void check_superuser();
int magent_start(char *);
int magent_restart(char *);
int magent_shutdown(int);
int magent_is_running();
int smux_port_is_used();
int snmp_port_is_used();

/*
 * Logging function for debugging
 * Logging is enabled by creating file <server-root>/admin-serv/logs/<cgi-name>.dbg
 */
static FILE * logfp=NULL;
static int enable_verified=0;
static void logMsg(char *format, ...) {
	char logfile[512];

	if (!logfp && enable_verified) return;

	/* Automatically enable logging if <cgi-name>.dbg file exists in the logs directory */
	if (!logfp && !enable_verified) {
		enable_verified = 1;
		sprintf(logfile, "%s/../logs/snmpmctl.dbg", getenv("ADMSERV_ROOT"));
		logfp = fopen(logfile, "r");
		if (logfp == NULL) {
			return;
		}
		fclose(logfp);
		logfp = fopen(logfile, "w");
	}

	if (logfp != NULL) {
		va_list ap;
		va_start(ap,format);
		vfprintf(logfp,format,ap);
		va_end(ap);
		fflush(logfp);
	}
}

/*
 * i18n conversions defines and function;
 * properties file  = "snmpmctl.properties"
 */
#define RESOURCE_FILE "snmpmctl"
#define resource_key(a,b)   a b

#define DBT_NO_METHOD 			resource_key(RESOURCE_FILE, "1")
#define DBT_UNKNOWN_METHOD 		resource_key(RESOURCE_FILE, "2")
#define DBT_NO_QUERY_STRING 		resource_key(RESOURCE_FILE, "3")
#define DBT_NO_OPERATION 		resource_key(RESOURCE_FILE, "4")
#define DBT_UNKNOWN_OPERATION		resource_key(RESOURCE_FILE, "5")
#define DBT_UNKNOWN_CMD			resource_key(RESOURCE_FILE, "6")
#define DBT_NOT_ROOTUID			resource_key(RESOURCE_FILE, "7")
#define DBT_TCP_PROTOCOL		resource_key(RESOURCE_FILE, "8")
#define DBT_CREATE_SOCK			resource_key(RESOURCE_FILE, "9")
#define DBT_TERMINATE			resource_key(RESOURCE_FILE, "10")
#define DBT_NOT_RUNNING			resource_key(RESOURCE_FILE, "11")
#define DBT_START			resource_key(RESOURCE_FILE, "12")
#define DBT_NO_PERM			resource_key(RESOURCE_FILE, "13")
#define DBT_FORK			resource_key(RESOURCE_FILE, "14")
#define DBT_EXECL			resource_key(RESOURCE_FILE, "15")
#define DBT_UDP_PROTOCOL		resource_key(RESOURCE_FILE, "16")
#define DBT_SMUX_PORT_USED		resource_key(RESOURCE_FILE, "17")
#define DBT_SNMP_PORT_USED		resource_key(RESOURCE_FILE, "18")
#define DBT_OPEN_PIDFILE  		resource_key(RESOURCE_FILE, "19")

static  char *acceptLanguage = (char*)"en";
static  Resource *i18nResource = NULL;

static void i18nInit() {
	char resPath[256], *execPath;
	execPath = getcwd(resPath, 256);
	if (execPath) {
		strcpy(resPath, execPath);
		strcat(resPath, "/property");
		i18nResource = res_init_resource(resPath, RESOURCE_FILE);
	}
	if (getenv("HTTP_ACCEPT_LANGUAGE")) {
		acceptLanguage = getenv("HTTP_ACCEPT_LANGUAGE");
	}
}
static void i18nCleanup() {
	if (i18nResource != NULL) {
	res_destroy_resource(i18nResource);
	}
}	
static const char *i18nMsg(char *msgid, char *defaultMsg) {
	const char *msg=NULL;

	if (i18nResource) {
            msg = res_getstring(i18nResource, msgid, acceptLanguage);
      }
	if (msg == NULL) {
		msg = (const char*)defaultMsg;
	}
	return msg;
}


int main(int argc, char *argv[])
{
    int _ai = ADMUTIL_Init();
    char line[BIG_LINE];
    char *method = getenv("REQUEST_METHOD");
	char *qs=0;
    char *root = getenv("NETSITE_ROOT");
	char error_info[128];   
    char *action_type;
	
    logMsg("In %s\n", argv[0]);

    i18nInit();

    /* GET or POST method */
    if (!method || !*method) {
        /* non UI CGI */
        rpt_err(SYSTEM_ERROR, i18nMsg(DBT_NO_METHOD,"No method is specified"), NULL, NULL);
    }
    else if (!strcmp(method, "GET")) {
        /* UI CGI - either upgrade or normal admin CGI */
        qs = getenv("QUERY_STRING");
	
        if (!qs || !*qs) {
            rpt_err(INCORRECT_USAGE, i18nMsg(DBT_NO_QUERY_STRING,"NO QUERY_STRING DATA"), NULL, NULL);
        }
        else {
            get_begin(qs);
        }
    }
    else if (!strcmp(method, "POST")) {
        post_begin(stdin);
    }
    else {
        sprintf(error_info, i18nMsg(DBT_UNKNOWN_METHOD, "Unknown Method (%s)"), method);
        rpt_err(INCORRECT_USAGE, error_info, NULL, NULL);
    }

    action_type = get_cgi_var("ACTION", NULL, NULL);
    if (!strcmp(action_type, "START")) {
       logMsg("START\n");
		check_superuser();
        magent_shutdown(0);
        magent_start(root);
    }
    else if (!strcmp(action_type, "STOP")) {
        logMsg("STOP\n");
		check_superuser();
        magent_shutdown(1);
    }
    else if (!strcmp(action_type, "RESTART")) {
        logMsg("RESTART\n");
		check_superuser();
        magent_restart(root);
    }
    else if (!strcmp(action_type, "STATUS")) {
	int running=magent_is_running();
        logMsg("STATUS\n");
	/*	check_superuser();*/
    	rpt_success(NULL);
        fprintf(stdout,"status:%s\n", running ? "ON": "OFF");
        logMsg("status:%s\n", running ? "ON": "OFF");
	return 0;
    }
    else {
        sprintf(error_info, i18nMsg(DBT_UNKNOWN_CMD,"Command (%s) not understood"), action_type);
        logMsg("%s\n", error_info);
		rpt_err(INCORRECT_USAGE, error_info, NULL, NULL);
    }
    rpt_success(NULL);
    exit(0);
    return 0;
}

/*
 * Check if running as a superuser
 */
void check_superuser() {
    if (getuid() != 0) {
		logMsg("Not a superuser, PID=%d\n", getuid());
        rpt_err(SYSTEM_ERROR,
        i18nMsg(DBT_NOT_ROOTUID, "Can not control SNMP Master Agent, "
        "Admin Server instance in not running under ROOT UID"),
        NULL,NULL);
    }
}

/*
 * Check if the SMUX port 199 is used
 *
 * This routine returns:
 *	0 if port is not used
 *	1 if port is used
 *
 */
int smux_port_is_used()
{
   struct servent  *pse;
   struct protoent *ppe;
   struct sockaddr_in sin;
   int    s, type;
   int    rc;

   sin.sin_family = AF_INET;
   sin.sin_addr.s_addr = inet_addr("127.0.0.1");

   if (pse = getservbyname("smux", "tcp")) {
      sin.sin_port = (u_short)pse->s_port;
   } else {
      sin.sin_port = htons(199);
   }

   logMsg("SMUX port=%d\n", ntohs(sin.sin_port));

   if ((ppe = getprotobyname("tcp")) == 0) {
      rpt_err(SYSTEM_ERROR, i18nMsg(DBT_TCP_PROTOCOL,"Can't get tcp protocol entry to check smux port status"), NULL, NULL);
      return(0);
   }

   if ((s = socket(AF_INET, SOCK_STREAM, ppe->p_proto)) < 0) {
      rpt_err(SYSTEM_ERROR, i18nMsg(DBT_CREATE_SOCK,"Can't create socket"), NULL, NULL);
      return(0);
   }

   if (rc = connect(s, (struct sockaddr *)&sin, sizeof(sin)) < 0) { 
     logMsg("connect() errno=%d => 199 is NOT used\n", errno);
     close(s);
     return 0;
   }
   else {
     logMsg("199 is used\n");
     close(s);
     return(1);
   }
}

/*
 *
 * Check if master agent for the current server root is running.
 *
 * This routine returns:
 *	0 if magt is NOT running
 *	1 if magt is running
 *
 * The run state is determined by reading magt pid from
 * <server-root>/admin-serv/logs/pid_masteragt
 * and checking the pid by sending signal 0 to it.
 *
 */
int magent_is_running()
{
   char command[BIG_LINE];
   char path[PATH_MAX];
   FILE *fhdl;
   char *ADMSERV_ROOT = getenv("ADMSERV_ROOT");
   pid_t  magtid;

   sprintf(path, "%s/../logs/%s", ADMSERV_ROOT, MASTERAGT_PID);
   logMsg("Looking for pid in (%s)\n", path);
   if ((fhdl = fopen(path, "r")) != (FILE *) NULL)
   {
     fscanf(fhdl, "%d\n", &magtid);
     fclose(fhdl);

     /* sending sig 0 checks if pid is a valid one */
     if (kill(magtid, 0) == 0)
     { 
       logMsg("Master pid %d is running\n", magtid);
       return 1;
     }
     else
     {
       logMsg("Master pid %d is not running, removing pid file\n", magtid);
       sprintf(command, "rm %s&", path);
       system(command);
       return 0;
     }
   }
   logMsg("No pid file, master is NOT running\n");
   return 0;
}


/*
 * Check if the SNMP port 161 is used
 *
 * This routine returns:
 *	0 if port is NOT used
 *	1 if port is used
 *
 */
int snmp_port_is_used()
{
    struct servent  *pse;
    struct protoent *ppe;
    struct sockaddr_in sin;
    int    s, rc;

    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = INADDR_ANY;

    if (pse = getservbyname("snmp", "udp"))
      sin.sin_port = (u_short)pse->s_port;
    else 
      sin.sin_port = htons(161);

    logMsg("SNMP port=%d\n", ntohs(sin.sin_port));
    
    if ((ppe = getprotobyname("udp")) == 0)
    {
      rpt_err(SYSTEM_ERROR, i18nMsg(DBT_UDP_PROTOCOL,"Can't get udp protocol entry to check snmp port status"), NULL, NULL);
      logMsg("getprotobyname(udp) failed\n");
      return 0;
    }
    
    s = socket(PF_INET, SOCK_DGRAM, ppe->p_proto);
    if(s<0) {
      rpt_err(SYSTEM_ERROR, i18nMsg(DBT_CREATE_SOCK,"Can't create socket"), NULL, NULL);
      logMsg("socket faild , rc=%d, errno=%d\n", s, errno);
      return 0;
    
    }

    if ((rc=bind(s, (struct sockaddr *)&sin, sizeof(sin)))<0) 
    {
        close(s);
	logMsg("bind() failed rc=%d, errno=%d => 161 is used \n", rc, errno);
        return 1;
    }
    logMsg("161 is not used \n", rc, errno);
    close(s);
    return 0;
}
/*
 *
 * Stop the master agent 
 *
 * This routine returns:
 *      0 success
 *     -1 error
 *
 * Read magt pid from <server-root>/admin-serv/logs/pid_masteragt
 * Send it signal 15
 * Remove logs/pid_masteragt
 *
 */
int magent_shutdown(int strict)
{
   char command[BIG_LINE];
   char path[PATH_MAX];
   FILE *fhdl;
   char buffer[20];
   char *ADMSERV_ROOT = getenv("ADMSERV_ROOT");
   pid_t  magtid;

   if (getuid() != 0)
     return(0);
 
   sprintf(path, "%s/../logs/%s", ADMSERV_ROOT, MASTERAGT_PID);
   logMsg("Looking for pid in (%s)\n", path);
   if ((fhdl = fopen(path, "r")) != (FILE *) NULL)
   {
     fscanf(fhdl, "%d\n", &magtid);
     fclose(fhdl);
     sprintf(command, "kill %d&", magtid);
     logMsg("Exceuting command: %s\n", command);
     if (!system(command))
     { 
       sprintf(command, "rm %s&", path);
       system(command);
       return(0);
     }
     else
     {
       rpt_err(SYSTEM_ERROR, i18nMsg(DBT_TERMINATE,"Failed to terminate Master agent"), NULL, NULL);
       return(-1);
     }
   }
   else {
     logMsg("Pid file not found => master agent is not running\n");
     if (strict) {
       rpt_err(SYSTEM_ERROR, i18nMsg(DBT_NOT_RUNNING,"Master agent is not running"), NULL, NULL);
	 }
     return(0);
   }
}

/*
 *
 * Restart the master agent 
 *
 * This routine returns:
 *      0 success
 *     -1 error
 *
 * Stop master agent if running
 * Start master agent
 *
 */
int magent_restart(char * root)
{
   if (getuid() != 0)
     return(0);

   if (magent_shutdown(0) != 0) {
     logMsg("In restart, shutdown failed\n");
     rpt_err(SYSTEM_ERROR, i18nMsg(DBT_TERMINATE,"Master agent failed to shut down"), NULL, NULL);
     return(-1);
   }

   if (magent_start(root) != 0) {
     logMsg("In restart, start failed\n");
     rpt_err(SYSTEM_ERROR, i18nMsg(DBT_START,"Master agent failed to start"), NULL, NULL );
     return(-1);
   }
}

/*
 *
 * Start the master agent 
 *
 * This routine returns:
 *      0 success
 *     -1 error
 *
 */
int magent_start(char * root)
{
  if (getuid() != 0)
    return(0);

  /* case 1: to see if you have permission to start master agent */
  {
    uid_t  uid;
    
    uid = geteuid();
    
    if(uid != 0){
      logMsg("In start, EUID != 0\n");
      rpt_err(SYSTEM_ERROR, i18nMsg(DBT_NO_PERM, "No Permission To Start Master Agent, EUID!=0"), NULL, NULL);
    }
  }

   /* case 2: to see if SMUX port is used. This tells if a smux agent running */
    if(smux_port_is_used()){
      logMsg("Can not start, smux port is used \n");
      rpt_err(SYSTEM_ERROR, i18nMsg(DBT_SMUX_PORT_USED, "Can't start master agent, SMUX port 199 is used by another process"), NULL, NULL);
    }

   /* case 3: to see if SNMP port is used. This tells if a snmp agent running */
    if(snmp_port_is_used()){
      logMsg("Can not start, snmp port is used \n");
      rpt_err(SYSTEM_ERROR, i18nMsg(DBT_SNMP_PORT_USED, "Can't start master agent, SNMP port 161 is used by another process"), NULL, NULL);
    }

  /* case 4: start our master agent if no exits from 3 cases above */
  {
    char *ADMSERV_ROOT = getenv("ADMSERV_ROOT");
    char command[BIG_LINE];
    char arg1[BIG_LINE];
    char arg2[BIG_LINE];
    char path[PATH_MAX];
    char errlog[PATH_MAX];
    FILE *fhdl;
    pid_t  magtid;
    char buffer[20];
	char err_text[128];
    int ret=0;

    sprintf(command, "%s/plugins/snmp/magt/magt", root);
    sprintf(arg1, "%s/plugins/snmp/magt/CONFIG", root);
    sprintf(arg2, "%s/plugins/snmp/magt/INIT", root);
    sprintf(errlog, "%s/plugins/snmp/magt/err.log", root);
    
    logMsg("Exceute start command: %s %s %s\n", command, arg1, arg2);

    /* master agent process running as a child should exit immediately
     * in the case of error (no zombie state) */
    signal(SIGCHLD, SIG_IGN);

    switch(magtid = fork()) {
      case -1:
      {
        rpt_err(SYSTEM_ERROR, i18nMsg(DBT_FORK,"Fork failed"), NULL, NULL);
        return -1;
      }
      case 0:
      {
        close(2);
        open(errlog, O_WRONLY|O_CREAT|O_TRUNC, 0644); 
        execl(command, command, arg1, arg2, NULL);
        sprintf(err_text, i18nMsg(DBT_EXECL, "Execl (%s) failed"), command);
        rpt_err(SYSTEM_ERROR, err_text, NULL, NULL);
        return -1;
      }
      default:
      {
        sprintf(path, "%s/../logs/%s", ADMSERV_ROOT, MASTERAGT_PID);
        logMsg("Write pid %d to %s\n", magtid, path);
        if ((fhdl = fopen(path, "w")) == (FILE *) NULL)
        {
          logMsg("Failed to open %s for writing\n", path);
          rpt_err(SYSTEM_ERROR,i18nMsg(DBT_OPEN_PIDFILE, "Can not open PID file for writing"), NULL, NULL);
          return -1;
        }

        fprintf(fhdl, "%d\n", magtid);  
        fclose(fhdl);
      }
    }
  }

  /* case 5: test to see if it really starts */
  {
    int magt_running, waittime=0;

    while (magt_running = magent_is_running()) {
      /* check if magt is listening on snmp and smux ports (wait up to 20 sec) */
      if (!smux_port_is_used() || !snmp_port_is_used()) {
         if(waittime < 20) {
           logMsg("sleep 1 sec\n");
            waittime++;
            sleep(1);
         }
         else {
             magt_running = 0;
             break;
	}
      }
      else {
        /* magt is running and listening on snmp and smux ports */
        break;
      }
   }
           
    if (!magt_running)
    {
      /* If the  err log file is not empty send its content to the client */
      struct stat buf;
      char errlog[PATH_MAX];
      sprintf(errlog, "%s/plugins/snmp/magt/err.log", root);
      if (stat(errlog, &buf) == 0 && buf.st_size > 0)
      {
        char inbuf[256];
        FILE *f = fopen(errlog, "r");
        if (f != NULL) {
          int line=1;
          fprintf(stdout, "Content-type: text/html\n\n");
          fprintf(stdout, "NMC_Status: 1\n");
          fprintf(stdout, "NMC_ErrInfo: %s\n", i18nMsg(DBT_START,"Master agent failed to start"));
          while(fgets(inbuf, sizeof(inbuf), f) != NULL) {
            logMsg("%s",inbuf);
            fprintf(stdout, "ERRLOG%d: %s",line++, inbuf);
          }
          fclose(f);
          exit (0);
        }
      }
    }              
    return (magt_running) ? 0 : -1 ;
  }
}
