/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * User/Group Directory Server get/set for the Admin Group
 *
 * Note:
 *     Terms "Global" and "Domain" are used interchangeable
 *     Terms "Local" and "Admin Group" are used interchangeable
 *
 * Get Operation
 * in:  op=getconfig
 * out: ugdsconfig.inforef=<reference (dn) to the global U/G directory>
 * out: ugdsconfig.globaldirurl=<url for the global U/G directory>
 * out: ugdsconfig.dirurl=<url for the local U/G directory>
 * out: ugdsconfig.binddn=<bind dn for the local U/G directory>
 * out: ugdsconfig.bindpw=<bind password for the local U/G directory>
 *
 * Set Operation
 * in: op=setconfig&ugdsconfig.inforef=<inforef dn | "default">
 * "default" is a special key-word which instructs that the inforef should be set
 * to reference the global domain U/G Directory entry
 *
 * or
 *
 * in: op=setconfig&ugdsconfig.dirurl=<url>&
 *      ugdsconfig.binddn=<binddn>&ugdsconfig.bindpw=<bindpw>
 *
 * If inforef parameter is present, the CGI does not look for the parameters dirurl, binddn, bindpw
 *
 */

#include "libadmin/libadmin.h"
#include "string.h"

#include "ldaputil/errors.h"
#include "ldaputil/ldaputil.h"
#include "ldaputil/dbconf.h"

#include "libadminutil/admutil.h"
#include "libadminutil/resource.h"
#include "libadminutil/admutil.h"
#include "libadminutil/distadm.h"

#include "libadmsslutil/psetcssl.h"
#include "libadmsslutil/admsslutil.h"
#include "libadmsslutil/certmgt.h"
#include "utf8.h"
#include "cert.h"

static void handle_getconfig();
static void handle_setconfig();
static char *getGlobalInfoDN(char *siedn);
static char *nonull_value(char *str);


/*
 * Logging function
 */
static FILE * logfp;
static int log_enabled = 0, enable_verified=0;
static void logMsg(char *format, ...) {
	char logfile[512];

	if (!log_enabled && enable_verified) return;

	/* Automatically enable logging if <cgi-name>.dbg file exists in the logs directory */
	if (!log_enabled && !enable_verified) {
		enable_verified = 1;
		sprintf(logfile, "%s/../logs/ugdsconfig.dbg", getenv("ADMSERV_ROOT"));
		logfp = fopen(logfile, "r");
		if (logfp == NULL) {
			return;
		}
		log_enabled = 1;
		fclose(logfp);
		logfp=NULL;
	}

	if (logfp==NULL) {
		sprintf(logfile, "%s/../logs/ugdsconfig.dbg", getenv("ADMSERV_ROOT"));
		logfp = fopen(logfile, "w");
	}

	if (logfp != NULL) {
		va_list ap;
		va_start(ap,format);
		vfprintf(logfp,format,ap);
		va_end(ap);
		fflush(logfp);
	}
}

/*
 * i18n conversions defines and function;
 * properties file  = "ugdsconfig.properties"
 */
#define RESOURCE_FILE "ugdsconfig"
#define resource_key(a,b)   a b

#define DBT_NO_METHOD 			resource_key(RESOURCE_FILE, "1")
#define DBT_UNKNOWN_METHOD 		resource_key(RESOURCE_FILE, "2")
#define DBT_NO_QUERY_STRING 		resource_key(RESOURCE_FILE, "3")
#define DBT_NO_OPERATION 		resource_key(RESOURCE_FILE, "4")
#define DBT_UNKNOWN_OPERATION		resource_key(RESOURCE_FILE, "5")
#define DBT_BUILD_LDAPINFO		resource_key(RESOURCE_FILE, "6")
#define DBT_GLOBAL_GETUSERDIR		resource_key(RESOURCE_FILE, "7")
#define DBT_LOCAL_GETUSERDIR		resource_key(RESOURCE_FILE, "8")
#define DBT_PARSE_SIEDN			resource_key(RESOURCE_FILE, "9")
#define DBT_GLOBAL_SETUSERDIR		resource_key(RESOURCE_FILE, "10")
#define DBT_LOCAL_SETUSERDIR		resource_key(RESOURCE_FILE, "11")
#define DBT_NO_SETPARAM			resource_key(RESOURCE_FILE, "12")
#define DBT_BAD_DIRURL			resource_key(RESOURCE_FILE, "13")
#define DBT_ADMSSL_INIT_FAILED	        resource_key(RESOURCE_FILE, "14")

static  char *acceptLanguage = (char*)"en";
static  Resource *i18nResource = NULL;

static void i18nInit() {
	char resPath[256], *execPath;
	execPath = getcwd(resPath, 256);
	if (execPath) {
		strcpy(resPath, execPath);
		strcat(resPath, "/property");
		i18nResource = res_init_resource(resPath, RESOURCE_FILE);
	}
	if (getenv("HTTP_ACCEPT_LANGUAGE")) {
		acceptLanguage = getenv("HTTP_ACCEPT_LANGUAGE");
	}
}
static void i18nCleanup() {
	if (i18nResource != NULL) {
	res_destroy_resource(i18nResource);
	}
}
static const char *i18nMsg(char *msgid, char *defaultMsg) {
	const char *msg=NULL;

	if (i18nResource) {
            msg = res_getstring(i18nResource, msgid, acceptLanguage);
	}
	if (msg == NULL) {
		msg = (const char*)defaultMsg;
	}
	return msg;
}

static char * nonull_value(char *str) {
	return (str!=NULL) ? str : (char *)"";
}

static char error_info[1024];

/*
 *  Main
 */
int main(int argc, char *argv[])
{
    int _ai=ADMUTIL_Init();

    char *method = getenv("REQUEST_METHOD");
    char *root = getenv("NETSITE_ROOT");
    char *qs = 0, *op=0;

    logMsg(" In %s\n", argv[0]);

    i18nInit();
    fprintf(stdout, "Content-type: text/html;charset=utf-8\n\n");

    /* GET or POST method */
    if (!method || !*method) {
        /* non UI CGI */
        rpt_err(SYSTEM_ERROR, i18nMsg(DBT_NO_METHOD,"No method is specified"), NULL, NULL);
    }
    else if (!strcmp(method, "GET")) {
        /* UI CGI - either upgrade or normal admin CGI */
        qs = getenv("QUERY_STRING");

        if (!qs || !*qs) {
            rpt_err(INCORRECT_USAGE, i18nMsg(DBT_NO_QUERY_STRING ,"NO QUERY_STRING DATA"), NULL, NULL);
        }
        else {
            get_begin(qs);
        }
    }
    else if (!strcmp(method, "POST")) {
        post_begin(stdin);
    }
    else {
        sprintf(error_info, i18nMsg(DBT_UNKNOWN_METHOD,"Unknown Method (%s)"), method);
        rpt_err(INCORRECT_USAGE, error_info, NULL, NULL);
    }

    logMsg("method=%s\n", method);

    op=get_cgi_var("op", NULL, NULL);

	logMsg("op=%s\n", op);

	if (op == NULL) {
        rpt_err(INCORRECT_USAGE, i18nMsg(DBT_NO_OPERATION,"No operation is defined"), NULL, NULL);
    }
	if (strcmp(op,"getconfig") == 0) {
        handle_getconfig();
    }
    else if (strcmp(op,"setconfig") == 0) {
		handle_setconfig();
    }
	else {
        sprintf(error_info, i18nMsg(DBT_UNKNOWN_OPERATION,"Unknown Operation (%s)"), op);
        rpt_err(INCORRECT_USAGE, error_info, NULL, NULL);
    }

    return 0;
}


/*
 * Return current U/G directory setting
 */
static void handle_getconfig()
{
	char *inforef=NULL, *globaldirurl=NULL, *dirurl=NULL, *binddn=NULL, *bindpw=NULL;
	AdmldapInfo adminfo;
	char *admroot = getenv("ADMSERV_ROOT");
	int rc;

	logMsg("In handle_getconfig\n");

	if(ADMSSL_InitSimple()) {
	  rpt_err(SYSTEM_ERROR, i18nMsg(DBT_ADMSSL_INIT_FAILED,"Cannot initialize SSL"), NULL, NULL);
	}

	adminfo = admldapBuildInfo(admroot, &rc);
	if (adminfo == NULL) {
		logMsg("admldapBuildInfo failed, rc=%d, admroot=%s\n", rc, nonull_value(admroot));
		sprintf(error_info, i18nMsg(DBT_BUILD_LDAPINFO,"Failed to build ldap info (err=%d)"), rc);
		rpt_err(SYSTEM_ERROR, error_info, NULL, NULL);
	}

	if (admldapGetSecurity(adminfo)) {
	        if (!admldapBuildInfoSSL(adminfo, &rc)) {
		        logMsg("admldapBuildInfo failed, rc=%d, admroot=%s\n", rc, nonull_value(admroot));
		        sprintf(error_info, i18nMsg(DBT_BUILD_LDAPINFO,"Failed to build ldap info (err=%d)"), rc);
			rpt_err(SYSTEM_ERROR, error_info, NULL, NULL);
		}
	}

	if (!admldapGetDomainUserDirectory(adminfo, &globaldirurl, &binddn, &bindpw, &inforef, &rc)) {
		logMsg("admldapGetDomainUserDirectory failed, rc=%d\n", rc);
		sprintf(error_info, i18nMsg(DBT_GLOBAL_GETUSERDIR,"Failed to read Domain User Directory Info (err=%d)"), rc);
		rpt_err(SYSTEM_ERROR, error_info, NULL, NULL);
	}

	if (!admldapGetAdmGrpUserDirectory(adminfo, &dirurl, &binddn, &bindpw, &inforef, &rc)) {
		logMsg("admldapGetAdmGrpUserDirectory failed, rc=%d\n", rc);
		sprintf(error_info, i18nMsg(DBT_LOCAL_GETUSERDIR,"Failed to read Admin Group User Directory Info (err=%d)"), rc);
		rpt_err(SYSTEM_ERROR, error_info, NULL, NULL);
	}

	rpt_success(NULL);

	fprintf(stdout, "ugdsconfig.inforef:%s\n", nonull_value(inforef));
	logMsg("ugdsconfig.inforef:%s\n", nonull_value(inforef));

	fprintf(stdout, "ugdsconfig.globaldirurl:%s\n", nonull_value(globaldirurl));
	logMsg("ugdsconfig.globaldirurl:%s\n", nonull_value(globaldirurl));

	fprintf(stdout, "ugdsconfig.dirurl:%s\n", nonull_value(dirurl));
	logMsg("ugdsconfig.dirurl:%s\n", nonull_value(dirurl));

	fprintf(stdout, "ugdsconfig.binddn:%s\n", nonull_value(binddn));
	logMsg("ugdsconfig.binddn:%s\n", nonull_value(binddn));

	fprintf(stdout, "ugdsconfig.bindpw:%s\n", nonull_value(bindpw));
	logMsg("ugdsconfig.bindpw:%s\n", nonull_value(bindpw));
}


/*
 * Modify U/G Directory setting
 */
static void handle_setconfig() {

	char *inforef=NULL, *dirurl=NULL, *binddn=NULL, *bindpw=NULL;
	char *admroot = getenv("ADMSERV_ROOT");
	char *siedn;
	AdmldapInfo adminfo;
	LDAPURLDesc *url;
	int rc;

	logMsg("In handle_setconfig\n");

	if(ADMSSL_InitSimple()) {
	  rpt_err(SYSTEM_ERROR, i18nMsg(DBT_ADMSSL_INIT_FAILED,"Cannot initialize SSL"), NULL, NULL);
	}

	inforef = get_cgi_var( "ugdsconfig.inforef", NULL, NULL );
	dirurl = localToUTF8((const char*)get_cgi_var( "ugdsconfig.dirurl", NULL, NULL ));
	binddn = get_cgi_var( "ugdsconfig.binddn", NULL, NULL );
	bindpw = get_cgi_var( "ugdsconfig.bindpw", NULL, NULL );

	logMsg("inforef=%s\n", nonull_value(inforef));
	logMsg("dirurl=%s\n",  nonull_value(dirurl));
	logMsg("binddn=%s\n", nonull_value(binddn));
	logMsg("bindpw size=%d\n", strlen(nonull_value(bindpw)));


	adminfo = admldapBuildInfo(admroot, &rc);
	if (adminfo == NULL) {
		logMsg("admldapBuildInfo failed, rc=%d, admroot=%s\n", rc, nonull_value(admroot));
		sprintf(error_info, i18nMsg(DBT_BUILD_LDAPINFO,"Failed to build ldap info (err=%d)"), rc);
		rpt_err(SYSTEM_ERROR, error_info, NULL, NULL);
	}

	if (admldapGetSecurity(adminfo)) {
	        if (!admldapBuildInfoSSL(adminfo, &rc)) {
		        logMsg("admldapBuildInfo failed, rc=%d, admroot=%s\n", rc, nonull_value(admroot));
		        sprintf(error_info, i18nMsg(DBT_BUILD_LDAPINFO,"Failed to build ldap info (err=%d)"), rc);
			rpt_err(SYSTEM_ERROR, error_info, NULL, NULL);
		}
	}

	if (inforef != NULL) {
		if (strcasecmp(inforef,"default")==0) {
			siedn = admldapGetSIEDN(adminfo);
			inforef = getGlobalInfoDN(siedn);
		}

		if (!admldapSetAdmGrpUserDirectoryCGI(NULL, NULL, NULL, inforef, &rc)) {
			logMsg("admldapSetAdmGrpDirectoryCGI failed, rc=%d, inforef=%s\n", rc, nonull_value(inforef));
			sprintf(error_info, i18nMsg(DBT_GLOBAL_SETUSERDIR,"Failed to set link to the Domain User Directory Info (err=%d)"), rc);
			rpt_err(SYSTEM_ERROR, error_info, NULL, NULL);
		}
	}
	else {
		if (dirurl==NULL) {
			logMsg("param ugdsconfig.dirurl missing");
			sprintf(error_info, i18nMsg(DBT_NO_SETPARAM,"Parameter %s missing"), "ugdsconfig.dirurl");
			rpt_err(INCORRECT_USAGE, error_info, NULL, NULL);
		}
		/*if (binddn==NULL) {
			logMsg("param ugdsconfig.binddn missing");
			sprintf(error_info, i18nMsg(DBT_NO_SETPARAM,"Parameter %s missing"), "ugdsconfig.binddn");
			rpt_err(INCORRECT_USAGE, error_info, NULL, NULL);
		}*/
		/*if (bindpw==NULL) {
			logMsg("param ugdsconfig.bindpw missing");
			sprintf(error_info, i18nMsg(DBT_NO_SETPARAM,"Parameter %s missing"), "ugdsconfig.bindpw");
			rpt_err(INCORRECT_USAGE, error_info, NULL, NULL);
		}*/

		if ((rc = ldap_url_parse(dirurl, & url)) != 0) {
			logMsg("Bad ugdsconfig.dirurl");
			sprintf(error_info, i18nMsg(DBT_BAD_DIRURL,"Bad LDAP URL value (%s)"), dirurl);
			rpt_err(INCORRECT_USAGE, error_info, NULL, NULL);
		}

		if (!admldapSetAdmGrpUserDirectoryCGI(dirurl, binddn, bindpw, NULL, &rc)) {
			logMsg("admldapSetAdmGrpDirectoryCGI failed, rc=%d\n", rc);
			sprintf(error_info, i18nMsg(DBT_LOCAL_SETUSERDIR,"Failed to set User Directory Parameters for the Admin Group (err=%d)"), rc);
			rpt_err(SYSTEM_ERROR, error_info, NULL, NULL);
		}
	}

	rpt_success(NULL);
}


/*
 * Return the U/G Directory Info DN for the domain:
 * "cn=UserDirectory, ou=Globalpreferences,ou=<Admin Domain>,o=NetscapeSuiteSpot"
 * The <Admin Domain> name is extractioned from the sie DN. The domain name is the second
 * RDN in the sie DN from right to left.
 *
 * Example: sie dn="admin-serv-goo,ou=Netscape Adminstration Server,cn=SuiteSpot 4.0,
 * cn=goo.mcom.com, ou=Netscape SuiteSpot, o=NetscapeRoot"
 * The admin domain name is "Netscape SuiteSpot"
 * Returning  cn=UserDirectory, ou=Globalpreferences,ou=Netscape SuiteSpot,o=NetscapeSuiteSpot
 */
static char *getGlobalInfoDN(char *siedn_p) {
	char *comma, *domain;
	char *siedn=strdup(siedn_p);
	char buf[1024];

	memset(buf,0,sizeof(buf));
	logMsg("getGlobalInfoDN for %s\n", nonull_value(siedn));
	comma = strrchr(siedn, ',');
	if (comma == NULL) {
		logMsg("getGlobalInfoDN comma1 not found\n");
		rpt_err(SYSTEM_ERROR, i18nMsg(DBT_PARSE_SIEDN,"Failed to parse SIE DN"), NULL, NULL);
	}
	*comma=0;

	comma = strrchr(siedn, ',');
	if (comma == NULL) {
		logMsg("getGlobalInfoDN comma2 not found\n");
		rpt_err(SYSTEM_ERROR, i18nMsg(DBT_PARSE_SIEDN,"Failed to parse SIE DN"), NULL, NULL);
	}

	domain=comma+1;

	sprintf(buf,"cn=UserDirectory,ou=Global Preferences,%s,o=NetscapeRoot", domain);
	logMsg("global info dn=\"%s\"\n", buf);
	return strdup(buf);
}
