/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/* Description (userauth.c)
 *
 *	This module contains routines used by administration CGI
 *	programs to return user authorization information
 * */
#ifdef XP_UNIX
#include <unistd.h>
#endif
#include <stdio.h>
#include <string.h>
#include "libadminutil/distadm.h"
#include "libadminutil/admutil.h"
#include "libadminutil/resource.h"
#include "libadmsslutil/admsslutil.h"
#include "ldap.h"
#include <ldap_ssl.h>

#define RESOURCE_FILE "userauth"

#define resource_key(a,b)   a b

#define DBT_NO_USER_NAME 	resource_key(RESOURCE_FILE, "1")
#define DBT_NO_LDAP_INFO 	resource_key(RESOURCE_FILE, "2")
#define DBT_NO_SR_INFO 	    resource_key(RESOURCE_FILE, "3")
#define DBT_ADMSSL_INIT_FAILED	resource_key(RESOURCE_FILE, "4")
#define DBT_BUILD_LDAPINFO	resource_key(RESOURCE_FILE, "5")
#define DBT_LOCAL_GETUSERDIR 	resource_key(RESOURCE_FILE, "6")


static char* getUserDirectory();

static Resource   *i18nResource = NULL;
static char       *acceptLanguage = "en";
static const char *i18nMsg(char *msgid, char *defaultMsg);

int main(int argc, char *argv[])
{
  int            _ai = ADMUTIL_Init();
  AdmldapInfo    ldapInfo;
  char           admconf[256];
  char           *method;
  int            rv;
  int            err;
  int            *errp = &err;
  char           *username = 0;
  char           *localAdmin = 0;
  char           *binddn = 0;
  char           *bindpw = 0;
  char           *basedn;
  char           *userdir;
  char           *nsroot = getenv("NETSITE_ROOT");
  char           *pw_expiring = getenv("PASSWORD_EXPIRING");
  char           *lang=getenv("HTTP_ACCEPT_LANGUAGE");
  char           resPath[256], *execPath;

  execPath = getcwd(resPath, 256);
  if (execPath) {
    strcpy(resPath, execPath);
    strcat(resPath, "/property");
    i18nResource = res_init_resource(resPath, RESOURCE_FILE);
  }

  if (lang) acceptLanguage = strdup(lang);

  memset((void *)errp, 0, sizeof(int));
  method = getenv("REQUEST_METHOD");

  /* Get UserDN and User Password  */

  rv = ADM_GetUserDNString(&err, &binddn);
  if (rv < 0 || !binddn || !*binddn) {
    /* No USER DN, */
    rv = ADM_GetCurrentUsername(&err, &username);
    if (rv < 0 || !username || !*username) {
      if (i18nResource) 
	rpt_err(APP_ERROR, 
		res_getstring(i18nResource, DBT_NO_USER_NAME, acceptLanguage),
		NULL, NULL);
      else rpt_err(APP_ERROR, "NO USER NAME", NULL, NULL);
    }
    else {
      /* No DN, maybe it is local super */
      localAdmin = admGetLocalAdmin(NULL, &rv);
      if (localAdmin &&  !strcmp(username, localAdmin)) {
	/* It is local Admin !!! */
	binddn = NULL;
      }
      else {
	if (i18nResource)
	  rpt_err(APP_ERROR, 
		  res_getstring(i18nResource, DBT_NO_USER_NAME, acceptLanguage),
		  NULL, NULL);
	else rpt_err(APP_ERROR, "NO USER NAME", NULL, NULL);
      }
    }
  }
  
  if (nsroot) { 
    sprintf(admconf, "%s%c%s%c%s", nsroot, FILE_PATHSEP, "admin-serv",
	    FILE_PATHSEP, "config");
    ldapInfo = admldapBuildInfo(admconf, &rv);

    if (!ldapInfo) {
      if (i18nResource) 
      	rpt_err(APP_ERROR, 
		res_getstring(i18nResource, DBT_NO_LDAP_INFO, acceptLanguage),
		NULL, NULL);
      else rpt_err(APP_ERROR, "NO LDAP INFO", NULL, NULL);
    }
    else {

	userdir=getUserDirectory();

      /*rpt_success(NULL);*/
	fprintf(stdout, "Content-type: text/html\n\n");
      if (binddn) fprintf(stdout, "UserDN: %s\n", binddn);
      else fprintf(stdout, "UserDN: %s\n", admldapGetSIEDN(ldapInfo));
	fprintf(stdout, "UserDirectory: %s\n", userdir);
      fprintf(stdout, "ldapHost: %s\n", admldapGetHost(ldapInfo));
      fprintf(stdout, "ldapPort: %d\n", admldapGetPort(ldapInfo));
      if (admldapGetSecurity(ldapInfo)) fprintf(stdout, "ldapSecurity: on\n");
      else fprintf(stdout, "ldapSecurity: off\n");
      if (basedn = admldapGetBaseDN(ldapInfo)) {
	fprintf(stdout, "ldapBaseDN: %s\n", basedn);
      }
      fprintf(stdout, "SIE: %s\n", admldapGetSIEDN(ldapInfo));
      if (pw_expiring) fprintf(stdout, "NW_PASSWD_EXPIRING: %s\n", pw_expiring);
	fprintf(stdout, "NMC_Status: 0\n");
      return(0);
    }
  }
  else {
    if (!ldapInfo)
      if (i18nResource) 
	rpt_err(APP_ERROR, 
		res_getstring(i18nResource, DBT_NO_SR_INFO, acceptLanguage),
		NULL, NULL);
      else rpt_err(APP_ERROR, "NO Server Root Infomation", NULL, NULL);
  }
}

/*
 * A helper function for retrieving i18n message
 */
static const char *i18nMsg(char *msgid, char *defaultMsg) {
	const char *msg=NULL;
	if (i18nResource) {
            msg = res_getstring(i18nResource, msgid, acceptLanguage);
      }
	return msg ? msg : (const char*)defaultMsg;
}

/*
 * Return U/G directory for the Server Group
 * If DS is down, do not report error, but return "?"
 */
static char* getUserDirectory()
{
	char *inforef=NULL, *dirurl=NULL, *binddn=NULL, *bindpw=NULL;
	AdmldapInfo adminfo;
	char *admroot = getenv("ADMSERV_ROOT");
	char error_info[1024];
	int rc;

	if(ADMSSL_InitSimple()) {
	  rpt_err(SYSTEM_ERROR, i18nMsg(DBT_ADMSSL_INIT_FAILED,"Cannot initialize SSL"), NULL, NULL);
	}

	adminfo = admldapBuildInfo(admroot, &rc);
	if (adminfo == NULL) {
		sprintf(error_info, i18nMsg(DBT_BUILD_LDAPINFO, "Failed to build ldap info (err=%d)"), rc);
		rpt_err(SYSTEM_ERROR, error_info, NULL, NULL);
	}


	if (admldapGetSecurity(adminfo)) {
	        if (!admldapBuildInfoSSL(adminfo, &rc)) {
		      sprintf(error_info, i18nMsg(DBT_BUILD_LDAPINFO, "Failed to build ldap info (err=%d)"), rc);
                      /*Miodrag said don't fail here, let it continue so cgi can complete it executation path*/
                      /*rpt_err(SYSTEM_ERROR, error_info, NULL, NULL);*/
                      /* system will continue executing the cgi and return 'dirurl="?"' rather then return error message*/
		}
	}

	if (!admldapGetAdmGrpUserDirectory(adminfo, &dirurl, &binddn, &bindpw, &inforef, &rc)) {
		if (rc==UG_NULL_LDAP) {
			/*
			 * Can not talk to Directory Server, probably is down
			 * Do not return error here, allow the console to proceed;
			 * The user can restart the DS remotly from the console
			*/
			dirurl="?";
		}
		else {
			sprintf(error_info, i18nMsg(DBT_LOCAL_GETUSERDIR,"Failed to read Admin Group User Directory Info (err=%d)"), rc);
			rpt_err(SYSTEM_ERROR, error_info, NULL, NULL);
		}
	}

	return strdup(dirurl);
}
