/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * viewdata.c:  View generic server parameters in HTML Admin Interface.
 * 
 * Initial version: 3/11/99 by Adam Prishtina
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <base/util.h>
#include "libadminutil/distadm.h"
#include "libadminutil/admutil.h"
#include "libadminutil/resource.h"
#include "libadmsslutil/admsslutil.h"
#include "libadmin/libadmin.h"
#include "libadmin/cluster.h"
#include "ldapu.h"
#include <ldap_ssl.h>
#include "setupinst.h"

#define MY_PAGE "viewdata.html"

#define NBUF_SIZE 1024

/*
 * i18n conversions defines and function;
 * properties file  = "viewdata.properties"
 */
#define RESOURCE_FILE "viewdata"
#define resource_key(a,b)   a b

#define DBT_INV_NUMBER 			resource_key(RESOURCE_FILE, "1")
#define DBT_INV_NUMBER_DESC 		resource_key(RESOURCE_FILE, "2")
#define DBT_SUBTITLE 		        resource_key(RESOURCE_FILE, "3")
#define DBT_WITH   		        resource_key(RESOURCE_FILE, "4")

/* netcenter links */
#define DBT_NETCENTER_GET_REQ             resource_key(RESOURCE_FILE, "10")
#define DBT_NETCENTER_HOST                resource_key(RESOURCE_FILE, "11")
#define DBT_NETCENTER_PORT                resource_key(RESOURCE_FILE, "12")
#define DBT_NETCENTER_SERVER_PRODUCT_URL  resource_key(RESOURCE_FILE, "13")
#define DBT_NETCENTER_ADMIN_ID            resource_key(RESOURCE_FILE, "14")

/* html resource */
#define DBT_OUTPUT_DATA_SERVER_PRODUCT_NAME	resource_key(RESOURCE_FILE, "20")
#define DBT_OUTPUT_DATA_DATE 		        resource_key(RESOURCE_FILE, "21")
#define DBT_OUTPUT_DATA_SERVER_ROOT 		resource_key(RESOURCE_FILE, "22")
#define DBT_OUTPUT_DATA_SERVER_PORT 		resource_key(RESOURCE_FILE, "23")
#define DBT_OUTPUT_DATA_PRODUCT_NAME 		resource_key(RESOURCE_FILE, "24")
#define DBT_OUTPUT_DATA_VENDOR 		        resource_key(RESOURCE_FILE, "25")
#define DBT_OUTPUT_DATA_PRODUCT_VERSION 	resource_key(RESOURCE_FILE, "26")
#define DBT_OUTPUT_DATA_BUILD_NUMBER 		resource_key(RESOURCE_FILE, "27")
#define DBT_OUTPUT_DATA_BUILD_SECURITY 		resource_key(RESOURCE_FILE, "28")
#define DBT_OUTPUT_DATA_REVISION_NUMBER 	resource_key(RESOURCE_FILE, "29")
#define DBT_OUTPUT_DATA_DESCRIPTION 		resource_key(RESOURCE_FILE, "30")
#define DBT_MAIN_TABLE_HEADER 		        resource_key(RESOURCE_FILE, "31")
#define DBT_MAIN_TABLE_FOOTER 		        resource_key(RESOURCE_FILE, "32")
#define DBT_MAIN_SERVER_ID 		        resource_key(RESOURCE_FILE, "33")
#define DBT_MAIN_PRODUCT_VERSION 		resource_key(RESOURCE_FILE, "34")
#define DBT_MAIN_UPGRADE_AVAILABLE 		resource_key(RESOURCE_FILE, "35")
#define DBT_MAIN_PRODUCT_URL 		        resource_key(RESOURCE_FILE, "36")


static  char *acceptLanguage = (char*)"en";
static  Resource *i18nResource = NULL;

char*
getResourceString(char *key) {
  return (char *)(res_getstring(i18nResource, key, acceptLanguage));
}

static void i18nInit() {
	char resPath[256], *execPath;
	execPath = getcwd(resPath, 256);
	if (execPath) {
		strcpy(resPath, execPath);
		strcat(resPath, "/property");
		i18nResource = res_init_resource(resPath, RESOURCE_FILE);
	}
	if (getenv("HTTP_ACCEPT_LANGUAGE")) {
		acceptLanguage = getenv("HTTP_ACCEPT_LANGUAGE");
	}
}


/*
 * Get the AdmldapInfo struct for ldap SDK API.
 */

AdmldapInfo get_adm_ldapinfo() {

  AdmldapInfo ldapInfo = NULL;
  int rv;
  char admroot[BIG_LINE];
  
  sprintf(admroot, "%s%cadmin-serv%cconfig",
	  getenv("NETSITE_ROOT"),
	  FILE_PATHSEP,
	  FILE_PATHSEP);
  
  ldapInfo = admldapBuildInfo(admroot, &rv);
  if(!ldapInfo)
    exit(0);
  
  rv = ADMSSL_InitSimple();
  if(rv)
    exit(0);
  
  return(ldapInfo);
}



/* 
 * Get bind DN and bind PW info. 
 */

int get_bindinfo(char **binddn, char **bindpw) {

  int rv;
  int err;
  char *username = 0;
  char *localAdmin = 0;

  rv = ADM_GetUserDNString(&err, binddn); /* & of (*binddn) */
  if (rv < 0 || !*binddn || !**binddn) {
    rv = ADM_GetCurrentUsername(&err, &username);
    if (rv < 0 || !username || !*username)
      exit(0);
    else {
      /* No DN, maybe it is local super */
      localAdmin = admGetLocalAdmin(NULL, &rv);
      if (localAdmin) {
	if (strcmp(username, localAdmin))
	  exit(0);
	else {
	  *binddn = NULL;
	  *bindpw = NULL;
	}
      }
      else
	exit(0);
    }
  }
  
  if (*binddn) rv = ADM_GetCurrentPassword(&err, bindpw);
  return 1;
}


LDAP *server_bind(char *host, int port, int security, char *binddn, char *bindpw) {

  int ver = LDAP_VERSION3;
  int rv;
  int ldapError;

  LDAP *server;

  if(security) {
    if(!(server = ldapssl_init(host, port, 1)))
      return NULL;
  }
  else {
    if(!(server = ldap_init(host, port)))
      return NULL;
  }

  rv = ldap_set_option(server, LDAP_OPT_PROTOCOL_VERSION, &ver);
  if(rv != LDAP_SUCCESS)
    return NULL;

  if ((ldapError = ldap_simple_bind_s(server, binddn, bindpw))
      != LDAP_SUCCESS ) {
    switch (ldapError) {
    case LDAP_INAPPROPRIATE_AUTH:
    case LDAP_INVALID_CREDENTIALS:
    case LDAP_INSUFFICIENT_ACCESS:
      /* authenticate failed: Should not continue */
#ifdef LDAP_DEBUG
      ldap_perror( ld, "ldap_simple_bind_s" );
#endif
      return NULL;
    case LDAP_NO_SUCH_OBJECT:
    case LDAP_ALIAS_PROBLEM:
    case LDAP_INVALID_DN_SYNTAX:
      /* Not a good DN */
      return NULL;
    default:
      ldap_unbind(server);
      return NULL;
    }
  }

  return server;
}


int get_product_url(LDAP *server, char *sie, char **text, char **url) {

  char *dn;
  char *domain;
  char *tmp, *tmp2;

  LDAPMessage *entry;
  int ldapError;
  char **vals;

  *url = NULL;
  *text = NULL;

  /* Get the domain */
  tmp = strdup(sie);
  tmp2 = strrchr(tmp, ',');
  tmp2[0] = '\0';
  domain = strrchr(tmp, ',');
  domain++;
  tmp2[0] = ',';

  while(*domain == ' ')
    domain++;

  /* form the rest of the url */

  if(strstr(sie, "Messaging")) {
    dn = (char *)malloc(strlen("ou=Messaging, ou=Global Preferences, ") + strlen(domain) + 1);
    sprintf(dn, "ou=Messaging, ou=Global Preferences, %s", domain);
  }
  else if(strstr(sie, "Administration")) {
    dn = (char *)malloc(strlen("ou=Admin, ou=Global Preferences, ") + strlen(domain) + 1);
    sprintf(dn, "ou=Admin, ou=Global Preferences, %s", domain);
  }
  else if(strstr(sie, "Directory")) {
    dn = (char *)malloc(strlen("ou=Directory, ou=Global Preferences, ") + strlen(domain) + 1);
    sprintf(dn, "ou=Directory, ou=Global Preferences, %s", domain);
  }
  else if(strstr(sie, "Enterprise")) {
    dn = (char *)malloc(strlen("ou=Enterprise, ou=Global Preferences, ") + strlen(domain) + 1);
    sprintf(dn, "ou=Enterprise, ou=Global Preferences, %s", domain);
  }
  else if(strstr(sie, "Certificate")) {
    dn = (char *)malloc(strlen("ou=Certificate, ou=Global Preferences, ") + strlen(domain) + 1);
    sprintf(dn, "ou=Certificate, ou=Global Preferences, %s", domain);
  }
  else if(strstr(sie, "Application")) {
    dn = (char *)malloc(strlen("ou=Application, ou=Global Preferences, ") + strlen(domain) + 1);
    sprintf(dn, "ou=Application, ou=Global Preferences, %s", domain);
  }
  else if(strstr(sie, "Biller")) {
    dn = (char *)malloc(strlen("ou=Biller, ou=Global Preferences, ") + strlen(domain) + 1);
    sprintf(dn, "ou=Biller, ou=Global Preferences, %s", domain);
  }

  if((ldapError = ldap_search_s(server, dn, LDAP_SCOPE_BASE,
				"(objectclass=*)", NULL, 0, &entry)) != LDAP_SUCCESS)
    return 1;

  if(vals = ldap_get_values(server, entry, "nshtmladminproducturl")) {
    *url = strdup(vals[0]);
    ldap_value_free(vals);
  }
  if(vals = ldap_get_values(server, entry, "nshtmladminproducttext")) {
    *text = strdup(vals[0]);
    ldap_value_free(vals);
  }

  if(*url && *text)
    return 0;
  return 1;
}


void output_data(LDAP *server, char *sie) {

  LDAPMessage *entry;
  char **vals;
  int ldapError;

  char *tmp = strdup(sie);
  char *isie;

  /* SIE has some data... */

  if((ldapError = ldap_search_s(server, sie, LDAP_SCOPE_BASE,
				"(objectclass=*)", NULL, 0, &entry)) != LDAP_SUCCESS)
    return;


  if(vals = ldap_get_values(server, entry, "serverproductname")) {
    fprintf(stdout,(const char*)getResourceString(DBT_OUTPUT_DATA_SERVER_PRODUCT_NAME), vals[0]);
    ldap_value_free(vals);
  }


  if(vals = ldap_get_values(server, entry, "installationtimestamp")) {
    struct tm tm;
    char buf[BIG_LINE];

    /* only PARSE YYYYmmddHHMMSS */
    sscanf(vals[0], "%4d%2d%2d%2d%2d%2d", &tm.tm_year, &tm.tm_mon, 
           &tm.tm_mday, &tm.tm_hour, &tm.tm_min, &tm.tm_sec);
        
    tm.tm_year -= 1900;  /* the number of years since 1900 */
    tm.tm_mon  -= 1;     /* The number of month since January, in the range 0 to 11 */

#ifdef LINUX
    strftime(buf, BIG_LINE, "%b %d, %Y %T %p", &tm);
    tzset();
    sprintf(buf, "%s %s", buf, daylight ? tzname[1] : tzname[0]);
#else
    strftime(buf, BIG_LINE, "%b %d, %Y %T %p %Z", &tm);
#endif


    fprintf(stdout, (const char*)getResourceString(DBT_OUTPUT_DATA_DATE), buf);
    ldap_value_free(vals);
  }


  if(vals = ldap_get_values(server, entry, "serverroot")) {
    fprintf(stdout, (const char*)getResourceString(DBT_OUTPUT_DATA_SERVER_ROOT), vals[0]);
    ldap_value_free(vals);
  }

  if(!(vals = ldap_get_values(server, entry, "nsserverport"))) {
    /* argh, port can be in the configuration object */

    char *config_buf = (char *)malloc(strlen("cn=configuration, ") + strlen(sie) + 1);
    sprintf(config_buf, "cn=configuration, %s", sie);
				      
    if((ldapError = ldap_search_s(server, config_buf, LDAP_SCOPE_BASE,
				  "(objectclass=*)", NULL, 0, &entry)) != LDAP_SUCCESS)
      return;

    vals = ldap_get_values(server, entry, "nsserverport");
    free(config_buf);
  }
  if(vals) {
    fprintf(stdout, (const char*)getResourceString(DBT_OUTPUT_DATA_SERVER_PORT), vals[0]);
    ldap_value_free(vals);
  }

  /* ... and ISIE has the rest. */

  isie = strtok(tmp, ",");
  isie = strtok(NULL, "\0");
  while(*isie == ' ')
    isie++; /* eliminate spaces */

  if((ldapError = ldap_search_s(server, isie, LDAP_SCOPE_BASE,
				"(objectclass=*)", NULL, 0, &entry)) != LDAP_SUCCESS)
    return;

  if(vals = ldap_get_values(server, entry, "nsproductname")) {
    fprintf(stdout, (const char*)getResourceString(DBT_OUTPUT_DATA_PRODUCT_NAME), vals[0]);
    ldap_value_free(vals);
  }

  if(vals = ldap_get_values(server, entry, "nsvendor")) {
    fprintf(stdout, (const char*)getResourceString(DBT_OUTPUT_DATA_VENDOR), vals[0]);
    ldap_value_free(vals);
  }

  if(vals = ldap_get_values(server, entry, "nsproductversion")) {
    fprintf(stdout, (const char*)getResourceString(DBT_OUTPUT_DATA_PRODUCT_VERSION), vals[0]);
    ldap_value_free(vals);
  }

  if(vals = ldap_get_values(server, entry, "nsbuildnumber")) {
    fprintf(stdout,(const char*)getResourceString(DBT_OUTPUT_DATA_BUILD_NUMBER), vals[0]);
    ldap_value_free(vals);
  }

  if(vals = ldap_get_values(server, entry, "nsbuildsecurity")) {
    fprintf(stdout, (const char*)getResourceString(DBT_OUTPUT_DATA_BUILD_SECURITY), vals[0]);
    ldap_value_free(vals);
  }

  if(vals = ldap_get_values(server, entry, "nsrevisionnumber")) {
    fprintf(stdout, (const char*)getResourceString(DBT_OUTPUT_DATA_REVISION_NUMBER), vals[0]);
    ldap_value_free(vals);
  }

  if(vals = ldap_get_values(server, entry, "description")) {
    fprintf(stdout, (const char*)getResourceString(DBT_OUTPUT_DATA_DESCRIPTION), vals[0]);
    ldap_value_free(vals);
  }

}

int main(int argc, char *argv[])
{

    int _ai = ADMUTIL_Init();
    char *qs = getenv("QUERY_STRING");
    char line[BIG_LINE];
    FILE *html = open_html_file(MY_PAGE);
    char *sie;

    char *request;
    int errorcode;
    PRFileDesc *sockd;
    bufstruct *nbuf;

    char *binddn, *bindpw;
    LDAP *server;
    LDAPMessage *entry;
    int ldapError;

    AdmldapInfo ldapInfo = get_adm_ldapinfo();

    i18nInit();

    if(!get_bindinfo(&binddn, &bindpw))
      exit(0);
    
    server = server_bind(admldapGetHost(ldapInfo),
			 admldapGetPort(ldapInfo),
			 admldapGetSecurity(ldapInfo),
			 binddn,
			 bindpw);
    
    fprintf(stdout, "Content-type: text/html;charset=utf-8\n\n");

    if(qs)  {
       get_begin(qs);
       sie=get_cgi_var("sie", NULL, NULL);
    }

    while(next_html_line(html, line))  {
        if(parse_line(line, NULL))  {
            if(directive_is(line, "SHOW_DATA"))  {
	      fprintf(stdout, getResourceString(DBT_MAIN_TABLE_HEADER));
	      output_data(server, sie);
	      fprintf(stdout, getResourceString(DBT_MAIN_TABLE_FOOTER));
            } 
            else if(directive_is(line, "ID_TITLE"))  {
	      char **vals;
	      if((ldapError = ldap_search_s(server, sie, LDAP_SCOPE_BASE,
					    "(objectclass=*)", NULL, 0, &entry)) != LDAP_SUCCESS)
		continue;
	      
	      if(vals = ldap_get_values(server, entry, "nsserverid")) {
		fprintf(stdout, (const char*)getResourceString(DBT_MAIN_SERVER_ID), vals[0]);
		ldap_value_free(vals);
	      }
	    }
            else if(directive_is(line, "CHECK_UPGRADE"))  {
	      char *tmp = strdup(sie);
	      char *isie;
	      char **vals;
	      char *sie_product_version;
	      char *html_line;
#if 0

	      /* First, get the current SIE's product version. */
	      isie = strtok(tmp, ",");
	      isie = strtok(NULL, "\0");
	      isie++;
	      while(*isie == ' ')
		isie++;

	      if((ldapError = ldap_search_s(server, isie, LDAP_SCOPE_BASE,
					    "(objectclass=*)", NULL, 0, &entry)) != LDAP_SUCCESS)
		return 1;
	      if(vals = ldap_get_values(server, entry, "nsproductversion")) {
		sie_product_version = strdup(vals[0]);
		ldap_value_free(vals);
	      }

	      /* 
	       * Next, get the currently-shipping product version of the product we're interested in.
	       * Check the secret Netcenter URL. No need to init SSL in this case since the site is always insecure.
	       */
		 
	      if(!get_bindinfo(&binddn, &bindpw))
		exit(0);

	      nbuf = (bufstruct *) new_buffer(NBUF_SIZE);

	      request = (char *)malloc(strlen(getResourceString(DBT_NETCENTER_GET_REQ)) +
				       strlen((char *)binddn) +
				       3); /* space + 2 newline characters */

	      sprintf(request, getResourceString(DBT_NETCENTER_GET_REQ), binddn);
	      sockd = make_http_request("http", 
					getResourceString(DBT_NETCENTER_HOST), 
					atoi(getResourceString(DBT_NETCENTER_PORT)), 
					request, 60, &errorcode);

	      if (sockd == NULL) {
		continue;
	      }
	      
	      /* Why is the third parameter not used in parse_http_header ?? */
	      if (parse_http_header(sockd, nbuf, "Enterprise") < 0) {
		end_http_request(sockd);
		continue;
	      }

	      while( (html_line = get_line_from_fd(sockd, nbuf)) != (char *) NULL)  {
		if((strstr(sie, "Messaging") && strstr(html_line, "Messaging")) ||
		   (strstr(sie, "Administration") && strstr(html_line, "Administration")) ||
		   (strstr(sie, "Directory") && strstr(html_line, "Directory")) ||
		   (strstr(sie, "Enterprise") && strstr(html_line, "Enterprise")) ||
		   (strstr(sie, "Certificate") && strstr(html_line, "Certificate")) ||
		   (strstr(sie, "Application") && strstr(html_line, "Application")) ||
		   (strstr(sie, "Biller") && strstr(html_line, "Biller"))) {
		  /* 
		   * Get the product version. If the version is higher than the server we're looking at,
		   * display an upgrade link.
		   */
		  if(tmp) free(tmp);
		  tmp = strtok(html_line, "=");
		  tmp = strtok(NULL, "\n");
		  if(setupVersionCompare(tmp, sie_product_version) > 0)
		    fprintf(stdout, (const char*)getResourceString(DBT_MAIN_PRODUCT_VERSION), getResourceString(DBT_NETCENTER_SERVER_PRODUCT_URL));
		  else
		    fprintf(stdout, getResourceString(DBT_MAIN_UPGRADE_AVAILABLE));

		  break;
		}
	      }
#endif
	    }
            else if(directive_is(line, "SHOW_URL"))  {
	      char *text;
	      char *url;

	      if(get_product_url(server, sie, &text, &url) == 0) {
		/* success - output link */
		fprintf(stdout, (const char*)getResourceString(DBT_MAIN_PRODUCT_URL), url, text);
	      }
	    }
	    else
	      fputs(line, stdout);
	}
    }

    return 0;	
}
