/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*********************************************************************
**
** NAME:
**   ux-config.cc
**
** DESCRIPTION:
**   Administration Server Pre-installation Program
**
** NOTES:
**   This program is intended for UNIX only and is NOT thread-safe.
**   Based on the original ux-config.c.
**
*********************************************************************/

extern "C" {
#include <stdio.h>
#include <string.h>
#include <pwd.h>
#include <signal.h>
}

#include "global.h"
#include "dialog.h"
#include "ux-config.h"
#include "ux-dialog.h"

extern const char *DEFAULT_SYSUSER = "root";
extern const char *DEFAULT_OLDROOT = "/usr/ns-home";


/*********************************************************************
**
** METHOD:
**   main
** DESCRIPTION:
**   This is the ns-config program. This program functions as
**     - The Pre-installation program used during the Installation
**       of the Administration Server. In this case, the program
**       is supposed to be executed by the common installer (setup)
**       and can be executed from anywhere.
**
**     - The stand-alone configuration program used to re-configure
**       the administration server. In this case, the program has
**       to be executed from the serverroot.
**
** SIDE EFFECTS:
**   None
** RESTRICTIONS:
**
** ALGORITHM:
**
**********************************************************************/

main(int argc, char **argv)
{
   int err = 0;

   AdminPreInstall program(argc, argv);

   err = program.init();
   if (!err)
   {
      err = program.start();
   }

   return err;
} 

AdminPreInstall::AdminPreInstall(int argc, char **argv)
   :_infoFile(0), _installInfo(0), _adminInfo(0), 
    _logFile(0), _installLog(0), _configured(False), _reconfig(False),
    _runFromShell(True)
{
   setInstallMode(Interactive);
   setInstallType(Typical);

   getOptions(argc, argv);

   _installLog->logMessage(START, "Admin", "Administration Server pre-configuration.");

}

AdminPreInstall::~AdminPreInstall()
{
   if (installMode() == Interactive)
   {
      disableWinMode();
   }
}

void
AdminPreInstall::getOptions(int argc, char **argv)
{

   int opt;

   while ((opt = getopt(argc,argv, "sl:m:f:r")) != -1)
   {
      switch (opt)
      {
         case 's':
           setInstallMode(Silent);
           break;
         case 'l':
           _logFile = strdup(optarg);
           break;
         case 'f':
           _infoFile = strdup(optarg);
           break;
         case 'm':
           setInstallType((InstallType) atoi(optarg));
           break;
         case 'r':
           _reconfig = True;
           break;
         default:
           break;
       }
   }
}

int
AdminPreInstall::init()
{
   char errMsg[BIG_BUF];
   int err = 0;

   umask(022);
   _serverRoot = InstUtil::getCurrentDir();

   if (installMode() == Interactive)
   {
      enableWinMode();
      Dialog::initDisplay("Administration");

      if (_infoFile == (char *) NULL)
      {
         // Not executing from the shell, check if this is the
         // server root
         if (InstUtil::isServerRoot(_serverRoot) == False)
         {
            err = -1;
            sprintf(errMsg, "ERROR: %s is not a server root\n",_serverRoot.data());
         }
         else
         {
            _runFromShell = False;
         }
      }
      else if (InstUtil::fileExists(_infoFile) == False)
      {
         sprintf(errMsg, "ERROR: answer cache not found\n");
         err = -1;
      }
   }
   else // Silent installation
   {
      if (_infoFile == (char *) NULL ||
          InstUtil::fileExists(_infoFile) == False)
      {
         sprintf(errMsg, "ERROR: installation script not found\n");
	 _installLog->logMessage(FATAL, "Admin", "Installation script not found");
         err = -1;
      }
      else 
      {
         if (_logFile == (char *) NULL)
         {
            _logFile = _serverRoot + "/" + SETUP_DIR + "/" + INSTALL_LOG;
         }
         _installLog = new InstallLog (_logFile);
      }
   }

   if (!err)
   {
      if (_runFromShell)
      {
         _installInfo = new InstallInfo(_infoFile);

         _installInfo->toUTF8(CONFIG_LDAP_URL);
         _installInfo->toUTF8(MC_ADMIN_ID);

         _adminInfo = _installInfo->getSection("Admin");

         if (installMode() == Silent)
         {
            if (_adminInfo == NULL)
            {
               sprintf(errMsg, "ERROR: Invalid installation script. No Admin section.\n");
	       _installLog->logMessage(FATAL, "Admin", "Invalid installation script. no Admin section");
               err = -1;
            }
         }
         else
         {
            _adminInfo = _installInfo->createSection("admin");
         }
         _serverRoot = _installInfo->get(SERVER_ROOT);
      }
      else
      {
         // Not run from the shell, this is a reconfiguration
         // attempt. Load configuration information from
         // installation database.
         // PVO
         char buf[SML_BUF];
//         _infoFile = _serverRoot + "/" + SETUP_DIR + "/" + INSTALL_INF; -- PVO
	 _infoFile = tmpnam(buf);
         _installInfo = new InstallInfo();
         _adminInfo = _installInfo->createSection("admin");
      }

      initDefaultConfig();

      setDefaultScript(_adminInfo);

      stopAdmin();

   }
   else
   {
      if (installMode() == Silent)
      {
         printf(errMsg);
      }
      else
      {
         DialogAlert alert(errMsg);
         alert.execute();
      }
   }

   return err;

}

void
AdminPreInstall::initDefaultConfig()
{
   char tmp[BIG_BUF];
   NVPair admConf;
   LdapError ldapError;
   char *ldapURL = NULL;
   char *ldapUser = NULL;
   char *installDN = NULL;
   NSString siePwd = NULL;
   NSString siePort = NULL;
   NSString hostName = InstUtil::guessHostname();

   /* First, determine whether Admin is already configured */
   sprintf(tmp, "%s/admin-serv/config/adm.conf", _serverRoot.data());

   admConf.setFormat(2);
   admConf.read(tmp);

   getDefaultLdapInfo(_serverRoot, &ldapURL, &ldapUser, &installDN );

   if (admConf.isEmpty() == False)
   {
      NSString sieDN;

      sieDN = admConf.get("sie");
      siePwd = admConf.get("siepid");
      siePort = admConf.get("port");

      if (sieDN != (char *) NULL && siePwd != (char *) NULL && ldapURL != NULL)
      {
         _configured = True;

         Ldap ldap (ldapError, ldapURL, sieDN, siePwd, sieDN, siePwd);

         if (ldapError == OKAY)
         {
            NSString configDN = NSString("cn=configuration") + LDAP_PATHSEP + sieDN;
            LdapEntry *configEntry = new LdapEntry(&ldap);

            if (configEntry->retrieve(configDN) == OKAY)
            {
               _adminInfo->set("Port", configEntry->getAttribute("nsServerPort"));
               _adminInfo->set("SysUser", configEntry->getAttribute("nsSuiteSpotUser"));
	       _adminInfo->set("ServerIpAddress", configEntry->getAttribute("nsServerAddress"));
            }
         }
      }
   }

   sprintf(tmp, "%s/admin-serv/config/admpw", _serverRoot.data());

   admConf.read(tmp);

   if (admConf.isEmpty())
   {
      _configured = False;
   }
   else
   {
      _adminInfo->set("ServerAdminID", admConf.name(0));
   }

   _adminInfo->set("ServerAdminPwd", siePwd);

   if (_adminInfo->get("SysUser") == NULL)
   {
      _adminInfo->set("SysUser", guessAdminSysUser());
   }

   if (_adminInfo->get("Port") == NULL)
   {
      if (siePort != (char*)NULL) {
           _adminInfo->set("Port", siePort);
      }
      else {
          _adminInfo->set("Port", (unsigned long)InstUtil::guessPort());
      }
   }

   if (_adminInfo->get("ServerIpAddress") == NULL)
   {
      _adminInfo->set("ServerIpAddress", " ");
   }

   if (_adminInfo->get("ServerAdminID") == NULL)
   {
      _adminInfo->set("ServerAdminID", DEFAULT_ADMINUSER);
   }

   if (_adminInfo->get("ServerAdminPwd") == NULL)
   {
      _adminInfo->set("ServerAdminPwd", DEFAULT_ADMINPWD);
   }

   if (_runFromShell == False)
   {
      _installInfo->set(SERVER_ROOT, _serverRoot.data());
      _installInfo->set(MACHINE_NAME, hostName);
      _installInfo->set(INSTALL_TYPE, DEFAULT_INSTALL_TYPE);
      if (ldapURL)
      {
         _installInfo->set(CONFIG_LDAP_URL, ldapURL);
      }
      else
      {
         _installInfo->set(CONFIG_LDAP_URL, NSString("ldap://") +
            _installInfo->get(MACHINE_NAME) +
            ":389/" +
            "o=" + InstUtil::guessDomain());

         ldapURL = (char *) _installInfo->get(CONFIG_LDAP_URL);
      }

      _installInfo->set(SS_USER_ID, InstUtil::getDefaultUser(_serverRoot));
      _installInfo->set(SS_GROUP, InstUtil::getDefaultGroup(_serverRoot));
   }
}

int
AdminPreInstall::start()
{
   DialogAction action;
   int err = -1;

   if (installMode() == Silent)
   {
      err = doSilentInstall();
      goto Finito;
   }

SS0:
   if (_configured == False)
   {
      goto SS1;
   }

SS0a:
#if 0
   askReconfig.registerDialogNext(this);
   action = askReconfig.execute();
   if (action == DIALOG_NEXT)
     goto SS1;
   else if (action == DIALOG_PREV)
     goto SS0a;
   else 
   {
     err = 7;
     goto Finito;
   }
#else
   // always update server information
   action = DIALOG_NEXT;
#endif
SS1:
SS2:
   if (installType() == Express)
   {
      goto SS3;
   }

   if (_configured != False)
   {
	   goto SS7;
   }

   askAdminPort.registerDialogNext(this);
   action = askAdminPort.execute();

   if (action == DIALOG_PREV)
     goto SS0;
   else if (action == DIALOG_SAME)
     goto SS1;
   else  if (action != DIALOG_NEXT)
     goto Finito;

SS2a:
   if (_runFromShell == False || installType() == Custom)
   {
      askHostIpAddress.registerDialogNext(this);
      action = askHostIpAddress.execute();
      if (action == DIALOG_PREV)
      {
         if (installType() == Express)
         {
            goto SS0;
         }
         else
         {
            goto SS2;
         }
      }
      else if (action == DIALOG_SAME)
      {
         goto SS2a;
      }
      else if (action != DIALOG_NEXT)
      {
         goto Finito;
      }
   }

SS3:
#if 1
   setupServerAdmin();
#else
   if (_runFromShell == False || installType() == Custom)
   {
      askAdminUser.registerDialogNext(this);
      action = askAdminUser.execute();
  
      if (action == DIALOG_PREV)
      {
         if (installType() == Express)
         {
            goto SS1;
         }
         else
         {
            goto SS2;
         }
      } 
      else if (action == DIALOG_SAME)
         goto SS3;
      else if (action != DIALOG_NEXT)
         goto Finito;
   }
   else
   {
      setupServerAdmin();
   }
#endif

SS4:
  if (installType() == Express)
  {
      NSString sysUser = guessAdminSysUser();
      if (sysUser != (char *) NULL)
      {
         _adminInfo->set("SysUser", sysUser);
         goto SS5;
      }
  }
  askSysUser.registerDialogNext(this);
  action = askSysUser.execute();
  if (action == DIALOG_PREV)
  {
      if (installType() == Custom || _runFromShell == False)
      {
         goto SS3;
      }
      else if (installType() == Express)
      {
         goto SS4;
      }
      else
      {
         goto SS2;
      }
  }
  else if (action == DIALOG_SAME)
      goto SS4;

SS5:

  if (_runFromShell == False)
  {
     askLdapInfo.registerDialogNext(this);
     action = askLdapInfo.execute();
     if (action == DIALOG_PREV)
     {
        goto SS4;
     }
     else if (action == DIALOG_SAME)
         goto SS5;
  }

SS6:

  if (_runFromShell == False)
  {
     askLdapUser.registerDialogNext(this);
     action = askLdapUser.execute();
     if (action == DIALOG_PREV)
     {
        goto SS5;
     }
     else if (action == DIALOG_SAME)
     {
        goto SS6;
     }
  }

SS7:

  err = 0;

  _installInfo->setFormat(1);

   _installInfo->toLocal(CONFIG_LDAP_URL);
   _installInfo->toLocal(MC_ADMIN_ID);

  _installInfo->write(_infoFile);

  if (_runFromShell == False)
  {
      err = updateAdmin();
  }

Finito:
  if (!err)
  {
      _installLog->logMessage(SUCCESS, "Admin", "Administration Server pre-configuration.");
  }

  return err;
}
   
int
AdminPreInstall::cont()
{
   return 0;
}

void
AdminPreInstall::clear()
{
}

void
AdminPreInstall::add(Dialog *p)
{
}
void
AdminPreInstall::addLast(Dialog *p)
{
}

void
AdminPreInstall::resetLast()
{
}

void
AdminPreInstall::setParent(void *parent)
{
   return;
}
void *
AdminPreInstall::parent() const
{
   return (void *) this;
}

int
AdminPreInstall::doSilentInstall()
{
   const char *s;
   NSString errMsg;
   int err = 0;

   s = _adminInfo->get("Port");

   errMsg = verifyAdminPort(s);

   if (errMsg != (char *) NULL)
   {
      err = -1;
      _installLog->logMessage(FATAL, "Admin", errMsg);
   }

   s = _adminInfo->get("SysUser");

   errMsg = verifyAdminSysUser(s);

   if (errMsg != (char *) NULL)
   {
      err = -1;
      _installLog->logMessage(FATAL, "Admin", errMsg);
   }

   s = _adminInfo->get("ServerAdminID");
   if (s == NULL)
   {
      err = -1;
      _installLog->logMessage(FATAL, "Admin", "ServerAdminID is not specified.");
   }
   
   s = _adminInfo->get("ServerAdminPwd");
   if (s == NULL)
   {
      err = -1;
      _installLog->logMessage(FATAL, "Admin", "ServerAdminPwd is not specified.");
   }
   return err;
}

NSString
AdminPreInstall::verifyAdminPort(const char *adminPort) const
{
   char errMsg[SML_BUF];
   int port;

   port = atoi(adminPort);
   sprintf(errMsg, "%d", port);

   if (strncmp(errMsg, adminPort, 6) || port > MAXPORT)
   {
      sprintf(errMsg, "OVERFLOW ERROR: Unable to bind to port %d\n"
                 "Please choose another port less than %d.\n",
                   port, MAXPORT);
   }
   else if (InstUtil::portAvailable(port) == False)
   {
      sprintf(errMsg,
"ERROR: Unable to bind to port %d\n"
"       Please choose another port.\n",
        port);
   }
   else
   {
      errMsg[0] = 0;
   }
   return errMsg;
}

NSString
AdminPreInstall::verifyAdminSysUser(const char *sysUser) const
{
   char errMsg[SML_BUF];
   int err = 0;

   err = InstUtil::availUser(sysUser);

   switch (err)
   {
      case -1:

         sprintf(errMsg,
"The user %s does not currently exist.\n"
"       Choose another, or create it an choose it again.\n",
        sysUser);
         break;
      case -2:
         sprintf(errMsg,
"The system will not allow you to run the Administration\n"
"       server as that user. Please choose another one.\n");
         break;
      default:
         errMsg[0] = 0;
         break;
   }

   return errMsg;
}

NSString
AdminPreInstall::guessAdminSysUser() const
{
   /*
    * Try to determine the default sys user to run admin as.
    * First, determine the sysuser from bin/admin/ns-admin
    * If ns-admin exists, use the owner of that file
    * otherwise default to the uid of the running process
    */
   struct stat fi;
   char tmp[BIG_BUF];
   NSString errMsg;
   NSString sysUser;

   sprintf(tmp, "%s/bin/admin/ns-admin", _serverRoot.data());

   if (stat(tmp, &fi) == 0)
   {
      sysUser = getpwuid(fi.st_uid)->pw_name;
   }
   else
   {
      sysUser = InstUtil::getSelfUserID();
   }

   errMsg = verifyAdminSysUser(sysUser);
   if (errMsg != (char *) NULL)
   {
      sysUser = (char *) NULL;
   }

   return sysUser;

}

void
AdminPreInstall::stopAdmin() const
{
   char pid[BIG_BUF];
   char stopProgram[BIG_BUF];
   struct stat fi;

   sprintf(pid, "%s/admin-serv/logs/pid", _serverRoot.data());

   if (stat(pid, &fi) == 0)
   {
      sprintf(stopProgram, "%s/stop-admin", _serverRoot.data());
      if (stat (stopProgram, &fi) != 0)
      {
         sprintf(stopProgram, "kill `cat %s`", pid);
      }
      system(stopProgram);
   }

   /* Stop SNMP Master Agent if running */
   int magpid;
   FILE* fhdl;
   sprintf(pid, "%s/admin-serv/logs/pid_masteragt", _serverRoot.data());
   if ((fhdl = fopen(pid, "r")) != NULL)
   {
      fscanf(fhdl, "%d\n", &magpid);
      fclose(fhdl);
      kill(magpid, SIGKILL);
      unlink(pid);
   }
}

int
AdminPreInstall::updateAdmin() const
{
   int err;
   char tmp[BIG_BUF];

   disableWinMode();

   sprintf(tmp, "%s/bin/admin/ns-update -f %s", _serverRoot.data(), _infoFile.data());

   err = system(tmp);

   enableWinMode();

   unlink(_infoFile.data());

   return err;
} 

void
AdminPreInstall::setupServerAdmin()
{
   const char *ldapUser;
   const char *ldapPwd;

   if (_configured  == False)
   {
      char **dn;
      ldapUser = _installInfo->get(MC_ADMIN_ID);
      ldapPwd  = _installInfo->get(MC_ADMIN_PWD);

      dn = ldap_explode_dn(ldapUser, 0);

      _adminInfo->set("ServerAdminID", dn[0]);
      _adminInfo->set("ServerAdminPwd", ldapPwd);

      ldap_value_free(dn);
   }
}
