/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

extern "C" {
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
}

#include "utf8.h"
#include "ux-util.h"
#include "dialog.h"
#include "ux-dialog.h"
#include "ux-config.h"

/*
** Forward References
*/

DialogAction yesNoDefaultNo (const char *answer);
DialogAction askReconfigNext (Dialog *me);
DialogAction askAdminPortSetup (Dialog *me);
DialogAction askAdminPortNext(Dialog *me);
DialogAction askSr2xInfoSetup(Dialog *me);
DialogAction askSr2xInfoNext(Dialog *me);
DialogAction askAdminUserSetup(Dialog *me);
DialogAction askAdminUserNext (Dialog *me);
DialogAction askSysUserSetup (Dialog *me);
DialogAction askSysUserNext (Dialog *me);
DialogAction askLdapInfoNext (Dialog *me);
DialogAction askLdapInfoSetup(Dialog *me);
DialogAction askLdapUserNext(Dialog *me);
DialogAction askLdapUserSetup(Dialog *me);
DialogAction askHostIpAddressSetup (Dialog *me);
DialogAction askHostIpAddressNext (Dialog *me);

DialogAction
yesNoDefaultNo(const char *answer)
{
   if (answer[0] == '\0' || answer[0] == '\n')
      return DIALOG_EXIT;
   else if (answer[0] != 'y' && answer[0] != 'Y')
      return DIALOG_EXIT;
   else
      return DIALOG_NEXT;
}

void
dialogSetup (Dialog *me, const char *which, InstallInfo *defaultScript)
{
   const char *ans = defaultScript->get(which);

   if (ans)
      me->setDefaultAns(ans);
}

//////////////////////////////////////////////////////////////////////
// AskAdminPort
//////////////////////////////////////////////////////////////////////
DialogYesNo askReconfig(
"WARNING: The Administration Server has already been configured!\n",
"Do you want to reconfigure this server?",
"No",
NULL,
askReconfigNext,
NULL
);

DialogAction
askReconfigNext(Dialog *me)
{
   const char *buf = me->input();
   DialogAction rc = yesNoDefaultNo(buf);

   return rc;
}

DialogInput askAdminPort(
"The Administration Server is separate from any of your application\n"
"servers since it listens to a different port and access to it is\n"
"restricted.\n\n"
"Pick a port number between 1024 and 65535 to run your Administration\n"
"Server on. You should NOT use a port number which you plan to\n"
"run an application server on, rather, select a number which you\n"
"will remember and which will not be used for anything else.\n\n"
"The default in brackets was randomly selected from the available\n"
"ports on your system. To accept the default, press return.",

"Administration port",

NULL,

askAdminPortSetup,
askAdminPortNext
);

DialogAction
askAdminPortSetup(Dialog *me)
{
   dialogSetup(me, "Port", me->manager()->getDefaultScript());

   return DIALOG_SAME;
}

DialogAction
askAdminPortNext(Dialog *me)
{
   const char *buf = me->input();
   const char *tmp;
   NSString errMsg;
   AdminPreInstall *installer = (AdminPreInstall *) me->manager()->parent();

   if (buf[0] == 0)
   {
      tmp = me->defaultAns();
   }
   else
   {
      tmp = buf;
   }

   errMsg = installer->verifyAdminPort(tmp);

   if (errMsg != (char *) NULL)
   {
      DialogAlert alert(errMsg);
      alert.execute();
      return DIALOG_SAME;
   }

   me->manager()->getDefaultScript()->set("Port", tmp);

   return DIALOG_NEXT;
}

DialogInput askSysUser(
"The Administration Server program runs as a certain user on your\n"
"system. This user should be different than the one which your\n"
"application servers run as. Only the user you select will be\n"
"able to write to your configuration files. If you run the\n"
"Administration Server as \"root\", you will be able to use the Server\n"
"Administration screen to start and stop your application servers.",

"Run Administration Server as",

DEFAULT_SYSUSER,

askSysUserSetup,
askSysUserNext
);

DialogAction
askSysUserSetup(Dialog *me)
{
   dialogSetup(me, "SysUser", me->manager()->getDefaultScript());

   return DIALOG_SAME;
}

DialogAction
askSysUserNext(Dialog *me)
{
   const char *buf = me->input();
   const char *sysUser;
   NSString errMsg;
   AdminPreInstall *installer = (AdminPreInstall *) me->manager()->parent();

   if (buf[0] == 0)
   {
      sysUser = me->defaultAns();
   }
   else
   {
      sysUser = buf;
   }
   
   errMsg = installer->verifyAdminSysUser(sysUser);

   if (errMsg != (char *) NULL)
   {
      errMsg = NSString("ERROR: ") + errMsg;
      DialogAlert alert(errMsg);
      alert.execute();
      return DIALOG_SAME;
   }

   me->manager()->getDefaultScript()->set("SysUser", sysUser);

   return DIALOG_NEXT;
}

DialogInput askAdminUser(
"Server administration is restricted to people who know the Server\n"
"Administrator user ID and password, and to people who access the\n"
"server from a host you specify.\n\n"
"When you access the Administration Server, it will prompt you\n"
"for the Server Administrator user ID and password. Please select a\n"
"user ID and password.",

"Server Administrator ID",

DEFAULT_ADMINUSER,

askAdminUserSetup,
askAdminUserNext
);

DialogAction
askAdminUserSetup(Dialog *me)
{
   dialogSetup(me, "ServerAdminID", me->manager()->getDefaultScript());
   return DIALOG_SAME;
}

DialogAction
askAdminUserNext(Dialog *me)
{
   const char *adminUser;
   const char *adminPwd;
   const char *buf;
   char *inp = NULL;

   buf = me->input();
   if (buf[0] == 0)
   {
      adminUser = me->defaultAns();
   }
   else
   {
      adminUser = buf;
   }

   me->manager()->getDefaultScript()->set("ServerAdminID", adminUser);

   while (1)
   {
      me->showString("Password: ");
      if (me->getPassword () == 0)
      {
         return DIALOG_PREV;
      }
      else
      {
         inp = strdup(me->input());

         if (inp[0] == 0)
         {
	    free(inp);
            continue;
         }
         else
         {
            me->showString("Password (again): ");
            if (me->getPassword() == 0)
            {
	       free(inp);
               return DIALOG_PREV;
            }
            else
            {
               adminPwd = me->input();
               if (strcmp(inp,adminPwd))
               {
                  DialogAlert alert("ERROR: Passwords don't match.");
                  alert.execute();
		  free(inp);
                  return DIALOG_SAME;
               }
               break;
            }
         }
      }
   }
   free(inp);
   me->manager()->getDefaultScript()->set("ServerAdminPwd", adminPwd);
   return DIALOG_NEXT;
}


DialogInput askLdapInfo(
"The Administration Server requires the use of an LDAP-based\n"
"directory server such as the Fedora Directory Server for \n"
"resource administration.\n"
"\n"
"Enter the URL of an existing LDAP-based directory server\n"
"in the form similar to the below:\n"
"\n"
"     ldap://<hostname>:<port>/<base dinstinguished name>"
"\n",

"URL of directory server",

NULL,

askLdapInfoSetup,
askLdapInfoNext
);

DialogAction
askLdapInfoSetup(Dialog *me)
{
   dialogSetup (me, CONFIG_LDAP_URL, ((AdminPreInstall *) me->manager()->parent())->getDefaultScript());
   return DIALOG_SAME;
}

DialogAction
askLdapInfoNext(Dialog *me)
{
   const char *ldapURL;
   const char *buf = me->input();
   int err = 0;
   char errMsg[BIG_BUF];
   char tmp[XSM_BUF];
   LDAPURLDesc *ludpp;
   DialogAction rc = DIALOG_NEXT;
   AdminPreInstall *setup = ((AdminPreInstall *)me->manager()->parent());
   char *localLdapURL;

   ldapURL = me->defaultAns();

   if (buf[0] != 0)
     ldapURL = buf;

   setup->getInstallInfo()->set(CONFIG_LDAP_URL,ldapURL);

   err = verifyLdapUrl(ldapURL, NULL, NULL);

   localLdapURL = UTF8ToLocal(ldapURL);

   switch (err)
   {
      case OKAY:
        errMsg[0] = 0;
        break;
      case INVALID_URL:
        sprintf(errMsg, "ERROR: The URL \"%s\" is not of valid format.\n", localLdapURL);
        break;
      case CONN_FAILED:
        sprintf(errMsg,
"ERROR: Cannot connect to URL \"%s\".\n"
"       The server may have been down. Please fix the problem\n"
"       before proceeding with installation.\n",
                localLdapURL);
        break;
      case INVALID_DN:
        sprintf(errMsg,
"ERROR: setup cannot verify the base suffix as specified in\n"
"            \"%s\".\n"
"       Please check the base suffix and re-enter the URL.\n",
                        localLdapURL);
        break;
      default:
        // Assume it's up and running
        break;
   }

   free(localLdapURL);

   if (errMsg[0] != 0)
   {
      DialogAlert alert(errMsg);
      alert.execute();
      rc =  DIALOG_SAME;
   }
   return rc;
}

DialogInput askLdapUser(
"Enter the user ID or Distinguished Name of the Configuration\n"
"Administrator who can access the Configuration Directory server\n"
"at",

"Configuration Admin ID or DN:",

DEFAULT_ADMINUSER,

askLdapUserSetup,
askLdapUserNext
);

DialogAction
askLdapUserSetup(Dialog *me)
{
   char text2[SML_BUF];
   InstallInfo *installInfo = ((AdminPreInstall *) me->manager()->parent())->getInstallInfo();
   char *localLdapURL;

   dialogSetup(me, MC_ADMIN_ID, installInfo);

   localLdapURL = UTF8ToLocal(installInfo->get(CONFIG_LDAP_URL));

   sprintf(text2, "    %s", localLdapURL);

   me->setText2(text2);

   free (localLdapURL);

   return DIALOG_SAME;
}

DialogAction
askLdapUserNext(Dialog *me)
{
   NSString    ldapUser;
   const char *ldapPwd;
   const char *ldapURL;
   const char *buf = me->input();
   char errMsg[BIG_BUF];
   DialogAction rc = DIALOG_NEXT;
   AdminPreInstall *setup = ((AdminPreInstall *)me->manager()->parent());
   LdapError err;

   if (buf[0] != 0)
   {
     ldapUser = buf;
   }
   else
   {
      ldapUser = me->defaultAns();
   }

   setup->getInstallInfo()->set(MC_ADMIN_ID,ldapUser);

   while (1)
   {
      me->showString("Password: ");
      if (me->getPassword () == 0)
      {
         return DIALOG_PREV;
      }
      else if (me->input()[0] == 0)
      {
         continue;
      }
      else
      {
         ldapPwd = me->input();
         break;
      }
   }

   setup->getInstallInfo()->set(MC_ADMIN_PWD, ldapPwd);

   ldapURL = setup->getInstallInfo()->get(CONFIG_LDAP_URL);

   // Got ldap User and Password, now authenticate
  
   errMsg[0] = 0;

   Ldap ldap(err, ldapURL, ldapUser, ldapPwd, NULL, NULL);

   char *localLdapURL = UTF8ToLocal(ldapURL);

   if (err)
   {
       switch(err)
       {
          case INVALID_INPUT:
             sprintf(errMsg, "ERROR: Invalid input.\n");
             break;
          case INVALID_URL:
             sprintf(errMsg, "ERROR: Invalid URL \"%s\".\n", localLdapURL);
             break;
          case INVALID_AUTH:
             sprintf(errMsg, "ERROR: Insufficient authorization.\n");
             break;
          case CONN_FAILED:
             sprintf(errMsg,
"ERROR: Cannot connect to LDAP server.\n"
"       The server may have been down. Please fix the problem\n"
"       before proceeding with installation.\n");
             break;
          case INVALID_USER:
             sprintf(errMsg, "ERROR: Invalid user and/or password.\n");
             break;
          default:
             sprintf(errMsg,
"ERROR: Authentication failed. Either you have entered\n"
"       an invalid user ID or password, or the directory server\n"
"       is having some problem. Please check and re-enter.\n");
             break;
       }
   }

   free(localLdapURL);

   if (errMsg [0] != 0)
   {
       DialogAlert alert(errMsg);
       alert.execute();
       rc =  DIALOG_SAME;
   }
   me->visited();

   return rc;
}

DialogInput askHostIpAddress(
"If you want to configure the Administration Server to bind\n"
"to a specific IP address, enter the address below.\n\n"
"To accept the default shown in brackets, press the Return key.\n",

"IP address",

NULL,
askHostIpAddressSetup,
askHostIpAddressNext
);

DialogAction
askHostIpAddressSetup(Dialog *me)
{
   dialogSetup(me, "ServerIpAddress", ((AdminPreInstall *) me->manager()->parent())->getDefaultScript());
   return DIALOG_SAME;
}

DialogAction
askHostIpAddressNext(Dialog *me)
{
   const char *buf = me->input();
   const char *hostIpAddress;
   NSString errMsg;
   AdminPreInstall *installer = (AdminPreInstall *) me->manager()->parent();

   if (buf[0] == 0)
   {
      hostIpAddress = me->defaultAns();
   }
   else
   {
      hostIpAddress = buf;
   }

   me->manager()->getDefaultScript()->set("ServerIpAddress", hostIpAddress);

   return DIALOG_NEXT;
}

