/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

extern "C" {
#include <stdio.h>		/* printf, file I/O */
#include <string.h>		/* strlen */
#include <strings.h>
#include <stdlib.h>		/* memset, rand stuff */
#include <time.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>

#include "ldap.h"

#if defined(AIX)
extern int seteuid(uid_t);
extern int setegid(gid_t);
#endif

}

#include "global.h"
#include "code.h"
#include "ux-util.h"
#include "ldapu.h"
#include "setupldap.h"
#include "uninstall.h"
#include "../common/include/servlet-conf.h"
#include "../common/include/acidef.h"

static int set_locate_admserv_aci(Ldap *ldap);
static int repair_group_expansion_aci(Ldap *ldap);

typedef struct {
char *section;
char *attr;
char *val;
} xml_file_mod_t;

static int set_xml_attr(char *line, char *section,  char *attr, char *val);
static int modify_xml_file(char *filename, xml_file_mod_t **mods);

#ifdef HPUX
  #define SHLIB_EXT "sl"
#else
  #define SHLIB_EXT "so"
#endif

#define EPOCH_32BIT_END_TIME  "20380118031417" /* <correct end time> minus 1 day */

const char *DEFAULT_SYSUSER = "root";
const char *DEFAULT_LCACHECONF = "userdb/ldap/config/lcache.conf";
const char *DEFAULT_LCACHEDIR  = "userdb/ldap/db";
const char *DEFAULT_OLDROOT	= "/opt/fedora";
const char *DEFAULT_TASKCONF = "admin-serv/ldif/tasks.ldif";
const char *DEFAULT_GLOBALCONF = "admin-serv/ldif/global.ldif";

const char *CHOWN_CMD = "chown -R %s %s >/dev/null 2>&1";
const char *CHGRP_CMD = "chgrp -R %s %s >/dev/null 2>&1";
/* Obsolete
const char *USERDB_CHGRP_CMD = "chgrp -R %s %s/userdb >/dev/null 2>&1";
const char *SSCONF_CHGRP_CMD = "chgrp -R %s %s/install/ssusers.conf >/dev/null 2>&1";
*/
const char *SHARED_CHGRP_CMD = "chgrp -R %s %s/shared >/dev/null 2>&1";
const char *SSCONF_CHGRP_CMD = "chgrp -R %s %s/shared/config/ssusers.conf >/dev/null 2>&1";
const char *SECALIAS_CHGRP_CMD = "chgrp -R %s %s/alias >/dev/null 2>&1";

const char *DEFAULT_ADMIN_CONFIG_ACI = "(targetattr=*)(version 3.0; acl \"Enable delegated admin to access configuration\"; allow (read,search) groupdn=\"ldap:///%s\";)";

InstallMode installMode = Interactive;
InstallInfo *installInfo = NULL;
InstallInfo *adminInfo = NULL;
InstallInfo *dsInfo = NULL;
char *infoFile = NULL;
char *logFile = NULL;
InstallLog    *installLog = NULL;
Bool reConfig = False;

char *
system_errmsg(void)
{
    return strerror(errno);
}

void
fatal_error(const char *format, const char *arg1, const char *arg2)
{
   if (installMode == Interactive)
   {
      printf(format,arg1, arg2);
      printf("\n");
   }
   else
   {
      printf("ERROR: Administration Server configuration failed. See install.log.\n");
      installLog->logMessage(FATAL, "Admin", format, arg1, arg2);
   }
   exit(1);
}

/*
 * The following escape LDAP URL is from ns/netsite/ldap/clients/dsgw/htmlout.c
 */

#define HREF_CHAR_ACCEPTABLE( c )       (( c >= '-' && c <= '9' ) ||    \
                                         ( c >= '@' && c <= 'Z' ) ||    \
                                         ( c == '_' ) ||                \
                                         ( c >= 'a' && c <= 'z' ))
void
escape_ldap_uri(char *s1, char *s2)
{
   unsigned char *q;
   char *p, *hexdig = "0123456789ABCDEF";

   p = s1 + strlen(s1);
   for (q = (unsigned char *) s2; *q != '\0'; ++q)
   {
      if (HREF_CHAR_ACCEPTABLE(*q))
      {
	 *p++ = *q;
      }
      else
      {
	 *p++ = '%';
	 *p++ = hexdig[*q >> 4];
	 *p++ = hexdig[*q & 0x0F];
      }
   }

   *p = '\0';
}

/* ----------------------- make a new admin server ------------------------ */

char *
make_shexp(char *source)
{
   char *answer;
   int hascomma, x, y;

   if (!source)
      return source;

   hascomma = (strchr(source, ',')) ? 1 : 0;

   /* + 2 = 1 for trailing \0 and one put in by Aruna for NT */
   answer = (char *) malloc(strlen(source) + 2);

   for (x = 0, y = 0; source[x]; x++)
   {
      if (source[x] == ',')
	 answer[y++] = '|';
      else if (source[x] != ' ')
	 answer[y++] = source[x];
      answer[y] = '\0';
   }
   if (!hascomma)
      return (answer);
   else
   {
      char *newanswer = (char *) malloc(strlen(answer) + 3);
      sprintf(newanswer, "(%s)", answer);
      return (newanswer);
   }
}

/* Be tricky and check the syntax of the URL at the same time it
   escapes the base DN.  Return escaped string if everything when according to
   plan, else NULL */
static char *
escapeLDAPurl(char *url)
{
   char *result = (char *) malloc(strlen(url) * 3 + 1);	/* overkill */
   int i;

   for (i = 0; url[i] && url[i] != ':'; ++i)
   {
      result[i] = url[i];
   }
   result[i] = '\0';
   if (strcmp(result, "ldap") != 0 && strcmp(result, "ldaps") != 0)
   {
      return NULL;
   }
   result[i++] = ':';
   if ((result[i] = url[i]) != '/')
      return NULL;
   ++i;
   if ((result[i] = url[i]) != '/')
      return NULL;
   ++i;
   while ((result[i] = url[i]) && (result[i] != '/'))
   {
      ++i;
   }
   if (!result[i])
      return NULL;
   result[++i] = '\0';
   escape_ldap_uri(result + i, url + i);

   return result;
}

void
configTasks(const Ldap *ldap, const char *sroot, const char *sieDN)
{
   char tmp[BIG_BUF];
   LdapErrorCode err;

   printf("Configuring Administration Tasks in Directory Server...\n");
   installLog->logMessage(INFO, "Admin", "Configuring Administration Tasks in Directory Server");

   sprintf(tmp, "%s/%s", sroot, DEFAULT_TASKCONF);

   err = setupReplaceLdifEntries(ldap, sieDN, tmp, sieDN);

   if (err != OKAY)
   {
      printf("Cannot write task configuration. Error (%ld).\n", err);
   }
   return;
}

void
configGlobalParameters(const Ldap *ldap, const char *sroot, const char *subTreeDN)
{
   char tmp[BIG_BUF];
   LdapErrorCode err;

   printf("Configuring Global Parameters in Directory Server...\n");
   installLog->logMessage(INFO, "Admin", "Configuring global parameters...");

   sprintf(tmp, "%s/%s", sroot, DEFAULT_GLOBALCONF);

   err = setupUpdateLdifEntries(ldap, subTreeDN, tmp, NULL);

   if (err != OKAY)
   {
      printf("Cannot write Admin global parameters. Error (%ld).\n", err);
   }
   return;
}


/*
 * Cleanup prevoius installation after upgrading from 6.0x
 */
int
cleanup_after_upgrade(const char *sroot) 
{
   struct stat fi;
   char buf[BIG_BUF];

   // Move admin-serv/config/secmod.db to the alias directory;
   // Starting with as 6.1 secmod.db is in the alias directory.
   sprintf(buf, "%s/admin-serv/config/secmod.db", sroot);
   if (stat(buf, &fi) == 0) {
       char to[BIG_BUF];
       sprintf(to, "%s/alias/secmod.db", sroot);
       if (link(buf, to) == 0) { 
           unlink(buf);
       }
   }

   // Remove admin-serv/config/libnssckbi.so;
   // Starting with as 6.1 secmod.db is in the alias directory.
   sprintf(buf, "%s/admin-serv/config/libnssckbi.%s", sroot, SHLIB_EXT);
   if (stat(buf, &fi) == 0) {
       unlink(buf);
   }

   // Remove admin-serv/config/res directory. Not packaged starting with as61.
   sprintf(buf, "%s/admin-serv/config/res", sroot);
   if (stat(buf, &fi) == 0) {
       char cmd[BIG_BUF];
       sprintf(cmd, "rm -fr %s", buf);
       system(cmd);
   }

   // Remove adminsdk 60
   sprintf(buf, "%s/lib/libadminutil60.%s", sroot, SHLIB_EXT);
   if (stat(buf, &fi) == 0) {
       unlink(buf);
   }
   sprintf(buf, "%s/lib/libadmsslutil60.%s", sroot, SHLIB_EXT);
   if (stat(buf, &fi) == 0) {
       unlink(buf);
   }

   // Remove admserv jars from the previous installation
   sprintf(buf, "%s/java/jars/admserv60.jar", sroot);
   if (stat(buf, &fi) == 0) {
       char cmd[BIG_BUF];
       sprintf(cmd, "rm %s/java/jars/admserv60*", sroot);
       system(cmd);
   }

   return 0;
}


void
start_admserv(const char *sroot)
{
   char pid[BIG_BUF];
   char startProgram[BIG_BUF];
   char startLog[SML_BUF];
   struct stat fi;
   int ret;

   sprintf(pid, "%s/admin-serv/logs/pid", sroot);

   tmpnam(startLog);

   if (stat(pid, &fi) == 0)
   {
      sprintf(startProgram, "%s/restart-admin > %s 2>&1", sroot, startLog);
   }
   else
   {
      sprintf(startProgram, "%s/start-admin > %s 2>&1", sroot, startLog);
   }

   ret = system(startProgram);

   if (ret)
   {
      printf("Can't start Admin server [%s] (error: %s)", startProgram, system_errmsg());
      installLog->logMessage(WARN, "Admin", "Can't start Admin server [%s] (error: %s)", startProgram, system_errmsg());
   }
   else
   {
      FILE *fp = fopen (startLog, "r");
      char buf[BIG_BUF];

      if (fp)
      {
         while (fgets(buf, BIG_BUF, fp))
         {
            printf("%s\n",buf);
         }
         fclose(fp);
         unlink(startLog);
      }
   }
}

/*
void
walk_dir_and_chown(const char *dirName, uid_t my_uid, uid_t admin_uid, uid_t tuid, gid_t tgid)
{
   if (chdir(dirName) == 0)
   {
      struct dirent *entry = NULL;
      DIR *dir;
      struct stat fi;

      dir = opendir(".");
      while ((entry = readdir(dir)) != NULL)
      {
         if (strcmp(entry->d_name, ".") && strcmp(entry->d_name, "..") &&
                lstat(entry->d_name, &fi) == 0)
         {
            if (S_ISDIR(fi.st_mode))
            {
               walk_dir_and_chown(entry->d_name,
                   my_uid, admin_uid, tuid, tgid);
            }

            if ((my_uid == fi.st_uid || admin_uid == fi.st_uid)
                    && tuid != fi.st_uid)
            {
                lchown(entry->d_name, tuid, tgid);
            }
         }
      }

      closedir(dir);

      if (dirName[0] != '/')
      {
         chdir("..");
      }
   }
}
*/

void
admin_chown(const char *sroot, uid_t tuid, const char *tuser, gid_t tgid, const char *tgrp)
{
   /*
    * Changing ownership is possible only if I am installing as
    * root and the directories/files are owned by me or
    * the original admin person
    */
   struct dirent *entry = NULL;
   DIR *dir;
   struct stat fi;
   char tmp[BIG_BUF];
   uid_t my_uid;
   uid_t admin_uid;

   /*
    * First, check to see who currently is the admin user
    */
   sprintf(tmp, "%s/admin-serv", sroot);
   stat(tmp, &fi);

   admin_uid = fi.st_uid;
   my_uid    = geteuid();

   chown(sroot, tuid, tgid);

   dir = opendir(sroot);
   while ((entry = readdir(dir)) != NULL)
   {
      if (strcmp(entry->d_name, ".") && strcmp(entry->d_name, "..") &&
           lstat(entry->d_name, &fi) == 0)
      {
         if (S_ISDIR(fi.st_mode))
         {
            if ((my_uid == fi.st_uid || admin_uid == fi.st_uid) &&
                 tuid != fi.st_uid)
            {
               sprintf(tmp, CHOWN_CMD, tuser, sroot);
               system(tmp);
               sprintf(tmp, CHGRP_CMD, tgrp, sroot);
               system(tmp);
            }
         }
      }
   }
}

void
create_admserv(char *hn, const char *sroot)
{
   char httpuser[SML_BUF], httppw[SML_BUF];
   char hosts[MED_BUF], addrs[MED_BUF];
   char *hexp, *aexp;
   char dir[BIG_BUF], tstr[BIG_BUF];
   char adminUser[SML_BUF];
   char SSuser[SML_BUF];
   char SSgroup[SML_BUF];
   char ldapURL[MED_BUF];
   NSString  hostip = InstUtil::getHostIpAddress(hn);
   char serverAddress[SML_BUF];
   struct stat statbuf;
   FILE *f;
   int port;
   const char *s;
   char *ss;

   chdir(sroot);

   // Get admin port from installInfo
   s = adminInfo->get("Port");
   port = atoi(s);

   s = adminInfo->get("SysUser");
   strcpy(adminUser, s);

   s = installInfo->get(SS_USER_ID);

   strcpy(SSuser, s);

   s = installInfo->get(SS_GROUP);
   strcpy(SSgroup,s);

   s = adminInfo->get("ServerIpAddress");
   if (s == NULL || s[0] == ' ')
      strcpy(serverAddress, "\0");
   else
      strcpy(serverAddress, s);

   char *p;

   p = strdup(installInfo->get(CONFIG_LDAP_URL));

   char *cp;

   cp = escapeLDAPurl(p);

   if (cp)
   {
      strcpy(ldapURL, cp);
      free(p);
   }

   s = adminInfo->get("ServerAdminID");
   if (s)
   {
      strcpy(httpuser, s);
   }
   else
   {
#if 0
      installLog->logMessage(WARN,"Admin","ServerAdminID not specified. Default to admin.");
#endif
      strcpy(httpuser, DEFAULT_ADMINUSER);
   }
   s = adminInfo->get("ServerAdminPwd");

   if (s)
   {
      strcpy(httppw, s);
   }
   else
   {
#if 0
      installLog->logMessage(WARN,"Admin","ServerAdminPwd not specified. Use default.");
#endif
      strcpy(httppw, DEFAULT_ADMINPWD);
   }

#ifdef ORIGINAL_HOST_IP_QUERIES
   // PVO
   s = adminInfo->get("HostIpAddress");
   strcpy(hosts,s);
#else
   strcpy(hosts, "*");
   if (strchr(hn, '.'))
   {
      strcat(hosts, strchr(hn, '.'));
   }

/*
 * PVO - temporary solution for nspr21 problem
   strcpy(addrs, "*");
*/
   if (hostip.data())
   {
      strcpy(addrs, hostip.data());
   }
   else
   {
      strcpy(addrs, "*");
   }

#endif

   if (installMode == Interactive)
   {
      printf("Your parameters are now entered into the Administration Server\n"
	     "database, and the Administration Server will be started.\n\n");
   }
   else
   {
      installLog->logMessage(INFO, "Admin", "Writing configuration data to %s/admin-serv/config", sroot);
      installLog->logMessage(INFO, "Admin", "Writing configuration data to %s/%s", sroot, DEFAULT_LCACHECONF);
   }

/* create alias subdirectory and cert.log file */
   sprintf(dir, "%s%calias", sroot, PATH_DELIM);
   if ((mkdir(dir, 0750) == -1) && (errno != EEXIST))
   {
      fatal_error("Can't create directory %s (error: %s)", dir, system_errmsg());
   }

   logUninstallInfo(sroot, "admin", "admin", dir);

   /*
    * Write new admpw.
    */
/* MLM - this should actually do more httpconf type of stuff
 *       (maybe an admconf.c?) */
   sprintf(tstr, "%s/admin-serv/config/admpw", sroot);
   if (!(f = fopen(tstr, "w")))
   {
      fatal_error("Can't write to %s (error: %s)", tstr, system_errmsg());
   }

   char *pw = sha1_pw_enc (httppw);
   fprintf(f, "%s:{SHA}%s", httpuser, pw);
   fclose(f);
   chmod(tstr, S_IRUSR | S_IWUSR);

   logUninstallInfo(sroot, "admin", "admin", tstr);

#ifdef SHOW_CONFIG_FILES
   if (installMode != Silent)
      printf("Wrote %s\n", tstr);
#endif

   /*
    * Write new LDAP local DB  default files
    */
   sprintf(tstr, "%s/%s", sroot, DEFAULT_LDAPSWITCH);
   if (!(f = fopen(tstr, "w")))
   {
      fatal_error("Can't write to %s (error: %s)", tstr, system_errmsg());
   }
   if (ldapURL[0] == '\0')
   {				/* No  external LDAP, user local DB */
      fprintf(f, "directory default ldapdb://%s/%s/\n", sroot, DEFAULT_LCACHECONF);
   }
   else
   {				/* Use external LDAP */
      fprintf(f, "directory default %s\n", ldapURL);
   }
   fclose(f);

   logUninstallInfo(sroot, "admin", "admin", tstr);


   sprintf(tstr, "chmod 640 %s/%s", sroot, DEFAULT_LDAPSWITCH);
   system(tstr);

   /*
    * Write new LDAP local DB  default files for Admin Server 45 core engine.
    */
   sprintf(tstr, "%s/userdb/dbswitch.conf", sroot);
   if (!(f = fopen(tstr, "w")))
   {
      fatal_error("Can't write to %s (error: %s)", tstr, system_errmsg());
   }
   if (ldapURL[0] == '\0')
   {				/* No  external LDAP, user local DB */
      fprintf(f, "directory default ldapdb://%s/%s/\n", sroot, DEFAULT_LCACHECONF);
   }
   else
   {				/* Use external LDAP */
      fprintf(f, "directory default %s\n", ldapURL);
   }
   fclose(f);

   logUninstallInfo(sroot, "admin", "admin", tstr);


   sprintf(tstr, "chmod 640 %s/userdb/dbswitch.conf", sroot);
   system(tstr);

   hexp = make_shexp(hosts);
   aexp = make_shexp(addrs);

   /*
    * Write new ns-cron.conf:
    *
    * ConfFile server-root/admin-serv/config/cron.conf Dir /tmp Status on
    */

   sprintf(tstr, "%s/admin-serv/config/ns-cron.conf", sroot);
   if (!(f = fopen(tstr, "w")))
      fatal_error("Can't write to %s (error: %s)", tstr, system_errmsg());

   fprintf(f, "ConfFile %s/admin-serv/config/cron.conf\n", sroot);
   fprintf(f, "Dir /tmp\n");
   fprintf(f, "Status on\n");
   fclose(f);

   logUninstallInfo(sroot, "admin", "admin", tstr);

#ifdef SHOW_CONFIG_FILES
   if (installMode == Interactive)
      printf("Wrote %s\n", tstr);
#endif


   /*
    * Write new cron.conf:
    *
    * <Object name=rmsearchtmp> Command
    * "server-root/admin-serv/config/clean-search-tmp" User adminUser
    * Time 3:28 Days Sun Mon Tue Wed Thu Fri Sat </Object>
    *
    * if NS_ENTERPRISE, blank otherwise
    */

   sprintf(tstr, "%s/admin-serv/config/cron.conf", sroot);
   if (!(f = fopen(tstr, "w")))
      fatal_error("Can't write to %s (error: %s)", tstr, system_errmsg());

   logUninstallInfo(sroot, "admin", "admin", tstr);

   /*
    * Write new start, stop and restart scripts for AS 4.5 core engine.
    */

   sprintf(tstr, "%s/start-admin", sroot);
   if (!(f = fopen(tstr, "w")))
   {
      fatal_error("Can't write to %s (error: %s)", tstr, system_errmsg());
   }
   if (chmod(tstr, 0755) == -1)
   {
      fatal_error("Can't make %s executable (error: %s)", tstr,
		  system_errmsg());
   }
   fprintf(f, "#!/bin/sh\n");
   fprintf(f, "\n");
   fprintf(f, "SERVER_ROOT=%s\n", sroot);
   fprintf(f, "PRODUCT_NAME=https\n");
   fprintf(f, "PRODUCT_BIN=uxwdog\n");
   fprintf(f, "PRODUCT_SUBDIR=%s/admin-serv\n", sroot);
   fprintf(f, "LD_LIBRARY_PATH=${SERVER_ROOT}/bin/${PRODUCT_NAME}/lib:${SERVER_ROOT}/lib:${LD_LIBRARY_PATH};export LD_LIBRARY_PATH\n");
   fprintf(f, "\n");
   fprintf(f, "# This sources global variables found in SERVER_ROOT/bin/https/bin/start-*\n");
   fprintf(f, "# scripts.  Modify at your own risk.\n");
   fprintf(f, "if test -r ${SERVER_ROOT}/bin/https/bin/start-*\n");
   fprintf(f, "then\n");
   fprintf(f, "\tfor STARTSCRIPT in ${SERVER_ROOT}/bin/https/bin/start-*\n");
   fprintf(f, "\t\tdo . $STARTSCRIPT\n");
   fprintf(f, "\tdone\n");
   fprintf(f, "fi\n");
   fprintf(f, "\n");
   fprintf(f, "LIBPATH=${LD_LIBRARY_PATH}:${LIBPATH}:/usr/threads/lib:/usr/ibmcxx/lib:/usr/lib:/lib; export LIBPATH\n");
   fprintf(f, "SHLIB_PATH=${LD_LIBRARY_PATH}:${SHLIB_PATH}; export SHLIB_PATH\n");
   fprintf(f, "\n");
   fprintf(f, "NS_SERVER_HOME=${SERVER_ROOT}; export NS_SERVER_HOME\n");
   fprintf(f, "NS_HTTPS_HOME=${NS_SERVER_HOME}/bin/https; export NS_HTTPS_HOME\n");
   fprintf(f, "PATH=${SERVER_ROOT}/bin/${PRODUCT_NAME}/bin:${PATH}; export PATH\n");
   fprintf(f, "\n");
   fprintf(f, "cd $SERVER_ROOT/bin/$PRODUCT_NAME/bin\n");
   fprintf(f, "\n");
   fprintf(f, "if [ $# -eq 0 ]\n");
   fprintf(f, "then\n");
   fprintf(f, "\targ=\"-start\";\n");
   fprintf(f, "elif [ $# -eq 1 -a x$1 = x-e ]\n");
   fprintf(f, "then\n");
   fprintf(f, "\targ=\"-start\";\n");
   fprintf(f, "else\n");
   fprintf(f, "\targ=$1;\n");
   fprintf(f, "\tshift;\n");
   fprintf(f, "fi\n");
   fprintf(f, "\n");
   fprintf(f, "case $arg in\n");
   fprintf(f, "\t-start)\n");
   fprintf(f, "\t\t./$PRODUCT_BIN -d $PRODUCT_SUBDIR/config $@\n");
   fprintf(f, "\t\tif test $? -ne 0 ; then\n");
   fprintf(f, "\t\t\texit 1\n");
   fprintf(f, "\t\tfi\n");
   fprintf(f, "\t\t;;\n");
   fprintf(f, "\t-shell)\n");
   fprintf(f, "\t\t/bin/sh\n");
   fprintf(f, "\t\t;;\n");
   fprintf(f, "\t*)\n");
   fprintf(f, "\t\techo \"Usage: start [-start|-shell]\"\n");
   fprintf(f, "\t\texit 1\n");
   fprintf(f, "\t\t;;\n");
   fprintf(f, "esac\n");
   fprintf(f, "\n");
   fprintf(f, "exit 0\n");
   fclose(f);
   
   logUninstallInfo(sroot, "admin", "admin", tstr);

   sprintf(tstr, "%s/stop-admin", sroot);
   if (!(f = fopen(tstr, "w")))
   {
      fatal_error("Can't write to %s (error: %s)", tstr, system_errmsg());
   }
   if (chmod(tstr, 0755) == -1)
   {
      fatal_error("Can't make %s executable (error: %s)", tstr,
		  system_errmsg());
   }
   fprintf(f, "#!/bin/sh\n");
   fprintf(f, "\n");
   fprintf(f, "PID_FILE=%s/admin-serv/logs/pid\n", sroot);
   fprintf(f, "if test -f $PID_FILE ; then\n");
   fprintf(f, "\tkill -TERM -`cat $PID_FILE`\n");
   fprintf(f, "\tif test $? -ne 0 ; then\n");
   fprintf(f, "\t\texit 1\n");
   fprintf(f, "\tfi\n");
   fprintf(f, "else\n");
   fprintf(f, "\techo server not running\n");
   fprintf(f, "\texit 1\n");
   fprintf(f, "fi\n");
   fprintf(f, "\n");
   fprintf(f, "loop_counter=1\n");
   fprintf(f, "max_count=30\n");
   fprintf(f, "while test $loop_counter -le $max_count; do\n");
   fprintf(f, "\tloop_counter=`expr $loop_counter + 1`\n");
   fprintf(f, "\tif test -f $PID_FILE ; then\n");
   fprintf(f, "\t\tsleep 2\n");
   fprintf(f, "\telse\n");
   fprintf(f, "\t\texit 0\n");
   fprintf(f, "\tfi\n");
   fprintf(f, "done\n");
   fprintf(f, "echo server not responding to exit command\n");
   fprintf(f, "exit 1\n");
   fclose(f);

   logUninstallInfo(sroot, "admin", "admin", tstr);

   sprintf(tstr, "%s/restart-admin", sroot);
   if (!(f = fopen(tstr, "w")))
   {
      fatal_error("Can't write to %s (error: %s)", tstr, system_errmsg());
   }
   if (chmod(tstr, 0755) == -1)
   {
      fatal_error("Can't make %s executable (error: %s)", tstr,
		  system_errmsg());
   }
   fprintf(f, "#!/bin/sh\n");
   fprintf(f, "\n");
   fprintf(f, "%s/stop-admin\n", sroot);
   fprintf(f, "sleep 3\n");
   fprintf(f, "%s/start-admin\n", sroot);
   fprintf(f, "\n");
   fclose(f);

   logUninstallInfo(sroot, "admin", "admin", tstr);

   /*
    * Add start-jvm script to set Java-related environment for servlets.
    * This code is based on the ES4.x installer. See function installJC()
    * in ES4.x source file ns/netsite/installer/src/src/httpconf.cpp
    */

   sprintf(tstr, "%s/bin/https/bin/start-jvm", sroot);
   if (!(f = fopen(tstr, "w")))
   {
      fatal_error("Can't write to %s (error: %s)", tstr, system_errmsg());
   }
   if (chmod(tstr, 0755) == -1)
   {
      fatal_error("Can't make %s executable (error: %s)", tstr,
		  system_errmsg());
   }
   fprintf(f, "NSES_JRE=%s/bin/base/jre; export NSES_JRE\n", sroot);
   fprintf(f, "NSES_SERVER_HOME=%s; export NSES_SERVER_HOME\n", sroot);

   // See platform specific defines for RUNTIME_LIBPATH  in servlet-conf.h
   fprintf(f, "NSES_JRE_RUNTIME_LIBPATH=%s; export NSES_JRE_RUNTIME_LIBPATH\n", NSES_JRE_RUNTIME_LIBPATH);

   fprintf(f, "# DO NOT CHANGE ANYTHING BELOW THIS LINE\n");

   // NSES_CLASSPATH is defined in servlet-conf.h 
   fprintf(f, "NSES_CLASSPATH=%s; export NSES_CLASSPATH\n", NSES_CLASSPATH);

   fprintf(f, "LD_LIBRARY_PATH=${NSES_JRE_RUNTIME_LIBPATH}:${LD_LIBRARY_PATH}; export LD_LIBRARY_PATH\n");
   fclose(f);

   logUninstallInfo(sroot, "admin", "admin", tstr);

   sprintf(tstr, "%s/%s", sroot, DEFAULT_SSUSERCONF);
   if (!(f = fopen(tstr, "w")))
      fatal_error("Can't write to %s (error: %s)", tstr,
		  system_errmsg());
   fprintf(f, "SuiteSpotUser\t%s\n", SSuser);
   fprintf(f, "SuiteSpotGroup\t%s\n", SSgroup);
   fclose(f);

   logUninstallInfo(sroot, "admin", "admin", tstr);

   if (!geteuid())
   {
      /* Current pw ent, target pw ent */
      struct passwd *cpw;
      struct group  *tgp;
      uid_t tuid;
      gid_t tgid;

      setpwent();
      if (!(cpw = getpwnam(adminUser)))
      {
	 fatal_error("Can't get password entry for user %s.", adminUser, NULL);
      }
      tuid = cpw->pw_uid;
      tgid = cpw->pw_gid;

      if (!(tgp = getgrgid(tgid)))
      {
	 fatal_error("Can't get group entry for target group.", NULL, NULL);
      }
      endpwent();

      if (installMode == Interactive)
         printf("Changing ownership to admin user %s...\n", adminUser);
      admin_chown(sroot, tuid, adminUser, tgid, tgp->gr_name);

#ifdef HPUX
      /*
      setresuid((uid_t) -1, tuid, (uid_t) -1);
      setresgid((gid_t) -1, tgid, (gid_t) -1);
      */
      setresuid(tuid, tuid, (uid_t) -1);
      setresgid(tgid, tgid, (gid_t) -1);
#else
      seteuid(tuid);
      setegid(tgid);
#endif
   }


   {
      /*
       * This is ok because running user is either root or part of the
       * SSgroup
       */
      sprintf(tstr, SHARED_CHGRP_CMD, SSgroup, sroot);
      system(tstr);
      sprintf(tstr, SSCONF_CHGRP_CMD, SSgroup, sroot);
      system(tstr);
      sprintf(tstr, SECALIAS_CHGRP_CMD, SSgroup, sroot);
      system(tstr);
   }

   /*
    * Turn of end-user access by default.
    */
   sprintf(tstr, "%s/adminacl", sroot);
   if (stat(tstr, &statbuf) < 0)
   {
      if ((mkdir(tstr, 0750) == -1) && (errno != EEXIST))
      {
         fatal_error("Can't create directory %s (error: %s)", tstr,
                     system_errmsg());
      }
   }

   logUninstallInfo(sroot, "admin", "admin", tstr);


   ////////////////////////////////////////////////////////////////
   // Now everything seems to be OKAY, install SIE for admin-server
   ////////////////////////////////////////////////////////////////

   LdapError ldapError;
   NSString  adminDomain;
   NSString  sieDN;
   NSString  isieDN;
   NSString  configDN;
   char *ldapStart = NULL;

   printf("Setting up Administration Server Instance...\n");
   installLog->logMessage(INFO, "Admin", "Setting up Administration Server Instance...\n");

   Ldap ldap (ldapError, ldapURL, installInfo->get(MC_ADMIN_ID),
                    installInfo->get(MC_ADMIN_PWD),
                    NULL, NULL); // NULL for now until figure out PVO

   if (ldapError != OKAY)
   {
      sprintf(dir, "%ld", ldapError.errorCode());
      fatal_error("ERROR: Ldap authentication failed (%s:%s)" , dir, ldapError.msg());
   }

   LdapEntry *sieEntry = new LdapEntry(&ldap);
   LdapEntry *appEntry = new LdapEntry(&ldap);
   LdapEntry *configEntry = new LdapEntry(&ldap);

   // Check if there's already an SIE for this Admin
   sprintf(tstr, "%s/admin-serv/config/adm.conf", sroot);
   NVPair admConf(tstr);

   adminDomain = installInfo->get(DS_ADMIN_DOMAIN);

   if (admConf.isEmpty() == False)
   {
      sieDN = admConf.get("sie");
      isieDN = admConf.get("isie");
      if (s=admConf.get("ldapStart"))
      {
         ldapStart = strdup(s);
      }
   }

   /*
    * Generate pwd for the SIE
    */

/*
   char password[9];

   srand(getpid() + time(NULL));
   for (int i = 0; i < 8 ; i++)
   {
      int charoff = rand() % 52;
      if (charoff < 26)
        password[i] = (char) charoff + 'A';
      else
        password[i] = (char) charoff - 26 + 'a';
   }

   password[8] = 0;

*/

   sprintf(tstr, "%s/%s/admin/admin.inf", sroot, SETUP_DIR);
   NVPair adminInf(tstr);
   NVPair *adminInfo = adminInf.getSection("admin");

   appEntry->addAttribute("objectclass","top");
   appEntry->addAttribute("objectclass","nsApplication");

   appEntry->setAttribute("cn", "Fedora Administration Server");

   appEntry->addAttribute("nsVendor", adminInfo->get(PKG_VENDOR));
   appEntry->addAttribute("description", adminInfo->get(PKG_DESCRIPTION));
   appEntry->addAttribute("nsProductName", adminInfo->get(PKG_NAME));
   appEntry->addAttribute("nsNickname", adminInfo->get(PKG_NICKNAME));
   appEntry->addAttribute("nsProductVersion", adminInfo->get(PKG_VERSION));
   appEntry->addAttribute("nsBuildNumber", adminInfo->get(PKG_BUILDNUMBER));
   appEntry->addAttribute("nsRevisionNumber", adminInfo->get(PKG_REVISION));
   appEntry->addAttribute("nsInstalledLocation", sroot);

   adminInfo->stampTime();
   appEntry->addAttribute("installationTimeStamp", adminInfo->timeStamp());
   appEntry->addAttribute("nsExpirationDate", adminInfo->get(PKG_EXPIRY));
   appEntry->addAttribute("nsBuildSecurity", adminInfo->get(PKG_SECURITY));
   appEntry->addAttribute("nsServerMigrationClassName", "com.netscape.management.admserv.AdminServerProduct@admserv70.jar");

   sieEntry->clear();

 
    /*
     * set password expiration to some later date so 
     * admin server will not stop working due to password expiration
     * see bug#343369 - http://scopus/bugsplat/show_bug.cgi?id=343369
     * 
     * passwordExpirationTime is a hidden attribute.
     * For some reason if we don't add this attribute to the top of the
     * attribute list we will end up with 2 passwordExpirationTime 
     * attribute...
     * leave it here don't move into setupCreateSIE.
     *
     */
   sieEntry->setAttribute("passwordExpirationTime", EPOCH_32BIT_END_TIME);

   sieEntry->addAttribute("objectclass", "top");
   sieEntry->addAttribute("objectclass", "netscapeServer");
   sieEntry->addAttribute("objectclass", "nsAdminServer");
   sieEntry->addAttribute("objectclass", "nsResourceRef");

   /* Form the CN for Admin Server, e.g. admin-serv-lupine */
   if (sieDN.data() == NULL)
   {
      sprintf(tstr, "%s-%s", "admin-serv", hn);
      ss = strchr(tstr, '.');
      if (ss) *ss = 0;
   }
   else
   {
      // Get the CN part
      char **tmpDN = ldap_explode_dn(sieDN.data(), 1);
      strcpy(tstr, tmpDN[0]);
      ldap_value_free(tmpDN);
   }

   sieEntry->setAttribute("cn", tstr);
   sieEntry->addAttribute("nsServerId", "admin-serv");
   sieEntry->addAttribute("userPassword", httppw);
   sieEntry->addAttribute("serverHostName", hn);
   sieEntry->addAttribute("serverRoot", sroot);
   sieEntry->addAttribute("serverProductName", "Administration Server");

   ldapError = createSIE(sieEntry, appEntry, hn, sroot, adminDomain);

   sieDN = sieEntry->entryDN();
   isieDN = appEntry->entryDN();

   if (ldapError != OKAY)
   {
      sprintf(dir, "%d", ldapError.errorCode());
      fatal_error("ERROR: cannot create server instance (%s:%s)", dir, ldapError.msg()); // PVO
   }


   NSString configTaskDN, opTaskDN;

   // Write configuration parameters (see ns-admin-schema.conf)
   configDN = NSString("cn=configuration") + LDAP_PATHSEP + sieDN;

   sprintf(dir, "%d", port);

   const char *userDirRef = strchr(isieDN.data(), ',') + 1;
   while (*userDirRef == ' ') ++userDirRef;

   if (configEntry->retrieve(configDN) != OKAY)
   {
      configEntry->addAttribute("objectclass", "nsConfig");
      configEntry->addAttribute("objectclass", "nsAdminConfig");
      configEntry->addAttribute("objectclass", "nsAdminObject");
      configEntry->addAttribute ("cn", "Configuration");

      configEntry->addAttribute ("nsServerPort", dir);
      configEntry->addAttribute ("nsSuiteSpotUser", adminUser);
      configEntry->addAttribute ("nsServerAddress", serverAddress);

      configEntry->addAttribute ("nsAdminEnableEnduser", "on");
      configEntry->addAttribute ("nsAdminEnableDSGW", "on");

      /*
       * Mandatory fields here
       */
      configEntry->addAttribute("objectclass", "nsDirectoryInfo");
      configEntry->setAttribute("nsDirectoryInfoRef", userDirRef);

      configEntry->addAttribute ("nsAdminUsers", "admin-serv/config/admpw");
      configEntry->addAttribute ("nsErrorLog", "admin-serv/logs/error");
      configEntry->addAttribute ("nsPidLog",   "admin-serv/logs/pid");
      configEntry->addAttribute ("nsAccessLog","admin-serv/logs/access");
      configEntry->addAttribute ("nsAdminCacheLifetime", "600");

      if (hosts[0])
         configEntry->addAttribute ("nsAdminAccessHosts", hexp);
      //if (addrs[0])
      //   configEntry->addAttribute ("nsAdminAccessAddresses", aexp);
      configEntry->addAttribute ("nsAdminAccessAddresses", "*");
      configEntry->addAttribute ("nsAdminOneAClDir", "adminacl");

      sprintf(tstr, DEFAULT_ADMIN_CONFIG_ACI, userDirRef);
      configEntry->addAttribute("aci",tstr);

      /*
       * Don't set this value - absent is default
       * configEntry->addAttribute ("nsAdminEnableDsgw",
       */

      configEntry->addAttribute ("nsDefaultAcceptLanguage", adminInfo->get(PKG_DEFAULT_LANGUAGE));

      sprintf(tstr, "com.netscape.management.admserv.AdminServer@admserv70.jar@%s", sieDN.data());
      configEntry->addAttribute ("nsClassname", tstr);

      // Set the appropriate ACI on for this branch
      createACIForConfigEntry(configEntry, sieDN);

      ldapError = configEntry->insert(configDN);
   }
   else
   {
      const char *s;
      sprintf(tstr, "com.netscape.management.admserv.AdminServer@admserv70.jar@%s", sieDN.data());
      configEntry->setAttribute ("nsClassname", tstr);
      configEntry->setAttribute ("nsServerPort", dir);
      configEntry->setAttribute ("nsSuiteSpotUser", adminUser);
      
      /*
       * in custom mode, update IP address in DS
       */
      if ((s = installInfo->get("InstallType")) && (strcasecmp(s, "3") == 0))
         configEntry->setAttribute ("nsServerAddress", serverAddress);

      ldapError = configEntry->replace(configDN);
   }

   if (ldapError != OKAY)
   {
      sprintf(dir, "%d", ldapError.errorCode());
      fatal_error("ERROR: cannot write SIE configuration (%s:%s)", dir, ldapError.msg());
   }

   /*
    *  Write Tasks node
    */

   configTasks(&ldap, sroot, sieDN);

   // Write global configuration paramters

   sprintf(tstr, "%s/shared/config/ds.conf", sroot);
   if (!(f = fopen(tstr,"w")))
   {
      printf("Warning - Can't write to %s (error: %s)", tstr, system_errmsg());
   }
   else
   {
      fprintf(f, "%s:%s\n", DS_ADMIN_DOMAIN, adminDomain.data());
      logUninstallInfo(sroot, "admin", "admin", tstr);
   }

   char *adminDomainDN;
   adminDomainDN = formAdminDomainDN(adminDomain);
   configGlobalParameters(&ldap, sroot, adminDomainDN);
   nsSetupFree (adminDomainDN);

   /*
    ***************************************************************
    * Fix ACIs under o=netscapeRoot
    ***************************************************************
    */
   /*ldapError = repair_group_expansion_aci(&ldap);
   if (ldapError != OKAY)
   {
      sprintf(dir, "%d", ldapError.errorCode());
      fatal_error("ERROR: cannot modify root ACIs (%s:%s)", dir, ldapError.msg()); 
   }*/

   /*ldapError = set_locate_admserv_aci(&ldap);
   if (ldapError != OKAY)
   {
      sprintf(dir, "%d", ldapError.errorCode());
      fatal_error("ERROR: cannot modify root ACIs (%s:%s)", dir, ldapError.msg()); 
   }*/

   /*
    ***************************************************************
    * Write new adm.conf (new for Admin40)
    ***************************************************************
    */

   LDAPURLDesc *ludpp = NULL;
   ldap_url_parse((char *) installInfo->get(CONFIG_LDAP_URL), &ludpp);

/*

   fprintf(f, "ldapHost: %s\n", ludpp->lud_host);
   fprintf(f, "ldapPort: %d\n", ludpp->lud_port);
   fprintf(f, "sie: %s\n", sieDN.data());
   fprintf(f, "siepid: %s\n", httppw);
   fprintf(f, "isie: %s\n", isieDN.data());
   fprintf(f, "port: %d\n", port);
   fclose(f);

*/
   admConf.set("ldapHost",  ludpp->lud_host);
   admConf.set("ldapPort",  (unsigned long)ludpp->lud_port);
   admConf.set("sie", sieDN.data());
   admConf.set("siepid", httppw);
   admConf.set("isie", isieDN.data());
   admConf.set("port",  (unsigned long)port);

   if (ldapStart == NULL)
   {
      const char *isConfigDS = NULL;
      const char *dsServerID = NULL;

      dsInfo = installInfo->getSection("slapd");
      if (dsInfo != NULL)
      {
          isConfigDS = dsInfo->get("SlapdConfigForMC");
          dsServerID = dsInfo->get("ServerIdentifier");
          if (isConfigDS && dsServerID && !strcasecmp(isConfigDS, "Yes"))
          {
             sprintf(tstr, "slapd-%s/start-slapd", dsServerID);
             ldapStart = strdup(tstr);
          }
      }
   }

   if (ldapStart)
   {
      admConf.set("ldapStart", ldapStart);
   }

   sprintf(tstr, "%s/admin-serv/config/adm.conf", sroot);

   admConf.setFormat(2);
   admConf.write(tstr);
   chmod(tstr, S_IRUSR | S_IWUSR);

   logUninstallInfo(sroot, "admin", "admin", tstr);

   // Mark to remove the following guys which are
   // created later on also
   sprintf(tstr, "%s/admin-serv/config/local.conf", sroot);
   logUninstallInfo(sroot, "admin", "admin", tstr);
   sprintf(tstr, "%s/admin-serv/logs/access", sroot);
   logUninstallInfo(sroot, "admin", "admin", tstr);
   sprintf(tstr, "%s/admin-serv/logs/error", sroot);
   logUninstallInfo(sroot, "admin", "admin", tstr);
   sprintf(tstr, "%s/admin-serv/logs/admpw.log", sroot);
   logUninstallInfo(sroot, "admin", "admin", tstr);
   sprintf(tstr, "%s/admin-serv/logs/dsconfig.log", sroot);
   logUninstallInfo(sroot, "admin", "admin", tstr);

   /* PVO remember to set access permission here */

   /*
    ***************************************************************
    * Write new magnus.conf (new for Admin45)
    * Note: These are the default values used in server startup.
    *       Some of these variables may be overwritten after
    *       magnus.conf has been processed by the server.
    ***************************************************************
    */

   sprintf(tstr, "%s/admin-serv/config/magnus.conf", sroot);
   if (!(f = fopen(tstr, "w")))
   {
      fatal_error("Can't write to %s (error: %s)", tstr, system_errmsg());
   }
   fprintf(f, "NetsiteRoot %s\n", sroot);
   fprintf(f, "ServerID admin-serv\n");
   fprintf(f, "ServerName %s\n", hn);
   fprintf(f, "ErrorLog %s/admin-serv/logs/error\n", sroot);
   fprintf(f, "PidLog %s/admin-serv/logs/pid\n", sroot);
   fprintf(f, "User %s\n", adminUser);
   fprintf(f, "AdminUsers %s/admin-serv/config/admpw\n", sroot);
   fprintf(f, "MtaHost localhost\n");
   fprintf(f, "DNS on\n");
   fprintf(f, "Security off\n");
   fprintf(f, "ClientLanguage %s\n", adminInfo->get(PKG_DEFAULT_LANGUAGE));
   fprintf(f, "AdminLanguage %s\n", adminInfo->get(PKG_DEFAULT_LANGUAGE));
   fprintf(f, "DefaultLanguage %s\n", adminInfo->get(PKG_DEFAULT_LANGUAGE));
   fprintf(f, "RqThrottle 64\n");
   fprintf(f, "ConnQueueSize 256\n");
   fprintf(f, "StackSize 131072\n");
   fprintf(f, "TempDir %s/admin-serv/tmp\n", sroot);
   if(serverAddress[0] != '\0')
     fprintf(f, "Address %s\n", serverAddress);

  fprintf(f, "\n");
  fprintf(f, "Init fn=flex-init access=%s\n\n", "\"$accesslog\" format.access=\"%Ses->client.ip% - %Req->vars.auth-user% [%SYSDATE%] \\\"%Req->reqpb.clf-request%\\\" %Req->srvhdrs.clf-status% %Req->srvhdrs.content-length%\"");

#if defined(HPUX)
   fprintf(f, "Init fn=\"load-modules\" shlib=\"%s/bin/https/lib/libAdmservPlugin.sl\"", sroot);
#else
   fprintf(f, "Init fn=\"load-modules\" shlib=\"%s/bin/https/lib/libAdmservPlugin.so\"", sroot);
#endif /* HPUX */
   fprintf(f, " funcs=\"admin40_task_eval_init,admin40_host_ip_init,admin40_host_ip_check,admin40_uri2path,admin40_authenticate,admin40_runtime_commands,admin40_ftype,admin40_isdiropen,admin40_flex_log,admin40_check_ds_availability_init\"\n");
   fprintf(f, "Init fn=\"admin40_check_ds_availability_init\" nsRoot=\"%s\"\n", sroot);
   fprintf(f, "Init fn=\"admin40_task_eval_init\" nsRoot=\"%s\" cacheLifeTime=\"600\"\n", sroot);
   fprintf(f, "Init fn=\"admin40_host_ip_init\"\n");

   fprintf(f, "\n#Servlet engine startup directives\n");
#if defined(HPUX)
   fprintf(f, "Init fn=\"load-modules\" shlib=\"%s/bin/https/lib/libNSServletPlugin.sl\" funcs=\"NSServletEarlyInit,NSServletLateInit,NSServletNameTrans,NSServletService\"\n", sroot);
#else
   fprintf(f, "Init fn=\"load-modules\" shlib=\"%s/bin/https/lib/libNSServletPlugin.so\" funcs=\"NSServletEarlyInit,NSServletLateInit,NSServletNameTrans,NSServletService\"\n", sroot);
#endif /* HPUX */
   fprintf(f, "Init fn=\"NSServletEarlyInit\" EarlyInit=\"yes\"\n");
   fprintf(f, "Init fn=\"NSServletLateInit\"  LateInit=\"yes\" CatchSignals=\"no\" Signals=\"15\" \n");

   fclose(f);

   chmod(tstr, S_IRUSR | S_IWUSR);
   logUninstallInfo(sroot, "admin", "admin", tstr);

   /* 
    ****************************************************************
    * Write new nsfc.conf (new for Admin62, tunes server-side
    * cache for static content
    ****************************************************************
    */

   sprintf(tstr, "%s/admin-serv/config/nsfc.conf", sroot);
   if (!(f = fopen(tstr, "w")))
   {
      fatal_error("Can't write to %s (error: %s)", tstr, system_errmsg());
   }
   fprintf(f, "MaxFiles=128\n", sroot);

   fclose(f);

   chmod(tstr, S_IRUSR | S_IWUSR);
   logUninstallInfo(sroot, "admin", "admin", tstr);

   /*
    ***************************************************************
    * Update server.xml, new for iWS6.  The file template is in
    * ns/netsite/admserv/newinst/common/config cvs directory.
    ***************************************************************
    */

   sprintf(tstr, "%s/admin-serv/config/server.xml", sroot);

   {
       char accesslog[BIG_BUF], aclfile[BIG_BUF], portnum[32];

       /* Set port */
       xml_file_mod_t port_mod =
           {"LS id=\"ls1\"", "port", portnum};

       /* Set server name */
       xml_file_mod_t srvname_mod =
           {"CONNECTIONGROUP", "servername", hn}; 
       xml_file_mod_t urlhosts_mod =
           {"VS id=\"admin-serv\"", "urlhosts", hn};

       xml_file_mod_t accesslog_mod =
       {"VARS", "accesslog", accesslog };

       xml_file_mod_t aclfile_mod =
       {"ACLFILE id=\"acl1\"", "file", aclfile };

       xml_file_mod_t *mods[6] = { 
           &port_mod, &srvname_mod, &urlhosts_mod,
           &accesslog_mod, &aclfile_mod, NULL};

       sprintf(portnum, "%d", port);
       sprintf(accesslog, "%s/admin-serv/logs/access", sroot);
       sprintf(aclfile,   "%s/adminacl/generated.https-admserv.acl", sroot);

       modify_xml_file(tstr, mods);
   }

   chmod(tstr, S_IRUSR | S_IWUSR);
   logUninstallInfo(sroot, "admin", "admin", tstr);

   /*
    ***************************************************************
    * Write new obj.conf (new for Admin45)
    ***************************************************************
    */

   sprintf(tstr, "%s/admin-serv/config/obj.conf", sroot);
   if (!(f = fopen(tstr, "w")))
   {
      fatal_error("Can't write to %s (error: %s)", tstr, system_errmsg());
   }

   fprintf(f, "\n");
   fprintf(f, "<Object name=\"default\">\n");

   fprintf(f, "AuthTrans fn=\"admin40_host_ip_check\"\n");
   fprintf(f, "AuthTrans fn=\"admin40_isdiropen\"\n");
   fprintf(f, "AuthTrans fn=\"admin40_authenticate\"\n");
   fprintf(f, "AuthTrans fn=\"admin40_uri2path\"\n");
/* fprintf(f, "NameTrans fn=\"pfx2dir\" from=\"/clients\" dir=\"%s/clients\"\n", sroot);*/
   fprintf(f, "ObjectType fn=\"admin40_ftype\"\n");
   fprintf(f, "Service fn=\"admin40_runtime_commands\" type=\"admin-internal/command\"\n");

   fprintf(f, "Service fn=\"admin-send-cgi\" type=\"magnus-internal/cgi\"\n");
   fprintf(f, "Service fn=\"NSServletService\" type=\"magnus-internal/servlet\"\n");
   fprintf(f, "Service fn=\"imagemap\" method=\"GET\" type=\"magnus-internal/imagemap\"\n");
   fprintf(f, "Service method=(GET|HEAD|POST) type=*~magnus-internal/* fn=send-file\n");
   fprintf(f, "Error fn=\"admin-error\" reason=\"server error\"\n");
   fprintf(f, "AddLog fn=\"admin40_flex_log\" name=\"access\"\n");
   fprintf(f, "NameTrans fn=\"NSServletNameTrans\" name=\"servlet\"\n");
   fprintf(f, "PathCheck fn=find-index index-names=\"index.html,home.html\"\n");
   fprintf(f, "Service type=\"magnus-internal/jsp\" fn=\"NSServletService\"\n");
   fprintf(f, "</Object>\n");
   fprintf(f, "\n");
   fprintf(f, "<Object name=\"servlet\">\n");
   fprintf(f, "ObjectType fn=force-type type=text/html\n");
   fprintf(f, "Service fn=\"NSServletService\"\n");
   fprintf(f, "</Object>\n");
   fprintf(f, "\n");
   fclose(f);

   chmod(tstr, S_IRUSR | S_IWUSR);
   logUninstallInfo(sroot, "admin", "admin", tstr);

   /*
    *******************************************************************
    * Write new ACL local file (new for Admin45, but not actually used)
    *******************************************************************
    */

   sprintf(tstr, "%s/adminacl/generated.https-admserv.acl", sroot);
   if (!(f = fopen(tstr, "w")))
   {
      fatal_error("Can't write to %s (error: %s)", tstr, system_errmsg());
   }
   fprintf(f, "version 3.0;\n");
   fprintf(f, "acl \"admin-defaults\";\n");
   fprintf(f, "authenticate (user, group) {\n");
   fprintf(f, "  prompt = \"WebServer Server\";\n");
   fprintf(f, "};\n");
   fprintf(f, "allow (read, list, execute,info) user = \"anyone\";\n");
   fprintf(f, "allow (write, delete) user = \"all\";\n");
   fprintf(f, "\n");
   fclose(f);

   chmod(tstr, S_IRUSR | S_IWUSR);
   logUninstallInfo(sroot, "admin", "admin", tstr);

   sprintf(dir, "%s/admin-serv/logs/access", sroot);
   if(InstUtil::fileExists(dir) == True) {
     /* back up the existing Admin Server access log file */
     sprintf(tstr, "mv %s %s.bak", dir, dir);
     system(tstr);
     
     sprintf(dir, "%s/admin-serv/logs/access.bak", sroot);
     logUninstallInfo(sroot, "admin", "admin", tstr);
   }

   /*
    ***********************************************************************
    * Write out new jvm12.conf file
    ***********************************************************************
    */
   sprintf(tstr, "%s/admin-serv/config/jvm12.conf", sroot);
   chmod(tstr, 0644);
   
   if (!(f = fopen(tstr, "w")))
   {
      fatal_error("Can't write to %s (error: %s)", tstr, system_errmsg());
   }
   fprintf(f, "[JVMConfig]\n");
#ifdef LINUX
   /*
    * Need to set IBM JRE specific bootclasspath on Linux 
    */
   fprintf(f, "jvm.option=-Xrs -Xbootclasspath:%s/bin/base/jre/lib/core.jar:%s/bin/base/jre/lib/server.jar:%s/bin/base/jre/lib/charsets.jar:%s/bin/base/jre/lib/security.jar:%s/bin/base/jre/lib/xml.jar:%s/bin/base/jre/lib/graphics.jar:%s/bin/https/jar/NSServletLayer.jar:%s/bin/https/jar/NSJavaUtil.jar:%s/bin/https/jar/NSJavaMiscUtil.jar:%s/bin/https/jar/servlet.jar:%s/bin/https/jar/jspengine.jar:%s/bin/https/jar/jakarta-naming.jar -Dsun.boot.library.path=%s/bin/base/jre/bin:%s/bin/https/lib\n", sroot, sroot, sroot, sroot, sroot, sroot, sroot, sroot, sroot, sroot, sroot, sroot, sroot, sroot);
#else
   fprintf(f, "jvm.option=-Xrs\n");
   fprintf(f, "jvm.option=-server\n");
#endif
   fclose(f);

   chmod(tstr, 0644);

   /*
    ***********************************************************************
    * 518630: Set file permissions, no write access for group and other
    ***********************************************************************
    */

   sprintf(tstr, "%s/admin-serv/config/contexts.properties", sroot);
   chmod(tstr, 0644);

   sprintf(tstr, "%s/admin-serv/config/rules.properties.default", sroot);
   chmod(tstr, 0644);

   sprintf(tstr, "%s/admin-serv/config/servlets.properties.default", sroot);
   chmod(tstr, 0644);

   sprintf(tstr, "%s/admin-serv/config/servlets.properties.default", sroot);
   chmod(tstr, 0644);

   sprintf(tstr, "%s/admin-serv/config/exempturis.conf", sroot);
   chmod(tstr, 0644);

   sprintf(tstr, "%s/plugins/snmp/netscape-main.mib", sroot);
   chmod(tstr, 0644);

   sprintf(tstr, "%s/java/jars", sroot);
   chmod(tstr, 0755);

   sprintf(tstr, "%s/bin/admin/admconfig", sroot);
   chmod(tstr, 0755);

   sprintf(tstr, "%s/uninstall", sroot);
   chmod(tstr, 0750);
}


int
update_schema(const char *sroot)
{
  char schema_file[BIG_BUF];
  
  sprintf(schema_file, "%s/setup/admin/ns-common-schema.conf", sroot);
  return setupInsertPluginSchemaEntries(installInfo->get(CONFIG_LDAP_URL),
					installInfo->get(MC_ADMIN_ID),
					installInfo->get(MC_ADMIN_PWD),
					schema_file);
}

/*
 * 359259: Add ACIs under o=netscapeRoot to allow regular, non-admin users to
 * locate admin servers. This is required for downloads of jars referenced from
 *  Users and Groups configuration under ou=Global Preferences.
 */
static int set_locate_admserv_aci(Ldap *ldap) {
    LDAPMod *mods[2];
    LdapErrorCode ldapError;
    LDAPMod mod;
    char *aciVal[2];

    // Allow read access to host, port, address and security attributes for SIEs
    aciVal[0] = (char*)LOCATE_ADMSERV_ACI;
    aciVal[1] = NULL;

    mod.mod_op = LDAP_MOD_ADD;
    mod.mod_type = "aci";
    mod.mod_values = aciVal;

    mods[0] = &mod;
    mods[1] = NULL;

    ldapError = ldap->modify("o=netscapeRoot", mods);

    if (ldapError != LDAP_SUCCESS &&
        ldapError != LDAP_TYPE_OR_VALUE_EXISTS) {
        fprintf(stderr, 
           "Could not modify ACI for o=netscapeRoot, \"%s\"", ldap_err2string(ldapError));
        return ldapError;
    }

    return OKAY;
}

/* 
 * 337487: Fix "Enable Group Expansion" ACI under o=netscapeRoot. The ACI
 * is defined as <groupdnattr="ldap:///o=NetscapeRoot?uniquemember?sub">.
 * This is instructing that all groups will be located under the
 * o=netscapeRoot suffix, which is true only for administrators. Set
 * <groupdnattr="uniquemember"> which allows a group to be under any suffix.
 */
static int repair_group_expansion_aci(Ldap *ldap) {
   LdapError ldapError;
   LdapEntry *root = new LdapEntry(ldap);
   char ** aciVals, *aci;

    ldapError = root->retrieve("o=netscapeRoot");
    if (ldapError != OKAY) {
        fprintf(stderr, "Can not read o=netscapeRoot err=%d\n", ldapError);
        return ldapError;
    }
    
    aciVals = root->getAttributes("aci");
    while (aci=*aciVals++) {

        if (strstr(aci, "Enable Group Expansion") &&
            strstr(aci, "groupdnattr=\"ldap:///o=NetscapeRoot?uniquemember?sub\"")) {
        
            LDAPMod *mods[3];
            LDAPMod mod1, mod2;
            char *prevVal[2], *newVal[2];

            prevVal[0] = aci;
            prevVal[1] = NULL;

            newVal[0]  = (char*)GROUP_EXPANSION_ACI;
            newVal[1]  = NULL;

            mod1.mod_op = LDAP_MOD_DELETE;
            mod1.mod_type = "aci";
            mod1.mod_values = prevVal;

            mod2.mod_op = LDAP_MOD_ADD;
            mod2.mod_type = "aci";
            mod2.mod_values = newVal;

            mods[0] = &mod1;
            mods[1] = &mod2;
            mods[2] = NULL;

            ldapError = ldap->modify("o=netscapeRoot", mods);

            if (ldapError != OKAY) {
                fprintf(stderr, "Failed to modify o=netscapeRoot ACI ldapError=%s\n", ldap_err2string(ldapError));
                return ldapError;
           }
       }
    }

    return OKAY;
}


int set_xml_attr(char *line, char *section,  char *attr, char *val) {
    /*
       Check if the line starts with the section id string. The id string may
       be prefixed with white spaces and '<'.
     */
    char *section_ref = strstr(line, section);
    char *c, *attr_ref;
    char out[BIG_BUF];
    if (section_ref == NULL) {
        return 0;
    }
    if (section_ref != line) {
        for (c=line; c != section_ref; c++) {
			if (*c != ' ' && *c != '\t' && *c!= '<') {
                return 0;
            }
        }
    }

    /* Find the attribute */
    attr_ref = strstr (line, attr);
    if (attr_ref == NULL) {
       return 0;
    }

    /* Replace the attribute value */
    memset(out, 0, BIG_BUF); 
    memcpy(out, line, attr_ref-line);
    for (c=attr_ref; c && *c != '"'; c++) {
       ; /* do nothing */ 
    }
    if (c == NULL) {
        return -1; /* malformed line */
    }
    else {
        c++;
    }

    /* Skip the current value */
    while ( c && *c != '"') {
        c++;
    }
    if (c == NULL) {
        return -1; /* malformed line */
    }
    sprintf(out,"%s%s=\"%s%s", out, attr, val, c);
    strcpy(line, out);
    return 1;
}
    
int modify_xml_file(char *filename, xml_file_mod_t **mods) {
    char filename0[BIG_BUF];
    char line[BIG_BUF];
    FILE *f0, *f;
    int i;

    /* Create temporary file with the current filename value */
    sprintf(filename0, "%s.0", filename);
    if (rename(filename, filename0) != 0) {
        fatal_error("Can't rename %s", filename, 0);
    }

   f0 = fopen (filename0, "r");
   if (f0 == NULL) {
       fatal_error("Can not open %s", filename0, 0);
   }
    
   /* The output file */
   f = fopen (filename, "w");
   if (f == NULL) {
       fatal_error("Can not open %s for writing", filename, 0);
   }

   while (fgets(line, BIG_BUF, f0)) {
       for (i=0; mods[i]; i++) {
           if (set_xml_attr(line, mods[i]->section, mods[i]->attr, mods[i]->val) < 0) {
               fatal_error("%s has invalid format", filename, 0);
           }
       }
       fprintf(f, "%s", line);
   }

   fclose(f);
  
   /* Remove temporary file */
   fclose(f0);
   remove(filename0);

   return 0;
}
       
/* -----------------------         main            ------------------------ */

main(int argc, char *argv[])
{
   char hn[SML_BUF];
   const char *sroot;
   struct stat fi;
   int opt;

   while ((opt = getopt(argc, argv, "sl:f:r")) != -1)
   {
      switch (opt)
	 {
            case 'r':
               reConfig = True;
               break;
	    case 's':
	       installMode = Silent;
	       break;
	    case 'l':
	       logFile = strdup(optarg);	/* Log file to use */
	       break;
	    case 'f':
	       infoFile = strdup(optarg);	/* Install script */
               installInfo = new InstallInfo(infoFile);
	       break;
	    default:
	       break;
	 }
   }

   /*
    * ns-update always expects the -f option.
    */
   if (installMode == Silent)
   {
      installLog = new InstallLog(logFile);
      installLog->logMessage(START, "Admin", "Administration Server post-configuration");
   }

   if (infoFile == NULL || installInfo == NULL ||
       installInfo->isEmpty() == True ||
       (adminInfo = installInfo->getSection("admin")) == NULL)
   {
      printf("ERROR: Answer cache not found or invalid\n");
      exit(1);
   }

   if (installMode == Silent)
   {
      if (infoFile == NULL || (stat(infoFile, &fi) == -1))
      {
	 installLog->logMessage(FATAL, "Admin", "Installation script not found");
	 printf("ERROR:Installation script not found.\n");
	 exit(1);
      }
      else
      {
         // PVO - why ???
	 installInfo = new InstallInfo (infoFile);
      }

   }

   // Some additional checking in case
   if (installInfo->get(MACHINE_NAME) == NULL ||
       installInfo->get(SERVER_ROOT) == NULL ||
       installInfo->get(CONFIG_LDAP_URL) == NULL ||
       installInfo->get(MC_ADMIN_ID) == NULL ||
       installInfo->get(MC_ADMIN_PWD) == NULL ||
       installInfo->get(DS_ADMIN_DOMAIN)==NULL)
   {
      installLog->logMessage(FATAL, "Admin", "Missing data from answer cache");
      printf("ERROR: missing data from answer cache\n");
      exit(1);
   }

   // Add config DN (o=NetscapeRoot) to CONFIG_LDAP_URL
   installInfo->set(CONFIG_LDAP_URL, formConfigLdapURL(installInfo->get(CONFIG_LDAP_URL)));

   installInfo->toUTF8(CONFIG_LDAP_URL);
   installInfo->toUTF8(MC_ADMIN_ID);
   installInfo->toUTF8(DS_ADMIN_DOMAIN);

   installLog->logMessage(INFO, "Admin", "Configuring Administration server...");
   printf("Configuring Administration Server...\n");

   hn[0] = '\0';

   if (stat(DEFAULT_ADMINCONF, &fi) == 0)
   {
      // PVO - already configured
      // Probably should check DS instead
   }

   sroot = installInfo->get(SERVER_ROOT);

   /*
    * Get the full hostname.
    */
   strcpy(hn,installInfo->get(MACHINE_NAME));

   cleanup_after_upgrade(sroot);

   create_admserv(hn, sroot);

#ifdef NEED_ADMIN_SERVER_IN_SERVERS_LIST
   addToServerList(sroot);
#endif

   start_admserv(sroot);

   installLog->logMessage(SUCCESS, "Admin", "Administration Server post-configuration");

   return 0;
}
