/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * adm-fn: Defines the administration server functions
 * 
 * Rob McCool
 */

#define ADMIN_HASH_SIZE 2

#include "base/pblock.h"
#include "base/session.h"
#include "base/ereport.h"
#include "frame/req.h"
#include "frame/func.h"
#include "frame/conf.h"
#include <ldap.h>
#include <libaccess/acl.h>     /* ACL_FAIL and friends */
#include <libaccess/las.h>     /* LAS_EVAL_FAIL and friends */
#include <libaccess/aclproto.h>
#include "libaccess/ldapacl.h"
#include "libaccess/nsuser.h"
#include "safs/auth.h"       /* get_authorization_basic */
#include "dbtadmserv.h"
#include "prlock.h"

/* DT - 4.0 task acl eval stuff */
#include "taskacl.h"

#ifdef XP_WIN32
#include "frame/http.h"
#define S_IXUSR 00100 /* execute permission: owner */
#define S_IXGRP 00010 /* execute permission: group */
#define S_IXOTH 00001 /* execute permission: other */
#endif /* XP_WIN32 */

#if !defined (PROP_REQUEST)
#define	PROP_REQUEST	"request"
#endif

/* Function table for library code */

/* Imports */
Func auth_basic1;
Func pcheck_require_auth;
Func simple_user;
Func pcheck_uri_clean;
#ifdef XP_WIN32
Func pcheck_ntcgicheck;
#endif /* XP_WIN32 */
Func pcheck_find_path;
Func pcheck_find_index;
Func cgi_send;
Func service_plain_file;
Func service_imagemap;
Func clf_record;
Func pcheck_check_acl;
Func ntrans_assign_name;
Func ntrans_redirect;

/* Admin only */
Func admin_check_admpw;
Func admin_uri2path;
Func admin_ftype;
Func admin_error;
Func admin40_authenticate;
Func admin40_isdiropen;
Func admin40_host_ip_check;
Func admin40_runtime_commands;
Func admin40_log_record;

struct FuncStruct func_standard[] = {
    {"basic-auth", auth_basic1, NULL},
    {"require-auth", pcheck_require_auth, NULL},
    {"simple-userdb", simple_user, NULL},
    {"find-index", pcheck_find_index, NULL},
#ifdef XP_UNIX
    {"uri-clean", pcheck_uri_clean, NULL},
#else /* XP_WIN32 */
    {"nt-uri-clean", pcheck_uri_clean, NULL},
	{"ntcgicheck", pcheck_ntcgicheck, NULL},
#endif /* XP_WIN32 */
    {"send-cgi", cgi_send, NULL},
    {"send-file", service_plain_file, NULL},
    {"imagemap", service_imagemap, NULL},
    {"common-log", admin40_log_record, NULL},

    {"admin-check-admpw", admin_check_admpw, NULL},
    {"admin-uri2path", admin_uri2path, NULL},
    {"admin-ftype", admin_ftype, NULL},
    {"admin-error", admin_error, NULL},
    {"check-acl", pcheck_check_acl, NULL},
    {"assign-name", ntrans_assign_name, NULL},
    {"redirect", ntrans_redirect, NULL},
    {"admin40-authenticate", admin40_authenticate, NULL},
    {"admin40-isdiropen", admin40_isdiropen, NULL},
    {"admin40-host-ip-check", admin40_host_ip_check, NULL},
    {"admin40-runtime-commands", admin40_runtime_commands, NULL},
    {NULL,NULL,NULL}
};

/* Code */


#include "base/shexp.h"     /* shexp_cmp */
#include "frame/protocol.h" /* protocol_status */
#include "base/util.h"      /* sprintf */

/* From adminit */
extern char *nsroot;
extern char *admroot;
extern char *admusers;
extern char *admhome;
/* Hack.... to be fixed later  */
char *admlog = NULL;
char *backups = NULL;


/* List of the installed server types, from servers.lst. */
extern char **installed;
extern int num_installed;
extern PRLock *serverListLock;

/* -------------------------- admin40_host_ip_check ---------------------------- */

/* Check if the caller hostname or ip address is disallowed */

int admin40_host_ip_check(pblock *pb, Session *sn, Request *rq)
{
    char *admdns = pblock_findval("admdns", pb);
    char *admip  = pblock_findval("admip", pb);

    if (admdns)
    {
       char *maxdns = session_maxdns(sn);
       if (maxdns && !shexp_cmp(maxdns, admdns))
          return REQ_NOACTION;
    }

    if (admip && !shexp_cmp(pblock_findval("ip", sn->client), admip))
       return REQ_NOACTION;

    pblock_nvinsert("admin-error", XP_GetClientStr(DBT_unauthorizedHost_), rq->vars);
    return REQ_ABORTED;
}

/* -------------------------- admin40_isdiropen ---------------------------- */

/* Check if the uri is for an open directory. If so, then bypass authentication. */

int admin40_isdiropen(pblock *pb, Session *sn, Request *rq)
{
    /* In kingpin, some directories are now supported without access control.
     * So, we check for them first.
     *
     * No basic auth for:
     *
     * /
     * /java
     * /dsgw
     * /manual
     * /dist
     *
     */
    
    char *uri = pblock_findval("uri", rq->reqpb);

#   define DSGW_DIR "/dsgw/"
#   define JAVA_DIR "/java/"
#   define MANUAL_DIR "/manual/"
#   define DIST_DIR "/dist"

    if ((!STRNCASECMP(uri, DSGW_DIR, strlen(DSGW_DIR))) ||
        (!STRNCASECMP(uri, JAVA_DIR, strlen(JAVA_DIR))) ||
        (!STRNCASECMP(uri, MANUAL_DIR, strlen(MANUAL_DIR))) ||
        (!STRNCASECMP(uri, DIST_DIR, strlen(DIST_DIR))) ||
        ((uri[0] == '/') && (uri[1] == '\0')))
    {
       pblock_nvinsert("ADM_AuthNotRequired", "true", rq->vars);
       return REQ_PROCEED;
    }

    return REQ_NOACTION;
}

/* -------------------------- admin_check_admpw ---------------------------- */

/*
 * Check the admpw file, if specified, along with the host and ip restrictions.
 * 
 * If they match, then this user is the super user, so treat them accordingly.
 * If they do not match, then allow the ACL code to deal with them.
 */
int admin_check_admpw(pblock *pb, Session *sn, Request *rq)
{
    pblock *execpb  = NULL;
    int ret         = 0;
    char *auth      = NULL;
    char *admpwfile = pblock_findval("admpwfile", pb);

    if (!admpwfile)
       return REQ_ABORTED;

    /* First, check to see if they authenticated at all. */
    if(request_header("authorization", &auth, sn, rq) == REQ_ABORTED)
       return REQ_ABORTED;

    /* Run:
     * AuthTrans fn="basic-auth" auth-type="basic" userfile=admpwfile */
    execpb = pblock_create(ADMIN_HASH_SIZE);
    pblock_nvinsert("fn", "basic-auth", execpb); 
    pblock_nvinsert("auth-type", "basic", execpb); 
    pblock_nvinsert("userfile", admpwfile, execpb); 
    ret = func_exec(execpb, sn, rq);
    pblock_free(execpb);

    if (ret == REQ_ABORTED)
      return admin40_error(sn, rq, PROTOCOL_UNAUTHORIZED, "invalid user credentials"); /*i18n*/

    if (ret == REQ_NOACTION)
    {
       /* Call require-auth to enforce admin user name. */
       int ret;
       pblock *execpb = pblock_create(ADMIN_HASH_SIZE);
       pblock_nvinsert("fn", "require-auth", execpb);
       pblock_nvinsert("auth-type", "basic", execpb);
       /* We don't really know what the user name is, so just list
        * a nonsense value. */
       pblock_nvinsert("auth-user", "NS-no-such-user", execpb);
       pblock_nvinsert("realm", MAGNUS_SERVER_NAME, execpb);
       ret = func_exec(execpb, sn, rq);
       pblock_free(execpb);
       return ret;
    }

    pblock_nvinsert("ADM_UserLevel", LOCAL_SUPER_NAME, rq->vars);
    return REQ_PROCEED;
}



/* ---------------------------- admin40_redirect ----------------------------- */

/*
 * Perform a redirect function exec. DT 12/12/97.
 */

static int
admin40_redirect(Session *sn, Request *rq, char *from, char *url)
{
   pblock *execpb = NULL;
   int     ret    = 0;

   execpb = pblock_create(ADMIN_HASH_SIZE);
   pblock_nvinsert("fn", "redirect", execpb);
   pblock_nvinsert("from", from, execpb);
   pblock_nvinsert("url", url, execpb);
   ret = func_exec(execpb, sn, rq);
   pblock_free(execpb);
   return ret;
}


/* ------------------------ admin40_runtime_commands ----------------------- */

/*
 * DT 3/15/98
 *
 * We need a way to send commands to a running admin server without simply
 * restarting the server (e.g. to resync to pick up new SIEs). The
 * /admin-serv/commands/* url namespace is to be used for this purpose.
 */

int admin40_runtime_commands(pblock *pb, Session *sn, Request *rq)
{
    char *name  = pblock_findval("command-name", rq->reqpb);
    char *query = pblock_findval("query", rq->reqpb);

    if (admin40_runtime_command_exec(name, query, rq))
    {
       pblock_nvinsert("content-length", "0", rq->srvhdrs);
       protocol_status(sn, rq, PROTOCOL_OK, NULL);
       protocol_start_response(sn, rq);
       return REQ_PROCEED;
    }

    return REQ_ABORTED;
}


/* ---------------------------- admin_uri2path ----------------------------- */

/*
 * Turns the requested URI into a path.
 */

int admin_uri2path(pblock *pb, Session *sn, Request *rq)
{
    char *uri = pblock_findval("uri", rq->reqpb);
    char *root = pblock_findval("root", pb);
#   define DSGW_ATTRIBUTE "dsgw"
#   define ENDUSER_ATTRIBUTE "enduser"
    char *disableDSGW = pblock_findval(DSGW_ATTRIBUTE, pb);
    char *disableENDUSER = pblock_findval(ENDUSER_ATTRIBUTE, pb);
    pb_param *path = pblock_find("ppath", rq->vars);
    char *t, *u, *uu, *ntrsubdir = NULL;
    int pass = 0;
    char *servid, *subdir, *pinfo;
    struct stat *finfo;

    /*
     * Stop robots from walking on unprotected admin servers
     */
    if(!STRNCASECMP(uri, "/robots.txt", strlen("/robots.txt")))  {
        /* See admin_error for the nitty gritty in this case. */
        pblock_nvinsert("admin-error", "robots.txt", rq->vars);
        return REQ_ABORTED;
    }

    /*
     * DT 12/01/97 - Map / to the applet index page, disabling direct
     * link to the 3.0 index cgi.
     * DT 4/27/98 - Map / to the download CGI for the AS home page.
     */
#   define DIST_CGI_PATH "/dist/download"
    if ((uri[0] == '/') && (uri[1] == '\0'))
       return admin40_redirect(sn, rq, "/", DIST_CGI_PATH);

    /* DT 11/11/97 Support for $SERVER_ROOT/java directory */
    if (!STRNCASECMP(uri, JAVA_DIR, strlen(JAVA_DIR)))
    {
       if (strlen(uri) == strlen(JAVA_DIR))
	  goto bong; /* can't request the directory */

       t = (char *) MALLOC(strlen(nsroot) + strlen(uri) + 1);
       util_sprintf(t, "%s%s", nsroot, uri);
       FREE(path->value);
       path->value = t;
       return REQ_PROCEED;
    }

    /* DT 3/4/98 Support for $SERVER_ROOT/dist directory */
#   define DOWNLOAD_CONTENT_TYPE "application/octet-stream"
    if (!STRNCASECMP(uri, DIST_DIR, strlen(DIST_DIR)))
    {
       if (strlen(uri) == strlen(DIST_DIR))
	  goto bong; /* can't request the directory */
       
       if (!STRCASECMP(uri, DIST_CGI_PATH))
          pblock_nvinsert("content-type", "magnus-internal/cgi", rq->srvhdrs);
       else
          pblock_nvinsert("content-type", DOWNLOAD_CONTENT_TYPE, rq->srvhdrs);

       t = (char *) MALLOC(strlen(nsroot) + strlen(uri) + 1);
       util_sprintf(t, "%s%s", nsroot, uri);
       FREE(path->value);
       path->value = t;
       return REQ_PROCEED;
    }

    /* DT 12/01/97 Support for $SERVER_ROOT/dsgw directory */
    if (!STRNCASECMP(uri, DSGW_DIR, strlen(DSGW_DIR)))
    {
       if (disableDSGW) {
           FREE(u);
           pblock_nvinsert("admin-error", XP_GetClientStr(DBT_DSGWAccessDisabled_), rq->vars);
           return REQ_ABORTED;
       }

       t = (char *) MALLOC(strlen(nsroot) + strlen(uri) + 1);
       util_sprintf(t, "%s%s", nsroot, uri);
       FREE(path->value);
       path->value = t;
       
       /* if execute bit is set, then set the cgi content-type */
#      define IS_EXECUTABLE(mode) (((mode)&S_IXUSR)||((mode)&S_IXGRP)||((mode)&S_IXOTH))
       if ((finfo = request_stat_path(path->value, rq)) != NULL)
       {
          if (IS_EXECUTABLE(finfo->st_mode))
             pblock_nvinsert("content-type", "magnus-internal/cgi", rq->srvhdrs);
       }
       /* what should we do if fstat fails? will a path check handle this? */

       return REQ_PROCEED;
    }

    /* DT 11/11/97 Support for $SERVER_ROOT/manual directory */
#   define HELP_CGI_PATH   "/manual/help/help"
#   define HELPSRCH_CGI_PATH   "/manual/help/helpsrch"
    if (!STRNCASECMP(uri, MANUAL_DIR, strlen(MANUAL_DIR)))
    {
       if (strlen(uri) == strlen(MANUAL_DIR))
	  goto bong; /* can't request the directory */

       if (!STRCASECMP(uri, HELP_CGI_PATH) || !STRCASECMP(uri, HELPSRCH_CGI_PATH))
          pblock_nvinsert("content-type", "magnus-internal/cgi", rq->srvhdrs);

       t = (char *) MALLOC(strlen(nsroot) + strlen(uri) + 1);
       util_sprintf(t, "%s%s", nsroot, uri);
       FREE(path->value);
       path->value = t;
       return REQ_PROCEED;
    }

    /* DT 12/1/97 Support for http://host:port/user redirect, and html directory */
    /* DT  1/6/98 Support for bin directory as well                              */
#   define ENDUSER_URI   "/user"
#   define ENDUSER_PATH  "/bin/user/admin/bin/enduser"
#   define ENDUSER_BASE  "/bin/user/admin/"
#   define ENDUSER_HTML  "html/"
#   define ENDUSER_BIN   "bin/"
#   define ENDUSER_GIFS  "gifs/"
    if (!STRCASECMP(uri, ENDUSER_URI)) {
       if (disableENDUSER) {
           FREE(u);
           pblock_nvinsert("admin-error", XP_GetClientStr(DBT_enduserAccessDisabled_), rq->vars);
           return REQ_ABORTED;
       }
       return admin40_redirect(sn, rq, ENDUSER_URI, ENDUSER_PATH);
    }
    if (!STRNCASECMP(uri, ENDUSER_BASE, strlen(ENDUSER_BASE)))
    {
       if (disableENDUSER) {
           FREE(u);
           pblock_nvinsert("admin-error", XP_GetClientStr(DBT_enduserAccessDisabled_), rq->vars);
           return REQ_ABORTED;
       }
       if (!STRNCASECMP(uri+strlen(ENDUSER_BASE), ENDUSER_BIN, strlen(ENDUSER_BIN)))
          pblock_nvinsert("content-type", "magnus-internal/cgi", rq->srvhdrs);
       else if ((STRNCASECMP(uri+strlen(ENDUSER_BASE), ENDUSER_HTML, strlen(ENDUSER_HTML))) && (STRNCASECMP(uri+strlen(ENDUSER_BASE), ENDUSER_GIFS, strlen(ENDUSER_GIFS))))
	  goto bong; /* only allow html gifs and bin directory contents */
       
       t = (char *) MALLOC(strlen(nsroot) + strlen(uri) + 1);
       util_sprintf(t, "%s%s", nsroot, uri);
       FREE(path->value);
       path->value = t;
       return REQ_PROCEED;
    }

    if (!STRNCASECMP(uri, "/admin-serv/tasks/icons", strlen("/admin-serv/tasks/icons"))) {
      char *icons;

      icons = (char *)MALLOC(strlen(nsroot) + 1 + 3 + 1 + 5 + 1 + 5 + 1 + 5 + 1 + strlen(strrchr(uri, '/') + 1) + 1);

      util_sprintf(icons, "%s%cbin%cadmin%cadmin%cicons%c%s", 
		   nsroot,
		   FILE_PATHSEP,
		   FILE_PATHSEP,
		   FILE_PATHSEP,
		   FILE_PATHSEP,
		   FILE_PATHSEP,
		   (strrchr(uri, '/') + 1));
      FREE(path->value);
      path->value = icons;

      pblock_nvinsert("content-type", "image/gif", rq->srvhdrs);

      return REQ_PROCEED;
    }

    /* Munge the URL and break it into serverid and subdir. */
    u = STRDUP(path->value);
    if((u[0]) != '/')
        goto bong;
    servid = u+1;
    if(!(subdir = strchr(servid, '/')))
        goto bong;
    *(subdir++) = '\0';
    
    /* set product in rq->vars, or everything breaks. DT 12/12/97 */
    pblock_nvinsert("product", servid, rq->vars);
   
    /* DT 4/8/98 Authentication is no longer a task. Hardcoded to userauth. */
#   define AUTH_IDENTIFIER "authenticate"
#   define AUTH_CGI        "userauth"
    if (!STRNCASECMP(subdir, AUTH_IDENTIFIER, strlen(AUTH_IDENTIFIER)))
    {
       char *authcgi;
 
       /* authcgi = ${SERVERROOT}/bin/admin/admin/bin/${execRef} */
       authcgi = (char *)MALLOC(strlen(nsroot)+1+3+1+5+1+5+1+3+1+strlen(AUTH_CGI)+1);
       sprintf(authcgi, "%s%cbin%cadmin%cadmin%cbin%c%s", nsroot, FILE_PATHSEP, FILE_PATHSEP, FILE_PATHSEP, FILE_PATHSEP, FILE_PATHSEP, AUTH_CGI);
 
       FREE(path->value);
       path->value = authcgi;
 
       pblock_nvinsert("content-type", "magnus-internal/cgi", rq->srvhdrs);
 
       return REQ_PROCEED;
    }
 
    /* DT 3/15/98 Runtime command support. */
#   define RUNTIME_COMMAND_BASE "commands/"
    if (!STRNCASECMP(subdir, RUNTIME_COMMAND_BASE, strlen(RUNTIME_COMMAND_BASE)))
    {
       char *name;

       if (admin40_task_access_eval(pb, sn, rq) != REQ_PROCEED)
	  goto bong;

       name = subdir + strlen(RUNTIME_COMMAND_BASE);
       pblock_nvinsert("command-name", name, rq->reqpb);
       pblock_nvinsert("content-type", RUNTIME_COMMAND_CONTENT_TYPE, rq->srvhdrs);
       return REQ_PROCEED;
    }

    /* DT - 2/2/98
     * New task uri2path code
     */
    if (!STRNCASECMP(subdir, TASK_IDENTIFIER, strlen(TASK_IDENTIFIER)))
    {
       char *truepath;
       char *execRef;
       char *execRefArgs;

       if (admin40_task_access_eval(pb, sn, rq) != REQ_PROCEED)
	  goto bong;

       execRef     = pblock_findval("execRef", pb);
       execRefArgs = pblock_findval("execRefArgs", pb);

       if ((t = strchr(servid, '-')))
	  *t = '\0';
       
       /* truepath = ${SERVERROOT}/bin/[server-type]/admin/bin/${execRef} */
       truepath = (char *)MALLOC(strlen(nsroot)+1+3+1+strlen(servid)+1+5+1+3+1+strlen(execRef)+1);
       sprintf(truepath, "%s%cbin%c%s%cadmin%cbin%c%s", nsroot, FILE_PATHSEP, FILE_PATHSEP, servid, FILE_PATHSEP, FILE_PATHSEP, FILE_PATHSEP, execRef);
    
       FREE(path->value);
       path->value = truepath;
    
       pblock_nvinsert("content-type", "magnus-internal/cgi", rq->srvhdrs);
    
       /* Make sure it's not a directory, Mike's way. */
       finfo = request_stat_path(path->value, rq);
       if (finfo)
       {
          if(S_ISDIR(finfo->st_mode))
          {
             pblock_nvinsert("admin-error", "Invalid URL - is a directory", rq->vars);
             return REQ_ABORTED;
          }
       }
    
       /* Append execRefArgs to query, if applicable. */
       if (execRefArgs != NULL)
       {
          char *new_query;
          pb_param *query = pblock_find("query", rq->reqpb);
    
          if (query != NULL)
          {
	     new_query = (char*)MALLOC(strlen(query->value) + 1 + strlen(execRefArgs) + 1);
	     sprintf(new_query, "%s&%s", execRefArgs, query->value);
             FREE(query->value);
             query->value = new_query;
          }
          else
          {
	     new_query = STRDUP(execRefArgs);
             pblock_nvinsert("query", new_query, rq->reqpb);
          }
       }
    
       return REQ_PROCEED;
    }
    
    bong:
    FREE(u);
    pblock_nvinsert("admin-error", XP_GetClientStr(DBT_invalidUrl_), rq->vars);
    return REQ_ABORTED;
}


/* ----------------------------- admin_ftype ------------------------------ */


/*
 * Allows a very fixed subset of types. Currently only JPEG, GIF, 
 * imagemap, and HTML. If a file is CGI, its type is set by uri2path.
 *
 * DT 11/11/97 Added support for .properties files.
 * DT 03/04/98 Added support for application/octet-stream download types,
 */

int admin_ftype(pblock *pb, Session *sn, Request *rq)
{
    char *path    = pblock_findval("path", rq->vars);
    char *curtype = pblock_findval("content-type", rq->srvhdrs);
    char *type    = NULL;
    char *t, *u;

    /* If the current content-type is already set to DOWNLOAD_CONTENT_TYPE
     * or RUNTIME_COMMAND_CONTENT_TYPE, don't override it
     */
    if (curtype && (!STRCMP(curtype, DOWNLOAD_CONTENT_TYPE) || !STRCMP(curtype, RUNTIME_COMMAND_CONTENT_TYPE)))
       return REQ_PROCEED;

    /* Get an extension (if there is one) first.  Can't trust the 
     * type set by the bin directory earlier because now there's 
     * class files in the bin directory too. */
    if((!(u = strrchr(path, FILE_PATHSEP))) || (!(t = strrchr(u, '.')))) {
        /* No extension?  If there's already a type, this must be the 
         * bin directory, proceed anyway. */
        if(curtype)  {
            return REQ_PROCEED;
        }  else  {
        /* Else you're a loser, get an extension. */
            pblock_nvinsert("admin-error",
                            XP_GetClientStr(DBT_noFilenameExtension_),
                            rq->vars);
            return REQ_ABORTED;
        }
    }
    ++t;

    if((!strcasecmp(t, "exe")) || (!strcasecmp(t, "cgi")))
        type = "magnus-internal/cgi";
    else if((!strcasecmp(t, "jpg")) || (!strcasecmp(t, "jpeg")))
        type = "image/jpeg";
    else if(!strcasecmp(t, "gif"))
        type = "image/gif";
    else if((!strcasecmp(t, "html")) || (!strcasecmp(t, "htm")))
        type = "text/html";
    else if(!strcasecmp(t, "map"))
        type = "magnus-internal/imagemap";
    else if((!strcasecmp(t, "js")) || (!strcasecmp(t, "mocha"))
	    || (!strcasecmp(t,"jss")))
        type = "application/x-javascript";
    else if((!strcasecmp(t, "class")) || (!strcasecmp(t, "zip")))
        type = "application/java-vm";
    else if(!strcasecmp(t, "properties"))
        type = "application/java-vm";
    else if(!strcasecmp(t, "ser"))
        type = "application/java-serialized-object";
    else if(!strcasecmp(t, "jar"))
        type = "application/java-archive";
    else {
        if(curtype)  {
            /* Must be a CGI from the bin directory.  
             * Don't override existing type. */
            return REQ_PROCEED;
        }  else  {
            pblock_nvinsert("admin-error",
                            XP_GetClientStr(DBT_unrecognizedType_),
                            rq->vars);
            return REQ_ABORTED;
        }
    }
    if(curtype)  {
        param_free(pblock_remove("content-type", rq->srvhdrs));
    }  
    pblock_nvinsert("content-type", type, rq->srvhdrs);
    return REQ_PROCEED;
}


/* ----------------------------- admin_error ------------------------------ */


/*
 * Sends custom error messages depending on the error status 
 */

int admin_error(pblock *pb, Session *sn, Request *rq)
{
    char *reason, buf[1024];
    int l;

    reason=pblock_findval("admin-error", rq->vars);

    if((reason) && (!strcmp(reason, "robots.txt")))  {
        /* Return: 
         *   # noone nohow
         *   User-Agent: *
         *   Disallow: /
         * Upon receipt of such a request.  
         */
        protocol_status(sn, rq, PROTOCOL_OK, NULL);
        protocol_start_response(sn, rq);
        l = util_sprintf(buf, "User-Agent: *\n"
                              "Disallow: /\n");
        net_write(sn->csd, buf, l);
        return REQ_PROCEED;
    }

    protocol_start_response(sn, rq);
    l = util_sprintf(buf,
                     XP_GetClientStr(DBT_TitleInternalErrorTitleNH1Intern_));
    if(reason)  {
        l = util_sprintf(buf, "%s%s <b>%s</b><p>\n",
                         buf, XP_GetClientStr(DBT_reason_), reason);
    }

    net_write(sn->csd, buf, l);
    return REQ_PROCEED;
}


/* ---------------------------- admin_cgivars ----------------------------- */

#define MAX_ADMIN_VARS 17
/* The current environment vars for admin programs:
 *    1. ADMSERV_PID - the pid of the currently running admin server
 *    2. ADMSERV_ROOT - the root of the admin server 
 *                      ([ServerRoot]/admin-serv/config)
 *    3. NETSITE_ROOT - the root of the current server instance
 *    4. CONFIG_DIR - the template for where the working changes should be made
 *    5. ADMIN_LOG - the full pathname of the verbose admin log
 *    6. COMMIT_LOG - the full pathname to the verbose log of commit stuff
 *    7. BACKUPS - the number of config file backups to keep
 *    8. SERVER_NAMES - a list of the name(s) of the servers currently 
 *                      being configured
 *    9. PASSWORD_PIPE - the location of the pipe to read password information
 *                       from
 *   10. PROGRAM_PERMISSIONS - a wildcard pattern of programs this user is
 *                             allowed to access.
 *   11. TEMP - Windows TEMP dir
 *   12. AUTH_USERTYPE - type of user that got authenticated
 *        (admin, postmaster, or mailuser)  -- lachman
 *   13. ClientLanguage  - Client Language from configuration file
 *   14. AdminLanguage   - Admin Language from configuration file
 *   15. DefaultLanguage - Default Language from configuration file
 *   16  SCRIPT_PATH - the full pathname of the script being executed.
 */

char **admin_cgivars(Session *sn, Request *rq, pid_t ppid, SYS_FILE pwpipe, char **env)
{
    char *t, *u, digit[16];
    int x;

    t = (char *) NULL;
	u = (char *) NULL;
    env = util_env_create(env, MAX_ADMIN_VARS, &x);

    util_sprintf(digit, "%d", ppid);
    env[x++] = util_env_str("ADMSERV_PID", digit);
    env[x++] = util_env_str("ADMSERV_ROOT", admroot);
    env[x++] = util_env_str("NETSITE_ROOT", nsroot);

#   ifdef USE_OLD_30_CODE
    if( (t = pblock_findval("product", rq->vars)) )
#   endif
    {
        char *z;

	/* This is hack to avoid "product was not set" */
        if (t) {
	  env[x++] = util_env_str("SERVER_NAMES", t);

	  u = (char *) MALLOC(strlen(nsroot) + 1 + strlen(t) + 1 + 7 + 1 + 2);
	  util_sprintf(u, "%s%c%%s%cconf_bk%c", nsroot, FILE_PATHSEP, 
		       FILE_PATHSEP, FILE_PATHSEP); 
	  env[x++] = util_env_str("CONFIG_DIR", u);
	  FREE(u);
	}

        if(admlog)  {
            u = STRDUP(admlog);
            env[x++] = util_env_str("ADMIN_LOG", u);
            FREE(u);
        }

        u = (char *) MALLOC(strlen(admroot) + 1 + 6 + 2);
        util_sprintf(u, "%s%ccommit",
                        admroot, FILE_PATHSEP);
        env[x++] = util_env_str("COMMIT_LOG", u);
        FREE(u);
 
        if(backups)  {
            u = STRDUP(backups);
            env[x++] = util_env_str("BACKUPS", u);
            FREE(u);
        }
#ifdef XP_WIN32
		if (u = getenv("TEMP")) {
			env[x++] = util_env_str("TEMP", u);
		}
#endif /* XP_WIN32 */
        /* MLM - send password pipe number in environment */
        u = (char *) MALLOC(64);
#ifdef XP_UNIX
        sprintf(u, "%d", PR_FileDesc2NativeHandle(pwpipe));
#else /* XP_WIN32 */
        sprintf(u, "%u", PR_FileDesc2NativeHandle(pwpipe));
#endif
        env[x++] = util_env_str("PASSWORD_PIPE", u);
        FREE(u);

        u = pblock_findval("ADM_Programs", rq->vars);
        if(u)  {
	  char	*isMulti;

	  u[0] = '(';
	  if ( ( isMulti = strchr( u, '|' ) ) ) {	/* multiple choice */
	    strcat( u, ")" );	/* extra space pre-allocated */
	  }
	  env[x++] = util_env_str("PROGRAM_PERMISSIONS", isMulti ? u : u + 1);
        }
    }

    util_sprintf(digit, "admin");
    env[x++] = util_env_str("AUTH_USERTYPE", digit);

    env[x++] = util_env_str("ClientLanguage", GetClientLanguage());
    env[x++] = util_env_str("AdminLanguage",  GetAdminLanguage());
    env[x++] = util_env_str("DefaultLanguage",GetDefaultLanguage());

    if ((t = pblock_findval("path", rq->vars)))
       env[x++] = util_env_str("SCRIPT_PATH", t);

    if ((t = pblock_findval(AUTHENTICATION_LDAP_URL, rq->vars)))
       env[x++] = util_env_str(AUTHENTICATION_LDAP_URL, t);

    
    if ((t = pblock_findval("auth-password-expire", rq->vars)))
       env[x++] = util_env_str("PASSWORD_EXPIRING", t);

    env[x] = NULL;
    return env;
}

/* ----------------------------- admin40_log_record ----------------------- */

/*
 * Allows for some custom logging handling.
 *
 * DT 3/20/98 Support for tasks' supressed logging.
 */

int
admin40_log_record(pblock *pb, Session *sn, Request *rq)
{
   if (pblock_findval("logSuppress", rq->reqpb))
      return REQ_PROCEED;

   return (clf_record(pb, sn, rq));
}
