/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * cron.c: Managed execution of batch jobs - ie UNIX cron.
 *
 *
 * Chuck Neerdaels
 * Copyright (c) 1996 Netscape Communcations Corporation
 *
 * modified and debugged by Alex Feygin
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <pwd.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>

#include "netsite.h"
#include "frame/req.h"
#include "base/ereport.h"
#include "base/daemon.h"

#include <errno.h>
#ifndef LINUX
extern char *sys_errlist[];
#endif
#undef system_errmsg
#define system_errmsg() (sys_errlist[errno]) /* from base/file.h */

extern char *optarg;
extern int optind;

#ifdef	INPROC_CRON
extern char *admroot;
extern char *nsroot;
#else	/* !INPROC_CRON */
char *admroot = (char *) NULL;
char *nsroot  = (char *) NULL;
#endif	/* INPROC_CRON */

#define	ADMSERV_DIR	   "admin-serv"
#define	ADMSERV_LOGDIR	   "admin-serv/logs"
#define	PIDFILE		   "pid.cron"

#define CRON_DEBUG          0

#define FALSE               0
#define TRUE                1
#define CRON_INITIAL_SLEEP  15
#define MAX_ARGS            20

#undef	OPEN_MAX
#define OPEN_MAX            256

#define BUF_SIZE            4096


static char *Days[] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};


typedef enum {
  ERROR = -1,
  OBJEND,
  COMMAND,
  USER,
  RDIR,
  TIME,
  DAYS
}CronDirective;
 
typedef struct CronObj {
  char *name;
  char *command;
  char *user;
  char *cwd;
  time_t start_t;
  time_t end_t;
  int days;                /* 7 bits used to turn on/off 7 day operation */
  int end_days;
} CronObj;
 
typedef struct CronObjList {
  time_t from_sun;
  CronObj *obj;
  pid_t current_pid;
  struct CronObjList *next;
  struct CronObjList *twin;
  int startstop;  /* 1 = start, 0 = stop */
} CronObjList;

typedef struct CronConf {
  char *conffile;
  char *status;
  char *dir;
  char *timeout;
} CronConf;

#ifndef MAX
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif
 
#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif


pid_t cron_pid = (pid_t)-1;
CronConf cronconf;

static void sig_chld_handler(int);
static void sig_term_handler(int);


/********** process list functions ************************/
typedef struct PidList {
  pid_t pid;
  CronObjList *cronobjlist;
  struct PidList *next;
} PidList;

static PidList *pidlisthead = NULL;
static PidList *pidlisttail = NULL;

static void Add_PidList(CronObjList *cronobjlist)
{
  PidList *t;

  t = (PidList*)MALLOC(sizeof(PidList));
  t->cronobjlist = cronobjlist;
  t->pid  = cronobjlist->current_pid;
  t->next = NULL;

  if (!pidlisthead) /* first in list */
    {
      pidlisthead = pidlisttail = t;
    }
  else /* other elements in list, put last */
    {
      pidlisttail->next = t;      
      pidlisttail = t;
    }
}

static int Find_PidList(pid_t pid)
{
  PidList *t = pidlisthead;
  int rval = 0;

  while(t)
    {
      if (t->pid == pid) { rval = 1; break; }
      t = t->next;
    }

  return rval;
}


#if CRON_DEBUG
static void Dump_PidList(FILE *fp)
{
  PidList *t = pidlisthead;

  fprintf(fp, "PIDLIST DUMP:\n");

  while(t)
    {
      fprintf(fp, "%d \n", (int)t->pid);
      t = t->next;
    }
}
#endif

static void Kill_PidList()
{
  PidList *t = pidlisthead;
  PidList *nt = NULL;

  while(t)
    {
      nt = t->next;

      kill(t->pid, SIGKILL);

      t = nt;
    }
}

static int Del_PidList(pid_t pid)
{
  PidList *t = NULL;
  PidList *p = NULL;
  int rval = 0;

  if (!pidlisthead)
    return 0;

  if (pidlisthead->pid == pid)
    {
      t = pidlisthead;

      if (pidlisthead == pidlisttail)
	{
	  pidlisttail = NULL;
	  pidlisthead = NULL;
	}
      else
	{
	  pidlisthead = pidlisthead->next;
	}
      
      t->cronobjlist->current_pid = 0;
      t->cronobjlist->twin->current_pid = 0;
      FREE(t);

      return 1;
    }

  p = pidlisthead;
  t = pidlisthead->next;

  while(t)
    {
      if (t->pid == pid)
	{
	  rval = 1;

	  p->next = t->next;
	  if (t == pidlisttail)
	    pidlisttail = p;

	  t->cronobjlist->current_pid = 0;
	  FREE(t);

	  break;
	}

      p = t;
      t = t->next;
    }

  return rval;
}

/********** end of process list functions *****************/



char *get_admroot()
{
  static char buf[BUF_SIZE];

  if (admroot) {
    return admroot;
  } else if ((admroot = getenv("ADMSERV_ROOT")) != (char *) NULL) {
    return admroot;
  } else {
    sprintf(buf, "%s/admin-serv/config", nsroot);
    admroot = buf;
    return buf;
  }
}

char *ctime_nocr(time_t *t)
{
  static char buf[40];
  strcpy(buf, ctime(t));
  if (buf[strlen(buf) - 1] == '\n')
    buf[strlen(buf) - 1] = '\0';

  return buf;  
}

int cron_error_init()
{
  char buf[BUF_SIZE];
  time_t now;
  FILE *fp;

  sprintf(buf, "%s/%s/cron.error", nsroot, ADMSERV_LOGDIR);

  if (!(fp = fopen(buf, "a")))
    return 0;

  sprintf(buf, "Cron initialized");

  now = time(NULL);

  fprintf(fp, "%s: %s\n", ctime_nocr(&now), buf);
 
  fclose(fp);
 
  return 1;
}

int cron_error(char *fmt, ...)
{
  va_list args;
  FILE *fp;
  char buf[BUF_SIZE];
  time_t now;

  sprintf(buf, "%s/%s/cron.error", nsroot, ADMSERV_LOGDIR);

  if (!(fp = fopen(buf, "a")))
    return 0;

  va_start(args, fmt);
  vsprintf(buf, fmt, args);
  va_end(args);

  now = time(NULL);

  fprintf(fp, "%s: %s\n", ctime_nocr(&now), buf);
  
  fclose(fp);

  return 1;
}

#if CRON_DEBUG
int cron_log_init()
{
  char buf[BUF_SIZE];
  time_t now;
  FILE *fp;

  sprintf(buf, "%s/cron.log", get_admroot());

  if (!(fp = fopen(buf, "w")))
    return 0;

  sprintf(buf, "Cron initialized");

  now = time(NULL);

  fprintf(fp, "%s: %s\n", ctime_nocr(&now), buf);
 
  fclose(fp);

  return 1;
}

int cron_log(char *fmt, ...)
{
  va_list args;
  FILE *fp;
  char buf[BUF_SIZE];
  time_t now;

  sprintf(buf, "%s/%s/cron.log", nsroot, ADMSERV_LOGDIR);

  if (!(fp = fopen(buf, "a")))
    return 0;

  va_start(args, fmt);
  vsprintf(buf, fmt, args);
  va_end(args);

  now = time(NULL);

  fprintf(fp, "%s: %s\n", ctime_nocr(&now), buf);
  
  fclose(fp);

  return 1;
}
#endif

int Read_CronConf(CronConf *cronconf, char *conf_file)
{
  FILE *cfp;
  char *p, *r;
  char  buf[BUF_SIZE];

  if (!(cfp = fopen(conf_file, "r")))
    {
      cron_error("Could not open %s", conf_file);
      return 0;
    }

  cronconf->conffile = NULL;
  cronconf->status   = NULL;
  cronconf->dir      = NULL;
  cronconf->timeout  = "23";

  while(fgets(buf, sizeof(buf), cfp))
    {
      r = strtok(buf, " \t\n");
      if (!r) /* bad line, ignore */
        continue;
 
      p = strtok(NULL, " \t\n");
      if (!p) /* bad line, ignore */
	continue;
 
      if (!strcmp(r, "ConfFile"))
        {
          cronconf->conffile = STRDUP(p);
        }
      else if (!strcmp(r, "Dir"))
        {
          cronconf->dir = STRDUP(p);
        }
      else if (!strcmp(r, "Status"))
        {
          cronconf->status = STRDUP(p);
        }
      else if (!strcmp(r, "Timeout"))
        {
          cronconf->timeout = STRDUP(p);
        }
      else /* bad line, ignore */
        continue;
    }

  fclose(cfp);  
  return 1;
}

void cron_set_uid( char *uid_name ) {
    struct passwd *pw;
 
    if( !getuid() ) {
         if (!(pw = getpwnam(uid_name))) {
	   cron_error("Could not find user \"%s\"\n", uid_name);
	   exit(1);
	 }
 
        if(setgid((gid_t)pw->pw_gid) == -1)
            cron_error("Could not set group id to %d (%s)\n",
		       (int)pw->pw_gid, system_errmsg());
 
        if(pw->pw_uid) {
            if(setuid((uid_t)pw->pw_uid) == -1)
	      cron_error("Could not set user id to %d (%s)\n",
			 (int)pw->pw_uid, system_errmsg());
        }
    }

}


int parse_command(char *cmd, char **argv)
{
  int i = 0;
  char *p;
  char cmd_cp[BUF_SIZE];
  char cmd_cp2[BUF_SIZE];
 
  strcpy(cmd_cp, cmd);

  if ((p = strstr(cmd_cp, "$NSROOT")))
    {
      /* replace $NSROOT with nsroot */
 
      *p = '\0';
      p++; /* N */
      p++; /* S */
      p++; /* R */
      p++; /* O */
      p++; /* O */
      p++; /* T */
      p++;
 
      sprintf(cmd_cp2, "%s%s%s", cmd_cp, nsroot, p);
      strcpy(cmd_cp, cmd_cp2);
    }
 
  if ((p = strstr(cmd_cp, "$ADMROOT")))
    {
      /* replace $ADMROOT with admroot */
 
      *p = '\0';
      p++; /* A */
      p++; /* D */
      p++; /* M */
      p++; /* R */
      p++; /* O */
      p++; /* O */
      p++; /* T */
      p++;
 
      sprintf(cmd_cp2, "%s%s%s", cmd_cp, get_admroot(), p);
      strcpy(cmd_cp, cmd_cp2);
    }
 
  argv[0] = "";
  argv[1] = NULL; /* in case cmd is "" */

  p = strtok(cmd_cp, " ");
 
  while(p)
    {
      if (i > (MAX_ARGS+1))
        return 0;
 
      argv[i] = strdup(p);
 
      argv[++i] = NULL;
      p = strtok(NULL, " ");
    }
 
  return 1;
}

void free_argv(char **argv)
{
  int i;
 
  for(i=0; argv[i]; i++)
    {
      free(argv[i]);
      argv[i] = NULL;
    }
}


int
util_get_day(time_t t)       /* Day of week [0-6] */
{
    struct tm *tp;
 
    if (!t)
        t = time(NULL);
 
    tp = localtime(&t);
 
    return (tp->tm_wday);
}

long
secs_since_sun( int day, time_t today_secs )
{
  return( day*86400 + today_secs );
}

long util_get_cur_secs(time_t t)
{
    struct tm *tp;

    if (!t)
	t = time(NULL);

    tp = localtime(&t);

    return (3600 * tp->tm_hour + 60 * tp->tm_min + tp->tm_sec);
}


static void
free_cron_obj( CronObj *obj )
{
  if( obj->name )
      free( obj->name );
  if( obj->command )
      free( obj->command );
  free( obj );
}

static void
cron_cleanup( CronObjList *obj_list )
{
  CronObjList *tmpobj, *oval=obj_list;

  if( !obj_list )
    return;

  /* Remember this is circularly linked */
  do {
      if( oval->obj ) 
          free_cron_obj( oval->obj );
      tmpobj = oval;
      oval = oval->next;
      free( tmpobj );
  } while( oval != obj_list );
}

static void
cron_rotate( CronObjList **obj_list_ptr )
{
 
  *obj_list_ptr = (*obj_list_ptr)->next;
}
 

#if CRON_DEBUG
static void convert_time(time_t from_sun, 
			 int *day, int *hour, int *minute, int *second)
{
  long t;

  *day = (int)((long)from_sun / 86400L); /* (60 * 60 * 24) */
  t = (long)from_sun % 86400L;

  *hour = (int)(t / 3600L);
  t = t % 3600L;

  *minute = (int)(t / 60L);
  *second = (int)(t % 60L);
}

static char *get_converted_time(time_t from_sun)
{
  static char buf[40];
  int d, h, m, s;

  convert_time(from_sun, &d, &h, &m, &s);
  sprintf(buf, "%s %d:%d:%d\n", Days[d], h, m, s);

  return buf;
}

static void cron_dump2(CronObjList *obj_list)
{
  char buf[BUF_SIZE];
  FILE *fp;
  CronObjList *list_elt = obj_list;
  int d, h, m, s;
 
  sprintf(buf, "%s/%s/cron.dump", nsroot, ADMSERV_LOGDIR);

  if (!(fp = fopen(buf, "w")))
    {
      cron_error("Couldn't open %s", buf);
      return;
    }

  if( !obj_list )
    {
      cron_error("cron_dump called with null object list");
      return;
    }

  while(list_elt->obj)
    {
      fprintf(fp, "Object \"%s\"\n", list_elt->obj->name);
      
      if (list_elt->startstop)
	{
	  fprintf(fp, "execute command \"%s\"\n", list_elt->obj->command);
	}
      else
	{
	  fprintf(fp, "kill command \"%s\"\n", list_elt->obj->command);
	}

      convert_time(list_elt->from_sun, &d, &h, &m, &s);
      fprintf(fp, "Scheduled for %s, %d:%d:%d\n\n", Days[d], h, m, s);

      list_elt = list_elt->next;
    }

  fclose(fp);
}


static void
cron_dump( CronObjList *obj_list )
{
  CronObj *cronobj;
  int i;
  char buf[BUF_SIZE];
  FILE *fp;
 
  sprintf(buf, "%s/%s/cron.dump", nsroot, ADMSERV_LOGDIR);

  if (!(fp = fopen(buf, "w")))
    {
      cron_error("Couldn't open %s", buf);
      return;
    }

  if( !obj_list )
    {
      cron_error("cron_dump called with null object list");
      return;
    }

  while( obj_list->obj )
    cron_rotate(&obj_list);

  cron_rotate(&obj_list);

  while( (cronobj = obj_list->obj) ) 
    {
      fprintf(fp, "<Object name=%s>\n", cronobj->name);
      fprintf(fp, "    Command \"%s\"\n", cronobj->command);
      fprintf(fp, "    User %s\n", cronobj->user);
      if( cronobj->cwd ) 
          fprintf(fp, "    Dir %s\n", cronobj->cwd);
 
      if( cronobj->days ) 
	{
	  fprintf(fp, "    Days ");
	  for( i=0; i< 7; i++ ) 
	    {
	      int sbit;
              if( i==0 )
		sbit = 1;
              else
		sbit = 2<<(i-1);
              if( cronobj->days & sbit )
		fprintf(fp, "%s ", Days[i]);
	    }
          fprintf(fp, "\n");
	}
 
      if( cronobj->start_t ) 
	{
          time_t h1 =  cronobj->start_t/3600L;
          time_t m1 =  (cronobj->start_t - h1*3600L)/60L;
          fprintf(fp, "    Time %2ld:%02ld\n", h1, m1);
	}

      if( cronobj->end_t ) 
	{
          time_t h1 =  cronobj->end_t/3600L;
          time_t m1 =  (cronobj->end_t - h1*3600L)/60L;
          fprintf(fp, "    End time %2ld:%02ld\n", h1, m1);
	}

      fprintf(fp, "    Seconds from sunday %ld\n", obj_list->from_sun);
      
      if (obj_list->startstop)
	fprintf(fp, "    Start this process\n");
      else
	fprintf(fp, "    Stop this process\n");
 
      printf("</Object>\n");
 
      cron_rotate(&obj_list);
    }

  fclose(fp);
}
#endif

static CronObj *
getObject( FILE *fd )
{
  int c;
  char line_in[BUFSIZ];
  char *oname=NULL;
  CronObj *newobj;
  char w_space[] = " =>\t\n";

  while( (c=fgetc(fd)) != EOF ) {
      /* Tokenize, and look for < Object name=foo > */
      if( c == '<' ) {
          fgets(line_in, BUFSIZ, fd);
          if( !strcasecmp(strtok(line_in, w_space),"object") &&
              !strcasecmp(strtok(NULL, w_space),"name") &&
              (oname = strtok(NULL, w_space)) ){
              break;
          }
      }
  }

  if( c == EOF ) 
      return NULL;

  /* Found a named object, create obj struct and return */
  newobj = (CronObj*) calloc(1,sizeof(struct CronObj));
  newobj->name = strdup(oname);
  newobj->command = NULL;

  return newobj;
}

static CronDirective
getCronDirective( FILE *fd )
{
  int c, foundstr=0, indx=0;
  char buf[BUFSIZ];
  char w_space[] = " >\t\n";
  long err_rewind=0L;

  while( ((c=fgetc(fd)) != ' ' && c != '\n') || !foundstr ) {
      if( isalpha(c) ) {
          foundstr = 1;
          buf[indx++] = c;
      }
      else if( c == '<' ) {
          fgets(buf, BUFSIZ, fd);
          err_rewind = (long)strlen(buf) + 1L;
          if( !strcasecmp(strtok(buf, w_space),"/object") )
	    return OBJEND;

          else if( !strcasecmp(strtok(buf, w_space),"object") ) {
              fprintf(stderr, "Nested batch update objects nor allowed, "
		      "assuming end of object, and trying to recover.\n");
	      cron_error("Nested batch update objects nor allowed, "
                      "assuming end of object, and trying to recover.");
              if( !fseek(fd,ftell(fd)-err_rewind,SEEK_SET) )
		return OBJEND;
              else
		return ERROR;
          }
          else {
              fprintf(stderr, "Unknown batch update directive.\n");
              cron_error("Unknown batch update directive.");
              return ERROR;
          }
      }
      else if( c == EOF ) {
          fprintf(stderr, "Unexpected batch update end of file.\n");
          cron_error("Unexpected batch update end of file.");
          return ERROR;
      }
      if( indx>BUFSIZ )
          break;
  }

  buf[indx] = '\0';

  if( !strcasecmp( buf, "command" ) )
      return COMMAND;
  if( !strcasecmp( buf, "user" ) )
      return USER;
  if( !strncasecmp( buf, "dir", 3 ) )
      return RDIR;
  else if( !strcasecmp( buf, "time" ) )
      return TIME;
  else if( !strcasecmp( buf, "days" ) )
      return DAYS;
  else 
      return ERROR;
}

static void
getcommand(FILE *fd, CronObj *objstr)
{
  char buf[BUFSIZ];
  char w_space[] = "\"\n";

  fgets(buf, BUFSIZ, fd);
  objstr->command = strdup(strtok(buf, w_space));
}

static void
getuser(FILE *fd, CronObj *objstr)
{
  char buf[BUFSIZ];
  char w_space[] = " \t\n";

  fgets(buf, BUFSIZ, fd);
  objstr->user = strdup(strtok(buf, w_space));
}

static void
getrdir(FILE *fd, CronObj *objstr)
{
  char buf[BUFSIZ];
  char w_space[] = "\"\n";

  fgets(buf, BUFSIZ, fd);
  objstr->cwd = strdup(strtok(buf, w_space));
}

static void
gettim(FILE *fd, CronObj *objstr)
{
  char buf[BUFSIZ];
  char w_space[] = " :-\t\n";
  char *h1, *m1;
  char *h2, *m2;
  
  fgets(buf, BUFSIZ, fd);
  if( (h1 = strtok(buf, w_space)) && (m1 = strtok(NULL,w_space)) ) 
    {
      objstr->start_t = atoi(h1)*3600L + atoi(m1)*60L;
      if( (h2 = strtok(NULL, w_space)) && (m2 = strtok(NULL,w_space)))
	{
	  objstr->end_t = atoi(h2)*3600L + atoi(m2)*60L;
	}
      else
	{
	  objstr->end_t =  atoi(cronconf.timeout)*3600L + objstr->start_t;
	  if ((24*3600L) < objstr->end_t)
	    objstr->end_t -= (24*3600L);
	}
#if CRON_DEBUG
      cron_log("%s: end_t = %d", objstr->name, objstr->end_t);
#endif
    }
  else
    cron_error("Cron config has bogus a Time value");
}

static void
getdays(FILE *fd, CronObj *objstr)
{
  char buf[BUFSIZ];
  char w_space[] = " ,\t\n";
  char *tok;

  fgets(buf, BUFSIZ, fd);
  tok = strtok(buf, w_space);

  while( tok ) {
      if( !strncasecmp(tok, "sun", 3) )
          objstr->days |= 1;
      else if( !strncasecmp(tok, "mon", 3) )
          objstr->days |= 2;
      else if( !strncasecmp(tok, "tue", 3) )
          objstr->days |= 4;
      else if( !strncasecmp(tok, "wed", 3) )
          objstr->days |= 8;
      else if( !strncasecmp(tok, "thu", 3) )
          objstr->days |= 16;
      else if( !strncasecmp(tok, "fri", 3) )
          objstr->days |= 32;
      else if( !strncasecmp(tok, "sat", 3) )
          objstr->days |= 64;
      else
          fprintf(stderr, "Cron config has unknown Days value\n");

      tok=strtok(NULL, w_space);
  }
}


static CronObjList *
cron_create_new_objlist(CronObj *obj, time_t from_sun, int startstop,
			CronObjList *next)
{
  CronObjList *rval;

  rval = (CronObjList*)calloc(1,sizeof(struct CronObjList));
  rval->obj = obj;
  rval->from_sun = from_sun;
  rval->startstop = startstop;
  rval->next = next;

  return rval;
}


static CronObjList *
cron_put_one_object_in_list(CronObjList **obj_list_head_p, CronObj *objstr,
			    int startstop, int cur_day) /* i == day of starting
							   process */
{
  CronObjList *rval   = NULL;
  CronObjList *oval   = NULL;
  int i;
  time_t *the_time;
  time_t from_sun;

  if (startstop)
    {
      the_time = &(objstr->start_t);
      i = cur_day;
    }
  else
    {
      the_time = &(objstr->end_t);

      if (objstr->end_t > objstr->start_t)
	i = cur_day;
      else
	{
	  if (cur_day == 6)
	    i = 0;
	  else
	    i = cur_day + 1;
	}
    }

  from_sun = secs_since_sun(i, *the_time);

  if( !(*obj_list_head_p)) /* I am the list, all bow before me */
    {
      rval = cron_create_new_objlist(objstr, from_sun, startstop,
					     NULL);
      *obj_list_head_p = rval;

#if CRON_DEBUG
      cron_log("Iteration: %d", i);
      cron_log("  Placing object in list FIRST EVER: %s, %s", 
	       objstr->name, get_converted_time(rval->from_sun));
#endif
			
    } 
  else 
    if( (*obj_list_head_p)->from_sun > from_sun ) /* New list head, as
						     I'm earlier than 
						     the head */
      {
	
	rval = cron_create_new_objlist(objstr, from_sun, startstop,
				       (*obj_list_head_p));
	*obj_list_head_p = rval;
	
#if CRON_DEBUG
	cron_log("Iteration: %d", i);
	cron_log("  Placing object in list: %s, %s", 
		 objstr->name, get_converted_time(rval->from_sun));
	cron_log("  First in list before %s, %s", 
		 rval->next->obj->name, 
		 get_converted_time(rval->next->from_sun));
#endif
      } 
  
    else  /* Insert or append onto later list */
      {
	oval = (*obj_list_head_p);
	
	while(oval->next && (oval->next->from_sun <= from_sun))
	  {
	    oval = oval->next;
	  }
	
#if CRON_DEBUG
	cron_log("About to insert something after %s, %s",
		 oval->obj->name, get_converted_time(oval->from_sun));
#endif

	if( oval->next ) /* Insert */
	  { 

	    rval = cron_create_new_objlist(objstr, from_sun, startstop,
					   oval->next);
	    oval->next = rval;

#if CRON_DEBUG
	    cron_log("Iteration: %d", i);
	    cron_log("  Placing object in list: %s, %s", 
		     objstr->name, get_converted_time(rval->from_sun));
	    cron_log("  After %s, %s", 
		     oval->obj->name, 
		     get_converted_time(oval->from_sun));
#endif
	  }
	else /* New tail, append */
	  {
	    rval = cron_create_new_objlist(objstr, from_sun, startstop,
						 NULL);
	    oval->next = rval;

#if CRON_DEBUG
	    cron_log("Iteration: %d", i);
	    cron_log("  Placing object in list: %s, %s", 
		     objstr->name, get_converted_time(rval->from_sun));
	    cron_log("  Last in list, after %s, %s",
		     oval->obj->name, 
		     get_converted_time(oval->from_sun));
#endif
	  }
      }

  return rval;
}


void
cron_put_objects_in_list(CronObjList **obj_list_head_p, CronObj *objstr)
{
  CronObjList *startObjList = NULL;
  CronObjList *stopObjList = NULL;
  int i;

  for( i=0; i< 7; i++ ) 
    {
      int sbit;

      if( i==0 ) sbit = 1;
      else	 sbit = 2<<(i-1); 

      if( (objstr->days) & sbit ) 
	{
#if CRON_DEBUG
          cron_log("startObj...");
#endif
          startObjList = cron_put_one_object_in_list(obj_list_head_p, 
						     objstr, 1, i);

#if CRON_DEBUG
          cron_log("stopObj...");
#endif
          stopObjList  = cron_put_one_object_in_list(obj_list_head_p, 
						     objstr, 0, i);
          
          if (startObjList)
            {
              startObjList->current_pid = 0;
              startObjList->twin        = stopObjList;
            }
            
          if (stopObjList)
            {
              stopObjList->current_pid = 0;
              stopObjList->twin        = startObjList;
            }
	}
    }
}


static CronObjList *
cron_read_conf( char *conf )
{

    int haveCommand       = FALSE;
    int haveTime          = FALSE;
    int haveDays          = FALSE;
    int haveUser          = FALSE;
    int haveDir           = FALSE;
    int skip_bad          = FALSE;
    CronObj *objstr       = NULL;
    CronObjList *oval     = NULL;
    CronDirective dir;
    CronObjList *obj_list = NULL;
  
    FILE *fd = fopen( conf, "r");

    while( (objstr = getObject(fd)) ) {
        haveCommand = haveUser = haveDir = FALSE;
        haveTime = haveDays = FALSE;
    
        skip_bad = FALSE;
  
        while( !skip_bad && (dir = getCronDirective(fd)) ) {

          switch(dir) {
            case ERROR:
              skip_bad = TRUE;
              fprintf(stderr, "skipping unknown directive...\n");
              break;
            case COMMAND:
              if( haveCommand )
                 fprintf(stderr,"Cron (%s) has multiple Command directives,"
                                 " ignoring this Object.\n", objstr->name );
              else 
                 getcommand(fd,objstr);
              haveCommand = TRUE;
              break;
            case USER:
              if( haveUser )
                 fprintf(stderr, "Cron (%s) has multiple User directives,"
                                 " ignoring this Object.\n", objstr->user );
              else 
                 getuser(fd,objstr);
              haveUser = TRUE;
              break;
            case RDIR:
              if( haveDir )
                 fprintf(stderr, "Cron (%s) has multiple Dir directives,"
                                 " ignoring this Object.\n", objstr->user );
              else 
                 getrdir(fd,objstr);
              haveDir = TRUE;
              break;
            case TIME:
              if( haveTime )
                fprintf(stderr, "Cron (%s) has multiple Time directives, "
                                "ignoring the later.\n", objstr->name );
              else
		{
		  gettim(fd,objstr);
		}
              haveTime = TRUE;
              break;
            case DAYS:
              if( haveDays )
                fprintf(stderr, "Cron (%s) has multiple Days directives, "
                                "ignoring the later.\n", objstr->name );
              else
                getdays(fd,objstr);
              haveDays = TRUE;
              break;
            default:
                fprintf(stderr, "Cron (%s) has an undefined directive.\n",
                                objstr->name );
          }
        }
    
        /* Look for next object if this one was bogus */
        if( !skip_bad && haveCommand && haveTime ) {
  
            if( !haveUser )
                objstr->user = strdup("nobody");
  
            if( !haveDays )
                objstr->days = 127;               /* 7 days a week */
            else if( !objstr->days )
                objstr->days = 127;
    
            objstr->start_t = MIN( objstr->start_t, 864400L );
            objstr->start_t = MAX( objstr->start_t, 0L );

	    objstr->end_t   = MIN( objstr->end_t, 864400L );
	    objstr->end_t   = MAX( objstr->end_t, 0L );

	    if (objstr->end_t > objstr->start_t)
	      objstr->end_days = objstr->days;
	    else
	      {
		objstr->end_days = objstr->days << 1;  /* shift over a day */
		
		if (objstr->days & 64) /* on saturday? */
		  objstr->end_days |= 1;		  
	      }
    
      /* At this point we have what appears to be a valid configuration,
         so we'll add it to the master CronObjList (note: sorted in ascending
         order by start time minutes since 0:00 Sunday).  Objects running
         on multiple days will be added to the list multiple times. */
	    
#if CRON_DEBUG
	  cron_log("Putting %s objects in list...", objstr->name);
#endif
	  cron_put_objects_in_list(&obj_list, objstr);
        } 
	else 
	  {
	    free_cron_obj( objstr );
	  }
      }

  /* If we found any objects, create circularly linked list, with
     a NULL object marking the tail */

    if( (oval = obj_list) ) {
        while( oval->next ) 
            oval = oval->next;

        oval->next = (CronObjList*) calloc(1,sizeof(struct CronObjList));
        oval->next->obj = NULL;
        oval->next->from_sun = 0;
        oval->next->next = obj_list;
        /* Rotate until next up is one after the NULL object (start of week) */
        while( obj_list->obj )
            cron_rotate(&obj_list);
        cron_rotate(&obj_list);
    }

    fclose( fd );

#if CRON_DEBUG
    cron_dump2(obj_list);
#endif
  
    return obj_list;
}

static time_t
cron_next_batch_sleep( CronObjList **obj_list_ptr, time_t curr_t )
{
  time_t cur_sec = util_get_cur_secs(curr_t);
  int cur_day = util_get_day(curr_t);
  CronObjList *oval = *obj_list_ptr;
  time_t sleepsecs=0, from_sun = secs_since_sun(cur_day,cur_sec);

  /* Search for next up, if we hit a null object, we're crossing a week 
     boundary, and need to add the appropriate time delta. */
  while( oval->obj && from_sun > oval->from_sun ) {
      cron_rotate(obj_list_ptr);
      oval = *obj_list_ptr;
  }
  /* Check if at null obj, cron_rotate to 1st entry of next week. */
  if( !oval->obj ) {
      cron_rotate(obj_list_ptr);
      sleepsecs = 604800L - from_sun + (*obj_list_ptr)->from_sun;
  } else {
      sleepsecs = (*obj_list_ptr)->from_sun - from_sun;
  }

  return sleepsecs;
}

void cron_terminate(void *data)
{
  CronObjList *conf_obj_list = (CronObjList*)data;

  Kill_PidList();  /* kill all child processes still out there */

  /* XXXrobm this code has a lot of problems with freeing things twice */
  /* cron_cleanup( conf_obj_list ); */

  if (cron_pid > 0) 
    {
      kill(cron_pid, SIGTERM);
      cron_pid = (pid_t)-1;
    }
  cron_error("Cron terminated");

#if CRON_DEBUG
  cron_log("Cron terminated");
#endif

}

static CronObjList *conf_obj_list_static;

void cron_terminate_static(void)
{
    cron_terminate(conf_obj_list_static);
}


void cron_main( CronObjList *conf_obj_list )
{
    time_t now;
    time_t nsecs = 0;
    time_t wake_up;
    pid_t  pid;
    CronObj *cronobj;
    char *argv[MAX_ARGS+1];
    struct sigaction act, oact;

#if CRON_DEBUG
    cron_log("In cron_main pid:%d", (int)getpid());
#endif
    sleep(CRON_INITIAL_SLEEP);	/* Allow the server to start up */

#if 0
    cron_dump(conf_obj_list);
#endif

    act.sa_handler = sig_chld_handler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
 
    if(sigaction(SIGCHLD, &act, &oact) < 0)
      cron_error("signal error.");

    act.sa_handler = sig_term_handler;
    if(sigaction(SIGTERM, &act, &oact) < 0)
      cron_error("signal error.");

    while( 1 ) 
      {
	int continue_loop = 0;

	do
	  {
	    now = time(NULL);
	    nsecs = cron_next_batch_sleep( &conf_obj_list, now );
	  }
	while(nsecs < 1);

	wake_up = now + nsecs;

	do
	  {
#if CRON_DEBUG
	    cron_log("In cron_main, sleeping %ld secs", nsecs);
#endif
	    sleep( nsecs );

	    nsecs = wake_up - time(NULL);
	  }
	while( nsecs > 0);

        now = time(NULL);

	do 
	  {

	    cronobj = conf_obj_list->obj;

#if CRON_DEBUG
	    cron_log("In the do loop: %s, %d", cronobj->name, 
		     conf_obj_list->startstop);
#endif

	    if (conf_obj_list->startstop) /* the actual process */
	      {
		if (!parse_command(cronobj->command, argv))
		  {
		    cron_error("Could not parse '%s'", cronobj->command);
		    continue;
		  }

#if CRON_DEBUG
		cron_log("looking to execute '%s'", cronobj->command);
#endif

		/* Fork, set uid, run cmd, log finish, ... */
		if ((pid = child_fork()) < 0)
		  {
		    cron_error("fork error");
#if CRON_DEBUG
		    cron_log("fork error");
#endif
		  }

		else if (pid == 0) /* child */
		  {
#if CRON_DEBUG
		    cron_log("In the child");
#endif
		    cron_set_uid( cronobj->user );
		    
#if CRON_DEBUG
		    cron_log("After cron_set_uid");
#endif

		    if( cronobj->cwd )
		      chdir( cronobj->cwd );
		    else if( cronconf.dir)
		      chdir( cronconf.dir );
		    
#if CRON_DEBUG
		    cron_log("Pid %d, trying to exec '%s'", (int)getpid(),
			     cronobj->command);
#endif

		    execvp(argv[0], argv);
		    cron_error("Couldn't execute '%s'", cronobj->command);
		    exit(127);
		  }
	    
		/* parent */

		conf_obj_list->current_pid = pid;
		Add_PidList(conf_obj_list);

		if (conf_obj_list->twin)
		  conf_obj_list->twin->current_pid = pid;
		else
		  cron_error("Process '%d' has no corresponding timeout", 
			     (long)pid);

		free_argv(argv);
	      }

	    else /* conf_obj_list->startstop == 0, the time-out of a process */
	      {
		if (Find_PidList(conf_obj_list->current_pid))
		  /* Uh oh...  process hasn't finished by time-out time..
		     gotta kill the damn thing */
		  {
		    kill(conf_obj_list->current_pid, SIGKILL);
		    cron_error("Warning: Process %d didn't finish in time, "
			       "had to terminate it", 
			       conf_obj_list->current_pid);
		    conf_obj_list->current_pid = 0;

		    if (conf_obj_list->twin)
		      conf_obj_list->twin->current_pid = 0;
		    else
		      cron_error("Process '%d' has no corresponding timeout", 
				 (long)pid);
		  }
	      }

	    if ( (conf_obj_list->next != conf_obj_list)
		&& (conf_obj_list->next->from_sun == conf_obj_list->from_sun) )
	      {
		continue_loop = 1;

		conf_obj_list = conf_obj_list->next;
	      }
	    else
	      {
		continue_loop = 0;
	      }

	    sleep( 1 ); /* get timing right */
	  }
	while( continue_loop );

      } /* while (1) */
    
    cron_cleanup( conf_obj_list );
}

int cron_start( CronObjList *conf_obj_list )
{

    switch ((cron_pid = fork())) {
      case -1:
	  fprintf(stderr, "Can't fork server cron process (%s)",
	  	system_errmsg());
	  cron_error("Can't fork server cron process (%s)", system_errmsg());
	  return REQ_PROCEED;
      case 0:
	  {
	    int i;
	    char path[1024];
	    FILE *fp;
	    char *err;

            if (setsid() < 0) {
	       fprintf(stderr, "Failed to create session (%d)", errno);
	       exit(1);
	    }

	    fclose(stdin);
	    fclose(stdout);
	    fclose(stderr);
 
	    i = open("/dev/null", O_RDWR);
	    if (i < 0) 
	      {
		perror("open: /dev/null:");
		exit(1);
	      }
#if 0
	    dup2(i, 0);
	    dup2(i, 1);
#endif
 
	    for (i = 3; i < OPEN_MAX; ++i)
	      close(i);

	    sprintf(path, "%s/%s/%s", nsroot, ADMSERV_LOGDIR, PIDFILE);
            if ((fp = fopen(path, "w")) == (FILE *) NULL) {
	       cron_error("failed to open pid file:%s (%s)\n", path, system_errmsg());
	       exit(1);
	    }
	    (void) fprintf(fp, "%d\n", getpid());
	    (void) fclose(fp);

	   signal(SIGTERM, exit);
           conf_obj_list_static = conf_obj_list;
	   cron_main( conf_obj_list );
	   exit(0);
	  }
	  break;
      default:
	  fprintf(stdout, "server cron daemon started");
	  cron_error("Server cron daemon started");
	  return REQ_PROCEED;
    }
}

#ifdef	CRON_AS_PLUGIN
/* Add {"init-cron", cron_init, NULL}, to func_standard[] in server main.c 
 *
 * Add this to obj.conf:
 * Init conf-file="cron.conf" fn="init-cron" dir="/tmp" status="on"
 */
int cron_init(pblock *pb, Session *sn, Request *rq)
#else	/* !CRON_AS_PLUGIN */
int cron_init()
#endif	/* CRON_AS_PLUGIN */
{
    struct stat st;
    CronObjList *conf_obj_list;
    char *status;
    char *dir;
    char *conf;
    char  cronconffile[1024];

#if CRON_DEBUG
    cron_log_init();
#endif
    
    cron_error_init();

    sprintf(cronconffile, "%s/ns-cron.conf", get_admroot());

    if (!Read_CronConf(&cronconf, cronconffile))
      {
	cron_error("Unable to read cron config file: '%s'", cronconffile);
        return REQ_PROCEED; /* no ns-cron.conf file, no cron */
      }

    status = cronconf.status;
    dir    = cronconf.dir;
    conf   = cronconf.conffile;

    if (status && strcmp(status, "on"))
	  return REQ_PROCEED;

    if (dir && stat(dir, &st) == -1 ) 
      {
        cron_error("Can't stat cron working directory %s", dir);
        fprintf(stderr, "Can't stat cron working directory %s", dir);
      }

    if (conf) 
      {
	if (stat(conf, &st) == -1) 
	  {
	    cron_error("Can't stat cron config file %s", conf);
	    fprintf(stderr, "Can't stat cron config file %s", conf);
	  }
      }
    else 
      {
	cron_error("The \"conf-file\" parameter not specified"
		   " for init-cron; cron jobs disabled");
	fprintf(stderr, "the \"conf-file\" parameter not specified"
		" for init-cron; cron jobs disabled");
      }

    /* try to parse config file and build CronObjList before the fork */
    if( (conf_obj_list = cron_read_conf( conf )) ) 
      {
	/* So we've read the configuration, not it's time to daemonize
	   and close down all the open fd's */

        chdir( dir );

        return cron_start( conf_obj_list );

    } 
    else {
      cron_error("No cron job is scheduled, ns-cron exiting ...\n");
      fprintf(stdout, "No cron job is scheduled, ns-cron exiting ...\n");
      return REQ_NOACTION;
    }
}


void sig_chld_handler(int sig)
{
  pid_t pid;
  int status;
 
  if ( (pid = wait(&status))  < 0)
    cron_error("wait error");

#if CRON_DEBUG
  cron_log("INTERRUPTED: Process %d finished", pid); 
#endif

  if (!Del_PidList(pid))
    cron_error("Warning: cron has no information about finished child "
	       "process %d", pid);
  
  return;
}

static void sig_term_handler(int unused)
{
    cron_terminate_static();
    exit(0);
}

void cron_kill(void)
{
    if(cron_pid > 0)
        kill(cron_pid, SIGTERM);
}


#ifndef	INPROC_CRON
int main( int argc, char **argv )
{
    CronObjList *conf_obj_list;
#if 0
    pblock *pb = get_mag_init("init-batch-update");
    char *conf   = pblock_findval("conf-file", pb);
#endif
    char *conf   =  "cron.conf";
    int c;
    char path[1024];
    FILE *fp;

#ifdef HPUX
  #ifndef HPUX_64
    _main();
  #endif
#endif

   if (geteuid() != 0) {
      fprintf(stderr, "ns-cron is started with non-root privilege, ");
      fprintf(stderr, "it may fail in carrying out cron tasks where it needs ");
      fprintf(stderr, "to act on behalf of another user!<BR>\n");
      fflush(stderr);
   }

   while ((c = getopt(argc, argv, "d:")) != -1) {
      switch(c) {
      case 'd':
         nsroot = STRDUP(optarg);
         chdir(optarg);
         break;
      case '?':
         fprintf(stderr, "Usage: %s [-d ns-home-dir]\n", argv[0]);
         return 1;
      }
   }

   if (!nsroot) {
      fprintf(stderr, "Assuming current directory as the ns-home\n");
      nsroot = STRDUP(".");
   }

   sprintf(path, "%s/%s/%s", nsroot, ADMSERV_LOGDIR, PIDFILE);
   if (file_exists(path)) {
      pid_t pid;
      if ((fp = fopen(path, "r")) == (FILE *) NULL) {
         fprintf(stderr, "%s:failed to open pid file:%s (%d)\n", argv[0], path, errno);
	 exit(1);
      }
      if (fscanf(fp, "%d\n", &pid) != EOF) {
         if (kill(pid, 0) == 0) {
	    fprintf(stderr, "%s:already running for this server root, pid=%d\n", argv[0], pid);
	    (void) fclose(fp);
	    exit(1);
	 }
      }
      (void) fclose(fp);
   }

   cron_init();
}
#endif	/* !INPROC_CRON */
