/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*********************************************************************
**
** NAME:
**   AdmservPlugin.c
**
** DESCRIPTION:
**   Starting with the 5.0 version, the admin server is implemented as an
**   ES plugin. The plugin implements custom functions for authentication,
**   access control, resource download and CGI/servlet execution.
**    
**
** NOTES:
**  Functions in this file are referenced from the 
**  <server-root>/admin-serv/config/obj.conf configuration file.
**
** HISTORY:
**
*********************************************************************/

#include "libadminutil/admutil.h"
#include "libadmsslutil/admsslutil.h"
#include "libadmsslutil/srvutilssl.h"
#include "libadmsslutil/psetcssl.h"
#include "libadmin/libadmin.h"

#include <stdio.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <netsite.h>
#include <base/util.h>
#include <base/pblock.h>
#include <base/ereport.h>
#include <base/eventlog.h>
#include <frame/func.h>

#include "AdmservPlugin.h"
#include "hashtable.h"
#include "frame/protocol.h"

#include "errors.h"
#include "extcmap.h"

#include "aclapi.h"
#include "plistdef.h"

#include "version.h"

#include "nspr.h"
#include "prnetdb.h"
#include "seccomon.h"
#include "ssl.h"
#include "nss.h"

/* Globals...hack. */
static char           *serverroot    = NULL;
static long            cacheLifetime = 0; /* Defaults to 0 (entries immediately expire) */
static LdapServerData  registryServer;
static LdapServerData  userGroupServer;

static char *accessHosts = NULL;
static char *accessAddresses = NULL;
static FILE *servlets_properties, *rules_properties;

static int servlet_env_init(LDAP *server, AdmldapInfo *ldapInfo);
static int start_servlet_engine(Session *sn, Request *rq);
static int populate_SIE_table(HashTable *table, AdmldapInfo *ldapInfo);
static int populate_ISIE_table(HashTable *table, AdmldapInfo *ldapInfo);
static int set_servlet_properties(pblock *pb, Session *sn, Request *rq);

#ifdef XP_WIN32
#define SERVLET_PLUGIN_PATH "bin/https/bin/NSServletPlugin.dll"
#else
#ifdef HPUX
#define SERVLET_PLUGIN_PATH "bin/https/lib/libNSServletPlugin.sl"
#else /* All other Unix platforms */
#define SERVLET_PLUGIN_PATH "bin/https/lib/libNSServletPlugin.so"
#endif /* HPUX */
#endif /* XP_WIN32 */

#define PROP_HEADER1 "#\n\
# DO NOT EDIT THIS FILE!\n\
# This file is generated automatically by combining the contents of\n\
# %s file with the servlet task information in\n\
# the configuration directory. To add some additional information\n\
# modify %s.\n#\n\n"

#define PROP_HEADER2 "\n#\n\
# The information below this line is generated from the configuration directory.\n#\n"
   
#ifdef XP_UNIX
#define PATH_SEP ':'
#else /* XP_WIN32 */
#define PATH_SEP ';'
#endif

/* extern int adm_secureLDAP;   no longer needed...*/


#define NS_ADMIN_ACCESS_HOSTS (char*)"configuration.nsAdminAccessHosts"
#define NS_ADMIN_ACCESS_ADDRESSES (char*)"configuration.nsAdminAccessAddresses"

#ifdef XP_UNIX
#define ERROR_EXIT(str)     \
    fprintf(stderr, str);   \
    exit(1) 
#define ERROR_MSG(str) fprintf(stderr, str)
#else /* XP_WIN32 */
#include "nt/messages.h"
#define ERROR_EXIT(str) \
    LogErrorEvent(NULL,EVENTLOG_ERROR_TYPE,0,MSG_BAD_STARTUP,str,NULL); \
    exit(1)
#define ERROR_MSG(str) LogErrorEvent(NULL,EVENTLOG_ERROR_TYPE,0,MSG_BAD_STARTUP,str,NULL)
#endif /* XP_WIN32 */

struct HashTable {
  PList_t plist;
};
 
HashTable*
HashTableCreate()
{
  HashTable *ht = (HashTable *)malloc(sizeof(HashTable));
    
  if (!ht)
    return NULL;

  ht->plist = PListNew(0);

  return ht;
}

int
HashTableInsert(HashTable *ht, char *name, void *value)
{
  int rv;

  if ((rv = PListFindValue(ht->plist, name, 0, 0)) >= 0)
    return (PListSetValue(ht->plist, rv, value, 0) >= 0);

  return (PListInitProp(ht->plist, 0, name, value, 0) >= 0);
}

void *
HashTableFind(HashTable *ht, char *name)
{
  void *value = NULL;

  (void)PListFindValue(ht->plist, name, &value, 0);

  return value;
}

void
HashTableEnumerate(HashTable *ht, HashEnumFunc *fn, void *data)
{
  PListEnumerate(ht->plist, fn, data);
}


/* Caches for [I]SIE DNs, authenticated users, and available tasks */
static HashTable *servers    = NULL;
static HashTable *auth_users = NULL;
static HashTable *auth_tasks = NULL;
static HashTable *servlets   = NULL;

static int servletCnt = 0;

/* Runtime command lookup hashtable */
static HashTable *commands   = NULL;

/* Config file listing URIs not subject to task evaluation */
#define EXEMPTION_FILE "exempturis.conf"
/* List of URIs not subject to task evaluation */
static HashTable *exempt_uris = NULL;
/* Initialize the table of exempt URIs */
static int get_exempted_uris(pblock *param, Request *rq);
static int is_exempt_uri(const char *uri);

#define NS_EXEC_REF           (char*)"nsExecRef"
#define NS_EXEC_REF_QUERY     (char*)"(nsExecRef=*)"
#define SERVLET_QUERY         (char*)"(nsExecRef=java:*)"
#define NS_LOG_SUPPRESS       (char*)"nsLogSuppress"
#define LOG_SUPPRESS_ON_VALUE (char*)"true"

static char *searchAttributes[] = { NS_EXEC_REF, NS_LOG_SUPPRESS, NULL };
static char *searchServletAttrs[] = { NS_EXEC_REF, (char*)"cn", NULL };


/* ------------------------------ _uudecode ------------------------------- */

const unsigned char pr2six[256]={
  64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,
  64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,62,64,64,64,63,
  52,53,54,55,56,57,58,59,60,61,64,64,64,64,64,64,64,0,1,2,3,4,5,6,7,8,9,
  10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,64,64,64,64,64,64,26,27,
  28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51,
  64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,
  64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,
  64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,
  64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,
  64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,
  64,64,64,64,64,64,64,64,64,64,64,64,64
};

char *_uudecode(char *bufcoded)
{
  register char *bufin = bufcoded;
  register unsigned char *bufout;
  register int nprbytes;
  unsigned char *bufplain;
  int nbytesdecoded;

  /* Find the length */
  while(pr2six[(int)*(bufin++)] <= 63);
  nprbytes = bufin - bufcoded - 1;
  nbytesdecoded = ((nprbytes+3)/4) * 3;

  bufout = (unsigned char *) MALLOC(nbytesdecoded + 1);
  bufplain = bufout;

  bufin = bufcoded;

  while (nprbytes > 0) {
    *(bufout++) = (unsigned char)
      (pr2six[(int)(*bufin)] << 2 | pr2six[(int)bufin[1]] >> 4);
    *(bufout++) = (unsigned char)
      (pr2six[(int)bufin[1]] << 4 | pr2six[(int)bufin[2]] >> 2);
    *(bufout++) = (unsigned char)
      (pr2six[(int)bufin[2]] << 6 | pr2six[(int)bufin[3]]);
    bufin += 4;
    nprbytes -= 4;
  }

  if(nprbytes & 03) {
    if(pr2six[(int)bufin[-2]] > 63)
      nbytesdecoded -= 2;
    else
      nbytesdecoded -= 1;
  }
  bufplain[nbytesdecoded] = '\0';

  return (char *)bufplain;
}


enum {
  LDAPU_STR_FILTER_DEFAULT,
  LDAPU_STR_FILTER_USER,
  LDAPU_STR_FILTER_GROUP,
  LDAPU_STR_FILTER_MEMBER,
  LDAPU_STR_FILTER_MEMBER_RECURSE,
  LDAPU_STR_ATTR_USER,
  LDAPU_STR_ATTR_CERT,
  LDAPU_STR_MAX_INDEX
};


static char *ldapu_strings[] = {
  (char*)"objectclass=*",		/* LDAPU_STR_DEFAULT */
  (char*)"uid=%s",			/* LDAPU_STR_FILTER_USER */
  (char*)"(& (cn=%s) (| (objectclass=groupofuniquenames) (objectclass=groupofnames)))", /* LDAPU_STR_FILTER_GROUP */
  (char*)"(| (uniquemember=%s) (member=%s))",	/* LDAPU_STR_FILTER_MEMBER */
  (char*)"(& %s (| (objectclass=groupofuniquenames) (objectclass=groupofnames))", /* LDAPU_STR_FILTER_MEMBER_RECURSE */
  (char*)"uid",			/* LDAPU_STR_ATTR_USER */
  (char*)"userCertificate;binary"	/* LDAPU_STR_ATTR_CERT */
};


/* If we are not interested in the returned attributes, just ask for one
 * attribute in the call to ldap_search.  Also don't ask for the attribute
 * value -- just the attr.
 */
static const char *default_search_attrs[] = { "c" , 0 };
static int default_search_attrsonly = 1;


/*
 * ldapu_find
 *   Description:
 *	Caller should free res if it is not NULL.
 *   Arguments:
 *	ld		Pointer to LDAP (assumes connection has been
 *	    		established and the client has called the
 *	    		appropriate bind routine)
 *	base		basedn (where to start the search)
 *	scope		scope for the search.  One of
 *	    		LDAP_SCOPE_SUBTREE, LDAP_SCOPE_ONELEVEL, and
 *	    		LDAP_SCOPE_BASE
 *	filter		LDAP filter
 *	attrs		A NULL-terminated array of strings indicating which
 *	    		attributes to return for each matching entry.  Passing
 *	    		NULL for this parameter causes all available
 *	    		attributes to be retrieved.
 *	attrsonly	A boolean value that should be zero if both attribute
 *	    		types and values are to be returned, non-zero if only
 *	    		types are wanted.
 *	res		A result parameter which will contain the results of
 *			the search upon completion of the call.
 *   Return Values:
 *	LDAPU_SUCCESS	if entry is found
 *	LDAPU_FAILED	if entry is not found
 *	<rv>		if error, where <rv> can be passed to
 *			ldap_err2string to get an error string.
 */
int 
ldapu_find (LDAP *ld, const char *base, int scope,
	    const char *filter, const char **attrs,
	    int attrsonly, LDAPMessage **res)
{
  int retval;
#ifdef USE_THIS_CODE /* ASYNCHRONOUS */
  int msgid;
#endif
  int numEntries;

  *res = 0;

  /* If base is NULL set it to null string */
  if (!base) {
    base = "";
  }

  if (!filter || !*filter) {
    filter = ldapu_strings[LDAPU_STR_FILTER_DEFAULT];
  }

  retval = ldap_search_s(ld, base, scope, filter, (char **)attrs,
			 attrsonly, res);

  if (retval != LDAP_SUCCESS)
    {
      /* retval = ldap_result2error(ld, *res, 0); */
      return(retval);
    }

  numEntries = ldap_count_entries(ld, *res);

  if (numEntries == 1) {
    /* success */
    return LDAPU_SUCCESS;
  }
  else if (numEntries == 0) {
    /* not found -- but not an error */
    return LDAPU_FAILED;
  }
  else if (numEntries > 0) {
    /* Found more than one entry! */
    return LDAPU_ERR_MULTIPLE_MATCHES;
  }
  else {
    /* should never get here */
    ldap_msgfree(*res);
    return LDAP_OPERATIONS_ERROR;
  }
}


/*
 * ldapu_find_uid_attrs
 *   Description:
 *	Maps the given uid to a user dn.  Caller should free res if it is not
 *	NULL.  Accepts the attrs & attrsonly args.
 *   Arguments:
 *	ld		Pointer to LDAP (assumes connection has been
 *	    		established and the client has called the
 *	    		appropriate bind routine)
 *	uid		User's name
 *	base		basedn (where to start the search)
 *	attrs		list of attributes to retrieve
 *	attrsonly	flag indicating if attr values are to be retrieved
 *	res		A result parameter which will contain the results of
 *			the search upon completion of the call.
 *   Return Values:
 *	LDAPU_SUCCESS	if entry is found
 *	LDAPU_FAILED	if entry is not found
 *	<rv>		if error, where <rv> can be passed to
 *			ldap_err2string to get an error string.
 */
int 
ldapu_find_uid_attrs (LDAP *ld, const char *uid, const char *base,
		      const char **attrs, int attrsonly,
		      LDAPMessage **res)
{
  int		scope = LDAP_SCOPE_SUBTREE;
  char	filter[ BUFSIZ ];
  int		retval;
  int size;

  size = sizeof(ldapu_strings[LDAPU_STR_FILTER_USER]) + strlen(uid);

  if (size < sizeof(filter)) {
    /* setup filter as (uid=<uid>) */
    sprintf(filter, ldapu_strings[LDAPU_STR_FILTER_USER], uid);
    retval = ldapu_find(ld, base, scope, filter, attrs, attrsonly, res);
  } else {
    char * scratch = (char *) MALLOC(size);
    if (scratch) {
      sprintf(scratch, ldapu_strings[LDAPU_STR_FILTER_USER], uid);
      retval = ldapu_find(ld, base, scope, scratch, attrs, attrsonly, res);
      FREE(scratch);
    } else {
      retval = LDAPU_ERR_OUT_OF_MEMORY;
    }
  }

  return retval;
}



/*
 * ldapu_find_userdn
 *   Description:
 *	Maps the given uid to a user dn.  Caller should free dn if it is not
 *	NULL. 
 *   Arguments:
 *	ld		Pointer to LDAP (assumes connection has been
 *	    		established and the client has called the
 *	    		appropriate bind routine)
 *	uid		User's name
 *	base		basedn (where to start the search)
 *	dn 		user dn
 *   Return Values:
 *	LDAPU_SUCCESS	if entry is found
 *	LDAPU_FAILED	if entry is not found
 *	<rv>		if error, where <rv> can be passed to
 *			ldap_err2string to get an error string.
 */
int 
ldapu_find_userdn (LDAP *ld, const char *uid, const char *base,
		   char **dn)
{
  LDAPMessage *res = 0;
  int		retval;

  retval = ldapu_find_uid_attrs(ld, uid, base, default_search_attrs,
				default_search_attrsonly, &res);

  if (retval == LDAPU_SUCCESS) {
    LDAPMessage *entry;

    entry = ldap_first_entry(ld, res);
    *dn = ldap_get_dn(ld, entry);
  }
  else {
    *dn = 0;
  }

  if (res) ldap_msgfree(res);

  return retval;
}




static char*
adm_normalize_dn(char* origDN, char* newDN) {

  char *orig = origDN, *dest = newDN;

  if (!newDN) return NULL;

  while (*orig != '\0') {
    if (*orig == ',') {
      *dest++ = *orig++;
      while (*orig == ' ') orig++;
    }
    else *dest++ = *orig++;
  }
  *dest = '\0';
  return newDN;
}

static void
convert_to_lower_case(char *s)
{
  register char *p = s;

  while (*p)
    {
      if ((((unsigned char)*p)<0x80)&&(isupper(*p)))
	*p = _tolower(*p);
      p++;
    }
}

static
void
closeLDAPConnection(LDAP *server)
{
  ldap_unbind_s(server);
}

static LDAP *
openLDAPConnection(LdapServerData *data)
{
  static int falseint = 0;

  LDAP *server;

  if (data->secure) {
    if (!(server = ldapssl_init(data->host, data->port, 1))) {
      ereport(LOG_FAILURE, (char*)"openLDAPConnection(): ldapssl_init failed");
      return NULL;
    }
  }
  else {
    if (!(server = ldap_init(data->host, data->port))) {
      ereport(LOG_FAILURE, (char*)"openLDAPConnection(): ldap_init failed");
      return NULL;
    }
  }

  if (ldap_set_option(server, LDAP_OPT_CACHE_ENABLE, (void*)&falseint) != LDAP_SUCCESS)
    {
      ereport(LOG_FAILURE, (char*)"openLDAPConnection(): ldap_set_option failed");
      closeLDAPConnection(server);
      return NULL;
    }

  return (server);
}

#if defined(__cplusplus) && defined(SOLARIS)
extern "C"
#endif
static
int LDAP_CALL LDAP_CALLBACK
admin40_ldap_rebind_proc(LDAP *ld, char **who, char **pw, int *auth, int freeit, void *arg)
{
  RebindData *data = (RebindData*)arg;

  if (!freeit)
    {
      *who  = data->user;
      *pw   = data->pw;
      *auth = LDAP_AUTH_SIMPLE;
    }

  return LDAP_SUCCESS;
}

static
void
setLDAPRebindProc(LDAP *server, char *user, char *pw)
{
  RebindData *data = (RebindData*)malloc(sizeof(RebindData));

  if (user) data->user = STRDUP(user);
  if (pw) data->pw   = STRDUP(pw);

  /* free old (void*)data, mem leak if not freed 
     if (NSLDAPI_VALID_LDAP_POINTER(server)) {
     free(server->ld_rebind_arg);
     }*/

  ldap_set_rebind_proc(server, admin40_ldap_rebind_proc, (void *)data);
}

static int
admin40_ldap_auth_userdn_password(LDAP *server, 
				  char *userdn, 
				  char *pw, 
				  int *pw_expiring)
{
  int ver = LDAP_VERSION3; 
  LDAPMessage *res = NULL;           /* LDAP result message */ 
  LDAPControl **ctrls = NULL;
  int ldapError;
  int msgid = 0;              /* async bind id */ 
  int rc = 0;                 /* result code of LDAP operation */ 
  int tries = 0;

  *pw_expiring = -1;

  setLDAPRebindProc(server, userdn, pw);

#if 1
  /* This is for password expiration check */
  /* set the LDAP version -- only v3 knows about controls */ 
  ldap_set_option(server, LDAP_OPT_PROTOCOL_VERSION, &ver); 

  /* DT 9/18/98 - Fix for bind problem. Previously, if pw == null,
   * then the bind does not occur (connection is still anonymous),
   * but ldap_simple_bind() returns LDAP_SUCCESS. We want the
   * bind to fail if pw == null, or the pw does not match.
   */
  if (pw == NULL)
    pw = (char*)"";

  /* bind user asynchronously */ 
  if ( (msgid = ldap_simple_bind(server, userdn, pw)) == -1 ) {
    /*return LDAP_OPERATIONS_ERROR;*/
    return LDAP_SERVER_DOWN;
  }

  rc = ldap_result(server, msgid, 1, NULL, &res);

  switch (rc) {
  case -1: 
    if (res) ldap_msgfree(res);
    return ldap_get_lderrno(server, NULL, NULL);
  default: 
    if ((ldapError = ldap_result2error(server, res, 0)) != LDAP_SUCCESS) {
      if (res) ldap_msgfree(res);
      return ldapError;
    }
  }

  if ((ldapError = ldap_parse_result(server, res, NULL, NULL, NULL, NULL, &ctrls, 0))
      != LDAP_SUCCESS) {
    if (res) ldap_msgfree(res);
    return ldapError;
  }

  if ((ldapError = ldap_result2error(server, res, 0)) != LDAP_SUCCESS) {
    if (res) ldap_msgfree(res);
    return ldapError;
  }

  if (ctrls) {
    int i; 
    for (i = 0; ctrls[i] != NULL; ++i) {
      if(!(strcmp(ctrls[i]->ldctl_oid, LDAP_CONTROL_PWEXPIRED))) {
	*pw_expiring = 0;
      } 
      else if(!(strcmp(ctrls[i]->ldctl_oid, LDAP_CONTROL_PWEXPIRING))) {
	/* "The password is expiring in n seconds" */ 
	if ((ctrls[i]->ldctl_value.bv_val != NULL ) && 
	    (ctrls[i]->ldctl_value.bv_len > 0)) {
	  *pw_expiring = atoi(ctrls[i]->ldctl_value.bv_val); 
	} 
      } 
    }
    ldap_controls_free(ctrls); 
  }

  if (res) ldap_msgfree(res);
  return ldapError;
#else
  /* This is fallback, no password expiration is checked  */
  return ldapu_auth_userdn_password(server, userdn, pw);
#endif
}

static int
admin40_ldap_auth_server(LDAP *server, LdapServerData *data)
{
  int ignored;
  if (!(data->bindDN) && !(data->bindPW))
    return LDAPU_SUCCESS;

  return admin40_ldap_auth_userdn_password(server, data->bindDN, data->bindPW, &ignored);
}

static char *
formLdapURL(LdapServerData *data)
{
  char *s = (char*)malloc(LDAPS_BASE_PREFIX_LENGTH + strlen(data->host) + 1 + 5 + 1 + strlen(data->baseDN) + 1);
  sprintf(s, "%s%s:%d/%s", (data->secure?LDAPS_BASE_PREFIX:LDAP_BASE_PREFIX), data->host, data->port, data->baseDN);
  return s;
}

static int
extractLdapError()
{
  ereport(LOG_FAILURE, (char*)"extractLdapServerData(): invalid ldapURL");
  return FALSE;
}

static int
extractLdapServerData(LdapServerData *data, char *ldapURL)
{
  char *p=NULL;

  if (!ldapURL)
    return extractLdapError();

  if (strstr(ldapURL, LDAP_BASE_PREFIX)) {
    ldapURL += strlen(LDAP_BASE_PREFIX);
    data->secure = 0;
    data->port = 389;   /* default non-SSL port */
  }
  else if (strstr(ldapURL, LDAPS_BASE_PREFIX)) {
    ldapURL += strlen(LDAPS_BASE_PREFIX);
    data->secure = 1;
    data->port = 636;   /* default SSL port */
  }
  else
    return extractLdapError();

  if ((p = strtok(ldapURL, "/")) == NULL)
    return extractLdapError();
  data->host = strdup(p);

  if ((p = strtok(NULL, "?")) == NULL)
    return extractLdapError();
  data->baseDN = strdup(p);

  /* Currently unused */
  data->bindDN = NULL;
  data->bindPW = NULL;

  return TRUE;
}

static int onlyOnceSwitch = 0;

#if defined(WINNT)
static char const prefixMask[] = "\\alias\\%s-";
static char const secmodName[] = "\\alias\\secmod.db";
#else
static char const prefixMask[] = "/alias/%s-";
static char const secmodName[] = "/alias/secmod.db";
#endif

static int
sslinit(AdmldapInfo info, char const * svrroot, int forceInit) {
  int rc = PR_FALSE;
  
  if (onlyOnceSwitch) {
    rc = PR_TRUE;
  } else {
    char * x;
    /*
     * Usually one would find that the path to the cert and key db
     * is alais/admin-serv-hostname- but one can not assume that
     * to be the case. Thus we must disassemble the SIEDN to find
     * the instance's actual name.
     */
    x = admldapGetSIEDN(info);
    if (x) {
      if (x[0]) {
	int i;
	for (i=1; x[i] && !(x[i] == '=' && x[i-1] != '\\'); ++i); /* scan past attr name */
	if (x[i] == '=') {
	  char * sie;
	  for (++i; x[i] && x[i] == ' '; ++i); /* drop value's leading spaces */
	  sie = &x[i];
	  for ( ; x[i] && !(x[i] == ',' && x[i-1] != '\\'); ++i); /* till end of value */
	  if (x[i] == ',') {
	    char * scratch;
	    for (x[i--] = 0; x[i] == ' '; --i) x[i] = 0; /* drop value's trailing spaces */
	    scratch = (char *) malloc(sizeof(prefixMask) + strlen(sie));
	    if (scratch) {
	      sprintf(scratch, prefixMask, sie);
	      if (SECSuccess == NSS_Initialize(serverroot, scratch, scratch, secmodName, 0)) {
		if (SSL_OptionSetDefault(SSL_ENABLE_SSL2, PR_FALSE)
		    || SSL_OptionSetDefault(SSL_ENABLE_SSL3, PR_TRUE)) {
		  ereport(LOG_FAILURE, (char*)"SSL_OptionSetDefault failed.");
		} else {
		  if (SSLPLCY_Install() == PR_SUCCESS) {
		    onlyOnceSwitch = 1;
		    rc = PR_TRUE;
		  } else {
		    ereport(LOG_FAILURE, (char*)"SSLPLCY_Install() failed.");
		  }
		}
	      }
	      free((void *) scratch);
	    }
	  }
	}
	free((void *) x);
      }
    }
  }

  return rc;
}

static int
buildUGInfo(char** errorInfo) {
  AdmldapInfo info;
  int  error;
  char  path[PATH_MAX];
  char *userGroupLdapURL;
  char *userGroupBindDN;
  char *userGroupBindPW;
  char *dirInfoRef;

  *errorInfo = (char*)"";

  /* Check whether data is available already */
  if (userGroupServer.host) return TRUE;

  if (!serverroot) {
    *errorInfo = (char*)"NULL ns root";
    return FALSE;
  }

  sprintf(path, "%s%cadmin-serv%cconfig", serverroot, FILE_PATHSEP, FILE_PATHSEP);


  if (!(info = admldapBuildInfo(path, &error))) {
    char const msg[] = "buildUGInfo(): unable to create AdmldapInfo (error code = %d)";
    char scratch[sizeof(msg) + 20];
    sprintf(scratch, msg, error);
    ereport(LOG_FAILURE, scratch);
    return FALSE;
  }

  if (admldapGetSecurity(info)) {

    sslinit(info, serverroot, PR_FALSE);
    /*if(ADMSSL_InitSimple2(serverroot)) {
     *errorInfo = "unable to initialize the security database (secomd, key, and cert db)";
     return FALSE;
     }*/

    if (admldapBuildInfoSSL(info, &error)) {
    } else {
      destroyAdmldap(info);
      return FALSE;
    }
  }
  
  userGroupServer.host   = NULL;
  userGroupServer.port   = 0;
  userGroupServer.secure = 0;
  userGroupServer.baseDN = NULL;

  admldapGetLocalUserDirectory(info, 
			       &userGroupLdapURL, 
			       &userGroupBindDN, 
			       &userGroupBindPW, 
			       &dirInfoRef, 
			       &error);

  if (error != UG_OP_OK) {
    *errorInfo = (char*)"unable to set User/Group baseDN";
    destroyAdmldap(info);
    return FALSE;
  }
   
  if (!extractLdapServerData(&userGroupServer, userGroupLdapURL)) {
    destroyAdmldap(info);
    *errorInfo = (char*)"unable to extract User/Group LDAP info";
    return FALSE;
  }
  userGroupServer.bindDN = userGroupBindDN ? STRDUP(userGroupBindDN) : NULL;
  userGroupServer.bindPW = userGroupBindPW ? STRDUP(userGroupBindPW) : NULL;
  destroyAdmldap(info);
   
  return TRUE;
}


int
task_register_server(char *serverid, char *sieDN)
{
   
  HashTableInsert(servers, serverid, sieDN);
  return TRUE;
}


static int
reverse_uri(char **storage, char *limit, char *taskuri)
{
  char *p = strchr(taskuri, '/');

  if (p)
    {
      *(p++) = '\0';
      if (!reverse_uri(storage, limit, p))
	return FALSE;
    }

  /* magic number 4 = "cn=" and "," */
  if ((*storage + strlen(taskuri)+4) > limit)
    {
      ereport(LOG_FAILURE, (char*)"reverse_uri():taskDN exceeds LINE_LENGTH, taskDN truncated...");
      return FALSE;
    }

  sprintf(*storage, "cn=%s, ", taskuri);
  *storage += strlen(*storage);
  return TRUE;
}

static int
build_full_DN(char **storage, char *limit, char *taskuri, char *sieDN)
{
  if (!reverse_uri(storage, limit, taskuri))
    return FALSE;

  if ((*storage + strlen(sieDN)) > limit)
    {
      ereport(LOG_FAILURE, (char*)"build_full_DN():taskDN exceeds LINE_LENGTH, taskDN truncated...");
      return FALSE;
    }

  sprintf(*storage, "%s", sieDN);
  return TRUE;
}


/* ---------------------------- admin40_redirect ----------------------------- */

/*
 * Perform a redirect function exec. DT 12/12/97.
 */


static int
admin40_redirect(Session *sn, Request *rq, char *from, char *url)
{
  pblock *execpb = NULL;
  int     ret    = 0;

  execpb = pblock_create(ADMIN_HASH_SIZE);

  pblock_nvinsert("fn", "redirect", execpb);
  pblock_nvinsert("from", from, execpb);
  pblock_nvinsert("url", url, execpb);
  ret = func_exec(execpb, sn, rq);
  pblock_free(execpb);
  return ret;
}


static int
check_auth_tasks_cache(char *dn, char *userdn, pblock *pb, pblock *reqpb, long now)
{
  TaskCacheEntry *cache_entry;
  char            normEntryDN[1024];
  long            createTime;

  adm_normalize_dn(dn, normEntryDN);

  /* Server is down. Attempt to resolve task entry from auth_tasks cache. */
  cache_entry = (TaskCacheEntry*)HashTableFind(auth_tasks, normEntryDN);

  if (!cache_entry || !(createTime = (long)HashTableFind(cache_entry->auth_userDNs, userdn)) || ((now - createTime) > cacheLifetime))
    return REQ_ABORTED; /* userdn not authorized for task, or task not found */

  pblock_nvinsert("execRef", cache_entry->execRef, pb);
  if (cache_entry->execRefArgs)
    pblock_nvinsert("execRefArgs", cache_entry->execRefArgs, pb);
  if (cache_entry->logSuppress)
    pblock_nvinsert("logSuppress", "true", reqpb);
  return REQ_PROCEED;
}



static void
populate_tasks_from_server(char *serverid, const void *sieDN,  void *userdata)
{
  PopulateTasksData *data   = (PopulateTasksData *)userdata;
  LDAP              *server = data->server;
  LDAPMessage       *result, *e;
  int                ldapError;
  char              normDN[1024];
  int tries = 0;

#if 0
  if ((ldapError = ldap_search_s(server, sieDN, LDAP_SCOPE_SUBTREE, NS_EXEC_REF_QUERY, searchAttributes, 0, &result)) != LDAP_SUCCESS)
    {
      ldap_perror(server, "populate_tasks_from_server():task search error");
      return;
    }
#endif /* 0 */

  if(data->server == NULL) {
    if(!(server = openLDAPConnection(&registryServer))) {
      ldap_perror(server, "populate_tasks_from_server(): Unable to open LDAPConnection");
      return;
    }
    data->server = server;
  }

  tries = 0;
  do {
    ldapError = ldap_search_s(server, (char *)sieDN, LDAP_SCOPE_SUBTREE, NS_EXEC_REF_QUERY, searchAttributes, 0, &result);
    if(ldapError != LDAP_SERVER_DOWN && ldapError != LDAP_CONNECT_ERROR)
      break;
     
    closeLDAPConnection(server);
    if(!(server = openLDAPConnection(&registryServer))) {
      ldap_perror(server, "populate_tasks_from_server(): Unable to open LDAPConnection");
      return;
    }
    data->server = server;
  } while (server != NULL && ++tries < 2);

  if (ldapError != LDAPU_SUCCESS) {
    ldap_perror(server, "populate_tasks_from_server():task search error");
    return;
  }
   
  for (e = ldap_first_entry(server, result) ; e != NULL ; e = ldap_next_entry(data->server, e))
    {
      char *dn, **vals, **vals2, *execRefArgs;
      TaskCacheEntry *cache_entry;

      dn = ldap_get_dn(server, e);

      if (!dn) continue;

      vals = ldap_get_values(server, e, NS_EXEC_REF);

      if (!vals) {
	ldap_memfree(dn);
	continue;
      }

      vals2 = ldap_get_values(server, e, NS_LOG_SUPPRESS);

      if ((execRefArgs = strchr(vals[0], '?')) != NULL)
	*(execRefArgs++) = '\0';

      adm_normalize_dn(dn, normDN);
      convert_to_lower_case(normDN);

      if (!(cache_entry = (TaskCacheEntry *)HashTableFind(auth_tasks, normDN))) {
	cache_entry = (TaskCacheEntry*)malloc(sizeof(TaskCacheEntry));
	cache_entry->auth_userDNs = HashTableCreate();
	HashTableInsert(auth_tasks, normDN, cache_entry);
      }
      else {
	free(cache_entry->execRef);
	if (cache_entry->execRefArgs)
	  free(cache_entry->execRefArgs);
      }

      cache_entry->execRef     = strdup(vals[0]);
      cache_entry->execRefArgs = execRefArgs ? strdup(execRefArgs) : NULL;
      cache_entry->logSuppress = (vals2 && !STRCASECMP(vals2[0], LOG_SUPPRESS_ON_VALUE));
      HashTableInsert(cache_entry->auth_userDNs, data->userDN, (char*)(data->now));

      ldap_value_free(vals);
      if (vals2)
	ldap_value_free(vals2);

      ldap_memfree(dn);
    }

  ldap_msgfree(result);
}


static void
populate_task_cache_entries(char *userDN, LDAP *server)
{
  PopulateTasksData data;

  data.server = server;
  data.userDN = userDN;
  data.now    = time(0);

  HashTableEnumerate(servers, populate_tasks_from_server, &data);
}


#   define CLIENTS_DIR "/clients/"
#   define JAVA_DIR "/java/"
#   define MANUAL_DIR "/manual/"
#   define DIST_DIR "/dist"
#   define FAVICON "/favicon.ico"


int
admin40_task_access_eval(pblock *pb, Session *sn, Request *rq)
{
  int             ldapError;
  char            entryDN[LINE_LENGTH];
  char           *p;
  LDAP           *server;
  LDAPMessage    *result, *e;
  char           **vals, **vals2;
  char           *userdn;
  char           *pw;
  char           *ulevel;
  char           *serverid;
  char           *storage = entryDN;
  char           *uri = pblock_findval("uri", rq->reqpb);
  long            now;
  int tries = 0;

  uri = STRDUP(uri);

  if (!(p = strchr(uri+1, '/')))
    {
      pblock_nvinsert("admin-error", "Invalid URL", rq->vars);
      return REQ_ABORTED;
    }

  serverid = uri+1;
  uri      = p+1;
  *p       = '\0';

  if (!(p = (char*)HashTableFind(servers, serverid)))
    {
      /* DT 4/6/98 -- If we're seeing a serverid for the first time, then we try to do
       *              a resync to pull in new data for the serverid. If it still fails,
       *              then the client is out of luck. */

      admin40_runtime_command_exec(RUNTIME_RESYNC_COMMAND, NULL, NULL);

      if (!(p = (char*)HashTableFind(servers, serverid)))
	{
	  ereport(LOG_FAILURE, (char*)"admin40_task_access_eval(): unable to find registered server (%s)", serverid);
	  return admin40_error(sn, rq, PROTOCOL_BAD_REQUEST, (char*)"server not registered"); /*i18n*/
	}
    }

  if (!build_full_DN(&storage, entryDN+LINE_LENGTH, uri, p))
    {
      pblock_nvinsert("admin-error", "Invalid URL", rq->vars);
      return REQ_ABORTED;
    }

  convert_to_lower_case(entryDN);

  if ((ulevel = pblock_findval("ADM_UserLevel", rq->vars)) && !STRCMP(ulevel, LOCAL_SUPER_NAME))
    return check_auth_tasks_cache(entryDN, LOCAL_SUPER_NAME, pb, rq->reqpb, 0);

  userdn = pblock_findval("userdn", rq->vars);
  pw     = pblock_findval("auth-password", rq->vars);

  if (!userdn || !pw)
    {
      ereport(LOG_FAILURE, (char*)"admin40_task_access_eval(): invalid userdn/pw");
      return REQ_ABORTED;
    }

  /* DT 3/10/98
   * Turn on the task cache for evals, and not just for DS failover.
   */

  now = time(0);

  if (check_auth_tasks_cache(entryDN, userdn, pb, rq->reqpb, now) == REQ_PROCEED)
    return REQ_PROCEED;

  /* Cache lookup failed or the entry expired. Try the DS. */

  if (!(server = openLDAPConnection(&registryServer)))
    return admin40_error_std(sn, rq, (char*)"unable to open LDAPConnection"); /*i18n*/

#if 0
  if ((ldapError = admin40_ldap_auth_server(server, &registryServer)) != LDAPU_SUCCESS)
    {
      closeLDAPConnection(server);

      if ((ldapError == LDAP_CONNECT_ERROR) || (ldapError == LDAP_SERVER_DOWN))
	return check_auth_tasks_cache(entryDN, userdn, pb, rq->reqpb, 0); /* Try the cache, ignoring expiration */

      return admin40_error(sn, rq, PROTOCOL_UNAUTHORIZED, "invalid user credentials"); /*i18n*/
    }
#endif /* 0 */

  tries = 0;
  do {
    /*
     * Bug fix #340829 - do NOT bind as the local admin here, bind as user DN.
     */
    int ignored;
    ldapError = admin40_ldap_auth_userdn_password(server, userdn, pw, &ignored);
    if(ldapError != LDAP_SERVER_DOWN && ldapError != LDAP_CONNECT_ERROR)
      break;
     
    closeLDAPConnection(server);
    if(!(server = openLDAPConnection(&registryServer)))
      return admin40_error_std(sn, rq, (char*)"unable to open LDAPConnection"); /*i18n*/
  } while (server != NULL && ++tries < 2);

  if(ldapError != LDAPU_SUCCESS)
    {
      closeLDAPConnection(server);

      if ((ldapError == LDAP_CONNECT_ERROR) || (ldapError == LDAP_SERVER_DOWN))
	return check_auth_tasks_cache(entryDN, userdn, pb, rq->reqpb, 0); /* Try the cache, ignoring expiration */

      return admin40_error(sn, rq, PROTOCOL_UNAUTHORIZED, (char*)"invalid user credentials"); /*i18n*/
    }     

  /* Attempt to read the entry data -- involves DS eval of entry ACI for userDN.
   * If we are able to read it, then access is granted.
   */

#if 0
  if ((ldapError = ldap_search_s(server, entryDN, LDAP_SCOPE_BASE, NS_EXEC_REF_QUERY, searchAttributes, 0, &result)) != LDAP_SUCCESS)
    {
      closeLDAPConnection(server);
      return admin40_error(sn, rq, PROTOCOL_FORBIDDEN, "task invocation denied"); /*i18n*/
    }
#endif /* 0 */

  tries = 0;
  do {
    ldapError = ldap_search_s(server, entryDN, LDAP_SCOPE_BASE, NS_EXEC_REF_QUERY, searchAttributes, 0, &result);
    if(ldapError != LDAP_SERVER_DOWN && ldapError != LDAP_CONNECT_ERROR)
      break;
     
    closeLDAPConnection(server);
    if(!(server = openLDAPConnection(&registryServer)))
      return admin40_error_auth(sn, rq, (char*)"unable to open LDAPConnection"); /*i18n*/
  } while (server != NULL && ++tries < 2);

  if (ldapError != LDAPU_SUCCESS) {
    closeLDAPConnection(server);
    ereport(LOG_FAILURE, (char*)"admin40_task_access_eval(): task search error, error = %d, entryDN = %s", ldapError, entryDN);
    return REQ_ABORTED;
  }

  /* push execRef and execRefArgs into request pblock */
  if (!(e = ldap_first_entry(server, result)))
    {
      /* 
       * Search succeeded, but no entries returned - this user wasn't allowed to access this task
       */
      closeLDAPConnection(server);
      return admin40_error(sn, rq, PROTOCOL_FORBIDDEN, (char*)"task invocation denied"); /*i18n*/
    }

  if ((vals = ldap_get_values(server, e, NS_EXEC_REF)) == NULL)
    {
      closeLDAPConnection(server);
      ereport(LOG_FAILURE, (char*)"admin40_task_access_eval(): unable to find %s", NS_EXEC_REF);
      return REQ_ABORTED;
    }

  if ((p = strchr(vals[0], '?')) != NULL)
    {
      *(p++) = '\0';
      pblock_nvinsert("execRefArgs", p, pb);
    }
  pblock_nvinsert("execRef", vals[0], pb);
  {
    char *test;
    test = pblock_findval("execRef", pb);
  }

  if ((vals2 = ldap_get_values(server, e, NS_LOG_SUPPRESS)) && !STRCASECMP(vals2[0], LOG_SUPPRESS_ON_VALUE))
    pblock_nvinsert("logSuppress", "true", rq->reqpb);

  ldap_msgfree(result);

  /* setup static task cache entries for userdn */
  populate_task_cache_entries(userdn, server);

  closeLDAPConnection(server);

  /* won't get this far if vals is NULL.  Play safe
     for future changes, check for null any way. */
  if (vals) {
    ldap_value_free(vals);
  }

  if (vals2) {
    ldap_value_free(vals2);
  }

  return REQ_PROCEED;
}


#ifdef XP_WIN32
#include "frame/http.h"
#define S_IXUSR 00100 /* execute permission: owner */
#define S_IXGRP 00010 /* execute permission: group */
#define S_IXOTH 00001 /* execute permission: other */
#endif /* XP_WIN32 */

#define IS_EXECUTABLE(mode) (((mode)&S_IXUSR)||((mode)&S_IXGRP)||((mode)&S_IXOTH))


/* ---------------------------- admin_uri2path ----------------------------- */

/*
 * Turns the requested URI into a path.
 */

#ifdef __cplusplus
extern "C"
#endif
NSAPI_PUBLIC int 
admin40_uri2path(pblock *pb, Session *sn, Request *rq)
{
  char *uri = pblock_findval("uri", rq->reqpb);
  char *root = pblock_findval("root", pb);
  pb_param *path = pblock_find("ppath", rq->vars);
  char *t, *u, *ntrsubdir = NULL;
  int pass = 0;
  char *servid, *subdir;
  struct stat *finfo;
  char *siepwd;
  char *ulevel;

  /*
   * Insert SIE password value discovered during server startup
   * (either by adm.conf or prompting the user).
   */

  siepwd = admGetCachedSIEPWD();
  if(siepwd)
    pblock_nvinsert("siepwd", siepwd, rq->vars);

  /*
   * Insert special "Admin-Server" name/value pair in the server header 
   * so Console knows what type of Admin Server this is.
   */

  pblock_nvinsert("Admin-Server", ADMSERV_VERSION_STRING, rq->srvhdrs);

  /*
   * DT 12/01/97 - Map / to the applet index page, disabling direct
   * link to the 3.0 index cgi.
   * DT 4/27/98 - Map / to the download CGI for the AS home page.
   */
#   define DIST_CGI_PATH (char*)"/dist/download"
  /* The server calls all NameTrans functions once before the doc root
	 is defined so this fails
  */
  if (!pblock_findval("host", rq->headers)) {
	  return REQ_PROCEED;
  } else {
	  if ((uri[0] == '/') && (uri[1] == '\0')) {
		  return admin40_redirect(sn, rq, (char*)"/", DIST_CGI_PATH);
	  }
  }

  /* DT 11/11/97 Support for $SERVER_ROOT/java directory */
  if (!STRNCASECMP(uri, JAVA_DIR, strlen(JAVA_DIR)))
    {
      if (strlen(uri) == strlen(JAVA_DIR))
	goto bong; /* can't request the directory */

      t = (char *) MALLOC(strlen(serverroot) + strlen(uri) + 1);
      util_sprintf(t, "%s%s", serverroot, uri);
      FREE(path->value);
      path->value = t;
      return REQ_PROCEED;
    }

  /* DT 3/4/98 Support for $SERVER_ROOT/dist directory */
  if (!STRNCASECMP(uri, DIST_DIR, strlen(DIST_DIR)))
    {

      if (strlen(uri) == strlen(DIST_DIR))
	goto bong; /* can't request the directory */
       
      if (!STRCASECMP(uri, DIST_CGI_PATH))
	pblock_nvinsert("content-type", "magnus-internal/cgi", rq->srvhdrs);
      else
	pblock_nvinsert("content-type", DOWNLOAD_CONTENT_TYPE, rq->srvhdrs);

      t = (char *) MALLOC(strlen(serverroot) + strlen(uri) + 1);
      util_sprintf(t, "%s%s", serverroot, uri);
      FREE(path->value);
      path->value = t;
      return REQ_PROCEED;
    }

  /* Support for $SERVER_ROOT/clients directory */
  if (!STRNCASECMP(uri, CLIENTS_DIR, strlen(CLIENTS_DIR)))
    {
      if (strlen(uri) == strlen(CLIENTS_DIR)) { 
          goto bong; /* can't request the directory */
      }

      t = (char *) MALLOC(strlen(serverroot) + strlen(uri) + 1 + 4); /* +4 is for appending .bat or .exe */
      util_sprintf(t, "%s%s", serverroot, uri);
       
      /* if execute bit is set, then set the cgi content-type */
	  finfo = request_stat_path(t, rq);

	  /* On Windows, check if there is CGI.exe or CGI.bat if CGI not found */
#ifdef XP_WIN32
      if (finfo==NULL) {
        util_sprintf(t, "%s%s.exe", serverroot, uri);
        if ((finfo = request_stat_path(t, rq)) == NULL) {
            util_sprintf(t, "%s%s.bat", serverroot, uri);
            finfo = request_stat_path(t, rq);
        }
      }
#endif
      if (finfo) {
          FREE(path->value);
          path->value = t;
          if (IS_EXECUTABLE(finfo->st_mode)) {
              pblock_nvinsert("content-type", "magnus-internal/cgi", rq->srvhdrs);
          }
          return REQ_PROCEED;
      }
      else { /* file not found means client app is not installed */
          FREE(t);
#     define CLIENT_APP_NOT_INSTALLED "/clients/not_installed.html"
          return admin40_redirect(sn, rq, uri, CLIENT_APP_NOT_INSTALLED);
      }      
    }

  /* DT 11/11/97 Support for $SERVER_ROOT/manual directory */
#   define HELP_CGI_PATH   "/manual/help/help"
#   define HELPSRCH_CGI_PATH   "/manual/help/helpsrch"
  if (!STRNCASECMP(uri, MANUAL_DIR, strlen(MANUAL_DIR)))
    {
      if (strlen(uri) == strlen(MANUAL_DIR))
	goto bong; /* can't request the directory */

      if (!STRCASECMP(uri, HELP_CGI_PATH) || !STRCASECMP(uri, HELPSRCH_CGI_PATH))
	pblock_nvinsert("content-type", "magnus-internal/cgi", rq->srvhdrs);

      t = (char *) MALLOC(strlen(serverroot) + strlen(uri) + 1);
      util_sprintf(t, "%s%s", serverroot, uri);
      FREE(path->value);
      path->value = t;
      return REQ_PROCEED;
    }

  /* 03/23/05 Support for $SERVER_ROOT/favicon.ico */
  if (!STRNCASECMP(uri, FAVICON, strlen(FAVICON)))
    {
      pblock_nvinsert("content-type", "image/x-icon", rq->srvhdrs);

      t = (char *) MALLOC(strlen(serverroot) + strlen(uri) + 1);
      util_sprintf(t, "%s%s", serverroot, uri);
      FREE(path->value);
      path->value = t;
      return REQ_PROCEED;
    }

  /* Initialize the table of exempt URIs */
  if (!exempt_uris) {
	  get_exempted_uris(pb, rq);
  }

  /* Check if the URI is exempted from task evaluation */
  if (is_exempt_uri(uri)) {
      return REQ_NOACTION;
  }

  if (!STRNCASECMP(uri, "/admin-serv/tasks/icons", strlen("/admin-serv/tasks/icons"))) {
      char *icons;

      icons = (char *)MALLOC(strlen(serverroot) + 1 + 3 + 1 + 5 + 1 + 5 + 1 + 5 + 1 + strlen(strrchr(uri, '/') + 1) + 1);

      util_sprintf(icons, "%s%cbin%cadmin%cadmin%cicons%c%s", 
		   serverroot,
		   FILE_PATHSEP,
		   FILE_PATHSEP,
		   FILE_PATHSEP,
		   FILE_PATHSEP,
		   FILE_PATHSEP,
		   (strrchr(uri, '/') + 1));
      FREE(path->value);
      path->value = icons;

      pblock_nvinsert("content-type", "image/gif", rq->srvhdrs);

      return REQ_PROCEED;
    }

    /* Munge the URL and break it into serverid and subdir. */
    u = STRDUP(path->value);
    if((u[0]) != '/')
      goto bong;
    servid = u+1;
    if(!(subdir = strchr(servid, '/')))
      goto bong;
    *(subdir++) = '\0';
    
    /* set product in rq->vars, or everything breaks. DT 12/12/97 */
    pblock_nvinsert("product", servid, rq->vars);
   
    /* DT 4/8/98 Authentication is no longer a task. Hardcoded to userauth. */
#   define AUTH_IDENTIFIER "authenticate"
#   define AUTH_CGI        "userauth"
    if (!STRNCASECMP(subdir, AUTH_IDENTIFIER, strlen(AUTH_IDENTIFIER)))
      {
	char *authcgi;
 
	/* authcgi = ${SERVERROOT}/bin/admin/admin/bin/${execRef} */
	authcgi = (char *)MALLOC(strlen(serverroot)+1+3+1+5+1+5+1+3+1+strlen(AUTH_CGI)+1);
	sprintf(authcgi, "%s%cbin%cadmin%cadmin%cbin%c%s", serverroot, FILE_PATHSEP, FILE_PATHSEP, FILE_PATHSEP, FILE_PATHSEP, FILE_PATHSEP, AUTH_CGI);
 
	FREE(path->value);
	path->value = authcgi;
 
	pblock_nvinsert("content-type", "magnus-internal/cgi", rq->srvhdrs);
 
	return REQ_PROCEED;
      }
 
    /* MK 8/1/00
     * Treat StartConfigDS as a special task. This will be the case if the user was
     * authenticated as the local superuser (admpw) and the server-id is "admin-serv"
     */
#   define STARTDS_IDENTIFIER "tasks/operation/StartConfigDS"
#   define STARTDS_CGI        "start_config_ds"
    if (!STRNCASECMP(subdir, STARTDS_IDENTIFIER, strlen(STARTDS_IDENTIFIER)) &&
        !STRCMP(servid, "admin-serv") &&  
        /* is local sueruser */
        (ulevel = pblock_findval("ADM_UserLevel", rq->vars)) && !STRCMP(ulevel, LOCAL_SUPER_NAME))
      {
	char *cgi;
 
	cgi = (char *)MALLOC(strlen(serverroot)+1+3+1+5+1+5+1+3+1+strlen(STARTDS_CGI)+1);
	sprintf(cgi, "%s%cbin%cadmin%cadmin%cbin%c%s", serverroot, FILE_PATHSEP, FILE_PATHSEP, FILE_PATHSEP, FILE_PATHSEP, FILE_PATHSEP, STARTDS_CGI);
 
	FREE(path->value);
	path->value = cgi;
 
	pblock_nvinsert("content-type", "magnus-internal/cgi", rq->srvhdrs);
 
	/* Show this in the error log*/
	log_error(LOG_INFORM, (char*)"admin40_uri2path", NULL, rq, (char*)"StartConfigDS requested by the local superuser"); 

	return REQ_PROCEED;
      }
 
    /* DT 3/15/98 Runtime command support. */
#   define RUNTIME_COMMAND_BASE (char*)"commands/"
    if (!STRNCASECMP(subdir, RUNTIME_COMMAND_BASE, strlen(RUNTIME_COMMAND_BASE)))
      {
	char *name;

	if (admin40_task_access_eval(pb, sn, rq) != REQ_PROCEED)
	  goto bong;

	name = subdir + strlen(RUNTIME_COMMAND_BASE);
	pblock_nvinsert("command-name", name, rq->reqpb);
	pblock_nvinsert("content-type", RUNTIME_COMMAND_CONTENT_TYPE, rq->srvhdrs);
	return REQ_PROCEED;
      }

    /* DT - 2/2/98
     * New task uri2path code
     */
    if (!STRNCASECMP(subdir, TASK_IDENTIFIER, strlen(TASK_IDENTIFIER)))
      {
	char *truepath;
	char *execRef;
	char *execRefArgs;

	if (admin40_task_access_eval(pb, sn, rq) != REQ_PROCEED)
	  goto bong;

	execRef     = pblock_findval("execRef", pb);
	execRefArgs = pblock_findval("execRefArgs", pb);

	/* For servlets execRef will start with "java:" */
	if (strstr(execRef, "java:") == execRef) {
            
	  pblock_nvinsert("content-type", "magnus-internal/servlet", rq->srvhdrs);
	  set_servlet_properties(pb, sn, rq);

	  {
	    /* 
	     * 442309 The task URIs should be case insensitive.
	     * Convert url to lower-case letters, as mappings in 
	     * rules.properties and servlets.properties are lower-case.
	     */
	    char *uri =  pblock_findval("uri", rq->reqpb);
	    char *c = uri;
	    for (c=uri; c && *c; c++) {
	      *c = tolower(*c);
	    }
	  }

	  return REQ_NOACTION; 
	}

	if ((t = strchr(servid, '-')))
	  *t = '\0';
       
	/* truepath = ${SERVERROOT}/bin/[server-type]/admin/bin/${execRef} */
	truepath = (char *)MALLOC(strlen(serverroot)+1+3+1+strlen(servid)+1+5+1+3+1+strlen(execRef)+1);
	sprintf(truepath, "%s%cbin%c%s%cadmin%cbin%c%s", serverroot, FILE_PATHSEP, FILE_PATHSEP, servid, FILE_PATHSEP, FILE_PATHSEP, FILE_PATHSEP, execRef);

	FREE(path->value);
	path->value = truepath;
    
	pblock_nvinsert("content-type", "magnus-internal/cgi", rq->srvhdrs);
    
	/* Make sure it's not a directory, Mike's way. */
	finfo = request_stat_path(path->value, rq);
	if (finfo)
	  {
	    if(S_ISDIR(finfo->st_mode))
	      {
		pblock_nvinsert("admin-error", "Invalid URL - is a directory", rq->vars);
		return REQ_ABORTED;
	      }
	  }

	/* Append execRefArgs to query, if applicable. */
	if (execRefArgs != NULL)
	  {
	    char *new_query;
	    pb_param *query = pblock_find("query", rq->reqpb);
    
	    if (query != NULL)
	      {
		new_query = (char*)MALLOC(strlen(query->value) + 1 + strlen(execRefArgs) + 1);
		sprintf(new_query, "%s&%s", execRefArgs, query->value);
		FREE(query->value);
		query->value = new_query;
	      }
	    else
	      {
		new_query = STRDUP(execRefArgs);
		pblock_nvinsert("query", new_query, rq->reqpb);
	      }
	  }
    
	return REQ_NOACTION;
      }
    
 bong:
    FREE(u);
    pblock_nvinsert("admin-error", "uri2path: Invalid URL!", rq->vars);
    return REQ_ABORTED;
}

/**
 * Initializes the hashtable of URIs which are not evaluated as DS tasks
 * by reading a file in the admin-serv/config folder
 * Returns the number of exempted URIs
 */
static int get_exempted_uris(pblock *param, Request *rq)
{
	FILE *fp = NULL;
    char line[BIG_LINE];
    char filename[BIG_LINE];
	int nExempt = 0;

	/* Form the config file name */
    sprintf(filename, "%s%cadmin-serv%cconfig%c%s", serverroot,
			FILE_PATHSEP, FILE_PATHSEP, FILE_PATHSEP, EXEMPTION_FILE);
	exempt_uris = HashTableCreate();
	fp = fopen(filename, "r");
	if (NULL != fp) {
		while(!feof(fp)) {
			char name[BIG_LINE];
			char *value;
			int len;
			fgets(line, sizeof(line), fp);
			/* Consider anything with # to be a comment line */
			if (strchr(line, '#')) {
				continue;
			}
			len = strlen(line) - 1;
			while((len >= 0) && isspace(line[len])) {
				line[len] = 0;
				len--;
			}
			if (len < 1) {
				continue;
			}
			strcpy(name, line);
			value = (char *)MALLOC(strlen(name+1));
			strcpy(value, name);
			HashTableInsert(exempt_uris, name, value);
			nExempt++;
/*
			log_error(LOG_INFORM, "get_exempted_uris", NULL, rq,
			  "Hashtable value of [%s] is [%s]", name, value);
*/
		}
		fclose(fp);
	} else {
/*
		log_error(LOG_INFORM, (char*)"get_exempted_uris", NULL, rq,
				  (char*)"Could not open file [%s]", filename);
*/
	}
	return nExempt;
}

static int
is_exempt_uri(const char *uri)
{
	int is_exempt = 0;
	if (uri) {
		char base[BIG_LINE];
		char *term;
		strcpy(base, uri);
		term = strrchr(base, '/');
		if (term && (term > base)) {
			term[1] = 0;
		}
		is_exempt = (NULL != HashTableFind(exempt_uris, base));
	}
	return is_exempt;
}

static int
check_auth_users_cache(char *user, char *pw, Request *rq, long now)
{
  /* Attempt to authenticate using auth_users cache */

  UserCacheEntry *cache_entry = NULL;

  if (!auth_users) return REQ_NOACTION; /* Failed. */

  cache_entry =  (UserCacheEntry*)HashTableFind(auth_users, user);

  if (!cache_entry || STRCMP(cache_entry->userPW, pw) || ((now - cache_entry->createTime) > cacheLifetime))
    return REQ_NOACTION; /* Failed. */

  pblock_nvinsert("userdn", cache_entry->userDN, rq->vars);
  pblock_nvinsert("auth-type", "basic", rq->vars);
  pblock_nvinsert("auth-password", pw, rq->vars);
  if (!strchr(user, '='))
    pblock_nvinsert("auth-user", user, rq->vars);
  pblock_nvinsert(AUTHENTICATION_LDAP_URL, cache_entry->ldapURL, rq->vars);

  return REQ_PROCEED;
}

static int
create_auth_users_cache_entry(char *user, char *userDN, char *userPW, char *ldapURL)
{
  UserCacheEntry *cache_entry;

  if (!(cache_entry = (UserCacheEntry *)HashTableFind(auth_users, user)))
    {
      cache_entry = (UserCacheEntry*)malloc(sizeof(UserCacheEntry));
      HashTableInsert(auth_users, user, cache_entry);
    }
  else
    {
      free(cache_entry->userDN);
      free(cache_entry->userPW);
      free(cache_entry->ldapURL);
    }

  cache_entry->userDN     = strdup(userDN);
  cache_entry->userPW     = strdup(userPW);
  cache_entry->ldapURL    = strdup(ldapURL);
  cache_entry->createTime = time(0);

  return 1;
}



/* ------------------------------ auth_basic ------------------------------ */

int 
auth_basic(pblock *param, Session *sn, Request *rq)
{
  char *pwfile, *grpfile, *type, *auth, *user, *pw;
  char *pwfn, *grpfn;
  pblock *npb;
  pb_param *pp;
  int ret;

  /* Although this is authorization (which is not cacheable) the
   * check is not actually done until we call require-auth.  So 
   * this part is cacheable; require-auth can be cacheable if the
   * user has limited the auth to only affect a certain set of 
   * paths.
   */
  rq->directive_is_cacheable = 1;

  if(request_header((char*)"authorization", &auth, sn, rq) == REQ_ABORTED)
    return REQ_ABORTED;

  if(!auth)
    return REQ_NOACTION;

  type = pblock_findval("auth-type", param);
  pwfile = pblock_findval("userdb", param);
  grpfile = pblock_findval("groupdb", param);
  pwfn = pblock_findval("userfn", param);
  grpfn = pblock_findval("groupfn", param);

  if((!type) || (!pwfile) || (!pwfn) || (grpfile && !grpfn)) {
    ereport(LOG_FAILURE, (char*)"missing parameter to basic-auth (need type, userdb, userfn)");
    protocol_status(sn, rq, PROTOCOL_SERVER_ERROR, NULL);
    return REQ_ABORTED;
  }

  /* Skip leading whitespace */
  while(*auth && (*auth == ' '))
    ++auth;
  if(!(*auth)) {
    protocol_status(sn, rq, PROTOCOL_FORBIDDEN, NULL);
    return REQ_ABORTED;
  }

  /* Verify correct type */
  if((strlen(auth) < 6) || strncasecmp(auth, "basic ", 6))
    return REQ_NOACTION;

  /* Skip whitespace */
  auth += 6;
  while(*auth && (*auth == ' '))
    ++auth;

  if(!*auth)
    return REQ_NOACTION;

  /* Uuencoded user:password now */
  if(!(user = _uudecode(auth)))
    return REQ_NOACTION;

  if(!(pw = strchr(user, ':'))) {
    FREE(user);
    return REQ_NOACTION;
  }
  *pw++ = '\0';

  npb = pblock_create(4);
  pblock_nvinsert("user", user, npb);
  pblock_nvinsert("pw", pw, npb);
  pblock_nvinsert("userdb", pwfile, npb);
  if(grpfile) 
    pblock_nvinsert("groupdb", grpfile, npb);
  pblock_nvinsert("fn", pwfn, npb);

  if ((ret = func_exec(npb, sn, rq)) != REQ_PROCEED)
    {
      goto bye;
    }

  pblock_nvinsert("auth-type", "basic", rq->vars);
  pblock_nvinsert("auth-user", user, rq->vars);
  pblock_nvinsert("auth-db", pwfile, rq->vars);
#if defined(XP_WIN32) || defined(MCC_ADMSERV)
  /* MLM - the admin server needs this password information,
   *       so I'm putting it back   */
  pblock_nvinsert("auth-password", pw, rq->vars);
#endif /* XP_WIN32 */

  if(grpfile) {
    pblock_nvinsert("groupdb", grpfile, npb);
    pp = pblock_find("fn", npb);
    FREE(pp->value);
    pp->value = STRDUP(grpfn);

    if( (ret = func_exec(npb, sn, rq)) != REQ_PROCEED )
      goto bye;
  }
  ret = REQ_PROCEED;
 bye:
  pblock_free(npb);
  FREE(user);
  return ret;
}


/* ----------------------------- auth_basic1 ------------------------------ */


/*
 * Compatibility function for 1.0 servers, to more powerful 1.1 SAF
 * accessible auth trans function
 */
int 
auth_basic1(pblock *param, Session *sn, Request *rq)
{
  pblock *npb = pblock_create(4);
  char *pwfile = pblock_findval("userfile", param);
  char *grpfile = pblock_findval("grpfile", param);
  char *type = pblock_findval("auth-type", param);
  char *pwdbm = pblock_findval("dbm", param);
  int ret;

  rq->directive_is_cacheable = 1;

  if(type) pblock_nvinsert("auth-type", type, npb);

  if(grpfile) {
    pblock_nvinsert("groupdb", grpfile, npb);
    pblock_nvinsert("groupfn", "simple-groupdb", npb);
  }
  if(pwfile) {
    pblock_nvinsert("userdb", pwfile, npb);
    pblock_nvinsert("userfn", "simple-userdb", npb);
  }
  else if(pwdbm) {
    pblock_nvinsert("userdb", pwdbm, npb);
    pblock_nvinsert("userfn", "dbm-userdb", npb);
  }
  ret = auth_basic(npb, sn, rq);
  pblock_free(npb);
  return ret;
}



/*
 * Check the admpw file, if specified, along with the host and ip restrictions.
 * 
 * If they match, then this user is the super user, so treat them accordingly.
 * If they do not match, then allow the ACL code to deal with them.
 */
int 
admin40_check_admpw(pblock *pb, Session *sn, Request *rq)
{
  pblock *execpb  = NULL;
  int ret         = 0;
  char *auth      = NULL;
  char *admpwfile = pblock_findval("admpwfile", pb);

  if(!admpwfile) {
    admpwfile = (char *)malloc(strlen(serverroot) + strlen("/admin-serv/config/admpw") + 2);
    sprintf(admpwfile, "%s%cadmin-serv%cconfig%cadmpw", serverroot, FILE_PATHSEP, FILE_PATHSEP, FILE_PATHSEP);
  }

  if (!admpwfile)
    return REQ_ABORTED;

  /* First, check to see if they authenticated at all. */
  if(request_header((char*)"authorization", &auth, sn, rq) == REQ_ABORTED)
    return REQ_ABORTED;

  /* Run:
   * AuthTrans fn="basic-auth" auth-type="basic" userfile=admpwfile */
  execpb = pblock_create(ADMIN_HASH_SIZE);
  pblock_nvinsert("auth-type", "basic", execpb); 
  pblock_nvinsert("userfile", admpwfile, execpb); 
  ret = auth_basic1(execpb, sn, rq);
  pblock_free(execpb);

  if (ret == REQ_ABORTED)
    return admin40_error(sn, rq, PROTOCOL_UNAUTHORIZED, (char*)"invalid user credentials"); /*i18n*/

  if (ret == REQ_NOACTION)
    {
      /* Call require-auth to enforce admin user name. */
      int ret;
      pblock *execpb = pblock_create(ADMIN_HASH_SIZE);
      pblock_nvinsert("fn", "require-auth", execpb);
      pblock_nvinsert("auth-type", "basic", execpb);
      /* We don't really know what the user name is, so just list
       * a nonsense value. */
      pblock_nvinsert("auth-user", "NS-no-such-user", execpb);
      pblock_nvinsert("realm", "Fedora Administration Server", execpb);
      ret = func_exec(execpb, sn, rq);
      pblock_free(execpb);
      return ret;
    }

  pblock_nvinsert("ADM_UserLevel", LOCAL_SUPER_NAME, rq->vars);
  return REQ_NOACTION;
}



static int
authenticate_user(LdapServerData *data, char *baseDN, char *user, char *pw, Session *sn, Request *rq)
{
  LDAP           *server;
  char           *uid, *userdn, *ldapURL;
  int             ldapError;
  int             pw_expiring = 0;
  char            pw_expiring_str[256];
  int tries = 0;

  if (!(server = openLDAPConnection(data)))
    return admin40_error_auth(sn, rq, (char*)"unable to open LDAPConnection to server"); /*i18n*/

#if 0
  if ((ldapError = admin40_ldap_auth_server(server, data)) != LDAPU_SUCCESS)
    {
      closeLDAPConnection(server);
      return admin40_error_auth(sn, rq, "unable to bind to server"); /*i18n*/
    }
#endif /* 0 */

  tries = 0;
  do {
    ldapError = admin40_ldap_auth_server(server, data);
    if(ldapError != LDAP_SERVER_DOWN && ldapError != LDAP_CONNECT_ERROR)
      break;
     
    closeLDAPConnection(server);
    if(!(server = openLDAPConnection(data)))
      return admin40_error_auth(sn, rq, (char*)"unable to open LDAPConnection to server"); /*i18n*/
  } while (server != NULL && ++tries < 2);
   
  if(ldapError != LDAPU_SUCCESS)
    {
      closeLDAPConnection(server);
      return admin40_error_auth(sn, rq, (char*)"unable to bind to server"); /*i18n*/
    }     

  /* The basic auth data may be either uid:pw or userDN:pw. The test for '='
   * is hopefully adequate to detect a DN...
   */
  if (!strchr(user, '='))
    {
      /* not a DN, so resolve the DN from the uid */

#if 0
      if ((ldapError = ldapu_find_userdn(server, user, baseDN ? baseDN : data->baseDN, &userdn)) != LDAPU_SUCCESS)
	{
	  closeLDAPConnection(server);

	  if ((ldapError == LDAP_CONNECT_ERROR) || (ldapError == LDAP_SERVER_DOWN))
	    return check_auth_users_cache(user, pw, rq, 0); /* DS down. Use the cache, ignoring entry expiration. */

	  return REQ_NOACTION; /* fall back to final check against admpw */
	}
#endif /* 0 */

      tries = 0;
      do {
	ldapError = ldapu_find_userdn(server, user, baseDN ? baseDN : data->baseDN, &userdn);
	if(ldapError != LDAP_SERVER_DOWN && ldapError != LDAP_CONNECT_ERROR)
	  break;
	
	closeLDAPConnection(server);
	if(!(server = openLDAPConnection(data)))
	  return admin40_error_auth(sn, rq, (char*)"unable to open LDAPConnection to server"); /*i18n*/
      } while (server != NULL && ++tries < 2);

      if (ldapError != LDAPU_SUCCESS)
	{
	  closeLDAPConnection(server);

	  if ((ldapError == LDAP_CONNECT_ERROR) || (ldapError == LDAP_SERVER_DOWN))
	    return check_auth_users_cache(user, pw, rq, 0); /* DS down. Use the cache, ignoring entry expiration. */

	  return REQ_NOACTION; /* fall back to final check against admpw */
	}

      uid = user;
    }
  else
    {
      /* it's a DN */

      userdn = user;
      uid    = NULL;

      /* strip the leading "ldap:", if present */

      if (!STRNCASECMP(userdn, LDAP_PREFIX, LDAP_PREFIX_LENGTH))
	{
	  if (strlen(userdn) > LDAP_PREFIX_LENGTH)
            userdn += LDAP_PREFIX_LENGTH;
	}
    }

#if 0
  if ((ldapError = admin40_ldap_auth_userdn_password(server, userdn, pw, &pw_expiring)) != LDAPU_SUCCESS)
    {
      closeLDAPConnection(server);

      if ((ldapError == LDAP_CONNECT_ERROR) || (ldapError == LDAP_SERVER_DOWN))
	return check_auth_users_cache(user, pw, rq, 0); /* DS down. Look in the cache, ignoring entry expiration. */

      return REQ_NOACTION; /* fall back to final check against admpw */
    }
#endif /* 0 */

  tries = 0;
  do {
    ldapError = admin40_ldap_auth_userdn_password(server, userdn, pw, &pw_expiring);
    if(ldapError != LDAP_SERVER_DOWN && ldapError != LDAP_CONNECT_ERROR)
      break;
     
    closeLDAPConnection(server);
    if(!(server = openLDAPConnection(data)))
      return admin40_error_auth(sn, rq, (char*)"unable to open LDAPConnection to server"); /*i18n*/
  } while (server != NULL && ++tries < 2);

  if (ldapError != LDAP_SUCCESS)
    {
      closeLDAPConnection(server);

      if ((ldapError == LDAP_CONNECT_ERROR) || (ldapError == LDAP_SERVER_DOWN))
	return check_auth_users_cache(user, pw, rq, 0); /* DS down. Look in the cache, ignoring entry expiration. */

      return REQ_NOACTION; /* fall back to final check against admpw */
    }

  closeLDAPConnection(server);

  pblock_nvinsert("userdn", userdn, rq->vars);
  pblock_nvinsert("auth-type", "basic", rq->vars);
  pblock_nvinsert("auth-password", pw, rq->vars);
  if (uid) pblock_nvinsert("auth-user", uid, rq->vars);
  pblock_nvinsert(AUTHENTICATION_LDAP_URL, ldapURL = formLdapURL(data), rq->vars);
  if (pw_expiring >= 0) {
    sprintf(pw_expiring_str, "%d", pw_expiring);
    pblock_nvinsert("auth-password-expire", pw_expiring_str, rq->vars);
  }

  create_auth_users_cache_entry(user, userdn, pw, ldapURL);

  if (ldapURL) {
    free(ldapURL);
  }

  return REQ_PROCEED;
}


#ifdef __cplusplus
extern "C"
#endif
NSAPI_PUBLIC int
admin40_authenticate(pblock *pb, Session *sn, Request *rq)
{
  char *auth, *user, *pw, *dummy = NULL;
  long  now;
  int   ret;
  char *auth_required;
  
  /* Initialize the table of exempt URIs */
  if (!exempt_uris) {
	  get_exempted_uris(pb, rq);
  }
  /* Check if the URI is exempted from task evaluation, if this is
     not the server bootstrap phase (not a real request) */
  if (pblock_findval("host", rq->headers)) {
	  char *uri = pblock_findval("uri", rq->reqpb);
	  if (uri && is_exempt_uri(uri)) {
		  return REQ_NOACTION;
	  }
  }
  auth_required = pblock_findval("ADM_AuthNotRequired", rq->vars);
  if(auth_required && (!strcmp(auth_required, "true"))) {
    param_free(pblock_remove("ADM_AuthNotRequired", rq->vars));
    pblock_nvinsert("ADM_AuthNotRequired", "false", rq->vars);
    return REQ_NOACTION;
  }
  if (auth_required) {
    param_free(pblock_remove("ADM_AuthNotRequired", rq->vars));
  }
  pblock_nvinsert("ADM_AuthNotRequired", "false", rq->vars);
  
  if (request_header((char*)"authorization", &auth, sn, rq) == REQ_ABORTED)
    return REQ_ABORTED;
  
  if (!auth)
    {
      /* Call require-auth to force basic auth */
      int ret;
      pblock *execpb = pblock_create(ADMIN_HASH_SIZE);
      pblock_nvinsert("fn", "require-auth", execpb);
      pblock_nvinsert("auth-type", "basic", execpb);
      /* We don't really know what the user name is, so just list
       * a nonsense value. */
      pblock_nvinsert("auth-user", "NS-no-such-user", execpb);
      pblock_nvinsert("realm", "Fedora Administration Server", execpb);
      ret = func_exec(execpb, sn, rq);
      pblock_free(execpb);
      return ret;
    }
  
  /* Skip leading whitespace */
  while(*auth && (*auth == ' '))
    ++auth;
  if(!*auth)
    return REQ_ABORTED;

  /* Verify correct type */
  if((strlen(auth) < 6) || strncasecmp(auth, "basic ", 6))
    return REQ_ABORTED;

  /* Skip whitespace */
  auth += 6;
  while(*auth && (*auth == ' '))
    ++auth;

  if(!*auth)
    return REQ_ABORTED;

  /* Uudecode user:password now */
  if(!(user = _uudecode(auth)))
    return REQ_ABORTED;

  /* find last ':' in auth field (since a DN might have a ldap: prefix) */
  for (pw = user + strlen(user) - 1 ; pw > user ; pw--)
    {
      if (*pw == ':')
	break;
    }
  if (pw == user) /* ':' not found */
    {
      FREE(user);
      return REQ_ABORTED;
    }
  *pw++ = '\0';

  /* DT 3/10/98
   * Enable the caches for lookups, and not just as a DS failover.
   */

  now = time(0);

  if (check_auth_users_cache(user, pw, rq, now) == REQ_PROCEED) {
    return REQ_NOACTION;
  }

  /* Cache lookup failed, or cache entry is expired. Try DS. */

  /* 5/18/98 DT -- new user/group separation code */

  /* First try o=NetscapeRoot in the registry server */
#  define NETSCAPE_ROOT_BASEDN (char*)"o=NetscapeRoot"

  ret = authenticate_user(&registryServer, NETSCAPE_ROOT_BASEDN, user, pw, sn, rq);
  if (ret != REQ_NOACTION) {
    return REQ_NOACTION;
  }

  /* Then, try the user/group server */

  /* If U/G Info is not available, try to get it */
  if (!(userGroupServer.host)) {
    buildUGInfo(&dummy);
  }

  if (userGroupServer.host) {
    ret = authenticate_user(&userGroupServer, NULL, user, pw, sn, rq);
    if (ret != REQ_NOACTION) {
      return REQ_NOACTION;
    }
  }

  /* last resort, check admpw file to see if s/he's the "emergency" user */
  ret = admin40_check_admpw(pb, sn, rq);

  /* Lastly, try the baseDN of the registry server 
     ret = authenticate_user(&registryServer, NULL, user, pw, sn, rq);
  */

  return ret;
}

int
admin40_runtime_command_init()
{
  commands = HashTableCreate();
  
  return (TRUE);
}

int
admin40_register_runtime_command(char *name, RuntimeCommandFn fn, void *arg)
{
  RuntimeCommandRecord *rcr;
  
  if (HashTableFind(commands, name))
    {
      ereport(LOG_FAILURE, (char*)"admin40_register_runtime_command(): attempt to register duplicate command (%s)", name);
      return FALSE;
    }

  rcr = (RuntimeCommandRecord*)malloc(sizeof(RuntimeCommandRecord));
  rcr->fn  = fn;
  rcr->arg = arg;
  HashTableInsert(commands, name, rcr);

  return TRUE;
}

int
admin40_runtime_command_exec(char *name, char *query, Request *rq)
{
  RuntimeCommandRecord *rcr = (RuntimeCommandRecord *)HashTableFind(commands, name);

  return (rcr && rcr->fn(name, query, rcr->arg, rq));
}

int
admin40_error_auth(Session *sn, Request *rq, char *reason)
{
  return REQ_NOACTION;
}

int
admin40_error_std(Session *sn, Request *rq, char *reason)
{
  return admin40_error(sn, rq, PROTOCOL_SERVER_ERROR, reason);
}

int
admin40_error(Session *sn, Request *rq, int code, char *reason)
{
  protocol_status(sn, rq, code, reason);
  return REQ_ABORTED;
}


/* Check if the caller hostname or ip address is disallowed */

#ifdef __cplusplus
extern "C"
#endif
NSAPI_PUBLIC int 
admin40_host_ip_check(pblock *pb, Session *sn, Request *rq)
{
  char * clientIP = pblock_findval("ip", sn->client);
  char * clf = pblock_findval("clf-request", rq->reqpb);

  /* Remove bogus error on startup */
  if (!clf)
    return REQ_NOACTION;

  if (clientIP) {
  } else {
    log_error(LOG_SECURITY, (char*)"admin40_host_ip_check", sn, rq,
	      (char*)"Unauthorized and unknown host, connection rejected");
    
    pblock_nvinsert("admin-error", "Unauthorized Host!", rq->vars);
    return REQ_ABORTED;
  }

  if (accessHosts && *accessHosts) {
    char *maxdns;
    maxdns = session_maxdns(sn);
    if (maxdns) {
      int rc;
      rc = shexp_casecmp(maxdns, accessHosts);
      FREE(maxdns);
      if (rc) {
      } else {
	FREE(clientIP);
	return REQ_NOACTION;
      }
    } else {
      PRNetAddr addr;
      ereport(LOG_WARN, (char*)"admin40_host_ip_check: session_maxdns could not resolve %s", clientIP);
      if (PR_SUCCESS == PR_StringToNetAddr(clientIP, &addr)) {
	char buf[PR_NETDB_BUF_SIZE];
	PRHostEnt hEntry;
	if (PR_SUCCESS == PR_GetHostByAddr(&addr, buf, sizeof(buf), &hEntry)) {
	  if (shexp_casecmp(hEntry.h_name, accessHosts)) {
	    char ** x;
	    for (x = hEntry.h_aliases; x && *x; x++) {
	      if (shexp_casecmp(*x, accessHosts)) {
	      } else {
		FREE(clientIP);
		return REQ_NOACTION;
	      }
	    }
	  } else {
	    FREE(clientIP);
	    return REQ_NOACTION;
	  }
	}
      }
    }
  }
  
  if (accessAddresses && *accessAddresses) {
    int rc;
    rc = shexp_casecmp(clientIP, accessAddresses);
    if (rc) {
    } else {
      FREE(clientIP);
      return REQ_NOACTION;
    }
  }
  
  log_error(LOG_SECURITY, (char*)"admin40_host_ip_check", sn, rq,
	    (char*)"Unauthorized host ip=%s, connection rejected", clientIP);
  
  pblock_nvinsert("admin-error", "Unauthorized Host!", rq->vars);
  FREE(clientIP);
  return REQ_ABORTED;
}

int sync_task_sie_data(char *name, char *query, void *arg, Request *rq);
int change_sie_password(char *name, char *query, void* arg, Request* rq);

#ifdef __cplusplus
extern "C"
#endif
NSAPI_PUBLIC int
admin40_task_eval_init(pblock *param, Session *sn, Request *rq)
{

  char *nsRoot;
  AdmldapInfo info;
  char  path[PATH_MAX];
  int   error;
  LDAP *server;
  char *userGroupLdapURL = NULL;
  char *userGroupBindDN = NULL;
  char *userGroupBindPW = NULL;
  char *dirInfoRef = NULL;
  char *errorInfo = NULL;
  int tries = 0;

  /* Set up caches */
  servers    = HashTableCreate();
  auth_users = HashTableCreate();
  auth_tasks = HashTableCreate();

  nsRoot = pblock_findval("nsRoot", param);

  /* cache entry expiration */
  cacheLifetime = atol(pblock_findval("cacheLifeTime", param));
  /*printf("Info: Cache expiration set to %d seconds\n", cacheLifetime);*/

  if (nsRoot) {
  } else {
    ereport(LOG_FAILURE, (char*)"admin40_task_eval_init(): NULL nsRoot");
    return FALSE;
  }

  {
    /* First, register the runtime commands */
    char *admin_root;

    admin_root = (char *)malloc(strlen(nsRoot) + strlen("/admin-serv/config") + 2);
    sprintf(admin_root, "%s%cadmin-serv%cconfig", nsRoot, FILE_PATHSEP, FILE_PATHSEP);
     
    admin40_runtime_command_init();
    admin40_register_runtime_command(RUNTIME_RESYNC_COMMAND, sync_task_sie_data, (void*)admin_root);
    admin40_register_runtime_command(CHANGE_SIEPWD_COMMAND, change_sie_password, (void*)admin_root);
  }

  sprintf(path, "%s%cadmin-serv%cconfig", nsRoot, FILE_PATHSEP, FILE_PATHSEP);

  info = admldapBuildInfo(path, &error);
  
  if (info) {
  } else {
    ereport(LOG_FAILURE, (char*)"admin40_task_eval_init(): unable to create AdmldapInfo");
    return FALSE;
  }
  
  serverroot = STRDUP(nsRoot);

  /* Registry DS setup */
  registryServer.host   = admldapGetHost(info);
  registryServer.port   = (admldapGetPort(info) < 0) ? 389 : admldapGetPort(info);
  registryServer.secure = (admldapGetSecurity(info)) ? 1 : 0;
  registryServer.baseDN = admldapGetBaseDN(info);
  registryServer.bindDN = admldapGetSIEDN(info);
  registryServer.bindPW = admldapGetSIEPWD(info);
  
  /* DT 5/18/98 Change for new User/Group stuff */
  
  /* Populate U/G Info */
  
  userGroupServer.host   = NULL;
  
  if (buildUGInfo(&errorInfo) != TRUE) {
    char scratch[2048];
    sprintf(scratch, "Warning! admin40_task_eval_init(): %s", errorInfo);
    ereport(LOG_FAILURE, scratch);
    return FALSE;
  }
  
  /* Register the admin server tasks */
  task_register_server(ADMIN_SERVER_ID, registryServer.bindDN);

  /* Populate the auth_tasks cache for the Local Admin */

  server = openLDAPConnection(&registryServer);

  if (server) {
  } else {
    destroyAdmldap(info);
    ereport(LOG_FAILURE, (char*)"Unable to open initial LDAPConnection to populate LocalAdmin tasks into cache.");
    return TRUE;
  }

  tries = 0;
  do {
    error = admin40_ldap_auth_server(server, &registryServer);
    if(error != LDAP_SERVER_DOWN && error != LDAP_CONNECT_ERROR)
      break;
     
    closeLDAPConnection(server);
    if(!(server = openLDAPConnection(&registryServer))) {
      destroyAdmldap(info);
      ereport(LOG_FAILURE, (char*)"Unable to open LDAPConnection to populate LocalAdmin tasks into cache.");
      return TRUE;
    }
  } while (server != NULL && ++tries < 2);
   
  if (error != LDAPU_SUCCESS)
    {
      destroyAdmldap(info);
      closeLDAPConnection(server);
      ereport(LOG_FAILURE, (char*)"Unable to bind as LocalAdmin to populate LocalAdmin tasks into cache.");
      return TRUE;
    }     

  populate_task_cache_entries(LOCAL_SUPER_NAME, server);

  /* Init for servlet tasks */
  servlet_env_init(server, &info);

  /* In 6.1 the servlet engine is always started always through directives in magnus.conf */ 
  /*servletsEnabled = pblock_findval("servletsEnabled", param);
  if (servletsEnabled && !STRNCASECMP(servletsEnabled, "yes", strlen("yes"))) {
    servlet_env_init(server, &info);
    if (servletCnt > 0) {
      start_servlet_engine(sn, rq);
    }
  }
  else {
    log_error(LOG_INFORM, "admin40_task_eval_init", sn, rq,
	      "servlets disabled, set parameter servletsEnabled=\"yes\" in obj.conf to enable"); 
  }*/

  closeLDAPConnection(server);
  destroyAdmldap(info);   

  return TRUE;
}

#ifdef __cplusplus
extern "C"
#endif
NSAPI_PUBLIC int 
admin40_host_ip_init(pblock *param, Session *sn, Request *rq) {

  int error;

  AdmldapInfo info;
  PsetHndl       pset;
  char *val;

  char path[PATH_MAX];

  sprintf(path, "%s%cadmin-serv%cconfig", serverroot, FILE_PATHSEP, FILE_PATHSEP);

  info = admldapBuildInfo(path, &error);
  
  if (info) {
  } else {
    char const msg[] = "admin40_host_ip_init(): unable to create AdmldapInfo (error code = %d)";
    char scratch[sizeof(msg)+ 20];
    sprintf(scratch, msg, error);
    ereport(LOG_FAILURE, scratch);
    return FALSE;
  }
  
  /*error = ADMSSL_Init(info, path);
    if (error) {
    fprintf(stderr, "admin40_host_ip_init(): LDAP security initialization failed (error code = %d)\n", error);
    return FALSE;
    }*/

  if (admldapGetSecurity(info)) {
    sslinit(info, serverroot, PR_FALSE);
    if (admldapBuildInfoSSL(info, &error)) {
    } else {
      char const msg[] = "admin40_host_ip_init(): unable to create secure AdmldapInfo (error code = %d)";
      char scratch[sizeof(msg) + 20];
      sprintf(scratch, msg, error);
      ereport(LOG_FAILURE, scratch);
      return FALSE;
    }
  }

  pset = psetCreateSSL((char*)"admin-serv", 
		       path,
		       NULL,
		       NULL,
		       &error);
  if (pset) {
  } else {
    char scratch[2048];
    sprintf(scratch, "admin40_host_ip_init(): PSET failure: Failed to create PSET handle (pset error = %s)", psetErrorString(error, NULL));
    ereport(LOG_FAILURE, scratch);
    return FALSE;
  }

  val = psetGetAttrSingleValue(pset, NS_ADMIN_ACCESS_HOSTS, &error);
  if (val) {
  } else {
    char scratch [2048];
    sprintf(scratch, "admin40_host_ip_init(): PSET failure: Could not retrieve access hosts attribute (pset error = %s)", psetErrorString(error, NULL));
    ereport(LOG_FAILURE, scratch);
    return FALSE;
  }

  accessHosts = strdup(val);

  val = psetGetAttrSingleValue(pset, NS_ADMIN_ACCESS_ADDRESSES, &error);
  if(val) {
  } else {
    char scratch[2048];
    sprintf(scratch,"admin40_host_ip_init(): PSET failure: Could not retrieve access addresses attribute (pset error = %s)", psetErrorString(error, NULL));
    ereport(LOG_FAILURE, scratch);
    return FALSE;
  }
  accessAddresses = strdup(val);

  if (accessHosts[0] || accessAddresses[0]) {
  } else {
    /* the configuration hasn't specified if anyone can connect.
     * That is a contradiction as an Admin Server that can't
     * let anyone connect is pointless! Consequently, I have
     * added code here to scan through the local.conf file
     * and extract the serverHostName and add it to the allowed
     * hosts file. rwagner
     */
    char const myAttr[] = "serverHostName";
    char line[BIG_LINE];
    char filename[BIG_LINE];
    FILE *f;
    
    sprintf(filename, "%s%cadmin-serv%cconfig%clocal.conf", serverroot, FILE_PATHSEP, FILE_PATHSEP, FILE_PATHSEP);
    if((f  = fopen(filename, "r" )) != NULL) {
      while(!feof(f)) {
	fgets(line, sizeof(line), f);
	
	if (STRNCASECMP(line, myAttr, sizeof(myAttr) - 1) == 0) {
	  char * c;
	  for (c = &line[sizeof(myAttr) - 1]; *c && *c != ':'; c++);
	  if (*c == ':') {
	    for (c++; *c && *c == ' '; c++);
	    if (*c) {
	      char * x;
	      for (x = c; *x; x++) {
		if (*x < ' ') {
		  *x = '\0';
		  break;
		}
	      }
	      FREE(accessHosts);
	      accessHosts = strdup(c);
	    }
	  }
	}
      }
      fclose(f);
      
      if (accessHosts[0]) {
	char const msg[] = "admin40_host_ip_init(): no hosts allowed or ip addresses allowed specified. Allowing %s nonetheless.";
	char scratch[2048];
	char buf[PR_NETDB_BUF_SIZE];
	PRHostEnt hEntry;
	sprintf(scratch, msg, accessHosts);
	ereport(LOG_INFORM, scratch);
	if (PR_SUCCESS == PR_GetHostByName(accessHosts, buf, sizeof(buf), &hEntry)) {
	  PRNetAddr addr;
	  if (PR_EnumerateHostEnt(0, &hEntry, 0, &addr) > 0) {
	    if (PR_SUCCESS == PR_NetAddrToString(&addr, buf, sizeof(buf))) {
	      FREE(accessAddresses);
	      accessAddresses = strdup(buf);
	    }
	  }
	}
      } else {
	ereport(LOG_WARN, (char*)"admin40_host_ip_init(): could not locate %s with %s", myAttr, filename);
      }
    }
  }
  
  if (accessHosts[0] || accessAddresses[0]) {
    if (accessHosts[0]) ereport(LOG_INFORM, (char*)"Access Host filter is: %s", accessHosts);
    if (accessAddresses[0]) ereport(LOG_INFORM, (char*)"Access Address filter is: %s", accessAddresses);
  } else {
    ereport(LOG_FAILURE, (char*)"admin40_host_ip_init(): the given configuration denies all hosts, thus nothing can connect!");
    return FALSE;
  }
  
  return TRUE;
}

/*
 * DT 3/15/98
 *
 * We need a way to send commands to a running admin server without simply
 * restarting the server (e.g. to resync to pick up new SIEs). The
 * \admin-serv\commands\* url namespace is to be used for this purpose.
 */

#ifdef __cplusplus
extern "C"
#endif
NSAPI_PUBLIC int
admin40_runtime_commands(pblock *pb, Session *sn, Request *rq)
{
  char *name  = pblock_findval("command-name", rq->reqpb);
  char *query = pblock_findval("query", rq->reqpb);

  if (name == NULL || query == NULL) {
        return REQ_ABORTED;
  }

  if (admin40_runtime_command_exec(name, query, rq))
    {
      pblock_nvinsert("content-length", "0", rq->srvhdrs);
      protocol_status(sn, rq, PROTOCOL_OK, NULL);
      protocol_start_response(sn, rq);
      return REQ_PROCEED;
    }

  return REQ_ABORTED;
}


/* ----------------------------- admin40_ftype ------------------------------ */


/*
 * Allows a very fixed subset of types. Currently only JPEG, GIF, 
 * imagemap, and HTML. If a file is CGI, its type is set by uri2path.
 *
 * DT 11/11/97 Added support for .properties files.
 * DT 03/04/98 Added support for application/octet-stream download types,
 */

#ifdef __cplusplus
extern "C"
#endif
NSAPI_PUBLIC int
admin40_ftype(pblock *pb, Session *sn, Request *rq)
{
  char *path    = pblock_findval("path", rq->vars);
  char *curtype = pblock_findval("content-type", rq->srvhdrs);
  char *type    = NULL;
  char *uri = pblock_findval("uri", rq->reqpb);
  char *t, *u;
    
  /* If the current content-type is already set to DOWNLOAD_CONTENT_TYPE
   * or RUNTIME_COMMAND_CONTENT_TYPE, don't override it
   */
  if (curtype && (!STRCMP(curtype, DOWNLOAD_CONTENT_TYPE) ||
		  !STRCMP(curtype, RUNTIME_COMMAND_CONTENT_TYPE) ||
		  !STRCMP(curtype, SERVLET_CONTENT_TYPE)))
    {
      return REQ_PROCEED;
    }

  /* Get an extension (if there is one) first.  Can't trust the 
   * type set by the bin directory earlier because now there's 
   * class files in the bin directory too. */
  if((!(u = strrchr(path, FILE_PATHSEP))) || (!(t = strrchr(u, '.')))) {
    /* No extension?  If there's already a type, this must be the 
     * bin directory, proceed anyway. */
    if(curtype)  {
      return REQ_PROCEED;
    }  else  {
		/* Also OK if it is an exempted URI */
		if (uri && is_exempt_uri(uri)) {
			return REQ_PROCEED;
		}

      /* Else you're a loser, get an extension. */
      pblock_nvinsert("admin-error",
		      "no filename extension",
		      rq->vars);
      return REQ_ABORTED;
    }
  }
  ++t;

  if((!strcasecmp(t, "exe")) || (!strcasecmp(t, "cgi")))
    type = (char*)"magnus-internal/cgi";
  else if((!strcasecmp(t, "jpg")) || (!strcasecmp(t, "jpeg")))
    type = (char*)"image/jpeg";
  else if(!strcasecmp(t, "gif"))
    type = (char*)"image/gif";
  else if(!strcasecmp(t, "css"))
    type = (char*)"text/css";
  else if(!strcasecmp(t, (char*)"jsp"))
    type = (char*)"magnus-internal/jsp";
  else if((!strcasecmp(t, (char*)"html")) || (!strcasecmp(t, (char*)"htm")))
    type = (char*)"text/html";
  else if(!strcasecmp(t, (char*)"map"))
    type = (char*)"magnus-internal/imagemap";
  else if((!strcasecmp(t, (char*)"js")) || (!strcasecmp(t, (char*)"mocha"))
	  || (!strcasecmp(t,(char*)"jss")))
    type = (char*)"application/x-javascript";
  else if((!strcasecmp(t, (char*)"class")) || (!strcasecmp(t, (char*)"zip")))
    type = (char*)"application/java-vm";
  else if(!strcasecmp(t, (char*)"properties"))
    type = (char*)"application/java-vm";
  else if(!strcasecmp(t, (char*)"ser"))
    type = (char*)"application/java-serialized-object";
  else if(!strcasecmp(t, (char*)"jar"))
    type = (char*)"application/java-archive";
  else {
    if(curtype)  {
      /* Must be a CGI from the bin directory.  
       * Don't override existing type. */
      return REQ_PROCEED;
    }  else  {
      pblock_nvinsert((char*)"admin-error",
		      (char*)"unrecognized type",
		      rq->vars);
      return REQ_ABORTED;
    }
  }
  if(curtype)  {
    param_free(pblock_remove("content-type", rq->srvhdrs));
  }  
  pblock_nvinsert("content-type", type, rq->srvhdrs);
  return REQ_PROCEED;
}




/* -------------------------- admin40_isdiropen ---------------------------- */

/* Check if the uri is for an open directory. If so, then bypass authentication. */

#ifdef __cplusplus
extern "C"
#endif
NSAPI_PUBLIC int 
admin40_isdiropen(pblock *pb, Session *sn, Request *rq)
{
  /* In kingpin, some directories are now supported without access control.
   * So, we check for them first.
     *
     * No basic auth for:
     *
     * /
     * /java
     * /clients
     * /manual
     * /dist
     *
     */
    
  char *uri = pblock_findval("uri", rq->reqpb);

#   define CLIENTS_DIR "/clients/"
#   define JAVA_DIR "/java/"
#   define MANUAL_DIR "/manual/"
#   define DIST_DIR "/dist"

  if ((!STRNCASECMP(uri, CLIENTS_DIR, strlen(CLIENTS_DIR))) ||
      (!STRNCASECMP(uri, JAVA_DIR, strlen(JAVA_DIR))) ||
      (!STRNCASECMP(uri, MANUAL_DIR, strlen(MANUAL_DIR))) ||
      (!STRNCASECMP(uri, DIST_DIR, strlen(DIST_DIR))) ||
      (!STRNCASECMP(uri, FAVICON, strlen(FAVICON))) ||
      ((uri[0] == '/') && (uri[1] == '\0')))
    pblock_nvinsert("ADM_AuthNotRequired", "true", rq->vars);
  else
    pblock_nvinsert("ADM_AuthNotRequired", "false", rq->vars);
  return REQ_NOACTION;
}


int
sync_task_sie_data(char *name, char *query, void *arg, Request *rq)
{
  AttrNameList serverlist = NULL, nl=NULL;
  AttributeList installlist = NULL;
  AdmldapInfo ldapInfo = NULL;
  int errorCode;
  PsetHndl tmp;
  char *admroot = (char*)arg;
  char errmsg[1024];
  int servercnt, i;
    
  ldapInfo = admldapBuildInfo(admroot, &errorCode);

  if (!ldapInfo) return FALSE;

  task_register_server(ADMIN_SERVER_ID, admldapGetSIEDN(ldapInfo));

  serverlist = getServerDNListSSL(ldapInfo);
  servercnt=0;
  if (serverlist) {
    while (serverlist[servercnt]) servercnt++;
  }
  if (servercnt) {
    for (i=0; i < servercnt; i++) {
      /* Create Pset for each individual server */
      tmp = psetRealCreateSSL(admldapGetHost(ldapInfo),
			      admldapGetPort(ldapInfo),
			      admldapGetSecurity(ldapInfo),
			      serverlist[i],
			      admldapGetSIEDN(ldapInfo),
			      admldapGetSIEPWD(ldapInfo),
			      NULL,
			      &errorCode);
      if (tmp) {
#          define SERVER_ID_ATTRIBUTE (char*)"nsServerID"

	int errorcode;
	char*  serverid = psetGetAttrSingleValue(tmp, SERVER_ID_ATTRIBUTE, &errorcode);

	if (!serverid) {
	  sprintf(errmsg, "sync_task_sie_data: Unable to find serverid for dn=\"%s\" (error code = %d)", serverlist[i], errorCode);
	  ereport(LOG_WARN, errmsg);
	  continue;
	}

	task_register_server(serverid, serverlist[i]);
      }
      else {
	sprintf(errmsg, "AdmInit: Failed to create psetHandle for %s (error code = %d)", serverlist[i], errorCode);
	ereport(LOG_FAILURE, errmsg);
      }
    }
  }

  /* DT 02/02/98
   * Register installed product tasks
   */
  if ((installlist = getInstalledServerDNListSSL(ldapInfo)))
    {
      int ii = 0;
      while (installlist[ii])
	{
	  char *productID = installlist[ii]->attrName;
	  char *productDN = installlist[ii]->attrVal[0];
	  task_register_server(productID, productDN);
          ii++;
	}
    }

  destroyAdmldap(ldapInfo);

  return TRUE;
}


static int update_ds(char *admroot, char *pwd, Request *rq);
static int update_admpwd(char *admroot, char *newuid, char *newpw);
static int update_adm_conf(char *admroot, char *newpw);

/* 
 * Miodrag (06-15-98)
 * The following metthod is called from the runtime command 
 * "change_sie_password" after the sie password is changed
 */
void
task_update_registry_server_bindpw(char *password)
{
  registryServer.bindPW = password;
}


/*
 * Modify SIE password 
 */
int change_sie_password(char *name, char *query, void* arg, Request* rq) {

  FILE *f;
  char *uid=NULL ,*pw=NULL, *col=NULL;
  char *newpw=query;
  static char filename[BIG_LINE];
  static char inbuf[BIG_LINE];
  char *admroot = (char *) arg;

  sprintf(filename, "%s/admpw", admroot);

  if (newpw==NULL) {
    ereport(LOG_FAILURE, (char*)"Wrong usage, password param missing");
    return 0;
  }

  if ((f=fopen(filename, "r"))==NULL) {
    ereport(LOG_FAILURE, (char*)"Can not open admpw file");
    return 0;
  }
  if (fgets(inbuf, sizeof(inbuf), f) == NULL) {
    fclose(f);
    ereport(LOG_FAILURE, (char*)"Can not read admpw file");
    return 0;
  }
  fclose(f);

  col = strchr(inbuf, ':');
  if (col == NULL) {
    ereport(LOG_FAILURE, (char*)"admpw file corrupted");
    return 0;
  }
	
  uid = inbuf; *col=0; pw=col+1;
	
  if (!update_ds(admroot, newpw, rq))  {
    return 0;
  }
  if (!update_adm_conf(admroot, newpw)) {
    return 0;
  }
  if (!update_admpwd(admroot, uid, admin40_pw_enc(newpw))) {
    return 0;
  }

  task_update_registry_server_bindpw(strdup(newpw));
  return 1;
}

/*
 * Modify admpw.
 */
static int update_admpwd(char *admroot, char *newuid, char *newpw) {

  FILE *f;
  int cnt;
  static char filename[BIG_LINE];
  static char inbuf[BIG_LINE];
  static char outbuf[BIG_LINE];

  sprintf(filename, "%s/admpw", admroot);

  f = fopen(filename, "w");
  if (f==NULL) {
    ereport(LOG_FAILURE, (char*)"Can not open admpw file for writing");
    return 0;
  }

  sprintf(outbuf,"%s:%s", newuid, newpw);

  cnt = fprintf(f,"%s", outbuf);

  if (cnt != (int)strlen(outbuf)) {
    ereport(LOG_FAILURE, (char*)"Failed to write to admpw");
    fclose(f);
    return 0;
  }
	
  fclose(f);
  return 1;
}

/*
 * Modify adm.conf
 */
static int update_adm_conf(char *admroot, char *newpw) {

  FILE *f;
  int i, modified=0;
  static char filename[BIG_LINE];
  char buf[80];
  int linecnt=0;
  char *lines[50];
  static char inbuf[BIG_LINE];


  sprintf(filename, "%s/adm.conf", admroot);

  f = fopen(filename, "r");
  if (f==NULL) {
    return 0;
  }

  while(fgets(inbuf, sizeof(inbuf), f) != NULL) {
    if (strstr(inbuf,"siepid:") == inbuf) { /* Line starts with "Password:" */
      sprintf(buf, "siepid: %s\n", newpw);
      lines[linecnt++] = strdup(buf);
      modified=1;
    }
    else {
      lines[linecnt++] = strdup(inbuf);
    }
  }	
  fclose(f);

  if (!modified) {
    admSetCachedSIEPWD(newpw);

    return 1;
  }

  f = fopen(filename, "w");
  if (f==NULL) {
    return 0;
  }

  for (i=0; i < linecnt; i++) {
    fprintf(f, "%s", lines[i]);
  }

  fclose(f);
  return 1;
}

/*
 * Modify userpassword in the DS
 */
static int update_ds(char *admroot, char *pwd, Request *rq) {
  
  int   errorCode;
  PsetHndl pset;
  char *binddn = 0;
  char *bindpw = 0;
  AdmldapInfo ldapInfo = NULL;

  /* Initialize the pset  */

  ldapInfo = admldapBuildInfo(admroot, &errorCode);

  if (!ldapInfo) {
    ereport(LOG_FAILURE,(char*)"AdmInit: Failed to read data from adm.conf");
    return 0;
  }

  if (admldapGetSecurity(ldapInfo)) {
    sslinit(ldapInfo, serverroot, PR_FALSE);
  }
  /*errorCode = ADMSSL_Init(ldapInfo, admroot);
    if (errorCode) {
    ereport(LOG_FAILURE, "Security Initialization Failed");
    return 0;
    }*/

  binddn = pblock_findval("userdn", rq->vars);
  bindpw = pblock_findval("auth-password", rq->vars);

  pset = psetCreateSSL((char*)"admin-serv", 
		       admroot,
		       binddn,
		       bindpw,
		       &errorCode);

  if (pset) {
  } else {
    const char msg[] = "PSET Creation Failed, err=%d";
    char scratch[sizeof(msg) + 20];
    sprintf(scratch, msg, errorCode);
    ereport(LOG_FAILURE, scratch);
    return 0;
  }

  errorCode = psetSetSingleValueAttr(pset, (char*)"userpassword", pwd);

  if (errorCode) {
    char const msg[] = "PSET SET Failed, err=%d";
    char scratch[sizeof(msg) + 20];
    psetDelete(pset);
    sprintf(scratch, msg, errorCode);
    ereport(LOG_FAILURE, scratch);
    return 0;
  }
  psetDelete(pset);
  return 1;
}


#ifdef __cplusplus
extern "C"
#endif
NSAPI_PUBLIC int 
admin40_flex_log(pblock *pb0, Session *sn, Request *rq)
{
  pblock *pb;
  int rc;

  if (pblock_findval("logSuppress", rq->reqpb)) {
    return REQ_PROCEED;
  }

  /* Must create a new pb for func_exec (see 600068) */
  pb = pblock_dup(pb0);

  param_free(pblock_remove("fn", pb));
  pblock_nvinsert("fn", "flex-log", pb);
   
  rc = func_exec(pb, sn, rq);

  pblock_free(pb);
  return rc;
}

/*
 * Set servlet propertes as request headers. 
 * 
 */
static int set_servlet_properties(pblock *pb, Session *sn, Request *rq){

  char buf[PATH_MAX];
  char *str;
  char *product = pblock_findval("product", rq->vars);

  pblock_nvinsert("admin.serverroot", serverroot,  rq->headers);

  sprintf(buf, "%s%cadmin-serv%cconfig", serverroot, FILE_PATHSEP, FILE_PATHSEP);
  pblock_nvinsert("admin.adminroot", buf,  rq->headers);

  str = pblock_findval("userdn", rq->vars);
  pblock_nvinsert("admin.user.dn", (str ? str : ""),  rq->headers);
  str = pblock_findval("auth-type", rq->vars);
  pblock_nvinsert("admin.auth.type", (str ? str : ""), rq->headers);
  str = pblock_findval("auth-user", rq->vars);
  pblock_nvinsert("admin.auth.user", (str ? str : ""), rq->headers);
  str = pblock_findval("auth-password", rq->vars);
  pblock_nvinsert("admin.auth.password", (str ? str : ""), rq->headers);
    
  pblock_nvinsert("admin.configds.host", registryServer.host,  rq->headers);
  sprintf(buf, "%d", registryServer.port);
  pblock_nvinsert("admin.configds.port", buf,  rq->headers);
  pblock_nvinsert("admin.configds.secure", (registryServer.secure ? "on" : "off"),  rq->headers); 
  pblock_nvinsert("admin.configds.basedn", registryServer.baseDN,  rq->headers);

  pblock_nvinsert("admin.userds.host", userGroupServer.host,  rq->headers);
  sprintf(buf, "%d", userGroupServer.port);
  pblock_nvinsert("admin.userds.port", buf,  rq->headers);
  pblock_nvinsert("admin.userds.secure", (userGroupServer.secure ? "on" : "off"),  rq->headers); 
  pblock_nvinsert("admin.userds.basedn", userGroupServer.baseDN,  rq->headers);

  pblock_nvinsert("admin.product.id", product,  rq->headers);
  str = (char*)HashTableFind(servers, product);
  pblock_nvinsert("admin.product.dn", (str ? str : ""),  rq->headers);

  return 0;
}

/*
 * Starts servlet engine
 */
static int start_servlet_engine(Session *sn, Request *rq){

  pblock *execpb = NULL;
  int     ret    = 0;
  char  shlib[PATH_MAX];
  char *funcs= (char*)
    "NSServletEarlyInit,NSServletLateInit,NSServletNameTrans,NSServletService";

  sprintf(shlib, "%s/" SERVLET_PLUGIN_PATH, serverroot);
  execpb = pblock_create(ADMIN_HASH_SIZE);

  pblock_nvinsert("fn", "load-modules", execpb);
  pblock_nvinsert("shlib", shlib, execpb);
  pblock_nvinsert("funcs", funcs, execpb);
  pblock_nvinsert("shlib_flags", "(global|now)", execpb);
  ret = func_exec(execpb, sn, rq);
  if (ret != 0) {
    char msg[1024];
    sprintf(msg, "load-modules failed shlib=%s funcs=%s", shlib, funcs);
    ereport(LOG_FAILURE,msg);
    pblock_free(execpb);
    return ret;
  }

  param_free(pblock_remove("fn", execpb));
  param_free(pblock_remove("shlib", execpb));
  param_free(pblock_remove("funcs", execpb));
  param_free(pblock_remove("shlib_flags", execpb));

  pblock_nvinsert("fn", "NSServletEarlyInit", execpb);
  pblock_nvinsert("EarlyInit", "yes", execpb);
  ret = func_exec(execpb, sn, rq);
  if (ret != 0) {
    ereport(LOG_FAILURE, (char*)"NSServletEarlyInit failed");
    pblock_free(execpb);
    return ret;
  }

  param_free(pblock_remove("fn", execpb));
  param_free(pblock_remove("EarlyInit", execpb));

  pblock_nvinsert("fn", "NSServletLateInit", execpb);
  pblock_nvinsert("LateInit", "yes", execpb);
  ret = func_exec(execpb, sn, rq);
   
  if (ret != 0) {
    ereport(LOG_FAILURE, (char*)"NSServletLateInit failed");
  }

  pblock_free(execpb);
  return ret;
}

/*
 * A HashEnumFunc for processing sie and isie table, searches servlet tasks.
 * Find all servlets for the thaget serverDN using the filter (nsExecRef=java:*)
 * Populate the servlets table to be processed later by create_servlet_config()
 */
static void lookup_servlet_tasks(char *serverID, const void *serverDN, void *userdata)
{
  LDAP *server = (LDAP*) userdata;
  LDAPMessage  *result, *e;
  int           ldapError;
  char          *normServerDN;

  /* normalized DN is the same size or shorter than the original DN */
  normServerDN=(char*) MALLOC(strlen((char *)serverDN) + 1);
  adm_normalize_dn((char*)serverDN, normServerDN);

  ldapError = ldap_search_s(server, (char *)serverDN, LDAP_SCOPE_SUBTREE,
			    SERVLET_QUERY, searchServletAttrs, 0, &result);

  if (ldapError != LDAPU_SUCCESS) {
    ldap_perror(server, "lookup_servlet_tasks():task search error");
    return;
  }

  for (e = ldap_first_entry(server, result) ; e != NULL ; e = ldap_next_entry(server, e)) {

    char *taskDN, **vals;
    char *normTaskDN;
    ServletLookupData *servletData;
      
    taskDN = ldap_get_dn(server, e);
    if (!taskDN) {
      continue;  /* should never be null, but just in case */
    }

    normTaskDN = (char*) MALLOC(strlen(taskDN)+1);
    adm_normalize_dn(taskDN, normTaskDN);
    convert_to_lower_case(normTaskDN);
    ldap_memfree(taskDN);

    servletData = (ServletLookupData*)HashTableFind(servlets, normTaskDN);
    if (servletData) {
      /*
       * The servlet task entry is already in the table. It must belong to a SIE
       * and we are now processing ISIE which is a parent for the SIE. Keep the 
       * servlet as the SIE servlet.
       */
      continue;
    }

    servletData = (ServletLookupData*)MALLOC(sizeof(ServletLookupData));
    servletData->serverID = STRDUP(serverID);
    servletData->serverDN = normServerDN;

    vals = ldap_get_values(server, e, NS_EXEC_REF);
    if (!vals) {
      continue;  /* should never be null, but just in case */
    }
    servletData->nsExecRef = STRDUP(vals[0]);
    ldap_value_free(vals);

    vals = ldap_get_values(server, e, "cn");
    if (!vals) {
      continue;  /* should never be null, but just in case */
    }
    servletData->taskcn = STRDUP(vals[0]);
    convert_to_lower_case(servletData->taskcn);
    ldap_value_free(vals);

    HashTableInsert(servlets, normTaskDN, servletData);
    FREE(normTaskDN);
      
    servletCnt++;

  }

  ldap_msgfree(result);
}


/*
 * A HashEnumFunc for processing servlets table, creates servlet URI mapping files
 * servlets.properties and rules.properties.
 * A servlet task entry has nsExecRef attribute with a value in the following format:
 * java:<className>@<jar1>,<jar2>,..?<arg1=val1>,<arg2=val2>
 * This value is mapped to servlet properties 
 * servlets.<serverID>.code=className
 * servlets.<serverID>.classpath=<serverroot>/bin/<serverID>/admin/servlets:<serverroot>/bin/<serverID>/admin/servlets/jar1
 * servlets.<serverID>.initArgs=arg1=val1,arg2=val2
 * The jar list and the arg list are optional.
 */
static void create_servlet_config(char *servletDN, const void *servletData, void *userdata)
{
  ServletLookupData *data = (ServletLookupData *) servletData;
  char *uri;
  char *val, *className, *jarList, *initArgs, *c;
  char *configID;
  char classpath [2048];
  static int servletIdx = 1;    

  /* copy nsExecRef into a buffer skipping first 5 char (the "java:" prefix) */
  val = STRDUP(data->nsExecRef+5);

  className = val;
            
  if ((initArgs = strchr(val, '?')) != NULL) {         
    *initArgs++ = 0;
  }

  if ((jarList = strchr(className, '@')) != NULL) {
    *jarList++ = 0;
  }

  /* remove SIE/ISIE part from the servlet task DN */
  servletDN[ strlen(servletDN) - strlen(data->serverDN) -1 ] = 0;
     
  /*
   * servletDN is now in the form e.g. "cn=stop,cn=operation,cn=tasks" and we need to
   * convert it into URI format /tasks/operation/stop
   */
   
  /* First remove all "cn=" */
  c = servletDN;
  do {
    while (*c && *c != '=') {
      memmove(c, c+1, strlen(c+1)+1); /* shift left 1 char, removing all chars before = */
    }
    if (*c == '=') {
      memmove(c, c+1, strlen(c+1)+1); /* shift left 1 char to remove = */

      /* advance to the next element in the DN delimitted by ',' */
      while(*c && *c != ',') c++;
      if (*c == ',') c++; /* skip ',' */
    }
  } while (*c);

  /* ... then create URI by reversing the DN elements and replacing ',' with '/' */
      
  uri = (char*) MALLOC(1 + strlen(data->serverID) + 1 + strlen(servletDN) +1 ); 
  sprintf(uri, "/%s", data->serverID);
  while (c=strrchr(servletDN, ',')) {
    *c++ = 0;
    sprintf(uri,"%s/%s", uri, c);
  }

  sprintf(uri,"%s/%s", uri, servletDN); /* add the last element */

  /* configID is the name rules.properties is using to reference entries in servlets.properties */
  configID = (char*) MALLOC(strlen(data->serverID) + 5);
  sprintf(configID, "%s-%02d", data->serverID, servletIdx++);

  fprintf(rules_properties, "%s=%s\n",uri,configID);

  fprintf(servlets_properties, "servlet.%s.code=%s\n", configID,className);
 
  {
    /* servlets are located in <serverroot>/bin/<serverID>/admin/servlets directory */
    char *serverID = STRDUP(data->serverID);

    if (c=strchr(serverID, '-')) {
      *c=0;
    }
    sprintf(classpath,"%s/bin/%s/admin/servlets", serverroot, serverID);

    if (jarList) {
      char *jar;
      for (jar = strtok(jarList, ","); jar; jar=strtok(NULL, ",")) {
	sprintf(classpath,"%s%c%s/bin/%s/admin/servlets/%s",
		classpath, PATH_SEP, serverroot, serverID, jar);
      }
    }
    fprintf(servlets_properties, "servlet.%s.classpath=%s\n", configID,classpath);

    FREE(serverID);
  }

  if (initArgs) {
    fprintf(servlets_properties, "servlet.%s.initArgs=%s\n", configID,initArgs);
  }
            
  FREE(val);
  FREE(uri);
  FREE(configID);
}

/*
 * Find in the server root all tasks implemented as servlets. Create servlet URI mapping
 * files servlets.properties and rules.properties.
 */
static int servlet_env_init(LDAP *server, AdmldapInfo *ldapInfo)
{
  HashTable * sieTab =  HashTableCreate();
  HashTable * isieTab = HashTableCreate();
  char servlets_properties_path[PATH_MAX];
  char rules_properties_path[PATH_MAX];

  servlets = HashTableCreate();
  servletCnt = 0;

  /* Populate the hash table 'servers' with all SIEs and ISIEs in the current server root */
  populate_SIE_table(sieTab, ldapInfo);
  populate_ISIE_table(isieTab, ldapInfo);

  /*
   * Search servlets for each SIE/ISIE (A task can be created per SIE as well as per ISIE).
   * lookup_servlet_tasks() will populate the servlets table. The sieTab MUST be processed
   * before the isieTab as when searching the directory the tasks under a SIE will be also
   * visible under it's parent ISIE. 
   */
  HashTableEnumerate(sieTab,  lookup_servlet_tasks, server);
  HashTableEnumerate(isieTab, lookup_servlet_tasks, server);

  sprintf(servlets_properties_path, "%s/admin-serv/config/servlets.properties", serverroot);
  sprintf(rules_properties_path, "%s/admin-serv/config/rules.properties", serverroot);

  servlets_properties = fopen(servlets_properties_path, "w");
  if (servlets_properties == NULL) {
    char msg[PATH_MAX];
    sprintf(msg, "Servlet initialization failed: Can not open %s for writing", servlets_properties_path);
    ereport(LOG_FAILURE, msg);
    return FALSE;
  }

  rules_properties = fopen(rules_properties_path, "w");
  if (rules_properties == NULL) {
    char msg[PATH_MAX];
    sprintf(msg, "Servlet initialization failed: Can not open %s for writing", rules_properties_path);
    ereport(LOG_FAILURE, msg);
    fclose(servlets_properties);
    return FALSE;
  }
 
  fprintf(servlets_properties, PROP_HEADER1, "servlets.properties.default", "servlets.properties.default");
  fprintf(rules_properties, PROP_HEADER1, "rules.properties.default", "rules.properties.default");

  /* Copy the contents of servlets.proerties.default to servlets.properties. */
  {
    char defaults_path[PATH_MAX];
    FILE *defaults;
    char line [1024];
    sprintf(defaults_path, "%s.default", servlets_properties_path);
    defaults = fopen(defaults_path, "r");
    if (defaults) {
      while(fgets(line, sizeof(line), defaults)) {
	fprintf(servlets_properties, "%s", line);
      }
      fclose(defaults);
    }
  }
            
  /* Copy the contents of rules.proerties.default to rules.properties. */
  {
    char defaults_path[PATH_MAX];
    FILE *defaults;
    char line [1024];
    sprintf(defaults_path, "%s.default", rules_properties_path);
    defaults = fopen(defaults_path, "r");
    if (defaults) {
      while(fgets(line, sizeof(line), defaults)) {
	fprintf(rules_properties, "%s", line);
      }
      fclose(defaults);
    }
  }

  fprintf(servlets_properties, PROP_HEADER2);
  fprintf(rules_properties,    PROP_HEADER2);

  HashTableEnumerate(servlets, create_servlet_config, NULL);

  fclose(servlets_properties);
  fclose(rules_properties);

  return TRUE;
}

static int populate_SIE_table(HashTable *table, AdmldapInfo *ldapInfo)
{
  AttrNameList serverlist = NULL, nl=NULL;
  int errorCode;
  PsetHndl psetHndl;
  char errmsg[1024];
  int servercnt=0, i;

  HashTableInsert(table, ADMIN_SERVER_ID, admldapGetSIEDN(*ldapInfo));

  serverlist = getServerDNListSSL(*ldapInfo);

  if (serverlist) {
    while (serverlist[servercnt]) servercnt++;
  }

  for (i=0; i < servercnt; i++) {

    /* Create Pset for each individual server */
    psetHndl = psetRealCreateSSL(admldapGetHost(*ldapInfo),
				 admldapGetPort(*ldapInfo),
				 admldapGetSecurity(*ldapInfo),
				 serverlist[i],
				 admldapGetSIEDN(*ldapInfo),
				 admldapGetSIEPWD(*ldapInfo),
				 NULL,
				 &errorCode);

    if (psetHndl) {
      int errorcode;
      char*  serverid = psetGetAttrSingleValue(psetHndl, (char*)"nsServerID", &errorcode);
      if (!serverid) {
          sprintf(errmsg, "populate_SIE_table: Unable to find serverid for dn=\"%s\" (error code = %d)", serverlist[i], errorCode);
          ereport(LOG_WARN, errmsg);
          continue;
      }

      HashTableInsert(table, serverid, serverlist[i]);
    }
    else {
      sprintf(errmsg, "AdmInit: Failed to create psetHandle for %s (error code = %d)", serverlist[i], errorCode);
      ereport(LOG_FAILURE, errmsg);
    }
  }

  return TRUE;
}

static int populate_ISIE_table(HashTable *table, AdmldapInfo *ldapInfo)
{
  AttributeList installlist = NULL;    

  if ((installlist = getInstalledServerDNListSSL(*ldapInfo)))
    {
      int ii = 0;
      while (installlist[ii])
        {
	  char *productID = installlist[ii]->attrName;
	  char *productDN = installlist[ii]->attrVal[0];
	  HashTableInsert(table, productID, productDN);
	  ii++;
        }
    }

  return TRUE;
}


#ifdef __cplusplus
extern "C"
#endif
NSAPI_PUBLIC int 
admin40_check_ds_availability_init(pblock *pb, Session *sn, Request *rq) {


  char *nsRoot;
  char path[PATH_MAX];
  AdmldapInfo info;
  LDAP *server;
  int pw_exp;
  char message[BIG_LINE];

  int error, ldapError;
  char *errorInfo;

  nsRoot = pblock_findval("nsRoot", pb);
  if (!nsRoot)
    {
      ereport(LOG_WARN, (char*)"admin40_check_ds_availability_init(): WARNING: NULL nsRoot");
      return FALSE;
    }

  sprintf(path, "%s%cadmin-serv%cconfig", nsRoot, FILE_PATHSEP, FILE_PATHSEP);

  if (!(info = admldapBuildInfo(path, &error)))
    {
      sprintf(message, "admin40_check_ds_availability_init(): WARNING: unable to create AdmldapInfo (error code = %d)", error);
      ereport(LOG_WARN, message);
      return FALSE;
    }

  serverroot = STRDUP(nsRoot);

  /*error = ADMSSL_Init(info, path);*/  

  if (admldapGetSecurity(info)) {
    sslinit(info, serverroot, PR_FALSE);
    if (!admldapBuildInfoSSL(info, &error)) {
      sprintf(message, "admin40_check_ds_availability_init(): unable to create ssl AdmldapInfo (error code = %d)", error);
      ereport(LOG_WARN, message);
      return PR_FALSE;
    }
  }

  /* Registry DS setup */
  registryServer.host   = admldapGetHost(info);
  registryServer.port   = (admldapGetPort(info) < 0) ? 389 : admldapGetPort(info);
  registryServer.secure = (admldapGetSecurity(info)) ? 1 : 0;
  registryServer.baseDN = admldapGetBaseDN(info);
  registryServer.bindDN = admldapGetSIEDN(info);
  registryServer.bindPW = admldapGetSIEPWD(info);

  if (!(server = openLDAPConnection(&registryServer)))
    {
      destroyAdmldap(info);
      ereport(LOG_WARN, (char*)"admin40_check_ds_availability_init(): WARNING: Unable to open LDAP connection to Configuration DS.");
      return FALSE;
    }

  ldapError = admin40_ldap_auth_userdn_password(server, NULL, NULL, &pw_exp);
  if(ldapError == LDAP_SERVER_DOWN || ldapError == LDAP_CONNECT_ERROR) {
    sprintf(message, "admin40_check_ds_availability_init(): WARNING: Configuration Directory Server is down or unreachable (%s)",
	    ldap_err2string(ldapError));
    ereport(LOG_WARN, message);
    return FALSE;
  }

  /* Config DS is running, see if U/G DS is running */

  userGroupServer.host   = NULL;

  if (buildUGInfo(&errorInfo) != TRUE) {
    sprintf(message, "admin40_check_ds_availability_init(): WARNING: Unable to determine the location of the User and Group DS (%s)", errorInfo);
    ereport(LOG_WARN, message);
    return FALSE;
  }

  if (userGroupServer.secure) {
    sslinit(info, serverroot, PR_TRUE);
  }

  if (!(server = openLDAPConnection(&userGroupServer)))
    {
      destroyAdmldap(info);
      ereport(LOG_WARN, (char*)"admin40_check_ds_availability_init(): Unable to open LDAP connection to User and Group DS.");
      return FALSE;
    }

  ldapError = admin40_ldap_auth_userdn_password(server, NULL, NULL, &pw_exp);

  if(ldapError == LDAP_SERVER_DOWN || ldapError == LDAP_CONNECT_ERROR) {
    sprintf(message, "admin40_check_ds_availability_init(): WARNING: User and Group Directory Server is down or unreachable (%s)",
	    ldap_err2string(ldapError));
    ereport(LOG_WARN, message);
    return FALSE;
  }

  return TRUE;
}
