/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*********************************************/
/*                                           */
/* taskacl.c : routines for ACL evaluation   */
/*             in Kingpin.                   */
/*                                           */
/* David Tompkins 9/29/97                    */
/*                                           */
/*********************************************/

/*
The current design utilizes the directory
server as an evaluation service; the ability
to read a task entry denotes the ability to
execute the task.
*/

#include <stdio.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <netsite.h>
#include <ldaputil/certmap.h>
#include <ldaputil/errors.h>
#include <ldaputil/ldapauth.h>
#include <ldap_ssl.h>
#include <base/util.h>
#include <base/pblock.h>
#include <base/ereport.h>
#include <frame/func.h>
#include <libadminutil/admutil.h>

#include "dbtadmserv.h"
#include "taskacl.h"
#include "hashtable.h"
#include "frame/protocol.h"

/* Globals...hack. */
static char           *serverroot    = NULL;
static long            cacheLifetime = 0; /* Defaults to 0 (entries immediately expire) */
static LdapServerData  registryServer;
static LdapServerData  userGroupServer;

/* extern int adm_secureLDAP;   no longer needed...*/


/* Caches for [I]SIE DNs, authenticated users, and available tasks */
static HashTable *servers    = NULL;
static HashTable *auth_users = NULL;
static HashTable *auth_tasks = NULL;

/* Runtime command lookup hashtable */
static HashTable *commands   = NULL;

#define NS_EXEC_REF           "nsExecRef"
#define NS_EXEC_REF_QUERY     "(nsExecRef=*)"
#define NS_LOG_SUPPRESS       "nsLogSuppress"
#define LOG_SUPPRESS_ON_VALUE "true"

static char *searchAttributes[] = { NS_EXEC_REF, NS_LOG_SUPPRESS, NULL };

/* from adminit.c  */
extern char *nsroot;

/* ------------------------------ _uudecode ------------------------------- */

extern const unsigned char pr2six[256];

char *_uudecode(char *bufcoded)
{
    register char *bufin = bufcoded;
    register unsigned char *bufout;
    register int nprbytes;
    unsigned char *bufplain;
    int nbytesdecoded;

    /* Find the length */
    while(pr2six[(int)*(bufin++)] <= 63);
    nprbytes = bufin - bufcoded - 1;
    nbytesdecoded = ((nprbytes+3)/4) * 3;

    bufout = (unsigned char *) MALLOC(nbytesdecoded + 1);
    bufplain = bufout;

    bufin = bufcoded;

    while (nprbytes > 0) {
        *(bufout++) = (unsigned char)
            (pr2six[(int)(*bufin)] << 2 | pr2six[(int)bufin[1]] >> 4);
        *(bufout++) = (unsigned char)
            (pr2six[(int)bufin[1]] << 4 | pr2six[(int)bufin[2]] >> 2);
        *(bufout++) = (unsigned char)
            (pr2six[(int)bufin[2]] << 6 | pr2six[(int)bufin[3]]);
        bufin += 4;
        nprbytes -= 4;
    }

    if(nprbytes & 03) {
        if(pr2six[(int)bufin[-2]] > 63)
            nbytesdecoded -= 2;
        else
            nbytesdecoded -= 1;
    }
    bufplain[nbytesdecoded] = '\0';

    return (char *)bufplain;
}

static char*
adm_normalize_dn(char* origDN, char* newDN) {

  char *orig = origDN, *dest = newDN;

  if (!newDN) return NULL;

  while (*orig != '\0') {
    if (*orig == ',') {
      *dest++ = *orig++;
      while (*orig == ' ') orig++;
    }
    else *dest++ = *orig++;
  }
  *dest = '\0';
  return newDN;
}

static void
convert_to_lower_case(char *s)
{
   register char *p = s;

   while (*p)
   {
      if ((((unsigned char)*p)<0x80)&&(isupper(*p)))
	 *p = _tolower(*p);
      p++;
   }
}

static
void
closeLDAPConnection(LDAP *server)
{
   ldap_unbind_s(server);
}

static LDAP *
openLDAPConnection(LdapServerData *data)
{
   static int false = 0;

   LDAP *server;

   if (data->secure) {
     if (!(server = ldapssl_init(data->host, data->port, 1))) {
       ereport(LOG_FAILURE, "openLDAPConnection(): ldapssl_init failed\n");
       return NULL;
     }
   }
   else {
     if (!(server = ldap_init(data->host, data->port))) {
       ereport(LOG_FAILURE, "openLDAPConnection(): ldap_init failed\n");
       return NULL;
     }
   }

   if (ldap_set_option(server, LDAP_OPT_CACHE_ENABLE, (void*)&false) != LDAP_SUCCESS)
   {
       ereport(LOG_FAILURE, "openLDAPConnection(): ldap_set_option failed\n");
      closeLDAPConnection(server);
      return NULL;
   }

   return (server);
}

static void
populate_tasks_from_server(char *serverid, void *sieDN,  void *userdata)
{
   PopulateTasksData *data   = (PopulateTasksData *)userdata;
   LDAP              *server = data->server;
   LDAPMessage       *result, *e;
   int                ldapError;
   char              normDN[1024];
   int tries = 0;

#if 0
   if ((ldapError = ldap_search_s(server, sieDN, LDAP_SCOPE_SUBTREE, NS_EXEC_REF_QUERY, searchAttributes, 0, &result)) != LDAP_SUCCESS)
   {
      ldap_perror(server, "populate_tasks_from_server():task search error");
      return;
   }
#endif /* 0 */

   if(data->server == NULL) {
     if(!(server = openLDAPConnection(&registryServer))) {
       ldap_perror(server, "populate_tasks_from_server(): Unable to open LDAPConnection");
       return;
     }
     data->server = server;
   }

   tries = 0;
   do {
     ldapError = ldap_search_s(server, sieDN, LDAP_SCOPE_SUBTREE, NS_EXEC_REF_QUERY, searchAttributes, 0, &result);
     if(ldapError != LDAP_SERVER_DOWN && ldapError != LDAP_CONNECT_ERROR)
       break;
     
     closeLDAPConnection(server);
     if(!(server = openLDAPConnection(&registryServer))) {
       ldap_perror(server, "populate_tasks_from_server(): Unable to open LDAPConnection");
       return;
     }
     data->server = server;
   } while (server != NULL && ++tries < 2);

   if (ldapError != LDAPU_SUCCESS) {
     ldap_perror(server, "populate_tasks_from_server():task search error");
     return;
   }
   
   for (e = ldap_first_entry(server, result) ; e != NULL ; e = ldap_next_entry(data->server, e))
   {
      char *dn, **vals, **vals2, *execRefArgs;
      TaskCacheEntry *cache_entry;

      dn = ldap_get_dn(server, e);

      if (!dn) continue;

      vals = ldap_get_values(server, e, NS_EXEC_REF);

      if (!vals) {
	 ldap_memfree(dn);
         continue;
      }

      vals2 = ldap_get_values(server, e, NS_LOG_SUPPRESS);

      if ((execRefArgs = strchr(vals[0], '?')) != NULL)
         *(execRefArgs++) = '\0';

      adm_normalize_dn(dn, normDN);
      convert_to_lower_case(normDN);

      if (!(cache_entry = HashTableFind(auth_tasks, normDN))) {
	 cache_entry = (TaskCacheEntry*)malloc(sizeof(TaskCacheEntry));
	 cache_entry->auth_userDNs = HashTableCreate();
	 HashTableInsert(auth_tasks, normDN, cache_entry);
      }
      else {
	 free(cache_entry->execRef);
	 if (cache_entry->execRefArgs)
	    free(cache_entry->execRefArgs);
      }

      cache_entry->execRef     = strdup(vals[0]);
      cache_entry->execRefArgs = execRefArgs ? strdup(execRefArgs) : NULL;
      cache_entry->logSuppress = (vals2 && !STRCASECMP(vals2[0], LOG_SUPPRESS_ON_VALUE));
      HashTableInsert(cache_entry->auth_userDNs, data->userDN, (char*)(data->now));

      ldap_value_free(vals);
      if (vals2)
         ldap_value_free(vals2);

      ldap_memfree(dn);
    }

   ldap_msgfree(result);
}

static void
populate_task_cache_entries(char *userDN, LDAP *server)
{
   PopulateTasksData data;

   data.server = server;
   data.userDN = userDN;
   data.now    = time(0);

   HashTableEnumerate(servers, populate_tasks_from_server, &data);
}

static
int LDAP_CALL LDAP_CALLBACK
admin40_ldap_rebind_proc(LDAP *ld, char **who, char **pw, int *auth, int freeit, void *arg)
{
   RebindData *data = (RebindData*)arg;

   if (!freeit)
   {
      *who  = data->user;
      *pw   = data->pw;
      *auth = LDAP_AUTH_SIMPLE;
   }

   return LDAP_SUCCESS;
}

static
void
setLDAPRebindProc(LDAP *server, char *user, char *pw)
{
   RebindData *data = (RebindData*)malloc(sizeof(RebindData));

   if (user) data->user = STRDUP(user);
   if (pw) data->pw   = STRDUP(pw);

   /* free old (void*)data, mem leak if not freed 
   if (NSLDAPI_VALID_LDAP_POINTER(server)) {
     free(server->ld_rebind_arg);
   }*/

   ldap_set_rebind_proc(server, admin40_ldap_rebind_proc, (void *)data);
}

static int
admin40_ldap_auth_userdn_password(LDAP *server, 
				  char *userdn, 
				  char *pw, 
				  int *pw_expiring)
{
  int ver = LDAP_VERSION3; 
  LDAPMessage *res = NULL;           /* LDAP result message */ 
  LDAPControl **ctrls = NULL;
  int ldapError;
  int msgid = 0;              /* async bind id */ 
  int rc = 0;                 /* result code of LDAP operation */ 
  int tries = 0;

  *pw_expiring = -1;

  setLDAPRebindProc(server, userdn, pw);

#if 1
  /* This is for password expiration check */
  /* set the LDAP version -- only v3 knows about controls */ 
  ldap_set_option(server, LDAP_OPT_PROTOCOL_VERSION, &ver); 

  /* DT 9/18/98 - Fix for bind problem. Previously, if pw == null,
   * then the bind does not occur (connection is still anonymous),
   * but ldap_simple_bind() returns LDAP_SUCCESS. We want the
   * bind to fail if pw == null, or the pw does not match.
   */

  /* 
   * adam 12/08/99 - Put in pvo's fix on top of DT's, which
   * completely fixes the problem described above. pvo's fix
   * checked only into admin4_branch but not in 41 or above. (379159)
   */
  if (pw == NULL || *pw == '\0')
     pw = "";

  /* bind user asynchronously */ 
  if ( (msgid = ldap_simple_bind(server, userdn, pw)) == -1 ) {
    /*return LDAP_OPERATIONS_ERROR;*/
    return LDAP_SERVER_DOWN;
  }

  rc = ldap_result(server, msgid, 1, NULL, &res);

  switch (rc) {
  case -1: 
    if (res) ldap_msgfree(res);
    return ldap_get_lderrno(server, NULL, NULL);
  default: 
    if ((ldapError = ldap_result2error(server, res, 0)) != LDAP_SUCCESS) {
      if (res) ldap_msgfree(res);
      return ldapError;
    }
  }

  if ((ldapError = ldap_parse_result(server, res, NULL, NULL, NULL, NULL, &ctrls, 0))
      != LDAP_SUCCESS) {
    if (res) ldap_msgfree(res);
    return ldapError;
  }

  if ((ldapError = ldap_result2error(server, res, 0)) != LDAP_SUCCESS) {
    if (res) ldap_msgfree(res);
    return ldapError;
  }

  if (ctrls) {
    int i; 
    for (i = 0; ctrls[i] != NULL; ++i) {
      if(!(strcmp(ctrls[i]->ldctl_oid, LDAP_CONTROL_PWEXPIRED))) {
	*pw_expiring = 0;
      } 
      else if(!(strcmp(ctrls[i]->ldctl_oid, LDAP_CONTROL_PWEXPIRING))) {
	/* "The password is expiring in n seconds" */ 
	if ((ctrls[i]->ldctl_value.bv_val != NULL ) && 
	    (ctrls[i]->ldctl_value.bv_len > 0)) {
	  *pw_expiring = atoi(ctrls[i]->ldctl_value.bv_val); 
	} 
      } 
    }
    ldap_controls_free(ctrls); 
  }

  if (res) ldap_msgfree(res);
  return ldapError;
#else
  /* This is fallback, no password expiration is checked  */
  return ldapu_auth_userdn_password(server, userdn, pw);
#endif
}

static int
admin40_ldap_auth_server(LDAP *server, LdapServerData *data)
{
  int ignored;
  if (!(data->bindDN) && !(data->bindPW))
    return LDAPU_SUCCESS;

  return admin40_ldap_auth_userdn_password(server, data->bindDN, data->bindPW, &ignored);
}

static char *
formLdapURL(LdapServerData *data)
{
   char *s = (char*)malloc(LDAPS_BASE_PREFIX_LENGTH + strlen(data->host) + 1 + 5 + 1 + strlen(data->baseDN) + 1);
   sprintf(s, "%s%s:%d/%s", (data->secure?LDAPS_BASE_PREFIX:LDAP_BASE_PREFIX), data->host, data->port, data->baseDN);
   return s;
}

static int
extractLdapError()
{
   ereport(LOG_FAILURE, "extractLdapServerData(): invalid ldapURL\n");
   return FALSE;
}

static int
extractLdapServerData(LdapServerData *data, char *ldapURL)
{
   char *p=NULL;

   if (!ldapURL)
      return extractLdapError();

   if (strstr(ldapURL, LDAP_BASE_PREFIX)) {
      ldapURL += strlen(LDAP_BASE_PREFIX);
      data->secure = 0;
      data->port = 389;   /* default non-SSL port */
   }
   else if (strstr(ldapURL, LDAPS_BASE_PREFIX)) {
      ldapURL += strlen(LDAPS_BASE_PREFIX);
      data->secure = 1;
      data->port = 636;   /* default SSL port */
   }
   else
      return extractLdapError();

   if ((p = strtok(ldapURL, "/")) == NULL)
     return extractLdapError();
   data->host = strdup(p);

   if ((p = strtok(NULL, "?")) == NULL)
      return extractLdapError();
   data->baseDN = strdup(p);

   /* Currently unused */
   data->bindDN = NULL;
   data->bindPW = NULL;

   return TRUE;
}

static int
buildUGInfo(char** errorInfo) {
  AdmldapInfo info;
  int  error;
  char  path[PATH_MAX];
  char *userGroupLdapURL;
  char *userGroupBindDN;
  char *userGroupBindPW;
  char *dirInfoRef;

  /* Check whether data is available already */
  if (userGroupServer.host) return TRUE;

  if (!nsroot) {
    *errorInfo = "NULL ns root";
    return FALSE;
  }

  sprintf(path, "%s%cadmin-serv%cconfig", nsroot, FILE_PATHSEP, FILE_PATHSEP);

  if (!(info = admldapBuildInfo(path, &error))) {
    *errorInfo = "unable to create AdmldapInfo";
    return FALSE;
  }

  if (admldapGetSecurity(info)) {
    if (!admldapBuildInfoSSL(info, &error)) {
      destroyAdmldap(info);
      *errorInfo = "unable to create SSL to LDAP";
      return FALSE;
    }
  }

  userGroupServer.host   = NULL;
  userGroupServer.port   = 0;
  userGroupServer.secure = 0;
  userGroupServer.baseDN = NULL;

  admldapGetLocalUserDirectory(info, 
			       &userGroupLdapURL, 
			       &userGroupBindDN, 
			       &userGroupBindPW, 
			       &dirInfoRef, 
			       &error);

  if (error != UG_OP_OK) {
    *errorInfo = "unable to set User/Group baseDN";
    destroyAdmldap(info);
    return FALSE;
  }
   
  if (!extractLdapServerData(&userGroupServer, userGroupLdapURL)) {
    destroyAdmldap(info);
    *errorInfo = "unable to extract User/Group LDAP info";
    return FALSE;
  }
  userGroupServer.bindDN = userGroupBindDN ? STRDUP(userGroupBindDN) : NULL;
  userGroupServer.bindPW = userGroupBindPW ? STRDUP(userGroupBindPW) : NULL;
  destroyAdmldap(info);
   
  return TRUE;
}


int
task_eval_init(char *nsRoot, long cache_lifetime)
{
   AdmldapInfo info;
   char  path[PATH_MAX];
   int   error;
   LDAP *server;
   char *userGroupLdapURL = NULL;
   char *userGroupBindDN = NULL;
   char *userGroupBindPW = NULL;
   char *dirInfoRef = NULL;
   char *errorInfo = NULL;
   int tries = 0;

   /* Set up caches */
   servers    = HashTableCreate();
   auth_users = HashTableCreate();
   auth_tasks = HashTableCreate();

   /* cache entry expiration */
   cacheLifetime = cache_lifetime;
   printf("Info: Cache expiration set to %d seconds\n", cacheLifetime);

   if (!nsRoot)
   {
      fprintf(stderr, "task_eval_init(): NULL nsRoot\n");
      return FALSE;
   }

   sprintf(path, "%s%cadmin-serv%cconfig", nsRoot, FILE_PATHSEP, FILE_PATHSEP);

   if (!(info = admldapBuildInfo(path, &error)))
   {
      fprintf(stderr, "task_eval_init(): unable to create AdmldapInfo\n");
      return FALSE;
   }

   serverroot = STRDUP(nsRoot);

   /* Registry DS setup */
   registryServer.host   = admldapGetHost(info);
   registryServer.port   = (admldapGetPort(info) < 0) ? 389 : admldapGetPort(info);
   registryServer.secure = (admldapGetSecurity(info)) ? 1 : 0;
   registryServer.baseDN = admldapGetBaseDN(info);
   registryServer.bindDN = admldapGetSIEDN(info);
   registryServer.bindPW = admldapGetSIEPWD(info);

   /* DT 5/18/98 Change for new User/Group stuff */

   /* Populate U/G Info */

   userGroupServer.host   = NULL;

   if (buildUGInfo(&errorInfo) != TRUE) {
     fprintf(stderr, "Warning! task_eval_init(): %s\n", errorInfo);
     return FALSE;
   }

   /* Register the admin server tasks */
   task_register_server(ADMIN_SERVER_ID, registryServer.bindDN);

   /* Populate the auth_tasks cache for the Local Admin */
   if (!(server = openLDAPConnection(&registryServer)))
   {
     destroyAdmldap(info);
     fprintf(stderr, "Unable to open LDAPConnection to populate LocalAdmin tasks into cache.\n");
     return TRUE;
   }

#if 0
   /* bind as the LocalAdmin to set up task cache entries */
   if ((error = admin40_ldap_auth_server(server, &registryServer)) != LDAPU_SUCCESS)
   {
     destroyAdmldap(info);
     closeLDAPConnection(server);
     fprintf(stderr, "Unable to bind as LocalAdmin to populate LocalAdmin tasks into cache.\n");
     return TRUE;
   }
#endif /* 0 */

   tries = 0;
   do {
     error = admin40_ldap_auth_server(server, &registryServer);
     if(error != LDAP_SERVER_DOWN && error != LDAP_CONNECT_ERROR)
       break;
     
     closeLDAPConnection(server);
     if(!(server = openLDAPConnection(&registryServer))) {
       destroyAdmldap(info);
       fprintf(stderr, "Unable to open LDAPConnection to populate LocalAdmin tasks into cache.\n");
       return TRUE;
     }
   } while (server != NULL && ++tries < 2);
   
   if(error != LDAPU_SUCCESS)
   {
     destroyAdmldap(info);
     closeLDAPConnection(server);
     fprintf(stderr, "Unable to bind as LocalAdmin to populate LocalAdmin tasks into cache.\n");
     return TRUE;
   }     

   populate_task_cache_entries(LOCAL_SUPER_NAME, server);
   closeLDAPConnection(server);
   destroyAdmldap(info);
   return TRUE;
}

/* 
 * Miodrag (06-15-98)
 * The following metthod is called from the runtime command 
 * "change_sie_password" after the sie password is changed
 */
void
task_update_registry_server_bindpw(char *password)
{
   registryServer.bindPW = password;
}

int
task_register_server(char *serverid, char *sieDN)
{
   HashTableInsert(servers, serverid, sieDN);
   return TRUE;
}

static int
reverse_uri(char **storage, char *limit, char *taskuri)
{
   char *p = strchr(taskuri, '/');

   if (p)
   {
      *(p++) = '\0';
      if (!reverse_uri(storage, limit, p))
	 return FALSE;
   }

   /* magic number 4 = "cn=" and "," */
   if ((*storage + strlen(taskuri)+4) > limit)
   {
      ereport(LOG_FAILURE, "reverse_uri():taskDN exceeds LINE_LENGTH, taskDN truncated...\n");
      return FALSE;
   }

   sprintf(*storage, "cn=%s, ", taskuri);
   *storage += strlen(*storage);
   return TRUE;
}

static int
build_full_DN(char **storage, char *limit, char *taskuri, char *sieDN)
{
   if (!reverse_uri(storage, limit, taskuri))
      return FALSE;

   if ((*storage + strlen(sieDN)) > limit)
   {
      ereport(LOG_FAILURE, "build_full_DN():taskDN exceeds LINE_LENGTH, taskDN truncated...\n");
      return FALSE;
   }

   sprintf(*storage, "%s", sieDN);
   return TRUE;
}

static int
check_auth_tasks_cache(char *dn, char *userdn, pblock *pb, pblock *reqpb, long now)
{
   TaskCacheEntry *cache_entry;
   char            normEntryDN[1024];
   long            createTime;

   adm_normalize_dn(dn, normEntryDN);

   /* Server is down. Attempt to resolve task entry from auth_tasks cache. */
   cache_entry = HashTableFind(auth_tasks, normEntryDN);

   if (!cache_entry || !(createTime = (long)HashTableFind(cache_entry->auth_userDNs, userdn)) || ((now - createTime) > cacheLifetime))
      return REQ_ABORTED; /* userdn not authorized for task, or task not found */

   pblock_nvinsert("execRef", cache_entry->execRef, pb);
   if (cache_entry->execRefArgs)
      pblock_nvinsert("execRefArgs", cache_entry->execRefArgs, pb);
   if (cache_entry->logSuppress)
      pblock_nvinsert("logSuppress", "true", reqpb);
   return REQ_PROCEED;
}

int
admin40_task_access_eval(pblock *pb, Session *sn, Request *rq)
{
   int             ldapError;
   char            entryDN[LINE_LENGTH];
   char           *p;
   LDAP           *server;
   LDAPMessage    *result, *e;
   char           **vals, **vals2;
   char           *userdn;
   char           *pw;
   char           *ulevel;
   char           *serverid;
   char           *storage = entryDN;
   char           *uri = pblock_findval("uri", rq->reqpb);
   long            now;
   int tries = 0;

   uri = STRDUP(uri);

   if (!(p = strchr(uri+1, '/')))
   {
      pblock_nvinsert("admin-error", XP_GetClientStr(DBT_invalidUrl_), rq->vars);
      return REQ_ABORTED;
   }

   serverid = uri+1;
   uri      = p+1;
   *p       = '\0';

   if (!(p = (char*)HashTableFind(servers, serverid)))
   {
      /* DT 4/6/98 -- If we're seeing a serverid for the first time, then we try to do
       *              a resync to pull in new data for the serverid. If it still fails,
       *              then the client is out of luck. */

      admin40_runtime_command_exec(RUNTIME_RESYNC_COMMAND, NULL, NULL);

      if (!(p = (char*)HashTableFind(servers, serverid)))
      {
         ereport(LOG_FAILURE, "admin40_task_access_eval(): unable to find registered server (%s)\n", serverid);
	 return admin40_error(sn, rq, PROTOCOL_BAD_REQUEST, "server not registered"); /*i18n*/
      }
   }

   if (!build_full_DN(&storage, entryDN+LINE_LENGTH, uri, p))
   {
      pblock_nvinsert("admin-error", XP_GetClientStr(DBT_invalidUrl_), rq->vars);
      return REQ_ABORTED;
   }
   convert_to_lower_case(entryDN);

   if ((ulevel = pblock_findval("ADM_UserLevel", rq->vars)) && !STRCMP(ulevel, LOCAL_SUPER_NAME))
      return check_auth_tasks_cache(entryDN, LOCAL_SUPER_NAME, pb, rq->reqpb, 0);

   userdn = pblock_findval("userdn", rq->vars);
   pw     = pblock_findval("auth-password", rq->vars);

   if (!userdn || !pw)
   {
      ereport(LOG_FAILURE, "admin40_task_access_eval(): invalid userdn/pw\n");
      return REQ_ABORTED;
   }

   /* DT 3/10/98
    * Turn on the task cache for evals, and not just for DS failover.
    */

   now = time(0);

   if (check_auth_tasks_cache(entryDN, userdn, pb, rq->reqpb, now) == REQ_PROCEED)
      return REQ_PROCEED;

   /* Cache lookup failed or the entry expired. Try the DS. */

   if (!(server = openLDAPConnection(&registryServer)))
      return admin40_error_std(sn, rq, "unable to open LDAPConnection"); /*i18n*/

#if 0
   if ((ldapError = admin40_ldap_auth_server(server, &registryServer)) != LDAPU_SUCCESS)
   {
      closeLDAPConnection(server);

      if ((ldapError == LDAP_CONNECT_ERROR) || (ldapError == LDAP_SERVER_DOWN))
	 return check_auth_tasks_cache(entryDN, userdn, pb, rq->reqpb, 0); /* Try the cache, ignoring expiration */

      return admin40_error(sn, rq, PROTOCOL_UNAUTHORIZED, "invalid user credentials"); /*i18n*/
   }
#endif /* 0 */

   tries = 0;
   do {
     /*
      * Bug fix #340829 - do NOT bind as the local admin here, bind as user DN.
      */
     int ignored;
     ldapError = admin40_ldap_auth_userdn_password(server, userdn, pw, &ignored);
     if(ldapError != LDAP_SERVER_DOWN && ldapError != LDAP_CONNECT_ERROR)
       break;
     
     closeLDAPConnection(server);
     if(!(server = openLDAPConnection(&registryServer)))
       return admin40_error_std(sn, rq, "unable to open LDAPConnection"); /*i18n*/
   } while (server != NULL && ++tries < 2);
   
   if(ldapError != LDAPU_SUCCESS)
   {
      closeLDAPConnection(server);

      if ((ldapError == LDAP_CONNECT_ERROR) || (ldapError == LDAP_SERVER_DOWN))
	 return check_auth_tasks_cache(entryDN, userdn, pb, rq->reqpb, 0); /* Try the cache, ignoring expiration */

      return admin40_error(sn, rq, PROTOCOL_UNAUTHORIZED, "invalid user credentials"); /*i18n*/
   }     


   /* Attempt to read the entry data -- involves DS eval of entry ACI for userDN.
    * If we are able to read it, then access is granted.
    */

#if 0
   if ((ldapError = ldap_search_s(server, entryDN, LDAP_SCOPE_BASE, NS_EXEC_REF_QUERY, searchAttributes, 0, &result)) != LDAP_SUCCESS)
   {
      closeLDAPConnection(server);
      return admin40_error(sn, rq, PROTOCOL_FORBIDDEN, "task invocation denied"); /*i18n*/
   }
#endif /* 0 */

   tries = 0;
   do {
     ldapError = ldap_search_s(server, entryDN, LDAP_SCOPE_BASE, NS_EXEC_REF_QUERY, searchAttributes, 0, &result);
     if(ldapError != LDAP_SERVER_DOWN && ldapError != LDAP_CONNECT_ERROR)
       break;
     
     closeLDAPConnection(server);
     if(!(server = openLDAPConnection(&registryServer)))
       return admin40_error_auth(sn, rq, "unable to open LDAPConnection"); /*i18n*/
   } while (server != NULL && ++tries < 2);

   if (ldapError != LDAPU_SUCCESS) {
     closeLDAPConnection(server);
     ereport(LOG_FAILURE, "admin40_task_access_eval(): task search error\n");
     return REQ_ABORTED;
   }

   /* push execRef and execRefArgs into request pblock */
   if (!(e = ldap_first_entry(server, result)))
   {
      /* 
       * Search succeeded, but no entries returned - this user wasn't allowed to access this task
       */
      closeLDAPConnection(server);
      return admin40_error(sn, rq, PROTOCOL_FORBIDDEN, "task invocation denied"); /*i18n*/
   }

   if ((vals = ldap_get_values(server, e, NS_EXEC_REF)) == NULL)
   {
      closeLDAPConnection(server);
      ereport(LOG_FAILURE, "admin40_task_access_eval(): unable to find %s\n", NS_EXEC_REF);
      return REQ_ABORTED;
   }

   if ((p = strchr(vals[0], '?')) != NULL)
   {
      *(p++) = '\0';
      pblock_nvinsert("execRefArgs", p, pb);
   }
   pblock_nvinsert("execRef", vals[0], pb);

   if ((vals2 = ldap_get_values(server, e, NS_LOG_SUPPRESS)) && !STRCASECMP(vals2[0], LOG_SUPPRESS_ON_VALUE))
      pblock_nvinsert("logSuppress", "true", rq->reqpb);

   ldap_msgfree(result);

   /* setup static task cache entries for userdn */
   populate_task_cache_entries(userdn, server);

   closeLDAPConnection(server);

    /* won't get this far if vals is NULL.  Play safe
       for future changes, check for null any way. */
   if (vals) {
     ldap_value_free(vals);
   }

   if (vals2) {
     ldap_value_free(vals2);
   }

   return REQ_PROCEED;
}

static int
check_auth_users_cache(char *user, char *pw, Request *rq, long now)
{
   /* Attempt to authenticate using auth_users cache */

   UserCacheEntry *cache_entry = NULL;

   if (!auth_users) return REQ_NOACTION; /* Failed. */

   cache_entry =  (UserCacheEntry*)HashTableFind(auth_users, user);

   if (!cache_entry || STRCMP(cache_entry->userPW, pw) || ((now - cache_entry->createTime) > cacheLifetime))
      return REQ_NOACTION; /* Failed. */

   pblock_nvinsert("userdn", cache_entry->userDN, rq->vars);
   pblock_nvinsert("auth-type", "basic", rq->vars);
   pblock_nvinsert("auth-password", pw, rq->vars);
   if (!strchr(user, '='))
      pblock_nvinsert("auth-user", user, rq->vars);
   pblock_nvinsert(AUTHENTICATION_LDAP_URL, cache_entry->ldapURL, rq->vars);

   return REQ_PROCEED;
}

static int
create_auth_users_cache_entry(char *user, char *userDN, char *userPW, char *ldapURL)
{
   UserCacheEntry *cache_entry;

   if (!(cache_entry = HashTableFind(auth_users, user)))
   {
      cache_entry = (UserCacheEntry*)malloc(sizeof(UserCacheEntry));
      HashTableInsert(auth_users, user, cache_entry);
   }
   else
   {
      free(cache_entry->userDN);
      free(cache_entry->userPW);
      free(cache_entry->ldapURL);
   }

   cache_entry->userDN     = strdup(userDN);
   cache_entry->userPW     = strdup(userPW);
   cache_entry->ldapURL    = strdup(ldapURL);
   cache_entry->createTime = time(0);
   return 1;
}

static int
authenticate_user(LdapServerData *data, char *baseDN, char *user, char *pw, Session *sn, Request *rq)
{
   LDAP           *server;
   char           *uid, *userdn, *ldapURL;
   int             ldapError;
   int             pw_expiring = 0;
   char            pw_expiring_str[256];
   int tries = 0;

   if (!(server = openLDAPConnection(data)))
      return admin40_error_auth(sn, rq, "unable to open LDAPConnection to server"); /*i18n*/

#if 0
   if ((ldapError = admin40_ldap_auth_server(server, data)) != LDAPU_SUCCESS)
   {
      closeLDAPConnection(server);
      return admin40_error_auth(sn, rq, "unable to bind to server"); /*i18n*/
   }
#endif /* 0 */

   tries = 0;
   do {
     ldapError = admin40_ldap_auth_server(server, data);
     if(ldapError != LDAP_SERVER_DOWN && ldapError != LDAP_CONNECT_ERROR)
       break;
     
     closeLDAPConnection(server);
     if(!(server = openLDAPConnection(data)))
	return admin40_error_auth(sn, rq, "unable to open LDAPConnection to server"); /*i18n*/
   } while (server != NULL && ++tries < 2);
   
   if(ldapError != LDAPU_SUCCESS)
   {
      closeLDAPConnection(server);
      return admin40_error_auth(sn, rq, "unable to bind to server"); /*i18n*/
   }     


   /* The basic auth data may be either uid:pw or userDN:pw. The test for '='
    * is hopefully adequate to detect a DN...
    */
   if (!strchr(user, '='))
   {
      /* not a DN, so resolve the DN from the uid */

#if 0
      if ((ldapError = ldapu_find_userdn(server, user, baseDN ? baseDN : data->baseDN, &userdn)) != LDAPU_SUCCESS)
      {
	 closeLDAPConnection(server);

         if ((ldapError == LDAP_CONNECT_ERROR) || (ldapError == LDAP_SERVER_DOWN))
	    return check_auth_users_cache(user, pw, rq, 0); /* DS down. Use the cache, ignoring entry expiration. */

         return REQ_NOACTION; /* fall back to final check against admpw */
      }
#endif /* 0 */

      tries = 0;
      do {
	ldapError = ldapu_find_userdn(server, user, baseDN ? baseDN : data->baseDN, &userdn);
	if(ldapError != LDAP_SERVER_DOWN && ldapError != LDAP_CONNECT_ERROR)
	  break;
	
	closeLDAPConnection(server);
	if(!(server = openLDAPConnection(data)))
	  return admin40_error_auth(sn, rq, "unable to open LDAPConnection to server"); /*i18n*/
      } while (server != NULL && ++tries < 2);

      if (ldapError != LDAPU_SUCCESS)
      {
	 closeLDAPConnection(server);

         if ((ldapError == LDAP_CONNECT_ERROR) || (ldapError == LDAP_SERVER_DOWN))
	    return check_auth_users_cache(user, pw, rq, 0); /* DS down. Use the cache, ignoring entry expiration. */

         return REQ_NOACTION; /* fall back to final check against admpw */
      }

      uid = user;
   }
   else
   {
      /* it's a DN */

      userdn = user;
      uid    = NULL;

      /* strip the leading "ldap:///", if present */

      if (!STRNCASECMP(userdn, LDAP_PREFIX, LDAP_PREFIX_LENGTH))
      {
	 if (strlen(userdn) > LDAP_PREFIX_LENGTH)
            userdn += LDAP_PREFIX_LENGTH;
      }
   }

#if 0
   if ((ldapError = admin40_ldap_auth_userdn_password(server, userdn, pw, &pw_expiring)) != LDAPU_SUCCESS)
   {
      closeLDAPConnection(server);

      if ((ldapError == LDAP_CONNECT_ERROR) || (ldapError == LDAP_SERVER_DOWN))
	 return check_auth_users_cache(user, pw, rq, 0); /* DS down. Look in the cache, ignoring entry expiration. */

      return REQ_NOACTION; /* fall back to final check against admpw */
   }
#endif /* 0 */

   tries = 0;
   do {
     ldapError = admin40_ldap_auth_userdn_password(server, userdn, pw, &pw_expiring);
     if(ldapError != LDAP_SERVER_DOWN && ldapError != LDAP_CONNECT_ERROR)
       break;
     
     closeLDAPConnection(server);
     if(!(server = openLDAPConnection(data)))
       return admin40_error_auth(sn, rq, "unable to open LDAPConnection to server"); /*i18n*/
   } while (server != NULL && ++tries < 2);

   if (ldapError != LDAP_SUCCESS)
   {
      closeLDAPConnection(server);

      if ((ldapError == LDAP_CONNECT_ERROR) || (ldapError == LDAP_SERVER_DOWN))
	 return check_auth_users_cache(user, pw, rq, 0); /* DS down. Look in the cache, ignoring entry expiration. */

      return REQ_NOACTION; /* fall back to final check against admpw */
   }

   closeLDAPConnection(server);

   pblock_nvinsert("userdn", userdn, rq->vars);
   pblock_nvinsert("auth-type", "basic", rq->vars);
   pblock_nvinsert("auth-password", pw, rq->vars);
   if (uid) pblock_nvinsert("auth-user", uid, rq->vars);
   pblock_nvinsert(AUTHENTICATION_LDAP_URL, ldapURL = formLdapURL(data), rq->vars);
   if (pw_expiring >= 0) {
     sprintf(pw_expiring_str, "%d", pw_expiring);
     pblock_nvinsert("auth-password-expire", pw_expiring_str, rq->vars);
   }

   create_auth_users_cache_entry(user, userdn, pw, ldapURL);

   if (ldapURL) {
     ldap_memfree(ldapURL);
   }

   return REQ_PROCEED;
}

int
admin40_authenticate(pblock *pb, Session *sn, Request *rq)
{
   char *auth, *user, *pw, *dummy = NULL;
   long  now;
   int   ret;

   if (request_header("authorization", &auth, sn, rq) == REQ_ABORTED)
      return REQ_ABORTED;

   if (!auth)
   {
      /* Call require-auth to force basic auth */
      int ret;
      pblock *execpb = pblock_create(ADMIN_HASH_SIZE);
      pblock_nvinsert("fn", "require-auth", execpb);
      pblock_nvinsert("auth-type", "basic", execpb);
      /* We don't really know what the user name is, so just list
       * a nonsense value. */
      pblock_nvinsert("auth-user", "NS-no-such-user", execpb);
      pblock_nvinsert("realm", MAGNUS_SERVER_NAME, execpb);
      ret = func_exec(execpb, sn, rq);
      pblock_free(execpb);
      return ret;
   }

   /* Skip leading whitespace */
   while(*auth && (*auth == ' '))
      ++auth;
   if(!*auth)
      return REQ_ABORTED;

   /* Verify correct type */
   if((strlen(auth) < 6) || strncasecmp(auth, "basic ", 6))
      return REQ_ABORTED;

   /* Skip whitespace */
   auth += 6;
   while(*auth && (*auth == ' '))
      ++auth;

   if(!*auth)
      return REQ_ABORTED;

   /* Uudecode user:password now */
   if(!(user = _uudecode(auth)))
      return REQ_ABORTED;

   /* find last ':' in auth field (since a DN might have a ldap:/// prefix) */
   for (pw = user + strlen(user) - 1 ; pw > user ; pw--)
   {
      if (*pw == ':')
	 break;
   }
   if (pw == user) /* ':' not found */
   {
      FREE(user);
      return REQ_ABORTED;
   }
   *pw++ = '\0';

   /* DT 3/10/98
    * Enable the caches for lookups, and not just as a DS failover.
    */

   now = time(0);

   if (check_auth_users_cache(user, pw, rq, now) == REQ_PROCEED)
      return REQ_PROCEED;

   /* Cache lookup failed, or cache entry is expired. Try DS. */

   /* 5/18/98 DT -- new user/group separation code */

   /* First try o=NetscapeRoot in the registry server */
#  define NETSCAPE_ROOT_BASEDN "o=NetscapeRoot"
   ret = authenticate_user(&registryServer, NETSCAPE_ROOT_BASEDN, user, pw, sn, rq);
   if (ret != REQ_NOACTION)
      return ret;
   
   /* Then, try the user/group server */

   /* If U/G Info is not available, try to get it */
   if (!(userGroupServer.host)) {
     buildUGInfo(&dummy);
   }

   if (userGroupServer.host) {
     ret = authenticate_user(&userGroupServer, NULL, user, pw, sn, rq);
     if (ret != REQ_NOACTION) return ret;
   }

   /* Lastly, try the baseDN of the registry server 
   ret = authenticate_user(&registryServer, NULL, user, pw, sn, rq);
   */
   
   return ret;
}

int
admin40_runtime_command_init()
{
   commands = HashTableCreate();

   return (TRUE);
}

int
admin40_register_runtime_command(char *name, RuntimeCommandFn fn, void *arg)
{
   RuntimeCommandRecord *rcr;

   if (HashTableFind(commands, name))
   {
      ereport(LOG_FAILURE, "admin40_register_runtime_command(): attempt to register duplicate command (%s)\n", name);
      return FALSE;
   }

   rcr = (RuntimeCommandRecord*)malloc(sizeof(RuntimeCommandRecord));
   rcr->fn  = fn;
   rcr->arg = arg;
   HashTableInsert(commands, name, rcr);

   return TRUE;
}

int
admin40_runtime_command_exec(char *name, char *query, Request *rq)
{
   RuntimeCommandRecord *rcr = HashTableFind(commands, name);

   return (rcr && rcr->fn(name, query, rcr->arg, rq));
}

int
admin40_error_auth(Session *sn, Request *rq, char *reason)
{
   return REQ_NOACTION;
}

int
admin40_error_std(Session *sn, Request *rq, char *reason)
{
   return admin40_error(sn, rq, PROTOCOL_SERVER_ERROR, reason);
}

int
admin40_error(Session *sn, Request *rq, int code, char *reason)
{
   protocol_status(sn, rq, code, reason);
   return REQ_ABORTED;
}
