/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/* time_cache.h
 *
 * Module for converting the time to a string quickly.
 * When profiling, we discovered that converting the time to a string can
 * actually be quite costly. 
 *
 * HTTP specific notes:
 * There are two places where the timecache is used.  First, the timecache
 * is used to cache timestamps which are commonly converted.  For example,
 * The last-modified time of a file is often converted.  This module 
 * implements a cache for storing the time_t to string translation.  The
 * second use of the timecache is for quickly converting the current time
 * to a string.  Every request converts the current time to a string for
 * returning the Date: header.  By caching the last conversion and then
 * only recomputing the string if the string has changed, we save lots of
 * instructions...
 * 
 * Note:  With the new accelerator cache which caches the HTTP headers, this
 * cache is almost obsolete;  if the accelerator cache gets lots of hits,
 * this cache almost never even gets used.  Should we take it out? XXXMB
 *
 * Mike Belshe
 * 9-23-95
 */


#ifndef _BASE_TIME_H_
#define _BASE_TIME_H_

#include "netsite.h"
#include "base/cache.h"

#ifdef XP_UNIX 
#define TIME_FMT	"%a, %d %b %Y %T GMT"
#define TIME_MAX_LEN	30
#define TIME_HOUR_POS	17
#define TIME_MIN_POS	20
#define TIME_SEC_POS	23
#else /* XP_WIN32 */
#define TIME_FMT	"%a, %d %b %Y %H:%M:%S GMT"
#define TIME_MAX_LEN	30
#define TIME_HOUR_POS	17
#define TIME_MIN_POS	20
#define TIME_SEC_POS	23
#endif

#define TIME_SIZE	29

typedef struct time_cache_entry_t {
	cache_t cache;
	time_t	time_val;		/* integer representation of time */
	char	*time_str;		/* string representation of time */
} time_cache_entry_t;

NSPR_BEGIN_EXTERN_C

int time_cache_internal_init(void);

/* time_cache_init()
 * Initializes the time cache.
 */
NSAPI_PUBLIC int time_cache_init(pblock *pb, Session *sn, Request *rq);

/* timecache_destroy()
 *
 */
NSAPI_PUBLIC void time_cache_destroy();

/* timecache_curr_strftime()
 * Returns the time in TIME_FMT format.
 */
NSAPI_PUBLIC char *time_cache_curr_strftime(time_t);
NSAPI_PUBLIC char *time_cache_curr_strftime_logfmt(time_t);

/* time_cache_strftime()
 * Checks for a string representation of the time in the cache.
 */
NSAPI_PUBLIC char *time_cache_strftime(time_t);

/* time_cache_curr_localtime()
 * time_cache_curr_gmtime()
 * Faster versions of these routines.  Only works with increasing values of 
 * new_time (in other words, don't use it for file access times)
 */
#if !defined(SCO)
int time_cache_curr_localtime(struct tm *lt, time_t new_time);
int time_cache_curr_gmtime(struct tm *lt, time_t new_time);
#endif

NSPR_END_EXTERN_C

#endif /* _BASE_TIME_H_ */

