/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * accel_cache.h
 *
 * Accelerator cache for cacheing static data.
 *
 *
 * Mike Belshe
 * 10-27-95
 */

#ifndef _ACCEL_CACHE_H_
#define _ACCEL_CACHE_H_

#ifdef ACCELERATOR_CACHE

#ifdef __cplusplus
extern "C" {
#endif

#include <errno.h>
#include "base/cache.h"
#include "frame/file_cache.h"
#include "frame/accel_file_cache.h"
#include "base/session.h"
#include "frame/req.h"
#include "frame/func.h"
#include "frame/objset.h"

#define ACCEL_CACHE_MAGIC	0x22222222

typedef struct accel_file_cache_entry_t accel_file_cache_entry_t;

/* accel_cache_log_t 
 * keeps track of a logging routine which needs to be called for
 * this entry
 */
typedef struct accel_cache_log_t {
	FuncPtr			log_fn;
	char			*logname;
	int			iponly;	/*1=true, 0=false*/
	struct accel_cache_log_t *next;
} accel_cache_log_t;

/* --- user callback functions ---------------------------------- */
/* Users of the cache have a few entry points for where they can be
 * called back by the cache.
 */

/* FUNCTION: accel_cache_user_service_callback
** DESCRIPTION:
**    Function pointer registered with the cache for servicing a cached
**    request.  Must handle its own work; including error cases.
** INPUTS:
**    pCacheEntry:  A pointer to the cache entry. 
**        NB: THIS IS FOR READ-ONLY DATA AND MUST NOT BE MODIFIED 
**    request:  A pointer to the entire request from the client.
**    request_length:  The length of the request.
**    user_data:  Any user data given to the cache when the cache entry
**        was inserted.
** RETURNS:
**    TRUE if the connection is still open.
**    FALSE if the connection has been closed.
*/
typedef PRBool (*accel_cache_user_service_callback)(
                   accel_file_cache_entry_t *pCacheEntry,
                   char *request,
                   int request_length,
                   void *user_data);

/* FUNCTION: accel_cache_user_notification_callback
** DESCRIPTION:
**    Function pointer registered with the cache to be called after
**    the cache has serviced the request in its normal manner.
** INPUTS:
**    pCacheEntry:  A pointer to the cache entry. 
**        NB: THIS IS FOR READ-ONLY DATA AND MUST NOT BE MODIFIED 
**    request:  A pointer to the entire request from the client.
**    request_length:  The length of the request.
**    user_data:  Any user data given to the cache when the cache entry
**        was inserted.
*/
typedef void (*accel_cache_user_notification_callback)(
                           accel_file_cache_entry_t *pCacheEntry,
                           char *request,
                           int request_length,
                           void *user_data);

/* FUNCTION: accel_cache_user_cleanup_callback
** DESCRIPTION:
**    Function pointer registered with cache to be called when the 
**    cache entry is destroyed.
** INPUTS:
**    user-data:  data passed into the accel insert function.
*/
typedef void (*accel_cache_user_cleanup_callback)(void *user_data);

/* accel_cache_entry_t
 * An object in the accelerator cache.  
 */
typedef struct accel_cache_entry_t {
#ifdef CACHE_DEBUG
	unsigned long		magic;
#endif
        cache_entry_t		*cache;		/* we're a subclass of cache */
	struct accel_file_cache_key_t	*key;
	char			*uri;
	char			*path;
	char			*headers;
	char			*date_header;
	char			*last_modified;
	unsigned int		header_len;	/* total length of the headers*/
	char			*headers_keepalive;
	unsigned int		header_keepalive_len;	/* total length of the headers*/
	accel_cache_log_t	*log_os;
    accel_cache_user_service_callback user_service_func;
    accel_cache_user_notification_callback user_notification_func;
    accel_cache_user_cleanup_callback user_cleanup_func;
    void *user_data;
} accel_cache_entry_t;

/* accel_cache_init()
 * Initializes the accelerator cache.  Cache must be initialized before use.
 *
 * This is just another NSAPI initializer function.
 *
 * Returns REQ_PROCEED on success, REQ_ABORTED on failure.
 */
NSAPI_PUBLIC int accel_cache_init(pblock *pb, Session *sn, Request *rq);

/* accel_cache_destroy()
 * Cleans up and deletes all entries in the cache.
 * Returns 0 on success, less-than-zero on failure.
 */
NSAPI_PUBLIC void accel_cache_destroy(void *unused);

/* accel_cache_create()
 * Creates a new cache entry but does not add it to the cache.  
 * Returns non-NULL on success, NULL on failure.
 */
NSAPI_PUBLIC accel_cache_entry_t *accel_cache_create(
        char *uri, 
        char *host,
        int local_ip, 
        int port, 
        char *path, 
        char *headers, 
        int header_len, 
        httpd_objset *log_os,
        accel_cache_user_service_callback user_service_func,
        accel_cache_user_notification_callback user_notification_func,
        accel_cache_user_cleanup_callback user_cleanup_func,
        void *user_data);


/* accel_cache_insert()
 * Inserts a new cache entry.  access_count is initialized to 1, so the
 * caller must call accel_cache_use_decrement() for this entry when 
 * done using the entry. accel_fn is the "accelerated" function which will
 * be called when we discover that a URI can take the accelerated path.
 *
 * Returns non-NULL on success, NULL on failure.
 */
NSAPI_PUBLIC 
accel_cache_entry_t *accel_cache_insert(
        char *uri, 
        char *host,
        struct sockaddr_in *local_addr, 
        char *path, 
        char *headers, 
        int header_len, 
        httpd_objset *log_os,
        accel_cache_user_service_callback user_service_func,
        accel_cache_user_notification_callback user_notification_func,
        accel_cache_user_cleanup_callback user_cleanup_func,
        void *user_data);

/* accel_cache_delete()
 * Attempts to manually delete an entry from the cache.  Normally, the 
 * cache is capable of maintaining itself.  This routine should only be
 * called if the data in the cache has somehow become invalid.
 *
 * Caller must increment the use count for this entry before attempting to
 * delete it.  If the access_count is > 1, then this routine returns -1,
 * and marks the entry as delete-pending.  The entry will be deleted when
 * the access count drops to 0.
 *
 * Returns 0 on success, less-than-zero on failure.
 */
NSAPI_PUBLIC 
int accel_cache_delete(void *entry);

/* accel_cache_use_increment()
 * Increments the use count for this entry.
 * Returns 0 on success, less-than-zero on failure.
 */
NSAPI_PUBLIC 
int accel_cache_use_increment(accel_cache_entry_t *entry);

/* accel_cache_use_decrement
 * Decrements the use count for this entry.
 * Returns 0 on success, less-than-zero on failure.
 */
NSAPI_PUBLIC 
int accel_cache_use_decrement(accel_cache_entry_t *entry);

/* accel_cache_lookup_uri()
 * Search for a URI in the cache.  If the entry is found, the use count
 * is automatically incremented and it is the caller's responsibility to
 * decrement it.
 *
 * Return non-NULL if found, NULL if not found.
 */
NSAPI_PUBLIC 
accel_cache_entry_t *accel_cache_lookup_uri(char *uri);

/* accel_cache_valid()
 * Checks to see if an entry in the cache is valid.  
 * The caller MUST have already incremented the use count for this 
 * entry before calling this routine.
 *
 * Returns 0 if an entry is still "valid",  less-than-zero on error.
 */
NSAPI_PUBLIC 
int accel_cache_valid(accel_cache_entry_t *entry);

/* accel_cache_touch()
 * Moves a cache entry to the top of the MRU chain.  The caller MUST have
 * already incremented the use count for this entry before calling this
 * routine.
 * Returns 0 on success, less-than-zero on failure.
 */
NSAPI_PUBLIC 
int accel_cache_touch(accel_cache_entry_t *entry);

/* accel_cache_cleanup()
 * Function to "cleanup" all data within an accel_cache_entry.
 * Returns 0 on sucess; less-than-zero on failure.
 */
NSAPI_PUBLIC int accel_cache_cleanup(void *entry);

/* accel_cache_print()
 * Function to print the data in the entry to the file fd.
 */
NSAPI_PUBLIC int accel_cache_print(void *entry, SYS_NETFD fd);

#ifdef __cplusplus
}
#endif

#endif /* ACCELERATOR_CACHE */
#endif /* _ACCEL_CACHE_H_ */
