/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * accel_file_cache.h
 *
 *
 * Mike Belshe
 * 02-07-96
 */

#ifndef _ACCEL_FILE_CACHE_H_
#define _ACCEL_FILE_CACHE_H_

#ifdef ACCELERATOR_CACHE

#include <errno.h>
#include "base/cache.h"
#include <libaccess/acl.h>
#include <libaccess/aclproto.h>
#include "frame/accel_cache.h"
#include "frame/file_cache.h"
#include "base/session.h"
#include "frame/req.h"
#include "frame/func.h"
#include "frame/objset.h"

#define ACCEL_FILE_CACHE_MAGIC	0x33333333
#define ACCEL_FILE_KEY_MAGIC	0x44444444

NSPR_BEGIN_EXTERN_C

/* --- globals -------------------------------------------------- */
/* The accel_file_cache pointer is provided so that the file,
 * accel, and accel_file cache can all share a single cache.
 */
extern cache_t *accel_file_cache;


/* --- key manipulation routines -------------------------------- */
/* The accelerator cache, file cache, and combined accelerator/file
 * cache will all use the same cache.  So, they'll need to use common
 * key manipulation routines in order to do lookups properly.
 */
#define URI_KEY		0xaa
#define PATH_KEY	0xbb
typedef struct accel_file_cache_key_t {
#ifdef CACHE_DEBUG
	unsigned long		magic;
#endif
	int 			type;
	char 			*key;
		/* host or local_ip or both can be NULL */
	char			*host;		/* software multihoming */
	int				local_ip;	/* hardware multihoming */
    int				port;      /* hardware multihoming */
} accel_file_cache_key_t;

int accel_file_cache_key_cmp(void *voidkey1, void *voidkey2);
unsigned int accel_file_cache_hash_key(unsigned int hashsize, void *voidkey);
NSAPI_PUBLIC accel_file_cache_key_t *
	accel_file_cache_key_create(int type, char *key, char *host, int local_ip, int port); 
NSAPI_PUBLIC int accel_file_cache_key_delete(accel_file_cache_key_t *key);

struct accel_cache_entry_t;
struct file_cache_entry_t;

/* accel_file_cache_entry_t
 * An object in the accelerator cache.  
 */
typedef struct accel_file_cache_entry_t {
#ifdef CACHE_DEBUG
	unsigned long		magic;
#endif
	cache_entry_t		*cache;		/* we're a subclass of cache */
	accel_file_cache_key_t	*key;
	unsigned long		hit_count;
		/* flag_action is an action which is pending as soon as the
		 * use-count drops to 0.  We can't update the entry until
		 * the use count is 0.
		 * ACTION_NONE - no action is pending
		 * ACTION_CACHE_FILE - when the count reaches 0, create the
		 *   file_cache_entry for this entry.
		 * ACTION_UNCACHE_FILE - when the count reaches 0, delete the
		 *   file_cache_entry for this entry.
		 */
#define ACTION_NONE		0
#define ACTION_CACHE_FILE	1
#define ACTION_UNCACHE_FILE	2
	int			action_pending;
	int			on_hit_list;
	struct accel_cache_entry_t	*accelcache;
	struct file_cache_entry_t	*filecache;
#ifndef USE_DIRECTORY_CHANGES
        PRUint32                last_access;    /* time this file was cached */
        PRUint32                last_checked;
        PRUint32                creation_date;
        PRUint32                modify_date;
#endif
	struct accel_file_cache_entry_t *less_hit, *more_hit;
        ACLListHandle_t		*acllist;
} accel_file_cache_entry_t;


/* --- accel_file cache routines -------------------------------- */

int accel_file_cache_internal_init(void);

/* accel_file_cache_init()
 * Initializes the accelerator cache.  Cache must be initialized before use.
 *
 * This is just another NSAPI initializer function.
 *
 * Returns REQ_PROCEED on success, REQ_ABORTED on failure.
 */
NSAPI_PUBLIC int accel_file_cache_init(pblock *pb, Session *sn, Request *rq);

/* accel_file_cache_destroy()
 * Cleans up and deletes all entries in the cache.
 * Returns 0 on success, less-than-zero on failure.
 *
 */
NSAPI_PUBLIC void accel_file_cache_destroy(void *unused);

/* accel_file_cache_create()
 * Creates a new cache entry but does not add it to the cache.  Does
 * not set the filecache entry.  The (uri, host, local_ip) trio makes 
 * identifies a single entity in the cache.
 * Returns non-NULL on success, NULL on failure.
 */
NSAPI_PUBLIC 
accel_file_cache_entry_t *accel_file_cache_create(
        char *uri, 
        char *host,
        int local_ip, 
        int port, 
        char *path, 
        char *headers, 
        int header_len, 
        httpd_objset *log_os,
        accel_cache_user_service_callback user_service_func,
        accel_cache_user_notification_callback user_nofitication_func,
        accel_cache_user_cleanup_callback user_cleanup_func,
        void *user_data,
	ACLListHandle_t *acllist);

/* accel_file_cache_insert()
 * Inserts a new cache entry.  access_count is initialized to 1, so the
 * caller must call accel_file_cache_use_decrement() for this entry when 
 * done using the entry. accel_fn is the "accelerated" function which will
 * be called when we discover that a URI can take the accelerated path.
 *
 * User must have incremented the use_count for filecache before calling
 * this routine; user must also decrement use_count for filecache after
 * calling this routine.
 *
 * Returns non-NULL on success, NULL on failure.
 */
NSAPI_PUBLIC
accel_file_cache_entry_t *accel_file_cache_insert(
        char *uri, 
        char *host,
        struct sockaddr_in *local_addr, 
        char *path, 
        char *headers, 
        int header_len, 
        httpd_objset *log_os,
        accel_cache_user_service_callback user_service_func,
        accel_cache_user_notification_callback user_nofitication_func,
        accel_cache_user_cleanup_callback user_cleanup_func,
        void *user_data,
	ACLListHandle_t *acllist);

/* accel_file_cache_delete()
 * Attempts to manually delete an entry from the cache.  Normally, the 
 * cache is capable of maintaining itself.  This routine should only be
 * called if the data in the cache has somehow become invalid.
 *
 * Caller must increment the use count for this entry before attempting to
 * delete it.  If the access_count is > 1, then this routine returns -1,
 * and marks the entry as delete-pending.  The entry will be deleted when
 * the access count drops to 0.
 *
 * Returns 0 on success, less-than-zero on failure.
 */
NSAPI_PUBLIC int accel_file_cache_delete(void *entry);

/* accel_file_cache_use_increment()
 * Increments the use count for this entry.
 * Returns 0 on success, less-than-zero on failure.
 */
NSAPI_PUBLIC 
int accel_file_cache_use_increment(accel_file_cache_entry_t *entry);

/* accel_file_cache_use_decrement
 * Decrements the use count for this entry.
 * Returns 0 on success, less-than-zero on failure.
 */
NSAPI_PUBLIC 
int accel_file_cache_use_decrement(accel_file_cache_entry_t *entry);

/* accel_file_cache_lookup_uri()
 * Search for a URI in the cache.  If the entry is found, the use count
 * is automatically incremented and it is the caller's responsibility to
 * decrement it.
 *
 * Return non-NULL if found, NULL if not found.
 */
NSAPI_PUBLIC 
accel_file_cache_entry_t *accel_file_cache_lookup_uri(char *uri, char *host,
	struct sockaddr_in *localAddress);

/* accel_file_cache_valid()
 * Checks to see if an entry in the cache is valid.  To do so, it checks
 * that the filecache entry is not out of date.  The caller MUST have already
 * incremented the use count for this entry before calling this routine.
 *
 * Returns 0 if an entry is still "valid",  less-than-zero on error.
 */
NSAPI_PUBLIC 
int accel_file_cache_valid(accel_file_cache_entry_t *entry);

/* accel_file_cache_touch()
 * Moves a cache entry to the top of the MRU chain.  The caller MUST have
 * already incremented the use count for this entry before calling this
 * routine.
 * Returns 0 on success, less-than-zero on failure.
 */
NSAPI_PUBLIC 
int accel_file_cache_touch(accel_file_cache_entry_t *entry);

/* accel_file_cache_cleanup()
 * Function to "cleanup" all data within an entry.
 * Returns 0 on sucess; less-than-zero on failure.
 */
NSAPI_PUBLIC 
int accel_file_cache_cleanup(void *entry);

/* accel_file_cache_print()
 * Function to print the data in the entry to the file fd.
 */
NSAPI_PUBLIC 
int accel_file_cache_print(void *entry, SYS_NETFD fd);

/* accel_file_cache_acl_flush()
 * Flushes all entries with ACLs
 */
NSAPI_PUBLIC 
void accel_file_cache_acl_flush(void *unused);

NSAPI_PUBLIC
PRUint32 accel_file_cache_get_size(void);

NSPR_END_EXTERN_C

#endif /* ACCELERATOR_CACHE */
#endif /* _ACCEL_FILE_CACHE_H_ */
