/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * file_cache.h
 *
 * Cache of open files.
 *
 * Mike Belshe
 * 01-08-96
 */

#ifndef _FILE_CACHE_H_
#define _FILE_CACHE_H_

#ifdef ACCELERATOR_CACHE

#include <errno.h>
#include "base/cache.h"
#include "frame/accel_file_cache.h"
#include "base/buffer.h"
#include "base/session.h"
#include "frame/req.h"
#include "base/crit.h"

#define FILE_CACHE_MAGIC	0x55555555

struct accel_file_cache_key_t;

#define STR_BYTES_LEN		16

/* file_cache_entry_t
 * An object in the file cache.  
 */
typedef struct file_cache_entry_t {
#ifdef CACHE_DEBUG
	unsigned long	magic;
#endif 
	cache_entry_t	*cache;		/* we are a subclass of cache */
	struct accel_file_cache_key_t *key;
	char		*path;
	SYS_FILE	fd;
	caddr_t		mmap;		/* can be NULL */
	unsigned long	mmap_len;	/* valid even if mmap is NULL */
	unsigned long	cache_len;	/* size taken from mmap_max for this */
	char		strbytes[STR_BYTES_LEN];/* string number of bytes */
	unsigned int 	async;		/* true if open for async i/o */
	unsigned long	ioport;		/* associated completion port, if any (async i/o) */
#ifdef USE_DIRECTORY_CHANGES
	struct file_cache_entry_t *prev_path,
				*next_path;	/* hash chain for hash by path */
#endif
#ifdef XP_WIN32
	void*		hmap;		/* handle for memory-mapped region */
	int		valid;		/* is this entry valid? */
#endif
#ifndef USE_DIRECTORY_CHANGES
	PRUint32		last_access;	/* time this file was cached */
	PRUint32		last_checked;
	PRUint32		creation_date;
	PRUint32		modify_date;
#endif
} file_cache_entry_t;

NSPR_BEGIN_EXTERN_C

int file_cache_internal_init(void);

/* file_cache_init()
 * Initializes the accelerator cache.  Must be initialized before use.
 *
 * This is just another NSAPI initializer function
 *
 * Returns REQ_PROCEED on success, REQ_ABORTED on failure.
 */
NSAPI_PUBLIC int file_cache_init(pblock *pb, Session *sn, Request *rq);

/* file_cache_destroy()
 * Cleans up and deletes all entries in the cache.
 * Returns 0 on success, less-than-zero on failure.
 */
NSAPI_PUBLIC void file_cache_destroy(void *unused);

/* file_cache_create()
 * Creates a new file cache entry but does not add it to the cache. 
 * Returns non-NULL on success, NULL on failure.
 */
NSAPI_PUBLIC file_cache_entry_t *file_cache_create(char *path);

/* file_cache_insert()
 * Inserts a new cache entry.  access_count is initialized to 1, so the
 * caller must call filel_cache_use_decrement() for this entry when 
 * done using the entry.
 * Returns non-NULL on success, NULL on failure.
 */
NSAPI_PUBLIC file_cache_entry_t *file_cache_insert(char *path);

/* file_cache_delete()
 * Attempts to manually delete an entry from the cache.  Normally, the 
 * cache is capable of maintaining itself.  This routine should only be
 * called if the data in the cache has somehow become invalid.
 *
 * Caller must increment the use count for this entry before attempting to
 * delete it.  If the access_count is > 1, then this routine returns -1.
 * <enhancement- mark the entry as delete-pending and allow the caller to 
 * move on>
 *
 * Returns 0 on success, less-than-zero on failure.
 */
NSAPI_PUBLIC int file_cache_delete(void *entry);

/* file_cache_use_increment()
 * Increments the use count for this entry.
 * Returns 0 on success, less-than-zero on failure.
 */
NSAPI_PUBLIC int file_cache_use_increment(void *entry);

/* file_cache_use_decrement
 * Decrements the use count for this entry.
 * Returns 0 on success, less-than-zero on failure.
 */
NSAPI_PUBLIC int file_cache_use_decrement(void *entry);


/* file_cache_lookup_path()
 * Search for a PATH in the cache.  If the entry is found, the use count
 * is automatically incremented and it is the caller's responsibility to
 * decrement it.
 *
 * Return non-NULL if found, NULL if not found.
 */
NSAPI_PUBLIC file_cache_entry_t *file_cache_lookup_path(char *path);

/* file_cache_valid()
 * Chacks to see if an entry in the cache is valid.  To do so, it checks
 * that the file has not been modified on disk.  The caller MUST have already
 * incremented the use count for this entry before calling this routine.
 * 
 * Returns 0 if an entry is still "valid",  less-than-zero on error.
 */
NSAPI_PUBLIC int file_cache_valid(file_cache_entry_t *entry);

/* file_cache_touch()
 * Moves a cache entry to the top of the MRU chain.  The caller MUST have
 * already incremented the use count for this entry before calling this
 * routine.
 *
 * Returns 0 on success, less-than-zero on failure.
 */
NSAPI_PUBLIC int file_cache_touch(file_cache_entry_t *entry);

/* file_cache_cleanup()
 * Function to "cleanup" all data within an entry.
 * Returns 0 on sucess; less-than-zero on failure.
 */
NSAPI_PUBLIC int file_cache_cleanup(void *entry);

/* file_cache_print()
 * Function to print the data in the entry to the file fd.
 */
NSAPI_PUBLIC int file_cache_print(void *entry, SYS_NETFD fd);

NSPR_END_EXTERN_C

#endif /* ACCELERATOR_CACHE */
#endif /* _FILE_CACHE_H_ */
