/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
#ifndef FRAME_HTTP_H
#define FRAME_HTTP_H

#ifndef NOINTNSAPI
#define INTNSAPI
#endif /* !NOINTNSAPI */

/*
 * http.h: Deals with HTTP-related issues
 * 
 * Rob McCool
 */

#ifndef NETSITE_H
#include "../netsite.h"
#endif /* !NETSITE_H */

#ifndef BASE_SESSION_H
#include "base/session.h"
#endif /* !BASE_SESSION_H */

#ifndef BASE_PBLOCK_H
#include "base/pblock.h"
#endif /* !BASE_PBLOCK_H */

#ifndef FRAME_REQ_H
#include "frame/req.h"               /* REQ_MAX_LINE, Request structure */
#endif /* !FRAME_REQ_H */

/*  #include <libares/arapi.h>  */

#ifndef PUBLIC_FRAME_HTTP_H
#include "public/frame/http.h"
#endif /* !PUBLIC_FRAME_HTTP_H */

/* ------------------------------ Constants ------------------------------- */

/* Number of seconds to allow a second request to come over a persistent
   connection, zero to disable */
#ifdef MCC_PROXY
#define HTTP_DEFAULT_KEEPALIVE_TIMEOUT 5
#else
#define HTTP_DEFAULT_KEEPALIVE_TIMEOUT 15
#endif

/* The maximum number of RFC-822 headers we'll allow */
/* This would be smaller if a certain browser wasn't so damn stupid. */
#define HTTP_MAX_HEADERS 200

#ifdef INTNSAPI

NSPR_BEGIN_EXTERN_C

NSAPI_PUBLIC void INThttp_handle_keepalive(Request *rq, char *t, int *posptr);

/* HTTP/1.1 specific validation routine */
NSAPI_PUBLIC int INThttp_handle_11(Session *sn, Request *rq);
NSAPI_PUBLIC int INThttp_handle_options(Session *sn, Request *rq);
NSAPI_PUBLIC int INThttp_handle_trace(Session *sn, Request *rq);

NSAPI_PUBLIC int INThttp_set_method_num(char *t);
NSAPI_PUBLIC int INThttp_cache_validation(Session *sn, Request *rq,
                                        struct stat *finfo);
NSAPI_PUBLIC char * INThttp_parse_uri(Request *rq, char *uri);
NSAPI_PUBLIC char * INThttp_method_string(int method_num);
NSAPI_PUBLIC void INThttp_set_allowed_methods();
NSAPI_PUBLIC char * INThttp_allowed_methods(Request *rq);
NSAPI_PUBLIC int  INThttp_handle_trace(Session *sn, Request *rq);
NSAPI_PUBLIC int  INThttp_handle_options(Session *sn, Request *rq);


/*
 * gets the first line of an HTTP request
 */
NSAPI_PUBLIC Request *INThttp_find_request(netbuf *buf, Session *sn, int *reply);

/*
 * parses the first line of an HTTP request
 */

NSAPI_PUBLIC int INThttp_parse_request(char *t, Request *rq, Session *sn);

/*
 * Scans HTTP headers from the given netbuf, and places them in headers.
 * If netbuf is NULL, the session's inbuf is used.
 * 
 * Folded lines are joined and the linefeed removed (but not the whitespace).
 * If there are any repeat headers they are joined and the two field bodies
 * separated by a comma and space.
 *
 * t should be a string of length REQ_MAX_LINE. This is a convenience to
 * req.c so that we don't use too much runtime stack.
 *
 * Session is an optional parameter. Use NULL if you wish. It's used for
 * error logs.
 */

NSAPI_PUBLIC int INThttp_scan_headers(Session *sn, netbuf *buf, char *t, 
                                      pblock *headers);


/*
 * Starts the HTTP response. If HTTP/0.9, does nothing. If 1.0, sends header.
 * If this returns REQ_NOACTION, the method was head and no body should be
 * sent. Otherwise, it will return REQ_PROCEED.
 */

NSAPI_PUBLIC int INThttp_start_response(Session *sn, Request *rq);


/*
 * INThttp_hdrs2env takes the entries from the given pblock and converts them
 * to an environment. 
 *
 * Each name entry will be made uppercase, prefixed with HTTP_ and any
 * occurrence of - will be converted to _.
 */

NSAPI_PUBLIC char **INThttp_hdrs2env(pblock *pb);


/*
 * INThttp_status sets status to the code n, with reason string r. If r is
 * NULL, the server will attempt to find one for the given status code.
 * If it finds none, it will give "Because I felt like it."
 */

NSAPI_PUBLIC void INThttp_status(Session *sn, Request *rq, int n, char *r);

/*
 * INThttp_set_finfo sets content-length and last-modified
 */

NSAPI_PUBLIC
int INThttp_set_finfo(Session *sn, Request *rq, struct stat *finfo);


/*
 * Takes the given pblock and prints headers into the given buffer at 
 * position pos. Returns the buffer, reallocated if needed. Modifies pos.
 */

NSAPI_PUBLIC char *INThttp_dump822(pblock *pb, char *t, int *pos, int tsz);

/*
 * Finishes a request. For HTTP, this just closes the socket.
 */

NSAPI_PUBLIC void INThttp_finish_request(Session *sn, Request *rq);


/*
 * INThttp_handle_session processes each request generated by Session
 */

NSAPI_PUBLIC void INThttp_handle_session(Session *sn);

/*
 * INThttp_uri2url takes the give URI prefix and URI suffix and creates a 
 * newly-allocated full URL from them of the form
 * http://(server):(port)(prefix)(suffix)
 * 
 * If you want either prefix or suffix to be skipped, use "" instead of NULL.
 *
 * Normally, the server hostname is taken from the ServerName parameter in
 * magnus.conf. The newer function INThttp_uri2url_dynamic should be used when 
 * a Session and Request structure are available, to ensure that the browser
 * gets redirected to the exact host they were originally referencing.
 */

NSAPI_PUBLIC char *INThttp_uri2url(const char *prefix, const char *suffix);
NSAPI_PUBLIC char *INThttp_uri2url_dynamic(const char *prefix,
                                           const char *suffix,
                                           Session *sn, Request *rq);

/*
 * INThttp_set_keepalive_timeout sets the number of seconds to wait for a new
 * request to come from a persistent connection. Returns nothing. Intended
 * to be called at server startup only.
 *
 * Specifying a timeout of zero will disable persistent connections and allow
 * browsers to request only one file per connection.
 */
NSAPI_PUBLIC void INThttp_set_keepalive_timeout(int secs);
NSAPI_PUBLIC int INThttp_get_keepalive_timeout();

NSPR_END_EXTERN_C

#define http_handle_keepalive INThttp_handle_keepalive
#define http_handle_11 INThttp_handle_11
#define http_handle_options INThttp_handle_options
#define http_handle_trace INThttp_handle_trace
#define http_set_method_num INThttp_set_method_num
#define http_cache_validation INThttp_cache_validation
#define http_parse_uri INThttp_parse_uri
#define http_method_string INThttp_method_string
#define http_set_allowed_methods INThttp_set_allowed_methods
#define http_allowed_methods INThttp_allowed_methods
#define http_handle_trace INThttp_handle_trace
#define http_handle_options INThttp_handle_options

#define http_find_request INThttp_find_request
#define http_parse_request INThttp_parse_request
#define http_scan_headers INThttp_scan_headers
#define http_start_response INThttp_start_response
#define http_hdrs2env INThttp_hdrs2env
#define http_status INThttp_status
#define http_set_finfo INThttp_set_finfo
#define http_dump822 INThttp_dump822
#define http_finish_request INThttp_finish_request
#define http_handle_session INThttp_handle_session
#define http_uri2url INThttp_uri2url
#define http_uri2url_dynamic INThttp_uri2url_dynamic
#define http_set_keepalive_timeout INThttp_set_keepalive_timeout
#define http_get_keepalive_timeout INThttp_get_keepalive_timeout

#endif /* INTNSAPI */

#endif /* !FRAME_HTTP_H */
