/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
#ifndef FRAME_OBJSET_H
#define FRAME_OBJSET_H

#ifndef NOINTNSAPI
#define INTNSAPI
#endif /* !NOINTNSAPI */

/*
 * objset.h: Handles object sets
 * 
 * Each object is produced by reading a config file of some form. See the
 * server documentation for descriptions of the directives that are 
 * recognized, what they do, and how they are parsed.
 * 
 * This module requires the pblock and buffer modules from the base library.
 *
 * Rob McCool
 */

#ifndef NETSITE_H
#include "../netsite.h"
#endif /* !NETSITE_H */

#ifndef BASE_PBLOCK_H
#include "../base/pblock.h"
#endif /* !BASE_PBLOCK_H */

#ifndef BASE_BUFFER_H
#include "../base/buffer.h"
#endif /* !BASE_BUFFER_H */

#ifndef FRAME_OBJECT_H
#include "object.h"
#endif /* !FRAME_OBJECT_H */

#ifndef PUBLIC_FRAME_OBJSET_H
#include "public/frame/objset.h"
#endif /* !PUBLIC_FRAME_OBJSET_H */

/* ------------------------------ Constants ------------------------------- */


/* 
 * The default number of objects to leave room for in an object set,
 * and the number of new entries by which to increase the size when that 
 * room is filled.
 */

#define OBJSET_INCSIZE 8

/*
 * When parsing config. files, httpd will put a limit on how long
 * the parameters to a directive can be (in characters).
 * 
 * Currently set to 10 lines (80 chars/line).
 */

#ifdef MCC_PROXY
#define MAX_DIRECTIVE_LEN 100000
#else
#define MAX_DIRECTIVE_LEN 800
#endif /* MCC_PROXY */

/*
 * The size of the hash tables that store a directive's parameters
 */

#define PARAMETER_HASH_SIZE 3


#ifdef INTNSAPI

/* --- Begin function prototypes --- */

NSPR_BEGIN_EXTERN_C

/*
 * INTobjset_copydirs takes a given objset and copies all of the 
 * directive in directive set <dir> into a new objset, dst_os.  If 
 * dst_os is NULL, it creates a new objset.
 * The newly created objset only has one object in it.
 */
httpd_objset *INTobjset_copydirs(httpd_objset *dst_os,
                                 httpd_objset *os, int dir);

/*
 * INTobjset_scan_buffer will scan through buffer, looking for object 
 * configuration information, and adding them to the object set os if it 
 * finds any. If os is NULL it will allocate a new object set.
 *
 * If any error occurs (syntax error, premature EOF) this function will
 * free os, print an error message into errstr, and return NULL.
 * This is because a config. file error is viewed as a catastrophic error
 * from which httpd should not try to recover. If httpd were to continue
 * after an error, it would not behave as the admin. expected and he/she
 * may not notice until it's too late.
 *
 * Upon EOF the file will not be closed.
 */

NSAPI_PUBLIC httpd_objset *INTobjset_scan_buffer(filebuffer *buf,
                                                 char *errstr, 
                                                 httpd_objset *os);

/*
 * INTobjset_create creates a new object set and returns a pointer to it.
 */

NSAPI_PUBLIC httpd_objset *INTobjset_create(void);

/*
 * INTobjset_free will free an object set, any associated objects, and any
 * associated Init functions.
 */

NSAPI_PUBLIC void INTobjset_free(httpd_objset *os);

/*
 * INTobjset_free_setonly frees only the object set, and not the associated
 * objects or init functions.
 */
NSAPI_PUBLIC void INTobjset_free_setonly(httpd_objset *os);

/*
 * INTobjset_new_object will add a new object to objset with the specified
 * name. It returns a pointer to the new object (which may be anywhere in 
 * the objset).
 */

NSAPI_PUBLIC
httpd_object *INTobjset_new_object(pblock *name, httpd_objset *os);

/*
 * INTobjset_add_object will add the existing object to os.
 */

NSAPI_PUBLIC void INTobjset_add_object(httpd_object *obj, httpd_objset *os);


/*
 * INTobjset_add_init will add the initialization function specified by 
 * initfn to the given object set. Modifies os->initfns.
 */
NSAPI_PUBLIC void INTobjset_add_init(pblock *initfn, httpd_objset *os);

/*
 * INTobjset_findbyname will find the object in objset having the given name,
 * and return the object if found, and NULL otherwise.
 * ign is a set of objects to ignore.
 */

NSAPI_PUBLIC httpd_object *INTobjset_findbyname(char *name, httpd_objset *ign, 
                                                httpd_objset *os);

/*
 * INTobjset_findbyppath will find the object in objset having the given 
 * partial path entry. Returns object if found, NULL otherwise.
 * ign is a set of objects to ignore.
 */

NSAPI_PUBLIC httpd_object *INTobjset_findbyppath(char *ppath,
                                                 httpd_objset *ign, 
                                                 httpd_objset *os);

NSPR_END_EXTERN_C

/* --- End function prototypes --- */

#define objset_scan_buffer INTobjset_scan_buffer
#define objset_create INTobjset_create
#define objset_free INTobjset_free
#define objset_free_setonly INTobjset_free_setonly
#define objset_new_object INTobjset_new_object
#define objset_add_object INTobjset_add_object
#define objset_add_init INTobjset_add_init
#define objset_findbyname INTobjset_findbyname
#define objset_findbyppath INTobjset_findbyppath

#endif /* INTNSAPI */

#endif /* !FRAME_OBJSET_H */
