/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
#ifndef __NSCNOUT_H
#define __NSCNOUT_H 1

#include "nsirinc.h"
#include "nsirmle.h"
#include "libnscore/nsbase.h"
#include "libnscore/nserrmsg.h"
#include "prstreams/prstrms.h"

// Class: ClientStream
// Description: 
//    Encapsulate SAF(NSAPI/SPAPI) output to the client with error logging
//    and keyed messages and an ostream i/f.
// 
//    Note: by default, the class does not use an NSerrmsg (error message file)
//    class to get it's keys.  The SetMsgFile function, once invoked, will set
//    the error message file for all subsequent Log calls.  Also, by default,
//    messages are logged with mode LOG_WARN.  Like SetMsgFile, the SetMode
//    function can be used to set the log mode.  Both functions return *this
//    to allow for daisy-chaining the calls.
//
//    The Put function call is used to send messages to the user (through the
//    session client socket descriptor.  The userStream returns the ofstream that
//    is attached to the socket descriptor, while Put sends the message.  Put
//    returns the ofstream, to allow for use of the insertion operator.
//
//    The error message file is passed in, rather than creating one privately 
//    using a filepath, in order to make use of a global error file, if one
//    existed, to avoid re-reading the file on each instance.


class ClientStream : public NSbase {
public:
   ClientStream(CStr fnx, Session*, Request*);
   ClientStream(CStr fnx, Session*, Request*, int delayFlag);
  ~ClientStream();

   // calls used to interface with the client

   PRofstream&    UserStream();
   ClientStream&  Put(CStr msgKey, NSavpair* opts=0, ostrstream* pOut = 0);
   NSbool         PutLogged() const;
   void           ClearPutLogged();

   // calls used to interface with the error log

   ClientStream&  SetMsgFile(NSerrmsg*);
   ClientStream&  SetContentHeader(CStr header);
   ClientStream&  SendHeader();
   ClientStream&  SetMode(int logMode = LOG_WARN);
   
   ClientStream&  Log(CStr msgKey);
   ClientStream&  Log(CStr msgKey, CStr parm);
   ClientStream&  Log(CStr msgKey, CStr parm, int rc);
   ClientStream&  Log(CStr msgKey, int parm);
    
   NSbool    ErrLogged() const;
   void      ClearErrLogged();

   // accessor functions

   Session* SessionBlock() const;
   Request* RequestBlock() const;  

protected:
   void  Dump(ostream&) const;   
   CStr  Name() const; 


private:
   Str        _contentHeader;
   NSbool     _errLogged;
   NSerrmsg*  _errMsgFile;
   NSbool     _headerSent;
   CStr       _funcname;
   int        _logMode;
   PRofstream _o;
   NSbool     _putLogged;
   Request*   _rq;
   Session*   _sn;
   
   ClientStream(const ClientStream&);        	    // make inaccessible
   ClientStream& operator=(const ClientStream&);    // make inaccessible
   };

// public member functions

inline void      
ClientStream::ClearErrLogged() {
   _errLogged = FALSE;
   }


inline void      
ClientStream::ClearPutLogged() {
   _putLogged = FALSE;
   }


inline NSbool    
ClientStream::ErrLogged() const {
   return(_errLogged);
   }


inline ClientStream&
ClientStream::Log(CStr msgKey) {
   Str omsg = _errMsgFile ? (Str)_errMsgFile->Get(msgKey) : (Str)msgKey;
   _errLogged = (_logMode != LOG_INFORM);
   log_error(_logMode, (Str)_funcname, _sn, _rq, omsg);
   return(*this);
   }
   

inline ClientStream&
ClientStream::Log(CStr msgKey, CStr parm) {
   Str omsg = _errMsgFile ? (Str)_errMsgFile->Get(msgKey) : (Str)msgKey;
   _errLogged = (_logMode != LOG_INFORM);
   log_error(_logMode, (Str)_funcname, _sn, _rq, "%s %s", omsg, parm);
   return(*this);
   }
   

inline ClientStream&
ClientStream::Log(CStr msgKey, CStr parm, int rc) {
   Str omsg = _errMsgFile ? (Str)_errMsgFile->Get(msgKey) : (Str)msgKey;
   _errLogged = (_logMode != LOG_INFORM);
   log_error(_logMode, (Str)_funcname, _sn, _rq, "%s %s: %d", omsg, parm, rc);
   return(*this);
   }
   

inline ClientStream&
ClientStream::Log(CStr msgKey, int parm) {
   Str omsg = _errMsgFile ? (Str)_errMsgFile->Get(msgKey) : (Str)msgKey;
   _errLogged = TRUE;
   log_error(_logMode, (Str)_funcname, _sn, _rq, "%s %d", omsg, parm);
   return(*this);
   }
   

inline NSbool    
ClientStream::PutLogged() const {
   return(_putLogged);
   }


inline Request*
ClientStream::RequestBlock() const {
   return(_rq);
   }


inline Session*
ClientStream::SessionBlock() const {
   return(_sn);
   }

inline PRofstream&  
ClientStream::UserStream() {
   return(_o);
   }


inline ClientStream&       
ClientStream::SetMode(int logMode) {
   _logMode = logMode;
   return(*this);
   }


inline ClientStream& 
ClientStream::SetMsgFile(NSerrmsg* msgFile) {
   _errMsgFile = msgFile;
   return(*this);
   }
   

#endif
