/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
//
// Name:    nssrch.h
// Contact: mjb<barbarin@netscape.com>
// Notes:
//   Encapsulates text searching.  Uses NSsearchArgs and NSsearchResults
//   as parameters.
//
//   Instantiating NSsearch creates a search session.  Once a session is
//   created, then all collections to be search must be opened.  Once this
//   is done, then searches can be done.  The search calls do the following:
//      *Search:  searches the specified collections (specified in NSsearchArgs
//          for the request) for documents matching the query and returns the
//          requested results in the NSsearchResults object.
//      *GetMore: redoes the search, but returns more results, beginning with
//          the specified offset (note that Search returns a number of documents
//          with an upper limit determined by NSsearchArgs (which defaults to 
//          20, if no upper limit is specified).
//      *SearchDoc: processes the query against a specified document.  this is 
//          most often used to retrieve a highlighted document (via the GetDoc
//          call), if the search results are not available (and only the query
//          and document are available, such as in the stateless environment
//          of the HTTP server)
//      *GetDoc: retrieves a highlighted document.
//
//   Additionally, there is a MapAttributes call that can be used to map a 
//   set of attribute name aliases for a specified collection.  This call was a
//   last minute modification for WebPublishing, and to expedite that mod I 
//   allowed for the introduction of an NSavpair object in the interface.  This 
//   needs to be removed in order to preserve the abstract interface that we'd
//   require in order to abstract the underlying search engine for other 
//   vendors.
//
//   There are two optional callbacks that are used by NSsearch.  One for the
//   ctor, which allows for installing a callback for intercepting the error
//   number and message from the underlying search engine.  One for intercepting
//   the document stream in GetDoc.  This last is useful for translating the
//   output document (for example, output of ascii documents to a browser 
//   requires escaping special characters such as '<' and '>').
//
//   Opening a collection can optionally fill in a collection status structure.
//   For the nsir.so plugin, this is used to determine collection status and 
//   size for the contents (collection description) display.
// -------------------------------------------------------------------------------
#ifndef __NSSRCH_H
#define __NSSRCH_H 1

#include "libir/nsirutil.h"
#include "libir/nssrchargs.h"
#include "libir/nssrchres.h"
#include "libnscore/nsavpair.h"

typedef void (*NSsearchCB)(int errNum, CStr errMsg, void* cbParm);

typedef int (*NSsearchGetDocCB)(const char* buf, ostream&, void* parm);


enum	{COLLECTION_STATUS_UNKNOWN=0, 
	COLLECTION_OPEN, 
	COLLECTION_CLOSED, 
	COLLECTION_BROKEN};


struct NSsearchOpenStruct {
   int collectionSize;
   int status;
   };


class NSsearch : public NSbase {
public:
   NSsearch(CStr initParms, 
      NSsearchCB=0, void* cbParm=0, CStr langParms=0, void* parm= 0);
  ~NSsearch();

   NSsearch&  OpenCollection(CStr collectionPath, void*& collHdl, 
      NSsearchOpenStruct* = 0);
   NSsearch&  MapAttributes(void* collHdl, NSavpair&);

   // basic search interface -- note that GetDoc will redo the search against the
   // selected document (this is in case the request is a different session)
   // GetDoc is used primarily to retrieve a document with highlighting 
   // characters inserted.

   NSsearch&  Search(NSsearchResults&, NSsearchArgs&);
   NSsearch&  GetMore(NSsearchResults&, NSsearchArgs&, int searchOffset);
   NSsearch&  SearchDoc(NSsearchResults&, NSsearchArgs&, CStr docPath);
   NSsearch&  GetDoc(NSsearchResults&, NSsearchArgs&, ostream&, 
      NSsearchGetDocCB=0, int offset=0);
   
protected:
   void  Dump(ostream&) const;   
   CStr  Name() const; 

private:
   void*  _sessionData;

   NSsearch(const NSsearch&);         		// make inaccessible
   NSsearch& operator=(const NSsearch&);	// make inaccessible
   };
#endif
