/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

#ifndef nsbase_h
#define nsbase_h

/***********************************************************************
** Includes
***********************************************************************/
#include <string.h>

#include "nserrno.h"


//**********************************************************************
// CLASS:	NSBase
// DESCRIPTION:
//   Base class for most of the object in the core system (as well as
//   others).  This class, when derived from, provides a consistent model
//   and interface for error handling that is lightweight and easy to use.
//
//   This is accomplished by defining an interface with accessor functions
//   "Good()" and "Bad()" to check whether the object is Good or Bad; the
//   "Errnum()" and "Errsrc()" functions return the current error number and
//   the Name() of the object that last generated this error (more on Errsrc
//   later).  There is a Clear() function that is used to clear the error
//   state.  Objects should NEVER set Errnum(0) as a replacement for Clear.
//   This is dangerous since it implies knowledge of what the underlying
//   implementation uses for the non-error state.  And, of course, this
//   could change.
//
//   In addition to the public functions, there are several protected
//   members that are only accessible to derived classes (and are therefore
//   not part of the public interface of any class derived from NSbase).
//   These functions are used to set the error state: Errnum(u16),
//   Errnum(NSbase*), and Errnum(NSbase&).  The last two functions will set
//   the error number to the error number of the passed-in object AND will
//   set Errsrc to the Name() of the passed-in-object.  Subsequent calls to
//   these last two functions will not reset the error number or error
//   source.  The Clear function will reset these to the non-error state and
//   set the error source to the current object Name().
//
//   The Errnum(u16) function will always reset the error number and error
//   source (the last will be set back to the Name of the current object).
//
//   Use Errnum(u16) at the lowest levels to set a particular error state
//   and use Errnum(NSbase*) and Errnum(NSbase&) to percolate error state
//   from captive objects.
//
//   This class requires derived classes to implement two protected
//   functions: first is "Dump" and second is "Name".  The Dump function is
//   invoked via the friend function "operator<<".  This gives users of
//   derived objects the ability to "cout" their objects for debugging
//   purposes (see example).  Note the the dump function is implemented for
//   NSbase&, so it doesn't work with pointers-to-objects
//
//   The second function, "Name", returns the name of the object -- i.e.,
//   the class name.  This value, as well as the error, the object causing
//   the last error (more on this later), and a count of all objects
//   instantiated in the program (more or less -- since this is a static,
//   and for efficiency purposes it is not gated, there may be inaccurate
//   counts in a threaded program -- which is why the count is a s32, and
//   not a u32).
//
//   In the example program, for ObjectA, the Test1 function, sets the
//   errnum to 1, then forces the b object to have an errnum of 2, then sets
//   ObjectA's errnum to b's errnum to 2.  Note that since the last Errnum
//   call in Test1 passes an Errnum (from b), the Errsrc is ObjectA.
//
//   Test2 just sets the ObjectA Errnum and Errsrc to Object B's state.
//
//   Test3 sets the Errnum and Errsrc to Object B's, but then resets the
//   Errsrc to ObjectA and sets the Errnum to Object B's by invoking Errnum
//   with a error number.
//
// EXAMPLE:
//    #include "nsbase.h"
//    
//    class ObjectB : public NSbase{
//    public:
//       ObjectB() {};
//       void SetErr() {Errnum(2);}
//    protected:
//       void  Dump(ostream&) const {};
//       CStr  Name() const {return("ObjectB");}   
//       };
//    
//    class ObjectA : public NSbase {
//    public:
//       ObjectA() {}
//       void Test1(){Clear(); Errnum(1); b.SetErr(); Errnum(b.Errnum());}
//       void Test2(){Clear(); b.SetErr(); Errnum(b);}
//       void Test3(){Clear(); b.SetErr(); Errnum(&b);  Errnum(3);}
//    protected:
//       void  Dump(ostream&) const {};
//       CStr  Name() const {return("ObjectA");}
//    private:
//       ObjectB b;
//       };
//    
//    int 
//    main(int argc, char** argv) {
//       ObjectA a;
//       a.Test1();
//       cout << a;
//       cout << "Error source object=" << a.Errsrc() << endl;
//       a.Test2();
//       cout << a;
//       cout << "Error source object=" << a.Errsrc() << endl;
//       a.Test3();
//       cout << a;
//       cout << "Error source object=" << a.Errsrc() << endl;
//       }
//
//    /* Sample output:
//    05/23/96 03:04:09::ObjectA
//    Error = 2, source object causing error = ObjectA, objcount = 2
//    Error source object=ObjectA
//    05/23/96 03:04:09::ObjectA
//    Error = 2, source object causing error = ObjectB, objcount = 2
//    Error source object=ObjectB
//    05/23/96 03:04:09::ObjectA
//    Error = 3, source object causing error = ObjectA, objcount = 2
//    Error source object=ObjectA
//    */
//
// RESTRICTIONS:
// _objcount - May become inconsistent when using with threads.
//
// RETURN CODE TABLE:
//
//**********************************************************************
class NSbase
{
    // FRIENDS
        friend ostream& operator<< (ostream&, const NSbase&);

public:
    // CONSTRUCTORS
        NSbase( );
        NSbase( const NSbase& );

    // DESTRUCTORS
        virtual
	~NSbase( );

    // ATTRIBUTE ACCESS
        NSbool  	Good() const;
        NSbool 		Bad() const;
        u16     	Errnum() const;
        CStr		Errsrc() const;

    // ATTRIBUTE MODIFICATION
	void    	Clear();

    // OPERATORS
        void		operator=(const NSbase&);

protected:
	u16		Errnum(u16);
        u16		Errnum(NSbase*);
        u16		Errnum(NSbase&);

        virtual
	void		Dump(ostream&) const = 0;
        virtual
	CStr		Name() const = 0;

private:
        enum {
	    MAXSZ_SRC = 32
	};

        u16		_err;
        char		_errSrc[ MAXSZ_SRC ];

        static
	s32		_objcount;
};


/***********************************************************************
** NSBase Inline Functions:
***********************************************************************/

/***********************************************************************
** FUNCTION:	NSbase
** DESCRIPTION:
**   Increments the static total object count and sets the error source
**   to be the value in Name() 
***********************************************************************/
inline NSbase::NSbase() : _err(0) {
   _errSrc[0] = 0;
   _objcount++;
   }

/***********************************************************************
** FUNCTION:	NSbase
** DESCRIPTION:
**   This is the copy constructor.  It does the same as the default
**   constructor for NSbase.
***********************************************************************/
inline NSbase::NSbase(const NSbase& from) 
:  _err(from._err) {
   strcpy(_errSrc, from._errSrc);
   _objcount++;
   }

/***********************************************************************
** FUNCTION:	Bad
** DESCRIPTION:
**   Returns TRUE if the object is bad (an error has been set);
**   otherwise FALSE.
***********************************************************************/
inline NSbool  NSbase::Bad() const {
   return(_err != 0);
   }

/***********************************************************************
** FUNCTION:	Clean
** DESCRIPTION:
**   Clears the error state and sets the source to Name()
***********************************************************************/
inline void NSbase::Clear() {
   _err = 0;
   _errSrc[0] = 0;
   }

/***********************************************************************
** FUNCTION:	Errnum
** DESCRIPTION:
**   Returns the current error number
***********************************************************************/
inline u16 NSbase::Errnum() const {
   return(_err);
   }

/***********************************************************************
** FUNCTION:	Errnum
** DESCRIPTION:
**   Sets the error state to the passed in value and sets the error src
**   to 0 (basically).  This last is used as a flag by Errsrc to
**   determine whether the Errsrc should be Name() or a passed-in value
**   (via the Errnum(NSbase*) or Errnum(NSbase&) functions).
***********************************************************************/
inline u16 NSbase::Errnum(u16 x) {
   u16 prev = _err;
   _err = x;
   _errSrc[0] = 0;
   return(prev);
   }

/***********************************************************************
** FUNCTION:	Errnum
** DESCRIPTION:
**   Takes a pointer to an object derived from NSbase and sets the
**   error number and error source to the values from the passed-in
**   object
***********************************************************************/
inline u16 NSbase::Errnum(NSbase* pObj) {
   u16 prev = _err;
   if (_err == 0) {
      strncpy(_errSrc, pObj->Name(), sizeof(_errSrc));
      _err = pObj->Errnum();
      }
   return(prev);
   }

/***********************************************************************
** FUNCTION:	Errnum
** DESCRIPTION:
**   Takes a reference to an object derived from NSbase and sets the
**   error number and error source to the values from the passed-in
**   object
***********************************************************************/
inline u16 NSbase::Errnum(NSbase& rObj) {
   u16 prev = _err;
   if (_err == 0) {
      strncpy(_errSrc, rObj.Name(), sizeof(_errSrc));
      _err = rObj.Errnum();
      }
   return(prev);
   }

/***********************************************************************
** FUNCTION:	Errsrc
** DESCRIPTION:
**   Returns the Name() of the object that caused the error condition
***********************************************************************/
inline CStr NSbase::Errsrc() const {
   return(_errSrc[0] == 0 ? Name() : _errSrc);
   }

/***********************************************************************
** FUNCTION:	Good
** DESCRIPTION:
**   Implemented with respect to Bad().  Returns TRUE if the object is
**   not in an error state, otherwise FALSE.
***********************************************************************/
inline NSbool  NSbase::Good() const {
   return(!Bad());
   }

/***********************************************************************
** FUNCTION:	operator=
** DESCRIPTION:
**   Assignment operator.
***********************************************************************/
inline void NSbase::operator=(const NSbase& from) {
   _err = from._err;
   strcpy(_errSrc, from._errSrc);
   }

#endif /* nsbase_h */
