/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

#ifndef nscfg_h
#define nscfg_h

/***********************************************************************
** Includes
***********************************************************************/
#include <fstream.h>

#include "nsbase.h"


/***********************************************************************
** Functional Typedef
***********************************************************************/
typedef NSbool  (*NScfgGet)(CStr attr, CStr val, PGen parm); 


//**********************************************************************
// CLASS:	NScfgNode
// DESCRIPTION:	Node to encapsulate data for NScfg class.
//
// EXAMPLE:
//
// RESTRICTIONS:
//
// RETURN CODE TABLE:
//**********************************************************************
class NScfgNode {
   friend class NScfg;
public:
   NScfgNode() :data(0), lineType('V'), nextNode(0) {}
   NScfgNode(size_t size) :  data(new char[size]), lineType('V'), nextNode(0) {}
  ~NScfgNode() { delete []data;}

private:
   Str        data;
   char       lineType;
   NScfgNode* nextNode;   
   };


//**********************************************************************
// CLASS:	NScfg
// DESCRIPTION:	Implementation for system.ini-type initialization files
//
// EXAMPLE:
//    Implementation of ".INI" facility.  Used for reading/writing .INI
//    files of the form:
//
//    [section-name1]
//    tag1=value
//    tag2=value2
//
//    [section-name2]
//    tag1=value
//    tag2=value2
//
//    ... and so on.
//
// RESTRICTIONS:
//    Using root.data as temporary storage in several member functions,
//    especially the Init() function.
//
// RETURN CODE TABLE:
//**********************************************************************
class NScfg : public NSbase {
public:
   enum Filemode {READONLY, READWRITE};

   NScfg();
   NScfg(CStr fname, Filemode=READONLY);
   NScfg(CStr fname, CStr* paths, Filemode=READONLY);
  ~NScfg();

   NScfg& Open(CStr fname, Filemode=READONLY);
   NScfg& Open(CStr fname, CStr* paths, Filemode=READONLY);
   void   Clear();
   NScfg& Close();  // does a Flush as well
   NScfg& Flush();
   NScfg& RequireFlush(NSbool  isrequired=TRUE); // no implicit changes

   NScfg& Get(CStr sec, CStr ent, u32& integer);
   NScfg& Get(CStr sec, CStr ent, Str& buf);
   NScfg& Get(CStr sec, NScfgGet, PGen parm);

   NScfg& Put(Str sec);
   NScfg& Put(Str sec, Str ent, Str buf);

   NScfg& Delete(CStr sec);
   NScfg& Delete(CStr sec, CStr ent);

   Str Section(NSbool  reset = FALSE);

protected:
   void  Dump(ostream&) const;
   CStr  Name() const;

private:
   fstream      _cfgfile;
   NSbool 	_flushRequired;
   Filemode     _filemode;
   char         _fname[_POSIX_PATH_MAX];
   NSbool       _isMod;
   NSbool       _isOpen;
   NScfgNode*   _lastNode;
   NScfgNode*   _pSection;
   NScfgNode*   _root;
   Str          _secname;
   size_t       _secnamelen;

   void        CleanUp();
   NScfgNode*  Find(CStr sec, NScfgNode*& prevnode);
   NScfgNode*  Find(CStr sec, CStr ent, NScfgNode*& prevnode);
   NSbool      Init();
   NScfgNode*  PutEntry(Str ent, Str buf, NScfgNode* prevnode);
   NScfgNode*  PutSection(Str sec);
   Str         RTB(Str srcEnd);
   void        SkipWS(Str& src);

   NScfg& operator=(const NScfg&);       // make inaccessible
   NScfg(const NScfg&);                  // make inaccessible
   };


/***********************************************************************
** NScfg Inline Functions
***********************************************************************/
inline void NScfg::Clear() {
   NSbase::Clear();
   _cfgfile.clear();
   }

inline Str NScfg::RTB(Str src) {
   while (*src == ' ' || *src == '\t')
      --src;
   return(src+1);
   }


inline void NScfg::SkipWS(Str& src) {
   while (*src == ' ' || *src == '\t')
      src++;
   }

#endif /* nscfg_h */
