/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
//
// Name:    nscfgn.h
// Contact: mjb<barbarin@netscape.com>
// Notes:
//    Class to provide access to a configuration file similar to the Win3.1
//    system.ini files.  This is not MT-safe (in the strictest sense).  This
//    class encapsulates a file, and the functions that require open/write
//    functionality shouldn't be used in a threaded program unless access to the
//    functions is protected.  The read-only functions should be thread-safe,
//    though.  Note that debug functions may not be thread-safe (just didn't
//    want to spend the time).  Also, the in-memory data is stored as a linked
//    list, so searches are linear.  Which leads to the admonition of keeping the
//    configuration file reasonable in length (okay, we could use a btree to 
//    speed things up, but for now let's keep things simple).
//
//    Basically, the class still derives from NSbase.  Instead of having the
//    member functions returning "*this", though, they return an int.  In the
//    MT environment you might create a "global" instance of NScfgn in readonly
//    mode (which generates a linked-list in memory from the contents of the
//    file).  Different threads may be accessing this list through the Get 
//    functions.  Using the Good/Bad/Errnum functions of IAbase would mean that
//    all of the Get functions in each of the threads would be using the same
//    error number integer from the global instance -- which is not good.
//
//    So I changed the "Get"-type functions to return integers instead of using
//    the error number from NSbase.
//
// Example:
//    Implementation of ".INI" facility.  Used for reading/writing .INI
//    files of the form:
//
//    [section-name1]
//    tag1=value
//    tag2=value2
//
//    [section-name2]
//    tag1=value
//    tag2=value2
//
//    ... and so on.
//
// -------------------------------------------------------------------------------

#ifndef __NSCFGN_H
#define __NSCFGN_H 1

#include <fstream.h>
#include "nsbase.h"


class   NScfgnNode; // forward declartion (see end of this file);
typedef NSbool (*NScfgnGetCB)(CStr attr, CStr val, void* parm); 

class NScfgn : public NSbase {
public:
   enum Filemode {READONLY, READWRITE};

   NScfgn();
   NScfgn(CStr fname, Filemode=READONLY);
   NScfgn(CStr fname, CStr* paths, Filemode=READONLY);
  ~NScfgn();

   // functions to open/close the configuration file

   NScfgn& Open(CStr fname, Filemode=READONLY);
   NScfgn& Open(CStr fname, CStr* paths, Filemode=READONLY);
   NScfgn& Close(); 
   void    DoNotCommit();

   // functions to retrieve values from the configuration file

   NSbool Get(CStr sec, CStr ent, int& val);
   NSbool Get(CStr sec, CStr ent, Str& buf);
   NSbool Get(CStr sec, NScfgnGetCB, void* parm);
   NSbool Section(Str secname, int seclen, void*& cookie);


   // functions to set values for the configuration file (make sure these are in a
   // critical section, if you're threading)

   NSbool Put(CStr sec);
   NSbool Put(CStr sec, CStr ent, CStr buf);
   NSbool Delete(CStr sec);
   NSbool Delete(CStr sec, CStr ent);

protected:
   void  Dump(ostream&) const;
   CStr  Name() const;

private:
   fstream      _cfgfile;
   Filemode     _filemode;
   char         _fname[_POSIX_PATH_MAX];
   NSbool       _isMod;
   NSbool       _isOpen;
   NScfgnNode*  _lastNode;
   NScfgnNode*  _root;

   NScfgnNode*  Find(CStr sec, NScfgnNode*& prevnode);
   NScfgnNode*  Find(CStr sec, CStr ent, NScfgnNode*& prevnode);
   NSbool       Flush();
   NSbool       Init();
   NScfgnNode*  PutEntry(CStr ent, CStr buf, NScfgnNode* prevnode);
   NSbool       PutSection(CStr sec);
   Str          RTB(Str srcEnd);

   NScfgn& operator=(const NScfgn&);        // make inaccessible
   NScfgn(const NScfgn&);                   // make inaccessible
   };


// public member functions

inline  void  
NScfgn::DoNotCommit() {
   _isMod = FALSE;
   }


// private classes

class NScfgnNode {
   friend class NScfgn;
private:
   NScfgnNode() : data(0), lineType('V'), nextNode(0) {}
   NScfgnNode(int size) : data(new char[size]), lineType('V'), nextNode(0) {}
  ~NScfgnNode() {delete []data;}

   Str         data;
   char        lineType;
   NScfgnNode* nextNode;   
   };


#endif
