/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

#ifndef nsfile_h
#define nsfile_h

/***********************************************************************
** Includes
***********************************************************************/
// Someone at some point, somewhere, will have to define OS_windows
// Currently the makefiles do not, since they are in a UNIX environment
#ifdef OS_windows
#include <io.h>
#endif

#include "nsbase.h"


//**********************************************************************
// CLASS:	NSfile
// DESCRIPTION:	Encapsulation of file pointer.
//
// EXAMPLE:
//
// RESTRICTIONS:
//
// RETURN CODE TABLE:
//**********************************************************************
class NSfile: public NSbase {
public:
   // APPEND is READWRITE but the file is not truncated when opened
   enum NSfilemode {READONLY, READWRITE, APPEND, UPDATE};
   enum NSbuffermode {UNBUFFERED, BUFFERED};

   NSfile();
   NSfile(CStr name, NSfilemode=READONLY);
   NSfile(CStr name, CStr* paths, NSfilemode=READONLY);
  ~NSfile();

   virtual NSfile&  Open(CStr name, NSfilemode=READONLY);
   virtual NSfile&  Open(CStr name, CStr* paths, NSfilemode=READONLY);
   virtual NSfile&  Close();

   NSfile&  Flush();
   NSfile&  Read(Str buf);
   NSfile&  Readln(Str buf, s16 maxlen);
   NSfile&  Read(s16 size, Str buf);
   NSfile&  Read(s32 offset, s16 size, Str buf);
   NSfile&  Seek(s32 offset=0, s16=SEEK_SET);
   NSfile&  SeekToBegin();
   NSfile&  SeekToEnd();
   NSfile&  Truncate(s32 newsz);
   NSfile&  Write(s16 size, CStr buf);
   NSfile&  Write(s32 offset, s16 size, CStr buf);

   // accessor functions

   s32      BytesTransferred();
   s16      FileHdl() const;
   NSbool   IsOpen() const;
   s32      Size();
   s32      Tell();

   static NSbool  Exists(CStr name);

protected:
   void  Dump(ostream&) const;         // from NSbase
   CStr  Name() const;                 // from NSbase

private:
   enum {BUFSZ = 8192};

   s32       _bytecnt;
   Str       _filebuf;
   NSbool    _isopen;
   FILE*     _pFile;

   NSfile(const NSfile&);		// make inaccessible
   NSfile& operator=(const NSfile&);    // make inaccessible
   };


/***********************************************************************
** NSfile Inline Functions
***********************************************************************/
// public member functions
inline s32 NSfile::BytesTransferred() {
   return(_bytecnt);
   }


inline NSbool  NSfile::Exists(CStr name) {
   return(access(name, 0) != -1);
   }


inline s16 NSfile::FileHdl() const {
   return(fileno(_pFile));
   }


inline NSfile& NSfile::Flush() {
   #ifdef XP_UNIX
      if (fflush(_pFile) != 0)
         Errnum(fileFlushErr);
   #else
      int duphandle = dup(fileno(_pFile));  // forces update of FAT on DOS
      if (close(duphandle) != 0)
         Errnum(fileCloseErr);
   #endif

   return *this;
   }


inline NSbool  NSfile::IsOpen() const {
   return(_isopen);
   }


inline NSfile& NSfile::Read(Str buf) {
   while ((_bytecnt = fread((void*)buf, 1, 1, _pFile)) > 0) {
      if (*buf == '\n') {
         *buf = '\0';
         break;
         }
      buf++;
      }

   if (_bytecnt < 0)
      Errnum(fileReadErr);
   else if (_bytecnt == 0) 
      Errnum(fileEOF);
   return(*this);
   }


inline NSfile& NSfile::Read(s16 bytes, Str buf) {
   if ((_bytecnt = fread((void*)buf, 1, bytes, _pFile)) < 0)
      Errnum(fileReadErr);
   else if (_bytecnt == 0) 
      Errnum(fileEOF);
   return(*this);
   }


inline NSfile& NSfile::Seek(s32 offset, s16 from) {
   if (fseek(_pFile, offset, from) != 0)
      Errnum(fileSeekErr);
   return(*this);
   }


inline NSfile& NSfile::Read(s32 offset, s16 bytes, Str buf) {
   if (Seek(offset).Good())
      Read(bytes, buf);
   return(*this);
   }


inline NSfile& NSfile::Readln(Str buf, s16 maxlen) {
   if (fgets(buf, maxlen, _pFile) == 0) 
      Errnum(fileEOF);
   _bytecnt = strlen(buf);

   if (buf[_bytecnt - 1] == '\n')
      buf[_bytecnt - 1] = '\0';  // get rid of end of line char

   return(*this);
   }


inline NSfile& NSfile::SeekToBegin() {
   if (fseek(_pFile, 0L, SEEK_SET) != 0)
      Errnum(fileSeekErr);
   return(*this);
   }


inline NSfile& NSfile::SeekToEnd() {
   if (fseek(_pFile, 0L, SEEK_END) != 0)
      Errnum(fileSeekErr);
   return(*this);
   }


inline s32 NSfile::Tell() {
   s32 loc = ftell(_pFile);
   if (loc == -1L)
      Errnum(fileTellErr);
   return(loc);
   }


inline s32 NSfile::Size() {
   s32 loc = Tell();
   SeekToEnd();
   s32 size = Tell();
   Seek(loc);
   return(Good() ? size : 0);
   }


inline NSfile& NSfile::Truncate(s32 sz) {
   if (nsChsize(fileno(_pFile), sz) != 0)
      Errnum(fileChsizeErr);
   return(*this);
   }


inline NSfile& NSfile::Write(s16 bytes, CStr buf) {
   if ((_bytecnt = fwrite((void*)buf, 1, bytes, _pFile)) <= 0)
      Errnum(fileWriteErr);
   return(*this);
   }


inline NSfile& NSfile::Write(s32 offset, s16 bytes, CStr buf) {
   if (Seek(offset).Good())
      Write(bytes, buf);
   return(*this);
   }

#endif /* nsfile_h */
