/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

#ifndef nsskplst_h
#define nsskplst_h

/***********************************************************************
** Includes
***********************************************************************/
#include "nsbase.h"


//**********************************************************************
// CLASS:	NSskiplist
// DESCRIPTION:	Simple skiplist implementation
//
// EXAMPLE:
//
// RESTRICTIONS:
//
// RETURN CODE TABLE:
//**********************************************************************
class NSskiplist : public NSbase {
public:
   NSskiplist();
  ~NSskiplist();

   NSskiplist&  Delete(CStr key);
   NSskiplist&  Get(CStr key, Str& data);
   NSskiplist&  Get(CStr key, Str& data, PGen& cookie);
   NSskiplist&  Put(CStr key, CStr data, NSbool  allowDupes = FALSE);

   NSskiplist&  First(Str& key, Str& data);
   NSskiplist&  First(Str& key, Str& data, PGen& cookie);
   NSskiplist&  Last(Str& key, Str& data);
   NSskiplist&  Next(Str& key, Str& data, PGen& cookie);

   u16 Count() const;

   struct SLnode {
      Str      key;
      Str      data;
      SLnode*  fwd[1];
      };

protected:
   void  Dump(ostream&) const;   
   CStr  Name() const; 

private:
   enum {LEVELS = 16, BITS_IN_RANDOM = 31};

   u16      _count;
   s16      _currlevel;
   s32      _randomBits;
   s32      _randomsLeft;
   SLnode*  _start;
   SLnode*  _update[LEVELS];

   NSskiplist& Get(CStr key, SLnode*&);
   PGen        NewNode(u16 lvls);
   s16         RandomLevel();

   NSskiplist&  operator=(const NSskiplist&);    // make inaccessible
   NSskiplist(const NSskiplist&);       	 // make inaccessible
   };


/***********************************************************************
** NSskiplist Inline Functions
***********************************************************************/
inline PGen
NSskiplist::NewNode(u16 lvls) {
   u16    spaceToAllocate = sizeof(SLnode) + (lvls * sizeof(SLnode*));
   SLnode* node = (SLnode*) new char[spaceToAllocate];

   if (node == 0)
      Errnum(outOfMem);

   memset(node, 0, spaceToAllocate);

   return(node);
   }

inline u16 
NSskiplist::Count() const {
   return(_count);
   }

inline NSskiplist& 
NSskiplist::First(Str& key, Str& data) {
   Clear();
   PGen p;
   First(key, data, p);
   return(*this);
   }

inline NSskiplist& 
NSskiplist::Get(CStr key, Str& data) {
   SLnode* p;
   Get(key, p);
   data = (p != 0) ? p->data : 0;
   return(*this);
   }

inline NSskiplist& 
NSskiplist::Get(CStr key, Str& data, PGen& cookie) {
   SLnode* p;
   Get(key, p);
   data = (p != 0) ? p->data : 0;
   cookie = (p != 0) ? p->fwd[0] : 0;
   return(*this);
   }

#endif /* nsskplst_h */
