/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

#ifndef NS_SNMP_HTTP_POLL_H
#define NS_SNMP_HTTP_POLL_H

#ifdef XP_WIN32
#include <windows.h>
#include <winsock.h>
#include <time.h>
#else
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#ifdef AIX
#include <sys/select.h>
#endif
#endif


#define MYTRACE(x)
#define MYTRACE2(x)



/*
 * Don't bother reading more than this in case the HTTP
 * server attempts to send more.
 *
 */
#define MAX_READ_SIZE		8096


/*
 * Quit reading more if already this much happens to be read.
 *
 */
#define SUFFICIENT_READ_SIZE	8000


/*
 * Time out if request not completed in this many seconds.
 *
 */
#define NS_SNMP_HTTP_TIMEOUT	5

/*
 * Poll the HTTP server every this many seconds.
 *
 */
#define NS_SNMP_HTTP_POLL_INTERVAL	30


/*
 * Current status of the server.
 *
 */
typedef enum _ServerStatus {
    SERVER_STATUS_UNKNOWN	= 0,	/* status unknown */
    SERVER_STATUS_DOWN		= 1,	/* server down */
    SERVER_STATUS_HUNG_CONNECT	= 2,	/* hangs connecting (all srvrs busy) */
    SERVER_STATUS_BROKEN	= 3,	/* closes connection abnormally */
    SERVER_STATUS_HUNG_SERVICE	= 4,	/* hangs before any resp recv'd */
    SERVER_STATUS_UP		= 5,	/* up and running */
    SERVER_STATUS_MAX		= 6	/* used for table dimension */
} ServerStatus;


/*
 * The state of the current connection.
 *
 *
 */
typedef enum _ConnState {
    CS_INVALID	= 0,
    CS_IDLE	= 1,	/* idle -- not connected or active at the time */
    CS_CONNECT	= 2,	/* start and finish connecting */
    CS_SEND	= 3,	/* start and finish sending request */
    CS_RECV	= 4,	/* start and finish receiving response */
    CS_RECV_CONT= 5,	/* start and finish receiving response */
    CS_ERROR	= 6,	/* error state; go to close */
    CS_CLOSE	= 7,	/* close socket and go idle */
    CS_MAX	= 8	/* used for table dimension */
} ConnState;


/*
 * Context for polling.
 *
 */
typedef struct _PollContext {

    /* for debugging purposes */
    char *host;
    int   port;

    /* data needed for the connection state machine */
    ConnState		state;	/* state of current connection */
    struct timeval	state_time_stamps[CS_MAX];
    int			sd;
    struct sockaddr_in	sa;
    struct sockaddr_in	sa_clean;
    int			use_security;
    char *		url;	/* URL to request (path part only) */
    char		buf[MAX_READ_SIZE + 1];
    int			outind;
    int			outcnt;
    int			incnt;

    /* statistics */
    ServerStatus	status;	/* current status of server */
    u_long		poll_stats[SERVER_STATUS_MAX];
    time_t		status_time_stamps[SERVER_STATUS_MAX];

} PollContext;



/*
 * Socket descriptors used by this module.
 * These can be used to do select() over them, and then call
 * http_handle_events() to actually handle those sockets.
 *
 */
extern fd_set poll_rd_fds;
extern fd_set poll_wr_fds;
extern int    poll_n_fds;



/*
 * Set up polling of the server on the specified host and port.
 * If use_security is true, uses HTTPS, otherwise HTTP.
 * Asks for the URL specified (only the path part of the URL).
 *
 * Returns context that is passed in subsequent calls
 * to this module.
 *
 * Returns NULL if host doesn't DNS resolve, or if notices
 * right away that the server is down.
 *
 * This function doesn't actually trigger the poll yet, merely
 * allocates and sets up the context to do so.
 *
 * http_start_poll() should be called every time a poll is to
 * be triggered.
 *
 * http_handle_events() has to be called every time there is
 * an event ready for processing in the socket descriptor
 * cx->sd.
 *
 */
PollContext *http_setup_poll(char *hostname, int port,
			     int use_security, char *url);


/*
 * Terminate any connections, and free memory pointed to by cx.
 *
 *
 */
void http_setup_free(PollContext *cx);


/*
 * Start the next poll.
 *
 * If the previous poll is still in progress, it will be interrupted
 * and marked as a timed out connection.
 *
 *
 */
int http_start_poll(PollContext *cx);


/*
 * Handles async HTTP events, until nothing more to process.
 * The caller should call this every time the socket is ready
 * for read or write (whichever it is that is blocking this).
 *
 */
int http_handle_events(PollContext *cx);


#endif

