/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
#ifndef PUBLIC_LIBPROXY_PUTIL_H
#define PUBLIC_LIBPROXY_PUTIL_H

/*
 * File:        putil.h
 *
 * Description: Proxy utility function
 * Ari Luotonen
 */

#ifndef PUBLIC_NETSITE_H
#include "../netsite.h"
#endif /* !PUBLIC_NETSITE_H */

#include "base/pblock.h"
#include <time.h>

/* --- Begin constant definitions --- */

/*
 * Max hostname length after it's sterilized -- that is, not allowing
 * longer than this, plus not allowing any non-alpha non-punct chars.
 */
#define MAX_STERILIZED_HLEN     64

/* --- End constant definitions --- */

/* --- Begin macro definitions --- */

#define STRALLOCCOPY(s,d)   system_StrAllocCopy(&s,d)
#define STRALLOCCAT(s,d)   system_StrAllocCat(&s,d)

/* --- End macro definitions --- */

/* --- Begin function prototypes --- */

NSPR_BEGIN_EXTERN_C

char *system_StrAllocCopy(char **dest, char *src);
char *system_StrAllocCat(char **dest, char *src);


/*
 * Returns the given localtime in GMT, or if (time_t)0 given
 * returns current time in GMT.
 */
NSAPI_PUBLIC time_t SPutil_make_gmt(time_t t);


/*
 * Returns the given GMT time in localtime.
 */
NSAPI_PUBLIC time_t SPutil_make_local(time_t t);


/*
 * Returns current time in GMT.
 */
NSAPI_PUBLIC time_t SPutil_get_current_gmt(void);


/*
 * Format the time to format that suits access logs.
 * If tt is (time_t)0, will get the current time.
 */
NSAPI_PUBLIC char *SPutil_make_log_time(time_t tt);


/*
 * Parses the HTTP time string into time_t.
 *
 */
NSAPI_PUBLIC time_t SPutil_parse_http_time(char *date_string);


/*
 * As pblock_nvinsert(), but take an integer or a long
 * instead of a string value.
 */
NSAPI_PUBLIC pb_param *SPpblock_nlinsert(char *name, long value, pblock *pb);

/*
 * Finds a long int value from pblock.
 */
NSAPI_PUBLIC long SPpblock_findlong(char *name, pblock *pb);


/*
 * Replace the name of the pblock parameter, retaining the value.
 */
NSAPI_PUBLIC void SPpblock_replace_name(char *oname, char *nname, pblock *pb);


/*
 * Parse a comma-separated list of strings, such as:
 *
 *      "one,two,three"
 *
 * and place them to an array of strings, terminated by a NULL pointer,
 * e.g.
 *      a[0] == "one"
 *      a[1] == "two"
 *      a[2] == "three"
 *      a[3] == NULL
 */
NSAPI_PUBLIC char ** SPstring_array_parse(char *str);


/*
 * Free an array of strings, including the strings.
 */
NSAPI_PUBLIC void SPstring_array_free(char **a);


/*
 * Match a header line's name against the given shell regexp.
 * The header name in the string MUST BE in all-lower-case, except
 * the first character can be either lower or upper-case letter.
 *
 * The header name need's not be there alone, i.e, the colon and
 * the header content may be present.
 */
NSAPI_PUBLIC int SPwildpat_header_match(char *hdr, char *regexp);


/*
 * As strdup(), but also turns all characters into lower-case.
 */
NSAPI_PUBLIC char *SPlowcase_strdup(char *str);


/*
 * Turns hostname in URL to all-lower-case, and removes redundant
 * port numbers, i.e. 80 for http, 70 for gopher and 21 for ftp.
 * Modifies its parameter string directly.
 *
 * Also turns the protocol specifier to all lower-case.
 */
NSAPI_PUBLIC void SPutil_url_fix_hostname(char * url);

/*
 * Returns 0 or 1 depending on whether or not URL references a
 * Fully Qualified Domain Name. Useful, because the proxy doesn't
 * cache non-FQDN's.
 */

NSAPI_PUBLIC int util_url_has_FQDN(char * url);


/*
 * Compare two URLs.
 * Return values as for strcmp().
 */
NSAPI_PUBLIC int SPutil_url_cmp(char *s1, char *s2);


/*
 * Check that the URL looks ok.
 */
NSAPI_PUBLIC int SPutil_uri_check(char *uri);


/*
 * Returns true iff the process with pid exists.
 */
NSAPI_PUBLIC int SPutil_does_process_exist(int pid);

/*
 * Move a file.
 * Preservs permissions, and creation and last access times.
 *
 * First try rename(); if that fails because files are on different
 * filesystems, does it by copying.
 *
 *
 */
NSAPI_PUBLIC int util_move_file(char *src, char *dst);


/*
 * Move a directory.
 * Preservs permissions, and creation and last access times.
 *
 * First try rename(); if that fails because directories are on
 * different filesystems, does it by copying.
 *
 *
 */
NSAPI_PUBLIC int util_move_dir(char *src, char *dst);


/*
 * Looks at a hostname and makes sure that it's no longer than
 * MAX_STERILIZED_HLEN (64) and that it only has alphanumeric and
 * punctuation characters.
 *
 * Truncates longer hostnames, and at the first bad character.
 *
 * Returns 1 if was ok as passed in.
 * Returns 0 if intervention was done (hostname was bad).
 *
 */
NSAPI_PUBLIC int util_sterilize_hostname(char *host);

NSPR_END_EXTERN_C

#ifdef SPAPI20

#define util_make_gmt SPutil_make_gmt
#define util_make_local SPutil_make_local
#define util_get_current_gmt SPutil_get_current_gmt
#define util_make_log_time SPutil_make_log_time
#define util_parse_http_time SPutil_parse_http_time
#define pblock_nlinsert SPpblock_nlinsert
#define pblock_findlong SPpblock_findlong
#define pblock_replace_name SPpblock_replace_name
#define string_array_parse SPstring_array_parse
#define string_array_free SPstring_array_free
#define wildpat_header_match SPwildpat_header_match
#define lowcase_strdup SPlowcase_strdup
#define util_url_fix_hostname SPutil_url_fix_hostname
#define util_url_cmp SPutil_url_cmp
#define util_uri_check SPutil_uri_check
#define util_does_process_exist SPutil_does_process_exist
#define util_move_file SPutil_move_file
#define util_move_dir SPutil_move_dir
#define util_sterilize_hostname SPutil_sterilize_hostname

#endif /* SPAPI20 */

#endif /* !PUBLIC_LIBPROXY_PUTIL_H */

