/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
#ifndef PUBLIC_ONEACL_ACLDEF_H
#define PUBLIC_ONEACL_ACLDEF_H

/*
 * File:        acldef.h
 *
 * Description:
 *
 *      This file contains constant and type definitions for the ACL API.
 */

#ifndef PUBLIC_ONEACL_NSERRDEF_H
#include "nserrdef.h"
#endif /* !PUBLIC_ONEACL_NSERRDEF_H */

NSPR_BEGIN_EXTERN_C

/*
 * Type:        ACLCachable_t
 *
 * Description:
 *
 *      This type is used to specify whether and how long something
 *      may be safely cached.  A value of zero (ACL_NOT_CACHABLE)
 *      indicates that the item is not cachable.  Any other value is
 *      a time, in seconds since 00:00:00 UTC, January 1, 1970, after
 *      which the cached information should be discarded.
 */

typedef unsigned long ACLCachable_t;

#define ACL_NOT_CACHABLE        0
#define ACL_INDEF_CACHABLE      ((unsigned long)(-1))

/*
 * Type:        ACLListHandle_t
 *
 * Description:
 *
 *      This type represents a list of ACLs in their in-memory form.
 */

typedef struct ACLListHandle ACLListHandle_t;
#ifndef INTONEACL
struct ACLListHandle {
};
#endif /* !INTONEACL */

/* The object has been checked for ACLs and has none attached */
#define	ACL_LIST_NO_ACLS	((ACLListHandle_t *)-1)

/*
 * Type:        ACLHandle_t
 *
 * Description:
 *
 *      This type represents the in-memory form of an ACL.
 */
typedef struct ACLHandle ACLHandle_t;
#ifndef INTONEACL
struct ACLHandle {
};
#endif /* !INTONEACL */

/*
 * Type:        ACLExprHandle_t
 *
 * Description:
 *
 *      This type represents a single ACL entry, e.g. allow, deny, etc.
 */
typedef struct ACLExprHandle ACLExprHandle_t;
#ifndef INTONEACL
struct ACLExprHandle {
};
#endif /* !INTONEACL */

/*
 * Type:        ACLEvalHandle_t
 *
 * Description:
 *
 *      This type represents an ACL evaluation context, which includes
 *      an ACL list and property lists for the subject and resource.
 */
typedef struct ACLEvalHandle ACLEvalHandle_t;
#ifndef INTONEACL
struct ACLEvalHandle {
};
#endif /* !INTONEACL */

/*
 * Type:        PFlags_t
 *
 * Description:
 *
 *      This type represents a set of processing flags for an ACL entry.
 */
typedef int PFlags_t;

#define ACL_PFLAG_ABSOLUTE  	0x1
#define ACL_PFLAG_TERMINAL  	0x2
#define ACL_PFLAG_CONTENT  	0x4

#define IS_ABSOLUTE(x)		((x) & ACL_PFLAG_ABSOLUTE)
#define IS_STATIC(x)		((x) & ACL_PFLAG_STATIC)
#define IS_CONTENT(x)		((x) & ACL_PFLAG_CONTENT)

/*
 * Type:        CmpOp_t
 *
 * Description:
 *
 *      This type represents a comparison operator in an ACL attribute
 *      expression.
 */
typedef enum	{
		CMP_OP_EQ,
		CMP_OP_NE,
		CMP_OP_GT,
		CMP_OP_LT,
		CMP_OP_GE,
		CMP_OP_LE
		} CmpOp_t;

/*
 * Type:        ACLExprType_t
 *
 * Description:
 *
 *      This type represents the type of an ACL entry.
 */
typedef enum 	{
		ACL_EXPR_TYPE_ALLOW,
		ACL_EXPR_TYPE_DENY,
		ACL_EXPR_TYPE_AUTH,
		ACL_EXPR_TYPE_RESPONSE
		} ACLExprType_t;

/*
 * Type:        ACLEvalRes_t
 *
 * Description:
 *
 *      This type represents the result of ACL evaluation.
 */
typedef enum	{
		ACL_RES_ALLOW,
		ACL_RES_DENY,
		ACL_RES_FAIL,
		ACL_RES_INVALID,
		ACL_RES_NONE
		} ACLEvalRes_t;

/*
 * Type:        ACLMethod_t
 *
 * Description:
 *
 *      This type represents a reference to an authentication method.
 */
typedef	void * ACLMethod_t;

#define	ACL_METHOD_ANY		((ACLMethod_t)-1)
#define	ACL_METHOD_INVALID	((ACLMethod_t)-2)

/*
 * Type:        ACLDbType_t
 *
 * Description:
 *
 *      This type represents a reference to a type of authentication
 *      database.
 */
typedef	void * ACLDbType_t;

#define	ACL_DBTYPE_ANY		((ACLDbType_t)-1)
#define	ACL_DBTYPE_INVALID	((ACLDbType_t)-2)

/*
 * Type:        ACLAttrGetterFn_t
 *
 * Description:
 *
 *      This type describes a kind of callback function that obtains
 *      a value for an ACL attribute and enters the attribute and value
 *      into the subject property list.
 */
typedef int (*ACLAttrGetterFn_t)(NSErr_t *errp, PList_t subject,
                                 PList_t resource, PList_t auth_info,
                                 PList_t global_auth);

typedef struct ACLAttrGetter ACLAttrGetter_t;
typedef void *ACLAttrGetterList_t;

/*
 * Type:        AclModuleInitFunc
 *
 * Description:
 *
 *      This type describes a kind of callback function that is
 *      specified to ACL_RegisterModule() and called from there.
 */
typedef int (*AclModuleInitFunc)(NSErr_t *errp);

/*
 * Type:        DbParseFn_t
 *
 * Description:
 *
 *      This type describes a kind of callback function that parses
 *      a reference to an authentication database of a particular
 *      database type.  It is called when ACL_RegisterDbName() is
 *      called for a database which is that database type.
 */
typedef int (*DbParseFn_t)(NSErr_t *errp, ACLDbType_t dbtype,
			   const char *name, const char *url,
			   PList_t plist, void **db);

/*
 * Type:        AclCacheFlushFunc_t
 *
 * Description:
 *
 *      This type describes a kind of callback function that is called
 *      when ACL_CacheFlush() is called.
 */
typedef int (*AclCacheFlushFunc_t)(void);

/*
 * Type:        LASEvalFunc_t
 *
 * Description:
 *
 *      This type describes a kind of callback function that is called
 *      to evaluate an attribute value expression in an ACL statement.
 */
typedef int (*LASEvalFunc_t)(NSErr_t *errp, char *attr_name,
                             CmpOp_t comparator, char *attr_pattern,
                             ACLCachable_t *cachable, void **cookie,
                             PList_t subject, PList_t resource,
                             PList_t auth_info, PList_t global_auth);

/*
 * Type:        LASFlushFunc_t
 *
 * Description:
 *
 *      This type describes a kind of callback function that is called
 *      when a previously cached LAS cookie is being flushed from
 *      the ACL cache.
 */
typedef void (*LASFlushFunc_t)(void **cookie);

/*
 * Type:        LDAP
 *
 * Description:
 *
 *      This is an opaque type that represents an open LDAP connection.
 *      It is used mostly via the LDAP SDK API.
 */
typedef struct ldap LDAP;

/*  Reserved property list entry names  */
#define	PROP_SESSION	"session"       /* subject property */
#define INDEX_SESSION	1

#define	PROP_REQUEST	"request"       /* resource property */
#define INDEX_REQUEST	2

#define PROP_IP		"ip"            /* subject property */
#define	INDEX_IP	3

#define PROP_DNS	"dns"           /* subject property */
#define	INDEX_DNS	4

#define PROP_ERROR	"error"
#define	INDEX_ERROR	5

#define PROP_PROGRAMS	"programs"      /* resource property */
#define	INDEX_PROGRAMS	6

#define PROP_WWW_AUTH   "www-auth"
#define INDEX_WWW_AUTH	7

#define INDEX_MAX	8	/* Must be larger than the highest idx */

/*  Flags to ACL_ListFind  */
#define ACL_CASE_INSENSITIVE	 0x1
#define ACL_CASE_SENSITIVE	 0x2

#define	ACL_MAX_TEST_RIGHTS	32
#define	ACL_MAX_GENERIC		32

/*
 *	Command values for the "position" argument to ACL_RegisterGetter
 *	Any positive >0 value is the specific position in the list to insert
 *	the new function.
 */
#define	ACL_AT_FRONT		0
#define	ACL_AT_END		-1
#define	ACL_REPLACE_ALL 	-2
#define	ACL_REPLACE_MATCHING	-3

#define ACL_ATTR_GROUP          "group"
#define ACL_ATTR_RAW_USER_LOGIN "user-login"
#define ACL_ATTR_AUTH_USER	"auth-user"
#define ACL_ATTR_AUTH_TYPE	"auth-type"
#define ACL_ATTR_AUTH_DB	"auth-db"
#define ACL_ATTR_AUTH_PASSWORD  "auth-password"
#define ACL_ATTR_USER	        "user"
#define ACL_ATTR_PASSWORD	"pw"
#define ACL_ATTR_USERDN	        "userdn"
#define ACL_ATTR_RAW_USER	"raw-user"
#define ACL_ATTR_RAW_PASSWORD   "raw-pw"
#define ACL_ATTR_USER_ISMEMBER  "user-ismember"
#define ACL_ATTR_DATABASE	"database"
#define ACL_ATTR_DBTYPE	        "dbtype"
#define ACL_ATTR_DBNAME	        "dbname"
#define ACL_ATTR_DATABASE_URL   "url"
#define ACL_ATTR_METHOD	        "method"
#define ACL_ATTR_AUTHTYPE	"authtype"
#define ACL_ATTR_AUTHORIZATION  "authorization"
#define ACL_ATTR_PARSEFN	"parsefn"
#define ACL_ATTR_ATTRIBUTE	"attr"
#define ACL_ATTR_GETTERFN	"getterfunc"
#define ACL_ATTR_IP		"ip"
#define ACL_ATTR_DNS	        "dns"
#define ACL_ATTR_MODULE	        "module"
#define ACL_ATTR_MODULEFUNC	"func"
#define ACL_ATTR_GROUPS	        "groups"
#define ACL_ATTR_IS_VALID_PASSWORD "isvalid-password"
#define ACL_ATTR_CERT2USER	"cert2user"
#define ACL_ATTR_USER_CERT	"cert"
#define ACL_ATTR_PROMPT	        "prompt"
#define ACL_ATTR_TIME	        "time"
#define ACL_ATTR_USERS_GROUP    "users-group"

#define ACL_DBTYPE_LDAP         "ldap"

#define METHOD_DEFAULT          "default"

/*  Errors must be < 0 */
#define ACL_RES_ERROR      	-1

/* LAS return codes - Must all be negative numbers */
#define	LAS_EVAL_TRUE		-1
#define	LAS_EVAL_FALSE		-2
#define	LAS_EVAL_DECLINE	-3
#define	LAS_EVAL_FAIL		-4
#define	LAS_EVAL_INVALID	-5
#define	LAS_EVAL_NEED_MORE_INFO	-6

NSPR_END_EXTERN_C

#ifndef PUBLIC_ONEACL_ACLAPI_H
#include "aclapi.h"
#endif /* !PUBLIC_ONEACL_ACLAPI_H */

#endif /* !PUBLIC_ONEACL_ACLDEF_H */
