/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * ntlogbuf.h: Handles log buffering
 * 
 * In the beginning this was a NT specific solution (hence ntlogbuf.h).
 * It has now been ported to UX, so all platforms use it.
 *
 * Mike Belshe
 */


#ifndef _NTLOGBUF_H
#define _NTLOGBUF_H

#define LOG_BUFFERING		/* XXXMB move to systems.h */

NSPR_BEGIN_EXTERN_C

#include "base/crit.h"
#include "base/pblock.h"
#include "base/file.h"

#include "prtypes.h"

#define DEFAULT_THREAD_BUFFER_SIZE	(8*1024)
#define DEFAULT_LARGE_BUFFER_SIZE	(DEFAULT_THREAD_BUFFER_SIZE * 8)
#define DEFAULT_NUM_BUFFERS			2

#define DEFAULT_FLUSH_TIMEOUT		30

#define LOGBUF_T	log_file_buf_t

#define LOGBUF_DIRECTIVE(str) \
   ((!strcmp(str, "buffer-size")) || (!strcmp(str, "thread-buffer-size ")) || (!strcmp(str, "num-buffers ")))



/* TLS storage with a linked list of all open buffers for this thread.
 * Used so the thread can properly cleanup.
 */
extern int log_buf_tls;


struct log_file_buf_t;

/* 
 * Each thread will maintain a log_thread_buf_t for each log file it is
 * maintaining.
 */
#define LB_STATE_WRITING	0x10	/* buffer is being written to disk */
#define LB_STATE_IDLE		0x20	/* buffer is idle or being filled */
typedef struct log_thread_buf_t {
	char *data;
	int current_size;
	int max_size;
	int state;						/* only used if we are a log_file_buf_t
									 * buffer.   LB_STATE_WRITING, LB_STATE_IDLE
									 */
	struct log_file_buf_t *log_file_buf;
	struct log_thread_buf_t *tls_next;	/* linked list of this thread's buffers */
	struct log_thread_buf_t *next;		/* global list of all buffers */
} log_thread_buf_t;

/* 
 * Each logfile being maintained will have a log_file_buf_t which is a 
 * set of very large buffers which the log_thread_buf_t feed into.
 */
typedef struct log_file_buf_t {
	SYS_FILE fd;
	int tls_key;		/* tls key each thread should use to find
						 * its own log_thread_buf_t
						 */

	unsigned int thread_buf_size;	/* threads should make their buffers
									 * this size 
									 */

	CRITICAL crit;		/* list, data, current_size, max_size, and
						 * current_buffer are protected by crit.
						 */

	int num_buffers;
	log_thread_buf_t **buf;
    int current_buffer;

	CRITICAL cv_crit;
    CONDVAR cv;				/* If a thread attempts to write to the
    						 * large buffers but the buffers are busy
							 * being written, the thread will block
							 * on cv.
				 			 */
	log_thread_buf_t *list;	/* Used to flush all buffers on log closure */
	struct log_file_buf_t *next;	/* linked list of all log files */
} log_file_buf_t;

/* Initialize everything...
 */
int					logbuf_init(pblock *pb);

/* Initialize logbuf thread */
PR_EXTERN(PRBool) logbufthread_init(void);

/* Create and destroy the per-logfile data structures */
LOGBUF_T*			logbuf_new(pblock *pb, SYS_FILE fd);
void				logbuf_destroy(LOGBUF_T *buf);

/* Create and destroy the per-thread data structures.
 * These will be called when a thread calls into log_buf_write()
 * the first time.
 */
log_thread_buf_t *	_log_thread_buf_create(LOGBUF_T *buf, int size);
void				_log_thread_buf_destroy(log_thread_buf_t *buf);


/* log_buf_write()
 * Called by a thread to do a log write; write may go to the thread's
 * TLS buffer, or it might get flushed to the main log buffers.
 *
 */
int 				logbuf_write(LOGBUF_T *lb, SYS_FILE fd, char *buf, int len);

/* log_buf_flush()
 * Called to flush a specific thread buffer to the main buffer
 */
void				logbuf_flush(log_thread_buf_t *lb);

/* log_buf_flush_all()
 * Called to flush all outstanding thread buffers to the main buffers.
 */
void				logbuf_flush_all();

/* log_buf_cleanup_thread()
 * Called to cleanup and destroy a thread's buffers.
 */
void				logbuf_cleanup_thread();

/* log_buf_cleanup_all()
 * Flushes and destroys all outstanding TLS buffers for all log files.
 */
void				logbuf_cleanup_all();

/* log_buf_close()
 * Flushes last buffer for a file and deletes the logbuf
 */
void				logbuf_close(LOGBUF_T *buf);

/* log_buf_isempty()
 * Checks if the calling thread's tls buffer is empty.  Returns 1 if empty,
 * 0 otherwise.
 */
int				logbuf_isempty(LOGBUF_T *buf);

/* log_buf_allempty()
 * Checks if all outstanding buffers for this thread are empty.  Returns
 * 1 if all empty, 0 otherwise.
 */
int				logbuf_allempty();


NSPR_END_EXTERN_C

#endif  /* LOGBUF_H */
