/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

#include "dsgw.h"
#include "dbtdsgw.h"
#include "../lib/libsi18n/gsadmserv.h"

#ifdef DSGW_DEBUG
#include <time.h>
#include <stdarg.h>
#endif /* DSGW_DEBUG */

static char **vpmap = NULL;

extern char *Versionstr;	/* from Versiongw.c */

char		*progname;		/* set by dsgw_init() */
dsgwconfig	*gc;			/* set by dsgw_init() */
int		http_hdr_sent = 0;	/* non-zero if header has been sent */
char		**header_lines = NULL;	/* null-terminated array of hdr lines */
char *dsgw_html_body_colors = "";	/* reset by dsgw_init() */


/*
 * dsgw_init -- initialize a dsgw CGI program:
 *	set "progname" global based on "progpath" (normally argv[0])
 *	check that REQUEST_METHOD is in "methods_handled" mask
 *	if request method is "POST", read HTML form variables from stdin
 *
 * If an fatal error occurs, -1 is returned.
 * If all goes well, returns either DSGW_METHOD_GET or DSGW_METHOD_POST
 */
int
dsgw_init( int argc, char **argv, int methods_handled )
{
    char	*m, *s;
    int		method;
    int		c, err;

    (void)ADM_Init();

    /* initialize the string database */
    XP_InitStringDatabase("../../../admin",DATABASE_NAME);

    if (( progname = strchr( argv[0], '/' )) == NULL ) {
	progname = dsgw_ch_strdup( argv[0] );
#ifdef _WIN32
	if (( s = strrchr( progname, '.' )) != NULL
		&& strcasecmp( s, ".EXE" ) == 0 ) {
	    *s = '\0';
	}
#endif /* _WIN32 */
    } else {
	++progname;
    }

    while (( c = getopt( argc, argv, "v" )) != EOF ) {
	if ( c == 'v' ) {
	    printf( "%s\n", Versionstr );
	}
	exit( 0 );
    }

#ifdef DSGW_DEBUG
    dsgw_log( "%s started\n", Versionstr );
#endif

    err = method = 0;
    gc = dsgw_read_config();

    /* set languages for string database */
    SetLanguage(CLIENT_LANGUAGE,gc->gc_ClientLanguage);
    SetLanguage(ADMIN_LANGUAGE,gc->gc_AdminLanguage);
    SetLanguage(DEFAULT_LANGUAGE,gc->gc_DefaultLanguage);

    if (( m = getenv( "REQUEST_METHOD" )) != NULL ) {
	if ( strcasecmp( m, "GET" ) == 0 || strcasecmp( m, "HEAD" ) == 0 ) {
	    method = DSGW_METHOD_GET;
	} else if ( strcasecmp( m, "POST" ) == 0 ) {
	    method = DSGW_METHOD_POST;
	    if (( err = dsgw_post_begin( stdin )) == 0 ) {
		if (( s = dsgw_get_cgi_var( "ldapport",
			DSGW_CGIVAR_OPTIONAL )) != NULL ) {
		    gc->gc_ldapport = atoi( s );
		    free( s );
		}
		if (( s = dsgw_get_cgi_var( "ldapserver",
			DSGW_CGIVAR_OPTIONAL )) != NULL ) {
		    gc->gc_ldapserver = s;
		}
	    }
	}
    }

    if ( method == 0 || ( methods_handled & method ) == 0 ) {
	err = DSGW_ERR_BADMETHOD;
    }

    if (( s = getenv( "HTTPS" )) == NULL || strcasecmp( s, "on" ) == 0 ||
	    ( s = getenv( "HTTPS_KEYSIZE" )) == NULL ) {
	gc->gc_httpskeysize = 0;
    } else {
	gc->gc_httpskeysize = atoi( s );
    }

    /* set default color scheme */
    if ( method == DSGW_METHOD_POST && ( s = dsgw_get_cgi_var( "colors",
	    DSGW_CGIVAR_OPTIONAL )) != NULL ) {
	dsgw_html_body_colors = s;
    } else {			/* use same color scheme as libadmin */
	dsgw_html_body_colors = "BGCOLOR=\"#C0C0C0\" LINK=\"#0000EE\" "
		"VLINK=\"#551A8B\" ALINK=\"#FF0000\"";
    }

    if ( err != 0 ) {
	dsgw_error( err, NULL, DSGW_ERROPT_EXIT, 0 );
    }

    return( method );
}


/*
 * function called back by dsgw_parse_line() to evaluate IF directives.
 * return non-zero for true, zero for false.
 */
int
dsgw_simple_cond_is_true( int argc, char **argv, void *arg /* UNUSED */ )
{
    if ( strcasecmp( argv[0], DSGW_COND_ADMSERV ) == 0 ) {
	return( gc->gc_admserv );
    }
 
    if ( strcasecmp( argv[0], DSGW_COND_LOCALDB ) == 0 ) {
	return( gc->gc_localdbconf != NULL );
    }

    if ( strcasecmp( argv[0], DSGW_COND_POSTEDFORMVALUE ) == 0 ) {
	/*
	 * format of IF statment is:
	 *    <!-- IF "PostedFormValue" "VARNAME" "VALUE" -->
	 * where VARNAME is the name of a POSTed CGI variable to look for and
	 * VALUE is an optional value to test it against.  If VALUE is omitted,
	 * the test is just for the presence of a variable named VARNAME.
	 */
	char	*postedvalue;

	if ( argc < 2 || ( postedvalue = dsgw_get_cgi_var( argv[1],
		DSGW_CGIVAR_OPTIONAL )) == NULL ) {
	    return( 0 );	/* VARNAME is missing or not posted */
	} else if ( argc < 3 ) {
	    return( 1 );	/* VALUE is missing, so return true */
	} else {
	    return( strcasecmp( postedvalue, argv[ 2 ] ) == 0 );
	}
    } 

    if ( strcasecmp( argv[0], DSGW_COND_LANGUAGETYPE ) == 0 ) {
	int	type;
	/*
	 * format of IF statment is:
	 *    <!-- IF "LanguageType" "TYPENUM" -->
	 * where TYPENUM matches the dslang_type field of the current input
	 * language.  TYPENUM of 0 can be used to test for no input language.
	 */
	if ( argc < 2 ) {
	    type = 0;
	} else {
	    type = atoi( argv[1] );
	}
	if ( gc == NULL || gc->gc_inputlanguage == NULL ) {
	    return( type == 0 );
	}
	return( type == gc->gc_inputlanguage->dslang_type );
    }

    if ( strcasecmp( argv[0], DSGW_COND_LDAPVERSION ) == 0 ) {
	int	ver;
	char	*op;
	/*
	 * format of IF statment is:
	 *    <!-- IF "LdapVersion" "<|<=|==|=|>=|>" "NUM" -->
	 * where NUM is the version number to compare with.
	 * NUM of 0 can be used to test for no version information.
	 */
	if ( argc < 3 ) {
	    return 0;
	} else {
	    op = argv[1];
	    ver = atoi( argv[2] );
	}
	if ( gc == NULL || gc->gc_ldapversion == 0 ) {
	    return( ver == 0 );
	}
	if ( *op == '<' ) {
	    if ( *++op == '=' )
		return( gc->gc_ldapversion <= ver );
	    else
		return( gc->gc_ldapversion < ver );
	} else if ( *op == '=' ) {
		return( gc->gc_ldapversion == ver );
	} else if ( *op == '>' ) {
	    if ( *++op == '=' )
		return( gc->gc_ldapversion >= ver );
	    else
		return( gc->gc_ldapversion > ver );
	} else {
	    return 0;
	}
    }

    return( 0 );
}


/*
 * return a pointer to a malloc'd string containing the path to
 * config. file "filename", based on the DSGW_CONFIGDIR define.
 * If "filename" contains "..", or "//" this is treated as a fatal
 * error.  If "prefix" is not NULL, it is pre-pended to "filename"
 */
char *
dsgw_file2path( char *prefix, char *filename )
{
    char	*path, *pattern, *languages;
    int		len;

    if ( strstr( filename, "//" ) != NULL ||
	    strstr( filename, ".." ) != NULL ) {
	dsgw_error( DSGW_ERR_BADFILEPATH, filename, DSGW_ERROPT_EXIT, 0 );
    }

    if ( prefix == NULL ) {
	prefix = "";
    }

    /* allocate buffers with enough extra room to fit "$$LANGDIR/" */
    len =  strlen( prefix ) + strlen( filename ) + 11;
    path = dsgw_ch_malloc( len );
    pattern = dsgw_ch_malloc( len );

    /* call GetFileForLanguage() to do its I18n magic */
    sprintf( pattern, "%s$$LANGDIR/%s", prefix, filename );
    if ( GetFileForLanguage( pattern, gc->gc_ClientLanguage, path ) < 0 ) {
	sprintf( path, "%s%s", prefix, filename );	/* fallback */
    }
    free( pattern );

    return( path );
}



/*
 * return a pointer to a malloc'd string containing the path to
 * config. file "filename", based on the DSGW_HTMLDIR define.
 * If "filename" contains "..", or "//" this is treated as a fatal
 * error.  If "prefix" is not NULL, it is pre-pended to "filename"
 */
char *
dsgw_file2htmlpath( char *prefix, char *filename )
{
    char	*path, *pattern, *languages;
    int		len;

    if ( strstr( filename, "//" ) != NULL ||
	    strstr( filename, ".." ) != NULL ) {
	dsgw_error( DSGW_ERR_BADFILEPATH, filename, DSGW_ERROPT_EXIT, 0 );
    }

    if ( prefix == NULL ) {
	prefix = "";
    }

    /* allocate buffers with enough extra room to fit "$$LANGDIR/" */
    len = strlen( DSGW_HTMLDIR ) + strlen( prefix ) + strlen( filename ) + 11;
    path = dsgw_ch_malloc( len );
    pattern = dsgw_ch_malloc( len );

    /* call GetFileForLanguage() to do its I18n magic */
    sprintf( pattern, "%s%s$$LANGDIR/%s", DSGW_HTMLDIR, prefix, filename );
    if ( GetFileForLanguage( pattern, gc->gc_ClientLanguage, path ) < 0 ) {
	sprintf( path, "%s%s", prefix, filename );	/* fallback */
    }
    free( pattern );

    return( path );
}


/*
 * malloc that checks for NULL return value and exits upon failure
 */
void *
dsgw_ch_malloc( size_t n )
{
    void *p;

    if (( p = malloc( n )) == NULL ) {
	dsgw_error( DSGW_ERR_NOMEMORY, NULL, DSGW_ERROPT_EXIT, 0 );
    }

    return( p );
}


/*
 * realloc that checks for NULL return value and exits upon failure
 * we also handle p == NULL by doing a malloc
 */
void *
dsgw_ch_realloc( void *p, size_t n )
{
    if ( p == NULL ) {
	p = malloc( n );
    } else {
	p = realloc( p, n );
    }

    if ( p == NULL ) {
	dsgw_error( DSGW_ERR_NOMEMORY, NULL, DSGW_ERROPT_EXIT, 0 );
    }

    return( p );
}


/*
 * strdup that checks for NULL return value and exits upon failure
 */
char *
dsgw_ch_strdup( char *s )
{
    int		len;
    char	*p;

    len = strlen( s ) + 1;
    p = dsgw_ch_malloc( len );
    memcpy( p, s, len );
    return( p );
}



/*
 * Escape any single- or double-quotes with a '\'.  Used when generating
 * JavaScript code.  Returns a malloc'd string which the caller is
 * responsible for freeing.
 */
char *
dsgw_escape_quotes( char *in )
{
    char *out;
    char *p, *t;
    int nq = 0;


    if ( in == NULL ) {
	return NULL;
    }
    /* count number of quotes */
    for ( p = in; *p != '\0'; p++ ) {
	if ( *p == '\'' || *p == '"' ) {
	    nq++;
	}
    }
    out = dsgw_ch_malloc(( p - in ) + nq + 1 );
    for ( p = in, t = out; *p != '\0'; p++ ) {
	if ( *p == '\'' || *p == '"' ) {
	    *t++ = '\\';
	}
	*t++ = *p;
    }
    *t = '\0';
    return out;
}


void
dsgw_escape_quotes_output( FILE *fp, char *in )
{
    char	*s;

    if (( s = dsgw_escape_quotes( in )) != NULL ) {
	fputs( s, fp );
	free( s );
    }
}
    

/*
 * Send the headers we've accumulated.
 */
void
dsgw_send_header()
{
    int i;

    if ( http_hdr_sent ) {
	return;
    }

    if ( header_lines == NULL ) {
	fputs( "Content-type: text/html", stdout );
	if ( gc != NULL && gc->gc_charset != NULL && *gc->gc_charset != '\0' ) {
	    fprintf( stdout, ";charset=%s", gc->gc_charset );
#ifdef DSGW_DEBUG
            dsgw_log( "send_header: charset=%s\n", gc->gc_charset );
#endif
	}
	fputs( "\n\n", stdout );
	http_hdr_sent = 1;
	return;
    }

    for ( i = 0; header_lines[ i ] != NULL; i++ ) {
	fputs( header_lines[ i ], stdout );
	fputs( "\n", stdout );
#ifdef DSGW_DEBUG
        dsgw_log( "send_header: %s\n", header_lines[ i ] );
#endif
    }
    fputs( "\n", stdout );

    http_hdr_sent = 1;
}
	

/*
 * Add a line to the array of header lines.
 */
void
dsgw_add_header( char *line )
{
    int i;
    char *ct_prefix = "Content-type: text/html";
    char *cs_prefix = ";charset=";

    if ( header_lines == NULL ) {
	header_lines = ( char ** ) dsgw_ch_malloc( 2 * sizeof( char * ));
	if ( gc != NULL && gc->gc_charset != NULL && *gc->gc_charset != '\0' ) {
	    header_lines[ 0 ] = dsgw_ch_malloc( strlen( ct_prefix ) +
		    strlen( cs_prefix ) + strlen( gc->gc_charset ) + 1 );
	    sprintf( header_lines[ 0 ], "%s%s%s", ct_prefix, cs_prefix,
		    gc->gc_charset );
	} else {
	    header_lines[ 0 ] = dsgw_ch_strdup( ct_prefix );
	}
	header_lines[ 1 ] = NULL;
    }
    for ( i = 0; header_lines[ i ] != NULL; i++ );
    header_lines = (char **) dsgw_ch_realloc( header_lines,
	   ( i + 2 ) * sizeof( char * ));
    header_lines[ i ] = dsgw_ch_strdup( line );
    header_lines[ i + 1 ] = NULL;
}


/*
 * Check the environment for an authentication cookie.  Returns the
 * entire auth cookie if present, or returns NULL if no such cookie
 * exists.  The returned string must be freed by the caller.
 */
char *
dsgw_get_auth_cookie()
{
    char *p, *e, *ckhdr;
 
    ckhdr = getenv( "HTTP_COOKIE" );

    if ( ckhdr == NULL ) {
        return NULL;
    } else {
	ckhdr = strdup( ckhdr );
    }
 
    if (( p = strstr( ckhdr, DSGW_AUTHCKNAME )) == NULL ) {
	free( ckhdr );
        return NULL;
    }
 
    if (( e = strchr( p, ';' )) != NULL ) {
        *e = '\0';
    }
 
    p = strdup( p );
    free( ckhdr );
    return p;
}



/* 
 * Break a cookie into its random string and DN parts.  The DN is returned
 * unescaped.  The caller is responsible for freeing the returned DN
 * and random string.  Returns 0 on success, -1 on error.  If the
 * cookie has the value "[unauthenticated]", then 0 is returned and
 * dn is set to NULL;
 */
int
dsgw_parse_cookie( char *cookie, char **rndstr, char **dn )
{
    char 	*p,  *r;
    int		rlen;

    if ( cookie == NULL ) {
	*rndstr = *dn = NULL;
	return -1;
    }

    /* Make sure cookie starts with "nsdsgwauth" */
    if ( strncmp( cookie, DSGW_AUTHCKNAME, strlen( DSGW_AUTHCKNAME ))) {
	/* Cookie didn't start with "nsdsgwauth" */
	*rndstr = *dn = NULL;
	return -1;
    }
    
    r = cookie + strlen( DSGW_AUTHCKNAME );
    if ( *r == '=' ) {
	r++;
    }

    /* Is cookie value "[unauthenticated]" ? */
    if ( !strncmp( r, DSGW_UNAUTHSTR, strlen( DSGW_UNAUTHSTR ))) {
	*rndstr = strdup( DSGW_UNAUTHSTR );
	*dn = NULL;
	return 0;
    }

    /* find start of DN */
    if (( p = strrchr( cookie, ':' )) == NULL ) {
	*rndstr = *dn = NULL;
	return -1;
    }
    
    rlen = p - r + 1;
    *(rndstr) = dsgw_ch_malloc( rlen );
    *(rndstr)[ 0 ] = '\0';
    strncat( *rndstr, r, rlen );
    (*rndstr)[ rlen - 1 ] = '\0'; 

    p++;
    *dn = strdup( p );
    dsgw_form_unescape( *dn );

    return 0;
}
    
/*
 * Generate a "go home" button with a link to the main entry point for
 * the gateway.  The caller is responsible for any surrounding
 * HTML, e.g. <FORM> and <TABLE> tags.
 */
void
dsgw_emit_homebutton()
{
    printf( "<INPUT TYPE=\"button\" VALUE=\"%s\" "
	"onClick=\"top.location.href='%s'\">", XP_GetClientStr(DBT_returnToMain_), DSGW_URLPREFIX_MAIN );
}
   

/*
 * Same as dsgw_emit_homebutton(), but all single quotes are escaped
 * so the text can be included inside a JavaScript definition.
 * This is really gross.  I'm sorry.
 */
void
dsgw_emit_homebutton_qe()
{
    printf( "<INPUT TYPE=\"button\" VALUE=\"%s\" "
	"onClick=\"top.location.href=\\'%s\\'\">", XP_GetClientStr(DBT_returnToMain_1),DSGW_URLPREFIX_MAIN );
}
   

/*
 * Generate a help button with a link to the tutor program for
 * the given help topic.   The caller is responsible for any surrounding
 * HTML, e.g. <FORM> and <TABLE> tags.
 */
void
dsgw_emit_helpbutton( char *topic )
{
    if ( topic == NULL ) {
	return;
    }

    if ( gc->gc_admserv ) {
	char	*jscript;

	if (( jscript = helpJavaScriptForTopic( topic )) == NULL ) {
	    return;
	}

	printf( "<INPUT TYPE=\"button\" VALUE=\"%s\" onClick=\"%s\">",
		XP_GetClientStr(DBT_help_), jscript );
    } else {
	printf( "<INPUT TYPE=\"button\" VALUE=\"%s\" "
	    "onClick=\"var hwin=window.open('%s?%s', 'infowin_dsgw', "
	    "'resizable=1,width=400,height=500')\">\n",
	    XP_GetClientStr(DBT_help_1), 
	    dsgw_getvp( DSGW_CGINUM_TUTOR ), topic );
    }
}


/*
 * Same as dsgw_emit_helpbutton(), but all single quotes are escaped
 * so the text can be included inside a JavaScript definition.
 */
void
dsgw_emit_helpbutton_qe( char *topic )
{
    if ( topic == NULL ) {
	return;
    }
    if ( gc->gc_admserv ) {
	printf( "<INPUT TYPE=\"button\" VALUE=\"%s\" onClick=\""
		"alert(\\'%s\\');\">",
		XP_GetClientStr(DBT_help_2), XP_GetClientStr(DBT_helpIsNotYetAvailable_) );
    } else {
	printf( "<INPUT TYPE=\"button\" VALUE=\"%s\" "
	    "onClick=\"var hwin=window.open(\\'%s?%s\\', \\'infowin_dsgw\\', "
	    "\\'resizable=1,width=400,height=500\\')\">",
	    XP_GetClientStr(DBT_help_3), dsgw_getvp( DSGW_CGINUM_TUTOR ), topic );
    }
}


/*
 * I18n: display language selection popup.
 */
void
dsgw_emit_language_popup( dsgwtmplinfo *tip, int argc, char **argv )
{
    char	line[BIG_LINE];
    char	*varname, *prefix, *suffix, *onchange;
    char	*selected;
    int 	sflag;
    dsgwlang	*lp;
    char	**vals = NULL;

    if (( varname = get_arg_by_name( "name", argc, argv )) == NULL ) {
	varname = "base";
    }
    if (( prefix = get_arg_by_name( "prefix", argc, argv )) == NULL ) {
	prefix = "";
    }
    if (( suffix = get_arg_by_name( "suffix", argc, argv )) == NULL ) {
	suffix = "";
    }
    if (( onchange = get_arg_by_name( "onchange", argc, argv )) == NULL ) {
	onchange = "";
    }

    /*
     * Default selection order:
     *   Choose this one.
     *   1. inputlanguage is available (user input)
     *   2. the user has the preferredLanguage
     *   3. default selection from the dsgw.conf
     *   To change the preferred language by selecting the lang pop-up,
     *   inputlanguage should be stronger than preferred language.
     */
    selected = gc->gc_defaultselectedlang;
    if ( gc->gc_inputlanguage ) {	/* user selection */
	selected = gc->gc_inputlanguage->dslang_tag;
    } else if ( tip && tip->dsti_entry ) {
        /* get language info from the person's preferredLanguage, if given */
	vals = ldap_get_values( tip->dsti_ld, tip->dsti_entry,
					"preferredlanguage" );
        if ( vals == NULL ) {	/* no preferred language -- group/ou */
          BerElement *p;
          char *ret, *q;
#define LANGDASH "lang-"
          selected = NULL;
          for ( ret = ldap_first_attribute( tip->dsti_ld, tip->dsti_entry, &p );
		ret != NULL;
                ret = ldap_next_attribute( tip->dsti_ld, tip->dsti_entry, p )) {
	    q = strstr( ret, LANGDASH );
	    if ( q == NULL ) {
	      ldap_memfree( ret );
	      continue;
	    }
	    q += 5;

	    if ( !strchr( q, ';' ) && strcasecmp( q, "en" )) {
	      /* not english && not phonetic */
	      selected = dsgw_ch_strdup( q );	
	      ldap_memfree( ret );
	      break;
	    }
	    ldap_memfree( ret );
          }
	  if ( !selected )
	      selected = DSGW_DEFAULT_SELECTED_LANG;
        } else			/* has preferred language */
	  selected = *vals;
	gc->gc_inputlanguage = dsgw_tag2lang( gc->gc_languages, selected );
    }
    lp = gc->gc_languages;
    if ( lp == NULL ) {	/* No language info */
	util_snprintf( line, BIG_LINE,
		"<INPUT TYPE=\"hidden\" NAME=\"%s\" VALUE=\">", varname );
        fputs( line, stdout );
    } else {
        util_snprintf( line, BIG_LINE, "%s\n<SELECT NAME=\"%s\" ONCHANGE=\"%s\">\n",
    		prefix, varname, onchange );
        fputs( line, stdout );
    
        sflag = 0;
        for ( ; lp != NULL; lp = lp->dslang_next ) {
            fputs( "<OPTION ", stdout );
    	    if ( !sflag && !strcasecmp( lp->dslang_tag, selected )) {
    	        fputs( "SELECTED ", stdout );
		sflag = 1;
	    }
    	    util_snprintf( line, BIG_LINE, "VALUE=\"%s\">%s\n",
    			     lp->dslang_tag, lp->dslang_name );
    	    fputs( line, stdout );
        }
    
        util_snprintf( line, BIG_LINE, "</SELECT>\n%s\n", suffix );
        fputs( line, stdout );
    }

    if ( vals )
        ldap_value_free( vals );
    return;
}

/*
 * I18n: display language map table (b/w tag and type).
 */
void
dsgw_emit_language_tables( int argc, char **argv )
{
    char	line[BIG_LINE];
    char	*varname, *prefix, *suffix;
    dsgwlang	*lp;

    if (( varname = get_arg_by_name( "name", argc, argv )) == NULL ) {
	varname = "base";
    }
    if (( prefix = get_arg_by_name( "prefix", argc, argv )) == NULL ) {
	prefix = "";
    }
    if (( suffix = get_arg_by_name( "suffix", argc, argv )) == NULL ) {
	suffix = "";
    }

    lp = gc->gc_languages;
    if ( lp != NULL ) {
	util_snprintf( line, BIG_LINE, "%s\n<FORM NAME=\"%s\">\n",
			     prefix, varname );
        fputs( line, stdout );

        for ( ; lp != NULL; lp = lp->dslang_next ) {
            fputs( "<INPUT TYPE=HIDDEN ", stdout );
            util_snprintf( line, BIG_LINE, "NAME=\"%s\" VALUE=\"%d\">\n",
            		     lp->dslang_tag, lp->dslang_type );
            fputs( line, stdout );
        }
	util_snprintf( line, BIG_LINE, "</FORM>\n%s\n", suffix );
	fputs( line, stdout );
    }

    return;
}

/* cleanup the language selection popup from dsgw.conf */
void
dsgw_free_language_popup()
{
    dsgwlang *lp, *nextlp;

    lp = gc->gc_languages;
    while ( lp != NULL ) {
        nextlp = lp->dslang_next;
        free( lp );
        lp = nextlp;
    }
    gc->gc_languages = NULL;

    return;
}

/* prepare a default selection "Default" for a language selection popup */
void
dsgw_setup_default_laguage_popup()
{
    dsgwlang *lp;

    lp = (dsgwlang *)dsgw_ch_malloc( sizeof( dsgwlang ));
    memset( lp, 0, sizeof( dsgwlang ));
    lp->dslang_tag = dsgw_ch_strdup ( "en" );
    lp->dslang_name = dsgw_ch_strdup ( "Default" );
    lp->dslang_type = 1;

    if ( gc->gc_languages )
	dsgw_free_language_popup();

    gc->gc_languages = lp;

    return;
}


/*
 * Return malloc'd URL prefix that consists of:
 *	prefix + '/' + HOST:PORT
 */
char *
dsgw_build_urlprefix( char *prefix )
{
    char	*p, *urlprefix;

    p = ( gc->gc_ldapserver == NULL ? "" : gc->gc_ldapserver );
    urlprefix = dsgw_ch_malloc( 8 /* room for "/:port#" + zero-term. */
            + strlen( prefix ) + strlen( p ));
    sprintf( urlprefix, "%s/%s", prefix, p );
    if ( gc->gc_ldapport != 0 && gc->gc_ldapport != LDAP_PORT ) {
        sprintf( urlprefix + strlen( urlprefix ), ":%d", gc->gc_ldapport );
    }

    return( urlprefix );
}


void
dsgw_addtemplate( dsgwtmpl **tlpp, char *template, int count, char **ocvals )
{
    int		i;
    dsgwtmpl	*prevtp, *tp;

    tp = (dsgwtmpl *)dsgw_ch_malloc( sizeof( dsgwtmpl ));
    memset( tp, 0, sizeof( dsgwtmpl ));
    tp->dstmpl_name = dsgw_ch_strdup( template );

    /* each argument is one objectClass */
    tp->dstmpl_ocvals = dsgw_ch_malloc(( count + 1 ) * sizeof( char * ));
    for ( i = 0; i < count; ++i ) {
	tp->dstmpl_ocvals[ i ] = dsgw_ch_strdup( ocvals[ i ] );
    }
    tp->dstmpl_ocvals[ count ] = NULL;

    if ( *tlpp == NULL ) {
	*tlpp = tp;
    } else {
	for ( prevtp = *tlpp; prevtp->dstmpl_next != NULL;
		prevtp = prevtp->dstmpl_next ) {
	    ;
	}
	prevtp->dstmpl_next = tp;
    } 
}


dsgwtmpl *
dsgw_oc2template( char **ocvals )
{
    int			i, j, needcnt, matchcnt;
    dsgwtmpl		*tp;

    for ( tp = gc->gc_templates; tp != NULL; tp = tp->dstmpl_next ) {
	needcnt = matchcnt = 0;
	for ( i = 0; tp->dstmpl_ocvals[ i ] != NULL; ++i ) {
	    for ( j = 0; ocvals[ j ] != NULL; ++j ) {
		if ( strcasecmp( ocvals[ j ], tp->dstmpl_ocvals[ i ] ) == 0 ) {
		    ++matchcnt;
		}
	    }
	    ++needcnt;
	}

	if ( matchcnt == needcnt ) {
	    return( tp );
	}
    }

    return( NULL );
}


/*
 * Given a language tag, find and return a language structure.  If we
 * don't know about this language we will return NULL.
 */
dsgwlang *
dsgw_tag2lang( dsgwlang *languages, char *tag )
{
    for ( ; languages != NULL; languages = languages->dslang_next ) {
	if ( strcasecmp( tag, languages->dslang_tag ) == 0 ) {
	    return( languages );
	}
    }

    return( NULL );
}


void
dsgw_init_searchprefs( struct ldap_searchobj **solistp )
{
    char	*path;

    path = dsgw_file2path( gc->gc_configdir, DSGW_SEARCHPREFSFILE );
    if ( ldap_init_searchprefs( path, solistp ) != 0 ) {
	dsgw_error( DSGW_ERR_BADCONFIG, path, DSGW_ERROPT_EXIT, 0 );
    }
    free( path );
}


void
dsgw_remove_leading_and_trailing_spaces( char **sp )
{
    char	*s, *p;

    if ( sp == NULL || *sp == NULL ) {
        return;
    }

    s = *sp;

    /* skip past any leading spaces */
    while ( XP_IS_SPACE( *s )) {
	++s;
    }

    /* truncate to remove any trailing spaces */
    if ( *s != '\0' ) {
	p = s + strlen( s ) - 1;
	while ( p >= s && XP_IS_SPACE( *p )) {
	    --p;
	}
	*++p = '\0';
    }

    *sp = s;
}


/*
 * Return the virtual path prefix for the CGI program specified by
 * cginum.
 */
char *
dsgw_getvp( int cginum )
{
    char	*cginame;
    char	*surl;
    char	*extpath;
    int		i;

    if ( cginum < 1 || cginum > DSGW_MODE_NUMMODES ) {
	return "";
    }
    if ( vpmap == NULL ) {
	/* note: slot zero of vpmap isn't used */
	vpmap = dsgw_ch_malloc(( DSGW_MODE_NUMMODES + 1 ) * sizeof( char * ));
	for ( i = 0; i <= DSGW_MODE_NUMMODES; i++ ) {
	    vpmap[ i ] = NULL;
	}
    }

    if ( vpmap[ cginum ] == NULL ) {
	switch ( cginum ) {
	case DSGW_CGINUM_DOSEARCH:
	    cginame = DSGW_CGINAME_DOSEARCH;
	    break;
	case DSGW_CGINUM_BROWSE:
	    cginame = DSGW_CGINAME_BROWSE;
	    break;
	case DSGW_CGINUM_SEARCH:
	    cginame = DSGW_CGINAME_SEARCH;
	    break;
	case DSGW_CGINUM_CSEARCH:
	    cginame = DSGW_CGINAME_CSEARCH;
	    break;
	case DSGW_CGINUM_AUTH:
	    cginame = DSGW_CGINAME_AUTH;
	    break;
	case DSGW_CGINUM_EDIT:
	    cginame = DSGW_CGINAME_EDIT;
	    break;
	case DSGW_CGINUM_DOMODIFY:
	    cginame = DSGW_CGINAME_DOMODIFY;
	    break;
	case DSGW_CGINUM_DNEDIT:
	    cginame = DSGW_CGINAME_DNEDIT;
	    break;
	case DSGW_CGINUM_RENAME:
	    cginame = DSGW_CGINAME_RENAME;
	    break;
	case DSGW_CGINUM_AVAEDIT:
	    cginame = DSGW_CGINAME_AVAEDIT;
	    break;
	default:
	    return "";
	}
	if (( surl = getenv( "SERVER_URL" )) == NULL ) {
	    surl = ""; 	/* XXX probably should abort */
	}
	if ( gc->gc_admserv ) {
	    char	*server_names;
	    /*
	     * include "/admin-serv/" or "/user-environment/" if appropriate
	     */
	    if ( gc->gc_enduser ) {
		extpath = DSGW_USER_ADM_BINDIR;
	    } else {
		extpath = DSGW_ADMSERV_BINDIR;
	    }
	} else {
	    extpath = "";
	}
	vpmap[ cginum ] = dsgw_ch_malloc( strlen( gc->gc_urlpfxcgi ) +
		strlen( surl ) + strlen( extpath ) + strlen( cginame ) + 2 );
	sprintf( vpmap[ cginum ], "%s/%s%s%s", surl, extpath,
		gc->gc_urlpfxcgi, cginame );
    }
    return( vpmap[ cginum ]);
}



char * 
dsgw_escape_for_shell( char *s ) 
{ 
    char        *escaped; 
    char        tmpbuf[4]; 
    size_t x,l; 
  
    if ( s == NULL ) { 
        return( s ); 
    } 
  
    l = 3 * strlen( s ) + 1; 
    escaped = dsgw_ch_malloc( l ); 
#ifdef _WIN32 
    memset( escaped, 0, l ); 
    for ( x = 0; s[x]; x++ ) { 
        if (( (unsigned char)s[x] & 0x80 ) == 0 ) { 
           strncat( escaped, &s[x], 1 ); 
        } else { 
           /* not an ASCII character - escape it */ 
           sprintf( tmpbuf, "\\%x", (unsigned)(((unsigned char)(s[x])) & 0xff) ); 
           strcat( escaped, tmpbuf ); 
        } 
    } 
#else 
    strcpy( escaped, s ); 
    escape_for_shell( escaped );        /* in libadmin */ 
#endif 
    return( escaped ); 
}


/*
 * if "path" does not look like a fully qualified path, make it so.
 * on WIN32, we also convert '/' characters in "cwd" to '\' characters.
 * return a malloc'd string.
 */
char *
dsgw_fully_qualified_path( char *path, char *cwd )
{
    char	*p, *q;

    if ( path == NULL ) {
	path = "";
    }

    if ( path[ 0 ] != '/'
#ifdef _WIN32
	&& path[ 0 ] != '\\' && ( !isalpha( path[ 0 ] ) || path[ 1 ] != ':' )
#endif
    ) {
	/* partial path */
	p = dsgw_ch_malloc( strlen( path ) + strlen( cwd ) + 2 );
	strcpy( p, cwd );
	for ( q = p; *q != '\0'; ++q ) {
#ifdef _WIN32
	    if ( *q == '/' ) {
		*q = '\\';
	    }
#endif
	}
	*q++ = DSGW_PATHSEP_CHAR;
	strcpy( q, path );
    } else {
	/* full path */
	p = dsgw_ch_strdup( path );
    }

    return( p );
}


#ifdef DSGW_DEBUG

/* Returns a directory path used for tmp files. */
char * 
dsgw_get_tmp_dir()
{
	static char	tmp[] = "/tmp";

#if defined( XP_WIN32 )
	int ilen;
	char *pch;
	CHAR *szTempPath;

	szTempPath = malloc( _MAX_PATH +1 );
	GetTempPath( _MAX_PATH +1, szTempPath );
	/* Remove trailing slash. */
	ilen = strlen(szTempPath);
	pch = *(szTempPath+(ilen-1));
	if( pch == '\\' || pch == '/' )
		*(szTempPath+ilen-1) = '\0';
	return szTempPath;
#else
	return( tmp );
#endif
}

/*
 * logging function -- called like printf(); syslog-like output is written
 *	to a file called /tmp/progname where progname is derived from argv[0]
 */
void
dsgw_log( char *fmt, ... )
{
    static FILE	*logfp = NULL;
    time_t	t;
    char	timebuf[ 20 ];
    va_list	ap;

    t = time( NULL );

    if ( logfp == NULL ) {
	char	fname[ 56 ];
	char	*tmppath;

	tmppath = dsgw_get_tmp_dir();
#if defined( XP_WIN32 )
	strcat( tmppath, "/log%.50s" );
#else
	strcat( tmppath, "/%.50s" );
#endif

	sprintf( fname, tmppath, progname );

	if (( logfp = fopen( fname, "a+" )) == NULL ) {
	    return;
	}
    }

    memcpy( timebuf, ctime( &t ), 19 );
    timebuf[ 19 ] = '\0';
    fprintf( logfp, "%s %s: ", timebuf, progname );

    va_start( ap, fmt );
    (void)vfprintf( logfp, fmt, ap );
    va_end( ap );
    fflush( logfp );
}


/*
 * log the contents of a NULL-terminated array of character strings
 */
void
dsgw_logstringarray( char *arrayname, char **strs )
{
    int		i;

    if ( strs == NULL || strs[ 0 ] == NULL ) {
	dsgw_log( "Array %s: empty\n", arrayname );
    } else {
	dsgw_log( "Array %s:\n", arrayname );

	for ( i = 0; strs[ i ] != NULL; ++i ) {
	    dsgw_log( "\t%2d: \"%s\"\n", i, strs[ i ] );
	}
    }
}
#endif /* DSGW_DEBUG */
