/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

#include <stdlib.h>

#include "dsgw.h"
#include "dbtdsgw.h"
#include "libadminutil/admutil.h"
#include "libadminutil/distadm.h"
#include "libadmsslutil/admsslutil.h"
#include "libadmsslutil/psetcssl.h"
#include "../../include/disptmpl.h"

static void get_request();
static int entry2htmlwrite( void *fp, char *buf, int len );

static char *binddn;
static char *bindpw;

int LDAP_CALL LDAP_CALLBACK
ldap_rebind_proc( LDAP *ld, char **whop, char **credp, int *methodp, int freeit, void *arg )
{
    if ( !freeit ) {
        *whop = binddn;
        *credp = bindpw;
        *methodp = LDAP_AUTH_SIMPLE;
    }
    return( LDAP_SUCCESS );
}           

static int
entry2htmlwrite( void *fp, char *buf, int len )
{
        return( fwrite( buf, len, 1, (FILE *)fp ) == 0 ? -1 : len );
}


main( argc, argv, env )
    int		argc;
    char	*argv[];
#ifdef DSGW_DEBUG
    char	*env[];
#endif
{
    (void)dsgw_init( argc, argv,  DSGW_METHOD_GET );

    /*
     * Note that normally we would call dsgw_send_header() here, but
     * we need to do it after examining the query string since there could
     * be a REAUTH flag present (in which case we need to send additional
     * headers).
     */

#ifdef DSGW_DEBUG
   dsgw_logstringarray( "env", env ); 
#endif

    get_request();

    exit( 0 );
}

static void
extractLdapError()
{
	dsgw_error( DSGW_ERR_LDAPURL_BAD, NULL, DSGW_ERROPT_EXIT, 0 );
}


/* Returns an LDAP handle. Return NULL if failed to initialize SSL */
static LDAP *
getLdapConnection(char *admroot, AdmldapInfo ldapInfo,
                                char *ldapUrl) {
        int rv;
	char *ldapHost;
	char *ldapPort;
	char *ptr;
	int port;
	int secure = 0;

	if(!(ldapHost = strstr(ldapUrl, "ldap://"))) 
		if(!(ldapHost = strstr(ldapUrl, "ldaps://"))) 
			extractLdapError();
		else {
			secure = 1;
			ldapHost += strlen("ldaps://");
		}
	else
		ldapHost += strlen("ldap://");
	if (!(ldapPort = strchr(ldapHost, ':')))
		extractLdapError();
	ldapPort[0] = '\0'; 
	if (!(ptr = strchr(++ldapPort, '/')))
		extractLdapError();
	ptr[0] = '\0';
	port = atoi(ldapPort);


        if (secure) {
                rv = ADMSSL_Init(ldapInfo, admroot);
                if (rv) {
			dsgw_error( DSGW_ERR_SSLINIT, NULL, DSGW_ERROPT_EXIT, 0);
                }
                return ldapssl_init(ldapHost, port, 1);
        }
        else
                return ldap_init(ldapHost, port);
}


static void
get_request()
{
    LDAP		*ld;
    char		hostDN[1024], *tmplname, *p, *parent;
    unsigned long	options;
int i=2;

    char                *nsroot = getenv("NETSITE_ROOT");
    int                 rv, error;

    char                admconf[256], *ldapHost, *basedn;
    AdmldapInfo         ldapInfo;
    int                 ldapPort, err, rc;
    LDAPMessage         *msgp, *entry;
    dsgwtmplinfo        *tip;

    char *userGroupLdapURL;
    char *userGroupBindDN;
    char *userGroupBindPW;
    char *dirInfoRef;

    /*
     *
     * Note: original form http://host/edit/dn[/...]?template[&...] is
     *       supported for keeping backward compatibility.
     *       But passing DN as PATH_INFO is NOT recommended.
     *       Since PATH_INFO is passed to CGI as is (non-escaped),
     *       the content has a risk to get broken especially when
     *       it contains 8-bit UTF-8 data.  (This is a known problem
     *       on localized Windows machines.)
     *
     * If the QUERY_STRING is non-NULL, it looks like this:
     *
     *     template [ &INFO=infostring ] [ &ADD ] [ &REAUTH ] \
     *              [&DNATTR=attrname&DNDESC=description] [ &LANG=language ]
     *
     * where:
     *   "template" is the name of the edit template to use for display,
     *   "infostring" is a message used to replace DS_LAST_OP_INFO directives
     *   "attrname" is the name of a DN-valued attribute
     *   "dndesc" is the destriptive name of the above DN-valued attribute
     *   "language" is the current language to use for display, e.g., en-us
     *
     * If "&ADD" is present, we check to make sure the entry
     * does not exist, then we check that the parent entry exists, and then
     * we present an "add entry" form.
     *
     * If "&REAUTH" is send HTTP headers to indicate that HTTP basic
     * authentication failed.  This causes the browser to ask the user to
     * reauthenticate.  It is useful to invoke edit this way immediately
     * following a password change transaction.
     *
     */
    options = DSGW_DISPLAY_OPT_EDITABLE;
    binddn = NULL;

    if (( tmplname = getenv( "QUERY_STRING" )) != NULL && *tmplname != '\0' ) {

      tmplname = dsgw_ch_strdup( tmplname );
      while (( p = strrchr( tmplname, '&' )) != NULL ) {
	*p++ = '\0';
	if ( strcasecmp( p, "add" ) == 0 ) {
	  options |= DSGW_DISPLAY_OPT_ADDING;
	  if (( p = strrchr( tmplname, '&' )) != NULL ) {
	    *p++ = '\0';
	  }
	}

        if ( p != NULL && strncasecmp( p, "dn=", 3 ) == 0 ) {
          binddn = dsgw_ch_strdup( p + 3 );
          dsgw_form_unescape( binddn );
	} else if ( p != NULL && strncasecmp( p, "info=", 5 ) == 0 ) {
	  dsgw_last_op_info = dsgw_ch_strdup( p + 5 );
	  dsgw_form_unescape( dsgw_last_op_info );
	} else if ( p != NULL && strncasecmp( p, "dnattr=", 7 ) == 0 ) {
	  dsgw_dnattr = dsgw_ch_strdup( p + 7 );
	  dsgw_form_unescape( dsgw_dnattr );
	} else if ( p != NULL && strncasecmp( p, "dndesc=", 7 ) == 0 ) {
	  dsgw_dndesc = dsgw_ch_strdup( p + 7 );
	  dsgw_form_unescape( dsgw_dndesc );
	} else if ( p != NULL && strncasecmp( p, "reauth", 6 ) == 0 ) {
	  dsgw_add_header( "Status: 401 Unauthorized" );
	  dsgw_add_header( "WWW-authenticate: basic realm=\""
			   DSGW_ADMIN_AUTH_REALM "\"" );
	} else if ( p != NULL && strncasecmp( p, "lang=", 5 ) == 0 ) {
	  if ( gc->gc_international )
	    gc->gc_inputlanguage = dsgw_tag2lang( gc->gc_languages, p + 5 );
	  else
	    gc->gc_inputlanguage = gc->gc_languages;
	}
      }

      if ( *tmplname == '\0' ) {
	tmplname = NULL;
      }
    } else {
	tmplname = NULL;
    }

    /* Support PATH_INFO */
    if ( binddn == NULL ) {
        if (( binddn = getenv( "PATH_INFO" )) != NULL ) {
            if ( *binddn == '/' ) {
                ++binddn;   /* skip past leading slash */
            }
        }
    }
    /*  atulb:6/3/98  Get binddn and password from the Admin pipe
        This is what the user logged in as from the browser/console */
    if (nsroot) {
      rv = ADM_GetUserDNString(&err, &binddn);

      if (rv < 0 || !binddn || !*binddn) {
	 dsgw_error(DSGW_ERR_MISSING_DS_ENTRY_FOR_USERNAME, NULL, DSGW_ERROPT_EXIT, 0);
      }
      ADM_GetCurrentPassword(&err, &bindpw);

      sprintf(admconf, "%s%c%s%c%s", nsroot, FILE_PATHSEP, "admin-serv",
	      FILE_PATHSEP, "config");

      ldapInfo = admldapBuildInfo(admconf, &rv);

      if (!ldapInfo) {
	 dsgw_error(DSGW_ERR_FAILEDTOBUILD_LDAPINFO, NULL, DSGW_ERROPT_EXIT, 0);
      }
      else {
        /* atulb:6/3/98  Use the AuthenticationLdapURL env var to determine
           which directory to talk to.  No longer use the one in dbswitch.conf*/
        if (getenv("AuthenticationLdapURL")) 
          userGroupLdapURL = dsgw_ch_strdup(getenv("AuthenticationLdapURL"));
	else {
	  dsgw_error(DSGW_ERR_ENV_AUTHLDAPURL_NOTSET, NULL, DSGW_ERROPT_EXIT, 0);
	}
      }
    }
    
    dsgw_send_header();

#ifdef DSGW_DEBUG
    dsgw_log( "get_request: dn: \"%s\", tmplname: \"%s\" "
	      "dnattr: \"%s\", dndesc: \"%s\"\n", userGroupLdapURL,
	    ( tmplname == NULL ) ? "(null)" : tmplname, 
	    ( dsgw_dnattr == NULL ) ? "(null)" : dsgw_dnattr, 
	    ( dsgw_dndesc == NULL ) ? "(null)" : dsgw_dndesc );
#endif

    if ((ld = getLdapConnection(admconf, ldapInfo, userGroupLdapURL )) == NULL ) {
        dsgw_ldap_error( ld, DSGW_ERROPT_EXIT );
    }

    (void)dsgw_init_ldap_woSSL( &ld, NULL );

    ldap_set_rebind_proc (ld, ldap_rebind_proc, NULL);

    if ((rv = ldap_simple_bind_s(ld, binddn, bindpw)) != LDAP_SUCCESS) {
	dsgw_ldap_error( ld, DSGW_ERROPT_EXIT );
    }

    if (( options & DSGW_DISPLAY_OPT_ADDING ) == 0 ) {
      /*
       * editing an existing entry -- if no DN is provided and we are running
       * under the admin server, try to get DN from admin. server
       */
      if ( *binddn == '\0' ) {
	(void)dsgw_get_adm_identity( ld, NULL, (char**) &binddn, NULL,
				     DSGW_ERROPT_EXIT );
      }

      tip = dsgw_display_init(DSGW_TMPLTYPE_DISPLAY, tmplname, options);

      if ((ldap_search_s( ld, binddn, LDAP_SCOPE_BASE, "objectClass=*", 
			  tip->dsti_attrs,0,&msgp)) != LDAP_SUCCESS )
	{
	  dsgw_ldap_error(ld, DSGW_ERROPT_EXIT);
	}
	
	if (( entry = ldap_first_entry (ld, msgp )) == NULL )
	  {
	    ldap_msgfree(msgp);
	    dsgw_ldap_error(ld, DSGW_ERROPT_EXIT);
	  }

	/* display it (finally!) */
	if ( tip == NULL ) {
	  /* no template available -- display in an ugly but complete manner */
	  if (( rc = ldap_entry2html( ld, NULL, entry, NULL, NULL, NULL, entry2htmlwrite, 
				      stdout, "\n", 0, LDAP_DISP_OPT_HTMLBODYONLY,
				      NULL, NULL )) != LDAP_SUCCESS ) {
	    ldap_msgfree( msgp );
	    dsgw_ldap_error( ld, DSGW_ERROPT_EXIT );
	  }
	  dsgw_html_end();
	} else {
	  /* use template to create a nicely formatted display */
	  dsgw_display_entry( tip, ld, msgp, binddn );
	  dsgw_display_done( tip );
	}
	
	ldap_msgfree( msgp );

    } else {
    	char		    *matched;

	/*
	 * new entry -- check to make sure it doesn't exist
	 */
	if ( dsgw_ldap_entry_exists( ld, binddn, &matched, DSGW_ERROPT_EXIT )) {
	    char	**rdns;

	    dsgw_html_begin( XP_GetClientStr(DBT_entryAlreadyExists_), 1 );
	    fputs( XP_GetClientStr(DBT_anEntryNamed_), stdout );
	    rdns = ldap_explode_dn( binddn, 1 );
	    dsgw_html_href(
		    dsgw_build_urlprefix( dsgw_getvp( DSGW_CGINUM_DOSEARCH )),
		    binddn, ( rdns == NULL || rdns[ 0 ] == NULL ) ? binddn : rdns[ 0 ],
		    NULL, XP_GetClientStr(DBT_onmouseoverWindowStatusClickHere_) );
	    if ( rdns != NULL ) {
		ldap_value_free( rdns );
	    }
	    fputs( XP_GetClientStr(DBT_alreadyExistsPPleaseChooseAnothe_), stdout );

	    fputs( "<FORM>\n<CENTER><TABLE border=2 width=\"100%\"><TR>\n",
		    stdout );
	    fputs( "<TD WIDTH=\"50%\" ALIGN=\"center\">\n", stdout );
	    fprintf( stdout, "<INPUT TYPE=\"button\" VALUE=\"%s\" "
		"onClick=\"parent.close()\">", XP_GetClientStr(DBT_closeWindow_1) );
	    fputs( "<TD WIDTH=\"50%\" ALIGN=\"center\">\n", stdout );
	    dsgw_emit_helpbutton( "ENTRYEXISTS" );
	    fputs( "\n</TABLE></CENTER></FORM>\n", stdout );
	    dsgw_html_end();
	} else if ( !dsgw_is_dnparent( matched, binddn ) &&
		!dsgw_dn_cmp( binddn, gc->gc_ldapsearchbase )) {
	    /*
	     * The parent entry does not exist, and the dn being added is not
	     * the same as the suffix for which the gateway is configured.
	     */
	    dsgw_html_begin( XP_GetClientStr(DBT_parentEntryDoesNotExist_), 1 );
	    printf( XP_GetClientStr(DBT_youCannotAddAnEntryByTheNamePBSB_),
		    binddn );
	    parent = dsgw_dn_parent( binddn );
	    if ( parent == NULL || strlen( parent ) == 0 ) {
		fputs( XP_GetClientStr(DBT_itsParentN_), stdout );
	    } else {
		printf( XP_GetClientStr(DBT_anEntryNamedPBSBN_), parent );
		free( parent );
	    }
	    fputs( "<FORM>\n<CENTER><TABLE border=2 width=\"100%\"><TR>\n",
		    stdout );
	    fputs( "<TD WIDTH=\"50%\" ALIGN=\"center\">\n", stdout );
	    fprintf( stdout, "<INPUT TYPE=\"button\" VALUE=\"%s\" "
		"onClick=\"parent.close()\">", XP_GetClientStr(DBT_closeWindow_2) );
	    fputs( "<TD WIDTH=\"50%\" ALIGN=\"center\">\n", stdout );
	    dsgw_emit_helpbutton( "ADD_NOPARENT" );
	    fputs( "\n</TABLE></CENTER></FORM>\n", stdout );
	    dsgw_html_end();
	} else {
	    /*
	     * The parent exists, or the user is adding the entry whose DN
	     * is the same as the suffix for which the gateway is configured.
	     * Display the "add entry" form.
	     */

	    if ( tmplname == NULL ) {
#ifdef DSGW_DEBUG
                dsgw_log( "NULL tmplname\n" );
#endif
		dsgw_error( DSGW_ERR_MISSINGINPUT, XP_GetClientStr(DBT_missingTemplate_),
			DSGW_ERROPT_EXIT, 0 );
	    }

	    tip = dsgw_display_init( DSGW_TMPLTYPE_DISPLAY, tmplname, options );
     
	    dsgw_display_entry( tip, ld, NULL, binddn );
	    dsgw_display_done( tip );
	}
    }

    ldap_unbind( ld );
}

