/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * ldif.c - routines for parsing and creating LDIF data.
 *
 * based on ns/netsite/ldap/libraries/libldif/line64.c rev 1.2.16.6 which was
 *	on the tip of server3_directory_branch on 10/20/97.  Changes:
 *	1) changed "ldif_" function prefix to "dsgw_ldif_"
 *	2) replaced all #include's with #include "dsgw.h"
 *	3) added #if 0/#endif around ldif_get_entry() function
 *	4) copied definitions from ldif.h rev 1.2.22.2 to dsgw.h
 *	5) removed all calls to LDAPDebug()
 */

#include "dsgw.h"

#ifndef isascii
#define isascii( c )	(!((c) & ~0177))
#endif

#define RIGHT2			0x03
#define RIGHT4			0x0f
#define CONTINUED_LINE_MARKER	'\001'

#define ISBLANK(c) (c == ' ' || c == '\t' || c == '\n') /* not "\r\v\f" */

static char nib2b64[0x40f] =
        "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

static unsigned char b642nib[0x80] = {
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0x3e, 0xff, 0xff, 0xff, 0x3f,
	0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b,
	0x3c, 0x3d, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06,
	0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e,
	0x0f, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16,
	0x17, 0x18, 0x19, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0x20,
	0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28,
	0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f, 0x30,
	0x31, 0x32, 0x33, 0xff, 0xff, 0xff, 0xff, 0xff
};

/*
 * ldif_parse_line - takes a line of the form "type:[:] value" and splits it
 * into components "type" and "value".  if a double colon separates type from
 * value, then value is encoded in base 64, and parse_line un-decodes it
 * (in place) before returning.
 */

int
dsgw_ldif_parse_line(
    char	*line,
    char	**type,
    char	**value,
    int		*vlen
)
{
	char	*p, *s, *d;
	int	b64;

	/* skip any leading space */
	while ( ISBLANK( *line ) ) {
		line++;
	}
	*type = line;

	for ( s = line; *s && *s != ':'; s++ )
		;	/* NULL */
	if ( *s == '\0' ) {
		return( -1 );
	}

	/* trim any space between type and : */
	for ( p = s - 1; p > line && ISBLANK( *p ); p-- ) {
		*p = '\0';
	}
	*s++ = '\0';

	/* check for double : - indicates base 64 encoded value */
	if ( *s == ':' ) {
		s++;
		b64 = 1;

	/* single : - normally encoded value */
	} else {
		b64 = 0;
	}

	/* skip space between : and value */
	while ( ISBLANK( *s ) ) {
		s++;
	}

	/* 
	 * If no value is present, return a zero-length string for
	 * *value, with *vlen set to zero.
	 */
	if ( *s == '\0' ) {
		*value = s;
		*vlen = 0;
		return( 0 );
	}

	/* check for continued line markers that should be deleted */
	for ( p = s, d = s; *p; p++ ) {
		if ( *p != CONTINUED_LINE_MARKER )
			*d++ = *p;
	}
	*d = '\0';

	*value = s;
	if ( b64 ) {
		if (( *vlen = dsgw_ldif_base64_decode( s, (unsigned char *)s ))
		    < 0 ) {
			return( -1 );
		}
		s[ *vlen ] = '\0';
	} else {
		*vlen = (int) (d - s);
	}

	return( 0 );
}


/*
 * ldif_base64_decode - take the BASE64-encoded characters in "src"
 * (a zero-terminated string) and decode them into the the buffer "dst".
 * "src" and "dst" can be the same if in-place decoding is desired.
 * "dst" must be large enough to hold the decoded octets.  No more than
 *	3 * strlen( src ) / 4 bytes will be produced.
 * "dst" may contain zero octets anywhere within it, but it is not
 *	zero-terminated by this function.
 *
 * The number of bytes copied to "dst" is returned if all goes well.
 * -1 is returned if the BASE64 encoding in "src" is invalid.
 */

int
dsgw_ldif_base64_decode( char *src, unsigned char *dst )
{
	char		*p, *stop;
	unsigned char	nib, *byte;
	int		i, len;

	stop = strchr( src, '\0' );
	byte = dst;
	for ( p = src, len = 0; p < stop; p += 4, len += 3 ) {
		for ( i = 0; i < 3; i++ ) {
			if ( p[i] != '=' && (p[i] & 0x80 ||
			    b642nib[ p[i] & 0x7f ] > 0x3f) ) {
				return( -1 );
			}
		}

		/* first digit */
		nib = b642nib[ p[0] & 0x7f ];
		byte[0] = nib << 2;

		/* second digit */
		nib = b642nib[ p[1] & 0x7f ];
		byte[0] |= nib >> 4;

		/* third digit */
		if ( p[2] == '=' ) {
			len += 1;
			break;
		}
		byte[1] = (nib & RIGHT4) << 4;
		nib = b642nib[ p[2] & 0x7f ];
		byte[1] |= nib >> 2;

		/* fourth digit */
		if ( p[3] == '=' ) {
			len += 2;
			break;
		}
		byte[2] = (nib & RIGHT2) << 6;
		nib = b642nib[ p[3] & 0x7f ];
		byte[2] |= nib;

		byte += 3;
	}

	return( len );
}

/*
 * ldif_getline - return the next "line" (minus newline) of input from a
 * string buffer of lines separated by newlines, terminated by \n\n
 * or \0.  this routine handles continued lines, bundling them into
 * a single big line before returning.  if a line begins with a white
 * space character, it is a continuation of the previous line. the white
 * space character (nb: only one char), and preceeding newline are changed
 * into CONTINUED_LINE_MARKER chars, to be deleted later by the
 * ldif_parse_line() routine above.
 *
 * it takes a pointer to a pointer to the buffer on the first call,
 * which it updates and must be supplied on subsequent calls.
 *
 * XXX need to update this function to also support <CR><LF> as EOL.
 */

char *
dsgw_ldif_getline( char **next )
{
	char	*l;
	char	c;

	if ( *next == NULL || **next == '\n' || **next == '\0' ) {
		return( NULL );
	}

	while ( **next == '#' ) {	/* skip comment lines */
		if (( *next = strchr( *next, '\n' )) == NULL ) {
			return( NULL );
		}
		(*next)++;
	}

	l = *next;
	while ( (*next = strchr( *next, '\n' )) != NULL ) {
		c = *(*next + 1);
		if ( ISBLANK( c ) && c != '\n' ) {
			**next = CONTINUED_LINE_MARKER;
			*(*next+1) = CONTINUED_LINE_MARKER;
		} else {
			*(*next)++ = '\0';
			break;
		}
		(*next)++;
	}

	return( l );
}

void
dsgw_ldif_put_type_and_value( char **out, char *t, char *val, int vlen )
{
	unsigned char	*p, *byte, *stop;
	char		*save;
	int		b64, len, savelen;
	len = 0;

	/* put the type + ": " */
	for ( p = (unsigned char *) t; *p; p++, len++ ) {
		*(*out)++ = *p;
	}
	*(*out)++ = ':';
	len++;
	save = *out;
	savelen = len;
	*(*out)++ = ' ';
	b64 = 0;

	stop = (unsigned char *) (val + vlen);
	if ( isascii( val[0] ) && isspace( val[0] ) || val[0] == ':' ) {
		b64 = 1;
	} else {
		for ( byte = (unsigned char *) val; byte < stop;
		    byte++, len++ ) {
			if ( !isascii( *byte ) || (!isprint( *byte ) &&
			    *byte != '\t') ) {
				b64 = 1;
				break;
			}
			if ( len > DSGW_LDIF_MAX_LINE_WIDTH ) {
				*(*out)++ = '\n';
				*(*out)++ = ' ';
				len = 1;
			}
			*(*out)++ = *byte;
		}
	}
	if ( b64 ) {
		*out = save;
		*(*out)++ = ':';
		*(*out)++ = ' ';
		len = dsgw_ldif_base64_encode( (unsigned char *)val, *out, vlen,
		    savelen + 2 );
		*out += len;
	}

	*(*out)++ = '\n';
}


/*
 * ldif_base64_encode - encode "srclen" bytes in "src", place BASE64
 * encoded bytes in "dst" and return the length of the BASE64
 * encoded string.  "dst" is also zero-terminated by this function.
 *
 * If "lenused" >= 0, newlines will be included in "dst" and "lenused" if
 * appropriate.  "lenused" should be a count of characters already used
 * on the current line.  The LDIF lines we create will contain at most
 * "DSGW_LDIF_MAX_LINE_WIDTH" characters on each line.
 *
 * If "lenused" < 0, no newlines will be included, and the DSGW_LDIF_BASE64_LEN()
 * macro can be used to determine how many bytes will be placed in "dst."
 */

int
dsgw_ldif_base64_encode( unsigned char *src, char *dst, int srclen, int lenused )
{
	unsigned char	*byte, *stop;
	unsigned char	buf[3];
	char		*out;
	unsigned long	bits;
	int		i, pad, len;

	len = 0;
	out = dst;
	stop = src + srclen;

	/* convert to base 64 (3 bytes => 4 base 64 digits) */
	for ( byte = src; byte < stop - 2; byte += 3 ) {
		bits = (byte[0] & 0xff) << 16;
		bits |= (byte[1] & 0xff) << 8;
		bits |= (byte[2] & 0xff);

		for ( i = 0; i < 4; i++, bits <<= 6 ) {
			if ( lenused >= 0 && lenused++ > DSGW_LDIF_MAX_LINE_WIDTH ) {
				*out++ = '\n';
				*out++ = ' ';
				lenused = 2;
			}

			/* get b64 digit from high order 6 bits */
			*out++ = nib2b64[ (bits & 0xfc0000L) >> 18 ];
		}
	}

	/* add padding if necessary */
	if ( byte < stop ) {
		for ( i = 0; byte + i < stop; i++ ) {
			buf[i] = byte[i];
		}
		for ( pad = 0; i < 3; i++, pad++ ) {
			buf[i] = '\0';
		}
		byte = buf;
		bits = (byte[0] & 0xff) << 16;
		bits |= (byte[1] & 0xff) << 8;
		bits |= (byte[2] & 0xff);

		for ( i = 0; i < 4; i++, bits <<= 6 ) {
			if ( lenused >= 0 && lenused++ > DSGW_LDIF_MAX_LINE_WIDTH ) {
				*out++ = '\n';
				*out++ = ' ';
				lenused = 2;
			}

			if (( i == 3 && pad > 0 ) || ( i == 2 && pad == 2 )) {
				/* Pad as appropriate */
				*out++ = '=';
			} else {
				/* get b64 digit from low order 6 bits */
				*out++ = nib2b64[ (bits & 0xfc0000L) >> 18 ];
			}
		}
	}

	*out = '\0';

	return( out - dst );
}


/*
 * return malloc'd, zero-terminated LDIF line
 */
char *
dsgw_ldif_type_and_value( char *type, char *val, int vlen )
{
    char	*buf, *p;
    int		tlen;

    tlen = strlen( type );
    if (( buf = (char *)malloc( DSGW_LDIF_SIZE_NEEDED( tlen, vlen ) + 1 )) !=
	    NULL ) {
	p = buf;
	dsgw_ldif_put_type_and_value( &p, type, val, vlen );
	*p = '\0';
    }

    return( buf );
}


#if 0
/*
 * ldif_get_entry - read the next ldif entry from the FILE referenced
 * by fp. return a pointer to a malloc'd, null-terminated buffer. also
 * returned is the last line number read, in *lineno.
 */
char *
ldif_get_entry( FILE *fp, int *lineno )
{
	char	line[BUFSIZ];
	char	*buf;
	int	max, cur, len, gotsome;

	buf = NULL;
	max = cur = gotsome = 0;
	while ( fgets( line, sizeof(line), fp ) != NULL ) {
		if ( lineno != NULL ) {
			(*lineno)++;
		}
		/* ldif entries are terminated by a \n on a line by itself */
		if ( line[0] == '\0' || line[0] == '\n' ) {
			if ( gotsome ) {
				break;
			} else {
				continue;
			}
		}
		gotsome = 1;
		len = strlen( line );
		while ( cur + (len + 1) > max ) {
			max += BUFSIZ;
			if ( buf == NULL ) {
				buf = (char *) malloc( max );
			} else {
				buf = (char *) realloc( buf, max );
			}
			if ( buf == NULL ) {
				return( NULL );
			}
		}

		memcpy( buf + cur, line, len + 1 );
		cur += len;
	}

	return( buf );
}
#endif /* 0 */
