/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * cache.c
 *
 * Abstract cache class.
 *
 * Warning:	procedures needing to lock the cache_crit and a cache_entry lock
 *		must always lock the cache_crit first in order to avoid 
 *		deadlocks.
 *
 * Note:	When inserting a new entry into a full cache where all 
 *		entries are currently in use (access_count >0), new entries 
 *		will be inserted.  Entries at the LRU will be marked as
 *		delete_pending.  Basically this means the cache will grow
 *		indefinitely up to the working set size. XXXMB
 *
 * ------------------
 *
 * Mike Belshe
 * 01-01-96
 */

#include "netsite.h"
#include "base/systems.h"

#include <sys/types.h>
#include <errno.h>
#include "base/ereport.h"
#include "base/util.h"
#include "base/cache.h"
#include "base/session.h"
#include "frame/req.h"
#include "frame/http.h"
#include "base/crit.h"
#include "base/daemon.h"  /* daemon_atrestart */
#include "nsassert.h"
#include "base/dbtbase.h"

/* The cache_list is really only here for debugging the multiple caches which
 * may be present in the system.  cache_crit is only used to protect access
 * to the cache_list.
 */
static cache_t *cache_list = NULL;
static CRITICAL cache_crit = NULL;

cache_entry_t * _cache_entry_lookup(cache_t *cache, void *key);

NSAPI_PUBLIC cache_t *
cache_create(unsigned int cache_max, unsigned int hash_size,
	public_cache_functions_t *fnlist)
{
	cache_t *newentry;
	unsigned int index;

SOLARIS_PROBE(cache_create_start, "cache");

	NS_ASSERT(cache_max > 0);
	NS_ASSERT(hash_size > 0);
	NS_ASSERT(fnlist);
	if ( (newentry = (cache_t *)PERM_MALLOC(sizeof(cache_t))) == NULL) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_insufficientMemoryToCreateHashTa_));
SOLARIS_PROBE(cache_create_end, "cache");
		return NULL;
	}

#ifdef CACHE_DEBUG
	newentry->magic = CACHE_MAGIC;
#endif
	newentry->cache_size	= 0;
	newentry->hash_size	= hash_size;
	newentry->max_size	= cache_max;
    newentry->insert_ok = 0;
    newentry->insert_fail = 0;
    newentry->deletes = 0;
	newentry->cache_hits	= 0;
	newentry->cache_lookups = 0;
	newentry->lock 	= crit_init();
	newentry->virtual_fn	= fnlist;
	if ( (newentry->table = (cache_entry_t **)
		PERM_MALLOC((sizeof(cache_entry_t *)) * (newentry->hash_size)))
		== NULL) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_insufficientMemoryToCreateHashTa_1));
SOLARIS_PROBE(cache_create_end, "cache");
		return NULL;
	}
	for (index=0; index<newentry->hash_size; index++) 
		newentry->table[index] = NULL;
	newentry->mru_head = newentry->lru_head = NULL;

	if (!cache_crit) 
		cache_crit = crit_init();
	
	crit_enter(cache_crit);
	newentry->next = cache_list;
	cache_list = newentry;
	crit_exit(cache_crit);

SOLARIS_PROBE(cache_create_end, "cache");
	return newentry;
}

NSAPI_PUBLIC void 
cache_destroy(void *cache_ptr)
{
	cache_t *cache = (cache_t *)cache_ptr;
	cache_t *search, *last;
	cache_entry_t *ptr;
SOLARIS_PROBE(cache_destroy_start, "cache");

	NS_ASSERT(cache_crit);
	NS_ASSERT(cache_ptr);
	NS_ASSERT(cache->magic == CACHE_MAGIC);

	crit_enter(cache_crit);
	crit_enter(cache->lock);

	ptr = cache->lru_head;
	while(ptr) {
		/* Caller MUST bump the access_count before calling delete
		 * We can do this since we hold the cache lock.  
		 */
		cache_use_increment(cache, ptr);
		cache_delete(cache, ptr);
		ptr = cache->lru_head;
	}

	PERM_FREE(cache->table);

	cache->max_size = 0;
	cache->hash_size = 0;

	for ( last = NULL, search = cache_list; search; last = search,
		search = search->next)
		if (search == cache)
			break;

	if (search) {
		if (last) 
			last->next = search->next;
		else
			cache_list = search->next;
	}
	else {
		ereport(LOG_WARN, XP_GetAdminStr(DBT_cacheDestroyCacheTablesAppearCor_));
	}
	crit_exit(cache_crit);
	crit_exit(cache->lock);

	crit_terminate(cache->lock);

	PERM_FREE(cache);
SOLARIS_PROBE(cache_destroy_end, "cache");
}

/* _cache_make_mru
 * Put this element at the MRU element.
 * Caller must hold cache lock.
 */
static int
_cache_make_mru(cache_t *cache, cache_entry_t *entry) 
{
SOLARIS_PROBE(cache_make_mru_start, "cache");
	NS_ASSERT(cache);
	NS_ASSERT(entry);
	NS_ASSERT(cache->magic == CACHE_MAGIC);
	NS_ASSERT(entry->magic == CACHE_ENTRY_MAGIC);

	if (cache->mru_head == NULL)
		cache->mru_head = cache->lru_head = entry;
	else {
		if (cache->mru_head != entry) {
			if (cache->lru_head == entry)
				cache->lru_head = entry->mru;
			/* only check entry->lru if we aren't the lru_head */
			else if (entry->lru) 
				entry->lru->mru = entry->mru;
			if (entry->mru)
				entry->mru->lru = entry->lru;
			entry->lru = cache->mru_head;
			entry->mru = NULL;
			/* cache->mru_head cannot be NULL here */
			cache->mru_head->mru = entry;
			cache->mru_head = entry;
		}
	}

SOLARIS_PROBE(cache_make_mru_end, "cache");
	return 0;
}

/* _cache_remove_mru()
 * Remove an entry from the MRU list. 
 * Caller MUST hold the cache->lock in order to call this routine.
 * Caller MUST be on the MRU list; otherwise this routine will segfault.
 */
static int
_cache_remove_mru(cache_t *cache, cache_entry_t *entry)
{

SOLARIS_PROBE(cache_remove_mru_start, "cache");
	NS_ASSERT(cache);
	NS_ASSERT(cache->mru_head);
	NS_ASSERT(cache->lru_head);
	NS_ASSERT(entry);
	NS_ASSERT((cache->mru_head == cache->lru_head) || 
		(entry->mru || entry->lru));
	NS_ASSERT(cache->magic == CACHE_MAGIC);
	NS_ASSERT(entry->magic == CACHE_ENTRY_MAGIC);

	if (cache->mru_head == entry) 
		cache->mru_head = entry->lru;
	else 
		/* if we are not the mru_head, entry->mru cannot be NULL */
		entry->mru->lru = entry->lru;
	if (cache->lru_head == entry) 
		cache->lru_head = entry->mru;
	else 
		/* if we are not the lru_head, entry->lru cannot be NULL */
		entry->lru->mru = entry->mru;

	entry->mru = entry->lru = NULL;

#ifdef CACHE_DEBUG
	{
		cache_entry_t *ptr;
		for (ptr = cache->mru_head; ptr; ptr = ptr->lru)
			NS_ASSERT(ptr != entry);
		for (ptr = cache->lru_head; ptr; ptr = ptr->mru)
			NS_ASSERT(ptr != entry);
	}
#endif /* CACHE_DEBUG */

SOLARIS_PROBE(cache_remove_mru_end, "cache");
	return 0;
}

NSAPI_PUBLIC int 
cache_touch(cache_t *cache, cache_entry_t *entry)
{
SOLARIS_PROBE(cache_touch_start, "cache");
	NS_ASSERT(cache_crit);
	NS_ASSERT(cache);
	NS_ASSERT(entry);
	NS_ASSERT(entry->access_count > 0);
	NS_ASSERT(cache->magic == CACHE_MAGIC);
	NS_ASSERT(entry->magic == CACHE_ENTRY_MAGIC);

	/* Since entries which have the access count > 0 (and the caller
	 * must hold the access count by definition, does this routine
	 * make sense anymore? XXXMB
	 */

SOLARIS_PROBE(cache_touch_end, "cache");
	return 0;
}

cache_entry_t *
cache_create_entry(void)
{
	cache_entry_t *newentry;

SOLARIS_PROBE(cache_create_entry_start, "cache");
	if ( (newentry = (cache_entry_t *)PERM_MALLOC(sizeof(cache_entry_t))) == NULL) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_unableToAllocateHashEntry_));
SOLARIS_PROBE(cache_create_entry_end, "cache");
		return NULL;
	}

#ifdef CACHE_DEBUG
	newentry->magic = CACHE_ENTRY_MAGIC;
#endif

SOLARIS_PROBE(cache_create_entry_end, "cache");
	return newentry;
}

NSAPI_PUBLIC int
cache_insert_p(cache_t *cache, cache_entry_t *entry, void *key, void *data,
	cache_entry_functions_t *fn)
{
	cache_entry_t *tmp;
	int bucket;
	cache_entry_t *delete_ptr;

SOLARIS_PROBE(cache_insert_p_start, "cache");
	NS_ASSERT(cache_crit);
	NS_ASSERT(cache);
	NS_ASSERT(entry);
	NS_ASSERT((cache->mru_head && cache->lru_head) || 
		(!cache->mru_head && !cache->lru_head));
	NS_ASSERT(cache->magic == CACHE_MAGIC);
	NS_ASSERT(entry->magic == CACHE_ENTRY_MAGIC);

	crit_enter(cache->lock);
	/* Reserve a space in the cache if possible; if not; try to delete
	 * the oldest guy...  Note that since we increment to reserve
	 * space; it is possible that we have maxed the cache when there
	 * are no entries yet added; so we need to check the lru_head ptr
	 * to see if the list is empty.  If the list is empty, there is 
	 * nothing we can delete, so just fail the insert request.  This
	 * condition should go away momentarily.
	 */
	if (cache->cache_size >= cache->max_size) {
		if (!cache->lru_head) {
			/* No space in the cache */
            cache->insert_fail++;
			crit_exit(cache->lock);
SOLARIS_PROBE(cache_insert_p_end, "cache");
			return -1;
		}

		/* Caller MUST bump the access_count before calling delete
		 * We can do this since we hold the cache lock.
		 */
		delete_ptr = cache->lru_head;
		cache_use_increment(cache, delete_ptr);
		if ( cache_delete(cache, delete_ptr) < 0 ) {
			cache_use_decrement(cache, delete_ptr);
			NS_ASSERT(delete_ptr->access_count > 0);
            cache->insert_fail++;
			crit_exit(cache->lock);
SOLARIS_PROBE(cache_insert_p_end, "cache");
			return -1;
		}
	}
	cache->cache_size++;
	crit_exit(cache->lock);

	entry->key = key;
	entry->data = data;
	entry->access_count = 1;
	entry->delete_pending = 0;
	entry->fn_list = fn;
	entry->next = NULL;
	entry->lru = NULL;
	entry->mru = NULL;
	bucket = cache->virtual_fn->hash_fn(cache->hash_size, key);

	crit_enter(cache->lock);
	/* Don't add duplicate entries in the cache */
	if ( (tmp = _cache_entry_lookup(cache, key)) ) {
                /* Try to delete original element */
                if ( cache_delete(cache, tmp) < 0) {
			cache->cache_size--;
			cache_use_decrement(cache, tmp);
			NS_ASSERT(tmp->access_count > 0);
            cache->insert_fail++;
			crit_exit(cache->lock);
SOLARIS_PROBE(cache_insert_p_end, "cache");
			return -1;
		}
	}

	/* Insert in hash table */
	entry->next = cache->table[bucket];
	cache->table[bucket] = entry;

    cache->insert_ok++;
	crit_exit(cache->lock);

SOLARIS_PROBE(cache_insert_p_end, "cache");
	return 0;
}

NSAPI_PUBLIC cache_entry_t *
cache_insert(cache_t *cache, void *key, void *data, cache_entry_functions_t *fn)
{
	cache_entry_t *newentry;

SOLARIS_PROBE(cache_insert_start, "cache");
	NS_ASSERT(cache_crit);
	NS_ASSERT(cache);
	NS_ASSERT(cache->magic == CACHE_MAGIC);

	if ( (newentry = cache_create_entry()) == NULL) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_cacheInsertUnableToCreateCacheEn_));
SOLARIS_PROBE(cache_insert_end, "cache");
		return NULL;
	}

	if ( cache_insert_p(cache, newentry, key, data, fn) < 0) {
		PERM_FREE(newentry);
SOLARIS_PROBE(cache_insert_end, "cache");
		return NULL;
	}


SOLARIS_PROBE(cache_insert_end, "cache");
	return newentry;
}

NSAPI_PUBLIC int
cache_delete(cache_t *cache, cache_entry_t *entry)
{
	int bucket;
	cache_entry_t *ptr, *last;

SOLARIS_PROBE(cache_delete_try_start, "cache");
	NS_ASSERT(cache);
	NS_ASSERT(entry);
	NS_ASSERT(cache->magic == CACHE_MAGIC);
	NS_ASSERT(entry->magic == CACHE_ENTRY_MAGIC);
	NS_ASSERT(cache_crit);

	/* the caller is supposed to bump access count to before calling us.  
	 * If the count is less than 1, something is wrong...
	 */
	NS_ASSERT(entry->access_count >= 1);

	crit_enter(cache->lock);

	entry->delete_pending = 1;

	/* Check the access_count now that we have the lock.  If
	 * we got context switched just after we checked, it is possible
	 * that someone else is now using the entry.
	 */
	if (entry->access_count > 1) {
		crit_exit(cache->lock);
SOLARIS_PROBE(cache_delete_try_end, "cache");
		return -1;
	}

SOLARIS_PROBE(cache_delete_start, "cache");
	/* Delete from the hash table */
	bucket = cache->virtual_fn->hash_fn(cache->hash_size, entry->key);
	for (last = NULL, ptr = cache->table[bucket]; ptr; 
		last = ptr, ptr= ptr->next)
		if (entry == ptr)
			break;
	NS_ASSERT(ptr);
	if (ptr) {
		if (last)
			last->next = ptr->next;
		else
			cache->table[bucket] = ptr->next;
	}
	cache->cache_size--;

	/* don't need to remove from the MRU/LRU list because
	 * the access count is > 0 in order to call delete 
	 */

	/* Technically, someone else could kill the whole cache
	 * right now, and we would be left with a bum pointer.   XXXMB
	 */
	entry->fn_list->cleanup_fn(entry->data);

    cache->deletes++;
	crit_exit(cache->lock);

	PERM_FREE(entry);

SOLARIS_PROBE(cache_delete_end, "cache");
	return 0;
}


NSAPI_PUBLIC int 
cache_use_increment(cache_t *cache, cache_entry_t *entry)
{
	int ret;
SOLARIS_PROBE(cache_use_increment_start, "cache");
	NS_ASSERT(cache_crit);
	NS_ASSERT(cache);
	NS_ASSERT(entry);
	NS_ASSERT(cache->magic == CACHE_MAGIC);
	NS_ASSERT(entry->magic == CACHE_ENTRY_MAGIC);

	/* 
	 * Why do we do a cache_do_lookup()? Solves the problem where one
	 * thread calls delete() while another called increment().  If
	 * the delete happened first, then we need to lookup the element
	 * before we try to bump the access count.
	 */
	ret = cache_do_lookup(cache, entry->key)?0:-1;
SOLARIS_PROBE(cache_use_increment_end, "cache");
	return ret;
}

NSAPI_PUBLIC int
cache_use_decrement(cache_t *cache, cache_entry_t *entry)
{
SOLARIS_PROBE(cache_use_decrement_start, "cache");
	NS_ASSERT(cache_crit);
	NS_ASSERT(cache);
	NS_ASSERT(entry);
	NS_ASSERT(entry->access_count > 0);
	NS_ASSERT(cache->magic == CACHE_MAGIC);
	NS_ASSERT(entry->magic == CACHE_ENTRY_MAGIC);

	crit_enter(cache->lock);
	/* If we are the last user of this entry and the delete
	 * is pending, cleanup now!
	 */
	if ((entry->access_count == 1) && (entry->delete_pending)) {
		if (cache_delete(cache, entry) < 0) {
			/* can't delete right now; should never happen
			 * since we have the cache lock and we know the
			 * access count is 1
			 */
			entry->access_count--;
			(void)_cache_make_mru(cache, entry);
		}
	} else {
		if (entry->access_count == 1)
			(void)_cache_make_mru(cache, entry);
		entry->access_count--;
	}
	crit_exit(cache->lock);

SOLARIS_PROBE(cache_use_decrement_end, "cache");
	return 0;
}

cache_entry_t *
_cache_entry_lookup(cache_t *cache, void *key)
{
	int bucket;
	cache_entry_t *ptr;

SOLARIS_PROBE(cache_entry_lookup_start, "cache");
	NS_ASSERT(cache_crit);
	NS_ASSERT(cache);
	NS_ASSERT(cache->magic == CACHE_MAGIC);

	crit_enter(cache->lock);

	cache->cache_lookups++;

	bucket = cache->virtual_fn->hash_fn(cache->hash_size, key);

	for (ptr = cache->table[bucket]; ptr; ptr = ptr->next) {
		if (!ptr->delete_pending && !ptr->fn_list->key_cmp_fn(key, ptr->key))
			break;
	}

	if (ptr) {
		cache->cache_hits++;
		if (ptr->access_count == 0) 
			(void)_cache_remove_mru(cache, ptr);
		ptr->access_count++;
		crit_exit(cache->lock);
SOLARIS_PROBE(cache_entry_lookup_end, "cache");
		return ptr;
	}
	crit_exit(cache->lock);

SOLARIS_PROBE(cache_entry_lookup_end, "cache");
	return NULL;
}

NSAPI_PUBLIC void *
cache_do_lookup(cache_t *cache, void *key)
{
	cache_entry_t *ptr;

SOLARIS_PROBE(cache_do_lookup_start, "cache");
	NS_ASSERT(cache_crit);
	NS_ASSERT(cache);
	NS_ASSERT(cache->magic == CACHE_MAGIC);

	ptr = _cache_entry_lookup(cache, key);

SOLARIS_PROBE(cache_do_lookup_end, "cache");
	if (ptr) 
		return ptr->data;
	else
		return NULL;
    
}

NSAPI_PUBLIC int 
cache_valid(cache_t *cache, cache_entry_t *entry)
{
	NS_ASSERT(cache_crit);
	NS_ASSERT(cache);
	NS_ASSERT(entry);
	NS_ASSERT(entry->access_count > 0);

	return entry->delete_pending?-1:0;
}

NSAPI_PUBLIC void
cache_lock(cache_t *cache)
{
SOLARIS_PROBE(cache_lock_start, "cache");
	crit_enter(cache->lock);
SOLARIS_PROBE(cache_lock_end, "cache");
}

NSAPI_PUBLIC void
cache_unlock(cache_t *cache)
{
SOLARIS_PROBE(cache_unlock_start, "cache");
	crit_exit(cache->lock);
SOLARIS_PROBE(cache_unlock_end, "cache");
}

NSAPI_PUBLIC unsigned int
cache_get_use_count(cache_t *cache, cache_entry_t *entry)
{
	unsigned int result;

	crit_enter(cache->lock);
	result = entry->access_count;
	crit_exit(cache->lock);

	return result;
}


#define MAX_DEBUG_LINE	1024
#ifdef DEBUG_CACHES
/* cache_service_debug()
 * To force the creation of good debugging tools, all known caches must
 * have a debugging routine.  The cache_service_debug() routine is a common
 * routine that can be inserted via NSAPI to view the status of all
 * known caches.  This is a hidden entry point which can be enabled when
 * debugging is needed.
 *
 */
NSAPI_PUBLIC int 
cache_service_debug(pblock *pb, Session *sn, Request *rq)
{
	cache_t *ptr;
	char buf[MAX_DEBUG_LINE];
	int len;

	NS_ASSERT(cache_crit);

	param_free(pblock_remove("content-type", rq->srvhdrs));
	pblock_nvinsert("content-type", "text/html", rq->srvhdrs);

	len = util_sprintf(buf, XP_GetClientStr(DBT_http10200OkNcontentTypeTextHtmlN_));
	net_write(sn->csd, buf, len);

	len = util_sprintf(buf, XP_GetClientStr(DBT_H2NetscapeCacheStatusReportH2N_));
	net_write(sn->csd, buf, len);

	crit_enter(cache_crit);
	if (cache_list) {
		len = util_sprintf(buf, "<HR>");
		net_write(sn->csd, buf, len);
		for (ptr = cache_list; ptr; ptr = ptr->next) {
			if (ptr->virtual_fn->debug_fn)
				if ( ptr->virtual_fn->debug_fn(pb, sn, rq) == REQ_ABORTED ) 
					return REQ_ABORTED;
		}
	} else {
		len = util_sprintf(buf, XP_GetClientStr(DBT_noCachesOnSystemP_));
		net_write(sn->csd, buf, len);
	}
	crit_exit(cache_crit);

	return REQ_PROCEED;
}

#endif /* 0 */

/* cache_dump()
 * A generic routine to dump a cache to a socket
 */
NSAPI_PUBLIC int
cache_dump(cache_t *cache, char *cache_name, SYS_NETFD fd)
{
#ifdef DEBUG_CACHES /* XXXrobm this causes ball of string effect */
	char buf[MAX_DEBUG_LINE];
	cache_entry_t *ptr;
	int index, len;

	NS_ASSERT(cache_crit);
	NS_ASSERT(cache);
	NS_ASSERT(cache_name);

	if (!cache_crit)
		return -1;

	crit_enter(cache->lock);
	len = util_sprintf(buf, XP_GetClientStr(DBT_H2SCacheH2N_), cache_name);
	net_write(fd, buf, len);
	len = util_sprintf(buf, XP_GetClientStr(DBT_cacheHitRatioDDFPNPN_),
		cache->cache_hits, cache->cache_lookups, 
		(cache->cache_lookups > 0)?
		((cache->cache_hits)/(cache->cache_lookups)):0.0);
	net_write(fd, buf, len);
	len = util_sprintf(buf, XP_GetClientStr(DBT_cacheSizeDDPNPN_),
		cache->cache_size, cache->max_size);
	net_write(fd, buf, len);
	len = util_sprintf(buf, XP_GetClientStr(DBT_hashTableSizeDPNPN_),
		cache->hash_size);
	net_write(fd, buf, len);
	len = util_sprintf(buf, XP_GetClientStr(DBT_mruDPNlruDPN_),
                cache->mru_head, cache->lru_head);
        net_write(fd, buf, len);

	/* Create an HTML table */
	len = util_sprintf(buf, XP_GetClientStr(DBT_UlTableBorder4ThBucketThThAddres_));
	net_write(fd, buf, len);

	for (index = 0; index < cache->hash_size; index++) {
		if (cache->table[index]) {
			for (ptr = cache->table[index]; ptr; ptr = ptr->next) {

				len = util_snprintf(buf, MAX_DEBUG_LINE, "\
<tr align=right> \
<td>%d</td> \
<td>%d</td> \
<td>%d</td> \
<td>%d</td> \
<td>%d</td> \
<td>%d</td> \
<td>%d</td> \
<td>%d</td> <td>",
					index, ptr, ptr->key,ptr->access_count, 
					ptr->delete_pending, ptr->next, 
					ptr->lru, ptr->mru);

				net_write(fd, buf, len);

				(void)ptr->fn_list->print_fn(ptr->data, fd);

				len = util_sprintf(buf, "</td></tr>\n");
				net_write(fd, buf, len);

			}
		}
	}

	len = util_sprintf(buf, "</TABLE></UL>\n");
	net_write(fd, buf, len);
	crit_exit(cache->lock);
	
#endif
	return REQ_PROCEED;
}
