/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * cinfo.c: Content Information for a file, i.e. its type, etc.
 * 
 * See cinfo.h for full details.
 *
 * Rob McCool
 */


#include "cinfo.h"

#include "netsite.h"            /* MALLOC, FREE, STRDUP */
#include "buffer.h"             /* filebuf_t and its ilk */
#include "util.h"               /* getline */

#include <string.h>             /* strrchr */
#include <ctype.h>              /* isspace */
#include <sys/types.h>


/* -------------------------- Module structures --------------------------- */


struct _cdata {
    int num_exts;
    char **exts;

    cinfo ci;
    struct _cdata *next;
};

struct _chash {
    char *ext;
    struct _cdata *info;
    struct _chash *next;
};


/* ----------------------------- Static data ------------------------------ */


/* The master database of content info (linked list) */
static struct _cdata *cimlist = NULL;

/* The database hash table, hashed on extension */
static struct _chash *cimhash[CINFO_HASHSIZE];


/* ---------------------------- _cdata_insert ----------------------------- */


void _cdata_insert(struct _cdata *cd) {
    struct _chash *ch;
    register int x, n = cd->num_exts;
    int hv;

    cd->next = cimlist;
    cimlist = cd;

    if(n) {
        for(x = 0; x < n; x++) {
            ch = (struct _chash *) MALLOC(sizeof(struct _chash));

            hv = CINFO_HASH(cd->exts[x]);
            ch->ext = cd->exts[x];
            ch->info = cd;
            ch->next = cimhash[hv];
            cimhash[hv] = ch;
        }
    }
}


/* ----------------------------- cinfo_create ----------------------------- */


void cinfo_zero(cinfo *ci)
{
    ci->type = NULL;
    ci->encoding = NULL;
    ci->language = NULL;
}

cinfo *cinfo_create(void) {
    cinfo *ci = (cinfo *) MALLOC(sizeof(cinfo));

    cinfo_zero(ci);
    return ci;
}


/* ------------------------------ cinfo_free ------------------------------ */


void cinfo_free(cinfo *ci) {
    if(ci->type) FREE(ci->type);
    if(ci->encoding) FREE(ci->encoding);
    if(ci->language) FREE(ci->language);
}


/* ---------------------------- _cdata_create ----------------------------- */


struct _cdata *_cdata_create(void) {
    struct _cdata *cd = (struct _cdata *) MALLOC(sizeof(struct _cdata));

    cd->num_exts = 0;
    cd->exts = NULL;
    cinfo_zero(&cd->ci);

    cd->next = NULL;

    return cd;
}


/* ------------------------- cinfo_dump_database -------------------------- */


NSAPI_PUBLIC void cinfo_dump_database(FILE *dump) {
    struct _cdata *cd = cimlist;
    register int x;

    fprintf(dump,"The content info database contains the following:\n");

    while(cd) {
        fprintf(dump,"For exts ");
        fputs((cd->num_exts ? cd->exts[0] : " (none)"), dump);
        for(x = 1; x < cd->num_exts; x++) 
            fprintf(dump,", %s", cd->exts[x]);

        fputs(":\n", dump);

        if(cd->ci.type)
            fprintf(dump, "Content-type: %s\n", cd->ci.type);
        if(cd->ci.encoding)
            fprintf(dump, "Content-encoding: %s\n", cd->ci.encoding);
        if(cd->ci.language)
            fprintf(dump, "Content-Language: %s\n", cd->ci.language);

        cd = cd->next;
    }
}


/* ----------------------------- _cdata_free ------------------------------ */


static void _cdata_free(struct _cdata *cd) {
    register int x;

    if(cd->exts) {
        for(x = 0; x < cd->num_exts; x++)
            FREE(cd->exts[x]);

        FREE(cd->exts);
    }
    cinfo_free(&cd->ci);

    FREE(cd);
}


/* ---------------------------- _cdata_new_ext ---------------------------- */


static void _cdata_new_ext(char *ext, struct _cdata *cd) {
    int n = cd->num_exts;

    cd->exts = (char **) (n ? REALLOC(cd->exts, (n+1)*sizeof(char *)) 
                            : MALLOC(sizeof(char *)));

    cd->exts[n] = STRDUP(ext);
    cd->num_exts = ++n;
}


/* ------------------------------ cinfo_init ------------------------------ */


NSAPI_PUBLIC void cinfo_init(void) {
    register int x;

    cimlist = NULL;
    for(x = 0; x < CINFO_HASHSIZE; x++)
        cimhash[x] = NULL;
}


/* --------------------------- cinfo_terminate ---------------------------- */


NSAPI_PUBLIC void cinfo_terminate(void) {
    struct _cdata *dp = cimlist, *dq;
    struct _chash *hp, *hq;
    register int x;

    while(dp) {
        dq = dp->next;
        _cdata_free(dp);
        dp = dq;
    }
    cimlist = NULL;

    for(x = 0; x < CINFO_HASHSIZE; x++) {
        hp = cimhash[x];
        while(hp) {
            hq = hp->next;
            FREE(hp);
            hp = hq;
        }
        cimhash[x] = NULL;
    }
}


/* ----------------------------- cinfo_merge ------------------------------ */


static char *_cinfo_parse_mcc(filebuf_t *buf, char *t);

NSAPI_PUBLIC char *cinfo_merge(char *fn) {
    SYS_FILE fd;
    filebuf_t *buf;
    char t[CINFO_MAX_LEN], err[MAGNUS_ERROR_LEN], *ret;

    if((fd = system_fopenRO(fn)) == SYS_ERROR_FD) {
        util_snprintf(err, MAGNUS_ERROR_LEN, "can't open types file %s (%s)", fn, 
                     system_errmsg());
        return STRDUP(err);
    }
    if(!(buf = filebuf_open(fd, FILE_BUFFERSIZE))) {
        util_snprintf(err, MAGNUS_ERROR_LEN, "can't mmap types file %s (%s)", fn, 
                     system_errmsg());
        return STRDUP(err);
    }
    ret = NULL;

    switch(util_getline(buf, 1, CINFO_MAX_LEN, t)) {
      case 0:
        /* Look for magic number */
        if((!strncmp(t, MCC_MT_MAGIC, MCC_MT_MAGIC_LEN)) || 
           (!strncmp(t, NCC_MT_MAGIC, NCC_MT_MAGIC_LEN)))
        {
            ret = _cinfo_parse_mcc(buf, t);
            break;
        }
        /* NCSA or unknown file: drop through */
      case 1:
        /* EOF: do nothing */
        ret = STRDUP("unknown file format");
        break;
      case -1:
        /* Error */
        ret = STRDUP(t);
    }

    filebuf_close(buf);
    return ret;
}


/* ----------------------------- cinfo_lookup ----------------------------- */


NSAPI_PUBLIC cinfo *cinfo_lookup(char *type)
{
    struct _cdata *cd=cimlist;

    while(cd) {
        if(cd->ci.type) {
            if(!strcasecmp(type, cd->ci.type))
                return &cd->ci;
        }
        cd = cd->next;
    }
    return NULL;
}


/* ------------------------------ cinfo_find ------------------------------ */


NSAPI_PUBLIC cinfo *cinfo_find(char *uri) {
    char *es, *et, c;
    cinfo *ci = NULL;
    struct _cdata *cd;
    struct _chash *ch;
    int hv;

    if(!(es = strrchr(uri,FILE_PATHSEP)))
        es = uri;
    else
        ++es;

    /* Check for extensions, or no database */
    if((!cimlist) || (!(es = strchr(es,CINFO_SEPARATOR))))
        return NULL;

    ++es;

    while(*es) {
        et = es;
        while((*et) && (*et != CINFO_SEPARATOR)) ++et;
        if(et != es) {
            c = *et;
            *et = '\0';

            hv = CINFO_HASH(es);
            ch = cimhash[hv];
            while(ch) {
                if(!strcasecmp(ch->ext, es)) {
                    if(!ci) ci = cinfo_create();

                    cd = ch->info;
                    if(cd->ci.type)
                        ci->type = cd->ci.type;
                    if(cd->ci.encoding)
                        ci->encoding = cd->ci.encoding;
                    if(cd->ci.language)
                        ci->language = cd->ci.language;

                    break;
                }
                ch = ch->next;
            }
            if( (*et = c) )
                ++et;
        }
        else {
            /* Some jarhead sent us foo..gif or similar, try to skip. */
            ++et;
        }
        es = et;
    }
    return ci;
}


/* ----------------------------- File parsers ----------------------------- */


/* Parse sexy new MCC mime types files */

static char *_cinfo_parse_mcc(filebuf_t *buf, char *line) 
{
    int ln = 1;
    int eof = 0;
#define LOCAL_ERROR_LEN 256
    char err[LOCAL_ERROR_LEN], *t, *u, *v, *name, *value, c;
    struct _cdata *cd;

    /* t is already filled. */
    while(1) {
        cd = NULL;

        t = line;
        if((*t) && (*t != '#')) {
            while(1) {
                while(*t && isspace(*t)) ++t;
                if(!(*t))
                    break;
                name = t;

                while(*t && (*t != '=')) ++t;
                if(!(*t)) {
                    util_snprintf(err, LOCAL_ERROR_LEN, "line %d: name without =", ln);
                    return STRDUP(err);
                }
                for(u = t - 1; isspace(*u); --u);
                *(u+1) = '\0';   /* terminate name */
                ++t;
                while(*t && (*t != '\"') && (isspace(*t))) ++t;
                if(!(*t)) {
                    util_snprintf(err, LOCAL_ERROR_LEN, "line %d: empty value", ln);
                    return STRDUP(err);
                }
                if(*t == '\"') {
                    value = ++t;
                    while(*t && (*t != '\"')) ++t;
                    if(!(*t)) {
                        util_snprintf(err, LOCAL_ERROR_LEN, "line %d: missing closing quotes", 
                                     ln);
                        return STRDUP(err);
                    }
                }
                else {
                    value = t;
                    while(*t && (!isspace(*t))) ++t;
                    /* null t is okay */
                }
                if(*t) *t++ = '\0';
                if(!cd) cd = _cdata_create();

                if(!strcasecmp(name, "type"))
                    cd->ci.type = STRDUP(value);
                else if(!strcasecmp(name, "enc"))
                    cd->ci.encoding = STRDUP(value);
                else if(!strcasecmp(name, "lang"))
                    cd->ci.language = STRDUP(value);
                else if(!strcasecmp(name, "exts")) {
                    u = value;
                    while(1) {
                        v = u;
                        while(*u && (*u != ',')) ++u;
                        c = *u;
                        *u = '\0';
                        _cdata_new_ext(v, cd);
                        if(c)
                            ++u;
                        else
                            break;
                    }
                }
            }
            if(cd)
                _cdata_insert(cd);
        }
        if(eof) break;
        ++ln;
        switch(util_getline(buf, ln, CINFO_MAX_LEN, line)) {
          case -1:
            /* Error */
            return STRDUP(line);
          case 1:
            /* EOF */
            eof = 1;
        }
    }

    return NULL;
}

