/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * dns_cache.c
 *
 * Cache of DNS entries
 *
 * Mike Belshe
 * 02-03-96
 */

#include "netsite.h"

#include <errno.h>
#include "base/cache.h"
#include "base/dns_cache.h"
#include "base/buffer.h"
#include "base/ereport.h"
#include "base/util.h"
#include "base/daemon.h"

#include "base/dbtbase.h"

/* ----- Macros ------------------------------------------------------- */
#define DEFAULT_EXPIRE_TIME	(20*60)		/* 20 minutes */
#define ABSOLUTE_EXPIRE_MAX	(60*60*24*365)	/* 1 year */

#define DEFAULT_CACHE_MAX	1024
#define ABSOLUTE_CACHE_MIN	32	/* disallow smaller sizes */
#define ABSOLUTE_CACHE_MAX	32768	/* disallow larger sizes */
#define DEFAULT_HASH_SIZE	(2*DEFAULT_CACHE_MAX)

/* ----- Forwards ----------------------------------------------------- */
static unsigned int dns_cache_hash_ip(unsigned int, void *);
static int dns_cache_compare_keys(void *, void *);
static int dns_cache_cleanup(void *);
static int dns_cache_debug(pblock *pb, Session *sn, Request *rq);
static int dns_cache_print(void *data, SYS_NETFD fd);

/* ----- Static globals ----------------------------------------------- */
static cache_t			*dns_cache = NULL;
static unsigned long		dns_expire_time = DEFAULT_EXPIRE_TIME;

static	public_cache_functions_t dns_cache_functions = {
		dns_cache_hash_ip,
		dns_cache_debug
	};

static	cache_entry_functions_t dns_cache_entry_functions = {
		dns_cache_compare_keys,
		dns_cache_cleanup,
		dns_cache_print
	};

/* ----- Functions ----------------------------------------------------- */

static unsigned int 
dns_cache_hash_ip(unsigned int hash_size, void *voidint)
{
	register unsigned int *intptr = (unsigned int *)voidint;

	return (*intptr) % hash_size;
}

static int
dns_cache_compare_keys(void *voidint1, void *voidint2)
{
	return  !( *(unsigned int *)voidint1 == *(unsigned int *)voidint2);
}

static int 
dns_cache_cleanup(void *voiddata)
{
	dns_cache_entry_t *data = (dns_cache_entry_t *)voiddata;

	if (!dns_cache || !data)
		return -1;

	if (data->host)
		PERM_FREE(data->host);

	/* It is imperative that we do not FREE the dns_cache_entry_t (data)
     * structure here.  cache_delete() will do that for us.
     */

	return 0;
}

static int
dns_cache_debug(pblock *pb, Session *sn, Request *rq)
{
	if (!dns_cache)
		return 0;

	return cache_dump(dns_cache, "DNS Cache", sn->csd);
}

#define MAX_DEBUG_LINE	1024
static int
dns_cache_print(void *voiddata, SYS_NETFD fd)
{
#ifdef DEBUG_CACHES	/* XXXhep this causes ball of string effect */
	dns_cache_entry_t *data = (dns_cache_entry_t *)voiddata;
	char buf[MAX_DEBUG_LINE];
	int len;

	if (!dns_cache || !data)
		return 0;

	len = util_snprintf(buf, MAX_DEBUG_LINE,  "%s %d\n",
		data->host?data->host:"null", data->ip);
	net_write(fd, buf,len);
#endif /* DEBUG_CACHES */

	return 0;
}

#ifdef MCC_ADMSERV
/* dns_cache_admin_init()
 * quick hack for initializing dns cache for admin server 
 */
NSAPI_PUBLIC int
dns_cache_admin_init()
{
	int dns_cache_size = 32;
	dns_expire_time = DEFAULT_EXPIRE_TIME;

	if ( (dns_cache = cache_create(dns_cache_size, 2*dns_cache_size, 
		&dns_cache_functions)) ==NULL){
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_errorCreatingDnsCache_));
		return -1;
	}

	return 0;
}
#endif
	
/* NOTE:  I'm using NSAPI Init routines to initialize the cache... 
 *	  This is probably not purist nsapi...
 */
NSAPI_PUBLIC int
dns_cache_init(pblock *pb, Session *sn, Request *rq)
{
	char *str_hash_size = pblock_findval("hash-size", pb);
	char *str_cache_size = pblock_findval("cache-size", pb);
	char *str_expire_time = pblock_findval("expire", pb);
	char *str_disable = pblock_findval("disable", pb);
	int hash_size;
	int cache_size;

	/* Check if cache already initialized */
	if (dns_cache)
		return REQ_PROCEED;

    if (str_disable)
        return REQ_PROCEED;

	if (str_hash_size == NULL)
		hash_size = DEFAULT_HASH_SIZE;
	else 
		if ( (hash_size = atoi(str_hash_size)) <= 0) {
			ereport(LOG_WARN, 
				XP_GetAdminStr(DBT_dnsCacheInitHashSize0UsingD_), 
				DEFAULT_HASH_SIZE);
			hash_size = DEFAULT_HASH_SIZE;
		}

	if (str_cache_size == NULL) 
		cache_size = DEFAULT_CACHE_MAX;
	else {
		if ( (cache_size = atoi(str_cache_size)) < ABSOLUTE_CACHE_MIN) {
			ereport(LOG_WARN, 
				XP_GetAdminStr(DBT_dnsCacheInitCacheSizeDUsingD_),
				ABSOLUTE_CACHE_MIN, DEFAULT_CACHE_MAX);
			cache_size = ABSOLUTE_CACHE_MIN;
		} else {
			if (cache_size > ABSOLUTE_CACHE_MAX) {
				ereport(LOG_WARN, 
					XP_GetAdminStr(DBT_dnsCacheInitCacheSizeIsDIsTooLar_),
					cache_size, ABSOLUTE_CACHE_MAX);
				cache_size = ABSOLUTE_CACHE_MAX;
			}
		}
		 
	}

	if (str_expire_time == NULL) 
		dns_expire_time = DEFAULT_EXPIRE_TIME;
	else 
		if ( (dns_expire_time = atoi(str_expire_time)) <= 0) {
			ereport(LOG_WARN, 
				XP_GetAdminStr(DBT_dnsCacheInitExpireTime0UsingD_),
				DEFAULT_EXPIRE_TIME);
			dns_expire_time = DEFAULT_EXPIRE_TIME;
		} else {
			if (dns_expire_time > ABSOLUTE_EXPIRE_MAX) {
				ereport(LOG_WARN, 
					XP_GetAdminStr(DBT_dnsCacheInitExpireIsDIsTooLargeU_),
					dns_expire_time, ABSOLUTE_EXPIRE_MAX);
				dns_expire_time = ABSOLUTE_EXPIRE_MAX;
			}
		}

	if ( (dns_cache = cache_create(cache_size, hash_size, 
		&dns_cache_functions)) ==NULL){
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_errorCreatingDnsCache_1));
		return REQ_ABORTED;
	}

	return REQ_PROCEED;
}

NSAPI_PUBLIC void 
dns_cache_destroy(cache_t *cache)
{
	/* cache module automatically cleans up the cache itself */
	dns_cache = NULL;
}


NSAPI_PUBLIC dns_cache_entry_t *
dns_cache_insert(char *host, unsigned int ip, unsigned int verified)
{
	dns_cache_entry_t *newentry;

	if ( !dns_cache || !ip)  {
		return NULL;
    }

	if ( (newentry = (dns_cache_entry_t *)PERM_MALLOC(sizeof(dns_cache_entry_t))) == NULL) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_dnsCacheInsertErrorAllocatingEnt_));
		goto error;
	}

	if (host) {
		if ( (newentry->host = PERM_STRDUP(host)) == NULL) {
			ereport(LOG_FAILURE, XP_GetAdminStr(DBT_dnsCacheInsertMallocFailure_));
			goto error;
		}
	} else
		newentry->host = NULL;

#ifdef CACHE_DEBUG
	newentry->cache.magic = CACHE_ENTRY_MAGIC;
#endif

	newentry->ip = ip;
	newentry->verified = verified;

	newentry->last_access = time(NULL);

	if ( cache_insert_p(dns_cache, &(newentry->cache), (void *)&(newentry->ip), 
		(void *)newentry, &dns_cache_entry_functions) < 0) {
		/* Not a bad error; it just means the cache is full */
		goto error;
	}

	return newentry;

error:
	if (newentry) {
		if (newentry->host)
			PERM_FREE(newentry->host);
		PERM_FREE(newentry);
	}
	return NULL;
}

NSAPI_PUBLIC int
dns_cache_delete(void *entry)
{
	if ( !dns_cache || !entry ) 
		return -1;

	return cache_delete(dns_cache, (cache_entry_t *)entry);
}


NSAPI_PUBLIC int 
dns_cache_use_increment(void *entry)
{
	if ( !dns_cache || !entry ) 
		return -1;

	return cache_use_increment(dns_cache, (cache_entry_t *)entry);
}

NSAPI_PUBLIC int
dns_cache_use_decrement(void *entry)
{
	if ( !dns_cache || !entry) 
		return -1;

	return cache_use_decrement(dns_cache, (cache_entry_t *)entry);
}

NSAPI_PUBLIC dns_cache_entry_t *
dns_cache_lookup_ip(unsigned int ip)
{
	dns_cache_entry_t *ptr;

	if (!dns_cache || !ip ) 
		return NULL;

	ptr = (dns_cache_entry_t *)cache_do_lookup(dns_cache, (void *)&ip);

	if (ptr) {
		if (dns_cache_valid(ptr) < 0) {
			if (dns_cache_delete(ptr) < 0)
				/* couldn't delete right now */
				dns_cache_use_decrement(ptr);
			return NULL;
		}

		(void)dns_cache_touch(ptr);
	}

	return ptr;
}

NSAPI_PUBLIC int
dns_cache_valid(dns_cache_entry_t *entry)
{
	time_t now;

	now = time(NULL);

	if (now > (time_t)(entry->last_access + dns_expire_time))
		return -1;

	return cache_valid(dns_cache, (cache_entry_t *)entry);
}

NSAPI_PUBLIC int
dns_cache_touch(dns_cache_entry_t *entry)
{
	if (!entry)
		return -1;

	return cache_touch(dns_cache, (cache_entry_t *)entry);
}

#include "httpdaemon/internalstats.h"

PRBool
GetDNSCacheInfo(DNSCacheInfo *info)
{
    if (!info)
        return PR_FALSE;

    memset(info, 0, sizeof(DNSCacheInfo));

    info->enabled = (dns_cache == NULL)?PR_FALSE:PR_TRUE;

    if (info->enabled) {
        info->maxCacheEntries = dns_cache->max_size;
        info->numCacheEntries = dns_cache->cache_size;

        info->numCacheHits = dns_cache->cache_hits;
        info->numCacheMisses = dns_cache->cache_lookups - dns_cache->cache_hits;

        info->numCacheInsertsOk = dns_cache->insert_ok;
        info->numCacheInsertsFail = dns_cache->insert_fail;
        info->numCacheDeletes = dns_cache->deletes;
    }

    return PR_TRUE;
}
