/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
** dstat.c: Support for daemon statistics
** 
** 
*/


#include "prtypes.h"
#include "prlock.h"
#include "prthread.h"
#include "httpdaemon/dstats.h"
#include "httpdaemon/httpdaemon.h"
#include "base/nsassert.h"
#include "base/util.h"
#include "frame/conf.h"
#include "base/ereport.h"

/* --- data structures -------------------------------------------- */
typedef struct _dstat_t {
    unsigned    statsize;
    unsigned    maxthreads;
    shmem_s    *statmem;	/* handle to stats shared memory */
    StatHeader *stats_hdr;	/* header block for statistics */
    StatSlot   *stats;		/* statistics data */
    PRLock     *crit;
    unsigned    last_slot;	/* last slot assigned */
} _dstat_t;

/* --- globals ---------------------------------------------------- */
PRUint32         dstat_tls_key = DSTATS_INVALID_KEY;
PRBool           dstat_initialized = PR_FALSE; 
static _dstat_t *ds = NULL;

/* --- routines --------------------------------------------------- */

static void
dstat_process_restart(void *unused)
{
    PRUint32 x;

    ds->stats_hdr->restart_cookie++;
    for(x = 0; x < ds->maxthreads; x++) {
        ds->stats[x].mode = (char)CHILD_EMPTY_SLOT;
    }
}


NSAPI_PUBLIC PRInt32
dstat_init(PRInt32 port, char *addr, PRUint32 threads, struct passwd *pw)
{
    char fn[256];
    PRUint32 x;

    NS_ASSERT(!ds);

    PR_NewThreadPrivateIndex(&dstat_tls_key, NULL);

    if (!(ds = (_dstat_t *)PERM_MALLOC(sizeof(_dstat_t))) ) {
        return -1;
    }

    if ( (ds->crit = PR_NewLock()) == NULL) {
        PERM_FREE(ds);
        ds = NULL;
        return -1;
    }

    ds->maxthreads = threads;
    ds->statsize = sizeof(StatHeader) + 
               (ds->maxthreads * sizeof(StatSlot));

#ifdef XP_UNIX
    util_sprintf(fn, "/tmp/daemonstat.%s.%d", addr ? addr : "any", port);
#else
    util_sprintf(fn, "%s%s", DSTATS_FILE, server_id);
#endif
    if(!(ds->statmem = shmem_alloc(fn, ds->statsize, 1))) {
        fprintf(stderr, "warning: statistics disabled (%s)\n", 
                system_errmsg());
        PERM_FREE(ds);
        ds = NULL;
        return -1;
    }

#ifdef XP_UNIX
    if (pw) {
        if (chown(fn, pw->pw_uid, pw->pw_gid) < 0) {
            fprintf(stderr,
                    "warning: failed to set server user as owner of %s (%s)\n",
                    fn, system_errmsg());
        }
    }
#endif

    ds->last_slot = 0;

    ZERO(ds->statmem->data, ds->statsize);

    ds->stats_hdr = (StatHeader *)ds->statmem->data;
    ds->stats_hdr->header_size    = sizeof(StatHeader);
    ds->stats_hdr->slot_size      = sizeof(StatSlot);
    ds->stats_hdr->software       = DSTATS_SOFTWARE_ENTERPRICE;
    ds->stats_hdr->version_major  = DSTATS_MAJOR_VERSION;
    ds->stats_hdr->version_minor  = DSTATS_MINOR_VERSION;
    ds->stats_hdr->obsolete1      = 1;
    ds->stats_hdr->maxthreads     = ds->maxthreads;
    ds->stats_hdr->start_time     = time(NULL);

    srand(time(NULL));
    ds->stats_hdr->restart_cookie = (PRInt32)rand();
    ds->stats_hdr->active         = (char)1;

    ds->stats = (StatSlot *)(((char *)(ds->statmem->data)) + 
        sizeof(StatHeader));
    for(x = 0; x < ds->maxthreads; x++)
        ds->stats[x].mode = (char)CHILD_EMPTY_SLOT;

    dstat_initialized = PR_TRUE;
    return 0;
}

NSAPI_PUBLIC PRUint32
dstat_get_thread_slot()
{
    PRUint32 rv;
    PRUint32 sentinel;

    if (!ds)
        return DSTATS_INVALID_SLOT;


    PR_Lock(ds->crit);

    sentinel = ds->last_slot;
    while(ds->stats[ds->last_slot].mode != (char)CHILD_EMPTY_SLOT) {
        ds->last_slot++;
        if (ds->last_slot > ds->maxthreads)
            ds->last_slot = 0;
 
        if (ds->last_slot == sentinel) {
            /* Argh; we are looping! */
            ereport(LOG_FAILURE, "dstats: unable to find empty statistics slot.  Possible thread resource problem?");
            ds->last_slot = 0;
            break;
        }
    }

    rv = ds->last_slot++;
    PR_Unlock(ds->crit);

    PR_SetThreadPrivate(dstat_tls_key, (void *)rv);
    return rv;
}

NSAPI_PUBLIC void
dstat_destroy()
{
    if (!ds)
        return;
 
    ds->stats_hdr->active = (char)0;

    shmem_free(ds->statmem);
    PR_DestroyLock(ds->crit);
    PERM_FREE(ds);
    ds = NULL;
}

static void 
dstat_record_child_status(PRUint32 slot, PRInt32 http_status, PRUint32 bytes)
{
    NS_ASSERT(!ds || slot < ds->last_slot);

    if (ds) {
        if (bytes > 0)
            ds->stats[slot].tot_bytes += bytes;
        ds->stats[slot].tot_reqs++;
        switch (http_status / 100) {
          case 2:  ds->stats[slot].tot_2xx++; break;
          case 3:  ds->stats[slot].tot_3xx++; break;
          case 4:  ds->stats[slot].tot_4xx++; break;
          case 5:  ds->stats[slot].tot_5xx++; break;
          default: ds->stats[slot].tot_xxx++;
        }
        switch (http_status) {
          case 200: ds->stats[slot].tot_200++; break;
          case 302: ds->stats[slot].tot_302++; break;
          case 304: ds->stats[slot].tot_304++; break;
          /*
           * tot_errs is displayed in sitemon as "Bad Requests"
           * which corresponds to status code 400.
           */
          case 400: ds->stats[slot].tot_errs++; break;
          case 401: ds->stats[slot].tot_401++; break;
          case 403: ds->stats[slot].tot_403++; break;
          default:  ;
	    }
    }
}

NSAPI_PUBLIC void 
child_status(PRInt32 mode)
{
    PRUint32 myslot;

    if(ds) {
        NS_ASSERT(dstat_tls_key != DSTATS_INVALID_KEY);

        myslot = (PRUint32)(long)PR_GetThreadPrivate(dstat_tls_key);

        NS_ASSERT(myslot < ds->last_slot);
        ds->stats[myslot].mode = mode;
    }
}


NSAPI_PUBLIC void 
child_status2(PRInt32 http_status, PRUint32 bytes)
{
    PRInt32 myslot;

    if (ds) {
        NS_ASSERT(dstat_tls_key != DSTATS_INVALID_KEY);
        myslot = (PRInt32)(long)PR_GetThreadPrivate(dstat_tls_key);
        dstat_record_child_status(myslot, http_status, bytes);
    }
}

