/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * netbuf.c: Handles buffered I/O on network sockets
 * 
 * Rob McCool
 */


#include "buffer.h"
#include "ereport.h"      /* ereport */



/* ----------------------------- netbuf_open ------------------------------ */


NSAPI_PUBLIC netbuf *netbuf_open(SYS_NETFD sd, int sz) {
    netbuf *buf = (netbuf *) MALLOC(sizeof(netbuf));

    buf->rdtimeout = NET_READ_TIMEOUT;
    buf->pos = sz;
    buf->cursize = sz; /* force buffer_next on first getc */
    buf->maxsize = sz;
    buf->sd = sd;

    buf->inbuf = NULL;
    buf->errmsg = NULL;

    return buf;
}


/* ----------------------------- netbuf_close ----------------------------- */


NSAPI_PUBLIC void netbuf_close(netbuf *buf) {
    if(buf->inbuf)
        FREE(buf->inbuf);
    FREE(buf);
}


/* ----------------------------- netbuf_next ------------------------------ */


NSAPI_PUBLIC int netbuf_next(netbuf *buf, int advance) {
    int n;

    if(!buf->inbuf)
        buf->inbuf = (unsigned char *) MALLOC(buf->maxsize);

    while(1) {
        switch(n = net_read(buf->sd,(char *)(buf->inbuf),buf->maxsize,buf->rdtimeout)) {
        case IO_EOF:
            return IO_EOF;
        case IO_ERROR:
            buf->errmsg = system_errmsg();
            return IO_ERROR;
        default:
            buf->pos = advance;
            buf->cursize = n;
            return buf->inbuf[0];
        }
    }
}

NSAPI_PUBLIC int netbuf_getbytes(netbuf *buf, char *buffer, int size)
{
    int bytes;

    if (!buf->inbuf) {
        buf->inbuf = (unsigned char *) MALLOC(buf->maxsize);
    } else {
        if (buf->pos < buf->cursize) {
            int bytes_in_buffer = buf->cursize - buf->pos;

            if (bytes_in_buffer > size)
                bytes_in_buffer = size;

            memcpy(buffer, &(buf->inbuf[buf->pos]), bytes_in_buffer);

            buf->pos += size;
            return bytes_in_buffer;
        }
    }

    /* The netbuf is empty.  Read data directly into the caller's buffer */
    bytes = net_read(buf->sd, buffer, size, buf->rdtimeout);
    if (bytes == 0)
        return NETBUF_EOF;
    if (bytes < 0) {
        buf->errmsg = system_errmsg();
        return NETBUF_ERROR;
    }
    return bytes;
}


/* ----------------------------- netbuf_grab ------------------------------ */


NSAPI_PUBLIC int netbuf_grab(netbuf *buf, int sz) {
    int n;

    if(!buf->inbuf) {
        buf->inbuf = (unsigned char *) MALLOC(sz);
        buf->maxsize = sz;
    }
    else if(sz > buf->maxsize) {
        buf->inbuf = (unsigned char *) REALLOC(buf->inbuf, sz);
        buf->maxsize = sz;
    }

    while(1) {
        switch(n = net_read(buf->sd,(char *)(buf->inbuf),sz,buf->rdtimeout)) {
        case IO_EOF:
            return IO_EOF;
        case IO_ERROR:
            buf->errmsg = system_errmsg();
            return IO_ERROR;
        default:
            buf->pos = 0;
            buf->cursize = n;
            return n;
        }
    }
}


/* ---------------------------- netbuf_buf2sd ----------------------------- */


NSAPI_PUBLIC int netbuf_buf2sd(netbuf *buf, SYS_NETFD sd, int len)
{
    register int n = len, t, ns;

    ns = 0;

    /* First, flush the current buffer */
    t = buf->cursize - buf->pos;

    if((n != -1) && (t > n))
        t = n;
    if((t) && (sd != SYS_NET_ERRORFD)) {
#if defined(OSF1) || defined(HPUX) || defined(SNI)
      fuckDEC:
#endif
        if(net_write(sd, (char *)&buf->inbuf[buf->pos], t) == IO_ERROR) {
#if defined(OSF1) || defined(HPUX) || defined(SNI)
            if(errno == EINTR) goto fuckDEC;
#endif
            buf->errmsg = system_errmsg();
            return IO_ERROR;
        }
        ns += t;
    }
    buf->pos += t;

    if(n != -1) {
        n -= t;
        if(!n)
            return ns;
    }
    else 
        t = buf->maxsize;

    /* Now, keep blasting until done */

    while(1) {
        if(n != -1)
            t = (n < buf->maxsize ? n : buf->maxsize);

#if defined(OSF1) || defined(HPUX) || defined(SNI)
      OSFsuck:
#endif
        switch(netbuf_grab(buf, t)) {
          case IO_ERROR:
#if defined(OSF1) || defined(HPUX) || defined(SNI)
            if(errno == EINTR) goto OSFsuck;
#endif
            return IO_ERROR;
          case IO_EOF:
            if(n == -1)
                return ns;
            else {
                buf->errmsg = "premature EOF";
                return IO_ERROR;
            }
          default:
            if(sd != SYS_NET_ERRORFD) {
#if defined(OSF1) || defined(HPUX)
              OSFmustdie:
#endif
                if(net_write(sd, (char *)(buf->inbuf), buf->cursize) == IO_ERROR) {
#if defined(OSF1) || defined(HPUX)
                    if(errno == EINTR) goto OSFmustdie;
#endif
                    buf->errmsg = system_errmsg();
                    return IO_ERROR;
                }
                ns += buf->cursize;
            }
            if(n != -1) {
                n -= buf->cursize;
                if(!n)
                    return ns;
            }
            break;
        }
    }
}


NSAPI_PUBLIC int netbuf_buf2file(netbuf *buf, SYS_FILE sd, int len)
{
    register int n = len, t, ns;

    ns = 0;

    /* First, flush the current buffer */
    t = buf->cursize - buf->pos;

    if((n != -1) && (t > n))
        t = n;
    if((t) && (sd != SYS_NET_ERRORFD)) {
#if defined(OSF1) || defined(HPUX) || defined(SNI)
      fuckDEC:
#endif
        if(system_fwrite(sd, (char *)&buf->inbuf[buf->pos], t) == IO_ERROR) {
#if defined(OSF1) || defined(HPUX) || defined(SNI)
            if(errno == EINTR) goto fuckDEC;
#endif
            buf->errmsg = system_errmsg();
            return IO_ERROR;
        }
        ns += t;
    }
    buf->pos += t;

    if(n != -1) {
        n -= t;
        if(!n)
            return ns;
    }
    else 
        t = buf->maxsize;

    /* Now, keep blasting until done */

    while(1) {
        if(n != -1)
            t = (n < buf->maxsize ? n : buf->maxsize);

#if defined(OSF1) || defined(HPUX) || defined(SNI)
      OSFsuck:
#endif
        switch(netbuf_grab(buf, t)) {
          case IO_ERROR:
#if defined(OSF1) || defined(HPUX) || defined(SNI)
            if(errno == EINTR) goto OSFsuck;
#endif
            return IO_ERROR;
          case IO_EOF:
            if(n == -1)
                return ns;
            else {
                buf->errmsg = "premature EOF";
                return IO_ERROR;
            }
          default:
            if(sd != SYS_NET_ERRORFD) {
#if defined(OSF1) || defined(HPUX)
              OSFmustdie:
#endif
                if(system_fwrite(sd, (char *)(buf->inbuf), buf->cursize) == IO_ERROR) {
#if defined(OSF1) || defined(HPUX)
                    if(errno == EINTR) goto OSFmustdie;
#endif
                    buf->errmsg = system_errmsg();
                    return IO_ERROR;
                }
                ns += buf->cursize;
            }
            if(n != -1) {
                n -= buf->cursize;
                if(!n)
                    return ns;
            }
            break;
        }
    }
}



/* ---------------------------- buffer_buf2sd ----------------------------- */


NSAPI_PUBLIC int filebuf_buf2sd(filebuf_t *buf, SYS_NETFD sd)
{
#ifndef FILE_MMAP
    int n = 0;
    while(filebuf_grab(buf, NET_BUFFERSIZE) != IO_EOF) {
        if(net_write(sd, (char *)(buf->inbuf), buf->cursize) == IO_ERROR)
            return IO_ERROR;
        n += buf->cursize;
    }
    return n;
#else
#define CHUNKSIZE (64*1024)
    /* Write out in 64K chunks */
    int buflen = buf->len;
    char *bp = (char *)buf->fp;

    if (buf->pos > 0) {
        bp += buf->pos;
        buflen -= buf->pos;
    }

    while (buflen > 0) {
        int bytes_written;
        int chunksize;

        if (buflen > CHUNKSIZE)
            chunksize = CHUNKSIZE;
        else
            chunksize = buflen;

        if((bytes_written = net_write(sd, bp, chunksize)) == IO_ERROR)
            return IO_ERROR;

        buflen -= bytes_written;
        bp += bytes_written;
            
    }
    return buf->len;
#endif
}


