/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * pblock.c: Handles Parameter Blocks
 * 
 * See pblock.h for public documentation.
 * 
 * Rob McCool
 * 
 * Note: Use property lists (plist.h) instead of pblocks.  This code has
 * been changed so that pblocks are actually property lists.
 */


#include "base/pblock.h"

#include "plist_pvt.h"
#include "base/plist.h"

#include "base/util.h"   /* util_itoa */
#ifdef MALLOC_POOLS
#include "base/pool.h"
#endif

#if defined(MALLOC_POOLS) && defined(MCC_HTTPD) && defined(THREAD_ANY)
#include "base/systhr.h"

#define MALLOC_POOL_HANDLE (thread_malloc_key != -1 ? (pool_handle_t *)systhread_getdata(thread_malloc_key) : getThreadMallocPool())

static int thread_malloc_key = -1;

static pool_handle_t *getThreadMallocPool()
{
    pool_handle_t *thread_malloc_pool = 0;

    thread_malloc_key = getThreadMallocKey();
    if (thread_malloc_key != -1) {
        thread_malloc_pool = (pool_handle_t *)systhread_getdata(thread_malloc_key);
    }

    return thread_malloc_pool;
}
#else
#define MALLOC_POOL_HANDLE ((pool_handle_t *)0)
#endif

/* ----------------------------- param_create ----------------------------- */


NSAPI_PUBLIC pb_param *param_create(char *name, char *value)
{
    PLValueStruct_t *ret;

    /*
     * Allocate a PLValueStruct_t containing the pb_param that will
     * be returned.  Normally PLValueStruct_ts are allocated from the
     * memory pool associated with a property list, but we don't have
     * that here, so we just use MALLOC().  Then we set a flag indicating
     * that the PLValueStruct_t was allocated this way.
     */
    ret = (PLValueStruct_t *)MALLOC(sizeof(PLValueStruct_t));

    ret->pv_pbentry.param = &ret->pv_pbparam;
    ret->pv_pbentry.next = 0;
    ret->pv_name = (name ? STRDUP(name) : NULL);
    ret->pv_value = (value ? STRDUP(value) : NULL);
    ret->pv_next = 0;
    ret->pv_type = 0;
    ret->pv_pi = 0;
    ret->pv_flags = PVF_MALLOC;

    return &ret->pv_pbparam;
}


/* ------------------------------ param_free ------------------------------ */


NSAPI_PUBLIC int param_free(pb_param *pp)
{
    PLValueStruct_t *pv;

    if (pp) {

        /* Get pointer to PLValueStruct_t */
        pv = PATOPV(pp);

        /* This only works if MALLOC() was used to create the structure */
        if (pv->pv_flags & PVF_MALLOC) {
            FREE(pv->pv_name);
            FREE(pv->pv_value);
            FREE(pv);
            return 1;
        }
    }
    return 0;
}


/* ---------------------------- pblock_create ----------------------------- */

NSAPI_PUBLIC pblock *pblock_create(int n)
{
    PListStruct_t *plist;       /* pointer to property list structure */
    int i;

    plist = (PListStruct_t *)PListCreate(MALLOC_POOL_HANDLE, n, 0, 0);

    plist->pl_resvpi = 0;
    
    return (plist) ? &plist->pl_pb : 0;
}

/* ----------------------------- pblock_free ------------------------------ */


NSAPI_PUBLIC void pblock_free(pblock *pb)
{
    PListStruct_t *pl = PBTOPL(pb);
    PLValueStruct_t **ppval;
    PLValueStruct_t *pv;
    int i;

#ifdef MALLOC_POOLS
    /* If the pools are enabled, this routine has no effect anyway, so 
     * just return. 
     */
    if (pool_enabled()) {
        return;
    }
#endif

    /* Free the property name symbol table if any */
    if (pl->pl_symtab) {
        pool_free(pl->pl_mempool, (void *)(pl->pl_symtab));
    }

    ppval = (PLValueStruct_t **)(pl->pl_ppval);

    /* Loop over the initialized property indices */
    for (i = 0; i < pl->pl_initpi; ++i) {

        /* Got a property here? */
        pv = ppval[i];
        if (pv) {

            param_free(&pv->pv_pbparam);
        }
    }

    /* Free the array of pointers to property values */
    pool_free(pl->pl_mempool, (void *)ppval);

    /* Free the property list head */
    pool_free(pl->pl_mempool, (void *)pl);
}


/* ---------------------------- pblock_pinsert ---------------------------- */

NSAPI_PUBLIC void pblock_pinsert(pb_param *pp, pblock *pb)
{
    PListStruct_t *pl = PBTOPL(pb);
    PLValueStruct_t **ppval;
    PLValueStruct_t *pv = PATOPV(pp);
    PLSymbolTable_t *pt;
    int cursize;
    int i;
    int wrapped;

    if (((pv->pv_flags & PVF_MALLOC) == 0) ||
        (pl->pl_mempool != MALLOC_POOL_HANDLE)) {

        /* XXX - not allowed, probably should panic */
        return;
    }

    ppval = (PLValueStruct_t **)(pl->pl_ppval);

    /*
     * Look for a free property index, starting at pl_lastpi + 1.
     * (Note that i is the property index - 1)
     */
    for (wrapped = 0, i = pl->pl_lastpi; ;) {

        /* Are we in an initialized part of the array? */
        if (i < pl->pl_initpi) {

            /* Yes, use this index if it's free */
            if (ppval[i] == 0) break;

            /* Otherwise step to the next one */
            ++i;
        }
        else {

            /* Have we reached the end yet? */
            if (i < pl->pl_cursize) {

                /*
                 * We are above the highest initialized index, but
                 * still within the allocated size.  An index in
                 * this range can be used with no further checks.
                 */
                ppval[i] = 0;
            }
            else {

                /*
                 * It's looking like time to grow the array, but
                 * first go back and look for an unused, unreserved
                 * index that might have been freed.
                 */
                if (!wrapped) {

                    i = pl->pl_resvpi;
                    wrapped = 1;
                    continue;
                }

                /*
                 * Grow the array unless there is a specified maximum
                 * size and we've reached it.
                 */
                i = pl->pl_cursize;
                if (pl->pl_maxprop && (i > pl->pl_maxprop)) {

                    /* Error - property list is full */
                    return;
                }

                /* Increase planned size of list */
                cursize = i + PLIST_DEFGROW;

                /* Reallocate the array of property value pointers */
                ppval = (PLValueStruct_t **)pool_realloc(pl->pl_mempool,
                                   (void *)ppval,
                                   (cursize * sizeof(PLValueStruct_t *)));
                if (!ppval) {

                    /* Error - insufficient memory */
                    return;
                }

                /* Initialize the first new entry and select it */
                ppval[i] = NULL;
                pl->pl_ppval = (pb_entry **)ppval;
                pl->pl_cursize = cursize;
            }

            /* Update the highest initialized index value */
            pl->pl_initpi = i + 1;
            break;
        }
    }

    /* Set the starting point for the next allocation */
    pl->pl_lastpi = i + 1;

    pv->pv_pbentry.param = &pv->pv_pbparam;
    pv->pv_pi = i + 1;
    ppval[i] = pv;

    /* Is there a hash table? */
    pt = pl->pl_symtab;

    if (!pt) {

        /* No, create one */
        pt = (PLSymbolTable_t *)pool_calloc(pl->pl_mempool, 1, PLHASHSIZE(0));
        if (!pt) {
            return;
        }

        pl->pl_symtab = pt;
    }
    else {

        /* Is it time to grow the hash table? */
        i = PLSIZENDX(pt->pt_sizendx);
        if ((pt->pt_sizendx < PLMAXSIZENDX) && pt->pt_nsyms >= (i + i)) {

            PLSymbolTable_t *npt;

            /* Yes, allocate the new table */
            npt = (PLSymbolTable_t *)pool_calloc(pl->pl_mempool, 1,
                                                 PLHASHSIZE(pt->pt_sizendx+1));
            if (npt) {
                PLValueStruct_t *opv;
                PLValueStruct_t *npv;
                int j;

                npt->pt_sizendx = pt->pt_sizendx + 1;
                npt->pt_nsyms = pt->pt_nsyms;

                /* Rehash all the names into the new table */
                for (i = 0; i < PLSIZENDX(pt->pt_sizendx); ++i) {
                    for (opv = pt->pt_hash[i]; opv; opv = npv) {
                        npv = opv->pv_next;
                        j = PListHashName(npt, opv->pv_name);
                        opv->pv_next = npt->pt_hash[j];
                        npt->pt_hash[j] = opv;
                    }
                }

                pl->pl_symtab = npt;

                /* Free the old symbol table */
                pool_free(pl->pl_mempool, (void *)pt);
                pt = npt;
            }
        }
    }

    /* Add name to symbol table */
    i = PListHashName(pt, pv->pv_name);
    pv->pv_next = pt->pt_hash[i];
    pt->pt_hash[i] = pv;
}


/* --------------------------- pblock_nvinsert ---------------------------- */


NSAPI_PUBLIC pb_param *pblock_nvinsert(const char *name,
                                       const char *value, pblock *pb)
{
    PListStruct_t *pl = PBTOPL(pb);
    PLValueStruct_t **ppval;
    PLValueStruct_t *pv;
    char *dupval;
    int pindex;

    dupval = pool_strdup(pl->pl_mempool, value);

    /* Create a new property */
    pindex = PListInitProp((PList_t)pl, 0, name, (void *)dupval, 0);
    if (pindex < 0) {
        return NULL;
    }

    ppval = (PLValueStruct_t **)(pl->pl_ppval);
    pv = ppval[pindex - 1];

    if (pl->pl_mempool == MALLOC_POOL_HANDLE) {

        pv->pv_flags |= PVF_MALLOC;
    }

    return &pv->pv_pbparam;
}

NSAPI_PUBLIC pb_param *pblock_nninsert(const char *name, int value, pblock *pb)
{
    char num[64];

    util_itoa(value, num);
    return pblock_nvinsert(name, num, pb);
}


/* ---------------------------- pblock_findval ---------------------------- */


NSAPI_PUBLIC char *pblock_findval(const char *name, pblock *pb)
{
    void *pvalue = 0;

    (void)PListFindValue((PList_t)(PBTOPL(pb)), name, &pvalue, 0);

    return (char *)pvalue;
}


/* ------------------------------ pblock_fr ------------------------------ */

NSAPI_PUBLIC pb_param *pblock_fr(const char *name, pblock *pb, int remove)
{
    PListStruct_t *pl = PBTOPL(pb);
    PLValueStruct_t **ppval;
    PLValueStruct_t **pvp;
    PLValueStruct_t *pv = NULL;
    void *pvalue = 0;
    int pindex;
    int i;

    if (pl->pl_symtab) {

        /* Compute hash of specified property name */
        i = PListHashName(pl->pl_symtab, name);

        /* Search hash collision list for matching name */
        for (pvp = &pl->pl_symtab->pt_hash[i];
             (pv = *pvp); pvp = &(*pvp)->pv_next) {

            if (!strcmp(name, pv->pv_name)) {

                if (remove) {

                    /* Remove PLValueStruct_t from symbol table */
                    *pvp = pv->pv_next;

                    /* Remove it from pl_ppval too */
                    ppval = (PLValueStruct_t **)(pl->pl_ppval);
                    pindex = pv->pv_pi;
                    ppval[pindex - 1] = 0;
                }
                break;
            }
        }
    }

    return (pv) ? &pv->pv_pbparam : NULL;
}


/* -------------------------- pblock_str2pblock --------------------------- */


int _verify_pbstr(const char *str)
{
    const char *cp;
    const char *scp;
    int np;
    int state;
    int quote;

    for(cp = str, np = 0, state = 0; *cp; ) {
        switch (state) {
        case 0:                 /* skipping leading spaces */

            while (*cp && isspace(*cp)) ++cp;
            if (*cp == '=') {
                return -1;
            }
            if (*cp) state = 1;
            break;

        case 1:                 /* scanning parameter name */

            scp = cp;
            while (*cp && (*cp != '=') && !isspace(*cp)) ++cp;
            if (*cp == '=') ++cp;
            else cp = scp;
            state = 2;
            break;

        case 2:                 /* scanning parameter value */
            quote = 0;
            if (*cp == '\"') {
                quote = 1;
                ++cp;
            }
            for (;;) {
                if (*cp == '\\') {
                    ++cp;
                    if (*cp == 0) {
                        return -1;
                    }
                }
                else if (*cp == '\"') {
                    if (!quote) {
                        return -1;
                    }
                    ++np;
                    ++cp;
                    quote = 0;
                    state = 0;
                    break;
                }
                else if (!quote && (!*cp || isspace(*cp))) {
                    ++np;
                    if (*cp) ++cp;
                    state = 0;
                    break;
                }
                else if (*cp == 0) {
                    return -1;
                }
                ++cp;
            }
            if (quote) {
                return -1;
            }
            break;
        }
    }

    return (state == 0) ? np : -1;
}

NSAPI_PUBLIC int pblock_str2pblock(const char *str, pblock *pb)
{
    char *cpy;
    char *cp;
    char *dp;
    char *pname;
    char *pvalue;
    int np;
    int quote;
    int state;
    char numbuf[16];

    if((np = _verify_pbstr(str)) < 1)
        return -1;

    while (*str && isspace(*str)) ++str;

    cpy = STRDUP(str);

    for (np = 0, cp = cpy, state = 0; *cp; ) {
        switch (state) {

        case 0:                 /* skipping leading spaces */

            while (*cp && isspace(*cp)) ++cp;
            if (*cp) state = 1;
            break;

        case 1:                 /* scanning parameter name */

            pname = cp;
            while (*cp && (*cp != '=') && !isspace(*cp)) ++cp;
            if (*cp == '=') {
                *cp++ = 0;
            }
            else {
                cp = pname;
                pname = numbuf;
                util_itoa(np+1, numbuf);
            }
            state = 2;
            break;

        case 2:                 /* scanning parameter value */
            quote = 0;
            if (*cp == '\"') {
                quote = 1;
                ++cp;
            }
            for (pvalue = cp, dp = cp; ; ++cp, ++dp) {
                if (*cp == '\\') {
                    ++cp;
                }
                else if (*cp == '\"') {
                    ++np;
                    ++cp;
                    *dp = 0;
                    quote = 0;
                    state = 0;
                    break;
                }
                else if (!quote && ((*cp == 0) || isspace(*cp))) {
                    ++np;
                    if (*cp != 0) {
                        ++cp;
                    }
                    *dp = 0;
                    state = 0;
                    break;
                }
                if (cp != dp) *dp = *cp;
            }
            pblock_nvinsert(pname, pvalue, pb);
            break;
        }
    }

    FREE(cpy);

    return np;
}


/* -------------------------- pblock_pblock2str --------------------------- */


NSAPI_PUBLIC char *pblock_pblock2str(pblock *pb, char *str)
{
    register char *s = str, *t, *u;
    PListStruct_t *pl = PBTOPL(pb);
    PLValueStruct_t **ppval;
    PLValueStruct_t *pv;
    int i;
    int sl;
    int xlen;

    ppval = (PLValueStruct_t **)(pl->pl_ppval);

    /* Loop over the initialized property indices */
    for (i = 0, xlen = 0; i < pl->pl_initpi; ++i) {

        /* Got a property here? */
        pv = ppval[i];
        if (pv && pv->pv_name) {

            int ln = strlen(pv->pv_name);
            int lv = strlen((char *)(pv->pv_value));

            /* Check for " or \ because we'll have to escape them */
            for (t = (char *)(pv->pv_value); *t; ++t) {
                if ((*t == '\"') || (*t == '\\')) ++lv;
            }

            /* 4: two quotes, =, and a null */
            xlen += (ln + lv + 4);
        }
    }

    /* Allocate string to hold parameter settings, or increase size */
    if (!s) {
        s = (char *)MALLOC(xlen);
        s[0] = '\0';
        t = &s[0];
        sl = xlen;
    }
    else {
        sl = strlen(s);
        t = &s[sl];
        sl += xlen;
        s = (char *)REALLOC(s, sl);
    }

    /* Loop over the initialized property indices */
    for (i = 0; i < pl->pl_initpi; ++i) {

        /* Got a property here? */
        pv = ppval[i];
        if (pv && pv->pv_name) {

            if (t != s) *t++ = ' ';

            for (u = pv->pv_name; *u; ) *t++ = *u++;

            *t++ = '=';
            *t++ = '\"';

            for (u = (char *)(pv->pv_value); *u; ) {
                if ((*u == '\\') || (*u == '\"')) *t++ = '\\';
                *t++ = *u++;
            }

            *t++ = '\"';
            *t = '\0';
        }
    }

    return s;
}


/* ----------------------------- pblock_copy ------------------------------ */


NSAPI_PUBLIC void pblock_copy(pblock *src, pblock *dst)
{
    PListStruct_t *pl = PBTOPL(src);
    PLValueStruct_t **ppval;
    PLValueStruct_t *pv;
    int i;

    ppval = (PLValueStruct_t **)(pl->pl_ppval);

    for (i = 0; i < pl->pl_initpi; ++i) {
        pv = ppval[i];
        if (pv) {
            pblock_nvinsert(pv->pv_name, (char *)(pv->pv_value), dst);
        }
    }
}

/* ---------------------------- pblock_dup -------------------------------- */

NSAPI_PUBLIC pblock *pblock_dup(pblock *src)
{
    pblock *dst;

    if (!src)
        return NULL;

    if ( (dst = pblock_create(src->hsize)) )
        pblock_copy(src, dst);

    return dst;
}


/* ---------------------------- pblock_pb2env ----------------------------- */


NSAPI_PUBLIC char **pblock_pb2env(pblock *pb, char **env)
{
    PListStruct_t *pl = PBTOPL(pb);
    PLValueStruct_t **ppval;
    PLValueStruct_t *pv;
    int i;
    int nval;
    int pos;

    /* Find out how many there are. */
    ppval = (PLValueStruct_t **)(pl->pl_ppval);

    for (i = 0, nval = 0; i < pl->pl_initpi; ++i) {
        if (ppval[i]) ++nval;
    }

    env = util_env_create(env, nval, &pos);

    for (i = 0; i < pl->pl_initpi; ++i) {
        pv = ppval[i];
        if (pv) {
            env[pos++] = util_env_str(pv->pv_name, (char *)(pv->pv_value));
        }
    }
    env[pos] = NULL;

    return env;
}

/*
 * Replace the value of the given <name> with the <new_value>.
 */ 
NSAPI_PUBLIC char * pblock_replace(const char *name,
                                   char * new_value, pblock *pb)
{
    PListStruct_t *pl = PBTOPL(pb);
    void *old_value = 0;
    int pindex;

    pindex = PListFindValue((PList_t)pl, name, &old_value, 0);
    if (pindex < 0) {

        return NULL;
    }

    FREE(old_value);

    pindex = PListSetValue((PList_t)pl, pindex, (const void *)new_value, 0);

    return new_value;
}
