/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * regexp.h: Wrapper for libregex
 *
 *
 *
 *
 *
 *
 * Ari Luotonen
 * Copyright (c) 1995 Netscape Communcations Corporation
 *
 */

#if defined(MCC_PROXY) && defined(USE_REGEX)

#include "base/shexp.h"
#include "base/regexp.h"
#include "base/ereport.h"

#include <sys/types.h>
#include "libregex/regex.h"

#include "base/dbtbase.h"

#define MAX_REGEX_ERR_LEN  200
#define MAX_REGEX_EXPAND   256


typedef struct _MyRegEx MyRegEx;
struct _MyRegEx {
    char *	orig;	/* original RE string as it came in */
    char *	pruned;	/* pruned RE string, but equal in meaning */
    int		case_flag;	/* case sense? */
    int		cnt;	/* # of RE's after expansion */
    char **	strs;	/* expanded RE strings */
    regex_t *	regs;	/* expanded compiled RE's */
    char *	oks;	/* RE compiled ok */
    MyRegEx *	except;	/* RE exceptions */
    MyRegEx *	next;	/* next RE in RE cache -- not part of RE */
};


/* looks up a cached RE, or compiles it and puts it into RE cache */
MyRegEx *my_regcomp_lookup(char *exp, int case_flag);


/*
 * Merges three strings together, allocates enough new memory.
 *
 *
 *
 */
static char *merge3(char *s1, char *s2, char *s3)
{
    char *buf = (char *)MALLOC((s1 ? strlen(s1) : 0) +
			       (s2 ? strlen(s2) : 0) +
			       (s3 ? strlen(s3) : 0) + 1);

    sprintf(buf, "%s%s%s",
	    s1 ? s1 : "", s2 ? s2 : "", s3 ? s3 : "");

    return buf;
}


/*
 * Frees a NULL terminated string array and its strings.
 *
 *
 */
static void free_str_array(char **arr)
{
    if (arr) {
	char **p = arr;
	while (*p) {
	    FREE(*p);
	    p++;
	}
	FREE(arr);
    }
}


/*
 * Takes an uncompiled regular expression, and expands it into multiple
 * (uncompiled) regular expressions each of which can be sequentially
 * tried on its own.
 *
 * The reason for this is that it turned out that the regex library
 * performs poorly with patterns containing (...|...|...) constructs,
 * but if those are separated into their own RE's, it works relatively
 * efficiently.
 *
 * Example (whitespace added for clarity):
 *
 *	x ( A | B ) x ( 1 | 2 | 3 ) x
 *
 * expands to:
 *
 *	x   A       x   1           x
 *	x   A       x       2       x
 *	x   A       x           3   x
 *	x       B   x   1           x
 *	x       B   x       2       x
 *	x       B   x           3   x
 *
 */
int expand_regex(char *regex, char ***expanded_arr, int max_cnt)
{
    char *p, *q, *s;
    int depth = 0;
    int parts = 0;
    int x, cnt = 0, rn = 0;
    int siz = 0;
    char *rest_str = NULL;
    char **rest = NULL;
    char **arr = NULL;

    if (!regex || !expanded_arr || max_cnt <= 1)
	return 0;

    /* find starting paren */
    for(p=regex; *p && *p != '('; p++) {
	if (*p == '\\' && *(p+1))
	    p++;
    }

    /* doesn't have (...|...) => doesn't expand */
    if (!*p)
	return 0;

    /* find out how many parts it has */
    depth = 1;
    parts = 1;
    for (q=p+1; *q && depth > 0; q++) {
	switch (*q) {
	  case '(':
	    depth++;
	    break;
	  case ')':
	    depth--;
	    break;
	  case '|':
	    if (depth == 1)
		parts++;
	    break;
	  case '\\':
	    if (*(q+1))
		q++;
	    break;
	  default:
	    ;
	}
    }

    if (parts > max_cnt)
	return 0;	/* too many parts -- don't expand */

    /* expand the rest */
    rest_str = q;
    rn = expand_regex(rest_str, &rest, max_cnt / parts);

    if (rn > 1)
	siz = rn * parts + 1;
    else
	siz = parts + 1;

    arr = (char **)MALLOC(siz * sizeof(char *));
    ZERO(arr, siz * sizeof(char *));

    depth = 1;
    cnt = 0;
    *p = '\0';
    for (q=s=p+1; *q && depth > 0; q++) {
	if (*q == '\\') {
	    if (*(q+1))
		q++;
	}
	else if (*q == '(') {
	    depth++;
	}
	else if (depth > 1 && *q == ')') {
	    depth--;
	}
	else if (depth == 1 && (*q == ')' || *q == '|')) {
	    if (*q == ')')
		depth--;
	    *q = '\0';
	    if (rn > 0) {
		for (x=0; x<rn; x++) {
		    arr[cnt++] = merge3(regex, s, rest[x]);
		}
	    }
	    else {
		arr[cnt++] = merge3(regex, s, rest_str);
	    }
	    *q = (depth > 0) ? '|' : ')';
	    s = q + 1;
	}
    }
    *p = '(';

    if (rest)
	free_str_array(rest);

    *expanded_arr = arr;
    return cnt;
}


char *prune(char *exp)
{
    int len;
    char *pruned, *p;

    if (!exp)
	return NULL;

    len = strlen(exp);
    pruned = (char *)MALLOC(len + 3);
    pruned[0] = '\0';

    /* handle empty string specially so it matches anything for sure */
    if (!*exp) {
	strcpy(pruned, ".*");
	return pruned;
    }

    if (exp[0] == '^') {
	if (exp[1] == '.' && exp[2] == '*') {
	    p = &exp[3];
	}
	else {
	    p = exp;
	}
    }
    else {
	if (exp[0] == '.' && exp[1] == '*') {
	    p = &exp[2];
	}
	else {
	    p = exp;
	    pruned[0] = '^';
	    pruned[1] = '\0';
	}
    }

    strcat(pruned, p);

    len = strlen(pruned);
    if (pruned[len-1] == '$') {
	if (pruned[len-3] == '.' && pruned[len-2] == '*') {
	    pruned[len-3] = '\0';
	}
    }
    else {
	if (pruned[len-2] == '.' && pruned[len-1] == '*') {
	    pruned[len-2] = '\0';
	}
	else {
	    pruned[len] = '$';
	    pruned[len+1] = '\0';
	}
    }

#ifdef REG_NEG_DEBUG
    printf(" | | | PRUNED: %s ---> %s\n", exp, pruned);
#endif

    return pruned;
}


char *split_except(char *exp)
{
    if (!exp)
	return NULL;

    while (*exp) {
	if (*exp == '~') {
	    *exp++ = '\0';
	    return exp;
	}
	else if (*exp == '\\') {
	    if (!(*(++exp)))
		return NULL;
	}
	exp++;
    }

    return NULL;
}


MyRegEx *my_regcomp(char *exp, int case_flag)
{
    MyRegEx *myreg = NULL;
    char **arr = NULL;
    char *copy  = NULL;
    char *copyp = NULL;
    char *pattern = NULL;
    char *except  = NULL;
    int cnt = 0;
    int err;
    int i;

#ifdef REG_NEG_DEBUG
    printf(" | | | COMPILING %s\n", exp);
#endif

    if (!exp)
	return NULL;

    /* split the pattern to the actual pattern and the exception pattern;
     * then prune it to take care of implicit rules that apply.
     */
    copy  = STRDUP(exp);
    copyp = split_except(copy);

#ifdef REG_NEG_DEBUG
    printf(" | | | PART 1 = '%s'\n", copy);
    printf(" | | | PART 2 = '%s'\n", copyp);
#endif

    pattern = prune(copy);
    if (copyp)
	except  = STRDUP(copyp);

#ifdef REG_NEG_DEBUG
    printf(" | | | pattern = '%s'\n", pattern);
    printf(" | | | except  = '%s'\n", except);
#endif

    FREE(copy);

    cnt = expand_regex(pattern, &arr, MAX_REGEX_EXPAND);
    if (cnt == 0) {
	/* fake it */
	cnt = 1;
	arr = (char **)MALLOC(2 * sizeof(char *));
	arr[0] = STRDUP(pattern);
	arr[1] = NULL;
    }

    myreg = (MyRegEx *)MALLOC(sizeof(MyRegEx));
    ZERO(myreg, sizeof(MyRegEx));

    myreg->orig = STRDUP(exp);
    myreg->pruned = pattern;
    myreg->case_flag = case_flag;
    myreg->cnt = cnt;
    myreg->strs = arr;

    myreg->regs = (regex_t *)MALLOC(cnt * sizeof(regex_t));
    ZERO(myreg->regs, cnt * sizeof(regex_t));

    myreg->oks = (char *)MALLOC(cnt * sizeof(char));
    ZERO(myreg->oks, cnt * sizeof(char));

    for (i=0; i<cnt; i++) {
	err = regcomp(&myreg->regs[i], myreg->strs[i],
		      REG_EXTENDED | REG_NOSUB | case_flag);
	if (!err) {
	    myreg->oks[i] = (char)1;
	}
	else {
	    char errbuf[MAX_REGEX_ERR_LEN + 1];

	    if (regerror(err, &myreg->regs[i], errbuf, MAX_REGEX_ERR_LEN) > 0)
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_regexErrorSRegexS_), errbuf,
			myreg->strs[i]);
	}
    }

    if (except) {
#ifdef REG_NEG_DEBUG
	printf("COMPILING EXCEPTION OUT OF %s\n", except);
#endif
	myreg->except = my_regcomp_lookup(except, case_flag);
	FREE(except);
    }

/* BEGIN DEBUG */
#ifdef REG_NEG_DEBUG
    {
	int x;

	printf("---- REGEX BEGIN ----\n");
	printf("   Orig RE: %s\n", myreg->orig);
	printf("   Pattern: %s\n", myreg->pruned);
	printf(" Case flag: %d\n", myreg->case_flag);
	printf("Variations: %d\n", myreg->cnt);
	for (x=0; x < myreg->cnt; x++) {
	    printf("Variant[%d]: %s\n", x, myreg->strs[x]);
	}
	printf("    Except: %s\n",
	       myreg->except ? myreg->except->orig : "NONE");
	printf("----- REGEX END -----\n");
    }
#endif
/* END DEBUG */

    return myreg;
}


/*
 * Uses the string length of the uncompiled RE as a hash key.
 *
 *
 */

#define REGEX_HASH_SIZE 31

static MyRegEx *re_hash_table[REGEX_HASH_SIZE];

MyRegEx *my_regcomp_lookup(char *exp, int case_flag)
{
    int hash = strlen(exp) % REGEX_HASH_SIZE;
    MyRegEx *re = re_hash_table[hash];

    while (re) {
	if (re->case_flag == case_flag && !strcmp(re->orig, exp))
	    return re;
	re = re->next;
    }

    if (!re) {
	re = my_regcomp(exp, case_flag);
	re->next = re_hash_table[hash];
	re_hash_table[hash] = re;
    }

    return re;
}


int my_regexec(char *str, MyRegEx *myreg)
{
    int i;
    int err;

    if (!str || !myreg)
	return 0;

#ifdef REG_NEG_DEBUG
    printf(" | | | matching %s with %s ~ %s\n",
	   str, myreg->pruned,
	   myreg->except ? myreg->except->pruned : "NO EXCEPTION");
#endif

    if (myreg->except && !my_regexec(str, myreg->except)) {
#ifdef REG_NEG_DEBUG
	printf(" | | | NO MATCH because matches exception: %s\n",
	       myreg->except->pruned);
#endif
	return 1;	/* no match possible -- matches exception */
    }

    for (i=0; i < myreg->cnt; i++) {
	if (myreg->oks[i] &&
	    !(err = regexec(&myreg->regs[i], str, 0, NULL, 0))) {
#ifdef REG_NEG_DEBUG
	    printf(" | | | MATCHED %s\n", myreg->strs[i]);
#endif
	    return 0;	/* match */
	}
#ifdef REG_NEG_DEBUG
	else {
	    printf(" | | | didn't match  %s\n", myreg->strs[i]);
	}
#endif
    }

#ifdef REG_NEG_DEBUG
    printf(" | | | NO MATCH, didn't match anything\n");
#endif

    return 1;	/* no match */
}


/* ----------------------------- regexp_valid ------------------------------ */


NSAPI_PUBLIC int regexp_valid(char *xp)
{
    MyRegEx *myreg = my_regcomp_lookup(xp, 0);
    int i;

    if (!myreg || myreg->cnt <= 0)
	return INVALID_REGEXP;

    for(i=0; i < myreg->cnt; i++) {
	if (!myreg->oks[i])
	    return INVALID_REGEXP;
    }

    return VALID_REGEXP;
}


/* ----------------------------- regexp_match ----------------------------- */


static int _regexp_match(char *str, char *exp, int case_flag)
{
    MyRegEx *re = my_regcomp_lookup(exp, case_flag);

    return (!re || my_regexec(str, re)) ? 1 : 0;
}


NSAPI_PUBLIC int regexp_match(char *str, char *exp)
{
    return _regexp_match(str, exp, 0);
}


/* ------------------------------ regexp_cmp ------------------------------- */


NSAPI_PUBLIC int regexp_cmp(char *str, char *exp)
{
    if (regexp_valid(exp) != VALID_REGEXP)
	return -1;

    return _regexp_match(str, exp, 0);
}


/* ---------------------------- regexp_casecmp ----------------------------- */


NSAPI_PUBLIC int regexp_casecmp(char *str, char *exp)
{
    if (regexp_valid(exp) != VALID_REGEXP)
	return -1;

    return _regexp_match(str, exp, REG_ICASE);
}


#endif	/* MCC_PROXY && USE_REGEX */

