/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * sem.c: Attempt to provide multi-process semaphores across platforms
 * 
 * Rob McCool
 */


#include "systems.h"
#include "util.h"
#include "sem.h"

#ifdef THREAD_NSPR_USER
#include "systhr.h"
#endif

#if defined(SEM_FLOCK) && defined(SEM_IRIX)
#undef SEM_FLOCK
#endif

#if defined(SEM_FLOCK)


#include "file.h"


/* ------------------------------- sem_init ------------------------------- */


SEMAPHORE sem_init(char *name, int number)
{
    char tn[256];
    SYS_FILE lfd;

    util_sprintf(tn, "/tmp/%s.%d", name, number);
    unlink(tn);

    if( (lfd = PR_Open(tn, PR_RDWR|PR_CREATE_FILE, 0644)) == NULL)
        return NULL;
    unlink(tn);
    return lfd;
}


/* ---------------------------- sem_terminate ----------------------------- */


void sem_terminate(SEMAPHORE id)
{
    PR_Close(id);
}


/* ------------------------ sem_grab, sem_release ------------------------- */


int sem_grab(SEMAPHORE id)
{
#ifdef THREAD_NSPR_USER
    /* Poll to avoid blocking. XXXrobm If errno is wrong this may go awry. */
    while(system_tlock(id) == -1)
        systhread_sleep(1000);
#else
    return system_flock(id);
#endif
}

int sem_tgrab(SEMAPHORE id)
{
    return system_tlock(id);
}

int sem_release(SEMAPHORE id)
{
    return system_ulock(id);
}


#elif defined(SEM_IRIX)

#include <unistd.h>
#include <sys/types.h>
#include <bstring.h>
#include <sys/time.h>


/* ------------------------------- sem_init ------------------------------- */


SEMAPHORE sem_init(char *name, int number)
{
    char tn[256];
    SEMAPHORE ret = (SEMAPHORE) MALLOC(sizeof(semirix_s));

    usconfig(CONF_ARENATYPE, US_SHAREDONLY);
    usconfig(CONF_INITSIZE, 64*1024);
    usconfig(CONF_AUTOGROW, 1);

    if(!(ret->arena = usinit("/dev/zero"))) {
        FREE(ret);
        return NULL;
    }
    if(!(ret->sem = usnewsema(ret->arena, 1))) {
        usdetach(ret->arena);
        FREE(ret);
        return NULL;
    }
    return ret;
}


/* ---------------------------- sem_terminate ----------------------------- */


void sem_terminate(SEMAPHORE id)
{
    usfreesema(id->sem, id->arena);
    usdetach(id->arena);
    FREE(id);
}


/* ------------------------ sem_grab, sem_release ------------------------- */


int sem_grab(SEMAPHORE id)
{
    uspsema(id->sem);
    return 0;
}

int sem_tgrab(SEMAPHORE id)
{
    return uscpsema(id->sem) ? 0 : -1;
}

int sem_release(SEMAPHORE id)
{
    usvsema(id->sem);
}


#elif defined(SEM_WIN32)


/* ------------------------------- sem_init ------------------------------- */


NSAPI_PUBLIC SEMAPHORE sem_init(char *name, int number)
{
	return CreateSemaphore(NULL, 1, SEM_MAXVALUE, name);
}


/* ---------------------------- sem_terminate ----------------------------- */


NSAPI_PUBLIC void sem_terminate(SEMAPHORE id)
{
	CloseHandle(id);
}


/* ------------------------ sem_grab, sem_release ------------------------- */


NSAPI_PUBLIC int sem_grab(SEMAPHORE id)
{
    if(WaitForSingleObject(id, INFINITE) == WAIT_FAILED)
        return -1;
    else
	    return 0;
}

/* sem_tgrab is supposed to prevent a block on the acquisition of
 * the semaphore. NT uses kernel threads and so the block doesn't
 * matter */

NSAPI_PUBLIC int sem_tgrab(SEMAPHORE id)
{
	if(WaitForSingleObject(id, INFINITE) == WAIT_FAILED)
        return -1;
    else
	    return 0;
}

NSAPI_PUBLIC int sem_release(SEMAPHORE id)
{
    if(!(ReleaseSemaphore(id, 1, NULL)))
        return -1;
    else
        return 0;
}

#else
#error "undefined semaphore type"
#endif
