/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * session.c: Deal with virtual sessions
 * 
 * Rob McCool
 */


/* Removed for ns security integration
#include "sslio/sslio.h"
*/

#ifdef XP_UNIX
#include "base/systems.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <netinet/in.h> /* ntohl */
#include <arpa/inet.h>  /* inet_ntoa */
#include <netdb.h>      /* hostent stuff */
#endif /* XP_UNIX */

#include "nspr.h"
#include "private/pprio.h"
extern "C" {
#include <cert.h>
#include <ssl.h>
#include <key.h>
}
#include <base/daemon.h>	/* child_status */
#include <base/pool.h>
#include <base/systhr.h>
#include <base/plist.h>
#include <frame/conf.h>

#include "libaccess/nsauth.h"
#include "libaccess/nsuser.h"
#include "libaccess/nsgroup.h"

#include "netio.h"
#include "session.h"


/* ---------------------------- session_create ---------------------------- */

Session *session_alloc(SYS_NETFD csd, struct sockaddr_in *sac)
{
    Session *ns = (Session *)MALLOC(sizeof(Session));
	ns->csd = csd;
    ns->iaddr = sac->sin_addr;
	ns->csd_open = 1;
	ns->fill = 0;

    ns->subject = NULL;
	return ns;
}

NSAPI_PUBLIC Session *session_fill(Session *ns) 
{
#if defined(MALLOC_POOLS) && defined(THREAD_ANY)
    pool_handle_t *pool = pool_create();
#endif
    ClAuth_t *cla;
#ifdef NET_SSL
    PRInt32 secon = -1;
    PRInt32 keySize, secretKeySize;
    char *cipher;
    char *issuer_dn;
    char *user_dn;
#endif

#if defined(MALLOC_POOLS) && defined(THREAD_ANY) 
    ns->pool = pool;
    systhread_setdata(getThreadMallocKey(), ns->pool);
#endif

    ns->client = pblock_create(SESSION_HASHSIZE);
    ns->inbuf = netbuf_open(ns->csd,NET_BUFFERSIZE);
	ns->fill = 1;

#ifdef MCC_PROXY
    ns->req_cnt = 0;
#endif

    pblock_nvinsert("ip",inet_ntoa(ns->iaddr),ns->client);

    /* Initialize ACL authentication information */
    cla =  (ClAuth_t *) MALLOC(sizeof(ClAuth_t));
    ns->clauth = (void *) cla;
    cla->cla_realm = 0;
    cla->cla_ipaddr = (IPAddr_t)ntohl(ns->iaddr.s_addr);
    cla->cla_dns = 0;
    cla->cla_uoptr = 0;
    cla->cla_goptr = 0;
#ifdef CLIENT_AUTH
    cla->cla_cert = 0;
#endif /* CLIENT_AUTH */

#ifdef NET_SSL
    /* As a hack use the non-null presence of security_status to 
     * indicate that security is on...  XXXMB
     */
    if (security_active) {
	if(!SSL_SecurityStatus(ns->csd, &secon, &cipher, &keySize, 
			       &secretKeySize, &issuer_dn, &user_dn)) {
            if(secon > 0)  {
                pblock_nvinsert("cipher", cipher, ns->client);
                if (cipher)
                    PR_Free(cipher);
                pblock_nninsert("keysize", keySize, ns->client); 
                pblock_nninsert("secret-keysize", secretKeySize, ns->client); 
                pblock_nvinsert("issuer_dn", issuer_dn, ns->client); 
		if (issuer_dn)
		    PR_Free(issuer_dn);
                pblock_nvinsert("user_dn", user_dn, ns->client); 
		if (user_dn)
		    PR_Free(user_dn);
        	}
        }
    }
#endif
    return ns;
}

NSAPI_PUBLIC Session *session_create(SYS_NETFD csd, struct sockaddr_in *sac) 
{
	Session *sn = session_alloc(csd, sac);
	return session_fill(sn);
}

/* ----------------------------- session_free ----------------------------- */


NSAPI_PUBLIC void session_free(Session *sn)
{
#if defined(MALLOC_POOLS) && defined(THREAD_ANY) 
    pool_handle_t *pool = sn->pool;
#endif
    ClAuth_t *cla = (ClAuth_t *) sn->clauth;

    pblock_free(sn->client);
    if (sn->subject)
	PListDestroy(sn->subject);
    netbuf_close(sn->inbuf);

    if (cla->cla_uoptr != 0) {
	userFree(cla->cla_uoptr);
    }
    if (cla->cla_goptr != 0) {
	groupFree(cla->cla_goptr);
    }
#ifdef CLIENT_AUTH
    if (cla->cla_cert != 0) {
	CERT_DestroyCertificate((CERTCertificate*)cla->cla_cert);
    }
#endif /* CLIENT_AUTH */
    /* don't need to free realm */
    FREE(cla);
#if defined(MALLOC_POOLS) && defined(THREAD_ANY) 
    systhread_setdata(getThreadMallocKey(), NULL);
    pool_destroy(sn->pool);
#endif /* MALLOC_POOLS */
    FREE(sn);
}

/* ----------------------------- session_dns ------------------------------ */


#include "net.h"


NSAPI_PUBLIC char *session_dns_lookup(Session *s, int verify)
{
    pb_param *dns = pblock_find("dns", s->client);
    char *hn;

    if(!dns) {
        if(!(hn = net_ip2host(pblock_findval("ip", s->client), verify))) {
            pblock_nvinsert("dns", "-none", s->client);
            return NULL;
        }
        dns = pblock_nvinsert("dns", hn, s->client);
        FREE(hn);
    }
    else if(!strcmp(dns->value, "-none"))
        return NULL;
    ((ClAuth_t *) s->clauth)->cla_dns = dns->value;
    return dns->value;
}

