/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
   This code is assumed to be #included from daemon.c. It will not compile
   on its own. It is here to avoid cluttering up daemon.c, and is not stand
   alone so that it can access static vars in daemon.c.

   Entry points are conn_add_wait and conn_get_session. Assumes a threaded
   server.

   Rob McCool
 */

#include "base/systems.h"
#include "libaccess/nsauth.h"   /* ClAuth_t */
#include "libaccess/nsgroup.h"
#include "libaccess/nsuser.h"

static PR_fd_set connFDs;
static CRITICAL conn_crit;
struct conn_s {
    Session *sn;
    time_t to;  /* seconds past epoch when this connection should be flushed */
    struct conn_s *next;
};

/* The free list keeps these connection structs around so we don't thrash 
   malloc. */

/* conn_num_ready may only be safely used by the thread that holds 
   accept_crit. The other variables may only be safely used by the thread 
   that holds conn_crit. */

/* These variables are assumed to be initialized below. */
static struct conn_s *conn_s_free = NULL;
static struct conn_s *conn_s_wait = NULL;
static struct conn_s *conn_s_wait_end = NULL;
static int conn_num_ready = 0;
static int conn_num_wait = 0, conn_blocked = 0;

static int conn_wakeup[2];

#ifdef CONN_DEBUG
static int conn_debug = 0;
#include <stdarg.h>
#define CDEBUG(arg) if(conn_debug) conn_dprintf arg;
void conn_dprintf(char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    vfprintf(stderr, fmt, args);
}
#else
#define CDEBUG(arg) (void)0
#endif


static void conn_init(void)
{
    int rv, flgs;

    conn_crit = crit_init();

    PR_FD_ZERO(&connFDs);
    /* conn_num_ready, wait, conn_s_wait, free assumed to be initialized */

#ifdef USE_PIPE
    rv = pipe(conn_wakeup);
#else
    rv = socketpair(AF_UNIX, SOCK_STREAM, 0, conn_wakeup);
#endif

    /* Set non-blocking */
    if( (rv < 0) || ((flgs = fcntl(conn_wakeup[0], F_GETFL, 0)) == -1) ) {
        ereport(LOG_FAILURE, XP_GetAdminStr(DBT_canTCreateIpcPipeS_), system_errmsg());
        exit(1);
    }
    fcntl(conn_wakeup[0], F_SETFL, flgs | O_NONBLOCK);
#ifdef CONN_DEBUG
    if(getenv("NETSITE_DEBUG_CONN"))
        conn_debug = 1;
#endif
}

/* Assumes you've already locked the connection mutex */
static struct conn_s *conn_alloc(Session *sn, time_t to)
{
    struct conn_s *cs = conn_s_free;

    if(cs) {
        conn_s_free = cs->next;
    }
    else {
        cs = (struct conn_s *) PERM_MALLOC(sizeof(struct conn_s));
    }
    cs->sn = sn;
    cs->to = to;
    cs->next = NULL;
    return cs;
}

/* Assumes you've already locked the connection mutex */
static void conn_remove_wait(struct conn_s *csp, struct conn_s *csv)
{
    CDEBUG(("before: wait is 0x%x, wait_end is 0x%x, %d waiters\n", 
            conn_s_wait, conn_s_wait_end, conn_num_wait));
    if(!csv)
        conn_s_wait = csp->next;
    else
        csv->next = csp->next;
    /* If this was the end of the list, update it. */
    if(!csp->next) conn_s_wait_end = csv;
    --conn_num_wait;
    CDEBUG(("after: wait is 0x%x, wait_end is 0x%x, %d waiters\n", 
            conn_s_wait, conn_s_wait_end, conn_num_wait));
}


/*
   Add a connection to the queue of waiting connections, with the current 
   time as returned by time(NULL) in now.
 */
static void conn_add_wait(Session *sn, time_t now)
{
    struct conn_s *cs;

    crit_enter(conn_crit);
    cs = conn_alloc(sn, sn->inbuf->rdtimeout + now);
    /* cs->next is NULL */
    CDEBUG(("before: wait is 0x%x, wait_end is 0x%x, %d waiters\n", 
            conn_s_wait, conn_s_wait_end, conn_num_wait));

    if(conn_s_wait)
        conn_s_wait_end->next = cs;
    else
        conn_s_wait = cs;
    conn_s_wait_end = cs;
    ++conn_num_wait;
    CDEBUG(("after: wait is 0x%x, wait_end is 0x%x, %d waiters\n", 
            conn_s_wait, conn_s_wait_end, conn_num_wait));
    if(conn_blocked) {
        CDEBUG(("sending wakeup byte to blocked thread\n"));
        if(write(conn_wakeup[1], "a", 1) == -1) {
            ereport(LOG_FAILURE, XP_GetAdminStr(DBT_writeToWakeupPipeFailedS_), 
                    system_errmsg());
        }
    }
    crit_exit(conn_crit);
}


/* Assumes you've already locked the connection mutex */
static void conn_free(struct conn_s *cs)
{
    cs->sn = NULL;
    cs->next = conn_s_free;
    conn_s_free = cs;
}

/* 
   Flush out idle connections. Because list is sorted with 
   shortest timeouts first, we can stop once we're past now. This will also
   dump older connections that haven't timed out yet when we don't have the
   threads to handle them.
 */
/* Assumes you've already locked the connection mutex */
static void conn_flush(int *conn_num_ready)
{
    struct conn_s *csp;
    int numdump;
    time_t now;
#ifdef MALLOC_POOLS
    pool_handle_t *pool = (pool_handle_t *)systhread_getdata(getThreadMallocKey());
#endif

    numdump = (conn_num_wait > max_ka_threads ? conn_num_wait - max_ka_threads : 0);

#if 0
    ereport(LOG_INFORM, XP_GetAdminStr(DBT_flushingDConnectionsCurrentDTotD_),
		numdump, conn_num_wait, max_ka_threads);
#endif

    if( (csp = conn_s_wait) )
        now = time(NULL);
    while(csp) {
        if(numdump || (csp->to <= now)) {
            CDEBUG(("flushing session 0x%x, numdump %d\n", csp, numdump));
            if(PR_FD_ISSET(csp->sn->csd, &connFDs)) {
                --(*conn_num_ready);
                CDEBUG(("session 0x%x has data, num_ready is now %d\n", 
                        csp->sn, *(conn_num_ready)));
            }
            conn_remove_wait(csp, NULL);
#ifdef MALLOC_POOLS
            systhread_setdata(getThreadMallocKey(), csp->sn->pool);
#endif
            net_close(csp->sn->csd);
            session_free(csp->sn);
            csp->next = conn_s_free;
            conn_s_free = csp;

            if(numdump) --numdump;
        }
        else 
            break;
        csp = conn_s_wait;
    }
#ifdef MALLOC_POOLS
    systhread_setdata(getThreadMallocKey(), pool);
#endif
}


/* 
   Find a ready connection. This gives priority to existing 
   connections over new ones, to avoid over-accepting our
   capacity and tries make sure that if people see delays in
   service, the smallest set of people see it.
 */

static Session *conn_find_ready(int *conn_num_ready)
{
    struct conn_s *csp, *csv;
    Session *sn = NULL;
    ClAuth_t *cla;

    for(csp = conn_s_wait, csv = NULL; csp; csp = csp->next) {
        if(PR_FD_ISSET(csp->sn->csd, &connFDs)) {
            sn = csp->sn;
            CDEBUG(("handling connection 0x%x\n", csp));
#if defined(MALLOC_POOLS)
            systhread_setdata(getThreadMallocKey(), sn->pool);
#endif
            /* Reset the user information so that the new request gets 
               auth-type and auth-user */
            cla = (ClAuth_t *) sn->clauth;
            if(cla->cla_uoptr)
                userFree(cla->cla_uoptr);
            if(cla->cla_goptr)
                groupFree(cla->cla_goptr);
#ifdef CLIENT_AUTH
	    if(cla->cla_cert) {
		SEC_DestroyCertificate(cla->cla_cert);
	    }
#endif /* CLIENT_AUTH */

            /* don't need to free realm */

            cla->cla_realm = 0;
            cla->cla_uoptr = 0;
            cla->cla_goptr = 0;
#ifdef CLIENT_AUTH
	    cla->cla_cert = 0;
#endif /* CLIENT_AUTH */

            --(*conn_num_ready);
            conn_remove_wait(csp, csv);
            break;
        }
        csv = csp;
    }
    return sn;
}


/*
   Select over the "wakeup" pipe, the bound socket, and any idle sockets. 
   Flush any idle sockets that are idle for too long. Returns when there are
   sockets besides the wakeup pipe ready.
 */
static Session *conn_do_select(int *conn_num_ready)
{
    time_t now, mintimeout;
    struct conn_s *csp;
    struct timeval tv, *tvp;
    int maxfd;
    Session *sn;

	crit_enter(conn_crit);
	if (*conn_num_ready) {
		sn = conn_find_ready(conn_num_ready);
		crit_exit(conn_crit);
		return sn;
	}

    *conn_num_ready = 0;
    while(!(*conn_num_ready)) {
        PR_FD_ZERO(&connFDs);
        PR_FD_NSET(conn_wakeup[0], &connFDs);
        tvp = NULL;
        tv.tv_usec = 0;
        maxfd = conn_wakeup[0];
        CDEBUG(("maxfd is %d\n", maxfd));
        for(csp = conn_s_wait; csp; csp = csp->next) {
            PR_FD_SET(csp->sn->csd, &connFDs);
            if(!tvp) {
                tvp = &tv;
                tv.tv_sec = csp->to;
                mintimeout = csp->to;
                CDEBUG(("setting new timeout to %d\n", csp->to));
            }
            else if(csp->to < mintimeout) {
                mintimeout = csp->to;
                CDEBUG(("setting timeout to %d\n", csp->to));
            }
        }
        if(tvp) {
            now = time(NULL);
            if(mintimeout < now) {
                CDEBUG(("correcting timeout %d to %d\n", mintimeout, now));
                mintimeout = now;
            }
            tv.tv_sec = mintimeout - now;
        }
        /* Show that we're going to block, and then release the connection 
           crit so others can add waiters. */
        conn_blocked = 1;
        crit_exit(conn_crit);

        CDEBUG(("entering select maxfd=%d timeout=%d\n", maxfd, 
                (tvp ? tv.tv_sec : -1) ));
        *conn_num_ready = PR_Select(0, &connFDs, NULL, NULL, (tvp ? PR_SecondsToInterval(tvp->tv_sec) : PR_INTERVAL_NO_TIMEOUT));

        crit_enter(conn_crit);
        conn_blocked = 0;

        /* Flush out timed out connections, even if they have data. The
           code to check for ready connections won't notice set FDs that 
           aren't in the wait set anyway.
         */
        conn_flush(conn_num_ready);

        if(*conn_num_ready == -1) {
            CDEBUG(("select failed (%s)\n", system_errmsg()));
            *conn_num_ready = 0;
            continue;
        }
        /* 
           Assumes that conn_wakeup[0] is non-blocking. Pipe semantics 
           probably allow us to avoid the while loop and assume we got the
           whole batch on the first shot, but let's be safe...
         */
        if(PR_FD_NISSET(conn_wakeup[0], &connFDs)) {
            char garbage[1024];
            while(read(conn_wakeup[0], garbage, sizeof(garbage)) == sizeof(garbage)) {
                CDEBUG(("read garbage\n"));
            }
            --(*conn_num_ready);
        }
    }
    CDEBUG(("finished select with %d fds\n", *conn_num_ready));

    sn = conn_find_ready(conn_num_ready);
    crit_exit(conn_crit);
    return sn;
}


/*
   Just does the rigamorale of accepting a new connection. Because of 
   user-level threading, this assumes that the bound socket has a new 
   connection ready.
 */
static Session *conn_accept_session(worker_data_s *wd)
{
    int clen;

    clen = sizeof(wd->sa_client);
    CDEBUG(("entering accept\n"));
    wd->csd = net_accept(sd, (struct sockaddr *) &wd->sa_client, &clen);
    CDEBUG(("exited accept\n"));
    /* 
       Done accepting. Let someone else have fun.
     */
    if(wd->csd == SYS_NET_ERRORFD) {
        if (errno != EAGAIN)
            ereport(LOG_WARN, XP_GetAdminStr(DBT_acceptFailedS_1), system_errmsg());
        return NULL;
    }
    else
        return session_alloc(wd->csd, &wd->sa_client);
}

