/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/* time_cache.c
 *
 * Attempts to minimize CPU overhead with respect to time functions 
 * in the http server by being more intelligent about possible times.
 *
 * optimization 1:
 *	For use with convert the current time to a string format..  Assumes 
 *	that there is a high likelihood that the current time will be similar 
 *	to the "last" current time requested.  i.e.  most of the work computing
 *	the time will be duplicated.  Use this assumption to only calculate
 *	the changed portions of the timestamp.
 *
 * optimization 1a: (added 5-15-97)
 *      The same as optimization 1, except do it for the common log time fmt.
 *
 * optimization 2:
 *	assumes that the server is serving the same (mostly unchanged) 
 *	documents over and over again.  As such, why recompute the string
 *	timestamp with every request?  Instead why not just cache the 
 *	last 128 or 512 or whatever timestamps and not recompute them.
 *	For 512 cached stamps, will use about 25K memory.
 *
 *
 * Note:  Since the accelerator cache went into place, the timecache rarely
 *	  gets used.  I no longer believe that the optimization #2 listed
 *	  above is useful.  Optimization 1 still helps a bit, but #2 is just
 *	  a waste of memory-  MB 1-15-96
 *
 * Mike Belshe
 * 9-17-95
 */



#include <stdio.h>
#include <sys/types.h>
#include "netsite.h"
#include "base/time_cache.h"
#include "base/ereport.h"
#include "base/util.h"
#include "base/daemon.h"
#include "base/dbtbase.h"

NSAPI_PUBLIC void time_cache_destroy(void *);


/* time_cache_curr_strftime()
 * Try to convert time to string faster by using smarts.
 * Note: this should only be used for converting "current" time.
 *       It is optimized to base its time on the last get_time.
 */
	/* NOTE NOTE NOTE!
	 * use fixed length TIME FMT or we can't find our timestamp easily!
	 * This is RFC1123 date format instead of RFC 850
	 */
#ifdef XP_UNIX
#define TIME_FMT	"%a, %d %b %Y %T GMT"
#define TIME_MAX_LEN	30
#define TIME_HOUR_POS	17
#define TIME_MIN_POS	20
#define TIME_SEC_POS	23
#else /* XP_WIN32 */
#define TIME_FMT	"%a, %d %b %Y %H:%M:%S GMT"
#define TIME_MAX_LEN	30
#define TIME_HOUR_POS	17
#define TIME_MIN_POS	20
#define TIME_SEC_POS	23
#endif

NSAPI_PUBLIC char *
time_cache_curr_strftime(time_t new_time)
{
	static char last_time_str[TIME_MAX_LEN+1];
	static time_t last_time = (time_t)-1;	
	static int last_sec;		
	static int last_min;
	static int last_hour;
	static time_t last_time_eod = 0;	/* last_time end of day */
	struct tm *gmt, gmts;
	int diff_time;
	int diff_sec;
	int diff_min;
	int diff_hour;

	if ( new_time == (time_t)0 ) {
		new_time = time(NULL);
	}

	/* The quickest server can do multiple requests / sec */
	if (new_time == last_time)
		return STRDUP(last_time_str);

	/* If new_time is prior to last_time, or we have a new
	 * day, just recompute using the strftime() function
	 */
	if ((last_time == (time_t)-1) ||
            (new_time < last_time) ||
	    (new_time >= last_time_eod) ) {
		gmt = system_gmtime(&new_time, &gmts);
		util_strftime(last_time_str, TIME_FMT, gmt);
                last_time_str[TIME_SIZE] = '\0';
		last_time = new_time;
		last_time_eod = (new_time / 86400 + 1) * 86400;
		last_sec = new_time % 60;
		last_min = (new_time / 60) % 60;
		last_hour = (new_time / 3600) % 24;
		return STRDUP(last_time_str);
	}	

	/* The day has not changed;  calculate new hours, mins, secs
	 * and just change the string.  Should be much faster than
	 * recomputing the whole thing; especially since we are 
	 * probably within SECONDS of the old time.
	 */

	diff_time = new_time - last_time;
	diff_sec = diff_time %60;

	last_sec += diff_sec;
	if (last_sec >= 60) {
		last_sec -= 60;
		diff_time += 60;
	}

	last_time_str[TIME_SEC_POS] = last_sec / 10 + '0';
	last_time_str[TIME_SEC_POS+1] = last_sec % 10 + '0';

	if (diff_time >= 60 ) {
		diff_min = ((diff_time = diff_time / 60)) % 60;
		
		last_min += diff_min;
		if (last_min >= 60) {
			last_min -= 60;
			diff_time += 60;
		}
	
		last_time_str[TIME_MIN_POS] = last_min / 10 + '0';
		last_time_str[TIME_MIN_POS+1] = last_min % 10 + '0';

		if (diff_time >= 60)  {
			diff_hour = diff_time / 60;

			last_hour += diff_hour;
			last_time_str[TIME_HOUR_POS] = last_hour / 10 + '0';
			last_time_str[TIME_HOUR_POS+1] = last_hour % 10 + '0';
		}
	}
	last_time = new_time;
        /* Caller must free it */
	return STRDUP(last_time_str);
}


/* time_cache_curr_strftime_logfmt()
 * Try to convert time to string faster by using smarts.
 * Note: this should only be used for converting "current" time.
 *       It is optimized to base its time on the last get_time.
 */
	/* NOTE NOTE NOTE!
	 * use fixed length TIME FMT or we can't find our timestamp easily!
	 * This is RFC1123 date format instead of RFC 850
	 */
#define TIMELOG_FMT	"%d/%b/%Y:%H:%M:%S"
#define TIMELOG_MAX_LEN	    21
#define TIMELOG_HOUR_POS    12
#define TIMELOG_MIN_POS	    15
#define TIMELOG_SEC_POS	    18
#define TIMELOG_SIZE        20

NSAPI_PUBLIC char *
time_cache_curr_strftime_logfmt(time_t new_time)
{
	static char last_time_str[TIMELOG_MAX_LEN+1];
	static time_t last_time = (time_t)-1;	
	static int last_sec;		
	static int last_min;
	static int last_hour;
	static time_t last_time_eod = 0;	/* last_time end of day */
	struct tm *lt, lts;
	int diff_time;
	int diff_sec;
	int diff_min;
	int diff_hour;

	if ( new_time == (time_t)0 ) {
		new_time = time(NULL);
	}

	/* The quickest server can do multiple requests / sec */
	if (new_time == last_time)
		return STRDUP(last_time_str);

	/* If new_time is prior to last_time, or we have a new
	 * day, just recompute using the strftime() function
	 */
	if ((last_time == (time_t)-1) ||
            (new_time < last_time) ||
	    (new_time >= last_time_eod) ) {
        long offset;

		lt = system_localtime(&new_time, &lts);
#ifdef BSD_TIME
        offset = lts.tm_gmtoff;
#else
        offset = - timezone;
        if(lts.tm_isdst)
            offset += 3600;
#endif
		util_strftime(last_time_str, TIMELOG_FMT, lt);
                last_time_str[TIMELOG_SIZE] = '\0';
		last_time = new_time - offset;
		last_time_eod = (new_time / 86400 + 1) * 86400;
		last_sec = new_time % 60;
		last_min = (new_time / 60) % 60;
		last_hour = (new_time / 3600) % 24;
		return STRDUP(last_time_str);
	}	

	/* The day has not changed;  calculate new hours, mins, secs
	 * and just change the string.  Should be much faster than
	 * recomputing the whole thing; especially since we are 
	 * probably within SECONDS of the old time.
	 */

	diff_time = new_time - last_time;
	diff_sec = diff_time %60;

	last_sec += diff_sec;
	if (last_sec >= 60) {
		last_sec -= 60;
		diff_time += 60;
	}

	last_time_str[TIMELOG_SEC_POS] = last_sec / 10 + '0';
	last_time_str[TIMELOG_SEC_POS+1] = last_sec % 10 + '0';

	if (diff_time >= 60 ) {
		diff_min = ((diff_time = diff_time / 60)) % 60;
		
		last_min += diff_min;
		if (last_min >= 60) {
			last_min -= 60;
			diff_time += 60;
		}
	
		last_time_str[TIMELOG_MIN_POS] = last_min / 10 + '0';
		last_time_str[TIMELOG_MIN_POS+1] = last_min % 10 + '0';

		if (diff_time >= 60)  {
			diff_hour = diff_time / 60;

			last_hour += diff_hour;
			last_time_str[TIMELOG_HOUR_POS] = last_hour / 10 + '0';
			last_time_str[TIMELOG_HOUR_POS+1] = last_hour % 10 + '0';
		}
	}
	last_time = new_time;
        /* Caller must free it */
	return STRDUP(last_time_str);
}

int
time_cache_curr_gmtime(struct tm *lt, time_t new_time)
{
	static struct tm last_time_tm;
	static time_t last_time = (time_t)-1;
	static time_t last_time_eod = 0;
	static int last_sec;
	static int last_min;
	static int last_hour;
	int diff_time;
	int diff_sec;
	int diff_min;
	int diff_hour;

	if (new_time == (time_t)0) {
		new_time = time(NULL);
	}

	if (new_time == last_time)
		memcpy((char *)lt, (char *)&last_time_tm, sizeof(struct tm));

	if ( (last_time == (time_t)-1) ||
		 (new_time < last_time) ||
		 (new_time > last_time_eod) ) {
	
		system_gmtime(&new_time, &last_time_tm);
		last_time = new_time;
		last_time_eod = (new_time / 86400 + 1) * 86400;
		last_sec = new_time % 60;
		last_min = (new_time / 60) % 60;
		last_hour = (new_time / 3600) % 24;
		return memcpy((char *)lt, (char *)&last_time_tm, sizeof(struct tm))?0:1;
	}

	diff_time = new_time - last_time;

	/* --- quick check seconds --- */
	if (diff_time < 60) 
		goto seconds;

	/* --- hours --- */
	for (diff_hour = 0; diff_time > 3600; diff_time -=3600, diff_hour++);
	last_hour += diff_hour;

	/* --- minutes --- */
	for (diff_min =0; diff_time > 60; diff_time -=60, diff_min++);
	last_min += diff_min;
	if (last_min >= 60) {
		last_min -= 60;
		last_hour++;
	}

seconds:
	/* --- seconds --- */
	last_sec += diff_time;
	if (last_sec >= 60) {
		last_sec -= 60;
		last_min++;
		if (last_min >= 60) {
			last_min -= 60;
			last_hour++;
		}
	}

	last_time_tm.tm_hour = last_hour;
	last_time_tm.tm_min = last_min;
	last_time_tm.tm_sec = last_sec;
	last_time = new_time;

	return memcpy((char *)lt, (char *)&last_time_tm, sizeof(struct tm))?0:1;
}

int
time_cache_curr_localtime(struct tm *lt, time_t new_time)
{
	static struct tm last_time_tm;
	static time_t last_time = (time_t)-1;
	static time_t last_time_eod = 0;
	static int last_sec;
	static int last_min;
	static int last_hour;
	int diff_time;
	int diff_sec;
	int diff_min;
	int diff_hour;

	if (new_time == (time_t)0) {
		new_time = time(NULL);
	}

	if (new_time == last_time)
		memcpy((char *)lt, (char *)&last_time_tm, sizeof(struct tm));

	if ( (last_time == (time_t)-1) ||
		 (new_time < last_time) ||
		 (new_time > last_time_eod) ) {
		long offset;

		system_localtime(&new_time, &last_time_tm);

#ifdef BSD_TIME
		offset = last_time_tm.tm_gmtoff;
#else
		offset = - timezone;
		if(last_time_tm.tm_isdst)
			offset += 3600;
#endif

		last_time = new_time;
		last_time_eod = (new_time / 86400 + 1) * 86400;

		new_time = new_time + offset;
		last_sec = new_time % 60;
		last_min = (new_time / 60) % 60;
		last_hour = (new_time / 3600) % 24;
		return memcpy((char *)lt, (char *)&last_time_tm, sizeof(struct tm))?0:1;
	}

	diff_time = new_time - last_time;

	/* --- quick check seconds --- */
	if (diff_time < 60) 
		goto seconds;

	/* --- hours --- */
	for (diff_hour = 0; diff_time > 3600; diff_time -=3600, diff_hour++);
	last_hour += diff_hour;

	/* --- minutes --- */
	for (diff_min =0; diff_time > 60; diff_time -=60, diff_min++);
	last_min += diff_min;
	if (last_min >= 60) {
		last_min -= 60;
		last_hour++;
	}

seconds:
	/* --- seconds --- */
	last_sec += diff_time;
	if (last_sec >= 60) {
		last_sec -= 60;
		last_min++;
		if (last_min >= 60) {
			last_min -= 60;
			last_hour++;
		}
	}

	last_time_tm.tm_hour = last_hour;
	last_time_tm.tm_min = last_min;
	last_time_tm.tm_sec = last_sec;
	last_time = new_time;

	return memcpy((char *)lt, (char *)&last_time_tm, sizeof(struct tm))?0:1;
}



/* Make these configurable? */
#define DEFAULT_HASH_SIZE	256
#define ABSOLUTE_HASH_MIN	32
#define ABSOLUTE_HASH_MAX	8192
#define DEFAULT_CACHE_SIZE	256
#define ABSOLUTE_CACHE_MIN	32
#define ABSOLUTE_CACHE_MAX	8192

static int time_cache_size = DEFAULT_CACHE_SIZE;
static int time_hash_size = DEFAULT_HASH_SIZE;
static int time_cache_enabled = 1;


/* ----- Forwards ----------------------------------------------------- */
static unsigned int time_cache_hash_time(unsigned int, void *);
static int time_cache_compare_keys(void *, void *);
static int time_cache_cleanup(void *);
static int time_cache_debug(pblock *pb, Session *sn, Request *rq);
static int time_cache_print(void *data, SYS_NETFD fd);

/* ----- Static globals ----------------------------------------------- */
static cache_t	*time_cache = NULL;
static public_cache_functions_t time_cache_functions = {
	time_cache_hash_time,
	time_cache_debug
};

static cache_entry_functions_t time_cache_entry_functions = {
	time_cache_compare_keys, 
	time_cache_cleanup,
	time_cache_print 
};

/* ----- Functions ---------------------------------------------------- */
static unsigned int 
time_cache_hash_time(unsigned int hash_size, void *voidtime)
{
	time_t *time = (time_t *)voidtime;
	return (*time % hash_size);
}

static int
time_cache_compare_keys(void *voidtime1, void *voidtime2)
{
	return ( !(*((time_t *)voidtime1) == *((time_t *)voidtime2)));
}

static int
time_cache_cleanup(void *voiddata)
{
	time_cache_entry_t *data = (time_cache_entry_t *)voiddata;
	
	PERM_FREE(data->time_str);

	/* data gets freed with cache_entry - is that ugly? XXXMB */

	return 0;
}

static 
int time_cache_debug(pblock *pb, Session *sn, Request *rq)
{
	return cache_dump(time_cache, "Time Cache", sn->csd);
}

#define MAX_DEBUG_LINE 1024
static int
time_cache_print(void *voiddata, SYS_NETFD fd)
{
#if 0 /* XXXrobm net_write causes ball of string with admserv CGIs */
	time_cache_entry_t *data = (time_cache_entry_t *)voiddata;
	char buf[MAX_DEBUG_LINE];
	int len;
	
	len = util_sprintf(buf, "%d, %s", data->time_val, data->time_str);
	net_write(fd, buf, len);
#endif
	return 0;
}

/* time_cache_internal_init()
 * time_cache_init()
 *
 * The time cache has two initialization functions.  The first is a public 
 * function which can be inserted in obj.conf.  This allows the user to 
 * configure the attributes of the time cache (hash-size, cache-size, 
 * disable, etc).  The second is the internal initializer which actually
 * does the real work- based on the user configuration.
 */
int
time_cache_internal_init()
{
	if (time_cache_enabled) {
		if ( (time_cache = cache_create(time_cache_size, 
			time_hash_size, &time_cache_functions)) == NULL) {
			ereport(LOG_INFORM, XP_GetAdminStr(DBT_errorCreatingTimeCache_));
			return REQ_ABORTED;
		}
	} else
		ereport(LOG_INFORM, XP_GetAdminStr(DBT_timeCacheCacheDisabled_));
    return REQ_PROCEED;
}

NSAPI_PUBLIC int 
time_cache_init(pblock *pb, Session *sn, Request *rq)
{
	char *str_hash_size = pblock_findval("hash-size", pb);
	char *str_cache_size = pblock_findval("cache-size", pb);
	char *str_cache_disabled = pblock_findval("disable", pb);
	int hash_size;
	int cache_size;

	if ( str_cache_disabled && strcasecmp(str_cache_disabled, "false") )
		time_cache_enabled = 0;
	else
		time_cache_enabled = 1;

	if (str_hash_size == NULL)
		hash_size = DEFAULT_HASH_SIZE;
	else {
		if ( (hash_size = atoi(str_hash_size)) < ABSOLUTE_HASH_MIN) {
			ereport(LOG_WARN, XP_GetAdminStr(DBT_timeCacheInitHashSizeDUsingDefau_), 
				ABSOLUTE_HASH_MIN, ABSOLUTE_HASH_MIN);
			hash_size = ABSOLUTE_HASH_MIN;
		} 
		if ( hash_size > ABSOLUTE_HASH_MIN) {
			ereport(LOG_WARN, XP_GetAdminStr(DBT_timeCacheInitHashSizeDUsingDefau_1), 
				ABSOLUTE_HASH_MAX, ABSOLUTE_HASH_MAX);
			hash_size = ABSOLUTE_HASH_MAX;
		} 
	}

	if (str_cache_size == NULL) 
		cache_size = DEFAULT_CACHE_SIZE;
	else {
		if ( (cache_size = atoi(str_cache_size)) < ABSOLUTE_CACHE_MIN) {
			ereport(LOG_WARN, XP_GetAdminStr(DBT_timeCacheInitCacheSizeDUsingDefa_),
				ABSOLUTE_CACHE_MIN, ABSOLUTE_CACHE_MIN);
			cache_size = ABSOLUTE_CACHE_MIN;
		}
		if ( cache_size > ABSOLUTE_CACHE_MAX) {
			ereport(LOG_WARN, XP_GetAdminStr(DBT_timeCacheInitCacheSizeDUsingDefa_1),
				ABSOLUTE_CACHE_MAX, ABSOLUTE_CACHE_MAX);
			cache_size = ABSOLUTE_CACHE_MAX;
		}
	}


	return REQ_PROCEED;
}

NSAPI_PUBLIC void
time_cache_destroy(void *)
{
	/* cache module automatically cleans up the cache itself */
	time_cache = NULL;
	time_cache_size = DEFAULT_CACHE_SIZE;
	time_hash_size = DEFAULT_HASH_SIZE;
	time_cache_enabled = 1;
}

NSAPI_PUBLIC char *
time_cache_strftime(time_t now)
{
	time_cache_entry_t *found, *new_item;
	struct tm *gmt, gmts;
	char *str;

	if (!time_cache)
		goto error;

	if ( (found = (time_cache_entry_t *)cache_do_lookup(time_cache, (void *)&now)) != NULL ) {
		if ( (str = STRDUP(found->time_str)) == NULL) 
			ereport(LOG_FAILURE, XP_GetAdminStr(DBT_errorAllocatingMemoryForTimeCach_));
		cache_use_decrement(time_cache, (cache_entry_t *)found);
		return str;
	}

	if ( (new_item = (time_cache_entry_t *)PERM_MALLOC(sizeof(time_cache_entry_t))) == NULL) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_errorAllocatingMemoryForTimeCach_1));
		goto error;
	}

#ifdef CACHE_DEBUG
	new_item->cache.magic = CACHE_ENTRY_MAGIC;
#endif

	if ( (new_item->time_str = (char *)PERM_MALLOC(sizeof(char)
		*(TIME_MAX_LEN+1))) == NULL) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_errorAllocatingMemoryForTimeCach_2));
		PERM_FREE(new_item);
		goto error;
	}

	new_item->time_val = now;
	gmt = system_gmtime(&now, &gmts);
	util_strftime(new_item->time_str, TIME_FMT, gmt);

	if (cache_insert_p(time_cache, (cache_entry_t *)&(new_item->cache), 
		(void *)&(new_item->time_val), (void *)new_item,
		&time_cache_entry_functions) < 0) {
#ifdef DEBUG
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_errorInsertingNewTimeCacheEntry_));
#endif
		(void)time_cache_cleanup((void *)new_item);
		PERM_FREE(new_item);
		goto error;
	}

	/* NOTICE: This is not a permanent allocation; the caller must free
         * the timestamp! 
         */
	if ( (str = STRDUP(new_item->time_str)) == NULL) 
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_errorAllocatingMemoryForTimeCach_3));
	cache_use_decrement(time_cache, (cache_entry_t *)new_item);
	
	return str;

error:
	/* An error occurred; but we try to NEVER tell the user that; we just 
	 * go the real thing instead.
	 */
	if ( (str = (char *)MALLOC(sizeof(char) *(TIME_MAX_LEN+1))) == NULL) {
		return NULL;
	}
        str[TIME_SIZE] ='\0';
	gmt = system_gmtime(&now, &gmts);
	util_strftime(str, TIME_FMT, gmt);
	return str;
	
}
