/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * Description (cgidebug.c)
 *
 *	This file contains routines to support debugging CGI programs.
 */

#ifdef XP_UNIX

#include <sys/types.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <sys/errno.h>
#include <sys/stat.h>

extern int errno;

/*
 * Description (CGI_DEBUG_FILE)
 *
 *	This string defines the name of a debug executable file that may be
 *	found in the same directory as the CGI program.  Each line of this
 *	file contains a tag string, and optionally a command.  CGI_Debug()
 *	searches for a tag matching its argument, and forks a child process
 *	to execute the specified command, if any, or a default command
 *	otherwise.  The parent process enters a loop waiting for the gocgi
 *	variable (defined below) to be set to a non-zero value (using a
 *	debugger, presumably).
 *
 *	If the CGI_DEBUG_FILE does not exist in the directory with the
 *	CGI program, or if the tag passed to CGI_Debug() does not match
 *	the beginning of any lines in the file, then CGI_Debug() just
 *	returns.
 */
#define CGI_DEBUG_FILE	".cgidebug"

/*
 * Description (gocgi)
 *
 *	This variable is used to prevent CGI_Debug() from returning until
 *	the user has gained control of the CGI program through the debugger.
 *	The user will manually set this variable to a non-zero value to
 *	allow CGI_Debug() to return.
 */
int gocgi = 0;

/*
 * Description (CGI_Debug)
 *
 *	This function provides a mechanism for getting a debugger attached
 *	to a CGI program.  The mechanism can be enabled or disabled at
 *	run-time, through the use of the CGI_DEBUG_FILE, described above.
 *	By default, if debugging is enabled for the calling CGI program,
 *	an attempt is made to pop up an xterm window running gdb on the
 *	CGI program's process.  The window is directed to the client
 *	system, as identified by the REMOTE_HOST or REMOTE_ADDR CGI
 *	environment variables.
 *
 * Arguments:
 *
 *	tag	- tag string to match in the CGI_DEBUG_FILE
 *
 * Returns:
 *
 *	The return value is zero if debugging is not enabled for the
 *	specified tag.  It is -1 if an error is encountered.  If debugging
 *	is enabled, the value of gocgi is returned as soon as gocgi is
 *	set to a non-zero value.
 */

int CGI_Debug(char * tag)
{
    int rv;
    pid_t mypid = getpid();
    pid_t childpid;
    int taglen;
    FILE * dbfile;
    char * cp;
    char * cgiprog;
    char * pwd = getenv("PWD");
    char * script = getenv("SCRIPT_NAME");
    char * remhost = getenv("REMOTE_HOST");
    char * remaddr = getenv("REMOTE_ADDR");
    char * xargs[4];
    static char cmdbuf[256];
    static char dspbuf[128];

    if (!script) {
	/* Shouldn't happen */
	return -1;
    }

    if (!pwd) {
	pwd = getcwd(NULL, 128);
    }

    /* Looking for "/" before the final path component */
    cp = strrchr(script, '/');
    if (!cp) {
	/* No debug if not there */
	return 0;
    }

    /* Remember the name of the CGI program */
    cgiprog = cp + 1;

    /*
     * See if there's an executable .cgidebug file in the
     * current directory.
     */
    dbfile = fopen(CGI_DEBUG_FILE, "r");
    if (!dbfile) {

	/* The file isn't there, so don't debug */
	return 0;
    }

    /* Get the length of the tag we're looking for */
    taglen = strlen(tag);

    /* Read the file */
    for (;;) {

	/* Read a line from the file */
	cp = fgets(cmdbuf, sizeof(cmdbuf), dbfile);
	if (!cp) {
	    /* EOF and tag not found - no debug */
	    fclose(dbfile);
	    return 0;
	}

	/* Check for tag at beginning of line */
	if (!strncmp(tag, cp, taglen)) {

	    /* Got it */

	    fclose(dbfile);

	    /* Skip any whitespace after the tag, up to end-of-line */
	    cp += taglen;
	    while ((*cp != '\n') && isspace(*cp)) ++cp;

	    /* Did we hit a comment or the end-of-line? */
	    if ((*cp == '#') || (*cp == '\n')) {

		/*
		 * Yes, construct the default exec():
		 *
		 *	/bin/sh -c "xterm -e gdb <cgi-program> <mypid>"
		 */
		sprintf(cmdbuf, "xterm -e gdb %s/%s %d &",
			pwd, cgiprog, (int)mypid);
		xargs[0] = "/bin/sh";
		xargs[1] = "-c";
		xargs[2] = cmdbuf;
		xargs[3] = 0;

		/* Also add DISPLAY=<client address>:0 to the environment */
		if (remhost || remaddr) {
		    sprintf(dspbuf, "DISPLAY=%s:0",
			    (remhost) ? remhost : remaddr);
		    putenv(dspbuf);
		}
	    }
	    else {

		/*
		 * Execute the command on the rest of the line as:
		 *
		 *	/bin/sh -c "<cmdbuf> <cgi-program> <mypid>"
		 *
		 * Probably in most cases where this is used, the specified
		 * command will invoke a script.
		 */
		xargs[0] = "/bin/sh";
		xargs[1] = "-c";
		xargs[2] = cp;
		xargs[3] = 0;

		/* Append <cgi-program> <mypid> */
		cp = cmdbuf + strlen(cmdbuf) - 1;
		if (*cp != '\n') {
		    ++cp;
		}
		sprintf(cp, " %s/%s %d", pwd, cgiprog, (int)mypid);
	    }

	    /* Make a new process for the debug script */
	    childpid = fork();
	    if (childpid) {

		/* Parent process */

		if (childpid < 0) {
		    /* Not good */
		    return -1;
		}

		/* Wait for user to start us through the debugger */
		for (gocgi = 0; !gocgi; ) ;

		/* On with the program ... */
		return gocgi;
	    }

	    /* Child process */

	    /* Make a new session - disassociate from the parent */
	    setsid();
	    close(0);
	    close(1);
	    close(2);

	    rv = execvp(xargs[0], xargs);

	    /* Shouldn't get here */
	    exit(1);
	}
    }
}
#endif /* XP_UNIX */
