/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * accel.c
 *
 * Defines the httpd accelerator routines.
 * 
 * Accelerator routines are commonly made up of two parts.  The first
 * part is just a normal NSAPI-like function.  However, in order to 
 * operate in the accelerator, it must have a "back-end" function which
 * is a callback function to complete any IO which would have normally
 * blocked.
 * 
 * By convention, the "front-end" functions are being named accel_<func>
 * and the "back-end" are being named "acb_<func>".
 * 
 *  THis module is not finished
 * 
 * 
 * Mike Belshe
 * 12-21-95
 * 
 */

#ifdef MIKE_ACCELERATOR
#ifdef HPUX
#include <pthread.h>
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include "base/ereport.h"
#include "frame/accel.h"
#include "frame/req.h"			/* REQ_MAX_LINE */
#include "frame/http.h"			/* http_parse_request */
#include "base/util.h"			/* util_sprintf */
#include "nspr/prfile.h"		/* PR_ImportSocket */
#include "frame/accel_cache.h"		/* Accelerator cache... */
#include "base/nsassert.h"
#include "base/util.h"

#include "frame/dbtframe.h"

#define MY_READ_TIMEOUT		20
#define MY_WRITE_TIMEOUT	20


/* -------------------------- accept_connection ------------------------ */
NSAPI_PUBLIC int
accel_accept_connection(mr_pool_t *mrp, int sock) 
{
#ifdef MIKE_ACCELERATOR_DEBUG
	ereport(LOG_INFORM, XP_GetAdminStr(DBT_mrAcceptD_), mrp);
#endif

	if (mr_async_io(mrp, AIO_READ, sock, 0, 0, NULL, NULL, 0,
		acb_accept_connection, NULL, NULL) < 0) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_errorIssuingReadOnAcceptSocket_));
		return -1;
	}
	return 0;
}

NSAPI_PUBLIC int 
acb_accept_connection(overlapped_io_t *olp)
{
	int accept_sock, newsock;
	struct sockaddr_in sin;
	int sinlen = sizeof(sin);
	char *buffer, errbuf[DEF_ERRBUF];
	Session *session;

#ifdef MIKE_ACCELERATOR_DEBUG
	ereport(LOG_INFORM, XP_GetAdminStr(DBT_acbAcceptConnectionD_), olp->file);
#endif
	
	/* Check that there was no error on the read */
	if (olp->return_length < 0) {
		ereport(LOG_CATASTROPHE, XP_GetAdminStr(DBT_errorGettingAcceptSocketD_),
			olp->return_error);
	}

	net_accept_enter();
	if ((accept_sock = net_accept(olp->file, (struct sockaddr *)&sin, &sinlen)) <0){
		/* We knew that a new connection had arrived because we
		 * used select.  But since we get EWOULDBLOCK here, it means
		 * that another process called accept() and took the socket
		 * away from us.  
		 */
		if (errno != EWOULDBLOCK) {
			ereport(LOG_CATASTROPHE, XP_GetAdminStr(DBT_errorInAcceptDS_), 
				errno, util_strerror(errno, errbuf, DEF_ERRBUF));
			net_accept_exit();
			return -1;
		}
	}
	net_accept_exit();

	/* Issue a new "async" accept() */
	if ( accel_accept_connection(olp->pool, olp->file) < 0) {
		ereport(LOG_CATASTROPHE, XP_GetAdminStr(DBT_errorCreatingNewAcceptRequest_));
		if (newsock > 0)
			net_close(newsock);
		return -1;
	}

	if (accept_sock < 0) {
		/* Another process snatched the call away from us! */
		return 0;
	}

	/* That wild and crazy NSPR... */
	newsock = PR_ImportSocket(accept_sock);

#ifdef MIKE_ACCELERATOR_DEBUG
	ereport(LOG_INFORM, XP_GetAdminStr(DBT_acceptedConnectionDNsprD_), 
		accept_sock, newsock );
#endif

	if ( (session = session_create(newsock, &sin)) == NULL) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_errorCreatingNewSessionStructure_));
		net_close(newsock);
		return -1;
	}
	session->mrp = olp->pool;
	FREE(olp);

	return accel_read_request(session);

}


/* -------------------------- read_request -------------------------- */

NSAPI_PUBLIC int
accel_read_request(Session *session)
{
#ifdef MIKE_ACCELERATOR_DEBUG
	ereport(LOG_INFORM, XP_GetAdminStr(DBT_accelReadRequest_));
#endif

	/* Go ahead and start using the session netbuf, even though it
	 * isn't owned by us.  The "correct" alternative would probably
	 * be to open something in netbuf.c, but I am being lazy now...
	 */
	if ( (session->inbuf->inbuf = 
		(unsigned char *)MALLOC(session->inbuf->maxsize*sizeof(char))) == NULL) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_errorAllocatingRequestReadBuffer_));
		goto error;
	}

	if ( mr_async_io( session->mrp, AIO_READ, session->csd, 0, 
		session->inbuf->maxsize, NULL, (char *)(session->inbuf->inbuf), 
		MY_READ_TIMEOUT, acb_read_request, NULL, session) < 0)  {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_errorIssuingAsyncReadRequest_));
		goto error;
	}

	return 0;
error:
	net_close(session->csd);
	session_free(session);
	return -1;
}

/* acb_read_request()
 *
 * This function is responsible for reading just the request line.
 * It does not parse other request headers.
 *
 * When this routine is first called, the request structure has not yet
 * been built.  We store data read from the client until we get the 
 * end-of-line marker.  Once that is done we parse the request and
 * check the cache.  
 * 
 */
NSAPI_PUBLIC int 
acb_read_request(overlapped_io_t *olp)
{
	Session *session;
	char *buf;
	unsigned int space;
	data_list_t *new_data;
	Request *request;
	char *end_of_request;
	char errbuf[DEF_ERRBUF];

#ifdef MIKE_ACCELERATOR_DEBUG
	ereport(LOG_INFORM, XP_GetAdminStr(DBT_acbReadRequestDBytesD_), olp->file, olp->return_length);
#endif

	NS_ASSERT(olp);
	NS_ASSERT(olp->private);
	NS_ASSERT(olp->type == AIO_READ);

	session = (Session *)olp->private;

#ifdef MIKE_ACCELERATOR_DEBUG
	ereport(LOG_INFORM, XP_GetAdminStr(DBT_acbReadReq1SessionD_), session);
#endif

	/* Check if READ failed */
	if ( olp->return_length <= 0) {
		if (olp->return_length < 0)
			ereport(LOG_FAILURE, XP_GetAdminStr(DBT_errorReadingRequestDS_),
				olp->return_error, util_strerror(olp->return_error, errbuf, DEF_ERRBUF));
		else
			ereport(LOG_FAILURE, XP_GetAdminStr(DBT_clientAbortedConnection_));
		goto error;
	}

	olp->buffer[olp->return_length] = '\0';

	/* Check if we have the whole line */
	if ( (end_of_request = strchr(olp->buffer, '\n')) == NULL) {
		if ( mr_async_io( olp->pool, AIO_READ, session->csd, 0, 
			REQ_MAX_LINE - olp->return_length, NULL,  
			&(olp->buffer[olp->return_length]), MY_READ_TIMEOUT,
			acb_read_request, NULL, session) < 0)  {
			ereport(LOG_FAILURE, XP_GetAdminStr(DBT_errorReadingRequest_));
			FREE(olp);
			goto error;
		}
		FREE(olp);
		return 0;
	}

	/* Ok, finished reading the whole request */

	/* Terminate the request string with a NULL character */
	*end_of_request = '\0';
	if ( (end_of_request > olp->buffer) )
		if ( *((char *)(end_of_request-1)) == '\r')
			*((char *)(end_of_request-1)) = '\0';
	end_of_request++;

	/* Now doctor up the netbuf so that it looks legit 
	 * This is not a proper way to deal with netbufs... XXXMB
	 */ 
	session->inbuf->cursize = olp->return_length;
	session->inbuf->pos = end_of_request - (char *)(session->inbuf->inbuf);

	/* Setup a new request structure */
	if ( (request = request_create()) == NULL) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_errorCreatingNewRequest_));
		goto error;
	}
	request->session = session;
	olp->private = (void *)request;

	if (http_parse_request(olp->buffer, request, session) == REQ_ABORTED) {
		request_free(request);
		goto error;
	}

	/* NOTE: If we wanted to be speedy gonzales, we could bypass reading
	 *       the remaining headers and immediately start servicing
	 *	 the request.  The problem is that if the request specified
	 *	 a byte-range, we wouldn't notice.  For now, lets do the
	 *	 right thing- parse the headers and check for byte-range.
	 */

	FREE(olp);
	/* Do we want to parse remaining headers yet? */
	return accel_async_scan_headers(request);

error:
#ifdef MIKE_ACCELERATOR_DEBUG
	ereport(LOG_INFORM, XP_GetAdminStr(DBT_errorOccurredClosingDIoWasForD_),
		session->csd, olp->file);
#endif
	net_close(session->csd);
	session_free(session);
	FREE(olp);
	return -1;
}


/* ----------------------- async_scan_headers ---------------------------*/

/* scan_headers_state_t
 * State info while working on scan_headers
 */
typedef struct scan_headers_state_t {
	int		initialized;	/* flag if this struct is init'd */
	int		x;		/* x position in current "line" */
	int		y;
	int		buffer_len;	/* length of real data in buffer */
	int		buffer_pos;	/* position of first "unprocessed"
					 * character in the buffer 
					 */
	int		buffer_size;	/* total size of the buffer */
	char		*buffer;	/* data buffer */
	char		outbuf[REQ_MAX_LINE];	/* temp output data buffer */
	int		nh;		/* number of headers found */
	int		alen;		/* current accept headers length */
	int		asz;		/* current accept headers buffer size */
	pb_param	*acc;		/* accept headers */
	Request		*request;	/* point to our request structure */
} scan_headers_state_t;

/* accel_async_scan_headers()
 * Function to start scanning the headers.  Similar to the
 * http_scan_headers() function, except that it never blocks by using
 * async IO.
 *
 * olp points to the last read completion from the client.  data points
 * to the next "unprocessed" data within the olp buffer.
 *
 * I don't know quite which module is best for this function; it belongs
 * halfway between the callbacks module and the http module...  Maybe
 * C++ would have helped here...
 *
 */
NSAPI_PUBLIC int
accel_async_scan_headers(Request *request) 
{
	scan_headers_state_t *state;
	overlapped_io_t *olp;

#ifdef MIKE_ACCELERATOR_DEBUG
	ereport(LOG_INFORM, XP_GetAdminStr(DBT_accelAsyncScanHeaders_));
#endif

	NS_ASSERT(request);

	if ( (state = (scan_headers_state_t *)MALLOC(sizeof(*state))) == NULL){
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_outOfMemoryAccelAsyncScanHeaders_));
		return -1;
	}

	if ( (olp = (overlapped_io_t *)MALLOC(sizeof(*olp))) ==  NULL) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_outOfMemoryAccelAsyncScanHeaders_1));
		FREE(state);
		return -1;
	}

	state->nh = 0;
	state->x = 0;
	state->y = -1;
	state->buffer_pos = request->session->inbuf->pos;
	state->buffer_len = request->session->inbuf->cursize;
	state->buffer_size = request->session->inbuf->maxsize; 
	state->buffer	= (char *)(request->session->inbuf->inbuf);
	state->acc = NULL;
	state->alen = 0;
	state->asz = 0;
	state->request = request;
	state->initialized = 1;

	/* The olp is only created as a means to start the callback */
	olp->pool = state->request->session->mrp;
	olp->file = state->request->session->csd;
	olp->buffer = state->buffer;
	olp->return_length = state->buffer_len;
	olp->length = state->buffer_size;
	olp->private = (void *)state;

	return acb_async_scan_headers(olp);
}

/* taken from http.c */
static void 
_merge_header(char *name, char *val, pblock *headers) 
{
    pb_param *found;

    if(!(found = pblock_find(name,headers)))
        pblock_nvinsert(name,val,headers);
    else {
        char *newval;
        int ln = strlen(val), lv = strlen(found->value);
        register int x, y;

        /* 3 = ", " + '\0' */
        newval = (char *) REALLOC(found->value, ln + lv + 3);

        x = lv;
        newval[x++] = ','; newval[x++] = ' ';
        for(y = 0; y < ln; y++,x++)
            newval[x] = val[y];
        newval[x] = '\0';

        found->value = newval;
    }
}

#define ACCEPT_GUESS 2048
NSAPI_PUBLIC int 
acb_async_scan_headers(overlapped_io_t *olp)
{
    scan_headers_state_t *state;
    register int i;
    register char c;

    NS_ASSERT(olp);

    /* Check that the read completed ok */
    if ( olp->return_length < 0 ) {
        ereport(LOG_FAILURE, XP_GetAdminStr(DBT_errorDuringAsyncReadDS_),
            olp->return_error, system_errmsg(olp->return_error));
        goto error;
    }
    
    state = (scan_headers_state_t *)olp->private;

    /* The first time we enter this routine, state is already 
     * initialized, and we should not reinitialize based on the olp.
     * Subsequent calls will need to reinitialize.
     */
    if (!state->initialized) {
        state->buffer_pos = 0;
        state->buffer_len = olp->return_length;
    }
    else {
        state->initialized = 0;
    }

    while(1) {
        if(state->x > REQ_MAX_LINE) {
            ereport(LOG_FAILURE, XP_GetAdminStr(DBT_scanHeadersReportsLineTooLong_));
            goto error;
        }
        if(state->nh > HTTP_MAX_HEADERS) {
            ereport(LOG_FAILURE, XP_GetAdminStr(DBT_scanHeadersReportsTooManyHeaders_));
            goto error;
        }

	if (state->buffer_pos < state->buffer_len) {
            i = olp->buffer[state->buffer_pos];
            state->buffer_pos++;
        }
	else {
            /* refill_buffer */
            if ( mr_async_io( olp->pool, AIO_READ, olp->file, 0, olp->length,
                NULL, olp->buffer, MY_READ_TIMEOUT, acb_async_scan_headers, 
		NULL, state) < 0) {
                ereport(LOG_FAILURE, XP_GetAdminStr(DBT_errorReadingHeaders_));
		FREE(olp);
                goto error;
            }
            FREE(olp);
            return 0;
	}

        if(i == IO_EOF) {
            ereport(LOG_FAILURE, (state->nh ? 
                XP_GetAdminStr(DBT_scanHeadersReportsHeaderMissingT_) : 
                XP_GetAdminStr(DBT_scanHeadersReportsHeaderWasEmpty_)));
            goto error;
        }

        c = (char) i;
        switch(c) {
          case LF:
            if((!state->x) || ((state->x == 1) && (state->outbuf[0] == CR))) {
                /* We are all done; go to next step */

		/* what about that req->loadhdrs stuff??? XXXMB */
		FREE(olp);
		FREE(state);
		return accel_send_plain_file(state->request);
            }

            if(state->outbuf[(state->x)-1] == CR)
                --(state->x);
            state->outbuf[state->x] = '\0';
            if(state->y == -1) {
                char *u = STRDUP(state->outbuf);
                util_sprintf(state->outbuf, XP_GetClientStr(DBT_nameWithoutValueGotLineS_), u);
                FREE(u);
                goto error;
            }
            while(state->outbuf[state->y] && isspace(state->outbuf[state->y])) 
                ++(state->y);

            if(!strcmp(state->outbuf, "accept")) {
                if(!state->acc) {
                    char * valbuf = (char *) MALLOC(ACCEPT_GUESS);
                    valbuf[0] = 0;
                    state->acc = param_create("accept", 0);
                    state->acc->value = valbuf;
                    pblock_pinsert(state->acc, state->request->headers);
                    state->asz = ACCEPT_GUESS;
                    state->alen = state->x - state->y;
                    strcpy(state->acc->value, &state->outbuf[state->y]);
                }
                else {
                    i = state->x - state->y + 2;
                    if((state->alen + i + 1) >= state->asz) {
                        state->asz += ACCEPT_GUESS;
                        state->acc->value = (char *)REALLOC(state->acc->value, 
                            state->asz);
                    }
                    state->acc->value[state->alen] = ','; 
                    state->acc->value[state->alen+1] = ' ';
                    strcpy(&(state->acc->value[state->alen+2]), 
                        &state->outbuf[state->y]);
                    state->alen += i;
                }
            }
            else
                _merge_header(state->outbuf, &state->outbuf[state->y], 
                    state->request->headers);

            state->x = 0; state->y = -1;
            ++(state->nh);
            break;
          case ':':
            if(state->y == -1) {
                state->y = state->x+1;
                c = '\0';
            }
          default:
            state->outbuf[(state->x)++] = 
                ((state->y == -1) && isupper(c) ? tolower(c) : c);
        }
    }

error:
    /* need to be able to fully close connection here */
    net_close(olp->file);
    return -1;
}


/* --------------------------- send_plain_file------------------------ */

/* accel_send_plain_file()
 * Another function which doesn't really belong here... maybe in http.c.
 *
 * Sets up first pass at servicing function... If further work is required,
 * will be done via acb_send_plain_file() (the callback)...
 */
#define SEND_BUF_LEN	(8*1024)
#define HEADER		"Server: Fedora-screans\nContent-type: text/plain\n\n\0"
NSAPI_PUBLIC int 
accel_send_plain_file(Request *request)
{
	accel_cache_entry_t *cache_entry;
	static char *header_buf;
	char *uri;

#ifdef MIKE_ACCELERATOR_DEBUG
	ereport(LOG_INFORM, XP_GetAdminStr(DBT_accelSendPlainFile_));
#endif

#ifdef MIKE_ACCELERATOR_DEBUG
	ereport(LOG_INFORM, XP_GetAdminStr(DBT_accelSendPlainFileFoundRequestD_), request);
#endif
	if ((uri = pblock_findval("uri", request->reqpb)) == NULL) {
		ereport(LOG_CATASTROPHE, XP_GetAdminStr(DBT_parseHeadersLostTheUri_));
		goto error;
	}

#ifdef MIKE_ACCELERATOR_DEBUG
	ereport(LOG_INFORM, XP_GetAdminStr(DBT_accelSendPlainFileFoundUriS_), uri?uri:"NULL");
#endif
	/* Now check for the file in the cache */
	if ( (cache_entry = accel_cache_lookup_uri(uri)) == NULL) {
                return accel_pass_to_worker(request);
	}

#ifdef MIKE_ACCELERATOR_DEBUG
	ereport(LOG_INFORM, XP_GetAdminStr(DBT_accelSendPlainFileFoundInCache_));
#endif
	if (!header_buf) {
		if ((header_buf = MALLOC(strlen(HEADER)+1)) == NULL) {
			ereport(LOG_FAILURE, XP_GetAdminStr(DBT_mallocDied_));
			goto error;
		}
		memcpy(header_buf, HEADER, strlen(HEADER));
		header_buf[strlen(HEADER)] = '\0';
		
	}

	if ( mr_async_io( request->session->mrp, AIO_WRITE, 
		request->session->csd, 0, cache_entry->mmap_len, header_buf, 
		cache_entry->mmap, MY_WRITE_TIMEOUT, acb_send_plain_file, NULL,
		(void *)request) < 0) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_errorWritingBackFileN_));
		goto error;
	}

	return 0;
error:
	net_close(request->session->csd);
	session_free(request->session);
	request_free(request);
	return -1;
}

NSAPI_PUBLIC int 
acb_send_plain_file(overlapped_io_t *olp)
{
	static char *buffer = NULL;
	Request *request;
	char errbuf[DEF_ERRBUF];

#ifdef MIKE_ACCELERATOR_DEBUG
	ereport(LOG_INFORM, XP_GetAdminStr(DBT_acbSendPlainFileD_), olp->file);
#endif

	NS_ASSERT(olp);

	if (olp->return_length < 0) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_errorWritingInAcbSendPlainFileDS_),
			olp->return_error, util_strerror(olp->return_error,errbuf,DEF_ERRBUF));
		goto error;
	}

	return acb_close_connection(olp);
error:
	net_close(request->session->csd);
	session_free(request->session);
	request_free(request);
	FREE(olp);
	return -1;
}


/* ----------------------------- close_connection ---------------------- */


NSAPI_PUBLIC int 
acb_close_connection(overlapped_io_t *olp)
{
	unsigned int tmp_socket;
	Request *request;
	char errbuf[DEF_ERRBUF];

#ifdef MIKE_ACCELERATOR_DEBUG
	ereport(LOG_INFORM, XP_GetClientStr(DBT_acbCloseConnectionD_), olp->file);
	NS_ASSERT( olp );
	NS_ASSERT( !FD_ISSET(olp->file, &(olp->pool->read_set)) );
	NS_ASSERT( !FD_ISSET(olp->file, &(olp->pool->write_set)) );
#endif

	if (olp->return_length < 0) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_erroredIoInAcbCloseConnectionDS_),
			olp->return_error, util_strerror(olp->return_error, errbuf, DEF_ERRBUF));
	}

	tmp_socket = olp->file;

	if (net_close(tmp_socket) < 0) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_unableToCloseSocketD_), tmp_socket);
		return -1;
	}

	request = (Request *)olp->private;

	session_free(request->session);
	request_free(request);
	FREE(olp);
	
	return 0;
}

#endif /* MIKE_ACCELERATOR */
