/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * accel_cache.c
 *
 * Cache for the front-end accelerator.
 *
 * Mike Belshe
 * 01-01-96
 */

#ifdef HPUX
#include <pthread.h>
#endif
#include "netsite.h"
#include "base/systems.h"

#ifdef ACCELERATOR_CACHE
#include <sys/types.h>
#include <errno.h>
#include "base/cache.h"
#include "frame/accel_file_cache.h"
#include "frame/accel_cache.h"
#include "base/buffer.h"
#include "base/ereport.h"
#include "base/session.h"
#include "frame/req.h"
#include "base/util.h"
#include "frame/func.h"
#include "frame/objset.h"
#include "base/systhr.h"
#include "base/pool.h"
#include "base/nsassert.h"
#include "frame/dbtframe.h"
#include <frame/conf.h>

/* ----- Forwards ----------------------------------------------------- */
int accel_cache_cleanup(void *);
int accel_cache_debug(pblock *pb, Session *sn, Request *rq);
int accel_cache_print(void *data, SYS_NETFD fd);

/* ----- Static globals ----------------------------------------------- */
static public_cache_functions_t accel_cache_functions = {
	accel_file_cache_hash_key,
	accel_cache_debug
};
static cache_entry_functions_t accel_cache_entry_functions = {
	accel_file_cache_key_cmp, 
	accel_cache_cleanup,
	accel_cache_print 
};

static FuncPtr _ptr_clf_record;
static FuncPtr _ptr_flex_log;
static int (*_ptr_flex_log_tokens)(char *);


/* ----- Functions ---------------------------------------------------- */
static unsigned int 
accel_cache_hash_uri(unsigned int hash_size, void *voidstr)
{
	register unsigned int hash = 0;
	register char *strptr = (char *)voidstr;

	while (*strptr) {
		hash = (hash << 5) +hash + *strptr;
		strptr++;
	}
	return hash % hash_size;
}

void 
_accel_cache_cleanup_logs(accel_cache_entry_t *data)
{
	/* In order to copy the log_os, we want to make sure that we use
	 * PERM_MALLOC instead of the local thread's mallocs... but
	 * objset_free doesn't know about that stuff.  So temporarily
	 * set the thread's pool to NULL so it will use perm malloc...
	 * I think this is ugly....
	 */
	{
		accel_cache_log_t *dead;

		while(data->log_os) {
			dead = data->log_os;
			data->log_os = data->log_os->next;
			if (dead->logname)
				PERM_FREE(dead->logname);
			PERM_FREE(dead);
		}
	}
}

int
accel_cache_cleanup(void *voiddata)
{
	accel_cache_entry_t *data = (accel_cache_entry_t *)voiddata;
	
	NS_ASSERT(data);
	NS_ASSERT(data->headers);
	NS_ASSERT(data->uri);
	NS_ASSERT(data->magic == ACCEL_CACHE_MAGIC);

    /* allow the user to cleanup her own stuff */
    if (data->user_cleanup_func)
        data->user_cleanup_func(data->user_data);

	if (data->headers)
		PERM_FREE(data->headers);
	if (data->key)
		accel_file_cache_key_delete(data->key);
	if (data->path)
		PERM_FREE(data->path);
	if (data->uri)
		PERM_FREE(data->uri);
	if (data->headers_keepalive)
		PERM_FREE(data->headers_keepalive);

	_accel_cache_cleanup_logs(data);

	PERM_FREE(data);

	return 0;
}

int 
accel_cache_debug(pblock *pb, Session *sn, Request *rq)
{
	return cache_dump(accel_file_cache, "Accelerator Cache", sn->csd);
}

#define MAX_DEBUG_LINE 1024
int
accel_cache_print(void *voiddata, SYS_NETFD fd)
{
	accel_cache_entry_t *data = (accel_cache_entry_t *)voiddata;
	char buf[MAX_DEBUG_LINE];
	int len;
	
	NS_ASSERT(data);
	NS_ASSERT(data->magic == ACCEL_CACHE_MAGIC);

	len = util_snprintf(buf, MAX_DEBUG_LINE, "%s, %s, %d (%s)", 
                data->uri, data->path, data->headers, data->headers);
	net_write(fd, buf, len);
	
	return 0;
}

NSAPI_PUBLIC int
accel_cache_init(pblock *pb, Session *sn, Request *rq)
{
    _ptr_clf_record = func_find("common-log");
    _ptr_flex_log = func_find("flex-log");
    _ptr_flex_log_tokens = (int (*)(char *))func_find("flex-log-tokens");

	return REQ_PROCEED;
}

NSAPI_PUBLIC void 
accel_cache_destroy(void *unused)
{
	return;
}

#include "safs/flexlog.h"
#include "safs/addlog.h"

/* The NT accelerator does not build pblocks so it needs a custom
 * log handling function.  This is totally tied to the logging types
 * which are supported from the accelerator; what a hack!
 * 
 * Returns the number of logs found in numfound.  If the return value
 * is NULL and numfound is not 0, then logs were found which are not
 * cacheable.
 */
accel_cache_log_t *_accel_cache_get_logs(httpd_objset *log_os, int *numfound)
{
	int index;
	char *fn_name, *logname;
	FuncPtr fn;
	dtable *d;
	accel_cache_log_t *head = NULL, *newentry;

	*numfound = 0;

	/* Now we have to deal with the task of converting the 
	 * objset into a smaller struct; we could get by without it,
	 * but then the accelerator would have to do it over and over
	 * and we don't want that.
	 */

	/* The objset is a copied objset so there is only 1 object
	 * in the set.
	 */
	d = &(log_os->obj[0]->dt[6]);
	for (index=0; index < d->ni; index++) {
		fn_name = pblock_findval("fn", d->inst[index].param);

		if (!fn_name)
			continue;

		fn = func_find(fn_name);

		if ( (fn == _ptr_clf_record) || (fn == _ptr_flex_log) ) {
			(*numfound)++;
		    if (d->inst[index].client)
			    /* Doing the log entry will be dependant on IP */
			    goto cleanup;
			newentry = (accel_cache_log_t *)PERM_MALLOC(sizeof(accel_cache_log_t));
			logname = pblock_findval("name", d->inst[index].param);
			if (logname)
				newentry->logname = PERM_STRDUP(logname);
			else
				newentry->logname = PERM_STRDUP("global");

			/* For flexlog; check to see if we can run the
			 * log function from the cache 
			 */
			if ( fn == _ptr_flex_log && (!_ptr_flex_log_tokens || !_ptr_flex_log_tokens(newentry->logname)) ) 
				goto cleanup;

			newentry->log_fn = fn;
			newentry->iponly = 
				pblock_findval("iponly", 
					d->inst[index].param) ? 1:0;
			newentry->next = head;
			head = newentry;
		}
		else {
			/* Can only handle clf_record and flex_log from
			 * accelerator.  If they want to use something else,
			 * don't cache it.
 			 */
			goto cleanup;
		}
		
	}
	return head;

cleanup:
	while(head) {
		newentry = head;
		head = head->next;
		if (newentry->logname)
			PERM_FREE(newentry->logname);
		PERM_FREE(newentry);
	}

	return NULL;
}

#define LAST_MODIFIED_POSITION (sizeof(MAGNUS_VERSION_STRING) + 85)
#define LAST_MODIFIED "Last-modified"
static char *
_accel_cache_find_last_modified(char *headers, int len)
{
	char *ptr, *tail = headers + len;
	char *guess = headers + LAST_MODIFIED_POSITION;
	char last = '\0';

	/* But first, a quick hack;
	 * 99.9999% of the time, the Last-modified header will be right
	 * after the Date header, which is right after the Server header
	 * which is right after the status header....  So it is a constant
	 * place.  Precalculate that position and see if it is there.
	 * If so, we're done!
	 */
	if ( (guess+sizeof(LAST_MODIFIED)-1 <= tail) && (guess[0] == 'L') ) {
		if (!strncmp(guess, LAST_MODIFIED, sizeof(LAST_MODIFIED)-1)) {
			for (ptr=guess+sizeof(LAST_MODIFIED); ptr<tail && isspace(*ptr); ptr++);
			return ptr;
		}
	}

	tail = tail - sizeof(LAST_MODIFIED);
	for (ptr = headers; ptr<tail; ptr++) {
		if ( last == '\n')
			if (*ptr == 'L')
				if (!strncmp(ptr, LAST_MODIFIED,
					sizeof(LAST_MODIFIED)-1)) {
					for (ptr=ptr+sizeof(LAST_MODIFIED); ptr<tail && isspace(*ptr); ptr++);
					return ptr;
				}
		last = *ptr;
	}
	return NULL;
}

NSAPI_PUBLIC accel_cache_entry_t *
accel_cache_create(
        char *uri, 
        char *host,
        int local_ip, 
        int port, 
        char *path, 
        char *headers, 
        int header_len, 
        httpd_objset *log_os,
        accel_cache_user_service_callback user_service_func,
        accel_cache_user_notification_callback user_notification_func,
        accel_cache_user_cleanup_callback user_cleanup_func,
        void *user_data)
{
	accel_cache_entry_t *newentry;
	int num_logs;

	if (!accel_file_cache) 
		return NULL;

	if ( (newentry = (accel_cache_entry_t *)PERM_MALLOC(
		sizeof(accel_cache_entry_t))) == NULL) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_accelCacheInsertErrorAllocatingE_));
		return NULL;
	}

#ifdef CACHE_DEBUG
	newentry->magic = ACCEL_CACHE_MAGIC;
#endif
	newentry->cache = NULL;
	newentry->key = NULL;
    
    newentry->user_data = user_data;

	/* In order to copy the log_os, we want to make sure that we use
	 * PERM_MALLOC instead of the local thread's mallocs... but
	 * objset_copydirs doesn't know about that stuff.  So temporarily
	 * set the thread's pool to NULL so it will use perm malloc...
	 * I think this is ugly....
	 */
	{
		if (log_os) {
			newentry->log_os = _accel_cache_get_logs(log_os, &num_logs);
			if ( !newentry->log_os && num_logs > 0)  {
				PERM_FREE(newentry);
				return NULL;
			}
		}
	}

	if ( (newentry->path = PERM_STRDUP(path)) == NULL) {
		goto cleanup_logs;
	}
	if ( (newentry->uri = PERM_STRDUP(uri)) == NULL) {
		PERM_FREE(newentry->path);
		goto cleanup_logs;
	}
	if (headers) {
		int nextra = 3; /* nul character, CR, LF */
                const char * keepalive = "Connection: keep-alive\r\n";
                int keepalive_len = strlen (keepalive);
		newentry->headers = (char *)PERM_MALLOC(header_len+nextra);
                
		if (newentry->headers == NULL) {
			PERM_FREE(newentry->path);
			PERM_FREE(newentry->uri);
			goto cleanup_logs;
		};
                memcpy(newentry->headers, headers, header_len);
		/* Since we dont have writev on NT, try to make a 
		 * reasonable header snapshot in place for now 
		 */
                strcpy(newentry->headers + header_len, "\r\n");
		newentry->header_len = header_len + 2;

                /* Make a 2nd snapshot of headers with the keepalive confirmation in place
                 * (one is used on non-keepalive connections, one is used on keepalive connections).
                 * On NT, TransmitFile supports only one "header" buffer, and writev() is not supported.
                 * So it's important to get the headers buffered in one place, to avoid multiple
                 * writes, or, recomposing/restaging them for each request.
                 */
		newentry->headers_keepalive = (char *)PERM_MALLOC(header_len + keepalive_len + nextra);
		if (newentry->headers_keepalive == NULL) {
                        PERM_FREE(newentry->headers);
			PERM_FREE(newentry->path);
			PERM_FREE(newentry->uri);
			goto cleanup_logs;
		};
                memcpy(newentry->headers_keepalive, (void *)headers, header_len);
                memcpy(newentry->headers_keepalive + header_len, keepalive, keepalive_len);
                strcpy(newentry->headers_keepalive + header_len + keepalive_len, "\r\n");
                newentry->header_keepalive_len = header_len + keepalive_len + 2;
		/* We have to scan for this because there was no convenient
		 * place to get it within the server.  This should be nominal 
		 * overhead, however.
		 */
		newentry->last_modified = _accel_cache_find_last_modified(
			newentry->headers, newentry->header_len);

	} else {
		newentry->headers_keepalive = NULL;
		newentry->headers = NULL;
                newentry->last_modified = NULL;
	}

    newentry->user_service_func = user_service_func;
    newentry->user_notification_func = user_notification_func;
    newentry->user_cleanup_func = user_cleanup_func;

	return newentry;

cleanup_logs:
	_accel_cache_cleanup_logs(newentry);
	PERM_FREE(newentry);
	return NULL;
}

NSAPI_PUBLIC accel_cache_entry_t *
accel_cache_insert(
        char *uri, 
        char *host,
        struct sockaddr_in *local_addr,
        char *path, 
        char *headers, 
        int header_len, 
        httpd_objset *log_os,
        accel_cache_user_service_callback user_service_func,
        accel_cache_user_notification_callback user_notification_func,
        accel_cache_user_cleanup_callback user_cleanup_func,
        void *user_data)
{
	accel_cache_entry_t *newentry;

	if (!accel_file_cache) 
		return NULL;

	if ( (newentry = accel_cache_create(uri, 
                                        host, 
                                        local_addr->sin_addr.s_addr, 
                                        local_addr->sin_port, 
                                        path, 
                                        headers, 
                                        header_len, 
                                        log_os, 
                                        user_service_func, 
                                        user_notification_func, 
                                        user_cleanup_func,
                                        user_data)) == NULL) {
		return NULL;
	}

	if ( (newentry->key = accel_file_cache_key_create(URI_KEY, uri, host, 
		local_addr->sin_addr.s_addr, local_addr->sin_port)) == NULL) {
		accel_cache_cleanup(newentry);
		return NULL;
	}

	if ( (newentry->cache = cache_create_entry()) == NULL) {
		(void)accel_cache_cleanup((void *)newentry);
		return NULL;
	}

	if ( cache_insert_p(accel_file_cache, newentry->cache, (void *)newentry->key, 
		(void *)newentry, &accel_cache_entry_functions) < 0) {
                /* Not a serious error; just means cache is full */
		(void)accel_cache_cleanup((void *)newentry);
		return NULL;
	}

	return newentry;
}

NSAPI_PUBLIC int
accel_cache_delete(void *entry)
{
	if ( !accel_file_cache ) 
		return -1;

	NS_ASSERT( ((accel_cache_entry_t *)entry)->magic == ACCEL_CACHE_MAGIC);

	/* cache_delete will call file_cache_cleanup */
	return cache_delete(accel_file_cache, ((accel_cache_entry_t *)entry)->cache);
}


NSAPI_PUBLIC int 
accel_cache_use_increment(accel_cache_entry_t *entry)
{
	NS_ASSERT( entry->magic == ACCEL_CACHE_MAGIC);

	if ( !accel_file_cache || !entry ) 
		return -1;

	return cache_use_increment(accel_file_cache, entry->cache);
}

NSAPI_PUBLIC int
accel_cache_use_decrement(accel_cache_entry_t *entry)
{
	NS_ASSERT(entry->magic == ACCEL_CACHE_MAGIC);

	if ( !accel_file_cache || !entry ) 
		return -1;

	return cache_use_decrement(accel_file_cache, entry->cache);
}

NSAPI_PUBLIC accel_cache_entry_t *
accel_cache_lookup_uri(char *uri)
{
	accel_cache_entry_t *ptr;
	accel_file_cache_key_t key;

	if (!accel_file_cache || !uri) 
		return NULL;

#ifdef CACHE_DEBUG
	key.magic = ACCEL_FILE_KEY_MAGIC;
#endif
	key.type = URI_KEY;
	key.key = uri;

	ptr = (accel_cache_entry_t *)cache_do_lookup(accel_file_cache, 
		(void *)&key);

	if (ptr) {
		NS_ASSERT(ptr->magic == ACCEL_CACHE_MAGIC);
		if (accel_cache_valid(ptr) < 0) {
			/* This may be overkill, but it works */
			if (accel_cache_delete(ptr) < 0)
				/* Couldn't delete just yet */
				accel_cache_use_decrement(ptr);
			return NULL;
		}

	}

	return ptr;

}

NSAPI_PUBLIC int
accel_cache_valid(accel_cache_entry_t *entry)
{
	int rv = 0;
	NS_ASSERT(entry->magic == ACCEL_CACHE_MAGIC);

	return cache_valid(accel_file_cache, entry->cache);
}

NSAPI_PUBLIC int
accel_cache_touch(accel_cache_entry_t *entry)
{
	NS_ASSERT(entry->magic == ACCEL_CACHE_MAGIC);
	return cache_touch(accel_file_cache, entry->cache);
}

#endif /* MIKE_ACCELERATOR */
