/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * accel_file_cache.c
 *
 * Combined cache for the accel_cache and the file_cache.  The reasoning
 * behind this cache is to tightly integrate the two caches.  Initially,
 * the two were separate, independant caches.  This model works ok, but
 * has a potential problem when the cache overflows with keeping the two
 * caches in sync.  There is no evidence that this is a real problem in
 * practice.  I still argue that the situation is a corner case when
 * the cache is overflowing anyway, and we aren't going to see speedups
 * for those files anyway.  
 *
 * Mike Belshe
 * 02-07-96
 */

#ifdef HPUX
#include <pthread.h>
#endif

#include "netsite.h"

#ifdef ACCELERATOR_CACHE
#include <sys/types.h>
#include <errno.h>
#include "base/cache.h"
#include "frame/accel_file_cache.h"
#include "frame/file_cache.h"
#include "frame/accel_cache.h"
#include "base/buffer.h"
#include "base/ereport.h"
#include "base/session.h"
#include "frame/req.h"
#include "base/util.h"
#include "frame/func.h"
#include "frame/objset.h"
#include "base/systhr.h"
#include "base/daemon.h"  /* daemon_atrestart */
#include "prio.h"
#include "pratom.h"
#include "base/nsassert.h"
#include <libaccess/acl.h>
#include <libaccess/aclproto.h>
#include <frame/conf.h>
#include <frame/acl.h>
#include "frame/dbtframe.h"
#include "base/fasttime.h"
#include "httpdaemon/internalstats.h"

extern unsigned long file_cache_poll_interval;
extern unsigned long file_cache_max_file_size;
extern unsigned long file_cache_mmapped_size;
extern unsigned long file_cache_mmapped_max;


/* ----- Macros ------------------------------------------------------- */
#define ABSOLUTE_HASH_MIN	32
#define ABSOLUTE_HASH_MAX	(32 * 1024)
#define ABSOLUTE_CACHE_MIN	ABSOLUTE_HASH_MIN
#define ABSOLUTE_CACHE_MAX	ABSOLUTE_HASH_MAX
#define ABSOLUTE_FILES_MIN	32
#define ABSOLUTE_FILES_MAX	(32 *1024)
#define DEFAULT_CACHE_SIZE	4096
#define DEFAULT_MMAP_FILES	512
#define DEFAULT_HASH_SIZE	(2*DEFAULT_CACHE_SIZE)

/* ----- Forwards ----------------------------------------------------- */
int accel_file_cache_cleanup(void *);
int accel_file_cache_debug(pblock *pb, Session *sn, Request *rq);
int accel_file_cache_print(void *data, SYS_NETFD fd);
accel_file_cache_entry_t *
_accel_file_cache_raw_lookup_uri(char *, char *, struct sockaddr_in *);

/* ----- Static globals ----------------------------------------------- */
cache_t *accel_file_cache = NULL;
static unsigned long accel_file_cache_hash_size = DEFAULT_HASH_SIZE;
static unsigned long accel_file_cache_size = DEFAULT_CACHE_SIZE;
static unsigned long accel_file_cache_max_mmap_files = DEFAULT_MMAP_FILES;
static unsigned long accel_file_cache_enabled = 1;
	/* The accel-file cache also keeps a sorted list of pages which
     * have been hit by their hit count.  These are also protected
     * by the cache_lock(accel_file_cache).
     */
static accel_file_cache_entry_t 
			*hit_list_head = NULL, 
			*hit_list_tail = NULL,
			*hit_list_last_mapped = NULL;
static unsigned long accel_file_cache_open_file_count = 0;
static CRITICAL accel_file_cache_open_file_count_lock = NULL;

static public_cache_functions_t accel_file_cache_functions = {
	accel_file_cache_hash_key,
	accel_file_cache_debug
};
static cache_entry_functions_t accel_file_cache_entry_functions = {
	accel_file_cache_key_cmp, 
	accel_file_cache_cleanup,
	accel_file_cache_print
};

/* ----- Functions ---------------------------------------------------- */
unsigned int 
accel_file_cache_hash_key(unsigned int hash_size, void *voidkey)
{
	register unsigned int hash;
	register char *strptr;
	accel_file_cache_key_t *keyptr = (accel_file_cache_key_t *)voidkey;

	/* nb: make sure type is not a 0 or it won't have any effect here */
	hash = keyptr->type;
	strptr = keyptr->key;

	while (*strptr) {
		hash = (hash << 5) +hash + *strptr + keyptr->type;
		strptr++;
	}

	/* just add the first character of the host field to the hash */
	if (keyptr->host)
		hash = (hash << 5) +hash + *(keyptr->host);

    hash = (hash << 5) +hash + keyptr->local_ip;
    hash = (hash << 5) +hash + keyptr->port;

	return hash % hash_size;
}

int
accel_file_cache_key_cmp(void *voidkey1, void *voidkey2)
{
	accel_file_cache_key_t *key1 = (accel_file_cache_key_t *)voidkey1;
	accel_file_cache_key_t *key2 = (accel_file_cache_key_t *)voidkey2;

	NS_ASSERT(key1);
	NS_ASSERT(key2);
	NS_ASSERT(key1->magic == ACCEL_FILE_KEY_MAGIC);
	NS_ASSERT(key2->magic == ACCEL_FILE_KEY_MAGIC);

	if (key1->type != key2->type)
		return -1;

#ifdef XP_WIN32
	if (!strcasecmp((char *)key1->key, (char *)key2->key))
#else
	if (!strcmp((char *)key1->key, (char *)key2->key))
#endif
		if (( !(key1->host) && !(key2->host) ) ||
			( key1->host && key2->host &&!strcasecmp(key1->host, key2->host) ) )
			if ( key1->local_ip == key2->local_ip )
                if (key1->port == key2->port)
				    return 0;
	return -1;
}
NSAPI_PUBLIC accel_file_cache_key_t *
accel_file_cache_key_create(int type, char *key, char *host, int local_ip, int port)
{
	accel_file_cache_key_t *newentry;

	if ( (newentry = (accel_file_cache_key_t *)PERM_MALLOC(
		sizeof(accel_file_cache_key_t))) == NULL) {
		return NULL;
	}

#ifdef CACHE_DEBUG
	newentry->magic = ACCEL_FILE_KEY_MAGIC;
#endif
	newentry->key = NULL;
	newentry->host = NULL;
	newentry->local_ip = local_ip;
    newentry->port = port;

	if ( (newentry->key = PERM_STRDUP(key)) == NULL)  
		goto fail;

	if (host) {
		newentry->host = PERM_STRDUP(host);
		if (!newentry->host) 
			goto fail;
	}

	newentry->type = type;

	return newentry;
fail:
	NS_ASSERT(newentry);

	if (newentry->key)
		PERM_FREE(newentry->key);
	if (newentry->host)
		PERM_FREE(newentry->host);
	if (newentry->local_ip)
		PERM_FREE(newentry->local_ip);
	PERM_FREE(newentry);

	return NULL;
}

NSAPI_PUBLIC int
accel_file_cache_key_delete(accel_file_cache_key_t *key)
{
	NS_ASSERT(key);
	NS_ASSERT(key->magic == ACCEL_FILE_KEY_MAGIC);

	if (!key)
		return -1;

	PERM_FREE(key->key);
	if (key->host)
		PERM_FREE(key->host);
	PERM_FREE(key);

	return 0;
}


/* Caller must hold the cache_lock(accel_cache_lock) before calling
 * if data->on_hit_list is true.
 */
NSAPI_PUBLIC int
accel_file_cache_cleanup(void *voiddata)
{
	accel_file_cache_entry_t *data = (accel_file_cache_entry_t *)voiddata;
	int had_fc_entry = 0;

	NS_ASSERT(data);
	NS_ASSERT(data->key);
	NS_ASSERT(data->magic == ACCEL_FILE_CACHE_MAGIC);

	(void)accel_file_cache_key_delete(data->key);
	data->key = NULL;

	if (data->accelcache) {
		NS_ASSERT(data->accelcache->magic == ACCEL_CACHE_MAGIC);
		(void)accel_cache_cleanup(data->accelcache);
		data->accelcache = NULL;
	}

	if (data->filecache) {
		NS_ASSERT(data->filecache->magic == FILE_CACHE_MAGIC);
		if (data->action_pending != ACTION_UNCACHE_FILE)
			had_fc_entry = 1;
		(void)file_cache_cleanup(data->filecache);
		data->filecache = NULL;
	} else if (data->action_pending == ACTION_CACHE_FILE)
		had_fc_entry = 1;

	/* remove from the hit-list */
	if (data->on_hit_list) {
		if (data->less_hit) {
			accel_file_cache_entry_t *ptr; 

			data->less_hit->more_hit = data->more_hit;

			/* Go find the entry to pull in the filecache entry for */
			if (had_fc_entry) {
				ptr = hit_list_last_mapped?hit_list_last_mapped->less_hit:NULL;
				if (ptr) {
					if (cache_get_use_count(accel_file_cache, ptr->cache) == 0) {
						if ( (ptr->filecache = file_cache_create(ptr->accelcache->path)) ) {
							ptr->filecache->cache = ptr->cache;
							ptr->action_pending = ACTION_NONE;
						} else {
							ptr->action_pending = ACTION_CACHE_FILE;
                                                }
					} else {
						NS_ASSERT(!ptr->filecache);
						ptr->action_pending = ACTION_CACHE_FILE;
					}
					hit_list_last_mapped = ptr;
				} else {
					/* nobody to promote */
					NS_ASSERT(!hit_list_tail || (hit_list_tail->filecache||
						hit_list_tail->action_pending == ACTION_CACHE_FILE));
					if (hit_list_last_mapped == data)
						hit_list_last_mapped = hit_list_tail;
					crit_enter(accel_file_cache_open_file_count_lock);
					accel_file_cache_open_file_count--;
					crit_exit(accel_file_cache_open_file_count_lock);
				}
			}
		}
		else {
			NS_ASSERT(hit_list_tail == data);
			hit_list_tail = data->more_hit;
			if (hit_list_last_mapped == data)
				hit_list_last_mapped = data->more_hit;
			if ( had_fc_entry ) {
				NS_ASSERT(!hit_list_tail || (hit_list_tail->filecache||
				hit_list_tail->action_pending == ACTION_CACHE_FILE));
				crit_enter(accel_file_cache_open_file_count_lock);
				accel_file_cache_open_file_count--;
				crit_exit(accel_file_cache_open_file_count_lock);
			}
		}
		if (data->more_hit)
			data->more_hit->less_hit = data->less_hit;
		else {
			NS_ASSERT(hit_list_head == data);
			hit_list_head = data->less_hit;
		}
		data->more_hit = data->less_hit = NULL;
	}

        /* cleanup acl data */
        NS_ASSERT(ACL_AssertAcllist(data->acllist));
        if (data->acllist != ACL_LIST_NO_ACLS) {
            ACL_ListDecrement(NULL, data->acllist);
            data->acllist = NULL;
        }
	
	PERM_FREE(data);
	
	return 0;
}

#define MAX_DEBUG_LINE 1024
int 
accel_file_cache_debug(pblock *pb, Session *sn, Request *rq)
{
    char buf[MAX_DEBUG_LINE];
	int len;
	accel_file_cache_entry_t *ptr;

	crit_enter(accel_file_cache_open_file_count_lock);
	len = util_snprintf(buf, MAX_DEBUG_LINE, "<P>files cached- %d/%d <P>", 
		accel_file_cache_open_file_count, accel_file_cache_max_mmap_files);
	crit_exit(accel_file_cache_open_file_count_lock);
	net_write(sn->csd, buf, len);

	len = util_snprintf(buf, MAX_DEBUG_LINE, "<P>hit-list-head %x, hit-list-tail %x hit-list-last-mapped %x <P>",
		 hit_list_head, hit_list_tail, hit_list_last_mapped);

	net_write(sn->csd, buf, len);

	for (ptr = hit_list_head; ptr; ptr=ptr->less_hit)
		accel_file_cache_print(ptr, sn->csd);

	return cache_dump(accel_file_cache, "Accelerator Cache", sn->csd);
}

NSAPI_PUBLIC int
accel_file_cache_print(void *voiddata, SYS_NETFD fd)
{
	accel_file_cache_entry_t *data = (accel_file_cache_entry_t *)voiddata;
    char buf[MAX_DEBUG_LINE];
	int len;

	NS_ASSERT(data->magic == ACCEL_FILE_CACHE_MAGIC);
	NS_ASSERT(data->accelcache->magic == ACCEL_CACHE_MAGIC);
	NS_ASSERT(!data->filecache || data->filecache->magic == FILE_CACHE_MAGIC);

	len = util_sprintf(buf, "<TABLE BORDER=4> <th>me</th>\
<th>less_hit</th><th>more-hit</th> <th>hit-count</th> <th>action_pending</th> <th>fcpresent</th>\
<tr align=right>");
	net_write(fd, buf,len);

	len = util_snprintf(buf, MAX_DEBUG_LINE,  "<td>%x</td> <td>%x</td><td>%x</td>\
<td>%d</td> <td>%d</td> <td>%d</td> </tr></table>\n",
		data, data->less_hit, data->more_hit, data->hit_count, data->action_pending, 
		data->filecache?1:0); 
	net_write(fd, buf,len);

	(void)accel_cache_print(data->accelcache, fd);
	if (data->filecache)
		(void)file_cache_print(data->filecache, fd);
	
	return 0;
}

int
accel_file_cache_internal_init(void)
{
	if (accel_file_cache_enabled == 0) {
		ereport(LOG_INFORM, XP_GetAdminStr(DBT_cacheInitServerCacheDisabled_));
		return REQ_PROCEED;
	}

    accel_cache_init(NULL,NULL,NULL);

    accel_file_cache_open_file_count_lock = crit_init();

	if ( file_cache_internal_init() < 0) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_accelFileCacheErrorInitializingF_));
		return REQ_ABORTED;
	}

	if ( (accel_file_cache = cache_create(accel_file_cache_size, 
		accel_file_cache_hash_size, 
		&accel_file_cache_functions)) ==NULL){
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_accelFileCacheErrorCreatingCache_));
		return REQ_ABORTED;
	}

	return 0;
}

NSAPI_PUBLIC int
accel_file_cache_init(pblock *pb, Session *sn, Request *rq)
{
	char *str_hash_size = pblock_findval("CacheHashSize", pb);
	char *str_cache_max_mmap_files = pblock_findval("MaxNumberOfOpenCachedFiles", pb);
	char *str_cache_size = pblock_findval("MaxNumberOfCachedFiles", pb);
	char *str_cache_disabled = pblock_findval("disable", pb);

	if (str_cache_disabled && strcasecmp(str_cache_disabled, "false"))
		accel_file_cache_enabled = 0;
	else
		accel_file_cache_enabled = 1;

	if (str_hash_size == NULL)
		accel_file_cache_hash_size = DEFAULT_HASH_SIZE;
	else 
		if ( (accel_file_cache_hash_size = atoi(str_hash_size)) < ABSOLUTE_HASH_MIN) {
			ereport(LOG_WARN, 
				XP_GetAdminStr(DBT_accelFileCacheInitHashSizeDUsing_), 
				ABSOLUTE_HASH_MIN, ABSOLUTE_HASH_MIN);
			accel_file_cache_hash_size = ABSOLUTE_HASH_MIN;
		} else if ( accel_file_cache_hash_size > ABSOLUTE_HASH_MAX ) {
			ereport(LOG_WARN, 
				XP_GetAdminStr(DBT_accelFileCacheInitHashSizeDUsing_1),
				ABSOLUTE_HASH_MAX, ABSOLUTE_HASH_MAX);
			accel_file_cache_hash_size = ABSOLUTE_HASH_MAX;
		}

	if (str_cache_size == NULL) 
		accel_file_cache_size = DEFAULT_CACHE_SIZE;
	else 
		if ( (accel_file_cache_size = atoi(str_cache_size)) < ABSOLUTE_HASH_MIN) {
			ereport(LOG_WARN, 
				XP_GetAdminStr(DBT_accelFileCacheInitCacheSizeDUsin_),
				ABSOLUTE_HASH_MIN, ABSOLUTE_HASH_MIN);
			accel_file_cache_size = ABSOLUTE_HASH_MIN;
		} else if ( accel_file_cache_size > ABSOLUTE_CACHE_MAX ) {
			ereport(LOG_WARN, 
				XP_GetAdminStr(DBT_accelFileCacheInitCacheSizeDUsin_1),
				ABSOLUTE_HASH_MAX, ABSOLUTE_HASH_MAX);
			accel_file_cache_size = ABSOLUTE_HASH_MAX;
		}

	if (str_cache_max_mmap_files == NULL) 
                if (accel_file_cache_size > DEFAULT_CACHE_SIZE)
                    /* If cache size is increased, increase number of mmaped
                     * files to match.
                     */
		    accel_file_cache_max_mmap_files = accel_file_cache_size;
                else
		    accel_file_cache_max_mmap_files = DEFAULT_MMAP_FILES;
	else 
		if ( (accel_file_cache_max_mmap_files = atoi(str_cache_max_mmap_files)) < ABSOLUTE_FILES_MIN) {
			ereport(LOG_WARN, 
				XP_GetAdminStr(DBT_accelFileCacheInitMaxFilesDUsing_),
				ABSOLUTE_FILES_MIN, ABSOLUTE_FILES_MIN);
			accel_file_cache_max_mmap_files = ABSOLUTE_FILES_MIN;
		} else if ( accel_file_cache_max_mmap_files > ABSOLUTE_FILES_MAX ) {
			ereport(LOG_WARN, 
				XP_GetAdminStr(DBT_accelFileCacheInitMaxFilesDUsing_1),
				ABSOLUTE_FILES_MAX, ABSOLUTE_FILES_MAX);
			accel_file_cache_max_mmap_files = ABSOLUTE_FILES_MAX;
		}

	if ( file_cache_init(pb, sn, rq) != REQ_PROCEED) {
		ereport(LOG_FAILURE, 
			XP_GetAdminStr(DBT_accelFileCacheErrorInitializingF_1));
		return REQ_ABORTED;
	}

    ereport(LOG_VERBOSE, XP_GetAdminStr(DBT_fileCacheEnabledS_), 
        accel_file_cache_enabled?XP_GetAdminStr(DBT_on_):XP_GetAdminStr(DBT_off_));
    ereport(LOG_VERBOSE, XP_GetAdminStr(DBT_fileCacheCacheSizeD0xX_), 
        accel_file_cache_size, accel_file_cache_size);
    ereport(LOG_VERBOSE, XP_GetAdminStr(DBT_fileCacheHashSizeD0xX_),
        accel_file_cache_hash_size, accel_file_cache_hash_size);
    ereport(LOG_VERBOSE, XP_GetAdminStr(DBT_fileCacheMaxFilesD0xX_),
        accel_file_cache_max_mmap_files, accel_file_cache_max_mmap_files);

	return REQ_PROCEED;
}

/* ACLs need to be able to flush the cache; this routine does the cache 
 * flush in a specialized way for them.  This ought to be more generic-XXXMB.
 * This function is SLOW; and it grabs a global lock! 
 */
NSAPI_PUBLIC void
accel_file_cache_acl_flush(void *unused)
{
	accel_file_cache_entry_t *ptr, *nptr;

	if (!accel_file_cache)
		return;

	cache_lock(accel_file_cache);

	for (ptr = hit_list_head; ptr; ptr=nptr) {
        NS_ASSERT(ACL_AssertAcllist(ptr->acllist));
        nptr = ptr->less_hit;
        accel_file_cache_delete(ptr);
    }

	cache_unlock(accel_file_cache);
}

NSAPI_PUBLIC void 
accel_file_cache_destroy(void *unused)
{
	if (accel_file_cache) {
		/* cache module automatically cleans up the cache itself */
		(void)file_cache_destroy(NULL);
		(void)accel_cache_destroy(NULL);
	}
	accel_file_cache = NULL;
}

NSAPI_PUBLIC accel_file_cache_entry_t *
accel_file_cache_create(
        char *uri, 
        char *host,
        int local_ip, 
        int port, 
        char *path, 
        char *headers, 
        int header_len, 
        httpd_objset *log_os,
        accel_cache_user_service_callback user_service_func,
        accel_cache_user_notification_callback user_notification_func,
        accel_cache_user_cleanup_callback user_cleanup_func,
        void *user_data,
	ACLListHandle_t *acllist)
{
	accel_file_cache_entry_t *newentry;
	file_cache_entry_t *fc_entry = NULL;
	accel_cache_entry_t *ac_entry;
	accel_file_cache_key_t *key;
	int action_pending = ACTION_NONE;
    int last_checked = 0,
        last_access = 0,
        creation_date = 0,
        modify_date = 0;

	if (!accel_file_cache) 
		return NULL;

	crit_enter(accel_file_cache_open_file_count_lock);
	if ( accel_file_cache_open_file_count < accel_file_cache_max_mmap_files ) {
		crit_exit(accel_file_cache_open_file_count_lock);
		if ((fc_entry = file_cache_create(path))) {
			NS_ASSERT(fc_entry->magic == FILE_CACHE_MAGIC);
		} else {
			action_pending = ACTION_CACHE_FILE;
			/* Otherwise go ahead with the open; just the file cache is not
			 * filled in.
			 */
                }
	} else {
		crit_exit(accel_file_cache_open_file_count_lock);
	}

	{
		struct stat buf;

		if ( stat(path, &buf) < 0)  {
			ereport(LOG_FAILURE, XP_GetAdminStr(DBT_accelFileCacheCreateErrorStatIngFileS_),
			path, system_errmsg());
			return NULL;
		} else {
			last_checked = last_access = ft_time();
			creation_date = buf.st_ctime;
			modify_date = buf.st_mtime;
		}
	}

	if ( (ac_entry = accel_cache_create(uri, host, local_ip, port, path, 
		headers, header_len, log_os, user_service_func, 
        user_notification_func, user_cleanup_func, user_data)) == NULL) {
		if (fc_entry) {
			file_cache_cleanup(fc_entry);
			fc_entry = NULL;
		}
		return NULL;
	} 
	NS_ASSERT(ac_entry->magic == ACCEL_CACHE_MAGIC);

	if ( (key = accel_file_cache_key_create(URI_KEY, uri, host, 
		local_ip, port)) == NULL) {
		(void)accel_cache_cleanup(ac_entry);
		if (fc_entry) {
			(void)file_cache_cleanup(fc_entry);
			fc_entry = NULL;
		}
		return NULL;
	}
	NS_ASSERT(key->magic == ACCEL_FILE_KEY_MAGIC);


	if ( (newentry = (accel_file_cache_entry_t *)PERM_MALLOC(
		sizeof(accel_file_cache_entry_t))) == NULL) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_accelCacheInsertErrorAllocatingE_1));
		(void)accel_cache_cleanup(ac_entry);
		if (fc_entry) {
			(void)file_cache_cleanup(fc_entry);
			fc_entry = NULL;
		}
		(void)accel_file_cache_key_delete(key);
		return NULL;
	}

	newentry->last_checked = last_checked;
	newentry->last_access = last_access;
	newentry->creation_date = creation_date;
	newentry->modify_date = modify_date;
   
#ifdef CACHE_DEBUG
	newentry->magic = ACCEL_FILE_CACHE_MAGIC;
#endif

	newentry->cache = NULL;
	newentry->filecache = fc_entry;
	newentry->accelcache = ac_entry;
	newentry->key = key;
	newentry->hit_count = 1;
	newentry->action_pending = action_pending;
	newentry->less_hit = newentry->more_hit = NULL;
	newentry->on_hit_list = 0;
	newentry->acllist = NULL;

	NS_ASSERT(newentry->key);
	NS_ASSERT(newentry->accelcache);
	NS_ASSERT(newentry->cache == NULL);

	return newentry;
}

NSAPI_PUBLIC accel_file_cache_entry_t *
accel_file_cache_insert(
        char *uri, 
        char *host,
        struct sockaddr_in *local_addr, 
        char *path, 
        char *headers, 
        int header_len, 
        httpd_objset *log_os,
        accel_cache_user_service_callback user_service_func,
        accel_cache_user_notification_callback user_notification_func,
        accel_cache_user_cleanup_callback user_cleanup_func,
        void *user_data,
	ACLListHandle_t *acllist)
{
    int local_ip = local_addr->sin_addr.s_addr;
    int port = local_addr->sin_port;

	accel_file_cache_entry_t *newentry;

	if (!accel_file_cache) 
		return NULL;


	/* The cache base class uses "replace semantics" on insert- that is
	 * it will delete the old entry to insert a new one.  We definitely
	 * do not want that with this type of cache.  Use the raw lookup so
	 * that the entry is not hidden behind an action_pending.
	 */
	if ((newentry = _accel_file_cache_raw_lookup_uri(uri, host, local_addr)) ) {
		accel_file_cache_use_decrement(newentry);
		return NULL;
    }

	if ( (newentry = accel_file_cache_create(uri, 
                                             host, 
                                             local_ip, 
                                             port, 
                                             path, 
                                             headers, 
                                             header_len, 
                                             log_os,
                                             user_service_func,
                                             user_notification_func,
                                             user_cleanup_func,
                                             user_data,
					     acllist)) == NULL) {
		return NULL;
	}

	/* Get the ACL List for this uri */
	NS_ASSERT(acllist);
	if (acl_root_30) {
            if (ACLUriCacheOn) {
	        if (!ACL_CacheCheckGet(uri, &acllist)) {
		    return NULL;
                }
            } else {
                if (!ACL_ListHashCheck(&acllist)) {
                    return NULL;
                } 
            }
        }
        NS_ASSERT(ACL_AssertAcllist(acllist));

        /* See if the ACL is a NOP for GET requests */
        if (ACL_CachableAclList(acllist)) {
            ACL_ListDecrement(NULL, acllist);
            newentry->acllist = ACL_LIST_NO_ACLS;
        } else {
            newentry->acllist = acllist;
        }
        

	if ( (newentry->cache = cache_create_entry()) == NULL) {
		/* don't need lock since on_hit_list is false */
		(void)accel_file_cache_cleanup((void *)newentry);
		return NULL;
	}

	newentry->accelcache->cache = newentry->cache;
	if (newentry->filecache)
	    newentry->filecache->cache = newentry->cache;

	cache_lock(accel_file_cache);
	if ( cache_insert_p(accel_file_cache, newentry->cache, (void *)newentry->key, 
		(void *)newentry, &accel_file_cache_entry_functions) < 0) {
                /* Not a serious error; just means cache is full */
		cache_unlock(accel_file_cache);
		(void)accel_file_cache_cleanup((void *)newentry);
		return NULL;
	}

	/* While protected by the lock, add to the tail of the hit list */
	newentry->on_hit_list = 1;
	newentry->less_hit = NULL;
	newentry->more_hit = hit_list_tail;
	if (hit_list_tail)
		hit_list_tail->less_hit = newentry;
	else
		hit_list_head = newentry;
	hit_list_tail = newentry;

	if (newentry->filecache || newentry->action_pending == ACTION_CACHE_FILE) {
		crit_enter(accel_file_cache_open_file_count_lock);
		if ( accel_file_cache_open_file_count < accel_file_cache_max_mmap_files ) {
			accel_file_cache_open_file_count++;
			crit_exit(accel_file_cache_open_file_count_lock);
			NS_ASSERT(!newentry->more_hit || (newentry->more_hit->filecache ||
				newentry->more_hit->action_pending == ACTION_CACHE_FILE));
			hit_list_last_mapped = newentry;
		} else {
			/* crap, someone else took our spot... */
			crit_exit(accel_file_cache_open_file_count_lock);
			if (newentry->filecache) 
				(void)file_cache_cleanup(newentry->filecache);
			newentry->action_pending = ACTION_NONE;
			newentry->filecache = NULL;
		}
	}

	cache_unlock(accel_file_cache);

	return newentry;
}

NSAPI_PUBLIC int
accel_file_cache_delete(void *entry)
{
	NS_ASSERT(entry);
	NS_ASSERT(((accel_file_cache_entry_t *)entry)->magic == 
		ACCEL_FILE_CACHE_MAGIC);

	if ( !accel_file_cache ) 
		return -1;

	/* cache_delete will call file_cache_cleanup */
	return cache_delete(accel_file_cache, 
		((accel_file_cache_entry_t *)entry)->cache);
}


NSAPI_PUBLIC int 
accel_file_cache_use_increment(accel_file_cache_entry_t *entry)
{
	if ( !accel_file_cache || !entry ) 
		return -1;

	NS_ASSERT(entry->magic == ACCEL_FILE_CACHE_MAGIC);

	return cache_use_increment(accel_file_cache, entry->cache);
}

NSAPI_PUBLIC int
accel_file_cache_use_decrement(accel_file_cache_entry_t *entry)
{
	int result;

	if ( !accel_file_cache || !entry ) 
		return -1;

	NS_ASSERT(entry->magic == ACCEL_FILE_CACHE_MAGIC);

    if ( entry->action_pending == ACTION_NONE ) 
		return cache_use_decrement(accel_file_cache, entry->cache);

	cache_lock(accel_file_cache);
	result = cache_use_decrement(accel_file_cache, entry->cache);
	if (result == 0) {
		switch(entry->action_pending) {
			case ACTION_CACHE_FILE:
				NS_ASSERT(!entry->filecache);
				if ((entry->filecache = file_cache_create(entry->accelcache->path))) {
					entry->filecache->cache = entry->cache;
    				entry->action_pending = ACTION_NONE;
				}
				break;
			case ACTION_UNCACHE_FILE:
				NS_ASSERT(entry->filecache);
				(void)file_cache_cleanup(entry->filecache);
				entry->filecache = NULL;
   				entry->action_pending = ACTION_NONE;
				break;
			default:
				NS_ASSERT(0);
   				entry->action_pending = ACTION_NONE;
				break;
		}
	}
	cache_unlock(accel_file_cache);

    return result;
}

accel_file_cache_entry_t *
_accel_file_cache_raw_lookup_uri(char *uri, char *host, 
    struct sockaddr_in *local_addr)
{
	accel_file_cache_entry_t *ptr;
	accel_file_cache_key_t key;

#ifdef CACHE_DEBUG
	key.magic = ACCEL_FILE_KEY_MAGIC;
#endif
	key.type = URI_KEY;
	key.key = uri;
	key.host = host;
	key.local_ip = local_addr->sin_addr.s_addr;
	key.port = local_addr->sin_port;

	ptr = (accel_file_cache_entry_t *)cache_do_lookup(accel_file_cache, 
		(void *)&key);

    return ptr;

}

NSAPI_PUBLIC accel_file_cache_entry_t *
accel_file_cache_lookup_uri(char *uri, char *host, struct sockaddr_in *local_addr)
{
	accel_file_cache_entry_t *ptr, *demoted, *promoted_spot;

	if (!accel_file_cache || !uri) 
		return NULL;

    cache_lock(accel_file_cache);
    ptr = _accel_file_cache_raw_lookup_uri(uri, host, local_addr );

	if (ptr) {
		NS_ASSERT(ptr->magic == ACCEL_FILE_CACHE_MAGIC);
		if (accel_file_cache_valid(ptr) < 0) {
			/* This may be overkill, but it works */
			if (accel_file_cache_delete(ptr) < 0)
				/* Couldn't delete just yet */
				accel_file_cache_use_decrement(ptr);
            cache_unlock(accel_file_cache);
			return NULL;
		}

        ptr->hit_count++;

		/* Check to see if the position in the hit-list has changed */
		if ( ptr->more_hit && (ptr->more_hit->hit_count < ptr->hit_count) ) {

			NS_ASSERT(hit_list_head != ptr);

			/* Find out where to promote to */
			for (promoted_spot = ptr->more_hit; 
				promoted_spot && promoted_spot->hit_count < ptr->hit_count;
				promoted_spot = promoted_spot->more_hit);
			if (promoted_spot)
				promoted_spot = promoted_spot->less_hit;
			else
				promoted_spot = hit_list_head;
			
			/* Remove the links to the elt being promoted */
			if (ptr->less_hit)
				ptr->less_hit->more_hit = ptr->more_hit;
			else {
				NS_ASSERT(hit_list_tail == ptr);
				ptr->more_hit->less_hit = NULL;
				hit_list_tail = ptr->more_hit;
			}
			if (ptr->more_hit)
				ptr->more_hit->less_hit = ptr->less_hit;
			else {
				/* can't happen! */
				NS_ASSERT(1);
				hit_list_head = ptr->less_hit;
			}
			if (hit_list_last_mapped == ptr)
				hit_list_last_mapped = ptr->more_hit;


			/* Insert the promoted elt in front of promoted_spot */
			if (promoted_spot->more_hit)
				promoted_spot->more_hit->less_hit = ptr;
			else 
				hit_list_head = ptr;
			ptr->less_hit = promoted_spot;
			ptr->more_hit = promoted_spot->more_hit;
			promoted_spot->more_hit = ptr;

			NS_ASSERT(hit_list_tail != ptr);

			if (promoted_spot->filecache || promoted_spot->action_pending == ACTION_CACHE_FILE) {
				/* Try to map this file in now */
				if (!ptr->filecache && ptr->action_pending != ACTION_CACHE_FILE) {
					if (cache_get_use_count(accel_file_cache, ptr->cache) == 1) {
						if ( (ptr->filecache = file_cache_create(ptr->accelcache->path)) ) {
							ptr->filecache->cache = ptr->cache;
							ptr->action_pending = ACTION_NONE;
						}
						else {
							ptr->action_pending = ACTION_CACHE_FILE;
						}
					} else {
						ptr->action_pending = ACTION_CACHE_FILE;
					}

					/* Try to unmap to compensate for the promoted mapped file 
					 * Loop until we find one that isn't already unmapped.
					 */
					while (1) {
						demoted = hit_list_last_mapped;
						if (demoted == NULL) {
							break;
						}
						hit_list_last_mapped = demoted->more_hit;
						if (demoted->filecache) {
							int action = demoted->action_pending;
							if (cache_get_use_count(accel_file_cache, 
								demoted->cache) == 0) {
								(void)file_cache_cleanup(demoted->filecache);
								demoted->filecache = NULL;
								demoted->action_pending = ACTION_NONE;
							} else {
								demoted->action_pending = ACTION_UNCACHE_FILE;
							}
							if (action != ACTION_UNCACHE_FILE)
								break;
						} else if (demoted->action_pending == ACTION_CACHE_FILE){
							demoted->action_pending = ACTION_NONE;
							break;
						} else {
							NS_ASSERT(/* corrupt cache! */ 1);
							break;
						}
					}
				}
			}
		}
	}
    cache_unlock(accel_file_cache);

	return ptr;

}

NSAPI_PUBLIC int
accel_file_cache_valid(accel_file_cache_entry_t *entry)
{
    NS_ASSERT(entry->magic == ACCEL_FILE_CACHE_MAGIC);
    NS_ASSERT(entry->accelcache->magic == ACCEL_CACHE_MAGIC);
    NS_ASSERT(!entry->filecache || entry->filecache->magic == FILE_CACHE_MAGIC);

    if ( accel_cache_valid(entry->accelcache) < 0) 
        return -1;
	

    if (entry->filecache) {
    	if ( file_cache_valid(entry->filecache) < 0) {
    	    return -1;
        }
    } else {
        /* No filecache entry so we have to stat the file ourselves. */
        
        PRUint32 now, target;
        struct stat st;

        now = ft_time();
        target = entry->last_checked + file_cache_poll_interval;
        if (target < now) {

                if (stat(entry->accelcache->path, &st) < 0) {
                        /* File is effectively gone; log it */
                        ereport(LOG_WARN, XP_GetAdminStr(DBT_accelFileCacheValidCannotStatS_),
                                entry->accelcache->path);
                        return -1;
                }

                /* Checking st_mtime may be redundant; but better than
                 * missing a change.
                 */
                if ((st.st_ctime != entry->creation_date) ||
                    (st.st_mtime != entry->modify_date))
                        return -1;
        }
        PR_AtomicSet((PRInt32 *)&entry->last_checked, (PRInt32)now);

    }

    return 0;
}

NSAPI_PUBLIC int
accel_file_cache_touch(accel_file_cache_entry_t *entry)
{
	NS_ASSERT(entry->magic == ACCEL_FILE_CACHE_MAGIC);

	return cache_touch(accel_file_cache, entry->cache);
}

/* This function is just used for other modules that want to tap into the cache
 * size.
 */
NSAPI_PUBLIC PRUint32
accel_file_cache_get_size()
{
    return accel_file_cache_size;
}

PRBool
GetCacheInfo(cacheInfo *info)
{
    if (!info)
        return PR_FALSE;

    memset(info, 0, sizeof(cacheInfo));

    info->enabled = (accel_file_cache == NULL)?PR_FALSE:PR_TRUE;

    if (info->enabled) {
        info->maxCacheEntries = accel_file_cache->max_size;
        info->maxOpenCacheEntries = accel_file_cache_max_mmap_files;
        info->numCacheEntries = accel_file_cache->cache_size;
        info->numOpenCacheEntries = accel_file_cache_open_file_count;

        info->numCacheHits = accel_file_cache->cache_hits;
        info->numCacheMisses = accel_file_cache->cache_lookups - accel_file_cache->cache_hits;

        info->numCacheInsertsOk = accel_file_cache->insert_ok;
        info->numCacheInsertsFail = accel_file_cache->insert_fail;
        info->numCacheDeletes = accel_file_cache->deletes;

        info->pollInterval = file_cache_poll_interval;
        info->maxTotalCacheSize = file_cache_mmapped_max;
        info->maxCurrentTotalCacheSize = file_cache_mmapped_size;
        info->maxFileSize = file_cache_max_file_size;
    }

    return PR_TRUE;
}

#endif /* ACCELERATOR_CACHE */
