/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * file_cache.c
 *
 * Cache of open files.
 *
 * The file cache has to have a limit; that is we can't just arbitrarily
 * open files and MMAP them all as we'll eat up all the memory.  Ideally,
 * we'd like to manage the cache to make sure that the MMAPed files 
 * are the most frequently hit.  This is a bit of a knapsack problem. 
 * For now, I'm just going to use heuristics to manage what goes in the
 * cache and what doesn't.  We'll leave the cache overflow problem as 
 * sufficiently obscure so that we don't need to optimize it much (yet!).
 *
 * Rules:
 *	- Don't MMAP anything larger than MAX_MMAPPED_FILE
 *	- Only MMAP a total size of files up to the configured limit
 *
 * A note about FILE_MMAP
 *	This module works fine on systems which don't have MMAP support.
 *	As always, it is the caller's responsibility to see if a cached
 *	filehandle has also been mmapped, and that is how we do it.  
 *	On systems which don't have MMAP, the cache is useless anyway-
 *	because we can't share a file descriptor for two threads to read
 *	from simultaneously.  So each thread needs its own copy.
 *  (An exception is when using a system that supports async i/o
 *   requests.  The i/o control block, eg, OVERLAPPED on NT, encodes
 *   the file pointer.  Multiple threads are ok, as long as all the 
 *   threads are doing async i/o.  -jjb)
 *
 * Mike Belshe
 * 01-01-96
 */


#ifdef XP_WIN32
#ifndef MCC_NEWS
/* tempory removed for undefined symbol in libdirmon: ReadDirectoryChangeW
#define USE_DIRECTORY_CHANGES
*/
#undef DEBUG_DIRECTORY_CHANGES
#endif
#endif

#ifdef USE_DIRECTORY_CHANGES
#ifndef  _WIN32_WINNT
 #define _WIN32_WINNT 0x0400
#elif   (_WIN32_WINNT < 0x0400)
 #undef  _WIN32_WINNT
 #define _WIN32_WINNT 0x0400
#endif
#endif

#ifdef HPUX
#include <pthread.h>
#endif
#include "netsite.h"
#include "base/systems.h"
#ifdef XP_UNIX
#include "frame/conf.h"
#endif /* XP_UNIX */
#include "prio.h"
#include "private/pprio.h"     /* Added for nspr20 binary */
#include "base/fasttime.h"

#include "base/pathnames.h"    /* for pn_extract_* --MLM */

#ifdef USE_DIRECTORY_CHANGES
#include "dirmon/dirmon.h"
#endif

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#ifdef ACCELERATOR_CACHE

#include <sys/types.h>
#ifdef XP_UNIX
NSPR_BEGIN_EXTERN_C
#include <sys/mman.h>
NSPR_END_EXTERN_C
#include <fcntl.h>
#else
#include <windows.h>
#endif
#include <errno.h>
#include "base/cache.h"
#include "frame/accel_file_cache.h"
#include "frame/file_cache.h"
#include "frame/conf.h"
#include "base/buffer.h"
#include "base/ereport.h"
#include "base/util.h"
#include "base/nsassert.h"
#include "prtypes.h"


#ifdef USE_DIRECTORY_CHANGES
#include "base/pathnames.h"
#endif

#include "frame/dbtframe.h"

/* ----- Macros ------------------------------------------------------- */
#define MAX_MMAPPED_FILE	(525*1024)
#define ABSOLUTE_MMAP_MIN	1024		/* in kilobytes */
#define ABSOLUTE_MMAP_MAX	(16 * 1024 * 1024)
#define DEFAULT_MMAP_SIZE	(10 * 1024)	/* 10K kilobytes (10MB) */
#define NSCP_TMP_SUBDIR		"fedora/"
#define DEFAULT_POLL_INTERVAL	(5)		/* seconds */

#ifdef USE_DIRECTORY_CHANGES
#define FILE_CACHE_HASH_SIZE	64
#define WATCH_HASH_SIZE	32
//#define WATCH_HASH(x)	( (x[0] + x[strlen(x)-1]) % WATCH_HASH_SIZE )
#define WATCH_HASH(x)		file_cache_hash_path(WATCH_HASH_SIZE, x)
#endif

/* ----- Forwards ----------------------------------------------------- */
int file_cache_cleanup(void *);
int file_cache_debug(pblock *pb, Session *sn, Request *rq);
int file_cache_print(void *data, SYS_NETFD fd);
#ifdef USE_DIRECTORY_CHANGES
void watchdir_callback(char *directory, char *file);
#endif

/* ----- Globals ------------------------------------------------------ */


/* ----- Static globals ----------------------------------------------- */

CRITICAL file_cache_lock = NULL;
unsigned long file_cache_mmapped_size = 0;
unsigned long file_cache_mmapped_max = 0;
unsigned long file_cache_max_file_size = MAX_MMAPPED_FILE;
unsigned long file_cache_poll_interval = DEFAULT_POLL_INTERVAL;
#ifdef XP_UNIX
static int file_cache_mmap_flags = MAP_SHARED;
static int file_cache_mmap_prots = PROT_READ;
#endif /* XP_UNIX */
#ifdef XP_WIN32
char file_cache_tmp_path[2*MAX_PATH];
#ifdef USE_DIRECTORY_CHANGES
	/* In order to do the asynchronous file change notifications on NT
	 * we need to hash the cache entries by filename
	 */
static file_cache_entry_t	*file_cache_hash[FILE_CACHE_HASH_SIZE];
#ifdef DEBUG_DIRECTORY_CHANGES
static int debug_directory_changes = 0;
#endif /* DEBUG_DIRECTORY_CHANGES */
#endif
#endif

static	public_cache_functions_t file_cache_functions = {
		accel_file_cache_hash_key,
		file_cache_debug
	};
static	cache_entry_functions_t file_cache_entry_functions = {
		accel_file_cache_key_cmp,
		file_cache_cleanup,
		file_cache_print
	};


/* ----- Functions ----------------------------------------------------- */

unsigned int 
file_cache_hash_path(unsigned int hash_size, void *voidstr)
{
	register unsigned int hash = 0;
	register char *strptr = (char *)voidstr;

	while (*strptr) {
		hash = (hash << 5) +hash + tolower(*strptr);
		strptr++;
	}
	return hash % hash_size;
}

NSAPI_PUBLIC int 
file_cache_cleanup(void *voiddata)
{
	file_cache_entry_t *data = (file_cache_entry_t *)voiddata;
#ifdef USE_DIRECTORY_CHANGES
	unsigned int hash;
#endif

	NS_ASSERT(data);
	NS_ASSERT(data->magic == FILE_CACHE_MAGIC);

	if (!accel_file_cache || !data)
		return -1;

	if (data->key) 
		accel_file_cache_key_delete(data->key);
	if (data->mmap) {
#ifdef FILE_MMAP
#ifdef XP_WIN32
		UnmapViewOfFile (data->mmap);
		CloseHandle (data->hmap);
#elif XP_UNIX
		if (munmap(data->mmap, data->mmap_len) == -1) 
			ereport(LOG_FAILURE, 
				XP_GetAdminStr(DBT_fileCacheCleanupMunmapFailedS_),
				system_errmsg());
#endif
#else
		ereport(LOG_WARN, XP_GetAdminStr(DBT_fileCacheCleanupFoundMmappedFile_));
#endif
	}

	if ( data->cache_len ) {	
		crit_enter(file_cache_lock);
		file_cache_mmapped_size -= data->mmap_len;
		crit_exit(file_cache_lock);
		data->cache_len = 0;
	}

#ifdef XP_WIN32
	PR_Close(data->fd);
#else
	/* Unix doesn't keep handle open */
#endif

#ifdef USE_DIRECTORY_CHANGES
	hash = file_cache_hash_path(FILE_CACHE_HASH_SIZE, data->path);

	(void)PR_UnwatchDirectory(data->path);

	cache_lock(accel_file_cache);
	if (data->prev_path)
		data->prev_path->next_path = data->next_path;
	if (data->next_path)
		data->next_path->prev_path = data->prev_path;
	if (file_cache_hash[hash] == data)
		file_cache_hash[hash] = data->next_path;
	cache_unlock(accel_file_cache);
#endif
	if (data->path)
		PERM_FREE(data->path);
	PERM_FREE(data);

	return 0;
}

int
file_cache_debug(pblock *pb, Session *sn, Request *rq)
{
	char buf[1024];
	int len;

	if (!accel_file_cache)
		return 0;

	len = util_sprintf(buf, "MMAP cache size: %d/%d\n", 
		file_cache_mmapped_size, file_cache_mmapped_max);
	net_write(sn->csd, buf, len);
	
	return cache_dump(accel_file_cache, "Open File Cache", sn->csd);
}

#define MAX_DEBUG_LINE	1024
NSAPI_PUBLIC int
file_cache_print(void *voiddata, SYS_NETFD fd)
{
	file_cache_entry_t *data = (file_cache_entry_t *)voiddata;
	char buf[MAX_DEBUG_LINE];
	int len;

	NS_ASSERT(data->magic == FILE_CACHE_MAGIC);

	if (!accel_file_cache || !data)
		return 0;

#ifndef XP_UNIX
	len = util_sprintf(buf, "<TABLE BORDER=4> <th>Path</th>\
<th>fd</th> <th>mmap</th> <th>mmap_len</th> <th>last_access</th>\
<tr align=right>");
	net_write(fd, buf,len);

	len = util_snprintf(buf, MAX_DEBUG_LINE,  "<td>%s</td> <td>%d</td>\
<td>%d</td> <td>%d</td> <td>%d</td> </tr></table>\n",
		data->path, data->fd, data->mmap, data->mmap_len, 
		0 /*XXXJB data->last_access*/);
#else
	len = util_sprintf(buf, "<TABLE BORDER=4> <th>Path</th>\
<th>mmap</th> <th>mmap_len</th> <th>last_access</th>\
<tr align=right>");
	net_write(fd, buf,len);

	len = util_snprintf(buf, MAX_DEBUG_LINE,  "<td>%s</td> \
<td>%d</td> <td>%d</td> <td>%d</td> </tr></table>\n",
		data->path, data->mmap, data->mmap_len, 
		0 /*XXXJB data->last_access*/);
#endif
	net_write(fd, buf,len);

	return 0;
}

int
file_cache_mmap_max_init(unsigned int mmap_size)
{
	file_cache_mmapped_max = mmap_size;
	if ( file_cache_mmapped_max < ABSOLUTE_MMAP_MIN) {
		ereport(LOG_WARN,
			XP_GetAdminStr(DBT_fileCacheInitMmapMaxDUsingD_),
			ABSOLUTE_MMAP_MIN, ABSOLUTE_MMAP_MIN);
		file_cache_mmapped_max = ABSOLUTE_MMAP_MIN;
	} else if ( file_cache_mmapped_max > ABSOLUTE_MMAP_MAX ) {
		ereport(LOG_WARN,
			XP_GetAdminStr(DBT_fileCacheInitMmapMaxDUsingD_1),
			ABSOLUTE_MMAP_MAX, ABSOLUTE_MMAP_MAX);
		file_cache_mmapped_max = ABSOLUTE_MMAP_MAX;
	}

	file_cache_mmapped_max *= 1024;         /* convert KB to bytes */

#ifdef XP_UNIX
    if (conf_getglobals()->mmap_flags != -1) {
        file_cache_mmap_flags = conf_getglobals()->mmap_flags;
        ereport(LOG_VERBOSE, XP_GetAdminStr(DBT_fileCacheUsingMmapFlags0xX_), file_cache_mmap_flags);
    }
    if (conf_getglobals()->mmap_prots != -1) {
        file_cache_mmap_prots = conf_getglobals()->mmap_prots;
        ereport(LOG_VERBOSE, XP_GetAdminStr(DBT_fileCacheUsingMmapProts0xX_), file_cache_mmap_prots);
    }
#endif /* XP_UNIX */

	return file_cache_mmapped_size;
}

/* internal_init()
 * Routine to be called internally to initialize.  If the user adds a 
 * "Init fn=file-cache-init" function, that will be handled by file_cache_init.
 */
int
file_cache_internal_init(void)
{
#ifdef USE_DIRECTORY_CHANGES
	int index;

        PR_InitDirMon(file_cache_poll_interval);
#endif
	if (file_cache_mmapped_max == 0)
		(void)file_cache_mmap_max_init(DEFAULT_MMAP_SIZE);

	if ( file_cache_lock == NULL ) {
		if ( (file_cache_lock = crit_init()) == NULL) {
			ereport(LOG_FAILURE,
				XP_GetAdminStr(DBT_fileCacheInitCouldNotCreateLock_));
			return -1;
		}
	}

#ifdef USE_DIRECTORY_CHANGES
	for (index=0; index < FILE_CACHE_HASH_SIZE; index++)
		file_cache_hash[index] = NULL;
#endif

#ifdef XP_WIN32
	/* Create a temporary directory to store all cache files.
     */
	{
		int len;
		int success = 1;

		len = GetTempPath(MAX_PATH, file_cache_tmp_path);
		if (len == 0 || len > MAX_PATH ) 
			success = 0;
		if ( MAX_PATH - len < sizeof(NSCP_TMP_SUBDIR) )
			success = 0;

		/* First directory c:/temp/fedora */
		if (success) {	
			/* Turn all "\" to "/" */
			{
				char *cp = file_cache_tmp_path;

				while(*cp) {
					if (*cp == '\\')
						*cp = '/';
					cp++;
				}
			}

			strcat(file_cache_tmp_path, NSCP_TMP_SUBDIR);

			if (!CreateDirectory(file_cache_tmp_path, NULL)) {
				int err = GetLastError();
				if (err != ERROR_ALREADY_EXISTS) {
					ereport(LOG_WARN, XP_GetAdminStr(DBT_fileCacheUnableToCreateTemporary_),
						file_cache_tmp_path);
					success = 0;
				}
			}
		}

		/* Second directory c:/temp/fedora/<randdir> */
		if (success) {

//			extern char *server_id;

			strcat(file_cache_tmp_path, server_id /*global*/);
				
			if ( len == 0 )
				success = 0;
			else {
				if (!CreateDirectory(file_cache_tmp_path, NULL)) {
					int err = GetLastError();
					if (err != ERROR_ALREADY_EXISTS) {
						ereport(LOG_WARN, XP_GetAdminStr(DBT_fileCacheUnableToCreateTemporary_1),
							file_cache_tmp_path);
						success = 0;
					}
				} else {
					/* Directory already exists; wipe existing files */
					util_delete_directory(file_cache_tmp_path, FALSE);
				}
			}
		}
	
		if (!success) {
			ereport(LOG_WARN, XP_GetAdminStr(DBT_fileCacheGettemppathCannotFindTe_));
			file_cache_tmp_path[0] = '\0';
		}
	}

#endif

	return 0;
}
	
NSAPI_PUBLIC int
file_cache_init(pblock *pb, Session *sn, Request *rq)
{
	char *str_poll_int = pblock_findval("PollInterval", pb);
	char *str_mmap_max = pblock_findval("MaxTotalCachedFileSize", pb);
	char *str_file_max = pblock_findval("MaxCachedFileSize", pb);
#ifdef DEBUG_DIRECTORY_CHANGES
	char *str_debug = pblock_findval("debug_dir", pb);

	if (str_debug)
		debug_directory_changes++;
#endif /* DEBUG_DIRECTORY_CHANGES */

	/* Don't call directly; use accel_file_cache_init() */

	if (str_poll_int == NULL) 
		file_cache_poll_interval = DEFAULT_POLL_INTERVAL;
	else 
		file_cache_poll_interval = atoi(str_poll_int);

	if (str_mmap_max == NULL) 
		(void)file_cache_mmap_max_init(DEFAULT_MMAP_SIZE);
	else 
		(void)file_cache_mmap_max_init(atoi(str_mmap_max));

	if (str_file_max) {
		file_cache_max_file_size = atoi(str_file_max);
		ereport(LOG_INFORM, XP_GetAdminStr(DBT_fileCacheInitSetMaxCachedFileSiz_),
			file_cache_max_file_size);
	}

	if ( file_cache_internal_init() < 0) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_fileCacheInitCouldNotCreateLock_1));
		return REQ_ABORTED;
	}

	return REQ_PROCEED;
}

NSAPI_PUBLIC void 
file_cache_destroy(void *unused)
{
#ifdef FILECACHE_DEBUG
	ereport(LOG_INFORM, XP_GetAdminStr(DBT_fileCacheDestroy_));
#endif

	crit_enter(file_cache_lock);
	file_cache_mmapped_size = 0;
	file_cache_mmapped_max = 0;
	crit_exit(file_cache_lock);
	crit_terminate(file_cache_lock);
	file_cache_lock = NULL;
}

#ifdef XP_WIN32
#define COPY_BUFFER_SIZE	(32 * 1024)
HANDLE
_file_cache_duplicate_file(char *filename, HANDLE orig_file)
{
	HANDLE tmp_file = NULL;	/* handle to file which will be in cache */
	char buffer[COPY_BUFFER_SIZE];
	char *tmp_path = NULL;
	char *tmp_name = NULL;
	char tmp_full_name[MAX_PATH];
	unsigned long bytes_read, bytes_written;

	/* First translate the real file to a temp file name */
	tmp_name = pn_extract_filename(filename);
	
	if (file_cache_tmp_path[0] == '\0') {
		/* try to gracefully recover by mirroring files in docroot */
		tmp_path = pn_extract_directory(filename);
		if (!GetTempFileName(tmp_path, tmp_name, 0, tmp_full_name)) {
			ereport(LOG_FAILURE, XP_GetAdminStr(DBT_fileCacheUnableToGetTempFileName_),
				system_errmsg());
			goto err;
		}
	} else
		if (!GetTempFileName(file_cache_tmp_path, tmp_name, 0, tmp_full_name)) {
			ereport(LOG_FAILURE, XP_GetAdminStr(DBT_fileCacheUnableToGetTempFileName_1),
				system_errmsg());
			goto err;
		}
	
		

	tmp_file = CreateFile(
			tmp_full_name,
			GENERIC_READ | GENERIC_WRITE,
			FILE_SHARE_READ,
			NULL,			/* security */
			CREATE_ALWAYS,
			FILE_FLAG_SEQUENTIAL_SCAN,
			(HANDLE)NULL);

	if (tmp_file == INVALID_HANDLE_VALUE) 
		goto err;

	/* Do copy */
	do {
		if (ReadFile( orig_file, buffer, COPY_BUFFER_SIZE, &bytes_read,
			NULL) ) {
			WriteFile(tmp_file, buffer, bytes_read, &bytes_written, 
				NULL);

			if (bytes_written != bytes_read)
				goto err;
		}
	} while(bytes_read > 0);

	/* close and re-open for ASYNC mode */
	CloseHandle(tmp_file);
	tmp_file = CreateFile(
			tmp_full_name,
			GENERIC_READ,
			FILE_SHARE_READ,
			NULL,			/* security */
			OPEN_EXISTING,
			FILE_FLAG_SEQUENTIAL_SCAN|
			FILE_ATTRIBUTE_TEMPORARY|
			FILE_FLAG_DELETE_ON_CLOSE|
			FILE_FLAG_OVERLAPPED,
			(HANDLE)NULL);

	if (tmp_file == INVALID_HANDLE_VALUE) 
		goto err;

	if (tmp_path)
		FREE(tmp_path);
	if (tmp_name)
		FREE(tmp_name);

	return tmp_file;
err:
	if (tmp_file)
		CloseHandle(tmp_file);
	if (tmp_path)
		FREE(tmp_path);
	if (tmp_name)
		FREE(tmp_name);

	return INVALID_HANDLE_VALUE;
}
#endif /* XP_WIN32 */


NSAPI_PUBLIC file_cache_entry_t *
file_cache_create(char *path)
{
	file_cache_entry_t *newentry;
	int successful_open;
	int successful_stat;
	struct stat buf;
#ifndef XP_WIN32
	SYS_FILE fd = SYS_ERROR_FD;
#else
	HANDLE tmp_handle = INVALID_HANDLE_VALUE;
#endif
#ifdef USE_DIRECTORY_CHANGES
	unsigned int hash;
	PRInt32 set_watch = 0;
#endif

#if !defined(FILE_MMAP) && !defined(XP_WIN32)
	/* Cache is worthless if we don't have MMAP */
	return NULL;
#endif

	if ( !accel_file_cache || !path) 
		return NULL;

	if ( (newentry = (file_cache_entry_t *)PERM_MALLOC(
		sizeof(file_cache_entry_t))) == NULL) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_fileCacheInsertErrorAllocatingEn_));
		goto error;
	}

#ifdef CACHE_DEBUG
	newentry->magic = FILE_CACHE_MAGIC;
#endif

	newentry->cache = NULL;
	newentry->key = NULL;
	/* initialize so error will be able to cleanup properly */
	newentry->path = NULL;
	newentry->mmap = NULL;
	newentry->cache_len = 0;

#ifdef XP_WIN32
    newentry->fd = 0;

	tmp_handle = 
		CreateFile (
			path, 
			GENERIC_READ,
			FILE_SHARE_READ|FILE_SHARE_WRITE,
			NULL,
			OPEN_EXISTING,
			FILE_FLAG_SEQUENTIAL_SCAN,
			0);

	successful_open = (tmp_handle != INVALID_HANDLE_VALUE);

	newentry->hmap = 0;
#else
	fd = system_fopenRO(path);
	successful_open = (fd != SYS_ERROR_FD);
#endif

	if (! successful_open) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_fileCacheCreateErrorOpeningFileS_),
			 path, system_errmsg());
		goto error;
	}

#ifdef XP_WIN32
	if ((newentry->mmap_len = GetFileSize (tmp_handle, NULL)) == 0xFFFFFFFFL)
		successful_stat = FALSE;
	else {
		newentry->valid = 1; /* used with directory monitoring */
		successful_stat = TRUE;
	}
#else
	if ( fstat(PR_FileDesc2NativeHandle(fd), &buf) < 0)  {
		successful_stat = FALSE;
	}
	else {
		successful_stat = TRUE;
		newentry->last_access = ft_time();
		newentry->mmap_len = buf.st_size;
        newentry->creation_date = buf.st_ctime;
        newentry->modify_date = buf.st_mtime;
	}
#endif
    if (successful_stat)
		sprintf(newentry->strbytes, "%d", newentry->mmap_len);
    else {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_fileCacheCreateErrorStatIngFileS_),
			path, system_errmsg());
		goto error;
	}

	if (newentry->mmap_len == 0) 
		goto error;

	newentry->mmap = NULL; /* tentatively */

#ifdef FILE_MMAP
	newentry->cache_len = newentry->mmap_len;
	crit_enter(file_cache_lock);  
	file_cache_mmapped_size += newentry->cache_len;
	
	if ((newentry->mmap_len < file_cache_max_file_size) &&
	    (file_cache_mmapped_size<file_cache_mmapped_max)){
	    	int status_mmap = -1; /* tentatively */
		crit_exit(file_cache_lock);
#ifdef XP_UNIX

		if ( (newentry->mmap = (caddr_t)mmap(NULL, buf.st_size, file_cache_mmap_prots, file_cache_mmap_flags,
			PR_FileDesc2NativeHandle(fd), 0)) != (caddr_t) -1) 
			status_mmap = 0;
		PR_Close(fd);
		fd = SYS_ERROR_FD;
#elif XP_WIN32
		/* Now create the backup copy for the cache */
		newentry->fd = PR_ImportFile((int32)_file_cache_duplicate_file(path, tmp_handle));
		if (newentry->fd != NULL)
			status_mmap = 0;

		if (!status_mmap && SetHandleInformation((HANDLE)PR_FileDesc2NativeHandle(newentry->fd), HANDLE_FLAG_INHERIT, 0) == FALSE)
			status_mmap = -2;
		/* Close the old copy */
		CloseHandle(tmp_handle);
		tmp_handle = INVALID_HANDLE_VALUE;

		status_mmap = 0;

#if 0
		if (newentry->hmap = CreateFileMapping ((HANDLE)newentry->fd, NULL, PAGE_READONLY, 0, 0, NULL))
			if (newentry->mmap = (char *) MapViewOfFile (newentry->hmap, FILE_MAP_READ, 0, 0, 0))
				status_mmap = 0;
			else {
				status_mmap = -3;
				UnmapViewOfFile (newentry->hmap);
			}
#endif
#endif
		if (status_mmap != 0) {
			ereport(LOG_FAILURE, XP_GetAdminStr(DBT_fileCacheCreateErrorMmapIngFileS_),
				path, strerror(errno));
			newentry->mmap = NULL;
			goto error;
		}
	} else {
		file_cache_mmapped_size -= newentry->cache_len;
		crit_exit(file_cache_lock);
		newentry->cache_len = 0;
#ifdef XP_WIN32
		/* If we are here, there wasn't enough room in the cache */
		goto error;
#else
		/* On unix, we can still use just the file handle open. */
		newentry->mmap = NULL;
        if (fd != SYS_ERROR_FD) {
            PR_Close(fd);
            fd = SYS_ERROR_FD;
        }
#endif
	}
#endif

	if ( (newentry->path = PERM_STRDUP(path)) == NULL) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_fileCacheCreateMallocFailure_));
		goto error;
	}

#ifdef USE_DIRECTORY_CHANGES
	{
		char *dir;
		dir = pn_extract_directory(path);
		if (!dir)
			goto error;
		set_watch = PR_WatchDirectory(dir, watchdir_callback);
		FREE(dir);
	}

	hash = file_cache_hash_path(FILE_CACHE_HASH_SIZE, newentry->path);

	cache_lock(accel_file_cache);
	newentry->prev_path = NULL;
	newentry->next_path = file_cache_hash[hash];
	if (file_cache_hash[hash])
		file_cache_hash[hash]->prev_path = newentry;
	file_cache_hash[hash] = newentry;
	cache_unlock(accel_file_cache);
#endif

	return newentry;

error:
#ifdef XP_WIN32
	if (tmp_handle != INVALID_HANDLE_VALUE)
		CloseHandle(tmp_handle);
#endif
	if (newentry) {
		if (newentry->path)
			PERM_FREE(newentry->path);
#ifdef FILE_MMAP
		if (newentry->mmap) {
#ifdef XP_UNIX
			if (munmap(newentry->mmap, newentry->mmap_len) < 0)
				ereport(LOG_FAILURE, XP_GetAdminStr(DBT_fileCacheCreateErrorCaseFailedTo_),
					newentry->mmap, newentry->mmap_len, strerror(errno));

#elif XP_WIN32
			UnmapViewOfFile (newentry->mmap);
			CloseHandle (newentry->hmap);
#endif
		}

		if (newentry->cache_len) {
			crit_enter(file_cache_lock);
			file_cache_mmapped_size -= newentry->mmap_len;
			crit_exit(file_cache_lock);
			newentry->cache_len = 0;
		}
#endif
#ifdef XP_WIN32
		if (newentry->fd)
			CloseHandle (newentry->fd);
#else
		if (fd != SYS_ERROR_FD)
			PR_Close(fd);
#endif

#ifdef USE_DIRECTORY_CHANGES
		if (set_watch)
			PR_UnwatchDirectory(newentry->path);
#endif
		PERM_FREE(newentry);
	}
	return NULL;
}

NSAPI_PUBLIC file_cache_entry_t *
file_cache_insert(char *path)
{
	file_cache_entry_t *newentry;
#ifndef XP_WIN32
	struct stat buf;
#endif

#if !defined(FILE_MMAP) && !defined(XP_WIN32)
	/* Cache is worthless if we don't have MMAP */
	return NULL;
#endif

	if ( !accel_file_cache || !path) 
		return NULL;

	if ( (newentry = file_cache_create(path)) == NULL)
		return NULL;

	if ( (newentry->key = accel_file_cache_key_create(PATH_KEY, path, 
		NULL, 0, 0 )) == NULL){
		file_cache_cleanup(newentry);
		return NULL;
	}

	if ( (newentry->cache = cache_create_entry()) == NULL) {
		(void)file_cache_cleanup(newentry);
		return NULL;
	}

	if ( cache_insert_p(accel_file_cache, newentry->cache, (void *)newentry->key, 
		(void *)newentry, &file_cache_entry_functions) < 0) {
		/* Not a bad error; it just means the cache is full */
		file_cache_cleanup(newentry);
		return NULL;
	}

	return newentry;
}

NSAPI_PUBLIC int
file_cache_delete(void *entry)
{
	NS_ASSERT(((file_cache_entry_t *)entry)->magic == FILE_CACHE_MAGIC);

	if ( !accel_file_cache || !entry ) 
		return -1;

	return cache_delete(accel_file_cache, ((file_cache_entry_t *)entry)->cache);
}


NSAPI_PUBLIC int 
file_cache_use_increment(void *entry)
{
	if ( !accel_file_cache || !entry ) 
		return -1;

	NS_ASSERT(((file_cache_entry_t *)entry)->magic == FILE_CACHE_MAGIC);

	return cache_use_increment(accel_file_cache, 
		((file_cache_entry_t *)entry)->cache);
}

NSAPI_PUBLIC int
file_cache_use_decrement(void *entry)
{
	if ( !accel_file_cache || !entry) 
		return -1;

	NS_ASSERT(((file_cache_entry_t *)entry)->magic == FILE_CACHE_MAGIC);

	return cache_use_decrement(accel_file_cache, 
		((file_cache_entry_t *)entry)->cache);
}

NSAPI_PUBLIC file_cache_entry_t *
file_cache_lookup_path(char *path)
{
	file_cache_entry_t *ptr;
	accel_file_cache_key_t key;

	if (!accel_file_cache || !path ) 
		return NULL;

#ifdef CACHE_DEBUG
	key.magic = ACCEL_FILE_KEY_MAGIC;
#endif
	key.type = PATH_KEY;
	key.key = path;
	key.local_ip = 0;
	key.host = 0;
	ptr = (file_cache_entry_t *)cache_do_lookup(accel_file_cache, (void *)&key);

	if (ptr) {
		NS_ASSERT(ptr->magic == FILE_CACHE_MAGIC);
		if (file_cache_valid(ptr) < 0) {
			if (file_cache_delete(ptr) < 0)
				/* couldn't delete right now */
			file_cache_use_decrement(ptr);
			return NULL;
		}
	}

	return ptr;
}

NSAPI_PUBLIC int
file_cache_valid(file_cache_entry_t *entry)
{
#ifdef USE_DIRECTORY_CHANGES
	if (!entry->valid)
		return -1;
#else
	struct stat st;
	PRUint32 now;
	PRUint32 target;

	if (!entry)
		return -1;

	NS_ASSERT(entry->magic == FILE_CACHE_MAGIC);

	now = ft_time(); 
    target = entry->last_checked + file_cache_poll_interval;
	if (target < now) {

		if (stat(entry->path, &st) < 0) {
			/* File is effectively gone; log it */
			ereport(LOG_WARN, XP_GetAdminStr(DBT_fileCacheValidCannotStatS_), 
				entry->path);
			return -1;
		}

		/* Checking st_mtime may be redundant; but better than 
		 * missing a change.
		 */
		if ((st.st_ctime != entry->creation_date) ||
			(st.st_mtime != entry->modify_date) ||
			(st.st_size != entry->mmap_len) )   
			return -1;
	}
	/* Technically two threads can update this at the same time; but the
	 * worst case is that we do an extra stat call... XXXMB
	 */
	entry->last_checked = now;
#endif

	return cache_valid(accel_file_cache, entry->cache);
}

NSAPI_PUBLIC int
file_cache_touch(file_cache_entry_t *entry)
{
	if (!entry)
		return -1;

	NS_ASSERT(entry->magic == FILE_CACHE_MAGIC);

	return cache_touch(accel_file_cache, entry->cache);
}

#ifdef USE_DIRECTORY_CHANGES

/* _file_cache_lookup_invalidate()
 * Invalidates entries with entry->path == full_name.
 * Returns the count of the number of entries invalidated.
 */
int
_file_cache_lookup_invalidate(char *full_name)
{
	unsigned int hash;
	file_cache_entry_t *ptr;
	unsigned int count = 0;

	NS_ASSERT(full_name);

	hash = file_cache_hash_path(FILE_CACHE_HASH_SIZE, full_name);

	cache_lock(accel_file_cache);
	ptr = file_cache_hash[hash];
	for (ptr = file_cache_hash[hash]; ptr; ptr = ptr->next_path) {
		if ( !pn_pathname_equal(ptr->path, full_name) ) {
#ifdef DEBUG_DIRECTORY_CHANGES
			if (debug_directory_changes)
				ereport(LOG_INFORM, XP_GetAdminStr(DBT_dirChangeInvalidatingSD_),
					ptr->path, ptr->valid);
#endif /* DEBUG_DIRECTORY_CHANGES */
			if (ptr->valid) {
				ptr->valid = 0;
				count++;
			}
		}
	}
	cache_unlock(accel_file_cache);

	return count;
}
#endif


#ifdef USE_DIRECTORY_CHANGES 

/* watchdir_callback()
 * Function that will get called by directory monitoring module 
 * when a file changes
 */
void
watchdir_callback(char *dir, char *file)
{
	char *full_name;

	full_name = pn_concatenate(dir, -1, file, -1);
	if (full_name == NULL) {
		ereport(LOG_FAILURE, XP_GetAdminStr(DBT_unableToCheckAsyncFileStatus_));
                return;
	}

	_file_cache_lookup_invalidate(full_name);
	FREE(full_name);

	return;
}

#endif /* USE_DIRECTORY_CHANGES */

#endif /* ACCELERATOR_CACHE */
