/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * func.c: Code to deal with function hash tables
 * 
 * Rob McCool
 */


#include "netsite.h"
#include "frame/func.h"
#include "base/nsassert.h"
#include "base/systhr.h"

/* nspr20 binary integration */
#include "nspr.h" 
/* #include "private/pprmisc.h" removed for nspr21 binary (19971028) */
/* Looks like AIX can't live with private/primpl.h and the others can't live without it */
/*
#if ! (defined(AIX) && !defined(XP_WIN32))
#include "private/primpl.h"
#endif
*/

#include "frame/log.h"

#include "base/shexp.h"   /* shexp_casecmp */
#include "frame/conf.h"   /* security_active */
#include "base/util.h"    /* sprintf and itoa */

#ifdef NOACL
#include "frame/ipfilter.h"		/* ip_filter_check() */
#include "frame/dnfilter.h"		/* dns_filter_check() */
#endif

#include "frame/dbtframe.h"

/* XXXrobm there is no abstraction for inet_ntoa */
#ifdef XP_WIN32
#include <winsock.h>  /* inet_ntoa prototype */

#else
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif

/*
 * Put NSAPI function vector pointers here to force them to be linked
 * in on platforms where explicit references control linkage.
 */

/* Set up pointer to static NSAPI 3.0 function vector */
extern nsapi_dispatch_t __nsapi30_init;
NSAPI_PUBLIC nsapi_dispatch_t *__nsapi30_table = &__nsapi30_init;

/*
 * On Unix systems nsapi20.o is part of libframe.a.  The following code is a
 * hack to force nsapi20.o to be pulled in when init.o is linked, since
 * there are no references to symbols in nsapi20.o otherwise.
 *
 * On NT any plugins built against Enterprise/FastTrack 2.0 will contain a
 * reference to ns-httpd20.dll, and will not load unless it is present.
 * So nsapi20.o is used to build ns-httpd20.dll on NT, and is not included
 * in libframe.lib.
 */
#ifndef XP_WIN32
/* pull in nsapi20.o */
NSPR_BEGIN_EXTERN_C
NSAPI_PUBLIC extern int nsapi20_present;
NSPR_END_EXTERN_C
static int *nsapi20_linkage = &nsapi20_present;
#endif /* !XP_WIN32 */

/* -------------------------- Static hash table --------------------------- */


static struct FuncStruct *func_tab[NUM_HASH];


/* ------------------------------ func_init ------------------------------- */

void _add_table(struct FuncStruct tab[]) {
    register int x=0;
    register int hn;
    register struct FuncStruct *p,*fs;

    for(x = 0; (tab[x].name); x++) {
        fs = &tab[x];

        hn = FUNC_HASH(fs->name);
        fs->next = NULL;    /* in case we're restarting */

        if(!(p = func_tab[hn]))
            func_tab[hn] = fs;
        else {
            while(p->next)
                p = p->next;
            p->next = fs;
        }
    }
}

PRInt32 func_native_pool_init();

NSAPI_PUBLIC void func_init(struct FuncStruct *func_standard)
{
    int x;
    PRInt32 rv;

    for(x = 0; x < NUM_HASH; x++)
        func_tab[x] = NULL;

    _add_table(func_standard);

    rv = func_native_pool_init();
    NS_ASSERT(rv == 0);
}


/* ------------------------------ func_find ------------------------------- */

extern int java_run(pblock *argpb, Session *sn, Request *rq);

NSAPI_PUBLIC struct FuncStruct *func_find_str(char *name) {
    int hv = FUNC_HASH(name);
    register struct FuncStruct *p;
    p = func_tab[hv];
    while(p) {
        if(!strcmp(name, p->name))
            return p;
        p = p->next;
    }
    return NULL;
}

NSAPI_PUBLIC FuncPtr func_find(char *name) {
    int hv = FUNC_HASH(name);
    register struct FuncStruct *p;
    p = func_tab[hv];
    while(p) {
        if(!strcmp(name, p->name))
            return p->func;
        p = p->next;
    }
    return NULL;
}

NSAPI_PUBLIC FuncPtr func_replace(char *name, FuncPtr newfunc) {
    int hv = FUNC_HASH(name);
    register struct FuncStruct *p;
    p = func_tab[hv];
    while(p) {
        if(!strcmp(name, p->name)) {
            FuncPtr oldfunc = p->func;
            p->func = newfunc;
            return oldfunc;
        }
        p = p->next;
    }
    return NULL;
}

NSAPI_PUBLIC int func_exec(pblock *pb, Session *sn, Request *rq)
{
    struct FuncStruct *f;
    char *fname;
    int rv;

    if(!(fname = pblock_findval("fn", pb))) {
        log_error(LOG_MISCONFIG, XP_GetAdminStr(DBT_funcExec_), sn, rq, 
                  XP_GetAdminStr(DBT_noHandlerFunctionGivenForDirecti_));
        return REQ_ABORTED;
    }
        
    if(!(f = func_find_str(fname))) {
        log_error(LOG_MISCONFIG, XP_GetAdminStr(DBT_funcExec_1), sn, rq, 
                  XP_GetAdminStr(DBT_cannotFindFunctionNamedS_), fname);
        return REQ_ABORTED;
    }

    if (rq)
        rq->directive_is_cacheable = 0;

    if (f->flags & FUNC_USE_NATIVE_THREAD)
        rv = func_native_pool_wait_work(f->func, pb, sn, rq);
    else
        rv = (f->func)(pb, sn, rq);

    if (rq)
        rq->request_is_cacheable &= rq->directive_is_cacheable;

    return rv;
}


/* ----------------------------- func_insert ------------------------------ */


NSAPI_PUBLIC struct FuncStruct *func_insert(char *name, FuncPtr fn)
{
    register int hn;
    register struct FuncStruct *p, *fs;

    hn = FUNC_HASH(name);

    fs = (struct FuncStruct *) MALLOC(sizeof(struct FuncStruct));
    fs->name = name;
    fs->func = fn;
    fs->next = NULL;
    fs->flags = 0;

    if(!(p = func_tab[hn]))
        func_tab[hn] = fs;
    else {
        while(p->next)
            p = p->next;
        p->next = fs;
    }
    return fs;
}

NSAPI_PUBLIC struct FuncStruct *func_insert2(char *name, FuncPtr fn, int flags)
{
    struct FuncStruct *fs;

    fs = func_insert(name, fn);
    if (fs)
        fs->flags = flags;
    return fs;
}

NSAPI_PUBLIC int func_set_native_thread_flag(char *name, int flags)
{
    FuncStruct *f = func_find_str(name);
    if (f == NULL) {
        return (1);
    } else {
        f->flags = flags;
        return (0);
    }
}

/* ---------------------------- native thread pool -------------------------*/

#include <prthread.h>
#include <prlock.h>
#include <prcvar.h>

static PRBool func_native_pool_initialized = PR_FALSE;
static PRInt32 func_native_pool_thread_count;
static PRInt32 func_native_pool_thread_max = 100;
static PRInt32 func_native_pool_thread_available;
static PRInt32 func_native_pool_work_available;
static PRLock *func_native_pool_work_lock;
static PRCondVar *func_native_pool_work_cvar;
static func_work_item_t *func_native_pool_queue_head,
                        *func_native_pool_queue_tail;
static PRUint32 func_native_pool_stacksize = 0;

NSAPI_PUBLIC void func_native_pool_create_thread();

NSAPI_PUBLIC PRInt32 
func_native_pool_init()
{
    char *str_stacksize;
    char *str_threadmax;

    func_native_pool_thread_count = 0;
    func_native_pool_thread_available = 0;
    func_native_pool_work_lock = PR_NewLock();
    if (!func_native_pool_work_lock)
        return -1;
    func_native_pool_work_cvar = PR_NewCondVar(func_native_pool_work_lock);
    if (!func_native_pool_work_cvar)
        return -1;

    func_native_pool_queue_head = NULL;
    func_native_pool_queue_tail = NULL;

    func_native_pool_initialized = PR_TRUE;

    /* the native thread pool gets initialized before we read magnus.conf,
     * so we have to use env variables.
     */
    if ( (str_stacksize = getenv("NSCP_POOL_STACKSIZE")) != 0) {
        func_native_pool_stacksize = atoi(str_stacksize);
    }
    if ( (str_threadmax = getenv("NSCP_POOL_THREADMAX")) != 0) {
        func_native_pool_thread_max = atoi(str_threadmax);
    }

    return 0;
}

NSAPI_PUBLIC PRInt32
func_native_pool_wait_work(FuncPtr fn, pblock *pb, Session *sn, Request *rq)
{
    func_work_item_t work_item;
    PRThread *thread = PR_CurrentThread();
    PRBool need_thread = PR_FALSE;

    /* If we already are a native thread, or if the native thread pool is
     * not initialized, just run the function here.
     */
    /* Since private/primpl.h is gone, this solaris special stuff is gone.
#ifndef SOLARIS
    if (func_native_pool_initialized == PR_FALSE ||
	(PR_GetThreadScope(thread) == PR_GLOBAL_THREAD))
#else
    if (func_native_pool_initialized == PR_FALSE ||
	(PR_GetThreadScope(thread) == PR_GLOBAL_THREAD && (thread->flags &
			_PR_BOUND_THREAD)))
#endif
        return fn(pb, sn, rq);
	*/

    if (func_native_pool_initialized == PR_FALSE ||
	(PR_GetThreadScope(thread) == PR_GLOBAL_THREAD))
      return fn(pb, sn, rq);

    work_item.func = fn;
    work_item.sn = sn;
    work_item.rq = rq;
    work_item.pb = pb;
    work_item.work_complete = PR_FALSE;
    work_item.waiter_lock = PR_NewLock();
    work_item.waiter_cvar = PR_NewCondVar(work_item.waiter_lock);
    
    PR_Lock(func_native_pool_work_lock);
    work_item.next = NULL;
    if (func_native_pool_queue_tail == NULL)
        func_native_pool_queue_head = &work_item;
    else
        func_native_pool_queue_tail->next = &work_item;
    func_native_pool_queue_tail = &work_item;
    
    PR_NotifyCondVar(func_native_pool_work_cvar);

    /* Now see if there are no threads available to do work */
    if (func_native_pool_thread_available == 0 && 
		(!func_native_pool_thread_max ||
		func_native_pool_thread_count < func_native_pool_thread_max)) {
        need_thread = PR_TRUE;
        func_native_pool_thread_available++;
    }
 
    PR_Unlock(func_native_pool_work_lock);

    if (need_thread == PR_TRUE)
        func_native_pool_create_thread();

    /* Now wait for work to be completed */
    PR_Lock(work_item.waiter_lock);
    while(work_item.work_complete == PR_FALSE)
        PR_WaitCondVar(work_item.waiter_cvar, PR_INTERVAL_NO_TIMEOUT);
    PR_Unlock(work_item.waiter_lock);

    /* Cleanup and continue */
    PR_DestroyCondVar(work_item.waiter_cvar);
    PR_DestroyLock(work_item.waiter_lock);

    return work_item.work_status;
}

/* Caller must hold the func_native_pool_work_lock */
NSAPI_PUBLIC func_work_item_t *
func_native_pool_get_work()
{
    func_work_item_t *rv;

    if (!func_native_pool_queue_head)
        return NULL;

    rv = func_native_pool_queue_head;

    func_native_pool_queue_head = func_native_pool_queue_head->next;
    if (!func_native_pool_queue_head) 
        func_native_pool_queue_tail = NULL;
        
    return rv;
}

NSAPI_PUBLIC void
func_native_pool_thread_main(void *unused)
{
    func_work_item_t *work_item = NULL;

    PR_Lock(func_native_pool_work_lock);
    /* decrement before going into the main loop; since we had to increment
     * before having created the thread
     */
    func_native_pool_thread_available--;
    PR_Unlock(func_native_pool_work_lock);

    while(1) {
        PR_Lock(func_native_pool_work_lock);
 
        func_native_pool_thread_available++;

        work_item = func_native_pool_get_work();
        while (!work_item) {
            PR_WaitCondVar(func_native_pool_work_cvar, PR_INTERVAL_NO_TIMEOUT);
            work_item = func_native_pool_get_work();
        }

        func_native_pool_thread_available--;

        PR_Unlock(func_native_pool_work_lock);

        if (work_item->sn)
            systhread_setdata(getThreadMallocKey(), work_item->sn->pool);

        /* do work */
        work_item->work_status = 
            work_item->func(work_item->pb, work_item->sn, work_item->rq);

        if (work_item->sn)
            systhread_setdata(getThreadMallocKey(), NULL);

        PR_Lock(work_item->waiter_lock);
        work_item->work_complete = PR_TRUE;
        PR_NotifyCondVar(work_item->waiter_cvar);
        PR_Unlock(work_item->waiter_lock);
    }
}

NSAPI_PUBLIC void
func_native_pool_create_thread()
{
    PRThread *thread;

    PR_Lock(func_native_pool_work_lock);
    func_native_pool_thread_count++;
    PR_Unlock(func_native_pool_work_lock);

    /* In solaris, all the threads  which are going to run java needs
     * to be bound thread so that we can reliably get the register
     * state to do GC
     */

    /* Since private/primpl.h is gone, solaris special stuff is GONE
#ifndef SOLARIS
    thread = PR_CreateThread(PR_USER_THREAD,
                             func_native_pool_thread_main,
                             NULL,
                             PR_PRIORITY_NORMAL,
                             PR_GLOBAL_THREAD,
                             PR_UNJOINABLE_THREAD,
                             func_native_pool_stacksize);
#else
    thread = _PR_CreateThread(PR_USER_THREAD,
			      func_native_pool_thread_main,
			      NULL,
			      PR_PRIORITY_NORMAL,
			      PR_GLOBAL_THREAD,
			      PR_UNJOINABLE_THREAD,
			      func_native_pool_stacksize,
			      _PR_BOUND_THREAD);
#endif
*/

    thread = PR_CreateThread(PR_USER_THREAD,
                             func_native_pool_thread_main,
                             NULL,
                             PR_PRIORITY_NORMAL,
                             PR_GLOBAL_THREAD,
                             PR_UNJOINABLE_THREAD,
                             func_native_pool_stacksize);

    if (!thread) {
        PR_Lock(func_native_pool_work_lock);
        func_native_pool_thread_count++;
        PR_Unlock(func_native_pool_work_lock);
        ereport(LOG_FAILURE, "Error creating native worker thread\n");
    }
 
    return;
}

/* ---------------------------- object_execute ---------------------------- */


/* 
 * Used by directive handlers, just executes the function for a given 
 * directive if applicable to the current client.
 */

/* XXX WHAT TO DO IF KEYSZ AND IP DNS THERE? XXX */

/*
 * This prototype uses void * for Request * in order to avoid including
 * all of req.h. 
 */
NSAPI_PUBLIC int object_execute(directive *inst, Session *sn, void *rq)
{
    if(inst->client) {
        /* In these variables, "d" is from directive, and "c" is from client */
#ifdef NET_SSL
        char *dkeysz, *ckeysz;
#endif
        char *cdns, *ddns, *cip, *dip;
        pblock *client = inst->client;
        int dapply, iapply, drev, irev;

        /* Check host header, and see if they're accessing the right host */
        char *urlhost, urlport[32], *urltmp, *dhost, *dport;
        int rv;

        /* If we are doing special client filtering for specific directives,
         * this request cannot be cached...
         */
        ((Request *)rq)->request_is_cacheable = 0;

        dhost = pblock_findval("urlhost", client);
        dport = pblock_findval("urlport", client);

        if(dhost || dport) {
            urlhost = pblock_findval("host", ((Request *) rq)->headers);
            if(urlhost) {
                if( (urltmp = strchr(urlhost, ':')) ) 
                    util_snprintf(urlport, sizeof(urlport), urltmp + 1);
                else
                    util_itoa(conf_getglobals()->Vport, urlport);

                if(dhost) {
                    if(urltmp) *urltmp = '\0';
                    rv = shexp_casecmp(urlhost, dhost);
                    if(urltmp) *urltmp = ':';
                    if(rv)
                        return REQ_NOACTION;
                }
                if(dport) {
                    if(strcmp(urlport, dport))
                        return REQ_NOACTION;
                }
            }
            else
                return REQ_NOACTION;
        }

	cip = NULL;
	cdns = NULL;

#ifdef NET_SSL
        dkeysz = pblock_findval("secret-keysize", client);
        if(dkeysz) {
            if(!security_active)
                return REQ_NOACTION;

            ckeysz = pblock_findval("secret-keysize", sn->client);

            if(shexp_casecmp(ckeysz, dkeysz))
                return REQ_NOACTION;
        }
#endif
        ddns = pblock_findval("dns", client);
        dip = pblock_findval("ip", client);

        if(ddns || dip) {
            dapply = 0; iapply = 0;

            if(ddns) {
                drev = ((ddns[0] == '*') && (ddns[1] == '~')) ? 1 : 0;
                if( (cdns = session_maxdns(sn)) ) {
                    if(!shexp_casecmp(cdns, ddns))
                        dapply = 1;
                }
                else if(drev) dapply = 1;
            }
            if(dip) {
                irev = ((dip[0] == '*') && (dip[1] == '~')) ? 1 : 0;
                cip = pblock_findval("ip", sn->client);
                if(!shexp_casecmp(cip, dip))
                    iapply = 1;
            }
            if((ddns && dip) && (dapply != iapply)) {
                switch(drev + irev) {
                  case 1:
                    if((!dapply) && (!drev) && (iapply) && (irev))
                        return REQ_NOACTION;
                  case 2:
                    return REQ_NOACTION;
                }
            }
            else if(!(dapply || iapply))
                return REQ_NOACTION;
        }

#ifdef NOACL
	/* Get client IP address if it's not there yet */
	if (!cip) {
	    cip = pblock_findval("ip", sn->client);
	}
	/* Got the IP address now? */
	if (cip != NULL) {
            IPAddr_t cipval;

	    /* Convert IP address string to binary */
	    cipval = inet_addr(cip);

	    /* If the address is not valid, don't execute the function */
	    if (cipval == (IPAddr_t)-1) {
		return REQ_NOACTION;
	    }

	    /* Check for a ipfilter parameter */
	    rv = ip_filter_check(client, cipval);
	    if (rv < 0) {
		/*
		 * A value of -2 indicates a broken filter.  Otherwise
		 * the client IP address was rejected by the filter.
		 */
		return ((rv == -2) ? REQ_ABORTED : REQ_NOACTION);
	    }
	}
	/* Get the client domain name if it's not there */
	if (!cdns) {
	    cdns = session_maxdns(sn);
	}

	/* Check for a dnsfilter parameter */
	rv = dns_filter_check(client, cdns);
	if (rv < 0) {
	    /*
	     * A value of -2 indicates a broken filter.  Otherwise
	     * there was a filter, the client's DNS name was available,
	     * and the filter rejected the client's DNS name.
	     */
	    return ((rv == -2) ? REQ_ABORTED : REQ_NOACTION);
	}
#endif /* NOACL */
    }
    return(func_exec(inst->param, sn, (Request *) rq));
}
