/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * http.c: Deal with HTTP protocol related things
 * 
 * Rob McCool
 */

/* Removed for ns security integration
#include "xp_error.h"
*/
#include "prerror.h"         /* Add for nspr20 binary */
#include "frame/http.h"

#include "netsite.h"         /* MALLOC, FREE */
#include "frame/log.h"       /* log_error */
#include "base/pblock.h"     /* parameter structure */
#include "base/cinfo.h"      /* cinfo_find */
#include "base/util.h"       /* env. functions, sprintf */
#include "frame/req.h"       /* request_handle */
#include "frame/cmstat.h"    /* set_link_header */
#include "frame/conf.h"      /* globals for server port and name */
#include "frame/servact.h"   /* servact_handle */
#include "base/time_cache.h"

#include "frame/dbtframe.h"

#include <stdarg.h>          /* Heh. */
#include <ctype.h>           /* toupper */
#include <time.h>            /* struct tm */

#include <base/fasttime.h>

#ifdef MCC_PROXY
int sockd_service     (Session *sn, Request *rq);
int sockd_read_request(Session *sn, Request *rq);
#endif

#ifndef XP_UNIX
#include <errno.h>
#endif /* XP_UNIX */

void util_uri_unescape_and_normalize(char *s);

#define http_assbackwards(rq) \
    ((!strcmp(pblock_findval("protocol",rq->reqpb),"HTTP/0.9")))

int http_keepalive_timeout = -1;
int HTTPVersion = 101;	/* Default to HTTP/1.1 */
char *sHTTPVersion = "1.1";

NSAPI_PUBLIC 
void http_set_keepalive_timeout(int secs) { http_keepalive_timeout = secs; }
NSAPI_PUBLIC 
int http_get_keepalive_timeout(void) { return http_keepalive_timeout; }


/* -------------------------- _scan_http_headers -------------------------- */


#define BAD_REQUEST(s)  \
   strcpy(t, s); \
   return REQ_ABORTED;


void _merge_header(char *name, char *val, pblock *headers) 
{
    pb_param *found;

    if(!(found = pblock_find(name,headers)))
        pblock_nvinsert(name,val,headers);
    else {
        char *newval;
        int ln = strlen(val), lv = strlen(found->value);
        register int x, y;

        /* 3 = ", " + '\0' */
        newval = (char *) REALLOC(found->value, ln + lv + 3);

        x = lv;
        newval[x++] = ','; newval[x++] = ' ';
        for(y = 0; y < ln; y++,x++)
            newval[x] = val[y];
        newval[x] = '\0';

        found->value = newval;
    }
}


#define ACCEPT_GUESS 2048


NSAPI_PUBLIC int http_scan_headers(Session *sn, netbuf *nb, char *t, 
                                   pblock *headers) 
{
    register int x,y;
    register char c;
    int nh, i, alen, asz;
    netbuf *buf = (nb ? nb : sn->inbuf);
    pb_param *acc;

    nh = 0;
    x = 0; y = -1;
    acc = NULL; alen = 0; asz = 0;

    while(1) {
        if(x >= REQ_MAX_LINE) {
            BAD_REQUEST("line too long");
        }
        if(nh > HTTP_MAX_HEADERS) {
            BAD_REQUEST("too many headers");
        }
        i = netbuf_getc(buf);

        if(i == IO_ERROR) {
#ifdef ETIMEDOUT
            if(PR_GetOSError() == ETIMEDOUT)
                strcpy(t, "timeout");
            else
#endif
                util_sprintf(t, "read failed, error is %s", 
                             system_errmsg());
            return REQ_ABORTED;
        }
        else if(i == IO_EOF) {
            BAD_REQUEST((nh ? "header missing terminator (an empty line)" : 
                "header was empty"));
        }

        c = (char) i;
        switch(c) {
          case LF:
            if((!x) || ((x == 1) && (t[0] == CR))) 
                return REQ_PROCEED;

            if(t[x-1] == CR)
                --x;
            t[x] = '\0';
            if(y == -1) {
                char *u = STRDUP(t);
                util_sprintf(t, "name without value: got line \"%s\"", u);
                FREE(u);
                return REQ_ABORTED;
            }
            while(t[y] && isspace(t[y])) ++y;

            if(!strcmp(t, "accept")) {
                if(!acc) {
                    char * valbuf = (char *) MALLOC(ACCEPT_GUESS);
                    valbuf[0] = 0;
                    acc = param_create("accept", 0);
                    acc->value = valbuf;
                    pblock_pinsert(acc, headers);
                    asz = ACCEPT_GUESS;
                    alen = x - y;
                    strcpy(acc->value, &t[y]);                    
                }
                else {
                    i = x - y + 2;
                    if((alen + i + 1) >= asz) {
                        asz += ACCEPT_GUESS;
                        acc->value = (char *) REALLOC(acc->value, asz);
                    }
                    acc->value[alen] = ','; acc->value[alen+1] = ' ';
                    strcpy(&acc->value[alen+2], &t[y]);
                    alen += i;
                }
            }
            else
                _merge_header(t, &t[y], headers);

            x = 0; y = -1;
            ++nh;
            break;
          case ':':
            if(y == -1) {
                y = x+1;
                c = '\0';
            }
          default:
            t[x++] = ((y == -1) && isupper(c) ? tolower(c) : c);
        }
    }
}


/* -------------------------- http_parse_request -------------------------- */


/*
 * Parses a request and inserts the data into rq->reqpb and rq->headers.
 * Should check for protocol later. Returns a REQ status code.
 *
 * The t string is destroyed by this routine.
 */

#ifndef MCC_PROXY	/* Proxy defines it's own */

NSAPI_PUBLIC int http_parse_request(char *t, Request *rq, Session *sn) {
    char *uri, *query, *protocol;
	int maj,min;	

    uri = t;
    if(!(*uri)) {
        /* The client jumped ship. */
        return REQ_ABORTED;
    }

    /* Split into parts */
    while(*uri && (*uri != ' '))
        uri++;

    if(!(*uri)) {
        log_error(LOG_WARN,XP_GetAdminStr(DBT_handleRequest_), sn, rq, XP_GetAdminStr(DBT_methodWithoutUri_));
        /* report error to client */
        return REQ_ABORTED;
    }

    pblock_nvinsert("clf-request", t, rq->reqpb);

    *uri++ = '\0';

	/* 
	 * It's a lot more efficient to compare integer than having to find the
	 * name value pair for the method - so we get the integer value now.
	 */
	rq->method_num  = http_set_method_num(t);

    pblock_nvinsert("method", t, rq->reqpb);  /* we can delete this later */

    query = NULL;
    for(protocol = uri; *protocol && (*protocol != ' '); protocol++) {
        if(!query) {
            if(*protocol == '?') {
                *protocol++ = '\0';
                query = protocol;
                if(*protocol == ' ')
                    break;
            }
        }
    }

    if(!(*protocol)) {
        protocol = "HTTP/0.9";
        pblock_nvinsert("protocol", "HTTP/0.9", rq->reqpb);
    }
    else {
        *protocol++ = '\0';
        pblock_nvinsert("protocol", protocol, rq->reqpb);
        rq->loadhdrs = 1;
    }

	/* Convert protocol version to number format */
	if ( sscanf(protocol,"HTTP/%d.%d",&maj,&min) == 2) {
		if (maj > 1) {	/* Hummm.. we can't handle this protocol */
	 		http_status(sn, rq, PROTOCOL_VERSION_NOT_SUPPORTED, NULL);
			return REQ_ABORTED;
		}
		rq->protv_num = 100*maj+min;
        /* Override request if in 1.0 mode */
        if ( rq->protv_num >= 101 && (HTTPVersion == 100) )
            rq->protv_num = 100;
	}
	else
		rq->protv_num = 100; /* Default to HTTP/1.0 */

    if(query)
        pblock_nvinsert("query", query, rq->reqpb);

#ifdef XP_WIN32
    util_uri_unescape_and_normalize(uri);
#else
    util_uri_unescape(uri);
#endif

	/* parse uri to obtain optional host:port */
	if ( (uri = http_parse_uri(rq, uri)) == NULL) {
      	http_status(sn, rq, PROTOCOL_BAD_REQUEST, NULL);
		return REQ_ABORTED;
	}
    pblock_nvinsert("uri", uri, rq->reqpb);

    if(rq->loadhdrs) {
        rq->loadhdrs = 0;
        if(http_scan_headers(sn, NULL, t, rq->headers) == REQ_ABORTED) {
            if(!strcmp(t, "timeout"))
                return REQ_ABORTED;
            log_error(LOG_WARN, XP_GetAdminStr(DBT_httpParseRequest_), sn, NULL, 
                      XP_GetAdminStr(DBT_whileScanningHttpHeadersS_), t);
            http_status(sn, rq, PROTOCOL_BAD_REQUEST, NULL);
            return REQ_ABORTED;
        }
    }
	if ( rq->protv_num >= 101 )  {

        /* Check for pipelined requests */
        if ( ( 
               ISMGET(rq) || 
               ISMHEAD(rq)||
               ISMTRACE(rq)||
               ISMOPTIONS(rq)
             ) &&
              (sn->inbuf->cursize - sn->inbuf->pos) > 3)  
              PIPELINED(rq) = 1;
		return http_handle_11(sn,rq);
    }

    return REQ_PROCEED;
}
#endif /* !MCOM_PROXY */


/* -------------------------- http_find_request --------------------------- */


/* 
 * Retrieves the first line of a request, and then calls appropriate 
 * request handlers.
 * 
 * This function allows HTTP lines to be terminated by either CRLF, or just 
 * LF. It also allows CR in the middle of a line for cretins who may want 
 * it.
 * 
 * Reply is set to 0 or 1 depending on whether or not an error message should
 * be sent back to the browser.
 *
 * Returns NULL upon error and a Request structure upon success.
 */

NSAPI_PUBLIC Request *http_find_request(netbuf *buf, Session *sn, int *reply)
{
    register int x, i;
    Request *rq;
    char t[REQ_MAX_LINE];

    *reply = 1;
    /* Get first line of request */

    x=0;
    while(1) {
        i = netbuf_getc(buf);
#ifdef MCC_PROXY
#define SOCKS_VERSION 0x04
	if (x == 0 && i == SOCKS_VERSION)
	    break;
#endif
        if(i == IO_ERROR) {
#ifdef ETIMEDOUT
            if(PR_GetOSError() != ETIMEDOUT)
#endif
                log_error(LOG_INFORM, XP_GetAdminStr(DBT_handleRequest_1), sn, NULL, 
                          XP_GetAdminStr(DBT_readFromSFailedErrorIsS_),
                          pblock_findval("ip",sn->client), 
                          system_errmsg());
            *reply = 0;
            return NULL;
        }

        else if((i == IO_EOF) || (i == LF)) {
            if(x && (t[x-1] == CR)) --x;
            t[x] = '\0';
            break;
        }
        else t[x++] = (char)i;
        if(x == REQ_MAX_LINE) {
            log_error(LOG_WARN, XP_GetAdminStr(DBT_handleRequest_2), sn, NULL, 
                      XP_GetAdminStr(DBT_requestTooLong_));
            return NULL;
        }
    }

    rq = request_create();

    /* Find request */
#ifdef MCC_PROXY
    if (x == 0 && i == SOCKS_VERSION)
      {
	  if (sockd_read_request(sn, rq) == REQ_ABORTED) {
	      request_free(rq);
	      return NULL;
	  }
      }
    else
#endif
    if((protocol_parse_request(t,rq,sn)) == REQ_ABORTED) {
        request_free(rq);
        return NULL;
    }

    return rq;
}


/* ----------------------------- http_dump822 ----------------------------- */


NSAPI_PUBLIC char *http_dump822(pblock *pb, char *t, int *ip, int tsz)
{
    struct pb_entry *p;
    register int x, y, z;
    int pos = *ip;

    for(x = 0; x < pb->hsize; x++) {
        p = pb->ht[x];
        while(p) {
            t[pos] = toupper(p->param->name[0]);
            for(y = pos+1, z = 1; (t[y] = p->param->name[z]); y++, z++) {
                if(y == (tsz - 4)) {
                    tsz += REQ_MAX_LINE;
                    t = (char *) REALLOC(t, tsz + 1);
                }
            }

            /* Skip internal status: information */
            if(((y - pos) == 6) && (!strncasecmp(&t[pos], "status", 6))) {
                p = p->next;
                continue;
            }

            t[y++] = ':';
            t[y++] = ' ';

            for(z = 0; (t[y] = p->param->value[z]); ++y, ++z) {
                if(y == (tsz - 2)) {
                    tsz += REQ_MAX_LINE;
                    t = (char *) REALLOC(t, tsz + 1);
                }
            }

            t[y++] = CR;
            t[y++] = LF;

            pos = y;
            p = p->next;
        }
    }
    t[pos] = '\0';
    *ip = pos;
    return t;
}


/* ------------------------- http_start_response -------------------------- */


/* Moved into its own function to keep start_response readable */

#ifdef MCC_PROXY
#define _CONNECTION_HEADER "proxy-connection"
#define _CONNECTION_REPLY \
  "Connection: keep-alive%c%cProxy-connection: keep-alive%c%c",CR,LF,CR,LF
#define _CONNECTION_CLOSE \
  "Connection: close%c%cProxy-connection: close%c%c",CR,LF,CR,LF
#else
#define _CONNECTION_HEADER "connection"
#define _CONNECTION_REPLY "Connection: keep-alive%c%c",CR,LF
#define _CONNECTION_CLOSE "Connection: close%c%c",CR,LF
#endif

NSAPI_PUBLIC void http_handle_keepalive(Request *rq, char *t, int *posptr)
{
    char *conn, sep[128], *st, *m;
    int seplen, pos = *posptr;

    if(http_keepalive_timeout < 1)
        return;

	if (rq->protv_num < CURRENT_PROTOCOL_VERSION) {
		/* Don't do it if it's not GET or HEAD. */
        if ( !ISMGET(rq) && !ISMHEAD(rq)) {
			KEEP_ALIVE(rq) = 0;
			return;
		}
   		conn = pblock_findval(_CONNECTION_HEADER, rq->headers);
   		if(conn && strcasecmp(conn, "close")) 
			KEEP_ALIVE(rq) = 1;
		else
			KEEP_ALIVE(rq) = 0;
	}
    else {
        /* Since we don't honor pipelined request, close connection after
           sending the first request. */ 
        if ( PIPELINED(rq) )
            KEEP_ALIVE(rq) = 0;
    } 

    
	/* If HTTP/1.1 - do keep-alive unless the request specified a "close" */
	if ( KEEP_ALIVE(rq) ) {
        /* Check for 304 response */
		if (rq->status_num == PROTOCOL_NOT_MODIFIED)
            pblock_nvinsert("content-length", "0", rq->srvhdrs);

        if( (rq->protv_num < CURRENT_PROTOCOL_VERSION) &&
            !pblock_find("content-length", rq->srvhdrs)) {
            /* If there is no content-length, send as multipart MIME */
            seplen = util_mime_separator(sep);

#if 0   /* XXXrobm 2.0 client doesn't support this yet */

            pos += util_sprintf(&t[pos], "Content-type: multipart/mixed; "
                                "boundary=\"%s\"%c%c", &sep[4], CR, LF);
#endif
            util_sprintf(&sep[seplen++], "%c%c", CR, LF);
#if 0   /* XXXrobm 2.0 client doesn't support this yet */
            pos += util_sprintf(&t[pos], "%s", sep);
#endif
            pblock_nvinsert("boundary", sep, rq->vars);
#if 0    /* XXXrobm 2.0 client doesn't support this yet */
            pos += util_sprintf(&t[pos], _CONNECTION_REPLY);
#endif
        }
        else {
            pos += util_sprintf(&t[pos], _CONNECTION_REPLY);
        }
    }
	else {
		pos += util_sprintf(&t[pos], _CONNECTION_CLOSE);
	}
    *posptr = pos;
}


NSAPI_PUBLIC int http_start_response(Session *sn, Request *rq) {
    char *status = pblock_findval("status", rq->srvhdrs);
    char *t, *ua, *now;
    pb_param *charset, *ct;
    int pos, l;
    PRUint32 tt;

	/* Sets "Link" header, if content manager is on and the status is 304 */
   if (rq->status_num == PROTOCOL_NOT_MODIFIED)
	 CM_set_link_header(rq);
	
    if((!rq->senthdrs) && (!(http_assbackwards(rq)))) {
        t = (char *) MALLOC(REQ_MAX_LINE);
        pos = util_sprintf(t, "HTTP/%s %s%c%c", sHTTPVersion, status, CR, LF);
#ifdef MCC_PROXY
	pos += util_sprintf(&t[pos], "Proxy-agent: %s%c%c",
			    MAGNUS_VERSION_STRING, CR, LF);
#else	/* Darn HP preprosessor doesn't understand defined() */
#ifdef MCC_HTTPD
        pos += util_sprintf(&t[pos], "Server: %s%c%c", MAGNUS_VERSION_STRING, 
                            CR, LF);
#endif
#endif
        tt = ft_time();
        now = time_cache_curr_strftime(tt);
        pos += util_sprintf(&t[pos], "Date: %s%c%c", now, CR, LF);
        FREE(now);

        /* 
         * Handle charset. Problem is that most browsers don't grok 
         * parameters on content-types. So we send it only to Mozilla/1.1 
         * or browsers that send accept-charset
         */
        if( (ct = pblock_find("content-type", rq->srvhdrs)) &&
            (charset = pblock_remove("magnus-charset", rq->srvhdrs)) )
        {
            char *acs;

            if(request_header("user-agent", &ua, sn, rq) == REQ_ABORTED) {
                FREE(t);
                return REQ_ABORTED;
            }
            if(request_header("accept-charset", &acs, sn, rq) == REQ_ABORTED) {
                FREE(t);
                return REQ_ABORTED;
            }

            if(util_is_mozilla(ua, "1", "1") || acs) {
                l = strlen(ct->value);
                ct->value = (char *) REALLOC(ct->value, l + 10 + 
                                             strlen(charset->value) + 1);
                util_sprintf(&ct->value[l], "; charset=%s", charset->value);
            }
            param_free(charset);
        }

        t = http_dump822(rq->srvhdrs, t, &pos, REQ_MAX_LINE);

        /* BEWARE:  Cacheing the request headers is tricky business;
         *          actually is a mess.  But what is a poor cache to do?
         *          The purpose of the cache was to avoid calling all of
         *          this NSAPI stuff.  So, we had to cache the headers.
         *          But any header which has to do with the connection
         *          rather than the data itself MUST NOT BE CACHED!
         *
         *          Therefore, notice the convenient place where we 
         *          snapshot the headers; just before we do keepalive
         *          stuff...
         */
        if (rq->request_is_cacheable) {
            rq->cached_headers = t;
            rq->cached_headers_len = pos;
        }

        /* Handle multiple requests. */
       	http_handle_keepalive(rq, t, &pos);

        t[pos++] = CR;
        t[pos++] = LF;

        if(net_write(sn->csd, t, pos) == IO_ERROR) {
            if(PR_GetError() != PR_CONNECT_RESET_ERROR) {
                log_error(LOG_WARN, XP_GetAdminStr(DBT_startHttpResponse_), sn, rq, 
                          XP_GetAdminStr(DBT_writeFailedS_), system_errmsg());
            }
            FREE(t);
            return REQ_EXIT;
        }

        /* If this is cacheable, save t for the cache to use.  Ugly? XXXMB*/
        if (!rq->request_is_cacheable)
            FREE(t);
#ifdef MCC_PROXY
	pblock_nninsert("p2c-hl", pos, rq->vars);
#endif
   }
    rq->senthdrs = 1;
	if (ISMHEAD(rq))
        return REQ_NOACTION;
    else 
        return REQ_PROCEED;
}


/* ---------------------------- http_hdrs2env ----------------------------- */


NSAPI_PUBLIC char **http_hdrs2env(pblock *pb)
{
    char *t, *n, **env;
    struct pb_entry *p;
    pb_param *pp;
    register int x, y, z;
    int pos, ts, ln;

    /* Find out how many there are. */
    for(x = 0, y = 0; x < pb->hsize; x++) {
        p = pb->ht[x];
        while(p) {
            ++y;
            p = p->next;
        }
    }

    env = util_env_create(NULL, y, &pos);

    ts = 1024;
    t = (char *) MALLOC(ts);

    for(x = 0; x < pb->hsize; x++) {
        p = pb->ht[x];

        while(p) {
            pp = p->param;

            ln = strlen(pp->name) + 6;

            if(ln >= ts) {
                ts = ln;
                t = (char *) REALLOC(t, ts);
            }
            n = pp->name;

            /* Skip authorization for CGI */
            if(strcasecmp(n, "authorization")) {
                if(strcasecmp(n, "content-type") &&
                   strcasecmp(n, "content-length")) 
                {
                    strncpy(t, "HTTP_", 5);
                    z = 5;
                }
                else
                    z = 0;

                for(y = 0; n[y]; ++y, ++z)
                    t[z] = (n[y] == '-' ? '_' : toupper(n[y]));
                t[z] = '\0';

                env[pos++] = util_env_str(t, pp->value);
            }
            p = p->next;
        }
    }
    env[pos] = NULL;
    FREE(t);
    return env;
}


/* ---------------------------- http_set_finfo ---------------------------- */


NSAPI_PUBLIC int http_set_finfo(Session *sn, Request *rq, struct stat *finfo)
{
    char *t, lm[HTTP_DATE_LEN], *ims;
    struct tm *tms = 0;
   	struct tm tmss;
	char *if_modified_since = pblock_findval("if-modified-since",rq->headers);
	/* HTTP/1.1 stuff */
    char *if_unmodified_since = pblock_findval("if-unmodified-since", rq->headers);
    char *if_match = pblock_findval("if-match", rq->headers);
    char *if_nonematch = pblock_findval("if-none-match", rq->headers);
    char *etagp, etag[MAX_ETAG];
	short int	use_local = 1;

    if (rq->status_num == PROTOCOL_OK) {
   	tms = system_gmtime(&finfo->st_mtime, &tmss);

	/* Check for weak/strong validator if HTTP/1.1 or greater */
	if (rq->protv_num >= CURRENT_PROTOCOL_VERSION) {
		/*
		 * Create an etag out of the metadata (ino#, size, mtime) to obtain
		 * ETag uniqueness.
		 */
		if (finfo->st_mode != 0) {
			util_sprintf(etag, "W/\"%lx-%lx-%lx\"", finfo->st_ino,
				finfo->st_size, finfo->st_mtime);
		}
		else {
			util_sprintf(etag, "W/\"%lx\"", finfo->st_mtime);
		}
		 
		/*
		 * Heuristic for weak/strong validator: if the request was made within
		 * a second of the last-modified date, then we send a weak ETag.
		 */
		etagp = etag + ((rq->req_start - finfo->st_mtime > 1) ? 2 : 0);
		pblock_nvinsert("etag", etagp, rq->srvhdrs);

	}	

    if(if_modified_since && util_later_than(tms, if_modified_since)) {
     	http_status(sn, rq, PROTOCOL_NOT_MODIFIED, NULL);
       	return REQ_ABORTED;
    }
	else if (if_unmodified_since) {
		if (util_later_than(tms, if_unmodified_since)) 
     		http_status(sn, rq, PROTOCOL_PRECONDITION_FAIL, NULL);
		else
     		http_status(sn, rq, PROTOCOL_NOT_MODIFIED, NULL);
		return REQ_ABORTED;
	}
	else if (if_nonematch) {
		if (if_nonematch[0] == '*' || strstr(if_nonematch, etagp)) {
			if ( !ISMGET(rq) && !ISMHEAD(rq) ) 
     			http_status(sn, rq, PROTOCOL_PRECONDITION_FAIL, NULL);
			else 
     			http_status(sn, rq, PROTOCOL_NOT_MODIFIED, NULL);
			return REQ_ABORTED;
		}
		/* Send a local copy */
	}
	else if (if_match) {
		if (!(if_match[0] == '*' || strstr(if_match, etagp))) 
			http_status(sn, rq, PROTOCOL_PRECONDITION_FAIL, NULL);
		else
     		http_status(sn, rq, PROTOCOL_NOT_MODIFIED, NULL);
        return REQ_ABORTED;
	}


    /* check return code? */
    if ( (t = time_cache_strftime(finfo->st_mtime)) == NULL) {
        t = (char *)MALLOC(HTTP_DATE_LEN+1);
        t[HTTP_DATE_LEN] = '\0';
        (void)util_strftime(t, HTTP_DATE_FMT, tms);
    }
    pblock_nvinsert("last-modified", t, rq->srvhdrs);
    FREE(t);
    }

    util_itoa((int)finfo->st_size, lm);
    pblock_nvinsert("content-length", lm, rq->srvhdrs);

    return REQ_PROCEED;
}


/* ----------------------------- http_status ------------------------------ */


NSAPI_PUBLIC void http_status(Session *sn, Request *rq, int n, char *r)
{
    char t[256];

    if((n < 0) || (n > 999)) {
        log_error(LOG_MISCONFIG, XP_GetAdminStr(DBT_httpStatus_), sn, rq, 
                  XP_GetAdminStr(DBT_DIsNotAValidHttpStatusCode_), n);
        n = PROTOCOL_SERVER_ERROR;
    }
	rq->status_num = n;
    util_itoa(n, t);  /* take our chances with dummies */
    if(!r) {
        switch(n) {
		  case PROTOCOL_CONTINUE:				/* HTTP/1.1 */
			r = "Continue";
			break;
		  case PROTOCOL_SWITCHING:				/* HTTP/1.1 */
			r = "Switching Protocols";
			break;
          case PROTOCOL_OK:
            r = "OK";
            break;
          case PROTOCOL_CREATED:
            r = "Created";
            break;
          case PROTOCOL_NO_RESPONSE:
            r = "No response";
            break;
          case PROTOCOL_PARTIAL_CONTENT:
            r = "Partial content";
            break;
          case PROTOCOL_REDIRECT:
            r = "Moved Temporarily";
            break;
          case PROTOCOL_NOT_MODIFIED:
            r = "Use local copy";
            break;
          case PROTOCOL_BAD_REQUEST:
            r = "Bad request";
            break;
          case PROTOCOL_UNAUTHORIZED:
            r = "Unauthorized";
            break;
          case PROTOCOL_FORBIDDEN:
            r = "Forbidden";
            break;
          case PROTOCOL_NOT_FOUND:
            r = "Not found";
            break;
	  	  case PROTOCOL_PROXY_UNAUTHORIZED:
	    	r = "Proxy authorization required";
	    	break;
          case PROTOCOL_SERVER_ERROR:
            r = "Server Error";
            break;
          case PROTOCOL_NOT_IMPLEMENTED:
            r = "Not Implemented";
            break;
		  case PROTOCOL_METHOD_NOT_ALLOWED:				/* HTTP/1.1 */
			r = "Method Not Allowed";
			break;
		  case PROTOCOL_LENGTH_REQUIRED:				/* HTTP/1.1 */
			r = "Length Required";
			break;
		  case PROTOCOL_CONFLICT:
			r = "Conflict";								/* HTTP/1.1 */
			break;
		  case PROTOCOL_PRECONDITION_FAIL:				/* HTTP/1.1 */
			r = "Precondition Failed";
			break;
		  case PROTOCOL_ENTITY_TOO_LARGE:				/* HTTP/1.1 */
			r = "Request Entity Too Large";
			break;
		  case PROTOCOL_URI_TOO_LARGE:					/* HTTP/1.1 */
			r = "Request-URI Too Long";
			break;
		  case PROTOCOL_VERSION_NOT_SUPPORTED:			/* HTTP/1.1 */
			r = "HTTP Version Not Supported";
			break;
          default:
            r = "Unknown reason";
            break;
        }
    }
    else {
        if(strlen(r) > 250)
            r[250] = '\0';
    }
    t[3] = ' ';
    strcpy(&t[4], r);

    pblock_nvinsert("status", t, rq->srvhdrs);
}


/* ------------------------- http_finish_request -------------------------- */


NSAPI_PUBLIC void http_finish_request(Session *sn, Request *rq) 
{
    pb_param *pp;

#ifdef DAEMON_STATS
    if (rq) {
	char *status = pblock_findval("status", rq->srvhdrs);
	char *bytes  = pblock_findval("content-length", rq->srvhdrs);

	child_status2(status ? atoi(status) : -1,
		      bytes  ? atol(bytes)  : 0L);
    }
#endif

    if((sn->csd_open == 1) && 
       rq && (rq->senthdrs) && (http_keepalive_timeout > 0)) 
    {
        /* See if we decided to keep it alive before. */
        if ( KEEP_ALIVE(rq) ) {
            /* Terminate the message. boundary is set by start_response */
            if( (rq->protv_num < CURRENT_PROTOCOL_VERSION) &&
                (pp = pblock_remove("boundary", rq->vars)) ) {
#if 0   /* XXXrobm 2.0 client doesn't support this yet */
                pp->value = (char *) REALLOC(pp->value, strlen(pp->value) + 3);
                t = pp->value;
                util_sprintf(&t[strlen(t) - 2], "--%c%c", CR, LF);
                if(net_write(sn->csd, t, strlen(t)) == IO_ERROR) {
                    param_free(pp);
                    goto close;
                }
#else
                param_free(pp);
                goto close;
#endif
            }
            /* sn->csd_open is 1, so the session will continue */
            sn->inbuf->rdtimeout = http_keepalive_timeout;
            return;
        }
    }
  close:

#ifdef XP_WIN32
    if ( net_shutdown(sn->csd, 1) < 0 ) {
        ereport(LOG_FAILURE, XP_GetAdminStr(DBT_unableToCloseSocketForWriting_));
    } else {
	    while(1) {
#define CLEANUP_BUF_SIZE	64
			int rv;
char buf[CLEANUP_BUF_SIZE];
			if ( (rv = net_read(sn->csd, buf, CLEANUP_BUF_SIZE, NET_ZERO_TIMEOUT)) <= 0  ) {
				/* rv == 0 is normal; if -1, socket is presumably closed */
				break;
			}
			else {
				/* read data from socket */
				rv = 0;  /* a place to set a breakpoint */
			}
		}
	}
#endif

    if(net_close(sn->csd) == -1)
        log_error(LOG_FAILURE, XP_GetAdminStr(DBT_finishRequest_), sn, rq, XP_GetAdminStr(DBT_closeFailedS_),
                  system_errmsg());
    sn->csd_open = 0;
}


/* ------------------------- http_handle_session -------------------------- */


NSAPI_PUBLIC void http_handle_session(Session *sn) 
{
    /* robm The threaded Web servers handle keepalives at the daemon level to
       avoid tying up a thread while waiting for the next request, and to keep 
       centralized information about current connections.
     */
#if ! (defined(THREAD_ANY) && defined(MCC_HTTPD))
    /* Keepalive makes csd stay open sometimes */
    while(sn->csd_open)
#endif
        servact_handle(sn);
#if (defined(THREAD_ANY) && defined(MCC_HTTPD))
    /* Make sure that if keepalives are off, the socket doesn't get back */
    if((!http_keepalive_timeout) && sn->csd_open) {
        (void) net_close(sn->csd);
        sn->csd_open = 0;
    }
#endif
}


/* ----------------------------- http_uri2url ----------------------------- */


static char *_http_uri2url_host(const char *host, const char *prefix, const char *suffix)
{
    register char *u;

#ifdef NET_SSL
    if(security_active) {
        u = (char *) MALLOC(strlen(HTTPS_URL) + 3 + strlen(host) +
                            17 + strlen(prefix) + strlen(suffix) + 1);

        if(strchr(host, ':') || (port == HTTPS_PORT))
            util_sprintf(u, "%s://%s%s%s", HTTPS_URL, host, 
                         prefix, suffix);
        else
            util_sprintf(u, "%s://%s:%d%s%s", HTTPS_URL, host, port, 
                         prefix, suffix);
    }
    else {
#endif
        u = (char *) MALLOC(strlen(HTTP_URL) + 3 + strlen(host) + 
                            17 + strlen(prefix) + strlen(suffix) + 1);

        if(strchr(host, ':') || (port == HTTP_PORT))
            util_sprintf(u, "%s://%s%s%s", HTTP_URL, host, 
                         prefix, suffix);
        else
            util_sprintf(u, "%s://%s:%d%s%s", HTTP_URL, host, port, 
                         prefix, suffix);
#ifdef NET_SSL
    }
#endif
    return u;
}

NSAPI_PUBLIC char *http_uri2url(const char *prefix, const char *suffix)
{
    return _http_uri2url_host(server_hostname, prefix, suffix);
}


NSAPI_PUBLIC char *http_uri2url_dynamic(const char *prefix, const char *suffix,
                                        Session *sn, Request *rq)
{
    char *host;

    host = pblock_findval("host", rq->headers);
    if(!host) {
        host = sn->inbuf->address;
        if(!(*host)) host = server_hostname;
    }
    return _http_uri2url_host(host, prefix, suffix);
}

/*
 * Handle OPTIONS method by sending allowed methods.
 */
NSAPI_PUBLIC int http_handle_options(Session *sn, Request *rq)
{
	http_status(sn, rq, PROTOCOL_OK, NULL);
    pblock_nvinsert("content-length", "0", rq->srvhdrs);
	pblock_nvinsert("allow", http_allowed_methods(rq), rq->srvhdrs);

    return REQ_PROCEED;
} 


/*
 * Handle TRACE method by echoing the request.
 */
NSAPI_PUBLIC int http_handle_trace(Session *sn, Request *rq)
{
    char sdigit[HTTP_DATE_LEN];  /* length in string */
    char *outbuf = (char*)MALLOC(REQ_MAX_LINE);
    int  pos=0;


    if ( sn->csd_open != 1 )
        return REQ_ABORTED;

    /* rebuild the  request */
    pos += util_sprintf(&outbuf[pos], "%s%c%c", 
             pblock_findval("clf-request",rq->reqpb), CR, LF);
    (void)http_dump822(rq->headers, outbuf, &pos, REQ_MAX_LINE - pos);
    pos += util_sprintf(&outbuf[pos], "%c%c", CR, LF);

    util_itoa(pos, sdigit);

	http_status(sn, rq, PROTOCOL_OK, NULL);
    pblock_nvinsert("content-length", sdigit, rq->srvhdrs);
    pblock_nvinsert("content-type", "message/http", rq->srvhdrs);
    http_start_response(sn, rq);

    /* Now just send it back */
    if( (net_write(sn->csd, outbuf, pos)  == IO_ERROR) || !KEEP_ALIVE(rq) ) {
        net_close(sn->csd);
        sn->csd_open = 0;
        return REQ_ABORTED;
    }
    return REQ_PROCEED;
} 

/*
 * Handle HTTP/1.1 specific stuff. It also does some header validation, so
 * can return an REQ_ABORTED when it detects inconsistent request/header 
 * values. 
 */
NSAPI_PUBLIC int http_handle_11( Session *sn, Request *rq)
{
	char *tval;
	char *host, *sname = server_hostname;

    if ( rq->protv_num > CURRENT_PROTOCOL_VERSION ) {
        http_status(sn, rq, PROTOCOL_VERSION_NOT_SUPPORTED, NULL);
        return REQ_ABORTED;
    }


	/* Must now specify a "Host" header if abs_uri was not used  */
	if (!ABS_URI(rq)){
		if ((host =  pblock_findval("host", rq->headers)) == NULL){
			http_status(sn, rq, PROTOCOL_BAD_REQUEST, NULL);
			return REQ_ABORTED;
		}
	}
	else {
		/* We'll need to replace host with hostname extractred from */
		/* abs_uri */
		if (pblock_replace("host",rq->hostname, rq->headers) == NULL) {
			http_status(sn, rq, PROTOCOL_BAD_REQUEST, NULL);
			FREE(rq->hostname);
			rq->hostname = NULL;
			return REQ_ABORTED;
		}
	}

	/* 
	 * By default, connection is persistent unless the client explicitly 
	 * say otherwise - "close" */
	tval = pblock_findval("connection", rq->headers);
	if (tval && !strcmp(tval,"close"))
		KEEP_ALIVE(rq) = 0;
	else
		KEEP_ALIVE(rq) = 1;

	/* Set transfer-encoding  "chunked" flag */
    tval = pblock_findval("transfer-encoding",rq->headers);
	if (tval) {
        if (strcmp(tval,"chunked")) {
            protocol_status(sn, rq, PROTOCOL_NOT_IMPLEMENTED, NULL);
            log_error(LOG_FAILURE, "http_handle_11",sn,rq,
                "Transfer-Encoding <%s> is not currently implemented.",
                tval);
            return REQ_ABORTED;
        }
		CHUNKED(rq) = 1;
	}
	else
		CHUNKED(rq) = 0;

	/* Validate method */
	if (rq->method_num == -1) {
	 	http_status(sn, rq, PROTOCOL_NOT_IMPLEMENTED, NULL);
		pblock_nvinsert("allow", http_allowed_methods(rq), rq->srvhdrs);
		return REQ_ABORTED;
	}

    if ( ISMOPTIONS(rq) )
       return http_handle_options(sn, rq);

    if ( ISMTRACE(rq) )
       return http_handle_trace(sn, rq);

	return REQ_PROCEED;
}


static char *method_list[] = {
                        "HEAD",
                        "GET",
                        "PUT",
                        "POST",
                        "DELETE",
                        "TRACE",
                        "OPTIONS",
                        "MOVE",
                        "INDEX",
                        "MKDIR",
                        "RMDIR",
                        "COPY"
                        };

/*
 * Routine to return the numeric value of the current method.  An -1
 * is returned for an invalid method.
 */                           
NSAPI_PUBLIC int http_set_method_num(char *t)
{
	int i;

	for (i=0;i<METHOD_MAX;i++) {
		if (strcmp(method_list[i],t) == 0)
			return(i);
	}
	return (-1);
}

/*
 * Return the string value of the method.
 */
NSAPI_PUBLIC char * http_method_string(int method_num)
{
	if (method_num >= METHOD_MAX || method_num < 0)
		return NULL;
	else
		return method_list[method_num];
}



/*
 * Routine to scan the Request-URI for an optional host:port.
 * Returns a pointer to the absolute path.
 */
NSAPI_PUBLIC char *
http_parse_uri(Request *rq, char *uri)
{
	char *host, *name, *tptr;
	int port;

	/* Does it have full URLs?? */
	if ( strncmp(uri,"http://", 7) )
		return uri;		/* NO */

	host = STRDUP(uri+7);  /* point to host name */
	tptr = strchr(host, '/');
	if (tptr) 
		*tptr = '\0';
	/* Adjust URI pointer */
	uri = uri+strlen(host)+7;

	/* Get port */
	tptr = strchr(host, ':');
	if (tptr) {
		port = atoi(tptr+1);
		*tptr = '\0';
	
		/* Validate port */
		if (port != server_portnum){
			FREE(host);
			return NULL;
		}
	}
	
	/* Validate hostname */
	if ( !strcasecmp(host, server_hostname)) {
		FREE(host);
		return NULL;
	}
	/* SHOULD also check to see if host is in the form of an IP address */
	/* TBD */

	ABS_URI(rq)=1;
	rq->hostname = host;
	return uri;
}

/*
 * Called once after server initialization to set allowable methods on
 * this server.
 */
NSAPI_PUBLIC void http_set_allowed_methods()
{
	/* Minimum  support */
	AllowedMethods |= (1 << METHOD_HEAD);
	AllowedMethods |= (1 << METHOD_GET);
	AllowedMethods |= (1 << METHOD_PUT);
	AllowedMethods |= (1 << METHOD_POST);

	/* Do we want to publish the NS extensions as well ??? */
}

/* 
 * Returns a string of allowable methods.
 */
NSAPI_PUBLIC char * http_allowed_methods(Request *rq)
{
	static char allowed[1024];
	int i;

	memset(allowed,0,1024);

	for (i=0;i<METHOD_MAX;i++) {
		if ( rq->allowed & (1 << i) ) { 
			if (i != 0)
				strcat(allowed, ", ");
			strcat(allowed, method_list[i]);
		}
	}
	return &allowed[0];
}

