/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
 /*
 * httpact.c: Defines the actions which compose the NSAPI for the HTTP
 * servers
 *
 * Rob McCool
 */


#include "netsite.h"
#include "base/pblock.h"    /* various */
#include "base/session.h"   /* Session struct */
#include "base/util.h"
#include <base/plist.h>
#include "base/shexp.h"     /* shexp_casecmp */
#include "libaccess/aclstruct.h" /* ACD_ALWAYS def */
#include <libaccess/acl.h>
#include <libaccess/aclproto.h>
#include "frame/httpact.h"
#include "frame/req.h"      /* Request structure, objset, object, etc */
#include "frame/log.h"      /* error logging */
#include "frame/func.h"     /* func_exec */
#include "frame/protocol.h" /* protocol_finish_request */
#include "frame/conf.h"     /* std_os global */
#include <frame/acl.h>

#ifdef PUMPKIN_HOUR
#include <signal.h>
#ifdef XP_WIN32
#include "nt/regparms.h"
#endif /* XP_WIN32 */
#endif

#include "frame/dbtframe.h"

int error_report(Session *sn, Request *rq);

#ifdef MCC_PROXY
/* Awful hack for the proxy to make it work with libnet */
Request *glob_rq = NULL;
Session *glob_sn = NULL;

int record_proxy_access(pblock *, Session *, Request *);
int sockd_service(Session *sn, Request *rq);
#include <libproxy/putil.h>
#endif /* MCC_PROXY */

int _perform_pathchecks(Session *sn, Request *rq, httpd_object *obj, int doacl);


/* ------------------------------------------------------------------------ */
/* ----------------------- Directive execution code ----------------------- */
/* ------------------------------------------------------------------------ */



/* ------------------------- _directive_applyone -------------------------- */



/* Apply the first applicable instance of a directive and return */

int _directive_applyone(dtable *d, Session *sn, Request *rq) 
{
    register int x;

    for(x = 0; x < d->ni; x++) {
        switch(object_execute(&d->inst[x], sn, (void *)rq)) {
          case REQ_ABORTED:
            return REQ_ABORTED;
          case REQ_PROCEED:
            return REQ_PROCEED;
        }
    }
    return REQ_NOACTION;
}


/* ------------------------- _directive_applyall -------------------------- */


int _directive_applyall(dtable *d, Session *sn, Request *rq)
{
    register int x;

    for(x = 0; x < d->ni; x++) {
        if(object_execute(&d->inst[x], sn, (void *) rq) == REQ_ABORTED)
            return REQ_ABORTED;
    }
    return REQ_PROCEED;
}


/* ------------------------------------------------------------------------ */
/* ----------------------- AuthTrans and NameTrans ------------------------ */
/* ------------------------------------------------------------------------ */


/* --------------------------- _object_findnext --------------------------- */


/*
 * object_findnext finds the object configured to follow the given object,
 * and stores the variables in rq->vars. It returns REQ_PROCEED if more 
 * objects should be processed, or REQ_NOACTION if it did not find any
 * further objects to process. If something bad happens, REQ_ABORTED is
 * returned.
 *
 * Handles all DIRECTIVE_CONSTRUCT type directives such as NameTrans and
 * AuthType.
 */

int _object_findnext(Session *sn, Request *rq, httpd_object *obj) 
{
    if(_directive_applyone(&obj->dt[0], sn, rq) == REQ_ABORTED)
        return REQ_ABORTED;
    return _directive_applyone(&obj->dt[1], sn, rq);
}


/* --------------------------- request_uri2path --------------------------- */


NSAPI_PUBLIC int servact_uri2path(Session *sn, Request *rq) 
{
    conf_global_vars_s *vars = conf_getglobals();
    char *name = vars->Vroot_object;
    char *ppath = pblock_findval("uri", rq->reqpb);
    httpd_objset *standard_os = vars->Vstd_os;
    httpd_object *obj;
    pb_param *pp;

    /* Use NameTrans directives to find object set */

    /* Initially, look for default object. */
    pblock_nvinsert("name", name, rq->vars);
    pblock_nvinsert("ppath", ppath, rq->vars);

    rq->os = objset_create();

    while(1) {
        /* Name takes precedence over ppath */
        if(name) {
            if(!(obj = objset_findbyname(name, rq->os, standard_os))) {
                log_error(LOG_MISCONFIG, XP_GetAdminStr(DBT_processUriObjects_), sn, rq, 
                          XP_GetAdminStr(DBT_cannotFindTemplateS_), name);
                return REQ_ABORTED;
            }
            name = NULL;
            param_free(pblock_remove("name", rq->vars));
        }
        else if(ppath) {
            if(!(obj = objset_findbyppath(ppath, rq->os, standard_os)))
                break;
        }
        else
            break;

        objset_add_object(obj, rq->os);

        if(_object_findnext(sn, rq, obj) == REQ_ABORTED)
            return REQ_ABORTED;

        name = pblock_findval("name", rq->vars);
        ppath = pblock_findval("ppath", rq->vars);
    }

    /* Have path, will travel. */
    if( (pp = pblock_remove("ppath", rq->vars)) ) {
        strcpy(pp->name, "path");
        pblock_pinsert(pp, rq->vars);
    }
    else {
        log_error(LOG_MISCONFIG, XP_GetAdminStr(DBT_processUriObjects_1), sn, rq, 
                  XP_GetAdminStr(DBT_noPartialPathAfterObjectProcessi_));
        return REQ_ABORTED;
    }
    return REQ_PROCEED;
}


/* ------------------------ servact_translate_uri ------------------------- */


NSAPI_PUBLIC char *servact_translate_uri(char *uri, Session *sn)
{
    Request *rq = request_restart_internal(uri, NULL);
    char *path;

    if(servact_uri2path(sn, rq) == REQ_ABORTED)
        path = NULL;
    else
        path = STRDUP(pblock_findval("path", rq->vars));

    request_free(rq);
    return path;
}


/* ------------------------ ACL_UriToAclList ------------------------- */


/* RETURNS:
 *	0 	on success
 *	-1 	on failure
 *	and stuffs the ACL List handle in *acllistp.
 */
NSAPI_PUBLIC int ACL_UriToAclList(char *uri, Session *sn, ACLListHandle_t **acllistp)
{
    
    if (!ACL_CacheCheckGet(uri, acllistp)) {    

        Request *rq = request_restart_internal(uri, NULL);
	// populate host into rq->headers since http_uri2url_dynamic expects it
	pblock_nvinsert("host", "", rq->headers);
	// Don't allow pcheck_find_index to add index.html to the uri.
	pblock_nvinsert("query", "", rq->reqpb);

        if(servact_uri2path(sn, rq) == REQ_ABORTED) {
            request_free(rq);
            return -1;
	}

//====================

	{		/* extracted from servact_pathchecks */
		httpd_objset *os;
		register int x;
		httpd_object *obj;
		char *luri, *uri, *path, *method;
		int critical = 0;
		NSErr_t *errp = 0;

		os = rq->os;

		/* XXXrobm PathCheck functions need ACL information. However, the ACL 
		   functions don't necessarily execute before ones that need the info.
		   The "real" solution is to make a new directive between NameTrans and
		   PathCheck that does the ACLs, but we're too far along into 2.0 to
		   make such a change. So I've hard-coded it so that check-acl gets called
		   first. */

		/* If we cached the ACL list for this URI, skip the check-acl
		 * pathchecks.
		 */
		if (acl_root_30) {
			uri = pblock_findval("uri", rq->reqpb);
			method = pblock_findval("method", rq->reqpb);
			if (ACLUriCacheOn) {	/* Don't cache if any virtual servers */
				if (strcmp(method, "GET") == 0) {
					if (!ACL_CacheCheckGet(uri, &rq->acllist)) {
					critical = 1;
					}
				} else {
					if (!ACL_CacheCheck(uri, &rq->acllist)) {
					critical = 1;
					}
				}
			} else {	/* Always rebuild the ACL List if virtual servers */
				critical = 1;
			}
		} else {
			rq->acllist = ACL_LIST_NO_ACLS;
		}

		if (acl_root_30 && critical) {
			for(x = 0; x < os->pos; ++x) {
				obj = os->obj[x];

				if(_perform_pathchecks(sn, rq, os->obj[x], 1) == REQ_ABORTED) {
					ACL_ListDestroy(errp, rq->acllist);
					return REQ_ABORTED;
				}
			}
		}

		if (acl_root_30) {
			/* After getting the PathCheck ACLs, look for any affiliated with any
			 * component of the URI.  Finally look for an FS pathname-bound ACLs.
			 */
			if (critical) {
				/* Use the long URI so we locate ACLs bound to implicit filenames
			 	* like index.html.
			 	*/
				luri = pblock_findval("luri", rq->reqpb);
				if (!luri)	/* only set if there was a find-index that took */
				luri = uri;
				ACL_GetPathAcls(luri, &rq->acllist, "uri=", (ACLListHandle_t *)acl_root_30);
				path = pblock_findval("path", rq->vars);
				ACL_GetPathAcls(path, &rq->acllist, "path=", (ACLListHandle_t *)acl_root_30);

				if (ACLUriCacheOn) {	/* Don't URI cache if any Virtual Servers */
					if (strcmp(method, "GET") == 0) {
						ACL_CacheEnterGet(uri, &rq->acllist);
					} else {
						ACL_CacheEnter(uri, &rq->acllist);
					}
				} else {			/* Only ACL List cache please */
					ACL_ListHashEnter(&rq->acllist);
				}
			}
		}
	}

//=====================


        if (!ACLUriCacheOn) {	/* Do everything manually */
            *acllistp = rq->acllist;
            request_free(rq);
            if (!*acllistp) {
                *acllistp = ACL_LIST_NO_ACLS;
            }
            if (*acllistp != ACL_LIST_NO_ACLS) {
                if (!ACL_ListHashCheck(acllistp)) {
                    return -1;
                }
            }
            return 0;
        }

        request_free(rq);
    }
	else
		return 0;

	// Didnt get it in cache the first time
	if (!ACL_CacheCheckGet(uri, acllistp)) 
		return -1;
	else
    	return 0;
}


/* ------------------------------------------------------------------------ */
/* ------------------------------ PathCheck ------------------------------- */
/* ------------------------------------------------------------------------ */


/* 
   doacl defines if we should call check-acl only. 1 == do only it. 
   0 == do everything else 
 */
int _perform_pathchecks(Session *sn, Request *rq, httpd_object *obj, int doacl)
{
    register int x, isacl;
    register char *fn;
    dtable *d = &obj->dt[2];

    for(x = 0; x < d->ni; x++) {
        fn = pblock_findval("fn", d->inst[x].param);
        if(fn) {
            isacl = (!strcmp(fn, "check-acl") ? 1 : 0);
            if(doacl != isacl)
                continue;
        }
        if(object_execute(&d->inst[x], sn, (void *) rq) == REQ_ABORTED)
            return REQ_ABORTED;
    }
    return REQ_PROCEED;
}


NSAPI_PUBLIC int servact_pathchecks(Session *sn, Request *rq)
{
    httpd_objset *os;
    register int x;
    httpd_object *obj;
    char *luri, *uri, *path, *method;
    int critical = 0;
    NSErr_t *errp = 0;

    os = rq->os;

    /* XXXrobm PathCheck functions need ACL information. However, the ACL 
       functions don't necessarily execute before ones that need the info.
       The "real" solution is to make a new directive between NameTrans and
       PathCheck that does the ACLs, but we're too far along into 2.0 to
       make such a change. So I've hard-coded it so that check-acl gets called
       first. */

    /* If we cached the ACL list for this URI, skip the check-acl
     * pathchecks.
     */
    if (acl_root_30) {
        uri = pblock_findval("uri", rq->reqpb);
        method = pblock_findval("method", rq->reqpb);
        if (ACLUriCacheOn) {	/* Don't cache if any virtual servers */
            if (strcmp(method, "GET") == 0) {
                if (!ACL_CacheCheckGet(uri, &rq->acllist)) {
	            critical = 1;
                }
            } else {
                if (!ACL_CacheCheck(uri, &rq->acllist)) {
	            critical = 1;
                }
            }
        } else {	/* Always rebuild the ACL List if virtual servers */
            critical = 1;
        }
    } else {
        rq->acllist = ACL_LIST_NO_ACLS;
    }

    if (acl_root_30 && critical) {
        for(x = 0; x < os->pos; ++x) {
            obj = os->obj[x];

            if(_perform_pathchecks(sn, rq, os->obj[x], 1) == REQ_ABORTED) {
                ACL_ListDestroy(errp, rq->acllist);
                return REQ_ABORTED;
	    }
        }
    }

    if (!(acl_root_30)) {
        /* See if any 2.0 ACLs applied */
        if (rq->aclstate != 0) {
	    pblock * npb = pblock_create(4);
    
	    (void)pblock_nvinsert("fn", "acl-state", npb);
	    rq->aclstate |= ACD_ALWAYS;
	    x = func_exec(npb, sn, rq);
	    pblock_free(npb);
    
	    if (x == REQ_ABORTED) {
	        return REQ_ABORTED;
	    }
        }
    }

    /* Perform the non-acl pathchecks */
    for(x = 0; x < os->pos; ++x) {
        obj = os->obj[x];
        if(_perform_pathchecks(sn, rq, os->obj[x], 0) == REQ_ABORTED) {
	    if (critical) {
		ACL_ListDestroy(errp, rq->acllist);
	    }
            return REQ_ABORTED;
	}
        /* nsconfig stuff modifies os. Make sure we stay on the right object */
        for(x = 0; x < os->pos; ++x)
            if(os->obj[x] == obj)
                break;
    }

    /* Finally, evaluate any ONE ACLs - must be done *after* all flavors of
     * pathchecks so that the pblock/"path" value is final.
     */
    if (acl_root_30) {
        /* After getting the PathCheck ACLs, look for any affiliated with any
         * component of the URI.  Finally look for an FS pathname-bound ACLs.
         */
        if (critical) {
	    /* Use the long URI so we locate ACLs bound to implicit filenames
	     * like index.html.
	     */
	    luri = pblock_findval("luri", rq->reqpb);
	    if (!luri)	/* only set if there was a find-index that took */
		luri = uri;
            ACL_GetPathAcls(luri, &rq->acllist, "uri=", (ACLListHandle_t *)acl_root_30);
	    path = pblock_findval("path", rq->vars);
	    ACL_GetPathAcls(path, &rq->acllist, "path=", (ACLListHandle_t *)acl_root_30);

	    /* An awful hack to allow WAI to work.  The WAI objects must be
             * allowed to get the IOR regardless of any ACLs.
             */
            if (!strcmp(uri, "/NameService")) {
		ACL_ListDestroy(errp, rq->acllist);
                rq->acllist = NULL; /* CacheEnter will set it to *NO_ACLS */
            }

            if (ACLUriCacheOn) {	/* Don't URI cache if any Virtual Servers */
                if (strcmp(method, "GET") == 0) {
                    ACL_CacheEnterGet(uri, &rq->acllist);
                } else {
                    ACL_CacheEnter(uri, &rq->acllist);
                }
            } else {			/* Only ACL List cache please */
                ACL_ListHashEnter(&rq->acllist);
            }
	}

	if (rq->acllist != ACL_LIST_NO_ACLS) {	/* Any ACLs to begin with? */
#define STRBUF_LEN 32
#define STRBUF_RIGHTS "http_"
	    char	strbuf[30];	/* temp space for method name */
            char *      rights[2];      /* access right name */
            int         rv;             /* result value */
         
            /* Look up the needed access right by name */
	    rights[0] = strbuf;
	    strcpy(rights[0], STRBUF_RIGHTS);
            strncat(rights[0], pblock_findval("method", rq->reqpb), STRBUF_LEN - sizeof(STRBUF_RIGHTS) -2 );
            rights[1] = NULL;
         
	    rv = ACL_SetupEval(rq->acllist, sn, rq, rights, http_generic, NULL);

	    if ( rv != ACL_RES_ALLOW ) {
		if (!pblock_findval("status", rq->srvhdrs)) {
		    /* Set the status only if it is not already set */
		    protocol_status(sn, rq, PROTOCOL_NOT_FOUND, NULL);
		}
                return REQ_ABORTED;
            }
	}
    }

    return REQ_PROCEED;
}


/* ------------------------------------------------------------------------ */
/* ------------------------------ ObjectType ------------------------------ */
/* ------------------------------------------------------------------------ */


/* -------------------------- _perform_findinfo --------------------------- */


int _perform_findinfo(Session *sn, Request *rq, httpd_object *obj) 
{
    return _directive_applyall(&obj->dt[3], sn, rq);
}


/* --------------------------- request_fileinfo --------------------------- */


NSAPI_PUBLIC int servact_fileinfo(Session *sn, Request *rq) 
{
    httpd_objset *os = rq->os;
    register int x;

    for(x = os->pos - 1; x != -1; --x) {
        /* REQ_PROCEED: multiple instances supercede each other */
        if(_perform_findinfo(sn, rq, os->obj[x]) == REQ_ABORTED)
            return REQ_ABORTED;
    }
    return REQ_PROCEED;
}


/* ------------------------------------------------------------------------ */
/* ------------------------------- Service -------------------------------- */
/* ------------------------------------------------------------------------ */


/* --------------------------- _perform_service --------------------------- */


#define _CHECK_PARAM(tn, un, upb) \
    if((t = pblock_findval(tn, d->inst[y].param))) { \
        if((u = pblock_findval(un, upb))) { \
            switch(shexp_cmp(u, t)) { \
              case -1: \
                log_error(LOG_MISCONFIG, XP_GetAdminStr(DBT_findService_), sn, rq, \
                          XP_GetAdminStr(DBT_invalidShexpS_), t); \
                return REQ_ABORTED; \
              case 1: \
                continue; \
            } \
        } \
    }

#define _CHECK_PARAM_SKIP(tn, un, upb) \
    if((t = pblock_findval(tn, d->inst[y].param))) { \
        if((u = pblock_findval(un, upb))) { \
            switch(shexp_cmp(u, t)) { \
              case -1: \
                log_error(LOG_MISCONFIG, XP_GetAdminStr(DBT_findService_1), sn, rq, \
                          XP_GetAdminStr(DBT_invalidShexpS_1), t); \
                return REQ_ABORTED; \
              case 1: \
                continue; \
            } \
        } \
        else continue; \
    }


int _perform_service(Session *sn, Request *rq, httpd_object *obj)
{
    register int y, r;
    char *t, *u;
    dtable *d;

    d = &obj->dt[4];
    for(y = 0; y < d->ni; y++) {
        _CHECK_PARAM("type", "content-type", rq->srvhdrs);
        _CHECK_PARAM("method", "method", rq->reqpb);
        _CHECK_PARAM_SKIP("query", "query", rq->reqpb);

        /* Success: execute function */
        if( (r = object_execute(&d->inst[y], sn, (void *)rq)) != REQ_NOACTION)
            return r;
    }
    return REQ_NOACTION;
}

/* --------------------------- request_service ---------------------------- */


NSAPI_PUBLIC int servact_service(Session *sn, Request *rq)
{
    httpd_objset *os = rq->os;
    register int x, r;

    for(x = os->pos - 1; x != -1; --x)
        if( (r = _perform_service(sn, rq, os->obj[x])) != REQ_NOACTION)
            return r;
    return REQ_NOACTION;
}


/* ------------------------------------------------------------------------ */
/* -------------------------------- AddLog -------------------------------- */
/* ------------------------------------------------------------------------ */


/* --------------------------- object_findlogs ---------------------------- */


int _perform_logs(Session *sn, Request *rq, httpd_object *obj)
{
    return _directive_applyall(&obj->dt[6], sn, rq);
}


/* ------------------------------ _find_logs ------------------------------ */


NSAPI_PUBLIC int servact_addlogs(Session *sn, Request *rq)
{
    httpd_objset *os = rq->os;
    register int x;

    for(x = os->pos - 1; x != -1; --x) {
        if(_perform_logs(sn, rq, os->obj[x]) == REQ_ABORTED)
            return REQ_ABORTED;
    }
    return REQ_PROCEED;
}


/* ------------------------------------------------------------------------ */
/* -------------------------------- Error --------------------------------- */
/* ------------------------------------------------------------------------ */


/* Assumes that status is properly formatted. */
int _find_error(Session *sn, Request *rq, httpd_object *obj)
{
    char *status = pblock_findval("status", rq->srvhdrs);
    register int y, r;
    register char *t;
    dtable *d;

    d = &obj->dt[5];
    for(y = 0; y < d->ni; y++) {
        if( (t = pblock_findval("reason", d->inst[y].param)) ) {
            if(strcasecmp(t, &status[4]))
                continue;
        }
        else if ( (t = pblock_findval("code", d->inst[y].param)) )
            if(strncmp(t, status, 3))
                continue;
	/* One of those passed. Try to execute.
	 * <Client> restrictions are applied by object_execute(),
	 * which will return NO_ACTION on mismatch.
	 */
	if ((r = object_execute(&d->inst[y], sn, (void*)rq)) != REQ_NOACTION)
	    return r;
    }
    return REQ_NOACTION;
}

NSAPI_PUBLIC int servact_finderror(Session *sn, Request *rq)
{
    register int x;

    for(x = rq->os->pos - 1; x != -1; --x) {
        if(_find_error(sn, rq, rq->os->obj[x]) == REQ_PROCEED)
            return REQ_PROCEED;
    }
    return REQ_NOACTION;
}


#ifdef MCC_PROXY
/* ------------------------------------------------------------------------ */
/* -------------------------------- ROUTE --------------------------------- */
/* ------------------------------------------------------------------------ */


/* ---------------------------- _perform_route ---------------------------- */
/*
 * Performs proxy routing.
 *
 *
 */

int _perform_route(Session *sn, Request *rq, httpd_object *obj)
{
    return _directive_applyall(&obj->dt[7], sn, rq);
}


/* ---------------------------- servact_route ----------------------------- */

int _servact_route(Session *sn, Request *rq)
{
    httpd_objset *os = rq->os;
    register int x;

    for(x = os->pos - 1; x != -1; --x) {
	/* REQ_PROCEED: multiple instances supercede each other */
	if(_perform_route(sn, rq, os->obj[x]) == REQ_ABORTED)
	    return REQ_ABORTED;
    }
    return REQ_NOACTION;
}


char *servact_route(char *safe_url, char *host, int port,
		    Session *sn, Request *rq)
{
    if (sn && rq) {
	pblock_nvinsert("route-url", safe_url, rq->vars);
	pblock_nvinsert("route-host", host, rq->vars);
	pblock_nninsert("route-port", port, rq->vars);

	_servact_route(sn, rq);

	return pblock_findval("route", rq->vars);
    }

    return NULL;
}


/* ------------------------------------------------------------------------ */
/* --------------------------------- DNS ---------------------------------- */
/* ------------------------------------------------------------------------ */


/* ----------------------------- _perform_dns ----------------------------- */
/*
 * Performs custom DNS resolution.
 *
 *
 */

int _perform_dns(Session *sn, Request *rq, httpd_object *obj)
{
    register int y, r;
    dtable *d;

    d = &obj->dt[8];
    for(y = 0; y < d->ni; y++) {
	if ( (r = object_execute(&d->inst[y], sn, (void *)rq)) != REQ_NOACTION)
	    return r;
    }
    return REQ_NOACTION;
}


/* ----------------------------- servact_dns ------------------------------ */


int servact_dns(Session *sn, Request *rq)
{
    httpd_objset *os = rq->os;
    register int x, r;

    for(x = os->pos - 1; x != -1; --x)
	if( (r = _perform_dns(sn, rq, os->obj[x])) != REQ_NOACTION)
	    return r;

    return REQ_NOACTION;
}


struct hostent *optimized_gethostbyname(char *host, pblock *vars)
{
    int cnt = 0;
    int dots = 1;
    int len = 0;
    char *tmp, *p;
    PRHostEnt   hent;
    char        buf[PR_NETDB_BUF_SIZE];
    PRStatus    err;


    if (!host || !*host) return NULL;

    if (vars && (tmp = pblock_findval("local-domain-levels", vars))) {
	if ((dots = atoi(tmp)) < 0) dots = 0;
    }

    for(p=host; *p; p++) {
	if (*p == '.') cnt++;
    }

    len = (int)(p - host);
    if (cnt > dots && host[len-1] != '.') {
      char *dothost = (char *)MALLOC(len + 2);
      strcpy(dothost, host);
      dothost[len] = '.';
      dothost[len+1] = '\0';
      err = PR_GetHostByName(dothost,
				buf,
				PR_NETDB_BUF_SIZE,
				&hent);
      FREE(dothost);
    }
    else {
      err = PR_GetHostByName(host,
				buf,
				PR_NETDB_BUF_SIZE,
				&hent);
    }

    /* This is really bad!!!!  */
    if (err == PR_SUCCESS) return &hent;
    else return NULL;
}


struct hostent *servact_gethostbyname(char *host, Session *sn, Request *rq)
{
    struct hostent *rv = NULL;

    if (host) {
	/* safeguard against bad libc gethostbyname() */
	if (strlen(host) >= 64)
	    host[63] = '\0';

	if (sn && rq) {
	    char *pushed = rq->host;

	    rq->host = host;
	    rq->hp = NULL;

	    switch (servact_dns(sn, rq)) {
	      case REQ_PROCEED:
		rv = rq->hp;
		break;

	      case REQ_NOACTION:
		rv = optimized_gethostbyname(host, rq->vars);
		break;

	      default:
		rv = NULL;
	    }
	    rq->host = pushed;
	}
	else {
	    rv = optimized_gethostbyname(host, rq->vars);
	}
    }

    return rv;
}


/* ------------------------------------------------------------------------ */
/* ------------------------------- Connect -------------------------------- */
/* ------------------------------------------------------------------------ */


/* --------------------------- _perform_connect --------------------------- */
/*
 * Returns the connected socket descriptor.
 *
 * RELIES HEAVILY ON THE RIGHT VALUES SET FOR REQ_* DEFINES:
 *
 * REQ_PROCEED ==  0	which may also be a valid returned sd (in practice
 *			zero is always already in use).  The success can't be
 *			tested against REQ_PROCEED, but with condition > -1.
 *
 * REQ_ABORTED == -1	which is the same as when the connection fails, and
 *			the only legal negative value for a connect function
 *			to return.
 *
 * REQ_NOACTION == -2	when no function was set, and the default should be
 *			used.
 *
 * Other negative values will be treated as -1 return value
 * (failure to connect).
 *
 */

int _perform_connect(Session *sn, Request *rq, httpd_object *obj)
{
    register int y, r;
    dtable *d;

    d = &obj->dt[9];
    for(y = 0; y < d->ni; y++) {
	if ( (r = object_execute(&d->inst[y], sn, (void *)rq)) != REQ_NOACTION)
	    return r;
    }
    return REQ_NOACTION;
}


/* --------------------------- servact_connect ---------------------------- */


int servact_connect(char *host, int port, Session *sn, Request *rq)
{
    httpd_objset *os = rq->os;
    register int x, r;
    char *pushed = rq->host;

    rq->host = host;
    rq->port = port;

    for(x = os->pos - 1; x != -1; --x) {
	if( (r = _perform_connect(sn, rq, os->obj[x])) != REQ_NOACTION) {
	    rq->host = pushed;
	    return (r > -1) ? r : -1;	/* return sd, or SYS_ERROR_FD */
	}
    }

    rq->host = pushed;
    return REQ_NOACTION;
}


/* ------------------------------------------------------------------------ */
/* -------------------------------- Filter -------------------------------- */
/* ------------------------------------------------------------------------ */


/* --------------------------- servact_filter ----------------------------- */

int _perform_filter(Session *sn, Request *rq, httpd_object *obj)
{
    register int y, r;
    char *t, *u;
    dtable *d;

    d = &obj->dt[10];
    for(y = 0; y < d->ni; y++) {
	/* currently always execute function */
	if( (r = object_execute(&d->inst[y], sn, (void *)rq)) != REQ_NOACTION)
	    return r;
    }
    return REQ_NOACTION;
}


int servact_filter(Session *sn, Request *rq)
{
    httpd_objset *os = rq->os;
    register int x, r;

    for(x = os->pos - 1; x != -1; --x)
	if( (r = _perform_filter(sn, rq, os->obj[x])) != REQ_NOACTION)
	    return r;
    return REQ_NOACTION;
}

#endif /* MCC_PROXY */


/* ----------------------- servact_handle_processed ----------------------- */


NSAPI_PUBLIC int servact_handle_processed(Session *sn, Request *rq)
{
    while(1) {
        /* Figure out what objects it uses */
        switch(servact_uri2path(sn,rq)) {
          case REQ_ABORTED:
            return REQ_ABORTED;
          case REQ_RESTART:
            continue;
        }
        /* Process PathChecks */
        switch(servact_pathchecks(sn, rq)) {
          case REQ_ABORTED:
            return REQ_ABORTED;
          case REQ_RESTART:
            continue;
        }
        /* Find that file's typing information */
        switch(servact_fileinfo(sn, rq)) {
          case REQ_ABORTED:
            return REQ_ABORTED;
          case REQ_RESTART:
            continue;
        }
#if 0	/* These get called as callbacks */
	/* Find the Connect function to use */
	switch (servact_connect(sn, rq)) {
	  case REQ_ABORTED:
	    return REQ_ABORTED;
	  case REQ_RESTART:
	    continue;
	}
#endif
#ifdef MCC_PROXY
	/* Find the Filters to pipe the data through */
	switch (servact_filter(sn, rq)) {
	  case REQ_ABORTED:
	    return REQ_ABORTED;
	  case REQ_RESTART:
	    continue;
	}
#endif
        /* Service object (actually send it) */
        switch(servact_service(sn, rq)) {
          case REQ_RESTART:
            continue;
          case REQ_NOACTION:
            log_error(LOG_MISCONFIG, XP_GetAdminStr(DBT_handleProcessed_), sn, rq, 
                      XP_GetAdminStr(DBT_noWayToServiceRequestForS_),
                      pblock_findval("uri", rq->reqpb)); 
            /* fallthrough */
          case REQ_ABORTED:
            error_report(sn, rq);
          case REQ_EXIT:
            /* Make sure the socket is closed */
            sn->csd_open = -1;
        }
        /* No restart was issued. */
        break;
    }

    protocol_finish_request(sn, rq);
    servact_addlogs(sn, rq); /* return doesn't matter */

    return REQ_PROCEED;
}

/* ---------------------------- servact_handle ---------------------------- */


NSAPI_PUBLIC void servact_handle(Session *sn) 
{
    Request *rq;
    int reply;

    if(!(rq = protocol_find_request(sn->inbuf, sn, &reply))) {
        if(reply)
            error_report(sn, NULL);
        protocol_finish_request(sn, rq);
        return;
    }

    /*
     * For HTTP/1.1 - if the status has been set then there is no need
     * to proceed. Send response
     */
    if (rq->protv_num >= CURRENT_PROTOCOL_VERSION && rq->status_num ) {
       if (rq->status_num == PROTOCOL_BAD_REQUEST)
           KEEP_ALIVE(rq) = 0; /* force disconnect */

       if ( ISMTRACE(rq) )  /* response is already sent */
           return;
    
       protocol_start_response(sn, rq);
       protocol_finish_request(sn, rq);
       return;
    }

#ifdef MCC_PROXY
    /* Awful hack for the proxy to make it work with netlib */
    glob_rq = rq;
    glob_sn = sn;

    /* Use the time of incoming request in the log, instead of
     * the time when the transaction was completed.
     */
    {
        extern char *util_make_log_time(time_t tt);
	char *tbuf = util_make_log_time((time_t)0);
	if (tbuf) {
	    pblock_nvinsert("clf-time", tbuf, rq->vars);
	    FREE(tbuf);
	}
    }
    /* Handle SOCKS; this is an abomination, but so is SOCKS */
    if (rq->socks_rq)
      {
	  sockd_service(sn, rq);
	  if(net_close(sn->csd) == -1)
	      log_error(LOG_FAILURE, XP_GetAdminStr(DBT_finishSocksRequest_), sn, rq,
			XP_GetAdminStr(DBT_closeFailedSCsdD_), system_errmsg(), sn->csd);
	  sn->csd_open = 0;
      }
    else
#endif
    if(servact_handle_processed(sn, rq) == REQ_ABORTED) {
        error_report(sn, rq);
        protocol_finish_request(sn, rq);
        servact_addlogs(sn, rq);
    }
    /* ignore REQ_EXIT */

    request_free(rq);
}

