/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * File:        nsapi20.c
 *
 * Description:
 *
 *      This is a plugin for the 3.0 web servers which is intended to
 *      provide binary compatibility for plugins developed for the
 *      2.0 web servers.  It defines all of the functions of the 2.0
 *      server that were exposed in public header files, and uses 3.0
 *      functions to implement them.
 *
 *      In most cases the mapping is direct from a 2.0 function to
 *      a completely equivalent 3.0 function, but there are some
 *      functions where the argument types have changed.
 *
 * BUILD NOTE:
 *
 * On Unix systems nsapi20.o is part of libframe.a.
 *
 * On NT any plugins built against Enterprise/FastTrack 2.0 will contain a
 * reference to ns-httpd20.dll, and will not load unless it is present.
 * So nsapi20.o is used to build ns-httpd20.dll on NT, and is not included
 * in libframe.lib.
 */

#include "private/pprio.h"
#include "prio.h"

#include "nsapi20/netsite.h"
#include "nsapi20/version.h"
#include "nsapi20/base/buffer.h"
#include "nsapi20/base/cinfo.h"
#include "nsapi20/base/crit.h"
#include "nsapi20/base/daemon.h"
#include "nsapi20/base/ereport.h"
#include "nsapi20/base/file.h"
#include "nsapi20/base/net.h"
#include "nsapi20/base/pblock.h"
#include "nsapi20/base/pool.h"
#include "nsapi20/base/regexp.h"
#include "nsapi20/base/sem.h"
#include "nsapi20/base/session.h"
#include "nsapi20/base/shexp.h"
#include "nsapi20/base/shmem.h"
#include "nsapi20/base/systems.h"
#include "nsapi20/base/systhr.h"
#include "nsapi20/base/util.h"
#include "nsapi20/frame/conf.h"
#include "nsapi20/frame/func.h"
#include "nsapi20/frame/http.h"
#include "nsapi20/frame/log.h"
#include "nsapi20/frame/object.h"
#include "nsapi20/frame/objset.h"
#include "nsapi20/frame/protocol.h"
#include "nsapi20/frame/req.h"

#include "nsapi20.h"

NSAPI_PUBLIC int nsapi20_present = 1;

#ifdef XP_UNIX
#define CONV_30_TO_20_SYS_FILE(fd30) PR_FileDesc2NativeHandle(fd30)
#else
#define CONV_30_TO_20_SYS_FILE(fd30) conv30_sys_file(fd30)

static SYS_FILE conv30_sys_file(PRFileDesc *fd30)
{
    SYS_FILE rfd;

    if (fd30 == (PRFileDesc *)(-1)) {
        return (SYS_FILE)NULL;
    }
    rfd = (SYS_FILE)MALLOC(sizeof(*rfd));
    rfd->fname = STRDUP("(unknown)");
    rfd->fh = (HANDLE)(PR_FileDesc2NativeHandle(fd30));
    rfd->flsem = SEM_ERROR;

    return rfd;
}
#endif /* XP_UNIX */

static PRFileDesc *temp30_file(SYS_FILE fd, PRFileDesc *fd30)
{
#ifdef XP_WIN32
    PR_ChangeFileDescNativeHandle(fd30, (PRInt32)(fd->fh));
#else
    PR_ChangeFileDescNativeHandle(fd30, (PRInt32)fd);
#endif /* XP_WIN32 */
    fd30->methods = PR_GetFileMethods();
    return fd30;
}

static PRFileDesc *temp30_socket(PRInt32 sd, PRFileDesc *sd30)
{
    /*
     * Do you frighten easily?  This test attempts to catch the
     * case where someone has picked up sn->csd and started
     * using it in net_xxxxx() calls.  Hopefully they won't also
     * try to use it in native socket calls.
     *
     * Assumes pointers are 4-byte aligned.
     */
    if (((unsigned)sd > 0xffff) && (((unsigned)sd & 3) == 0) &&
        (((PRFileDesc *)sd)->methods == PR_GetTCPMethods())) {

        return (PRFileDesc *)sd;
    }

    PR_ChangeFileDescNativeHandle(sd30, sd);
    sd30->methods = PR_GetTCPMethods();

    return sd30;
}

NSAPI_PUBLIC char *system_version()
{
    return INTsystem_version();
}

NSAPI_PUBLIC void system_version_set(char *server_version)
{
    INTsystem_version_set(server_version);
}

NSAPI_PUBLIC void *system_malloc(int size)
{
    return INTsystem_malloc(size);
}

NSAPI_PUBLIC void *system_calloc(int size)
{
    return INTsystem_calloc(size);
}

NSAPI_PUBLIC void *system_realloc(void *ptr, int size)
{
    return INTsystem_realloc(ptr, size);
}

NSAPI_PUBLIC void system_free(void *ptr)
{
    INTsystem_free(ptr);
}

NSAPI_PUBLIC char *system_strdup(char *ptr)
{
    return INTsystem_strdup(ptr);
}

NSAPI_PUBLIC void *system_malloc_perm(int size)
{
    return INTsystem_malloc_perm(size);
}

NSAPI_PUBLIC void *system_calloc_perm(int size)
{
    return INTsystem_calloc_perm(size);
}

NSAPI_PUBLIC void *system_realloc_perm(void *ptr, int size)
{
    return INTsystem_realloc_perm(ptr, size);
}

NSAPI_PUBLIC void system_free_perm(void *ptr)
{
    INTsystem_free_perm(ptr);
}

NSAPI_PUBLIC char *system_strdup_perm(char *ptr)
{
    return INTsystem_strdup_perm(ptr);
}

#if !defined(HAVE_TIME_R ) && defined(THREAD_ANY)
/* reentrant time routines for those OSs that do not define
 * gmtime_r or localtime_r. 
 */

static CRITICAL localtime_lock = NULL;
static CRITICAL gmtime_lock = NULL;

NSAPI_PUBLIC struct tm *
system_localtime_r(const time_t *timer, struct tm *result)
{
    struct tm *tmp;

	if (localtime_lock == NULL) {
		localtime_lock = crit_init();
	}
	crit_enter(localtime_lock);
    tmp = localtime(timer);
    *result = *tmp;
	crit_exit(localtime_lock);

    return result;
}

NSAPI_PUBLIC struct tm *
system_gmtime_r(const time_t *timer, struct tm *result)
{
    struct tm *tmp;

	if (gmtime_lock == NULL) {
		gmtime_lock = crit_init();
	}
	crit_enter(gmtime_lock);
    tmp = gmtime(timer);
    *result = *tmp;
	crit_exit(gmtime_lock);

    return result;
}

#endif /* !HAVE_TIME_R && THREAD_ANY*/

void filebuf_close_buffer(filebuffer *buf, int clean_mmap)
{
    PRFileDesc *fd30 = (PRFileDesc *)(buf->fd);
    INTfilebuf_close_buffer(buf, clean_mmap);
    free(fd30);
}

NSAPI_PUBLIC filebuffer *filebuf_open_nostat(SYS_FILE fd, int sz, 
                                             struct stat *finfo) 
{
    PRFileDesc fd30;

    return (filebuffer *)INTfilebuf_open_nostat(temp30_file(fd, &fd30), sz, finfo);
}

NSAPI_PUBLIC filebuffer *filebuf_open(SYS_FILE fd, int sz)
{
#ifdef XP_WIN32
    PRFileDesc *fd30 = PR_ImportFile((PRInt32)(fd->fh));
#else
    PRFileDesc *fd30 = PR_ImportFile(fd);
#endif /* XP_WIN32 */

    return INTfilebuf_open(fd30, sz);
}

NSAPI_PUBLIC void filebuf_close(filebuffer *buf) 
{
    INTfilebuf_close(buf);
}

#ifndef FILE_MMAP
NSAPI_PUBLIC int filebuf_next(filebuffer *buf, int advance)
{
    return INTfilebuf_next(buf, advance);
}
#endif /* !FILE_MMAP */

NSAPI_PUBLIC int filebuf_grab(filebuffer *buf, int sz)
{
    return INTfilebuf_grab(buf, sz);
}

filebuffer *filebuf_create(SYS_FILE fd,
                           caddr_t mmap_ptr, int mmap_len, 
                           int bufsz)
{
#ifdef XP_WIN32
    PRFileDesc *fd30 = PR_ImportFile((PRInt32)(fd->fh));
#else
    PRFileDesc *fd30 = PR_ImportFile(fd);
#endif /* XP_WIN32 */

    return INTfilebuf_create(fd30, mmap_ptr, mmap_len, bufsz);
}

NSAPI_PUBLIC netbuf *netbuf_open(SYS_NETFD sd, int sz)
{
    PRFileDesc *sd30 = PR_ImportTCPSocket(sd);

    return INTnetbuf_open(sd30, sz);
}

NSAPI_PUBLIC void netbuf_close(netbuf *buf)
{
    PRFileDesc *sd30 = (PRFileDesc *)(buf->sd);

    INTnetbuf_close(buf);
    free(sd30);
}

NSAPI_PUBLIC int netbuf_next(netbuf *buf, int advance)
{
    return INTnetbuf_next(buf, advance);
}

NSAPI_PUBLIC int netbuf_grab(netbuf *buf, int sz)
{
    return INTnetbuf_grab(buf, sz);
}

NSAPI_PUBLIC int netbuf_buf2sd(netbuf *buf, SYS_NETFD sd, int len)
{
    PRFileDesc sd30;

    return INTnetbuf_buf2sd(buf, temp30_socket(sd, &sd30), len);
}

NSAPI_PUBLIC int filebuf_buf2sd(filebuf *buf, SYS_NETFD sd)
{
    PRFileDesc sd30;

    return INTfilebuf_buf2sd(buf, temp30_socket(sd, &sd30));
}

#ifdef XP_WIN32
NSAPI_PUBLIC filebuffer *pipebuf_open(SYS_FILE fd, int sz, struct stat *finfo)
{
#ifdef XP_WIN32
    PRFileDesc *fd30 = PR_ImportFile((PRInt32)(fd->fh));
#else
    PRFileDesc *fd30 = PR_ImportFile(fd);
#endif /* XP_WIN32 */

    return INTpipebuf_open(fd30, sz, finfo);
}

NSAPI_PUBLIC void pipebuf_close(filebuffer *buf) 
{
    INTpipebuf_close(buf);
}

NSAPI_PUBLIC int pipebuf_next(filebuffer *buf, int advance)
{
    return INTpipebuf_next(buf, advance);
}

NSAPI_PUBLIC int pipebuf_grab(filebuffer *buf, int sz)
{
    return INTpipebuf_grab(buf, sz);
}

NSAPI_PUBLIC int pipebuf_buf2sd(filebuffer *buf, SYS_NETFD sd, int len)
{
    PRFileDesc sd30;

    return INTpipebuf_buf2sd(buf, temp30_socket(sd, &sd30), len);
}

NSAPI_PUBLIC int pipebuf_netbuf2sd(netbuf *buf, SYS_FILE sd, int len)
{
    PRFileDesc fd30;

    return INTpipebuf_netbuf2sd(buf, temp30_file(sd, &fd30), len);
}

NSAPI_PUBLIC int pipebuf_netbuf2pipe(netbuf *buf, SYS_FILE sd, int len)
{
    PRFileDesc fd30;

    return INTpipebuf_netbuf2pipe(buf, temp30_file(sd, &fd30), len);
}

#endif /* XP_WIN32 */

NSAPI_PUBLIC void cinfo_init(void)
{
    INTcinfo_init();
}

NSAPI_PUBLIC void cinfo_terminate(void)
{
    INTcinfo_terminate();
}

NSAPI_PUBLIC char *cinfo_merge(char *fn)
{
    return INTcinfo_merge(fn);
}

NSAPI_PUBLIC cinfo *cinfo_find(char *uri)
{
    return INTcinfo_find(uri);
}

NSAPI_PUBLIC cinfo *cinfo_lookup(char *type)
{
    return INTcinfo_lookup(type);
}

NSAPI_PUBLIC void cinfo_dump_database(FILE *dump)
{
    INTcinfo_dump_database(dump);
}

NSAPI_PUBLIC CRITICAL crit_init(void)
{
    return INTcrit_init();
}

NSAPI_PUBLIC void crit_enter(CRITICAL id)
{
    INTcrit_enter(id);
}

NSAPI_PUBLIC void crit_exit(CRITICAL id)
{
    INTcrit_exit(id);
}

NSAPI_PUBLIC void crit_terminate(CRITICAL id)
{
    INTcrit_terminate(id);
}

NSAPI_PUBLIC CONDVAR condvar_init(CRITICAL id)
{
    return INTcondvar_init(id);
}

NSAPI_PUBLIC void condvar_wait(CONDVAR cv)
{
    INTcondvar_wait(cv);
}

NSAPI_PUBLIC void condvar_notify(CONDVAR cv)
{
    INTcondvar_notify(cv);
}

NSAPI_PUBLIC void condvar_terminate(CONDVAR cv)
{
    INTcondvar_terminate(cv);
}

NSAPI_PUBLIC COUNTING_SEMAPHORE cs_init(int initial_count)
{
    return INTcs_init(initial_count);
}

NSAPI_PUBLIC void cs_terminate(COUNTING_SEMAPHORE csp)
{
    INTcs_terminate(csp);
}

NSAPI_PUBLIC int cs_wait(COUNTING_SEMAPHORE csp)
{
    return INTcs_wait(csp);
}

NSAPI_PUBLIC int cs_trywait(COUNTING_SEMAPHORE csp)
{
    return INTcs_trywait(csp);
}

NSAPI_PUBLIC int cs_release(COUNTING_SEMAPHORE csp)
{
    return INTcs_release(csp);
}


NSAPI_PUBLIC void magnus_atrestart(void (*fn)(void *), void *data)
{
    INTdaemon_atrestart(fn, data);
}

NSAPI_PUBLIC void daemon_atrestart(void (*fn)(void *), void *data)
{
    INTdaemon_atrestart(fn, data);
}

#ifdef XP_UNIX
NSAPI_PUBLIC pid_t child_fork(void)
{
    return INTchild_fork();
}
#endif /* XP_UNIX */

NSAPI_PUBLIC SYS_FILE ereport_getfd(void)
{
    PRFileDesc *fd30 = INTereport_getfd();
    return CONV_30_TO_20_SYS_FILE(fd30);
}

NSAPI_PUBLIC int ereport(int degree, char *fmt, ...)
{
    va_list args;
    int rv;
    va_start(args, fmt);
    rv = INTereport_v(degree, fmt, args);
    va_end(args);
    return rv;
}

NSAPI_PUBLIC int ereport_v(int degree, char *fmt, va_list args)
{
    return INTereport_v(degree, fmt, args);
}

NSAPI_PUBLIC char *ereport_init(char *err_fn, char *email, struct passwd *pw)
{
    return INTereport_init(err_fn, email, (PASSWD)pw, "NSAPI20");
}

NSAPI_PUBLIC void ereport_terminate(void)
{
    INTereport_terminate();
}

NSAPI_PUBLIC int system_stat(char *name, struct stat *finfo)
{
    return INTsystem_stat(name, finfo);
}

#ifdef FILE_WIN32
NSAPI_PUBLIC int system_fread(SYS_FILE fd, char *buf, int sz)
{
    PRFileDesc fd30;
    return INTsystem_fread(temp30_file(fd, &fd30), buf, sz);
}

NSAPI_PUBLIC int system_pread(SYS_FILE fd, char *buf, int sz)
{
    PRFileDesc fd30;
    return INTsystem_pread(temp30_file(fd, &fd30), buf, sz);
}

NSAPI_PUBLIC SYS_FILE system_fopenRO(char *path)
{
    PRFileDesc *fd30 = INTsystem_fopenRO(path);
    SYS_FILE fd = CONV_30_TO_20_SYS_FILE(fd30);
    if (fd30 != (PRFileDesc *)(-1)) {
        free(fd30);
    }
    return fd;
}

NSAPI_PUBLIC SYS_FILE system_fopenWA(char *path)
{
    PRFileDesc *fd30 = INTsystem_fopenWA(path);
    SYS_FILE fd = CONV_30_TO_20_SYS_FILE(fd30);
    if (fd30 != (PRFileDesc *)(-1)) {
        free(fd30);
    }
    return fd;
}

NSAPI_PUBLIC SYS_FILE system_fopenRW(char *path)
{
    PRFileDesc *fd30 = INTsystem_fopenRW(path);
    SYS_FILE fd = CONV_30_TO_20_SYS_FILE(fd30);
    if (fd30 != (PRFileDesc *)(-1)) {
        free(fd30);
    }
    return fd;
}

NSAPI_PUBLIC SYS_FILE system_fopenWT(char *path)
{
    PRFileDesc *fd30 = INTsystem_fopenWT(path);
    SYS_FILE fd = CONV_30_TO_20_SYS_FILE(fd30);
    if (fd30 != (PRFileDesc *)(-1)) {
        free(fd30);
    }
    return fd;
}

NSAPI_PUBLIC void system_fclose(SYS_FILE fd)
{
    PRFileDesc fd30;
    INTsystem_fclose(temp30_file(fd, &fd30));
}

NSAPI_PUBLIC int file_notfound(void)
{
    return INTfile_notfound();
}

NSAPI_PUBLIC int system_initlock(SYS_FILE fd)
{
    return 0;
}

#endif /* FILE_WIN32 */

NSAPI_PUBLIC int system_nocoredumps(void)
{
    return INTsystem_nocoredumps();
}

#ifdef XP_WIN32
NSAPI_PUBLIC char *system_winsockerr(void)
{
    return INTsystem_winsockerr();
}

NSAPI_PUBLIC char *system_winerr(void)
{
    return INTsystem_winerr();
}

NSAPI_PUBLIC void file_unix2local(char *path, char *p2)
{
    INTfile_unix2local(path, p2);
}

NSAPI_PUBLIC SYS_DIR dir_open(char *path)
{
    return INTdir_open(path);
}

NSAPI_PUBLIC SYS_DIRENT *dir_read(SYS_DIR ds)
{
    return INTdir_read(ds);
}

NSAPI_PUBLIC void dir_close(SYS_DIR ds)
{
    INTdir_close(ds);
}

#endif

NSAPI_PUBLIC int system_fwrite(SYS_FILE fd,char *buf,int sz)
{
    PRFileDesc fd30;
    return INTsystem_fwrite(temp30_file(fd, &fd30), buf, sz);
}

NSAPI_PUBLIC int system_fwrite_atomic(SYS_FILE fd, char *buf, int sz)
{
    PRFileDesc fd30;
    return INTsystem_fwrite_atomic(temp30_file(fd, &fd30), buf, sz);
}

NSAPI_PUBLIC char *system_errmsg_fn(void)
{
    char *emsg;
    int rv;
    static char errbuf[256];

    emsg = errbuf;
    rv = INTsystem_errmsg_fn(&emsg, 256);

    return emsg;
}

NSAPI_PUBLIC int file_setinherit(SYS_FILE fd, int value)
{
    PRFileDesc fd30;
    return INTfile_setinherit(temp30_file(fd, &fd30), value);
}

NSAPI_PUBLIC int dir_create_all(char *dir)
{
    return INTdir_create_all(dir);
}

NSAPI_PUBLIC SYS_NETFD net_socket(int domain, int type, int protocol)
{
    PRFileDesc *sd30 = INTnet_socket(domain, type, protocol);
    PRInt32 sd = PR_FileDesc2NativeHandle(sd30);
    /* XXX can we do anything about async mode? */
    free(sd30);
    return sd;
}

NSAPI_PUBLIC int net_listen(SYS_NETFD s, int backlog)
{
    PRFileDesc sd30;
    return INTnet_listen(temp30_socket(s, &sd30), backlog);
}

NSAPI_PUBLIC SYS_NETFD net_create_listener(char *ipaddr, int port)
{
    PRFileDesc *sd30 = INTnet_create_listener(ipaddr, port);
    PRInt32 sd = PR_FileDesc2NativeHandle(sd30);
    /* XXX can we do anything about async mode? */
    free(sd30);
    return sd;
}

NSAPI_PUBLIC
int net_connect(SYS_NETFD s, const void *sockaddr, int namelen)
{
    PRFileDesc sd30;
    return INTnet_connect(temp30_socket(s, &sd30), sockaddr, namelen);
}

NSAPI_PUBLIC
int net_getpeername(SYS_NETFD s, struct sockaddr *name, int *namelen)
{
    PRFileDesc sd30;
    return INTnet_getpeername(temp30_socket(s, &sd30), name, namelen);
}

NSAPI_PUBLIC int net_close(SYS_NETFD s)
{
    PRFileDesc *sd30 = PR_ImportTCPSocket(s);
    return INTnet_close(sd30);
}

NSAPI_PUBLIC
int net_bind(SYS_NETFD s, const struct sockaddr *name, int namelen)
{
    PRFileDesc sd30;
    return INTnet_bind(temp30_socket(s, &sd30), name, namelen);
}

NSAPI_PUBLIC
SYS_NETFD net_accept(SYS_NETFD s, struct sockaddr *addr, int *addrlen)
{
    PRFileDesc sd30;
    PRFileDesc *ret30 = INTnet_accept(temp30_socket(s, &sd30), addr, addrlen);
    PRInt32 ret = PR_FileDesc2NativeHandle(ret30);
    free(ret30);
    return ret;
}

NSAPI_PUBLIC int net_read(SYS_NETFD sd, char *buf, int sz, int timeout)
{
    PRFileDesc sd30;
    return INTnet_read(temp30_socket(sd, &sd30), buf, sz, timeout);
}

NSAPI_PUBLIC int net_write(SYS_NETFD sd, char *buf, int sz)
{
    PRFileDesc sd30;
    return INTnet_write(temp30_socket(sd, &sd30), buf, sz);
}

NSAPI_PUBLIC int net_writev(SYS_NETFD sd, struct iovec *iov, int iovlen)
{
    PRFileDesc sd30;
    return INTnet_writev(temp30_socket(sd, &sd30), iov, iovlen);
}

NSAPI_PUBLIC int net_isalive(SYS_NETFD sd)
{
    PRFileDesc sd30;
    return INTnet_isalive(temp30_socket(sd, &sd30));
}

NSAPI_PUBLIC char *net_ip2host(char *ip, int verify)
{
    return INTnet_ip2host(ip, verify);
}

NSAPI_PUBLIC
int net_getsockopt(SYS_NETFD s, int level,
                   int optname, void *optval, int *optlen)
{
    PRFileDesc sd30;
    return INTnet_getsockopt(temp30_socket(s, &sd30),
                            level, optname, optval, optlen);
}

NSAPI_PUBLIC
int net_setsockopt(SYS_NETFD s, int level,
                   int optname, const void *optval, int optlen)
{
    PRFileDesc sd30;
    return INTnet_setsockopt(temp30_socket(s, &sd30),
                            level, optname, optval, optlen);
}

NSAPI_PUBLIC
int net_select(int nfds,
               fd_set *r, fd_set *w, fd_set *e, struct timeval *timeout)
{
    /* ???? */
    return INTnet_select(nfds, r, w, e, timeout);
}

NSAPI_PUBLIC int net_ioctl(SYS_NETFD s, int tag, void *result)
{
    PRFileDesc sd30;
    return INTnet_ioctl(temp30_socket(s, &sd30), tag, result);
}

NSAPI_PUBLIC pb_param *param_create(char *name, char *value)
{
    return INTparam_create(name, value);
}

NSAPI_PUBLIC int param_free(pb_param *pp)
{
    return INTparam_free(pp);
}

NSAPI_PUBLIC pblock *pblock_create(int n)
{
    return INTpblock_create(n);
}

NSAPI_PUBLIC void pblock_free(pblock *pb)
{
    INTpblock_free(pb);
}

NSAPI_PUBLIC char *pblock_findval(char *name, pblock *pb)
{
    return INTpblock_findval(name, pb);
}

NSAPI_PUBLIC
pb_param *pblock_nvinsert(char *name, char *value, pblock *pb)
{
    return INTpblock_nvinsert(name, value, pb);
}

NSAPI_PUBLIC
pb_param *pblock_nninsert(char *name, int value, pblock *pb)
{
    return INTpblock_nninsert(name, value, pb);
}

NSAPI_PUBLIC void pblock_pinsert(pb_param *pp, pblock *pb)
{
    INTpblock_pinsert(pp, pb);
}

NSAPI_PUBLIC int pblock_str2pblock(char *str, pblock *pb)
{
    return INTpblock_str2pblock(str, pb);
}

NSAPI_PUBLIC char *pblock_pblock2str(pblock *pb, char *str)
{
    return INTpblock_pblock2str(pb, str);
}

NSAPI_PUBLIC void pblock_copy(pblock *src, pblock *dst)
{
    INTpblock_copy(src, dst);
}

NSAPI_PUBLIC pblock *pblock_dup(pblock *src)
{
    return INTpblock_dup(src);
}

NSAPI_PUBLIC char **pblock_pb2env(pblock *pb, char **env)
{
    return INTpblock_pb2env(pb, env);
}

NSAPI_PUBLIC pb_param *_pblock_fr(char *name, pblock *pb, int remove)
{
    return INTpblock_fr(name, pb, remove);

}

NSAPI_PUBLIC
char * pblock_replace(const char *name,char * new_value,pblock *pb)
{
    return INTpblock_replace(name, new_value, pb);
}

NSAPI_PUBLIC pool_handle_t *pool_create(void)
{
    return INTpool_create();
}

NSAPI_PUBLIC void pool_destroy(pool_handle_t *pool_handle)
{
    INTpool_destroy(pool_handle);
}

NSAPI_PUBLIC int pool_enabled(void)
{
    return INTpool_enabled();
}

NSAPI_PUBLIC void *pool_malloc(pool_handle_t *pool_handle, size_t size )
{
    return INTpool_malloc(pool_handle, size);
}

NSAPI_PUBLIC void pool_free(pool_handle_t *pool_handle, void *ptr )
{
    INTpool_free(pool_handle, ptr);
}

NSAPI_PUBLIC 
void *pool_calloc(pool_handle_t *pool_handle, size_t nelem, size_t elsize)
{
    return INTpool_calloc(pool_handle, nelem, elsize);
}

NSAPI_PUBLIC 
void *pool_realloc(pool_handle_t *pool_handle, void *ptr, size_t size )
{
    return INTpool_realloc(pool_handle, ptr, size);
}

NSAPI_PUBLIC char *pool_strdup(pool_handle_t *pool_handle, char *orig_str )
{
    return INTpool_strdup(pool_handle, orig_str);
}

#ifdef USE_REGEX
NSAPI_PUBLIC int regexp_valid(char *exp)
{
    return INTregexp_valid(exp);
}

NSAPI_PUBLIC int regexp_match(char *str, char *exp)
{
    return INTregexp_match(str, exp);
}

NSAPI_PUBLIC int regexp_cmp(char *str, char *exp)
{
    return INTregexp_cmp(str, exp);
}

NSAPI_PUBLIC int regexp_casecmp(char *str, char *exp)
{
    return INTregexp_casecmp(str, exp);
}
#endif /* USE_REGEX */

NSAPI_PUBLIC SEMAPHORE sem_init(char *name, int number)
{
    return INTsem_init(name, number);
}

NSAPI_PUBLIC void sem_terminate(SEMAPHORE id)
{
    INTsem_terminate(id);
}

NSAPI_PUBLIC int sem_grab(SEMAPHORE id)
{
    return INTsem_grab(id);
}

NSAPI_PUBLIC int sem_tgrab(SEMAPHORE id)
{
    return INTsem_tgrab(id);
}

NSAPI_PUBLIC int sem_release(SEMAPHORE id)
{
    return INTsem_release(id);
}

NSAPI_PUBLIC Session *session_alloc(SYS_NETFD csd, struct sockaddr_in *sac);

NSAPI_PUBLIC Session *session_fill(Session *sn)
{
    return INTsession_fill(sn);
}

NSAPI_PUBLIC Session *session_create(SYS_NETFD csd, struct sockaddr_in *sac);

NSAPI_PUBLIC void session_free(Session *sn)
{
    INTsession_free(sn);
}

NSAPI_PUBLIC char *session_dns_lookup(Session *sn, int verify)
{
    return INTsession_dns_lookup(sn, verify);
}

NSAPI_PUBLIC int shexp_valid(char *exp)
{
    return INTshexp_valid(exp);
}

NSAPI_PUBLIC int shexp_match(char *str, char *exp)
{
    return INTshexp_match(str, exp);
}

NSAPI_PUBLIC int shexp_cmp(char *str, char *exp)
{
    return INTshexp_cmp(str, exp);
}

NSAPI_PUBLIC int shexp_casecmp(char *str, char *exp)
{
    return INTshexp_casecmp(str, exp);
}

NSAPI_PUBLIC shmem_s *shmem_alloc(char *name, int size, int expose)
{
    return INTshmem_alloc(name, size, expose);
}

NSAPI_PUBLIC void shmem_free(shmem_s *region)
{
    INTshmem_free(region);
}

NSAPI_PUBLIC
SYS_THREAD systhread_start(int prio, int stksz, void (*fn)(void *), void *arg)
{
    return INTsysthread_start(prio, stksz, fn, arg);
}

NSAPI_PUBLIC SYS_THREAD systhread_current(void)
{
    return INTsysthread_current();
}

NSAPI_PUBLIC void systhread_yield(void)
{
    INTsysthread_yield();
}

NSAPI_PUBLIC SYS_THREAD systhread_attach(void)
{
    return INTsysthread_attach();
}

NSAPI_PUBLIC void systhread_terminate(SYS_THREAD thr)
{
    INTsysthread_terminate(thr);
}

NSAPI_PUBLIC void systhread_sleep(int milliseconds)
{
    INTsysthread_sleep(milliseconds);
}

NSAPI_PUBLIC void systhread_init(char *name)
{
    INTsysthread_init(name);
}

NSAPI_PUBLIC void systhread_timerset(int usec)
{
    INTsysthread_timerset(usec);
}

NSAPI_PUBLIC int systhread_newkey(void)
{
    return INTsysthread_newkey();
}

NSAPI_PUBLIC void *systhread_getdata(int key)
{
    return INTsysthread_getdata(key);
}

NSAPI_PUBLIC void systhread_setdata(int key, void *data)
{
    INTsysthread_setdata(key, data);
}

NSAPI_PUBLIC
int util_getline(filebuffer *buf, int lineno, int maxlen, char *l)
{
    return INTutil_getline(buf, lineno, maxlen, l);
}

#ifdef XP_UNIX

NSAPI_PUBLIC int util_can_exec(struct stat *finfo, uid_t uid, gid_t gid)
{
    return INTutil_can_exec(finfo, uid, gid);
}

#endif /* XP_UNIX */

NSAPI_PUBLIC char **util_env_create(char **env, int n, int *pos)
{
    return INTutil_env_create(env, n, pos);
}

NSAPI_PUBLIC char *util_env_str(char *name, char *value)
{
    return INTutil_env_str(name, value);
}

NSAPI_PUBLIC void util_env_replace(char **env, char *name, char *value)
{
    INTutil_env_replace(env, name, value);
}

NSAPI_PUBLIC void util_env_free(char **env)
{
    INTutil_env_free(env);
}

NSAPI_PUBLIC char **util_env_copy(char **src, char **dst)
{
    return INTutil_env_copy(src, dst);
}

NSAPI_PUBLIC char *util_env_find(char **env, char *name)
{
    return INTutil_env_find(env, name);
}

NSAPI_PUBLIC char *util_hostname(void)
{
    return INTutil_hostname();
}

NSAPI_PUBLIC int util_chdir2path(char *path)
{
    return INTutil_chdir2path(path);
}

NSAPI_PUBLIC int util_is_mozilla(char *ua, char *major, char *minor)
{
    return INTutil_is_mozilla(ua, major, minor);
}

NSAPI_PUBLIC int util_is_url(char *url)
{
    return INTutil_is_url(url);
}

NSAPI_PUBLIC int util_later_than(struct tm *lms, char *ims)
{
    return INTutil_later_than(lms, ims);
}

NSAPI_PUBLIC int util_time_equal(struct tm *lms, char *ims)
{
    return INTutil_time_equal(lms, ims);
}

NSAPI_PUBLIC int util_str_time_equal(char *t1, char *t2)
{
    return INTutil_str_time_equal(t1, t2);
}

NSAPI_PUBLIC int util_uri_is_evil(char *t)
{
    return INTutil_uri_is_evil(t);
}

NSAPI_PUBLIC void util_uri_parse(char *uri)
{
    INTutil_uri_parse(uri);
}

NSAPI_PUBLIC void util_uri_unescape(char *s)
{
    INTutil_uri_unescape(s);
}

NSAPI_PUBLIC char *util_uri_escape(char *d, char *s)
{
    return INTutil_uri_escape(d, s);
}

NSAPI_PUBLIC char *util_url_escape(char *d, char *s)
{
    return INTutil_url_escape(d, s);
}

NSAPI_PUBLIC char *util_sh_escape(char *s)
{
    return INTutil_sh_escape(s);
}

NSAPI_PUBLIC int util_mime_separator(char *sep)
{
    return INTutil_mime_separator(sep);
}

NSAPI_PUBLIC int util_itoa(int i, char *a)
{
    return INTutil_itoa(i, a);
}

NSAPI_PUBLIC int util_vsprintf(char *s, register char *fmt, va_list args)
{
    return INTutil_vsprintf(s, fmt, args);
}

NSAPI_PUBLIC int util_sprintf(char *s, char *fmt, ...)
{
    va_list args;
    int rv;
    va_start(args, fmt);
    rv = INTutil_vsprintf(s, fmt, args);
    va_end(args);
    return rv;
}

NSAPI_PUBLIC
int util_vsnprintf(char *s, int n, register char *fmt, va_list args)
{
    return INTutil_vsnprintf(s, n, fmt, args);
}

NSAPI_PUBLIC int util_snprintf(char *s, int n, char *fmt, ...)
{
    va_list args;
    int rv;
    va_start(args, fmt);
    rv = INTutil_vsnprintf(s, n, fmt, args);
    va_end(args);
    return rv;
}

#ifdef XP_WIN32
NSAPI_PUBLIC VOID util_delete_directory(char *FileName, BOOL delete_directory)
{
    INTutil_delete_directory(FileName, delete_directory);
}
#endif

NSAPI_PUBLIC int util_strftime(char *s, const char *format, const struct tm *t)
{
    return INTutil_strftime(s, format, t);
}

#ifdef NEED_STRCASECMP
NSAPI_PUBLIC int util_strcasecmp(CASECMPARG_T char *s1, CASECMPARG_T char *s2)
{
    return INTutil_strcasecmp(s1, s2);
}

NSAPI_PUBLIC int util_strncasecmp(CASECMPARG_T char *s1, CASECMPARG_T char *s2,
				  int n)
{
    CASECMPARG_T char *pA;
    CASECMPARG_T char *pB;

    for(pA=s1, pB=s2;; pA++, pB++) {
        int tmp;
        if (pA == s1+n)
            return 0;
        if (!(*pA && *pB))
            return *pA - *pB;
        tmp = tolower(*pA) - tolower(*pB);
        if (tmp)
            return tmp;
    }
}
#endif /* NEED_STRCASECMP */

NSAPI_PUBLIC char *conf_init(char *cfn)
{
    return INTconf_init(cfn);
}

NSAPI_PUBLIC void conf_terminate(void)
{
    INTconf_terminate();
}

NSAPI_PUBLIC conf_global_vars_s *conf_getglobals(void)
{
    return INTconf_getglobals();
}

NSAPI_PUBLIC void conf_vars2daemon(daemon_s *d)
{
    INTconf_vars2daemon(d);
}

#ifdef XP_WIN32
NSAPI_PUBLIC BOOLEAN GetDomainAddresses(PCHAR ErrMsg)
{
	return TRUE;
}
#endif /* XP_WIN32 */

#ifndef XP_WIN32
NSAPI_PUBLIC void func_init(void)
{
    /*
     * We'd call INTfunc_init(func_standard) here, but func_standard
     * is in netsite/httpd/src.  Since magnus_init() calls
     * INTfunc_init(func_standard), we'll hope that's sufficient.
     */
}
#endif /* !XP_WIN32 */

NSAPI_PUBLIC FuncPtr func_find(char *name)
{
    return INTfunc_find(name);
}

NSAPI_PUBLIC int func_exec(pblock *pb, Session *sn, Request *rq)
{
    return INTfunc_exec(pb, sn, rq);
}

NSAPI_PUBLIC struct FuncStruct *func_insert(char *name, FuncPtr fn)
{
    return INTfunc_insert(name, fn);
}

NSAPI_PUBLIC Request *http_find_request(netbuf *buf, Session *sn, int *reply)
{
    return INThttp_find_request(buf, sn, reply);
}

NSAPI_PUBLIC int http_parse_request(char *t, Request *rq, Session *sn)
{
    return INThttp_parse_request(t, rq, sn);
}

NSAPI_PUBLIC int http_start_response(Session *sn, Request *rq)
{
    return INThttp_start_response(sn, rq);
}

NSAPI_PUBLIC char **http_hdrs2env(pblock *pb)
{
    return INThttp_hdrs2env(pb);
}

NSAPI_PUBLIC void http_status(Session *sn, Request *rq, int n, char *r)
{
    INThttp_status(sn, rq, n, r);
}

NSAPI_PUBLIC int http_set_finfo(Session *sn, Request *rq, struct stat *finfo)
{
    return INThttp_set_finfo(sn, rq, finfo);
}

NSAPI_PUBLIC char *http_dump822(pblock *pb, char *t, int *pos, int tsz)
{
    return INThttp_dump822(pb, t, pos, tsz);
}

NSAPI_PUBLIC void http_finish_request(Session *sn, Request *rq)
{
    INThttp_finish_request(sn, rq);
}

NSAPI_PUBLIC void http_handle_session(Session *sn)
{
    INThttp_handle_session(sn);
}

NSAPI_PUBLIC char *http_uri2url(char *prefix, char *suffix)
{
    return INThttp_uri2url(prefix, suffix);
}

NSAPI_PUBLIC char *http_uri2url_dynamic(char *prefix, char *suffix,
                                        Session *sn, Request *rq)
{
    return INThttp_uri2url_dynamic(prefix, suffix, sn, rq);
}

NSAPI_PUBLIC void http_set_keepalive_timeout(int secs)
{
    INThttp_set_keepalive_timeout(secs);
}

NSAPI_PUBLIC
int log_error(int degree, char *func, Session *sn, Request *rq, char *fmt, ...)
{
    va_list args;
    int rv;
    va_start(args, fmt);
    rv = INTlog_error_v(degree, func, sn, rq, fmt, args);
    va_end(args);
    return rv;
}

NSAPI_PUBLIC int log_ereport(int degree, char *fmt, ...)
{
    va_list args;
    int rv;
    va_start(args, fmt);
    rv = INTlog_ereport_v(degree, fmt, args);
    va_end(args);
    return rv;
}

NSAPI_PUBLIC httpd_object *object_create(int nd, pblock *name)
{
    return INTobject_create(nd, name);
}

NSAPI_PUBLIC void object_free(httpd_object *obj)
{
    INTobject_free(obj);
}

NSAPI_PUBLIC void object_add_directive(int dc, pblock *p, pblock *c, 
                                       httpd_object *obj)
{
    INTobject_add_directive(dc, p, c, obj);
}

NSAPI_PUBLIC int object_execute(directive *inst, Session *sn, void *rq)
{
    return INTobject_execute(inst, sn, rq);
}

NSAPI_PUBLIC httpd_objset *objset_scan_buffer(filebuffer *buf, char *errstr,
                                              httpd_objset *os)
{
    return INTobjset_scan_buffer(buf, errstr, os);
}

NSAPI_PUBLIC httpd_objset *objset_create(void)
{
    return INTobjset_create();
}

NSAPI_PUBLIC void objset_free(httpd_objset *os)
{
    INTobjset_free(os);
}

NSAPI_PUBLIC void objset_free_setonly(httpd_objset *os)
{
    INTobjset_free_setonly(os);
}

NSAPI_PUBLIC httpd_object *objset_new_object(pblock *name, httpd_objset *os)
{
    return INTobjset_new_object(name, os);
}

NSAPI_PUBLIC void objset_add_object(httpd_object *obj, httpd_objset *os)
{
    INTobjset_add_object(obj, os);
}

NSAPI_PUBLIC void objset_add_init(pblock *initfn, httpd_objset *os)
{
    INTobjset_add_init(initfn, os);
}

NSAPI_PUBLIC httpd_object *objset_findbyname(char *name, httpd_objset *ign, 
                                             httpd_objset *os)
{
    return INTobjset_findbyname(name, ign, os);
}

NSAPI_PUBLIC httpd_object *objset_findbyppath(char *ppath, httpd_objset *ign, 
                                              httpd_objset *os)
{
    return INTobjset_findbyppath(ppath, ign, os);
}

#ifdef XP_WIN32

/*
 * XXXhep The following functions were in ntobjset.cpp, which is no
 * longer in the server.
 */

/* Read the registry configuration parameters into memory */

NSAPI_PUBLIC 
httpd_objset *objset_scan_registry(char *errstr, httpd_objset *os, 
                                   char *product_key)
{
    return NULL;
}

NSAPI_PUBLIC VOID ProcessObject(httpd_objset *nos, HKEY hObjectKey,
                                CHAR *Object, DWORD ObjectIndex	)
{
}

NSAPI_PUBLIC BOOLEAN ProcessClient(httpd_object *cobj, HKEY hObjectKey,
                                   CHAR *Client, DWORD ObjectIndex)
{
    return TRUE;
}

NSAPI_PUBLIC BOOLEAN ProcessDirective(httpd_object *cobj, pblock *param,
                                      pblock *client, HKEY hObjectKey,
                                      CHAR *Directive, DWORD ObjectIndex,
                                      DWORD DirectiveIndex, DWORD *clrefs,
                                      BOOLEAN IsObject)
{
    return TRUE;
}

NSAPI_PUBLIC BOOLEAN ProcessFunction(pblock **param, HKEY hDirectiveKey,
                                     CHAR *Function, DWORD ObjectIndex,
                                     DWORD DirectiveIndex,
                                     DWORD FunctionIndex)
{
    return TRUE;
}

NSAPI_PUBLIC int _ntverify_parameter(char *str)
{
    return -1;
}

NSAPI_PUBLIC int ntpblock_str2pblock(int ParameterIndex, char *name,
                                     char *value, int namelength,
                                     int valuelength, pblock *pb)
{
    return -1;
}

NSAPI_PUBLIC int ntpblock_pblock2str(struct pb_entry *p, PCHAR *ParameterName,
                                     PCHAR *ParameterValue)
{
    return -1;
}

NSAPI_PUBLIC BOOLEAN ClearConfigurationParameters(HKEY hParentKey, 
                                                  PCHAR CompleteKeyName,
                                                  PCHAR KeyName)
{
    return TRUE;
}

NSAPI_PUBLIC char *GetProductKey()
{
    return NULL;
}

#endif /* XP_WIN32 */

NSAPI_PUBLIC Request *request_create(void)
{
    return INTrequest_create();
}

NSAPI_PUBLIC void request_free(Request *req)
{
    INTrequest_free(req);
}

NSAPI_PUBLIC Request *request_restart_internal(char *uri, Request *rq)
{
    return INTrequest_restart_internal(uri, rq);
}

NSAPI_PUBLIC int request_header(char *name, char **value, 
                                Session *sn, Request *rq)
{
    return INTrequest_header(name, value, sn, rq);
}

NSAPI_PUBLIC struct stat *request_stat_path(char *path, Request *rq)
{
    return INTrequest_stat_path(path, rq);
}

NSAPI_PUBLIC int servact_uri2path(Session *sn, Request *rq)
{
    return INTservact_uri2path(sn, rq);
}

NSAPI_PUBLIC int servact_pathchecks(Session *sn, Request *rq)
{
    return INTservact_pathchecks(sn, rq);
}

NSAPI_PUBLIC int servact_fileinfo(Session *sn, Request *rq)
{
    return INTservact_fileinfo(sn, rq);
}

NSAPI_PUBLIC int servact_service(Session *sn, Request *rq)
{
    return INTservact_service(sn, rq);
}

NSAPI_PUBLIC int servact_handle_processed(Session *sn, Request *rq)
{
    return INTservact_handle_processed(sn, rq);
}

NSAPI_PUBLIC char *servact_translate_uri(char *uri, Session *sn)
{
    return INTservact_translate_uri(uri, sn);
}

NSAPI_PUBLIC int servact_finderror(Session *sn, Request *rq)
{
    return INTservact_finderror(sn, rq);
}

#ifdef XP_WIN32
NSAPI_PUBLIC int stat(const char *name, struct stat *finfo)
{
    return INTsystem_stat((char *)name, finfo);
}
#endif /* XP_WIN32 */
