/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * objset.c: Parses a config. file into an object set
 * 
 * See objset.h for details.
 * 
 * Rob McCool
 */


#include "frame/objset.h"

#include "netsite.h"		/* safe malloc, free, strdup */

#include "base/shexp.h"		/* shexp_cmp */
#include "base/util.h"		/* util_itoa, util_sprintf */
#include "base/ereport.h"	/* ereport() */

#ifdef NOACL
#include "frame/ipfilter.h"	/* ip_filter_setup() */
#include "frame/dnfilter.h"	/* dns_filter_setup() */
#endif /* NOACL */

#include "frame/servact.h"	/* NUM_DIRECTIVES, directive_name2num */

#include <string.h>		/* strlen, strncmp, strcpy */
#include <ctype.h>		/* isalpha, toupper */

#include "frame/dbtframe.h"

/* ---------------------------- objset_create ----------------------------- */


NSAPI_PUBLIC httpd_objset *objset_create(void) {
    httpd_objset *os = (httpd_objset *) MALLOC(sizeof(httpd_objset));

    os->pos = 0;
    os->obj = (httpd_object **) MALLOC(OBJSET_INCSIZE*sizeof(httpd_object *));
    os->obj[0] = NULL;

    os->initfns = NULL;
    return os;
}


/* ------------------------- objset_free_setonly -------------------------- */


NSAPI_PUBLIC void objset_free_setonly(httpd_objset *os) 
{
    FREE(os->obj);
    FREE(os);
}


/* ----------------------------- objset_free ------------------------------ */


NSAPI_PUBLIC void objset_free(httpd_objset *os) 
{
    register int x;

    for(x=0; x < os->pos; x++)
        object_free(os->obj[x]);
    if(os->initfns) {
        for(x = 0; os->initfns[x]; ++x)
            pblock_free(os->initfns[x]);
        FREE(os->initfns);
    }
    objset_free_setonly(os);
}


/* -------------------------- objset_add_object --------------------------- */


NSAPI_PUBLIC void objset_add_object(httpd_object *obj, httpd_objset *os) 
{
    char *t, *ppath = pblock_findval("ppath", obj->name);
    register httpd_object **objs;
    register int x, y, l;

    if((os->pos) && (!((os->pos+1) % OBJSET_INCSIZE)))
        os->obj = (httpd_object **) 
            REALLOC(os->obj, (os->pos+OBJSET_INCSIZE) * sizeof(httpd_object*));

    objs = os->obj;
    x = os->pos;
    if(ppath) {
        l = strlen(ppath);
        for(x = 0; x < os->pos; x++) {
            if( (t = pblock_findval("ppath", objs[x]->name)) &&
                (strlen(t) > strlen(ppath)))
                break;
        }
    }
    for(y = ++os->pos; y != x; --y)
        objs[y] = objs[y-1];
    objs[x] = obj;
}


/* -------------------------- objset_new_object --------------------------- */


NSAPI_PUBLIC httpd_object *objset_new_object(pblock *name, httpd_objset *os) 
{
    httpd_object *obj = object_create(NUM_DIRECTIVES, name);

    objset_add_object(obj, os);
    return obj;
}


/* --------------------------- objset_add_init ---------------------------- */


NSAPI_PUBLIC void objset_add_init(pblock *initfn, httpd_objset *os)
{
    register int x;

    if(!os->initfns) {
        os->initfns = (pblock **) MALLOC(2*sizeof(pblock *));
        x = 0;
    }
    else {
        for(x = 0; os->initfns[x]; ++x); /* no action */
        os->initfns = (pblock **) 
            REALLOC(os->initfns, (x+2) * sizeof(pblock *));
    }
    os->initfns[x] = initfn;
    os->initfns[x+1] = NULL;
}


/* -------------------------- objset_scan_buffer -------------------------- */


#define syntax_error(oops) \
  { objset_free(nos);  util_sprintf(errstr, "%d: %s", ln, oops); return NULL; }

/* What kind of line we just got done parsing */
#define SAW_NOTHING 0
#define SAW_OBJECT 1
#define SAW_CLIENT 2
#define SAW_DIRECTIVE 3
#define SAW_INIT 4

NSAPI_PUBLIC
httpd_objset *objset_scan_buffer(filebuf_t *buf, char *errstr, httpd_objset *os)
{
    httpd_objset *nos = (os ? os : objset_create());
    int eof, ln, clrefs, dnum, lastline;
    int rv;
    char line[MAX_DIRECTIVE_LEN], *t;
    pblock *client, *param;
    httpd_object *cobj;
#ifdef NOACL
    IPFilterErr_t iperr;
    DNSFilterErr_t dnserr;
#endif /* NOACL */
    client = NULL; param = NULL;
    clrefs = 0; ln = 0; eof = 0; 
    lastline = SAW_NOTHING;
    cobj = NULL;

    /* Getline can return EOF and still return data at the same time. */
    while(!eof) {
        eof = util_getline(buf, ++ln, sizeof(line), line);

        /* If the line was only a linefeed, just close off prior directive */
        if(!line[0]) {
            lastline = SAW_NOTHING;
            continue;
        }

        /* If the line was a comment, don't do anything */
        else if(line[0] == '#')
            continue;

        /* Client or Object open or close tag */
        else if(line[0] == '<') {
            /* We already know strlen(line) > 0 */
            t = &line[strlen(line) - 1];
            if(*t == '>')
                *t = '\0';

            /* Close tag */
            if(line[1] == '/') {
                lastline = SAW_NOTHING;

                /* Laziness: O == Object, C == Client */
                if(line[2] == 'O')
                    cobj = NULL;
                else if((line[2] == 'C') || (line[2] == 'c')) {
                    if(clrefs)
                        pblock_nninsert("refcount", clrefs, client);
                    else
                        pblock_free(client);
                    client = NULL;
                }
#ifdef MCC_BATMAN
                /* F == Filter */
                else if ((line[2] == 'F') || (line[2] == 'f')) {
                    cobj = NULL;
                }
#endif
                else
                    syntax_error("unknown closing directive");
            }

            /* Open tag */
            else {
                /* Laziness: O == Object, C == Client */
#ifdef MCC_BATMAN
                /* Laziness: F == Filter */
                if ((line[1] == 'O') || (line[1] == 'o') || 
                    (line[1] == 'F') || (line[1] == 'f')) {
#else
                if((line[1] == 'O') || (line[1] == 'o')) {
#endif
                    if(cobj)
                        syntax_error("new object before previous one closed");

                    lastline = SAW_OBJECT;
                    param = pblock_create(PARAMETER_HASH_SIZE);

                    /* 8 == strlen("<Object ") == strlen("<Filter ") */
                    pblock_str2pblock(&line[8], param);
                    cobj = objset_new_object(param, nos);
                }
                else if((line[1] == 'C') || (line[1] == 'c')) {
                    lastline = SAW_CLIENT;
                    if(client)
                        syntax_error("new client tag before previous closed");

                    client = pblock_create(PARAMETER_HASH_SIZE);
                    clrefs = 0;
                    /* 8 == strlen("<Client ") */
                    pblock_str2pblock(&line[8], client);

#ifdef NOACL
		    /* Process IP filter specifications, if any */
		    rv = ip_filter_setup(client, &iperr);
		    if (rv < 0) {
			/* Error in IP filter specifications */
			/* Log the error and keep going */
			ereport(LOG_MISCONFIG,
	XP_GetAdminStr(DBT_objConfLineDErrorInFilterFileSAt_),
				ln, iperr.filename, iperr.lineno,
				iperr.errstr);
			if (iperr.filename) FREE(iperr.filename);
		    }

		    /* Process DNS filter specifications, if any */
		    rv = dns_filter_setup(client, &dnserr);
		    if (rv < 0) {
			/* Error in DNS filter specifications */
			/* Log the error and keep going */
			ereport(LOG_MISCONFIG,
	XP_GetAdminStr(DBT_objConfLineDErrorInFilterFileSAt_1),
				ln, dnserr.filename, dnserr.lineno,
				dnserr.errstr);
			if (dnserr.filename) FREE(dnserr.filename);
		    }
#endif /* NOACL */
                }
                else
                    syntax_error("unknown container directive");
            }
        }

        /* New directive */
        else if(isalpha(line[0])) {
            /* Truncate the string so we can look up the directive name. */
            if(!(t = strchr(line, ' ')))
                syntax_error("Directives must have at least one parameter");
            /* Increment t so that it is a pointer to the directive's parms */
            *t++ = '\0';

            /* Create the directive's parameters and store them */
            param = pblock_create(PARAMETER_HASH_SIZE);
            pblock_str2pblock(t, param);

            /* Special case for Init directives */
            if(!strcasecmp(line, "Init")) {
                if(cobj || client)
                    syntax_error("init functions must be outside Client or "
                                 "Object regions");
               objset_add_init(param, nos);
                lastline = SAW_INIT;
            }
            /* Standard directive */
            else {
                if((dnum = directive_name2num(line)) == -1)
                    syntax_error("unrecognized directive");
                if(client)
                    ++clrefs;

                if(!cobj)
                    syntax_error("directive outside object");
				pblock_nvinsert("Directive", line, param); // Tags how the directive will execute
                object_add_directive(dnum, param, client, cobj);
                lastline = SAW_DIRECTIVE;
            }
        }

        /* Continuation of previous directive */
        else if(isspace(line[0])) {
            switch(lastline) {
              case SAW_NOTHING:
                syntax_error("Lines must only begin with space if they are"
                             "continuations of a previous directive line.");

              case SAW_CLIENT:
                /* fallthrough */
              case SAW_OBJECT:
                /* We already know strlen(line) > 0 */
                t = &line[strlen(line) - 1];
                if(*t == '>')
                    *t = '\0';
                pblock_str2pblock(line, lastline == SAW_OBJECT ? cobj->name 
                                                               : client);
                break;

              case SAW_DIRECTIVE:
              case SAW_INIT:
                /* Add this line to the previous directive's parameters */
                pblock_str2pblock(line, param);
                break;
            }
        }
    }
    return nos;
}


/* -------------------------- objset_findbyname --------------------------- */


/* 
 * These should be read from hash tables build in the scan-buffer 
 * routine. 
 */
NSAPI_PUBLIC
httpd_object *objset_findbyname(char *name, httpd_objset *ign,
                                httpd_objset *os) 
{
    register int x, y;
    pb_param *pp;

    for(x = 0; x < os->pos; x++) {
        if((pp = pblock_find("name", os->obj[x]->name))) {
            if(!strcasecmp(name, pp->value)) {
                if(ign) {
                    for(y = 0; y < ign->pos; y++)
                        if(ign->obj[y] == os->obj[x])
                            break;
                    if(y != ign->pos)
                        continue;
                }
                return os->obj[x];
            }
        }
    }
    return NULL;
}



/* -------------------------- objset_findbyppath -------------------------- */

NSAPI_PUBLIC
httpd_object *objset_findbyppath(char *ppath, httpd_objset *ign, 
                                 httpd_objset *os) 
{
    register int x, y;
    pb_param *pp;
    char *tmp_path = NULL;

    for(x = 0; x < os->pos; x++) {
        if((pp = pblock_find("ppath", os->obj[x]->name))) {
	    /* httpd uses shexp's, proxy regexp's */
            if(WILDPAT_CMP(ppath, pp->value)) {
	        /* If the path wasn't found try adding a '/' (anton) */
	        if (!tmp_path) {
		    /* 2 for '/' and '\0' */
		    tmp_path = (char *) MALLOC(strlen(ppath) + 2);
		    util_sprintf(tmp_path, "%s%c", ppath, '/');
		}

		if(WILDPAT_CMP(tmp_path, pp->value))
		    continue;
	    }

	    if(ign) {
	        for(y = 0; y < ign->pos; y++)
		    if(ign->obj[y] == os->obj[x])
		        break;
		if(y != ign->pos)
		    continue;
	    }
	    if (tmp_path) FREE(tmp_path);
	    return os->obj[x];
        }
    }
    if (tmp_path) FREE(tmp_path);
    return NULL;
}

/* -------------------------- objset_copydirs -------------------------- */


httpd_objset *objset_copydirs(httpd_objset *dst_os, httpd_objset *os, int dir)
{
    pblock *name = pblock_create(PARAMETER_HASH_SIZE);
    httpd_object *no;
    int obj_x, dir_x;

    if (name)
        pblock_nvinsert("name", "log-list", name);

    no = object_create(NUM_DIRECTIVES, name);

    if (dst_os == NULL)
        dst_os = objset_create();

    if (dst_os) {
        if (no)
            objset_add_object(no, dst_os);
        else {
            /* Error occurred creating object */
            objset_free(dst_os);
            return NULL;
        }

        for (obj_x=os->pos-1; obj_x != -1; obj_x--) {
            for (dir_x = 0; dir_x < os->obj[obj_x]->dt[dir].ni; dir_x++) {
                object_add_directive(dir,
                    pblock_dup(os->obj[obj_x]->dt[dir].inst[dir_x].param),
                    pblock_dup(os->obj[obj_x]->dt[dir].inst[dir_x].client),
                    no);
            }
        }
    }

    return dst_os;
}
