/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * req.c: Request-specific stuff
 * 
 * Rob McCool
 * 
 */

#include <prerror.h>
#include "base/util.h"            /* unescape */
#include "base/plist.h"
#include "frame/req.h"
#include "frame/log.h"
#include "frame/protocol.h"
#include "frame/conf.h"           /* default object */

#include "frame/object.h"

#include <base/fasttime.h>

#include <signal.h>        /* signal, alarm */
#include <ctype.h>         /* isupper, tolower */
/* Removed for ns security integration
#include <xp_error.h>
*/

/* From error.c */
NSAPI_PUBLIC int error_report(Session *sn, Request *rq);

#ifdef MCC_PROXY
NSAPI_PUBLIC int sockd_service     (Session *sn, Request *rq);
NSAPI_PUBLIC int sockd_read_request(Session *sn, Request *rq);
#endif


/* ---------------------------- request_create ---------------------------- */


NSAPI_PUBLIC Request *request_create(void) {
    Request *rq = (Request *) MALLOC(sizeof(Request));

    rq->vars = pblock_create(REQ_HASHSIZE);
    rq->headers = pblock_create(REQ_HASHSIZE);
    rq->loadhdrs = 0;
    rq->senthdrs = 0;
    rq->reqpb = pblock_create(REQ_HASHSIZE);
    rq->srvhdrs = pblock_create(REQ_HASHSIZE);
    rq->os = NULL;
    rq->tmpos = NULL;
    rq->statpath = NULL;
    rq->staterr = NULL;
    rq->finfo = NULL;
    rq->aclstate = 0;
    rq->acldirno = 0;
    rq->aclname = 0;
    rq->aclpb = 0;
    rq->acllist = 0;
#ifdef MCC_PROXY
    rq->hp = NULL;
    rq->host = NULL;
    rq->port = 0;
    rq->socks_rq = NULL;
#endif
    rq->request_is_cacheable = 1;
    rq->directive_is_cacheable = 0;
    rq->cached_headers = NULL;
	/* HTTP/1.1 defaults */
	rq->req_start = ft_time();
	KEEP_ALIVE(rq) = 1;
	CHUNKED(rq) = 0;
	ABS_URI(rq) = 0;
    PIPELINED(rq)=0;
	rq->method_num = -1;
	rq->protv_num = 0;
	rq->status_num = 0;
	rq->hostname = NULL;
	rq->allowed = AllowedMethods;
    return rq;
}


/* ----------------------- request_restart_internal ----------------------- */


NSAPI_PUBLIC Request *request_restart_internal(char *uri, Request *rq)
{
    pb_param *protocol, *method, *rql;
    Request *ret;
    char *qs;

    if(rq) {
        method = pblock_remove("method", rq->reqpb);
        protocol = pblock_remove("protocol", rq->reqpb);
        rql = pblock_remove("clf-request", rq->reqpb);

        pblock_free(rq->reqpb);
        pblock_free(rq->vars);
/*         PListDeleteProp(sn->subject, ACL_ATTR_USER_INDEX, ACL_ATTR_USER);  */
        pblock_free(rq->srvhdrs);
        rq->reqpb = pblock_create(REQ_HASHSIZE);
        rq->srvhdrs = pblock_create(REQ_HASHSIZE);
        rq->vars = pblock_create(REQ_HASHSIZE);

        pblock_pinsert(protocol, rq->reqpb);
        if(strcmp(method->value, "HEAD")) {
            FREE(method->value);
            method->value = STRDUP("GET");
        }
        pblock_pinsert(rql, rq->reqpb);
        pblock_pinsert(method, rq->reqpb);
        rq->request_is_cacheable = 1;
        rq->cached_headers = NULL;
        ret = rq;
    }
    else {
        ret = request_create();
        pblock_nvinsert("protocol", "HTTP/0.9", ret->reqpb);
        pblock_nvinsert("method", "GET", ret->reqpb);
    }
    if( (qs = strchr(uri, '?')) ) {
        *qs++ = '\0';
        pblock_nvinsert("query", qs, ret->reqpb);
    }
    pblock_nvinsert("uri", uri, ret->reqpb);

    return ret;
}


/* ----------------------------- request_free ----------------------------- */


NSAPI_PUBLIC void request_free(Request *rq) {

    pblock_free(rq->vars);
/*    PListDeleteProp(sn->subject, ACL_ATTR_USER_INDEX, ACL_ATTR_USER);  */
    pblock_free(rq->headers);
    pblock_free(rq->reqpb);
    pblock_free(rq->srvhdrs);
    if(rq->os)
        objset_free_setonly(rq->os);
    if(rq->tmpos)
        objset_free(rq->tmpos);

    if(rq->statpath)
        FREE(rq->statpath);
    if(rq->staterr)
        FREE(rq->staterr);
    if(rq->finfo)
        FREE(rq->finfo);

#ifdef MCC_PROXY
    if (rq->socks_rq)
	FREE(rq->socks_rq);
#endif

    if (rq->cached_headers)
        FREE(rq->cached_headers);

    FREE(rq);
}


/* ---------------------------- request_header ---------------------------- */


NSAPI_PUBLIC int request_header(char *name, char **value, 
                                Session *sn, Request *rq)
{
    *value = pblock_findval(name, rq->headers);
	return REQ_PROCEED;
}


/* -------------------------- request_stat_path --------------------------- */


#ifndef XP_WIN32
#define _GET_SYSERRNO() errno
#define _SET_SYSERRNO(x) (errno = x)
#else
#define _GET_SYSERRNO() PR_GetError()
#define _SET_SYSERRNO(x) PR_SetError(x, x)
#endif

NSAPI_PUBLIC struct stat *request_stat_path(char *tpath, Request *rq) {
    char *path = (tpath ? tpath : pblock_findval("path", rq->vars));

    if(!path) {
        if(rq->staterr)
            FREE(rq->staterr);
        rq->staterr = STRDUP("no path given to stat");
        return NULL;
    }
    if(rq->statpath) {
        if(!strcmp(rq->statpath, path)) {
            _SET_SYSERRNO(rq->staterrno);
            return rq->finfo;
        }
        FREE(rq->statpath);
        rq->statpath = NULL;
    }
    if(!rq->finfo)
        rq->finfo = (struct stat *) MALLOC(sizeof(struct stat));

    if(system_stat(path, rq->finfo) == -1) {
        /* Tell future stats that this was not found */
        rq->statpath = STRDUP(path);
        rq->staterrno = _GET_SYSERRNO();
        if(rq->finfo) 
            FREE(rq->finfo); 
        rq->finfo = NULL; 
        if(rq->staterr)
            FREE(rq->staterr);
        rq->staterr = STRDUP(system_errmsg());
        return NULL;
    }
    rq->statpath = STRDUP(path);
    if(rq->staterr)
        FREE(rq->staterr);
    rq->staterr = NULL;
    return rq->finfo;
}

