/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
#ifdef XP_WIN32
#include <process.h>
#endif
#include "prthread.h"
#include "prinit.h"
#include "prio.h"
#include <prnetdb.h>
#include "private/pprio.h"
extern "C" {
#include "ssl.h"
}
#include "daemonprocessor.h"
#include "daemonrequest.h"
#include "daemonsession.h"
#include "httprequest.h"
#include "daemon.h"
#include "base/daemon.h"
#include "frame/conf.h"
#include "safs/ntlogbuf.h"
#include "frame/accel_file_cache.h"
#include "base/net.h"
#include "../base/netio.h"
#include "httpdaemon/dbthttpdaemon.h"
#include "base/util.h"
#ifdef XP_WIN32
#include "daemonprocessornt.h"
#elif defined(XP_UNIX)
#include "daemonprocessorux.h"
#else
    Cannot find appropriate daemon type;
#endif
#include "httpdaemon/multihome.h"
#include "frame/log.h"

#include "libadmsslutil/psetcssl.h"
#include "libadmsslutil/admsslutil.h"

#define DEFAULT_MAX_REQ_IN_PROGRESS     512

/* Bug fix #342507 */
#ifdef LINUX
#define DEFAULT_MIN_REQ_IN_PROGRESS     24
#else
#define DEFAULT_MIN_REQ_IN_PROGRESS     48
#endif /* LINUX */

#define DEFAULT_RECYCLE_THRESHOLD       256

#ifdef XP_UNIX
#define DEFAULT_WDOG_BASE_FD	10
void _daemon_exit();
extern PRBool _global_main_terminating;
#endif

extern "C" void PR_SetThreadRecycleMode(PRUint32);
extern "C" PRInt32 PR_GetThreadAffinityMask(PRThread *, PRUint32 *);
extern "C" PRInt32 PR_SetThreadAffinityMask(PRThread *, PRUint32);

const PRUint32 iProcessorNil = ((PRUint32) -1);

listenSocket_t     *DaemonProcessor::vpSocketListen = 0;
PRLock             *DaemonProcessor::socketListenLock = 0;
PRUint32            DaemonProcessor::cSocketListen = 0;

PRUint32            DaemonProcessor::mAffinityProcess = 0;
PRUint32            DaemonProcessor::mAffinityProcessorIo = 0;
PRUint32            DaemonProcessor::mAffinityDefault = 0;
PRUint32            DaemonProcessor::iProcessorIo = iProcessorNil;

PRUint32            DaemonProcessor::cConcurrency = nThreadIoPerProcessorMax;
PRUint32            DaemonProcessor::cRecycleThreshold = DEFAULT_RECYCLE_THRESHOLD;
PRInt32             DaemonProcessor::cMaxThreads = 0;
PRInt32             DaemonProcessor::cTotalBusyThreads = 0;
PRInt32             DaemonProcessor::cMaxThreadsPerSocket = 0;
PRInt32             DaemonProcessor::cMinThreadsPerSocket = 0;

PRLock             *DaemonProcessor::lockPostMinThreads;

PRUint32            DaemonProcessor::cSessionInProgress;
PRLock             *DaemonProcessor::sessionInProgressLock;
PRCondVar          *DaemonProcessor::sessionInProgressCvar;

PRUint32            nSocketSndBufSize = 0;
PRUint32            nSocketRcvBufSize = 0;

#ifdef DAEMON_NEEDS_SEMAPHORE
PRInt32             single_accept = SINGLE_ACCEPT_ON;
#endif

PRBool              DaemonProcessor::fTerminating = PR_FALSE;

PRUint32 CountBits (PRUint32 l);

PRBool DaemonProcessor::Initialize(daemon_s *pDaemon)
{
    PRBool fSuccess = PR_TRUE;

#ifdef XP_UNIX
    fSuccess = DaemonProcessorUX::Initialize(pDaemon);
#else
    fSuccess = DaemonProcessorNT::Initialize(pDaemon);
#endif
    if (!fSuccess) {
        ereport(LOG_FAILURE, "Failure initializing MD DaemonProcessor");
        return PR_FALSE;
    }
    
    fSuccess = DaemonRequest::Initialize ();
    if (!fSuccess) {
        ereport(LOG_FAILURE, "Failure initializing DaemonRequest");
        return PR_FALSE;
    }

    fSuccess = HttpRequest::Initialize ();
    if (!fSuccess) {
        ereport(LOG_FAILURE, "Failure initializing HttpRequest");
        return PR_FALSE;
    }

    fSuccess = DaemonSession::Initialize ();
    if (!fSuccess) {
        ereport(LOG_FAILURE, "Failure initializing DaemonSession");
        return PR_FALSE;
    }

    cMaxThreads = DEFAULT_MAX_REQ_IN_PROGRESS;
    cMaxThreads = GetConfigParm("RqThrottle", cMaxThreads);
    cMaxThreadsPerSocket = cMaxThreads;
    cMaxThreadsPerSocket = GetConfigParm("RqThrottleMaxPerSocket", cMaxThreadsPerSocket);
    cMinThreadsPerSocket = DEFAULT_MIN_REQ_IN_PROGRESS;
    cMinThreadsPerSocket = GetConfigParm("RqThrottleMinPerSocket", cMinThreadsPerSocket);

    fSuccess = InitializeListenSockets (pDaemon);
    if (!fSuccess) {
        ereport(LOG_FAILURE, "Failure initializing Listen Sockets");
        return PR_FALSE;
    }

    if ( (cMinThreadsPerSocket * cSocketListen) > cMaxThreads)
        cMinThreadsPerSocket = max (cMaxThreads / cSocketListen, 1);

#ifdef XP_UNIX
    fSuccess =  DaemonProcessorUX::LateInitialize(pDaemon);
    if (!fSuccess) {
        ereport(LOG_FAILURE, "Failure in late UX initialization");
        return PR_FALSE;
    }
#endif


    // Find the default affinity mask for this process
    PR_GetThreadAffinityMask (PR_CurrentThread(), &mAffinityProcess);

    // We will avoid using the IO processor in an MP configuration, if 
    // possible.  The i/o processor is currently set by an .ini argument, 
    // in one of two ways - either an affinity mask, 
    // "IoProcessorAffinity=xxx", or a numeric designation "IoProcessor=i".
    // (i = 0..nCpu-1).  The latter is easier, the former is provided for 
    // systems with multiple io processors.
    //
    mAffinityProcessorIo = (PRUint32)GetConfigParm("IoProcessorAffinity", 0); 
    if (! mAffinityProcessorIo) { 
        iProcessorIo = (PRUint32) GetConfigParm("IoProcessor", iProcessorIo); 
        mAffinityProcessorIo = 
            ((iProcessorIo != iProcessorNil)? (1 << iProcessorIo) : 0);
        if (mAffinityProcessorIo == mAffinityProcess) 
            mAffinityProcessorIo = 0; // never disable all available CPU's...
    }
    mAffinityDefault = (mAffinityProcess & ~mAffinityProcessorIo);

    PR_SetThreadAffinityMask(PR_CurrentThread(), mAffinityDefault);

    cConcurrency = CountBits(mAffinityDefault);
    cConcurrency = GetConfigParm("Concurrency", cConcurrency);

    cRecycleThreshold = DEFAULT_RECYCLE_THRESHOLD;
    cRecycleThreshold = GetConfigParm("RecycleThreshold",cRecycleThreshold);

    // Initialize static globals for DaemonProcessor

    lockPostMinThreads = PR_NewLock();
    if (!lockPostMinThreads) {
        ereport(LOG_FAILURE, "Unable to create lockPostMinThreads");
        return PR_FALSE;
    }

    socketListenLock = PR_NewLock();
    if (!socketListenLock) {
        ereport(LOG_FAILURE, "Unable to create socketListenLock");
        return PR_FALSE;
    }

    // Check for tunable socket buffer sizes
    nSocketSndBufSize = GetConfigParm("SndBufSize", nSocketSndBufSize);
    nSocketRcvBufSize = GetConfigParm("RcvBufSize", nSocketRcvBufSize);

    return PR_TRUE;
}

DaemonProcessor::DaemonProcessor () 
{
    fTerminating = PR_FALSE;

    mAffinityProcessor = mAffinityDefault;
    iSocketListenNext = 0;

    child_status(CHILD_AWAIT_CONNECT);

#ifndef IRIX
    //
    // Don't create multiple processors on irix until after the fork 
    //
    PR_SetConcurrency(cConcurrency);
#endif

    /* enable nspr thread recycling */
    PR_SetThreadRecycleMode(cRecycleThreshold);

    cSessionInProgress = 0;
    sessionInProgressLock = PR_NewLock();
    sessionInProgressCvar = PR_NewCondVar(sessionInProgressLock);

}

DaemonProcessor::~DaemonProcessor ()
{
    /* Can we terminate all the threads we created? */
}

#define TERMINATE_WAIT	10
#define DEFAULT_TERMINATE_TIMEOUT	3
PRBool DaemonProcessor::Run() 
{
    // Calling PostMinThreads() for each listen socket to start off
    // the DaemonSessions.
    for (PRUint32 cSocketPost = cSocketListen; cSocketPost--; )
        PostMinThreads(cSocketPost);

    // Now just wait for all our requests to die.  Can't wait for all 
    // Sessions to die because on UNIX some sessions are not-wakeupable.
    // The problem is that if they are in select, and we close the socket,
    // select doesn't wakeup.  This was confirmed by sun as normal, standards-
    // based behavior..  Gotta love it!
#ifdef XP_UNIX
    while(!_global_main_terminating) 
	PR_Sleep(PR_SecondsToInterval(1));
    if (cTotalBusyThreads > 0) {
    	char *waitTime_s = conf_findGlobal("TerminateTimeout");
    	PRInt32 waitTime = DEFAULT_TERMINATE_TIMEOUT;
    	if (waitTime_s) 
       	    waitTime = atoi(waitTime_s);
	while ((waitTime > 0) && (cTotalBusyThreads > 0)) {
	    PR_Sleep(PR_SecondsToInterval(1));
	    waitTime--;
	}
    }
    _daemon_exit();	/* no return */
#else
    PR_Lock(sessionInProgressLock);
    while(!fTerminating || cTotalBusyThreads > 0) {
        if (fTerminating)
            log_ereport(LOG_VERBOSE, "DaemonProcessor::Run terminating with %d sessions still in progress", cTotalBusyThreads);
        PR_WaitCondVar(sessionInProgressCvar, PR_SecondsToInterval(TERMINATE_WAIT));
    }
    PR_Unlock(sessionInProgressLock);
#endif

    log_ereport(LOG_VERBOSE, "DaemonProcessor::Run sessions all cleaned up");

    // Flush any outstanding log buffers
    logbuf_cleanup_all();

    // Cleanup the accelerator cache; otherwise tmp dirs may be left around
    accel_file_cache_destroy(NULL);

    return PR_TRUE;
}

PRBool DaemonProcessor::Terminate ()
{
    // Set the death flag 
    fTerminating = PR_TRUE;

    log_ereport(LOG_VERBOSE, "DaemonProcessor::Terminate set fTerminating flag");

    // XXXMB - warning! 
    // Shutdown isn't working right.  With the move to fibers and nspr20,
    // we have nspr fibers which are blocked in IO which we can't send a 
    // signal to.  So there is no way to kill them.  I tried just closing the
    // socket on them, hoping they would awaken and return, but NT 
    // goes into blue screen when I do that.
    //
    // NSPR is supposed to implement interrupts which allow you to abort 
    // an IO call.  Once those are implemented and working, we should be
    // able to manually abort the IO.

    // Shut down the listen sockets.  
    // This will wake all of the daemonsessions; they will see the 
    // fTerminating flag and go down.
    log_ereport(LOG_VERBOSE, "DaemonProcessor::Terminate calling TerminateListenSock");

#ifndef XP_UNIX
    TerminateListenSockets();
#endif

    log_ereport(LOG_VERBOSE, "DaemonProcessor::Terminated Listen Sockets");

    // Terminate any open session sockets.  Otherwise NT can't terminate 
    // correctly.
    DaemonSession::KillOpenSessionSockets();

    log_ereport(LOG_VERBOSE, "DaemonProcessor::Terminated Open Session Sockets");

    // The DaemonProcessor::Run() thread is waiting for sessions to die.  Poke
    // the condvar so that it wakes up quickly.
    NotifySessionInProgressWaiter();

    return PR_TRUE;
}

void DaemonProcessor::NotifySessionInProgressWaiter()
{
    // The DaemonProcessor::Run() thread is waiting for sessions to die.  Poke
    // the condvar so that it wakes up quickly.
    PR_Lock(sessionInProgressLock);
    PR_NotifyCondVar(sessionInProgressCvar);
    PR_Unlock(sessionInProgressLock);
}

// allocate a listen sockets for each ip domain home this server will accept
PRBool DaemonProcessor::InitializeListenSockets (daemon_s *pDaemon)
{
    PRBool fSuccess = PR_TRUE;
    PRStatus status;
    int    sslStatus;
    multiple_domains *domain_array;
    PRUint32 iSocketListen;
#ifdef XP_UNIX
    char *wdstr = getenv("WD_PIPE_FD");
    PRFileDesc *sd;
    int wdfd = DEFAULT_WDOG_BASE_FD;
    int flags;
#endif /* XP_UNIX */

    PRSocketOptionData *option_data;

    // Add the "primary" IP
    AddMultihomeDomain(conf_getglobals()->Vaddr,
                       NULL,
                       conf_getglobals()->Vport,
                       conf_getglobals()->Vsecurity_active?PR_TRUE:PR_FALSE,
                       conf_getglobals()->Vsecure_auth?PR_TRUE:PR_FALSE,
                       conf_getglobals()->Vssl3_active?PR_TRUE:PR_FALSE,
                       conf_getglobals()->Vssl2_active?PR_TRUE:PR_FALSE);

    cSocketListen = GetCountMultihomeDomains();
    domain_array = GetMultihomeDomains();

    if (cSocketListen <= 0)
        return PR_FALSE;

    // allocate array of listen socket pointers (one socket for each ip domain)
    vpSocketListen = new listenSocket_t[cSocketListen];
    for (iSocketListen=0; iSocketListen < cSocketListen; iSocketListen++) {
        vpSocketListen[iSocketListen].listenSocket = NULL;
        vpSocketListen[iSocketListen].cMinThreads = cMinThreadsPerSocket;
        vpSocketListen[iSocketListen].cMaxThreads = cMaxThreadsPerSocket;
        vpSocketListen[iSocketListen].cActiveThreads = 0;
        vpSocketListen[iSocketListen].cWaitingThreads = 0;
        vpSocketListen[iSocketListen].cBusyThreads = 0;
    }

    // open all listen sockets
    for (iSocketListen = 0; fSuccess && iSocketListen < cSocketListen; ++iSocketListen) {

        char *pzServerIp = domain_array [iSocketListen].address;
        int serverPort = domain_array[iSocketListen].port;
        PRBool serverSecurity = domain_array[iSocketListen].security;

#ifdef XP_UNIX
        if (wdstr) {
            if (serverSecurity) {
                sd = SSL_ImportFD(NULL, PR_ImportTCPSocket(wdfd++));
            }
            else {
                sd = PR_ImportTCPSocket(wdfd++);
            }
#ifdef SOLARIS
            /*
             * unset NONBLOCK flag; 
             */
	    /* Have no idea why Solaris want to unset NONBLOCK flag when it should 
	       be in NON-BLOCK mode, and new NSPR20 does not give file descriptor
	       back, so the code are removed --- yjh
            flags = fcntl(sd->osfd, F_GETFL, 0);
            fcntl(sd->osfd, F_SETFL, flags & ~O_NONBLOCK);
	    */
#endif
            net_listen(sd, net_listenqsize);
            domain_array[iSocketListen].sd = sd;
        }
        else {
            domain_array[iSocketListen].sd = net_create_listener(pzServerIp, serverPort);
        }
#else
        domain_array[iSocketListen].sd = net_create_listener(pzServerIp, serverPort);
#endif /* XP_UNIX */

        vpSocketListen[iSocketListen].listenSocket = domain_array[iSocketListen].sd;

        if (vpSocketListen[iSocketListen].listenSocket != (PRFileDesc *)-1) {
            fSuccess = PR_TRUE;
        } else {
            vpSocketListen[iSocketListen].listenSocket = 0;
            fSuccess = PR_FALSE;
        }

#ifdef XP_UNIX
        if (nSocketSndBufSize > 0) {
	  /* 19980120 NSPR uses data struct, not enum - adam */
    	  option_data = (PRSocketOptionData *)MALLOC(sizeof(PRSocketOptionData));
	  option_data->option = PR_SockOpt_SendBufferSize;
	  option_data->value.send_buffer_size = (PRSize)nSocketSndBufSize;
	  status = PR_SetSocketOption(vpSocketListen[iSocketListen].listenSocket,
				      option_data);
	  FREE(option_data);

	  /* for ns security/ nspr21 integration
            setsockopt( PR_FileDesc2NativeHandle(vpSocketListen[iSocketListen].listenSocket),
                        SOL_SOCKET, 
                        SO_SNDBUF, 
                        (const char *)&nSocketSndBufSize, 
                        sizeof(nSocketSndBufSize));
			*/
        }
        if (nSocketRcvBufSize > 0) {
	  /* 19980120 NSPR uses data struct, not enum - adam */
    	  option_data = (PRSocketOptionData *)MALLOC(sizeof(PRSocketOptionData));
	  option_data->option = PR_SockOpt_RecvBufferSize;
	  option_data->value.recv_buffer_size = nSocketRcvBufSize;
	  status = PR_SetSocketOption(vpSocketListen[iSocketListen].listenSocket,
				      option_data);
	  FREE(option_data);

	   /* for ns security/ nspr21 integration
            setsockopt( PR_FileDesc2NativeHandle(vpSocketListen[iSocketListen].listenSocket),
                        SOL_SOCKET, 
                        SO_RCVBUF, 
                        (const char *)&nSocketRcvBufSize, 
                        sizeof(nSocketRcvBufSize));
			*/
        }
#endif

        if (fSuccess && serverSecurity == PR_TRUE) {
            PsetHndl pset;
            int errorCode;

            PRFileDesc *sock = vpSocketListen[iSocketListen].listenSocket;

            pset = psetCreateSSL("admin-serv", 
		                 /* configRoot */ pDaemon->admroot,
		                 /* userDN */ NULL,
		                 /* passwd */ NULL,
		                 &errorCode);

            /* Set SSL for each socket - adam */
            servssl_init(pset, sock, pDaemon->admroot);            

// << SSL_Enable in not available in NSS3> 
#if 0

            sslStatus = SSL_Enable(sock, SSL_SECURITY, PR_TRUE);
            sslStatus = SSL_Enable(sock, SSL_ENABLE_SSL3, pDaemon->ssl3 ? PR_TRUE : PR_FALSE);
            sslStatus = SSL_Enable(sock, SSL_ENABLE_SSL2, pDaemon->ssl2 ? PR_TRUE : PR_FALSE);
	    /* This option is removed from ns security 19971028 release
            SSL_Enable(sock, SSL_DELAYED_HANDSHAKE, 1);
	    */
            if(domain_array[iSocketListen].secure_auth == PR_TRUE)
                sslStatus = SSL_Enable(sock, SSL_REQUEST_CERTIFICATE, PR_TRUE);
#endif
        }

#ifdef XP_UNIX
	/* Since this only assign PR_Accept as ssl accept function, it may not be needed with the
	   new ns security. nspr21 integration
        if (net_accept_init(serverPort) == -1) {
            AdminFprintf(stderr, XP_GetAdminStr(DBT_seminitFailedSN_), 
                system_errmsg());
            fSuccess = PR_FALSE;
        }
	*/

        if (fSuccess) {
	  if(pzServerIp) {
	    PRStatus  s;
	    PRBool    hostflag;
	    PRNetAddr addr;
	    char      buf[PR_NETDB_BUF_SIZE];
	    PRHostEnt hent;

	    hostflag = PR_FALSE;
	    s = PR_StringToNetAddr(pzServerIp, &addr);

	    if (s == PR_SUCCESS) {
	      s = PR_GetHostByAddr(&addr, 
				   buf,
				   PR_NETDB_BUF_SIZE,
				   &hent);
	      if (s == PR_SUCCESS) {
		if (hent.h_name) hostflag = PR_TRUE;
	      }
	    }

	    AdminFprintf(stderr, 
			 XP_GetAdminStr(DBT_startupListeningToSPortDAsSN_), 
			 serverSecurity==PR_TRUE?"https":"http",
			 hostflag?hent.h_name:pzServerIp, 
			 serverPort, 
			 pDaemon->pw->pw_name);
	  }
	  else {
	    char *lhost;

	    lhost = util_hostname();
	    AdminFprintf(stderr, 
			 XP_GetAdminStr(DBT_startupListeningToPortDAsSN_), 
			 serverSecurity==PR_TRUE?"https":"http", 
			 lhost?lhost:"unknown", 
			 serverPort, 
			 pDaemon->pw->pw_name); 
            } 
        } else {
	    if(pzServerIp)
	       AdminFprintf(stderr, XP_GetAdminStr(DBT_couldNotBindToPortDH_), serverPort, pzServerIp, system_errmsg());
            else
	       AdminFprintf(stderr, XP_GetAdminStr(DBT_couldNotBindToPortD_), serverPort, system_errmsg());
        }
#endif
        if (!fSuccess) {
            if(pzServerIp)
                ereport(LOG_FAILURE, XP_GetAdminStr(DBT_couldNotBindToPortDH_), 
                    serverPort, pzServerIp, system_errmsg());
            else 
                ereport(LOG_FAILURE, XP_GetAdminStr(DBT_couldNotBindToPortD_), 
                    serverPort, system_errmsg());
        }
    }

    return fSuccess;
}

void DaemonProcessor::TerminateListenSockets ()
{
    for (PRUint32 iSocketListen = 0; iSocketListen < cSocketListen; ++iSocketListen) {
        if (vpSocketListen [iSocketListen].listenSocket != NULL) {
            PR_Close(vpSocketListen[iSocketListen].listenSocket);
            /* Don't set ptr to NULL here; there is a race where the
             * accept threads might try to use the listen socket still.
             * If they do, they'll crash.  This race is bad, but there isn't
             * much we can do without putting a big lock in the accept thread
             * as well.... XXXMB... Maybe we need a "close but don't free" 
             * function?
             */
        }
    }
    cSocketListen = 0;
}

void DaemonProcessor::SetListenSocketsNonInheritable()
{
    for (PRUint32 iSocketListen=0; 
        iSocketListen < cSocketListen; 
        iSocketListen++)
        if (vpSocketListen[iSocketListen].listenSocket)
            file_setinherit(vpSocketListen[iSocketListen].listenSocket, 0);
}

PRInt32 DaemonProcessor::PostMinThreads (PRInt32 socketIndex)
{
    PRInt32 numCount = 0;

    if ((vpSocketListen[socketIndex].cWaitingThreads < cMinThreadsPerSocket) ) {
        PR_Lock(socketListenLock);
        if (FThrottleOpen(socketIndex)) {
            numCount = cMinThreadsPerSocket - vpSocketListen[socketIndex].cWaitingThreads;
	    if (numCount > 0)
                vpSocketListen[socketIndex].cWaitingThreads = cMinThreadsPerSocket;
        }
        PR_Unlock(socketListenLock);
        while (numCount > 0) {
            DaemonSession * pSession = NewAcceptSession(socketIndex);
            numCount--;
        }
    }

    return 0;
}

PRUint32 CountBits (PRUint32 i)
{
    PRInt32 c = 0;

    while (i) {
        #if 1
            if (i & 1)
                ++c;
            i = i >> 1;
        #else
            ++c;
            i = i & ((~i) - 1); // turns off a single bit
        #endif
    }

    return c;
}

DaemonProcessor *DaemonProcessor::NewDaemonProcessor()
{
#ifdef XP_WIN32
    return new DaemonProcessorNT();
#elif defined(XP_UNIX)
    return new DaemonProcessorUX();
#else
    Cannot find appropriate daemon type;
#endif
}

PRBool 
DaemonProcessor::GetListenSocketInfo(PRUint32 *numSockets, 
                                     listenSocketInfo **infoArray)
{
    multiple_domains *domainArray = GetMultihomeDomains();
    int index;


    *numSockets = cSocketListen;

    *infoArray = 
        (listenSocketInfo *)MALLOC(*numSockets * sizeof(listenSocketInfo));
    if (!infoArray)
        return PR_FALSE;

    for (index=0; index < *numSockets; index++) {
        (*infoArray)[index].port = domainArray[index].port;
        (*infoArray)[index].ipAddr = domainArray[index].address;
        (*infoArray)[index].securityEnabled = domainArray[index].security;

        (*infoArray)[index].numActiveThreads = vpSocketListen[index].cActiveThreads;
        (*infoArray)[index].numWaitingThreads = vpSocketListen[index].cWaitingThreads;
        (*infoArray)[index].numBusyThreads = vpSocketListen[index].cBusyThreads;
        (*infoArray)[index].numIdleThreads = 0;
        
        (*infoArray)[index].numMinThreads = vpSocketListen[index].cMinThreads;
        (*infoArray)[index].numMaxThreads = vpSocketListen[index].cMaxThreads;
    }

    
    return PR_TRUE;
}
