/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
** DaemonProcessor - per CPU definitions
**
**
**  The accelerator is modelled as a set of DaemonProcessors - usually one 
**  per CPU on an SMP
**  system.  
**
**  Each DaemonProcessor fields incoming connections on one or more listen 
**  sockets; each connection instantiates an DaemonRequest/HttpRequest.  
**  DaemonRequest captures the primitive async socket connection, while 
**  HttpRequest extends that functionality with a state-machine for handling
**  the phases of http-request processing (eg, reading header, writing 
**  results, etc).  Only requests that can be delivered from a content 
**  cache - without continued NSAPI support, etc - are fully processed in 
**  the accelerator.  Other requests (and CGI, etc) are handed off to 
**  synchronous server threads that run the main gamut of httpd functionality.
**
**  The state machinery in HttpRequest is designed to be fairly extensible, 
**  so that it could be used to accelerate other servers (eg, proxy) with 
**  different state sequences.
**
**  To avoid lock/memory/etc contention across CPU's, each DaemonProcessor 
**  has it's own pool of threads, requests and state information.  When a 
**  request is received by an DaemonProcessor, it will be processed by that 
**  processor until it's complete.  The idea is that there should be very 
**  little memory (or lock) contention between different CPU's on an SMP 
**  system, since they have independent state and handle discrete pools of 
**  requests.  Also the requests/state/etc will ideally be able to reside 
**  within each CPU's fast-memory cache.  This design provides proportionally 
**  more advantages as the number of processors grows.
**
**  Some optimizations are in place to control the configuration on an SMP 
**  system, eg, whether all available processors are used, and how to throttle
**  them individually, balance between them, etc.  The biggest practical 
**  utility of this so far has been seen when simply identifying which 
**  processor appears to be running most of the TCP/IP stack on NT (not to 
**  be confused with the processor servicing the i/o interrupts).  For some 
**  reason most of the network stack seems to run on one processor on NT, 
**  which has proven to be an asymmetric bottleneck (the specific processor  
**  varies across systems; most of its CPU time is spent in priveleged mode).
**  When we explicitly avoid running an DaemonProcessor on this particular 
**  CPU, we see better performance.  So far, this is most profound on a 2 CPU 
**  system - where, ironically, we perform best when we allocate only one 
**  DaemonProcessor to avoid running any of our server code on the "IoProcessor"
**  bottleneck.  Watching perfmon CPU utilization with and without these 
**  assignments is revealing.  On a 4-CPU NT system, usually the IO processor 
**  is such a bottleneck that it cannot fully utilize a 100MB ethernet 
**  connection - the additional processors become a luxury for static HTML 
**  (on a system running a lot of CGI's, java, etc, the additional horsepower 
**  becomes significant, of course).
*/

#ifndef _DAEMONPROCESSOR_H_
#define _DAEMONPROCESSOR_H_

#include "prtypes.h"
#include "prlock.h"
#include "prcvar.h"
#include "pratom.h"

#include "libdaemon.h"
#include "base/daemon.h"
#include "base/nsassert.h"
#include "httpdaemon/internalstats.h"

// Some forward/external defns
class DaemonSession;
class DaemonRequest;

#define nThreadIoPerProcessorMax 1

typedef struct listenSocket_t {
    PRFileDesc *listenSocket;
    PRUint32    cActiveThreads;   /* # of threads active (wait+busy counts),
                                   * used for throttling
                                   */
    PRUint32    cWaitingThreads;  /* # of threads waiting in accept or read */
    PRUint32    cBusyThreads;     /* # of threads processing requests */
    PRUint32    cMinThreads;   /* Constant */
    PRUint32    cMaxThreads;   /* Constant */
} listenSocket_t;


class DaemonProcessor {

 public:

    // Initialize the daemonProcessor; read the config files, etc
    static PRBool Initialize(daemon_s *pDaemon);

    // Terminate the daemonProcessor
    PRBool Terminate ();

    // NewDaemonProcessor is the method for creating a daemonprocessor.
    // Depending on the platform we are on it will create either a 
    // DaemonProcessorUX or a DaemonProcessorNT; but the caller doesn't 
    // need to know; it just works.
    static DaemonProcessor *NewDaemonProcessor();

    // Create/Delete a Daemon Processor tied to the given CPU mask
    DaemonProcessor ();
    ~DaemonProcessor ();

    // Run/shutdown DaemonProcessor and component systems
    PRBool Run();

    // Create an appropriate DaemonSession for this DaemonProcessor
    // This session starts by waiting in select
    virtual DaemonSession *NewAcceptSession(PRInt32 index) = 0;

    // Some accessors
    PRBool                     FThrottleOpen(PRInt32 socketIndex);
    PRBool                     FSafeThrottleOpen(PRInt32 socketIndex);

    // Check the terminating flag to see if we are going down
    PRBool                     FTerminating();

    // Daemon Accept Counter Management
    void                       IncrementWaitingCount(PRInt32 socketIndex);
    void                       IncWaitAndDecBusyCount(PRInt32 socketIndex);
    void                       DecWaitAndIncBusyCount(PRInt32 socketIndex);
    PRBool                     TestDecWaitingCount(PRInt32 socketIndex);
    void                       IncrementActiveCount(PRInt32 socketIndex);
    void                       DecActiveAndWaitCount(PRInt32 socketIndex);
    void                       DecBusyCount(PRInt32 socketIndex);
    void                       IncBusyCount(PRInt32 socketIndex);

    // Session Counter Management
    void                       IncrementSessionsInProgress();
    void                       DecrementSessionsInProgress();
    void                       NotifySessionInProgressWaiter();

    void                       SetListenSocketsNonInheritable();

    // number of NSPR "CPU"s to allocate (damnit, needs to be public for IRIX)
    static PRUint32            	      cConcurrency;

    // threshold after which threads should not be recycled in NSPR
    static PRUint32		      cRecycleThreshold;
    
    // debugging ingo
    PRBool                     GetListenSocketInfo(PRUint32 *numSockets,
                                               listenSocketInfo **infoArray);
 private:

    // Make sure the minimum number of threads are posted for the
    // listen socket.
    PRInt32                        PostMinThreads (PRInt32 socketIndex);

    // When SSL is on, most of the accelerator is bypassed (SSL interfaces do 
    // not support async io.)  SecureThreads are somewhat simplified 
    // "worker/sync" threads.
    static void                SecureThread ();

    static PRBool              InitializeListenSockets (daemon_s *pDaemon);

    void                       TerminateListenSockets ();

    friend class DaemonRequest; // just for free list access, ug
    friend class DaemonSession; // just for free list access, ug

    // # of listen sockets (> 1 when ip-multihoming) 
    static PRUint32            cSocketListen;

    // vector of listen sockets
    static listenSocket_t      *vpSocketListen;
    static PRLock              *socketListenLock;
 
    // next listen socket for round-robin accepts
    PRUint32                   iSocketListenNext;

    // mask of processors for this process
    static PRUint32              mAffinityProcess;

    // default affinity for our threads 
    static PRUint32              mAffinityDefault;

    // bottleneck i/o processor 
    static PRUint32            iProcessorIo;

    // affinity of i/o processor
    static PRUint32            mAffinityProcessorIo;

    // mask of this specific processor
    PRUint32                   mAffinityProcessor;

    // Flag if this daemonprocessor has been asked to terminate
    static PRBool              fTerminating;

    // guard PostMinThreads for posting/counting # of outstanding accepts
    static PRLock                    *lockPostMinThreads;

    // # of requests ceiling: current flow limit
    static PRInt32             cMaxThreads;
    static PRInt32             cMinThreadsPerSocket;
    static PRInt32             cMaxThreadsPerSocket;  /* waiting or busy */
        // Total number of threads busy across all sockets 
    static PRInt32             cTotalBusyThreads;
    
public:
    static PRUint32            cSessionInProgress;
    static PRLock             *sessionInProgressLock;
    static PRCondVar          *sessionInProgressCvar;
};

 
inline PRBool DaemonProcessor::FThrottleOpen(PRInt32 socketIndex)
{ 
    listenSocket_t *pSocket = &vpSocketListen[socketIndex];

    /* Number of busy threads across all sockets must be less than RqThrottle.
     * Also, this particular socket cannot have more threads than it is 
     * allowed.
     */
    return ((cTotalBusyThreads < cMaxThreads) &&
            (pSocket->cActiveThreads < pSocket->cMaxThreads))?PR_TRUE:PR_FALSE;
}

inline PRBool DaemonProcessor::FSafeThrottleOpen(PRInt32 socketIndex)
{
    PRBool fOpen;

    PR_Lock(socketListenLock);
    if ((fOpen = FThrottleOpen(socketIndex)) == PR_FALSE) {
        /* Do what DecActiveAndWaitCount() does */
        vpSocketListen[socketIndex].cActiveThreads--;
        vpSocketListen[socketIndex].cWaitingThreads--;
    }
    PR_Unlock(socketListenLock);

    return fOpen;
}

inline void DaemonProcessor::IncrementWaitingCount(PRInt32 socketIndex)
{
    PR_Lock(socketListenLock);
    vpSocketListen[socketIndex].cWaitingThreads++;
    PR_Unlock(socketListenLock);
}

inline void DaemonProcessor::DecWaitAndIncBusyCount(PRInt32 socketIndex)
{
    PR_Lock(socketListenLock);
    vpSocketListen[socketIndex].cWaitingThreads--;
    vpSocketListen[socketIndex].cBusyThreads++;
    cTotalBusyThreads++;
    PR_Unlock(socketListenLock);
}

inline void DaemonProcessor::IncWaitAndDecBusyCount(PRInt32 socketIndex)
{
    PR_Lock(socketListenLock);
    vpSocketListen[socketIndex].cWaitingThreads++;
    vpSocketListen[socketIndex].cBusyThreads--;
    cTotalBusyThreads--;
    PR_Unlock(socketListenLock);
}

inline void DaemonProcessor::DecBusyCount(PRInt32 socketIndex)
{
    PR_Lock(socketListenLock);
    vpSocketListen[socketIndex].cBusyThreads--;
    vpSocketListen[socketIndex].cActiveThreads--;
    cTotalBusyThreads--;
    PR_Unlock(socketListenLock);
}

inline void DaemonProcessor::IncBusyCount(PRInt32 socketIndex)
{
    PR_Lock(socketListenLock);
    vpSocketListen[socketIndex].cBusyThreads++;
    vpSocketListen[socketIndex].cActiveThreads++;
    cTotalBusyThreads++;
    PR_Unlock(socketListenLock);
}

inline void DaemonProcessor::IncrementActiveCount(PRInt32 socketIndex)
{
    PR_Lock(socketListenLock);
    vpSocketListen[socketIndex].cActiveThreads++;
    PR_Unlock(socketListenLock);
}

inline void DaemonProcessor::DecActiveAndWaitCount(PRInt32 socketIndex)
{
    PR_Lock(socketListenLock);
    vpSocketListen[socketIndex].cActiveThreads--;
    vpSocketListen[socketIndex].cWaitingThreads--;
    PR_Unlock(socketListenLock);
}

/* Caller must be in the "waiting" state */
inline PRBool DaemonProcessor::TestDecWaitingCount(PRInt32 socketIndex)
{
    PRBool rv = PR_FALSE;

    PR_Lock(socketListenLock);
    if (vpSocketListen[socketIndex].cWaitingThreads > vpSocketListen[socketIndex].cMinThreads) {
        /* Do what DecActiveAndWaitCount() does */
        vpSocketListen[socketIndex].cActiveThreads--;
        vpSocketListen[socketIndex].cWaitingThreads--;
        rv = PR_TRUE;
    }
    PR_Unlock(socketListenLock);

    return rv;
}

inline PRBool DaemonProcessor::FTerminating()
{
    return fTerminating;
}

inline void DaemonProcessor::IncrementSessionsInProgress()
{
    PR_Lock(sessionInProgressLock);
    cSessionInProgress++;
    PR_Unlock(sessionInProgressLock);
}

inline void DaemonProcessor::DecrementSessionsInProgress()
{
    PR_Lock(sessionInProgressLock);
    cSessionInProgress--;
    if (cSessionInProgress == 0)
        PR_NotifyCondVar(sessionInProgressCvar);
    PR_Unlock(sessionInProgressLock);
}

#endif /* _DAEMONPROCESSOR_H_ */
