/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
//
// DaemonSession.cpp
//
#include "httpdaemon/dbthttpdaemon.h"

#include "daemonsession.h"
#include "httprequest.h"
#include "prthread.h"
#include "prio.h"
#include "prerror.h"
#include "frame/accel_file_cache.h"
#ifdef XP_UNIX
#include "daemonsessionux.h"
#else
#include "daemonsessionnt.h"
#endif
#include "httpdaemon/multihome.h"
#if defined( XP_WIN32 ) || defined( AIX )
#include <time.h>
#endif
#include "frame/log.h"
#include "prlock.h"

#include "private/pprio.h"
/*
#include "base/fasttime.h"
*/

extern int http_keepalive_timeout;
#ifdef DAEMON_NEEDS_SEMAPHORE
extern PRInt32 single_accept;
#endif

const PRUint32 sInetAddrPadded = sizeof(struct sockaddr_in) + 16;

PRInt32        DaemonSession::sBuffer   = 2048;  // sizeof buffer for initial read
PRInt32        DaemonSession::nbDataRaw = sBuffer + 2*sInetAddrPadded/*for AcceptEx addresses*/;
PRUint32       DaemonSession::iMagicAlloc = 0x32e45Fd1;
PRUint32       DaemonSession::iMagicFree  = 0x8a09078d;

PRUint32       DaemonSession::sessionStackSize  = 0;
PRLock        *DaemonSession::sessionListLock = 0; 
DaemonSession *DaemonSession::pSessionFreeHead = 0;
DaemonSession *DaemonSession::pSessionBusyHead = 0;
DaemonSession *DaemonSession::pSessionBusyTail = 0;
PRUint32       DaemonSession::cSessionFreeListSize = 0;
PRUint32       DaemonSession::cSessionBusyListSize = 0;

static PRInt32  fUseGlobalThreadsOnly = 0;

const int DEFAULT_ACCEPT_TIMEOUT = 30;
PRIntervalTime acceptTimeoutInterval = 0;

const int DEFAULT_KEEPALIVE_TIMEOUT = 30;
PRIntervalTime keepAliveTimeoutInterval = 0;

multiple_domains *DaemonSession::domainArray;

const int DEFAULT_NUM_KEEPALIVES = 200;
KeepAliveList *DaemonSession::KAList;


// Initialize global state for DaemonSession
PRBool DaemonSession::Initialize ()
{
    sBuffer   = GetConfigParm("HeaderBufferSize", sBuffer);
    nbDataRaw = sBuffer + 2 * sInetAddrPadded;
    PRInt32 maxKeepAlives = DEFAULT_NUM_KEEPALIVES;

    fUseGlobalThreadsOnly = 
        GetConfigParm("KernelThreads", fUseGlobalThreadsOnly);

    /* Free/Busy list management */
    sessionListLock = PR_NewLock();
    if (!sessionListLock) {
        ereport(LOG_FAILURE, "Unable to create sessionListLock");
        return PR_FALSE;
    }
    pSessionFreeHead = 0;
    pSessionBusyHead = 0;
    pSessionBusyTail = 0;
    cSessionFreeListSize = 0;
    cSessionBusyListSize = 0;

    /* Configure the accept timeout */
    acceptTimeoutInterval = 
        GetConfigParm("AcceptTimeout", acceptTimeoutInterval);
    if (acceptTimeoutInterval == 0)
        acceptTimeoutInterval = DEFAULT_ACCEPT_TIMEOUT;
    acceptTimeoutInterval = PR_SecondsToInterval(acceptTimeoutInterval);

    /* Configure the keeplalive timeout */
    keepAliveTimeoutInterval = 
        GetConfigParm("KeepAliveTimeout", keepAliveTimeoutInterval);
    if (keepAliveTimeoutInterval == 0)
        keepAliveTimeoutInterval = DEFAULT_KEEPALIVE_TIMEOUT;
    log_ereport(LOG_VERBOSE, "KeepAliveTimeout set to %ds", 
                keepAliveTimeoutInterval);
    keepAliveTimeoutInterval = PR_SecondsToInterval(keepAliveTimeoutInterval);

    sessionStackSize = 
        GetConfigParm("StackSize", sessionStackSize);

    domainArray = GetMultihomeDomains();

    maxKeepAlives = GetConfigParm("MaxKeepAliveConnections", maxKeepAlives);
    KAList = new KeepAliveList(maxKeepAlives);
    log_ereport(LOG_VERBOSE, "MaxKeepAliveConnections set to %d", 
                maxKeepAlives);

    return PR_TRUE;
}

DaemonSession::DaemonSession (
                   PRInt32 _socketIndex, 
                   DaemonProcessor *_daemonProcessor) : rqCurrent(_daemonProcessor, this)
{
    NS_ASSERT (iMagic == iMagicAlloc);
    NS_ASSERT (pbData && nbData == sBuffer); // buffer allocated in operator new

    daemonProcessor = _daemonProcessor;

    fKeepAliveRequested = PR_FALSE;
    socketIndex = _socketIndex;
}

PRBool
DaemonSession::StartThread()
{
    if (!thread) {
        thread = PR_CreateThread(PR_USER_THREAD,
                                 CThreadMain,
                                 this,
                                 PR_PRIORITY_NORMAL,
                                 fUseGlobalThreadsOnly?PR_GLOBAL_THREAD:PR_LOCAL_THREAD,
                                 PR_UNJOINABLE_THREAD,
                                 sessionStackSize);
        if (!thread) {
            ereport(LOG_FAILURE, "StartThread failed to create thread");
            return PR_FALSE;
        }
    }
    else {
        PR_Lock(workLock);
        NS_ASSERT(workState == WORKSTATE_IDLE);
        workState = WORKSTATE_READY;
        PR_NotifyCondVar(workCvar);
        PR_Unlock(workLock);
    }

    return PR_TRUE;
}

// Remove session, including destroying any inactive socket, etc
DaemonSession::~DaemonSession ()
{
}

extern "C" 
void CThreadMain(void *vDaemonSession)
{
    DaemonSession *daemonSession = (DaemonSession *)vDaemonSession;

    daemonSession->ThreadMain();

    delete daemonSession;
}

/* GetConnection()
 *
 * Will get the first "chunk" of data from the client.
 * If this is a keepalive connection, will just do a read from the socket.
 * If no connection is open, will do an accept()
 *
 * Returns TRUE if it got data and is ready to proceed; 
 * Returns FALSE otherwise.
 */
#ifdef DAEMON_NEEDS_SEMAPHORE
PRBool
DaemonSession::GetConnection()
{
    PRInt32 err, oserr;
    char    errorText[256];
    memset(errorText, '\0', 256);

    NS_ASSERT(socketIndex >= 0);

    /* Get a new connection */

    socketListen = daemonProcessor->vpSocketListen[socketIndex].listenSocket;
    serverIP = domainArray[socketIndex].address;
    serverPort = domainArray[socketIndex].port;

    if (single_accept == SINGLE_ACCEPT_ON) {
        PR_Lock(domainArray[socketIndex].sockLock);
    }

    if (daemonProcessor->FTerminating() == PR_TRUE) {
        if (single_accept == SINGLE_ACCEPT_ON)
            PR_Unlock(domainArray[socketIndex].sockLock);
        return PR_FALSE;
    }

    /* At the end of the data buffer we have space for a few PRNetAddr
     * structures.
     */
    remoteAddress = (struct sockaddr_in *)((char *)pbData + nbData);

    while ( (socketAccept  = PR_Accept(socketListen,
                                       (PRNetAddr *)remoteAddress,
                                       PR_INTERVAL_NO_TIMEOUT)) == NULL) {
        if (daemonProcessor->FTerminating() == PR_TRUE) {
            if (single_accept == SINGLE_ACCEPT_ON)
                PR_Unlock(domainArray[socketIndex].sockLock);
            return PR_FALSE;
        }

	err = 0;
        err = PR_GetError();
#ifdef SOLARIS
        /* solaris spits out spurrious EPROTO errors from accept() */
        oserr = PR_GetOSError();
        if (err != PR_IO_TIMEOUT_ERROR && 
            (err != PR_UNKNOWN_ERROR && oserr != EPROTO)) {
#else
        if (err != PR_IO_TIMEOUT_ERROR) {
#endif
	  PR_GetErrorText(errorText);
	  ereport(LOG_FAILURE, "Error accepting connection (error code: %d (%s))", 
		  err, errorText);
	  if (single_accept == SINGLE_ACCEPT_ON)
                PR_Unlock(domainArray[socketIndex].sockLock);
            return PR_FALSE;
        }
    }

    /* Release lock before calling recv */
    if (single_accept == SINGLE_ACCEPT_ON)
        PR_Unlock(domainArray[socketIndex].sockLock);

    cbData = -1;
    while (cbData < 0) {
      cbData = PR_Recv(socketAccept, 
		       pbData, 
		       nbData, 
		       0, 
		       acceptTimeoutInterval);
      if (cbData < 0) {
        err = PR_GetError();
	PR_GetErrorText(errorText);
	if (err == PR_WOULD_BLOCK_ERROR) {
	  log_ereport(LOG_FAILURE, "PR_WOULD_BLOCK_ERROR");
	}
	else {
	  if (err != PR_IO_TIMEOUT_ERROR && err != PR_CONNECT_RESET_ERROR) {
            log_ereport(LOG_FAILURE, "Error receiving connection (%s)", errorText);
	  }
	  PR_Close(socketAccept);
	  return PR_FALSE;
	}
      }
    }
 
    return PR_TRUE;
}

#else /* DAEMON_NEEDS_SEMAPHORE */

PRBool
DaemonSession::GetConnection()
{
    PRInt32 err;
    char    errorText[256];
    memset(errorText, '\0', 256);

    NS_ASSERT(socketIndex >= 0);

    /* Get a new connection */

    socketListen = daemonProcessor->vpSocketListen[socketIndex].listenSocket;
#ifndef HPUX
    /* On HPUX, any listen socket may accept a connection for any local IP */
    serverIP = domainArray[socketIndex].address;
    serverPort = domainArray[socketIndex].port;
#endif /* HPUX */

    if (daemonProcessor->FTerminating() == PR_TRUE) {
        return PR_FALSE;
    }
#ifdef XP_UNIX
	/* atulb: 6/4/98  bug#118016.  PR_AcceptRecv() timesout out the 
		connection incorrectly, leaving behind a lost HTTP request.  
		See also bug#122476 */

	/* adam: 7/31/98 bug#252382. *Sigh*... HPUX needs to do the same thing.. */
#if defined( IRIX ) || defined ( HPUX )
    remoteAddress = (struct sockaddr_in *)((char *)pbData + nbData);

    while ( (socketAccept  = PR_Accept(socketListen,
                                       (PRNetAddr *)remoteAddress,
                                       acceptTimeoutInterval)) == NULL) {
        if (daemonProcessor->FTerminating() == PR_TRUE) 
            return PR_FALSE;
    

        err = 0;
        if ((err = PR_GetError()) != PR_IO_TIMEOUT_ERROR) {
          PR_GetErrorText(errorText);
          ereport(LOG_FAILURE, "Error accepting connection (error code: %d (%s))", err, errorText);
          return PR_FALSE;
        } 
    }

    cbData = -1;
    while (cbData < 0) {
      cbData = PR_Recv(socketAccept,
                       pbData,
                       nbData,
                       0,
                       acceptTimeoutInterval);
      if (cbData < 0) {
        err = PR_GetError();
        PR_GetErrorText(errorText);
        if (err == PR_WOULD_BLOCK_ERROR) {
          log_ereport(LOG_FAILURE, "PR_WOULD_BLOCK_ERROR");
        }
        else {
          if (err != PR_IO_TIMEOUT_ERROR && err != PR_CONNECT_RESET_ERROR) {
            log_ereport(LOG_FAILURE, "Error receiving connection (%s)", errorText);
          }
          PR_Close(socketAccept);
          return PR_FALSE;
        }
      }
    }
#else /* For non Irix and HPUX Unix */
    while ( (cbData = PR_AcceptRead(socketListen,
                                    &socketAccept,
                                    (PRNetAddr **)&remoteAddress,
                                    pbData,
                                    nbData, acceptTimeoutInterval)) < 0) {
#endif  /* ifdef IRIX || HPUX */
#else  /* XP_UNIX  */
    for (;;) {
#ifdef WIN95
            cbData = PR_AcceptRead(socketListen,
                                   &socketAccept,
                                   (PRNetAddr **)&remoteAddress,
                                   pbData,
                                   nbData, acceptTimeoutInterval);
#else

        if (domainArray[socketIndex].security == PR_FALSE) {
            cbData = PR_NTFast_AcceptRead(
                                    socketListen,
                                    &socketAccept,
                                    (PRNetAddr **)&remoteAddress,
                                    pbData,
                                    nbData,
                                    acceptTimeoutInterval);
        }
        else {
            /*
             * timeout interval = none for SSL connections - adam
             */
            cbData = PR_AcceptRead(socketListen,
                                   &socketAccept,
                                   (PRNetAddr **)&remoteAddress,
                                   pbData,
                                   nbData, PR_INTERVAL_NO_TIMEOUT);
        }
#endif

        if (cbData >= 0) break;
#endif
#if !defined( IRIX ) && !defined( HPUX )  /* On Irix/HPUX, the error checking has been along with the PR_Recv */
        if (daemonProcessor->FTerminating() == PR_TRUE) {
            return PR_FALSE;
        }

        err = PR_GetError();
        if (err != PR_IO_TIMEOUT_ERROR) {
            ereport(LOG_FAILURE, "Error accepting connection %d (%s)",
                    err, system_errmsg());
        } else {
            return PR_FALSE;
        }
    }
#endif  /* if !IRIX && !HPUX */

#ifdef HPUX
    {
        PRNetAddr loc_addr;
        struct _fourbytes {
            unsigned char b1;
            unsigned char b2;
            unsigned char b3;
            unsigned char b4;
        } *ip4;

        /* Find out what local IP address the connection was for */
        PR_GetSockName(socketAccept, &loc_addr);
        ip4 = (struct _fourbytes *)&loc_addr.inet.ip;
        sprintf(localIP, "%d.%d.%d.%d", ip4->b1, ip4->b2, ip4->b3, ip4->b4);
        
        serverIP = localIP;
        serverPort = ntohs(loc_addr.inet.port);
    }
#endif /* HPUX */

    return PR_TRUE;
}

#endif /* DAEMON_NEEDS_SEMAPHORE */

void 
DaemonSession::ThreadMain()
{
    fNeedFlush = PR_FALSE;

    daemonProcessor->IncrementSessionsInProgress();

    PR_SetThreadPrivate(getThreadMallocKey(), NULL);
    dstat_get_thread_slot();
    child_status(CHILD_AWAIT_CONNECT);

    while( daemonProcessor->FTerminating() == PR_FALSE) {

        daemonProcessor->IncrementActiveCount(socketIndex);

        while(1) {

            if (!daemonProcessor->FThrottleOpen(socketIndex)) {

                /* FSafeThrottleOpen decrements Active Count atomically if
                 * throttle is closed.
                 */
                if (!daemonProcessor->FSafeThrottleOpen(socketIndex)) {
                    break;
                }
            }

            if (GetConnection() == PR_FALSE) {
                /* Check to see if this thread is needed on this socketIndex
                 * anymore.
                 */
                if (daemonProcessor->TestDecWaitingCount(socketIndex)) {
                    break;
                } 
                
                if( daemonProcessor->FTerminating() == PR_TRUE) {
                    break;
                }

                continue;
            }

            daemonProcessor->DecWaitAndIncBusyCount(socketIndex);
        
#if defined(PUMPKIN_HOUR) && !defined(MCC_ADMSERV)
            if(ft_time() > (PUMPKIN_HOUR + 10)) {
                ereport(LOG_FAILURE, XP_GetAdminStr(DBT_thisBetaSoftwareHasExpiredN_));
                return;
            }
#endif

            do {
                PRBool fSuccess;

                NS_ASSERT(socketAccept > 0);

                rqCurrent.InitializeRequest();
                fSuccess = Respond();

                if (fSuccess == PR_TRUE) {
                    // Before sleeping on a keepalive read, make sure threads
                    // are available for accept()
                    daemonProcessor->PostMinThreads(socketIndex);
                    fKeepAliveRequested = HandleKeepAlive();
                }
                else
                    fKeepAliveRequested = PR_FALSE;

            } while (fKeepAliveRequested == PR_TRUE);

            fNeedFlush = PR_TRUE;

            daemonProcessor->IncWaitAndDecBusyCount(socketIndex);

            daemonProcessor->PostMinThreads(socketIndex);

        }

        if (daemonProcessor->FTerminating() == PR_FALSE) {
            PR_Lock(workLock);
            workState = WORKSTATE_IDLE;
            MoveToSessionFreeList(this); 
            while(workState == WORKSTATE_IDLE) {
                PR_WaitCondVar(workCvar, PR_INTERVAL_NO_TIMEOUT);
            }
            workState = WORKSTATE_RUNNING;
            PR_Unlock(workLock);

            // "this" must now be valid again 
            NS_ASSERT (iMagic == iMagicAlloc);
        }
    }

    thread = NULL;
    daemonProcessor->DecrementSessionsInProgress();
}

PRBool DaemonSession::Respond ()
{
    PRBool fSocketOpen;

    fKeepAliveRequested = PR_FALSE;

    if (cbData)
        fSocketOpen = rqCurrent.Respond(pbData, cbData);
    else {
        PR_Close(socketAccept);
        return PR_FALSE;
    }

    if (fSocketOpen) {
        if (fKeepAliveRequested == PR_FALSE) {
            PR_Close(socketAccept);
        }
        return fKeepAliveRequested;
    }
    return PR_FALSE;
}

//
// Manage free list of DaemonSessions as necessary
//
void*    DaemonSession::operator new (size_t s)
{
    DaemonSession *pSessionAlloc = NULL;

    NS_ASSERT (s >= sizeof(DaemonSession));

    // pull next session-node from free list, or allocate a new one
    pSessionAlloc = MoveToSessionBusyList();

    if (pSessionAlloc) {
        NS_ASSERT (pSessionAlloc->iMagic == iMagicFree);
        NS_ASSERT (pSessionAlloc->pbData);
    }
    else {
        // allocate more request structures when necessary
        pSessionAlloc = (DaemonSession *) new char [s];

        if (pSessionAlloc) {

            memset (pSessionAlloc, 0, sizeof(DaemonSession));
    
            pSessionAlloc->workLock = PR_NewLock();
            pSessionAlloc->workCvar = PR_NewCondVar(pSessionAlloc->workLock);
            pSessionAlloc->workState = WORKSTATE_IDLE;

            // allocate associated i/o buffer for request

            // note, NT's AcceptEx function requires that we over 
            // allocate here, to allow for raw addresses in buffer
            // Note:  AcceptEx has a bug where it will overflow our buffers 
            // by 64 bytes; silently overallocate to avoid the problem.
            if (pSessionAlloc->pbData = new char [nbDataRaw + (2 * sInetAddrPadded) + 64])  {
                pSessionAlloc->nbData = sBuffer;
            }else {
                NS_ASSERT (0);
                delete[] (char*) pSessionAlloc;
                pSessionAlloc->nbData = 0;
            }
        }
        else
            NS_ASSERT (0);
    }
    if (pSessionAlloc) {
        pSessionAlloc->iMagic = iMagicAlloc;
    }


    return pSessionAlloc;
}

void DaemonSession::operator delete (void *p)
{
    DaemonSession *ptr = (DaemonSession *)p;
    MoveToSessionFreeList(ptr); 
}

PRBool DaemonSession::HandleKeepAlive ()
{
    KeepAliveList::keepAliveItem elt;
    elt.socket = socketAccept;

    // Adding ourselves to the keepalive list here will let us get closed by
    // another thread if there are too many keepalive connections outstanding.
    // The linked list element was allocated on our stack, so we must cleanup
    // before exiting this routine.

    KAList->AddKeepAliveItem(&elt);

    daemonProcessor->DecBusyCount(socketIndex);

    // Post the keepalive read
    cbData = PR_Recv (socketAccept, 
                      pbData, 
                      nbData, 
                      0, 
                      keepAliveTimeoutInterval);

    daemonProcessor->IncBusyCount(socketIndex);

    KAList->DelKeepAliveItem(&elt, cbData>0?PR_TRUE:PR_FALSE);
   
    if (elt.linked == -1)
        /* Socket was forcibly closed by another thread. */
        return PR_FALSE;

    if (cbData > 0) 
        return PR_TRUE;
    else {
#if 0
        /* There are a lot of different errors that we get here since we
         * closed the socket.  Everything from EAGAIN to EPIPE.
         * for now, I'm disabling the error reporting, since it isn't really
         * much of an error anyway  XXXMB
         */
        int err = PR_GetError();
        if (err != PR_IO_TIMEOUT_ERROR && err != PR_CONNECT_RESET_ERROR) {
            log_ereport(LOG_FAILURE, "Error receiving keepalive (%s)",
                system_errmsg());
        }
#endif
        PR_Close(socketAccept);
        return PR_FALSE;
    }
}


// Free/Busy list functions
void DaemonSession::AddToSessionFreeList(DaemonSession *pSession)
{
    NS_ASSERT(pSession->iMagic == iMagicAlloc);

    PR_Lock(sessionListLock);
    pSession->pNextFree = pSessionFreeHead;
    pSessionFreeHead = pSession;
    pSession->iMagic = iMagicFree;

    cSessionFreeListSize++;

    PR_Unlock(sessionListLock);
}

DaemonSession *DaemonSession::MoveToSessionBusyList()
{
    DaemonSession *pSession;
    
    PR_Lock(sessionListLock);

    // Remove a session from free list
    pSession = pSessionFreeHead;

    if (!pSession) {
        PR_Unlock(sessionListLock);
        return NULL;
    }
    pSessionFreeHead = pSessionFreeHead->pNextFree;
    pSession->pNextFree = 0;

    cSessionFreeListSize--;

    // Add that element to busy list
    pSession->pNextBusy = pSessionBusyHead;
    pSession->pPrevBusy = NULL;
    if (pSessionBusyHead)
        pSessionBusyHead->pPrevBusy = pSession;
    pSessionBusyHead = pSession;

    cSessionBusyListSize++;

    PR_Unlock(sessionListLock);

    return pSession;
}

void DaemonSession::MoveToSessionFreeList(DaemonSession *pSession)
{
    NS_ASSERT(pSession->iMagic == iMagicAlloc);

    PR_Lock(sessionListLock);
    
    // Remove the session from busy list
    if (pSession->pNextBusy)
        pSession->pNextBusy->pPrevBusy = pSession->pPrevBusy;
    if (pSession->pPrevBusy)
        pSession->pPrevBusy->pNextBusy = pSession->pNextBusy;
    if (pSessionBusyHead == pSession)
        pSessionBusyHead = pSession->pNextBusy;
    if (pSessionBusyTail == pSession)
        pSessionBusyTail = pSession->pPrevBusy;

    pSession->pNextBusy = 0;
    pSession->pPrevBusy = 0;

    cSessionBusyListSize--;

    // Add to free list
    pSession->iMagic = iMagicFree;
    pSession->pNextFree = pSessionFreeHead;
    pSessionFreeHead = pSession;

    cSessionFreeListSize++;
   
    PR_Unlock(sessionListLock);
}
void DaemonSession::KillOpenSessionSockets()
{
    DaemonSession *pSession;
    int rv;

    PR_Lock(sessionListLock);
    for (pSession = pSessionBusyHead; pSession; pSession=pSession->pNextBusy)
        // Don't use PR_Close here, or we'll delete the structure.
#ifndef XP_WIN32
        close(PR_FileDesc2NativeHandle(pSession->socketAccept));
        
#else
        rv = closesocket(PR_FileDesc2NativeHandle(pSession->socketAccept));
        if (rv != 0)
            log_ereport(LOG_VERBOSE, "Error closing sessions socket on shutdown");
#endif
    PR_Unlock(sessionListLock);
}




KeepAliveList::KeepAliveList(PRInt32 _cMaxKeepAlives)
{
    pKeepAliveHead = pKeepAliveTail = 0;
    cMaxKeepAliveItems = _cMaxKeepAlives;
    cNumKeepAliveItems = 0;
    cNumKeepAliveHits = 0;
    cNumKeepAliveFlushes = 0;
    lLock = PR_NewLock();
}

KeepAliveList::~KeepAliveList()
{
    PR_DestroyLock(lLock);
}

void
KeepAliveList::AddKeepAliveItem(keepAliveItem *elt)
{
    PRFileDesc *closeSocket = NULL;

    /* Add to tail */
    PR_Lock(lLock);
    elt->next = NULL;
    elt->prev = pKeepAliveTail;
    if (pKeepAliveTail)
        pKeepAliveTail->next = elt;
    else if (pKeepAliveHead == NULL)
        pKeepAliveHead = elt;
    pKeepAliveTail = elt;
    elt->linked = 1;
    cNumKeepAliveItems++;
   
    if (cNumKeepAliveItems >= cMaxKeepAliveItems) {
        keepAliveItem *pDead = pKeepAliveHead;

        closeSocket = pDead->socket;
        RemoveKeepAliveItem(pDead, PR_FALSE);
        cNumKeepAliveFlushes++; 
        pDead->linked = -1;
    }
    PR_Unlock(lLock);

    if (closeSocket)
        PR_Close(closeSocket);
}

void
KeepAliveList::RemoveKeepAliveItem(keepAliveItem *elt, PRBool gotKeepAlive)
{
    if (elt->linked == 1) {
        elt->linked = 0;
        if (pKeepAliveHead == elt)
            pKeepAliveHead = elt->next;
        else 
            elt->prev->next = elt->next;
        if (pKeepAliveTail == elt)
            pKeepAliveTail = elt->prev;
        else 
            elt->next->prev = elt->prev;
        cNumKeepAliveItems--;
    }
   
    if (gotKeepAlive)
        cNumKeepAliveHits++;
}

void
KeepAliveList::DelKeepAliveItem(keepAliveItem *elt, PRBool gotKeepAlive)
{
    PR_Lock(lLock);
    RemoveKeepAliveItem(elt, gotKeepAlive);
    PR_Unlock(lLock);
}

PRBool
KeepAliveList::GetKeepAliveInfo(keepAliveInfo *info)
{
    if (!info)
        return PR_FALSE;

    info->numKeepAlives = cNumKeepAliveItems;
    info->maxKeepAlives = cMaxKeepAliveItems;
  
    info->numKeepAliveHits = cNumKeepAliveHits;
    info->numKeepAliveFlushes = cNumKeepAliveFlushes;

    return PR_TRUE;
}
