/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
** DaemonSession.h
**
**  DaemonSession maintains sessions that are active on our listen socket.
**
**  DaemonSession will accept new connections and reading the initial data 
**  block on a connection, creating an individual request for each logical 
**  connection in turn.  DaemonSession will also keep this connection alive 
**  across multiple requests, when appropriate.
**
**  Because of the functioning of the Windows NT function AcceptEx, a 
**  DaemonSession may survive across a given client's connect() and the 
**  associated keepalives.  (AcceptEx is a Windows NT function that will 
**  accept() a new connection and read the initial data block in one 
**  asynchronous system call.)  AcceptEx reuses the same socket across 
**  connections, so our AcceptSession follows that lifetime too.
**
**  In other words, initially an DaemonSession will be Accept'ing a new 
**  connection.  After processing the new connection, it may do a Keepalive 
**  read.  Eventually the individual connection will be logically closed - 
**  either because keep-alive wasn't requested, the connection couldn't be 
**  kept alive, or the keep-alive timed out.
**
**  Then, the DaemonSession will go back to Accept'ing a new connection.
*/

#ifndef _DAEMONSESSION_H_ 
#define _DAEMONSESSION_H_ 

#include "libdaemon.h"
#include "daemonprocessor.h"
#include "prthread.h"
#include "frame/accel_file_cache.h"
#include "httpdaemon/multihome.h"
#include "httprequest.h"
#include "httpdaemon/internalstats.h"

#define SESSION_DEAD_KEY		0x87654321

class KeepAliveList;
class DaemonSession;

class DaemonSession {

 public:
    // Module initialization
    static PRBool Initialize ();

    DaemonSession (PRInt32 index,
                   DaemonProcessor *daemonProcessor);
    ~DaemonSession ();

    // Can't actually start the thread in the constructor because the
    // class is not fully initialized yet.  Call StartThread() after
    // Constructing the class.
    PRBool StartThread();

    // Function to get initial request data from client
    PRBool GetConnection();

    // HandleCachedResponse()
    // Defined by the platform-specific subclass to send back the cached
    // response in the optimal way for that platform.
    virtual PRBool      HandleCachedResponse(PRFileDesc *, 
                                             accel_file_cache_entry_t *,
                                             PRBool, PRInt32 *) = 0;

    // HandleKeepAlive()
    // Returns PR_TRUE if the socket is still open and has received a keepalive
    // request.  Returns PR_FALSE if no keepalive request was received.
    // Side effects:
    //    If returning PR_FALSE, closes the socket before returning.
    PRBool              HandleKeepAlive();

    void                RequestKeepAlive ();

    void                ThreadMain();

    // Misc accessors
    PRFileDesc *           GetSocket ();
    char               *GetServerIP();
    int                 GetServerPort();
    struct sockaddr_in *GetLocalAddress();
    struct sockaddr_in *GetRemoteAddress();

    // Retrieve the header buffer contents
    void *              PbData ();
    PRInt32             CbData ();

    // Manage free list of DaemonSessions as necessary
    void               *operator new (size_t);
    void	            operator delete (void *);

    // Should be private; daemonsessionux uses them for now.
    PRInt32             socketIndex;
    PRFileDesc         *socketListen; 
    PRFileDesc         *socketAccept;
    struct sockaddr_in *remoteAddress;

    // Should be private; daemonsessionnt uses them for now.
    PRInt32             nbData;     // physical size of allocated buffer, actual
    void               *pbData;     // content data from initial read
    PRInt32             cbData;     // logical amount of data in initial read

    static KeepAliveList *GetKeepAliveList();
   
    // Session list management; sessions are either on the free list or
    // on the busy list.
    static void           AddToSessionFreeList(DaemonSession *);
    static DaemonSession *MoveToSessionBusyList(void);
    static void           MoveToSessionFreeList(DaemonSession *);

    // Terminate open session sockets.
    static void KillOpenSessionSockets();

 private:

    PRUint32            iMagic;     // some magic # tags to veriify requests on free list, etc
    static PRUint32     iMagicAlloc;
    static PRUint32     iMagicFree;

    static PRInt32      sBuffer;    // logical sizeof buffer for initial read
    static PRInt32      nbDataRaw;  // physical size of allocated buffer, intended

    // Handle incoming request.
    // Returns PR_TRUE if the socket is still open.
    // Returns PR_FALSE if the socket has been closed.
    PRBool              Respond ();

    char               *serverIP;
    int                 serverPort;
    DaemonProcessor    *daemonProcessor;

    PRThread           *thread;

        // requested size in order to hold additional address
        // information that precedes the content information
        // (See AcceptEx buffer argument documentation.)
    void               *vbDataRaw; 

    PRBool              fNeedFlush;   /* TRUE if this thread has log data 
                                       * buffered up
                                       */
    PRBool              fKeepAliveRequested;
    static KeepAliveList *KAList;

    // For HTTP; this is hardcoded as a HttpRequest for now.
    HttpRequest         rqCurrent;    // current request-in-progress

    typedef enum {
         WORKSTATE_IDLE,        // thread is idle; no work waiting
         WORKSTATE_RUNNING,     // thread is running; doing work
         WORKSTATE_READY        // thread is idlel work is waiting
    } workstate_e;

    workstate_e         workState;
    PRLock             *workLock;
    PRCondVar          *workCvar;

    static multiple_domains   *domainArray;

    //
    // Free/Busy list mgmt
    //
    static PRLock        *sessionListLock;
    static DaemonSession *pSessionBusyHead;
    static DaemonSession *pSessionBusyTail;
    static DaemonSession *pSessionFreeHead;
    static PRUint32       cSessionFreeListSize;
    static PRUint32       cSessionBusyListSize;

    DaemonSession      *pNextFree; 
    DaemonSession      *pNextBusy; 
    DaemonSession      *pPrevBusy;

#ifdef DAEMON_NEEDS_SEMAPHORE
    static PRLock       *acceptLock;
#endif

    static PRUint32      sessionStackSize;

#ifdef HPUX
    char localIP[24];
#endif /* HPUX */
};

/* We don't want an infinite number of keepalive threads open simultaneously.
 * So at most we'll keep numKeepAlives threads waiting on a keepalive.  Once
 * that number is reached, we'll just overflow, killing the last keptalive
 * socket first.  This is just a circular list.
 */
class KeepAliveList
{
public:
    typedef struct keepAliveItem {
        PRFileDesc *socket;
        PRInt32     linked;  /* 1 if this elt is linked into the keepalive list,
                              * 0 if it has been unlinked from the keepalive list.
                              * -1 if it has been unlinked and the socket closed.
                              */
        struct keepAliveItem *prev, *next;
    } keepAliveItem;

    KeepAliveList(PRInt32 cMaxKeepAlives);
    ~KeepAliveList();

    void AddKeepAliveItem(keepAliveItem *);
    void DelKeepAliveItem(keepAliveItem *, PRBool);

    // Get statistics about the keepalives
    PRBool              GetKeepAliveInfo(keepAliveInfo *info);
private:
    void RemoveKeepAliveItem(keepAliveItem *, PRBool);

    keepAliveItem *pKeepAliveHead;
    keepAliveItem *pKeepAliveTail;
    PRUint32       cMaxKeepAliveItems;
    PRUint32       cNumKeepAliveItems;
    PRUint32       cNumKeepAliveHits;
    PRUint32       cNumKeepAliveFlushes;
    PRLock        *lLock;
};

inline void *DaemonSession::PbData()
{
    return pbData;
}

inline PRInt32 DaemonSession::CbData()
{
    return cbData;
}

inline PRFileDesc * DaemonSession::GetSocket()
{
    return socketAccept;
}

inline void DaemonSession::RequestKeepAlive()
{
    fKeepAliveRequested = PR_TRUE;
}

inline char *DaemonSession::GetServerIP()
{
    return serverIP;
}

inline int DaemonSession::GetServerPort()
{
    return serverPort;
}

inline struct sockaddr_in *DaemonSession::GetLocalAddress()
{
    return &(domainArray[socketIndex].addr);
}

inline struct sockaddr_in *DaemonSession::GetRemoteAddress()
{
    return remoteAddress;
}

inline KeepAliveList *DaemonSession::GetKeepAliveList()
{
    return DaemonSession::KAList;
}

extern "C" void CThreadMain(void *vDaemonSession);

#endif // _DaemonSession_h_
