/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
#include "daemonsessionux.h"
#include "daemonprocessorux.h"
#include "prlock.h"
#include "frame/accel_file_cache.h"
#include "httpdaemon/dbthttpdaemon.h"
#ifdef AIX
#include <time.h>
#endif

extern DaemonProcessor *globalDaemonProcessor;

DaemonSessionUX::DaemonSessionUX(
                   PRInt32 _index, 
                   DaemonProcessor *_pDaemon) :
    DaemonSession(_index, _pDaemon)
{
}

DaemonSessionUX::~DaemonSessionUX()
{
}

PRBool DaemonSessionUX::HandleCachedResponse(PRFileDesc *socket,
         accel_file_cache_entry_t *entry, PRBool _fKeepAliveRequested, PRInt32 *bytesSent)
{
    PRIOVec iov[2];
    int iovlen;
    char *header_buffer;
    int header_len;
    PRInt32 rv;

    if (_fKeepAliveRequested) {
        header_buffer = entry->accelcache->headers_keepalive;
        header_len = entry->accelcache->header_keepalive_len;
    } else {
        header_buffer = entry->accelcache->headers;
        header_len = entry->accelcache->header_len;
    }

    /* If we have everything we need in the cache, quickly writev() back */
    if (entry->filecache && entry->filecache->mmap) {
        iov[0].iov_base = header_buffer;
        iov[0].iov_len  = header_len;
        iov[1].iov_base = entry->filecache->mmap;
        iov[1].iov_len  = entry->filecache->mmap_len;
        iovlen = 2;

        if ( (rv = PR_Writev(socket, iov, iovlen, PR_INTERVAL_NO_TIMEOUT)) < 0) {
            ereport(LOG_FAILURE, "Error sending to client (%s)", system_errmsg());
            *bytesSent = -1;
            goto error;
        }

        *bytesSent = rv;
    } else {
        struct stat diskInfo;
        PRFileDesc *diskFile;
        PRInt32 bytesRead, bytesWritten;
        PRUint32 amount = -1;
        #define DATABUF_SIZE (16*1024)
        char dataBuf[DATABUF_SIZE];
        PRStatus status;

        /* If we know the length of the file, use it to avoid an extra
         * read() later.
         */
        if (entry->filecache)
            amount = entry->filecache->mmap_len;

        diskFile = PR_Open(entry->accelcache->path, PR_RDONLY, 0);
        if (diskFile == NULL) {
            ereport(LOG_FAILURE, 
                "HandleResponse(): Error opening file %s (%s)",
                entry->accelcache->path, system_errmsg());
            *bytesSent = -1;
            goto error;
        }

        /* Read the first chunk of data from the disk file and then writev
         * that back with the headers.
         */
        bytesRead = PR_Read(diskFile, dataBuf, DATABUF_SIZE);
        if (bytesRead < 0) {
            ereport(LOG_FAILURE, 
                "HandleResponse(): Error reading file %s (%s)",
                entry->accelcache->path, system_errmsg());
            PR_Close(diskFile);
            *bytesSent = -1;
            goto error;
        }

        iov[0].iov_base = header_buffer;
        iov[0].iov_len  = header_len;
        iov[1].iov_base = dataBuf;
        iov[1].iov_len  = bytesRead;
        iovlen = 2;

        if ((rv = PR_Writev(socket, iov, iovlen, PR_INTERVAL_NO_TIMEOUT)) < 0) {
            ereport(LOG_FAILURE, "Error sending to client (%s)", system_errmsg());
            *bytesSent = -1;
            goto error;
        }

        *bytesSent = rv;

        if (amount > 0)
            amount -= rv;


        /* Now, continue with the remaining chunks of data */

        while( amount != 0 ) { 
            bytesRead = PR_Read(diskFile, dataBuf, DATABUF_SIZE); 
            if (bytesRead < 0) { 
                ereport(LOG_FAILURE, "HandleResponse(): Error reading file %s (%s)",
                    entry->accelcache->path, system_errmsg());
                PR_Close(diskFile);
                goto error;
            } else if (bytesRead == 0)
                /* We got an EOF */
                break;

            if (bytesRead > 0) {
                bytesWritten = PR_Send(socket, dataBuf, bytesRead, 0, 
                                       PR_INTERVAL_NO_TIMEOUT);
                if (bytesWritten < 0) {
                    ereport(LOG_FAILURE, 
                        "HandleResponse(): Error writing to socket (%s)",
                        system_errmsg());
                    PR_Close(diskFile);
                    goto error;
                }

                if (amount != -1)
                    amount -= bytesWritten;

                *bytesSent += bytesWritten;
            }
        } 

        PR_Close(diskFile);
    }

error:
    PR_Close(socket);
    return PR_FALSE;
}
