/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
** HttpHeader.cpp
*/

#include "libdaemon.h"
#include "httpheader.h"
#include "base/nsassert.h"
#include "base/systems.h"
#include "base/util.h"

#define mytolower(x) (x>='A'&&x<='Z'?x|0x20:x)

PRBool HttpHeader::ScanHeader (void *pHeader, size_t sHeader)
{
    pzRequest = (char *)pHeader;
    pzUri = NULL;
    pzProtocol = NULL;
    pzReferer = NULL;
    pzUserAgent = NULL;
    pzIfModifiedSince = NULL;
    pzHost = NULL;
    pzAuthenticate = NULL;
    iRequestLen = 0;

    fRangeRequested = PR_FALSE;

    fKeepAliveRequested = PR_FALSE;
    
    fPipeLined  = PR_FALSE;
    sVersionNumber = 0;

    if (sHeader >= sHttpHeaderMax)
        return PR_FALSE;
    else {
        memcpy (vzHeader, pHeader, sHeader);
        vzHeader [sHeader] = '\0';  // place sentinel to terminate header, for easier parsing
                                    // note: we allocated to sHttpHeaderMax+1 to allow room for this

        pzHeader = vzHeader; // initialize parser's cursor

        return (ScanHeadline () && ScanAttributes ())?PR_TRUE:PR_FALSE;
    }
}

PRBool HttpHeader::ScanHeadline ()
{
    const char *pzMethod = "GET";     // header must be start with GET then whitespace
    const PRInt32 szMethod = strlen (pzMethod);

    const char *pzExpectedProtocol = "HTTP/1"; // [sic] protocol must be >= 1.0 for accel
    const PRInt32 nzExpectedProtocol = strlen (pzExpectedProtocol);
    char *pzUriTail;
    int   sMajor=0, sMinor=0;

    // Verify "GET" method for accel
    if (strneq (pzHeader, pzMethod, szMethod)) {

        pzHeader += szMethod;

        if (isspace (*pzHeader)) {
            while (isspace (*pzHeader))
                ++pzHeader;

            // Delimit URI
            pzUri = pzHeader;
            while (*pzHeader && ! isspace (*++pzHeader))
                ;
            if (* pzHeader) {
                pzUriTail = pzHeader;        // mark end of URI
                while (isspace (*++pzHeader))
                    ;

                // Verify protocol in allowable range (ie, >= 1.0)
                pzProtocol = pzHeader;
                // Extract Protocol Version Number
                sscanf(pzProtocol,"HTTP/%d.%d",&sMajor,&sMinor);
                sVersionNumber = 100 * sMajor + sMinor;
                if (sVersionNumber >= 100) {
                    while (*pzHeader && *++pzHeader != '\n')
                        ;
                    if (*pzHeader == '\n')  {

                        if (*(pzHeader-1) == '\r') {// guaranteed to not underflow
                            *(pzHeader-1) = '\0';
                            iRequestLen = (PRUint32)(pzHeader-vzHeader-1);
                        } else {
                            *pzHeader = '\0';
                            iRequestLen = (PRUint32)(pzHeader-vzHeader);
                        }

                        ++pzHeader;
                        *pzUriTail = '\0';        // nul at end of URI
                        if ( sVersionNumber >= 101 ) // HTTP/1.1 or greater
                            fKeepAliveRequested = PR_TRUE;
                        return PR_TRUE;          // parsed headline AOK, URI is delimited in place.
                    }
                    else
                        return PR_FALSE;
                }
            }
        }
    }

    return PR_FALSE;
}

PRBool HttpHeader::ScanAttributes ()
{
    char *pNext = pzHeader;
    char *pzTag = NULL;
    char *pzValue = NULL;
    PRBool fInTag  = PR_FALSE;
    PRBool fInValue = PR_FALSE;

    while (*pNext) {

        char *pCurrent = pNext++;

        char c = *pCurrent;

        switch (c) {

            case '\r':    if (*pNext != '\n')
                            return PR_FALSE;
                        else
                            ++pNext;

            case '\n':    if (! pzTag) {
                            // if we're dealing with HTTP/1.1, we'll need to
                            // to check if the request packet is pipelined.
                            if ( GetVersionNumber() >= 101) {
                                fPipeLined = (strlen(pNext) > 3)?PR_TRUE:PR_FALSE;
                            }
                            return PR_TRUE; // blank line terminates header
                        }
                        else if (! fInValue)
                            return PR_FALSE;  // unexpected eol before value
                        else {
                            // newline terminating <tag: value> pair; process pair
                            NS_ASSERT (pzTag && pzValue);
                            *pCurrent = '\0';
                            if (! SetAttribute (pzTag, pzValue))
                                return PR_FALSE;
                            pzTag = pzValue = NULL;
                            fInValue = PR_FALSE;
                        }
                        break;

            case ':':    if (fInTag) {
                            // first colon ends tag-token field; ready for value token
                            *pCurrent = '\0';
                            fInTag = PR_FALSE;
                        }
                        break;

            default:    if (! fInTag && ! fInValue) {
                            // Start new token (tag or value) at first non-whitespace
                            if (! pzTag) {
                                fInTag = PR_TRUE;
                                pzTag  = pCurrent;
                            }
                            else {
                                fInValue = PR_TRUE;
                                pzValue  = pCurrent;
                            }
                        } 
                        if ( !fInValue )
                            *pCurrent = mytolower(c);
                        break;

            case ' ':
            case '\t':    if (fInTag) 
                            // truncate property-name at first wsp, if necessary
                            *pCurrent = '\0';
                        break;
        }
    }

    return PR_FALSE; // unexpected end-of-header (should end with extra newline)
}

PRBool HttpHeader::SetAttribute (const char *pzTag, const char *pzValue)
{
    // Interpret (tag, value) pair for this request.
    //
    // For most items, this means we make sure it's in a limited
    // set of simple items that we can ignore.  Complex headers
    // are passed to the back-end for handling.  
    //
    // Some properties will be handled in the async daemon; eg,
    // byte range and keep alives.  In the long run, accept: might
    // need special treatment too.
    //

    switch (*pzTag) {

        case 'a':
	    if (streq(pzTag, "authorization"))
		pzAuthenticate = pzValue;
            break;

        case 'c':
            if (streq (pzTag, "connection")) {
                if (strieq (pzValue, "keep-alive"))
                    fKeepAliveRequested = PR_TRUE;
            }
            else
                ;
            break;

        case 'h':
            if (streq (pzTag, "host"))
                pzHost = pzValue;
            break;

        case 'i':
            if (streq (pzTag, "if-modified-since"))
                pzIfModifiedSince = pzValue;
            break;

        case 'p':
            // pragmas?  XXXJB
            // if (streq (pzTag, "pragma")) ... ;
            break;

        case 'r':
            if (streq (pzTag, "referer"))
                pzReferer = pzValue;
            else if (streq (pzTag, "range")) 
                fRangeRequested = PR_TRUE;
            break;

        case 'u':
            if (streq (pzTag, "user-agent"))
                pzUserAgent = pzValue;
            break;

        default:
            break;
            // XXXJB but we can still return from cache for a matched static GET 
            // as long as we're not directed NOT to
    }

    return PR_TRUE; // this attribute does not prevent accelerated handling
}
