/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
** HttpRequest.cpp
*/

#include <stdio.h>

#include "httprequest.h"
#include "daemonprocessor.h"
#include "daemon.h"
#include "daemonsession.h"
#include <base/plist.h>
#include <libaccess/acl.h>
#include <libaccess/aclproto.h>

#include "frame/accel_file_cache.h"
#include "frame/cmstat.h"
#include "safs/flexlog.h"
#include "safs/addlog.h"
#include "base/util.h"

#include "prio.h"
#include "private/pprio.h"

// when a file is not memory mapped, use either transmit file or iterative reads/writes:
PRBool       HttpRequest::fTransmitFile   = PR_TRUE; 

// # of synchronous requests pending (queued or being handled by "backend" thread)

HttpRequest::HttpRequest (DaemonProcessor *_processor, DaemonSession *_pSession) 
: DaemonRequest (_processor, _pSession),
    pCacheEntry           (NULL),
    iStatus               (0),
    pzStatus              (NULL),
    fAcceleratedResponse  (PR_FALSE),
    fCleanup              (PR_FALSE),
    pSession              (_pSession)
{
}

HttpRequest::~HttpRequest ()
{
}

void
HttpRequest::InitializeRequest()
{
    pCacheEntry = NULL;
    iStatus = 0;
    pzStatus = NULL;
    fAcceleratedResponse = PR_FALSE;
    fCleanup = PR_FALSE;
    socket = pSession->GetSocket();
}

void
HttpRequest::Cleanup(const char *user, const char *authorization)
{
    if (fCleanup == PR_FALSE) {
        fCleanup = PR_TRUE;

        if (pCacheEntry) {  
            if (fAcceleratedResponse)
                Log (user, authorization);
            accel_file_cache_use_decrement (pCacheEntry);
            pCacheEntry = 0;
        }
    }
}

PRBool HttpRequest::Respond (void *pRequest, PRUint32 sRequest)
{
    PRBool rv;
    const char *pzIfModifiedSince;
    const char *pzAuthenticate = NULL;
    char *user = NULL;
    ACLEvalHandle_t *acleval;
    PList_t subject, resource;
    NSErr_t     err;
    NSErr_t     *errp = &err;
    char *rights[2] = {
        "http_get",
        NULL,
    };
    char *bong, *bong_type, *acl_tag;
    int expr_num;
 
    if (rqHdr.ScanHeader (pRequest, sRequest) != PR_TRUE)
        goto unaccel;
 
    pCacheEntry = accel_file_cache_lookup_uri ((char*) rqHdr.GetUri(), (char *)rqHdr.GetHost(), pSession->GetLocalAddress());
 
    //  Check for cache entry and a valid ACL set - which could be empty but
    //  not NULL.
    if (!pCacheEntry || !pCacheEntry->acllist)
        goto unaccel;
    NS_ASSERT(ACL_AssertAcllist(pCacheEntry->acllist));

    pzIfModifiedSince = rqHdr.GetIfModifiedSince();
 
    //  Evaluate the ACLs if there are any.
    if (pCacheEntry->acllist != ACL_LIST_NO_ACLS) {
	pool_handle_t *pool = 0;
        int rv;
 
        memset((void *)errp, 0, sizeof(NSErr_t));
        acleval = ACL_EvalNew(errp, NULL);
        rv = ACL_EvalSetACL(errp, acleval, pCacheEntry->acllist);
	pool = pool_create();
        subject = PListCreate(pool, ACL_ATTR_INDEX_MAX, NULL, NULL);
        PListInitProp(subject, ACL_ATTR_IP_INDEX, ACL_ATTR_IP, (const void *)(ntohl(((struct sockaddr_in *)pSession->GetRemoteAddress())->sin_addr.s_addr)), 0);
        pzAuthenticate = rqHdr.GetAuthenticate();
	/* Always set ACL_ATTR_ACCEL_AUTH since that determines whether this is
	 * in accelerated mode or not for functions in lib/safs/auth.cpp */
	PListInitProp(subject, ACL_ATTR_ACCEL_AUTH_INDEX, ACL_ATTR_ACCEL_AUTH,
		      (const void *)pzAuthenticate, 0);
        rv = ACL_EvalSetSubject(errp, acleval, subject);
        resource = PListCreate(pool, ACL_ATTR_INDEX_MAX, NULL, NULL);
        rv = ACL_EvalSetResource(errp, acleval, resource);
 
        rv = ACL_EvalTestRights(errp, acleval, rights, http_generic, &bong,
                                &bong_type, &acl_tag, &expr_num);

	/* Check if an authenticated user is available for logging */
	PListGetValue(subject, ACL_ATTR_USER_INDEX, (void **)&user, NULL);

        PListDestroy(resource);
        PListDestroy(subject);
	pool_destroy(pool);
	// Since we started with the acllist (without calling ACL_CacheCheck,
	// no need to decrement its ref_count implicitly in EvalDestroy.
        ACL_EvalDestroyNoDecrement(NULL, NULL, acleval);
        nserrDispose(errp);
 
        if (rv != ACL_RES_ALLOW)
            goto unaccel;
 
    }


    // Resume the pre-ACL code.

    if (pzIfModifiedSince && pCacheEntry->accelcache->last_modified)
        if ( !util_str_time_equal (pCacheEntry->accelcache->last_modified, (char *)pzIfModifiedSince)) {
            fAcceleratedResponse = PR_TRUE;
            rv =  RespondNotModified ();
            goto cleanup;
        }

    if (! rqHdr.FRangeRequested()) {
        fAcceleratedResponse = PR_TRUE;

        if (rqHdr.FKeepAlive ())
            pSession->RequestKeepAlive ();

        /* Service the request; if the user specified his own
         * handler function, use that; otherwise use default 
         * handler.
         */
        if ( pCacheEntry->accelcache->user_service_func ) {
            int ret;
            ret  = pCacheEntry->accelcache->user_service_func(
                             pCacheEntry,
                             rqHdr.GetHeader(),
                             rqHdr.GetHeaderLen(),
                             pCacheEntry->accelcache->user_data);
            if (ret < 0)
                PR_Close(socket);
            rv = ret>=0?PR_TRUE:PR_FALSE;
        }
        else
            rv = pSession->HandleCachedResponse(socket, 
                             pCacheEntry, rqHdr.FKeepAlive(), &bytesSent);

        /* Call user specified notification routine, if specified */
        if (pCacheEntry->accelcache->user_notification_func ) 
            pCacheEntry->accelcache->user_notification_func(
                             pCacheEntry,
                             rqHdr.GetHeader(),
                             rqHdr.GetHeaderLen(),
                             pCacheEntry->accelcache->user_data);
            
        goto cleanup;
    }
    else
        ; // let back end do ranges - the headers are dynamic, and it requires writev 
          // to accelerate it well in general.  note we do verify that we can't optimize 
          // it with if-modified-since before passing it off

unaccel:
#if defined(XP_WIN32)  && !defined(WIN95)
    // With the FastAccept socket on NT, we can't use the socket for general
    // purpose socket operations yet (like shutdown, setsockopt, etc).  Before
    // passing to unacceleratedrespond() (which might call shutdown), update
    // the socket to make it a general purpose socket.
    PR_NTFast_UpdateAcceptContext(pSession->socketAccept, pSession->socketListen);
#endif

    rv = UnacceleratedRespond();

cleanup:
    Cleanup(user, pzAuthenticate);
    return rv;
}

#define HEADER_304_L1 "HTTP/1.0 304 Use local copy\r\nServer: "
#define HEADER_304_L2 "\r\n\r\n"
#define IMS_BUFF_LEN    2048
PRBool HttpRequest::RespondNotModified ()
{
    // Send back a 304 Header

    char vzBuffer[IMS_BUFF_LEN];
	char linkHdr[IMS_BUFF_LEN];  
    iStatus = 304;
    PRInt32 len;
    PRInt32 rv;
	char *uri = (char *)rqHdr.GetUri();
	int  ret;
	
	/* gets "Link" header, if content manager is on. Null otherwise */	
	ret = CM_get_link_header(uri, linkHdr, IMS_BUFF_LEN );
	if ( ret ) 
	  util_snprintf(vzBuffer, IMS_BUFF_LEN, "%s%s \r\n%s%s",
					HEADER_304_L1, MAGNUS_VERSION_STRING, linkHdr, HEADER_304_L2);
	else
	  util_snprintf(vzBuffer, IMS_BUFF_LEN, "%s%s%s",
					HEADER_304_L1, MAGNUS_VERSION_STRING, HEADER_304_L2);

    len = strlen(vzBuffer);

    pzStatus = "304 Use local copy";

    // XXXMB - do we want to send the date too?

    if ((rv = PR_Send(socket, (void*) vzBuffer, len, 0, PR_INTERVAL_NO_TIMEOUT)) != len) {
        ereport(LOG_FAILURE, "RespondNotModified(): Error writing to socket (%s)",
            system_errmsg());
        PR_Close(socket);
        return PR_FALSE;
    }
    NS_ASSERT(rv == len);

    if (rqHdr.FKeepAlive ())
        pSession->RequestKeepAlive ();

    return PR_TRUE;
}

//
// Work detail for background threads
//

PRBool HttpRequest::UnacceleratedRespond()
{
    PRBool fKeepAliveRequested = PR_FALSE;

    // dispatch request to "normal" server mainline
    HandleSynchronousRequestSession (
        pSession->GetServerIP (), 
        pSession->GetLocalAddress(),
        (struct sockaddr *)pSession->GetRemoteAddress(),
        socket,
        pSession->PbData (), 
        pSession->CbData (), 
        &fKeepAliveRequested
    );

    if (fKeepAliveRequested)
        pSession->RequestKeepAlive ();

    /* Socket already closed if keepalive is off */
    return fKeepAliveRequested;
}


PRBool HttpRequest::Initialize ()
{
    return PR_TRUE;
}

struct _fourbytes {
    unsigned char b1;
    unsigned char b2;
    unsigned char b3;
    unsigned char b4;
};

void
_inet_ntoa(struct in_addr ip, char *strIp)
{
    struct _fourbytes *fbip;

    fbip = (struct _fourbytes *)&ip;

    sprintf(strIp, "%d.%d.%d.%d", fbip->b1, fbip->b2, fbip->b3, fbip->b4);
}

//
// Logging interface for accelerator
//
void HttpRequest::Log (const char *user, const char *authorization)
{
    accel_cache_log_t *pLog;
    PRInt32 iResult = REQ_PROCEED;
    char *pzHost = NULL;
    struct sockaddr_in *sin;
    char pzIp[24];
    char pzBytesSent[24];
    PRUint32 cbSent = 0;

    NS_ASSERT(pSession);
    sin = pSession->GetRemoteAddress();

    NS_ASSERT(pCacheEntry);
    for (
        pLog = pCacheEntry->accelcache->log_os; 
        pLog && iResult == REQ_PROCEED; 
        pLog = pLog->next) {

        /* XXXMB - we are converting this IP on every request basically! */
        _inet_ntoa(sin->sin_addr, pzIp);
        if ( !pLog->iponly ) {
            pzHost = net_ip2host(pzIp, 0);
        }

        if (!pzHost)
            pzHost = pzIp;

        if (pCacheEntry->filecache == NULL) {
            cbSent = bytesSent;

            if (cbSent > 0)
                util_itoa(cbSent, pzBytesSent);
            else 
                util_itoa(0, pzBytesSent);
                
        }
    
        if (pLog->log_fn == flex_log) {
            iResult = flex_log_accel( 
                pLog->logname, 
                pzStatus?(char *)pzStatus: ((char*)"200 OK"), 
                cbSent?(char *)pzBytesSent:pCacheEntry->filecache?pCacheEntry->filecache->strbytes:0, 
                pzHost, 
                rqHdr.GetRequest(), 
                "GET", 
                (char *)rqHdr.GetProtocol(), 
                NULL, 
                (char *)rqHdr.GetUri(), 
                (char *)rqHdr.GetReferer(), 
                (char *)rqHdr.GetUserAgent(), 
                (char *)authorization, 
                (char *)user
            );

        } else if (pLog->log_fn == clf_record) {
            iResult = clf_record_accel( 
                pLog->logname, 
                pzHost,
                rqHdr.GetRequest(), 
                (char *)user, 
                cbSent?(char *)pzBytesSent:pCacheEntry->filecache?pCacheEntry->filecache->strbytes:0, 
                pzStatus?(char *)pzStatus:((char*)"200 OK")
            );
        }
	if (pzHost && pzHost != pzIp) FREE(pzHost);
	pzHost = NULL;
    }

#ifdef DAEMON_STATS
    child_status2(iStatus?iStatus:200, cbSent?cbSent:pCacheEntry->filecache?pCacheEntry->filecache->mmap_len:0);
#endif
}


