/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
** multihome.cpp
**
** Multihoming module
**
** Warning- not thread safe.  Two threads calling Add() can be bad!
**
**/

#include "httpdaemon/multihome.h"
#include "frame/conf.h"
#include "frame/conf_api.h"
#ifdef XP_UNIX
#include <arpa/inet.h>  /* inet_ntoa */
#include <netdb.h>      /* hostent stuff */
#endif /* XP_UNIX */


#define INITIAL_DOMAIN_COUNT 16
static multiple_domains *domain_array;
static int domain_count;
static int domain_size;

/* FUNCTION: MultihomeInit
** DESCRIPTION:
**     Internal init routine for multihoming. Scans config
**     file objects for domains to add.
** RETURNS
**     PR_TRUE if initializes successfully, PR_FALSE otherwise
*/
PRBool MultihomeInit()
{
    char *name = root_object;
    httpd_object *obj;
	dtable *d;
	int index;

    domain_size = INITIAL_DOMAIN_COUNT;
    domain_count = 0;
    domain_array = (multiple_domains *)PERM_MALLOC(sizeof(multiple_domains) * INITIAL_DOMAIN_COUNT);
    if (!domain_array)
        return PR_FALSE;

    if(!(obj = objset_findbyname(name, NULL, std_os))) 
        return PR_FALSE;

	d = &obj->dt[1];
    for(index = 0; index < d->ni; index++) {
        char *funcName;

        /* Search for document-root directives and add them to the list
         * of virtual servers.
         */
        funcName = pblock_findval("fn", d->inst[index].param);
        if (funcName && !strcmp(funcName, "document-root")) {
            char *address = pblock_findval("address", d->inst[index].param);
            char *docroot;
            char *strPort, *strSecure, *strSecure_auth, *strSsl3, *strSsl2;
            PRUint32 port;
            PRBool secure, secure_auth, ssl3, ssl2;

            if (!address)
                continue;

            docroot = pblock_findval("docroot", d->inst[index].param);
            strPort = pblock_findval("port", d->inst[index].param);
            strSecure = pblock_findval("secure", d->inst[index].param);
            strSecure_auth = pblock_findval("secure_auth", d->inst[index].param);
            strSsl3 = pblock_findval("ssl3", d->inst[index].param);
            strSsl2 = pblock_findval("ssl2", d->inst[index].param);

            if (strPort)
                port = (PRUint32)atoi(strPort);
            else
                port = conf_getglobals()->Vport;
               
            if (strSecure)
                secure = atoi(strSecure)==0?PR_FALSE:PR_TRUE;
            else
                secure = conf_getglobals()->Vsecurity_active==0?PR_FALSE:PR_TRUE;
               
            if (strSecure_auth)
                secure_auth = atoi(strSecure_auth)==0?PR_FALSE:PR_TRUE;
            else
                secure_auth = conf_getglobals()->Vsecure_auth==0?PR_FALSE:PR_TRUE;
               
            if (strSsl3)
                ssl3 = atoi(strSsl3)==0?PR_FALSE:PR_TRUE;
            else
                ssl3 = conf_getglobals()->Vssl3_active==0?PR_FALSE:PR_TRUE;
               
            if (strSsl2)
                ssl2 = atoi(strSsl2)==0?PR_FALSE:PR_TRUE;
            else
                ssl2 = conf_getglobals()->Vssl2_active==0?PR_FALSE:PR_TRUE;
               
            (void)AddMultihomeDomain(address, 
                                     docroot, 
                                     port, 
                                     secure, 
                                     secure_auth, 
                                     ssl3, 
                                     ssl2);
        }
	}

    /* Add a variable to the conf-globals indicating virtual servers is "on" */
    if (GetCountMultihomeDomains() > 1)
        conf_setGlobal("VirtualServers", "on");

	return PR_TRUE;
}

multiple_domains *GetMultihomeDomains()
{
    return domain_array;
}


PRInt32 GetCountMultihomeDomains()
{
    return domain_count;
}

PRBool AddMultihomeDomain(char *address, char *docroot, PRUint32 port,
                          PRBool security, PRBool secure_auth, PRBool ssl3,
                          PRBool ssl2)
{
    int index;

    if (domain_count >= domain_size) {
        /* time to realloc */
        domain_size *= 2;
        domain_array = (multiple_domains *)PERM_REALLOC(domain_array, 
                             (sizeof(multiple_domains)) * domain_size);
        if (domain_array == 0)
            return PR_FALSE;
    }

    index = domain_count++;

    if (address)
        domain_array[index].address = PERM_STRDUP(address);
    else
        domain_array[index].address = NULL;

    if (docroot)
        domain_array[index].docroot = PERM_STRDUP(docroot);
    else
        domain_array[index].docroot = NULL;
    domain_array[index].port = port;
    domain_array[index].security = security;
    domain_array[index].secure_auth = secure_auth;
    domain_array[index].ssl3 = ssl3;
    domain_array[index].ssl2 = ssl2;

    domain_array[index].addr.sin_family = AF_INET;
    domain_array[index].addr.sin_port = htons(port);
    if (domain_array[index].address == NULL) 
        domain_array[index].addr.sin_addr.s_addr = htonl(INADDR_ANY);
    else
        domain_array[index].addr.sin_addr.s_addr = 
            inet_addr(domain_array[index].address);

    domain_array[index].sd = 0;

#ifdef DAEMON_NEEDS_SEMAPHORE
    domain_array[index].sockLock = PR_NewLock();
#else
    domain_array[index].sockLock = NULL;
#endif

    return PR_TRUE;
}

void MultihomeCleanup()
{
    int index;
   
    for (index=0; index<domain_count; index++) {
        PR_Close(domain_array[index].sd);
#ifdef DAEMON_NEEDS_SEMAPHORE
        PR_DestroyLock(domain_array[index].sockLock);
#endif
    }

    domain_count = 0;

    PERM_FREE(domain_array);
    domain_array = 0;
}
