/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/


/*
 * Description (acladmin.c)
 *
 *	This module contains functions for inspecting, modifying, and
 *	writing out to a file in text form ACL structures.
 */

#include <stdlib.h>
#include <assert.h>
#include <base/systems.h>
extern "C" {
#include <libadminutil/admutil.h>
}
#include <libadmin/libadmin.h>
#include <libaccess/nsadb.h>
#include <libaccess/aclerror.h>
#include <libaccess/aclparse.h>
#include <libaccess/aclbuild.h>
#include <libaccess/acladmin.h>
#include <libaccess/symbols.h>
#include <base/net.h>
#include <base/util.h>
#ifdef XP_UNIX
#include <sys/types.h>
#include <netinet/in.h>  /* ntohl */
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif /* XP_UNIX */

/* defined in dnsdmain.c */
NSPR_BEGIN_EXTERN_C
NSAPI_PUBLIC char *dns_guess_domain(char * hname);
NSPR_END_EXTERN_C

NSAPI_PUBLIC int aclDNSAddHost(char * newhost, char ***alist, int * asize)
{
    char * nhp = newhost;
    char **q;

    if (!strchr(newhost, '.')) {
	nhp = dns_guess_domain(newhost);

	/* If there's no domain name, just use the unqualified host name */
	if (nhp == 0) nhp = newhost;
    }

    /* Is it already in the list? */
    if (*alist != 0) {

	for (q = *alist; *q != 0; ++q) {

	    if (!strcasecmp(*q, nhp)) {

		/* Yes, forget it */
		return 0;
	    }
	}
    }
    else {

	/* Create the list */
	*alist = (char **)MALLOC(4*sizeof(char *));
	**alist = 0;
	*asize = 0;
	q = *alist;
    }

    /* Add new host name to list */
    if ((*asize & 3) == 3) {

	/* Grow the list */
	*alist = (char **)REALLOC(*alist, (*asize + 5)*sizeof(char *));
	q = *alist + *asize;
    }
    *q++ = STRDUP(nhp);
    *q = 0;
    *asize += 1;

    /* Indicate entry added */
    return 1;
}

NSAPI_PUBLIC int aclDNSAddAliases(char * host, char ***alist, int * asize)
{
    int cnt = 0;
    int ipcnt = 0;
    int i;
    char **p;
    unsigned long * ipaddrs = 0;
    struct hostent * he;
    PRHostEnt   hent;
    char        buf[PR_NETDB_BUF_SIZE];
    PRStatus    err;

    err = PR_GetHostByName(host,
			      buf,
			      PR_NETDB_BUF_SIZE,
			      &hent);
    if (err == PR_SUCCESS) he = (hostent *)&hent;
    else he = NULL;
 
    if (he) {
	/* Make a copy of the list of IP addresses if any */
	if (he->h_addr_list && he->h_addr_list[0]) {

	    /* Count the IP addresses */
	    for (p = he->h_addr_list; *p; ++p) {
		++ipcnt;
	    }

	    /* Allocate space */
	    ipaddrs = (unsigned long *)MALLOC(ipcnt * sizeof(unsigned long));

	    /* Copy IP addresses */
	    for (i = 0; i < ipcnt; ++i) {
		ipaddrs[i] = 0;
		memcpy((void *)&ipaddrs[i], he->h_addr_list[i], 4);
	    }
	}

	/* Add each of the aliases to the list */
	if (he->h_aliases && he->h_aliases[0]) {

	    for (p = he->h_aliases; *p; ++p) {
		cnt += aclDNSAddHost(*p, alist, asize);
	    }
	}

	for (i = 0; i < ipcnt; ++i) {
	  err = PR_GetHostByAddr((PRNetAddr *)&ipaddrs[i],
				    buf,
				    PR_NETDB_BUF_SIZE,
				    &hent);
	  if (err == PR_SUCCESS) he = (hostent *)&hent;
	  else he = NULL;
	  
	  if (he) {
	    if (he->h_name) {
	      cnt += aclDNSAddHost(he->h_name, alist, asize);
	    }

	    if (he->h_aliases && he->h_aliases[0]) {
	      for (p = he->h_aliases; *p; ++p) {
		cnt += aclDNSAddHost(*p, alist, asize);
	      }
	    }
	  }
	}

	if (ipaddrs) {
	    FREE(ipaddrs);
	}
    }

    return cnt;
}

NSAPI_PUBLIC int aclDNSPutHost(char * hname, int fqdn, int aliases,
			       char ***alist, int * asize)
{
    struct hostent * he;	/* returned by gethostbyname() */
    char **p;
    int cnt = 0;
    int bsize;
    PRHostEnt   hent;
    char        buf[PR_NETDB_BUF_SIZE];
    PRStatus    err;

    bsize = *asize;

    /* Need a fully-qualified DNS name? */
    if (fqdn && !strchr(hname, '.')) {

      /* Yes, look for other names */
      err = PR_GetHostByName(hname,
				buf,
				PR_NETDB_BUF_SIZE,
				&hent);
      if (err == PR_SUCCESS) he = (hostent *)&hent;
      else he = NULL;
      if (he) {
	if (he->h_name && strchr(he->h_name, '.')) {
	  hname = he->h_name;
	}
	else if (he->h_aliases && he->h_aliases[0]) {
	  for (p = he->h_aliases; *p; ++p) {
	    if (strchr(*p, '.')) {
	      hname = *p;
	      break;
	    }
	  }
	}
      }
    }
    
    cnt += aclDNSAddHost(hname, alist, asize);

    if (aliases) {

	while (bsize < *asize) {
	    cnt += aclDNSAddAliases((*alist)[bsize], alist, asize);
	    ++bsize;
	}
    }

    return cnt;
}

/*
 * Description (aclFindByName)
 *
 *	This function searches an ACL context for an ACL with a specified
 *	name.  The caller can specify a complete name or a prefix of a
 *	name.  An list of access right names can also be specified along
 *	with a flag to indicate whether a) all rights in the list must be
 *	controlled by the target ACL, b) any rights in the list are
 *	controlled by the target ACL, or c) exactly the rights in the
 *	list are controlled by the target ACL.  The caller can search
 *	from the beginning of the ACL context, or starting with the
 *	ACL following the last one found.
 *
 * Arguments:
 *
 *	acc		- pointer to ACL context (returned by accReadFile())
 *	aclname		- pointer to ACL name or name prefix string
 *	rights		- pointer to array of access right name strings
 *	flags		- bit flags:
 *		ACLF_NPREFIX:	aclname is a prefix string
 *		ACLF_RALL:	all rights in the rights list must be
 *				present in the target ACL
 *		ACLF_REXACT:	rights list must match target ACL rights
 *				exactly
 *	pacl		- pointer to returned ACL pointer
 *			  Note: if *pacl is non-null on entry to the
 *			  function, then the search begins with the ACL
 *			  following the referenced ACL.
 *
 * Returns:
 *
 *	If a matching ACL is found, the result value is one and a
 *	pointer to the ACL is returned via 'pacl'.  If a matching ACL
 *	is not found, the result is zero and the ACL pointer referenced
 *	by 'pacl' (if any) is unchanged.
 */

NSAPI_PUBLIC int aclFindByName(ACContext_t * acc, char * aclname,
			       char **rights, int flags, ACL_t **pacl)
{
    ACL_t * acl;			/* ACL pointer */
    char **rlist;			/* ACL right name list */
    unsigned long rspec;		/* specified rights mask */
    unsigned long rmask;		/* current ACL rights mask */
    int nlen;				/* length of aclname string */
    int rv;				/* result value */

    /* Determine what rights are specified */
    rspec = (rights == 0) ? (~0) : aclGetRightsMask(acc, rights);

    /* If the specified ACL name is a prefix, search the ACL list */
    if (flags & ACLF_NPREFIX) {

	nlen = strlen(aclname);

	/*
	 * Start looking at specified ACL, if any, and
	 * otherwise start from the beginning of the ACL list.
	 */
	if ((pacl != 0) && (*pacl != 0)) acl = *pacl;
	else acl = acc->acc_acls;
    }
    else {

	/* The aclname is the complete ACL name, so look it up */
	rv = symTableFindSym(acc->acc_stp,
			     aclname, ACLSYMACL, (void **)&acl);
	if (rv != 0) acl = 0;
    }

    while (acl != 0) {

	/* If a prefix was specified, find the next matching ACL */
	if (flags & ACLF_NPREFIX) {

	    for (; acl != 0; acl = acl->acl_next) {

		if (!strncasecmp(aclname, acl->acl_sym.sym_name, nlen)) break;
	    }

	    /* Did we find an ACL with a matching name? */
	    if (acl == 0) break;
	}


	/* Get the ACL rights mask */
	rlist = aclGetRights(acl);
	rmask = aclGetRightsMask(acc, rlist);
	free_strlist(rlist);

	/* Any matching rights bits? */
	if ((rmask & rspec) != 0) {

	    /* Want exact match? */
	    if (flags & ACLF_REXACT) {

		/* Yes, do we have it?  Exit loop if so */
		if (rmask == rspec) break;
	    }
	    else if (flags & ACLF_RALL) {

		/* All specified rights bits must be set */
		if ((rmask & rspec) == rspec) break;
	    }
	    else break;
	}
	else if ((flags & ACLF_NPREFIX) == 0) {

	    /* If a full ACL name was given, fail now */
	    acl = 0;
	}
    }

    if (acl != 0) {

	if (pacl != 0) *pacl = acl;
	return 1;
    }

    return 0;
}

/*
 * Description (aclGetAuthMethod)
 *
 *	This function returns the authentication method associated
 *	with an ACL "authenticate" directive.  The method name is
 *	returned as a string, "basic" or "SSL".
 *
 * Arguments:
 *
 *	acl			- pointer to ACL structure
 *	dirno			- directive number in ACL (1..n)
 *
 * Returns:
 *
 *	A pointer to a dynamically-allocated string is returned.
 *	If the specified directive is not an "authenticate" directive,
 *	or some other error occurs, a null pointer is returned.
 */

NSAPI_PUBLIC char * aclGetAuthMethod(ACL_t * acl, int dirno)
{
    ACDirective_t * acd;		/* ACL directive pointer */
    RealmSpec_t * rsp;			/* realm specification pointer */
    char * meth = 0;			/* return string pointer */
    char * mp;				/* method string pointer */
    int i;				/* current directive number */

    /* Locate the specified directive */
    for (i = 1, acd = acl->acl_dirf;
	 (i < dirno) && (acd != 0); ++i, acd = acd->acd_next) ;

    /* Did we find the specified directive? */
    if (acd != 0) {

	/* Yes, is it an "authenticate" directive? */
	if ((acd->acd_action & ACD_ACTION) == ACD_AUTH) {

	    /* Yes, get pointer to realm specification */
	    rsp = acd->acd_auth.au_realm;

	    if (rsp != 0) {

		/* Convert authentication method code to a string */
		switch (rsp->rs_realm.rlm_ameth) {
		  case AUTH_METHOD_BASIC:
		    mp = "basic";
		    break;

		  case AUTH_METHOD_SSL:
		    mp = "SSL";
		    break;

		  default:
		    mp = 0;
		    break;
		}
	    }
	}
    }

    /* Did we come up with a method string? */
    if (mp != 0) {

	/* Yes, copy it to a dynamically allocated string */
	meth = (char *)MALLOC(strlen(mp) + 1);
	if (meth != 0) {
	    strcpy(meth, mp);
	}
    }

    /* Return method string pointer */
    return meth;
}

/*
 * Description (aclGetDatabase)
 *
 *	This function returns the name of the authentication database
 *	associated with an ACL "authenticate" directive.  The database
 *	name is returned as a dynamically allocated string.
 *
 * Arguments:
 *
 *	acl			- pointer to ACL structure
 *	dirno			- directive number in ACL (1..n)
 *
 * Returns:
 *
 *	A pointer to a dynamically-allocated string is returned.
 *	If the specified directive is not an "authenticate" directive,
 *	or some other error occurs, a null pointer is returned.
 */

NSAPI_PUBLIC char * aclGetDatabase(ACL_t * acl, int dirno)
{
    ACDirective_t * acd;		/* ACL directive pointer */
    RealmSpec_t * rsp;			/* realm specification pointer */
    char * dbname = 0;			/* return string pointer */
    char * dp;				/* database name string pointer */
    int i;				/* current directive number */

    /* Locate the specified directive */
    for (i = 1, acd = acl->acl_dirf;
	 (i < dirno) && (acd != 0); ++i, acd = acd->acd_next) ;

    /* Did we find the specified directive? */
    if (acd != 0) {

	/* Yes, is it an "authenticate" directive? */
	if ((acd->acd_action & ACD_ACTION) == ACD_AUTH) {

	    /* Yes, get pointer to realm specification */
	    rsp = acd->acd_auth.au_realm;

	    if (rsp != 0) {

		/* Get pointer to authentication database name */
		dp = rsp->rs_realm.rlm_dbname;
	    }
	}
    }

    /* Did we come up with a database name string? */
    if (dp != 0) {

	/* Yes, copy it to a dynamically allocated string */
	dbname = (char *)MALLOC(strlen(dp) + 1);
	if (dbname != 0) {
	    strcpy(dbname, dp);
	}
    }

    /* Return database name string pointer */
    return dbname;
}

/*
 * Description (aclGetHosts)
 *
 *	This function retrieves a list of hosts associated with an ACL
 *	"allow" or "deny" directive.  The list is returned as an array
 *	of pointers to dynamically allocated strings.  Each string is
 *	a fully-qualified domain name, or a domain name suffix, or an
 *	IP host address, or an IP address with a netmask.  Examples:
 *
 *		"doon.mcom.com"
 *		"*.mcom.com"
 *		"205.217.235.57"
 *		"205.217.235.0 255.255.255.0"
 *
 * Arguments:
 *
 *	acl			- pointer to ACL structure
 *	dirno			- directive number in ACL (1..n)
 *	clsno			- client specification number (1..m)
 *
 * Returns:
 *
 *	A pointer to a dynamically-allocated string is returned.
 *	If the specified directive is not an "allow" or "deny" directive,
 *	or some other error occurs, a null pointer is returned.
 */

NSAPI_PUBLIC char **aclGetHosts(ACL_t * acl, int dirno, int clsno)
{
    HostSpec_t * hsp;			/* pointer to host specification */
    ACDirective_t * acd;		/* ACL directive pointer */
    ACClients_t * clp;			/* client specification pointer */
    char **hlist = 0;			/* pointer to returned list */
    int i;				/* current directive number */
    int dtype;				/* current directive type */
    int ipcnt;				/* count of IP filter entries */
    int dnscnt;				/* count of DNS filter entries */
    int rv;				/* result value */

    /* Locate the specified directive */
    for (i = 1, acd = acl->acl_dirf;
	 (i < dirno) && (acd != 0); ++i, acd = acd->acd_next) ;

    /* Did we find the specified directive? */
    if (acd != 0) {

	/* Yes, is it an "allow" or "deny" directive? */
	dtype = acd->acd_action & ACD_ACTION;

	if ((dtype == ACD_ALLOW) || (dtype == ACD_DENY)) {

	    /* Yes, get pointer to host specification */
	    for (i = 1, clp = acd->acd_cl;
		 (i < clsno) && (clp != 0); clp = clp->cl_next, ++i) ;

	    if (clp != 0) {

		hsp = clp->cl_host;

		/* Is there a host specification? */
		if (hsp != 0) {

		    /* Get count of IP filter host/network specifications */
		    ipcnt = aclIPFilterStrings(0, &hsp->hs_host.inh_ipf);

		    /* Get count of DNS filter specifications */
		    dnscnt = aclDNSFilterStrings(0, &hsp->hs_host.inh_dnf);

		    /* Allocate the array of strings */
		    hlist = new_strlist(ipcnt + dnscnt);

		    /* Put IP filter specifications in list */
		    rv = aclIPFilterStrings(hlist, &hsp->hs_host.inh_ipf);

		    /* Put DNS filter specifications in list */
		    rv = aclDNSFilterStrings(hlist + rv,
					     &hsp->hs_host.inh_dnf);
		}
		else {

		    /* Return an empty host list */
		    hlist = new_strlist(0);
		}
	    }
	}
    }

    return hlist;
}

/*
 * Description (aclGetPrompt)
 *
 *	This function returns the authentication prompt string
 *	associated with an ACL "authenticate" directive.
 *
 * Arguments:
 *
 *	acl			- pointer to ACL structure
 *	dirno			- directive number in ACL (1..n)
 *
 * Returns:
 *
 *	A pointer to a dynamically-allocated string is returned.
 *	If the specified directive is not an "authenticate" directive,
 *	or some other error occurs, a null pointer is returned.
 */

NSAPI_PUBLIC char * aclGetPrompt(ACL_t * acl, int dirno)
{
    ACDirective_t * acd;		/* ACL directive pointer */
    RealmSpec_t * rsp;			/* realm specification pointer */
    char * prompt = 0;			/* return string pointer */
    char * pp = 0;			/* prompt string pointer */
    int i;				/* current directive number */

    /* Locate the specified directive */
    for (i = 1, acd = acl->acl_dirf;
	 (i < dirno) && (acd != 0); ++i, acd = acd->acd_next) ;

    /* Did we find the specified directive? */
    if (acd != 0) {

	/* Yes, is it an "authenticate" directive? */
	if ((acd->acd_action & ACD_ACTION) == ACD_AUTH) {

	    /* Yes, get pointer to realm specification */
	    rsp = acd->acd_auth.au_realm;

	    if (rsp != 0) {

		/* Get pointer to prompt string */
		pp = rsp->rs_realm.rlm_prompt;
	    }
	}
    }

    /* Did we come up with a prompt string? */
    if (pp != 0) {

	/* Yes, copy it to a dynamically allocated string */
	prompt = (char *)MALLOC(strlen(pp) + 1);
	if (prompt != 0) {
	    strcpy(prompt, pp);
	}
    }

    /* Return prompt string pointer */
    return prompt;
}

/*
 * Description (aclGetRights)
 *
 *	This function returns a list of the names of access rights
 *	associated with a specified ACL.  The list is returned as
 *	a pointer to an array of dynamically allocated strings.
 *
 * Arguments:
 *
 *	acl			- pointer to ACL structure
 *
 * Returns:
 *
 *	A pointer to a dynamically allocated array of strings is
 *	returned.
 */

NSAPI_PUBLIC char **aclGetRights(ACL_t * acl)
{
    char **rlist = 0;			/* pointer to returned list */
    ACContext_t * acc = acl->acl_acc;	/* context for this ACL */
    RightDef_t * rdp;			/* pointer to right definition */
    USI_t * rid;			/* pointer to list of right ids */
    int rcnt;				/* number of rights */
    int i;				/* loop index */

    rcnt = 0;

    /* Get number of rights and pointer to list of right ids */
    if (acl->acl_rights != 0) {
	rcnt = UILCOUNT(&acl->acl_rights->rs_list);
	rid = UILLIST(&acl->acl_rights->rs_list);
    }

    /* Create the array of string pointers */
    rlist = new_strlist(rcnt);

    /* Make one pass through all rights defined in this ACL context */
    for (rdp = acc->acc_rights; rdp != 0; rdp = rdp->rd_next) {

	/* Translate the right ids for this ACL as we go */
	for (i = 0; i < rcnt; ++i) {
	    if (rid[i] == rdp->rd_id) {
		rlist[i] = STRDUP(rdp->rd_sym.sym_name);
		break;
	    }
	}
    }

    return rlist;
}

/*
 * Description (aclGetRightsMask)
 *
 *	This function returns a bit mask representing the access right
 *	names in a specified list, for a given ACL context.  It requires
 *	that the right ids be numerically smaller than the number of bits
 *	in an unsigned long.
 *
 * Arguments:
 *
 *	acc			- pointer to ACL context structure
 *	rlist			- pointer to array of right name strings
 *
 * Returns:
 *
 *	A bit mask containing a one bit for each right id.
 */

NSAPI_PUBLIC unsigned long aclGetRightsMask(ACContext_t * acc, char **rlist)
{
    RightDef_t * rdp;			/* pointer to right definition */
    int rv;				/* result value */
    unsigned long rmask = 0;		/* resulting mask */

    while (*rlist != 0) {

	rv = symTableFindSym(acc->acc_stp, *rlist, ACLSYMRIGHT, (void **)&rdp);
	if (rv == 0) {
	    if (rdp->rd_id < (8*sizeof(unsigned long))) {
		rmask |= (1 << rdp->rd_id);
	    }
	}

	++rlist;
    }

    return rmask;
}

/*
 * Description (aclGetSignature)
 *
 *	This function returns a signature string for a specified ACL.
 *	A signature string is a compact textual representation of the
 *	structure of an ACL, where each character represents one ACL
 *	directive, and the order of the characters indicates the order
 *	of the ACL directives.
 *
 *	The specific characters used in a signature string are:
 *
 *	a - default allow ...
 *	A - always allow ...
 *	d - default deny ...
 *	D - always deny
 *	r - default authenticate ...
 *	R - always authenticate ...
 *	x - default execute ...
 *	X - always execute ...
 *
 * Arguments:
 *
 *	acl		- pointer to ACL
 *
 * Returns:
 *
 *	The result value is a pointer to a dynamically allocated string
 *	containing the ACL signature.  Insufficient dynamic memory is
 *	indicated by a null return pointer.
 */

NSAPI_PUBLIC char * aclGetSignature(ACL_t * acl)
{
    ACDirective_t * acd;		/* pointer to ACL directive */
    char * sigptr;			/* pointer to signature buffer */
    char * sp;				/* pointer to signature character */
    int ndir;				/* number of directives */

    /* First count the number of directives in the ACL */
    ndir = 0;

    for (acd = acl->acl_dirf; acd != NULL; acd = acd->acd_next) ++ndir;

    /* Allocate a buffer for the signature string */
    sigptr = (char *)MALLOC(ndir + 1);
    if (sigptr != 0) {

	sp = sigptr;

	for (acd = acl->acl_dirf; acd != NULL; acd = acd->acd_next) {

	    /* Classify this directive and generate signature character */
	    switch (acd->acd_action & ACD_ACTION) {

	      case ACD_ALLOW:
		*sp++ = (acd->acd_flags & ACD_ALWAYS) ? 'A' : 'a';
		break;

	      case ACD_DENY:
		*sp++ = (acd->acd_flags & ACD_ALWAYS) ? 'D' : 'd';
		break;

	      case ACD_AUTH:
		*sp++ = (acd->acd_flags & ACD_ALWAYS) ? 'R' : 'r';
		break;

	      case ACD_EXEC:
		*sp++ = (acd->acd_flags & ACD_ALWAYS) ? 'X' : 'x';
		break;

	      default:
		*sp++ = '?';
		break;
	    }
	}

	/* Terminate string */
	*sp = 0;
    }

    return sigptr;
}

/*
 * Description (aclGetUsers)
 *
 *	This function returns a list of user and/or groups names
 *	associated with an "allow" or "deny" ACL directive.
 *
 * Arguments:
 *
 *	acl			- pointer to ACL structure
 *	dirno			- directive number in ACL (1..n)
 *	clsno			- client specification number (1..m)
 *
 * Returns:
 *
 *	A pointer to a dynamically-allocated array of strings is returned.
 *	If the specified directive is not an "allow" or "deny" directive,
 *	or some other error occurs, a null pointer is returned.
 */

NSAPI_PUBLIC char **aclGetUsers(ACL_t * acl, int dirno, int clsno)
{
    UserSpec_t * usp;			/* pointer to user specification */
    ACDirective_t * acd;		/* ACL directive pointer */
    Realm_t * rlm;			/* current authentication realm */
    ACClients_t * clp;			/* client specification pointer */
    char **ulist = 0;			/* pointer to returned list */
    int i;				/* current directive number */
    int dtype;				/* directive type */
    int ucnt;				/* user count */
    int gcnt;				/* group count */
    int rv;				/* result value */

    rlm = 0;

    /* Locate the specified directive */
    for (i = 1, acd = acl->acl_dirf;
	 (i < dirno) && (acd != 0); ++i, acd = acd->acd_next) {

	/* Keep track of authentication realm */
	if ((acd->acd_action & ACD_ACTION) == ACD_AUTH) {

	    if (acd->acd_auth.au_realm != 0) {

		/* Update current realm pointer */
		rlm = &acd->acd_auth.au_realm->rs_realm;
	    }
	}
    }

    /* Did we find the specified directive? */
    if (acd != 0) {

	/* Yes, is it an "allow" or "deny" directive? */
	dtype = acd->acd_action & ACD_ACTION;

	if ((dtype == ACD_ALLOW) || (dtype == ACD_DENY)) {

	    /* Yes, get pointer to user specification */
	    for (i = 1, clp = acd->acd_cl;
		 (i < clsno) && (clp != 0); clp = clp->cl_next, ++i) ;

	    if (clp != 0) {

		usp = clp->cl_user;

		/* Is there a user specification? */
		if (usp != 0) {

		    /* Does it refer to "all" authenticated users? */
		    if (usp->us_flags & ACL_USALL) {

			/* Yes, short list */
			ulist = new_strlist(1);
			ulist[0] = STRDUP("all");
		    }
		    else if (rlm != 0) {

			/* Open the realm authentication database */
			rv = nsadbOpen(0,
				       rlm->rlm_dbname, 0, &rlm->rlm_authdb);
			if (rv < 0) goto err_dbopen;
			rlm->rlm_aif = &NSADB_AuthIF;

			/*
			 * Allocate a list to hold the indicated number
			 * of users and groups.
			 */

			ucnt = UILCOUNT(&usp->us_user.uu_user);
			gcnt = UILCOUNT(&usp->us_user.uu_group);

			ulist = new_strlist(ucnt + gcnt);

			/* Convert user ids to user names */
			rv = aclIdsToNames(ulist,
					   &usp->us_user.uu_user, 1, rlm);

			/* Convert group ids to group names */
			rv = aclIdsToNames(ulist + rv,
					   &usp->us_user.uu_group, 0, rlm);

			/* Close the realm authentication database */
			nsadbClose(rlm->rlm_authdb, 0);
			rlm->rlm_authdb = 0;
			rlm->rlm_aif = 0;
		    }
		}
		else {

		    /* There is no user specification, which means "anyone" */
		    ulist = new_strlist(1);
		    ulist[0] = STRDUP("anyone");
		}
	    }
	}
    }

    return ulist;

  err_dbopen:
    return 0;
}

/* Argument structure for DNS hash table enumeration */
typedef struct DNSHelp_s DNSHelp_t;
struct DNSHelp_s {
    char **dh_list;		/* pointer to array of strings */
    int dh_cnt;			/* number of entries counted */
};

NSAPI_PUBLIC int aclDNSHelper(Symbol_t * sym, void * parg)
{
    DNSHelp_t * dhp = (DNSHelp_t *)parg;
    char * dnsp;			/* DNS spec string pointer */
    int len;				/* length of DNS spec string */

    /* If there's a list add a copy of this entry to it */
    if (dhp->dh_list != 0) {

	/* Get length of specification string */
	len = strlen(sym->sym_name);

	/* Is it a fully-qualified domain name or "*"? */
	if ((sym->sym_type == 0) && strcmp(sym->sym_name, "*")) {

	    /* No, prepend "*" */
	    dnsp = (char *)MALLOC(len + 2);
	    if (dnsp != 0) {
		dnsp[0] = '*';
		strcpy(&dnsp[1], sym->sym_name);
	    }
	}
	else {

	    /* Copy fully-qualified domain name or "*" */
	    dnsp = STRDUP(sym->sym_name);
	}

	if (dnsp != 0) {

	    /* Add entry to list */
	    dhp->dh_list[dhp->dh_cnt] = dnsp;

	    /* Count another entry */
	    dhp->dh_cnt += 1;
	}
    }
    else {

	/* Count another entry, but add nothing */
	dhp->dh_cnt += 1;
    }

    /* Continue enumeration */
    return 0;
}

NSAPI_PUBLIC int aclDNSFilterStrings(char **list, DNSFilter_t * dnf)
{
    void * table;		/* hash table pointer */
    DNSHelp_t dnshelp;		/* hash table enumeration arguments */

    /* Get the filter hash table pointer */
    table = dnf->dnf_hash;

    /* Return if no table */
    if (table == 0) {
	return 0;
    }

    /* Initialize arguments for hash table enumeration */
    dnshelp.dh_list = list;
    dnshelp.dh_cnt = 0;

    /* Now enumerate the hash table */
    symTableEnumerate(table, (void *)&dnshelp, aclDNSHelper);

    /* Return number of hash table entries */
    return dnshelp.dh_cnt;
}

NSAPI_PUBLIC int aclIPFilterStrings(char **list, IPFilter_t * ipf)
{
    IPLeaf_t * leaf;		/* radix tree leaf pointer */
    IPNode_t * ipn;		/* current internal node */
    IPNode_t * lastipn;		/* last lower node seen in search */
    IPNode_t * previpn;		/* last upper node seen in search */
    char * ips;			/* pointer to IP specification buffer */
    int lcnt;			/* count of leaf entries in IP filter */
    int i;

    /* Return if no IP filter radix tree */
    if ((ipf == 0) || (ipf->ipf_tree == 0)) {
	return 0;
    }

    lastipn = NULL;
    ipn = ipf->ipf_tree;
    lcnt = 0;

    while (ipn != NULL) {

	/*
	 * Work down the tree processing child links in order.
	 * Exit the loop when there are no more internal nodes.
	 */
	while ((ipn != NULL) && (ipn->ipn_type == IPN_NODE)) {

	    /*
	     * lastipn points to the last child of the current (ipn)
	     * node that was visited, if any.
	     */
	    if (lastipn == NULL) {

		/* No children of this node visited yet */
		i = 0;
	    }
	    else {

		/* Look for the last link we followed down the tree */
		for (i = 0; i < IPN_NLINKS; ++i) {
		    if (ipn->ipn_links[i] == lastipn) break;
		}

		/* Step to the next link down the tree */
		++i;
	    }

	    /* Find the next link down the tree */
	    for (; i < IPN_NLINKS; ++i) {
		if (ipn->ipn_links[i] != 0) break;
	    }

	    /* Did we find a link down the tree? */
	    if (i < IPN_NLINKS) {

		/* Yes follow it */
		previpn = ipn;
		lastipn = 0;
		ipn = ipn->ipn_links[i];
	    }
	    else {

		/*
		 * Otherwise go back up the tree, remembering where we
		 * came from so that we don't follow the same link back
		 * down.
		 */
		lastipn = ipn;
		ipn = ipn->ipn_parent;
	    }
	}

	/* Did we end up with a non-NULL node pointer? */
	if (ipn != NULL) {

	    /* It must be a leaf node */
	    assert(ipn->ipn_type == IPN_LEAF);
	    leaf = (IPLeaf_t *)ipn;

	    /* Is there a list to output specifications to? */
	    if (list != 0) {

		/* Is there a netmask associated with this entry? */
		if (leaf->ipl_netmask == 0xffffffff) {

		    /* No, just format the IP address */
		    ips = (char *)MALLOC(16);
		    sprintf(ips, "%d.%d.%d.%d",
			    (int)((leaf->ipl_ipaddr >> 24) & 0xff),
			    (int)((leaf->ipl_ipaddr >> 16) & 0xff),
			    (int)((leaf->ipl_ipaddr >> 8) & 0xff),
			    (int)(leaf->ipl_ipaddr & 0xff));
		}
		else if (leaf->ipl_netmask == 0xffffff00) {
		    ips = (char *)MALLOC(16);
		    sprintf(ips, "%d.%d.%d.*",
			    (int)((leaf->ipl_ipaddr >> 24) & 0xff),
			    (int)((leaf->ipl_ipaddr >> 16) & 0xff),
			    (int)((leaf->ipl_ipaddr >> 8) & 0xff));
		}
		else if (leaf->ipl_netmask == 0xffff0000) {
		    ips = (char *)MALLOC(16);
		    sprintf(ips, "%d.%d.*",
			    (int)((leaf->ipl_ipaddr >> 24) & 0xff),
			    (int)((leaf->ipl_ipaddr >> 16) & 0xff));
		}
		else if (leaf->ipl_netmask == 0xff000000) {
		    ips = (char *)MALLOC(16);
		    sprintf(ips, "%d.*",
			    (int)((leaf->ipl_ipaddr >> 24) & 0xff));
		}
		else {

		    /* Format IP address and netmask */
		    ips = (char *)MALLOC(32);
		    sprintf(ips, "%d.%d.%d.%d %d.%d.%d.%d",
			    (int)((leaf->ipl_ipaddr >> 24) & 0xff),
			    (int)((leaf->ipl_ipaddr >> 16) & 0xff),
			    (int)((leaf->ipl_ipaddr >> 8) & 0xff),
			    (int)(leaf->ipl_ipaddr & 0xff),
			    (int)((leaf->ipl_netmask >> 24) & 0xff),
			    (int)((leaf->ipl_netmask >> 16) & 0xff),
			    (int)((leaf->ipl_netmask >> 8) & 0xff),
			    (int)(leaf->ipl_netmask & 0xff));
		}

		list[lcnt] = ips;
	    }

	    /* Count another leaf */
	    ++lcnt;

	    /* Back up to last internal node */
	    lastipn = ipn;
	    ipn = previpn;
	}
    }

    /* Return number of IP specifications in the filter */
    return lcnt;
}

/*
 * Description (aclIdsToNames)
 *
 *	This function converts a list of user or group ids to user or
 *	group names.  The caller provides a pointer to an array of
 *	strings, which is filled in with pointers to dynamically
 *	allocated strings containing the user or group names.  The
 *	array is assumed to be large enough to hold the number of ids
 *	present.
 *
 * Arguments:
 *
 *	list			- pointer to name list
 *	uilptr			- pointer to list of ids
 *	uflag			- non-zero if list contains user ids,
 *				  zero if group ids
 *	rlm			- pointer to authentication realm for ids
 *
 * Returns:
 *
 *	The number of ids successfully converted to names is returned.
 */

NSAPI_PUBLIC int aclIdsToNames(char **list,
			       USIList_t * uilptr, int uflag, Realm_t * rlm)
{
    char * name;			/* current name pointer */
    USI_t * idp;			/* current id pointer */
    int cnt;				/* number of ids in list */
    int i;				/* loop index */
    int j;				/* list index */
    int rv;				/* result value */

    /* Check uflag and convert it */
    uflag = (uflag) ? AIF_USER : AIF_GROUP;

    /* Get list pointer and number of entries */
    idp = UILLIST(uilptr);
    cnt = UILCOUNT(uilptr);

    /* Initialize number of entries in list */
    j = 0;

    /* For each id in the list */
    for (i = 0; i < cnt; ++i) {

	/* Does this realm look useable? */
	if (rlm->rlm_aif != 0) {

	    /* Lookup name for this id */
	    rv = (*rlm->rlm_aif->aif_idtoname)(0, rlm->rlm_authdb, *idp,
					       uflag, &name);

	    /* Got it? */
	    if (name != 0) {

		/* Yes, add copy of name string to list */
		list[j++] = STRDUP(name);
	    }
	}

	++idp;
    }

    /* Return number of names added to list */
    return j;
}

/*
 * Description (aclMakeNew)
 *
 *	This function creates a new ACL in a specified ACL context.  The
 *	caller provides an ACL signature string, and a name or name
 *	prefix string string to name the new ACL.  If a name prefix is
 *	specified, a randomized suffix of numeric digits is appended to
 *	the prefix to create a unique ACL name.  The caller can also
 *	specify the access rights controlled by the new ACL.
 *
 * Arguments:
 *
 *	acc		- pointer to ACL context (returned by accReadFile())
 *	aclsig		- pointer to ACL signature string
 *			  (see aclGetSignature() for description)
 *	aclname		- pointer to ACL name or name prefix string
 *			  (may be null)
 *	rights		- pointer to array of ACL access right names
 *	flags		- bit flags:
 *		ACLF_NPREFIX:	aclname is a prefix string
 *	pacl		- pointer to returned ACL pointer
 *
 * Returns:
 *
 *	If the ACL is create successfully, the result value is zero and a
 *	pointer to the ACL is returned via 'pacl'.  An error is indicated
 *	by a negative return value (ACLERRxxxxx - see aclerror.h).
 */

NSAPI_PUBLIC int aclMakeNew(ACContext_t * acc, char * aclsig, char * aclname,
			    char **rights, int flags, ACL_t **pacl)
{
    ACL_t * acl;			/* pointer to new ACL structure */
    ACDirective_t * acd;		/* current directive pointer */
    char * sp;				/* ACL signature pointer */
    char * np = 0;			/* generated unique name pointer */
    int nlen;				/* ACL name or name prefix length */
    int rv;				/* result value */

    /* Get the length of the ACL name or name prefix string */
    nlen = strlen(aclname);

    /*
     * If the specified aclname is a prefix, create a buffer to hold
     * the generated unique name.
     */
    if (flags & ACLF_NPREFIX) {

	np = (char *)MALLOC(nlen + 10);
	if (np == 0) goto err_nomem;

	/* Copy prefix to new buffer */
	strcpy(np, aclname);

	/* Add a hyphen */
	np[nlen++] = '-';
	np[nlen] = 0;

	/* Seed the random number generator */
	srand((int)time(NULL));

	/* Loop until a unique name is generated */
	for (;;) {

	    /* Append a random number (up to 4 digits) to the name */
	    sprintf(&np[nlen], "%d", rand() % 10000);

	    /* Does an ACL with this name already exist? */
	    rv = symTableFindSym(acc->acc_stp, np, ACLSYMACL, (void **)&acl);

	    /* Exit loop if not */
	    if (rv != 0) break;
	}

	/* New ACL name is referenced by np */
	aclname = np;
    }
    else {

	/* Does an ACL with the specified name already exist? */
	rv = symTableFindSym(acc->acc_stp, aclname, ACLSYMACL, (void **)&acl);

	/* Error if so */
	if (rv == 0) goto err_dup;
    }

    /* Now create the new ACL */
    rv = aclCreate(0, acc, aclname, &acl);
    if (rv < 0) goto err_create;

    /* Set the specified rights in the ACL */
    rv = aclSetRights(acl, rights, 1);

    if (aclsig != 0) {

	/* Loop to create empty directives according to the ACL signature */
	for (sp = aclsig; *sp != 0; ++sp) {

	    /* Create an empty directive */
	    acd = aclDirectiveCreate();
	    if (acd == 0) goto err_nomem2;

	    /* Interpret signature character to set action and flags */
	    switch (*sp) {
	    
	      case 'A':
		acd->acd_flags = ACD_ALWAYS;
		acd->acd_action = ACD_ALLOW;
		break;

	      case 'a':
		acd->acd_action = ACD_ALLOW;
		break;

	      case 'D':
		acd->acd_flags = ACD_ALWAYS;
		acd->acd_action = ACD_DENY;
		break;

	      case 'd':
		acd->acd_action = ACD_DENY;
		break;

	      case 'R':
		acd->acd_flags = ACD_ALWAYS;
		acd->acd_action = ACD_AUTH;
		break;

	      case 'r':
		acd->acd_action = ACD_AUTH;
		break;

	      case 'X':
		acd->acd_flags = ACD_ALWAYS;
		acd->acd_action = ACD_EXEC;
		break;

	      case 'x':
		acd->acd_action = ACD_EXEC;
		break;

	      default:
		FREE(acd);
		goto err_badsig;
	    }

	    /* Add this directive to the ACL */
	    rv = aclDirectiveAdd(acl, acd);
	}
    }

    if (pacl) *pacl = acl;

    return 0;

  err_nomem:
    rv = ACLERRNOMEM;
    goto punt;

  err_dup:
    rv = ACLERRDUPSYM;
    goto punt;

  err_create:
    rv = ACLERRNOMEM;
    goto punt;

  err_nomem2:
    rv = ACLERRNOMEM;
    goto punt;

  err_badsig:
    rv = ACLERRUNDEF;
    goto punt;

  punt:
    return rv;
}

/*
 * Description (aclPutAllowDeny)
 *
 *	This function adds an "allow" or "deny" directive to a
 *	specified ACL.  The associated lists of users and hosts are
 *	specified as arrays of pointers to strings, each terminated
 *	by a null pointer.
 *
 * Arguments:
 *
 *	errp			- error frame list pointer (may be null)
 *	acl			- pointer to ACL structure
 *	always			- non-zero for "Always" instead of "Default"
 *	allow			- non-zero for "allow" instead of "deny"
 *	users			- pointer to array of user strings
 *				  (may be null --> "anyone")
 *	hosts			- pointer to array of host DNS/IP strings
 *				  (may be null --> no "at host-list" clause
 *				   --> all hosts match (quicker than "*"))
 *
 * Returns:
 *
 *	The function return value is zero if all goes well.
 */

NSAPI_PUBLIC int aclPutAllowDeny(NSErr_t * errp, ACL_t * acl, int always,
				 int allow, char **users, char **hosts)
{
    ACDirective_t * acd = 0;		/* current directive pointer */
    Realm_t * rlm = 0;			/* current realm pointer */
    ACClients_t * clp = 0;		/* client specification pointer */
    UserSpec_t * usp = 0;		/* user specification pointer */
    char **lst;				/* name entry pointer */
    char * cp;				/* string pointer */
    IPAddr_t ipaddr;			/* IP address */
    IPAddr_t netmask;			/* network IP mask */
    int haveusers = 0;			/* user list is not empty */
    int fqdn;				/* fully-qualified DNS name */
    int rv;				/* result value */

    if (acl == 0) goto punt;

    /* Is there a non-empty user list? */
    if ((users != 0) && (users[0] != 0)) {

	haveusers = 1;

	/* Scan the ACL directives to establish the current realm */
	for (acd = acl->acl_dirf; acd != 0; acd = acd->acd_next) {

	    if ((acd->acd_action == ACD_AUTH) &&
		(acd->acd_auth.au_realm != 0)) {

		/* Update realm pointer */
		rlm = &acd->acd_auth.au_realm->rs_realm;
	    }
	}
    }

    /* Create a new directive structure */
    acd = aclDirectiveCreate();
    if (acd == 0) goto punt;

    /* Set the directive type and flags */
    acd->acd_action = (allow) ? ACD_ALLOW : ACD_DENY;
    acd->acd_flags = (always) ? ACD_ALWAYS : 0;
    acd->acd_cl = 0;

    /* Create an ACClients_t structure and point the directive at it */
    clp = aclClientsDirCreate();
    if (clp == 0) goto punt;

    acd->acd_cl = clp;

    /* Do we have a non-empty user list? */
    if (haveusers) {

	/* Yes, create a user specification structure */
	usp = aclUserSpecCreate();
	if (usp == 0) goto punt;

	/* Point the ACClients_t structure at the UserSpec_t */
	clp->cl_user = usp;

	/* Is the first entry "all"? */
	if (!strcasecmp(users[0], "all")) {

	    /* Yes, just set the flag to indicate all users */
	    usp->us_flags = ACL_USALL;
	}
	else {

	    /* Convert user/group names to ids */
	    for (lst = users; *lst != 0; ++lst) {

		rv = aclAuthNameAdd(errp, usp, rlm, *lst);
		if (rv < 0) goto punt;
	    }

	    /* Close database in current realm if any */
	    if (rlm && rlm->rlm_authdb) {
		(*rlm->rlm_aif->aif_close)(rlm->rlm_authdb, 0);
		rlm->rlm_authdb = 0;
	    }
	}
    }

    /* Got anything in the host list? */
    if ((hosts != 0) && (hosts[0] != 0)) {

	/* Scan hosts entries */
	for (lst = hosts; *lst != 0; ++lst) {

	    /* Get pointer to host specification string */
	    cp = *lst;

	    /* Does it look like an IP address? */
	    if (isdigit(*cp)) {
		int dotcnt = 0;

		netmask = 0xffffffff;

		/* Terminate IP address if there's a netmask */
		while (*cp && (*cp != ',') && !isspace(*cp)) {
		    if (*cp == '.') ++dotcnt;
		    else if (*cp == '*') {
			*cp = '0';
			if (netmask == 0xffffffff) {
			    netmask <<= (4-dotcnt)*8;
			    netmask = htonl(netmask);
			}
		    }
		    ++cp;
		}
		if (*cp != 0) *cp++ = 0;

		/* Convert the IP address to binary form */
		ipaddr = (IPAddr_t)inet_addr(*lst);
		if (ipaddr == (IPAddr_t)-1) goto punt;

		/* Look for a netmask */
		if ((netmask == 0xffffffff) && (*cp != 0)) {
		    /* Put a space back */
		    cp[-1] = ' ';
		    while (*cp && isspace(*cp)) ++cp;
		    if ((*cp != 0) && isdigit(*cp)) {

			netmask = (IPAddr_t)inet_addr(cp);
			if (netmask == (IPAddr_t)-1) {
			    if (strcmp(cp, "255.255.255.255")) goto punt;
			}
		    }
		}

		/* Add IP entry to host specification */
		rv = aclAuthIPAdd(&clp->cl_host,
				  ntohl(ipaddr), ntohl(netmask));
		if (rv) goto punt;
	    }
	    else {

		/* Does it start with "*"? */
		fqdn = 1;
		if (*cp == '*') {

		    /* If so, it's not fully-qualified */
		    fqdn = 0;

		    /* Does it start with "*."? */
		    if (cp[1] == '.') {
			/* Yes, strip the leading "*" */
			++cp;
		    }
		}

		/* Add DNS filter entry */
		rv = aclAuthDNSAdd(&clp->cl_host, cp, fqdn);
		if (rv < 0) goto punt;
	    }
	}
    }

    rv = aclDirectiveAdd(acl, acd);

    /* All went well */
    return rv;

  punt:
    /* Error */
    if (acd != 0) {
	aclDirectiveDestroy(acd);
    }
    return -1;
}

/*
 * Description (aclPutAuth)
 *
 *	This function adds an "authenticate" directive to a specified
 *	ACL.  An authentication method code, database name, and prompt
 *	string may be specified.
 *
 * Arguments:
 *
 *	errp			- error frame list pointer (may be null)
 *	acl			- pointer to ACL structure
 *	always			- non-zero for "Always" instead of "Default"
 *	amethod			- authentication method
 *				  (AUTH_METHOD_BASIC or AUTH_METHOD_SSL)
 *	dbname			- pointer to authentication database name
 *				  (may be null)
 *	prompt			- pointer to prompt string
 *				  (may be null)
 *
 * Returns:
 *
 *	The function return value is zero if all goes well.
 */

NSAPI_PUBLIC int aclPutAuth(NSErr_t * errp, ACL_t * acl, int always,
			    int amethod, char * dbname, char * prompt)
{
    ACDirective_t * acd;		/* current directive pointer */
    RealmSpec_t * rsp;			/* realm specification pointer */
    Realm_t * rlm;			/* current realm pointer */
    int rv;				/* result value */

    if (acl == 0) goto punt;

    /* Create a new directive structure */
    acd = aclDirectiveCreate();
    if (acd == 0) goto punt;

    /* Set the directive type and flags */
    acd->acd_action = ACD_AUTH;
    acd->acd_flags = (always) ? ACD_ALWAYS : 0;

    /* Create a new realm specification structure */
    rsp = (RealmSpec_t *)MALLOC(sizeof(RealmSpec_t));
    if (rsp == 0) goto punt;

    memset((void *)rsp, 0, sizeof(RealmSpec_t));
    rsp->rs_sym.sym_type = ACLSYMREALM;
    
    /* Point the directive at it */
    acd->acd_auth.au_realm = rsp;

    rlm = &rsp->rs_realm;
    rlm->rlm_ameth = amethod;
    rlm->rlm_dbname = (dbname) ? STRDUP(dbname) : 0;
    rlm->rlm_aif = &NSADB_AuthIF;
    rlm->rlm_prompt = (prompt) ? STRDUP(prompt) : 0;

    rv = aclDirectiveAdd(acl, acd);

    /* All went well */
    return rv;

  punt:
    /* Error */
    return -1;
}

/*
 * Description (aclSafeIdent)
 *
 *	This function checks whether a given string is a valid TOKEN_IDENT
 *	token, and returns a pointer to an equivalent TOKEN_STRING if not.
 *	The returned string is static and overwritten by the next call.
 *	If the argument is a valid TOKEN_IDENT, the argument pointer is
 *	the return value.
 *
 * Arguments:
 *
 *	str			- pointer to possible TOKEN_IDENT
 *
 * Returns:
 *
 *	Pointer to argument or static string, as discussed above.
 *	Strings longer than QSMAX-3 will be truncated if they are not
 *	TOKEN_IDENT.
 */

#define QSMAX	256			/* maximum string length */

NSAPI_PUBLIC char * aclSafeIdent(char * str)
{
    char * sp;
    char * qp;
    char * rp = str;
    int len = 3;
    static char qstring[QSMAX];

    if (str == 0) {
	return 0;
    }

    /* TOKEN_IDENT must begin with a letter */
    if (!isalpha(*str)) rp = qstring;

    qstring[0] = '"';

    for (sp = str, qp = &qstring[1]; *sp; ++sp) {

	/* TOKEN_IDENT contains letters, digits, or underscores */
	if (!isalpha(*sp) && !isdigit(*sp) && (*sp != '_')) {

	    rp = qstring;

	    /* Quote certain characters in a TOKEN_STRING */
	    if ((*sp == '"') || (*sp == '\n') || (*sp == '\\')) {
		len += 2;
		if (len < QSMAX) {
		    *qp++ = '\\';
		    *qp++ = *sp;
		}
		continue;
	    }
	}

	len += 1;
	if (len < QSMAX) {
	    *qp++ = *sp;
	}
    }

    *qp++ = '"';
    *qp = 0;

    return rp;
}

/*
 * Description (aclSetRights)
 *
 *	This function adds to or replaces the access rights of a given
 *	ACL with those in a specified list.
 *
 * Arguments:
 *
 *	acl			- pointer to ACL structure
 *	rights			- pointer to array of right name strings
 *	replace			- non-zero to replace current rights
 *
 * Returns:
 *
 *	The return value is zero if the operation completes successfully.
 */

NSAPI_PUBLIC int aclSetRights(ACL_t * acl, char **rights, int replace)
{
    ACContext_t * acc = acl->acl_acc;	/* ACL context */
    char **rp;				/* pointer to rights list entry */
    RightSpec_t * rsp;			/* ACL rights spec pointer */
    RightDef_t * rdp;			/* access right definition pointer */
    int rv;				/* result value */

    rsp = acl->acl_rights;
    if (rsp == 0) {
	rsp = (RightSpec_t *)MALLOC(sizeof(RightSpec_t));
	rsp->rs_sym.sym_name = 0;
	rsp->rs_sym.sym_type = ACLSYMRDEF;
	UILINIT(&rsp->rs_list);
	acl->acl_rights = rsp;
    }
    else if (replace) {
	UILFREE(&rsp->rs_list);
    }

    for (rp = rights; *rp != 0; ++rp) {

	rv = symTableFindSym(acc->acc_stp, *rp, ACLSYMRIGHT, (void **)&rdp);
	if (rv == 0) {
	    usiInsert(&rsp->rs_list, rdp->rd_id);
	}
    }

    return 0;
}

/*
 * Description (accWriteFile)
 *
 *	This function writes out the ACLs in a specified ACL context
 *	to a specified text file.
 *
 * Arguments:
 *
 *	acc			- pointer to ACL context structure
 *	filename		- pointer to output filename
 *	flags			- bit flags (unused - must be zero)
 *
 * Returns:
 *
 *	The result value is zero if there are no errors.  Otherwise
 *	it is a negative error code (ACLERRxxxx - see aclerror.h).
 */

NSAPI_PUBLIC int accWriteFile(ACContext_t * acc, char * filename, int flags)
{
    FILE * ofp;			/* output file pointer */
    ACL_t * newacl;		/* pointer to reverse ACL list */
    ACL_t * acl;		/* current ACL pointer */
    ACDirective_t * acd;	/* current ACL directive pointer */
    ACClients_t * clp;		/* client specification pointer */
    Realm_t * rlm;		/* realm pointer */
    char **list;		/* list of string pointers */
    char **lp;			/* pointer to list entry */
    int ndir;			/* current directive number */
    int action;			/* current directive action code */
    int ncls;			/* current client specification number */

    ofp = fopen(filename, "w");
    if (ofp == 0) goto err_fopen;

    /* Reverse the list of ACLs since that's the order they were in */
    for (newacl = 0, acl = acc->acc_acls; acl != 0; acl = acc->acc_acls) {
	acc->acc_acls = acl->acl_next;
	acl->acl_next = newacl;
	newacl = acl;
    }
    acc->acc_acls = newacl;

    /* Scan list of ACLs in the ACL context */
    for (acl = acc->acc_acls; acl != 0; acl = acl->acl_next) {

	fprintf(ofp, "ACL %s (", acl->acl_sym.sym_name);

	/* Output ACL rights list */
	list = aclGetRights(acl);
	for (lp = list; *lp != 0; ++lp) {
	    fprintf(ofp, "%s%s", *lp, (lp[1] == 0) ? ")" : ", ");
	}
	free_strlist(list);
	fprintf(ofp, " {\n");

	/* Scan ACL directives */
	for (acd = acl->acl_dirf, ndir = 1;
	     acd != 0; acd = acd->acd_next, ++ndir) {

	    action = acd->acd_action & ACD_ACTION;

	    switch (action) {
	      case ACD_ALLOW:		/* allow access */
	      case ACD_DENY:		/* deny access */
		fprintf(ofp, "\t%s %s ",
			((acd->acd_flags & ACD_ALWAYS) ? "Always" : "Default"),
			((action == ACD_ALLOW) ? "allow" : "deny"));

		for (ncls = 1, clp = acd->acd_cl;
		     clp != 0; clp = clp->cl_next, ++ncls) {

		    if (clp->cl_user == 0) {
			fprintf(ofp, "anyone");
		    }
		    else if (clp->cl_user->us_flags & ACL_USALL) {
			fprintf(ofp, "all");
		    }
		    else {
			/* Get list of user/group names */
			list = aclGetUsers(acl, ndir, ncls);
			if (list[1] == 0) {
			    fprintf(ofp, "%s", aclSafeIdent(list[0]));
			}
			else {
			    fprintf(ofp, "(");
			    for (lp = list; *lp != 0; ++lp) {
				fprintf(ofp, "%s%s",
					aclSafeIdent(*lp),
					(lp[1] == 0) ? ")" : ", ");
			    }
			}
			free_strlist(list);
		    }

		    /* Got any host specifications? */
		    if ((clp != 0) && (clp->cl_host != 0)) {
			list = aclGetHosts(acl, ndir, ncls);
			if (list[1] == 0) {
			    fprintf(ofp, " at %s", list[0]);
			}
			else {
			    fprintf(ofp, " at (");
			    for (lp = list; *lp != 0; ++lp) {
				fprintf(ofp, "%s%s",
					*lp, (lp[1] == 0) ? ")" : ", ");
			    }
			}
			free_strlist(list);
		    }

		    if (clp->cl_next != 0) {
			fprintf(ofp, ", ");
		    }
		}
		fprintf(ofp, ";\n");
		break;

	      case ACD_AUTH:		/* specify authentication realm */
		fprintf(ofp, "\t%s authenticate",
			(acd->acd_flags & ACD_ALWAYS) ? "Always" : "Default");
		if (acd->acd_auth.au_realm != 0) {
		    fprintf(ofp, " in {\n");
		    rlm = &acd->acd_auth.au_realm->rs_realm;
		    if (rlm->rlm_dbname != 0) {
			fprintf(ofp, "\t\tDatabase \"%s\";\n",
				rlm->rlm_dbname);
		    }
		    if (rlm->rlm_ameth != 0) {
			fprintf(ofp, "\t\tMethod %s;\n",
				(rlm->rlm_ameth == AUTH_METHOD_BASIC) ? "basic"
								      : "SSL");
		    }
		    if (rlm->rlm_prompt != 0) {
			fprintf(ofp, "\t\tPrompt \"%s\";\n", rlm->rlm_prompt);
		    }
		    fprintf(ofp, "\t}");
		}
		fprintf(ofp, ";\n");
		break;

	      case ACD_EXEC:		/* execute (conditionally) */
		fprintf(ofp, "\t%s execute",
			(acd->acd_flags & ACD_ALWAYS) ? "Always" : "Default");
		if (acd->acd_flags & (ACD_EXALLOW|ACD_EXDENY|ACD_EXAUTH)) {
		    fprintf(ofp, " if ");
		    if (acd->acd_flags & ACD_EXALLOW) {
			fprintf(ofp, "allow");
		    }
		    if (acd->acd_flags & ACD_EXDENY) {
			if (acd->acd_flags & ACD_EXALLOW) {
			    fprintf(ofp, ", ");
			}
			fprintf(ofp, "deny");
		    }
		    if (acd->acd_flags & ACD_EXAUTH) {
			if (acd->acd_flags & (ACD_EXALLOW|ACD_EXDENY)) {
			    fprintf(ofp, ", ");
			}
			fprintf(ofp, "authenticate");
		    }
		}
		fprintf(ofp, ";\n");
		break;

	      default:
		break;
	    }
	}
	fprintf(ofp, "}\n\n");
    }

    fclose(ofp);

    return 0;

  err_fopen:
    return ACLERROPEN;
}

/*
 * Description (aclStringGet)
 *
 *	This function is the stream read function designated by
 *	aclStringOpen() to read the string associated with the LEX
 *	stream it creates.  It always returns zero to indicate EOF.
 *
 * Arguments:
 *
 *	lst			- pointer to LEX stream structure
 *
 * Returns:
 *
 *	A return value of zero to indicate end-of-file.
 */

NSAPI_PUBLIC int aclStringGet(LEXStream_t * lst)
{
    lst->lst_len = 0;
    return 0;
}

/*
 * Description (aclStringOpen)
 *
 *	This function creates an ACLFile_t structure that is used
 *	for parsing a string that is already in memory.  This
 *	includes a handle for a LEX data stream for the string.
 *
 * Arguments:
 *
 *	errp			- error frame list pointer (may be null)
 *	slen			- string length
 *	sptr			- string pointer
 *	flags			- bit flags (unused - must be zero)
 *	pacf			- pointer to returned ACLFile_t pointer
 *
 * Returns:
 *
 *	The return value is zero if the ACLFile_t is created successfully,
 *	and a pointer to the ACLFile_t is returned in the location
 *	specified by 'pacf'.  Otherwise a negative error code
 *	(ACLERRxxxx - see aclerror.h) is returned, and an error frame
 *	will be generated if an error list is provided.
 */

NSAPI_PUBLIC int aclStringOpen(NSErr_t * errp, int slen, char * sptr,
			       int flags, ACLFile_t **pacf)
{
    ACLFile_t * acf;		/* pointer to ACL file structure */
    int rv;			/* return value */
    int eid;			/* error identifier */

    *pacf = 0;

    /* Allocate the ACLFile_t structure */
    acf = (ACLFile_t *)MALLOC(sizeof(ACLFile_t));
    if (acf == 0) goto err_nomem1;

    memset((void *)acf, 0, sizeof(ACLFile_t));
    acf->acf_filename = STRDUP("form input");
    acf->acf_fd = SYS_ERROR_FD;
    acf->acf_lineno = 1;
    acf->acf_flags = flags;

    /* Create a LEX token object */
    rv = lex_token_new((pool_handle_t *)0, 32, 8, &acf->acf_token);
    if (rv < 0) goto err_nomem2;

    /* Create a LEX stream for the file */
    acf->acf_lst = lex_stream_create(aclStringGet, 0, sptr, slen);
    if (acf->acf_lst == 0) goto err_nomem3;

    /* Set up initial pointer and length for stream data */
    acf->acf_lst->lst_len = acf->acf_lst->lst_buflen;
    acf->acf_lst->lst_cp = acf->acf_lst->lst_buf;

    *pacf = acf;
    return 0;

  err_nomem1:				/* MALLOC of ACLFile_t failed */
    rv = ACLERRNOMEM;
    eid = ACLERR5000;
    goto err_mem;

  err_nomem2:				/* lex_token_new() failed */
    rv = ACLERRNOMEM;
    eid = ACLERR5020;
    goto err_mem;

  err_nomem3:				/* lex_stream_create() failed */
    rv = ACLERRNOMEM;
    eid = ACLERR5040;

  err_mem:
    nserrGenerate(errp, rv, eid, ACL_Program, 0);
    goto punt;

  punt:
    return rv;
}

NSAPI_PUBLIC int aclCheckUsers(NSErr_t * errp, char * dbpath, char * usernames,
			       char * groupnames, char ***uglist,
			       char ***badulist, char ***badglist)
{
    ACLFile_t * acf = 0;
    char **rlist = 0;
    UserSpec_t * usp = 0;
    char **badlist = 0;
    void * token;
    char * tokenstr;
    void * ugoptr;
    struct {
	char * names;
	char ***badlist;
	int aifflag;
    } ug[2];
    Realm_t realm;
    int i;
    int nnames = 0;
    int badsize;
    int badcnt;
    int gotbad = 0;
    int irv;
    int rv;

    /* Initialize the ACL parser */
    rv = aclParseInit();

    /* Set up Realm_t for user/group lookup */
    realm.rlm_dbname = dbpath;
    realm.rlm_aif = &NSADB_AuthIF;
    realm.rlm_authdb = 0;

    /* Set up for loop below */
    ug[0].names = usernames;
    ug[0].badlist = badulist;
    ug[0].aifflag = AIF_USER;
    ug[1].names = groupnames;
    ug[1].badlist = badglist;
    ug[1].aifflag = AIF_GROUP;

    for (i = 0; i < 2; ++i) {

	badlist = 0;
	badsize = 0;
	badcnt = 0;

	/* Is there a list of user names to process? */
	if (ug[i].names && *ug[i].names) {

	    /* Yes, set up to parse list */
	    rv = aclStringOpen(errp,
			       strlen(ug[i].names), ug[i].names, 0, &acf);
	    if (rv < 0) goto punt;

	    /* Get token handle */
	    token = acf->acf_token;

	    for (;;) {

		/* Get token from name list */
		rv = aclGetToken(errp, acf, 0);

		/* Looking for a user or group identifier */
		if ((rv == TOKEN_IDENT) || (rv == TOKEN_STRING)) {

		    /* Check for reserved words */
		    tokenstr = lex_token(token);

		    /* Check for keywords not allowed as a user name */
		    if (strcasecmp(tokenstr, KEYWORD_AT) &&
			strcasecmp(tokenstr, KEYWORD_ANY) &&
			strcasecmp(tokenstr, KEYWORD_ALL)) {

			/* None found.  Check for valid user/group name */
			if (realm.rlm_authdb == 0) {
			    irv = nsadbOpen(errp,
					    dbpath, 0, &realm.rlm_authdb);
			    if (irv < 0) goto err_open;
			}

			irv = nsadbFindByName(errp, realm.rlm_authdb,
					      tokenstr, ug[i].aifflag,
					      &ugoptr);

			/*
			 * Create a UserSpec_t structure if we haven't
			 * got one yet.
			 */
			if (usp == 0) {
			    usp = aclUserSpecCreate();
			    if (usp == 0) goto err_nomem1;
			}

			if (irv == ug[i].aifflag) {

			    if (irv == AIF_USER) {

				/* Add the user id to the user list */
				irv = usiInsert(&usp->us_user.uu_user,
						((UserObj_t *)ugoptr)->uo_uid);
			    }
			    else if (irv == AIF_GROUP) {

				/* Add group id to an_groups list. */
				irv = usiInsert(&usp->us_user.uu_group,
						((GroupObj_t *)ugoptr)->go_gid);
			    }

			    /* Examine the result of the insert operation */
			    if (irv >= 0) continue;
			}
		    }
		}

		if (rv == TOKEN_COMMA) continue;
		if (rv == TOKEN_EOF) break;

		/* Get everything to the next comma, whitespace, or the end */
		for (;;) {
		    rv = lex_scan_to(acf->acf_lst,
				     aclChTab, (CCM_SPECIAL|CCM_WS), token);
		    if ((rv == ',') ||
			lex_class_check(aclChTab, rv, CCM_WS) ||
			(rv == 0)) break;
		    (void)lex_token_append(token, 1, acf->acf_lst->lst_cp);
		    (void)lex_next_char(acf->acf_lst, aclChTab,
					(CCM_SPECIAL|CCM_WS));
		}

		tokenstr = lex_token(token);

		gotbad += 1;

		/* Not a valid name.  Add to list if specified */
		if (ug[i].badlist != 0) {
		    if (badlist == 0) {
			/* Create the list */
			badlist = (char **)MALLOC(4*sizeof(char *));
			badlist[0] = 0;
			badcnt = 1;
			badsize = 4;
			*ug[i].badlist = badlist;
		    }
		    else if (badcnt >= badsize) {
			/* Grow the list */
			badsize += 4;
			badlist = (char **)REALLOC(badlist,
						   badsize*sizeof(char *));
			*ug[i].badlist = badlist;
		    }

		    /* Add current token string to list */
		    badlist[badcnt-1] = STRDUP(tokenstr);
		    badlist[badcnt] = 0;
		    ++badcnt;
		}
	    }

	    if (rv < 0) goto err_parse;
	    if (rv != TOKEN_EOF) goto err_eof;

	    aclFileClose(acf, 0);
	    acf = 0;
	}
    }

    /* Convert lists in UserSpec_t back to name list */
    if (uglist != 0) {

	rlist = 0;

	if (usp != 0) {

	    nnames = UILCOUNT(&usp->us_user.uu_user)
			    + UILCOUNT(&usp->us_user.uu_group);

	    if (nnames > 0) {

		/* Allocate the name list to be returned */
		rlist = (char **)MALLOC((nnames+1)*sizeof(char *));
		if (rlist != 0) {
		    memset((void *)rlist, 0, (nnames+1)*sizeof(char *));

		    /* Convert user ids to names */
		    irv = aclIdsToNames(rlist,
					&usp->us_user.uu_user, 1, &realm);

		    /* Convert group ids to names */
		    irv += aclIdsToNames(rlist+irv,
					 &usp->us_user.uu_group, 0, &realm);

		}
	    }
	}

	/* Return the resulting list of users and groups */
	*uglist = rlist;
    }

    rv = gotbad;

  punt:

    if (realm.rlm_authdb != 0) {

	/* Close the authentication database */
	nsadbClose(realm.rlm_authdb, 0);
    }

    if (usp != 0) {
	aclUserSpecDestroy(usp);
    }

    if (acf != 0) {
	aclFileClose(acf, 0);
    }

    return rv;

  err_open:
    rv = ACLERROPEN;
    goto punt;

  err_nomem1:
    rv = ACLERRNOMEM;
    goto punt;

  err_parse:
    rv = ACLERRSYNTAX;
    goto punt;

  err_eof:
    rv = ACLERRSYNTAX;
    goto punt;
}

NSAPI_PUBLIC int aclCheckHosts(NSErr_t * errp, int hexpand,
			       char * dnsspecs, char * ipspecs,
			       char ***hlist, char ***baddns, char ***badip)
{
    ACLFile_t * acf = 0;
    HostSpec_t * hsp = 0;		/* pointer to host specification */
    char **rlist = 0;
    char **badlist;
    void * token;
    char * tokenstr;
    char **p;				/* pointer to current alias pointer */
    char **alist;			/* pointer to alias list */
    int asize;				/* count of entries in alias list+1 */
    int badsize;
    int badcnt;
    int gotbad = 0;
    int ipcnt = 0;			/* count of IP filter entries */
    int dnscnt = 0;			/* count of DNS filter entries */
    int fqdn;				/* fully qualified domain name */
    IPAddr_t ipaddr;			/* IP address value */
    IPAddr_t netmask;			/* IP netmask value */
    int arv;				/* alternate result value */
    int rv;				/* result value */

    /* Initialize the ACL parser */
    rv = aclParseInit();

    /* Are there any DNS specifications? */
    if (dnsspecs != 0) {

	badlist = 0;

	/* Yes, set up to parse list */
	rv = aclStringOpen(errp, strlen(dnsspecs), dnsspecs, 0, &acf);
	if (rv < 0) goto punt;

	/* Get token handle */
	token = acf->acf_token;

	for (;;) {

	    /* Get next token */
	    rv = aclGetToken(errp, acf, 0);
	    if (rv < 0) goto punt;

	    if ((rv == TOKEN_STAR) || (rv == TOKEN_IDENT)) {

		fqdn = 0;

		if (rv == TOKEN_IDENT) {
		    fqdn = 1;
		    tokenstr = lex_token(token);

		    if (!strcasecmp(tokenstr, KEYWORD_HOSTS)) goto bad_dns;
		}

		/* This looks like the start of an auth-dns-spec */
		arv = aclGetDNSString(errp, acf);
		tokenstr = lex_token(token);

		if (arv >= 0) {

		    /* If the DNS spec begins with "*.", strip the "*" */
		    if (tokenstr && (tokenstr[0] == '*')) {

			if (tokenstr[1] == '.') {
			    tokenstr += 1;
			}

			/* Add to HostSpec_t and continue loop if successful */
			arv = aclAuthDNSAdd(&hsp, tokenstr, fqdn);
		    }
		    else {

			alist = 0;
			asize = 0;
			arv = aclDNSPutHost(tokenstr, 1, hexpand,
					    &alist, &asize);
			if (arv == 0) goto bad_dns;

			/* For each alias ... */
			for (p = alist; *p; ++p) {

			    /* Add to HostSpec_t */
			    arv = aclAuthDNSAdd(&hsp, *p, fqdn);
			}

			for (p = alist; *p; ++p) {
			    FREE(*p);
			}
			FREE(alist);
		    }

		    if (arv >= 0) continue;
		}
	    }

	    if (rv == TOKEN_COMMA) continue;
	    if (rv == TOKEN_EOF) break;

	  bad_dns:

	    /* Get everything to the next comma or the end */
	    for (;;) {
		rv = lex_scan_to(acf->acf_lst, aclChTab, CCM_SPECIAL, token);
		if ((rv == ',') || (rv == 0)) break;
		(void)lex_token_append(token, 1, acf->acf_lst->lst_cp);
		(void)lex_next_char(acf->acf_lst, aclChTab, CCM_SPECIAL);
	    }

	    tokenstr = lex_token(token);

	    /* Not a valid dnsspec.  Add to list if specified */
	    gotbad += 1;

	    if (baddns != 0) {
		if (badlist == 0) {
		    /* Create the list */
		    badlist = (char **)MALLOC(4*sizeof(char *));
		    badlist[0] = 0;
		    badcnt = 1;
		    badsize = 4;
		    *baddns = badlist;
		}
		else if (badcnt >= badsize) {
		    /* Grow the list */
		    badsize += 4;
		    badlist = (char **)REALLOC(badlist,
					       badsize*sizeof(char *));
		    *baddns = badlist;
		}

		/* Add current token string to list */
		badlist[badcnt-1] = STRDUP(tokenstr);
		badlist[badcnt] = 0;
		++badcnt;
	    }
	}

	aclFileClose(acf, 0);
    }

    if (ipspecs != 0) {

	badlist = 0;

	/* Yes, set up to parse list */
	rv = aclStringOpen(errp, strlen(ipspecs), ipspecs, 0, &acf);
	if (rv < 0) goto punt;

	/* Get token handle */
	token = acf->acf_token;

	for (;;) {

	    /* Get next token */
	    rv = aclGetToken(errp, acf, 0);
	    if (rv < 0) goto punt;

	    /* Does this look like an auth-ip-spec? */
	    if (rv == TOKEN_NUMBER) {

		/* Yes, go parse it */
		rv = aclGetIPAddr(errp, acf, &ipaddr, &netmask);
		if (rv >= 0) {

		    /* Add the IP address and netmask to the HostSpec_t */
		    arv = aclAuthIPAdd(&hsp, ipaddr, netmask);
		    if (arv >= 0) {

			if (rv == TOKEN_COMMA) continue;
		    }
		}
	    }

	    if (rv == TOKEN_EOF) break;
	    if (rv == TOKEN_COMMA) continue;

	    /* Get everything to the next comma or the end */
	    for (;;) {
		rv = lex_scan_to(acf->acf_lst, aclChTab, CCM_SPECIAL, token);
		if ((rv == ',') || (rv == 0)) break;
		(void)lex_token_append(token, 1, acf->acf_lst->lst_cp);
		(void)lex_next_char(acf->acf_lst, aclChTab, CCM_SPECIAL);
	    }

	    tokenstr = lex_token(token);

	    /* Not a valid ipspec.  Add to list if specified */
	    gotbad += 1;

	    if (badip != 0) {
		if (badlist == 0) {
		    /* Create the list */
		    badlist = (char **)MALLOC(4*sizeof(char *));
		    badlist[0] = 0;
		    badcnt = 1;
		    badsize = 4;
		    *badip = badlist;
		}
		else if (badcnt >= badsize) {
		    /* Grow the list */
		    badsize += 4;
		    badlist = (char **)REALLOC(badlist,
					       badsize*sizeof(char *));
		    *badip = badlist;
		}

		/* Add current token string to list */
		badlist[badcnt-1] = STRDUP(tokenstr);
		badlist[badcnt] = 0;
		++badcnt;
	    }
	}

	aclFileClose(acf, 0);
    }

    /* Is there a host specification? */
    if (hsp != 0) {

	/* Get count of IP filter host/network specifications */
	ipcnt = aclIPFilterStrings(0, &hsp->hs_host.inh_ipf);

	/* Get count of DNS filter specifications */
	dnscnt = aclDNSFilterStrings(0, &hsp->hs_host.inh_dnf);

	/* Allocate the array of strings */
	rlist = new_strlist(ipcnt + dnscnt);

	/* Put IP filter specifications in list */
	rv = aclIPFilterStrings(rlist, &hsp->hs_host.inh_ipf);

	/* Put DNS filter specifications in list */
	rv = aclDNSFilterStrings(rlist + rv, &hsp->hs_host.inh_dnf);

	aclHostSpecDestroy(hsp);
    }

    if (hlist != 0) *hlist = rlist;

    return gotbad;

  punt:
    return rv;
}
