/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
#include "base/systems.h"
#include "netsite.h"
#include "base/shexp.h"
#include "libadmin/libadmin.h"
#include "nsumgmt.h"
#include "nsgmgmt.h"
#define __PRIVATE_NSADB
#include "nsamgmt.h"
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <ctype.h>

#define CMD_MKGROUP	0
#define CMD_RMGROUP	1
#define CMD_CHGROUP	2
#define CMD_SHGROUP	3

static char * progname;

static void sigint()
{
    static int intcnt = 0;

    if (++intcnt > 2) {
	fprintf(stderr, "%s: interrupted - database may be corrupted!\n",
		progname);
	exit(1);
    }
}

static void printUser(AuthDB_t * authdb, USI_t uid, int numeric, int * column)
{
    char * outp;
    int len;
    int rv;
    char idstring[16];

    if (numeric) {
	sprintf(idstring, "%d", uid);
	outp = idstring;
    }
    else {

	rv = nsadbIdToName(0, authdb, uid, AIF_USER, &outp);
	if (rv != 0) {
	    sprintf(idstring, "#%d", uid);
	    outp = idstring;
	}
    }

    len = strlen(outp) + 1;
    *column += len;

    if (*column > 80) {
	fprintf(stdout, "\n                  ");
	*column = 19 + len;
    }

    fprintf(stdout, "%s", outp);
}

static void printGroup(AuthDB_t * authdb, USI_t gid, int numeric, int * column)
{
    char * outp;
    int len;
    int rv;
    char idstring[16];

    if (numeric) {
	sprintf(idstring, "%d", gid);
	outp = idstring;
    }
    else {

	rv = nsadbIdToName(0, authdb, gid, AIF_GROUP, &outp);
	if (rv != 0) {
	    sprintf(idstring, "#%d", gid);
	    outp = idstring;
	}
    }

    len = strlen(outp) + 1;
    *column += len;

    if (*column > 80) {
	fprintf(stdout, "\n                  ");
	*column = 19 + len;
    }

    fprintf(stdout, "%s", outp);
}

static int listGroup(NSErr_t * errp,
		    void * authdb, void * parg, GroupObj_t * goptr)
{
    char * exp = (char *)parg;
    USI_t * list;
    int cnt;
    int i;
    int column;

    if (!shexp_match(goptr->go_name, exp)) {
	fprintf(stdout, "%s:%s:", goptr->go_name,
		(goptr->go_desc) ? (char *)(goptr->go_desc) : "");

	list = UILLIST(&goptr->go_groups);
	cnt = UILCOUNT(&goptr->go_groups);
	for (i = 0; i < cnt; ) {

	    column = 0;
	    printGroup(authdb, *list, 0, &column);
	    if (++i < cnt) {
		fprintf(stdout, ",");
	    }
	    ++list;
	}

	fprintf(stdout, ":");

	list = UILLIST(&goptr->go_users);
	cnt = UILCOUNT(&goptr->go_users);
	for (i = 0; i < cnt; ) {

	    column = 0;
	    printUser(authdb, *list, 0, &column);
	    if (++i < cnt) {
		fprintf(stdout, ",");
	    }
	    ++list;
	}

	fprintf(stdout, ":\n");
    }
    return 0;
}

static int showMany(NSErr_t * errp, void * authdb, char * exp)
{
    int rv;

    rv = nsadbEnumerateGroups(errp, authdb, (void *)exp, listGroup);

    return rv;
}

int main(int argc, char * argv[])
{
    char * dbname;
    void * padb;
    UserObj_t * uoptr;
    GroupObj_t * goptr;
    GroupObj_t * cgoptr;
    AuthDB_t * authdb;
    USI_t * gidlist;
    NTS_t descstr = 0;
    NTS_t name = 0;
    NTS_t userlist = 0;
    NTS_t grouplist = 0;
    char * gname;
    char * cp;
    char * member;
    USI_t gid = 0;
    int cmd;
    int opt;
    int i;
    int numeric;
    int add, remove;
    int user, group;
    int newdesc;
    int column;
    int rv;
    NSErr_t err = NSERRINIT;
    static char * cmdopt[] = {
	"d:g:hu:",			/* mkgroup options */
	"hi:",				/* rmgroup options */
	"ad:i:g:hru:",			/* chgroup options */
	"Nhi:"				/* shgroup options */
    };

    extern char * optarg;
    extern int optind, opterr;

    add = 0;
    remove = 0;
    user = 1;
    group = 0;
    numeric = 0;
    newdesc = 0;
    descstr = STRDUP("");

    NSPR_INIT("mkgroup");

    opterr = 0;
    optind = 1;

    progname = getprogname(argv[0]);

    if (!strcmp(progname, "mkgroup")) cmd = CMD_MKGROUP;
    else if (!strcmp(progname, "rmgroup")) cmd = CMD_RMGROUP;
    else if (!strcmp(progname, "chgroup")) cmd = CMD_CHGROUP;
    else if (!strcmp(progname, "shgroup")) cmd = CMD_SHGROUP;
    else {
	fprintf(stderr,
	  "%s: must be invoked as mkgroup, rmgroup, chgroup, or shgroup\n",
		progname);
	exit(1);
    }

    for (;;) {

	opt = getopt(argc, argv, cmdopt[cmd]);
	if (opt == EOF) break;
	switch (opt) {
	  case 'N':			/* display numeric user/group ids */
	    numeric = 1;
	    break;

	  case 'a':			/* add to group flag */
	    add = 1;
	    break;

	  case 'd':			/* group description */
	    if (descstr) {
		FREE(descstr);
	    }
	    descstr = optarg;
	    newdesc = 1;
	    break;

	  case 'g':			/* group list */
	    grouplist = optarg;
	    if (cmd == CMD_MKGROUP) add = 1;
	    break;

	  case 'h':			/* help */
	    goto do_usage;

	  case 'i':			/* group id */
	    gid = atoi(optarg);
	    break;

	  case 'r':			/* remove from group flag */
	    remove = 1;
	    break;

	  case 'u':			/* user list */
	    userlist = optarg;
	    if (cmd == CMD_MKGROUP) add = 1;
	    break;

	  default:
	    fprintf(stderr, "%s: unknown option - %s\n",
		    progname, argv[optind-1]);
	    goto do_usage;
	}
    }

    if (add & remove) {
	fprintf(stderr, "%s: use -a to add members, -r to remove, not both\n",
		progname);
	goto do_usage;
    }

    if ((userlist || grouplist) && !(add|remove)) {
	fprintf(stderr, "%s: specify -a to add members, or -r to remove\n",
		progname);
	goto do_usage;
    }

    if (gid == 0) {
	if ((argc - optind) < 1) goto do_usage;
	name = argv[optind++];
    }
    else {
	name = 0;
    }

    if ((argc - optind) < 1) {
	dbname = getenv("NSAUTHDB");
	if (dbname == 0) goto do_usage;
    }
    else dbname = argv[optind++];

    if (optind < argc) {
	fprintf(stderr, "%s: arguments starting with \"%s\" are extraneous\n",
		progname, argv[optind]);
	goto do_usage;
    }

    if (cmd == CMD_MKGROUP) {
	goptr = groupCreate(name, descstr);
	if (goptr == 0) {
	    fprintf(stderr, "%s: groupCreate() failed\n", progname);
	    exit(1);
	}
    }

    signal(SIGINT, sigint);

    rv = nsadbOpen(&err, dbname, (cmd == CMD_MKGROUP) ? AIF_CREATE : 0, &padb);
    if (rv < 0) {
	fprintf(stderr, "%s: error %d from nsadbOpen()\n", progname, rv);
	exit(1);
    }

    rv = 0;
    authdb = (AuthDB_t *)padb;

    if (cmd == CMD_MKGROUP) {
	rv = nsadbCreateGroup(&err, authdb, goptr);
	if (rv) {
	    fprintf(stderr, "%s: error %d from groupStore()\n", progname, rv);
	    rv = 1;
	}
	else {
	    fprintf(stderr, "%s: group \"%s\" created successfully.\n",
		    progname, goptr->go_name);
	    rv = 0;
	}
    }
    else {
	if (gid) {
	    rv = nsadbIdToName(&err, authdb, gid, AIF_GROUP, &gname);
	    if (rv < 0) {
		fprintf(stderr, "%s: error %d from nsadbIdToName()\n",
			progname, rv);
		rv = 1;
		goto close;
	    }
	    rv = nsadbFindByName(&err,
				 authdb, gname, AIF_GROUP, (void **)&goptr);
	    if (rv != AIF_GROUP) {
		if (rv == AIF_NONE) {
		    fprintf(stderr,
			"%s: group %s not found in the \"%s\" database.\n",
			    progname, gname, dbname);
		}
		else {
		    fprintf(stderr, "%s: error %d from nsadbFindByName()\n",
			    progname, rv);
		}
		rv = 1;
		goto close;
	    }
	    name = goptr->go_name;
	    rv = 0;
	}
	else {

	    /* Handle name as a shexp for shuser */
	    if (cmd == CMD_SHGROUP) {
		if (shexp_valid(name) == VALID_SXP) {
		    showMany(&err, authdb, name);
		    goto close;
		}
	    }

	    rv = nsadbFindByName(&err,
				 authdb, name, AIF_GROUP, (void **)&goptr);
	    if (rv != AIF_GROUP) {
		if (rv == AIF_NONE) {
		    fprintf(stderr,
			"%s: group \"%s\" not found in the \"%s\" database.\n",
			    progname, name, dbname);
		}
		else {
		    fprintf(stderr, "%s: error %d from nsadbFindByName()\n",
			    progname, rv);
		}
		rv = 1;
		goto close;
	    }
	    rv = 0;
	}

	if (cmd == CMD_RMGROUP) {
	    rv = nsadbRemoveGroup(&err, authdb, name);
	    if (rv) {
		fprintf(stderr, "%s: error %d from nsadbRemoveGroup()\n",
			progname, rv);
		rv = 1;
	    }
	    else {
		fprintf(stderr,
			"%s: group \"%s\" successfully removed.\n",
			progname, name);
		rv = 0;
	    }
	}
	else if (cmd == CMD_SHGROUP) {
	    fprintf(stdout, "Group name:  %s\n", goptr->go_name);
	    fprintf(stdout, "Group id:    %d\n", goptr->go_gid);
	    fprintf(stdout, "Flags:       0x%x\n", goptr->go_flags);
	    if (goptr->go_desc) {
		fprintf(stdout, "Description: %s\n", goptr->go_desc);
	    }
	    
	    if (UILCOUNT(&goptr->go_pgroups) > 0) {
		fprintf(stdout, "Parent groups:    ");
		gidlist = UILLIST(&goptr->go_pgroups);
		column = 19;
		for (i = 0; i < UILCOUNT(&goptr->go_pgroups); ++i) {
		    printGroup(authdb, gidlist[i], numeric, &column);
		    fprintf(stdout, " ");
		}
		fprintf(stdout, "\n");
	    }
	    if (UILCOUNT(&goptr->go_groups) > 0) {
		fprintf(stdout, "Member groups:    ");
		gidlist = UILLIST(&goptr->go_groups);
		column = 19;
		for (i = 0; i < UILCOUNT(&goptr->go_groups); ++i) {
		    printGroup(authdb, gidlist[i], numeric, &column);
		    fprintf(stdout, " ");
		}
		fprintf(stdout, "\n");
	    }
	    if (UILCOUNT(&goptr->go_users) > 0) {
		fprintf(stdout, "Member users:     ");
		gidlist = UILLIST(&goptr->go_users);
		column = 19;
		for (i = 0; i < UILCOUNT(&goptr->go_users); ++i) {
		    printUser(authdb, gidlist[i], numeric, &column);
		    fprintf(stdout, " ");
		}
		fprintf(stdout, "\n");
	    }
	    rv = 0;
	}
	else if (cmd == CMD_CHGROUP) {

	    if (newdesc && descstr) {
		if (goptr->go_desc) {
		    FREE(goptr->go_desc);
		}
		goptr->go_desc = STRDUP(descstr);

		rv = nsadbModifyGroup(&err, authdb, goptr);
		if (rv) {
		    fprintf(stderr, "%s: error %d from groupStore()\n",
			    progname, rv);
		    rv = 1;
		}
		else {
		    fprintf(stderr,
		      "%s: group \"%s\" description modified successfully.\n",
			    progname, goptr->go_name);
		}
	    }
	}
    }

    if ((cmd == CMD_MKGROUP) || (cmd == CMD_CHGROUP)) {
	while ((rv == 0) && (grouplist || userlist)) {
	    if (grouplist) {
		member = grouplist;
		cp = strchr(grouplist, ',');
		if (cp) {
		    *cp++ = 0;
		    grouplist = cp;
		}
		else {
		    grouplist = 0;
		}
		rv = nsadbFindByName(&err, authdb,
				     member, AIF_GROUP, (void **)&cgoptr);
		if ((rv == AIF_GROUP) && (cgoptr != 0)) {
		    if (add) {
			rv = nsadbAddGroupToGroup(&err, authdb, goptr, cgoptr);
		    }
		    else if (remove) {
			rv = nsadbRemGroupFromGroup(&err, authdb,
						    goptr, cgoptr);
		    }
		    if (rv == 0) {
			fprintf(stderr,
			  "%s: group \"%s\" successfully %s group \"%s\".\n",
				progname, cgoptr->go_name,
				((add) ? "added to" : "removed from"),
				goptr->go_name);
		    }
		    else {
			fprintf(stderr,
			   "%s: failed to %s group \"%s\" %s group \"%s\".\n",
				progname, ((add) ? "add" : "remove"),
				cgoptr->go_name, ((add) ? "to" : "from"),
				goptr->go_name);
		    }
		    groupFree(cgoptr);
		    cgoptr = 0;
		}
		else {
		    if (rv == AIF_NONE) {
			fprintf(stderr,
				"%s: group \"%s\" not found - ignored.\n",
				progname, member);
		    }
		    else {
			fprintf(stderr,
			  "%s: error %d from nsadbFindByName() for %s\n",
				progname, rv, member);
			rv = 1;
			goto close;
		    }
		}
	    }
	    else if (userlist) {

		member = userlist;
		cp = strchr(userlist, ',');
		if (cp) {
		    *cp++ = 0;
		    userlist = cp;
		}
		else {
		    userlist = 0;
		}
		rv = nsadbFindByName(&err, authdb,
				     member, AIF_USER, (void **)&uoptr);
		if ((rv == AIF_USER) && (uoptr != 0)) {
		    if (add) {
			rv = nsadbAddUserToGroup(&err, authdb, goptr, uoptr);
		    }
		    else if (remove) {
			rv = nsadbRemUserFromGroup(&err, authdb, goptr, uoptr);
		    }
		    if (rv == 0) {
			fprintf(stderr,
			  "%s: user \"%s\" successfully %s group \"%s\".\n",
				progname, uoptr->uo_name,
				((add) ? "added to" : "removed from"),
				goptr->go_name);
		    }
		    else {
			fprintf(stderr,
			   "%s: failed to %s user \"%s\" %s group \"%s\".\n",
				progname, ((add) ? "add" : "remove"),
				uoptr->uo_name, ((add) ? "to" : "from"),
				goptr->go_name);
		    }
		    userFree(uoptr);
		    uoptr = 0;
		}
		else {
		    fprintf(stderr, "%s: user \"%s\" not found - ignored.\n",
			    progname, member);
		}
	    }
	}
	if (rv) {
	    fprintf(stderr, "%s: error code %d from attempt to modify group\n",
		    progname, rv);
	    rv = 1;
	}
    }

  close:
    nsadbClose(authdb, 0);
    exit(rv);

  do_usage:
    if (cmd == CMD_MKGROUP) {
	fprintf(stderr,
		"Usage: %s [options] group-name [auth-db-dir]\n", progname);
    }
    else {
	fprintf(stderr,
		"Usage: %s [options] [group-name] [auth-db-dir]\n",
		progname);
    }
    fprintf(stderr, "Options:\n");
    if ((cmd == CMD_MKGROUP) || (cmd == CMD_CHGROUP)) {
	if (cmd == CMD_CHGROUP) {
	    fprintf(stderr,
		    "    -a                          add members to group\n");
	    fprintf(stderr,
		    "    -r                          remove members from group\n");
	}
	fprintf(stderr,
		"    -d group-description        set group description\n");
	fprintf(stderr,
		"    -g group-list               specify list of groups\n");
	fprintf(stderr,
		"    -u user-list                specify list of users\n");
    }
    if (cmd != CMD_MKGROUP) {
	fprintf(stderr,
		"    -i group-id                 group id instead of group name\n");
    }
    if (cmd == CMD_SHGROUP) {
	fprintf(stderr,
		"    -N                          display numeric user/group ids\n");
    }

    dbname = getenv("NSAUTHDB");
    if ((dbname == 0) || (*dbname == 0)) {
	fprintf(stderr,
    "auth-db-dir can be specified using the NSAUTHDB environment variable.\n"
		);
    }
    exit(1);
}
