/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
#include "base/systems.h"
#include "netsite.h"
#include "base/shexp.h"
#include "libadmin/libadmin.h"
#include "nsumgmt.h"
#include "nsgmgmt.h"
#define __PRIVATE_NSADB
#include "nsamgmt.h"
#include "nsautherr.h"
#include "nsadb.h"
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <ctype.h>

extern char * getprogname(char * argv0);

#define SSTR(s) ((s) ? (char *)(s) : "(null)")

#define CMD_MKUSER	0
#define CMD_RMUSER	1
#define CMD_CHUSER	2
#define CMD_SHUSER	3

static char * progname;

static char errbuf[2048];

static void sigint()
{
    static int intcnt = 0;

    if (++intcnt > 2) {
	fprintf(stderr, "%s: interrupted - database may be corrupted!\n",
		progname);
	exit(1);
    }
}

static void printGroup(void * authdb, USI_t gid, int numeric, int * column)
{
    char * name;
    int len;
    int rv;
    char idstring[16];

    if (numeric) {
	sprintf(idstring, "%d", gid);
	name = idstring;
    }
    else {

	rv = nsadbIdToName(0, authdb, gid, AIF_GROUP, &name);
	if (rv < 0) {
	    sprintf(idstring, "#%d", gid);
	    name = idstring;
	}
    }

    len = strlen(name) + 1;
    *column += len;

    if (*column > 80) {
	fprintf(stdout, "\n           ");
	*column = 12 + len;
    }

    fprintf(stdout, "%s ", SSTR(name));
}

static int listUser(NSErr_t * errp,
		    void * authdb, void * parg, UserObj_t * uoptr)
{
    char * exp = (char *)parg;

    if (!shexp_match(uoptr->uo_name, exp)) {
	fprintf(stdout, "%s:%s:::%s\n", uoptr->uo_name,
		(uoptr->uo_pwd) ? (char *)(uoptr->uo_pwd) : "",
		(uoptr->uo_rname) ? (char *)(uoptr->uo_rname) : "");
    }
    return 0;
}

static int showMany(NSErr_t * errp, void * authdb, char * exp)
{
    int rv;

    rv = nsadbEnumerateUsers(errp, authdb, (void *)exp, listUser);

    return rv;
}

int main(int argc, char * argv[])
{
    char * dbname;
    void * authdb;
    UserObj_t * uoptr;
    USI_t * gidlist;
    NTS_t newname = 0;
    NTS_t name = 0;
    NTS_t pwd = 0;
    NTS_t rname = 0;
    USI_t uid = 0;
    int cmd;
    int opt;
    int i;
    int numeric;
    int column;
    int rv;
    NSErr_t err = NSERRINIT;
    static char * cmdopt[] = {
	"hn:p:",			/* mkuser options */
	"hu:",				/* rmuser options */
	"a:hn:p:u:",			/* chuser options */
	"Nhu:"				/* shuser options */
    };

    extern char * optarg;
    extern int optind, opterr;

    NSPR_INIT("mkuser");

    numeric = 0;

    opterr = 0;
    optind = 1;

    progname = getprogname(argv[0]);

    if (!strcmp(progname, "mkuser")) cmd = CMD_MKUSER;
    else if (!strcmp(progname, "rmuser")) cmd = CMD_RMUSER;
    else if (!strcmp(progname, "chuser")) cmd = CMD_CHUSER;
    else if (!strcmp(progname, "shuser")) cmd = CMD_SHUSER;
    else {
	fprintf(stderr,
		"%s: must be invoked as mkuser, rmuser, chuser, or shuser\n",
		progname);
	exit(1);
    }

    for (;;) {

	opt = getopt(argc, argv, cmdopt[cmd]);
	if (opt == EOF) break;
	switch (opt) {
	  case 'N':			/* display numeric group ids */
	    numeric = 1;
	    break;

	  case 'a':			/* new user account name */
	    newname = optarg;
	    break;

	  case 'h':			/* help */
	    goto do_usage;

	  case 'n':			/* real user name */
	    rname = optarg;
	    break;

	  case 'p':			/* encrypted password */
	    pwd = optarg;
	    break;

	  case 'u':			/* user id */
	    uid = atoi(optarg);
	    break;

	  default:
	    fprintf(stderr, "%s: unknown option - %s\n",
		    progname, argv[optind-1]);
	    goto do_usage;
	}
    }

    if (uid == 0) {
	if ((argc - optind) < 1) goto do_usage;
	name = argv[optind++];
    }
    else name = 0;

    if ((argc - optind) < 1) {
	dbname = getenv("NSAUTHDB");
	if (dbname == 0) goto do_usage;
    }
    else dbname = argv[optind++];

    if (optind < argc) {
	fprintf(stderr, "%s: arguments starting with \"%s\" are extraneous\n",
		progname, argv[optind]);
	goto do_usage;
    }

    if (cmd == CMD_MKUSER) {
	uoptr = userCreate(name, pwd, rname);
	if (uoptr == 0) {
	    fprintf(stderr, "%s: userCreate() failed\n", progname);
	    exit(1);
	}
    }

    signal(SIGINT, sigint);

    rv = nsadbOpen(&err,
		   dbname, (cmd == CMD_MKUSER) ? AIF_CREATE : 0, &authdb);
    if (rv < 0) {
	fprintf(stderr, "%s: error opening authentication database\n",
		progname);
	nsadbErrorFmt(&err, errbuf, sizeof(errbuf), 4);
	fprintf(stderr, "%s\n", errbuf);
	exit(1);
    }

    if (cmd == CMD_MKUSER) {
	rv = nsadbCreateUser(&err, authdb, uoptr);
	if (rv < 0) {
	    fprintf(stderr, "%s: failed to create user\n",
		    progname);
	    nsadbErrorFmt(&err, errbuf, sizeof(errbuf), 4);
	    fprintf(stderr, "%s\n", errbuf);
	    rv = 1;
	}
	else {
	    fprintf(stderr, "%s: user \"%s\" created successfully.\n",
		    progname, uoptr->uo_name);
	    rv = 0;
	}
    }
    else {
	if (uid) {

	    /* Convert the user id to a user name */
	    rv = nsadbIdToName(&err, authdb, uid, AIF_USER, (char **)&name);
	    if (rv < 0) {
		fprintf(stderr,
			"%s: user id %d not found in the \"%s\" database.\n",
			progname, uid, SSTR(dbname));
		rv = 1;
		goto close;
	    }

	    /* Retrieve a user object for the user name */
	    rv = nsadbFindByName(&err,
				 authdb, name, AIF_USER, (void **)&uoptr);
	    if (rv <= 0) {
		if (rv == AIF_NONE) {
		    fprintf(stderr,
		  "%s: user entry for %s not found in the \"%s\" database.\n",
			    progname, SSTR(name), SSTR(dbname));
		}
		else {
		    fprintf(stderr, "%s: user %s not found\n",
			    progname, SSTR(name));
		    nsadbErrorFmt(&err, errbuf, sizeof(errbuf), 4);
		    if (errbuf[0]) {
			fprintf(stderr, "%s\n", errbuf);
		    }
		}
		rv = 1;
		goto close;
	    }
	    name = uoptr->uo_name;
	}
	else {

	    /* Handle name as a shexp for shuser */
	    if (cmd == CMD_SHUSER) {
		if (shexp_valid(name) == VALID_SXP) {
		    showMany(&err, authdb, name);
		    goto close;
		}
	    }

	    rv = nsadbFindByName(&err,
				 authdb, name, AIF_USER, (void **)&uoptr);
	    if (rv <= 0) {
		if (rv == AIF_NONE) {
		    fprintf(stderr,
			"%s: user \"%s\" not found in the \"%s\" database.\n",
			    progname, SSTR(name), SSTR(dbname));
		}
		else {
		    fprintf(stderr, "%s: user %s not found\n",
			    progname, SSTR(name));
		    nsadbErrorFmt(&err, errbuf, sizeof(errbuf), 4);
		    if (errbuf[0]) {
			fprintf(stderr, "%s\n", errbuf);
		    }
		}
		rv = 1;
		goto close;
	    }
	}

	if (cmd == CMD_RMUSER) {
	    rv = nsadbRemoveUser(&err, authdb, name);
	    if (rv) {
		fprintf(stderr, "%s: error removing user %s\n",
			progname, name);
		nsadbErrorFmt(&err, errbuf, sizeof(errbuf), 4);
		if (errbuf[0]) {
		    fprintf(stderr, "%s\n", errbuf);
		}
		rv = 1;
	    }
	    else {
		fprintf(stderr,
			"%s: user \"%s\" successfully removed.\n",
			progname, SSTR(name));
		rv = 0;
	    }
	}
	else if (cmd == CMD_SHUSER) {
	    fprintf(stdout, "Username:  %s\n", SSTR(uoptr->uo_name));
	    fprintf(stdout, "Password:  %s\n", SSTR(uoptr->uo_pwd));
	    fprintf(stdout, "User id:   %d\n", uoptr->uo_uid);
	    fprintf(stdout, "Flags:     0x%x\n", uoptr->uo_flags);
	    fprintf(stdout, "Real name: %s\n", SSTR(uoptr->uo_rname));
	    if (UILCOUNT(&uoptr->uo_groups) > 0) {
		fprintf(stdout, "Groups:    ");
		gidlist = UILLIST(&uoptr->uo_groups);
		column = 12;
		for (i = 0; i < UILCOUNT(&uoptr->uo_groups); ++i) {
		    printGroup(authdb, gidlist[i], numeric, &column);
		}
		fprintf(stdout, "\n");
	    }
	    rv = 0;
	}
	else if (cmd == CMD_CHUSER) {

	    if (pwd) {
		if (uoptr->uo_pwd) {
		    FREE(uoptr->uo_pwd);
		}
		uoptr->uo_pwd = STRDUP(pwd);
	    }
	    if (rname) {
		if (uoptr->uo_rname) {
		    FREE(uoptr->uo_rname);
		}
		uoptr->uo_rname = STRDUP(rname);
	    }
	    if (newname) {
		rv = userRename(&err, ((AuthDB_t *)authdb)->adb_userdb,
				uoptr, newname);
		if (rv) {
		    fprintf(stderr, "%s: error (%d) in renaming user\n",
			    progname, rv);
		    nsadbErrorFmt(&err, errbuf, sizeof(errbuf), 4);
		    if (errbuf[0]) {
			fprintf(stderr, "%s\n", errbuf);
		    }
		    rv = 1;
		}
	    }
	    else {
		rv = nsadbModifyUser(&err, authdb, uoptr);
		if (rv < 0) {
		    fprintf(stderr, "%s: error (%d) in modifying user \n",
			    progname, rv);
		    nsadbErrorFmt(&err, errbuf, sizeof(errbuf), 4);
		    if (errbuf[0]) {
			fprintf(stderr, "%s\n", errbuf);
		    }
		    rv = 1;
		}
	    }
	    if (rv == 0) {
		fprintf(stderr, "%s: user \"%s\" modified successfully.\n",
			progname, SSTR(uoptr->uo_name));
	    }
	}
    }

  close:
    nsadbClose(authdb, 0);
    exit(rv);

  do_usage:
    if (cmd == CMD_MKUSER) {
	fprintf(stderr,
		"Usage: %s [options] username [auth-db-dir]\n", progname);
    }
    else {
	fprintf(stderr,
		"Usage: %s [options] [username] [auth-db-dir]\n",
		progname);
    }
    fprintf(stderr, "Options:\n");
    if (cmd == CMD_CHUSER) {
	fprintf(stderr,
		"    -a new-username             changes account username\n");
    }
    if ((cmd == CMD_MKUSER) || (cmd == CMD_CHUSER)) {
	fprintf(stderr,
		"    -n real-user-name           actual name of person\n");
	fprintf(stderr,
		"    -p encrypted-password       Unix DES-encrypted password\n");
    }
    if (cmd != CMD_MKUSER) {
	fprintf(stderr,
		"    -u user-id                  user id instead of username\n");
    }
    if (cmd == CMD_SHUSER) {
	fprintf(stderr,
		"    -N                          display numeric group ids\n");
    }

    dbname = getenv("NSAUTHDB");
    if ((dbname == 0) || (*dbname == 0)) {
	fprintf(stderr,
    "auth-db-dir can be specified using the NSAUTHDB environment variable.\n"
		);
    }
    exit(1);
}
