/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * Description (tacl.c)
 *
 *	Test program for ACLs.  Reads and parses an ACL definition file.
 *	Then reads access requests in the form:
 *
 *	acl-name right-name client-IP client-DNS client-user
 *
 *	where:
 *
 *	acl-name is the name of an ACL defined in the ACL definition
 *	file.
 *
 *	right-name is the name of an access right associated with
 *	at least one of the defined ACLs.
 *
 *	client-IP is the IP address of a hypothetical client, in dotted
 *	notation.
 *
 *	client-DNS is the fully-qualified DNS name of the hypothetical
 *	client, or is '-' to indicate that the client DNS name is not
 *	available.
 *
 *	client-user is a hypothetical client user name, or is '-' to
 *	indicate that the user name is not available.
 */

#include "base/systems.h"
#include "netsite.h"
#include "nsdb.h"
#include "nsuser.h"
#include "aclerror.h"
#include "aclparse.h"
#include "acleval.h"
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <ctype.h>

extern char * getprogname(char * argv0);

static char * progname;

static char linebuf[256];
static char errbuf[2048];

static void leave(int rc)
{
    exit(rc);
}

static void sigint()
{
    static int intcnt = 0;

    if (++intcnt > 2) {
	fprintf(stderr, "%s: interrupted - database may be corrupted!\n",
		progname);
	exit(1);
    }
}

static char * getstring(char * cp, char * pname, char **ret)
{
    char * tp;

    for (;;) {
	while (*cp && isspace(*cp)) ++cp;

	if (!*cp) {
	    fprintf(stdout, "Enter a string value for %s: ", pname);
	    cp = fgets(linebuf, sizeof(linebuf), stdin);
	    if (!cp) leave(0);
	    continue;
	}
	break;
    }

    tp = cp;
    while (*cp && (*cp != '\n') && !isspace(*cp)) ++cp;

    if (*cp) *cp++ = 0;

    *ret = strdup(tp);
    return cp;
}

static char * getintval(char * cp, char * pname, int * ret)
{
    long val;

    for (;;) {
	while (*cp && isspace(*cp)) ++cp;

	if (!*cp) {
	    fprintf(stdout, "Enter an integer value for %s: ", pname);
	    cp = fgets(linebuf, sizeof(linebuf), stdin);
	    if (!cp) leave(0);
	    continue;
	}
	break;
    }

    val = strtol(cp, &cp, 0);
    *ret = (int)val;

    return cp;
}

void nserrTraceback(NSErr_t * errp)
{
    NSEFrame_t * efp;
    int i;

    fprintf(stderr, "Error traceback:\n");

    for (efp = errp->err_first; efp != 0; efp = efp->ef_next) {

	fprintf(stderr, "  %s error %d: return code %d", efp->ef_program,
		(int)(efp->ef_errorid), (int)(efp->ef_retcode));

	if (efp->ef_errc > 0) {
	    fprintf(stderr, ", args = %s", efp->ef_errv[0]);
	    for (i = 1; i < efp->ef_errc; ++i) {
		fprintf(stderr, ", %s", efp->ef_errv[i]);
	    }
	}

	fprintf(stderr, "\n");
    }
}

int main(int argc, char * argv[])
{
    char * cp;
    char * filename;
    char * testfilename = 0;
    void * stp;
    ACContext_t * acc;
    ACL_t * aclptr;
    char * aclname;
    char * rightname;
    char * ipaddrstr;
    char * dnsstr;
    char * username;
    RightDef_t * rdp;
    int mdn;
    int rv;
    int tfd;
    int lineno;
    Realm_t * rlm;
    UserObj_t * uoptr;
    ClAuth_t clauth;
    NSErr_t err = NSERRINIT;

    progname = getprogname(argv[0]);

    /* Initialize NSPR */
    PR_Init("tacl", 1, 1, 0);

    /* Get the name of the file containing ACL definitions */
    if (argc > 1) filename = argv[1];
    else {
	cp = getstring("", "ACL file name", &filename);
    }

    rv = accReadFile(&err, filename, &acc);
    if (rv < 0) {
	fprintf(stderr, "%s: error reading ACL file (%d)\n", argv[0], rv);
	aclErrorFmt(&err, errbuf, sizeof(errbuf), 4);
	if (*errbuf != 0) {
	    fprintf(stderr, "%s\n", errbuf);
	}
	exit(1);
    }

    fprintf(stdout, "ACL definitions processed successfully!\n");

    if (argc > 2) {
	testfilename = argv[2];
	tfd = open(testfilename, O_RDONLY, 0);
	if (tfd < 0) {
	    fprintf(stderr,
		    "%s: Could not open test file %s.  Defaulting to stdin.\n",
		    argv[0], testfilename);
	}
	else {
	    close(0);
	    dup2(tfd, 0);
	}
    }

    lineno = 0;
    for (;;) {

	memset((void *)&clauth, 0, sizeof(clauth));

	cp = fgets(linebuf, sizeof(linebuf), stdin);
	if (cp == NULL) break;
	++lineno;

	while (*cp && isspace(*cp)) ++cp;
	aclname = cp;
	while (*cp && !isspace(*cp)) ++cp;
	if (*cp == 0) {
	    fprintf(stderr,
		    "line %d: missing required parameters...skipping\n",
		    lineno);
	    continue;
	}
	*cp++ = 0;
	while (*cp && isspace(*cp)) ++cp;
	rightname = cp;
	while (*cp && !isspace(*cp)) ++cp;
	if (*cp == 0) {
	    fprintf(stderr,
		    "line %d: missing required parameters...skipping\n",
		    lineno);
	    continue;
	}
	*cp++ = 0;
	while (*cp && isspace(*cp)) ++cp;
	ipaddrstr = cp;
	while (*cp && !isspace(*cp)) ++cp;
	if (*cp == 0) {
	    fprintf(stderr,
		    "line %d: missing required parameters...skipping\n",
		    lineno);
	    continue;
	}
	*cp++ = 0;
	while (*cp && isspace(*cp)) ++cp;
	dnsstr = cp;
	while (*cp && !isspace(*cp)) ++cp;
	if (*cp == 0) dnsstr = 0;
	else {
	    *cp++ = 0;
	    if ((dnsstr[0] == '-') && (dnsstr[1] == 0)) dnsstr = 0;
	}
	while (*cp && isspace(*cp)) ++cp;
	username = cp;
	while (*cp && !isspace(*cp)) ++cp;
	if (*cp == 0) username = 0;
	else {
	    *cp++ = 0;
	    if ((username[0] == '-') && (username[1] == 0)) username = 0;
	}

	rv = symTableFindSym(stp, aclname, ACLSYMACL, (void **)&aclptr);
	if (rv < 0) {
	    fprintf(stderr, "line %d: ACL name %s not found...skipping.\n",
		    lineno, aclname);
	    continue;
	}

	rv = symTableFindSym(stp, rightname, ACLSYMRIGHT, (void **)&rdp);
	if (rv < 0) {
	    fprintf(stderr, "line %d: right name %s not found...skipping.\n",
		    lineno, aclname);
	    continue;
	}

	clauth.cla_ipaddr = (IPAddr_t)inet_addr(ipaddrstr);
	clauth.cla_dns = (dnsstr) ? STRDUP(dnsstr) : 0;

	rv = aclEvaluate(aclptr, rdp->rd_id, &clauth, &mdn);

	while ((rv & ACD_ACTION) == ACD_AUTH) {

	    fprintf(stdout,
		    "line %d: authentication requested %s by directive %d\n",
		    lineno, (rv & ACD_ALWAYS) ? "immediately" : "", mdn);

	    rlm = clauth.cla_realm;
	    if (rlm == 0) {
		fprintf(stderr,
			"line %d: authentication requested, but no realm\n",
			lineno);
		goto next_req;
	    }

	    if (username == 0) {
		fprintf(stderr, "line %d: no user name, so access denied\n",
			lineno);
		goto next_req;
	    }

	    signal(SIGINT, sigint);

	    rv = (*rlm->rlm_aif->aif_open)(&err, rlm->rlm_dbname, 0,
					   &rlm->rlm_authdb);
	    if (rv) {
		fprintf(stderr,
		 "line %d: error %d opening authentication database \"%s\"\n",
			lineno, rv, rlm->rlm_dbname);
		nserrTraceback(&err);
		nserrDispose(&err);
		goto next_req;
	    }

	    rv = (*rlm->rlm_aif->aif_findname)(&err, rlm->rlm_authdb, username,
					       AIF_USER, (void **)&uoptr);
	    if (rv != AIF_USER) {
		fprintf(stderr, "line %d: user database lookup returned %d\n",
			lineno, rv);
		nserrTraceback(&err);
		nserrDispose(&err);
		goto next_req;
	    }

	    (rlm->rlm_aif->aif_close)(rlm->rlm_authdb, 0);

	    signal(SIGINT, SIG_DFL);

	    clauth.cla_uoptr = uoptr;
	    rv = aclEvaluate(aclptr, rdp->rd_id, &clauth, &mdn);
	}

	if ((rv & ACD_ACTION) == ACD_DENY) {
	    fprintf(stdout, "line %d: access denied %sby directive %d\n",
		    lineno, (rv & ACD_ALWAYS) ? "immediately " : "", mdn);
	}
	else if ((rv & ACD_ACTION) == ACD_ALLOW) {
	    fprintf(stdout, "line %d: access allowed %sby directive %d\n",
		    lineno, (rv & ACD_ALWAYS) ? "immediately " : "", mdn);
	}
	else if (rv == 0) {
	    fprintf(stdout, "line %d: ACL does not apply to this access\n",
		    lineno);
	}
	else {
	    fprintf(stderr, "line %d: aclEvaluate() returned %d\n",
		    lineno, rv);
	}

      next_req:

	if (clauth.cla_uoptr != 0) {
	    userFree(uoptr);
	    clauth.cla_uoptr = 0;
	}
	if (clauth.cla_dns != 0) {
	    FREE(clauth.cla_dns);
	    clauth.cla_dns = 0;
	}
    }

    return 0;
}

