/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
#include "base/systems.h"
#include "netsite.h"
#include "nsdbmgmt.h"
#include "nsumgmt.h"
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

static UserObj_t * uoptr = 0;
static void * ndb = 0;
static char linebuf[256];

static void leave(int rc)
{
    if (ndb) {
	ndbClose(ndb, 0);
    }
    exit(rc);
}

static int require_db(void * userdb, char * funcname)
{
    if (userdb) return 0;

    fprintf(stdout, "%s requires a database to be open.  Try ndbOpen().\n",
	    funcname);
    return 1;
}

static int require_user(UserObj_t * uop, char * funcname)
{
    if (uop) return 0;

    fprintf(stdout, "%s requires a user object.  Try one of:\n", funcname);
    fprintf(stdout, "    userCreate(), userFindByName(), userFindByUid()\n");
    return 1;
}

static int showRecord(NSErr_t * errp,
		      void * parg, char * name, int reclen, char * recptr)
{
    int len;

    if (name) {
	if (name[0] == NDB_MDPREFIX) {

	    /* Metadata record */

	    fprintf(stdout, "name=\"%s\"  length=%d  data=\n",
		    name, reclen);
	    for (len = 0; len < reclen; ++len) {
		if ((len & 15) == 0) {
		    fprintf(stdout, "    ");
		}
		fprintf(stdout, "%02x ", *(unsigned char *)(recptr + len));
		if ((len & 3) == 3) {
		    fprintf(stdout, " ");
		}
		if ((len & 15) == 15) {
		    fprintf(stdout, "\n");
		}
	    }
	    if ((len & 15) != 15) {
		fprintf(stdout, "\n");
	    }
	}
	else {

	    /* Normal data record */

	    len = strlen(recptr);

	    if (reclen != (len+1)) {
		fprintf(stdout, "showRecord length mismatch (%d != %d)\n",
			reclen, len+1);
	    }

	    fprintf(stdout, "name=\"%s\"  data=\"%s\"\n", name, recptr);
	}
    }

    if (parg) {
	*(int *)parg += 1;
    }

    return NDB_ENUMCONT;
}

static int showUser(NSErr_t * errp, void * parg, UserObj_t * uoptr)
{
    int i;
    int gcnt;
    USI_t * gids;

    gcnt = UILCOUNT(&uoptr->uo_groups);
    fprintf(stdout, "name=%s, password=%s, uid=%d, flags=0x%x, %s\n",
	    uoptr->uo_name, uoptr->uo_pwd, uoptr->uo_uid,
	    uoptr->uo_flags, uoptr->uo_rname);
	    fprintf(stdout, "%d groups: ", gcnt);
    gids = UILLIST(&uoptr->uo_groups);
    for (i = 0; i < gcnt; ++i) {
	fprintf(stdout, "%d ", gids[i]);
    }
    fprintf(stdout, "\n");

    if (parg) {
	*(int *)parg += 1;
    }

    return NDB_ENUMCONT;
}

static char * getstring(char * cp, char * pname, char **ret)
{
    char * tp;

    for (;;) {
	while (*cp && isspace(*cp)) ++cp;

	if (!*cp) {
	    fprintf(stdout, "Enter a string value for %s: ", pname);
	    cp = fgets(linebuf, sizeof(linebuf), stdin);
	    if (!cp) leave(0);
	    continue;
	}
	break;
    }

    tp = cp;
    while (*cp && (*cp != '\n') && !isspace(*cp)) ++cp;

    if (*cp) *cp++ = 0;

    *ret = strdup(tp);
    return cp;
}

static char * getintval(char * cp, char * pname, int * ret)
{
    long val;

    for (;;) {
	while (*cp && isspace(*cp)) ++cp;

	if (!*cp) {
	    fprintf(stdout, "Enter an integer value for %s: ", pname);
	    cp = fgets(linebuf, sizeof(linebuf), stdin);
	    if (!cp) leave(0);
	    continue;
	}
	break;
    }

    val = strtol(cp, &cp, 0);
    *ret = (int)val;

    return cp;
}

int main(int argc, char * argv[])
{
    char * line;
    char * cp;
    int rv;
    NSErr_t err = NSERRINIT;

    for (;;) {

	line = fgets(linebuf, sizeof(linebuf), stdin);
	if (!line) leave(0);

	rv = strlen(line) - 1;

	if (line[rv] == '\n') line[rv] = 0;

	if (!strncmp(line, "ndbOpen", 7)) {
	    char * dbname;
	    int flags;
	    int dbtype;
	    int version;

	    cp = line + 7;
	    cp = getstring(cp, "dbname", &dbname);
	    cp = getintval(cp, "flags", &flags);
	    cp = getintval(cp, "dbtype", &dbtype);

	    ndb = ndbOpen(&err, dbname, flags, dbtype, &version);

	    fprintf(stdout, "ndbOpen returns 0x%x, version %d\n",
		    (unsigned int)ndb, version);
	    free(dbname);
	    continue;
	}

	if (!strncmp(line, "ndbAllocId", 10)) {
	    char * name;
	    unsigned int id;

	    if (require_db(ndb, "ndbAllocId")) continue;

	    cp = line + 10;
	    cp = getstring(cp, "name", &name);

	    rv = ndbAllocId(&err, ndb, name, &id);

	    fprintf(stdout, "ndbAllocId returns %d, id %d\n", rv, id);
	    free(name);
	    continue;
	}

	if (!strncmp(line, "ndbFreeId", 9)) {
	    char * name;
	    unsigned int id;

	    if (require_db(ndb, "ndbFreeId")) continue;

	    cp = line + 9;
	    cp = getstring(cp, "name", &name);
	    cp = getintval(cp, "id", &id);

	    rv = ndbFreeId(&err, ndb, name, id);

	    fprintf(stdout, "ndbFreeId returns %d\n", rv);
	    free(name);
	    continue;
	}

	if (!strncmp(line, "ndbIdToName", 11)) {
	    char * name;
	    unsigned int id;

	    if (require_db(ndb, "ndbIdToName")) continue;

	    cp = line + 11;
	    cp = getintval(cp, "id", &id);

	    rv = ndbIdToName(&err, ndb, id, &name);

	    if ((rv == 0) && (name != 0)) {
		fprintf(stdout, "ndbIdToName returns \"%s\"\n", name);
	    }
	    else {
		fprintf(stdout, "ndbIdToName returns %d\n", rv);
	    }
	    continue;
	}
	if (!strncmp(line, "ndbFindName", 11)) {
	    char * name;
	    int reclen;
	    char * recptr;

	    if (require_db(ndb, "ndbFindName")) continue;

	    cp = line + 11;
	    cp = getstring(cp, "name", &name);

	    reclen = 0;
	    recptr = 0;

	    rv = ndbFindName(&err, ndb, name, &reclen, &recptr);

	    fprintf(stdout, "ndbFindName returns %d, reclen = %d, \"%s\"\n",
		    rv, reclen, recptr);
	    free(name);
	    continue;
	}
	if (!strncmp(line, "ndbStoreName", 12)) {
	    int flags;
	    char * name;
	    int reclen;
	    char * recptr;

	    if (require_db(ndb, "ndbStoreName")) continue;

	    cp = line + 12;
	    cp = getintval(cp, "flags", &flags);
	    cp = getstring(cp, "name", &name);

	    while (cp && *cp && (isspace(*cp) || (*cp == '\n'))) ++cp;

	    if (!cp || !*cp) {
		fprintf(stdout, "record contents: ");
		cp = fgets(linebuf, sizeof(linebuf), stdin);
		if (!cp) leave(0);
	    }

	    recptr = cp;
	    reclen = strlen(recptr);
	    if (cp[reclen-1] == '\n') {
		cp[reclen-1] = 0;
	    }
	    else {
		/* Include null in reclen */
		++reclen;
	    }

	    rv = ndbStoreName(&err, ndb, flags, name, reclen, recptr);

	    fprintf(stdout, "ndbStoreName returns %d\n", rv);

	    free(name);
	    continue;
	}
	if (!strncmp(line, "ndbEnumerate", 12)) {
	    int flags;
	    int reccnt = 0;

	    if (require_db(ndb, "ndbEnumerate")) continue;

	    cp = line + 12;
	    cp = getintval(cp, "flags", &flags);

	    rv = ndbEnumerate(&err, ndb, flags, (void *)&reccnt, showRecord);

	    fprintf(stdout, "ndbEnumerate returns %d, %d records\n",
		    rv, reccnt);
	    continue;
	}
	if (!strncmp(line, "userCreate", 10)) {
	    char * name;
	    char * pwd;
	    char * rname;
	    int len;

	    cp = line + 10;
	    cp = getstring(cp, "name", &name);
	    cp = getstring(cp, "password", &pwd);

	    while (cp && *cp && (isspace(*cp) || (*cp == '\n'))) ++cp;

	    if (!cp || !*cp) {
		fprintf(stdout, "real name: ");
		cp = fgets(linebuf, sizeof(linebuf), stdin);
		if (!cp) leave(0);
	    }

	    rname = cp;
	    len = strlen(cp);
	    if (cp[len-1] == '\n') {
		cp[--len] = 0;
	    }

	    uoptr = userCreate(name, pwd, rname);

	    fprintf(stdout, "userCreate returns 0x%x\n", (USI_t)uoptr);
	    free(name);
	    free(pwd);
	    continue;
	}
	if (!strncmp(line, "userStore", 9)) {

	    if (require_db(ndb, "userStore")) continue;
	    if (require_user(uoptr, "userStore")) continue;

	    rv = userStore(&err, ndb, 0, uoptr);

	    fprintf(stdout, "userStore returns %d, uid %d\n",
		    rv, uoptr->uo_uid);
	    continue;
	}
	if (!strncmp(line, "userFindByName", 14)) {
	    char * name;

	    if (require_db(ndb, "userFindByName")) continue;

	    cp = line + 14;
	    cp = getstring(cp, "name", &name);

	    uoptr = userFindByName(&err, ndb, name);

	    fprintf(stdout, "userFindByName returns %x\n", (USI_t)uoptr);
	    free(name);
	    continue;
	}
	if (!strncmp(line, "userFindByUid", 13)) {
	    unsigned int uid;

	    if (require_db(ndb, "userFindByUid")) continue;

	    cp = line + 13;
	    cp = getintval(cp, "uid", &uid);

	    uoptr = userFindByUid(&err, ndb, uid);

	    fprintf(stdout, "userFindByUid returns 0x%x\n", (USI_t)uoptr);
	    continue;
	}
	if (!strncmp(line, "showUser", 8)) {
	    if (require_user(uoptr, "showUser")) continue;
	    showUser(0, 0, uoptr);
	    continue;
	}
	if (!strncmp(line, "userRemove", 10)) {
	    char * name;

	    if (require_db(ndb, "userRemove")) continue;
	    cp = line + 10;

	    cp = getstring(cp, "name", &name);

	    rv = userRemove(&err, ndb, 0, name);

	    fprintf(stdout, "userRemove returns %d\n", rv);
	    free(name);
	    continue;
	}
	if (!strncmp(line, "userAddGroup", 12)) {
	    unsigned int gid;

	    if (require_user(uoptr, "userAddGroup")) continue;

	    cp = line + 12;
	    for (;;) {
		cp = getintval(cp, "gid", &gid);
		if (gid == 0) break;

		rv = userAddGroup(uoptr, gid);
		fprintf(stdout, "userAddGroup returns %d\n", rv);
	    }
	    continue;
	}
	if (!strncmp(line, "userDeleteGroup", 16)) {
	    unsigned int gid;

	    if (require_user(uoptr, "userDeleteGroup")) continue;

	    cp = line + 16;
	    for (;;) {
		cp = getintval(cp, "gid", &gid);
		if (gid == 0) break;

		rv = userDeleteGroup(uoptr, gid);
		fprintf(stdout, "userDeleteGroup returns %d\n", rv);
	    }
	    continue;
	}
	if (!strncmp(line, "userEnumerate", 13)) {
	    int reccnt = 0;
	    int flags;

	    if (require_db(ndb, "userEnumerate")) continue;
	    cp = line + 13;
	    cp = getintval(cp, "flags", &flags);

	    rv = userEnumerate(&err, ndb, flags, (void *)&reccnt, showUser);

	    fprintf(stdout, "userEnumerate returns %d, %d users\n",
		    rv, reccnt);
	    continue;
	}
	fprintf(stdout, "unrecognized command\n");
    }
}
