/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * admlog.c: Log (verbosely) changes being made to server configuration
 *
 * All blame to Mike McCool
 */

#include "libadmin/libadmin.h"

#include "base/file.h"   /* FILE_PATHSEP */
#include "base/util.h"   /* sprintf */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#define TIME_FMT "%d/%b/%Y:%H:%M:%S"
#define DEF_ADMIN_USER "anonymous"

static int already_logged = 0;

/* Admin log format:
 *
 * <user>@<site>: <name> <time> 
 *     <script> <type>:<change>
 * 
 * admin@nightmare.mcom.com: https-nightmare-secure [16/Aug/1995:02:28:02]
 *     sec-act magnus.conf: 'Security' set to 'off'
 *
 * admin@nightmare.mcom.com: https-nightmare-secure [16/Aug/1995:02:32:45]
 *     commit admserv: Committing changes to /dir/foo/there
 */
 
/* iscommit = behave differently for commit log.  so shoot me */
void _do_log(char *logfile, char *msg)
{
    SYS_FILE f;
    char *filepath = NULL;
    int howmany = make_conflist();
    register int y;

    if(logfile)  {
        filepath = STRDUP(logfile);
        howmany=1;
    }

    for(y=0; y<howmany; y++)  {
        char *targetmsg = NULL, *srvnames = NULL;

        if(!logfile)  {
            filepath = (char *) MALLOC(strlen(CONFDIR(y))+strlen("commit")+2);
            sprintf(filepath, "%scommit", CONFDIR(y));
            srvnames = get_srvname(y);
        }  else  
            srvnames = get_srvname(-1);
 
        targetmsg = (char *)MALLOC(strlen(msg)+strlen(srvnames)+2);
        sprintf(targetmsg, msg, srvnames);

        f = system_fopenWA(filepath);
        if(f == SYS_ERROR_FD)  {
            report_error(FILE_ERROR, filepath, get_msg(ADMLOG_ERR_NOOPEN));
        }
#ifndef XP_WIN32 /* XXX aruna. Need to define global mutex on NT */
        if(system_flock(f) == -1)  {
            report_error(FILE_ERROR, filepath, get_msg(ADMLOG_ERR_NOLOCK));
        }
#endif /* XP_WIN32 */    
        system_fwrite_atomic(f, targetmsg, strlen(targetmsg));
#ifndef XP_WIN32 /* XXX aruna. Need to define global mutex on NT */
       if(system_ulock(f) == -1)  {
            report_error(FILE_ERROR, filepath, get_msg(ADMLOG_ERR_NOUNLOCK));
        }
#endif /* XP_WIN32 */
        system_fclose(f);
        FREE(filepath);
    }
}

char *_escape_percents(char *org)
{
    char *ans=(char *) MALLOC((strlen(org)*2)+4);
    register int x, y;
    for(x=0, y=0; org[x]; /* none */)  {
        if(org[x]=='%')  {
            ans[y++]=org[x];
            ans[y++]=org[x++];
        }  else  {
            ans[y++]=org[x++];
        }
    }
    ans[y]='\0';
    return ans;
}

NSAPI_PUBLIC void log_change(char *kind, char *change, ...)
{
    char *logfile;

    logfile = getenv("ADMIN_LOG");

    if(kind && change)  {
        char msg[BIG_LINE], *t1, *tmp;
        char *user, *host, *date, *script;
        va_list args;

        user = getenv("REMOTE_USER");
        if(!user) 
            user=DEF_ADMIN_USER;

        host = getenv("REMOTE_HOST");
        if(!host) {
            host = getenv("REMOTE_ADDR");
            if(!host)
                host = "";
        }

        {
            time_t tnl;
            struct tm *t;
            char tbuf[28];

            tnl = time(NULL);
            t = localtime(&tnl);

            strftime(tbuf, 28, TIME_FMT, t);

            date = STRDUP(tbuf);
        }

        tmp = STRDUP(getenv("SCRIPT_NAME"));
        script = STRDUP(strrchr(tmp, '/')+1);
        FREE(tmp);

        va_start(args, change);
        util_vsnprintf(msg, sizeof(msg), change, args);
        t1 = STRDUP(msg);
        tmp=_escape_percents(t1);

        if(!already_logged)  {
            util_snprintf(msg, sizeof(msg), 
                               "%s@%s: %%s [%s]%s  %s %s: %s%s",
                               user, host, /*sname,*/ date, ENDLINE, 
                               script, kind, tmp, ENDLINE);
            already_logged = 1;
        }  else  {
            util_snprintf(msg, sizeof(msg), "  %s %s: %s%s",
                                            script, kind, tmp, ENDLINE);
        }

        if(logfile) 
            _do_log(logfile, msg);

        if( ( (strcmp(kind, TO_STATUS)) && 
              (strcmp(kind, TO_SEC)) &&
              (strcmp(kind, TO_ADMIN)) &&
              (strcmp(kind, TO_BACKUP)) &&
              (strcmp(kind, TO_CACHE)) &&
              (strcmp(kind, TO_USERDB)) &&
              (strcmp(kind, TO_LDAP)) &&
              (!is_admserv())
           ))
            _do_log(NULL, msg);

    }
}

/* Get a pretty string for the current resource. */
NSAPI_PUBLIC char *log_curres(char **config)
{
    char *rtype = get_current_typestr(config);
    char *curres = get_current_resource(config);
    char ans[BIG_LINE];

    if(!strcmp(rtype, NAME))  {
        if(!strcmp(curres, "default"))
            sprintf(ans, get_msg(RES_ENTIRE_SERVER));
        else
            sprintf(ans, get_msg(RES_OBJECT_NAMED), curres);
    }
    else if(!strcmp(rtype, FILE_OR_DIR))  {
        int l = strlen(curres) - 1;
        if(curres[l] == '*')  {
            sprintf(ans, get_msg(RES_DIRECTORY), curres);
        }  else  {
            if(strchr(curres, '*'))
                sprintf(ans, get_msg(RES_FILES), curres);
            else
                sprintf(ans, get_msg(RES_FILE), curres);
        }
    }
    else if(!strcmp(rtype, TEMPLATE))  {
        sprintf(ans, get_msg(RES_TMPL), curres);
    }
    else if(!strcmp(rtype, WILDCARD))  {
        sprintf(ans, get_msg(RES_WILD), curres);
    }
    return STRDUP(ans);
}

