/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

/***********************************************************************
** Includes
***********************************************************************/

#include	<libadmin/libadmin.h>
#ifndef NSPR20
// new stuff added for confhttp.dll
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#define SYS_FILE FILE *
#endif
#include	<libadmin/backup.h>
#include	<string.h>
#include	<sys/stat.h>
#include	<errno.h>
#include	<time.h>
#ifndef NSPR20
// new stuff added for confhttp.dll
#include "subnspr.h"
#define system_fclose(ptr) fclose((FILE *)ptr)
#define system_ulock(X)
#define system_flock(X) 0
#define system_tlock(X) 0
#define strcasecmp(X, Y) stricmp((X), (Y))
#define system_gets(X, Y, Z)  fgets((X), (Y),(FILE *)(Z))
#define system_fwrite(stream, buffer, size)  fwrite((buffer), (size), 1, (FILE *)(stream))
#define filebuf_open(f, bufsiz) (filebuffer *)f
#define filebuf_close_buffer(X, Y)
#define system_fopenRW(confFile)  fopen(confFile, "r+")
#define system_unlink(X)  unlink(X)
#endif

static int flush( ADM_bkTree * );
static void deleteTag( ADM_bkTag * );
static void newFile( ADM_bkFile *, char *, char *, char *, int, long, char * );
static void deleteFile( ADM_bkFile * );
static int copyFile( char *src, char *dest );
static void reaganify( char *string );

/***********************************************************************
** FUNCTION:	ADM_bk_initTree
** DESCRIPTION:
**   Create a tree from the config file and root directory
** INPUTS:	root -		something like /usr/ns-home
**
**		confFile -	full path to the backups configuration file
**
**		readOnly -	0 if we intend to make changes, non-0 to snoop
**
**		waitForLock -	0 if we should return immediately if there is
**				lock
** OUTPUTS:	None
** RETURN:	Pointer to ADM_bkTree structure
** SIDE EFFECTS:
**	Build a data structure that maps the config file, and keep a lock
**	on it if not readOnly
** RESTRICTIONS:
**	Parser is rather rudimentary, so if the file has bad syntax, it will
**	probably crash.
** MEMORY:	Much memory is malloced.
** ALGORITHM:
**	Straightforward parser looks for expected patterns of text and
**	whitespace (important) to build the data structure.
***********************************************************************/

NSAPI_PUBLIC ADM_bkTree *
ADM_bk_initTree( char *root, char *confFile, int readOnly, int waitForLock )
{
  ADM_bkTree	*tree;
  SYS_FILE	f;
  filebuffer	*fb;
  unsigned	bufsiz = BUFSIZ;
  char		*buf;
  char		fullPath[BUFSIZ];
  int		section = 0;
  char		*token[5];
  int		i;
  int		j;
  int		srcLen;
  int		fileIndex;
  ADM_bkTag	*tag;
  int		lockStatus;
  int		failLockCount = 0;
  struct stat	statbuf;
  int		done = 0;

  if ( stat( confFile, &statbuf ) != 0 ) {
    return NULL;
  }
  buf = (char *)MALLOC( bufsiz );
  if ( ( f = system_fopenRW( confFile ) ) == NULL ) {
    return NULL;	/* bad things happened */
  }
  fb = filebuf_open( f, bufsiz );
  while ( ( lockStatus = ( waitForLock ? system_flock( f )
			   : system_tlock( f ) ) ) < 0) {
    sleep( 1 );
    if ( ++failLockCount > 5 ) {
      break;
    }
  }
  if ( lockStatus ) {
    return NULL;
  }

  tree = (ADM_bkTree *) MALLOC( sizeof( ADM_bkTree ) );

  /* Make the backup dirname from the confFile */
  tree->backupDir = STRDUP( confFile );
  for ( i = strlen( confFile ) - 2 ; tree->backupDir[i] != '/' && i ; --i )
    ;
  if ( i ) {
    tree->backupDir[i] = '\0';
  } else {
    tree->backupDir[0] = '.';
    tree->backupDir[1] = '\0';
  }

  tree->modifiable = ! readOnly;
  tree->quota = 0l;	/* default */
  tree->maxTags = 0;	/* default */
  tree->root = STRDUP( root );
  tree->f = f;
  tree->numFiles = 0;
  tree->numTags = 0;

  /* Get the version number */
  system_gets( buf, bufsiz, fb );
  /* QQQQ Maybe check version */

  while ( ( ! done ) && system_gets( buf, bufsiz, fb ) ) {
    while ( strlen( buf ) == bufsiz - 1 && buf[bufsiz-2] != '\n' ) {
      /* line longer than bufsiz */
      buf = (char *)REALLOC( buf, bufsiz * 2 );
      system_gets( buf + bufsiz - 1, bufsiz + 1, fb );
      bufsiz *= 2;
    }
    if ( buf[0] == '\n' ) {	/* section separator */
      ++section;
    } else {
      switch ( section ) {
      case 0:	/* config settings */
	srcLen = strlen( buf );
	for ( i = 0 ; i < srcLen ; ++i ) {
	  if ( buf[i] == '=' ) {
	    buf[i] = '\0';
	    /* kill any trailing whitespace before = */
	    for ( j = i - 1 ; j && ( buf[j] == ' ' || buf[j] == '\t' ) ;
		  --j ) {
	      buf[j] = '\0';
	    }
	    if ( strcasecmp( "diskquota", buf ) == 0 ) {
	      tree->quota = atol( buf + i + 1 );
	    } else if ( strcasecmp( "maxtags", buf ) == 0 ) {
	      tree->maxTags = atol( buf + i + 1 );
	    }
	    break;	/* leave for loop, we've done our work */
	  }
	}
	break;
      case 1:	/* files */
	srcLen = strlen( buf );
	token[0] = buf;
	token[4] = NULL;	/* in case of old file, with no excess info */
	for ( i = 0, j = 0 ; i < srcLen && j < 4 ; ++i ) {
	  if ( buf[i] == ':' ) {
	    token[++j] = buf + i + 1;
	    buf[i] = '\0';
	  }
	}
	if ( tree->numFiles++ ) {	/* already had some */
	  tree->files = (ADM_bkFile *) REALLOC( tree->files,
						sizeof( ADM_bkFile ) *
						tree->numFiles );
	} else {			/* first time for everything */
	  tree->files = (ADM_bkFile *) MALLOC( sizeof( ADM_bkFile ) );
	}
	if ( token[1][0] == '/' ) {	/* full path */
	  strcpy( fullPath, token[1] );
	} else {			/* relative path */
	  sprintf( fullPath, "%s/%s", root, token[1] );
	}
	newFile( tree->files + tree->numFiles - 1,
		 token[0], token[1], fullPath,
		 atoi( token[2] ), atol( token[3] ), token[4] );
	break;
      default:	/* tags */
	if ( buf[0] == '-' ) {
	  done = 1;
	  break;	/* Truncate not working on NT, gotta fake EOF */
	}
	srcLen = strlen( buf );
	token[0] = buf;
	for ( i = 0, j = 0 ; i < srcLen ; ++i ) {
	  if ( buf[i] == ':' ) {
	    buf[i] = '\0';		/* delimit something */
	    if ( token[0] == buf ) {	/* We've got a timestamp */
	      if ( tree->numTags == 0 ) {
		tree->tags = (ADM_bkTag *) MALLOC( sizeof( ADM_bkTag ) );
	      } else {
		tree->tags = (ADM_bkTag *) REALLOC( tree->tags,
						    sizeof( ADM_bkTag ) *
						    ( tree->numTags + 1 ) );
	      }
	      tree->tags[tree->numTags].timestamp = atol( buf );
	      tree->tags[tree->numTags].numCohorts = 0;
	      token[0] = buf + i + 1;	/* nickname */
	      ++tree->numTags;
	    } else {	/* not a timestamp, a nickname/version */
	      fileIndex = ADM_bk_findFile( tree, token[0] );
	      tag = tree->tags + tree->numTags - 1;
	      if ( tag->numCohorts == 0 ) {
		tag->cohorts = (ADM_bkTagEl *) MALLOC( sizeof( ADM_bkTagEl ) );
	      } else {
		tag->cohorts = (ADM_bkTagEl *) REALLOC( tag->cohorts,
							sizeof( ADM_bkTagEl ) *
							( tag->numCohorts +
							  1 ) );
	      }
	      tag->cohorts[tag->numCohorts].file = fileIndex;
	      tag->cohorts[tag->numCohorts].version = atoi( token[1] );
	      ++tag->numCohorts;
	      if ( buf[i+1] == ':' ) {	/* log section */
		tree->tags[tree->numTags - 1].log = STRDUP( buf + i + 2 );
		break;	/* don't bother parsing the log message */
	      }
	      token[0] = buf + i + 1;	/* next set of stuff */
	    }
	  } else if ( buf[i] == '/' ) {
	    token[1] = buf + i + 1;	/* start of version number */
	    buf[i] = '\0';		/* delimit nickname */
	  }
	}
	break;
      }
    }
  }
  /* All done */
  filebuf_close_buffer( fb, 0 );
  if ( readOnly ) {
    system_ulock( f );
    system_fclose( f );
  }
  FREE( buf );
  return tree;
}


/***********************************************************************
** FUNCTION:	ADM_bk_done
** DESCRIPTION:
**   Destructor for this class.  If the structure was modified, write it
**   out, and unlock the file.
** INPUTS:	tree to shut down
** OUTPUTS:	None
** RETURN:	None
** SIDE EFFECTS:
**	Rewrite conf file, if modified, and free up all the memory.
** RESTRICTIONS:
**	None known.
** MEMORY:	Cleans up previous MALLOCS
***********************************************************************/

NSAPI_PUBLIC void
ADM_bk_done( ADM_bkTree *tree )
{
  int	i;

  if ( tree->modifiable ) {
    if ( tree->modifiable > 1 ) {	/* was modified */
      flush( tree );
    }
    system_ulock( tree->f );
    system_fclose( tree->f );
  }
#ifdef NSPR20  /* TODO: figure out why it crashes on NT */
  FREE( tree->root );
  FREE( tree->backupDir );
#endif
  for ( i = 0 ; i < tree->numFiles ; ++i ) {
    deleteFile( tree->files + i );
  }
#ifdef NSPR20  /* TODO: figure out why it crashes on NT */
  FREE( tree->files );
#endif
  for ( i = 0 ; i < tree->numTags ; ++i ) {
    deleteTag( tree->tags + i );
  }
#ifdef NSPR20  /* TODO: figure out why it crashes on NT */
  FREE( tree->tags );
  FREE( tree );
#endif
}

static void
tag( ADM_bkTree *tree, long timestamp, char *log )
{
  int	i;

  if ( tree->numTags == 0 ) {
    tree->tags = (ADM_bkTag *) MALLOC( sizeof( ADM_bkTag ) );
  } else {
    tree->tags = (ADM_bkTag *) REALLOC( tree->tags,
					sizeof( ADM_bkTag ) *
					( tree->numTags + 1 ) );
  }
  tree->tags[tree->numTags].timestamp = timestamp;
  tree->tags[tree->numTags].log = STRDUP( log ? log : "\n" );
  tree->tags[tree->numTags].numCohorts = tree->numFiles;
  tree->tags[tree->numTags].cohorts =
    (ADM_bkTagEl *) MALLOC( sizeof( ADM_bkTagEl ) * ( tree->numFiles ) );
  for ( i = 0 ; i < tree->numFiles ; ++i ) {
    tree->tags[tree->numTags].cohorts[i].file = i;
    tree->tags[tree->numTags].cohorts[i].version = tree->files[i].version;
  }
  ++tree->numTags;
}

/***********************************************************************
** FUNCTION:	ADM_bk_addFile
** DESCRIPTION:
**   Put another file under backup management.  Do expiration of old
**   tags if quota is non-zero.
** INPUTS:	tree -	pointer to the tree structure initialized by
**			ADM_bk_initTree
**
**		nickname - Nickname to use for path.  If NULL, this
**			   function picks one for you.
**
**		path -	relative path from tree->root to file to manage.
**			If path starts with '/', it's the absolute path.
**
**		log -	a log message to associate with the tag created
**			by this new file being managed
** OUTPUTS:	None
** RETURN:	0 if good, otherwise, an error code
** SIDE EFFECTS:
**	Makes a copy of the file in the tree->backupDir, and modifies
**	the tree->files and tree->tags arrays.
** RESTRICTIONS:
**	If log is non-null, it must end in a new-line, and have no other
**	newlines in it.
**	If extraInfo is non-null, it must end in a new-line, and have no other
**	newlines in it.
** MEMORY:	Memory is realloced
** ALGORITHM:
**	Add a file to the tree->files block, and make a new tag to
**	include this file.  Copy the file into the tree->backupDir.
***********************************************************************/

NSAPI_PUBLIC int
ADM_bk_addFile( ADM_bkTree *tree, char *nickname, char *path,
		char *extraInfo, char *log )
{
  int		fileIndex;
  int		aliasCount = 0;
  char		*file;
  char		buf[BUFSIZ];
  char		filePath[BUFSIZ];
  struct stat	statBuf;

  if ( *path == '/' ) {	/* full path */
    strcpy( filePath, path );
  } else {		/* relative path */
    sprintf( filePath, "%s/%s", tree->root, path );
  }
  if ( ! tree->modifiable ) {
    return EACCES;		/* this is readonly */
  }
  if ( stat( filePath, &statBuf ) != 0 ) {
    return errno;
  }
  if ( nickname ) {
    strcpy( buf, nickname );
  } else {
    /* Get file name with no directory into buf */
    for ( file = path + strlen( path ) - 1 ;
	  *file != '/' && file >= path ; --file )
      ;
    nickname = file + 1;
    strcpy( buf, nickname );
  }
  reaganify( buf );	/* translate dangerous :'s to to _ */
  while ( ( fileIndex = ADM_bk_findFile( tree, buf ) ) != -1 ) {
    /* already exists */
    if ( strcmp( tree->files[fileIndex].path, path ) == 0 ) {
      /* already under management */
      return 0;
    } else {	/* same name, different path */
      sprintf( buf, "%s.%d", nickname, ++aliasCount );
      reaganify( buf );	/* translate dangerous :'s to to _ */
    }
  }
  /* if the while didn't return, we should have a unique name in buf */
  if ( tree->numFiles ) {
    tree->files = (ADM_bkFile *) REALLOC( tree->files,
					  sizeof( ADM_bkFile ) *
					  ( tree->numFiles + 1 ) );
  } else {
    tree->files = (ADM_bkFile *) MALLOC( sizeof( ADM_bkFile ) );
  }
  newFile( tree->files + tree->numFiles,
	   buf, path, filePath, 1, statBuf.st_mtime, extraInfo );
  ++tree->modifiable;
  sprintf( buf, "%s/%s.1", tree->backupDir, tree->files[tree->numFiles].name );
  copyFile( filePath, buf );
  ++tree->numFiles;
  if ( stat( buf, &statBuf ) != 0 ) {
    return errno;
  }
  tag( tree, statBuf.st_mtime, log );	/* tag to time of copied file */
  return 0;
}

/***********************************************************************
** FUNCTION:	ADM_bk_restoreFile
** DESCRIPTION:
**   restore a particular version of a file
** INPUTS:	tree -		pointer to tree created by ADM_bk_initTree
**
**		file -		index into tree->files array of file to restore
**
**		version -	version of file to restore
** OUTPUTS:	None
** RETURN:	0 if OK, otherwise errno from file copy.
** SIDE EFFECTS:
**	Backup version is copied over master file
** RESTRICTIONS:
**	None
** MEMORY:	No memory changes
** ALGORITHM:
**	Copy the appropriate backup copy
***********************************************************************/

NSAPI_PUBLIC int
ADM_bk_restoreFile( ADM_bkTree *tree, int file, int version )
{
  char	buf[BUFSIZ];

  sprintf( buf, "%s/%s.%d", tree->backupDir, tree->files[file].name, version );
  return copyFile( buf, tree->files[file].fullPath );
}

/***********************************************************************
** FUNCTION:	ADM_bk_restoreToTime
** DESCRIPTION:
**   restore all files back to the requisite time
** INPUTS:	tree -		pointer to tree created by ADM_bk_initTree
**
**		timestamp -	seconds since 1970 to find tag that tells
**				which versions of the backed up files to
**				restore
**
**		errorFunc -	pointer to function to call if there is an
**				error.  NULL if you don't care about errors.
**
**		successFunc -	pointer to function to call if there was
**				a successful backup.  NULL if you don't
**				care.
**
**		parameter -	something to pass to errorFunc and successFunc
**
** OUTPUTS:	None
** RETURN:	0 if good, otherwise, an errno value
** SIDE EFFECTS:
**	Potentially, lots of files get copied, but maybe none, if latest
**	tag is before timestamp
** RESTRICTIONS:
**	If you specify a time when not all the current files were being
**	tracked, those newer files are untouched.
** MEMORY:	No expected memory leakage
** ALGORITHM:
**	Look for the tag that is at or just after the timestamp, then
**	restore to those versions.  If the version is the same as the
**	current tag, the file is untouched.
***********************************************************************/

NSAPI_PUBLIC int
ADM_bk_restoreToTime( ADM_bkTree *tree, long timestamp,
		      ADM_bkAddResultFunc errorFunc,
		      ADM_bkAddResultFunc successFunc,
		      void *parameter )
{
  int		i;
  ADM_bkTag	*tag = NULL;
  int		result = 0;

  if ( ! tree->numTags ||
       ( tree->tags[tree->numTags - 1].timestamp < timestamp ) ) {
    /* no tags, or all tags older than this time */
    return 0;
  }
  for ( i = tree->numTags - 1 ; --i ; ) {
    if ( tree->tags[i].timestamp == timestamp ) {
      tag = tree->tags + i;
      break;	/* found precisely it */
    }
    if ( tree->tags[i].timestamp < timestamp ) {
      tag = tree->tags + i - 1;
      break;
    }
  }
  if ( ! tag ) {
    tag = tree->tags;	/* first tag, since they are asking for prehistory */
  }
  for ( i = 0 ; i < tag->numCohorts ; ++i ) {
    /* depends on tag and file to always have the files in the same order */
    if ( ADM_bk_restoreFile( tree, i, tag->cohorts[i].version ) != 0 ) {
      if ( errorFunc ) {
	(*errorFunc)( parameter, tree, i, tag->cohorts[i].version );
	result = 1;
      }
    } else {		/* it worked */
      if ( successFunc ) {
	(*successFunc )( parameter, tree, i, tag->cohorts[i].version );
      }
    }
  }
  return result;
}

static void
deleteTag( ADM_bkTag *tag )
{
#ifdef NSPR20  /* TODO: figure out why it crashes on NT */
  FREE( tag->cohorts );
  if ( tag->log ) {
    FREE( tag->log );
  }
#endif
}

static void
newFile( ADM_bkFile *file, char *name, char *path, char *fullPath,
	 int version, long timestamp, char *extraInfo )
{
  file->name = STRDUP( name );
  file->path = STRDUP( path );
  file->fullPath = STRDUP( fullPath );
  file->version = version;
  file->timestamp = timestamp;
  if ( extraInfo ) {
    file->extraInfo = STRDUP( extraInfo );
  } else {
    file->extraInfo = NULL;
  }
}

static void
deleteFile( ADM_bkFile *file )
{
#ifdef NSPR20  /* TODO: figure out why it crashes on NT */
  FREE( file->name );
  FREE( file->path );
  FREE( file->fullPath );
  if ( file->extraInfo ) {
    FREE( file->extraInfo );
  }
#endif
}

NSAPI_PUBLIC int
ADM_bk_findFile( ADM_bkTree *tree, char *name )
{
  int	i;

  for ( i = 0 ; i < tree->numFiles ; ++i ) {
    if ( strcmp( name, tree->files[i].name ) == 0 ) {
      return i;
    }
  }
  return -1;	/* didn't find it */
}

static int
copyFile( char *src, char *dest )
{
  FILE		*sf;
  FILE		*df;
  char		buf[BUFSIZ];
  size_t	charsRead;

  if ( ( sf = fopen( src, "r" ) ) == NULL ) {
    return errno;
  }
  if ( ( df = fopen( dest, "w" ) ) == NULL ) {
    return errno;
  }
  while ( ( charsRead = fread( buf, 1, BUFSIZ, sf ) ) != 0 ) {
    if ( fwrite( buf, 1, charsRead, df ) != charsRead ) {
      return errno;
    }
  }
  fclose( sf );
  fclose( df );

  return 0;	/* everything's good */
}

static int
flush( ADM_bkTree *tree )
{
  int	i;
  int	j;
  char	buf[BUFSIZ];

#ifndef NSPR20 
  rewind(tree->f);   /* no need to truncate file during installation */
#else
  system_zero( tree->f );	/* in case the file got shorter */
#endif
  system_fwrite( tree->f, "backups.conf:Version 3.0\n", 25 );
  
  if ( tree->quota ) {
    sprintf( buf, "DiskQuota=%ld\n", tree->quota );
    system_fwrite( tree->f, buf, strlen( buf ) );
  }
  if ( tree->maxTags ) {
    sprintf( buf, "MaxTags=%d\n", tree->maxTags );
    system_fwrite( tree->f, buf, strlen( buf ) );
  }
  system_fwrite( tree->f, "\n", 1 );	/* separator */
  /* Print file block */
  for ( i = 0 ; i < tree->numFiles ; ++i ) {
    sprintf( buf, "%s:%s:%d:%ld:%s",
	     tree->files[i].name, tree->files[i].path,
	     tree->files[i].version, tree->files[i].timestamp,
	     tree->files[i].extraInfo ? tree->files[i].extraInfo : "\n" );
    system_fwrite( tree->f, buf, strlen( buf ) );
  }
  system_fwrite( tree->f, "\n", 1 );	/* separator */
  /* print tags block */
  for ( i = 0 ; i < tree->numTags ; ++i ) {
    sprintf( buf, "%ld:", tree->tags[i].timestamp );
    system_fwrite( tree->f, buf, strlen( buf ) );
    for ( j = 0 ; j < tree->tags[i].numCohorts ; ++j ) {
      sprintf( buf, "%s/%d:",
	       tree->files[tree->tags[i].cohorts[j].file].name,
	       tree->tags[i].cohorts[j].version );
      system_fwrite( tree->f, buf, strlen( buf ) );
    }
    system_fwrite( tree->f, ":", 1 );
    /* newline still in log */
    system_fwrite( tree->f, tree->tags[i].log, strlen( tree->tags[i].log ) );
  }
  system_fwrite( tree->f, "--EOF--\n", 8 );
  return 0;
}

static int
backup( ADM_bkTree *tree, int file, int timestamp )
{
  char	buf[BUFSIZ];

  if ( ! tree->modifiable ) {
    return -1;
  }
  ++tree->files[file].version;
  tree->files[file].timestamp = timestamp;
  ++tree->modifiable;
  sprintf( buf, "%s/%s.%d", tree->backupDir, tree->files[file].name,
	   tree->files[file].version );
  return copyFile( tree->files[file].fullPath, buf );
}

/***********************************************************************
** FUNCTION:	Using the passed data structure, make backups
** DESCRIPTION:
**   Go through all the tree->files, and check to see if any master files
**   have changed.  If so, make new backups for them, and make a tag.
** INPUTS:	tree -	pointer to tree created by ADM_bk_initTree
**
**		log -	message to store with tag
** OUTPUTS:	None
** RETURN:	0 if everything went well
** SIDE EFFECTS:
**	tree structure might change, and files might get copied
** RESTRICTIONS:
**	If you don't call ADM_bk_done, the conf file will not reflect
**	the state of the backup directory.
** MEMORY:	No direct changes, but backup() will allocate memory
** ALGORITHM:
**	Checks the modified time for each file under management, and if
**	it's different from the last time it was backed up, it makes a
**	new backup.  If that happens at least once, a new tag is created
**	in memory.
***********************************************************************/

NSAPI_PUBLIC int
ADM_bk_doBackups( ADM_bkTree *tree, char *log )
{
  int		i;
  struct stat	statBuf;
  int		tagIt = 0;

  for ( i = 0 ; i < tree->numFiles ; ++i ) {
    if ( stat( tree->files[i].fullPath, &statBuf ) != 0 ) {
      return errno;
    }
    if ( tree->files[i].timestamp != statBuf.st_mtime ) {	/* modified */
      tagIt = 1;
      backup( tree, i, statBuf.st_mtime );
    }
  }
  if ( tagIt ) {
    tag( tree, time( 0 ), log );
  }
  ADM_bk_expire( tree );	/* keep our footprint light */

  return 0;
}

/***********************************************************************
** FUNCTION:	ADM_bk_makeBackups
** DESCRIPTION:
**   Make backups of files that have changed.
** INPUTS:	confFile	- full path to backup configurate file
**
**		root		- path of root dir that files are relative to
**
**		log		- message to log with tag, if any files get
**				  backed up
** OUTPUTS:	None
** RETURN:	0 if everything went well
** SIDE EFFECTS:
**	Files could get copied, and the confFile could be changed.
** RESTRICTIONS:
**	None known, other than if you give it bad files, it may crash.
** MEMORY:	Net memory change should be zero.  Things are created and
**		destroyed on the heap, though.
** ALGORITHM:
**	Easy function to call if all you want to do is make backups.
***********************************************************************/

NSAPI_PUBLIC int
ADM_bk_makeBackups( char *confFile, char *root, char *log )
{
  ADM_bkTree	*tree;
  int		result;

  if ( ( tree = ADM_bk_initTree( root, confFile, 0, 1 ) ) == NULL ) {
    return -1;
  }
  result = ADM_bk_doBackups( tree, log );
  ADM_bk_done( tree );
  return result;
}

/* Remove colons, safely replacing them with _ */
/************************************************
 * Q: What features was the Presidential typwriter missing in 1988?
 * A: No colon and no memory
 ***********************************************/
static void
reaganify( char *string )
{
  while ( *string ) {
    if ( *string == ':' ) {
      *string = '_';
    }
    ++string;
  }
}

static long
spaceUsed( ADM_bkTree *tree )
{
  int		i;
  int		j;
  int		*lastVersion = (int *)MALLOC( sizeof( int ) * tree->numFiles );
  long		size = 0l;
  struct stat	statBuf;
  char		buf[BUFSIZ];

  /* pretend to calloc */
  for ( i = 0 ; i < tree->numFiles ; ++i ) {
    lastVersion[i] = 0;
  }
  for ( i = 0 ; i < tree->numTags ; ++i ) {
    for ( j = 0 ; j < tree->tags[i].numCohorts ; ++j ) {
      if ( tree->tags[i].cohorts[j].version != lastVersion[j] ) {
	sprintf( buf, "%s/%s.%d", tree->backupDir, tree->files[j].name,
		 tree->tags[i].cohorts[j].version );
	stat( buf, &statBuf );
#ifdef XP_UNIX
	/* get the size in bytes, rounded up to the blocksize */
	size += ( ( statBuf.st_size + statBuf.st_blksize - 1 ) /
		  statBuf.st_blksize ) * statBuf.st_blksize;
#else /* Windows NT */
	/* get the size in bytes.  Can't round up, since windows stat
	   won't tell us how big the blocks are */
	size += statBuf.st_size;
#endif
	lastVersion[j] = tree->tags[i].cohorts[j].version;
      }
    }
  }
  FREE( lastVersion );
  return ( size + 1023L ) / 1024L;
}

static void
expireSpace( ADM_bkTree *tree )
{
  int	i;
  char	buf[BUFSIZ];

  if ( ! tree->modifiable ) {
    return;
  }
  if ( tree->quota && spaceUsed( tree ) > tree->quota ) {
    /* trim oldest tag and any files that aren't refered to any more */
    if ( tree->numTags < 2 ) {	/* not enough tags to expire anything */
      return;
    }
    for ( i = 0 ; i < tree->tags[0].numCohorts ; ++i ) {
      if ( tree->tags[0].cohorts[i].version !=
	   tree->tags[1].cohorts[i].version ) {
	/* changed, so we can expire it */
	sprintf( buf, "%s/%s.%d", tree->backupDir, tree->files[i].name,
		 tree->tags[0].cohorts[i].version );
	system_unlink( buf );
      }
    }
    deleteTag( tree->tags );	/* free tag info */
    ++tree->modifiable;
    /* remove tag 0, by the simple expediency of shifting the tags down */
    for ( i = 1 ; i < tree->numTags ; ++i ) {
      tree->tags[i - 1] = tree->tags[i];
    }
    --tree->numTags;
    expireSpace( tree );	/* again, in case we haven't deleted enough */
  }
}

static void
expireTags( ADM_bkTree *tree )
{
  if ( ! tree->modifiable ) {
    return;
  }
  if ( tree->maxTags && tree->numTags > tree->maxTags ) {
    int		i;
    int		j;
    int		keepRow;
    char	buf[BUFSIZ];

    keepRow = tree->numTags - tree->maxTags;
    for ( i = keepRow ; i-- ; ) {
      for ( j = tree->tags[i].numCohorts ; j-- ; ) {
	if ( tree->tags[i].cohorts[j].version !=
	     tree->tags[keepRow].cohorts[j].version ) {
	  /* obsolete */
	  sprintf( buf, "%s/%s.%d", tree->backupDir, tree->files[j].name,
		   tree->tags[i].cohorts[j].version );
	  system_unlink( buf );
	}
      }
      deleteTag( tree->tags + i );
    }
    /* shift the good tags down */
    for ( i = keepRow ; i < tree->numTags ; ++i ) {
      tree->tags[i - keepRow] = tree->tags[i];
    }
    tree->numTags -= keepRow;	/* should be maxTags */
    ++tree->modifiable;
  }
}

/***********************************************************************
** FUNCTION:	ADM_bk_expire
** DESCRIPTION:
**   Check space used by backup files, and trim if necessary
** INPUTS:	tree - pointer to tree created by ADM_bk_initTree
** OUTPUTS:	None
** RETURN:	None
** SIDE EFFECTS:
**	Less memory and disk space may be used after the call
** RESTRICTIONS:
**	None
** MEMORY:	Might free up some memory
** ALGORITHM:
**	Check the disk space, and remove the files associated
**	with the oldest tag if necessary.  Then recurse.
***********************************************************************/

extern void
ADM_bk_expire( ADM_bkTree *tree )
{
  if ( ! tree->modifiable ) {
    return;
  }
  if ( tree->quota ) {
    expireSpace( tree );
  }
  if ( tree->maxTags ) {
    expireTags( tree );
  }
}

NSAPI_PUBLIC void
ADM_bk_setQuota( ADM_bkTree *tree, long quota )
{
  tree->quota = quota;
  if ( tree->modifiable ) {
    ++(tree->modifiable);
  }
}

NSAPI_PUBLIC void
ADM_bk_setMaxTags( ADM_bkTree *tree, int maxTags )
{
  tree->maxTags = maxTags;
  if ( tree->modifiable ) {
    ++(tree->modifiable);
  }
}

NSAPI_PUBLIC int
ADM_bk_newConfFile( char *dir, char *fileName )
{
  char	buf[BUFSIZ];
  FILE	*file;

  strcpy( buf, dir );
  strcat( buf, "/" );
  strcat( buf, fileName );
  if ( ! ( file = fopen( buf, "w" ) ) ) {
    return errno;
  }
  fputs( "backups.conf:Version 3.0\n\n", file );
  fclose( file );
  return 0;
}

#ifdef TESTING

/* Dummy main for testing purposes QQQQ */
main( int argc, char **argv )
{
  ADM_bkTree	*tree;

  if ( argc == 3 && strcmp( argv[1], "-b" ) == 0 ) {
    ADM_bk_makeBackups( argv[2], "/tmp", NULL );
  } else if ( argc == 4 && strcmp( argv[1], "-a" ) == 0 ) {
    tree = ADM_bk_initTree( "/tmp", argv[3], 0, 1 );
    ADM_bk_addFile( tree, NULL, argv[2], NULL, NULL );
    ADM_bk_done( tree );
  } else {
    fprintf( stderr, "Usage: %s {-a file|-b} configFile\n", argv[0] );
    exit( 1 );
  }
}

#endif
