/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * cluster.c
 *
 * Revision History	By Whom		Note
 * 09/12/96		Adrian Chan	Initial Version
 */

#include "libadminutil/admutil.h"
#include "libadminutil/distadm.h"
#include "libadmin/libadmin.h"
#include "libadmin/cluster.h"
#include "base/shexp.h"
#include "base/util.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef XP_WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif 

#if defined(SOLARIS)
#include <sys/filio.h>
#endif	/* SOLARIS */

#include "ssl.h"

typedef	struct {		/* This is internal threaded structure for */
   CL_server *server;		/* cluster management */
   CL_request *req;
   int errcode;
} CL_thread_data;

/*
 * macro for synchronizing output
 */
#ifdef	XP_UNIX

#include "base/crit.h"
static CRITICAL ClusterCriticalSection = NULL;
static int InitCriticalSection = 1;
#define	INIT_CLUSTER_CRITICAL_SECTION	\
	if (InitCriticalSection) { \
	   systhread_init("cluster"); \
	   ClusterCriticalSection = crit_init(); \
	   InitCriticalSection = 0; \
	}
#define ENTER_CLUSTER_CRITICAL_SECTION crit_enter(ClusterCriticalSection)
#define LEAVE_CLUSTER_CRITICAL_SECTION crit_exit(ClusterCriticalSection)

#else	/* XP_WIN32 */

static CRITICAL_SECTION ClusterCriticalSection;
static BOOLEAN InitCriticalSection = TRUE;

#define	INIT_CLUSTER_CRITICAL_SECTION	\
	if (InitCriticalSection) { \
	   systhread_init("cluster"); \
	   InitializeCriticalSection(&ClusterCriticalSection); \
	   InitCriticalSection = FALSE; \
	}
#define ENTER_CLUSTER_CRITICAL_SECTION EnterCriticalSection(&ClusterCriticalSection);
#define LEAVE_CLUSTER_CRITICAL_SECTION LeaveCriticalSection(&ClusterCriticalSection)
                           
#endif	/* XP_UNIX */

#define	ADMIN_ID		"Netscape-Administrator"

#define	CL_GET_REQ_HEADER	"%s %s%c%s HTTP/1.0\nNS-Interactive: No\nAuthorization: %s\n\n"
#define	CL_POST_REQ_HEADER	"%s %s HTTP/1.0\nNS-Interactive: No\nAuthorization: %s\nContent-type: application/x-www-form-urlencoded\nContent-length: %d\n\n%s\n"

/*
 * parse_http_header:
 *
 * return: 0 on success (server is OK and user is authenticated)
 *	  -1 on unmatched remote server (wrt remote_server_id)
 *	  -2 on bad authentication.
 */
NSAPI_PUBLIC int
parse_http_header(PRFileDesc *sockd, bufstruct *nbuf, char *remote_server_id)
{
   char *line;		/* pointing to each received ASCII line */
   int server_OK;	/* indicates if remote admin server is good */
   int auth_OK;

   server_OK = 0;
   auth_OK = 1;

   while( (line = get_line_from_fd(sockd, nbuf)) != (char *) NULL)  {
      if (is_end_of_headers(line)) {
	 if (!server_OK) {
	    end_http_request(sockd);
	    return(-1);
         }
	 if (!auth_OK) {
	    end_http_request(sockd);
	    return(-2);
	 }
	 return(0);
      }

      if (!strncasecmp(line, "HTTP/", 5)) {
	 if (strstr(line, " 401 Unauthorized")) {
	    auth_OK = 0;
	 }
      }

      if (!strncasecmp(line, "Server: ", 8)) {
         if (strstr(line, remote_server_id)) {
            server_OK = 1;
         }
      }
   }

   return(-1);
}

NSAPI_PUBLIC int
parse_xfile_boundary(PRFileDesc *sockd, bufstruct *nbuf)
{
   char *line;		/* pointing to each received ASCII line */
   int server_OK;	/* indicates if remote admin server is good */

   server_OK = 0;

   while( (line = get_line_from_fd(sockd, nbuf)) != (char *) NULL)  {
      if (!strncmp(line, CF_ERRSTR, strlen(CF_ERRSTR)))  {
         return(-1);
      }
      /* designate begin/end of file */
      if (!strncmp(line, CF_BOUNDARY, strlen(CF_BOUNDARY))) {
         return(0);
      }
   }

   /* Bad protocol: No file boundary mark */
   return(-1);
}

NSAPI_PUBLIC int
parse_xfilename(PRFileDesc *sockd, bufstruct *nbuf, char *filename)
{
   char *line;

   while( (line = get_line_from_fd(sockd, nbuf)) != (char *) NULL)  {
      if (!strncmp(line, CF_ERRSTR, strlen(CF_ERRSTR)))  {
	 fprintf(stdout, "System error: remote server fails: %s\n", line);
         return(-1);
      }
      if (!strncmp(line, CF_NEWCONFIG, strlen(CF_NEWCONFIG))) {

         char *t, *u;
						/* doen't allow directory */
	 t = &(line[strlen(CF_NEWCONFIG)]);
	 if(strchr(t, '/') || strchr(t, '\\'))  {
	    end_http_request(sockd);
	    fprintf(stdout, "Incorrect Usage: received bad config file name: %s\n", t);
	    return(-1);
         }

         u = t;
         while(*u)  {			/* don't allow CR/LF */
	    if((*u == CR) || (*u == LF)) {
	       *u = '\0';
	    }
            ++u;
         }

	 if (!strcmp(filename, "*")) {
	    strcpy(filename, t);
	    return(0);
	 }

	 if (!strcmp(t, filename)) {
	    return(0);
	 }

	 end_http_request(sockd);
	 fprintf(stdout, "Incorrect Usage: received unexpected filename: %s\n", filename);
	 return(-1);
      }
   }
   return(-1);
}						/* if not the expected file */

NSAPI_PUBLIC int
CL_emptyDatabase()
{
   parse_line( "<!-- TITLE \"Cluster Database Empty\" -->\n", NULL );
   fprintf(stdout, "<center><table border=2 width=100%%>\n");
   fprintf(stdout, "<tr><td align=center width=100%%><hr size=0 width=0>");
   fprintf(stdout, "<FONT size=+2><b>Cluster Database Is Empty</b></FONT>");
   fprintf(stdout, "<hr size=0 width=0></th></tr>");
   fprintf(stdout, "</table></center>\n");
   fprintf(stdout, "<B><CENTER><PRE>\n\n");
   fprintf(stdout, "Use Add Server to populate database first\n");
   fprintf(stdout, "</PRE></CENTER></B>\n");
   fflush(stdout);
   return 0;
}

NSAPI_PUBLIC int
CL_getAdminURL(char *product, CL_server *server, int *errcode)
{
   char *ADMSERV_ROOT;   	/* ENV inited from ns-admin */
   char path[PATH_MAX];         /* full path to <product>/<server>/adminURL */
   FILE *fp;			/* open file to server's adminURL file */
   char *uri;			/* placeholder of uri in adminURL */
 
   if ((ADMSERV_ROOT = getenv("ADMSERV_ROOT")) == (char *) NULL) {
      *errcode = CLERR_NOT_INIT;
      return(-1);
   }

   server->nsAdmin_port = -1;
 
   sprintf(path, "%s%ccluster%c%s%c%s%cadminURL", ADMSERV_ROOT, FILE_PATHSEP,
                  FILE_PATHSEP, product, FILE_PATHSEP, server->server_ident, FILE_PATHSEP);
   if ((fp = fopen(path, "r")) == (FILE *) NULL) {
      server->nsAdmin_URL = (char *) MALLOC(1);
      *(server->nsAdmin_URL) = '\0';
      *errcode = ERR_FAIL_OPEN;
      return(-1);
   }
 
   server->nsAdmin_URL = (char *) MALLOC(get_file_size(path)+2);
   fgets(server->nsAdmin_URL, PATH_MAX, fp);
 
   fclose(fp);
 
   if ((server->nsAdmin_URL)[strlen(server->nsAdmin_URL)-1] == '\n') {
      (server->nsAdmin_URL)[strlen(server->nsAdmin_URL)-1] = '\0';
   }

   if (decompose_url(server->nsAdmin_URL,
		     &(server->nsAdmin_protocol),
		     &(server->nsAdmin_hostname),
		     (uint *) &(server->nsAdmin_port),
		     &uri) < 0) {
      *errcode = CLERR_BAD_URL;
      return(-1);
   }

   return(0);
}

/*
 * There should be a way for CGI to obtain user/password information
 * For now, get it from cgi variable list and turned that into uuencode
 * auth data!
 */
NSAPI_PUBLIC int
CL_getUserAuth(unsigned char **auth, int *errcode) 
{
   int len;
   char *user, *passwd;
   unsigned char *userpass;

   if (ADM_InitializePermissions(errcode) != 0) {
      return(-1);
   }

   if (ADM_GetAuthorizationString(errcode, (char **) auth) == 0) {
      return(0);
   }

   if (ADM_GetCurrentUsername(errcode, &user) != 0) {
      *errcode = CLERR_NO_USER;
      return(-1);
   }

   if (ADM_GetCurrentPassword(errcode, &passwd) != 0) {
      *errcode = CLERR_NO_PASS;
      return(-1);
   }

   len = strlen(user)+strlen(passwd)+2;
   userpass = (unsigned char *) MALLOC(len);
   *auth = (unsigned char *) MALLOC(2*len);

   (void) sprintf((char *) userpass, "%s:%s", user, passwd);
   (void) do_uuencode(userpass, *auth, strlen((char *) userpass));
   FREE(userpass);

   return(0);
}

NSAPI_PUBLIC int
CL_getRequest(CL_request *req, int *errcode)
{
   char **var;			/* array of CGI variable */
   char **var_value;		/* corresponding value of the post variable */
   char *cgivar;		/* pointer to a single CGI variable */
   int  num_vars;		/* number of variables to pass to target */
   int  len, i;			/* length of request string & loop index */
   char tmpvar[16];		/* tempoary storage for CL_VAR(x) */

   if (CL_getUserAuth(&(req->auth_info), errcode) < 0) return(-1);

   req->product = (char *) get_cgi_var("CL_PRODUCT", "", "");
   if (req->product == (char *) NULL) {
      *errcode = CLERR_NO_PRODUCT;
      return(-1);
   }

   req->method = (char *) get_cgi_var("CL_REQ_METHOD", "", "");
   if (req->method == (char *) NULL) {
      *errcode = CLERR_NO_METHOD;
      return(-1);
   }

   req->action = (char *) get_cgi_var("CL_REQ_ACTION", "", "");
   if (req->action == (char *) NULL) {
      *errcode = CLERR_NO_ACTION;
      return(-1);
   }

   req->expect = (char *) get_cgi_var("CL_REQ_EXPECT", "", "");
   if (req->expect == (char *) NULL) {
      *errcode = CLERR_NO_EXPECT;
      return(-1);
   }

   req->form_data = (char *) NULL;

   cgivar = (char *) get_cgi_var("CL_NVARS", "", "");
   if (cgivar == (char *) NULL) {
      num_vars = 0;
      var = (char **) MALLOC(1); var[0] = (char *) NULL;
      var_value = (char **) MALLOC(1); var_value[0] = (char *) NULL;
   } else {
      if ((num_vars = atoi(cgivar)) > 0) {
         var  = (char **) MALLOC(num_vars);
         var_value = (char **) MALLOC(num_vars);
      } else {
         num_vars = 0;
         var = (char **) MALLOC(1); var[0] = (char *) NULL;
         var_value = (char **) MALLOC(1); var_value[0] = (char *) NULL;
      }
   }

   len = 0;
   for (i = 0; i < num_vars; i++) {
      sprintf(tmpvar, "CL_VAR%d", i);
      if ((var[i]  = get_cgi_var(tmpvar, "", "")) != (char *) NULL) {
         len += strlen(var[i]);
      }
      if ((var_value[i] = get_cgi_var(var[i], "", "")) != (char *) NULL) {
         len += strlen(var_value[i]);
      }
      len += + 2;
   }
   len = len + strlen("CL_TARGET") + 256 + 2;

   req->form_data = (char *) MALLOC(len);
   req->form_data[0] = '\0';
   if (var[0] != (char *) NULL) {
      if (var_value[0] != (char *) NULL) {
         sprintf(req->form_data, "%s=%s", var[0], var_value[0]);
      } else {
         sprintf(req->form_data, "%s=", var[0]);
      }
   }
   for (i = 1; i < num_vars; i++) {
      if (var[i] != (char *) NULL) {
         if (var_value[i] != (char *) NULL) {
            sprintf(req->form_data, "%s&%s=%s", req->form_data, var[i], var_value[i]);
	 } else {
            sprintf(req->form_data, "%s&%s=", req->form_data, var[i]);
	 }
      }
   }

   FREE(var);
   FREE(var_value);

   return(0);
}

NSAPI_PUBLIC int
CL_getTargetServers(char *product, CL_server_list *server_list, int *errcode)
{
   char *cgivar;	/* pointer to CGI variables */
   int total_servers;	/* total number of servers presented to user */
   int i;		/* loop index */
   CL_server *server;	/* index pointer to server in server_list */
   char serverid[256];	/* placeholder for server identifier */

   /*
    * Total number of servers presented to user to select
    */
   cgivar = (char *) get_cgi_var("CL_NSERVER", "", "");
   if (cgivar == (char *) NULL) {
      *errcode = CLERR_NO_SERVER;
      return(-1);
   }
   total_servers = atoi(cgivar);

   /*
    * Total number of servers selected by user
    */
   cgivar = (char *) get_cgi_var("CL_NSELECT", "", "");
   if (cgivar == (char *) NULL) {
      *errcode = CLERR_NO_TARGET;
      return(-1);
   }
   server_list->total_selected = atoi(cgivar);
   server_list->server = (CL_server *) MALLOC(sizeof(CL_server) * server_list->total_selected);
   memset(server_list->server, '\0', sizeof(CL_server) * server_list->total_selected);

   /*
    * For now, go through the list exhaustively.  If proved to be slow, then
    * here may be a place to improve.
    */
   server = server_list->server;
   for (i = 0; i < total_servers; i++) {
      server->CLidx = i;
      sprintf(serverid, "server%d", i);
      cgivar = (char *) get_cgi_var(serverid, "", "");
      if (cgivar != (char *) NULL) {
         server->server_ident = (char *) STRDUP(cgivar);
         server++;
      }
   }

   return(0);
}

NSAPI_PUBLIC int
CL_reportLayout(CL_request *req, CL_server_list *server_list)
{
   int i, errcode;
   CL_server *server;
   char serverid[256];
   char *delimiter;

   /*
    * banner
    */
   fprintf(stdout, "<HTML><BODY BGCOLOR=\"#C0C0C0\">\n");
   fprintf(stdout, "<FORM NAME=CL_REPORT ACTION=clctrl>\n");
   fprintf(stdout, "<TABLE BORDER=2 WIDTH=100%%>\n");
   fprintf(stdout, "<TR><TD ALIGN=CENTER COLSPAN=3 WIDTH=100%%><FONT SIZE=+2><B>Cluster Execution Report</B></FONT></TD></TR>\n");
   fprintf(stdout, "<TR><TD>Server</TD><TD>Status</TD><TD>Details</TD></TR>\n");
   for (i = 0; i < server_list->total_selected; i++) {
      server = &(server_list->server[i]);
      if (CL_getAdminURL(req->product, server, &errcode) < 0) {
         server->request_status = errcode;
      }
      delimiter = strchr(server->server_ident, '@');
      if (delimiter) {
         if (strlen(delimiter) > 7) {
	    strcpy(delimiter+4, "...");
	 }
      } 
      (void) strcpy(serverid, server->server_ident);
      delimiter = strchr(serverid, '@');
      if (delimiter) {
	 *delimiter = '\0';
      }
      fprintf(stdout, "<TR>\n");
      fprintf(stdout, "<TD><A HREF=\"%s/%s/bin/index\" TARGET=\"%s\" >%s</A></TD>\n", server->nsAdmin_URL, serverid, server->server_ident, server->server_ident);
      fprintf(stdout, "<TD><INPUT TYPE=TEXT NAME=\"server%d\" ", server->CLidx);
      if (server->request_status == 0) {
         fprintf(stdout, "VALUE=\"START: remote execution\" SIZE=30</TD>\n");
      } else {
         fprintf(stdout, "VALUE=\"DONE: failed to obtain remote Admin Server URL\" SIZE=30</TD>\n");
      }
      fprintf(stdout, "<TD><INPUT TYPE=BUTTON NAME=VIEW ");
      fprintf(stdout, "VALUE=VIEW onClick=\"win_server%d=open('', 'report%d',", server->CLidx, server->CLidx);
      fprintf(stdout, "'scrollbars=1,resizable=1,width=400,height=200'); ");
      fprintf(stdout, "win_server%d.document.clear();", server->CLidx);
      fprintf(stdout, "win_server%d.document.write('<HEAD><TITLE>Execution Report For %s</TITLE></HEAD>');", server->CLidx, server->server_ident);
      fprintf(stdout, "win_server%d.document.writeln('<PRE>');", server->CLidx);
      fprintf(stdout, "win_server%d.document.writeln(document.CL_REPORT.report%d.value);", server->CLidx, server->CLidx);
      fprintf(stdout, "win_server%d.document.writeln('<CENTER><FORM>", server->CLidx);
      fprintf(stdout, "<INPUT TYPE=BUTTON NAME=CLOSE VALUE=CLOSE onClick=window.close()>");
      fprintf(stdout, "</FORM></CENTER>');\n");
      fprintf(stdout, "win_server%d.document.writeln('</PRE>');win_server%d.document.close();\"></TD>\n", server->CLidx, server->CLidx);
      fprintf(stdout, "</TR>\n");
      fflush(stdout);
   }
   fprintf(stdout, "</TABLE>\n");
   for (i = 0; i < server_list->total_selected; i++) {
      server = &(server_list->server[i]);
      fprintf(stdout, "<INPUT TYPE=HIDDEN NAME=\"report%d\" VALUE=\"<B>Detailed report for %s</B>\">\n", server->CLidx, server->server_ident);
   }
   fprintf(stdout, "</FORM>\n");
   fprintf(stdout, "<CENTER><A HREF=clctrl?CL_PRODUCT=%s>Return to Cluster Control Page</A></CENTER>\n", req->product);
   fprintf(stdout, "</BODY>\n");
   fflush(stdout);

   return(0);
}

CL_reportStatus(int style, CL_server *server, CL_request *request, int type, char *mesg)
{
   ENTER_CLUSTER_CRITICAL_SECTION;

   switch (style) {
   case CL_RPT:
      fprintf(stdout, "<SCRIPT LANUAGE=JavaScript>\n");
      fprintf(stdout, "document.CL_REPORT.server%d.value = \"%s\"\n", server->CLidx, mesg);
      fprintf(stdout, "</SCRIPT>\n");
      break;
   case CL_RPT_STDOUT:
      fprintf(stdout, "%s: %s\n", server->server_ident, mesg);
      break;
   default:
      break;
   }
   fflush(stdout);

   LEAVE_CLUSTER_CRITICAL_SECTION;

   return(0);
}

CL_reportDetails(int style, CL_server *server, CL_request *request, char *details)
{
   char *p, *from, *mesg;

   from = details;
   p = mesg = (char *)MALLOC(strlen(details)*2);
   while (*from != '\0') {
      if (!strncmp("<SCRIPT", from, strlen("<SCRIPT"))) {
	 from = strchr(from, '>') + 1;
	 continue;
      }
      if (!strncmp("</SCRIPT", from, strlen("</SCRIPT"))) {
	 from = strchr(from, '>') + 1;
	 continue;
      }
      if ((*from == LF) || (*from == CR)) {
	 from++;
	 continue;
      }
      if (*from == '"') {
         *p++ = '\\';
      }
      *p++ = *from++;
   }
   *p = *from;

   ENTER_CLUSTER_CRITICAL_SECTION;

   switch (style) {
   case CL_RPT:
      fprintf(stdout, "<SCRIPT LANGUAGE=JavaScript>\n");
      fprintf(stdout, "document.CL_REPORT.report%d.value += ", server->CLidx);
      fprintf(stdout, "\"<BR>%s\"\n", mesg);
      fprintf(stdout, "</SCRIPT>\n");
      break;
   case CL_RPT_STDOUT:
      fprintf(stdout, "%s: %s\n", server->server_ident, mesg);
      break;
   default:
      break;
   }
   fflush(stdout);

   LEAVE_CLUSTER_CRITICAL_SECTION;

   FREE(mesg);

   return(0);
}

NSAPI_PUBLIC int
nsAdmin_rexec(CL_server *server, CL_request *request, int rpt_style)
{
   PRFileDesc *sockd;		/* socket connected to remote admin server */
   bufstruct *buf;	/* private buffer to store returned data */
   char *line;		/* used to represent a line from buf */
   int retcode;		/* retcode from functions */

   INIT_CLUSTER_CRITICAL_SECTION;

   server->request_status = CL_FAIL;

   CL_reportStatus(rpt_style, server, request, CL_INFO, "Connecting to server");

   if ((sockd = make_http_request(server->nsAdmin_protocol,
			          server->nsAdmin_hostname,
   			          server->nsAdmin_port,
			          server->request,
				  60,
				  &retcode)) == NULL) {
      server->request_status = CLERR_NO_HANDSHAKE;
      CL_reportStatus(rpt_style, server, request, CL_FAIL, "DONE: command failed");
      CL_reportDetails(rpt_style, server, request, get_msg(retcode));
      return(-1);
   }

   buf = (bufstruct *) new_buffer(BUFSIZ);

   /*
    * Check HTTP header returned by remote admin server before continue
    */
   if ((retcode = parse_http_header(sockd, buf, ADMIN_ID)) < 0) {
      (void) delete_buffer(buf);
      CL_reportStatus(rpt_style, server, request, CL_FAIL, "DONE: command failed");
      switch (retcode) {
      case -1:
         server->request_status = CLERR_BAD_HEADER;
         CL_reportDetails(rpt_style, server, request, get_msg(CLERR_BAD_HEADER));
	 break;
      case -2:
         server->request_status = ERR_FAIL_AUTH;
         CL_reportDetails(rpt_style, server, request, get_msg(ERR_FAIL_AUTH));
	 break;
      }
      (void) delete_buffer(buf);
      return(-1);
   }

   CL_reportStatus(rpt_style, server, request, CL_INFO, "Collecting detail output");

   while ((line = (char *) get_line_from_fd(sockd, buf)) != (char *) NULL) {
      CL_reportDetails(rpt_style, server, request, line);
      if (request->expect != (char *) NULL) {
         if (!shexp_match(line, request->expect)) {
	    server->request_status = CL_SUCC;
	 }
      }
   }

   switch (server->request_status) {
   case CL_SUCC:
      CL_reportStatus(rpt_style, server, request, CL_SUCC, "DONE: command succeeded");
      break;

   case CL_FAIL:
   default:
      CL_reportStatus(rpt_style, server, request, CL_FAIL, "DONE: command might fail");
      break;
   }

   /* what to do when expect is not set? */

   (void) delete_buffer(buf);
   return(0);
}

void
CL_nsAdmin_rexec(void *data)
{
   CL_thread_data *d = (CL_thread_data *) data;

   (void) nsAdmin_rexec(d->server, d->req, CL_RPT);
   ENTER_CLUSTER_CRITICAL_SECTION;
   (d->req->servers_executing)--;
   LEAVE_CLUSTER_CRITICAL_SECTION;
}

NSAPI_PUBLIC int
CL_rexec(CL_request *req, CL_server_list *server_list, int *errcode)
{
   int i;			/* loop index */
   int req_len;			/* length of final request form */
   CL_server *server;		/* pointer to per server structure */
   char command[PATH_MAX];	/* where to locate the command */
   char *query_string;		/* url escaped form data */
   CL_thread_data *data;	/* data structure to be passed to thread */

   INIT_CLUSTER_CRITICAL_SECTION;

   req->servers_executing = server_list->total_selected;

   (void) CL_reportLayout(req, server_list);

   for (i = 0; i < server_list->total_selected; i++) {
      /*
       * Establish contact information about remote admin server which is
       * responsible for the target server.
       */
      server = &(server_list->server[i]);
      if (server->request_status != 0) {
	 continue;
      }

      /*
       * If remote admin is operating in secure mode, then this CGI
       * must also be invoked from a secure connection also.  We don't
       * allow insecure connection to initiate request to a secure
       * admin server.
       */
      if (((!strcasecmp(getenv("HTTPS"), "ON")) &&
          (strcasecmp("https", server->nsAdmin_protocol))) ||
	  (( strcasecmp(getenv("HTTPS"), "ON")) &&
          (!strcasecmp("https", server->nsAdmin_protocol))))
      {
	    server->request_status = CLERR_NO_MIXSEC;
            CL_reportStatus(CL_RPT, server, req, CL_FAIL, "DONE: command failed");
            CL_reportDetails(CL_RPT, server, req, get_msg(CLERR_NO_MIXSEC));
	    continue;
      }

      /*
       * Allow user to specific absolute path to the command.  If action
       * start with '/', then user is specifying where to find the command
       * in the 2.0 admin server style: /<server>/bin/<command>.
       */
      if (req->action[0] == '/') {
         sprintf(command, "%s", req->action);
      } else {
         sprintf(command, "/%s/%s", server->server_ident, req->action);
      }

      /*
       * Add CL_TARGET=<serverid> to form_data for use by remote CGI.
       * And uuescape form_data before sending it out!
       */
      if (req->form_data[0] == '\0') {
         sprintf(req->form_data, "%s=%s", "CL_TARGET", server->server_ident);
      } else {
         sprintf(req->form_data, "%s&%s=%s", req->form_data, "CL_TARGET", server->server_ident);
      }
      query_string = (char *) util_url_escape(NULL, req->form_data);
      FREE(req->form_data);
      req->form_data = query_string;

      /*
       * Construct remote request on per server information.
       * Request format including header information:
       *	<METHOD> /<server-id>/<ACTION> HTTP/1.0
       *	NS-Interactive: No
       *	Authorization: basic <auth-info>
       */
      if (!strcmp("GET", req->method)) {
         req_len = strlen(CL_GET_REQ_HEADER) +
                   strlen(req->method) +
                   strlen(command) +
		   strlen((char *) req->auth_info) +
		   4 + /* num of digits to represent strlen(form_data) */
		   strlen(req->form_data);
         if ((server->request = (char *) MALLOC(req_len)) == (char *) NULL) {
	    *errcode = CLERR_NO_MEM;
            CL_reportStatus(CL_RPT, server, req, CL_FAIL, "DONE: command failed");
            CL_reportDetails(CL_RPT, server, req, get_msg(CLERR_NO_MEM));
	    continue;
	 }
	 if (strchr(command, '?')) {
	    sprintf(server->request, CL_GET_REQ_HEADER,
	 		  req->method,
			  command, '&', req->form_data,
			  req->auth_info);
	 } else {
	    sprintf(server->request, CL_GET_REQ_HEADER,
	 		  req->method,
			  command, '?', req->form_data,
			  req->auth_info);
	 }
      } else if (!strcmp("POST", req->method)) {
         req_len = strlen(CL_POST_REQ_HEADER) +
                   strlen(req->method) +
                   strlen(command) +
		   strlen((char *) req->auth_info) +
		   4 + /* num of digits to represent strlen(form_data) */
		   strlen(req->form_data);
         if ((server->request = (char *) MALLOC(req_len)) == (char *) NULL) {
	    *errcode = CLERR_NO_MEM;
            CL_reportStatus(CL_RPT, server, req, CL_FAIL, "DONE: command failed");
            CL_reportDetails(CL_RPT, server, req, get_msg(CLERR_NO_MEM));
	    continue;
	 }
	 sprintf(server->request, CL_POST_REQ_HEADER,
	 		  req->method,
			  command,
			  req->auth_info,
		          strlen(req->form_data),
			  req->form_data);
      }

      data = (CL_thread_data *) MALLOC(sizeof(CL_thread_data));
      data->server = server;
      data->req = req;
      data->errcode = 0;

      if (systhread_start(SYSTHREAD_DEFAULT_PRIORITY, 262144, CL_nsAdmin_rexec,
                          data) == SYS_THREAD_ERROR) {
         (void) FREE(server->request);
      }
   }

   /*
    * Until we have something like systhread_join(), we have to do busy
    * looping to detect all threads stop executing.
    */
   while (1) {
      if (req->servers_executing) {
#ifdef	XP_WIN32
	 systhread_sleep(1000);
#else	/* XP_UNIX */
         sleep(1);
#endif	/* XP_WIN32 */
      } else {
	 break;
      }
   }
   return(0);
}
