/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "libadminutil/admutil.h"
#include "libadmin/ds_base.h"

/* -------------------------------------------------------- 
   		dynamic array
   An array object can grow automatically.
   Use defined function to insert array elements.
   --
   Create	create_dyn_array(int size, int delta)
   Destroy	destroy_dyn_array(dyn_array *ary, void (*free_elm)())
   Insert	dyn_add(dyn_array *ary, void *data)
   Delete	dyn_delete(dyn_array *ary, int inx)
   Get		dyn_data(dyn_array *ary, int index)
   Size		dyn_size(dyn_array *ary)
   -------------------------------------------------------- */

/* return a dynamic array with initial size, once overflow
   size will be increased by delta.
*/
dyn_array *create_dyn_array(int size, int delta)
{
    dyn_array *ary;
 
    ary = (dyn_array *)MALLOC(sizeof(dyn_array));
    /* initial size */
    ary->size = size;
    ary->entry = MALLOC(sizeof(void *) * size);
    ary->delta = delta;
    ary->index = 0;
 
    return(ary);
}
 
/* destroy the dynamic array, free every element by calling
   provided free function and free its own memeory.
*/
void destroy_dyn_array(dyn_array *ary, void (*free_elm)())
{
    int i;
 
    /* free each element */
    if(free_elm)
      for(i =0; i < ary->index; ++i) { free_elm(ary->entry[i]); } 
    
    FREE(ary->entry);
    FREE(ary);
}
 
/* put an element into dynamic array, if overflow,
   increase array size by delta. 
*/
int dyn_add(dyn_array *ary, void *data)
{
    if(ary->index >= ary->size) { /* overflow, enlarge size */
        ary->size += ary->delta;
        REALLOC(ary->entry, sizeof(void *) * ary->size);
    }
    ary->entry[ary->index] = data;
    return(++(ary->index));
}
 
/* get the address of an element from a dynamic array,
   return NULL if index out of range. 
*/
void *dyn_data(dyn_array *ary, int index)
{
    if(index >= ary->index) return(NULL);
    return(ary->entry[index]);
}
 
/* return number of elements in the array. 
*/
int dyn_size(dyn_array *ary)
{
    return(ary->index);
}

/* remove the specified element from array,
   move following elements up. 
*/
void dyn_delete(dyn_array *ary, int inx)
{
    int i;

    if(inx >= ary->index) return;
    FREE(ary->entry[inx]);
    ary->entry[inx] = NULL;
    for(i = inx; i < (ary->index - 1); ++i) {
	ary->entry[i] = ary->entry[i+1];
    }
    --(ary->index);
}	
/* -------------------------------------------------------- */


/* -------------------------------------------------------- 
		utilities for batman admin 
   There are 3 different configuration for batman, catalog server.
   Following routines privide a way to hide their differences.
   Some routines return a pointer to allocated memory, it is not
   good for memory management, no explicity for caller to free them.
   Anyway, this is just a simple cgi program, run and quit, no
   real harm will result.
   --
   Index	module_index(char *module)
   Name		module_name(int which
   Dir		module_dir(int which)
   Http Home	get_batman_http_dir(int whichone, int module)
   Home		get_ds_dir(int whichone, int module)
   Log		get_ds_logs_dir(int whichone, int module)
   Conf		get_ds_conf_dir(int whichone, int module)
   Pid		get_ds_pid_file(int whichone, int module)
   -------------------------------------------------------- */

/* map the module string into a index constant.
   return 0, if unmatched.
*/
int module_index(char *module)
{
    if(!module) return(0);

    if(!strcmp(module, "cs")) {
      return(CATALOG_SERVER);
    }
    else if(!strcmp(module,"rds")) {
      return(RDS_SERVER);
    }
    else if(!strcmp(module, "minirds")) {
      return(MINI_SERVER);

    }
    return(0);
}

/* return module string name according the index.
   return NULL, if out of range.
*/
char *module_name(int which)
{
      switch(which) {
        case CATALOG_SERVER:
          return("Catalog Server");
        case RDS_SERVER:
          return("Resource Description Server");
        case MINI_SERVER:
          return("AutoCatalog");
      }
      return(NULL);
}

/* The directory name for the module,
   NOT full path, only the last component.
*/
char *module_dir(int module)
{
        switch(module) {
          case CATALOG_SERVER:
            return("catalog");
          case RDS_SERVER:
            return("rds");
          case MINI_SERVER:
            return("ns-rds-mini");
        }
        return(NULL);
}

char *get_ns_home()
{
    char *srvroot, *slash;

    srvroot = get_mag_var("#ServerRoot");
    if(!srvroot)  {
        report_error(INCORRECT_USAGE, "No server root variable",
                     "The magnus.conf variable #ServerRoot was "
                     "not set.  Please set the value of your server "
                     "root through the administrative forms.");
    }
    CONVERT_TO_HTTP_FORMAT(srvroot)
    slash = strrchr(srvroot, FILE_PATHSEP);
    if(slash) *slash = '\0';

    return(srvroot);
}

char *get_server_home()
{
    char *srvroot;

    srvroot = get_mag_var("#ServerRoot");
    if(!srvroot)  {
        report_error(INCORRECT_USAGE, "No server root variable",
                     "The magnus.conf variable #ServerRoot was "
                     "not set.  Please set the value of your server "
                     "root through the administrative forms.");
    }
    CONVERT_TO_HTTP_FORMAT(srvroot)
    return(srvroot);
}



/* The full path to batman http server home directory
*/
char *get_batman_http_dir(int which, int module)
{
    char *confdir = NULL;
    char *home;
 
    home = get_ns_home();
    confdir = (char *) MALLOC(strlen(home) + 30);
    sprintf(confdir, "%s/bin/%s", home, module_dir(module));
    return confdir;
}

/* The full path to module home direcory.
*/
char *get_ds_dir(int whichone, int module)
{
    char *home, *confdir = NULL;

    if(module == MINI_SERVER) {
      home = get_ns_home();
      confdir = (char *) MALLOC(strlen(home) + 60);
      sprintf(confdir, "%s%cplugins%c%s%c", home, FILE_PATHSEP,
	FILE_PATHSEP, module_dir(module), FILE_PATHSEP);
    }
    else {
      home = get_server_home();
      confdir = (char *) MALLOC(strlen(home) + 4);
      sprintf(confdir, "%s%c", home, FILE_PATHSEP);
    }
    return(confdir);
}

/* log directory */
char *get_ds_logs_dir(int whichone, int module)
{
    char *tmp = get_ds_dir(whichone, module);
    char *confdir = NULL;

    confdir = (char *) MALLOC(strlen(tmp) + strlen("logs") + 4);
    sprintf(confdir, "%slogs%c", tmp, FILE_PATHSEP);
    FREE(tmp);
    return(confdir);
}

/* Full path for pid file */
char *get_ds_pid_file(int whichone, int module)
{
    char *tmp = get_ds_logs_dir(whichone, module);
    char *confdir = NULL;

    confdir = (char *) MALLOC(strlen(tmp) + strlen("robot.pid") + 4);
    sprintf(confdir, "%srobot.pid", tmp);
    FREE(tmp);
    return(confdir);
}

/* Full path to conf directory */
char *get_ds_conf_dir(int whichone, int module)
{
    char *tmp = get_ds_dir(whichone, module);
    char *confdir = NULL;
 
    confdir = (char *) MALLOC(strlen(tmp) + strlen("config") + 4);
    sprintf(confdir, "%sconfig%c", tmp, FILE_PATHSEP);
    FREE(tmp);
    return(confdir);
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- 
		-- robot_cf.conf --
   Routines for handling robot_cf.conf which are only used by 
   batman admin. User can change the configuration files for
   robot, here is the place to save user's selections.
   --
   1st line for process.conf, full path
   2nd line for filter.conf, full path
   if file isn't exist, create one.
   --
   Put		put_robot_conf(int module, char *process, char *filter)
   Get		get_robot_conf(int module, char *process, char *filter)
   -------------------------------------------------------- */

/* save the selections from robot_cf.conf */
int put_robot_conf(int module, char *process, char *filter)
{
        char robot_cf[BIG_LINE];
        char *dir = get_ds_conf_dir(0, module);
        FILE *p = NULL;
 
        sprintf(robot_cf,"%srobot_cf.conf", dir);
        p = fopen(robot_cf, "w");
        if(p) fprintf(p,"%s\n%s\n", process, filter);
	else return(0);
        fclose(p);
 
        /* update robot.conf for pointing correct filter.conf */
        sprintf(robot_cf,"%srobot.conf", dir);
        p = fopen(robot_cf, "w");
        if(p) fprintf(p,"LoadObjects %s\n", filter);
        fclose(p);
	return(1);
}

/* get the selections from robot_cf.conf */
void get_robot_conf(int module, char *process, char *filter)
{
        char robot_cf[BIG_LINE];
        char *dir = get_ds_conf_dir(0, module);
        FILE *p = NULL;
        int err_flag = 1;
 
        sprintf(robot_cf,"%srobot_cf.conf", dir);
        if(file_exists(robot_cf)) {
          p = fopen(robot_cf, "r");
          if(p) if ((fscanf(p, "%s\n", process)) > 0)
            if ((fscanf(p, "%s\n", filter))  > 0) err_flag = 0;
          fclose(p);
        }
        if(err_flag) {  /* set default value */
          sprintf(process,"%sprocess.conf", dir);
          sprintf(filter,"%sfilter.conf", dir);
	  put_robot_conf(module, process, filter);
        }
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- 
		-- server.conf --
   Routines for handling server.conf which are only used by batman.
   It is similiar to magnus.conf or obj.conf, name-value pairs.
   --
   Where	get_server_conf(int module)
   Read		read_server_conf(int module)
   Write	write_server_conf(int module)
   Get		get_server_var(int module, char *name)
   Put		set_server_var(int module, char *name, char *val)
   -------------------------------------------------------- */

/* Hide the char which is for continuing to next line.
   Because libadmin doesn't take care of "\" at end of line,
   hide it to make processing working.
*/
static char stub_line[BIG_LINE];
char *hide_esc(char *str)
{
    int l = strlen(str) - 1;
    if( l > 0 && str[l] == '\\') {
	strcpy(stub_line, str);
	stub_line[l] = '\0';
	return(stub_line);
    }
    return(str);
}

/* Return full path to server.conf.
   At first calling, it places the result in global.
   This routine has to be initialized before real usage.
*/
char *get_server_conf(int module)
{
    static char *server_conf = NULL;
    char *ds_conf_dir;

    if(!server_conf) {
      ds_conf_dir = get_ds_conf_dir(0, module);
      server_conf = MALLOC(strlen(ds_conf_dir) + 20);
      sprintf(server_conf, "%sserver.conf", ds_conf_dir);
    }
    return(server_conf);
}

/* read in the server.conf,
   it only load once for each run.
*/
static char ***srvconfarray = NULL;
static char **srvconf = NULL;
void read_server_conf(int module)
{
   char magpath[1024];

   if(srvconfarray == NULL)  {
       int howmany = make_conflist();
       register int x=0;
 
       srvconfarray = (char ***) MALLOC(howmany*sizeof(char **));
       for(x=0; x < howmany; x++)  {
           sprintf(magpath, "%s", get_server_conf(module));
           srvconfarray[x] = NULL;
           if(file_exists(magpath)) srvconfarray[x] = scan_tech(magpath);
           if(!srvconfarray[x]) {
             srvconfarray[x] = (char **) MALLOC(1024*sizeof(char *));
             srvconfarray[x][0] = NULL;
           }
       }
       /* Set to default for subsequent stuffs */
       srvconf = srvconfarray[0];
   }
}
 
/* write server.conf back.
*/
void write_server_conf(int module)
{
    register int x = 0;
    FILE *f;
    char magpath[512];
    int y, howmany;
 
    howmany = make_conflist();
 
    for(y=0; y<howmany; y++)  {
        sprintf(magpath, "%s", get_server_conf(module));
        srvconf = srvconfarray[y];
        f = fopen_l(magpath, "w");
        if(f == NULL)
           report_error(FILE_ERROR, magpath,
                    "Could not write configuration file server.conf.");
 
        for(x=0; srvconf[x]; x++) {
            if(strcmp(srvconf[x], "\n"))
                fprintf(f, "%s\n", srvconf[x]);
            else
                fprintf(f, "%s", srvconf[x]);
        }
        fclose_l(f);
 
        set_commit(y, 1);
    }
}

/* Return the value part of specified name.
*/
char *get_server_var(int module, char *name)
{
    register int x;
    pblock *answer;
    char *tmp;
 
    read_server_conf(module);
 
    for(x=0; srvconf[x]; x++)  {
       answer = pblock_create(PARAMETER_HASH_SIZE);
       pblock_str2pblock(hide_esc(srvconf[x]), answer);
 
       while(1) {
         int l;
         l = strlen(srvconf[x]) - 1;
         if(l < 0) break;
         if(srvconf[x][l] != '\\') break;
         if(!srvconf[++x]) break;
         pblock_str2pblock(hide_esc(srvconf[x]), answer);
       }
 
       if((tmp = pblock_findval(name, answer)))  {
         tmp = STRDUP(tmp);
         pblock_free(answer);
         return(tmp);
       }
       pblock_free(answer);
    }
    return NULL;
}
 
/* read in the server.conf (actually reading only at first calling), 
   put a new pair of name/vale, overwrite if name exist.
*/
int set_server_var(int module, char *name, char *val)
{
    int x;
    pblock *answer;
    char *tmp;
    pb_param *pp;
 
    read_server_conf(module);
 
    for(x=0; srvconf[x]; x++)  {
       answer = pblock_create(PARAMETER_HASH_SIZE);
       pblock_str2pblock(hide_esc(srvconf[x]), answer);
 
       while(1) {
         int l;
         l = strlen(srvconf[x]) - 1;
         if(l < 0) break;
         if(srvconf[x][l] != '\\') break;
         if(!srvconf[++x]) break;
         pblock_str2pblock(hide_esc(srvconf[x]), answer);
       }
 
       if((pp = pblock_find(name, answer))) {
         if(pp->value) FREE(pp->value);
         pp->value = STRDUP(val);
         srvconf[x] = pblock_pblock2str(answer, NULL);
         pblock_free(answer);
         return(0);
       }
       pblock_free(answer);
    }
 
    /* add a new line into */
    tmp = MALLOC(strlen(name)+strlen(val)+4);
    sprintf(tmp,"%s=%s", name, val);
    srvconf[x++] = tmp;
    srvconf[x] = NULL;
 
    return(1);
}
/* -------------------------------------------------------- */

/* -------------------------------------------------------- 
		-- OLD ROUTINES --
   Following routines are needed only for working with server 1.12.
   Because supporting riutines for server 1.12 were not available,
   so we need to wrote them for ourselves.
   The main problem is that it abandoned the Registry on NT after
   1.12. It also moved stuffs from magnus.conf to obj.conf some
   point after server1.12.
   -------------------------------------------------------- */
 
#ifdef USE112
/* following codes are for 1.12 only */
static char ***magconfarray = NULL;
static char **magconf = NULL;
 
 
#ifndef XP_WIN32
void old_delete_mag_var(int offset)
{
    int maglen;
    register int x;
    int y, l, do_next;
 
    for(maglen=0; magconf[maglen]; maglen++);
 
    /* check to see if continue to next line */
    x = offset;
    do_next = 1;
    while(do_next) {
      do_next = 0;
      l = strlen(magconf[x]) - 1;
      if(l >= 0 && magconf[x][l] == '\\') do_next = 1;
      FREE(magconf[x++]);
    }
 
    for(y = offset; x <= maglen; x++, y++)  {
        magconf[y] = magconf[x];
    }
 
    magconf[maglen] = NULL;
}
#endif

 
void old_read_mag_conf(void)
{
   char magpath[512];
#ifdef USE_ADMSERV
   if(srvconfarray == NULL)  {
       int howmany = make_conflist();
       register int x=0;
 
#ifndef XP_WIN32
       magconfarray = (char ***) MALLOC(howmany*sizeof(char **));
#endif
       srvconfarray = (char ***) MALLOC(howmany*sizeof(char **));
       for(x=0; x < howmany; x++)  {
#ifndef XP_WIN32
           sprintf(magpath, "%s%s", CONFDIR(x), MAGNUS_CONF);
	   /* fix bug of crashing while redaing empty file */
	   magconfarray[x] = NULL;
           magconfarray[x] = scan_tech(magpath);
	   if(!magconfarray[x]) {
             magconfarray[x] = (char **) MALLOC(1024*sizeof(char *));
             magconfarray[x][0] = NULL;
	   }
#endif
	   /* patch for server.conf */
           sprintf(magpath, "%s", get_server_conf(0));
	   srvconfarray[x] = NULL;
	   if(file_exists(magpath)) srvconfarray[x] = scan_tech(magpath);
	   if(!srvconfarray[x]) {
             srvconfarray[x] = (char **) MALLOC(1024*sizeof(char *));
             srvconfarray[x][0] = NULL;
	   }

#ifndef XP_WIN32
           if(!magconfarray[x])
               report_error(FILE_ERROR, magpath,
                       "Could not read configuration file magnus.conf.");
#endif
       }
       /* Set to default for subsequent stuffs */
#ifndef XP_WIN32
       magconf = magconfarray[0];
#endif
       srvconf = srvconfarray[0];
   }
#else
   if(magconf == NULL) {
 
       sprintf(magpath, "%s%s", CONFDIR(0), MAGNUS_CONF);
       magconf = scan_tech(magpath);

       /* patch for server.conf */
       sprintf(magpath, "%s", get_server_conf(0));
       if(file_exists(magpath))
         srvconf = scan_tech(magpath);
       ielse
         srvconf = NULL;
   }
   if(magconf == NULL)
       report_error(FILE_ERROR, magpath,
                    "Could not read configuration file magnus.conf.");
#endif
 }
 
/* add patch for server.conf,
   put all the lines begin with "Init" to server.conf
*/
 
void old_write_mag_conf(void)
{
    register int x = 0;
    FILE *f;
    char magpath[512];
 
#ifdef USE_ADMSERV
#ifndef XP_WIN32
    register int y, howmany;
 
    howmany=make_conflist();
 
    for(y=0; y<howmany; y++)  {
        sprintf(magpath, "%s%s", CONFDIR(y), MAGNUS_CONF);
        conf_backup(magpath, BK_MAGNUS, y);
        magconf = magconfarray[y];
 
        f = fopen_l(magpath, "w");
        if(f == NULL)
           report_error(FILE_ERROR, magpath,
                    "Could not write configuration file magnus.conf.");
 
        for(x=0; magconf[x]; x++) {
            if(strcmp(magconf[x], "\n"))
                fprintf(f, "%s\n", magconf[x]);
            else
                fprintf(f, "%s", magconf[x]);
	}
        fclose_l(f);
        set_commit(y, 1);
    }
#endif
#else
    sprintf(magpath, "%s%s", CONFDIR(0), MAGNUS_CONF);
    f = fopen_l(magpath, "w");
 
    if(f == NULL)
       report_error(FILE_ERROR, magpath,
                    "Could not write configuration file magnus.conf.");
 
    for(x=0; magconf[x]; x++)
        if(strcmp(magconf[x], "\n"))
            fprintf(f, "%s\n", magconf[x]);
        else
            fprintf(f, "%s", magconf[x]);
 
    fclose_l(f);
#endif
}

/* Deal with those crazy init directives. */
#ifdef XP_WIN32
pblock *old_get_mag_init(char *fn)
{
    int i;
    char *http = get_http_dir();
    pblock *answer;
    char buf[1024];
    char val[1024];
    LONG fOpen = ERROR_SUCCESS;
    HKEY hkey;
    DWORD len, lType;
 
 
    for(i = 1; fOpen == ERROR_SUCCESS; ++i) {
      sprintf(buf,
        "SOFTWARE\\Netscape\\%s\\CurrentVersion\\Startup\\InitFunction%i",
        http, i);
      fOpen=RegOpenKeyEx(HKEY_LOCAL_MACHINE, buf, 0, KEY_READ, &hkey);
      if(fOpen == ERROR_SUCCESS) {
        len = 1024;
        if(RegQueryValueEx(hkey,"fn",NULL,&lType,val,&len) == ERROR_SUCCESS)  {
          if(!strcasecmp(fn, val)) {
            RegCloseKey(hkey);
	    /* only return a pointer to indicate existence */
            answer = pblock_create(PARAMETER_HASH_SIZE);
            return(answer);
          }
        }
      }
      RegCloseKey(hkey);
    }
    return(NULL);
}

#else
pblock *old_get_mag_init(char *fn)
{
    register int x;
    pblock *answer;
    char *tmp;
 
    old_read_mag_conf();
 
    for(x=0; magconf[x]; x++)  {
        if(!strncasecmp(magconf[x], "Init ", 5))  {
            answer = pblock_create(PARAMETER_HASH_SIZE);
            pblock_str2pblock(hide_esc(&magconf[x][5]), answer);

            while(1) {
                  int l;
                  l = strlen(magconf[x]) - 1;
                  if(l < 0) break;
                  if(magconf[x][l] != '\\') break;
                  if(!magconf[++x]) break;
                  pblock_str2pblock(hide_esc(magconf[x]), answer);
            }

            if( (tmp = pblock_findval("fn", answer)) )  {
                if(!strcmp(tmp, fn))
                    return answer;
            }
            pblock_free(answer);
        }
    }
    return NULL;
}
#endif

#ifdef XP_WIN32
pblock *old_get_mag_init_ext(char *fn, char *key, char *data)
{
    int i;
    char *http = get_http_dir();
    pblock *answer;
    char buf[1024];
    char val[1024];
    LONG fOpen = ERROR_SUCCESS;
    HKEY hkey;
    DWORD len, lType;
 
 
    for(i = 1; fOpen == ERROR_SUCCESS; ++i) {
      sprintf(buf,
        "SOFTWARE\\Netscape\\%s\\CurrentVersion\\Startup\\InitFunction%i",
        http, i);
      fOpen=RegOpenKeyEx(HKEY_LOCAL_MACHINE, buf, 0, KEY_READ, &hkey);
      if(fOpen == ERROR_SUCCESS) {
        len = 1024;
        if(RegQueryValueEx(hkey,"fn",NULL,&lType,val,&len) == ERROR_SUCCESS)  {
          if(!strcasecmp(fn, val)) {
            len = 1024;
            if(RegQueryValueEx(hkey,key,NULL,&lType,val,&len) == ERROR_SUCCESS) {
              if(!strcasecmp(data, val)) {
                RegCloseKey(hkey);
                /* only return a pointer to indicate existence */
                answer = pblock_create(PARAMETER_HASH_SIZE);
                return(answer);
	      }
	    }
          }
        }
      }
      RegCloseKey(hkey);
    }
    return(NULL);
}

#else
pblock *old_get_mag_init_ext(char *fn, char *key, char *val)
{
    register int x;
    pblock *answer;
    char *tmp, *tmp2;
 
    old_read_mag_conf();
 
    for(x=0; magconf[x]; x++)  {
        if(!strncasecmp(magconf[x], "Init ", 5))  {
            answer = pblock_create(PARAMETER_HASH_SIZE);
            pblock_str2pblock(hide_esc(&magconf[x][5]), answer);

	    /* check to see if continue to next line */
	    while(1) {
		  int l;
		  l = strlen(magconf[x]) - 1;
		  if(l < 0) break;
		  if(magconf[x][l] != '\\') break;
		  if(!magconf[++x]) break;
                  pblock_str2pblock(hide_esc(magconf[x]), answer);
	    }
 
            if( (tmp  = pblock_findval("fn", answer)) &&
		(tmp2 = pblock_findval(key,  answer)) )  {
                if(!strcmp(tmp, fn) && !strcmp(tmp2, val))
                    return answer;
            }
            pblock_free(answer);
        }
    }
    return NULL;
}
#endif

 
#ifdef XP_WIN32

void old_set_mag_init(char *fn, char *key, char *val, int nargs, ...)
{
    va_list arglist;
    char *varname, *varval;
    int i, y;
    char *http = get_http_dir();
    char buf[1024];
    char data[1024];
    LONG fOpen = ERROR_SUCCESS;
    HKEY hkey;
    DWORD len, lType;
 
    for(i = 1; fOpen == ERROR_SUCCESS; ++i) {
      sprintf(buf,
        "SOFTWARE\\Netscape\\%s\\CurrentVersion\\Startup\\InitFunction%i",
        http, i);
      fOpen=RegOpenKeyEx(HKEY_LOCAL_MACHINE, buf, 0, KEY_READ, &hkey);
      if(fOpen == ERROR_SUCCESS) {
        len = 1024;
        if(RegQueryValueEx(hkey,"fn",NULL,&lType,data,&len) == ERROR_SUCCESS)  {
          if(!strcasecmp(fn, data)) {
	    if(!key || !val) break;
            len = 1024;
            if(RegQueryValueEx(hkey,key,NULL,&lType,data,&len) == ERROR_SUCCESS) {
              if(!strcasecmp(val, data)) break;
            }
          }
        }
      }
      RegCloseKey(hkey);
      if(fOpen != ERROR_SUCCESS) break;
    }
    if(fOpen != ERROR_SUCCESS) { /* not found, create a new key */
      sprintf(buf,
        "SOFTWARE\\Netscape\\%s\\CurrentVersion\\Startup\\InitFunction%i",
        http, i);
      fOpen = RegCreateKeyEx(HKEY_LOCAL_MACHINE, buf, 0, NULL, 
	REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL, &hkey, &len);
    }

    if(fOpen != ERROR_SUCCESS)
	report_error(SYSTEM_ERROR, "Write registry", "failed!");

    /* start putting values into */
    RegSetValueEx(hkey, "fn", 0, REG_SZ, fn, strlen(fn)+1);
    va_start(arglist, nargs);
    y = 0;
    while (y < nargs)  {
      varname = STRDUP(va_arg(arglist, char *));
      varval = STRDUP(va_arg(arglist, char *));
      RegSetValueEx(hkey, varname, 0, REG_SZ, varval, strlen(varval)+1);
      y += 2;
    }
    va_end(arglist);
    RegCloseKey(hkey);
}

#else
void old_set_mag_init(char *fn, char *key_nam, char *key_val, int nargs, ...)
{
    register int x, y, found = 0;
    pblock *pb;
    char *tmp1, *tmp2, *line;
    va_list arglist;
    char *varname, *varval;
    char *pbstring;
    int numservs = make_conflist(), z;
 
    old_read_mag_conf();
 
    if (key_nam && !key_val)
        key_val = "";
 
    for(z=0; z<numservs; z++)  {
        magconf=magconfarray[z];
 
        for(x=0, found = 0; magconf[x]; x++)  {
            if(!strncasecmp(magconf[x], "Init ", 5))  {
                pb = pblock_create(PARAMETER_HASH_SIZE);
                pblock_str2pblock(&magconf[x][5], pb);
                if ( (tmp1 = pblock_findval("fn", pb)) &&
                     (!strcmp(tmp1, fn)) &&
                     (!key_nam || ((tmp2 = pblock_findval(key_nam, pb)) &&
                                   (!strcmp(tmp2, key_val)))) )  {
                    found = 1;
                    break;
                }
                pblock_free(pb);
            }
        }
        if(found)  {
            pb_param *prm;
            y = 0;
 
            va_start(arglist, nargs);
            while (y < nargs)  {
                varname = STRDUP(va_arg(arglist, char *));
                varval = STRDUP(va_arg(arglist, char *));
 
                if((prm = pblock_find(varname, pb)))  {
                    if(strcmp(varval, "NONE"))  {
                        free(prm->value);
                        prm->value = varval;
                    }  else  {
                        param_free(pblock_remove(varname, pb));
                    }
                }  else  {
                    if(strcmp(varval, "NONE"))
                        pblock_nvinsert(varname, varval, pb);
                    /* else shrug */
                }
                y += 2;
            }
            va_end(arglist);
 
            pbstring = pblock_pblock2str(pb, NULL);
	    line = MALLOC(strlen(pbstring) + 20);
            sprintf(line, "Init %s", pbstring);
            magconf[x] = STRDUP(line);
	    FREE(line);
        }  else  {  /* !found */
            pb = pblock_create(PARAMETER_HASH_SIZE);
 
            x = 0;
            va_start(arglist, nargs);
            while(x < nargs)  {
                varname = STRDUP(va_arg(arglist, char *));
                varval = STRDUP(va_arg(arglist, char *));
 
                if(strcmp(varval, "NONE"))
                    pblock_nvinsert(varname, varval, pb);
                x+=2;
            }
            va_end(arglist);
 
            pbstring = pblock_pblock2str(pb, NULL);
	    line = MALLOC(strlen(pbstring) + strlen(fn) + 20);
            sprintf(line, "Init fn=%s %s", fn, pbstring);
            for(x=0; magconf[x]; x++);
            magconf = grow_strlist(magconf, x+2);

	    /* big bug, put the changes pointer back, erhyuan */
            magconfarray[z] = magconf;

            magconf[x] = STRDUP(line);
            magconf[x+1] = NULL;
	    FREE(line);
        }
    }
    old_write_mag_conf();
}
#endif
 
 
 
#ifdef XP_WIN32
void old_delete_mag_init(char *fn)
{
    int i;
    char *http = get_http_dir();
    char buf[1024];
    char val[1024];
    LONG fOpen = ERROR_SUCCESS;
    HKEY hkey;
    DWORD len, lType;
 
 
    for(i = 1; fOpen == ERROR_SUCCESS; ++i) {
      sprintf(buf,
        "SOFTWARE\\Netscape\\%s\\CurrentVersion\\Startup\\InitFunction%i",
        http, i);
      fOpen=RegOpenKeyEx(HKEY_LOCAL_MACHINE, buf, 0, KEY_READ, &hkey);
      if(fOpen == ERROR_SUCCESS) {
        len = 1024;
        if(RegQueryValueEx(hkey,"fn",NULL,&lType,val,&len) == ERROR_SUCCESS)  {
          if(!strcasecmp(fn, val)) {
            RegCloseKey(hkey);
	    RegDeleteKey(HKEY_LOCAL_MACHINE, buf);
	    break;
          }
        }
      }
      RegCloseKey(hkey);
    }
}
#else
void old_delete_mag_init(char *fn)
{
    register int x, found = 0;
    pblock *pb;
    char *tmp;
    int numservs=make_conflist(), y=0;
    int target;
 
    old_read_mag_conf();
 
    for(y=0; y<numservs; y++)  {
        magconf = magconfarray[y];
 
        for(x=0, found = 0; magconf[x]; x++)  {
	    target = x;
            if(!strncasecmp(magconf[x], "Init ", 5))  {
                pb = pblock_create(PARAMETER_HASH_SIZE);
                pblock_str2pblock(hide_esc(&magconf[x][5]), pb);

                /* check to see if continue to next line */
                while(1) {
                  int l;
                  l = strlen(magconf[x]) - 1;
                  if(l < 0) break;
                  if(magconf[x][l] != '\\') break;
                  if(!magconf[++x]) break;
                  pblock_str2pblock(hide_esc(magconf[x]), pb);
                }

                if( (tmp = pblock_findval("fn", pb)) )  {
                    if(!strcmp(tmp, fn))  {
                        found = 1;
			x = target;
                        break;
                    }
                }
                pblock_free(pb);
            }
        }
        if(found) old_delete_mag_var(x);
    }
    old_write_mag_conf();
}
#endif
 
#if 0
void old_delete_specific_mag_init(char *fn, char *key_name, char *key_val)
{
    register int x, found = 0;
    pblock *pb;
    char *tmp;
    int numservs=make_conflist(), y=0;
    int target;
 
    old_read_mag_conf();
 
    for(y=0; y<numservs; y++)  {
        magconf = magconfarray[y];
 
        for(x=0, found = 0; magconf[x]; x++)  {
	    target = x;
            if(!strncasecmp(magconf[x], "Init ", 5))  {
                pb = pblock_create(PARAMETER_HASH_SIZE);
                pblock_str2pblock(hide_esc(&magconf[x][5]), pb);

		/* check to see if continue to next line */
		while(1) {
		  int l;
		  l = strlen(magconf[x]) - 1;
		  if(l < 0) break;
		  if(magconf[x][l] != '\\') break;
		  if(!magconf[++x]) break;
                  pblock_str2pblock(hide_esc(magconf[x]), pb);
		} 

                if( (tmp = pblock_findval("fn", pb)) )  {
                    if(!strcmp(tmp, fn))  {
                      if( (tmp = pblock_findval(key_name, pb)) ) {
                          if(!strcmp(tmp, key_val)) {
			      x = target;
                              found = 1;
                              break;
                          }
                      }
                    }
                }
                pblock_free(pb);
            }
        }
        if(found) old_delete_mag_var(x);
    }
    old_write_mag_conf();
}
#endif

#ifdef XP_WIN32
void delete_mag_init_fun(char *fn, char *key, char *val)
{
    int i;
    char *http = get_http_dir();
    char buf[1024];
    char data[1024];
    LONG fOpen = ERROR_SUCCESS;
    HKEY hkey;
    DWORD len, lType;
 
 
    for(i = 1; fOpen == ERROR_SUCCESS; ++i) {
      sprintf(buf,
        "SOFTWARE\\Netscape\\%s\\CurrentVersion\\Startup\\InitFunction%i",
        http, i);
      fOpen=RegOpenKeyEx(HKEY_LOCAL_MACHINE, buf, 0, KEY_READ, &hkey);
      if(fOpen == ERROR_SUCCESS) {
        len = 1024;
        if(RegQueryValueEx(hkey,"fn",NULL,&lType,data,&len) == ERROR_SUCCESS)  {
          if(!strcasecmp(fn, data)) {
            len = 1024;
            if(RegQueryValueEx(hkey,key,NULL,&lType,data,&len) == ERROR_SUCCESS) {
              if(strstr(data, val)) {
                RegCloseKey(hkey);
	        RegDeleteKey(HKEY_LOCAL_MACHINE, buf);
	        break;
              }
            }
          }
        }
      }
      RegCloseKey(hkey);
    }
}
#else
void delete_mag_init_fun(char *fn, char *key_name, char *key_val)
{
    register int x, found = 0;
    pblock *pb;
    char *tmp;
    int numservs=make_conflist(), y=0;
    int target;
 
    old_read_mag_conf();
 
    for(y=0; y<numservs; y++)  {
        magconf = magconfarray[y];
 
        for(x=0, found = 0; magconf[x]; x++)  {
            target = x;
            if(!strncasecmp(magconf[x], "Init ", 5))  {
                pb = pblock_create(PARAMETER_HASH_SIZE);
                pblock_str2pblock(hide_esc(&magconf[x][5]), pb);
 
                /* check to see if continue to next line */
                while(1) {
                  int l;
                  l = strlen(magconf[x]) - 1;
                  if(l < 0) break;
                  if(magconf[x][l] != '\\') break;
                  if(!magconf[++x]) break;
                  pblock_str2pblock(hide_esc(magconf[x]), pb);
                }
 
                if( (tmp = pblock_findval("fn", pb)) )  {
                    if(!strcmp(tmp, fn))  {
                      if( (tmp = pblock_findval(key_name, pb)) ) {
			  if(strstr(tmp, key_val)) {
                              x = target;
                              found = 1;
                              break;
                          }
                      }
                    }
                }
                pblock_free(pb);
            }
        }
        if(found) old_delete_mag_var(x);
    }
    old_write_mag_conf();
}
#endif

#endif	/* !USE112 */
/* --------------------------------------------- */


/* --------------------------------------------------------
		-- NT Registry --
   These routine are to handling NT registry.
   It was needed for NT server 1.12, shall be obsoleted after.
   -------------------------------------------------------- */

#ifdef  USE112
/* --------------------------------------------- */
/* for manipulate Registry of NT */

#ifdef XP_WIN32

char *http_server = NULL;

/* env CONFIG_DIR has format ..../https-8000/config/ */
char *get_http_dir()
{
    char *tmp;
    char *slash;

    if(http_server) return(http_server);

    tmp = getenv("CONFIG_DIR");
    if(tmp) {
      tmp = STRDUP(tmp);
      CONVERT_TO_HTTP_FORMAT(tmp);
      slash = strrchr(tmp, '/');
      if(slash) {
	*slash = '\0';
        slash = strrchr(tmp, '/');
	if(slash) {
	  *slash = '\0';
          slash = strrchr(tmp, '/');
	  if(slash) {
	    http_server = STRDUP(++slash);
	    FREE(tmp);
	    return(http_server);
	  } 
	}
      }
      FREE(tmp);
    }
    report_error(SYSTEM_ERROR, "Environment variable CONFIG_DIR", "not exist!");
    return(NULL);
}

char *get_mag_var(char *name)
{
    char buf[1024];
    char val[1024];
    char *http;
    HKEY hkey;
    DWORD len, lType;

    http = get_http_dir();
    sprintf(buf,"SOFTWARE\\Netscape\\%s\\CurrentVersion\\Startup", http);
    if( RegOpenKeyEx(HKEY_LOCAL_MACHINE, buf, 0, KEY_READ, &hkey) == ERROR_SUCCESS)
    {
        len = 1024;
	if(RegQueryValueEx(hkey,name,NULL,&lType,val,&len) == ERROR_SUCCESS)  {
	  RegCloseKey(hkey);
	  if(lType == REG_DWORD) {
	    sprintf(buf, "%i", *((DWORD *)val));
	    return(STRDUP(buf));
	  }
	  return(STRDUP(val));
	}
	RegCloseKey(hkey);
    }
    return(NULL);
}

#if 0
/* not used */
void set_mag_var(char *name, char *val)
{
    char buf[1024];
    char val[1024];
    char *http;
    HKEY hkey;
 
    http = get_http_dir();
    sprintf(buf,"SOFTWARE\\Netscape\\%s\\CurrentVersion\\Startup", http);
    if( RegOpenKeyEx(HKEY_LOCAL_MACHINE, buf, 0, KEY_WRITE, &hkey) == ERROR_SUCCESS)
    {
        RegSetValueEx(hkey,name,NULL,REG_SZ,val,strlen(val)+1);
        RegCloseKey(hkey);
    }
}
#endif

void set_obj_var(char *action, char *fn, char *name, char *val)
{
    int i, j;
    char *http = get_http_dir();
    pblock *answer;
    char buf[1024];
    char data[1024];
    LONG fOpen = ERROR_SUCCESS;
    HKEY hkey;
    DWORD len, lType;
 
 
    for(i = 1; fOpen == ERROR_SUCCESS; ++i) {
      sprintf(buf,
        "SOFTWARE\\Netscape\\%s\\CurrentVersion\\Objects\\Object1\\Directive%i",
        http, i);
      fOpen=RegOpenKeyEx(HKEY_LOCAL_MACHINE, buf, 0, KEY_ALL_ACCESS, &hkey);
      if(fOpen == ERROR_SUCCESS) {
        len = 1024;
        if(RegQueryValueEx(hkey,"DirectiveName",NULL,&lType,data,&len) == ERROR_SUCCESS)  {
          if(!strcasecmp(action, data)) {
	    for(j = 1; fOpen == ERROR_SUCCESS ; ++j) {
              sprintf(buf,
        "SOFTWARE\\Netscape\\%s\\CurrentVersion\\Objects\\Object1\\Directive%i\\Function%i",
              http, i, j);
              fOpen=RegOpenKeyEx(HKEY_LOCAL_MACHINE, buf, 0, KEY_ALL_ACCESS, &hkey);
              if(fOpen == ERROR_SUCCESS) {
                len = 1024;
                fOpen=RegQueryValueEx(hkey,"fn",NULL,&lType,data,&len);
                if(!strcasecmp(fn, data)) break;
	      }
	      if(fOpen != ERROR_SUCCESS) break;
	    }
	    if(fOpen != ERROR_SUCCESS) {
              sprintf(buf,
        "SOFTWARE\\Netscape\\%s\\CurrentVersion\\Objects\\Object1\\Directive%i\\Function%i",
              http, i, j);
              fOpen = RegCreateKeyEx(HKEY_LOCAL_MACHINE, buf, 0, NULL,
                	REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL, &hkey, &len);

	    }
            RegSetValueEx(hkey,"fn",0,REG_SZ,fn,strlen(fn)+1);
	    if(name && val) RegSetValueEx(hkey,name,0,REG_SZ,val,strlen(val)+1);
            RegCloseKey(hkey);
	    break;
          }
        }
      }
      RegCloseKey(hkey);
    }
}

void del_obj_var(char *action, char *fn)
{
    int i, j;
    char *http = get_http_dir();
    pblock *answer;
    char buf[1024];
    char val[1024];
    LONG fOpen = ERROR_SUCCESS;
    HKEY hkey;
    DWORD len, lType;
 
 
    for(i = 1; fOpen == ERROR_SUCCESS; ++i) {
      sprintf(buf,
        "SOFTWARE\\Netscape\\%s\\CurrentVersion\\Objects\\Object1\\Directive%i",
        http, i);
      fOpen=RegOpenKeyEx(HKEY_LOCAL_MACHINE, buf, 0, KEY_ALL_ACCESS, &hkey);
      if(fOpen == ERROR_SUCCESS) {
        len = 1024;
        if(RegQueryValueEx(hkey,"DirectiveName",NULL,&lType,val,&len) == ERROR_SUCCESS)  {
          if(!strcasecmp(action, val)) {
            for(j = 1; fOpen == ERROR_SUCCESS ; ++j) {
              sprintf(buf,
        "SOFTWARE\\Netscape\\%s\\CurrentVersion\\Objects\\Object1\\Directive%i\\Function%i",
              http, i, j);
              fOpen=RegOpenKeyEx(HKEY_LOCAL_MACHINE, buf, 0, KEY_ALL_ACCESS, &hkey);
              if(fOpen == ERROR_SUCCESS) {
                len = 1024;
                fOpen=RegQueryValueEx(hkey,"fn",NULL,&lType,val,&len);
                if(!strcasecmp(fn, val)) {
      		   RegCloseKey(hkey);
		   RegDeleteKey(HKEY_LOCAL_MACHINE, buf);
		   return;
		}
              }
            }
          }
        }
      }
      RegCloseKey(hkey);
    }
}

#endif		/* WIN32 */
#endif		/* !USE112 */
/* ---------------------------------------------------------- */

/* --------------------------------------------------------
		-- csid.conf --
   These routines take care of csid.conf.
   Use can list, add, and delete CSID.
   --
   Where	char *get_csid_conf(int module)
   Read		dyn_array *read_csid_list(char *filename)
   Write	void write_csid_list(char *filename, dyn_array *ary)
   Get		csid_type *get_csid(dyn_array *ary, char *name)
   Delete	void csid_delete(dyn_array *csid_ary, char *csid)
   Dup		csid_type *dup_csid(csid_type *id)
   -------------------------------------------------------- */

/* Full path to csid.conf.
*/
char *get_csid_conf(int module)
{
    char *path;
    char *conf;
 
    conf = get_ds_conf_dir(0, module);
    path = MALLOC(strlen(conf) + 10);
    sprintf(path, "%scsid.conf", conf);
    FREE(conf);
    return(path);
}

/* return a new csid structure. */
csid_type *create_csid()
{
    int i;
    char *tmp;
 
    csid_type* a_csid = MALLOC(sizeof(csid_type));
    for(tmp=(char *)a_csid, i = 0; i < sizeof(csid_type); ++i, ++tmp)
        *tmp = '\0';
    return(a_csid);
}
 
/* free memory of csid structure. */
void destroy_csid(void *a_csid)
{
        FREE(a_csid);
}
 
/* Map name of database type into a int index */
int type_index(char *buf)
{
    if(!buf) return(0);
    if(!strncasecmp(buf, "rds", 3)) return(0);
    if(!strncasecmp(buf, "informix", 8)) return(1);
    if(!strncasecmp(buf, "oracle", 6)) return(2);
    if(!strncasecmp(buf, "sybase", 6)) return(3);
    if(!strncasecmp(buf, "ODBC", 4)) return(4);
    return(0);
}
 
/* Map a database type index into a string name */
char *type_name(int inx)
{
    switch(inx) {
        case 0:         return("rds");
        case 1:         return("informix");
        case 2:         return("oracle");
        case 3:         return("sybase");
        case 4:         return("ODBS");
        default:
                        return("None");
    }
    return(NULL);
}
 
/* extract string after "=" */
static char *get_name(char *buf)
{
    char *tmp = NULL;
    char *end;
 
    if(!buf) return(NULL);
    if(!(tmp = strchr(buf, '='))) return(NULL);
    ++tmp;
    while(*tmp) if(*tmp == '\"' || *tmp == ' ') ++tmp; else break;
    if(!(*tmp)) return(NULL);
    end = strchr(tmp, '\"');
    if(end) *end = '\0';
    return(STRDUP(tmp));
}
 
/* move pointer to next non-blank char */
char *skip_blank(char *buf)
{
    char *tmp;
 
    if(!buf) return(NULL);
    tmp = buf;
    while(*tmp) if(*tmp == ' ' || *tmp == '\t') ++tmp; else break;
    return(tmp);
}
 
/* read in csid.conf into a dynamic array.
*/
dyn_array *read_csid_list(char *filename)
{
    dyn_array *ary = NULL;
    csid_type *a_csid = NULL;
    FILE *fp;
    char buf[BIG_LINE];
    char *tmp, *line;
 
    if(!filename) return(NULL);
    ary = create_dyn_array(5, 5);
    if(!(fp = fopen(filename, "r")))
        report_error(FILE_ERROR, filename, "couldn't read file");
    while(fgets(buf, BIG_LINE, fp)) {
        if(buf[0] == '#') continue;
        if(!strncmp(buf, "csid", 4)) {
          tmp = get_name(buf + 4);
          if(tmp) {
            a_csid = create_csid();
            a_csid->csid = tmp;
            dyn_add(ary, (void *)a_csid);
          }
        }
        else {
          line = skip_blank(buf);
          if(!strncmp(line, "verity-home", 11)) {
                tmp = get_name(line+11);
                if(a_csid && tmp) a_csid->verity_home = tmp;
          }
          else if (!strncmp(line, "verity-collection", 17)) {
                tmp = get_name(line+17);
                if(a_csid && tmp) a_csid->verity_col = tmp;
          }
          else if (!strncmp(line, "alias-file", 10)) {
                tmp = get_name(line+10);
                if(a_csid && tmp) a_csid->alias = tmp;
          }
          else if (!strncmp(line, "catalog-database", 16)) {
                tmp = get_name(line+16);
                if(a_csid && tmp) a_csid->viewcache = tmp;
          }
          else if (!strncmp(line, "statistics-log", 14)) {
                tmp = get_name(line+14);
                if(a_csid && tmp) a_csid->statistics = tmp;
          }
          else if (!strncmp(line, "dbdir", 5)) {
                tmp = get_name(line+5);
                if(a_csid && tmp) a_csid->dbdir = tmp;
          }
          else if (!strncmp(line, "schema-description", 18)) {
                tmp = get_name(line+18);
                if(a_csid && tmp) a_csid->schema_d = tmp;
          }
          else if (!strncmp(line, "server-description", 18)) {
                tmp = get_name(line+18);
                if(a_csid && tmp) a_csid->server_d = tmp;
          }
          else if (!strncmp(line, "import-config", 13)) {
                tmp = get_name(line+13);
                if(a_csid && tmp) a_csid->import_c = tmp;
          }
        }
    }
    fclose(fp);
    if(dyn_size(ary) == 0) {
        destroy_dyn_array(ary, NULL);
        ary = NULL;
    }
    return(ary);
}

/*  write csid dynamic array back to file csid.conf. */
void write_csid_list(char *filename, dyn_array *ary)
{
    FILE *fp;
    int i;
    csid_type *id;
 
    if(!(fp = fopen(filename, "w")))
        report_error(FILE_ERROR, filename, "couldn't create file");

    if(ary) for(i = 0; i < dyn_size(ary); ++i) {
        id = (csid_type *)dyn_data(ary, i);
        fprintf(fp, "csid=\"%s\"\n", id->csid);
        if(id->verity_home) fprintf(fp, "\tverity-home=\"%s\"\n", id->verity_home);
        if(id->verity_col) fprintf(fp, "\tverity-collection=\"%s\"\n", id->verity_col);
        if(id->alias) fprintf(fp, "\talias-file=\"%s\"\n", id->alias);
        if(id->viewcache) fprintf(fp, "\tcatalog-database=\"%s\"\n", id->viewcache);
        if(id->statistics) fprintf(fp, "\tstatistics-log=\"%s\"\n", id->statistics);
        if(id->dbdir) fprintf(fp, "\tdbdir=\"%s\"\n", id->dbdir);
        if(id->schema_d) fprintf(fp, "\tschema-description=\"%s\"\n", id->schema_d);
        if(id->server_d) fprintf(fp, "\tserver-description=\"%s\"\n", id->server_d);
        if(id->import_c) fprintf(fp, "\timport-config=\"%s\"\n", id->import_c);
        fprintf(fp,"\n");
    }
    fclose(fp);
}
 
/* remove a csid from list.
   it is identified by csid string.
*/
void csid_delete(dyn_array *csid_ary, char *csid)
{
    int i;
 
    if(!csid_ary) return;
    for(i = 0; i < dyn_size(csid_ary); ++i) {
        char *tmp = ((csid_type *)dyn_data(csid_ary, i))->csid;
        if(!strcmp(tmp, csid)) {
          dyn_delete(csid_ary, i); 
          break;
        }
    }
}
 
/* if an old one exist, then overwrite it */
void csid_add(dyn_array *ary, csid_type *csid)
{
    int i;
    csid_type *id;

    for(i = 0; i < dyn_size(ary); ++i) {
	id = (csid_type *)dyn_data(ary, i);
	if(!id) continue;
	if(!strcmp(id->csid, csid->csid)) {
	  ary->entry[i] = csid;
	  FREE(id);
	  return;
	}
    }
    dyn_add(ary, (void *)csid);
    return;
}


/* return the pointer to csid according specified name.
*/
csid_type *get_csid(dyn_array *ary, char *name)
{
    csid_type *id;
    int i;

    if(ary) for(i = 0; i < dyn_size(ary); ++i) {
	id = (csid_type *)dyn_data(ary, i);
	if(!strcmp(name, id->csid)) return(id);
    }
    return(NULL);
}


/* duplicate a csid, allocate memory for new csid. */
csid_type *dup_csid(csid_type *id)
{
    csid_type *dup;

    if(!id) return(NULL);
    dup = create_csid();
    /* bcopy((char *)id, (char *)dup, sizeof(csid_type)); */
    {
	int i;
	char *src = (char *)id;
	char *dst = (char *)dup;
	for(i = 0; i < sizeof(csid_type); ++i, ++src, ++dst) 
		*dst = *src;
    }
    return(dup);
}


/*
  create a dynamic array for csid.conf with default csid only.
*/
dyn_array *set_default_csid()
{
    char buf[200];
 
    dyn_array *ary = create_dyn_array(3, 3);
    csid_type *a_csid = create_csid();
    sprintf(buf, "x-catalog://%s:%s/AutoCatalog",
                get_mag_var("ServerName"), get_mag_var("Port"));
    a_csid->csid = STRDUP(buf);
    dyn_add(ary, (void *)a_csid);
 
    return(ary);
}


/* --------------------------------------------- */

void batman_success(char *description)  {
    char line[BIG_LINE];
    char **config = get_adm_config();
    FILE *html=open_html_file("goodcsid.html");
 
    WSACleanup();
    /* Be sure headers are terminated. */
    fputs("\n", stdout);
 
    while(next_html_line(html, line))  {
        if(parse_line(line, config))  {
            if(directive_is(line, "SUCCESS_DESCRIPTION"))  {
                fprintf(stdout, "<dl><dt><FONT size=+1>Success!</FONT>\n");
                fprintf(stdout, "<dd>%s\n</dl>", description);
            }
        }
    }
    exit(0);
}

/* --------------------------------------------- */

/*
  Routines steal from netsite/lib/cs
*/

int cslist_set_compare(CSList *l, CSCompareFn cmp)
{
        if (!l || !cmp)
                return 1;
        l->cmp = cmp;
        return 0;
}
 
int cslist_set_free(CSList *l, CSFreeFn fr)
{
        if (!l || !fr)
                return 1;
        l->fr = fr;
        return 0;
}
 
int cslist_set_copy(CSList *l, CSCopyFn cp)
{
        if (!l || !cp)
                return 1;
        l->cp = cp;
        return 0;
}
 
void *cslist_copy_nop(void *data)
{
        return data;
}
 
void cslist_free_nop(void *data)
{
        return;
}
 
int cslist_cmp_nop(void *data1, void *data2)
{
        return memcmp(data1, data2, sizeof(void *));
}
 
int cslist_create(CSList **l)
{
        if ((*l = (CSList *) MALLOC(sizeof(CSList))) == NULL)
                return 1;
        (*l)->head = NULL;
        (*l)->tail = NULL;
        (*l)->nnodes = 0;
        cslist_set_free(*l, (CSFreeFn) free);
        cslist_set_compare(*l, (CSCompareFn) strcmp);
        cslist_set_copy(*l, (CSCopyFn) strdup);
        return 0;
}

 
int csqueue_create(CSQueue **q)
{
        int r = cslist_create((CSList **)q);
 
        if (!r) {
                cslist_set_compare(*q, cslist_cmp_nop);
                cslist_set_free(*q, cslist_free_nop);
                cslist_set_copy(*q, cslist_copy_nop);
        }
        return r;
}
 
/* append node to the list */
int cslist_append(CSList *l, void *data)
{
        CSListNode *node = NULL;
 
        if (!l || !data)
                return 1;
 
        node = (CSListNode *)MALLOC(sizeof(CSListNode));
        node->data = l->cp(data);
        node->next = NULL;
 
        if (l->head == NULL) {
                l->head = l->tail = node;
        } else {
                l->tail->next = node;
                l->tail = l->tail->next;
        }
        l->nnodes++;
        return 0;
}
 
 
int csqueue_put(CSQueue *q, void *data)
{
        return cslist_append((CSList *)q, data);
}
 
 CSQueue *processq = NULL;

/* ------------------------------------------------------------------ */
 
/* reads the work to be done */
void dsenum_load_process_cf(char *fn)
{
        FILE *fp = fopen(fn, "r");
        char buf[4096], *s;
        ProcessNode *pn = NULL;
 
        if (fp == NULL) {
                report_error(FILE_ERROR,"Robot Control","Cannot read configuration process.conf");
                return;
        }
 
        csqueue_create(&processq);
 
        while (fgets(buf, sizeof(buf), fp)) {
                if (buf[0] == '\n' || buf[0] == '#')
                        continue;
                if (!strncasecmp(buf, "<Process", 8)) {         /* START */
                        pn = MALLOC(sizeof(ProcessNode));
                        csqueue_create(&pn->urls);
                        pn->pb = pblock_create(3);
                        for (s = buf+8; *s && *s != '>'; s++);
                        *s = '\0';
                        (void)pblock_str2pblock(buf+9, pn->pb);
                } else if (!strncasecmp(buf, "</Process", 9)) { /* STOP */
                        csqueue_put(processq, (void *)pn);
                        pn = NULL;                      /* detach */
                } else if (isspace(buf[0])) {           /* Start con't */
                        for (s = buf+1; *s && *s != '>'; s++);
                        *s = '\0';
                        if (pn != NULL)
                                (void)pblock_str2pblock(buf, pn->pb);
                } else {
                        if (pn != NULL) {
                                buf[strlen(buf)-1] = '\0';      /* nuke \n */
                                csqueue_put(pn->urls, (void *)STRDUP(buf));
                        }
                }
        }
        fclose(fp);
}
 
/* ------------------------------------------------------------------ */
 
void do_stat_error(char *errstr, int print_errno)
{
	int x = 0;

        fprintf(stdout, "<FONT color=%s>[%s]:</FONT> %s",
                        HTML_ERRCOLOR, get_srvname(x), errstr);
        if(print_errno)
            fprintf(stdout, " (%d: %s)", errno, system_errmsg());
        fprintf(stdout, "<br>\n");
        fflush(stdout);
}
 
int do_stat_log(char *success, char *statline)
{
    int x = 0;
    int ans = 0;
 
    if( (!success) || (!strncmp(statline, success, strlen(success))) ) {
            ans = 0;
            fprintf(stdout, "[%s]: %s<br>\n", get_srvname(x), statline);
            fflush(stdout);
            return ans;
    }  else  {
            ans = 1;
            fprintf(stdout, "<FONT color=%s>[%s]:</FONT> %s<br>\n",
                            HTML_ERRCOLOR, get_srvname(x), statline);
            fflush(stdout);
            return ans;
    }
    return ans;
}
 
/* ----------------------------------------------------------------------------- */

