/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

#include "base/daemon.h"
#include "libadmin/dstats.h"
#include "base/systems.h"	/* ZERO */
#include "base/util.h"		/* sprintf, is_mozilla */
#include "base/file.h"		/* MMAP_FLAGS */
#include "base/daemon.h"	/* CHILD_*, StatSlot */
#include "frame/conf.h"		/* server_hostname */

#include <sys/stat.h>
#include <sys/types.h>
#include <string.h>
#ifdef XP_UNIX
#include <unistd.h>
#include <sys/mman.h>
#else /* XP_WIN32 */
#include <io.h>
#include <nt/regparms.h>
#endif /* XP_WIN32 */
#include <fcntl.h>


/*
 * Open the statistics file for the server running on the
 * specified port.
 *
 * Returns the handle to daemon statistics on success.
 * Returns NULL on failure, and sets 'err_title' and 'err_msg'
 * to point to an error title and message strings.
 *
 *
 */
NSAPI_PUBLIC DStatsHandle *dstats_open(int port, char *addr, char **err_title, char **err_msg)
{
    char fn[256];
#ifdef XP_UNIX
    int fd;
    struct stat fi;
    caddr_t fp;
#else
    HANDLE mapHandle;
    char *fp;
#endif
	int hsiz, ssiz;

    DStatsHandle *handle = NULL;


    /*
     * Open the file and mmap it.
     *
     *
     */
#ifdef XP_UNIX
    sprintf(fn, "/tmp/daemonstat.%s.%d", addr ? addr : "any", port);
    if( (fd = open(fn, O_RDONLY)) == -1) {
		*err_title = "Can't find daemon statistics file";
		*err_msg = (char *)MALLOC(1000);
		sprintf(*err_msg,
				"Opening of %s failed.  Either the file has been removed,\n"
				"or the system does not support daemon statistics.\n"
				"The error was %s", fn, system_errmsg());

		return NULL;
    }

    fstat(fd, &fi);
    fp = (caddr_t)mmap(NULL, (unsigned int)fi.st_size, PROT_READ, SHMEM_MMAP_FLAGS, fd, 0);
    if(fp == (caddr_t) -1) {
		*err_title = "Can't map daemon statistics file";
		*err_msg = (char *)MALLOC(1000);
		sprintf(*err_msg,
				"mmap of %s failed. Either the file has been removed,\n"
				"or the system doesn't support daemon statistics.\n"
				"The error was %s", fn, system_errmsg());

		return NULL;
    }

#else /* XP_WIN32 */

    sprintf( fn, "%s"SVR_ID_SERVICE"%s", DSTATS_FILE,
	     strchr( get_srvname(0), '-' ) );

    if ((mapHandle = OpenFileMapping(
			FILE_MAP_READ,
			0, 
			fn)) == NULL) {

		*err_title = "Can't open mapping of daemon statistics file";
		*err_msg = (char *)MALLOC(1000);
		sprintf(*err_msg,
				"mmap of %s failed. Either the file has been removed,"
				" or the system doesn't support daemon statistics. "
				"The error was %s", fn, system_errmsg());

		return NULL;
    }

    if (!(fp = MapViewOfFile(mapHandle, FILE_MAP_READ, 0, 0, 0))) {
		*err_title = "Can't mmap daemon statistics file";
		*err_msg = (char *)MALLOC(1000);
		sprintf(*err_title,
				"mmap of %s failed. Either the file has been removed,"
				" or the system doesn't support daemon statistics. "
				"The error was %s", fn, system_errmsg());

		return NULL;
    }

#endif /* XP_WIN32 */


    /*
     * Open and mmap succeeded -- create the context to be
     * passed in consequent calls to this module.
     *
     *
     */
    if (!(handle = (DStatsHandle *)MALLOC(sizeof(DStatsHandle)))) {
		*err_title = "Out of memory";
		*err_msg = "Ran out of memory trying to create the statistics handle";
		return NULL;
    }
    ZERO(handle, sizeof(DStatsHandle));


	/*
	 * Store platform specific handles to shared memory.
	 */
#ifdef XP_UNIX
	handle->fd = fd;
#else
	handle->mapHandle = mapHandle;
#endif


    /*
     * Store the actual shared memory.
     *
     */
	handle->mapped = fp;
#ifndef XP_WIN32
	handle->mapped_size = fi.st_size;
#endif


	/*
	 * Separate different slots from the shared memory.
	 * It first has a header section defining the sizes.
	 *
	 */
	handle->hdr = (StatHeader *)fp;
	hsiz = handle->hdr->header_size;
	ssiz = handle->hdr->slot_size;
#ifdef XP_UNIX
	if ((size_t)hsiz < handle->mapped_size && ssiz > 0) {
		handle->stats = (StatSlot *)(((caddr_t)fp) + hsiz);
		handle->nslots = (handle->mapped_size - hsiz) / ssiz;
	}
#else
	handle->stats = (StatSlot *)(((caddr_t)fp) + hsiz);
#if 0
	handle->nslots = handle->hdr->maxprocs * handle->hdr->maxthreads;
#else
	handle->nslots = handle->hdr->maxthreads;
#endif
#endif

	handle->last_restart_cookie = handle->hdr->restart_cookie;

    return handle;
}


/*
 * Grab and compile the current statistics from the statistics file.
 *
 *
 */
NSAPI_PUBLIC dstats *dstats_snapshot(DStatsHandle *handle)
{
	int x;
	dstats *snew     = &handle->snew;
	StatSlot *totals = &handle->totals;
	StatSlot *stats  = handle->stats;

	if (!stats)
		return NULL;

	handle->sold = handle->snew;
	ZERO(snew, sizeof(dstats));
	ZERO(totals, sizeof(StatSlot));

	for(x = 0; x < handle->nslots; x++) {
		unsigned int stx;

		stx = stats[x].mode & 0xff;

		if(stx == CHILD_EMPTY_SLOT)
		    snew->empty++;
		else {
			++snew->active;
			switch (stx) {
			  case CHILD_AWAIT_CONNECT:
			    snew->idle++;
				break;
			  case CHILD_PROCESSING:
			    snew->process++;
				break;
			  case CHILD_READING:
			    snew->read++;
				break;
			  case CHILD_WRITING:
				snew->write++;
				break;
			  case CHILD_RESOLVING:
				snew->resolv++;
				break;
			  default:
				;
			}

			totals->tot_bytes += stats[x].tot_bytes;
			totals->tot_reqs  += stats[x].tot_reqs;
			totals->tot_errs  += stats[x].tot_errs;

			totals->tot_2xx += stats[x].tot_2xx;
			totals->tot_3xx += stats[x].tot_3xx;
			totals->tot_4xx += stats[x].tot_4xx;
			totals->tot_5xx += stats[x].tot_5xx;
			totals->tot_xxx += stats[x].tot_xxx;

			totals->tot_200 += stats[x].tot_200;
			totals->tot_302 += stats[x].tot_302;
			totals->tot_304 += stats[x].tot_304;
			totals->tot_401 += stats[x].tot_401;
			totals->tot_403 += stats[x].tot_403;
		}
	}

	return &handle->snew;
}


/*
 * Un-mmap and close the statistics file, and free the statistics
 * context.
 *
 *
 */
NSAPI_PUBLIC void dstats_close(DStatsHandle *handle)
{
#ifdef XP_UNIX
	if (handle->mapped && handle->mapped_size > 0)
		munmap(handle->mapped, handle->mapped_size);

	close(handle->fd);
#else
	UnmapViewOfFile(handle->mapped);
	CloseHandle(handle->mapHandle);
#endif

    FREE(handle);
}


NSAPI_PUBLIC int dstats_poll_for_restart(DStatsHandle *handle)
{
	if (!handle || !handle->hdr || !handle->hdr->active)
		return 2;

	if (handle->last_restart_cookie != handle->hdr->restart_cookie) {
		/* has restarted */
		handle->last_restart_cookie = handle->hdr->restart_cookie;
		return 1;
	}

	return 0;
}

