/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * install.c: Routines shared among server (not to be confused with
 *            Netsite) installation programs
 * 
 * Rob McCool
 */


#include "libadmin/libadmin.h"
#include "libadmin/install.h"

#include "base/systems.h"      /* ZERO */
#include "base/util.h"         /* sprintf */
#include "base/net.h"          /* SYS_NETFD etc. */

#include <string.h>

#ifdef XP_WIN32
#define EACCES WSAEACCES
#define EADDRINUSE WSAEADDRINUSE
#include <winsock.h>
#include <io.h>
#else
#include <arpa/inet.h>
#endif /* XP_WIN32 */


int interactive = 0;
int upgrading = 0;


/* ------------------------- install_finish_error ------------------------- */


NSAPI_PUBLIC void install_finish_error(char *errpage, char *info, char *details)
{
    char line[BIG_LINE];
    char **config = get_adm_config();
    FILE *html=open_html_file(INSTERR_PAGE);

    while(next_html_line(html, line))  {
        if(parse_line(line, config))  {
            if(directive_is(line, "ERROR_DETAILS"))  {
                printf("<dl><dt>\n");
                printf("<b>Error: %s</b><p>\n", info);
                printf("<dd>\n%s\n</dl>\n", details);
                if(errpage)
                    printf(
"The error can be corrected on <a href=%s>this page</a>.<p>\n", errpage);
            }  else
                fputs(line, stdout);
        }
    }
    exit(0);
}


/* ------------------------------- myprintf ------------------------------- */


NSAPI_PUBLIC void myprintf(char *fmt, ...)
{
    va_list args;
    char buf[MAX_PRINTF];

    va_start(args, fmt);
    util_vsprintf(buf, fmt, args);
    OUT(buf);
}


/* --------------------------------- die ---------------------------------- */


NSAPI_PUBLIC void die(char *page, char *msg, char *info, ...)
{
    va_list args;
    char buf[MAX_PRINTF];

    va_start(args, info);
    util_vsprintf(buf, info, args);

    myprintf("</pre>\n");
    NEW_DOCUMENT();

    install_finish_error((upgrading ? "upgrade" : page), msg, buf);
}



/* ---------------------------- create_subdirs ---------------------------- */


#include <sys/types.h>
#include <sys/stat.h>


/* Simulate mkdir -p */
NSAPI_PUBLIC void create_subdirs(char *dir, char *page)
{
    struct stat fi;
    char *t;

#ifdef XP_UNIX
    t = dir + 1;
#else /* XP_WIN32 */
    t = dir + 3;
#endif /* XP_WIN32 */

    while(1) {
        t = strchr(t, FILE_PATHSEP);

        if(t) *t = '\0';
        if(stat(dir, &fi) == -1) {
            myprintf("mkdir %s\n", dir);
#ifdef XP_UNIX
            if(mkdir(dir, NEWDIR_MODE) == -1) {
#else /* XP_WIN32 */
			if(!CreateDirectory(dir, NULL)) {
				if (GetLastError() != ERROR_ALREADY_EXISTS)
#endif /* XP_WIN32 */
                die(page, "Can't create directory", 
                    "mkdir %s failed. The error was <code>%s</code>\n", 
                    dir, system_errmsg());
            }
        }
        if(t) *t++ = FILE_PATHSEP;
        else break;
    }
}



/* --------------------- copy_file and copy_binaries ---------------------- */

#ifdef XP_UNIX
/* Uses mmap to blast a file to a new location. Pretty UNIX specific. */
#include <sys/mman.h>

NSAPI_PUBLIC void copy_file(char *sfile, char *dfile, int mode)
{
    int sfd, dfd, len;
    struct stat fi;
    caddr_t fp;

    if( (sfd = open(sfile, O_RDONLY)) == -1)
        die(NULL, "Can't open file for reading",
            "Can't open %s. The error was <code>%s</code>\n",
            sfile, system_errmsg());

    fstat(sfd, &fi);
    if(!(S_ISREG(fi.st_mode))) {
        close(sfd);
        return;
    }
    len = fi.st_size;

    if( (dfd = open(dfile, O_RDWR | O_CREAT | O_TRUNC, mode)) == -1)
        die(NULL, "Can't open file for writing",
            "Can't write to %s. The error was <code>%s</code>\n",
            dfile, system_errmsg());

    if(len) {
        if( (fp=(caddr_t)mmap(NULL,len,PROT_READ,FILE_MMAP_FLAGS,sfd,0)) == (caddr_t)-1)
            die(NULL, "Can't map file into memory",
                "Can't scan %s for copy. The error was <code>%s</code>\n",
                sfile, system_errmsg());

        if(write(dfd, fp, len) == -1)
            die("server", "Can't write new file",
                "Can't write the copy %s. The error was <code>%s</code>\n",
                dfile, system_errmsg());
        munmap(fp, len);
    }
    close(sfd);
    close(dfd);
}
#else /* XP_WIN32 */
NSAPI_PUBLIC void copy_file(char *sfile, char *dfile, int mode)
{
    HANDLE sfd, dfd, MapHandle;
    PCHAR fp;
	DWORD BytesWritten = 0;
	DWORD len;

    //myprintf("%s -> %s\n", sfile, dfile);
    if( (sfd = CreateFile(sfile, GENERIC_READ,
    		FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
    		OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL))
    			== INVALID_HANDLE_VALUE) {
		if (GetLastError() == ERROR_ACCESS_DENIED)
			return;
        die(NULL, "Can't open file for reading",
            "Can't open %s. The error was <code>%s</code>\n",
            sfile, system_errmsg());
	}
	len = GetFileSize(sfd, NULL);
    if( (MapHandle = CreateFileMapping(sfd, NULL, PAGE_READONLY,
    		0, 0, NULL)) == NULL) {
        die(NULL, "Can't map file into memory",
            "Can't scan %s for copy. The error was <code>%s</code>\n",
            sfile, system_errmsg());
	}
	if (!(fp = MapViewOfFile(MapHandle, FILE_MAP_READ, 0, 0, 0))) {
        die(NULL, "Can't map file into memory",
            "Can't scan %s for copy. The error was <code>%s</code>\n",
            sfile, system_errmsg());
	}
    if( (dfd = CreateFile(dfile, GENERIC_READ | GENERIC_WRITE,
    	FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,OPEN_ALWAYS,
    	FILE_ATTRIBUTE_NORMAL, NULL)) == INVALID_HANDLE_VALUE) {
        die(NULL, "Can't open file for writing",
            "Can't write to %s. The error was <code>%s</code>\n",
            dfile, system_errmsg());
	}
	while ( len) {
    	if(!WriteFile(dfd, fp, len, &BytesWritten, NULL)) {
        	die("server", "Can't write new file",
            	"Can't write the copy %s. The error was <code>%s</code>\n",
            	dfile, system_errmsg());
		}
		len -= BytesWritten;
		fp += BytesWritten;
	}

    CloseHandle(sfd);
    UnmapViewOfFile(fp);
	CloseHandle(MapHandle);
	FlushFileBuffers(dfd);
    CloseHandle(dfd);
}
#endif /* XP_WIN32 */


NSAPI_PUBLIC void copy_dir(char *sdir, char *ddir, int sl, int dl)
{
    SYS_DIR ds;
    SYS_DIRENT *d;

    myprintf("directory %s -> %s\n", sdir, ddir);
    if(!(ds = dir_open(sdir)))
        die("server", "Can't scan directory",
            "Can't scan the directory %s. The error was <code>%s</code>\n",
            sdir, system_errmsg());

    while( (d = dir_read(ds)) ) {
        if(d->d_name[0] == '.')
            continue;
        strcpy(&sdir[sl], d->d_name);
        strcpy(&ddir[dl], d->d_name);
        copy_file(sdir, ddir, NEWSCRIPT_MODE);
    }
    dir_close(ds);
}


#ifdef XP_UNIX
NSAPI_PUBLIC void run_server(char *cmd, FILE *html)
{
    struct runcmd_s rm;
    char junk[1024];

    myprintf("%s\n", cmd);

    if(run_cmd(cmd, html, &rm) == -1) {
        if(rm.sysmsg) {
            sprintf(junk, "%s The error was <code>%s</code>.\n",
                         rm.msg, system_errmsg());
            die(NULL, rm.title, junk, rm.arg, system_errmsg());
        }
        else
            die(NULL, rm.title, rm.msg, rm.arg, system_errmsg());
    }
}


/* --------------------------- install_killadm ---------------------------- */


#include <unistd.h>
#include <signal.h>

NSAPI_PUBLIC void install_killadm(void)
{
    pid_t pid = atoi(getenv("ADMSERV_PID"));
    char cmd[1024];

    fflush(stdout);
    fflush(stderr);

    switch(fork()) {
      case -1:
        break;
      case 0:
        fclose(stdin); fclose(stdout); fclose(stderr);
        if(setsid() == -1)
            exit(1);

        sleep(10);   /* wait for a bit so the server can give feedback */
        /* Yeah, I should do this by hand but aw hell. */
        sprintf(cmd, "rm -rf %s", getenv("ADMSERV_ROOT"));
        system(cmd);
        kill(pid, SIGTERM);
        exit(0);
    }
}
#else /* XP_WIN32 */
NSAPI_PUBLIC void install_killadm(void) {}
#endif /* XP_WIN32 */


/* -------------------------------- try_* --------------------------------- */


/* robm This doesn't use net_ abstractions because they drag in SSL */
NSAPI_PUBLIC int try_bind(char *addr, int port)
{
    int sd;
    struct sockaddr_in sa_server;
    int one = 1, ret;

#ifdef XP_WIN32
    WSADATA wsd;

    if(WSAStartup(MAKEWORD(1, 1), &wsd) != 0)
        return -1;
#endif

    if ((sd = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP)) == -1)
        goto you_lose;

/*
    if((setsockopt(sd, SOL_SOCKET, SO_REUSEADDR,
                   (void *)&one, sizeof(one))) == -1) 
        goto you_lose;
*/

    ZERO((char *) &sa_server, sizeof(sa_server));
    sa_server.sin_family=AF_INET;
    sa_server.sin_addr.s_addr = (addr ? inet_addr(addr) : htonl(INADDR_ANY));
    sa_server.sin_port=htons(port);
    ret = bind(sd, (struct sockaddr *) &sa_server,sizeof(sa_server));
/*
    if(ret != -1)
        listen(sd, 5);
*/
#ifdef XP_UNIX
    close(sd);
#else
    closesocket(sd);
    WSACleanup();
#endif
    return ret;

  you_lose:
#ifdef XP_WIN32
    WSACleanup();
#endif
    return -1;
}


#ifdef XP_UNIX
#include <pwd.h>

NSAPI_PUBLIC int try_user(char *user)
{
    struct passwd *pw;
    char fn[128];
    int fd, ret;

    setpwent();
    if(!(pw = getpwnam(user)))
        return -1;

    endpwent();

    if(geteuid())
        return 0;

    sprintf(fn, "/tmp/trychown.%d", getpid());
    if( (fd = creat(fn, 0777)) == -1)
        return 0;         /* Hmm. */
    ret = chown(fn, pw->pw_uid, pw->pw_gid);
    close(fd);
    unlink(fn);
    return (ret == -1 ? -2 : 0);
}
#endif /* XP_UNIX */


/* -------------------------- install_checkport --------------------------- */


NSAPI_PUBLIC void install_checkport(char *addr, char *sport)
{
    int port;

    port = atoi(sport);
    if((port < 1) || (port > 65535)) {
        report_error(INCORRECT_USAGE, "Invalid port number", 
                     "Valid port numbers are between 1 and 65535");
    }
    if(try_bind(addr, port) == -1) {
        if(errno == EADDRINUSE)	{
            report_error(SYSTEM_ERROR, 
                         "That port is already in use", "");
        } else if(errno == EACCES) {
            report_error(SYSTEM_ERROR, 
                         "Ports below 1024 require super user access.", 
                         "You must run the installation as root to install "
                         "on that port.");
        } else {
            report_error(SYSTEM_ERROR, "That port is not available",
                         system_errmsg());
        }
    }
}

#ifdef XP_UNIX
NSAPI_PUBLIC void install_checkuser(char *user)
{
    switch(try_user(user)) {
      case -1:
        report_error(INCORRECT_USAGE, "User not found", 
                     "The server can't find that user, did you create one?");
      case -2:
        report_error(INCORRECT_USAGE, "User invalid", 
"The system will not allow you to use that user. Please select or create "
"another.");
    }
}
#else /* XP_WIN32 */
NSAPI_PUBLIC void install_checkuser(char *user) {}
#endif /* XP_WIN32 */


/* ------------------------------ admconf_* ------------------------------ */


NSAPI_PUBLIC struct admconf *admconf_create(void)
{
    struct admconf *dc = (struct admconf *) MALLOC(sizeof(struct admconf));

    ZERO(dc, sizeof(struct admconf));
    return(dc);
}


NSAPI_PUBLIC void admconf_write(struct admconf *ac, char *filename)
{
    char fn[1024], err[sizeof(fn) + 50];
    FILE *f;

    sprintf(fn, "%s%c%s", CONFIG_DIRECTORY, FILE_PATHSEP, 
            filename ? filename : DEFAULT_ADMCONF);

    if(!(f = fopen_l(fn, "w"))) {
        sprintf(err, "Can't write to file %s", fn);
        report_error(FILE_ERROR, system_errmsg(), err);
    }
    fprintf(f, "AdminUser %s\n", ac->adminuser);
    fprintf(f, "Password %s\n", ac->password);
    fprintf(f, "Hosts %s\n", ac->hosts ? ac->hosts : "");
    fprintf(f, "Addresses %s\n", ac->addresses ? ac->addresses : "");
    fprintf(f, "AdmservPort %s\n", ac->admport ? ac->admport : "");
    fprintf(f, "AdmservUser %s\n", ac->admuser ? ac->admuser : "");

    fclose_l(f);
}


NSAPI_PUBLIC struct admconf *admconf_scan(char *filename)
{
    struct admconf *ac = admconf_create();
    char fn[1024], err[sizeof(fn) + 50], **lines, *t;
    int x;

    sprintf(fn, "%s%c%s", CONFIG_DIRECTORY, FILE_PATHSEP, 
            filename ? filename : DEFAULT_ADMCONF);

    if(!(lines = scan_tech(fn))) {
        sprintf(err, "Can't open file %s", fn);
        report_error(FILE_ERROR, system_errmsg(), err);
    }
    ac->lines = lines;
    for(x = 0; lines[x]; x++) {
        if( (t = strchr(lines[x], ' ')) ) {
            *t++ = '\0';
            if(!strcasecmp(lines[x], "adminuser"))
                ac->adminuser = t;
            else if(!strcasecmp(lines[x], "password"))
                ac->password = t;
            else if(!strcasecmp(lines[x], "hosts"))
                ac->hosts = t;
            else if(!strcasecmp(lines[x], "addresses"))
                ac->addresses = t;
            else if(!strcasecmp(lines[x], "admservport"))
                ac->admport = t;
            else if(!strcasecmp(lines[x], "admservuser"))
                ac->admuser = t;
        }
    }
    return ac;
}

