/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * Utilities for dealing with httpd config files etc.
 * 
 * Rob McCool
 */


/* From magnus */
#include "frame/objset.h"
#include "frame/req.h"
#include "base/file.h"
#include "base/buffer.h"
#include "frame/conf.h"
#include "frame/servact.h"

#ifndef MCC_NEWS
#define TECH_MAXLINES 1024
#else
/* Active files require huge amounts of lines. */
#define TECH_MAXLINES 20000
#endif

#include <stdio.h>
#include "libadmin/libadmin.h"

/*
 * Define internal <Client> parameters,
 * i.e. those never written to obj.conf.
 */
#define NINTPARAM	3		/* count of internal parameters */
static char * int_params[NINTPARAM] = {
    "refcount",
    "dnsfilter",
    "ipfilter"
};

/* Define internal directive parameters too */
#define NDIRPARAM       1
static char *dir_params[NDIRPARAM] = {
    "Directive"
};

NSAPI_PUBLIC void dump_database_tofile(int whichsrv, char *fn, httpd_objset *os)
{
    register int x, y, z;
    int i;
    int nip;				/* count of entries in pp[] */
    pb_param *pp[NINTPARAM];		/* saved internal parameters */
    int ndp;				/* count of entries in dpp[] */
    pb_param *dpp[NDIRPARAM];		/* saved internal dir parameters */
    dtable *dt;
    const char *u;
    char *t;
    FILE *f;
    char objpath[PATH_MAX];

    if(fn)  {
        strcpy(objpath, fn);
    }  else  {
    /* Backup magnus.conf too to make sure numbers are synced. -MLM */
#ifdef NOTDEF /* MLM - was USE_ADMSERV - only back up at confirm time */
        sprintf(objpath, "%s%s", CONFDIR(whichsrv), MAGNUS_CONF);
        conf_backup(objpath, BK_MAGNUS, whichsrv);
#endif
        sprintf(objpath, "%s%s", CONFDIR(whichsrv), OBJ_DATABASE);
#ifdef NOTDEF /* MLM - was USE_ADMSERV - only back up at confirm time */
        conf_backup(objpath, BK_OBJ, whichsrv);
#endif
    }

    if(!(f = fopen_l(objpath, "w")))
        report_error(FILE_ERROR, NULL, 
                     "Can't open object database for writing.");

    fputs("# Fedora Project - obj.conf\n", f);
    fputs("# You can edit this file, but comments and formatting changes\n", f);
    fputs("# might be lost when the admin server makes changes.\n\n", f);

#if defined(MCC_HTTPD) && defined(XP_WIN32)
    fputs("# Use only forward slashes in pathnames--backslashes can cause\n", f);
    fputs("# problems. See the documentation for more information.\n\n", f);
#endif /* XP_WIN32 */

    /* Write out the Init directives first. */
    if(os->initfns)  {
        for(x = 0; os->initfns[x]; x++)  {
            char *direct = NULL;
            direct = pblock_pblock2str(os->initfns[x], NULL);
            fprintf(f, "Init %s\n", direct);
            FREE(direct);
        }
        fprintf(f, "\n");
    }

    for(x = 0; x < os->pos; x++) {
        t = pblock_pblock2str(os->obj[x]->name, NULL);
        fprintf(f, "<Object %s>\n", t);
        FREE(t);
        for(y = 0; y < os->obj[x]->nd; y++) {
            dt = &os->obj[x]->dt[y];
            u = directive_num2name(y);
            for(z = 0; z < dt->ni; z++) {
                if(dt->inst[z].client) {

		    /* Remove any internal parameters */
		    for (i = 0, nip = 0; i < NINTPARAM; ++i) {
			pp[nip] = pblock_remove(int_params[i],
						dt->inst[z].client);
			if (pp[nip]) ++nip;
		    }

                    t = pblock_pblock2str(dt->inst[z].client, NULL);

		    /* Restore internal parameters */
		    for (i = 0; i < nip; ++i) {
			pblock_pinsert(pp[i], dt->inst[z].client);
		    }

                    fprintf(f, "<Client %s>\n", t);
                    FREE(t);
                }

                /* Remove any internal directive parameters */
                for (i = 0, ndp = 0; i < NDIRPARAM; ++i) {
                    dpp[ndp] = pblock_remove(dir_params[i],
                                             dt->inst[z].param);
                    if (dpp[ndp]) ++ndp;
                }

                t = pblock_pblock2str(dt->inst[z].param, NULL);

                /* Restore internal parameters */
                for (i = 0; i < ndp; ++i) {
                    pblock_pinsert(dpp[i], dt->inst[z].param);
                }

                fprintf(f, "%s %s\n", u, t);
                FREE(t);
                if(dt->inst[z].client)
                    fputs("</Client>\n", f);
            }
        }
        fputs("</Object>\n\n", f);
    }
    fclose_l(f);
}

NSAPI_PUBLIC void dump_database(int whichsrv, httpd_objset *os)
{
    dump_database_tofile(whichsrv, NULL, os);
}


NSAPI_PUBLIC httpd_objset *read_config_from_file(char *objconf)
{
    char errstr[MAGNUS_ERROR_LEN];
    httpd_objset *os = NULL;
    filebuf_t *buf;
    SYS_FILE fd;
#ifdef XP_UNIX
    SYS_FILE lf;
#elif defined(XP_WIN32)
	HANDLE lf;
#endif
    char *lockpath;
    char *sn=get_srvname(0);
    char *flp=FILE_LOCK_PATH;

    if(!objconf)
	report_error(FILE_ERROR, NULL,"No obj.conf file location specified.\n");
    if((fd = system_fopenRO(objconf)) == SYS_ERROR_FD)
        report_error(FILE_ERROR, NULL, "Can't open object database.\n");
    lockpath=(char *) MALLOC(strlen(sn)+strlen(flp)+16);
    sprintf(lockpath, flp, sn);
#ifdef XP_UNIX
    if((lf = system_fopenRW(lockpath)) == SYS_ERROR_FD)
        report_error(FILE_ERROR, lockpath, "Can't open file.\n");
    if(system_flock(lf)==IO_ERROR)
        report_error(FILE_ERROR, lockpath, "Could not lock file.");
#elif defined(XP_WIN32)
    /* Using mutexes because if the CGI program dies, the mutex will be
     * automatically released by the OS for another process to grab.
     * Semaphores do not have this property; and if the CGI program crashes,
     * the admin server would be effectively crippled.
     */
    if ( (lf = CreateMutex(NULL, 0, lockpath)) == NULL) {
        report_error(FILE_ERROR, lockpath, "Could not create admin mutex.");
    } else {
        if ( WaitForSingleObject(lf, 60*1000) == WAIT_FAILED) {
            report_error(FILE_ERROR, lockpath, "Unable to obtain mutex after 60 seconds.");
        }
    }
#endif

    if(!(buf = filebuf_open(fd, FILE_BUFFERSIZE)))
        report_error(FILE_ERROR, NULL, "Can't open buffer from object file.\n");
    if(!(os = objset_scan_buffer(buf, errstr, NULL)))
        report_error(FILE_ERROR, NULL, 
                     "objset_scan returned NULL. Object file corrupt...\n");
    filebuf_close(buf);

#ifdef XP_UNIX
    if(system_ulock(lf)==IO_ERROR)
        report_error(FILE_ERROR, lockpath, "Could not unlock file.");
    system_fclose(lf);
#elif defined(XP_WIN32)
	if (lf) {
        ReleaseMutex(lf);
        CloseHandle(lf);
    }
#endif
    return os;
}


NSAPI_PUBLIC httpd_objset *read_config(int x)
{
    char objconf[512];

    sprintf(objconf, "%s%s", CONFDIR(x), OBJ_DATABASE);

    return read_config_from_file(objconf);
}


NSAPI_PUBLIC void insert_ntrans(pblock *p, pblock *c, httpd_object *obj)
{
    register dtable *d;
    register int x, y;
    register char *t;
#ifdef XP_UNIX
    int len;
#else /* XP_WIN32 */
    unsigned int len;
#endif /* XP_WIN32 */
    directive i;

    d = &obj->dt[directive_name2num("NameTrans")];
    t = pblock_findval("from", p);
    len = t ? strlen(t) : 0;

    for(x = 0; x < d->ni; x++) {
        if(!strcmp(pblock_findval("fn", d->inst[x].param), "document-root"))
            break;
        if((!(t = pblock_findval("from", d->inst[x].param))) ||
           ((strcmp(pblock_findval("fn", d->inst[x].param), "pfx2dir")) &&
            (strcmp(pblock_findval("fn", d->inst[x].param), "redirect"))))
            continue;
        else
            if(strlen(t) < len)
                break;
    }
    object_add_directive(directive_name2num("NameTrans"), p, c, obj);
    i = d->inst[d->ni - 1];

    for(y = d->ni - 1; y > x; y--)
        d->inst[y] = d->inst[y-1];

    d->inst[x] = i;
}

NSAPI_PUBLIC void insert_ntrans_an(pblock *p, pblock *c, httpd_object *obj)
{
    register dtable *d;
    register int x, y;
    register char *t;
#ifdef XP_UNIX
    int len;
#else /* XP_WIN32 */
    unsigned int len;
#endif /* XP_WIN32 */
    directive i;

    d = &obj->dt[directive_name2num("NameTrans")];
    
    if (strcmp(pblock_findval("fn", p), "mkssi-version") &&
        strcmp(pblock_findval("fn", p), "search-find") &&
        strcmp(pblock_findval("fn", p), "livewireNameTrans") )
      {
	len = strlen(pblock_findval("from", p));

	for(x = 0; x < d->ni; x++) 
	  {
	    if (!strcmp(pblock_findval("fn", d->inst[x].param), 
			"mkssi-version"))
	      continue;
	    if (!strcmp(pblock_findval("fn", d->inst[x].param), 
			"search-find"))
	      continue;
	    else if((!(t = pblock_findval("from", d->inst[x].param))) ||
		    (strcmp(pblock_findval("fn", d->inst[x].param), 
			    "assign-name")))
	      break;
	    else
	      if(strlen(t) < len)
		break;
	  }
      }
    else /* alex */
      x = 0;

    object_add_directive(directive_name2num("NameTrans"), p, c, obj);
    i = d->inst[d->ni - 1];

    for(y = d->ni - 1; y > x; y--)
        d->inst[y] = d->inst[y-1];

    d->inst[x] = i;
}


NSAPI_PUBLIC void insert_pcheck_mp(pblock *p, pblock *c, httpd_object *obj)
{
    register dtable *d;
    register int x, y;
    register char *t;
    directive i;

    d = &obj->dt[directive_name2num("PathCheck")];

    x = 0;

    object_add_directive(directive_name2num("PathCheck"), p, c, obj);
    i = d->inst[d->ni - 1];

    for(y = d->ni - 1; y > x; y--)
        d->inst[y] = d->inst[y-1];

    d->inst[x] = i;
}

NSAPI_PUBLIC void insert_alias(pblock *p, pblock *c, httpd_object *obj)
{
    register dtable *d;
    register int y;
    directive i;

    d = &obj->dt[directive_name2num("NameTrans")];
    object_add_directive(directive_name2num("NameTrans"), p, c, obj);
    i = d->inst[d->ni - 1];

    for(y = d->ni - 1; y; y--)
        d->inst[y] = d->inst[y-1];

    d->inst[0] = i;
}

NSAPI_PUBLIC char **scan_tech(char *fn)
{
    FILE *f;
    char **ret, buf[CONF_MAXLEN];
    register int x, l;

    if(!(f = fopen_l(fn, "r")))
        return NULL;

    ret = (char **) MALLOC(TECH_MAXLINES*sizeof(char *));
    x = 0;
    /* leave space at end for routine to insert a line */
    for(x = 0; x < (TECH_MAXLINES - 1); x++) {
        if(!(fgets(buf, CONF_MAXLEN, f)))
            break;
        /* F*** stdio. */
        l = strlen(buf) - 1;
        while(l && isspace(buf[l])) --l;
        buf[l+1] = '\0';
        ret[x] = STRDUP(buf);
    }
    fclose_l(f);
    ret[x] = NULL;
    return ret;
}


NSAPI_PUBLIC int write_tech(char *fn, char **lines)
{
    FILE *f;
    int x;

    if(!(f = fopen_l(fn, "w")))
        return -1;

    for(x = 0; lines[x]; x++) {
        if(strcmp(lines[x], "\n"))
            fprintf(f, "%s\n", lines[x]);
        else
            fprintf(f, "%s", lines[x]);
    }
    fclose_l(f);
    return 0;
}


NSAPI_PUBLIC httpd_object *findliteralppath(char *qs, httpd_objset *os)
{
    register int x;
    register char *t;

    for(x = 0; x < os->pos; x++) {
        if( (t = pblock_findval("ppath", os->obj[x]->name)) )
#ifdef XP_UNIX
            if(!strcmp(t, qs))
#else /*XP_WIN32 */
            if(!strnicmp(t, qs, strlen(qs)))
#endif /* XP_WIN32 */
                return os->obj[x];
    }
    return NULL;
}

