/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * objconf.c - Read in the obj.conf file, get info from it, and modify it
 *
 * All blame to Mike McCool
 */

#include "libadminutil/admutil.h"
#include "libadmin/libadmin.h"
#include "frame/servact.h"  /* NUM_DIRECTIVES */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

static httpd_objset **osarray = NULL;
static httpd_objset *os = NULL;

typedef struct reqstruct {
    int objtype;
    char *object;
    char *directive;
    char *function;
    char *id_type;
    char *id_value;
    pblock **pbs;
} reqstruct, *reqsptr;

static reqsptr *rqarray = NULL;
static int rqarraysize = 0;

/* Deal with multiple server config files.  When a pblock ptr is requested,
 * and it's found, make an array of what's in the other config files too.
 */ 

/* Note: Don't confuse these req structures with the ones used in NSAPI.
 * They're another beast altogether.  This is storing what the admin
 * program requested from the object set, not what the client requested
 * from the server. */
reqsptr new_rq_array(void)
{
    int numservs = make_conflist();

    if(rqarray == NULL)  {
        rqarraysize=1;
        rqarray = (reqsptr *) MALLOC(rqarraysize*sizeof(reqsptr)+2);
        rqarray[rqarraysize-1] = (reqsptr) MALLOC(sizeof(reqstruct)+2);
        rqarray[rqarraysize-1]->pbs = (pblock **) 
                                     MALLOC(numservs*sizeof(pblock *));
        return rqarray[rqarraysize-1];
    }  else  {
        rqarraysize++;
        rqarray = (reqsptr *) REALLOC(rqarray, rqarraysize*sizeof(reqsptr)+2);
        rqarray[rqarraysize-1] = (reqsptr) MALLOC(numservs*sizeof(reqstruct)+2);
        rqarray[rqarraysize-1]->pbs = (pblock **) 
                                     MALLOC(numservs*sizeof(pblock *));
        return rqarray[rqarraysize-1];
    }
    return NULL;
}

reqsptr find_rq_byptr(pblock *pb)
{
    reqsptr ans=NULL;
    register int x;

    for(x=0; x<rqarraysize; x++)  {
        ans = rqarray[x];
        if((ans->pbs)[0] == pb)
            return ans;
    }
    return NULL;
}

reqsptr find_rq_byreq(int objtype, char *object, 
                      char *directive, char *function,
                      int nargs, va_list arglist)
{
    register int x, y, numservs, found;
    reqsptr ans = NULL;
    reqsptr secondbest = NULL;

    for(x=0; x<rqarraysize; x++)  {
        ans = rqarray[x];
        if(ans->objtype != objtype) 
            continue;

        /* check for the NULLs put in by client ptrs. */
        if( ((!ans->object) && (object)) ||
            ((!ans->directive) && (directive)) ||
            ((!ans->function) && (function)) )
            continue;

        if( !( (!(ans->object) && !(object)) ||
               (!strcmp(ans->object, object)) ) )
            continue;
        if( !( (!(ans->directive) && !(directive)) ||
               (!strcmp(ans->directive, directive)) ) )
            continue;
        if( !( (!(ans->function) && !(function)) ||
               (!strcmp(ans->function, function)) ) )
            continue;

        /* Check to see if the pblocks are all NULL.  Arguably, this
         * should be taken care of by not creating the rq instead of
         * futzing around with it here, but I found it easier to just
         * do it here rather than special casing the ones that create
         * the rq. */
        for(numservs = make_conflist(), y=0, found=0; y<numservs; y++)  {
            if(ans->pbs[y] != NULL) found=1;
        }
        if(!found) continue;

        /* If there wasn't a specific pblock being looked for,
         * don't return it, because it may be ambiguous. */
        /* However, keep it around, because if you don't find
         * anything specific, then perhaps this should be it. */
        if((!ans->id_type) && (!ans->id_value))  {
            secondbest = ans;
            continue;
        }

        for(y=0; y<nargs; y+=2)  {
            char *curname, *curvalue;

            curname = STRDUP(va_arg(arglist, char *));
            curvalue = STRDUP(va_arg(arglist, char *));

            if(!strcmp(ans->id_type, curname))  {
                if(!strcmp(ans->id_value, curvalue))  {
                    return ans;
                }
            }
        }
    }
    return secondbest;
}

void read_obj_conf(void) 
{
#ifdef USE_ADMSERV
    if(osarray == NULL)  {
        int numservs = make_conflist();
        register int x=0;

        osarray = (httpd_objset **) MALLOC(numservs*sizeof(httpd_objset *));
        for(x=0; x < numservs; x++)  {
            osarray[x] = read_config(x);
        }
        os = osarray[0];
    }
#else
    httpd_object *obj = NULL;

    if(os == NULL)
        os = read_config(0);
    if(os == NULL)
        report_error(FILE_ERROR, NULL, 
                     "Could not read configuration file obj.conf.");

    if(!(obj = objset_findbyname("default", NULL, os)))
         report_error(FILE_ERROR, "Error in obj.conf", 
                      "There was no default object in your obj.conf file - "
                      "It must have gotten corrupted somehow.");
#endif
}

void write_obj_conf(void)
{
#ifdef USE_ADMSERV
    int numservs = make_conflist();
    register int x=0;
 
    for(x=0; x < numservs; x++)  {
        dump_database(x, osarray[x]);
	set_commit(x, 1);
    }
#else
    dump_database(os);
    set_commit(0, 1);
#endif
}

NSAPI_PUBLIC void add_object(int objtype, char *id)
{
    httpd_object *obj = NULL;
    pblock *newpb;
    int numservs = make_conflist();
    register int x;

    read_obj_conf();

    for(x=0; x<numservs; x++)  {
        os = osarray[x];

        if(objtype == PB_NAME)
            obj = objset_findbyname(id, NULL, os);
        else if(objtype == PB_PATH)
            obj = findliteralppath(id, os);
        else 
            report_error(INCORRECT_USAGE, NULL, 
                         "Unknown object type for pblock.");
        if(obj) continue;

        newpb = pblock_create(PARAMETER_HASH_SIZE);

        pblock_nvinsert((objtype == PB_PATH) ? "ppath" : "name", id, newpb);
        obj = objset_new_object(newpb, os);
    }

    write_obj_conf();
}

NSAPI_PUBLIC httpd_object *grab_object(int objtype, char *id)
{
    httpd_object *obj = NULL;

    read_obj_conf();

    if(objtype == PB_NAME)
        obj = objset_findbyname(id, NULL, os);
    else if(objtype == PB_PATH)
        obj = findliteralppath(id, os);
    else 
        report_error(INCORRECT_USAGE, NULL, "Unknown object type for pblock.");

    return obj;
}

NSAPI_PUBLIC void delete_object(int objtype, char *id)
{
    register int x;
    register int y;
    register int z;
    httpd_object *obj = NULL;
    int numservs = make_conflist();

    read_obj_conf();

    for(z=0; z<numservs; z++)  {
        os = osarray[z];

        if(objtype == PB_NAME)
            obj = objset_findbyname(id, NULL, os);
        else if(objtype == PB_PATH)
            obj = findliteralppath(id, os);
        else
            report_error(INCORRECT_USAGE, NULL, 
                         "Unknown object type for pblock.");

#if 0   
        /* BONEHEAD - what to do.  Do we make a special error type,
         * which says, warning, we couldn't find this object in this
         * server, what do we do. */
        if(!obj)
            report_error(INCORRECT_USAGE, NULL, 
                     "Object could not be deleted (because it was not found.)");
#else
        if(!obj)
            continue;
#endif

        os->pos--; /* decrement the resource count */
        for(x=0; os->obj[x]; x++)  {
            if (os->obj[x] == obj)  {
                for(y=x; y < os->pos; y++)
                    os->obj[y] = os->obj[y+1];
                break;
            }
        }
    }

    write_obj_conf();
}

NSAPI_PUBLIC char **list_objects(int objtype)
{
    register int x;
    int size = 0;
    char **answer = NULL;
    register char *t;
    char *type = (objtype == PB_NAME ? "name" : "ppath");

    read_obj_conf();

    for(x=0; x < os->pos; x++)  {
        if( (t = pblock_findval(type, os->obj[x]->name)) )  {
            if(!answer)  {
                answer=new_strlist(5);
            }  else  {
                if( !((size+1) % 5) )
                    answer = grow_strlist(answer, size+5);
            }
            answer[size++] = STRDUP(t);
            answer[size] = NULL;
        }
    }
    return answer;
}


NSAPI_PUBLIC int count_objects(int objtype)
{
    int cnt = 0;
    register int x;
    register char *t;
    char *type = (objtype == PB_NAME ? "name" : "ppath");

    read_obj_conf();

    for(x=0; x < os->pos; x++)  {
        if( (t = pblock_findval(type, os->obj[x]->name)) )  {
	    cnt++;
        }
    }

    return cnt;
}


NSAPI_PUBLIC int total_object_count(void)
{
    read_obj_conf();

    return os->pos;
}


NSAPI_PUBLIC pblock *grab_pblock_byid(int objtype,char *object, char *directive,
                     char *fname, char *function, char *id_type, char *id_value)
{
    httpd_object *obj = NULL;
    dtable *d;
    pblock *pb = NULL, *ans = NULL;
    register int x;
    int z, numservs = make_conflist();
    char *pbval;  /* 30886 */
#ifdef DO_CLUSTER
    reqsptr req=NULL;
#endif 

    read_obj_conf();
#ifdef XP_WIN32
        if(objtype == PB_PATH) {
                CONVERT_TO_HTTP_FORMAT(object);
        }
#endif /* XP_WIN32 */
 
    for(z=0; z<numservs; z++)  {
        os = osarray[z];
        pb = NULL;
 
        if(objtype == PB_NAME)
            obj = objset_findbyname(object, NULL, os);
        else if(objtype == PB_PATH)
            obj = findliteralppath(object, os);
        else
            report_error(INCORRECT_USAGE, NULL,
                         "Unknown object type for pblock.");
 
        if(obj == NULL)  {
            continue;
        }
 
        d = &obj->dt[directive_name2num(directive)];
        for(x = 0; x < d->ni; x++) {
            pbval=pblock_findval(fname, d->inst[x].param); /* 30886 */
            if(!pbval) 
                continue;  /* 30886 */

            if(!strcmp(pbval, function))  /* 30886 */
            {
                if(id_type)  {
                    char *tmp = pblock_findval(id_type, d->inst[x].param);
#ifdef XP_WIN32
                    CONVERT_TO_HTTP_FORMAT(id_value);
#endif /* XP_WIN32 */
                    if((!tmp && strcmp(id_value, "NONE")) || 
                       ( tmp && strcmp(tmp, id_value)) )
                        continue;
                }
                pb = d->inst[x].param;
                break;
            }
        }
        if(z==0) ans=pb;
    }
    return ans;
}

NSAPI_PUBLIC pblock *grab_pblock(int objtype, char *object, char *directive, 
                    char *function, char *id_type, char *id_value)
{
    return grab_pblock_byid(objtype, object, directive, "fn", function,
                            id_type, id_value);
}

/* prototype for later use */
pblock *_set_pblock_vals(pblock *curpb, int nargs, va_list arglist);

pblock *_add_pblock(int objtype, char *object, char *my_directive, 
                    char *function, int nargs, va_list arglist)
{
    httpd_object *obj = NULL;
    dtable *d;
    pblock *pb = NULL;
    register int x = 0;
    char *varname, *varval;

#ifdef XP_WIN32
    if(objtype == PB_PATH) {
        CONVERT_TO_HTTP_FORMAT(object);
    }
#endif /* XP_WIN32 */
 
    if(objtype == PB_NAME)
        obj = objset_findbyname(object, NULL, os);
    else if(objtype == PB_PATH)
        obj = findliteralppath(object, os);
    else
        report_error(INCORRECT_USAGE, NULL,
                     "Unknown object type for pblock.");

    if(obj == NULL)  {
        pblock *newpb = pblock_create(PARAMETER_HASH_SIZE);
        pblock_nvinsert((objtype == PB_PATH) ? "ppath" : "name",
                        object, newpb);
        obj = objset_new_object(newpb, os);
    }

    d = &obj->dt[directive_name2num(my_directive)];

    pb = pblock_create(PARAMETER_HASH_SIZE);
    pblock_nvinsert("fn", function, pb);

    /* va_start(arglist, nargs); */
    while(x < nargs)  {
        varname = STRDUP(va_arg(arglist, char *));
        varval = STRDUP(va_arg(arglist, char *));
        if(strcmp(varval, "NONE")) {
#ifdef XP_WIN32
        CONVERT_TO_HTTP_FORMAT(varval);
#endif /* XP_WIN32 */
        pblock_nvinsert(varname, varval, pb);
 
        }
        x += 2;
    }
    /* va_end(arglist); */

    if (strcmp(my_directive, "NameTrans") ||
         ( strcmp(function, "pfx2dir") && 
           strcmp(function, "redirect") &&
           strcmp(function, "unix-home") &&
           strcmp(function, "assign-name") && 
           strcmp(function, "home-page") &&
           strcmp(function, "java-find-applet") &&
           strcmp(function, "es-search-nametrans") &&
	   strcmp(function, "agent_name_trans") &&
           strcmp(function, "livewireNameTrans") ))
        object_add_directive(directive_name2num(my_directive),
                             pb, NULL, obj);
    else  {
/* Yet another special casing.  Bertha get my gun.  assign-names should be
 * first, before all else.  On top of that, they should be sorted amongst
 * themselves.
 */
      if( strcmp(function, "assign-name") &&
              strcmp(function, "livewireNameTrans") )
	insert_ntrans(pb, NULL, obj);
      else
	insert_ntrans_an(pb, NULL, obj);
    }

    if(!(strcmp(my_directive, "Service")))  {
    /* send-file has to be last and key-toosmall must be first */
        if(d->ni > 1)  {      /* d->ni is total number of instances */
            directive i;
            register int x;

    /* Get the new directive which was added last in the list by above code */
            i = d->inst[d->ni - 1];
            if(!strcmp(function, "key-toosmall")) {
    /* And do remember to go in top-down fashion. */
                for(x = (d->ni - 2); x >= 0; x--) {
                    d->inst[x + 1] = d->inst[x];
                }
                d->inst[0] = i;  /* add to beginning of list */
            }
            else 
              if (strcmp(function, "send-file")) {
    /* if send-file, nothing to be done. send-file *should* be last */

    /* d->ni-2 is index of the directive that's next to last in the list */
                char *tfn = pblock_findval("fn", d->inst[d->ni - 2].param);

    /* x is the index of the directive just before the new directive */
                x = d->ni - 2;

    /* check to see if a new Service directive was added after the send-file
       directive, and if there was, make sure that send-file is always last */
                if(!strcmp(tfn, "send-file")) {
                    d->inst[d->ni - 1] = d->inst[d->ni - 2];
                    d->inst[d->ni - 2] = i;
                    x = d->ni - 3;  /* since i is in d->ni-2 */
                }
    /* check for content manager directives, if any are found make sure
       they go later (but before send-file) */
                for (; x >= 0; x--) {
                    tfn = pblock_findval("fn", d->inst[x].param);
                    if (strncmp(tfn , "CM_", 3))
                       break;    /* non CM directive found */
                    else {
                       d->inst[x + 1] = d->inst[x];
                       d->inst[x] = i;
                    }
                }

            } /* not send-file */
        } /* d->ni > 1 */
    } /* Service directive */
    return pb;
}

NSAPI_PUBLIC pblock *add_pblock(int objtype, char *object, char *my_directive, 
                   char *function, int nargs, ...)
{
    va_list arglist;
    int y, numservs, foundold=0;
    pblock *pb = NULL, *ans = NULL;
    reqsptr rq=NULL;

    read_obj_conf();
    numservs = make_conflist();

    foundold=0;

    for(y=0; y<numservs; y++)  {
        os = osarray[y];

        va_start(arglist, nargs);
        if((foundold) && (rq->pbs[y]))  {
            pb = _set_pblock_vals(rq->pbs[y], nargs, arglist);
        } else {
            pb = _add_pblock(objtype, object, my_directive, function,
                             nargs, arglist);
        }
        va_end(arglist);
        if(y==0) ans=pb;
    }
    write_obj_conf();

    return ans;
}

NSAPI_PUBLIC void delete_pblock(int objtype, char *object, char *directive, char *function,
                    char *id_type, char *id_value)
{
    httpd_object *obj = NULL;
    dtable *d;
    register int x, y, found = 0;
    int z, numservs = make_conflist();

    read_obj_conf();
#ifdef XP_WIN32
	if(objtype == PB_PATH) {
		CONVERT_TO_HTTP_FORMAT(object);
	}
#endif /* XP_WIN32 */

    for(z=0; z<numservs; z++)  {
        os = osarray[z];

        if(objtype == PB_NAME)  
            obj = objset_findbyname(object, NULL, os);
        else if(objtype == PB_PATH)
            obj = findliteralppath(object, os);
        else 
            report_error(INCORRECT_USAGE, NULL, 
                         "Unknown object type for pblock.");

        if(obj == NULL)
            continue;

        d = &obj->dt[directive_name2num(directive)];

        for(x = 0; x < d->ni; x++) {
            if(!strcmp(pblock_findval("fn", d->inst[x].param), function)) 
            {
                 if(id_type)  {
                     char *tmp = pblock_findval(id_type, d->inst[x].param);
#ifdef XP_WIN32
                     CONVERT_TO_HTTP_FORMAT(id_value);
#endif /* XP_WIN32 */
                     if((!tmp && strcmp(id_value, "NONE")) || 
                        ( tmp && strcmp(tmp, id_value)))
                         continue;
                 }
                 found = 1;
                 break;
            }
        }

        if(!found) continue;

        --d->ni;
        for(y=x; y < d->ni; y++)
            d->inst[y] = d->inst[y+1];
    
        if(strcmp(object, "default"))  {
            for(x=0, found = 0; x < NUM_DIRECTIVES; x++)  {
                d = &obj->dt[x];
                found += d->ni;
            }
            if(!found)  {
                os->pos--; /* decrement the resource count */
                for(x=0; os->obj[x]; x++)  {
                    if (os->obj[x] == obj)  {
                        for(y=x; y < os->pos; y++)
                            os->obj[y] = os->obj[y+1];
                        break;
                    }
                }
            }
        }
    }
    write_obj_conf();

}

pblock *_set_pblock_vals(pblock *curpb, int nargs, va_list arglist)
{
    pb_param *prm;
    char *varname, *varval;
    register int x=0;

    if(!curpb)  {
        return NULL;
    }  else  {
        while(x < nargs)  {
            varname = STRDUP(va_arg(arglist, char *));
            varval = STRDUP(va_arg(arglist, char *));

            x += 2;

            if((prm = pblock_find(varname, curpb)))  {
                if(strcmp(varval, "NONE"))  {
                    FREE(prm->value);
                    prm->value = varval;
                }  else  {
                    param_free(pblock_remove(varname, curpb));
                }
            }  else  {
#ifdef XP_WIN32
                CONVERT_TO_HTTP_FORMAT(varval);
#endif /* XP_WIN32 */
                if(strcmp(varval, "NONE"))
                    pblock_nvinsert(varname, varval, curpb);
                else
                    continue;
            }
        }
    }
    return curpb;
}

NSAPI_PUBLIC void set_pblock_vals(pblock *pb, int nargs, ...)
{
    va_list arglist;
    int numservs=make_conflist();
    int y;

    {   /* Check to see if the pblock has _really_ changed. */
        register int x = 0;
        char *varname, *varval, *oldval;
        int haschanged = 0;

        va_start(arglist, nargs);
        while((x<nargs) && (!haschanged))  {
            varname = STRDUP(va_arg(arglist, char *));
            varval = STRDUP(va_arg(arglist, char *));
            x+=2;

            if(!strcmp(varval, "NONE"))
                varval = NULL;

            oldval = pblock_findval(varname, pb);

            if((!oldval) && (!varval))
                continue;
            if(((!oldval) && (varval)) || ((oldval) && (!varval)))  {
                haschanged = 1;
                continue;
            }
            if(strcmp(oldval, varval))  {
                haschanged = 1;
                continue;
            }
        }
        va_end(arglist);
        if(!haschanged) return;
    }
    
    for(y=0; y<numservs; y++)  {
        os = osarray[y];

        va_start(arglist, nargs);
        _set_pblock_vals(pb, nargs, arglist);
        va_end(arglist);
    }

    write_obj_conf();
}

NSAPI_PUBLIC pblock *new_pblock(char *function, int nargs, ...)
{
    pblock *pb = pblock_create(PARAMETER_HASH_SIZE);
    va_list arglist;
    char *varname;
    char *varval;
    register int x = 0;

    pblock_nvinsert("fn", function, pb);

    va_start(arglist, nargs);
    while(x < nargs)  {
        varname = STRDUP(va_arg(arglist, char *));
        varval = STRDUP(va_arg(arglist, char *));
        if(strcmp(varval, "NONE")) {
#ifdef XP_WIN32
			CONVERT_TO_HTTP_FORMAT(varval);
#endif /* XP_WIN32 */
            pblock_nvinsert(varname, varval, pb);

        }
        x += 2;
    }
    va_end(arglist);

    return pb;
}

NSAPI_PUBLIC pblock **list_pblocks(int objtype, char *object, char *direct, char *function)
{
    httpd_object *obj = NULL;
    dtable *d;
    register int x, total = 0;
    pblock **answer = NULL;

    read_obj_conf();
#ifdef XP_WIN32
	if(objtype == PB_PATH) {
		CONVERT_TO_HTTP_FORMAT(object);
	}
#endif /* XP_WIN32 */

    if(objtype == PB_NAME)
        obj = objset_findbyname(object, NULL, os);
    else if(objtype == PB_PATH)
        obj = findliteralppath(object, os);
    else
        report_error(INCORRECT_USAGE, NULL, "Unknown object type for pblock.");

    if(obj == NULL)
        return NULL;

    d = &obj->dt[directive_name2num(direct)];

    for(x = 0; x < d->ni; x++) {
        if(!strcmp(pblock_findval("fn", d->inst[x].param), function))
        {
            if(!answer) {
                answer = (pblock **)MALLOC((++total+1) * sizeof(pblock *));
            }  else  {
                answer = (pblock **)REALLOC(answer, 
                                            (++total+1) * sizeof(pblock *));
            }
            answer[total-1] = d->inst[x].param;
            answer[total] = NULL;
        }
    }
    return answer;
}

/* These client things will probably (I hope) have more functionality later.
 * Right now they're just kind of cut and paste from above. */

NSAPI_PUBLIC pblock *grab_client(int objtype, char *object, char *directive, 
                    char *function, char *id_type, char *id_value)
{
    httpd_object *obj = NULL;
    dtable *d;
    pblock *pb = NULL, *ans = NULL;
    register int x, y;
    int numservs=make_conflist();

    read_obj_conf();

    for(y=0; y<numservs; y++)  {
        os = osarray[y];
        pb = NULL;

#ifdef XP_WIN32
        if(objtype == PB_PATH) {
            CONVERT_TO_HTTP_FORMAT(object);
        }
#endif /* XP_WIN32 */

        if(objtype == PB_NAME)  
            obj = objset_findbyname(object, NULL, os);
        else if(objtype == PB_PATH)
            obj = findliteralppath(object, os);
        else 
            report_error(INCORRECT_USAGE, NULL, 
                         "Unknown object type for pblock.");
 
        if(obj == NULL)
            continue;

        d = &obj->dt[directive_name2num(directive)];

        for(x = 0; x < d->ni; x++) {
            if(!strcmp(pblock_findval("fn", d->inst[x].param), function)) 
            {
                if(id_type)  {
                    char *tmp = pblock_findval(id_type, d->inst[x].param);
#ifdef XP_WIN32
                    CONVERT_TO_HTTP_FORMAT(id_value);
#endif /* XP_WIN32 */
                    if((!tmp) || (strcmp(tmp, id_value)))
                        continue;
                }
                pb = d->inst[x].client;
                break;
            }
        }
        if(y==0) ans=pb;
    }
    return ans;
}

NSAPI_PUBLIC void add_client(int objtype, char *object, char *direct, 
                pblock *oldpb, int nargs, ...)
{
    httpd_object *obj = NULL;
    pblock *pb = NULL, *clnt = NULL;
    register int x = 0;
    va_list arglist;
    char *varname, *varval;
    int numservs, y;
    reqsptr req = NULL;

    read_obj_conf();
    numservs=make_conflist();

    for(y=0; y<numservs; y++)  {
        os = osarray[y];
        x = 0;

#ifdef XP_WIN32
        if (objtype == PB_PATH) {
            CONVERT_TO_HTTP_FORMAT(object);
        }
#endif /* XP_WIN32 */

        if(objtype == PB_NAME)
            obj = objset_findbyname(object, NULL, os);
        else if(objtype == PB_PATH)
            obj = findliteralppath(object, os);
        else
            report_error(INCORRECT_USAGE, NULL, 
                         "Unknown object type for pblock.");

        if(obj == NULL)  {
            pblock *newpb = pblock_create(PARAMETER_HASH_SIZE);
            pblock_nvinsert((objtype == PB_PATH) ? "ppath" : "name", 
                            object, newpb);
            obj = objset_new_object(newpb, os);
        }

        {
            if(oldpb)
                pb = oldpb;
            else  {
                pb = pblock_create(PARAMETER_HASH_SIZE);
                pblock_nvinsert("fn", "deny-existence", pb);
            }
        }

        clnt = pblock_create(PARAMETER_HASH_SIZE);

        va_start(arglist, nargs);
        while(x < nargs)  {
            varname = STRDUP(va_arg(arglist, char *));
            varval = STRDUP(va_arg(arglist, char *));
            if(strcmp(varval, "NONE")) {
#ifdef XP_WIN32
                CONVERT_TO_HTTP_FORMAT(varval);
#endif /* XP_WIN32 */
                pblock_nvinsert(varname, varval, clnt);
            }
            x += 2;
        }
        va_end(arglist);

        if((req && !req->pbs[y]) || (!oldpb))
            object_add_directive(directive_name2num(direct), pb, clnt, obj);
        else  {
            directive *drct = get_pb_directive(objtype, object, direct, 
                              req ? req->pbs[y] : pb);
            if (!drct)
                object_add_directive(directive_name2num(direct), 
                                     req ? req->pbs[y] : pb, clnt, obj);
            else
                drct->client = clnt;
        }
    }
 
    write_obj_conf();
}

NSAPI_PUBLIC directive **list_clients(int objtype, char *object, char *direct, 
                         char *function)
{
    httpd_object *obj = NULL;
    dtable *d;
    register int x, total = 0;
    directive **answer = NULL;

    read_obj_conf();
    os = osarray[0];

#ifdef XP_WIN32
    if(objtype == PB_PATH) {
        CONVERT_TO_HTTP_FORMAT(object);
    }
#endif /* XP_WIN32 */

    if(objtype == PB_NAME)
        obj = objset_findbyname(object, NULL, os);
    else if(objtype == PB_PATH)
        obj = findliteralppath(object, os);
    else
        report_error(INCORRECT_USAGE, NULL, 
                     "Unknown object type for pblock.");

    if(obj == NULL)
        return NULL;

    d = &obj->dt[directive_name2num(direct)];

    for(x = 0; x < d->ni; x++) {
        if(!strcmp(pblock_findval("fn", d->inst[x].param), function))
        {
            char *tmp;

            if(!answer) {
                answer = (directive **)
                         MALLOC((++total+1)*sizeof(directive *));
            }  else  {
                answer = (directive **)REALLOC(answer,
                                        (++total+1)*sizeof(directive *));
            }
            answer[total-1] = &d->inst[x];
            answer[total] = NULL;

            tmp=pblock_findval("path", d->inst[x].param);

	    if(tmp) {
	      if((grab_client(objtype, object, direct, function,
			      "path", STRDUP(tmp)))
		 != (d->inst[x].client))
                report_error(INCORRECT_USAGE, "Diagnostic",
                             "Incorrect ptr returned from grab_client");
	    }
	    else {
	      if((grab_client(objtype, object, direct, function,
			      NULL, NULL))
		 != (d->inst[x].client))
                report_error(INCORRECT_USAGE, "Diagnostic",
                             "Incorrect ptr returned from grab_client");
	    }

            os = osarray[0];
        }
    }
    return answer;
}

NSAPI_PUBLIC void delete_client(int objtype, char *object, char *direct, char *path, 
                   char *dns, char *ip)
{
    httpd_object *obj = NULL;
    dtable *d;
    register int x, y, found;
    char *tmp;
    int numservs, z;

    read_obj_conf();

    numservs=make_conflist();
    for(z=0; z<numservs; z++)  {
        os=osarray[z];
        found=0;
        x=0;
        y=0;

#ifdef XP_WIN32
        if(objtype == PB_PATH) {
            CONVERT_TO_HTTP_FORMAT(object);
        }
	if(path) {
            CONVERT_TO_HTTP_FORMAT(path);
	}
#endif /* XP_WIN32 */

        if(objtype == PB_NAME)
            obj = objset_findbyname(object, NULL, os);
        else if(objtype == PB_PATH)
            obj = findliteralppath(object, os);
        else
            report_error(INCORRECT_USAGE, NULL, 
                         "Unknown object type for pblock.");

        if(obj == NULL)
            return;

        d = &obj->dt[directive_name2num(direct)];

        for(x = 0; x < d->ni; x++) {
            if(!strcmp(pblock_findval("fn", d->inst[x].param), 
                "deny-existence"))  {
                tmp = pblock_findval("path", d->inst[x].param);
                if(!tmp || !path)  {
                    if( !(!tmp && !path) )
                        continue;
                }  else  {
                    if(strcmp(tmp, path))
                        continue;
                }
                tmp = pblock_findval("dns", d->inst[x].client);
                if(!tmp || !dns)  {
                    if( !(!tmp && !dns) )
                        continue;
                }  else  {
                    if(strcmp(tmp, dns))
                        continue;
                }
                tmp = pblock_findval("ip", d->inst[x].client);
                if(!tmp || !ip)  {
                    if( !(!tmp && !ip) )
                        continue;
                }  else  {
                    if(strcmp(tmp, ip))
                        continue;
                }
                /* wow.  this must be it. */
                --d->ni;
                for(y=x; y < d->ni; y++)
                    d->inst[y] = d->inst[y+1];
                found = 1;
                break;
            }
        }
        if(!found) continue;
    }

    write_obj_conf();
}

directive *get_pbcl_directive(int objtype, char *object, 
                              char *directive, pblock *pb, int use_client)
{
    httpd_object *obj = NULL;
    dtable *d;
    register int x;

    read_obj_conf();

    if(objtype == PB_NAME)
        obj = objset_findbyname(object, NULL, os);
    else if(objtype == PB_PATH)
        obj = findliteralppath(object, os);
    else
        report_error(INCORRECT_USAGE, NULL, "Unknown object type for pblock.");

    if(obj == NULL)
        return NULL;

    d = &obj->dt[directive_name2num(directive)];

    if(use_client) {
        for(x = 0; x < d->ni; x++) {
            if(d->inst[x].client == pb)  
                return &d->inst[x];
        }
    }
    else {
        for(x = 0; x < d->ni; x++) {
            if(d->inst[x].param == pb)  
                return &d->inst[x];
        }
    }
    return NULL;
}


NSAPI_PUBLIC directive *get_pb_directive(int objtype, char *object, 
                            char *directive, pblock *pb)
{
    return get_pbcl_directive(objtype, object, directive, pb, 0);
}

NSAPI_PUBLIC directive *get_cl_directive(int objtype, char *object, 
                            char *directive, pblock *cl)
{
    return get_pbcl_directive(objtype, object, directive, cl, 1);
}

NSAPI_PUBLIC void delete_pblock_byptr(int objtype, char *object, char *directive, pblock *pb)
{
    httpd_object *obj = NULL;
    dtable *d;
    register int x, found;
    int numservs, z;

    read_obj_conf();
    numservs=make_conflist();

    for(z=0; z<numservs; z++)  {
        os=osarray[z];
        found=0;

        if(objtype == PB_NAME)
            obj = objset_findbyname(object, NULL, os);
        else if(objtype == PB_PATH)
            obj = findliteralppath(object, os);
        else
            report_error(INCORRECT_USAGE, NULL, 
                         "Unknown object type for pblock.");

        if(obj == NULL)
            return;

        d = &obj->dt[directive_name2num(directive)];

        for(x = 0; x < d->ni; x++) {
            if(d->inst[x].param == pb)  {
                found = 1;
                d->inst[x].param = NULL;
            }
            else if(d->inst[x].client == pb)  {
                found = 1;
                d->inst[x].client = NULL;
            }
            if((found) && 
                (d->inst[x].param == NULL) && (d->inst[x].client == NULL))  {
                int y;
                --d->ni;
                for(y=x; y < d->ni; y++)
                   d->inst[y] = d->inst[y+1];
            }
        }
    }
    write_obj_conf();
}

/* Init directives now reside in obj.conf.  Deal with them here. */
/* ------------------------------------------------------------- */

/* Deal with those crazy init directives. */
NSAPI_PUBLIC pblock *get_mag_init(char *fn)
{
    register int x;
    char *tmp;
 
    read_obj_conf();
 
    os = osarray[0];
    if(!os->initfns)
        return NULL;
 
    for(x=0; os->initfns[x]; x++)  {
        if( (tmp = pblock_findval("fn", os->initfns[x])) )  {
            if(!strcmp(tmp, fn))  
                return os->initfns[x];
        }
    }
    return NULL;
}
 
/* Get only a particular mag init */
NSAPI_PUBLIC pblock *get_specific_mag_init(char *fn, char *name, char *value)
{
  int x;
  pblock *pb = NULL;
  pblock **pb_arr = NULL;
  char *val;
 
  if (fn && name && value)
    pb_arr = get_all_mag_inits(fn);
 
  if (pb_arr)
    for(x=0; pb_arr[x]; x++)
      {
        val = pblock_findval(name, pb_arr[x]);
 
        if (val)
          if (!strcmp(value, val))
            {
              pb = pb_arr[x];
              break;
            }
      }
 
  return pb;
}

/* Get all instances of the same Init function as an array of pblock ptrs */
NSAPI_PUBLIC pblock **get_all_mag_inits(char *fn)
{
    register int x, i=0, siz=4;
    pblock *pb, **ret=NULL;
    char *tmp;
 
    read_obj_conf();
 
    os = osarray[0];
    if(!os->initfns)
        return NULL;
 
    for(x=0; os->initfns[x]; x++) {
        pb = os->initfns[x];
        if( (tmp = pblock_findval("fn", pb)) && (!strcmp(tmp, fn)) ) {
            if (!ret) {
                ret = (pblock **)MALLOC(siz * sizeof(pblock*));
                ZERO(ret, siz * sizeof(pblock*));
            }
            else if (i + 1 >= siz) {
                siz *= 2;
                ret = (pblock **)REALLOC(ret, siz * sizeof(pblock*));
            }
            ret[i++] = pb;
        }
    }
 
    if(ret)
        ret[i] = NULL;
 
    return ret;
}
 
NSAPI_PUBLIC void set_mag_init(char *fn, char *key_nam, char *key_val, int nargs, ...)
{
    register int x, y, found = 0;
    pblock *pb = 0;
    char *tmp1, *tmp2;
    va_list arglist;
    char *varname, *varval;
    int numservs = make_conflist(), z;
 
    read_obj_conf();
 
    if (key_nam && !key_val)
        key_val = "";
 
    for(z=0; z<numservs; z++)  {
        os=osarray[z];
 
        if(os->initfns) {
            for(x=0, found = 0; os->initfns[x]; x++)  {
                pb=os->initfns[x];
                if ( (tmp1 = pblock_findval("fn", pb)) &&
                     (!strcmp(tmp1, fn)) &&
                     (!key_nam || ((tmp2 = pblock_findval(key_nam, pb)) &&
                                   (!strcmp(tmp2, key_val)))) )  {
                    found = 1;
                    break;
                }
            }
        }
        if(found)  {
            pb_param *prm;
            y = 0;
 
            va_start(arglist, nargs);
            while (y < nargs)  {
                varname = STRDUP(va_arg(arglist, char *));
                varval = STRDUP(va_arg(arglist, char *));
 
                if((prm = pblock_find(varname, pb)))  {
                    if(strcmp(varval, "NONE"))  {
                        free(prm->value);
                        prm->value = varval;
                    }  else  {
                        param_free(pblock_remove(varname, pb));
                    }
                }  else  {
                    if(strcmp(varval, "NONE"))
                        pblock_nvinsert(varname, varval, pb);
                    /* else shrug */
                }
                y += 2;
            }
            va_end(arglist);
        }  else  {  /* !found */
            pb = pblock_create(PARAMETER_HASH_SIZE);
 
            x = 0;
            va_start(arglist, nargs);
            while(x < nargs)  {
                varname = STRDUP(va_arg(arglist, char *));
                varval = STRDUP(va_arg(arglist, char *));
 
                if(strcmp(varval, "NONE"))
                    pblock_nvinsert(varname, varval, pb);
                x+=2;
            }
            va_end(arglist);

            if(!pblock_findval("fn", pb))
                pblock_nvinsert("fn", fn, pb);

            if(!os->initfns)  {
                pblock **pblist = NULL;
                pblist = (pblock **) MALLOC(4*sizeof(pblock *));
                pblist[0] = pb;
                pblist[1] = NULL;
                os->initfns = pblist;
            }  else  {
                int size;
                for(size=0; os->initfns[size]; size++);
                x=(size+4) * sizeof(pblock *);
                os->initfns = (pblock **) REALLOC(os->initfns, x);
                os->initfns[size++] = pb;
                os->initfns[size] = NULL;
            } 
        }
    }
    write_obj_conf();
}

NSAPI_PUBLIC void delete_mag_init(char *fn)
{
    register int x, found = 0;
    pblock *pb;
    char *tmp;
    int numservs=make_conflist(), y=0;
 
    read_obj_conf();
 
    for(y=0; y<numservs; y++)  {
        os = osarray[y];
 
        if(!os->initfns) continue;

        for(x=0, found = 0; os->initfns[x]; x++)  {
            pb = os->initfns[x];
            if( (tmp = pblock_findval("fn", pb)) )  {
                if(!strcmp(tmp, fn))  {
                    found = 1;
                    break;
                }
            }
        }
        if(found)  {
            int z = 0;
            for(z=x; os->initfns[z]; z++)  {
                os->initfns[z] = os->initfns[z+1];
            }
        }
    }
    write_obj_conf();
}


NSAPI_PUBLIC void delete_specific_mag_init(char *fn, char *key_name, char *key_val)
{
    register int x, found = 0;
    pblock *pb;
    char *tmp;
    int numservs=make_conflist(), y=0;
 
    read_obj_conf();
 
    for(y=0; y<numservs; y++)  {
        os = osarray[y];
 
        if(!os->initfns) continue;

        for(x=0, found = 0; os->initfns[x]; x++)  {
            pb = os->initfns[x];
            if( (tmp = pblock_findval("fn", pb)) )  {
                if(!strcmp(tmp, fn))  {
                    if( (tmp = pblock_findval(key_name, pb)) ) {
                        if(!strcmp(tmp, key_val)) {
                            found = 1;
                            break;
                        }
                    }
                }
            }
        }
        if(found)  {
            int z = 0;
            for(z=x; os->initfns[z]; z++)  {
                os->initfns[z] = os->initfns[z+1];
            }
        }
    }
    write_obj_conf();
}


/***********************************************************************
    Function : IsCurrentTemplateNSPlugin
	Takes the name of a template (cgi, agents, search etc - basicaaly objects in obj.conf)
	Returns 1 if the specified template is a one of our plugins
	Returns 0 otherwise
	Since there is no way of differentiating this, we just have to
	hard-code the strings used in here
	Used by the cgis which let you do style editing sich as tmplrm,
	tmpled, tmplapp etc to disallow our plugins from being edited
***********************************************************************/
NSAPI_PUBLIC int IsCurrentTemplateNSPlugin(char* currentTemplate)
{
    if (strcmp(currentTemplate, "agents") &&
        strcmp(currentTemplate, "agent-admin") &&
        strcmp(currentTemplate, "search") &&
        strcmp(currentTemplate, "ContentManager") &&
        strcmp(currentTemplate, "ServerSideJS"))
        return 0;
    else
        return 1;
}

