/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * password.c: Interfaces with the UNIX DES password encryption libs
 * 
 * Rob McCool
 */

#include "libadmin/libadmin.h"

#ifdef ENCRYPT_PASSWORDS

#include <sys/types.h>
#include <assert.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>

static unsigned char itoa64[] =         /* 0 ... 63 => ascii - 64 */
        "./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

#ifdef SHA1_PWD_ENCRYPT

#include "password.h"

/*
 * SHA-1 implementation
 *
 * Copyright  1995 Netscape Communications Corporation, all rights reserved.
 *
 * $Id: password.c,v 1.2 2005/03/02 23:31:53 nhosoi Exp $
 *
 * Reference implementation of NIST FIPS PUB 180-1.
 *   (Secure Hash Standard, revised version).
 *   Copyright 1994 by Paul C. Kocher.  All rights reserved.
 *
 * Comments:  This implementation is written to be as close as
 *   possible to the NIST specification.  No performance or size
 *   optimization has been attempted.
 *
 * Disclaimer:  This code is provided without warranty of any kind.
 */

struct SHA1ContextStr
{
   uint32 h[5];
   uint32 w[80];
   int lenW;
   uint32 sizeHi, sizeLo;
};

typedef struct SHA1ContextStr SHA1Context;

#define SHS_ROTL(X,n) (((X) << (n)) | ((X) >> (32-(n))))

static void
shsHashBlock(SHA1Context * ctx)
{
   int t;
   uint32 a, b, c, d, e, temp;

   for (t = 16; t <= 79; t++)
      ctx->w[t] = SHS_ROTL(
			   ctx->w[t - 3] ^ ctx->w[t - 8] ^ ctx->w[t - 14] ^ ctx->w[t - 16], 1);

   a = ctx->h[0];
   b = ctx->h[1];
   c = ctx->h[2];
   d = ctx->h[3];
   e = ctx->h[4];

   for (t = 0; t <= 19; t++)
   {
      temp = SHS_ROTL(a, 5) + (((c ^ d) & b) ^ d) + e + ctx->w[t] + 0x5a827999l;
      e = d;
      d = c;
      c = SHS_ROTL(b, 30);
      b = a;
      a = temp;
   }
   for (t = 20; t <= 39; t++)
   {
      temp = SHS_ROTL(a, 5) + (b ^ c ^ d) + e + ctx->w[t] + 0x6ed9eba1l;
      e = d;
      d = c;
      c = SHS_ROTL(b, 30);
      b = a;
      a = temp;
   }
   for (t = 40; t <= 59; t++)
   {
      temp = SHS_ROTL(a, 5) + ((b & c) | (d & (b | c))) + e + ctx->w[t] + 0x8f1bbcdcl;
      e = d;
      d = c;
      c = SHS_ROTL(b, 30);
      b = a;
      a = temp;
   }
   for (t = 60; t <= 79; t++)
   {
      temp = SHS_ROTL(a, 5) + (b ^ c ^ d) + e + ctx->w[t] + 0xca62c1d6l;
      e = d;
      d = c;
      c = SHS_ROTL(b, 30);
      b = a;
      a = temp;
   }

   ctx->h[0] += a;
   ctx->h[1] += b;
   ctx->h[2] += c;
   ctx->h[3] += d;
   ctx->h[4] += e;
}

static void
SHA1_Begin(SHA1Context * ctx)
{
   ctx->lenW = 0;
   ctx->sizeHi = ctx->sizeLo = 0;

   /*
    * initialize h with the magic constants (see fips180 for constants)
    */
   ctx->h[0] = 0x67452301l;
   ctx->h[1] = 0xefcdab89l;
   ctx->h[2] = 0x98badcfel;
   ctx->h[3] = 0x10325476l;
   ctx->h[4] = 0xc3d2e1f0l;
}


static SHA1Context *
SHA1_NewContext(void)
{
   SHA1Context *cx;

   cx = (SHA1Context *) malloc(sizeof(SHA1Context));
   return cx;
}

static SHA1Context *
SHA1_CloneContext(SHA1Context * cx)
{
   SHA1Context *clone = SHA1_NewContext();
   if (clone)
      *clone = *cx;
   return clone;
}

static void
SHA1_DestroyContext(SHA1Context * cx, int freeit)
{
   if (freeit)
   {
      free(cx);
   }
}

static void
SHA1_Update(SHA1Context * ctx, const unsigned char *dataIn, unsigned len)
{
   int i;

   /*
    * Read the data into W and process blocks as they get full
    */
   for (i = 0; i < len; i++)
   {
      ctx->w[ctx->lenW / 4] <<= 8;
      ctx->w[ctx->lenW / 4] |= (uint32) dataIn[i];
      if ((++ctx->lenW) % 64 == 0)
      {
	 shsHashBlock(ctx);
	 ctx->lenW = 0;
      }
      ctx->sizeLo += 8;
      ctx->sizeHi += (ctx->sizeLo < 8);
   }
}


static void
SHA1_End(SHA1Context * ctx, unsigned char *hashout,
	 unsigned int *pDigestLen, unsigned int maxDigestLen)
{
   unsigned char pad0x80 = 0x80;
   unsigned char pad0x00 = 0x00;
   unsigned char padlen[8];
   int i;

   assert(maxDigestLen >= SHA1_LENGTH);

   /*
    * Pad with a binary 1 (e.g. 0x80), then zeroes, then length
    */
   padlen[0] = (unsigned char) ((ctx->sizeHi >> 24) & 255);
   padlen[1] = (unsigned char) ((ctx->sizeHi >> 16) & 255);
   padlen[2] = (unsigned char) ((ctx->sizeHi >> 8) & 255);
   padlen[3] = (unsigned char) ((ctx->sizeHi >> 0) & 255);
   padlen[4] = (unsigned char) ((ctx->sizeLo >> 24) & 255);
   padlen[5] = (unsigned char) ((ctx->sizeLo >> 16) & 255);
   padlen[6] = (unsigned char) ((ctx->sizeLo >> 8) & 255);
   padlen[7] = (unsigned char) ((ctx->sizeLo >> 0) & 255);

   SHA1_Update(ctx, &pad0x80, 1);
   while (ctx->lenW != 56)
      SHA1_Update(ctx, &pad0x00, 1);
   SHA1_Update(ctx, padlen, 8);

   /*
    * Output hash
    */
   for (i = 0; i < SHA1_LENGTH; i++)
   {
      hashout[i] = (unsigned char) (ctx->h[i / 4] >> 24);
      ctx->h[i / 4] <<= 8;
   }
   *pDigestLen = SHA1_LENGTH;
   SHA1_Begin(ctx);
}


static int
SHA1_HashBuf(unsigned char *dataOut, const unsigned char *dataIn, uint32 src_len)
{
   SHA1Context ctx;
   unsigned int outLen;

   if (dataOut == NULL || dataIn == NULL || src_len == 0)
   {
      return 1;
   }

   SHA1_Begin(&ctx);
   SHA1_Update(&ctx, dataIn, src_len);
   SHA1_End(&ctx, dataOut, &outLen, SHA1_LENGTH);

   return 0;
}

static int
SHA1_Hash(unsigned char *dest, const char *src)
{
   if (dest == NULL || src == NULL)
   {
      return 1;
   }

   return SHA1_HashBuf(dest, (unsigned char *) src, strlen(src));
}


/*
 * Base-64 Encoding/Decoding implementation
 *
 * Copyright \051 1995 Netscape Communications Corporation, all rights reserved.
 *
 * Reference implementation of NIST FIPS PUB 180-1.
 *   (Secure Hash Standard, revised version).
 *   Copyright 1994 by Paul C. Kocher.  All rights reserved.
 *
 * Comments:  This implementation is written to be as close as
 *   possible to the NIST specification.  No performance or size
 *   optimization has been attempted.
 *
 * Disclaimer:  This code is provided without warranty of any kind.
 */

#define RIGHT2                  0x03
#define RIGHT4                  0x0f
#define CONTINUED_LINE_MARKER   '\001'
#define MAX_LINE_WIDTH     	76

static char nib2b64[0x40f] =
"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

static unsigned char b642nib[0x80] = {
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
   0xff, 0xff, 0xff, 0x3e, 0xff, 0xff, 0xff, 0x3f,
   0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b,
   0x3c, 0x3d, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
   0xff, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06,
   0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e,
   0x0f, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16,
   0x17, 0x18, 0x19, 0xff, 0xff, 0xff, 0xff, 0xff,
   0xff, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f, 0x20,
   0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28,
   0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f, 0x30,
   0x31, 0x32, 0x33, 0xff, 0xff, 0xff, 0xff, 0xff
};

/*
 * base64_decode - take the BASE64-encoded characters in "src"
 * (a zero-terminated string) and decode them into the the buffer "dst".
 * "src" and "dst" can be the same if in-place decoding is desired.
 * "dst" must be large enough to hold the decoded octets.  No more than
 *      3 * strlen( src ) / 4 bytes will be produced.
 * "dst" may contain zero octets anywhere within it, but it is not
 *      zero-terminated by this function.
 *
 * The number of bytes copied to "dst" is returned if all goes well.
 * -1 is returned if the BASE64 encoding in "src" is invalid.
 */

static int
base64_decode(const char *src, unsigned char *dst)
{
   const char *p, *stop;
   unsigned char nib, *byte;
   int i, len;

   stop = strchr(src, '\0');
   byte = dst;
   for (p = src, len = 0; p < stop; p += 4, len += 3)
   {
      for (i = 0; i < 3; i++)
      {
	 if (p[i] != '=' && (p[i] & 0x80 ||
			     b642nib[p[i] & 0x7f] > 0x3f))
	 {
	    return (-1);
	 }
      }

      /* first digit */
      nib = b642nib[p[0] & 0x7f];
      byte[0] = nib << 2;

      /* second digit */
      nib = b642nib[p[1] & 0x7f];
      byte[0] |= nib >> 4;

      /* third digit */
      if (p[2] == '=')
      {
	 len += 1;
	 break;
      }
      byte[1] = (nib & RIGHT4) << 4;
      nib = b642nib[p[2] & 0x7f];
      byte[1] |= nib >> 2;

      /* fourth digit */
      if (p[3] == '=')
      {
	 len += 2;
	 break;
      }
      byte[2] = (nib & RIGHT2) << 6;
      nib = b642nib[p[3] & 0x7f];
      byte[2] |= nib;

      byte += 3;
   }

   return (len);
}

/*
 * base64_encode - encode "srclen" bytes in "src", place BASE64
 * encoded bytes in "dst" and return the length of the BASE64
 * encoded string.  "dst" is also zero-terminated by this function.
 *
 * If "lenused" >= 0, newlines will be included in "dst" and "lenused" if
 * appropriate.  "lenused" should be a count of characters already used
 * on the current line.  The LDIF lines we create will contain at most
 * "LDIF_MAX_LINE_WIDTH" characters on each line.
 *
 * If "lenused" < 0, no newlines will be included, and the LDIF_BASE64_LEN()
 * macro can be used to determine how many bytes will be placed in "dst."
 */

static int
base64_encode(const unsigned char *src, char *dst, int srclen, int lenused)
{
   const unsigned char *byte, *stop;
   unsigned char buf[3];
   char *out;
   unsigned long bits;
   int i, pad, len;

   len = 0;
   out = dst;
   stop = src + srclen;

   /* convert to base 64 (3 bytes => 4 base 64 digits) */
   for (byte = src; byte < stop - 2; byte += 3)
   {
      bits = (byte[0] & 0xff) << 16;
      bits |= (byte[1] & 0xff) << 8;
      bits |= (byte[2] & 0xff);

      for (i = 0; i < 4; i++, bits <<= 6)
      {
	 if (lenused >= 0 && lenused++ > MAX_LINE_WIDTH)
	 {
	    *out++ = '\n';
	    *out++ = ' ';
	    lenused = 2;
	 }

	 /* get b64 digit from high order 6 bits */
	 *out++ = nib2b64[(bits & 0xfc0000L) >> 18];
      }
   }

   /* add padding if necessary */
   if (byte < stop)
   {
      for (i = 0; byte + i < stop; i++)
      {
	 buf[i] = byte[i];
      }
      for (pad = 0; i < 3; i++, pad++)
      {
	 buf[i] = '\0';
      }
      byte = buf;
      bits = (byte[0] & 0xff) << 16;
      bits |= (byte[1] & 0xff) << 8;
      bits |= (byte[2] & 0xff);

      for (i = 0; i < 4; i++, bits <<= 6)
      {
	 if (lenused >= 0 && lenused++ > MAX_LINE_WIDTH)
	 {
	    *out++ = '\n';
	    *out++ = ' ';
	    lenused = 2;
	 }

	 if ((i == 3 && pad > 0) || (i == 2 && pad == 2))
	 {
	    /* Pad as appropriate */
	    *out++ = '=';
	 }
	 else
	 {
	    /* get b64 digit from low order 6 bits */
	    *out++ = nib2b64[(bits & 0xfc0000L) >> 18];
	 }
      }
   }

   *out = '\0';

   return (out - dst);
}

static int
sha1_pw_cmp(const char *userpwd, const char *dbpwd)
{
   unsigned char userhash[SHA1_LENGTH];
   unsigned char dbhash[SHA1_LENGTH];

   if (strlen(dbpwd) > BASE64_LEN(SHA1_LENGTH))
   {
      return (1);		/* failure */
   }

   /* SHA1 hash the user's key */
   if (SHA1_Hash(userhash, userpwd) != 0)
   {
      return 1;
   }

   /* decode hash stored in database */
   if (base64_decode(dbpwd, dbhash) != SHA1_LENGTH)
   {
      return (1);		/* failure */
   }

   /* the proof is in the comparison... */
   return (memcmp(userhash, dbhash, SHA1_LENGTH));
}


static char *
sha1_pw_enc(const char *pwd)
{
   unsigned char hash[SHA1_LENGTH];
   char *enc;
   char *retval;

   /* SHA1 hash the user's key */
   if (SHA1_Hash(hash, pwd) != 0)
   {
      return NULL;
   }

   if ((enc = (char *) malloc(BASE64_LEN(SHA1_LENGTH))) == NULL)
   {
      return (NULL);
   }

   base64_encode(hash, enc, SHA1_LENGTH, -1);
   /* add "{SHA}" in front of string */
   if ((retval = (char *) malloc(strlen(enc) + 5)) == NULL)
   {
      return (NULL);
   }
   sprintf(retval, "{SHA}%s", enc);

   return (retval);
}

#else /* SHA1_PWD_ENCRYPT */

#ifdef CRYPT_H
#include <crypt.h>
#else
char *crypt(const char *key, const char *salt);
#endif

#endif /* SHA1_PWD_ENCRYPT */

NSAPI_PUBLIC char *admin40_pw_enc(const char *pw)
{

#ifdef SHA1_PWD_ENCRYPT

    return sha1_pw_enc(pw);

#else

    char *cpw, salt[3];
    long v;

    srand((int)time(NULL));
    v = rand();

    salt[0] = itoa64[v & 0x3f];
    v >>= 6;
    salt[1] = itoa64[v & 0x3f];
    cpw = crypt(pw, salt);
    return cpw;

#endif

}

NSAPI_PUBLIC int pw_cmp(const char *pw, const char *enc)
{

#ifdef SHA1_PWD_ENCRYPT
    return sha1_pw_cmp(pw, enc);

#else
    char *cpw = crypt(pw, enc); /* use salt from enc */
    return strcmp(enc, cpw);

#endif
}

#else /* ENCRYPT_PASSWORD */

NSAPI_PUBLIC char *admin40_pw_enc(const char *pw)
{
    return pw;
}

NSAPI_PUBLIC int pw_cmp(const char *pw, const char *enc)
{
    return strcmp(pw, enc);
}
#endif /* ENCRYPT_PASSWORD */
