/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/* usrlists.c - function so manipulate sorted arrays of ID/String 
 *              pairs - used for user and group handling.
 * Ben Polk
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>     

#include "netsite.h"
#include "usrlists.h"

#define START_NAME_ARRAY_SIZE 1000

struct ArEntry_t {
    int   id;
    char *name;
};
typedef struct ArEntry_t ArEntry;

struct NameAr_t {
    int        nUsed;
    int        nMax;
    ArEntry   *ar;
};
typedef struct NameAr_t NameAr;
static const int NAMaxNameLen = 200;

static int _dumbsort(const void *s1, const void *s2)
{
    ArEntry *e1 = (ArEntry *)s1;
    ArEntry *e2 = (ArEntry *)s2;
    return strcmp(e1->name, e2->name);
} 

/*
 * ulsAlloc - create a new list array object, return 0 if success,
 *            non-zero if it fails.  
 */
int  ulsAlloc(void **handle)
{
    int     rv = 1;
    NameAr *pNA = NULL; 

    pNA = (NameAr *)MALLOC(sizeof(NameAr));
    if (pNA) {
        pNA->nUsed = 0;
        pNA->nMax  = START_NAME_ARRAY_SIZE;
        pNA->ar    = (ArEntry *) MALLOC(pNA->nMax * sizeof(ArEntry));
        if (pNA->ar) {
            *handle = (void *)pNA;
            rv = 0;
        }
    }
    return rv;
}

/*
 * ulsAddToList - create a new list array object, return 0 if success,
 *            non-zero if it fails.  
 */
#define min(x,y) ((x)<(y)?(x):(y))
int  ulsAddToList(void *handle, int id, char *name)
{
    int     rv      = 1;
    NameAr *pNA     = (NameAr *)handle; 
    int     nStrLen = min(strlen(name), NAMaxNameLen)+1;

    if (pNA && name) {
        if(pNA->nUsed == (pNA->nMax-1)) {
               pNA->nMax   += START_NAME_ARRAY_SIZE;
               pNA->ar = (ArEntry *) REALLOC((void *)pNA->ar,
                                              pNA->nMax * sizeof(ArEntry));
        }
        if (pNA->ar) { 
            pNA->ar[pNA->nUsed].name = (char *) MALLOC(nStrLen);
            if (pNA->ar[pNA->nUsed].name) {
                pNA->ar[pNA->nUsed].id = id;
                strcpy(pNA->ar[pNA->nUsed].name, name);
                ++pNA->nUsed; 
                rv = 0;
            }
        }
    }
    return rv;
}

/*
 * ulsSortName - sorts by the name field of the entries.
 */
int ulsSortName(void *handle) 
{
    int     rv  = 1;
    NameAr *pNA = (NameAr *)handle; 

    if (pNA) {
        qsort((void *)pNA->ar, pNA->nUsed, sizeof(ArEntry), _dumbsort);  
        rv = 0;
    }

    return rv;
}

/*
 * ulsGetCount - passes back the number of entries currently in use.
 */
int ulsGetCount(void *handle, int *count) 
{ 
    int rv = 1;

    if (handle) {
        *count = ((NameAr *)handle)->nUsed;
        rv = 0;
    }
    return rv;
}

/*
 * ulsGetEntry - get the id and name for a particular index
 *               into the array.
 */
int ulsGetEntry(void *handle, int entry, int *id, char **name)
{
    int     rv  = 1;
    NameAr *pNA = (NameAr *)handle; 

    if (pNA && entry < pNA->nUsed) {
        *name = pNA->ar[entry].name;
        *id   = pNA->ar[entry].id;
    }
    return rv;
}

/*
 * ulsFree - frees both the array of entries and the string
 *           referred to by each entry.
 */
void ulsFree(void **handle)
{
    NameAr *pNA = (NameAr *)*handle; 
    int     i;

    if (pNA) {
        for (i=0; i<pNA->nUsed; ++i)
        {
            if (pNA->ar[i].name)
                FREE(pNA->ar[i].name);
        }
        FREE(pNA);
        *handle = NULL;
    }
    return;
}
