/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
//
// Name:    nssrchargs.cpp
// Contact: mjb<barbarin@netscape.com>
// Notes:
//    Note that the default for the number of documents returned from a search,
//    the maximum size for the value of an attribute, and the maximum total number
//    of records returned from a search are defined.
//
//    NSsearchArgs::MaxRecords() <= MAXSZ_SET_RETURNED_FROM_SEARCH
//
//    Note also that Verity allows a maximum table size of 32K.  This means that
//    the number or rows * the number of columns (attributes) in the results set
//    from a search must be <= 32K.
// -------------------------------------------------------------------------------

#include "nssrchargs.h"
#include "nssrchstructs.h"
#include "nsirutil.h"
#include "libnscore/nsavpair.h"

#define MAXRECS_RETURNED_FROM_SEARCH 20
#define MAXSZ_ATTRIBUTES_RETURNED_FROM_SEARCH 255
#define MAXSZ_SET_RETURNED_FROM_SEARCH VdkMaxDoc_All

// public member functions

NSsearchArgs::NSsearchArgs() 
:  _argsData((void*)new NSsearchArgsData) {
   if (!_argsData)
      Errnum(outOfMem);
   else {
      // set all the default for the arguments;
      NSsearchArgsData* pArgs = (NSsearchArgsData*)_argsData;

      pArgs->highlightBegin = 0;
      pArgs->highlightEnd = 0;
      pArgs->highlightOn = FALSE;
      pArgs->maxAttributeSize = MAXSZ_ATTRIBUTES_RETURNED_FROM_SEARCH;
      pArgs->largestSet = MAXSZ_SET_RETURNED_FROM_SEARCH;
      pArgs->maxrecords = MAXRECS_RETURNED_FROM_SEARCH;
      pArgs->query = 0;
      pArgs->queryParser = QP_BOOLEAN;
      pArgs->tableFields = 0;
      pArgs->tableFieldsCount = 0;
      pArgs->searchBase = 0;
      pArgs->sortby = 0;
      }
   }


NSsearchArgs::~NSsearchArgs() {
   if (_argsData) {
      NSsearchArgsData* pArgs = (NSsearchArgsData*)_argsData;

      delete []pArgs->highlightBegin;  pArgs->highlightBegin=0;
      delete []pArgs->highlightEnd;    pArgs->highlightEnd=0;
      delete []pArgs->query;           pArgs->query=0;
      delete []pArgs->sortby;          pArgs->sortby=0;
      delete []pArgs->tableFields;     pArgs->tableFields=0;

      delete (NSsearchArgsData*)_argsData;   _argsData=0;
      }
   }


// Name: CollectionCount
// Notes:
//    Return the number of Select()ed collections

int  
NSsearchArgs::CollectionCount() {
   if (!_argsData)
      return(0);
   return(((NSsearchArgsData*)_argsData)->idxPaths.Count());
   }


// Name: CollectionName
// Notes:
//    For a given collection path, return  the name of the collection

CStr  
NSsearchArgs::CollectionName(CStr path) {
   if (!_argsData || !path)
      return(0);
   return(((NSsearchArgsData*)_argsData)->idxPaths.Get(path));
   }


// Name: CollectionPath
// Notes:
//    For a given offset in the "list" of collection names, return the path of the 
//    collection

CStr  
NSsearchArgs::CollectionPath(int offset)  {
   NSsearchArgsData* pArgs = (NSsearchArgsData*)_argsData;
   if (!pArgs)
      return(0);

   Str  collectionPath = 0;
   if (offset < pArgs->idxPaths.Count()) {
      void* cookie = 0;
      Str   val = 0;
      pArgs->idxPaths.First(collectionPath, val, cookie);
      for (int i = 1; i <= offset; i++) {
         pArgs->idxPaths.Next(collectionPath, val, cookie);
         }
      if (pArgs->idxPaths.Bad()) 
         Errnum(pArgs->idxPaths.Errnum());
      }
   return(collectionPath);
   }


// Name: DefineResultFields
// Notes:
//    Define the lists of valid attributes that are to be returned for this search.
//    Note that there are always DocScore, DocPath, and VdkVgwKey attributes defined.

NSsearchArgs&  
NSsearchArgs::DefineResultFields(int fieldCnt, CStr fieldNames) {
   NSsearchArgsData* pArgs = (NSsearchArgsData*)_argsData;
   if (!pArgs) {
      Errnum(ctrGetErr);
      return(*this);
      }

   if (!fieldNames) {
      Errnum(invalidArgument);
      return(*this);
      }

   // allocate 2 extra bytes, one for the null terminator, and one for a 
   // sentinel field name delimiter
   int len = strlen(fieldNames);
   delete []pArgs->tableFields;  pArgs->tableFields=0;
   if ((pArgs->tableFields = new char[len+2]) == 0) {
      Errnum(outOfMem);
      return(*this);
      }

   strcpy(pArgs->tableFields, fieldNames);
   if (pArgs->tableFields[len-1] != ';') {
      pArgs->tableFields[len] = ';';
      pArgs->tableFields[len+1] = '\0';
      }
   
   pArgs->tableFieldsCount = fieldCnt;   
   return(*this);
   }


// Name: FieldCount
// Notes:
//    Return the number of fields defined by the DefineResultsFields call.

int  
NSsearchArgs::FieldCount() const {
   if (!_argsData)
      return(0);
   return(((NSsearchArgsData*)_argsData)->tableFieldsCount);
   }


// Name: Fields
// Notes:
//    Return the semicolon-separated list of result field names.

CStr  
NSsearchArgs::Fields() const {
   if (!_argsData)
      return(0);
   return(((NSsearchArgsData*)_argsData)->tableFields);
   }


// Name: HLbegin
// Notes:
//    Get/set the highlight begin character string that will be inserted into the
//    document stream when retrieving a highlighted document. 

CStr
NSsearchArgs::HLbegin() const {
   if (!_argsData)
      return(0);
   return(((NSsearchArgsData*)_argsData)->highlightBegin);
   }


NSsearchArgs&
NSsearchArgs::HLbegin(CStr beginStr) {
   NSsearchArgsData* pArgs = (NSsearchArgsData*)_argsData;
   if (!pArgs) {
      Errnum(ctrGetErr);
      return(*this);
      }

   if (StrArgCopy(pArgs->highlightBegin, beginStr) ==FALSE)
      Errnum(outOfMem);
   else
      pArgs->highlightOn = TRUE;
   return(*this);
   }


// Name: HLend
// Notes:
//    Get/set the highlight end character string that will be inserted into the
//    document stream when retrieving a highlighted document. 

CStr
NSsearchArgs::HLend() const {
   if (!_argsData)
      return(0);
   return(((NSsearchArgsData*)_argsData)->highlightEnd);
   }


NSsearchArgs&
NSsearchArgs::HLend(CStr endStr) {
   NSsearchArgsData* pArgs = (NSsearchArgsData*)_argsData;
   if (!pArgs) {
      Errnum(ctrGetErr);
      return(*this);
      }

   if (StrArgCopy(pArgs->highlightEnd, endStr) == FALSE)
      Errnum(outOfMem);
   return(*this);
   }


// Name: LargestSet
// Notes:
//    Get/set the largest set possible to be retrieved (less than or equal to the
//    constant MAXSZ_SET_RETURNED_FROM_SEARCH)

int
NSsearchArgs::LargestSet() const {
   if (!_argsData)
      return(0);
   return(((NSsearchArgsData*)_argsData)->largestSet);
   }


NSsearchArgs&
NSsearchArgs::LargestSet(int largestSet) {
   if (!_argsData) {
      Errnum(ctrGetErr);
      return(*this);
      }
   ((NSsearchArgsData*)_argsData)->largestSet = largestSet;
   return(*this);
   }


// Name: MaxAttributeSize
// Notes:
//    Get/set the maximum size for the value of an attribute

int  
NSsearchArgs::MaxAttributeSize() const {
   if (!_argsData)
      return(0);
   return(((NSsearchArgsData*)_argsData)->maxAttributeSize);
   }


NSsearchArgs&
NSsearchArgs::MaxAttributeSize(int maxAttributeSize) {
   if (!_argsData) {
      Errnum(ctrGetErr);
      return(*this);
      }
   ((NSsearchArgsData*)_argsData)->maxAttributeSize = maxAttributeSize;
   return(*this);
   }


// Name: MaxRecords
// Notes:
//    Get/set the maximum number of records that will be returned from a search.

int
NSsearchArgs::MaxRecords() const {
   if (!_argsData)
      return(0);
   return(((NSsearchArgsData*)_argsData)->maxrecords);
   }


NSsearchArgs&
NSsearchArgs::MaxRecords(int maxrecords) {
   if (!_argsData) {
      Errnum(ctrGetErr);
      return(*this);
      }
   ((NSsearchArgsData*)_argsData)->maxrecords = maxrecords;
   return(*this);
   }


// Name: Query
// Notes:
//    Get/set the query string that will be submitted to the search engine.

CStr 
NSsearchArgs::Query() const {
   if (!_argsData)
      return(0);
   return(((NSsearchArgsData*)_argsData)->query);
   }


NSsearchArgs& 
NSsearchArgs::Query(CStr query) {
   NSsearchArgsData* pArgs = (NSsearchArgsData*)_argsData;
   if (!pArgs) {
      Errnum(ctrGetErr);
      return(*this);
      }

   if (StrArgCopy(pArgs->query, query) == FALSE)
      Errnum(outOfMem);
   return(*this);
   }


// Name: QueryType
// Notes:
//    Get/set the default query parser for the search engine.

int 
NSsearchArgs::QueryType() const {
   if (!_argsData)
      return(0);
   return(((NSsearchArgsData*)_argsData)->queryParser);
   }


NSsearchArgs&
NSsearchArgs::QueryType(int querytype) {
   if (!_argsData) {
      Errnum(ctrGetErr);
      return(*this);
      }
   ((NSsearchArgsData*)_argsData)->queryParser = querytype;
   return(*this);
   }


// Name: QueryTypeString
// Notes:
//    Return a string representation of the default query parser for the search engine.

CStr  
NSsearchArgs::QueryTypeString() const {
   if (!_argsData)
      return(0);
   switch (((NSsearchArgsData*)_argsData)->queryParser) {
      case QP_BOOLEAN:
         return("BOOLEAN");
      case QP_FREETEXT:
         return("FREETEXT");
      case QP_OLDFREETEXT:
         return("OLDFREETEXT");
      case QP_PASSTHRU:
         return("PASSTHRU");
      }
   return("FREETEXT");     
   }


// Name: SearchBase
// Notes:
//    Get/set the base in the result set for returning results from a query to the 
//    search engine.  This is commonly used when doing a prev/next on a list of 
//    results.

int  
NSsearchArgs::SearchBase() const {
   if (!_argsData)
      return(0);
   return(((NSsearchArgsData*)_argsData)->searchBase);
   }


NSsearchArgs&
NSsearchArgs::SearchBase(int searchBase) {
   if (!_argsData) {
      Errnum(ctrGetErr);
      return(*this);
      }

   ((NSsearchArgsData*)_argsData)->searchBase = searchBase;
   return(*this);
   }


// Name: Select
// Notes:
//    Select a collection to search against, by defining a collection name and the
//    index path to the collection.

NSsearchArgs&  
NSsearchArgs::Select(CStr collectionName, CStr idxPath) {
   NSsearchArgsData* pArgs = (NSsearchArgsData*)_argsData;
   if (!pArgs) {
      Errnum(ctrGetErr);
      return(*this);
      }

   NormalizePathSlashes((Str)idxPath);
   if (pArgs->idxPaths.Put(idxPath, collectionName).Bad())
      Errnum(pArgs->idxPaths.Errnum());
   return(*this);
   }


// Name: SortBy
// Notes:
//    Set/get the sort parameter which is used to determine the sort order of the
//    results.  Sort order is given by [+|-]fieldName[,[+|-]fieldName]*

NSsearchArgs&  
NSsearchArgs::SortBy(CStr sortby) {
   NSsearchArgsData* pArgs = (NSsearchArgsData*)_argsData;
   if (!pArgs) {
      Errnum(ctrGetErr);
      return(*this);
      }

   if (StrArgCopy(pArgs->sortby, sortby) == FALSE)
      Errnum(outOfMem);
   return(*this);
   }


CStr	
NSsearchArgs::SortBy() const {
   if (!_argsData)
      return(0);
   return(((NSsearchArgsData*)_argsData)->sortby);
  }


// protected member functions

void
NSsearchArgs::Dump(ostream& o) const {
   NSsearchArgsData* pArgs = (NSsearchArgsData*)_argsData;
   if (_argsData) 
      o << "Max records to return = " << pArgs->maxrecords << endl
        << "Query type = " << QueryTypeString() 
        << endl
        << "Highlight start string: " 
        << (pArgs->highlightBegin ? pArgs->highlightBegin : "UNDEFINED") 
        << endl
        << "Highlight end string: " 
        << (pArgs->highlightEnd ? pArgs->highlightEnd : "UNDEFINED") 
        << endl
        << "Collection paths and names:" << pArgs->idxPaths 
        << endl;
   }


CStr
NSsearchArgs::Name() const {
   return("NSsearchArgs");
   }
