/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * A push-style consumer supports the pushConsumer interface to receive messages. 
 * This interface allows consumers to monitor objects. When a message is received
 * by an object, the consumer's notify procedure is called with the message ID, an 
 * object UID, a pointer to the supplier's data and a pointer to the consumer's 
 * data. 
 *
 * Daryoush Paknad
 *
 */
#include <netsite.h> 
#include <base/systhr.h>
#include <base/crit.h>
#include <base/net.h>

#include <time.h>
#include <mdtypes.h>
#include <pub_cons.h>
#include <msgchnel.h>
#include <orb.h>
#include <debug.h>
#include <imsg.h>

NSPR_BEGIN_EXTERN_C


/* 
 * Creates a new consumer object for a specified message, object UID, client 
 * data and notification proc. 
 *
 * input: msdID- The message that this consumer is interested in.
 *        guid- Object type
 *        objID- The object id of the object that this consumer is monitoring.
 *        consumerData - data to be associated with this object
 *        PushNotify- The consumer's notification proc.
 *
 * output: A null PushConsumer of failure.
 *
 */
NSAPI_PUBLIC PushConsumer ConsumerCreatePush( short msgID, long guid, long objID, 
				   void *consumerData, NotifyProc PushNotify, MD_DeleteProc PushDelete)
{
	PushConsumer pc; 
	IBase        *obj;
	IMessage     *msg;
	int          ret;
	
	DEBUG_MSG( "ConsumerCreatePush: ", "Creating a new monitor!" );

	/*
	 * Since the timer object is a different than all the other objects in the 
	 * system, we will not create a puch consumer object for it. The message is
	 * simply put in a timer pipe.
	 */
	if ( guid == IITIMER ) 	{
		/* Create a timer object */
		int ret;

		ret = TimerServiceProviderAddTimeObject( gChannel.timer,(time_t) objID,
						PushNotify,consumerData  );
		if ( ret != TIMER_SUCCESS )
			return (PushConsumer) MD_FAILED;

		  /*  
		   * Since service requester and provider are on separate threads,
		   * this function can not return status. We assume everything is fine 
		   */
		return (PushConsumer) MD_SUCCESS;
	}

	pc = ConsumerNewPush( msgID );
	if ( !pc )	{
		DEBUG_MSG( "ConsumerCreatePush", "ConsumerNewPush failed!" );
		return (PushConsumer) NULL;
	}
	pc->objID = objID;
	pc->msgID = msgID;
	pc->consumerData = consumerData;
	pc->PushNotify = PushNotify;
	pc->PushDelete = PushDelete;
	
	/* Load the object and set its message hook to point to this monitor */
	obj = NSORB_FindObject( guid,  objID );
	if ( obj ) {
		/* Does this object support message hooks */
		ret = obj->method->QueryInterface( obj, IIMESSAGE, (void **)&msg );
		if ( ret == CM_SUCCESS ) 
			IMSG(msg)->MsgAddMonitor( msg, pc ); 
	}
	if ( MCInsertConsumerPush( pc ) == MD_FAILED ) {
		DEBUG_MSG( "ConsumerCreatePush", "MCInsertConsumerPush failed!" );
		PERM_FREE( pc );
		return (PushConsumer) NULL;
	}		
		
	return pc;
}




NSAPI_PUBLIC int ConsumerDeletePush(  PushConsumer pc, long guid )
{
  IBase      *obj;
  IMessage   *msgHookObj;
  int        ret;
  
  if ( MCDeleteConsumerPush( pc ) == MD_FAILED )
	return MD_FAILED;
  
	obj = NSORB_FindObject( guid,  pc->objID );
	if ( obj ) {
		/* Does this object support message hooks */
		ret = obj->method->QueryInterface( obj, IIMESSAGE, (void **)&msgHookObj );
		if ( ret == CM_SUCCESS ) 
			ret = msgHookObj->method->MsgDeleteMonitor( msgHookObj, pc ); 
	}
  	return ret;
}




/*
 * Deletes the specified message from the resource message list.
 * 
 * Input:	res - The phisical path to the object.
 *			msgID - The message to delete
*/
NSAPI_PUBLIC int ConsumerDeleteResPush(  char *res, short msgID )
{
	IBase       *obj;
	IMessage	*msg;
	long		ret;
	
	obj = NSORB_GetInstance( IICONTENTSTORE, res );
	if ( !obj )
        return CM_FAILED;      /* Unable to find resource */

	ret = IBASE(obj)->QueryInterface( obj, IIMESSAGE, (void **) &msg );
	if ( ret == CM_FAILED )
        return CM_FAILED;      /* Unable to find resource */

	return IMSG(msg)->MsgDeleteMessage( msg, msgID );
}




/* 
 * Creates a new consumer (monitor) object for a specified message.
 *
 * input: msdID- The message that this consumer is interested in.
 *
 * output: A push consumer object
 */
NSAPI_PUBLIC PushConsumer ConsumerNewPush( short msgID )
{
	PushConsumer pc;
	
	pc = (PushConsumer) PERM_MALLOC( sizeof (  struct _t_PushConsumer ) );
	if ( !pc ) {
		DEBUG_MSG( "ConsumerNewPush", "Malloc failed!" );
		return (PushConsumer) NULL;
	}
	
	CS_Memset( (void *) pc, 0, sizeof( struct _t_PushConsumer ) );

	pc->msgID = msgID;
	
	return pc;
}




/* 
 * Removes the push consumer from the message channel. It also rleases the
 * object.
 *
 * input: pc - A PushConsumer
 *
 * output: MD_SUCCESS on success, MD_FAILED on failure
 *
 */
NSAPI_PUBLIC int ConsumerDisconnectPush( PushConsumer pc )
{
	if ( !pc )
	    return MD_FAILED;

	return MCDeleteConsumerPush( pc );
}




/* 
 * Attaches a push consumer to the message channel.
 *
 * input: pc - A PushConsumer
 *
 * output: MD_SUCCESS on success, MD_FAILED on failure
 *
 */
NSAPI_PUBLIC int ConsumerConnectPush( PushConsumer pc )
{
	if ( !pc )
	    return MD_FAILED;

	return MCInsertConsumerPush( pc );
}







/* 
 * Sets the object id of the specified consumer.
 *
 * input: pc - A PushConsumer
 *        objID- The object id of the object that this consumer is monitoring.
 *
 * output: MD_SUCCESS on success, MD_FAILED on failure
 *
 */
NSAPI_PUBLIC int ConsumerSetPushObj( PushConsumer pc, long objID )
{
	if ( !pc )
	    return MD_FAILED;
	
	pc->objID = objID;
	return MD_SUCCESS;	
}

 
/*  
 * Returns the UID of the object that the consumer is monitoring 
 *
 * input: pc - A PushConsumer
 *
 * output: object's uid, if successful. MD_FAILED if pc is null
 *
 */
NSAPI_PUBLIC long ConsumerGetPushObj( PushConsumer pc )
{
	if ( !pc )
	    return MD_FAILED;
	return pc->objID;
}
	


/*
 * Attaches the consumer's data to the consumer object.
 * Message channel calls the conumer's notify proc  with this pointer. 
 *
 * input: pc - A PushConsumer
 *        consumerData - data to be associated with this object
 *
 * output: MD_SUCCESS if successful, MD_FAILED otherwise
 *
 */
NSAPI_PUBLIC int ConsumerSetPushData( PushConsumer pc, void *consumerData )
{
	if ( !pc )
	    return MD_FAILED;

	pc->consumerData = consumerData;

	return MD_SUCCESS;
}



/*  
 * Returns the consumer's data that is attached to the specified object.
 *
 * input: pc - A PushConsumer
 *
 * output: object's data, if successful. NULL otherwise
 *
 */
NSAPI_PUBLIC void *ConsumerGetPushData( PushConsumer pc )
{
	if ( !pc )
	    return (void *) NULL;

	return pc->consumerData;
}	




/* 
 * Sets the notification procedure of the specified consumer.
 *
 * input: pc - A PushConsumer
 *        PushNotify- The consumer's notification proc.
 *
 * output: MD_SUCCESS if successful, MD_FAILED otherwise
 *
 */
NSAPI_PUBLIC int ConsumerSetPushNotifyProc( PushConsumer pc, NotifyProc PushNotify )
{
	if ( !pc )
	    return MD_FAILED;

	pc->PushNotify = PushNotify;

	return MD_SUCCESS;
}	



/* 
 * Returns the message ID of the consumer
 *
 * input: pc - A PushConsumer
 *
 * output: consumer's message id. MD_FAILED on failure
 *
 */
NSAPI_PUBLIC short ConsumerGetPushMsg( PushConsumer pc )
{
	if ( !pc )
	    return MD_FAILED;

	return pc->msgID;
}	
	
NSPR_END_EXTERN_C
