/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/* 
 * A non-recursive heterogeneous binary object tree
 *
 * Daryoush Paknad
*/
#include "prtypes.h"
#include "mdbtree.h"
#include "stdio.h"
#include "string.h"
#include "debug.h"

NSPR_BEGIN_EXTERN_C

static BNode *BTreePrivFindNode(BTree *tree, uint32 key, BNode **parent, char *dir, BNode **rel, void *cData);
static void BTreeInorder( BNode *aNode, TravelProc Traversal, void *cData );



/*
 * Creates a tree
 *
 * Output: A binary tree object
*/
NSAPI_PUBLIC BTree * CMNewBTree(KeyCollisionProc  CollisionProc)
{
	BTree  *bTree;

    bTree = (BTree *) PERM_CALLOC( sizeof (BTree) );
    if (!bTree) {
		DEBUG_MSG( "NewBTree", "Calloc failed..." );
		return (BTree *) NULL;
	}
	bTree->CollisionProc = CollisionProc;
    return bTree;
}


/*
 * Inserts a tree node into a binary sorted tree pointed by 
 * tree.
 * Input: tree -     A tree to add the node to
 *        key -      A key for this node
 *        nodeData - Dat associated to this node
 *        status -   return status
 *                   BTREE_SUCCESS        on success
 *                   BTREE_DUPLICATE_KEY  if there is an entry with the same key 
 *                                        in the system
 *                   BTREE_FAILED         on failure
 *
 * Output:
 *     A pointer to the parent node of the node inserted or NULL
 *     if this is the first (root) in the tree. Note that in this case
 *     status is set to BTREE_SUCCESS
 *     
*/
NSAPI_PUBLIC BNode *CMBTreeAddNode( BTree *tree, uint32 key, void *nodeData, int *status )
{
    BNode   *newNode, *cur;
	BNode   *parent;
    long     pathLen=0;
	long   ret;
    char    dir;

	*status =  BTREE_SUCCESS;

	DEBUG_MSG1L( "BTreeAddNode", "Adding a new node key =  ", key  );

	if ( !tree ) {
		DEBUG_MSG( "BTreeAddNode", "A null tree..." );
		*status = BTREE_FAILED;
		return (BNode *) NULL;
	}
	/* Allocate a new node */
    newNode = (BNode *) PERM_CALLOC( sizeof ( BNode ) );
	if ( !newNode ) {
		DEBUG_MSG( "BTreeAddNode", "Calloc failed..." );
		*status = BTREE_FAILED;
		return (BNode *) NULL;
	}
	newNode->key = key;
	newNode->data = nodeData;
	
    if ( !tree->root ) {    /* The first object in the tree */
		tree->root = newNode;
		++tree->nodeCnt;
		return  (BNode *) NULL;
	}
		
    parent = NULL;
    cur = tree->root;

    while ( cur ) {
       parent = cur;
       ret = key - parent->key;

       if (ret < 0) {
           dir = 'L';
           cur = cur->left;
       }
       else {
           if (ret > 0) {
               dir = 'R';
               cur = cur->right;
           }
           else  {
#if 0
				/* An entry with this key already exist */
				*status = BTREE_DUPLICATE_KEY;
				/* Get the client to handle the collision */
				if ( tree->CollisionProc )
					ret = (*tree->CollisionProc )( cur, newNode );
					if ( ret ) 
						PERM_FREE( newNode );

#ifdef CS_DEBUG
			   fprintf( stderr,"PANIC PANIC PANIC PANIC PANIC......\n" );
			   fprintf( stderr,"PANIC: Duplicate key in the system.\n" );
			   fprintf( stderr,"PANIC PANIC PANIC PANIC PANIC......\n" );
#endif			   
               return cur;
#endif
			   BNode   *curPtr;
			   
			   curPtr = cur;
			   
			   while (curPtr->next)
				 curPtr = curPtr->next;
			   
			   tree->numColl++;
			   curPtr->next = newNode;
			   return parent;
				 
           }
       }
       pathLen++;
    }

    /* If this is an empty tree. make the node to be the root */
    if (parent == NULL) {
	  DEBUG_MSG1L( "BTreeAddNode", "new node =  ", (long) newNode  );	  
	  tree->root = newNode;
	}
    else  {
       if (dir == 'L')
           parent->left = newNode;
       else
           parent->right = newNode;
    }
    if ( tree->treeHeight < (short) pathLen )
       tree->treeHeight = pathLen;

	++tree->nodeCnt;    

    return parent;
}       



/*
 * Returns a pointer to a node which has the same 
 * key value as the passed in parameter "key".  It also sets "parent"
 * to point to the the node's parent object
 *
 * Input: tree -     A tree to add the node to
 *        key -      A key for the node
 *        parent -   sets this pointer to the node's parent object
 *
 * Output: NULL on failure
*/
#ifdef XP_WIN32
#pragma optimize( "", off )
#endif
NSAPI_PUBLIC BNode *CMBTreeFindNode( BTree *tree, uint32 key, BNode **parent, BNode **rel, void *cData )
{
    BNode  *cur, *curPtr;
    long    ret;
	char    dir;


	DEBUG_MSG( "BTreeFind:", "............................" );

	if ( !tree ) {
		DEBUG_MSG( "BTreeFind", "A null tree..." );
		return (BNode *) NULL;
	}
    *parent = NULL;
    *rel = NULL;
    cur = tree->root;

    while ( cur ) {
	  
		ret = key - cur->key;

		if ( ret == 0 ) {

		  if ( !(tree->CollisionProc) )
			return cur;

		  curPtr = cur;
		  while ( curPtr ) {
			/* The caller selects node by returning 1 */
			if ( (*tree->CollisionProc)( curPtr, key, cData ) ) {
			  *rel = cur;
			  return curPtr;
			}
			
			curPtr = curPtr->next;
		  }
		  return (BNode *) NULL;			
		}
		
		*parent = cur;
		if ( ret < 0 ) {
            dir = 'L';
			cur = cur->left;
		}
		else {
			dir = 'R';
			cur = cur->right;
		}
    }

	DEBUG_MSG( "BTreeFind: ", "Object not found!" );

    return (BNode *) NULL;
}

#ifdef XP_WIN32
#pragma optimize( "", on )
#endif




/*
 * This function deletes the node with the key value equal to the
 * passed in key parameter.
 *
 * Input: tree -     A tree to delete the node from
 *        key -      The node's key
 *
 * Output:  NULL if node is not found; a pointer to the node's client data on
 *          success. Note that the client is responsible for freeing its node
 *          data.
*/
NSAPI_PUBLIC void * CMBTreeDeleteNode( BTree *tree, uint32 key, void *clientData )
{
    BNode  *parent, *cur, *q, *prevPtr;
    BNode  *nPtr, *suc, *pred, *rel;
    char   dir;
	void   *cData;
	
	if ( !tree ) {
		DEBUG_MSG( "BTreeFind", "A null tree..." );
		return (void *) NULL;
	} 

	cur = BTreePrivFindNode( tree, key, &parent, &dir, &rel, clientData );

	/* if rel is set to a valid ptr, cur is a collision node */
	prevPtr = nPtr = rel;
	if ( nPtr && (nPtr != cur) ) {
	  while ( nPtr && (nPtr != cur) ) {
		prevPtr = nPtr;
		nPtr = nPtr->next;
	  }
	
	  if ( nPtr ) {
		prevPtr->next = nPtr->next;
		cData = nPtr->data;	
		PERM_FREE( cur );
		return cData;
	  }
	}
	
	
    if ( !cur ) {
		DEBUG_MSG( "BTreeDeleteNode", "Node not found..." );
		return (void *) NULL;
    }
    cData = cur->data;
	
    if ( cur->left == NULL )
       /* empty left subtree */
       q = cur->right;
    else  {
       if ( cur->right == NULL )
           /* empty right subtree */
           q = cur->left;
       else  {
           /* check right child for successor */
           suc = cur->right;
           if ( suc->left == NULL ) {
               suc->left = cur->left;
               q = suc;
           }
           else  {
               pred = cur->right;
               suc = pred->left;
               while ( suc->left != NULL ) {
                   pred = suc;
                   suc = pred->left;
               }
               pred->left = suc->right;
               suc->left  = cur->left;
               suc->right = cur->right;
               q = suc;
           }
       }
    }
    if ( !parent ) {
       /* The node we are deleting happens to be the root node */
       tree->root = q;
	   DEBUG_MSG1L( "BTreeDeleteNode", "tree->root =  ", (long )tree->root  );	  
    
	}
    else {
       if ( dir == 'L' )
           parent->left = q;
       else
           parent->right = q;
    }

	if ( q && cur )
	  q->next = cur->next;
	
    PERM_FREE( cur );
	tree->nodeCnt--;

    return cData;
}




static BNode *BTreePrivFindNode( BTree *tree, uint32 key, BNode **parent, char *dir,BNode **rel, void *cData )
{
    BNode  *cur, *curPtr;
    long ret;

	if ( !tree ) {
		DEBUG_MSG( "BTreeFind", "A null tree..." );
		return (BNode *) NULL;
	}
    *parent = NULL;
    *rel = NULL;
    cur = tree->root;

    while ( cur ) {
       ret = key - cur->key;

		if ( ret == 0 ) {

		  if ( !(tree->CollisionProc) )
			return cur;

		  curPtr = cur;
		  while ( curPtr ) {
			/* The caller selects node by returning 1 */
			if ( (*tree->CollisionProc)( curPtr, key, cData ) ) {
			  *rel = cur;
			  return curPtr;
			}
			
			curPtr = curPtr->next;
		  }
		  return (BNode *) NULL;			
		}

       *parent = cur;
       if ( ret < 0 ) {
           *dir = 'L';
           cur = cur->left;
       }
       else {
           *dir = 'R';
           cur = cur->right;
       }
    }

    return (BNode *) NULL;
}




/*
 * Deletes a tree
 *
 * Input: tree - tree to be deleted
 *        CFree - client's free proc. This procedure is called with 
 *                node's data.
 *
 * Output: A BTREE_SUCCES on success and BTREE_FAILURE otherwise
 *
 * Note-  This function can be optimized
*/
NSAPI_PUBLIC int CMBTreeDestroy( BTree *tree, ClientFreeProc CFree )
{
    BNode  *parent, *cur, *pred;

	if ( !tree ) {
		DEBUG_MSG( "BTreeAddNode", "A null tree..." );
		return BTREE_SUCCESS;
	}

	parent = cur  = tree->root;
	
	while ( cur ) {

		if ( cur->left ) {
			if ( cur != parent )
			    pred = parent;  

			parent = cur;
			cur = cur->left;
			continue;
		}

		if ( cur->right ) {
			if ( cur != parent )
			    pred = parent;  

			parent = cur;
			cur = cur->right;
			continue;
		}
		if ( !cur->left && !cur->right ) {
			if ( cur->data )
			    (*CFree)( cur->data );

			if ( parent->left == cur ) 
			    parent->left = (BNode *) NULL;
			else
			    parent->right = (BNode *) NULL;

			if ( cur == tree->root ) {  /* all done! */
				PERM_FREE( cur );
				PERM_FREE( tree );
				return BTREE_SUCCESS;
			}
			PERM_FREE( cur );
			cur = parent;
			if ( !parent->left && !parent->right )
			    parent = pred;
		}

		if ( (cur == parent) && (cur == pred) )  /* lets start from the root again */
		    cur = parent = pred = tree->root;
	}			
	return BTREE_FAILED;
}



NSAPI_PUBLIC int CMBTreeGetNumNode( BTree *tree )
{
	return tree->nodeCnt;
}



/* travels a tree inorder */
NSAPI_PUBLIC void CMBTreeInorderTravel( BTree *tree, TravelProc Traversal, void *cData )
{
	if ( !tree || !tree->root )
	    return;
	
	BTreeInorder( tree->root, Traversal, cData );
}



/* Calls the Traversal proc with a tree node and the client data "cData" */
static void BTreeInorder( BNode *aNode, TravelProc Traversal, void *cData )
{
	BNode *node;
	
	node = aNode;
	
	if ( (node->left == (BNode *) NULL) && (node->right == (BNode *) NULL) )	{
		/* A leaf node */
		if ( !Traversal( node->data, cData ) )
			return;
	}
	else  {
		if ( node->left ) 
			BTreeInorder( node->left, Traversal, cData );

		if ( !Traversal( node->data, cData ) )
			return;

		if ( node->right ) 	
			BTreeInorder( node->right, Traversal, cData );
	}
}			
			
NSPR_END_EXTERN_C
