/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/* 
 * A message channel implementation.  It decouples the communication between 
 * component's objects. It provides asynchronous communication of message data
 * between suppliers and consumers.
 *
 * Daryoush Paknad
*/

#include <netsite.h> 
#include <base/systhr.h>
#include <base/crit.h>
#include <base/net.h>

#include "mdtypes.h"
#include "debug.h"
#include "msgchnel.h"
#include "nsarray.h"
#include "message.h"


#define MIN_REQ_ELAPS_TIME   300


MsgChannel gChannel;


NSPR_BEGIN_EXTERN_C

static BTree *MCPrivFindMsgTree(NSObjArray * msgList, short msgID );
static BTree *MCNewMessage( short msgID );
static int MCDumpMonitor( void * nodeData, void *cData );
static int CollisionProc( BNode *node, uint32 key, void *cData );

/*
 * Creates a new message channel.  It initializes the channel with Netscape's
 * pre-defined messages.  Message ids 0 through 10000 are Netscape static
 * ids. Client's can create their own messages with ids between 10001 to 65536.
 *
 * Output: MD_SUCCESS on success, MD_FAILED otherwise
 *
 */
NSAPI_PUBLIC int MsgChannelInit()
{
	int i;
	MsgNode *msgNode;
	
	DEBUG_MSG( "MsgChannelInit", "Initializing the message channel..." );
	/* Create a new message list */
	gChannel.msgList = NSObjArrayNew( 50, sizeof( MsgNode ) );
	if ( !gChannel.msgList ) {
		DEBUG_MSG( "MsgChannelInit", "Failed to create an array..." );
		return MD_FAILED;
	}

	/* Put the default messages in the list */
	for ( i=0; i< NUM_NS_MESSAGES; i++ )  {
		msgNode = (MsgNode *) NSObjArrayNewObj( gChannel.msgList );
		if ( !msgNode ) {
			DEBUG_MSG( "MsgChannelInit", "Failed to create a message node..." );
			return MD_FAILED;
		}
		msgNode->msgID = i;
		msgNode->consumerTree = CMNewBTree(CollisionProc);
		if ( !msgNode->consumerTree ) {
			DEBUG_MSG( "MsgChannelInit", "Failed to create a consumer tree..." );
			return MD_FAILED;
		}
	}
	gChannel.timer = NULL;
	/*************
	gChannel.timer = TimerServiceProviderInit(5);
	if ( !gChannel.timer ) {
		DEBUG_MSG( "MsgChannelInit", "Failed to create a timer..." );
		return MD_FAILED;
	}
	************/

	return MD_SUCCESS;
}



static BTree *MCNewMessage( short msgID )
{
	BTree *tree;
	MsgNode *msgNode;
	
	tree = MCFindMsgTree( msgID );
	if ( tree )  /* If the message is already in the channel, return its tree */
	    return tree;

	msgNode = (MsgNode *) NSObjArrayNewObj( gChannel.msgList );
	if ( !msgNode ) {
		DEBUG_MSG( "MCNewMessage", "Failed to create a message node..." );
		return (BTree *) NULL;
	}
	DEBUG_MSG( "MCNewMessage", "Creating a new message node..." );
	msgNode->msgID = msgID;
	msgNode->consumerTree = CMNewBTree(CollisionProc);
	if ( !msgNode->consumerTree ) {
		DEBUG_MSG( "MCNewMessage", "Failed to create a consumer tree..." );
		return (BTree *) NULL;
	}
	return msgNode->consumerTree;
}


/* 
 * Find the consumer tree for the specified message id.
 * Note that for performance reasons, this function assumes that all of the
 * Netscape specific messages are put in the list sequentialy.
 * Messsages starting from index 0 upto NUM_NS_MESSAGES are Netscape specific.
 * To find message ids bigger than  NUM_NS_MESSAGES, it walks up the list.
 *
 * Input: 
 *        msgID-    Message id
 *
 * Output: Null on failure
 *         The consumer's tree on success.
 *
*/
NSAPI_PUBLIC BTree *MCFindMsgTree( short msgID )
{
	return MCPrivFindMsgTree( gChannel.msgList, msgID );
}



/* 
 * Find the consumer tree for the specified message id.
 * Note that for performance reasons, this function assumes that all of the
 * Netscape specific messages are put in the list sequentialy.
 * Messsages starting from index 0 upto NUM_NS_MESSAGES are Netscape specific.
 * To find message ids bigger than  NUM_NS_MESSAGES, it walks up the list.
 *
 * Input: msgList - Channel's message table
 *        msgID-    Message id
 *
 * Output: Null on failure
 *         The consumer's tree on success.
 *
*/
static BTree *MCPrivFindMsgTree(NSObjArray * msgList, short msgID )
{
	int i, numObj;
	MsgNode *msgNode;
	
	if ( msgID < NUM_NS_MESSAGES ) {
		msgNode = (MsgNode *)NSObjArrayGetObjNum( msgList, msgID );
		if ( !msgNode ) {
			DEBUG_MSG( "MCPrivFindMsgTree", "Failed to get a message node..." );
			return (BTree *) NULL;
		}
		
		/* double check */
		if ( msgNode->msgID != msgID ) {
			DEBUG_MSG( "Panic", "MCPrivFindMsgTree: Bad message id..." );
			return (BTree *) NULL;
		}
		return msgNode->consumerTree;
	}

	numObj = NSObjArrayGetNumObj( msgList );
	for ( i=NUM_NS_MESSAGES; i<numObj; i++ ) {

		msgNode = (MsgNode *)NSObjArrayGetObjNum( msgList, i );
		if ( !msgNode ) {
			DEBUG_MSG( "MCPrivFindMsgTree", "Failed to get a message node..." );
			return (BTree *) NULL;
		}

		if ( msgNode->msgID == msgID ) 
			return msgNode->consumerTree;
	}		
	return (BTree *) NULL;
}


		
		
/*
 * Inserts the consumer object in the message channel
 *
 * Input: pc - A push consumer object
 *
 * Output: MD_SUCCESS on success, MD_FAILED otherwise
 *
*/
NSAPI_PUBLIC int MCInsertConsumerPush( PushConsumer pc )
{
	BTree *cnsTree;
	BNode   *pNode;
	int     status;
	long    key;
	
	/* Find message tree for the consumer's message */
	cnsTree = MCFindMsgTree( pc->msgID );
	if ( !cnsTree ) {     /* Don't have this message in the channel */
		/* Create a new message */
		cnsTree = MCNewMessage( pc->msgID );
		if ( !cnsTree ) {  
			DEBUG_MSG( "MCInsertConsumerPush","Unable to find the message tree.");
			return MD_FAILED;
		}
	}
	
	/* 
	 * XXXXX Need to be change this.The key to the btree used to be the objID.
	 * In order to eliminate the key conflict for objects that have multiple
	 * monitors (same msgId), I changed the key to be objID + NOtifyProc + cData.
	 * Note that this change breaks MCFindConsumerPush.
	 * pNode = CMBTreeAddNode( cnsTree, pc->objID, (void *) pc, &status );
	 */
	key = pc->objID + (long) pc->PushNotify + (long) pc->consumerData;
	
	pNode = CMBTreeAddNode( cnsTree, key, (void *) pc, &status );
	if ( status != BTREE_SUCCESS ) {
		DEBUG_MSG( "MCInsertConsumerPush", "Failed to add a node to the btreee.");
		return MD_FAILED;
	}

	DEBUG_MSG1( "MCInsertConsumerPush", "Added a new monitor to ", pc->objName );
	return MD_SUCCESS;
}


		
/*
 * Deletes the consumer object in the message channel
 *
 * Input: pc - A push consumer object
 *
 * Output: MD_SUCCESS on success, MD_FAILED otherwise
 *
*/
NSAPI_PUBLIC int MCDeleteConsumerPush( PushConsumer pc )
{
	BTree *cnsTree;
	void    *retC;
	
	/* Find message tree for the consumer's message */
	cnsTree = MCFindMsgTree( pc->msgID );
	if ( !cnsTree ) {
		DEBUG_MSG( "MCDeleteConsumerPush", "Unable to find the message tree..." );
		return MD_FAILED;
	}
	
	retC = CMBTreeDeleteNode( cnsTree, pc->objID, NULL );
	if ( retC ) {
		/* lets make sure we really have the right object in out hands */
		if ( retC != (void *) pc ) 	{  /* Oops! */
			DEBUG_MSG( "Panic", "MCDeleteConsumerPush: Deleted a wrong object!" );
			return false;
		}
		DEBUG_MSG1( "MCDeleteConsumerPush", "Delete a monitor from ", pc->objName );
		PERM_FREE( pc );
		return MD_SUCCESS;
	}
	
	DEBUG_MSG( "Panic", "MCDeleteConsumerPush: Couldn't delete the node!" );
	return MD_FAILED;
}


	
/*
 * Finds the consumer object in the message channel.  Consumer keys are: message id
 * and the object id.
 *
 * Input: msgId- consumer's message id
 *        objID- consumer's object
 *
 * Output: A push consumer object on success, null on failure.
 *
*/
NSAPI_PUBLIC PushConsumer MCFindConsumerPush( short msgID, long objID )
{
	BTree *cnsTree;
	BNode   *node, *parent;
	BNode   *buket;
	
	/* Find message tree for the consumer's message */
	cnsTree = MCFindMsgTree( msgID );
	if ( !cnsTree ) {
		DEBUG_MSG( "MCFindConsumerPush", "Unable to find the message tree..." );
		return (PushConsumer) NULL;
	}
	
    node = CMBTreeFindNode( cnsTree, objID, &parent, &buket, NULL );
	if ( !node ) {
		DEBUG_MSG( "MCFindConsumerPush", "Failed to find the node..." );
		return (PushConsumer) NULL;
	}

	return (PushConsumer) node->data;
}



NSAPI_PUBLIC void MCDumpMessageChannel( char *output )
{
	BTree   *tree;

	int     len, i;
	MsgNode *node;
	FILE    *fPtr;
	
	fPtr = fopen( output, "w" );
	if ( !fPtr )
	    return;
	
	len =  NSObjArrayGetNumObj( gChannel.msgList );	

	for ( i=0; i<len; i++ ) {
		node = (MsgNode *) NSObjArrayGetObj( gChannel.msgList, i );
		fprintf( fPtr, "msgID = %d\n", node->msgID );
		tree = node->consumerTree;

		CMBTreeInorderTravel( tree, MCDumpMonitor, (void *) fPtr );		
	}
	fclose( fPtr );
}



static int MCDumpMonitor( void * nodeData, void *cData )
{
	FILE          *fPtr = (FILE *) cData;
	PushConsumer  pc = (PushConsumer) nodeData;

	fprintf( fPtr,"\t%s ID = %ld\n", pc->objName, pc->objID );

	return 1;
}


static int CollisionProc( BNode *node, uint32 key, void *cData )
{
	/* XXX Need to fix this */
	return 1;
}

NSPR_END_EXTERN_C
