/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * This package is responsible for all of the mechanisms required to find an
 * object implementation for the request, to prepare the object implementation to
 * receive the request, and communicate the data making up the request.
 * 
 * Daryoush Paknad
*/
#include "base/systems.h"
#include <debug.h>
#include <mdtypes.h>
#include <orb.h>
#include <mdutil.h>
#include <mdbtree.h>

NSPR_BEGIN_EXTERN_C

static int NSORB_ValidGUID( long guid );
static NSInterface *NSORB_FindInterface( NSObjArray *IList, long guid );


NSObjArray *NSOrb;   /* An array of registered interfaces */

NSAPI_PUBLIC int NSORB_Init()
{
	DEBUG_MSG( "NSORB_Init", "Initializing the NS_ORB..." );
	NSOrb = NSObjArrayNew( 10, sizeof( NSInterface ) );	
	if ( !NSOrb ) {
		DEBUG_MSG( "NSORB_Init", "Failed to create an array..." );
		return CM_FAILED;
	}
	return CM_SUCCESS;
}






/* 
 * Looks for the instance; if not found, it will create an instance of the
 * specified interface.
 * returns the object id or CM_FAILED on failure.
*/
NSAPI_PUBLIC long NSORB_InstanceID( long guid, char *objName )
{
	IBase        *intr;
	IObjectStore *objs;
	long         ret;
	
	DEBUG_MSG1( "NSORB_InstanceID: ","Looking for " , objName );

	intr = NSORB_GetInterface( guid );
	if ( !intr ) {
		DEBUG_MSG( "NSORB_InstanceID: ","PANIC NSORB_GetInterface failed!" );
		return CM_FAILED;
	}

	ret = IBASE(intr)->QueryInterface( intr, IIOBJECT_STORE, (void **) &objs );
	if ( ret == CM_FAILED )
	    return CM_FAILED;
				
	return IOBJS(objs)->GetObjInstanceID( objs, guid, objName );
}




/* 
 * Looks for the instance; if not found, it will create an instance of the
 * specified interface.
 * returns the object id or CM_FAILED on failure.
*/
NSAPI_PUBLIC IBase *NSORB_GetInstance( long guid, char *objName, void *transaction )
{
	IBase        *intr;
	IObjectStore *objs;
	long         ret;
	
	DEBUG_MSG1( "NSORB_GetInstance: ","Looking for " , objName );

	intr = NSORB_GetInterface( guid );
	if ( !intr ) {
		DEBUG_MSG( "NSORB_GetInstance: ","PANIC NSORB_GetInterface failed!" );
		return (IBase *) NULL;
	}

	ret = IBASE(intr)->QueryInterface( intr, IIOBJECT_STORE, (void **) &objs );
	if ( ret == CM_FAILED )
		return (IBase *) NULL;
				
	return IOBJS(objs)->GetObjInstance( objs, guid, objName, transaction );
}




/*
 * Registers an interface with the broker.
 * Currently it only works with in-process objects.  It will not load object
 * servers that are in shared libs (not loaded).  It will not run executable
 * object servers.
 *
 * Input: guid- A global unique id
 *        serverType- DLL_SERVER Object servers are in a dynamicly loadable libs
 *                    EXEC_SERVER Object servers are in an executable
 *
 *        objName - Object's name (not really used anywhere)
 *        objServer - Points to the location of the object server
 *        ClassFactory - Should return a IBase pointer to the interface's class factory
 *                       (base object for the interface).
 *        cfData - Class factory data
 *
 * Output: Returns 1 for success and 0 otherwise
 *
*/
NSAPI_PUBLIC int
NSORB_RegisterInterface( long guid, short serverType, char *objName, 
char *objServer, ClassFactory InterfaceProc, void *cfData )
{
	NSInterface  *ifc;
	
	if ( !guid || !InterfaceProc || !NSOrb ) 
		return CM_FAILED;

	ifc = NSORB_FindInterface( NSOrb, guid );

	/* Create a new interface, if not found in the list */
	if ( !ifc )   
	    ifc = (NSInterface *) NSObjArrayNewObj( NSOrb );

	if ( !ifc ) {
		DEBUG_MSG( "NSORB_RegisterInterface", "Failed to create a message node..." );
		return CM_FAILED;
	}

	ifc->guid = guid;
	ifc->InterfaceProc = InterfaceProc;
	ifc->serverType = serverType;
	ifc->cfData = cfData;

	if ( objName )
	    ifc->iName = CS_strdup( objName );

	if ( objServer )
	    ifc->objServer = CS_strdup( objServer );

	return CM_SUCCESS;
}



/*
 * Takes a global unique ID and returns and interface pointer.
 *
 * Input: guid  A global unique interface identifier.
 *
 * Output: Returns an Interface pointer on success.
 *         Returns (void *) NULL if the guid is invalid or bad server type 
 *
*/
NSAPI_PUBLIC IBase *NSORB_GetInterface( long guid ) 
{
	NSInterface  *ifc;

	if ( !NSOrb )
	    return (IBase *) NULL;

	if ( !NSORB_ValidGUID( guid ) )
	    return (IBase *) NULL;
	
	DEBUG_MSG( "NSORB_GetInterface", "Getting interface pointer.." );
	ifc = NSORB_FindInterface( NSOrb, guid );
	
	if ( ifc && ifc->InterfaceProc )
	    return ifc->InterfaceProc( ifc->cfData );

	return (IBase *) NULL;
}		



static int NSORB_ValidGUID( long guid )
{
    return 1;
}




/* Looks for the specified interface in its list */
static NSInterface *NSORB_FindInterface( NSObjArray *IList, long guid )
{
	int i, numI;
	NSInterface *I;
	
	numI = NSObjArrayGetNumObj( IList );

	for ( i=0; i<numI; i++ ) {

		I = (NSInterface *) NSObjArrayGetObjNum( IList, i );
		
		if ( !I ) {
			DEBUG_MSG( "NSORB_FindInterface","Failed to get an interface pointer..");
			return (NSInterface *) NULL;
		}

		if ( I->guid == guid ) {
			DEBUG_MSG1L("NSORB_FindInterface", "Found interface ", guid ); 
			return I;
		}
	}	
	DEBUG_MSG( "NSORB_FindInterface","Failed to find the interface.");	
	return (NSInterface *) NULL;
}





/* Looks for the specified instance of an object */
NSAPI_PUBLIC IBase *NSORB_FindObject( long guid, long objID )
{

	IBase        *intr;
	IObjectStore *objs;
	long         ret;
	
	DEBUG_MSG1L( "NSORB_FindObject: ","Looking for " , objID );

	intr = NSORB_GetInterface( guid );
	if ( !intr ) {
		DEBUG_MSG( "NSORB_FindObject: ","PANIC NSORB_FindObject failed!" );
		return (IBase *) NULL;
	}

	ret = IBASE(intr)->QueryInterface( intr, IIOBJECT_STORE, (void **) &objs );
	if ( ret == CM_FAILED )
	    return (IBase *) NULL;
				
	return IOBJS(objs)->StoreGetObject( objs, guid, objID );
}


NSPR_END_EXTERN_C
