/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

/***********************************************************************
** Includes
***********************************************************************/
#ifndef _WINDOWS
#include <unistd.h>
#endif

#include "libnscore/nscfg.h"
#include "libnscore/nsfname.h"


//********************************************************************
// MEMBER:	NScfg
// DESCRIPTION:
//	Constructors.
//********************************************************************
NScfg::NScfg() 
:  _flushRequired(FALSE), _filemode(READONLY), _isMod(FALSE), _isOpen(FALSE),
   _lastNode(0),_pSection(0), _root(new NScfgNode), _secname(0) {
   if (_root == 0)
      Errnum(outOfMem);
   }


NScfg::NScfg(CStr fn, Filemode fm)
:  _flushRequired(FALSE), _isMod(FALSE), _isOpen(FALSE), _lastNode(0),
   _pSection(0), _root(new NScfgNode), _secname(0) {

   if (_root == 0)
      Errnum(outOfMem);
   Open(fn, fm);       
   }


NScfg::NScfg(CStr fn, CStr* paths, Filemode fm)
:  _flushRequired(FALSE), _isMod(FALSE), _isOpen(FALSE), _lastNode(0),
   _pSection(0), _root(new NScfgNode), _secname(0) {

   if (_root == 0)
      Errnum(outOfMem);
   Open(fn, paths, fm);       
   }


//********************************************************************
// MEMBER:	~NScfg
// DESCRIPTION:
//	Destructor
//********************************************************************
NScfg::~NScfg() {
   if (_flushRequired) _isMod = FALSE;
   Close();
   delete _root;
   delete []_secname;
   }


//********************************************************************
// MEMBER:	Close
// DESCRIPTION:
//	@+@member description
//********************************************************************
NScfg& NScfg::Close() {
   Clear();
   if (_isOpen == FALSE)
      return(*this);

   Flush();
   CleanUp();

   _cfgfile.close();
   _isOpen = FALSE;

   return(*this);
   }


//********************************************************************
// MEMBER:	Delete
// DESCRIPTION:
//	@+@member description
//********************************************************************
NScfg& NScfg::Delete(CStr sec) {
   Clear();
   if (_isOpen == FALSE)
      Errnum(fileOpenErr);
   else if (_filemode == READONLY) 
      Errnum(fileWriteErr);
   else {
      NScfgNode* prevnode = 0;
      NScfgNode* currnode = Find(sec, prevnode);
      while (currnode) {
         NScfgNode* tmp = currnode->nextNode;
         delete currnode;
         currnode = tmp;
         if (currnode && currnode->lineType == 'S')
            break;
         }
      prevnode->nextNode = currnode;
      _isMod = TRUE;
      }

   return(*this);
   }


NScfg& NScfg::Delete(CStr sec, CStr ent) {
   Clear();
   if (_isOpen == FALSE)
      Errnum(fileOpenErr);
   else if (_filemode == READONLY)
      Errnum(fileWriteErr);
   else {
      NScfgNode* prevnode = 0;
      NScfgNode* currnode = Find(sec, ent, prevnode);
      if (currnode) {
         NScfgNode* tmp = currnode->nextNode;
         delete currnode;
         prevnode->nextNode = tmp;
         _isMod = TRUE;
         }
      }
   return(*this);
   }


//********************************************************************
// MEMBER:	Flush
// DESCRIPTION:
//	@+@member description
//********************************************************************
NScfg& NScfg::Flush() {
   Clear();
   if (_isOpen == FALSE) {
      Errnum(fileFlushErr);
      return(*this);
      }

   if (_isMod == FALSE) 
      return(*this);

   if (_filemode == READWRITE) {
      char tmpfilename[_POSIX_PATH_MAX];
      tmpnam(tmpfilename);
      Str src = strrchr(tmpfilename, DIRSLASH);

      char tmpfilepath[_POSIX_PATH_MAX];
      strcpy(tmpfilepath, _fname);
      Str target = strrchr(tmpfilepath, DIRSLASH);
      if (target == 0) 
         strcpy(tmpfilepath, src+1);
      else while (*src != '\0')
         *target++ = *src++;

      ofstream out(tmpfilepath);

      if (out.fail())
         Errnum(fileOpenErr);
      else {
         NScfgNode* currnode = _root->nextNode;
         while (currnode) {
            out << currnode->data << endl;
            if (out.fail()) {
               Errnum(fileWriteErr);
               return(*this);
               }
            currnode = currnode->nextNode;
            }
#ifdef _WINDOWS
	 //release NT lock on open file
	 _cfgfile.close();
	 out.close();
#endif
         nsRemove(_fname);
         rename(tmpfilepath, _fname);

#ifdef _WINDOWS
	 //reopen the file for normal operation
	 _isOpen=FALSE;
	 Open(_fname,_filemode);
#endif
         }
      }

   _isMod = FALSE;
   return(*this);
   }


//********************************************************************
// MEMBER:	Get
// DESCRIPTION:
//	@+@member description
//********************************************************************
NScfg& NScfg::Get(CStr sec, CStr ent, u32& integer) {
   integer = 0;

   Clear();
   if (_isOpen == FALSE)
      Errnum(fileOpenErr);
   else {
      NScfgNode* prevnode = 0;
      NScfgNode* currnode = Find(sec, ent, prevnode);
      if (currnode == 0) 
         Errnum(ctrGetErr);
      else {
         Str valueStr = strchr(currnode->data, '=');
         if (valueStr != 0) {
            SkipWS(++valueStr);
            integer = atoi(valueStr);
            }           
         }
      }
   return(*this);
   }


NScfg& NScfg::Get(CStr sec, CStr ent, Str& buf) {
   Clear();
   if (_isOpen == FALSE)
      Errnum(fileOpenErr);
   else {
      NScfgNode* prevnode = 0;
      NScfgNode* currnode = Find(sec, ent, prevnode);
      if (currnode == 0) 
         Errnum(ctrGetErr);
      else {
         Str valueStr = strchr(currnode->data, '=');
         if (valueStr != 0) {
            SkipWS(++valueStr);
            buf = valueStr;
            }           
         }
      }
   return(*this);
   }


NScfg& NScfg::Get(CStr sec, NScfgGet cb, PGen parm) {
   Clear();
   if (_isOpen == FALSE)
      Errnum(fileOpenErr);
   else {
      NScfgNode* prevnode = 0;
      NScfgNode* currnode = Find(sec, prevnode);
      if (currnode == 0) 
         Errnum(ctrGetErr);
      else while ( (currnode = currnode->nextNode) ) {
         if (currnode->lineType == 'S')
            break;
         if (currnode->lineType == 'V') {
            Str endOfLine = RTB(currnode->data + strlen(currnode->data) - 1);
            *endOfLine = '\0';

            Str attr = currnode->data;
            Str attrEnd = strchr(currnode->data, '=');
            Str val = attrEnd+1;
            
            attrEnd--;
            Str  endOfAttr = RTB(--attrEnd);
            char saveCh = *endOfAttr;
            *endOfAttr = '\0';
            SkipWS(val);

            if ((cb)(attr, val, parm) == FALSE)
               Errnum(ctrPutErr);

            *endOfAttr = saveCh;
            }
         }
      }
   return(*this);
   }


//********************************************************************
// MEMBER:	Open
// DESCRIPTION:
//	@+@member description
//********************************************************************
NScfg& NScfg::Open(CStr fn, Filemode fm) {
   Clear();
   if (_isOpen == TRUE)
      if (Close().Bad())
         return(*this);

   strcpy(_fname, fn);
   _filemode = fm;

   if (_filemode == NScfg::READONLY)
#ifdef _WINDOWS
      _cfgfile.open(fn, ios::in | ios::nocreate);
#else
      _cfgfile.open(fn, ios::in);
#endif
   else {
      _cfgfile.open(fn, ios::in | ios::out);
      _cfgfile.seekg(0);
      }

   if (_cfgfile.fail()) 
      Errnum(fileOpenErr);
   else {
      _isOpen = TRUE;
      Init();
      }
   return(*this);
   }


NScfg& NScfg::Open(CStr fn, CStr *paths, Filemode fm) {
#if XP_UNIX
   //
   // UNIX specific check: if the filename starts with /, do not
   // loop over the paths, since this is an absolute pathname.
   //
   if (*fn == '/') return Open(fn, fm);
#endif

   while (*paths) {
      NSfname full(fn, *paths);
      if (Open(full.fullname(), fm).Good()) 
         break;
      paths += 1;
      }

   return(*this);
   }


//********************************************************************
// MEMBER:	Put
// DESCRIPTION:
//	@+@member description
//********************************************************************
NScfg& NScfg::Put(Str sec) {
   Clear();
   if (_isOpen == FALSE)
      Errnum(fileOpenErr);
   else if (_filemode == READONLY)
      Errnum(fileWriteErr);
   else {
      NScfgNode* prevnode = 0;
      if (Find(sec, prevnode) != 0)
         return(*this);
      PutSection(sec);
      _isMod = TRUE;
      }
   return(*this);
   }


NScfg& NScfg::Put(Str sec, Str ent, Str buf) {
   Clear();
   if (_isOpen == FALSE)
      Errnum(fileOpenErr);
   else if (_filemode == READONLY)
      Errnum(fileWriteErr);
   else {
      NScfgNode* prevnode = 0;
      NScfgNode* secnode = 0;
      NScfgNode* entnode = 0;
      if ((secnode = Find(sec, prevnode)) == 0) {
         if (PutSection(sec) != 0)
            _lastNode = PutEntry(ent, buf, _lastNode);
         }
      else if ((entnode = Find(sec, ent, prevnode)) != 0) {
	 delete [] entnode->data;
	 entnode->data = new char[strlen(ent) + strlen(buf) + 4];
	 if (!entnode->data) {
            Errnum(outOfMem);
	    return (*this);
	    }
         sprintf(entnode->data, "%s = %s", ent, buf);
         } 
      else 
         PutEntry(ent, buf, secnode);

      _isMod = TRUE;
      }
   return(*this);
   }


//********************************************************************
// MEMBER:	Section
// DESCRIPTION:
//	@+@member description
//********************************************************************
Str NScfg::Section(NSbool  reset) {
   Clear();
   if (_isOpen == FALSE)
      Errnum(fileOpenErr);
   else {
      if (reset == TRUE)
         _pSection = _root;
    
      NScfgNode* currnode = _pSection;
      while (currnode) {
         if (currnode->lineType == 'S') {
            _pSection = currnode->nextNode;
            Str    source = currnode->data + 1;
            size_t namelen = strlen(source)+1; 
            if (_secname == 0 || _secnamelen < namelen) {
               delete []_secname;
               _secnamelen = namelen;
               if ((_secname = new char[_secnamelen]) == 0) {
                  Errnum(outOfMem);
                  break;
                  }
               }
            Str target = _secname;
            while (*source != ']')
               *target++ = *source++;
            *target = '\0';
            return(_secname);
            }
         currnode = currnode->nextNode;
         }
      }

   return(0);
   }


//********************************************************************
// MEMBER:	RequireFlush
// DESCRIPTION:
//	Call this to prevent changes from being implicitly flushed by
//      the destructor. Instead client must explicitly Flush (or
//      Close).
//********************************************************************
NScfg& NScfg::RequireFlush(NSbool  isrequired) {
   _flushRequired = isrequired;
   return(*this);
   }


//********************************************************************
// MEMBER:	Dump
// DESCRIPTION:
//	Debugging aid to dump the tree.
//********************************************************************
void NScfg::Dump(ostream& o) const {
   o << "  filename=" << _fname << endl
     << "  flushRequired=" << _flushRequired
     << "  fmode=" << (int) _filemode << ", isOpen=" << _isOpen
     << ", isMod=" << _isMod << endl;

   if (_root) {
      NScfgNode* currnode = _root->nextNode;
      while (currnode) {
         o << "  type=" << currnode->lineType << ", " << currnode->data << endl;
         currnode = currnode->nextNode;
         }
      } 
   }


//********************************************************************
// MEMBER:	Name
// DESCRIPTION:
//	Returns the class name for the instance.
//********************************************************************
CStr NScfg::Name() const {
   return("NScfg");
   }


//********************************************************************
// MEMBER:	CleanUp
// DESCRIPTION:
//	@+@member description
//********************************************************************
void NScfg::CleanUp() {
   NScfgNode* currnode = _root->nextNode;
   while (currnode) {
      NScfgNode* nextnode = currnode->nextNode;
      delete currnode;
      currnode = nextnode;;
      }
   _root->nextNode = 0;
   }


//********************************************************************
// MEMBER:	Find
// DESCRIPTION:
//	@+@member description
//********************************************************************
NScfgNode* NScfg::Find(CStr sec, NScfgNode*& prev) {
   NScfgNode* currnode = _root->nextNode;
   prev = _root;
   u16  seclen = strlen(sec);
   while (currnode) {
      if (currnode->lineType == 'S') {
         if (strncmp(currnode->data+1, sec, seclen) == 0) {
            char nextCh = currnode->data[1+seclen];
            if (nextCh == ']' || nextCh == ' ' || nextCh == '\t')
               return(currnode);
            }
         }
      prev = currnode;
      currnode = currnode->nextNode;
      }
   return(0);
   }


NScfgNode* NScfg::Find(CStr sec, CStr ent, NScfgNode*& prev) {
   NScfgNode* pSection = Find(sec, prev);
   
   if (pSection) {
      prev = pSection;

      NScfgNode* currnode = pSection->nextNode;
      u16        entlen   = strlen(ent);

      while (currnode) {
         if (currnode->lineType == 'S')
            break;
         if (currnode->lineType == 'V') 
            if (strncmp(currnode->data, ent, entlen) == 0) {
               char nextCh = currnode->data[entlen];
               if (nextCh == '=' || nextCh == ' ' || nextCh == '\t') {
                  return(currnode);
                  }
               }
         prev = currnode;
         currnode = currnode->nextNode;
         }
      }

   return(0);
   }


//********************************************************************
// MEMBER:	Init
// DESCRIPTION:
//	@+@member description
//********************************************************************
NSbool  NScfg::Init() {
   NScfgNode*    currnode = _root;
   strstreambuf  buf; 
   Str           startData = 0;

   while (_cfgfile.get(buf, '\n'), _cfgfile.peek() != EOF) {
      buf.sputc('\0');
#if !defined(OS_irix)
      buf.seekoff(0, ios::beg, ios::out);
#endif
      startData = buf.str();

      while ((*startData != 0) && (*startData == ' ' || *startData == '\t'))
         startData++;

      NScfgNode* newnode = new NScfgNode(strlen(startData)+1);
      if (newnode == 0) {
         Errnum(outOfMem);
         return(FALSE);
         }

      switch (*startData) {
         case '#':
         case ';':  newnode->lineType = 'C';  // comment
                    break;
         case '[':  newnode->lineType = 'S';  // section
                    break;
         case 0:    newnode->lineType = 'B';  // blank line
                    break;
         }

      strcpy(newnode->data, startData);
      currnode->nextNode = newnode;
      currnode = newnode;
     
#if defined(OS_irix)
      buf.seekoff(0, ios::beg, ios::out);
#endif
      _cfgfile.ignore();
      buf.freeze(0);
      }

   _lastNode = currnode;
   _pSection = _root;
   return(TRUE);
   }


//********************************************************************
// MEMBER:	PutEntry
// DESCRIPTION:
//	@+@member description
//********************************************************************
NScfgNode* NScfg::PutEntry(Str ent, Str buf, NScfgNode* prevnode) {
   NScfgNode* newnode = new NScfgNode(strlen(ent)+strlen(buf)+6);
   if (newnode == 0) 
      Errnum(outOfMem);
   else {
      NScfgNode* nextnode = prevnode->nextNode;
      prevnode->nextNode = newnode;
      newnode->nextNode = nextnode;
      delete [] newnode->data;
      newnode->data = new char[strlen(ent) + strlen(buf) + 4];
      if (!newnode->data) {
         Errnum(outOfMem);
         return (newnode);
      }
      sprintf(newnode->data, "%s = %s", ent, buf);
      }
   return(newnode);
   }


//********************************************************************
// MEMBER:	PutSection
// DESCRIPTION:
//	@+@member description
//********************************************************************
NScfgNode* NScfg::PutSection(Str sec) {
   NScfgNode* newnode = new NScfgNode(strlen(sec)+4);
   if (newnode == 0) 
      Errnum(outOfMem);
   else {
      _lastNode->nextNode = newnode;
      _lastNode = newnode;
      newnode->lineType = 'S';
      delete [] newnode->data;
      newnode->data = new char[strlen(sec) + 3];
      if (!newnode->data) {
           Errnum(outOfMem);
	   return (newnode);
      }
      sprintf(newnode->data, "[%s]", sec);
     }
   return(newnode);
   }

