/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

#ifndef XP_WIN32
#include <unistd.h>
#endif

#include "libnscore/nscfgn.h"
#include "libnscore/nsfname.h"


// public member functions

// Member: NScfgn
// Description:
//    The default constructor, this member allocates a root node

NScfgn::NScfgn() 
:  _filemode(READONLY), _isMod(FALSE), _isOpen(FALSE), _lastNode(0), 
   _root(new NScfgnNode) {
   if (_root == 0)
      Errnum(outOfMem);
   }


// Member: NScfgn
// Description:
//    Constructor to create a configuration file for a specified filename and 
//    filemode (note that the default filemode is READONLY).

NScfgn::NScfgn(CStr fn, Filemode fm)
:  _filemode(fm), _isMod(FALSE), _isOpen(FALSE),  _lastNode(0),
   _root(new NScfgnNode) {

   if (_root == 0)
      Errnum(outOfMem);
   Open(fn, fm);       
   }


// Member: NScfgn
// Description:
//    Constructor to create a configuration file for a specified filename and 
//    filemode (note that the default filemode is READONLY) by looking for the
//    file in the array of paths.

NScfgn::NScfgn(CStr fn, CStr* paths, Filemode fm)
:  _filemode(fm), _isMod(FALSE), _isOpen(FALSE),  _lastNode(0),
   _root(new NScfgnNode) {

   if (_root == 0)
      Errnum(outOfMem);
   Open(fn, paths, fm);       
   }


// Member: ~NScfgn
// Description:
//    The destructor for the configuration class invokes the Close function and
//    deletes the root node.

NScfgn::~NScfgn() {
   Close();
   delete _root;  _root = 0;
   }


// Member: Close
// Description:
//    Close the configuration file, if it's open.  This function is called from
//    the dtor as well as the Open function, so it's probably not necessary to use
//    unless you're really into consistency.

NScfgn& 
NScfgn::Close() {
   if (_isOpen == FALSE)
      return(*this);

   if (_isMod)
     if (Flush() == FALSE)
        Errnum(fileCloseErr);

   NScfgnNode* currnode = _root->nextNode;
   while (currnode) {
      NScfgnNode* nextnode = currnode->nextNode;
      delete currnode;  currnode = 0;
      currnode = nextnode;;
      }
   _root->nextNode = 0;

   _isOpen = FALSE;
   _isMod = FALSE;

   return(*this);
   }


// Member: Delete
// Description:

NSbool
NScfgn::Delete(CStr sec) {
   if (_isOpen == FALSE || _filemode == READONLY)
      return(FALSE);

   NScfgnNode* prevnode = 0;
   NScfgnNode* currnode = Find(sec, prevnode);
   while (currnode) {
      NScfgnNode* tmp = currnode->nextNode;
      delete currnode;  currnode = 0;
      currnode = tmp;
      if (currnode && currnode->lineType == 'S')
         break;
      }
   prevnode->nextNode = currnode;
   _isMod = TRUE;

   return(TRUE);
   }


// Member: Delete
// Description:

NSbool 
NScfgn::Delete(CStr sec, CStr ent) {
   if (_isOpen == FALSE || _filemode == READONLY)
      return(FALSE);

   NScfgnNode* prevnode = 0;
   NScfgnNode* currnode = Find(sec, ent, prevnode);
   if (currnode) {
      NScfgnNode* tmp = currnode->nextNode;
      delete currnode;  currnode = 0;
      prevnode->nextNode = tmp;
      _isMod = TRUE;
      }

   return(TRUE);
   }


// Member: Get
// Description:
//    For a specified section and attribute, return the value (as a int).

NSbool 
NScfgn::Get(CStr sec, CStr ent, int& value) {
   value = 0;
   if (_isOpen == FALSE)
      return(FALSE);
   NScfgnNode* prevnode = 0;
   NScfgnNode* currnode = Find(sec, ent, prevnode);
   if (currnode == 0) 
      return(FALSE);
   Str valueStr = strchr(currnode->data, '=');
   if (valueStr != 0) 
      value = atoi(valueStr+1);
   return(TRUE);
   }


// Member: Get
// Description:
//    For a specified section and attribute, return the value (as a string).  Note
//    that we're returning a pointer.  Copy the value, especially if you intend to
//    add any more data to the configuration file (not really recommended in a 
//    thread environment -- you'll need to be careful)

NSbool 
NScfgn::Get(CStr sec, CStr ent, Str& buf) {
   buf = 0;
   if (_isOpen == FALSE)
      return(FALSE);
   NScfgnNode* prevnode = 0;
   NScfgnNode* currnode = Find(sec, ent, prevnode);
   if (currnode == 0) 
      return(FALSE);
   Str valueStr = strchr(currnode->data, '=');
   if (valueStr != 0) 
      buf = valueStr+1;
   return(TRUE);
   }


// Member: Get
// Description:
//    Get all of the name/value pairs for a particular section, invoking the 
//    specified callback for each, and passing the generic parameters.  Note that
//    a copy is made of the in-memory configuration file line, so that we can
//    cut it apart into attribute and value.  This is done to prevent problems in
//    a MT environment (after all, it's possible that many threads are sharing the
//    single in-memory version of this file).

NSbool 
NScfgn::Get(CStr sec, NScfgnGetCB cb, void* parm) {
   if (_isOpen == FALSE)
      return(FALSE);
   NScfgnNode* prevnode = 0;
   NScfgnNode* currnode = Find(sec, prevnode);
   if (currnode == 0) 
      return(FALSE);
   int buflen = 256;
   Str tmpStr = new char[buflen];
   if (tmpStr == 0)
      return(FALSE);
   while ( (currnode = currnode->nextNode) ) {
      if (currnode->lineType == 'S')
         break;
      if (currnode->lineType == 'V') {
         // don't like it, but it'll do for now.
         int len = strlen(currnode->data)+1;
         if (len > buflen) {
            delete []tmpStr; tmpStr = 0;
            buflen = len;
            if ((tmpStr = new char[buflen]) == 0)
               return(FALSE);
            }
         strcpy(tmpStr, currnode->data);
         Str attrEnd = strchr(tmpStr, '=');
         *attrEnd = '\0';
         if ((cb)(tmpStr, attrEnd+1, parm) == FALSE)
            return(FALSE);
         }
      }
   delete []tmpStr; tmpStr = 0;
   return(TRUE);
   }


// Member: Open
// Description:
//    Open the configuration file and create an in-memory representation of the 
//    configuration file for the Get/Put operations.  Close the file first, if it
//    is already open, then open the file for the specfied file and mode.

NScfgn& 
NScfgn::Open(CStr fn, Filemode fm) {
   _filemode = fm;
   if (_isOpen == TRUE)
      if (Close().Bad())
         return(*this);

   if (fm == NScfgn::READONLY)
#ifdef XP_WIN32
      _cfgfile.open(fn, ios::in | ios::nocreate);
#else
      _cfgfile.open(fn, ios::in);
#endif
   else {
      _cfgfile.open(fn, ios::in | ios::out);
      _cfgfile.seekg(0);
      }

   if (_cfgfile.fail()) 
      Errnum(fileOpenErr);
   else {
      strcpy(_fname, fn);
      _isOpen = TRUE;
      _isMod = FALSE;
      Init();
      _cfgfile.close();
      }
   return(*this);
   }


// Member: Open
// Description:
//    Helper function to invoke Open on an array of paths, stopping when the 
//    configuration file is finally found.

NScfgn& 
NScfgn::Open(CStr fn, CStr *paths, Filemode fm) {
#if XP_UNIX
   //
   // UNIX specific check: if the filename starts with /, do not
   // loop over the paths, since this is an absolute pathname.
   //
   if (*fn == '/') return Open(fn, fm);
#endif

   while (*paths) {
      NSfname full(fn, *paths);
      if (Open(full.fullname(), fm).Good()) 
         break;
      paths += 1;
      }

   return(*this);
   }


// Member: Put
// Description:
//   Add a new section into the configuration.  Note that in a MT program, this 
//   would be in a critical section, since (and you definitely don't want to mix
//   the "Get"-type functions with the "Put"-functions without being careful.

NSbool 
NScfgn::Put(CStr sec) {
   if (_isOpen == FALSE || _filemode == READONLY)
      return(FALSE);

   NScfgnNode* prevnode = 0;
   if (Find(sec, prevnode) != 0)
      return(TRUE);
   if (PutSection(sec) == FALSE)
      return(FALSE);
   _isMod = TRUE;
   return(TRUE);
   }


// Member: Put
// Description:
//   Add a new section and entry into the configuration.  Create a section if it
//   doesn't already exist, overwrite the entry if it does.  Note that in a MT
//   program, this would be in a critical section, since (and you definitely don't 
//   want to mix the "Get"-type functions with the "Put"-functions without being
//   careful.

NSbool 
NScfgn::Put(CStr sec, CStr ent, CStr buf) {
   if (_isOpen == FALSE || _filemode == READONLY)
      return(FALSE);

   NScfgnNode* prevnode = 0;
   NScfgnNode* secnode = 0;
   NScfgnNode* entnode = 0;

   if ((secnode = Find(sec, prevnode)) == 0) {
      // could not find the section, so add the section and entry
      if (PutSection(sec) == TRUE) {
         if ((_lastNode = PutEntry(ent, buf, _lastNode)) == 0)
            return(FALSE);
         }
      else 
         return(FALSE);
      }
   else if ((entnode = Find(sec, ent, prevnode)) != 0) {
      // found the section and entry, so overwrite the entry
      delete [] entnode->data; entnode->data = 0;
      entnode->data = new char[strlen(ent) + strlen(buf) + 4];
      if (entnode->data == 0) 
         return (FALSE);
      sprintf(entnode->data, "%s = %s", ent, buf);
      } 
   else {
      // found the section, but the entry wasn't there, so add it.
      if (PutEntry(ent, buf, secnode) == 0)
         return(FALSE);
      }

   _isMod = TRUE;
   return(TRUE);
   }




// Member: Section
// Description:
//    Iterate through all of the section headings in a configuration file. The 
//    caller passes a buffer to place the section name into, and a cookie for 
//    maintaining state for the iteration.  For the first call, set cookie to 0.
//    Subsequent calls should use the cookie returned by the call.  All section
//    names have been returned when cookie == 0.

NSbool 
NScfgn::Section(Str secname, int seclen, void*& cookie) {
   if (_isOpen == FALSE)
      return(FALSE);
   NScfgnNode* pSection = (NScfgnNode*)cookie ? (NScfgnNode*)cookie : _root;
   NScfgnNode* currnode = pSection;
   secname[0] = 0;
   while (currnode) {
      if (currnode->lineType == 'S') {
         pSection = currnode->nextNode;
         Str  source = currnode->data + 1;
         int  len = 0;
         while (*source++ != ']')
            len++;
         if (len > seclen)
            len = seclen;
         strncpy(secname, currnode->data+1, len);
         secname[len] = '\0';
         cookie = pSection;
         return(TRUE);
         }
      currnode = currnode->nextNode;
      }
   cookie = 0;

   return(FALSE);
   }


// protected member functions

void 
NScfgn::Dump(ostream& o) const {
   if (_root) {
      NScfgnNode* currnode = _root->nextNode;
      while (currnode) {
         o << "  type=" << currnode->lineType << ", " 
           << currnode->data << endl;
         currnode = currnode->nextNode;
         }
      } 
   }


CStr 
NScfgn::Name() const {
   return("NScfgn");
   }


// private member functions

NScfgnNode* 
NScfgn::Find(CStr sec, NScfgnNode*& prev) {
   NScfgnNode* currnode = _root->nextNode;
   prev = _root;
   int  seclen = strlen(sec);
   while (currnode) {
      if (currnode->lineType == 'S') {
         if (strncmp(currnode->data+1, sec, seclen) == 0) {
            char nextCh = currnode->data[1+seclen];
            if (nextCh == ']' || nextCh == ' ' || nextCh == '\t')
               return(currnode);
            }
         }
      prev = currnode;
      currnode = currnode->nextNode;
      }
   return(0);
   }


NScfgnNode* 
NScfgn::Find(CStr sec, CStr ent, NScfgnNode*& prev) {
   NScfgnNode* pSection = Find(sec, prev);
   
   if (pSection) {
      prev = pSection;

      NScfgnNode* currnode = pSection->nextNode;
      int        entlen   = strlen(ent);

      while (currnode) {
         if (currnode->lineType == 'S')
            break;
         if (currnode->lineType == 'V') 
            if (strncmp(currnode->data, ent, entlen) == 0) {
               char nextCh = currnode->data[entlen];
               if (nextCh == '=' || nextCh == ' ' || nextCh == '\t') {
                  return(currnode);
                  }
               }
         prev = currnode;
         currnode = currnode->nextNode;
         }
      }

   return(0);
   }


NSbool 
NScfgn::Flush() {
   // what we want to do here is remove filename.old, rename filename to filename.old,
   // then write out a new filename.  note that we closed the file, once the Init call
   // completed.

   // we'd only be called if the data had been modified (which would only happen if the
   // file had been opened READWRITE

   char backupFile[_POSIX_PATH_MAX];
   sprintf(backupFile, "%s.backup", _fname);
   nsRemove(backupFile);
   rename(_fname, backupFile);
      
   _cfgfile.open(_fname, ios::in | ios::out);
   _cfgfile.seekg(0);

   if (_cfgfile.fail()) {
      Errnum(fileOpenErr);
      return(FALSE);
      }

   NScfgnNode* currnode = _root->nextNode;
   while (currnode) {
      _cfgfile << currnode->data << endl;
      if (_cfgfile.fail()) {
         Errnum(fileWriteErr);
         return(FALSE);
         }
      currnode = currnode->nextNode;
      }

   _cfgfile.close();
   return(TRUE);
   }


NSbool  
NScfgn::Init() {
   NScfgnNode*   currnode = _root;
   strstreambuf  buf; 
   Str           startData = 0;
   int           lenTmpVal = 256;
   Str           tmpVal = new char[lenTmpVal];
   
   if (tmpVal == 0) {
      Errnum(outOfMem);
      return(FALSE);
      }

   while (_cfgfile.get(buf, '\n'), _cfgfile.peek() != EOF) {
      buf.sputc('\0');
#if !defined(OS_irix)
      buf.seekoff(0, ios::beg, ios::out);
#endif
      startData = buf.str();

      while ((*startData != 0) && (*startData == ' ' || *startData == '\t'))
         startData++;

      NScfgnNode* newnode = new NScfgnNode(strlen(startData)+1);
      if (newnode == 0 || newnode->data == 0) {
         delete []tmpVal; tmpVal = 0;
         Errnum(outOfMem);
         return(FALSE);
         }

      switch (*startData) {
         case '#':
         case ';':  newnode->lineType = 'C';  // comment
                    break;
         case '[':  newnode->lineType = 'S';  // section
                    break;
         case 0:    newnode->lineType = 'B';  // blank line
                    break;
         }

      strcpy(newnode->data, startData);
      Str equalSign = strchr(newnode->data, '=');
      if (newnode->lineType == 'V')
         if (equalSign == 0) 
            newnode->lineType = 'C';
         else {
            Str endOfAttr = RTB(equalSign-1);
            *endOfAttr = '=';
            Str valStart = equalSign+1;
            while (*valStart == ' ' || *valStart == '\t')
              valStart++;
            int  valLen = strlen(valStart)+1;
            if (valLen > lenTmpVal) {
               delete []tmpVal; tmpVal = 0;
               tmpVal = new char[valLen];
               if (tmpVal == 0) {
                  Errnum(outOfMem);
                  return(FALSE);
                  }
               lenTmpVal = valLen;
               }
            strcpy(tmpVal, valStart);
            strcpy(endOfAttr+1, tmpVal);
            }
      if (newnode->lineType != 'B') {
         Str endOfLine = RTB(newnode->data+strlen(newnode->data)-1);
         *endOfLine = '\0';
         }

      currnode->nextNode = newnode;
      currnode = newnode;
     
#if defined(OS_irix)
      buf.seekoff(0, ios::beg, ios::out);
#endif
      _cfgfile.ignore();
      buf.freeze(0);
      }

   _lastNode = currnode;
   delete []tmpVal; tmpVal = 0;
   return(TRUE);
   }


NScfgnNode* 
NScfgn::PutEntry(CStr ent, CStr buf, NScfgnNode* prevnode) {
   NScfgnNode* newnode = new NScfgnNode(strlen(ent)+strlen(buf)+6);
   if (newnode == 0 || newnode->data == 0) 
      return(0);

   NScfgnNode* nextnode = prevnode->nextNode;
   prevnode->nextNode = newnode;
   newnode->nextNode = nextnode;
   sprintf(newnode->data, "%s = %s", ent, buf);
   return(newnode);
   }


NSbool 
NScfgn::PutSection(CStr sec) {
   NScfgnNode* newnode = new NScfgnNode(strlen(sec)+4);
   if (newnode == 0 || newnode->data == 0) 
      return(FALSE);
   _lastNode->nextNode = newnode;
   _lastNode = newnode;
   newnode->lineType = 'S';
   sprintf(newnode->data, "[%s]", sec);

   return(TRUE);
   }

// private member functions

Str 
NScfgn::RTB(Str src) {
   while (*src == ' ' || *src == '\t')
      --src;
   return(src+1);
   }


