/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

#include "libnscore/nsgetmeta.h"

const char BLANK = ' '; 
const char NL = '\n';
const char DATA = '\0';         // don't care data
const char TAB = '\t';  
const char QUOTE = 0x27;        

const int  COMMENT = 6;         // checking for  HTML comment
const int  CONTENT = 43;        // checking for CONTENT attribute
const int  ATTR = 28;           // start of attribute (maybe a ")
const int  END = 11;            // handle HTML tag end
const int  EQUALS = 21;         // parsing equal sign (=)
const int  META = 2;            // HTML meta tag
const int  NAME = 39;           // checking for NAME attribute
const int  SKIPTOKEN = 16;      // skip to end of token
const int  SKIPTOKENDATA = 20;  // skip to end of token
const int  WS = 12;             // whitespace before an attribute

const int  DONTCARE = -1;       // don't care state, will never happen
const int  FOUND = -1;          // found what was being looked for
const int  REPEAT = -1;         // emit and redo check
const int  START = 0;           // tag start

const int  UNKNOWN_ATTR = 0;    // unknown meta tag attribute
const int  NAME_ATTR = 1;       // NAME meta tag attribute
const int  CONTENT_ATTR = 2;    // CONTENT meta tag attribute

struct FSMelement {
   char target;
   int  onMatch;
   int  onFail;
   int  attrType;
   };

const FSMelement FSM[] = {
   //           onMatch         onFail          attrType
   
   // START: find start of tag (<) and skip if it's a comment (!)
   {'<',        START+1,        REPEAT,         UNKNOWN_ATTR},  // 0       
   {'!',        COMMENT,        META,           UNKNOWN_ATTR},

   // META: this should be a meta tag
   {'m',        META+1,         END,            UNKNOWN_ATTR},  // 2
   {'e',        META+2,         END,            UNKNOWN_ATTR},
   {'t',        META+3,         END,            UNKNOWN_ATTR},
   {'a',        WS,             END,            UNKNOWN_ATTR},  

   // COMMENT: ignore HTML comments
   {'-',        COMMENT+1,      END,            UNKNOWN_ATTR},  // 6
   {'-',        COMMENT+2,      END,            UNKNOWN_ATTR},
   {'-',        COMMENT+3,      REPEAT,         UNKNOWN_ATTR},
   {'-',        COMMENT+4,      COMMENT+2,      UNKNOWN_ATTR},
   {'>',        START,          COMMENT+2,      UNKNOWN_ATTR},

   // END: find end tag
   {'>',        START,          REPEAT,         UNKNOWN_ATTR},  // 11

   // WS: parse out leading white space
   {BLANK,      WS,             WS+1,           UNKNOWN_ATTR},  // 12
   {TAB,        WS,             WS+2,           UNKNOWN_ATTR},
   {NL,         WS,             WS+3,           UNKNOWN_ATTR},
   {'>',        START,          NAME,           UNKNOWN_ATTR},

   // SKIPTOKEN: skip until the end of a token
   {BLANK,      WS,             SKIPTOKEN+1,    UNKNOWN_ATTR},  // 16
   {TAB,        WS,             SKIPTOKEN+2,    UNKNOWN_ATTR},
   {NL,         WS,             SKIPTOKEN+3,    UNKNOWN_ATTR},
   {'>',        START,          SKIPTOKENDATA,  UNKNOWN_ATTR},

   // SKIPTOKENDATA: token data to skip    
   {DATA,       SKIPTOKEN,      DONTCARE,       UNKNOWN_ATTR},  // 20

   // EQUALS: parse out the "=", with/without leading whitespace
   {BLANK,      EQUALS,         EQUALS+1,       UNKNOWN_ATTR},  // 21
   {TAB,        EQUALS,         EQUALS+2,       UNKNOWN_ATTR},
   {NL,         EQUALS,         EQUALS+3,       UNKNOWN_ATTR},
   {'=',        EQUALS+4,       SKIPTOKEN,      UNKNOWN_ATTR},  
   {BLANK,      EQUALS+4,       EQUALS+5,       UNKNOWN_ATTR},
   {TAB,        EQUALS+4,       EQUALS+6,       UNKNOWN_ATTR},  
   {NL,         EQUALS+4,       ATTR,           UNKNOWN_ATTR},

   // ATTR: parse out the attribute's value (with/without ")
   {'"',        ATTR+1,         ATTR+8,         UNKNOWN_ATTR},  // 28
   {'"',        FOUND,          ATTR+2,         UNKNOWN_ATTR},
   {DATA,       ATTR+1,         DONTCARE,       UNKNOWN_ATTR},
   {'>',        FOUND,          ATTR+4,         UNKNOWN_ATTR},
   {BLANK,      FOUND,          ATTR+5,         UNKNOWN_ATTR},
   {TAB,        FOUND,          ATTR+6,         UNKNOWN_ATTR},
   {NL,         FOUND,          ATTR+7,         UNKNOWN_ATTR},
   {DATA,       ATTR+3,         DONTCARE,       UNKNOWN_ATTR},
   {QUOTE,      ATTR+9,         ATTR+3,         UNKNOWN_ATTR},
   {QUOTE,      FOUND,          ATTR+10,        UNKNOWN_ATTR},
   {DATA,       ATTR+9,         DONTCARE,       UNKNOWN_ATTR},

   // here are the attributes we're looking for
   {'n',        NAME+1,         CONTENT,        UNKNOWN_ATTR},  // 39
   {'a',        NAME+2,         SKIPTOKEN,      UNKNOWN_ATTR},
   {'m',        NAME+3,         SKIPTOKEN,      UNKNOWN_ATTR},
   {'e',        EQUALS,         SKIPTOKEN,      NAME_ATTR},     

   {'c',        CONTENT+1,      SKIPTOKEN,      UNKNOWN_ATTR},  // 43
   {'o',        CONTENT+2,      SKIPTOKEN,      UNKNOWN_ATTR},
   {'n',        CONTENT+3,      SKIPTOKEN,      UNKNOWN_ATTR},
   {'t',        CONTENT+4,      SKIPTOKEN,      UNKNOWN_ATTR},
   {'e',        CONTENT+5,      SKIPTOKEN,      UNKNOWN_ATTR},
   {'n',        CONTENT+6,      SKIPTOKEN,      UNKNOWN_ATTR},
   {'t',        EQUALS,         SKIPTOKEN,      CONTENT_ATTR}
   };


NSgetMeta::NSgetMeta(char* start, int length, CBgetMeta cb, void* parm) {
   // initialized some state
   int      attrType = UNKNOWN_ATTR;
   int      currState = 0;
   char*    attr = 0;
   int      attrLen = 0;
   char*    p = start;
   char     target = DATA;

   int nameFound = 0;
   int contentFound = 0;

   // allocate buffers to hold data that we will pass to the callback
   int   nameAttrLen = 127;
   int   contentAttrLen = 127;
   char* nameAttr = new char[128];
   char* contentAttr = new char[128];
   if (nameAttr == 0 || contentAttr == 0) {
      delete []nameAttr;
      delete []contentAttr;
      return;
      }
   while (p != start+length) {
      ///cout << currState << ". " << *p << endl;
      if  (FSM[currState].attrType != UNKNOWN_ATTR) 
         attrType = FSM[currState].attrType;

      target = FSM[currState].target;
      if (target == DATA) {
         if (currState != SKIPTOKENDATA) {
            if (attr == 0)  
               attr = p;
            attrLen++;
            }
         *p++;
         currState = FSM[currState].onMatch;
         }
      else if (target == char(tolower(*p))) {
         *p++;
         currState = FSM[currState].onMatch;
         if (currState == FOUND) {
            if (attrType == NAME_ATTR) {
               nameFound = 1;
               if (attrLen > nameAttrLen) {
                  delete []nameAttr;
                  if ((nameAttr = new char[attrLen+1]) == 0)
                     goto CLEANUP;
                  }
               strncpy(nameAttr, attr, attrLen);
               nameAttr[attrLen] = '\0';
               }
            else if (nameFound) {
               contentFound = 1;
               if (attrLen > contentAttrLen) {
                  delete []contentAttr;
                  if ((contentAttr = new char[attrLen+1]) == 0)
                     goto CLEANUP;
                  }
               strncpy(contentAttr, attr, attrLen);
               contentAttr[attrLen] = '\0';
               }

            if (nameFound && contentFound) {
               int rcCB = (*cb)(nameAttr, contentAttr, parm);
               if (rcCB)
                  return;
               nameFound = 0;
               contentFound = 0;
               }
            
            attrType = UNKNOWN_ATTR;
            attr = 0;
            attrLen = 0;
            currState = WS;
            }
         }
      else if (FSM[currState].onFail == REPEAT)
         *p++;
      else
         currState = FSM[currState].onFail;
      }

CLEANUP:
   delete []nameAttr;
   delete []contentAttr;
   return;
   }
