/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

/***********************************************************************
** Includes
***********************************************************************/
#include "libnscore/nsskplst.h"


//********************************************************************
// MEMBER:	NSskiplist
// DESCRIPTION:
//	Default constructor.
//********************************************************************
NSskiplist::NSskiplist() 
:  _count(0), _currlevel(0), _start(0) {
   if ((_start = (SLnode*)NewNode(LEVELS)) != 0) {
      for (int i = 0; i < LEVELS; i++) _start->fwd[i] = 0;
      srand((unsigned int)(time(0)));
      _randomBits = rand();
      _randomsLeft = BITS_IN_RANDOM >> 1;
      }
   else
      Errnum(outOfMem);   
   }


//********************************************************************
// MEMBER:	~NSskiplist
// DESCRIPTION:
//	Destructor.
//********************************************************************
NSskiplist::~NSskiplist() {
   SLnode* p = _start;
   SLnode* q = 0;
   while (p != 0) {
      q = p->fwd[0];
      delete []p->key;
      delete []p->data;
      if (p != 0)
         delete [](char*)p;  
      p = q;
      }
   }


//********************************************************************
// MEMBER:	Delete
// DESCRIPTION:
//	Delete the given key from the skiplist.
//********************************************************************
NSskiplist& NSskiplist::Delete(CStr key) {
   Clear();
   SLnode* p = _start;
   SLnode* q = 0;
   s16     r = -1;
   s16     s = _currlevel;

   do {
      while ((q = p->fwd[s]) != 0) {
         if ((r = strcmp(q->key, key)) < 0)
            p = q;
         else
            break;
         }
      _update[s] = p;
      } while (--s >= 0);

   if (r == 0) {
      s16  t = _currlevel;
      for (s = 0; s <= t && (p=_update[s])->fwd[s] == q; s++) 
         p->fwd[s] = q->fwd[s];
      delete []q->key;
      delete []q->data;
      delete q;
      q = 0;
      while (_start->fwd[t] == 0 && t > 0)
         t--;
      _currlevel = t;
      _count--;
      }

   return(*this);
   }


//********************************************************************
// MEMBER:	First
// DESCRIPTION:
//	Returns the first element in the skiplist.
//********************************************************************
NSskiplist& NSskiplist::First(Str& key, Str& data, PGen& cookie) {
   Clear();
   if (_start != 0 && _start->fwd[0] != 0) {
      SLnode* tmp = _start->fwd[0];
      key = tmp->key;
      data = tmp->data;
      cookie = tmp->fwd[0];
      }
   else {
      data = key = 0;
      cookie = 0;
      }
   return(*this);
   }


//********************************************************************
// MEMBER:	Last
// DESCRIPTION:
//	Returns the last element in the skiplist.
//********************************************************************
NSskiplist& NSskiplist::Last(Str& key, Str& data) {
   Clear();
   s16     s = _currlevel;
   SLnode* p = _start;
   SLnode* q = 0;
   do {
      while ((q = p->fwd[s]) != 0)
         p = q;
      } while (--s >= 0);

   if (p != 0) {
      data = p->data;
      key = p->key;
      }
   else
      data = key = 0;
    
   return(*this);
   }


//********************************************************************
// MEMBER:	Next
// DESCRIPTION:
//	Returns the next element in the skiplist.
//********************************************************************
NSskiplist& NSskiplist::Next(Str& key, Str& data, PGen& cookie) {
   SLnode* p = (SLnode*)cookie;
   if (p != 0) {
      data = p->data;
      key = p->key;
      cookie = p->fwd[0];
      }
   else {
      data = key = 0;
      cookie = 0;
      }
   return(*this);
   }


//********************************************************************
// MEMBER:	Put
// DESCRIPTION: 
//	Create an entry in the skiplist using the given key and data
//	values.
//********************************************************************
NSskiplist& NSskiplist::Put(CStr key, CStr data, NSbool  allowDuplicates) {
   Clear();
   SLnode* p = _start;
   SLnode* q = 0;

   s16 r = -1;
   s16 s = _currlevel;

   do {
      while ((q = p->fwd[s]) != 0) {
         if ((r = strcmp(q->key, key)) < 0)
            p = q;
         else
            break;
         }
      _update[s] = p;
      } while (--s >= 0);


   if (r == 0 && !allowDuplicates) {
      delete []q->data;
      q->data = new char[strlen(data)+1];
      if (!q->data) 
         Errnum(outOfMem);
      else
         strcpy(q->data, data);
      return(*this);
      }

   s = RandomLevel();
   if (s > _currlevel) {
      s = ++_currlevel;
      _update[s] = _start;
      }

   if ((q = (SLnode*)NewNode(s)) == 0)
      return(*this);

   q->key = new char[strlen(key)+1];
   q->data = new char[strlen(data)+1];
   if (!q->key || !q->data) {
      delete []q->key;
      delete []q->data;
      Errnum(outOfMem);
      return(*this);
      }
   strcpy(q->key, key);
   strcpy(q->data, data);

   do {
      p = _update[s];
      q->fwd[s] = p->fwd[s];
      p->fwd[s] = q;
      } while (--s >= 0);

   _count++;
   return(*this);
   }


//********************************************************************
// MEMBER:	Dump
// DESCRIPTION:
//	Debugging aid to dump the contents of the current object.
//********************************************************************
void NSskiplist::Dump(ostream& o) const {
   SLnode* p = _start->fwd[0];
   while (p != 0) {
      o << "Key=" << p->key << ", data=" << p->data << endl;
      p = p->fwd[0];
      };
   }


//********************************************************************
// MEMBER:	Name
// DESCRIPTION:
//	Returns the class name for the instance.
//********************************************************************
CStr NSskiplist::Name() const {
   return("NSskiplist");
   }


//********************************************************************
// MEMBER:	Get
// DESCRIPTION:
//	Get the node from the skiplist for the given key.
//********************************************************************
NSskiplist& NSskiplist::Get(CStr key, SLnode*& node) {
   Clear();
   SLnode* p = _start;
   SLnode* q = 0;

   s16 r = -1;
   s16 s = _currlevel;

   do {
      while ((q = p->fwd[s]) != 0) {
         if ((r = strcmp(q->key, key)) < 0)
            p = q;
         else
            break;
         }
      } while (--s >= 0);

   node = (r == 0) ? q : 0;

   return(*this);
   }


//********************************************************************
// MEMBER:	RamdomLevel
// DESCRIPTION:
//	@+@member description
//********************************************************************
s16 NSskiplist::RandomLevel() {
   s16 level = 0;
   s32 b;
   do {
      if ((b = (_randomBits & 3)) == 0)
         level++;
      _randomBits >>= 2;
      if (--_randomsLeft == 0) {
         _randomBits = rand();
         _randomsLeft = BITS_IN_RANDOM >> 1;
         };
      } while (b == 0);
   return(level > (LEVELS-1) ? (LEVELS-1) : level);
   }

