/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "netsite.h"
#include "base/crit.h"
#include "base/systhr.h"

#include "mcom_db.h"
#include "nsres.h"
#include "i18n.h"

#define MAX_LIBNAME_LEN   32
#define NUM_DATA_BUCKETS 128 /* must be a power of 2 */
#define NUM_LANG_BUCKETS  16 /* must be a power of 2 */
#define NUM_LIB_BUCKETS   16 /* must be a power of 2 */

#define DATA_BUCKET_MASK NUM_DATA_BUCKETS-1
#define LANG_BUCKET_MASK NUM_LANG_BUCKETS-1
#define LIB_BUCKET_MASK  NUM_LIB_BUCKETS-1

/* End Of language List character and string; */
/* can not be first character in any filename */
#define EOL_CHAR   '\1'
#define EOL_STRING "\1"

static char emptyString[] = "";

#define COMPILE_STRINGS_IN_MEMORY
/*  Use #define COMPILE_STRINGS_IN_MEMORY to put message string tables in memory.
 *  This is temporary solution for product which doesn't know where
 *  to load message table from.
 */

char*
XP_GetStringFromMemory(char* strLibraryName,int iToken);


/*******************************************************************************/

/*  Global data structure and variable used in the file */

/*
 * pLangList -> LANGLIST       one for each language (link list)
 *                 |
 *                 V
 *   pDBList -> DBFILE         one for each database file,
 *                 |           may support several languages
 *                 V
 *  pNextLib -> LIBLIST        one for each library within database (link list)
 *                 |           (has hash table for resource strings)
 *                 V
 *    pCache -> RESCACHEDATA   one for each resource string
 *                             (link list from hash table)
 */

typedef struct RESCACHEDATA {
  int id;
  char* str;
  struct RESCACHEDATA* nextData;
} RESCACHEDATA;

typedef struct LIBLIST {
  char library[MAX_LIBNAME_LEN];
  struct LIBLIST* nextLib;
  struct RESCACHEDATA* firstData[NUM_DATA_BUCKETS];
} LIBLIST;

typedef struct DBFILE {
  char* filename;
  NSRESHANDLE hresdb;
  struct DBFILE* nextFile;
  struct LIBLIST* firstLib;
} DBFILE;

typedef struct LANGLIST {
  char language[MAX_ACCEPT_LENGTH];
  struct LANGLIST* nextLang;
  struct DBFILE* fileForLang;
} LANGLIST;

static LANGLIST LangListBase = { EOL_STRING,NULL,NULL }; /* first in link list */
static DBFILE DBListBase = { NULL,0,NULL,NULL };    /* first file in link list */

static CRITICAL reslock = NULL;    /* Mutex variable used by APIs in this file */

static char* pathDB = NULL;        /* keep path passed at initialization */
static char  existDBstr[1] = "";   /* path for database found to exist */
static char* existDB = existDBstr; /* path for database found to exist */

/*******************************************************************************/

NSAPI_PUBLIC
void
XP_InitStringDatabase(char* pathCWD, char* databaseName)
{
  pathDB = (char*)malloc(strlen(pathCWD)
                   +strlen("/$$LANGDIR/")
                   +strlen(databaseName)
                   +1);
  if (pathDB)
    {
      sprintf(pathDB,"%s%s%s",pathCWD,"/$$LANGDIR/",databaseName);
    }
  else
    {
      fprintf(stderr,"malloc failed for %s%s%s\n",
              pathCWD,"/$$LANGDIR/",databaseName);
      pathDB = NULL;
    }
  
  existDB = (char*)malloc(strlen(pathDB)+MAX_ACCEPT_LENGTH);
  if (!existDB)
    {
      fprintf(stderr,"malloc failed %s +%d\n",MAX_ACCEPT_LENGTH,pathDB);
    }

#ifndef NS_MAIL
  if ((reslock == NULL) && (systhread_current())) {
    reslock = crit_init();
  } else {
    reslock = NULL;
  }
#endif
}

/*******************************************************************************/

static char*
LoadStringFromCache(struct LIBLIST* pNextLib, char* library, int id)
{
    RESCACHEDATA* pCache;

    /* find the library */
    while (pNextLib->nextLib) {
      if (strcmp(pNextLib->library,library)) {
        /* look further */
        pNextLib = pNextLib->nextLib;
      } else {
        /* found the library; use hash table to find entry */
        pCache = pNextLib->firstData[id&DATA_BUCKET_MASK];
        while (pCache->id != 0) {
          if (pCache->id == id) {
            return pCache->str;
          }
          pCache = pCache->nextData;
        }
        return NULL;
      }
    }
    
    /* library was not found; caller will handle this */
    return NULL;
}

/*******************************************************************************/

static char*
SaveStringToCache(struct LIBLIST* pNextLib, char* library, int id, char* str)
{
    RESCACHEDATA* pCache;
    RESCACHEDATA* newItem;
    int i;
    
    /* find the library */
    while (pNextLib->nextLib) {
      if (strcmp(pNextLib->library,library)) {
        /* look further */
        pNextLib = pNextLib->nextLib;
      } else {
        /* found the library */
        break;
      }
    }
    
    if (pNextLib->nextLib==NULL) {
      /* library was not found; create new library entry */
      pNextLib->nextLib = (LIBLIST*)malloc(sizeof(LIBLIST));
      pNextLib->nextLib->nextLib = NULL;
      strcpy(pNextLib->library,library);
      pNextLib->firstData[0] = 
               (RESCACHEDATA*)malloc(NUM_DATA_BUCKETS*sizeof(RESCACHEDATA));
      for ( i=0 ; i<NUM_DATA_BUCKETS ; i++ ) {
        newItem = pNextLib->firstData[i] = pNextLib->firstData[0]+i;
        newItem->id = 0;
        newItem->str = NULL;
        newItem->nextData = NULL;
      }
    }
    
    /* use hash table to find first entry */
    pCache = pNextLib->firstData[id&DATA_BUCKET_MASK];
    while (pCache->id != 0) {
      pCache = pCache->nextData;
    }

    pCache->id = id;
    pCache->str = str;

    newItem = (RESCACHEDATA*)malloc(sizeof(RESCACHEDATA));
    newItem->id = 0;
    newItem->str = NULL;
    newItem->nextData = NULL;

    pCache->nextData = newItem;

    return str;
}

/*******************************************************************************/

NSAPI_PUBLIC
char*
XP_GetStringFromDatabase(char* strLibraryName,
                         char* strLanguage,
                         int iToken)
{
  int compResult;
  char* pRet;
  LANGLIST* pLangList = &LangListBase;
  DBFILE* pDBList;

  ACCEPT_LANGUAGE_LIST langList;
  int numLang;

  numLang = XP_AccLangList(strLanguage,langList);
  if (numLang==0) {
    strLanguage = emptyString;
  } else {
    strLanguage = langList[0];  /* temporary: just take first language */
  }
  
  /* The following code is not thread safe, and therefore needs to be locked. */
  /* In this turns out to be a performance problem, a more sophisticated      */
  /* locking scheme will be needed. For now, just lock the whole function.    */
  if (reslock) crit_enter(reslock);

  /* first find the language */
  while ((compResult=(*(pLangList->language)!=EOL_CHAR))
          && (strcmp(pLangList->language,strLanguage))) {
    pLangList = pLangList->nextLang;
  }
  if (compResult==0) {
    /* create entry for language if not found */
    pLangList->nextLang = (LANGLIST*)malloc(sizeof(LANGLIST));
    strcpy(pLangList->nextLang->language,EOL_STRING);
    strcpy(pLangList->language,strLanguage);
    GetFileForLanguage(pathDB,strLanguage,existDB);
    /* if no file is found, existDB=="" */
    /* see if file is already open */
    pDBList = &DBListBase;
    while ((pDBList->nextFile) && (strcmp(pDBList->filename,existDB))) {
      pDBList = pDBList->nextFile;
    }
    pLangList->fileForLang = pDBList;
    /* if not open, open database file, if possible */
    if (pDBList->nextFile==NULL) {
      pDBList->nextFile = (DBFILE*)malloc(sizeof(DBFILE));
      pDBList->nextFile->nextFile = NULL;
      pDBList->filename = strdup(existDB);
      pDBList->firstLib = (LIBLIST*)malloc(sizeof(LIBLIST));
      pDBList->firstLib->nextLib = NULL;
      pDBList->hresdb = NSResOpenTable(existDB, NULL);
    }
  } else {
    /* use previously created entry for language */
    pDBList = pLangList->fileForLang;
  }

#ifdef COMPILE_STRINGS_IN_MEMORY
    /* if open failed, last chance is to use default string in memory */
    if (pDBList->hresdb == 0)
      {
        if (reslock) crit_exit(reslock);
        return XP_GetStringFromMemory(strLibraryName,iToken) ;
      }
#else
    /* if open failed, there is nothing left to be done */
    if (pDBList->hresdb == 0)
      {
        if (reslock) crit_exit(reslock);
        return emptyString;  /* ERROR */
      }
#endif  
  
  /* Load string from cache if it's there */
  pRet = LoadStringFromCache(pDBList->firstLib, strLibraryName, iToken);
  if (pRet == NULL)
    {
      pRet = NSResLoadString(pDBList->hresdb, strLibraryName, iToken, 0, NULL);
      
      if (pRet)
        {
          
          SaveStringToCache(pDBList->firstLib,strLibraryName, iToken, pRet);
        }
    }
      
  if (reslock) crit_exit(reslock);

  return pRet ;
}

/*******************************************************************************/

#ifdef XP_DEBUG

void
XP_DumpCache(void)  /* debug routine */
{
  LANGLIST*     pLangList = &LangListBase;
  DBFILE*       pDBList = &DBListBase;
  LIBLIST*      pNextLib;
  RESCACHEDATA* pCache;

  int i;
  
  printf("String Cache:  Number of Buckets %d %X   Mask %X\n",
         NUM_DATA_BUCKETS,NUM_DATA_BUCKETS,DATA_BUCKET_MASK);

  while (*(pLangList->language)!=EOL_CHAR) 
    {
      puts("");
      printf("Language: %s\n",pLangList->language);
      printf("DB File:  %s\n",pLangList->fileForLang->filename);
      printf("Handle    %X\n",pLangList->fileForLang->hresdb);
      pLangList = pLangList->nextLang;
    }
  
  while (pDBList->nextFile) 
    {
      puts("");
      printf("DB File:  %s\n",pDBList->filename);
      printf("Handle    %X\n",pDBList->hresdb);

      if (pDBList->hresdb) {
        pNextLib = pDBList->firstLib;
        while (pNextLib->nextLib) {
          printf("        Library: %s\n",pNextLib->library);
          for ( i=0 ; i<NUM_DATA_BUCKETS ; i++ ) {
            pCache = pNextLib->firstData[i];
            while (pCache->id!=0) {
              printf("                %4d %4d %4d %s\n",
                     i,pCache->id,strlen(pCache->str),pCache->str);
              pCache = pCache->nextData;
            }
          }
          pNextLib = pNextLib->nextLib;
        }
      }

      pDBList = pDBList->nextFile;
    }
  
}

#endif /* XP_DEBUG */

/*******************************************************************************/
