/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

#include "libsnmp/conf.h"
#include "base/file.h"
#include "base/buffer.h"
#include "base/util.h"
#include "base/ereport.h"
#include "frame/conf.h"		/* CONF_MAXLEN */

#include <ctype.h>
#include <string.h>



#define REPLACE(x,y) do { if ((x)) FREE((x)); (x) = STRDUP((y)); } while(0)


/* ------------------------- subagt_conf_process -------------------------- */

/*
 * Parses a directive and value and takes action depending on the value.
 * Returns 0 upon success, and -1 upon error (with an error string in l).
 *
 */

static int subagt_conf_process(char *l, int ln, SubagtConfig *cfg)
{
    char *v, *d = l;

    if((*d) == '#')
	return 0;

    while((*d) && (isspace(*d))) ++d;
    if(!(*d))
	return 0;

    for(v = d; (*v) && (*v != ' '); ++v);
    if(!(*v))
	goto noval;

    *v++ = '\0';
    while((*v) && isspace(*v)) ++v;

    if(!(*v)) {
      noval:
	util_sprintf(l,"line %d: directive with no value",ln);
	return -1;
    }

    /* magnus.conf */
    if(!strcasecmp(d, "ServerName")) {
	REPLACE(cfg->snmp_host, v);
    }

    else if(!strcasecmp(d, "Address")) {
	REPLACE(cfg->snmp_addr, v);
    }

    else if(!strcasecmp(d, "Port")) {
	cfg->snmp_port = atoi(v);
    }

    else if(!strcasecmp(d, "Security")) {
	cfg->snmp_secure = (!strncasecmp(v, "On", 2)) ? 1 : 0;
    }

    /* snmp.conf */
    else if(!strcasecmp(d, "MasterHost")) {
	REPLACE(cfg->snmp_master_host, v);
    }

    else if(!strcasecmp(d, "Index")) {
	cfg->snmp_index = atoi(v);
    }

    else if(!strcasecmp(d, "Description")) {
	REPLACE(cfg->snmp_description, v);
    }

    else if(!strcasecmp(d, "Organization")) {
	REPLACE(cfg->snmp_organization, v);
    }

    else if(!strcasecmp(d, "Location")) {
	REPLACE(cfg->snmp_location, v);
    }

    else if(!strcasecmp(d, "Contact")) {
	REPLACE(cfg->snmp_contact, v);
    }

    else if(!strcasecmp(d, "SNMP")) {
	cfg->snmp_on = (!strncasecmp(v, "On", 2)) ? 1 : 0;
    }

    else if(!strcasecmp(d, "PollURL")) {
	REPLACE(cfg->snmp_poll_url, v);
    }

    else if(!strcasecmp(d, "PollTimeout")) {
	cfg->snmp_poll_timeout = atol(v);
    }

    else if(!strcasecmp(d, "PollInterval")) {
	cfg->snmp_poll_interval = atol(v);
    }

    else {
	/* ignore unknown fields for now */
    }

    return 0;
}


static char *subagt_conf_read(char *root, char *fn, SubagtConfig *cfg,
			      int is_snmp_config)
{
    filebuf_t *buf;
    SYS_FILE fd;
    int ln = 0, t, x;
    char line[CONF_MAXLEN], *err;
    char *cfn = NULL;

    if (root) {
	cfn = (char *)MALLOC(strlen(root) + strlen(fn) + 2);
	util_sprintf(cfn, "%s%c%s", root, FILE_PATHSEP, fn);
    }
    else {
	cfn = STRDUP(fn);
    }

    if((fd = system_fopenRO(cfn)) == SYS_ERROR_FD) {
	if (is_snmp_config) {
	    cfg->snmp_on = 0;
	    return NULL;
	}
	else {
	    util_sprintf(line,"error opening %s: %s", cfn, system_errmsg());
	    return STRDUP(line);
	}
    }

    buf = filebuf_open(fd, FILE_BUFFERSIZE);

    while(1) {
	++ln;
	switch((t = util_getline(buf, ln, CONF_MAXLEN, line))) {
	  case -1:
	    filebuf_close(buf);
	    return STRDUP(line);

	  case 1:
	    t = subagt_conf_process(line, ln, cfg);
	    filebuf_close(buf);
	    if(t) return STRDUP(line);

	    return NULL;

	  default:
	    if((t = subagt_conf_process(line, ln, cfg))) {
		filebuf_close(buf);
		return STRDUP(line);
	    }
	}
    }

    return NULL;
}


SubagtConfig *subagt_conf_create(char *root)
{
    SubagtConfig *cfg = (SubagtConfig *)MALLOC(sizeof(SubagtConfig));

    ZERO(cfg, sizeof(SubagtConfig));

    if (root)
	cfg->root = STRDUP(root);

    return cfg;
}


#define FREEIF(x) if(x) FREE(x)

void subagt_conf_destroy(SubagtConfig *cfg)
{
    if (cfg) {

	FREEIF(cfg->root);
	FREEIF(cfg->snmp_host);
	FREEIF(cfg->snmp_master_host);
	FREEIF(cfg->snmp_description);
	FREEIF(cfg->snmp_organization);
	FREEIF(cfg->snmp_location);
	FREEIF(cfg->snmp_contact);
	FREEIF(cfg->snmp_poll_url);
	FREE(cfg);
    }
}


SubagtConfig *subagt_conf_load(char *root, char **err)
{
    SubagtConfig *cfg = subagt_conf_create(root);

    cfg->snmp_on = 1;	/* on unless explicitly off */

    if ((*err = subagt_conf_read(root, "config/magnus.conf", cfg, 0)) ||
	(*err = subagt_conf_read(root, "config/snmp.conf", cfg, 1)) ||
	!cfg->snmp_on) {
	subagt_conf_destroy(cfg);
	return NULL;
    }

    if (!cfg->snmp_host) {
	ereport(LOG_MISCONFIG, "SNMP: ServerName not found for use by SNMP subagent");
	subagt_conf_destroy(cfg);
	return NULL;
    }
    if (!cfg->snmp_port) {
	ereport(LOG_MISCONFIG, "SNMP: Port number not found for use by SNMP subagent");
	subagt_conf_destroy(cfg);
	return NULL;
    }
    if (!cfg->snmp_master_host) {
	ereport(LOG_MISCONFIG, "SNMP: MasterHost not specified in snmp.conf");
	subagt_conf_destroy(cfg);
	return NULL;
    }

#define SET_NA(x)	if (!(x)) (x) = STRDUP("N/A")

    SET_NA(cfg->snmp_description);
    SET_NA(cfg->snmp_organization);
    SET_NA(cfg->snmp_location);
    SET_NA(cfg->snmp_contact);
    if (!cfg->snmp_poll_url)
	cfg->snmp_poll_url = STRDUP("/index.html");

    /* XXXari add other sanity checks here */

    return cfg;
}

