/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

#ifdef XP_WIN32

#include <windows.h>
#include <winsock.h>

#else /* ! XP_WIN32 */

#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <signal.h>

#endif /* ! XP_WIN32 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#ifdef SVR4
#include <sys/filio.h>
#endif

#include "netsite.h"
#include "libsnmp/httppoll.h"


#include "ssl.h"


#if 0
#define MYTRACE(x) fprintf x
#endif

#ifdef XP_WIN32
#define EWOULDBLOCK	WSAEWOULDBLOCK
#define EISCONN		WSAEISCONN
#define EINPROGRESS	WSAEINPROGRESS
#endif


fd_set poll_rd_fds;
fd_set poll_wr_fds;
int    poll_n_fds = 0;


static time_t first_up = 0;	/* subagt first verified that srv up */


/* forward prototypes */
static void http_interrupt(PollContext *cx);
static int http_connect(PollContext *cx);
static int http_send(PollContext *cx);
static int http_recv(PollContext *cx);
static int http_close(PollContext *cx);

static char *status_strings[7] = {
  "UNKNOWN       ",
  "DOWN          ",
  "HUNG (CONNECT)",
  "BROKEN        ",
  "HUNG (SERVICE)",
  "UP            ",
  "<MAX>         "
};
static char *state_strings[9] = {
  "INVALID",
  "IDLE   ",
  "CONNECT",
  "SEND   ",
  "RECV 1 ",
  "RECV 2 ",
  "ERROR  ",
  "CLOSE  ",
  "<MAX>  "
};


static u_long timediff(PollContext *cx, ConnState s1, ConnState s2)
{
    return (((cx->state_time_stamps[s2].tv_sec -
	      cx->state_time_stamps[s1].tv_sec) * 1000) +
	    ((cx->state_time_stamps[s2].tv_usec -
	      cx->state_time_stamps[s1].tv_usec + 500) / 1000));
}


static void update_statistics(PollContext *cx)
{
    long connect_delay;
    long send_delay;
    long response_delay;
    long xfer_time;
    long close_delay;
    long xaction_time;

    if (!cx)
	return;

    connect_delay  = timediff(cx, CS_CONNECT, CS_SEND);
    send_delay     = timediff(cx, CS_SEND, CS_RECV);
    response_delay = timediff(cx, CS_RECV, CS_RECV_CONT);
    xfer_time      = timediff(cx, CS_RECV, CS_CLOSE);
    close_delay    = timediff(cx, CS_CLOSE, CS_IDLE);
    xaction_time   = timediff(cx, CS_CONNECT, CS_IDLE);

    MYTRACE2((stderr, "[H[J-------------------------------\n"));
    MYTRACE2((stderr, " server status = %s\n",
	    cx->status == SERVER_STATUS_UNKNOWN      ? "unknown" :
	    cx->status == SERVER_STATUS_DOWN         ? "down" :
	    cx->status == SERVER_STATUS_HUNG_CONNECT ? "connect hangs" :
	    cx->status == SERVER_STATUS_BROKEN       ? "broken" :
	    cx->status == SERVER_STATUS_HUNG_SERVICE ? "service hangs" :
	    cx->status == SERVER_STATUS_UP           ? "up" : "invalid"));
    MYTRACE2((stderr, "\n"));
    MYTRACE2((stderr, "          down = %ld times\n", cx->poll_stats[SERVER_STATUS_DOWN]));
    MYTRACE2((stderr, "hung w/connect = %ld times\n", cx->poll_stats[SERVER_STATUS_HUNG_CONNECT]));
    MYTRACE2((stderr, "        broken = %ld times\n", cx->poll_stats[SERVER_STATUS_BROKEN]));
    MYTRACE2((stderr, "hung w/service = %ld times\n", cx->poll_stats[SERVER_STATUS_HUNG_SERVICE]));
    MYTRACE2((stderr, "            up = %ld times\n", cx->poll_stats[SERVER_STATUS_UP]));
    MYTRACE2((stderr, "\n"));
    if (cx->state_time_stamps[CS_CONNECT].tv_sec &&
	cx->state_time_stamps[CS_SEND].tv_sec) {
	MYTRACE2((stderr, " connect delay = %ld ms\n", connect_delay));

	if (cx->state_time_stamps[CS_RECV].tv_sec) {
	    MYTRACE2((stderr, " request delay = %ld ms\n", send_delay));

	    if (cx->state_time_stamps[CS_RECV_CONT].tv_sec) {
		MYTRACE2((stderr, "response delay = %ld ms\n", response_delay));

		if (cx->state_time_stamps[CS_CLOSE].tv_sec) {
		    MYTRACE2((stderr, "     xfer time = %ld ms\n", xfer_time));

		    if (cx->state_time_stamps[CS_IDLE].tv_sec) {
			MYTRACE2((stderr, "   close delay = %ld ms\n", close_delay));
			MYTRACE2((stderr, "  xaction time = %ld ms\n", xaction_time));
			MYTRACE2((stderr, " * COMPLETED *\n"));
		    }
		}
	    }
	}
    }

    MYTRACE2((stderr, "------------------------------\n"));
}


static void status_bean_counter(PollContext *cx,
				ServerStatus status)
{
    cx->poll_stats[status]++;

    MYTRACE((stderr, "OLD STATUS = %s\t\t*** [%s:%d] ***\n",
	     status_strings[cx->status], cx->host, cx->port));
    MYTRACE((stderr, "NEW STATUS = %s\t\t*** [%s:%d] ***\n",
	     status_strings[status],     cx->host, cx->port));

    if (cx->status != status) {
	cx->status = status;
	time(&cx->status_time_stamps[status]);
    }
}


static void change_state(PollContext *cx, ConnState state)
{
    MYTRACE((stderr, "[%s:%d fd %d] %s -> %s\n",
	     cx->host,
	     cx->port,
	     cx->sd,
	     state_strings[cx->state],
	     state_strings[state]));

#if defined(_SVID_GETTOD)
    gettimeofday(&cx->state_time_stamps[state]);
#else
    gettimeofday(&cx->state_time_stamps[state], 0);
#endif
    cx->state = state;
}


PollContext *http_setup_poll(char *hostname, int port, int use_security, char *url)
{
    PollContext *cx = (PollContext *)MALLOC(sizeof(PollContext));

    ZERO(cx, sizeof(PollContext));

    MYTRACE((stderr, "[%s:%d] poll setup for %s\n",
	     hostname, port, use_security ? "HTTPS" : "HTTP"));

    cx->host = STRDUP(hostname);
    cx->port = port;

    memset(&cx->sa_clean, 0, sizeof(struct sockaddr_in));
    cx->sa_clean.sin_family = AF_INET;
    cx->sa_clean.sin_port   = htons(port);

    if (isdigit(*hostname)) {		/* Numeric ip address */
	cx->sa_clean.sin_addr.s_addr = inet_addr(hostname);
    }
    else {				/* Hostname */
	struct hostent *hp = gethostbyname(hostname);
	if (!hp) {
	    FREE(cx);
	    return NULL;
	}
	memcpy(&cx->sa_clean.sin_addr, hp->h_addr, hp->h_length);
    }

    cx->use_security = use_security;
    cx->url = STRDUP(url);

    change_state(cx, CS_IDLE);

    return cx;
}


void http_setup_free(PollContext *cx)
{
    if (cx->state != CS_IDLE)
	http_interrupt(cx);

    if (cx->url)
	FREE(cx->url);

    FREE(cx);
}


static void http_interrupt(PollContext *cx)
{
    MYTRACE((stderr, "* * * INTERRUPT %d\n", cx->sd));

    switch (cx->state) {
      case CS_INVALID:
      case CS_IDLE:
	return;

      case CS_CONNECT:
	status_bean_counter(cx, SERVER_STATUS_HUNG_CONNECT);
	break;

      default:
	status_bean_counter(cx, SERVER_STATUS_HUNG_SERVICE);
    }

    http_close(cx);
}


int http_start_poll(PollContext *cx)
{
    int rv, one=1;

    MYTRACE((stderr, "* * * http_start_poll() -- state = %d\n", cx->state));

    if (cx->state != CS_IDLE)
	http_interrupt(cx);

    memset(cx->state_time_stamps, 0, CS_MAX * sizeof(struct timeval));
    cx->outind = 0;
    cx->outcnt = 0;
    cx->incnt = 0;

    memcpy(&cx->sa, &cx->sa_clean, sizeof(struct sockaddr_in));

    cx->sd = SSL_Socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    MYTRACE((stderr, "* * * Got socket %d\n", cx->sd));
    if (cx->sd < 0)
	goto bad_loser;

    if (cx->use_security) {
	rv = SSL_Enable(cx->sd, SSL_SECURITY, 1);
	if (rv < 0)
	    goto loser;
    }

    rv = SSL_Ioctl(cx->sd, FIONBIO, &one);
    if (rv == -1)
	goto loser;

    /* This will allow me to do my own buffering here, instead of have
     * SSL do busy-looping to ensure that the whole block of data gets
     * actually sent.
     */
    SSL_Enable(cx->sd, SSL_ASYNC_WRITES, 0);

    change_state(cx, CS_CONNECT);

    if (http_handle_events(cx) < 0) {	/* start the connect */
	MYTRACE((stderr, "poll start probably failed\n"));

	return 0;
    }
    else {
	MYTRACE((stderr, "poll start ok\n"));

	if (poll_n_fds <= cx->sd)
	    poll_n_fds = cx->sd + 1;

	return 1;
    }

loser:
    SSL_Close(cx->sd);

bad_loser:
    MYTRACE((stderr, "poll start failed badly (maybe)\n"));

    return 0;
}


static int http_connect(PollContext *cx)
{
    int rv;

    MYTRACE((stderr, "HTTP connect\n"));

    rv = SSL_Connect(cx->sd, (struct sockaddr *)&cx->sa,
		     sizeof(struct sockaddr_in));
    if (rv <= -1 && errno != 0 && errno != EISCONN) {
	MYTRACE((stderr, "connect returned %d, errno %d\n", rv, errno));
	if (errno != EINPROGRESS && errno != EWOULDBLOCK && errno != EINTR) {
	    MYTRACE((stderr, "connect failed\n"));

	    change_state(cx, CS_ERROR);
	    status_bean_counter(cx, SERVER_STATUS_DOWN);

	    return 1;
	}

	FD_SET(cx->sd, &poll_wr_fds);

	MYTRACE((stderr, "HTTP connect waiting\n"));
	return 0;	/* wait to connect */
    }

    FD_CLR(cx->sd, &poll_wr_fds);

   /* as long as connecting, we thought server is up
       This is also how serv mon tells server status */
    change_state(cx, CS_CLOSE);
    status_bean_counter(cx, SERVER_STATUS_UP);
    return 1; 

    /*
    change_state(cx, CS_SEND);

    MYTRACE((stderr, "HTTP connect done\n"));
    return 1;  */	/* connected -- continue */
}


static int http_send(PollContext *cx)
{
    int n;

    MYTRACE((stderr, "HTTP send\n"));

    if (!cx->outind && !cx->outcnt) {
	/* First call to this function -- construct request */

	sprintf(cx->buf,
		"HEAD %s HTTP/1.0\n"
		"User-agent: Netscape-SNMP-Agent/2.0\n"
		"\n",
		cx->url ? cx->url : "/snmp-ping");

	cx->outind = 0;
	cx->outcnt = strlen(cx->buf);
    }

    n = SSL_Write(cx->sd, cx->buf + cx->outind, cx->outcnt);

    if (n == -1) {
	if (errno != EWOULDBLOCK && errno != EINTR) {
	    MYTRACE((stderr, "HTTP send failed\n"));

	    change_state(cx, CS_ERROR);
	    status_bean_counter(cx, SERVER_STATUS_BROKEN);

	    return 1;
	}
    }
    else {
	cx->outcnt -= n;
	cx->outind += n;
    }

    if (cx->outcnt > 0) {
	FD_SET(cx->sd, &poll_wr_fds);
	MYTRACE((stderr, "HTTP send waiting\n"));
	return 0;	/* wait to become write-ready */
    }

    FD_CLR(cx->sd, &poll_wr_fds);

    change_state(cx, CS_RECV);

    MYTRACE((stderr, "HTTP send done\n"));
    return 1;	/* continue */
}


static int http_recv(PollContext *cx)
{
    int n;

    n = SSL_Read(cx->sd, cx->buf + cx->incnt, MAX_READ_SIZE - cx->incnt);

    if (n >= 0 && cx->state == CS_RECV)
	change_state(cx, CS_RECV_CONT);

    if (n == 0) {	/* socket closed, done */
	MYTRACE((stderr, "HTTP recv done\n"));

	change_state(cx, CS_CLOSE);
	status_bean_counter(cx, SERVER_STATUS_UP);
	if (!first_up)
	    first_up = cx->status_time_stamps[SERVER_STATUS_UP];

	return 1;
    }
    else if (n <= -1) {
	if (errno != EWOULDBLOCK && errno != EINTR && errno != 0) {

	    MYTRACE((stderr, "HTTP recv failed\n"));

	    change_state(cx, CS_ERROR);
	    status_bean_counter(cx, SERVER_STATUS_BROKEN);

	    return 1;
	}
	else {
	    MYTRACE((stderr, "%s\n",
		     errno == EWOULDBLOCK ? "EWOULDBLOCK" : "EINTR"));
	    if (errno == EINTR)
		return http_recv(cx);
	}
    }
    else {
	cx->incnt += n;
    }

    if (cx->incnt >= SUFFICIENT_READ_SIZE) {
	MYTRACE((stderr, "HTTP recv done enough (interrupting xfer)\n"));

	change_state(cx, CS_CLOSE);
	status_bean_counter(cx, SERVER_STATUS_UP);
	if (!first_up)
	    first_up = cx->status_time_stamps[SERVER_STATUS_UP];

	return 1;	/* continue */
    }

    FD_SET(cx->sd, &poll_rd_fds);

    MYTRACE((stderr, "HTTP recv waiting, state = %d\n", cx->state));
    return 0;	/* wait to become read-ready */
}


static int http_close(PollContext *cx)
{
    MYTRACE((stderr, "HTTP close %d\n", cx->sd));

    FD_CLR(cx->sd, &poll_rd_fds);
    FD_CLR(cx->sd, &poll_wr_fds);

    SSL_Close(cx->sd);

    change_state(cx, CS_IDLE);

    update_statistics(cx);

    return 1;
}


int http_handle_events(PollContext *cx)
{
    int rv;

    if (!cx)
	return -1;

    do {
	MYTRACE((stderr, "HTTP event\n"));

	switch(cx->state) {
	  case CS_CONNECT:
	    rv = http_connect(cx);
	    break;
	  case CS_SEND:
	    rv = http_send(cx);
	    break;
	  case CS_RECV:		/* waiting for first chuck */
	  case CS_RECV_CONT:	/* waiting for the rest of it */
	    rv = http_recv(cx);
	    break;
	  case CS_ERROR:
	    MYTRACE((stderr, " *** ERROR STATE ***\n"));
	    change_state(cx, CS_CLOSE);
	    rv = 1;
	    break;
	  case CS_CLOSE:
	    rv = http_close(cx);
	    rv = -1;
	    break;
	  default:
	    rv = -1;
	}
    } while (rv > 0);

    MYTRACE((stderr, "HTTP events done, state = %d\n", cx->state));

    return rv;
}

