/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * Description (acl.c)
 *
 *	This module contains routines involved in relating ACLs to
 *	server objects.
 */

#include <sec.h>
#include <ldap.h>
#include <ldap_ssl.h>

#include "netsite.h"
#include "safs/acl.h"
#include "safs/auth.h"
#include "base/shexp.h"
#include "frame/log.h"
#include "base/file.h"
#include "base/util.h"  /* util_uri_is_evil */
#include <base/pblock.h>
#include <base/crit.h>
#include <base/daemon.h>

#include "frame/protocol.h"
#include "frame/conf.h"
#include "frame/func.h"
#include "frame/http.h"

#include <safs/init.h>
#ifdef CLIENT_AUTH
#include <safs/clauth.h>
#endif /* CLIENT_AUTH */
#include <ldaputil/init.h>
#include <ldaputil/certmap.h>
#include <ldaputil/ldapdb.h>
#include <libaccess/acl.h>
#include <libaccess/aclproto.h>
#include <libaccess/las.h>
#include <libaccess/nserror.h>
#include <libaccess/nsauth.h>   /* ClAuth_t */
#include <libaccess/aclstruct.h>
#include <libaccess/acleval.h>
#include "libaccess/nsadb.h"
#include "libaccess/nsamgmt.h"
#include <libaccess/ldapacl.h>
#include <sys/stat.h>
#include <ctype.h>

#define BIG_LINE 1024

#ifdef MCC_ADMSERV
#define DEFAULT_ONLY 1
#else
#define DEFAULT_ONLY 0
#endif

/*
 * Description (pcheck_check_acl30)
 *
 *	This function adds an ACL to an intermediate ACL list.
 *
 * Directive format:
 *
 *	PathCheck fn=check-acl acl=aclname path=shexp bong-file=filename
 *
 * Arguments:
 *
 *	pb		- PathCheck directive parameter block
 *	sn		- session pointer
 *	rq		- request pointer
 *
 * Returns:
 *
 *	Returns REQ_NOACTION or REQ_ABORTED.  REQ_ABORTED is
 *	returned only if an ACL denies access with an "always"
 *	directive.  The rq->aclstate value is updated if any
 *	any of the ACLs apply to the request.
 */

int pcheck_check_acl30(pblock * pb, Session * sn, Request * rq)
{
    char * aclname;		/* ACL name */
    ACLHandle_t * acl;		/* named ACL */
    char *check_path;
    NSErr_t errp;

    /* Is there a path shexp to match? */
    check_path = pblock_findval("path", pb);

    if (check_path != 0) {
	pb_param *path = pblock_find("path", rq->vars);

	/* Yes, check request path against it */
	if (WILDPAT_CMP(path->value, check_path)) {

	    /* Request path does not match pattern, so do nothing */
	    return REQ_NOACTION;
	}
    }

    /* Get the ACL name */
    aclname = pblock_findval("acl", pb);
    if (aclname == 0) {
	return REQ_NOACTION;
    }

    /* Else get it from the ACL database */
    acl = ACL_ListFind((ACLListHandle_t *)acl_root_30, aclname, ACL_CASE_INSENSITIVE);
    if (!acl) {
	/* XXX server configuration error */
        log_error(LOG_MISCONFIG, "check-acl", sn, rq, 
                  "ACL name %s not defined", aclname);
        protocol_status(sn, rq, PROTOCOL_SERVER_ERROR, NULL);
	return REQ_ABORTED;
    }

    /* Add it to the ACL List */
    if (!rq->acllist)
	rq->acllist = ACL_ListNew();
    ACL_ListAppend(&errp, rq->acllist, acl, 0);

    return REQ_NOACTION;

}

/*
 * Description (pcheck_acl_state)
 *
 *	This function processes the results of ACL evaluation, stored in
 *	the aclstate field of the specified Request structure.  Most of
 *	the work involves handling an ACL request for authentication.  The
 *	authentication realm is specified by the clauth.cla_realm field
 *	in the Session structure.  The aclstate field in the Request
 *	structure indicates whether the request is immediate (ACD_ALWAYS)
 *	or not.  The function first checks the request variables to see
 *	if information is available from a previous call, and whether
 *	the authentication was in the specified realm.  If not, it checks
 *	for an "authorization" request header, and attempts to validate
 *	the indicated user and password in the specified realm, creating
 *	request variables to indicate a successful result.  If there is
 *	no "authorization" request header, or if the result of processing
 *	the user and password in the current realm was unsuccessful, then
 *	the request is aborted with a request for authentication if
 *	ACD_ALWAYS is set in aclstate.
 *
 *	This function is called by pcheck_check_acl() whenever an
 *	individual ACL applies to a request.  It is also called from
 *	servact_pathchecks() after all PathCheck directives have been
 *	processed.  The ACD_ALWAYS flag is set prior to this call.
 *
 * Arguments:
 *
 *	pb		- PathCheck directive parameter block
 *	sn		- session pointer
 *	rq		- request pointer
 *
 * Returns:
 *
 *	Returns REQ_PROCEED, REQ_NOACTION, or REQ_ABORTED.  REQ_ABORTED
 *	is returned to request immediate authentication, or to deny
 *	access.
 */

int pcheck_acl_state(pblock * pb, Session * sn, Request * rq)
{
    char * user;
    char * authdb;
    char * authtype;
    Realm_t * rlm;
    ClAuth_t *cla;
    pblock * npb;
    int decision;
    int rv;

    if (acl_root_30) {
	/* xxxx return error */
	log_error(LOG_SECURITY, "acl-state", sn, rq, 
		  "3.0 error, ACL %s directive %d",
		  rq->aclname, rq->acldirno);
        return REQ_NOACTION;
    }


    /* Get the current state of the ACL decision */
    decision = rq->aclstate;

    if (decision == 0) {
	/* No applicable ACL information */
	return REQ_NOACTION;
    }

    /* Does the ACL indicate that authentication is needed? */
    if ((decision & ACD_ACTION) == ACD_AUTH) {
        cla = (ClAuth_t *) sn->clauth;

	/* Get pointer to realm for requested authentication */
	rlm = cla->cla_realm;

	/* Is there an authenticated user name available? */
	user = pblock_findval("auth-user", rq->vars);
	if (user == 0) {
	    goto req_auth;
	}

	/* Was the right authentication database used? */
	authdb = pblock_findval("auth-db", rq->vars);
	if ((authdb == 0) || strcmp(authdb, rlm->rlm_dbname)) {
	    /* auth-db not present or doesn't match */
	    goto req_auth;
	}

	/* Was the right authentication method used? */
	authtype = pblock_findval("auth-type", rq->vars);
	if ((authtype == 0) ||
	    ((rlm->rlm_ameth == AUTH_METHOD_BASIC) && strcasecmp(authtype,
								 "basic")) ||
	    ((rlm->rlm_ameth == AUTH_METHOD_SSL) && strcasecmp(authtype,
							       "ssl"))) {
	    /* auth-type not present or doesn't match */
	    goto req_auth;
	}

	/* There should be a UserObj_t pointer here */
	if (cla->cla_uoptr == 0) {
	    /* XXX repeat lookup of user name? */
	    return REQ_ABORTED;
	}

	/* Authentication has succeeded */
	return REQ_PROCEED;
    }

    /* Is this an immediate decision to deny access? */
    if (decision == (ACD_ALWAYS|ACD_DENY)) {
	char *bong_file;
	pb_param *path = pblock_find("path", rq->vars);
#ifdef MCC_PROXY
	pb_param *escp = pblock_find("escaped", rq->vars);
#endif

	/* Yes, abort the request now */
	log_error(LOG_SECURITY, "acl-state", sn, rq, 
		  "access of %s denied by ACL %s directive %d",
		  path->value, rq->aclname, rq->acldirno);

	/* Check for a bong-file parameter on the PathCheck fn=check-acl */
	if ((rq->aclpb != 0) &&
	    (bong_file = pblock_findval("bong-file", rq->aclpb)) != 0) {
	    struct stat finfo;

	    if (!stat(bong_file, &finfo)) {
		/*
		 * Clear the ACL state for the request so it doesn't get
		 * applied to the bong file.
		 */
		rq->aclstate = 0;
		rq->acldirno = 0;
		rq->aclname = 0;
		rq->aclpb = 0;

		FREE(path->value);
		path->value = STRDUP(bong_file);
#ifdef MCC_PROXY
                if( escp ) {
                    FREE(escp->value);
                    escp->value = STRDUP(bong_file);
                }

#endif

		return REQ_PROCEED;
	    }

	    log_error(LOG_MISCONFIG, "acl-state", sn, rq,
		      "access denied response file %s is unreadable",
		      bong_file);
	}
#ifdef MCC_PROXY
	protocol_status(sn, rq, PROTOCOL_FORBIDDEN, NULL);
#else
	protocol_status(sn, rq, PROTOCOL_NOT_FOUND, NULL);
#endif
	return REQ_ABORTED;
    }

    /* Must be a decision to allow access */
    return (decision & ACD_ALWAYS) ? REQ_PROCEED : REQ_NOACTION;

  req_auth:

    /* Remove any request variables created from previous authentication */
    pblock_remove("auth-user", rq->vars);
    pblock_remove("auth-type", rq->vars);
    pblock_remove("auth-db", rq->vars);
#if defined(XP_WIN32) || defined(MCC_ADMSERV)
    /* MLM - the admin server needs this password, so I'm adding it back for
     *       admin server as well.  */
    pblock_remove("auth-password", rq->vars);
#endif /* XP_WIN32 */

    /* Set up fake AuthTrans directive parameters for current realm */
    npb = pblock_create(4);

    if (rlm->rlm_dbname != 0) {
	pblock_nvinsert("userdb", rlm->rlm_dbname, npb);
	pblock_nvinsert("userfn", "dbm-userdb", npb);
    }

    if (rlm->rlm_ameth == AUTH_METHOD_BASIC) {

	/* Do basic authentication */
	pblock_nvinsert("auth-type", "basic", npb);

	rv = auth_basic(npb, sn, rq);

	pblock_free(npb);

	/* Did authentication succeed? */
	if (rv == REQ_NOACTION) {

	    /* No, is it required immediately? */
	    if (decision & ACD_ALWAYS) {
		char buf[256];

		/* Yes, request authentication for current realm */
		if (rlm == 0) {
		    log_error(LOG_MISCONFIG, "acl-state", sn, rq, 
			      "missing realm");
		    return REQ_ABORTED;
		}

		util_sprintf(buf, "basic realm=\"%s\"",
			     (rlm->rlm_prompt == 0) ? "unspecified"
						    : rlm->rlm_prompt);

		pblock_nvinsert("WWW-authenticate", buf, rq->srvhdrs);
		protocol_status(sn, rq, PROTOCOL_UNAUTHORIZED, NULL);

		return REQ_ABORTED;
	    }
	}
    }
    else {
#ifdef CLIENT_AUTH
	/* Do SSL client authentication */
	pblock_nvinsert("realm",
			(rlm && rlm->rlm_prompt) ? rlm->rlm_prompt
						 : "unspecified", npb);

	rv = CA_GetUserID(npb, sn, rq);
#else
	log_error(LOG_MISCONFIG, "acl-state", sn, rq, 
		  "SSL client authentication not supported");
	rv = REQ_ABORTED;
#endif /* CLIENT_AUTH */

	pblock_free(npb);
    }

    return rv;
}

/*
 * Description (pcheck_check_acl)
 *
 *	This function evaluates a list of ACLs attached to an object
 *	during PathCheck processing.  The result of the evaluation
 *	may be an immediate decision to allow or deny access, or to
 *	request authentication, or it may be a tentative decision
 *	which is subject to further PathCheck processing and possibly
 *	other ACLs.  The state of the decision is maintained in the
 *	Request structure.
 *
 *	If authentication is requested, this function will attempt
 *	to process authentication information, including an attempt
 *	to validate and translate raw authentication information
 *	supplied in the request as would normally be done during
 *	AuthTrans processing.  AuthTrans is not used because the
 *	decision to request authentication depends on how the ACLs
 *	specified in this PathCheck directive relate to the client
 *	host.  [An alternative would be to evaluate the ACLs at
 *	both AuthTrans and PathCheck time, but that seemed less
 *	desirable.]
 *
 * Directive format:
 *
 *	PathCheck fn=check-acl acl=aclname path=shexp bong-file=filename
 *
 * Arguments:
 *
 *	pb		- PathCheck directive parameter block
 *	sn		- session pointer
 *	rq		- request pointer
 *
 * Returns:
 *
 *	Returns REQ_NOACTION or REQ_ABORTED.  REQ_ABORTED is
 *	returned only if an ACL denies access with an "always"
 *	directive.  The rq->aclstate value is updated if any
 *	any of the ACLs apply to the request.
 */

int pcheck_check_acl(pblock * pb, Session * sn, Request * rq)
{
    ACL_t * acl;		/* current ACL pointer */
    RightDef_t * rdp;		/* required access right pointer */
    char * acllist;		/* ACL name */
    char * right;		/* access right name */
    int adn;			/* applicable ACL directive number */
    int rv;			/* result value */
    char *check_path;
    ClAuth_t *cla;

    if (acl_root_30)
	return(pcheck_check_acl30(pb, sn, rq));

    /* Has a final decision already been made by a previous ACL? */
    if (rq->aclstate & ACD_ALWAYS) {

	/* Yes, do nothing */
	return REQ_NOACTION;
    }

    /* Is there a path shexp to match? */
    check_path = pblock_findval("path", pb);

    if (check_path != 0) {
	pb_param *path = pblock_find("path", rq->vars);

	/* Yes, check request path against it */
	if (WILDPAT_CMP(path->value, check_path)) {

	    /* Request path does not match pattern, so do nothing */
	    return REQ_NOACTION;
	}
    }

    /* Get the ACL name */
    acllist = pblock_findval("acl", pb);

    /* Make sure there is an ACL root structure and an ACL name */
    if ((acl_root == 0) || (acllist == 0)) {
	return REQ_NOACTION;
    }

    /* Look up the ACL by name */
    rv = symTableFindSym( ((ACContext_t *) acl_root)->acc_stp, 
                         acllist, ACLSYMACL, (void **)&acl);
    if (rv < 0) {
	/* XXX server configuration error */
        log_error(LOG_MISCONFIG, "check-acl", sn, rq, 
                  "ACL name %s not defined", acllist);
        protocol_status(sn, rq, PROTOCOL_SERVER_ERROR, NULL);
	return REQ_ABORTED;
    }

    /* Look up the needed access right by name */
    right = pblock_findval("method", rq->reqpb);

    rv = symTableFindSym(  ((ACContext_t *) acl_root)->acc_stp, 
                         right, ACLSYMRIGHT, (void **)&rdp);
    if (rv < 0) {
	/* XXX server configuration error */
        log_error(LOG_MISCONFIG, "check-acl", sn, rq, 
                  "access right name %s not defined", right);
        protocol_status(sn, rq, PROTOCOL_SERVER_ERROR, NULL);
	return REQ_ABORTED;
    }

    /* Get the client DNS name if we don't have it yet */
    cla = (ClAuth_t *) sn->clauth;
    if (cla->cla_dns == 0) {
	(void)session_maxdns(sn);
    }

    /* Loop while authentication is required */
    for (;;) {

	/* Evaluate access against the ACL */
	rv = aclEvaluate(acl, rdp->rd_id, cla, &adn);

	if (rv != 0) {

	    /* Store latest ACL decision in the Request structure */
	    rq->aclstate = rv;
	    rq->acldirno = adn;
	    rq->aclname = acllist;
	    rq->aclpb = pb;

	    /* Process latest aclstate value */
	    rv = pcheck_acl_state(pb, sn, rq);

	    /* If authentication succeeded, evaluate the ACL again */
	    if (rv == REQ_PROCEED) {
		if ((rq->aclstate & ACD_ACTION) == ACD_AUTH) {
		    rq->aclstate = 0;
		    continue;
		}

		/* Otherwise we're not ready for REQ_PROCEED yet */
		rv = REQ_NOACTION;
	    }

	    /* Otherwise return the value from pcheck_acl_state() */
	    break;
	}

	/* Take no action yet */
	rv = REQ_NOACTION;
	break;
    }

    return rv;
}



NSAPI_PUBLIC void
ACL_InitSafs(void)
{
    pblock *pb = pblock_create(4);

    init_acl_modules(pb, 0, 0);
    return;
}

static int dbEmptyCB (NSErr_t * errp, void *authdb, void *parg,
		      UserObj_t *uoptr)
{
    int *empty = (int *)parg;

    /* If this function gets called, the authdb is not empty */
    *empty = 0;
    return ADBF_STOPENUM;
}

static CRITICAL ldb_crit = NULL; /* Controls crit sections for LDAP calls */

static void ldb_crit_enter (void)
{
    crit_enter(ldb_crit);
}

static void ldb_crit_exit (void)
{
    crit_exit(ldb_crit);
}

static void ldb_crit_init (void)
{
    ldb_crit = crit_init();
}

static 	SECCertDBHandle ldapssl_authdb;

NSAPI_PUBLIC void
ACL_InitSafsPostMagnus(void)
{
    if (server_root) {
	/* See if there are any entries in the 2.0 default database.  
	 * If there are any entries there then set that as the default
	 * database.  That means the user has not migrated 2.0 databases to
	 * LDAP databases. 
	 */
	char *sroot = server_root;
	char db_map_file[BIG_LINE];
	char default_20[BIG_LINE];
	char certmap_file[BIG_LINE];
	void *authdb;
	int empty = 1;
	int rv;

	sprintf(db_map_file, "%s%cshared%cconfig%cdbswitch.conf",
		sroot, FILE_PATHSEP, FILE_PATHSEP, FILE_PATHSEP);

	rv = ACL_ReadDbMapFile(NULL, db_map_file, DEFAULT_ONLY);

	if (rv != LAS_EVAL_TRUE) return; /* Error */

	sprintf(certmap_file, "%s%cshared%cconfig%ccertmap.conf", 
		sroot, FILE_PATHSEP, FILE_PATHSEP, FILE_PATHSEP);

	rv = ldaputil_init(certmap_file);
	
	/* Set the critical section functions */
	ldapu_crit_enter_fn = ldb_crit_enter;
	ldapu_crit_exit_fn = ldb_crit_exit;
	ldb_crit_init();

	if (rv != LDAPU_SUCCESS) return; /* Error */

	sprintf(default_20, "%s%cauthdb%cdefault", 
		sroot, FILE_PATHSEP, FILE_PATHSEP);

	rv = nsadbOpen(NULL, default_20, 0, &authdb);

	if (rv >= 0) {

	    rv = nsadbEnumerateUsers(NULL, authdb, (void *)&empty, dbEmptyCB);
	    nsadbClose(authdb, 0);

	    if (!empty && ACL_RegisterDbName(NULL, DBTYPE_20, "default_20",
					     default_20, NULL))
	    {
		ACL_SetDefaultDatabase(NULL, "default_20");
	    }
	}
    }

    if (ACL_GetDefaultDbtype(NULL)) {

	if (ACL_GetDefaultMethod(NULL) == NULL) {
	    /*  default method is not set thro' magnus.conf */
	    ACL_SetDefaultMethod((NSErr_t *)NULL, AUTHTYPE_BASIC);
	}

	ACL_ListPostParseForAuth((ACLListHandle_t *)acl_root_30,
				 ACL_GetDefaultDbtype(NULL));
    }

    if (!security_active && ACL_NeedLDAPOverSSL()) {
	/* Initialize the cert database */
	char certdb[BIG_LINE];
	void *authdb;

	if (!secure_certfn || !*secure_certfn) {
	    /* Certificate file is required if using LDAP over SSL */
	    fprintf(stderr, "Certificate file is required if using LDAP over SSL\n");
	    return;
	}

	if (servssl_certinit(secure_certfn, &authdb) < 0) {
	    fprintf(stderr, "Initialization for LDAP over SSL failed\n");
	    return;
	}
    }

    return;
}
