/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * addlog.c: Handles request-level logging
 * 
 * There are the beginnings of support for multiple logfiles here...
 *
 * Rob McCool
 */

#include "netsite.h"
#include "safs/addlog.h"
#include "base/session.h"
#include "frame/req.h"
#include "base/file.h"         /* system_fopenWA, system_write_atomic */
#include "frame/conf.h"        /* user_pw */
#include "frame/log.h"         /* log_error */
#include "base/util.h"         /* util_sprintf */

#include <time.h>
#ifdef XP_UNIX
#include <signal.h>
#include <unistd.h>
#endif /* XP_UNIX */


#include "safs/ntlogbuf.h"

/* ------------------------ Common Logfile Format ------------------------- */


struct clog {
    SYS_FILE fd;
    char *name;
    /* Pathname of the logfile */
    char *CompleteName;
    LOGBUF_T *logbuf;
    struct clog *next;
};

struct clog *cl = NULL;

#ifdef XP_WIN32
void clf_rotate(void)
{
    struct clog *p;
    char err[256], *errmsg;
    long Result;
    char TmpFile[1024], tmpbuf[128];
    time_t current_time;
    struct tm tms;
    struct tm *tmsp;

    time(&current_time);
    tmsp = system_localtime(&current_time, &tms);

    (void)util_strftime(tmpbuf, "%d%h-%I%p", tmsp);

    p = cl;
    while(p) {
    	if(system_flock(p->fd) == IO_ERROR)
        	return;
        system_fclose(p->fd);
    	sprintf(TmpFile, "%s.%s", p->CompleteName, tmpbuf);
    	if(! MoveFileEx(p->CompleteName, TmpFile, MOVEFILE_REPLACE_EXISTING))
    		Result = GetLastError();
      	if((p->fd = system_fopenWA(p->CompleteName)) == SYS_ERROR_FD) {
     		util_sprintf(err, "can't open logfile %s (%s)",
      			p->CompleteName, system_errmsg());
        }
     	p = p->next;
     }
     /* Now rotate the errors log */
	if(system_flock(ereport_getfd()) == IO_ERROR)
    	return;
    system_fclose(ereport_getfd());
	sprintf(TmpFile, "%s.%s", master_error_log, tmpbuf);
	if(! MoveFileEx(master_error_log, TmpFile, MOVEFILE_REPLACE_EXISTING))
		Result = GetLastError();
    if((errmsg = ereport_init(master_error_log, NULL, NULL, BUILD_NUM))) {
 		util_sprintf(err, "can't open logfile %s (%s)",
  			master_error_log, system_errmsg());
    }
}
#endif


void clf_terminate(void *data)
{
    struct clog *p;

    while(cl) {
        FREE(cl->name);
    	FREE(cl->CompleteName);
        logbuf_close(cl->logbuf);
        system_fclose(cl->fd);

        p = cl;
        cl = cl->next;
        FREE(p);
    }
}


int clf_init(pblock *pb, Session *sn, Request *rq)
{
    struct pb_entry *p;
    struct clog *c;
    register int x;
    register char *t;
    char err[MAGNUS_ERROR_LEN];
#ifdef XP_UNIX
    int root = (!geteuid());
#endif /* XP_UNIX*/

    if(logbuf_init(pb) == REQ_ABORTED)
        return REQ_ABORTED;
    param_free(pblock_remove("server-version", pb));

    cl = NULL;

    for(x = 0; x < pb->hsize; x++) {
        p = pb->ht[x];
        while(p) {
            t = p->param->name;

            if(strcmp(t, "fn") && (!(LOGBUF_DIRECTIVE(t)))) {
                c = (struct clog *) MALLOC(sizeof(struct clog));
                c->name = STRDUP(t);
                if((c->fd = system_fopenWA(p->param->value)) == SYS_ERROR_FD) {
                    util_snprintf(err, MAGNUS_ERROR_LEN, "can't open logfile %s (%s)",
                                 p->param->value, system_errmsg());
                    pblock_nvinsert("error", err, pb);
                    return REQ_ABORTED;
                }
                c->CompleteName = STRDUP(p->param->value);
                c->logbuf = logbuf_new(pb, c->fd);
                c->next = cl;
                cl = c;

                /* 
                 * If we're running as root, make sure that when we restart 
                 * we can re-open the log files.
                 */
#ifdef XP_UNIX
                if(root && user_pw)
                    chown(p->param->value, user_pw->pw_uid, user_pw->pw_gid);
#endif /* XP_UNIX */
            }
            p = p->next;
        }
    }

    return REQ_PROCEED;
}


#ifdef NET_SSL
int _clf_record(char *rql, char *user, char *bytes, char *status, 
    char *rhost, char *logname, char *cipher, char *user_dn)
#else
int _clf_record(char *rql, char *user, char *bytes, char *status, 
    char *rhost, char *logname)
#endif
{
    char tbuf[28], sign, *buf;
    struct clog *p;

    time_t tnl;
    long tz;
    struct tm *tmsp, tms;
    int len;

    for(p = cl; p; p = p->next)
        if(!strcmp(p->name, logname))
            break;

    if(!p) {
        ereport(LOG_MISCONFIG, "common-log: cannot find log named %s", logname);
        return REQ_ABORTED;
    }

    if(!bytes)  bytes = "-";
    if(!user)   user = "-";
    if(!status) 
        status = "-";
    else
        status[3] = '\0';

    tnl = time(NULL);
    tmsp = system_localtime(&tnl, &tms);

#ifdef BSD_TIME
    tz = tmsp->tm_gmtoff;
#else
    tz = - timezone;
    if(tmsp->tm_isdst)
        tz += 3600;
#endif

    sign = (tz >= 0 ? '+' : '-');
    if(tz < 0) 
        tz = -tz;

    len = strlen(rhost) + 1 + 2 + strlen(user) + 1 + 28 + 1 +
                          strlen(rql) + 2 + 1 + strlen(status) + 1 +
                          strlen(bytes) + 2 + 1;

#ifdef NET_SSL
    len += strlen(cipher?cipher:"-") + strlen(user_dn?user_dn:"-") + 6;
#endif

    buf = (char *) MALLOC(len);

    (void)util_strftime(tbuf, "%d/%b/%Y:%H:%M:%S", tmsp);


#ifdef NET_SSL
    /* Can't use util-sprintf because of 02 */
    sprintf(buf, "%s %s %s [%s %c%02d%02d] \"%s\" %s %s %s { %s }%s",
            rhost, "-", user, tbuf, sign, (int)(tz/3600), 
            (int)(tz%3600), rql, status, bytes, 
	    cipher ? cipher : "-", 
	    user_dn ? user_dn : "-", 
	    ENDLINE);
#else
    sprintf(buf, "%s %s %s [%s %c%02d%02d] \"%s\" %s %s%s",
            rhost, "-", user, tbuf, sign, (int)(tz/3600), 
            (int)(tz%3600), rql, status, bytes, ENDLINE);
#endif

    len = strlen(buf); /* grumble mumble stupid BSD */

    if(status != "-")
        status[3] = ' ';

    logbuf_write(p->logbuf, p->fd, buf, len);
    FREE(buf);

    return REQ_PROCEED;
}

int clf_record(pblock *pb, Session *sn, Request *rq)
{
    pb_param *rqlp = pblock_find("clf-request", rq->reqpb);
    char *rql;
    char *user = pblock_findval("auth-user", rq->vars);
    char *bytes = pblock_findval("content-length", rq->srvhdrs);
    char *status = pblock_findval("status", rq->srvhdrs);
    char *iponly = pblock_findval("iponly", pb);
    char *logname = pblock_findval("name", pb);
    char *host;
#ifdef NET_SSL
    char *cipher = pblock_findval("cipher", sn->client);
    char *user_dn = pblock_findval("user_dn", sn->client);
#endif

    if((!rqlp) || (!cl)) {
        return REQ_NOACTION;
    }
    else
        rql = rqlp->value;

    if(!logname)
        logname = "global";

    if(iponly || (!(host = session_dns(sn))))
        host = pblock_findval("ip", sn->client);

#ifdef NET_SSL
    return _clf_record(rql, user, bytes, status, host, logname, cipher, user_dn);
#else
    return _clf_record(rql, user, bytes, status, host, logname);
#endif
}

/* clf_record_accel()
 * Interface to logging routines for NT front-end accelerator.
 */
int clf_record_accel(char *logname, char *host, char *rql, char *user, 
	char *bytes, char *status)
{
    if (!logname)
        logname = "global";

    if (!host)
        host = "error getting host";

#ifdef NET_SSL
    return _clf_record(rql, user, bytes, status, host, logname, NULL, NULL);
#else
    return _clf_record(rql, user, bytes, status, host, logname);
#endif
}

/* ---------------------------- log_useragent ----------------------------- */


int record_useragent(pblock *pb, Session *sn, Request *rq)
{
    char *logname = pblock_findval("name", pb), *rhost, *ua, buf[512];
    struct clog *p;

    if(!logname)
        logname = "global";

    for(p = cl; p; p = p->next)
        if(!strcmp(p->name, logname))
            break;

    if(!p) {
        log_error(LOG_MISCONFIG, "record-useragent", sn, rq, 
                  "cannot find log named %s", logname);
        return REQ_ABORTED;
    }

    rhost = pblock_findval("ip", sn->client);
    if(request_header("user-agent", &ua, sn, rq) == REQ_ABORTED)
        return REQ_ABORTED;
    if(!ua)
        ua = "(none given)";

    system_fwrite_atomic(p->fd, buf, util_snprintf(buf, sizeof(buf),
                                                   "%s %s%s", rhost, ua,
                                                   ENDLINE));
    return REQ_PROCEED;
}


/* ---------------------------- record_keysize ----------------------------- */

#ifdef NET_SSL

int record_keysize(pblock *pb, Session *sn, Request *rq)
{
    char *logname = pblock_findval("name", pb), *rhost, buf[512];
    struct clog *p;

    char *keySize = pblock_findval("keysize", sn->client);

    char tbuf[28], sign;
    time_t tnl;
    long tz;
    struct tm *t, tms;

    if(!security_active)  
        return REQ_NOACTION;

    tnl = time(NULL);
    t = system_localtime(&tnl, &tms);

#ifdef BSD_TIME
    tz = t->tm_gmtoff;
#else
    tz = - timezone;
    if(t->tm_isdst)
        tz += 3600;
#endif

    sign = (tz >= 0 ? '+' : '-');
    if(tz < 0) 
        tz = -tz;

    (void)util_strftime(tbuf, "%d/%b/%Y:%H:%M:%S", t);

    if(!logname)
        logname = "global";

    for(p = cl; p; p = p->next)
        if(!strcmp(p->name, logname))
            break;

    if(!p) {
        log_error(LOG_MISCONFIG, "record-keysize", sn, rq, 
                  "cannot find log named %s", logname);
        return REQ_ABORTED;
    }

    rhost = pblock_findval("ip", sn->client);
    if(keySize)  {
	sprintf(buf, "%s: [%s %c%02d%02d] using keysize %s%s", rhost, tbuf, sign, (int)(tz/3600), (int)(tz%3600), keySize, ENDLINE);
	system_fwrite_atomic(p->fd, buf, strlen(buf));
    }  else  {
        log_error(LOG_WARN, "record-keysize", sn, rq, 
                  "Cannot get security info for session");
    }

    return REQ_PROCEED;
}

#endif
