/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * auth.c: Handle AuthTrans related things. Currently, only stupid HTTP
 * basic authentication.
 * 
 * Rob McCool
 */


#include "netsite.h"
/* removed for ns security integration
#include "xp_mem.h"
*/
#include "nspr.h"
#include "private/pprio.h"
/* Removed for new nspr21 and ns security integration
#include "sslio/sslio.h"
*/
extern "C" {
#include "key.h"
#include "cert.h"
#include "ssl.h"
#include "base64.h"
}

#include "base/daemon.h"
#include "base/session.h"
#include "frame/req.h"

#include "frame/protocol.h"
#include "frame/log.h"
#include "frame/func.h"
#include "frame/http.h"
#include "frame/conf.h"

#include "base/buffer.h"
#include "base/util.h"

#include "libaccess/nsuser.h"
#include "libaccess/nsgroup.h"
#include "libaccess/nsamgmt.h"
#include "libaccess/acl.h"
#include "libaccess/aclproto.h"
#include "libaccess/las.h"
#include "libaccess/authdb.h"

#include "libadmin/libadmin.h"

#include "safs/auth.h"
#include "safs/init.h"

/* Libcrypt in NT is not thread-safe and needs a critical section
 * wrapper */

#ifdef XP_UNIX
/* XXXrobm This NSPR section can probably be shared by NT and UNIX but I 
   don't want to rock the boat right now. */
#ifdef USE_NSPR
#include "base/crit.h"
static CRITICAL auth_crit = NULL;
#define ENTER_AUTH_CRITICAL_SECTION crit_enter(auth_crit)
#define LEAVE_AUTH_CRITICAL_SECTION crit_exit(auth_crit)
extern "C" void auth_init_crits(void) { auth_crit = crit_init(); }

#else
#define ENTER_AUTH_CRITICAL_SECTION
#define LEAVE_AUTH_CRITICAL_SECTION
#endif
#else /* XP_WIN32 */

static CRITICAL_SECTION authCriticalSection;
static BOOLEAN InitCriticalSection = TRUE;

#define ENTER_AUTH_CRITICAL_SECTION\
    if (InitCriticalSection) {                    \
        InitializeCriticalSection(&authCriticalSection);\
        InitCriticalSection = FALSE;                     \
    }                                                   \
    EnterCriticalSection(&authCriticalSection);

#define LEAVE_AUTH_CRITICAL_SECTION LeaveCriticalSection(&authCriticalSection)

#endif /* XP_WIN32 */

/*
 * _cmpCryptPasswd is called to protect crypt() in a threaded
 * world. The result of crypt() is only needed to do the
 * string compare, so that's the limit of the scope of protection.
 * 
 * This routine is called from clauth.c as well as the two
 * direct references from within this file.
 */
int _cmpCryptPasswd(const char *encrypted, const char* cleartext)
{
    int rv;
    ENTER_AUTH_CRITICAL_SECTION;
    // The following really treats the args as const

    rv = pw_cmp(cleartext, encrypted);

    LEAVE_AUTH_CRITICAL_SECTION;
    return rv;
}  /* _cmpCryptPasswd */

#define AUTH_MAX_LINE 1024


/* ------------------------------ _uudecode ------------------------------- */

static const unsigned char pr2six[256]={
    64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,
    64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,62,64,64,64,63,
    52,53,54,55,56,57,58,59,60,61,64,64,64,64,64,64,64,0,1,2,3,4,5,6,7,8,9,
    10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,64,64,64,64,64,64,26,27,
    28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51,
    64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,
    64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,
    64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,
    64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,
    64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,64,
    64,64,64,64,64,64,64,64,64,64,64,64,64
};

char *_uudecode(char *bufcoded)
{
    register char *bufin = bufcoded;
    register unsigned char *bufout;
    register int nprbytes;
    unsigned char *bufplain;
    int nbytesdecoded;

    /* Find the length */
    while(pr2six[(int)*(bufin++)] <= 63);
    nprbytes = bufin - bufcoded - 1;
    nbytesdecoded = ((nprbytes+3)/4) * 3;

    bufout = (unsigned char *) MALLOC(nbytesdecoded + 1);
    bufplain = bufout;

    bufin = bufcoded;
    
    while (nprbytes > 0) {
        *(bufout++) = (unsigned char) 
            (pr2six[(int)(*bufin)] << 2 | pr2six[(int)bufin[1]] >> 4);
        *(bufout++) = (unsigned char) 
            (pr2six[(int)bufin[1]] << 4 | pr2six[(int)bufin[2]] >> 2);
        *(bufout++) = (unsigned char) 
            (pr2six[(int)bufin[2]] << 6 | pr2six[(int)bufin[3]]);
        bufin += 4;
        nprbytes -= 4;
    }
    
    if(nprbytes & 03) {
        if(pr2six[(int)bufin[-2]] > 63)
            nbytesdecoded -= 2;
        else
            nbytesdecoded -= 1;
    }
    bufplain[nbytesdecoded] = '\0';

    return (char *)bufplain;
}


/* ------------------------------ auth_basic ------------------------------ */

int auth_basic(pblock *param, Session *sn, Request *rq)
{
    char *pwfile, *grpfile, *type, *auth, *user, *pw;
    char *pwfn, *grpfn;
    pblock *npb;
    pb_param *pp;
    int ret;

    /* Although this is authorization (which is not cacheable) the
     * check is not actually done until we call require-auth.  So 
     * this part is cacheable; require-auth can be cacheable if the
     * user has limited the auth to only affect a certain set of 
     * paths.
     */
    rq->directive_is_cacheable = 1;

    if(request_header("authorization", &auth, sn, rq) == REQ_ABORTED)
        return REQ_ABORTED;

    if(!auth)
        return REQ_NOACTION;

    type = pblock_findval("auth-type", param);
    pwfile = pblock_findval("userdb", param);
    grpfile = pblock_findval("groupdb", param);
    pwfn = pblock_findval("userfn", param);
    grpfn = pblock_findval("groupfn", param);

    if((!type) || (!pwfile) || (!pwfn) || (grpfile && !grpfn)) {
        log_error(LOG_MISCONFIG, "basic-auth", sn, rq,
              "missing parameter to basic-auth (need type, userdb, userfn)");
        protocol_status(sn, rq, PROTOCOL_SERVER_ERROR, NULL);
        return REQ_ABORTED;
    }

    /* Skip leading whitespace */
    while(*auth && (*auth == ' '))
        ++auth;
    if(!(*auth)) {
        protocol_status(sn, rq, PROTOCOL_FORBIDDEN, NULL);
        return REQ_ABORTED;
    }

    /* Verify correct type */
    if((strlen(auth) < 6) || strncasecmp(auth, "basic ", 6))
        return REQ_NOACTION;

    /* Skip whitespace */
    auth += 6;
    while(*auth && (*auth == ' '))
        ++auth;

    if(!*auth)
        return REQ_NOACTION;

    /* Uuencoded user:password now */
    if(!(user = _uudecode(auth)))
        return REQ_NOACTION;

    if(!(pw = strchr(user, ':'))) {
        FREE(user);
        return REQ_NOACTION;
    }
    *pw++ = '\0';

    npb = pblock_create(4);
    pblock_nvinsert("user", user, npb);
    pblock_nvinsert("pw", pw, npb);
    pblock_nvinsert("userdb", pwfile, npb);
    if(grpfile) 
        pblock_nvinsert("groupdb", grpfile, npb);
    pblock_nvinsert("fn", pwfn, npb);

    if ((ret = func_exec(npb, sn, rq)) != REQ_PROCEED)
    {
        goto bye;
    }

    pblock_nvinsert("auth-type", "basic", rq->vars);
    pblock_nvinsert("auth-user", user, rq->vars);
    pblock_nvinsert("auth-db", pwfile, rq->vars);
#if defined(XP_WIN32) || defined(MCC_ADMSERV)
    /* MLM - the admin server needs this password information,
     *       so I'm putting it back   */
    pblock_nvinsert("auth-password", pw, rq->vars);
#endif /* XP_WIN32 */

    if(grpfile) {
        pblock_nvinsert("groupdb", grpfile, npb);
        pp = pblock_find("fn", npb);
        FREE(pp->value);
        pp->value = STRDUP(grpfn);

        if( (ret = func_exec(npb, sn, rq)) != REQ_PROCEED )
            goto bye;
    }
    ret = REQ_PROCEED;
  bye:
    pblock_free(npb);
    FREE(user);
    return ret;
}


/* ----------------------------- auth_basic1 ------------------------------ */


/*
 * Compatibility function for 1.0 servers, to more powerful 1.1 SAF
 * accessible auth trans function
 */
int auth_basic1(pblock *param, Session *sn, Request *rq)
{
    pblock *npb = pblock_create(4);
    char *pwfile = pblock_findval("userfile", param);
    char *grpfile = pblock_findval("grpfile", param);
    char *type = pblock_findval("auth-type", param);
    char *pwdbm = pblock_findval("dbm", param);
    int ret;

    rq->directive_is_cacheable = 1;

    if(type) pblock_nvinsert("auth-type", type, npb);

    if(grpfile) {
        pblock_nvinsert("groupdb", grpfile, npb);
        pblock_nvinsert("groupfn", "simple-groupdb", npb);
    }
    if(pwfile) {
        pblock_nvinsert("userdb", pwfile, npb);
        pblock_nvinsert("userfn", "simple-userdb", npb);
    }
    else if(pwdbm) {
        pblock_nvinsert("userdb", pwdbm, npb);
        pblock_nvinsert("userfn", "dbm-userdb", npb);
    }
    ret = auth_basic(npb, sn, rq);
    pblock_free(npb);
    return ret;
}


#include "libaccess/nsadb.h"
#include "libaccess/nsauth.h"  /* ClAuth_t */

static int check_password (const char *user, const char *pw,
			   const char *pwfile, Session *sn, Request *rq)
{
    void * authdb;
    UserObj_t * uoptr;
    ClAuth_t * cla = (ClAuth_t *) sn->clauth;
    int auth_result;
    int rv;

    rv = nsadbOpen(0, (char *)pwfile, 0, &authdb);
    if (rv < 0) {
	log_error(LOG_MISCONFIG, "basic-ncsa", sn, rq, 
		  "cannot open database %s", pwfile);
	return REQ_ABORTED;
    }

    rv = nsadbFindByName(0, (char *)authdb, (char *)user, AIF_USER,
			 (void **)&uoptr);
    if (rv != AIF_USER) {
	nsadbClose(authdb, 0);
	return REQ_NOACTION;
    }

    nsadbClose(authdb, 0);

#ifdef ENCRYPT_PASSWORDS
    auth_result = _cmpCryptPasswd((const char *)(uoptr->uo_pwd), pw);
#else
    auth_result = strcmp((const char *)uoptr->uo_pwd, pw);
#endif /* ENCRYPT_PASSWORDS */

    if (auth_result) {
	userFree(uoptr);
	log_error(LOG_SECURITY, "basic-ncsa", sn, rq, 
		  "user %s password did not match database %s", user, pwfile);
	return REQ_NOACTION;
    }

    /* Free any previous user object */
    if (cla->cla_uoptr != 0) {
 	userFree(cla->cla_uoptr);
    }

    /* Store pointer to the latest user object */
    cla->cla_uoptr = uoptr;

    return REQ_PROCEED;
}

#ifdef AUTH_DBM


/* ------------------------------- dbm_user ------------------------------- */


int dbm_user(pblock *param, Session *sn, Request *rq)
{
    char * pwfile = pblock_findval("userdb", param);
    char * user = pblock_findval("user", param);
    char * pw = pblock_findval("pw", param);

    return check_password(user, pw, pwfile, sn, rq);
}
#endif


/* ------------------ NCSA httpd compatibility functions ------------------ */



/* ----------------------------- simple_group ----------------------------- */


int simple_group(pblock *param, Session *sn, Request *rq)
{
    SYS_FILE fd;
    filebuf_t *buf;
    int eof, ln;
    char *t, *u;
    char line[AUTH_MAX_LINE];

    char *gfn = pblock_findval("groupdb", param);
    char *user = pblock_findval("user", param);

    rq->directive_is_cacheable = 1;

    if( (fd = system_fopenRO(gfn)) == SYS_ERROR_FD) {
        log_error(LOG_FAILURE, "basic-ncsa", sn, rq, 
                  "cannot open group file %s", gfn);
        protocol_status(sn, rq, PROTOCOL_SERVER_ERROR, NULL);
        return REQ_ABORTED;
    }
    if(!(buf = filebuf_open(fd, FILE_BUFFERSIZE))) {
        log_error(LOG_FAILURE, "basic-ncsa", sn, rq, 
                  "cannot open buffer to group file %s", gfn);
        protocol_status(sn, rq, PROTOCOL_SERVER_ERROR, NULL);
        return REQ_ABORTED;
    }

    ln = 0;
    while(1) {
        eof = util_getline(buf, ln, AUTH_MAX_LINE, line);
        if(!line[0]) {
            if(eof) {
                filebuf_close(buf);
                return REQ_NOACTION;
            }
            else
                continue;
        }

        for(t = line; *t && (*t != ':'); t++);
        if(!(*t))
            continue;
        *t++ = '\0';
        while(1) {
            while((*t) && (*t == ' ')) ++t;
            if(!(*t))
                break;
            u = t;

            while((*u) && (*u != ' ')) ++u;
            if(!strncmp(user, t, u - t)) {
                pblock_nvinsert("auth-group", line, rq->vars);
                filebuf_close(buf);
                return REQ_PROCEED;
            }
            if(!(*u))
                break;
            else
                t = u;
        }
    }
}

int simple_user(pblock *param, Session *sn, Request *rq)
{
    register char *t;
    char line[AUTH_MAX_LINE];
    SYS_FILE fd;
    filebuf_t *buf;
    int ln, eof;

    char *pwfile = pblock_findval("userdb", param);
    char *user = pblock_findval("user", param);
    char *pw = pblock_findval("pw", param);

    rq->directive_is_cacheable = 1;

    /* Open user file and verify password */
    if((fd = system_fopenRO(pwfile)) == SYS_ERROR_FD) {
        log_error(LOG_FAILURE, "basic-ncsa", sn, rq, 
                  "can't open basic password file %s (%s)", pwfile, 
                  system_errmsg());
        protocol_status(sn, rq, PROTOCOL_SERVER_ERROR, NULL);
        return REQ_ABORTED;
    }
    if(!(buf = filebuf_open(fd, FILE_BUFFERSIZE))) {
        log_error(LOG_FAILURE, "basic-ncsa", sn, rq, 
                  "can't open buffer from password file %s (%s)", pwfile, 
                  system_errmsg());
        protocol_status(sn, rq, PROTOCOL_SERVER_ERROR, NULL);
        return REQ_ABORTED;
    }
    ln = 1;
    while(1) {
        eof = util_getline(buf, ln, AUTH_MAX_LINE, line);
        if(!line[0]) {
            if(eof) {
                filebuf_close(buf);
                return REQ_NOACTION;
            }
            else
                continue;
        }

        if(!(t = strchr(line, ':')))
            continue;
        *t++ = '\0';
        if(!strcmp(line, user)) {
	    if (_cmpCryptPasswd(t, pw)) {
                log_error(LOG_SECURITY, "basic-ncsa", sn, rq, 
                          "user %s password did not match pwfile %s", 
                          user, pwfile);
                filebuf_close(buf);
#ifdef MCC_ADMSERV
                /* MLM - Admin server wants this to stop if the user gives
                 *       the wrong password or if they don't exist.  Leave
                 *       a message for the admin server and return noaction. */
                pblock_nvinsert("admin-rejection", "bad password", 
                                rq->vars);
#endif
                return REQ_NOACTION;
            }
            else {
                filebuf_close(buf);
                return REQ_PROCEED;
            }
        }
        if(eof) {
            /* if user not found, keep going */
            filebuf_close(buf);
            return REQ_NOACTION;
        }
    }
}


/*
 * get_auth_user_basic --
 *   Description:
 *	This getter calls the getter for ACL_ATTR_IS_VALID_PASSWORD to verify the
 *	password and store authenticated user information on the Session
 *	property list for backward compatibilty.
 */
int get_auth_user_basic (NSErr_t *errp, PList_t subject, PList_t resource,
			 PList_t auth_info, PList_t global_auth, void *unused)
{
    char *raw_user;
    char *raw_pw;
    char *userdn;
    int rv;
    pblock *pb;
    Session *sn;
    Request *rq;
    int accel;			/* accelerated path? - no request struct! */
    char *cached_user = 0;

    /* Authenticate the raw_user and raw_pw against the given database. */

    rv = PListGetValue(subject, ACL_ATTR_CACHED_USER_INDEX, 
		       (void **)&cached_user, NULL);

    if ((rv >= 0) && cached_user && *cached_user) {
        PListInitProp(subject, ACL_ATTR_USER_INDEX, ACL_ATTR_USER, 
		      cached_user, 0);
        return LAS_EVAL_TRUE;
    }

    rv = PListGetValue(resource, ACL_ATTR_REQUEST_INDEX, (void **)&rq, NULL);

    accel = (rv < 0 || !rq);

    if (!accel && !rq) {
        ereport(LOG_SECURITY, "get_auth_user_basic unable to get request address %d", rv);
        return LAS_EVAL_FAIL;
    }

    PListDeleteProp(subject, ACL_ATTR_IS_VALID_PASSWORD_INDEX,
		    ACL_ATTR_IS_VALID_PASSWORD);
    rv = ACL_GetAttribute(errp, ACL_ATTR_IS_VALID_PASSWORD, (void **)&raw_user,
			  subject, resource, auth_info, global_auth);

    if (!accel && rv == LAS_EVAL_FALSE) {
	/* incorrect password */
	char* realm = 0;

	rv = PListGetValue(subject, ACL_ATTR_SESSION_INDEX, (void **)&sn, NULL);
	if (rv < 0 || !sn) {
	    ereport(LOG_SECURITY, "unable to get session address %d", rv);
	    return LAS_EVAL_FAIL;
	}
    
	if (auth_info)
	    rv = PListGetValue(auth_info, ACL_ATTR_PROMPT_INDEX, (void **)&realm, NULL);

	/* This may fail if somebody has requested www-auth already */
	PListInitProp(resource, ACL_ATTR_WWW_AUTH_PROMPT_INDEX,
		      ACL_ATTR_WWW_AUTH_PROMPT, realm, 0);

	return LAS_EVAL_NEED_MORE_INFO;
    }
    else if (rv != LAS_EVAL_TRUE) {
	return rv;
    }

    /* If the password is correct, set the auth-user attributes.
     * For backward compatibilty, set the auth-* attributes in the
     * request's vars pblock
     */

    if (!accel) {
	char *dbname = 0;

	pb = rq->vars;
	pblock_nvinsert(ACL_ATTR_AUTH_TYPE, ACL_AUTHTYPE_BASIC, pb);
	pblock_nvinsert(ACL_ATTR_AUTH_USER, raw_user, pb);

	rv = PListGetValue(subject, ACL_ATTR_RAW_PASSWORD_INDEX,
			   (void **)&raw_pw, NULL);
	
	if (rv >= 0 && raw_pw) {
	    pblock_nvinsert(ACL_ATTR_AUTH_PASSWORD, raw_pw, pb);
	}


	rv = PListGetValue(subject, ACL_ATTR_USERDN_INDEX, (void **)&userdn,
			   NULL);

	if (rv >= 0 && userdn && *userdn) {
	    pblock_nvinsert(ACL_ATTR_USERDN, userdn, pb);
	}

	rv = ACL_AuthInfoGetDbname(auth_info, &dbname);
	if ((rv >= 0) && dbname) {
	    pblock_nvinsert(ACL_ATTR_AUTH_DB, dbname, rq->vars);
	}
    }

    /* pblock_nvinsert(ACL_ATTR_AUTH_DB, pwfile, subject); */

    PListInitProp(subject, ACL_ATTR_USER_INDEX, ACL_ATTR_USER,
		  pool_strdup(PListGetPool(subject), raw_user), 0);
    return LAS_EVAL_TRUE;
}

int parse_basic_user_login (const char *auth, char **user,
			    char **password)
{
    char *uid;
    char *pw;

    /* Skip leading whitespace */
    while (*auth && (*auth == ' ')) ++auth;

    if (!(*auth)) return LAS_EVAL_FAIL;

    /* Verify correct type */
    if((strlen(auth) < 6) || strncasecmp(auth, "basic ", 6))
        return LAS_EVAL_FAIL;

    /* Skip whitespace */
    auth += 6;
    while(*auth && (*auth == ' '))
        ++auth;

    if(!*auth)
        return LAS_EVAL_FAIL;

    /* Uuencoded user:password now */
    if(!(uid = _uudecode((char *)auth)))
        return LAS_EVAL_FAIL;

    if(!(pw = strchr(uid, ':'))) {
        FREE(uid);
        return LAS_EVAL_FAIL;
    }
    *pw++ = '\0';

    *user = uid;
    *password = pw;

    return LAS_EVAL_TRUE;
}


int get_user_login_basic (NSErr_t *errp, PList_t subject, 
			  PList_t resource, PList_t auth_info,
			  PList_t global_auth, void *unused)
{
    /* Get the authorization attribute and parse it into raw_user and
     * raw_password.
     */
    char *auth;
    char *raw_user;
    char *raw_pw;
    int rv;

    rv = ACL_GetAttribute(errp, ACL_ATTR_AUTHORIZATION, (void **)&auth, subject,
			  resource, auth_info, global_auth);

    if (rv != LAS_EVAL_TRUE) {
	return rv;
    }
    else if (!auth) {
        ereport(LOG_SECURITY, "get_user_login_basic getter returned NULL auth header");
	return LAS_EVAL_FAIL;
    }

    /* Parse the authorization attribute to get raw_user_login and
     * raw_user_password.
     */
    rv = parse_basic_user_login(auth, &raw_user, &raw_pw);

    if (rv != LAS_EVAL_TRUE) {
	return rv;
    }

    PListInitProp(subject, ACL_ATTR_RAW_USER_INDEX, ACL_ATTR_RAW_USER,
		  pool_strdup(PListGetPool(subject), raw_user), 0);
    PListInitProp(subject, ACL_ATTR_RAW_PASSWORD_INDEX, ACL_ATTR_RAW_PASSWORD,
		  pool_strdup(PListGetPool(subject), raw_pw), 0);
    FREE(raw_user);		/* also frees raw_pw */

    return LAS_EVAL_TRUE;
}

int get_authorization_basic (NSErr_t *errp, PList_t subject, 
			     PList_t resource, PList_t auth_info,
			     PList_t global_auth, void *unused)
{
    /* request basic auth if authorization attr doesn't exist */
    int rv;
    Session *sn = 0;
    Request *rq = 0;
    char *auth = 0;

    rv = PListGetValue(subject, ACL_ATTR_ACCEL_AUTH_INDEX, (void **)&auth, NULL);
    if (rv <0) {			/* Unaccellerated path */

        rv = PListGetValue(subject, ACL_ATTR_SESSION_INDEX, (void **)&sn, NULL);
        if (rv < 0 || !sn) {
            ereport(LOG_SECURITY, "get_authorization_basic unable to get session address %d", rv);
            return LAS_EVAL_FAIL;
        }
 
        rv = PListGetValue(resource, ACL_ATTR_REQUEST_INDEX, (void **)&rq, NULL);
        if (rv < 0 || !rq) {
            ereport(LOG_SECURITY, "get_authorization_basic unable to get request address %d", rv);
            return LAS_EVAL_FAIL;
        }

        /* request_header always returns REQ_PROCEED! */
        request_header("authorization", &auth, sn, rq);
    } else {				/* Via the accel_file_cache */
	if (!auth || !*auth) {
	    /* Don't bother with response header */
	    /* return LAS_EVAL_FALSE to avoid error messages since
	     * unaccelerated path will be tried anyway. */
	    return LAS_EVAL_FALSE;
	}
    }


    if (!auth || !*auth) {
	char* realm = 0;
    
	if (auth_info)
	    rv = PListGetValue(auth_info, ACL_ATTR_PROMPT_INDEX, (void **)&realm, NULL);
	PListInitProp(resource, ACL_ATTR_WWW_AUTH_PROMPT_INDEX, ACL_ATTR_WWW_AUTH_PROMPT, realm, 0);

	return LAS_EVAL_NEED_MORE_INFO;
    }
    else {
	PListInitProp(subject, ACL_ATTR_AUTHORIZATION_INDEX,
		      ACL_ATTR_AUTHORIZATION,
		      pool_strdup(PListGetPool(subject), auth), 0);
	return LAS_EVAL_TRUE;
    }
}


#if defined(CLIENT_AUTH)

/* Get base64 encoding of SSL session id into sn->client pblock */
int auth_get_sslid(pblock *param, Session *sn, Request *rq)
{
    char *idstr;
    SECItem *iditem;

    /* This requires SSL to be active */
    if (security_active) {

        /* Have we already got the session id in the client pblock? */
        idstr = pblock_findval("ssl-id", sn->client);
        if (!idstr) {

            /* No, look it up based on the connection fd */
            iditem = SSL_GetSessionID(sn->csd);
            if (iditem) {

                /* Convert to base64 ASCII encoding */
                idstr = BTOA_DataToAscii(iditem->data, iditem->len);
                if (idstr) {

                    /* Add encoding to client pblock */
                    pblock_nvinsert("ssl-id", idstr, sn->client);

                    /* Free the encoding buffer (pblock_nvinsert dups it) */
                    PR_Free(idstr);
                }

                SECITEM_FreeItem(iditem, PR_TRUE);
            }
        }
    }

    return REQ_NOACTION;
}

#endif /* defined(CLIENT_AUTH) */
