/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * cgi.c: Handle Common Gateway Interface scripts
 *
 * CGI is an interface definition for external programs to the server which
 * can be executed to produce documents on the fly. Handy for forms etc.
 * 
 * Rob McCool
 */

#ifdef HPUX
#include "pthread.h"
#endif

#include "safs/cgi.h"
#include "base/systems.h"
#include "base/pblock.h"
#include "base/session.h"
#include "frame/req.h"
#include "frame/log.h"
#include "frame/httpact.h"  /* for servact_translate_uri */

#include "base/util.h"      /* environment functions, can_exec */
#include "base/net.h"       /* SYS_NETFD */
#include "base/buffer.h"    /* netbuf */
#include "base/daemon.h"    /* child_fork */
#include "frame/protocol.h" /* protocol_status */
#include "frame/http.h"     /* http_hdrs2env */
#include "frame/conf.h"     /* server action globals */
#include "frame/cmstat.h"
/* Removed for new nspr and ns security integration
#include "sslio/sslio.h"
*/

#include <sys/stat.h>
#include <stdlib.h>         /* getenv */
#include <sys/types.h>
#include <unistd.h>         /* pipe, fork, exec */
#include <ctype.h>          /* isdigit */

#include <signal.h>

#include <fcntl.h>  /* O_NONBLOCK */

#include <private/pprio.h>
#ifdef NET_SSL
#include "ssl.h"
#endif

#ifdef FORTEZZA
/* removed for ns security integration
#include "sec.h"
*/
#include <key.h>
#include <cert.h>
#endif /* FORTEZZA */

#ifdef FORTEZZA
/* XXX Nasty hack to get the libsec headers */
/* integration with ns security
extern "C" {
#include "../../lib/libsec/preenc.h"
#include "../../lib/libsec/fortezza.h"
}
*/
#endif


#include "base/systhr.h"

#define port conf_getglobals()->Vport

#ifdef MCC_ADMSERV
#ifdef __cplusplus
extern "C" char **admin_cgivars(Session *sn, Request *rq, pid_t ppid, SYS_FILE pwpipe, char **env);
#else
extern char **admin_cgivars(Session *sn, Request *rq, pid_t ppid, SYS_FILE pwpipe, char **env);
#endif
#endif

#define CGI_VERSION "CGI/1.1"
static int cgi_rdtimeout = 300;
#define DEFAULT_PATH "/bin:/usr/bin:/usr/ucb:/usr/bsd:/usr/local/bin"

static pblock *initenv = NULL;
static char *_cgi_env_path = NULL, *_cgi_env_tz = NULL, *_cgi_env_lang = NULL;

#define WAITPID(a,b,c) if (conf_getglobals()->wait_for_cgi) util_waitpid(a,b,c)



/* ----------------------------- _create_argv ----------------------------- */


/*
 * _create_argv takes a query string (URL-encoded) and converts it to 
 * command line parameters. Doesn't handle "foo bar" being one argument, 
 * etc. or $LOGNAME stuff.
 */

char **_create_argv(char *av0, char *args) {
    register int x,n;
    char **av, *w, *l, c;

    /* MLM - Removing the side effect change on args - Before, the
     *       strcpy didn't exist, but the MALLOC did, which leaked
     *       memory.  */
    n = strlen(args);
    l = (char *) MALLOC(n + 1);
    strcpy(l, args);

    for(x=0,n=2;l[x];x++) {
        if(l[x] == '+') {
            l[x] = ' ';
            ++n;
        }
    }

    av = (char **) MALLOC((n+1)*sizeof(char *));
    av[0] = av0;
    /* MLM - deleted l=args */
    util_uri_unescape(l);
    n = 1; x = 0; w = l;

    while(1) {
        if((!l[x]) || (l[x] == ' ')) {
            c = l[x];
            l[x] = '\0';
            av[n++] = util_sh_escape(w);
            if(c)
                w = &l[x+1];
            else
                break;
        }
        ++x;
    }
    /* MLM - plug leak */
    FREE(l);
    av[n] = NULL;
    return av;
}


/* -------------------------- cgi_specific_vars --------------------------- */


#define MAX_COMMON_VARS 18
#define MAX_CGI_VARS (MAX_COMMON_VARS+8)
#define MAX_CGI_SSI_VARS (MAX_CGI_VARS+7)

char **cgi_specific_vars(Session *sn, Request *rq, char *args, char **env)
{
    int x, y;
    register pblock *pb;
    char c, *t, *u;

    env = util_env_create(env, MAX_CGI_VARS, &x);

    pb = rq->reqpb;

    env[x++] = util_env_str("GATEWAY_INTERFACE", CGI_VERSION);

    env[x++] = util_env_str("SERVER_PROTOCOL", 
                            pblock_findval("protocol", pb));

    env[x++] = util_env_str("REQUEST_METHOD",
                            pblock_findval("method", pb));

    t = pblock_findval("uri", pb);

    /* Simulate CGI URIs by truncating path info */
    if( (u = pblock_findval("path-info", rq->vars)) ) {
        y = strlen(t) - strlen(u);
        c = t[y];
        t[y] = '\0';
        env[x++] = util_env_str("SCRIPT_NAME", t);
        t[y] = c;

        env[x++] = util_env_str("PATH_INFO", u);
        if((t = request_translate_uri(u, sn))) {
            env[x++] = util_env_str("PATH_TRANSLATED", t);
            FREE(t);
        }
    }
    else
        env[x++] = util_env_str("SCRIPT_NAME", t);

    if(args)
        env[x++] = util_env_str("QUERY_STRING", args);

    env[x] = NULL;
    return env;
}

/* passing SSI variables as environment vars for SSI CGIs -- gdong */
/* --------------------------- cgi_ssi_vars ---------------------------- */
char **cgi_ssi_vars(Session *sn, Request *rq, char *args, char **env)
{
    int x;
    register pblock *pb;
    char *t;
 
   env = util_env_create(env, MAX_CGI_SSI_VARS,  &x);

   pb = rq->reqpb;

   if (t = pblock_findval("DOCUMENT_URI", pb))	
     env[x++] = util_env_str("DOCUMENT_URI", t);

   if (t = pblock_findval("DOCUMENT_NAME",  pb))
     env[x++] = util_env_str("DOCUMENT_NAME", t);

   if (t = pblock_findval("DATE_LOCAL", pb))
     env[x++] = util_env_str("DATE_LOCAL", t);

   if (t = pblock_findval("DATE_GMT", pb))
     env[x++] = util_env_str("DATE_GMT", t);

   if (t = pblock_findval("LAST_MODIFIED",  pb))
     env[x++] = util_env_str("LAST_MODIFIED", t);

   if (t = pblock_findval("QUERY_STRING_UNESCAPED",  pb))
     env[x++] = util_env_str("QUERY_STRING_UNESCAPED", t);

   env[x] = NULL;
   return env;
}
 

/* --------------------------- cgi_common_vars ---------------------------- */


char **cgi_common_vars(Session *sn, Request *rq, char **env) 
{
    char *t, *ip, tp[15];
    int x;

    env = util_env_create(env,MAX_COMMON_VARS,&x);

    if(_cgi_env_path)
        env[x++] = util_env_str("PATH", _cgi_env_path);
    if(_cgi_env_tz)
        env[x++] = util_env_str("TZ", _cgi_env_tz);
    if(_cgi_env_lang)
        env[x++] = util_env_str("LANG", _cgi_env_lang);

    env[x++] = util_env_str("SERVER_SOFTWARE", MAGNUS_VERSION_STRING);

    util_itoa(port, tp);
    env[x++] = util_env_str("SERVER_PORT", tp);

    env[x++] = util_env_str("SERVER_NAME", server_hostname);

    t = http_uri2url_dynamic("","", sn, rq);
    env[x++] = util_env_str("SERVER_URL", t);
    FREE(t);

    ip = pblock_findval("ip", sn->client);
    t = session_dns(sn);
    env[x++] = util_env_str("REMOTE_HOST", (t ? t : ip));
    env[x++] = util_env_str("REMOTE_ADDR", ip);

    if((t = pblock_findval("auth-user", rq->vars))) {
        env[x++] = util_env_str("REMOTE_USER", t);
        env[x++] = util_env_str("AUTH_TYPE", 
                                pblock_findval("auth-type", rq->vars));
	if((t = pblock_findval("auth-userdn", rq->vars))) {
	    env[x++] = util_env_str("REMOTE_USERDN", t);
	}
    }

#ifdef CLIENT_AUTH
    if ((t = pblock_findval("auth-cert", rq->vars))) {
	/* Alan made up this environment variable name */
        env[x++] = util_env_str("CLIENT_CERT", t);
    }
#endif    
	


#ifdef NET_SSL
    if(security_active)  {
        char *sslid = pblock_findval("ssl-id", sn->client);

        env[x++] = util_env_str("HTTPS", "ON");

        if (pblock_findval("keysize", sn->client) != NULL)
          env[x++] = util_env_str("HTTPS_KEYSIZE", 
                                   pblock_findval("keysize", sn->client));

        if (pblock_findval("secret-keysize", sn->client) != NULL)
          env[x++] = util_env_str("HTTPS_SECRETKEYSIZE", 
                                   pblock_findval("secret-keysize", sn->client));

        env[x++] = util_env_str("HTTPS_SESSIONID",
                                sslid?sslid:(char *)"");
    }  else  {
        env[x++] = util_env_str("HTTPS", "OFF");
    }
#endif

    env[x] = NULL;
    return env;
}


/* --------------------------- cgi_scan_headers --------------------------- */


#define BAD_CGI(s1, s2) \
    util_sprintf(err, s1, s2); \
    log_error(LOG_FAILURE, "cgi-parse-output", sn, rq, \
              "the CGI program %s did not produce a valid header (%s)", \
              pblock_findval("path", rq->vars), err); \
    return REQ_ABORTED;


int cgi_scan_headers(Session *sn, Request *rq, netbuf *nb)
{
    register int x,y;
    register char c;
    int nh, i;
    netbuf *buf = (nb ? nb : sn->inbuf);
    char t[REQ_MAX_LINE], err[REQ_MAX_LINE];

    nh = 0;
    x = 0; y = -1;

    while(1) {
        if(x > REQ_MAX_LINE) {
            BAD_CGI("CGI header line too long (max %d)", REQ_MAX_LINE);
        }
        if(nh > HTTP_MAX_HEADERS) {
            BAD_CGI("too many headers from CGI script (max %d)",
                        HTTP_MAX_HEADERS);
        }
        i = netbuf_getc(buf);

        if(i == IO_ERROR) {
            BAD_CGI("read failed, error is %s", system_errmsg());
        }
        else if(i == IO_EOF) {
            BAD_CGI("program terminated without a valid CGI header. Check for core dump or other abnormal termination", "");
        }

        c = (char) i;
        switch(c) {
          case LF:
            if((!x) || ((x == 1) && (t[0] == CR)))
                return REQ_PROCEED;

            if(t[x-1] == CR)
                --x;
            t[x] = '\0';
            if(y == -1) {
                BAD_CGI("name without value: got line \"%s\"", t);
            }
            while(t[y] && isspace(t[y])) ++y;

            pblock_nvinsert(t, &t[y], rq->srvhdrs);

            x = 0; y = -1;
            ++nh;
            break;
          case ':':
            if(y == -1) {
                y = x+1;
                c = '\0';
            }
          default:
#if defined(SNI)
	    {
              int d = tolower(c);
              t[x++] =  ((y == -1) && isupper(c) ? d : c);
	    }
#else
            t[x++] = ((y == -1) && isupper(c) ? tolower(c) : c);
#endif
        }
    }
}


/* --------------------------- cgi_send_content --------------------------- */


/* 
 * This routine is not forgiving of pinheads: If content-length
 * or content-type is not right, it punts.
 */

int cgi_send_content(Session *sn, Request *rq, SYS_NETFD s2c)
{
    char *cls, *cts;
    int cl, x;

    if((request_header("content-length", &cls, sn, rq) == REQ_ABORTED) || 
       (request_header("content-type", &cts, sn, rq) == REQ_ABORTED))
        return REQ_ABORTED;

    if((!cls) || (!cts))
        return REQ_PROCEED;

    while(isspace(*cls)) ++cls;
    for(x = 0; cls[x] && (!isspace(cls[x])); ++x) {
        if(!isdigit(cls[x])) {
            log_error(LOG_WARN, "send-cgi", sn, rq, 
                      "client sent bad content-length (%s)", cls);
            return REQ_ABORTED;
        }
    }

    cl = atoi(cls);
    if(cl < 0) {
        log_error(LOG_WARN, "send-cgi", sn, rq, 
                  "client sent bad content-length (%s)", cls);
        return REQ_ABORTED;
    }

    if(netbuf_buf2sd(sn->inbuf, s2c, cl) == IO_ERROR) {
        if(errno != EPIPE) {
            log_error(LOG_FAILURE, "send-cgi", sn, rq, 
                      "error occurred while sending content to script (%s)",
                      sn->inbuf->errmsg);
            return REQ_ABORTED;
        }
    }
    return REQ_PROCEED;
}

#ifdef MCC_ADMSERV
/* MLM - send the password to the admin cgi */
int adm_cgi_send_password(Session *sn, Request *rq, int fd)
{
    char *user, *pass, *auth, *userDN, *ans;
    char *siepwd, *sievalue;
#ifdef FORTEZZA
    char *pin;
#endif /* FORTEZZA */
    int r;

    /* Yuck.  Well, if there's no user, I guess we'll send a nonsense
     * string rather than tearing down the pipe. */
    user = pblock_findval("auth-user", rq->vars);
    pass = pblock_findval("auth-password", rq->vars);
    userDN = pblock_findval("userdn", rq->vars);
    r    = request_header("authorization", &auth, sn, rq);
    if( (r == REQ_ABORTED) )
        return REQ_ABORTED;

    if(!user)  {
        user = ADM_NO_VALUE_STRING;
    }
    if(!pass)  {
        pass = ADM_NO_VALUE_STRING;
    }
    if(!auth)  {
        auth = ADM_NO_VALUE_STRING;
    }
    if(!userDN)  {
        userDN = ADM_NO_VALUE_STRING;
    }

    siepwd = admGetCachedSIEPWD();
    sievalue = (char *)(siepwd?siepwd:ADM_NO_VALUE_STRING);

#ifdef FORTEZZA

    pin = NULL;
    pin = FortezzaGetAuthPin();
    if(!pin)  {
	pin = ADM_NO_VALUE_STRING;
    }

    ans = (char *) MALLOC(strlen(user) + strlen(ADM_USER_STRING) +
                          strlen(pass) + strlen(ADM_PASS_STRING) + 
                          strlen(auth) + strlen(ADM_AUTH_STRING) +
			              strlen(pin) + strlen(ADM_PIN_STRING) +
                          strlen(userDN) + strlen(ADM_USERDN_STRING) +
                          strlen(sievalue) + strlen(ADM_SIEPWD_STRING) +
                          20);
    sprintf(ans, "%s%s\n%s%s\n%s%s\n%s%s\n%s%s\n%s%s\n",
        ADM_USER_STRING, user, 
        ADM_PASS_STRING, pass, 
        ADM_AUTH_STRING, auth,
        ADM_PIN_STRING, pin,
        ADM_USERDN_STRING, userDN,
        ADM_SIEPWD_STRING, sievalue,
        );
#else

    ans = (char *) MALLOC(strlen(user) + strlen(ADM_USER_STRING) +
                          strlen(pass) + strlen(ADM_PASS_STRING) + 
                          strlen(auth) + strlen(ADM_AUTH_STRING) +
                          strlen(userDN) + strlen(ADM_USERDN_STRING) +
                          strlen(sievalue) + strlen(ADM_SIEPWD_STRING) +
                          16);
    sprintf(ans, "%s%s\n%s%s\n%s%s\n%s%s\n%s%s\n",
        ADM_USER_STRING, user, 
        ADM_PASS_STRING, pass, 
        ADM_AUTH_STRING, auth,
        ADM_USERDN_STRING, userDN,
        ADM_SIEPWD_STRING, sievalue
        );
#endif /* FORTEZZA */

    /* strlen(ans)+1 to include null */
    r = write(fd, ans, strlen(ans)+1);
    FREE(ans);

#ifdef FORTEZZA
    if(strcmp(pin, ADM_NO_VALUE_STRING) != 0)
       FortezzaDestroyAuthPin(pin);
#endif /* FORTEZZA */

    if (siepwd) free(siepwd);

    if(r==-1)  {
        if(errno != EPIPE)  {
            log_error(LOG_FAILURE, "send-cgi", sn, rq, 
                      "error occurred while sending username to script (%s)",
                      system_errmsg());
            return REQ_ABORTED;
        }
    }
    return REQ_PROCEED;
}
#endif


/* ---------------------------- cgi_start_exec ---------------------------- */


/*
 * Takes a path and request structures, and a pid_t to store the child pid
 * in. Returns -1 on error, and a file descriptor on success.
 * 
 * Note that NPH is no longer taken care of here. Data from the CGI
 * program is still buffered in case SSL is active.
 */

typedef struct child_parms {
    char *path;
    PRFileDesc *s2c[2];
    PRFileDesc *c2s[2];
    int ex[2];
    PRFileDesc *csd;
    char *args;
    char *argv0;
    char **env;
} child_parms;

void child_code(void *_parms)
{
    child_parms *parms = (child_parms *)_parms;

    if(util_chdir2path(parms->path) == -1)
        ereport(LOG_FAILURE, "send-cgi reports: could not chdir (%s)",
                system_errmsg());

    /* For mob rules, make it so we don't get sighup'd */
    setsid();

    /* refer to the comments below from robm */
    /* changed for new nspr21 
    close(parms->s2c[1]->osfd);
    close(parms->c2s[0]->osfd);
    */
    PR_Close(parms->s2c[1]);
    PR_Close(parms->c2s[0]);
    close(parms->ex[0]);

    if(PR_FileDesc2NativeHandle(parms->c2s[1]) != STDOUT_FILENO) {
        PRInt32 oldFD = PR_FileDesc2NativeHandle(parms->c2s[1]);
        dup2(oldFD, STDOUT_FILENO);
        close(oldFD);
    }
    /* HOT DAMN! Finally nailed that bug!!! */
    /* XXXrobm I need to be able to close the socket without going into
       a lock, because with fork1() the locks are in an undefined state.
       That's why this is close() and not net_close(). */
    /* changed for new nspr21 
    close(parms->csd->osfd);
    */
    /* Well with the new NSPR21 and ns security, this close cause a lot of problem...yjh
    PR_Close(parms->csd);
    */

    if(PR_FileDesc2NativeHandle(parms->s2c[0]) != STDIN_FILENO) {
        PRInt32 oldFD = PR_FileDesc2NativeHandle(parms->s2c[0]);
        dup2(oldFD, STDIN_FILENO);
        close(oldFD);
    }
    dup2(STDOUT_FILENO, STDERR_FILENO);

    /* Reset SIGPIPE so /bin/sh and stdio still die when we cut them off */
#ifdef BSD_SIGNALS
    signal(SIGPIPE, SIG_DFL);
#else
    {
        struct sigaction sa;
        sa.sa_flags = 0;
        sigemptyset(&sa.sa_mask);
        sa.sa_handler = SA_HANDLER_T(SIG_DFL);
        if ( sigaction(SIGPIPE, &sa, NULL) == -1)
            fprintf(stderr, "Failed to create action for SIGPIPE(errno=%d)\n",
                    errno);
    }
#endif

    /* Only ISINDEX scripts get decoded arguments. */
    if((!parms->args) || (strchr(parms->args,'=')))
        execle(parms->path, parms->argv0, (char *)0, parms->env);
    else
        execve(parms->path, _create_argv(parms->argv0, parms->args), parms->env);

#if !defined(SOLARIS) || !defined(THREAD_NSPR_KERNEL)
    /* Under Solaris kernel threads, post-fork1() the locks are in an 
       undefined state. strftime() needs a lock. */
    ereport(LOG_FAILURE, "send-cgi reports: cannot exec %s (%s)",
            parms->path, system_errmsg());
#endif
    write(parms->ex[1], "", 1);
    exit(1);
}

int create_child(child_parms *parms)
{
    int pid;

    pid = child_fork();
    if (pid == 0) 
        child_code(parms);
    return pid;
}



SYS_NETFD cgi_start_exec(Session *sn, Request *rq, pid_t *pid, char *path)
{
    char **env, *argv0, *args, c;
    PRFileDesc *c2s[2], *s2c[2];
    int ex[2];
    int n;
    struct stat *finfo;
#ifdef MCC_ADMSERV  
    /* MLM - send username and password via IPC pipe to  *
     *       CGI for libadmin usage.                     */
    pid_t ppid = getppid();
    int pwPipe[2];
    PRFileDesc *pwPipe_prfd[2];
#endif
    child_parms childparms;

#ifdef THREAD_NSPR_USER
    /* When we're using the companion process DNS crud, NSPR doesn't like 
       us using PR_Read from a child process. Use this to start the
       DNS call, and later we'll pick it up. 
     */
    session_dns(sn);
#endif

    args = pblock_findval("query", rq->reqpb);

    if(!(finfo = request_stat_path(path, rq))) {
        log_error(LOG_WARN, "send-cgi", sn, rq, 
                  "cannot find CGI program %s (%s)", path, rq->staterr);
        protocol_status(sn, rq, PROTOCOL_NOT_FOUND, NULL);
        return SYS_NET_ERRORFD;
    } 
    else if(!S_ISREG(finfo->st_mode)) {
        log_error(LOG_WARN, "send-cgi", sn, rq, 
                  "cannot execute %s (is a directory)", path);
        protocol_status(sn, rq, PROTOCOL_NOT_FOUND, NULL);
        return SYS_NET_ERRORFD;
    }
    if(!util_can_exec(finfo, geteuid(), getegid())) {
        log_error(LOG_FAILURE, "send-cgi", sn, rq, 
                  "cannot execute CGI script %s (permission denied)", path);
        return SYS_NET_ERRORFD;
    }

    if((argv0 = strrchr(path,'/')) != NULL)
        argv0++;
    else argv0 = path;

    n = PR_NewTCPSocketPair(c2s);
    if((n != -1) && (PR_NewTCPSocketPair(s2c) == -1)) {
        n = -1;
        PR_Close(c2s[0]); PR_Close(c2s[1]);
    }
    if((n != -1) && (pipe(ex) == -1)) {
        n = -1;
        PR_Close(c2s[0]); PR_Close(c2s[1]);
        PR_Close(s2c[0]); PR_Close(s2c[1]);
    }        
#ifdef MCC_ADMSERV
    /* MLM - create the pipe for the password stuff */
    if((n != -1) && (pipe(pwPipe) == -1))  {
        n = -1;
        PR_Close(c2s[0]); PR_Close(c2s[1]);
        PR_Close(s2c[0]); PR_Close(s2c[1]);
        close(ex[0]);  close(ex[1]);
    }
#endif
    if(n == -1) {
        log_error(LOG_CATASTROPHE, "send-cgi", sn, rq, 
                  "could not create IPC pipe (%s)", system_errmsg());
        return SYS_NET_ERRORFD;
    }
    /* ex is a close-on-exec pipe which we use to see if the exec worked. */
    fcntl(ex[0], F_SETFD, 1);
    fcntl(ex[1], F_SETFD, 1);

    /* 
     * PR_NewTCPSocketPair()/PR_ImportFile() make both ends of the pipes
     * nonblocking, which is bad for the CGI program's end.  Undo this now, 
     * because Solaris may hang if we try it after the fork. - mikep
     */
    if( (n = fcntl(PR_FileDesc2NativeHandle(c2s[1]), F_GETFL, 0)) > -1)
	 fcntl(PR_FileDesc2NativeHandle(c2s[1]), F_SETFL, n & (~O_NONBLOCK));
    if( (n = fcntl(PR_FileDesc2NativeHandle(s2c[0]), F_GETFL, 0)) > -1)
	 fcntl(PR_FileDesc2NativeHandle(s2c[0]), F_SETFL, n & (~O_NONBLOCK));
#ifdef MCC_ADMSERV
    pwPipe_prfd[0] = PR_ImportFile(pwPipe[0]);
    pwPipe_prfd[1] = PR_ImportFile(pwPipe[1]);
    if( (n = fcntl(pwPipe[1], F_GETFL, 0)) > -1) /* see above */
	 fcntl(pwPipe[1], F_SETFL, n & (~O_NONBLOCK));
#endif
    env = http_hdrs2env(rq->headers);
#ifdef MCC_ADMSERV
    /* MLM - add pipe info */
    env = admin_cgivars(sn, rq, ppid, pwPipe_prfd[0], env);
#endif
    env = cgi_common_vars(sn, rq, env);
    env = cgi_specific_vars(sn, rq, args, env);
    env = cgi_ssi_vars(sn, rq, args, env);
    if(initenv)
	env = pblock_pb2env(initenv, env);

    childparms.path = path;
    childparms.s2c[0] = s2c[0];
    childparms.s2c[1] = s2c[1];
    childparms.c2s[0] = c2s[0];
    childparms.c2s[1] = c2s[1];
    childparms.ex[0] = ex[0];
    childparms.ex[1] = ex[1];
    childparms.csd = sn->csd;
    childparms.args = args;
    childparms.argv0 = argv0;
    childparms.env = env;
    
    if((*pid = create_child(&childparms)) < 0) {
        log_error(LOG_CATASTROPHE, "send-cgi", sn, rq, 
                  "could not fork new process (%s)", system_errmsg());
	util_env_free(env);
	PR_Close(c2s[0]); PR_Close(c2s[1]);
	PR_Close(s2c[0]); PR_Close(s2c[1]);
	close(ex[0]); close(ex[1]);
#ifdef MCC_ADMSERV 
	/* MLM password pipe close too */
	close(pwPipe[0]); close(pwPipe[1]);
#endif
	return SYS_NET_ERRORFD;
    }

    util_env_free(env);

    PR_Close(c2s[1]);
    PR_Close(s2c[0]);
    close(ex[1]);

    n = read(ex[0], &c, 1);
    close(ex[0]);

    if(n == 1) {
#if defined(SOLARIS)
        /* Under Solaris kernel threads, post-fork1() the locks are in an 
           undefined state. strftime() needs a lock. */
        log_error(LOG_FAILURE, "send-cgi", sn, rq, "can't exec %s", path);
#endif
        /* execution failed */
        WAITPID(*pid, NULL, 0);
        PR_Close(s2c[1]);
        PR_Close(c2s[0]);
        return SYS_NET_ERRORFD;
    }
    n = cgi_send_content(sn, rq, s2c[1]);
    PR_Close(s2c[1]);
#ifdef MCC_ADMSERV
    /* MLM - send the password to the CGI program */
    n = adm_cgi_send_password(sn, rq, pwPipe[1]);
    PR_Close(pwPipe_prfd[0]);
    PR_Close(pwPipe_prfd[1]);
#endif
    if(n == REQ_ABORTED) {
        PR_Close(c2s[0]);
	/* Now wait for child to die... */
        WAITPID(*pid, NULL, 0);
        return SYS_NET_ERRORFD;
    }

    return c2s[0];
}


/* --------------------------- cgi_parse_output --------------------------- */


/*
 * Returns -1 on error, 0 for proceed normally, and 1 for dump output and 
 * restart request.
 */

int cgi_parse_output(netbuf *buf, Session *sn, Request *rq)
{
    char *s, *l;

    if(cgi_scan_headers(sn, rq, buf) == REQ_ABORTED)
        return -1;

    l = pblock_findval("location", rq->srvhdrs);

    if((s = pblock_findval("status", rq->srvhdrs))) {
        if((strlen(s) < 3) || (!isdigit(s[0]) || (!isdigit(s[1])) || 
                                                           (!isdigit(s[2]))))
        {
            log_error(LOG_WARN, "send-cgi", sn, rq, 
                      "ignoring CGI status line %s (no 3-digit code)",s);
            s = NULL;
        }
        else pblock_nvinsert("status", s, rq->srvhdrs);
    }
    if(l && (!util_is_url(l)))
        return 1;

    if(!s) {
        if (l)
            pblock_nvinsert("url", l, rq->vars);
        protocol_status(sn, rq, (l ? PROTOCOL_REDIRECT : PROTOCOL_OK), NULL);
    }

    return 0;
}


/* ------------------------------- cgi_send ------------------------------- */


/*
 * Execute a CGI script and return its output to the client. A Service 
 * class function.
 */

int cgi_send(pblock *pb, Session *sn, Request *rq)
{
    char *t, *path = pblock_findval("path", rq->vars), num[16];
    pid_t pid = 0;
    SYS_NETFD cfd, tfd = 0;
    netbuf *buf;
    pb_param *pp;
    int restart = 0, ns;

    if((pp = pblock_remove("content-type", rq->srvhdrs)))
        param_free(pp);

    if((cfd = cgi_start_exec(sn, rq, &pid, path)) == SYS_NET_ERRORFD)
        return REQ_ABORTED;

    cgiwatch_add_client(sn, rq, pid);

    buf = netbuf_open(cfd, NET_BUFFERSIZE);
    buf->rdtimeout = cgi_rdtimeout;
    t = strrchr(path, '/');

    if(t && (!strncmp(t+1, "nph-", 4))) {
        tfd = sn->csd;
        restart = 0;
    }
    else {

        switch(cgi_parse_output(buf, sn, rq)) {
          case -1:
            net_close(buf->sd);
            netbuf_close(buf);
	    cgiwatch_remove_client(sn, rq, pid);
            WAITPID(pid, NULL, 0);
            return REQ_ABORTED;
          case 0:
            if(protocol_start_response(sn, rq) == REQ_NOACTION)
                tfd = SYS_NET_ERRORFD;
            else
                tfd = sn->csd;
            restart = 0;
            break;
          case 1:
            tfd = SYS_NET_ERRORFD;
            restart = 1;
        }
    }
    if( (ns = netbuf_buf2sd(buf, tfd, -1)) == IO_ERROR) {
        if(errno != EPIPE) {
            log_error(LOG_FAILURE, "send-cgi", sn, rq, 
                      "error occurred while sending script output (%s)",
                      buf->errmsg);
            net_close(buf->sd);
            netbuf_close(buf);
	    cgiwatch_remove_client(sn, rq, pid);
            WAITPID(pid, NULL, 0);
            return REQ_ABORTED;
        }
    }
    else {
        util_itoa(ns, num);
        pblock_nvinsert("content-length", num, rq->srvhdrs);
    }
    net_close(buf->sd);
    netbuf_close(buf);

    cgiwatch_remove_client(sn, rq, pid);
    WAITPID(pid, NULL, 0);
    if(restart) {
        pb_param *l = pblock_remove("location", rq->srvhdrs);

        request_restart_internal(l->value, rq);
        param_free(l);
        return REQ_RESTART;
    }
    return REQ_PROCEED;
}


/* ------------------------------ cgi_query ------------------------------- */


int cgi_query(pblock *pb, Session *sn, Request *rq)
{
    pb_param *rqpath = pblock_find("path", rq->vars);
    char *path = pblock_findval("path", pb);

	if ( CM_get_status() )  {     /* if cm is on */
	  char *queryStr = pblock_findval("query", rq->reqpb);
	  if ( queryStr && CM_is_cm_query( queryStr ) ) 
		return REQ_NOACTION;    /* let CM handle this */
	}

    FREE(rqpath->value);
    rqpath->value = STRDUP(path);

    if(!(pblock_findval("path-info", rq->vars)))
        pblock_nvinsert("path-info", pblock_findval("uri", rq->reqpb), 
                        rq->vars);

    return cgi_send(NULL, sn, rq);
}


/* ------------------------------- cgi_init ------------------------------- */


#ifdef XP_UNIX
/* XXXrobm The getenvs don't belong here but it was convenient */
extern "C" void cgi_grabstds(void)
{
    int fd, dfd;

    _cgi_env_path = getenv("PATH");
    _cgi_env_tz = getenv("TZ");
    _cgi_env_lang = getenv("LANG");

    if( (fd = open("/dev/null", O_RDONLY)) == -1) {
        fprintf(stderr, "error: could not open /dev/null (%s)\n", 
                system_errmsg());
        exit(1);
    }
    dfd = 0;
    while(dfd < 3) {
        if( (dfd = fcntl(fd, F_DUPFD, 0)) == -1) {
            fprintf(stderr, "error: could not dup file descriptor (%s)\n",
                    system_errmsg());
            exit(1);
        }
    }
    close(dfd);
    if(fd > 2)
        close(fd);
}
#endif

void cgi_terminate(void *unused)
{
    if(initenv) {
        pblock_free(initenv);
        initenv = NULL;
    }
}

int cgi_init(pblock *pb, Session *sn, Request *rq)
{
    pb_param *timeout;

    param_free(pblock_remove("server-version", pb));
    if(!initenv)
        initenv = pblock_create(4);
    pblock_copy(pb, initenv);

    timeout = pblock_remove("timeout", initenv);
    if(timeout) {
        cgi_rdtimeout = atoi(timeout->value);
        param_free(timeout);
    }
    param_free(pblock_remove("fn", initenv));

#if 0
    magnus_atrestart(cgi_terminate, NULL);
#endif
    return 0;
}


#ifdef FORTEZZA

int service_preencrypted(pblock *param, Session *sn, Request *rq)
{
    PEHeader *header;
    char *path, *t;
    SYS_FILE fd;
    int l;
    filebuf_t *buf;
    struct stat *finfo;
    char tmpbuf[128];
    int ret;

    path = pblock_findval("path", rq->vars);
    if( (t = pblock_findval("path-info", rq->vars)) ) {
        protocol_status(sn, rq, PROTOCOL_NOT_FOUND, NULL);
        log_error(LOG_WARN, "send-file", sn, rq,
                  "%s%s not found", path, t);
        return REQ_ABORTED;
    }
    if(!(finfo = request_stat_path(path, rq))) {
        protocol_status(sn, rq, (rtfile_notfound() ? PROTOCOL_NOT_FOUND :
                                 PROTOCOL_FORBIDDEN), NULL);
        log_error(LOG_WARN, "send-file", sn, rq, "can't find %s (%s)",
                  path, rq->staterr);
        return REQ_ABORTED;
    }

    if(protocol_set_finfo(sn, rq, finfo) == REQ_ABORTED)
        return REQ_ABORTED;

    if((fd = system_fopenRO(path)) == SYS_ERROR_FD) {
        protocol_status(sn, rq, (file_notfound() ? PROTOCOL_NOT_FOUND :
                                 PROTOCOL_FORBIDDEN), NULL);
        log_error(LOG_WARN, "send-file", sn, rq, "error opening %s (%s)",
                  path, system_errmsg());
        return REQ_ABORTED;
    }

    if(!(buf = filebuf_open_nostat(fd, FILE_BUFFERSIZE, finfo))) {
        system_fclose(fd);
        protocol_status(sn, rq, PROTOCOL_FORBIDDEN, NULL);
        log_error(LOG_WARN,"send-file", sn, rq,
                  "error opening buffer from %s (%s)", path,
              system_errmsg());
        return REQ_ABORTED;
    }

    pblock_nvinsert("accept-ranges", "bytes", rq->srvhdrs);
    if(!(pblock_find("status", rq->srvhdrs)))
        protocol_status(sn, rq, PROTOCOL_OK, NULL);

#ifdef notdef
    /* Turn off caching for pre-encrypted files until we get the
     * Navigators working right */
    pblock_nvinsert("Pragma", "no-cache", rq->srvhdrs);
#endif

    if(protocol_start_response(sn,rq) != REQ_NOACTION) {

        if (filebuf_grab(buf,sizeof(tmpbuf)) == IO_EOF) {
            filebuf_close(buf);
            return REQ_EXIT;
        }
        memcpy(tmpbuf,(char *)buf->inbuf,buf->cursize);
        /*
         * this is broken, but it works because the only version of
         * SSL_PreencryptedFileToStream will actually return tmpbuf..
         * We need to fix this later...... */
        header = SSL_PreencryptedFileToStream(sn->csd->osfd,
                                                 (PEHeaderStr *)tmpbuf);
        if (header == NULL) {
            filebuf_close(buf);
            return REQ_EXIT;
        }
        if (net_write(sn->csd, (char *)header, buf->cursize) == IO_ERROR) {
            filebuf_close(buf);
            return REQ_EXIT;
        }
        while (filebuf_grab(buf,NET_BUFFERSIZE) != IO_EOF) {
            if (net_write(sn->csd,(char *)buf->inbuf,buf->cursize)==IO_ERROR) {
                filebuf_close(buf);
                return REQ_EXIT;
            }
        }
    }
    filebuf_close(buf);

    return REQ_PROCEED;
}

#endif /* FORTEZZA */

