/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * cgiwatch.cpp: Handle client disconnects when running long cgi's
 *
 * These routines are used by the cgi-watchdog thread which will sense
 * when a client hits stop and kill the associated CGI app.
 *
 * Mike Patnode
 */

#include "safs/cgi.h"
#include "base/systems.h"
#include "base/session.h"
#include "base/crit.h"
#include "base/systhr.h"
#include "frame/conf.h"
#include "frame/log.h"

/* NSPR21 19971028 move this from prio.h to private/pprio.h
   It may be removed later .... Keep you finger cross.... */
#include "private/pprio.h"

#include <signal.h>

static cgi_info *watch_list_head;
static CRITICAL watch_lock;
static watch_timeout;

PRPollDesc *poll_desc = NULL;
int poll_index = 0;

#ifdef XP_WIN32
void wait_for_child_death(HANDLE pid, int wait_flag, int terminate_flag);
#endif

void add_poll_desc(PRFileDesc *fd, PRPollDesc **poll_desc) {

   int i;

   if(!fd) return;
   for(i=0; i < poll_index; i++) {
      if(((*poll_desc)[i].fd) && ((*poll_desc)[i].fd = fd))
	/* alread there - we're done */
	return;
   }

   if(poll_index == 0)
      *poll_desc = (PRPollDesc *)MALLOC(sizeof(PRPollDesc));
   else
      *poll_desc = (PRPollDesc *)REALLOC(*poll_desc, (poll_index+1)*sizeof(PRPollDesc));

   poll_index++;
   (*poll_desc)[poll_index-1].fd = fd;
   (*poll_desc)[poll_index-1].in_flags = PR_POLL_READ;
}

void remove_poll_desc(PRFileDesc *fd, PRPollDesc **poll_desc) {

   int i;

   if(!fd) return;
   for(i=0; i < poll_index; i++) {
      if((*poll_desc)[i].fd = fd) {
	/* remove it */
	(*poll_desc)[i].fd = NULL;
	(*poll_desc)[i].in_flags = 0;
	return;
      }
   }
}

void clear_poll_desc(PRPollDesc *poll_desc) {

   if(poll_desc)
     FREE(poll_desc);
   poll_index = 0;
}

int exists_poll_desc(PRFileDesc *fd, PRPollDesc *poll_desc) {

   int i;

   if(!fd) return 0;
   for(i=0; i < poll_index; i++) {
      if((poll_desc[i].fd) && (poll_desc[i].fd = fd))
	return 1;
   }
   return 0;
}

void
cgiwatch_add_client(Session *sn, Request *rq, pid_t pid)
{
    cgi_info *ci;

    if (!watch_timeout) return;

    ci = (cgi_info *)MALLOC(sizeof(cgi_info));
    ci->sn = sn;
    ci->rq = rq;
    ci->pid = pid;

    crit_enter(watch_lock);
    ci->next = watch_list_head;
    watch_list_head = ci;
    add_poll_desc(sn->csd, &poll_desc);
    crit_exit(watch_lock);
}

void
cgiwatch_remove_client(Session *sn, Request *rq, pid_t pid)
{
    cgi_info *ci, *prev = NULL;

    if (!watch_timeout) return;

    crit_enter(watch_lock);
    remove_poll_desc(sn->csd, &poll_desc);

    for (ci = watch_list_head; ci && ci->pid != pid; ci = ci->next)
	prev = ci;

    if (!ci) /* list corruption? */
	{
	log_error(LOG_WARN, "cgiwatch-remove-client", ci->sn, ci->rq, 
                      "pid (%d) not found", pid);
	crit_exit(watch_lock);
	return;
	}

    if (prev) 
	prev->next = ci->next;
    else
	watch_list_head = ci->next;
    crit_exit(watch_lock);

    FREE(ci);
}

/* ------------------------- cgi_watch_client ------------------------ */

/*
 * Handle client disconnects when running long cgi's.  Create
 * a new thread which select's on a list of client connections for reads. 
 * If select returns ready and there are 0 bytes, the client has hit Stop.
 */

static void
cgiwatch_check_clients(void *unused)
{
    int ready;
    cgi_info *ci;

    if (!(watch_timeout = conf_getglobals()->cgiwatch_timeout)) return;

    clear_poll_desc(poll_desc);

    watch_lock = crit_init();

    while (1)
	{

	crit_enter(watch_lock);
	crit_exit(watch_lock);
/*
	ready = PR_Select(0, &rd, NULL, NULL, watch_timeout);
*/
	ready = PR_Poll(poll_desc, poll_index, watch_timeout);

	/*
	 * It's possible the whole transaction finished prior to us entering
	 * select.  In that case we get EBADF.  cgi_send() will of removed
	 * the bad fd from the watch_clients set, so we just ignore this.
	 */
#ifdef XP_WIN32
	if (ready < 0)
	    {
	    log_error(LOG_WARN, "cgi-watch-client", NULL, NULL, 
			  "error occurred in select (%d)",
			  WSAGetLastError());
	    /* Whatever went wrong, don't let it repeat */
	    clear_poll_desc(poll_desc);
	    }
#else
	if (ready < 0 && errno != EBADF && errno != EINTR) 
	    {
	    log_error(LOG_WARN, "cgi-watch-client", NULL, NULL, 
			  "error occurred in select (%s)",
			  system_errmsg());
	    /* Whatever went wrong, don't let it repeat */
	    clear_poll_desc(poll_desc);
	    }
#endif
	else if (ready > 0)
	    {
	    /*
	     * Now map a fd returned from select back to
	     * the right cgi_info struct
	     */
	    crit_enter(watch_lock);
	    for (ci = watch_list_head; ci; ci = ci->next)
		{
		if (exists_poll_desc(ci->sn->csd, poll_desc) && !net_isalive(ci->sn->csd))
		    {
		    log_error(LOG_VERBOSE, "cgi-watch-client", ci->sn, ci->rq, 
			      "client dropped connection, killing %d", ci->pid);
#if defined(XP_UNIX)
		    kill(ci->pid, SIGHUP);
#elif defined(XP_WIN32)
		    /* This doesn't work very well */
		    (void)wait_for_child_death(ci->pid, 0, 1);
#else
#error "writeme: How to kill a CGI?"
#endif
		    /* Do this now, else select will just return again */
		    remove_poll_desc(ci->sn->csd, &poll_desc);
		    }
		}
	    crit_exit(watch_lock);
	    }
	}
}

void
cgiwatch_internal_init()
{
    (void)systhread_start(0, 0, cgiwatch_check_clients, NULL);
}
