/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * Implementation for client authentication in a secure server.
 * 
 */

#if defined(CLIENT_AUTH)

#define __PRIVATE_NSDB
#define __PRIVATE_NSADB

#include <netsite.h>

#include <nspr.h>
#include <prio.h>

/* removed for ns security integration
#include <sec.h>
*/
extern "C" {
#include <key.h>
#include <cert.h>
#include <base64.h>
#include <ssl.h>
}

/*
#include <xp_error.h>
#include <xp_file.h>
*/
/* removed for ns security integration
#include "sslio/sslio.h"
*/

#include "frame/req.h"
#include "frame/log.h"
#include "frame/http.h"
#include "frame/conf.h"
#include "base/pblock.h"
#include "base/session.h"
#include "base/util.h"
#include "base/shexp.h"
#include <ldaputil/certmap.h>
#include <ldaputil/ldapdb.h>
#include <ldaputil/errors.h>
#include <libaccess/nsdb.h>
#include <libaccess/nsadb.h>
#include <libaccess/nsauth.h>
#include <libaccess/nsuser.h>
#include <libaccess/nscert.h>
#include <libaccess/acl.h>
#include <libaccess/aclproto.h>
#include <libaccess/las.h>
#include <libaccess/ldapacl.h>
#include "safs/clauth.h"
#include "safs/init.h"

/* Want enum string ids to allow a later array initializer to work */
#ifndef WANT_ENUM_STRING_IDS
#define WANT_ENUM_STRING_IDS 1
#endif
/* removed for ns security integration, looks like all the symbols are now defined in secerr.h
#include <allxpstr.h>
*/
#include <secerr.h>
#include <sslerr.h>

#include <string.h>
#include <sys/stat.h>
#include <stdio.h>

#define BIG_LINE 1024

extern int _cmpCryptPasswd(	    	/* thread safe cover crypt() */
    const char *encrypted, const char* cleartext);

/* prototypes imported from other libraries */

char *_uudecode(char *bufcoded);    /* from auth.c */

/* private prototypes implemented here */

static int ca_GetClientCert(Session* sn, Request* rq);

static void ca_HandshakeDone(PRFileDesc *fd, void *clientData);

static void ca_SetRqVars(pblock *pb, Session *sn, Request *rq);
static int ca_ProcessBasicAuth(pblock *pb, Session *sn, Request *rq);
static void *ca_OpenDatabase(pblock *param, Session* sn, Request *rq);
static void ca_RequestWWWAuth(pblock *param, Session* sn, Request *rq);
static const char *mapSecError(int32 code);


/*
 * Initialization
 * --------------
 *  Assumptions:
 *	Still single threaded (no need to protect)
 */
NSAPI_PUBLIC int CA_Initialize(const char *path, const char *key)
{
    int rv;
    rv = nsadbCertInitialize();
    return rv;
}  /* CA_Initialize */


/*
 * Getting a Cert
 * --------------
 * Get the current SSL peer certificate, putting it onto
 * sn->clauth->cla_cert. If it is not present, return failure
 * immediately.
 *
 * Return value:
 *    REQ_ABORTED on failure
 *    REQ_PROCEED on success with a CERTCertificate* on sn->clauth->cla_cert.
 */
NSAPI_PUBLIC int CA_GetClientCert(pblock *pb, Session *sn, Request *rq)
{
  ClAuth_t *cla = (ClAuth_t*)sn->clauth;
  /* Removed for ns security integration
  CERTCertificate *cert = SSL_PEERCERTIFICATE(sn->csd);
  */
  CERTCertificate *cert = SSL_PeerCertificate(sn->csd);
  if (cla->cla_cert != NULL)
    CERT_DestroyCertificate(cla->cla_cert);
  cla->cla_cert = cert;
  return (cert == NULL) ? REQ_ABORTED : REQ_PROCEED; 
}  /* CA_GetClientCert */

/*
 * Requesting a Cert
 * ------------------
 *
 * Request the cert, redo the handshake (even if a cert is
 * already present).
 *
 * Return value:
 *    REQ_ABORTED on failure with error code obtainable by XP_GetError()
 *    REQ_PROCEED on success with a SECCertificate* on sn->clauth->cla_cert.
 */
NSAPI_PUBLIC int CA_RequestClientCert(
	pblock *pb, Session *sn, Request *rq)
{
	int rv = ca_GetClientCert(sn, rq);
	if (rv == REQ_PROCEED)
	{
		rv = CA_GetClientCert(pb, sn, rq);
	}
	return rv;
}  /* CA_RequestClientCert */



/*
 * Mapping a cert
 * --------------
 *
 * Map the cert on sn->clauth_t->cla_cert using the "userdb" in pb
 * and put the result onto the user object on sn->clauth.
 *
 * IF NO MAPPING EXISTS return an appropriate error code, don't do
 * basic auth automatically.
 *
 * Return values:
 *    REQ_ABORTED on failure with an error code obtainable by XP_GetError()
 *    REQ_PROCEED on success with a UserObj_t* on sn->clauth_t->cla_uoptr
 */
NSAPI_PUBLIC int CA_MapCertToUser(
	pblock *pb, Session *sn, Request *rq)
{
    char *user = NULL;
	ClAuth_t *cla = (ClAuth_t*)sn->clauth;
    void *authdb = ca_OpenDatabase(pb, sn, rq);
    int rv = (authdb == NULL) ? REQ_ABORTED : REQ_PROCEED;

	/* Free any previous user object */
	if (cla->cla_uoptr != NULL) {
	    userFree(cla->cla_uoptr);
	    cla->cla_uoptr = NULL;
	}

    if (authdb != NULL)
    {
    	(void)nsadbGetUserByCert(0, authdb, cla->cla_cert, &user);

    	/* seems like there should be more to say here */
    	if (user == NULL)  /* that means no mapping */
    	{
	    /*  Let the caller log as appropriate.
    	 *  log_error(LOG_WARN, CA_PROGRAM, sn, rq,
	     *		  "No mapping for certificate to a user");
	     */
	    rv = REQ_ABORTED;
    	}
    	else if (nsadbFindByName(NULL, authdb, user, AIF_USER,
				 (void**)&cla->cla_uoptr) != AIF_USER)
    	{
    		/* cert maps to a user, but the user doesn't exist */
			log_error(LOG_MISCONFIG, CA_PROGRAM, sn, rq, 
				"Certificate maps to non-existent user: %s", user);
			rv = REQ_ABORTED;
		}
		nsadbClose(authdb, 0);
	}
	return rv;
}  /* CA_MapCertToUser */


/*
 * Full featured function
 * ----------------------
 * The goal here is to find the "user name" of the client
 * involved in the current secure session. That is done by
 * matching his current certificate with out data base and
 * ferreting out the user name.
 */
NSAPI_PUBLIC int CA_GetUserID(
    pblock *param, Session *sn, Request* rq)
{
    int rv;
    /*
     * Ask libsec for the certificate of the client of this
     * session. If one is returned, libsec has already performed
     * rudimentary validation on that certificate.
     */
    rv = CA_GetClientCert(param, sn, rq);
    if (rv == REQ_ABORTED) {
      /*
       * If no certificate is available, make SSL redo
       * the handshake and get one. If that completes
       * successfully, then there should be a certificate.
       */
      rv = CA_RequestClientCert(param, sn, rq);
      if (rv == REQ_ABORTED) {
	/* removed for ns security integration
	SSL_INVALIDATESESSION(sn->csd);
	*/
	SSL_InvalidateSession(sn->csd);
	protocol_status(sn, rq, PROTOCOL_FORBIDDEN, NULL);
	log_error(LOG_SECURITY, CA_PROGRAM, sn, rq, 
		  "no client certificate available");
	return REQ_ABORTED;
      }
    }

	/*
	 * Try to map this cert to a user. If that works,
	 * we finished early and can go home.
	 */
	rv = CA_MapCertToUser(param, sn, rq);
	/*
		If we haven't succeeded by this time, then we need to
		create a new mapping. That might require another round
		trip at the http level.
	*/
	if (rv != REQ_PROCEED) {
		rv = ca_ProcessBasicAuth(param, sn, rq);
        if (rv != REQ_PROCEED) {
	  /* removed for ns security integration
	     SSL_INVALIDATESESSION(sn->csd);
	     */
	  SSL_InvalidateSession(sn->csd);
        }
    }

	if (rv == REQ_PROCEED) ca_SetRqVars(param, sn, rq);

    return rv;
}  /* CA_GetUserID */


static int ca_PutUserByCert(const char *userdb, void *authdb,
			    CERTCertificate *cert, const char *user)
{
    int rv;
    rv = nsadbPutUserByCert(0, authdb, cert, user);

    /* See if mapping was created successfully */
    if (!rv) {
	/* Report the certificate registration */
	ereport(LOG_INFORM, "registered client certificate for user %s in "
		"database %s", user, userdb);
    }

    return rv;
}

static int ca_ProcessBasicAuth(pblock *pb, Session *sn, Request *rq)
{
    int rv;
    void * authdb;
    char * auth = 0;
    char * user = 0;
    UserObj_t * uo = 0;
    char * password = 0;
    CERTCertificate * cert;
    ClAuth_t * cla = (ClAuth_t *)sn->clauth;
    int log = LOG_FAILURE;
    const char * format = 0;
    int httpstatus = PROTOCOL_SERVER_ERROR;

    /* Make sure we have a client certificate at this point */
    if (!cla || !(cert = cla->cla_cert)) goto err_cert;

    /* Request HTTP authorization if we haven't got that info yet */
    auth = pblock_findval("authorization", rq->headers);
    if (auth == NULL) {
	ca_RequestWWWAuth(pb, sn, rq);  /* ask for WWW-Auth */
	return REQ_ABORTED;  /* presumably this will be retried */
    }

    /* XXX ANIL POINTS OUT THAT THIS CODE IS DUPLICATED IN auth.c */

    /*
     * We have authorization in the request, hopefully basic.
     */

    /* Skip leading blanks in authorization info */
    while(*auth && (*auth == ' ')) ++auth;

    /* Anything left? */
    if(!(*auth)) goto err_auth;

    /* Is this "basic" authentication info? */
    if((strlen(auth) >= 6) && !strncasecmp(auth, "basic ", 6)) {
	for (auth += 6; *auth != '\0'; ++auth) {
	    if (*auth != ' ') {
		/* Note: decoded user string is allocated by _uudecode */
		user = _uudecode(auth);  /* decode it */
		if (user != NULL) {
		    password = strchr(user, ':');
		    if (password) *password++ = '\0';
		}
		break;
	    }
	}
    }

    /* Both username and password must be specified */
    if (user == 0) goto err_user;
    if (password == 0) goto err_pass;

    /* XXX ANIL POINTS OUT THAT THIS CODE IS DUPLICATED IN auth.c */

    /* Open the authentication database specified in the parameter block */
    authdb = ca_OpenDatabase(pb, sn, rq);
    if (authdb == 0) goto err_open;

    /* Look up the specified user in the database */
    if ((nsadbFindByName(0, authdb, user, AIF_USER, (void**)&uo) != AIF_USER)
	|| (uo == 0)) {
	nsadbClose(authdb, 0);
	goto err_dbusr;
    }

    /* Check for the correct password */
    if (_cmpCryptPasswd((const char *)uo->uo_pwd, password) != 0) {
	nsadbClose(authdb, 0);
	userFree(uo);
	uo = NULL;
	goto err_badpwd;
    }

    /*
     * The user is authorized and the certificate is assumed good
     * at this point.  Now create a mapping from the certificate
     * to the user, i.e. register the certificate.  Note that this
     * operation acquires a global lock that is not released until
     * the authentication database is closed.
     */
    rv = ca_PutUserByCert(pblock_findval("userdb", pb), authdb,
			  cert, user);
    nsadbClose(authdb, 0);

    /* See if mapping was created successfully */
    if (rv) {
	userFree(uo);
	if (rv == NSAERRCMAP) goto err_cmap;
	goto err_maperr;
    }

    /* Replace pointer to the latest user object */
    if (cla->cla_uoptr != 0) {
	userFree(cla->cla_uoptr);
    }
    cla->cla_uoptr = uo;

    /* "user" is always allocated by _uudecode(), right? */
    FREE(user);

    /* All has gone well */
    return REQ_PROCEED;

  punt:					/* Common error processing */

    /* If we're returning PROTOCOL_UNAUTHORIZED request basic auth again */
    if (httpstatus == PROTOCOL_UNAUTHORIZED) {
	ca_RequestWWWAuth(pb, sn, rq);
    }
    else {
	/* Otherwise just set the indicated HTTP status */
	protocol_status(sn, rq, httpstatus, NULL);
    }

    if (format != 0) {
	log_error(log, CA_PROGRAM, sn, rq,
		  " %s (name: %s in database %s)", format,
		  (user == NULL) ? "unknown" : user,
		  pblock_findval("userdb", pb));
    }

    /* "user" is always allocated by _uudecode(), right? */
    if (user) {
	FREE(user);
    }

    return REQ_ABORTED;

  err_cert:
    /* No client certificate */
    httpstatus = PROTOCOL_SERVER_ERROR;
    format = "client certificate missing";
    log = LOG_FAILURE;
    goto punt;

  err_auth:
    /* HTTP authorization header contains no information */
    httpstatus = PROTOCOL_UNAUTHORIZED;
    format = "empty HTTP authorization header";
    log = LOG_SECURITY;
    goto punt;

  err_user:
    /* No user name in HTTP authorization header */
    httpstatus = PROTOCOL_UNAUTHORIZED;
    format = "missing user name";
    log = LOG_SECURITY;
    goto punt;

  err_pass:
    /* No password in HTTP authorization header */
    httpstatus = PROTOCOL_UNAUTHORIZED;
    format = "missing password";
    log = LOG_SECURITY;
    goto punt;

  err_open:
    /* Error opening authentication database */
    httpstatus = PROTOCOL_SERVER_ERROR;
    format = "error opening authentication database";
    log = LOG_FAILURE;
    goto punt;

  err_dbusr:
    /* The specified user name was not found in the authentication database */
    httpstatus = PROTOCOL_UNAUTHORIZED;
    format = "user not found";
    log = LOG_SECURITY;
    goto punt;

  err_badpwd:
    /* An incorrect password was specified */
    httpstatus = PROTOCOL_UNAUTHORIZED;
    format = "invalid password for user";
    log = LOG_SECURITY;
    goto punt;

  err_cmap:
    /* A certificate is already registered for the indicated user. */
    httpstatus = PROTOCOL_FORBIDDEN;
    format = "user already has a registered certificate";
    log = LOG_SECURITY;
    goto punt;

  err_maperr:
    /* Failed to create the certificate-to-user mapping */
    httpstatus = PROTOCOL_SERVER_ERROR;
    format = "certificate registration failed for user";
    log = LOG_FAILURE;
    goto punt;

}  /* ca_ProcessBasicAuth */


static void *ca_OpenDatabase(
	pblock *pb, Session* sn, Request *rq)
{
    void *authdb = NULL;
    char *authdbName = pblock_findval("userdb", pb);
    if ((authdbName == NULL)
    || (nsadbOpen(NULL, authdbName, 0, &authdb) < 0))
    {
		log_error(
	    	LOG_MISCONFIG, CA_PROGRAM, sn, rq, 
	    	"cannot open database %s",
	    	(authdbName == NULL) ? "unknown" : authdbName);
    }
	return authdb;
}  /* ca_OpenDatabase */

static void ca_RequestWWWAuth(
	pblock *param, Session* sn, Request *rq)
{
	char buf[256];
	char* realm = pblock_findval("realm", param);

	if (!realm) realm = "";

	util_snprintf(buf, sizeof(buf), "Basic realm=\"%s\"", realm);

	pblock_nvinsert("WWW-authenticate", buf, rq->srvhdrs);
	protocol_status(sn, rq, PROTOCOL_UNAUTHORIZED, NULL);
}  /* ca_RequestWWWAuth */

static void ca_SetRqVars(pblock *pb, Session *sn, Request *rq)
{
	/* Provide authentication info to subsequent directives */
	ClAuth_t *cla = (ClAuth_t*)sn->clauth;
	pblock_nvinsert("auth-type", "ssl", rq->vars);
	pblock_nvinsert("auth-user", (char *)cla->cla_uoptr->uo_name, rq->vars);
	pblock_nvinsert(
		"auth-db", pblock_findval("userdb", pb), rq->vars);
}  /* ca_SetRqVars */

/***********************************************************/

/*
 * The following section is support for redoing the SSL
 * Handshake in order to acquire a client certificate.
 * 
 * The particularly messy part is reading the client data
 * to get it out of the way so SSL can use the transport to
 * run its protocol. That data is read into VM buffers, then
 * pumped back into httpd's buffer when it's all over. It is
 * more important to be able to finish the handshake than to
 * successfully capture all the client data.
 * 
 * If the handshake completes, but for some reason the user
 * data cannot be replayed, the operation can be aborted and
 * retried. The next time around, the certificate will already
 * be on board and we won't have to redo the handshake. Sure,
 * that's far from optimal, but it is progress.
 */

enum {ca_allocate=500};		/* allocation quantum */

typedef struct CA_ReadBuffer_s
{
    struct CA_ReadBuffer_s *next;
    uint8 data[ca_allocate];
} CA_ReadBuffer;

typedef struct CA_HandshakeData_s
{
    int done;
    uint32 total, position;
    CA_ReadBuffer *head, *tail;
} CA_HandshakeData;

static void ca_HandshakeDone(PRFileDesc *fd, void *clientData)
{
    /*
     * This routine is called by SSL when the handshake is
     * complete. It sets local state so that the loop doing
     * reads from the socket can exit.
     */
    ((CA_HandshakeData*)clientData)->done = 1;
}  /* ca_HandshakeDone */

static int ca_GetClientCert(Session* sn, Request* rq)
{
    int rv = REQ_PROCEED;  /* some return value */
    unsigned char *to = NULL;  /* where we copy data to */
    
    CA_HandshakeData *clientData = (CA_HandshakeData*)MALLOC(
		sizeof(struct CA_HandshakeData_s));
    clientData->done = 0;  /* will be set true by callback*/
    clientData->total = 0;  /* total # bytes cubby-holed away */
    clientData->position = ca_allocate;  /* within this structure */
    clientData->head = clientData->tail = NULL;  /* list thereof */

    /*
     * Set up the machinery to redo the handshake. If all that succeeds,
     * then initiate the handshake as well.
     */
    /* rewrite for ns security integration
    if (SSL_HANDSHAKECALLBACK(sn->csd, ca_HandshakeDone, clientData)
	|| SSL_ENABLE(sn->csd, SSL_REQUEST_CERTIFICATE,
		      SSL_SECURITY_STATUS_ON_HIGH)
	|| SSL_REDOHANDSHAKE(sn->csd))
	*/
    if (SSL_HandshakeCallback(sn->csd, ca_HandshakeDone, clientData)
	|| SSL_OptionSet(sn->csd, SSL_REQUEST_CERTIFICATE,
		      PR_TRUE)
	|| SSL_ReHandshake(sn->csd, /*FlushCache=*/PR_TRUE))
    {
	int errCode = PR_GetError();
	if (errCode == SEC_ERROR_INVALID_ARGS) {
	    /* Treat this code specially - it means it's not an SSL3 socket */
	    if (!conf_getglobals()->Vsecurity_active ||
		!conf_getglobals()->Vssl3_active) {
		/* SSL3 is not enabled */
		log_error(LOG_MISCONFIG, CA_PROGRAM, sn, rq, 
		  "SSL v3 must be enabled to use client authentication.");
	    } else {
		/* Client couldn't/wouldn't step up to SSL3 */
		log_error(LOG_FAILURE, CA_PROGRAM, sn, rq,
		  "Unable to get certificate. Client is not using SSL v3.");
	    }
	} else {
	    log_error(LOG_FAILURE, CA_PROGRAM, sn, rq, 
		      "SSL library call failed %s (%d)",
		      mapSecError(errCode), errCode);
	}
	return REQ_ABORTED;
    }

    while (!clientData->done) {
      int bytes;
      if (ca_allocate == clientData->position) {
	CA_ReadBuffer *rb = (CA_ReadBuffer*)MALLOC(ca_allocate + sizeof(struct CA_ReadBufferStr*));
	if (clientData->tail != NULL) clientData->tail->next = rb;
	if (clientData->head == NULL) clientData->head = rb;
	rb->next = NULL;
	clientData->tail = rb;
	clientData->position = 0;
      }
      /*
       * We can't use PR_Read() here because it blocks until
       * application data arrives, and there may not be any
       * application data coming.  Use SSL_Recv() instead.
       */

      /* For new ns security, SSL_Recv is no longer supported, change SSL_Recv to PR_Recv */
      bytes = PR_Recv(sn->csd,
		      &clientData->tail->data[clientData->position],
		      (PRInt32)(ca_allocate - clientData->position),
		      (PRIntn)0, 
		      (PRIntervalTime)0);
      if (bytes == 0) {
	log_error(LOG_FAILURE, CA_PROGRAM, sn, rq,
		  "Unexpected end of file. "
		  "Client may have closed connection");
	rv = REQ_ABORTED;  /* an EOF now isn't good */
	break;
      }
      if (bytes < 0) {
	int errCode = PR_GetError();
	/* EWOULDBLOCK is expected and acceptable */
	if (errCode == PR_WOULD_BLOCK_ERROR) continue;
	/* some other error */
	if (IS_SEC_ERROR(errCode) || IS_SSL_ERROR(errCode)) {
	  /* SEC or SSL Error */
	  log_error(LOG_FAILURE, CA_PROGRAM, sn, rq,
		    "Error completing handshake: (%d) %s",
		    errCode, mapSecError(errCode));
	} else if (errCode == PR_CONNECT_RESET_ERROR) {
	  /* Connection reset by peer */
	  log_error(LOG_FAILURE, CA_PROGRAM, sn, rq,
		    "Connection closed by client.");
	} else {
	  log_error(LOG_FAILURE, CA_PROGRAM, sn, rq, 
		    "Unexpected error receiving data: %d",
		    errCode);
	}
	/* get out now */
	rv = REQ_ABORTED;
	break;
      }
      clientData->position += bytes;
      clientData->total += bytes;
    }

	/* Rest the ssl callback machinery */
    (void)SSL_HandshakeCallback(sn->csd, NULL, NULL);
    (void)SSL_GetClientAuthDataHook(sn->csd, NULL, NULL);

    /*
     * If everything's still okay, then copy the bits.
     */
    if ((rv == REQ_PROCEED) && (clientData->total != 0)) {
      /*
       * Append this data into the net buffer of the session.
       * Then it will be consumed by the reads performed by
       * the higher level code.
       */
      netbuf *input = sn->inbuf;  /* get closer */
      if (clientData->total > ((uint32)(input->cursize - input->pos))) {
	/*
	 * Bummer! The buffer isn't big enough to hold
	 * our massive bag of bits. See if it's okay to
	 * reallocate.
	 */
	int needed = input->pos + clientData->total;
	if (needed > input->maxsize) {
	  rv = REQ_ABORTED;
	  log_error(LOG_FAILURE, CA_PROGRAM, sn, rq, 
		    "Cannot buffer client data");
	}
	else {
	  /*
	   * Yes! Yes! Oh, the thrill of it!
	   */
	  input->inbuf = (unsigned char*)REALLOC(input->inbuf, needed);
	}
      }
      to = input->inbuf + input->pos;
    }

    /*
     * If we read any data, slosh it over, but only if the
     * operation succeeded. In any case, delete our linked list.
     * $$$
     * It occurs to me that we could have been reading into the
     * session's input buffer all this time. There might be an
     * opportinity there.
     * For alternate thoughts--
     * In order to make progress, this operation needs to complete.
     * We need to get the client's certificate in our hands. Then
     * even if the operation blows off, when the user retries it
     * the certificate will already be here. I'm concerned that
     * the 'inbuf' is assumed pretty small and may not be able to
     * handle the data we get. What we're looking at could be
     * a few packet's worth, but maybe no more.
     * $$$
     */
    while (clientData->head != NULL)
    {
		CA_ReadBuffer *rb = clientData->head;
		if (to != NULL)
		{
	    	int nbytes = (rb->next == NULL) ?
			clientData->position : ca_allocate;
	    	if (nbytes > 0)
	    	{
				memcpy(to, rb->data, nbytes);
				to += nbytes;
	    	}
		}
		clientData->head = rb->next;
		FREE(rb);
    }
    FREE(clientData);

    return rv;
}  /* ca_GetClientCert */



/*
 * This is a mapping function for SEC and SSL error codes to strings
 * that are reported in the server log for client authentication.
 * Not all error codes may be mapped.  Unmapped codes are mapped to the
 * empty string (const char *) "".  Many of the strings here are
 * context dependent, (i.e. they expect to be used after cert verification
 * fails).
 */
     

typedef struct ErrMapEntryStr { 
    int32 code;
    char *str;
} ErrMapEntry;


static
ErrMapEntry _secErrMap[] = {

    /* Mapped SEC Errors */
    { SEC_ERROR_IO, "I/O Error" },
    { SEC_ERROR_LIBRARY_FAILURE, "Library Failure" },
    { SEC_ERROR_BAD_DATA, "Bad data was received" },
    { SEC_ERROR_INVALID_ALGORITHM,
      "Certificate contains invalid encryption or signature algorithm" },
    { SEC_ERROR_INVALID_TIME, "Certificate contains an invalid time value" },
    { SEC_ERROR_BAD_DER, "Certificate is improperly DER encoded" },
    { SEC_ERROR_BAD_SIGNATURE, "Certificate has invalid signature" },
    { SEC_ERROR_EXPIRED_CERTIFICATE, "Certificate has expired" },
    { SEC_ERROR_REVOKED_CERTIFICATE, "Certificate has been revoked" },
    { SEC_ERROR_UNKNOWN_ISSUER, "Certificate is signed by an unknown issuer" },
    { SEC_ERROR_BAD_KEY, "Invalid public key in certificate." },
    { SEC_ERROR_BAD_DATABASE, "Problem using certificate or key database" },
    { SEC_ERROR_NO_MEMORY, "Out of Memory" },
    { SEC_ERROR_UNTRUSTED_ISSUER,
      "Certificate is signed by an untrusted issuer" },

    /* Mapped SSL Errors */
    { SSL_ERROR_EXPORT_ONLY_SERVER, "Client required high grade encryption" },
    { SSL_ERROR_US_ONLY_SERVER,
      "Client unable to provide high grade encryption" },
    { SSL_ERROR_NO_CYPHER_OVERLAP,
      "Unable to agree with client on an acceptable cypher" },
    { SSL_ERROR_NO_CERTIFICATE,
      "The server is unable to find its own certificate or key" },
    { SSL_ERROR_BAD_CERTIFICATE, "Client's certificate is bad" },
    /* note -- the library sometimes sets the wrong one.  report one string */
    { SSL_ERROR_BAD_CLIENT, "Protocol error." },
    { SSL_ERROR_BAD_SERVER, "Protocol error." },
    { SSL_ERROR_UNSUPPORTED_CERTIFICATE_TYPE, "Unsupported certificate type" },
    { SSL_ERROR_UNSUPPORTED_VERSION, "Unsupported SSL version" },
    { SSL_ERROR_WRONG_CERTIFICATE,
      "The public key in the server's own certificate does not match its "
      "private key" },

    /* Trailer -- DO NOT REMOVE */
    { 0 }
};



static
const char *
mapSecError(int32 code)
{
    int32 c;
    int i=0;
    for (i=0; (c = _secErrMap[i].code) != 0; i++) {
	if (code == c) {
	    return _secErrMap[i].str;
	}
    }
    /* unmapped */
    return "";
}





/*****************************************************************/

/* NSAPI SAF Support 
 * 
 * The following section contains the entry points for the SAFs directly
 * configurable from obj.conf.
 *
 */


static
int checkmethod(pblock *pb, Session *sn, Request *rq)
{
    int rv = REQ_PROCEED;
    char *fn = 0;
    char *method = 0;
    char *methodexp = 0;

    methodexp = pblock_findval("method", pb);
    if (methodexp) {

	method = pblock_findval("method", rq->reqpb);
	if (method == 0) method = "";   /* method should always be there */

	switch (WILDPAT_CASECMP(method, methodexp)) {

	  case -1:		/* methodexp is invalid */
	    fn = pblock_findval("fn",pb);
	    if (fn == 0) fn = "unknown function";
	    log_error(LOG_MISCONFIG, CA_PROGRAM, sn, rq,
		      "%s: invalid pattern in \"method\" argument.", fn);
	    rv = REQ_ABORTED;
	    break;

	  case 0:		/* method matches methodexp */
	    rv = REQ_PROCEED;
	    break;

	  case 1:		/* method doesn't match methodexp */
	    rv = REQ_NOACTION;
	    break;

	}
    }

    return rv;
}





/*
 * CA_getcert -- (implements the "get-client-cert" SAF)
 *	Get the client cert.  If it is not already present on the session,
 *	and the value of "dorequest" is nonzero, then redo the handshake
 * 	and get it.
 *
 *      If a "method" parameter is present in the directive parameter block
 *      it should be a shell expression, and this operation is only done on
 *      requests where the method matches the shell expression specified in
 *      the directive.
 *
 *	Success: If a cert is obtained successfully put the base64
 *	DER-encoded cert in "auth-cert" in the rq->vars pblock,
 *	return REQ_PROCEED.
 *	
 *	Failure: If the value of "require" is present and is 0,
 *	in the pblock return REQ_NOACTION, otherwise it is REQ_ABORTED
 *      with protocol status set to FORBIDDEN.	(Note this means "require"
 *      is the default).  
 *
 *	e.g.
 *	# This gets a cert, requesting it if it is not already present and
 *	# failing the request if it can't obtain an acceptable cert.
 *      # only on POST requests.
 *	PathCheck fn="get-client-cert" method="POST" dorequest="1" require="1"
 *
 */


NSAPI_PUBLIC int CA_getcert(pblock *pb, Session *sn, Request *rq)
{
    int rv = REQ_PROCEED;
    CERTCertificate *cert = 0;
    char *certb64 = 0;

    rv = checkmethod(pb,sn,rq);
    if (rv != REQ_PROCEED) {
	return rv;
    }

    if (!conf_getglobals()->Vsecurity_active ||
	!conf_getglobals()->Vssl3_active) {
	log_error(LOG_MISCONFIG, CA_PROGRAM, sn, rq, 
	  "get-client-cert requires that security and SSL3 be enabled.");
	/* Tentatively fail, but actual value depends on "require" below */
	rv = REQ_ABORTED;
    } else {
	rv = CA_GetClientCert(pb, sn, rq);
	if (rv != REQ_PROCEED) {
	    /* don't already have a cert */
	    char *dorequest_val;
	    dorequest_val = pblock_findval("dorequest", pb);
	    if (dorequest_val && strcmp(dorequest_val,"0") != 0) {
		/* we are asked to request one from client */
		rv = CA_RequestClientCert(pb, sn, rq);
	    }
	}
    }
    
    if (rv != REQ_PROCEED) {
	/*
	 * Don't have the client's cert.  We return either REQ_NOACTION
	 * or REQ_ABORTED depending on whether a cert was supposed to
	 * be required to get past this directive or not.
	 */
	char *require_val;
	require_val = pblock_findval("require", pb);
	if (require_val && strcmp(require_val,"0") == 0) {
	    /* we DONT require that a cert be present at this point*/
	    return REQ_NOACTION;
        } else {
	    /* "require" is not present in pb or is 1, go no further */
	    protocol_status(sn, rq, PROTOCOL_FORBIDDEN, NULL);
	    return REQ_ABORTED;
	}
    }
    
    /* if we get here we have a cert on sn->clauth cla_cert */
    /* base64 encode it and put it into "auth-cert" in rq->vars */
    cert = ((ClAuth_t*)sn->clauth)->cla_cert;
    certb64 = BTOA_DataToAscii(cert->derCert.data, cert->derCert.len);
    if (certb64 == 0) {
	log_error(LOG_FAILURE, CA_PROGRAM, sn, rq, 
		  "Certificate conversion to base64 failed.");
	return REQ_ABORTED;
    }
    param_free(pblock_remove("auth-cert", rq->vars));
    pblock_nvinsert("auth-cert", certb64, rq->vars);
    /* pblock_nvinsert will strdup() the value, we need to free ours */
    /* XXX ??? is this the right way to call free? */
    PR_Free(certb64);
    return REQ_PROCEED;
    
}			/* CaGetCert */


/*
 * CA_cert2user (implements the "cert2user" SAF)
 * ---------
 *	Maps cert on sn->clauth->cla_cert to a user name, using the
 *      db "userdb" and puts the result in "auth-user" in the rq->vars
 *      pblock. If the mapping doesn't exist and "makefrombasic" is
 *		present and nonzero, make it by the basic auth thing.
 *
 *
 *      If a "method" parameter is present in the directive parameter block
 *      it should be a shell expression, and this operation is only done on
 *      requests where the method matches the shell expression specified in
 *      the directive.
 *
 *	Success: REQ_PROCEED with "auth-user" and "auth-type" and
 *	"authdb" in rq->vars.
 *
 *  Failure: If "require" is not present or is present and is "1" then
 *  the return value on failure is REQ_ABORTED, otherwise it is
 *  REQ_NOACTION with protocol status set to FORBIDDEN.
 *	(Note this means "require" is the default).  
 *
 *	e.g.
 *		PathCheck fn="cert2user" makefrombasic="0" require="1"
 *
 */
     
NSAPI_PUBLIC int CA_cert2user(pblock *pb, Session *sn, Request *rq)
{
    int rv;
    ClAuth_t *cla = (ClAuth_t*)sn->clauth;
    
    rv = checkmethod(pb,sn,rq);
    if (rv != REQ_PROCEED) {
	return rv;
    }

    /* Check to see if there's a cert in the session */
    if (cla->cla_cert != NULL) {
	/* try to map cert to a user using userdb iff there's a cert */
	rv = CA_MapCertToUser(pb, sn, rq);
	
	if (rv != REQ_PROCEED) { /* there is no current mapping */
	  /* are we allowed to go get it? */
	  char *value = pblock_findval("makefrombasic", pb);
	  if (value && strcmp(value,"0") != 0) {
	    /* we're allowed to create it -- try */
	    rv = ca_ProcessBasicAuth(pb, sn, rq);
	    if (rv != REQ_PROCEED &&
		pblock_findval("WWW-authenticate",rq->srvhdrs)) {
	      /* it set WWW-Authenticate/unauthorized for basic auth */
	      /*
	       * Make the user select the certificate again, in
	       * case they accidentally provided the wrong one.
	       * Ideally we'd like to invalidate the session only if
	       * the user hits "Cancel" on the authentication request,
	       * but that doesn't seem to send anything to the server.
	       * This makes the one-time login more awkward, but this
	       * seems to be the only way to give the user a chance
	       * to pick another certificate if they make a mistake.
	       */
	      SSL_InvalidateSession(sn->csd);
	      return rv;
	    }
	  } 
	  else {
	    /* it doesn't exist and we're not allowed to create it */
	    log_error(LOG_SECURITY, CA_PROGRAM, sn, rq,
		      "Client certificate to user mapping unavailable");
	    rv = REQ_ABORTED;
	  }

	  /* Give the user a chance to pick a different certificate */
	  SSL_InvalidateSession(sn->csd);
	}
    } 
    else {
      /* No certificate, tentative failure, but depends on "require" below */
      rv = REQ_ABORTED;
    }
    
    
    /* At this point rv should be REQ_PROCEED iff the cert there and mapped */
    if (rv != REQ_PROCEED) {
	char *require_val = pblock_findval("require", pb);
	if (require_val && strcmp(require_val,"0") == 0) {
	    /* we DONT require that a mapping be present at this point*/
	    return REQ_NOACTION;
	} else {
	    /* "require" is not present in pb or is 1, go no further */
	    protocol_status(sn, rq, PROTOCOL_FORBIDDEN, NULL);
	    return REQ_ABORTED;
	}
    }
    
    /*
     * we get here only if the cert was successfully mapped to a user
     * and rq->vars should already have the auth info in it.
     */
    
    if (rv == REQ_PROCEED) ca_SetRqVars(pb, sn, rq);
    return rv;
}			/* CACert2user */


int get_user_cert_ssl (NSErr_t *errp, PList_t subject, PList_t resource,
		       PList_t auth_info, PList_t global_auth, void *unused)
{
    /* Get the user certificate */
    pblock *pb;
    Session *sn;
    Request *rq;
    CERTCertificate *cert;
    int rv;

    rv = PListGetValue(subject, ACL_ATTR_SESSION_INDEX, (void **)&sn, NULL);
    if (rv < 0 || !sn) {
        ereport(LOG_SECURITY, "get_user_cert_ssl unable to get session address %d\n", rv);
        return LAS_EVAL_FAIL;
    }
 
    rv = PListGetValue(resource, ACL_ATTR_REQUEST_INDEX, (void **)&rq, NULL);
    if (rv < 0 || !rq) {
        ereport(LOG_SECURITY, "get_user_cert_ssl unable to get request address %d\n", rv);
        return LAS_EVAL_FAIL;
    }

    rq->request_is_cacheable = 0;	/* Don't use accel_file_cache */

    pb = pblock_create(4);
    pblock_nvinsert("dorequest", "1", pb);
    rv = CA_getcert(pb, sn, rq);
    pblock_free(pb);

    if (rv != REQ_PROCEED) {
	return LAS_EVAL_FAIL;
    }

    cert = ((ClAuth_t*)sn->clauth)->cla_cert;

    PListInitProp(subject, ACL_ATTR_USER_CERT_INDEX, ACL_ATTR_USER_CERT, cert, 0);
    return LAS_EVAL_TRUE;
}


int ldapu_benign_error (const int rv)
{
    /* Certmap ('ldaputil') errors are all -ve */
    /* LDAP errors are all +ve */
    /* treat all LDAP errors as serious except for the
       LDAP_NO_SUCH_OBJECT error */
    /* All 'ldaputil' errors to be treated benign */
    return (rv < 0 || rv == LDAP_NO_SUCH_OBJECT);
}

int get_auth_user_ssl (NSErr_t *errp, PList_t subject, 
		       PList_t resource, PList_t auth_info,
		       PList_t global_auth, void *unused)
{
    Session *sn = 0;
    Request *rq = 0;
    char *dbname = 0;
    ACLDbType_t dbtype;
    int rv;
    char *user = 0;
    CERTCertificate *cert;
    LDAPDatabase_t *ldb;
    char *userdn;
    char* cached_user = NULL;
    pool_handle_t *subj_pool = PListGetPool(subject);

    rv = PListGetValue(subject, ACL_ATTR_CACHED_USER_INDEX, 
		       (void **)&cached_user, NULL);

    if ((rv >= 0) && cached_user && *cached_user) {
        PListInitProp(subject, ACL_ATTR_USER_INDEX, ACL_ATTR_USER, 
		      cached_user, 0);
        return LAS_EVAL_TRUE;
    }

    rv = ACL_GetAttribute(errp, ACL_ATTR_USER_CERT, (void **)&cert, subject,
			  resource, auth_info, global_auth);

    if (rv != LAS_EVAL_TRUE) {
        ereport(LOG_SECURITY, "cert2user_ldap unable to get user certificate %d\n", rv);
        return rv;
    }

    rv = PListGetValue(resource, ACL_ATTR_REQUEST_INDEX, (void **)&rq, NULL);
    if (rv < 0 || !rq) {
        ereport(LOG_SECURITY, "cert2user_ldap unable to get request address %d\n", rv);
        return LAS_EVAL_FAIL;
    }

    rv = ACL_AuthInfoGetDbname(auth_info, &dbname);
    if (rv < 0 || !dbname) {
        ereport(LOG_SECURITY, "cert2user_ldap unable to get database name %d", rv);
        return LAS_EVAL_FAIL;
    }

    rv = ACL_DatabaseFind(errp, dbname, &dbtype, (void **)&ldb);

    if (rv != LAS_EVAL_TRUE) {
	ereport(LOG_SECURITY, "cert2user_ldap unable to get parsed database %s", dbname);
	return rv;
    }

    rv = acl_map_cert_to_user(errp, dbname, ldb, (void *)cert,
			      resource, subj_pool, &user, &userdn);

    if (rv != LDAPU_SUCCESS || !user) {
	char *subjectDN = 0;
	char *issuerDN = 0;
	int  tmp_rv;

	ldapu_get_cert_subject_dn((void *)cert, &subjectDN);
	ldapu_get_cert_issuer_dn((void *)cert, &issuerDN);

	ereport(LOG_SECURITY, "cert2user_ldap unable to map cert to LDAP entry. Reason: %s, Issuer: \"%s\", User: \"%s\"", 
		ldapu_err2string(rv), issuerDN ? issuerDN : "unknown",
		subjectDN ? subjectDN : "unknown");
	ldapu_free(subjectDN);
	ldapu_free(issuerDN);
	/* ask the user to give another cert */
	tmp_rv = PListGetValue(subject, ACL_ATTR_SESSION_INDEX, (void **)&sn, NULL);
	if (tmp_rv < 0 || !sn) {
	    ereport(LOG_SECURITY, "cert2user_ldap unable to get session address %d\n", rv);
	    return LAS_EVAL_FAIL;
	}

	SSL_InvalidateSession(sn->csd);
	/* Return FAIL only if there was LDAP error.  By returning FALSE we
	 * keep the option open for redirection (bong file) */
	return ldapu_benign_error(rv) ? LAS_EVAL_FALSE : LAS_EVAL_FAIL;
    }

    /* do similar work as ca_SetRqVars for backward compatibility */
    pblock_nvinsert(ACL_ATTR_AUTH_TYPE, ACL_AUTHTYPE_SSL, rq->vars);
    pblock_nvinsert(ACL_ATTR_AUTH_USER, user, rq->vars);
    pblock_nvinsert(ACL_ATTR_AUTH_DB, dbname, rq->vars);

    if (userdn) {
	pblock_nvinsert(ACL_ATTR_USERDN, userdn, rq->vars);
	PListInitProp(subject, ACL_ATTR_USERDN_INDEX, ACL_ATTR_USERDN,
		      userdn, 0);
    }

    PListInitProp(subject, ACL_ATTR_USER_INDEX, ACL_ATTR_USER, user, 0);
    return LAS_EVAL_TRUE;
}

/* clauth.cpp */

#endif				/* defined(CLIENT_AUTH) */
