/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/* 
 * flexlog.c: Code to handle flexible logfile format
 *
 * Don Eastman, imported by Mike McCool 
 */

#include <nspr.h>
#include <plhash.h>
#include <plstr.h>

#include "netsite.h"
#include "safs/flexlog.h"
#include "base/session.h"
#include "frame/req.h"
#include "base/file.h"         /* system_fopenWA, system_write_atomic */
#include "frame/conf.h"        /* user_pw */
#include "frame/log.h"         /* log_error */
#include "base/util.h"         /* util_sprintf */

#include <time.h>
#ifdef XP_UNIX
#include <signal.h>
#include <unistd.h>
#endif /* XP_UNIX */

#include "safs/ntlogbuf.h"
#include "base/fasttime.h"

/* Removed for ns security integration
#include "xp_hash.h"
#include "xp_mcom.h"
*/


#undef TIMELEN
#define TIMELEN 28
/* B_SIZE-1 is the maximum size line that the flexlog will generate.   The
 * outbufs which are slopily passed through this code are all assumed to be
 * B_SIZE bytes long; overflow checks are done based on B_SIZE.
 */
#define B_SIZE 4096

# define COMMON_FORMAT "%Ses->client.ip% - %Req->vars.auth-user% [%SYSDATE%] \"%Req->reqpb.clf-request%\" %Req->srvhdrs.clf-status% %Req->srvhdrs.content-length%"ENDLINE
#define TEXTMODE        0
#define KEYWORDMODE     1
# define PBLOCK        "pblock."
# define PBLOCK_LEN    7
# define CLF_CLIENT    "Ses->client."
# define CLF_CLIENT_LEN 12
# define CLF_REQVARS   "Req->vars."
# define CLF_REQVARS_LEN 10
# define CLF_REQPB     "Req->reqpb."
# define CLF_REQPB_LEN 11
# define CLF_REQHDRS   "Req->headers."
# define CLF_REQHDRS_LEN 13
# define CLF_REQSHDRS  "Req->srvhdrs."
# define CLF_REQSHDRS_LEN 13
# define CLF_DATE      "SYSDATE"
# define CLF_IP             "ip"
# define CLF_AUTH_USER      "auth-user"
# define CLF_FULL_REQ       "clf-request"
# define CLF_METHOD         "method"
# define CLF_URI            "uri"
# define CLF_QUERY          "query"
# define CLF_PROTOCOL       "protocol"
# define CLF_STATUS         "clf-status"
# define CLF_CONTENT_LENGTH "content-length"
# define CLF_REFERER        "referer"
# define CLF_USER_AGENT     "user-agent"
# define CLF_AUTHORIZATION  "authorization"

#define MAXTOKENS       256
#define SEPARATOR       '%'

/* Removed for ns security integration
#include "xp_hash.h"
#include "xp_mcom.h"
*/

typedef struct flex_tokens
{
    int  offset;        /* its offset in buffer */
    int  len;           /* Length */
    int  mode;          /* TEXTMODE or KEYWORDMODE */
} FlexToken;

#define TOKEN_NONE			0
#define TOKEN_STATUS			1
#define TOKEN_CONTENT_LENGTH		2
#define TOKEN_CLIENT_HOST		3
#define TOKEN_FULL_REQ			4
#define TOKEN_METHOD			5
#define TOKEN_PROTOCOL			6
#define TOKEN_QUERY_STRING		7
#define TOKEN_URI			8
#define TOKEN_REFERER			9
#define TOKEN_USER_AGENT		10
#define TOKEN_AUTHORIZATION		11
#define TOKEN_AUTH_USER			12
#define TOKEN_DATE			13
#define TOKEN_TEXTMODE			14
 
struct fclog 
{
    SYS_FILE fd;
    PRBool   delimit;
    char *name;
    char *format;
    LOGBUF_T *logbuf;
    char *CompleteName;
    FlexToken tokens[MAXTOKENS];
    int num_tokens;
    int *int_tokens;
    struct fclog *next;
};

int flexOffsets(char *buffer, FlexToken flex[],  char separator);
int flexTokens2intTokens(struct fclog *log);

/* ----------------------- flexRotate ------------------------- */

struct fclog *fcl = NULL;

#ifdef XP_WIN32
void flexRotate(void)
{
    struct fclog *p;
    char err[B_SIZE];
    DWORD Result;
    char TmpFile[B_SIZE], tmpbuf[128], outbuf[B_SIZE];
    time_t current_time;
    struct tm tms;
    struct tm *tmsp;

    current_time = ft_time();
    tmsp = system_localtime(&current_time, &tms);

    (void)util_strftime(tmpbuf, "%d%h-%I%p", tmsp);
 
    p = fcl;
    
    while (p) 
    {
        if (system_flock(p->fd) == IO_ERROR) return;

        system_fclose(p->fd);
        PR_snprintf(TmpFile, B_SIZE, "%s.%s", p->CompleteName, tmpbuf);
        
	if (! MoveFileEx(p->CompleteName, TmpFile, MOVEFILE_REPLACE_EXISTING))
	{
                Result = GetLastError();
	}

        if ((p->fd = system_fopenWA(p->CompleteName)) == SYS_ERROR_FD) 
	{
                PR_snprintf(err, B_SIZE, "can't open logfile %s (%s)", p->CompleteName, system_errmsg());
        }
        
        
	PR_snprintf(outbuf, B_SIZE, "format=%s%s", p->format, ENDLINE);
        system_fwrite(p->fd, outbuf, PL_strlen(outbuf));
        p->logbuf->fd = p->fd;
	p = p->next;
     }
 }
#endif
 
/* --------------------- flexTerminate ------------------------ */


void flexTerminate(void *parameter)
{
    struct fclog *p;
 
    while (fcl) 
    {
        system_fclose(fcl->fd);
        FREE(fcl->name);
        FREE(fcl->format);

        if (fcl->int_tokens)
            FREE(fcl->int_tokens);

		/*
		 * This has already been done in DaemonProcessor::Run.
		 */
        // logbuf_close(fcl->logbuf);
        FREE(fcl->CompleteName);
        
	p = fcl;
        fcl = fcl->next;
        FREE(p);
    }
    fcl = NULL;
}

/* ----------------------- flex_init -------------------------- */

int flex_init(pblock *pb, Session *sn, Request *rq)
{
    int              i;
    int		     len;
    char             err[MAGNUS_ERROR_LEN];
    char             outbuf[B_SIZE], inbuf[B_SIZE];
    char            *t, *x;
    
    struct pb_entry *p;
    struct fclog     *c;

#ifdef XP_UNIX
    int root = !geteuid();
#endif /* XP_UNIX*/
 
    /* Removed because it makes the flex-init code not be able to be called
     * more than once.  Initialization takes place in the variable 
     * declaration.  --MLM */
    /* fcl = NULL; */
 
    if(logbuf_init(pb) == REQ_ABORTED)
        return REQ_ABORTED;
    param_free(pblock_remove("server-version", pb));

    for (i = 0; i < pb->hsize; i++) {
      p = pb->ht[i];
      while (p) {
	t = p->param->name;

	if (PL_strcmp(t, "fn") && 
	    PL_strncmp(t, "format.", PL_strlen("format.")) &&
	    PL_strncmp(t, "delimit.", PL_strlen("delimit.")) &&
	    (!LOGBUF_DIRECTIVE(t))) {
	  c = (struct fclog *) PR_Malloc(sizeof(struct fclog));
	  c->name = PL_strdup(t);

	  PR_snprintf(inbuf, B_SIZE, "delimit.%s", t);
	  if ((x = pblock_findval(inbuf, pb)) != 0) c->delimit = PR_TRUE;
	  else  c->delimit = PR_FALSE;

	  PR_snprintf(inbuf, B_SIZE, "format.%s", t);
	  if ((x = pblock_findval(inbuf, pb)) != 0) c->format = PL_strdup(x);
	  else c->format = PL_strdup(COMMON_FORMAT);

	  if ((c->fd = system_fopenWA(p->param->value)) == SYS_ERROR_FD) {
	    PR_snprintf(err, MAGNUS_ERROR_LEN, "can't open logfile %s (%s)", 
		      p->param->value, system_errmsg());
	    pblock_nvinsert("error", err, pb);
	    return REQ_ABORTED;
	  }
    
	  struct stat sbuf;
	  SYS_FILE fd;

	  PR_snprintf(outbuf, B_SIZE, "format=%s%s", c->format, ENDLINE);

	  if (system_stat(p->param->value, &sbuf) != 0 || !sbuf.st_size){
	    if (system_flock(c->fd) != IO_ERROR){
	      system_fwrite(c->fd, outbuf, PL_strlen(outbuf));
	      (void) system_ulock(c->fd);
	    }
	  }
	  else {
            if ((fd = system_fopenRO(p->param->value)) == SYS_ERROR_FD) {
	      PR_snprintf(err, MAGNUS_ERROR_LEN, "can't open logfile %s (%s)",
			    p->param->value, system_errmsg());
	      pblock_nvinsert("error", err, pb);
	      return REQ_ABORTED;
            }
	    len = system_fread(fd, inbuf, B_SIZE-1);
            if (len == IO_ERROR) {
                PR_snprintf(err, MAGNUS_ERROR_LEN, "can't read logfile %s (%s)",
                    p->param->value, system_errmsg());
                pblock_nvinsert("error", err, pb);
                system_fclose(fd);
		return REQ_ABORTED;
            }
            system_fclose(fd);
	    inbuf[len] = 0;
	    /* no nspr support on strtok ..... should change to nspr when it is available */
	    (void) strtok(inbuf,ENDLINE);
	    if (PL_strncmp(inbuf, outbuf, PL_strlen(inbuf))) {
	      PR_snprintf(err, MAGNUS_ERROR_LEN, "must cycle logfile %s before changing format", 
			  p->param->value);
	      pblock_nvinsert("error", err, pb);
	    }
	  }
	  /*
	   * Don't make cgi's inherit all the log fd's
	   */
	  file_setinherit(c->fd, PR_FALSE);

	  c->logbuf = logbuf_new(pb, c->fd);
	  c->CompleteName = PL_strdup(p->param->value);
                
	  c->next = fcl;
	  fcl = c;
 
	  /*
	   * If we're running as root, make sure that when we restart
	   * we can re-open the log files.
	   */
#ifdef XP_UNIX
	  if(root && user_pw) {
	    chown(p->param->value, user_pw->pw_uid, user_pw->pw_gid);
	  }
#endif /* XP_UNIX */
	}
	p = p->next;
      }
    }
    c->int_tokens = NULL;
    c->num_tokens = flexOffsets(c->format, c->tokens, SEPARATOR);
    if ( flexTokens2intTokens(c) == REQ_ABORTED) 
        c->int_tokens = NULL;
 
#ifndef MCC_ADMSERV
    if(fcl) {
      daemon_atrestart(flexTerminate, NULL);
    }
#endif
 
    return REQ_PROCEED;
}

/* ---------------------- flexOffsets ------------------------- */
/* Tokenize format/log record - return number of tokens found */

int flexOffsets(char *buffer, FlexToken flex[],  char separator)
{
    char *position = buffer;
    char *positionend = buffer + PL_strlen(buffer);
    int   mode = TEXTMODE;
    int   tcnt = 0;

    while ((tcnt < MAXTOKENS) && (position < positionend))
    {
        switch(mode)
        {
        case TEXTMODE:

            flex[tcnt].mode = mode;
            flex[tcnt].offset = position - buffer;

            while (position < positionend)
	    {
	        if (*position != separator) position++;
		else if ((position == buffer) || (position > buffer && *(position - 1) != '\\')) break;
	        else position++;
	    }
  
            flex[tcnt].len = (position - buffer) - flex[tcnt].offset;

            tcnt++;

            mode = KEYWORDMODE;
            break;

        case KEYWORDMODE:

            position++;

            flex[tcnt].mode = mode;
            flex[tcnt].offset = position - buffer;

            while (position < positionend)
	    {
	        if (*position != separator) position++;
		else if ((position == buffer) || (position > buffer && *(position - 1) != '\\')) break;
	        else position++;
	    }

            flex[tcnt].len = (position - buffer) - flex[tcnt].offset;

            position++;         /* move beyond separator */

            tcnt++;

            mode = TEXTMODE;
            break;
        }
    }
    flex[tcnt].offset = -1;
    if (tcnt >= MAXTOKENS)
        ereport(LOG_WARN, "Truncating log to %d tokens", MAXTOKENS);

    return tcnt;
}

/* -------------------- flexTokens2intTokens ------------------ */
/* Try to convert the string-based tokens into integer tokens
 * for speed...
 */
int flexTokens2intTokens(struct fclog *log)
{
    char *format = log->format;
    FlexToken *tokens = log->tokens;
    int **int_tokens = &(log->int_tokens);
    int token_count = log->num_tokens;
    int index = 0;

    *int_tokens = (int *)MALLOC(sizeof(int) * (token_count + 1));

    while(index < token_count) {
        switch (tokens[index].mode) {
            case TEXTMODE: 
                (*int_tokens)[index] = TOKEN_TEXTMODE;
                break;
            case KEYWORDMODE:
                switch(format[tokens[index].offset]) {
                    case 'S':   /* Ses->client || SYSDATE */
                        switch (format[tokens[index].offset+1]) {
                            case 'e': /* Ses->client */
                                if (!PL_strncmp(&format[tokens[index].offset],
                                    CLF_CLIENT CLF_IP, sizeof(CLF_CLIENT CLF_IP)-1)) 
                                    (*int_tokens)[index] = TOKEN_CLIENT_HOST;
                                else
                                    goto punt;
                                break;
                            case 'Y': /* SYSDATE */
                                if (!PL_strncmp(&format[tokens[index].offset],
                                    CLF_DATE, sizeof(CLF_DATE)-1)) 
                                    (*int_tokens)[index] = TOKEN_DATE;
                                else
                                    goto punt;
                                break;
                            default: 
                                goto punt;
                        }
                        break;
                    case 'R':   /* Req->xxx */
                        switch(format[tokens[index].offset+5]) {
                            case 'v':  /* Req->vars */
                                if (!PL_strncmp(&format[tokens[index].offset],
                                    CLF_REQVARS CLF_AUTH_USER, sizeof(CLF_REQVARS CLF_AUTH_USER)-1)) 
                                    (*int_tokens)[index] = TOKEN_AUTH_USER;
                                else 
                                    goto punt;
                                break;
                            case 'r':  /* Req->reqpb */
                                if (!PL_strncmp(&format[tokens[index].offset],
                                    CLF_REQPB CLF_FULL_REQ, sizeof(CLF_REQPB CLF_FULL_REQ)-1)) 
                                    (*int_tokens)[index] = TOKEN_FULL_REQ;
                                else if (!PL_strncmp(&format[tokens[index].offset],
                                    CLF_REQPB CLF_METHOD, sizeof(CLF_REQPB CLF_METHOD)-1)) 
                                    (*int_tokens)[index] = TOKEN_METHOD;
                                else if (!PL_strncmp(&format[tokens[index].offset],
                                    CLF_REQPB CLF_PROTOCOL, sizeof(CLF_REQPB CLF_PROTOCOL)-1)) 
                                    (*int_tokens)[index] = TOKEN_PROTOCOL;
                                else if (!PL_strncmp(&format[tokens[index].offset],
                                    CLF_REQPB CLF_URI, sizeof(CLF_REQPB CLF_URI)-1)) 
                                    (*int_tokens)[index] = TOKEN_URI;
                                else if (!PL_strncmp(&format[tokens[index].offset],
                                    CLF_REQPB CLF_QUERY, sizeof(CLF_REQPB CLF_QUERY)-1)) 
                                    (*int_tokens)[index] = TOKEN_QUERY_STRING;
                                else 
                                    goto punt;
                                break;
                            case 'h':  /* Req->headers */
                                if (!PL_strncmp(&format[tokens[index].offset],
                                    CLF_REQHDRS CLF_REFERER, sizeof(CLF_REQHDRS CLF_REFERER)-1)) 
                                    (*int_tokens)[index] = TOKEN_REFERER;
                                else if (!PL_strncmp(&format[tokens[index].offset],
                                    CLF_REQHDRS CLF_USER_AGENT, sizeof(CLF_REQHDRS CLF_USER_AGENT)-1)) 
                                    (*int_tokens)[index] = TOKEN_USER_AGENT;
                                else if (!PL_strncmp(&format[tokens[index].offset],
                                    CLF_REQHDRS CLF_AUTHORIZATION, sizeof(CLF_REQHDRS CLF_AUTHORIZATION)-1)) 
                                    (*int_tokens)[index] = TOKEN_AUTHORIZATION;
                                else 
                                    goto punt;
                                break;
                            case 's':  /* Req->srvhdrs */
                                if (!PL_strncmp(&format[tokens[index].offset],
                                    CLF_REQSHDRS CLF_STATUS, sizeof(CLF_REQSHDRS CLF_STATUS)-1)) 
                                    (*int_tokens)[index] = TOKEN_STATUS;
                                else if (!PL_strncmp(&format[tokens[index].offset],
                                    CLF_REQSHDRS CLF_CONTENT_LENGTH, sizeof(CLF_REQSHDRS CLF_CONTENT_LENGTH)-1)) 
                                    (*int_tokens)[index] = TOKEN_CONTENT_LENGTH;
                                else 
                                    goto punt;
                                break;
                            default: 
                                goto punt;
                        }
                        break;
                    default: 
                        /* user custom thing which we can't do quickly! */
                        goto punt;
                }
                break;
        }

        index++;
    }

    (*int_tokens)[index] = TOKEN_NONE; /*sentinel */

    return REQ_PROCEED;
punt:
    FREE(*int_tokens);
    return REQ_ABORTED;
}

/* -------------------- flexTokenReplace ---------------------- */

int flexTokenReplace(char *rql, char *tokenreplace, char *token, pblock *pb, Session *sn, Request *rq)
{
    char  *p;
    char   buf[B_SIZE];
    char   timebuf[TIMELEN];
    char   sign;
    time_t secondssince;
    long   tz;
    struct tm *tm, tms;

    if (!PL_strncmp(token, PBLOCK, PBLOCK_LEN))
    {
	p = pblock_findval(token + PBLOCK_LEN, pb);
    }
    else if (!PL_strncmp(token, CLF_CLIENT, CLF_CLIENT_LEN))
    {
	p = token + CLF_CLIENT_LEN;

	if (!PL_strcmp(p, "ip"))
	{
	    char *iponly = pblock_findval("iponly", pb);

	    if (iponly || (!(p = session_dns(sn))))
	    {
		p = pblock_findval("ip", sn->client);
	    }
	}
	else
	{
	    p = pblock_findval(token + CLF_CLIENT_LEN, sn->client);
	}
    }
    else if (!PL_strncmp(token, CLF_REQVARS, CLF_REQVARS_LEN))
    {
	p = pblock_findval(token + CLF_REQVARS_LEN, rq->vars);
    }
    else if (!PL_strncmp(token, CLF_REQPB, CLF_REQPB_LEN))
    {
	p = token + CLF_REQPB_LEN;

	if (!PL_strcmp(p, "clf-request"))
	{
	    p = rql;
	}
	else
	{
	    p = pblock_findval(token + CLF_REQPB_LEN, rq->reqpb);
	}
    }
    else if (!PL_strncmp(token, CLF_REQHDRS, CLF_REQHDRS_LEN))
    {
	p = pblock_findval(token + CLF_REQHDRS_LEN, rq->headers);
    }
    else if (!PL_strncmp(token, CLF_REQSHDRS, CLF_REQSHDRS_LEN))
    {
	p = token + CLF_REQSHDRS_LEN;
        /* MLM */
        if(!PL_strcmp(p, "clf-status"))  {
            char *space=NULL;
            p = pblock_findval("status", rq->srvhdrs);
            if(p)  {
                space=PL_strchr(p, ' ');
                if(space) *space='\0';
            }
        }  else  {
            p = pblock_findval(token + CLF_REQSHDRS_LEN, rq->srvhdrs);
        }
    }
    else if (!PL_strcmp(token, CLF_DATE))   /* CLF_DATE comforms to common format */
    {
	secondssince = ft_time();
	tm  = system_localtime(&secondssince, &tms);

#ifdef BSD_TIME
	tz = tm->tm_gmtoff;
#else
	tz = - timezone;
	if(tm->tm_isdst) tz += 3600;
#endif

	sign = tz >=0 ? '+' : '-';
	if (tz < 0) tz = -tz;
	
	(void)util_strftime(timebuf, "%d/%b/%Y:%H:%M:%S", tm);
	PR_snprintf(buf, B_SIZE, "%s %c%02d%02d", timebuf, sign, (int)(tz/3600), (int)(tz%3600));
	p = buf;
    }
    else
    {
	p = 0;
    }

    if (!p) PL_strcpy(tokenreplace, "-"); else PL_strcpy(tokenreplace, p);

    return 0;
}

/* -------------------- flexParseFormat ----------------------- */

int flexParseFormat(int delimit, char *informat, struct flex_tokens *flexformat, 
		    char *rql, char *outbuf, pblock *pb, Session *sn, Request *rq)
{
  char format[B_SIZE], tokenreplace[B_SIZE];

  int  tcnt = 0;
  int  outbufindex = 0;
  int  i, mode = TEXTMODE;
  int  tmplen;

  PL_strncpy(format, informat, B_SIZE);

  while (flexformat[tcnt].offset != -1) {
    switch(flexformat[tcnt].mode) {
    case TEXTMODE:
      if (outbufindex + flexformat[tcnt].len +4 > B_SIZE) {
	ereport(LOG_WARN, "flex log buffer overflow- greater than %d characters",
		B_SIZE);
	return outbufindex;
      }
      PL_strncpy(outbuf + outbufindex, format + flexformat[tcnt].offset, flexformat[tcnt].len);
      outbufindex += flexformat[tcnt].len;
      mode = KEYWORDMODE;
      break;

    case KEYWORDMODE:
      *(format + flexformat[tcnt].offset + flexformat[tcnt].len) = 0;
      flexTokenReplace(rql, tokenreplace, format + flexformat[tcnt].offset, pb, sn, rq);

      if (delimit) outbuf[outbufindex++] = SEPARATOR;

      i = 0; 

      tmplen = strlen(tokenreplace);
      if (outbufindex + tmplen +4 > B_SIZE) {
	ereport(LOG_WARN, "flex log buffer overflow- greater than %d characters",
		B_SIZE);
	return outbufindex;
      }
      memcpy(&outbuf[outbufindex], tokenreplace, tmplen);
      outbufindex += tmplen;
		
      if (delimit) outbuf[outbufindex++] = SEPARATOR;

      mode = TEXTMODE;
      break;
    }
    tcnt++;
  }
    
  outbuf[outbufindex++] = '\n';
  /* Probably not needed */
  outbuf[outbufindex] = '\0';
  return outbufindex;
}

/* -------------------- _flex_log_gen ----------------------- */
/* _flex_log_gen()
 * Two problems exist with flexParseFormat().  First, it is intricately
 * tied to pblocks, and the NT accelerator does not want to deal with
 * pblocks at all.  Having to deal with pblocks would be extremely expensive.
 * The second problem is that the flex logs are fairly slow.  So, if the
 * user is only asking for a few key items, we can dramatically speed it up.
 * That is the theory of flex_log_fast, anyway.
 *
 * _flex_log_gen is called by the flex_log_accel() and the flex_log_fast()
 * routine.  flex_log_accel is NT specific; flex_log_fast() will be called
 * whenever the logging is "simple" enough that we know how to deal with
 * it without having to do all those strcmp()s...
 */

static 
int _flex_log_gen(  char *outbuf,
                    int *tl,
                    char *flexformat,
                    struct flex_tokens *flextokens,
                    int delimit,
                    char *status,
                    char *content_length,
                    char *client_host,
                    char *full_request,
                    char *method,
                    char *protocol,
                    char *query_string,
                    char *uri,
                    char *referer,
                    char *user_agent,
                    char *authorization,
                    char *auth_user)
{
    int pos = 0;
    int token = 0;
    int tmplen;
    char *cp;
    char timebuf[TIMELEN], tmpbuf[TIMELEN+24];
    time_t secondssince;
    long tz;
    char sign;
    struct tm *tm, tms;
 
    while ( tl[token] != TOKEN_NONE ) {
        tmplen = 0;
        switch(tl[token]) {
            case TOKEN_STATUS:
		/* status could be a read-only const */
                if (status && (cp = PL_strdup(status))) {
                    char *ch;
                    for (ch=cp; *ch && *ch != ' '; ch++);
                    if(*ch) 
                        *ch = '\0';
                    tmplen = PL_strlen(cp);
		    /*FREE(cp);*/
                }
                break;
            case TOKEN_CONTENT_LENGTH:
                cp = content_length;
                if (cp)
                    tmplen = PL_strlen(cp);
                break;
            case TOKEN_CLIENT_HOST:
                cp = client_host;
                if (cp)
                    tmplen = PL_strlen(cp);
                break;
            case TOKEN_FULL_REQ:
                cp = full_request;
                if (cp)
                    tmplen = PL_strlen(cp);
                break;
            case TOKEN_METHOD:
                cp = method;
                if (cp)
                    tmplen = PL_strlen(cp);
                break;
            case TOKEN_PROTOCOL:
                cp = protocol;
                if (cp)
                    tmplen = PL_strlen(cp);
                break;
            case TOKEN_QUERY_STRING:
                cp = query_string;
                if (cp)
                    tmplen = PL_strlen(cp);
                break;
            case TOKEN_URI:
                cp = uri;
                if (cp)
                    tmplen = PL_strlen(cp);
                break;
            case TOKEN_REFERER:
                cp = referer;
                if (cp)
                    tmplen = PL_strlen(cp);
                break;
            case TOKEN_USER_AGENT:
                cp = user_agent;
                if (cp)
                    tmplen = PL_strlen(cp);
                break;
            case TOKEN_AUTHORIZATION:
                cp = authorization;
                if (cp)
                    tmplen = PL_strlen(cp);
                break;
            case TOKEN_AUTH_USER:
                cp = auth_user;
                if (cp)
                    tmplen = PL_strlen(cp);
                break;
            case TOKEN_DATE:
	        secondssince = ft_time();
        	tm  = system_localtime(&secondssince, &tms);

#ifdef BSD_TIME
        	tz = tm->tm_gmtoff;
#else
        	tz = - timezone;
        	if(tm->tm_isdst) tz += 3600;
#endif

        	sign = tz >=0 ? '+' : '-';
        	if (tz < 0) tz = -tz;
	
        	(void)util_strftime(timebuf, "%d/%b/%Y:%H:%M:%S", tm);
        	PR_snprintf(tmpbuf, TIMELEN+24, "%s %c%02d%02d", 
			    timebuf, sign, (int)(tz/3600), (int)(tz%3600));
        	cp = tmpbuf;
                tmplen = PL_strlen(cp);
                break;
            case TOKEN_TEXTMODE:
                /* use data from flextokens */
                cp = &(flexformat[flextokens[token].offset]);
                tmplen = flextokens[token].len;
                break;
            default:
                break;
        }
        if (delimit && tl[token] != TOKEN_TEXTMODE)
            outbuf[pos++] = SEPARATOR;
        if (cp) {
            if (pos+tmplen+2 > B_SIZE) {
                ereport(LOG_WARN, "flex log buffer overflow- greater than %d characters",
                    B_SIZE);
                outbuf[pos++] = '\0';
                return pos;
            }
            memcpy(&outbuf[pos], cp, tmplen+1);
            pos += tmplen;
	    if(tl[token] == TOKEN_STATUS)
		FREE(cp);
        }
        else 
            outbuf[pos++] = '-';
        if (delimit && tl[token] != TOKEN_TEXTMODE)
            outbuf[pos++] = SEPARATOR;
        token++;
    }
    outbuf[pos++] = '\n';
    outbuf[pos] = '\0';

    return pos;
}

/* -------------------- flex_log_fast   ----------------------- */
/* flex_log_fast()
 * When a flex_log() caller discovers that they have a "simple"
 * log which can be done quickly, it just calls this routine rather
 * than the "big" flex logger.
 */
int flex_log_fast(struct fclog *p, char *logname, char *full_request, 
    char *outbuf, pblock *pb, Session *sn, Request *rq)
{
    char *status = NULL;
    char *content_length = NULL;
    char *client_host = NULL;
    char *method = NULL;
    char *protocol = NULL;
    char *query_string = NULL;
    char *uri = NULL;
    char *referer = NULL;
    char *user_agent = NULL;
    char *authorization = NULL;
    char *auth_user = NULL;
    int index;
   
    /* Walk through the tokens list and figure out which elements we
     * need to "get".
     */
    for (index=0; p->int_tokens[index] != TOKEN_NONE; index++) {
        switch(p->int_tokens[index]) {
            case TOKEN_STATUS:
                status = pblock_findval("status", rq->srvhdrs);
                break;
            case TOKEN_CONTENT_LENGTH:
                content_length = pblock_findval("content-length", rq->srvhdrs);
                break;
            case TOKEN_CLIENT_HOST:
                 
                if (pblock_findval("iponly", pb) || !(client_host = session_dns(sn)) )
                    client_host = pblock_findval("ip", sn->client);
                break;
            case TOKEN_FULL_REQ:
                /* already have it */
                break;
            case TOKEN_METHOD:
                method = pblock_findval("method", rq->reqpb );
                break;
            case TOKEN_PROTOCOL:
                protocol = pblock_findval("protocol", rq->reqpb);
                break;
            case TOKEN_QUERY_STRING:
                query_string = pblock_findval("query", rq->reqpb);
                break;
            case TOKEN_URI:
                uri = pblock_findval("uri", rq->reqpb);
                break;
            case TOKEN_REFERER:
                referer = pblock_findval("referer", rq->headers);
                break;
            case TOKEN_USER_AGENT:
                user_agent = pblock_findval("user-agent", rq->headers);
                break;
            case TOKEN_AUTHORIZATION:
                authorization = pblock_findval("authorization", rq->headers);
                break;
            case TOKEN_AUTH_USER:
                auth_user = pblock_findval("auth-user", rq->vars);
                break;
            case TOKEN_DATE:
            case TOKEN_TEXTMODE:
                /* nothing to get */
                break;
        }
    }

    return _flex_log_gen(outbuf, p->int_tokens, p->format, p->tokens, 
        p->delimit, status, content_length, client_host, full_request, 
        method, protocol, query_string, uri, referer, user_agent, 
        authorization, auth_user);
}

/* -------------------- flex_log_accel  ----------------------- */
/* flex_log_accel
 * Interface for the accelerator to get into the flex logging
 * facility
 */
int flex_log_accel( char *logname,
                    char *status,
                    char *content_length,
                    char *client_host,
                    char *full_request,
                    char *method,
                    char *protocol,
                    char *query_string,
                    char *uri,
                    char *referer,
                    char *user_agent,
                    char *authorization,
                    char *auth_user)
{

    struct fclog *p;
    char outbuf[B_SIZE];
    int len;

    for (p = fcl; p; p = p->next)
    {
	if (!PL_strcmp(p->name, logname)) break;
    }

    if (!p)
    {
        ereport(LOG_MISCONFIG, "flex-log cannot find log named %s", logname);
        return REQ_ABORTED;
    }

    len = _flex_log_gen(outbuf, p->int_tokens, p->format, p->tokens, 
        p->delimit, status, content_length, client_host, full_request, 
        method, protocol, query_string, uri, referer, user_agent, 
        authorization, auth_user);

    logbuf_write(p->logbuf, p->fd, outbuf, len);
    return REQ_PROCEED;
}

int flex_log_tokens( char *logname ) 
{
    struct fclog *p;

    for (p = fcl; p; p = p->next)
    {
        if (!PL_strcmp(p->name, logname)) break;
    }

    if (!p || !p->int_tokens)
        return 0;

    return 1;
}

/* ------------------------ flex_log -------------------------- */
int flex_log(pblock *pb, Session *sn, Request *rq)
{


    char  outbuf[B_SIZE];
    char *format;

    struct fclog *p;
    char *logname = pblock_findval("name", pb);
    char *rql;

    pb_param *rqlp = pblock_find("clf-request", rq->reqpb);
    int len;

    if (!rqlp || !fcl)
    {
        return REQ_NOACTION;
    }
    
    rql = rqlp->value;

    if (!logname) logname = "global";

    for (p = fcl; p; p = p->next)
    {
	if (!PL_strcmp(p->name, logname)) break;
    }

    if (!p)
    {
        log_error(LOG_MISCONFIG, "common-log", sn, rq, "cannot find log named %s", logname);
        return REQ_ABORTED;
    }
    
    if (!(format = p->format)) format = COMMON_FORMAT;

    if (!p->int_tokens)
        len = flexParseFormat(p->delimit, format, p->tokens, rql, outbuf, 
            pb, sn, rq);
    else
        len = flex_log_fast(p, logname, rql, outbuf, pb, sn, rq);

    logbuf_write(p->logbuf, p->fd, outbuf, len);
    return REQ_PROCEED;
}
