/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
#include "netsite.h"
#include <base/ereport.h>
#include <libaccess/authdb.h>
#include <libaccess/acl.h>
#include <libaccess/aclproto.h>
#include <libaccess/las.h>
#include <libaccess/ldapacl.h>
#include <frame/func.h>
#include <frame/conf.h>
#include <safs/auth.h>
#include <safs/clauth.h>
#include <safs/init.h>

ACLMethod_t ACL_MethodBasic = ACL_METHOD_INVALID;
ACLMethod_t ACL_MethodSSL = ACL_METHOD_INVALID;

#define BIG_LINE 1024

#define ACL_REG(x, y, z) \
    { \
	int rv = (x); \
	if (rv < 0) { \
	    ereport(LOG_MISCONFIG, y, z); \
	    return REQ_ABORTED; \
	} \
    }

int register_attribute_getter (pblock *pb, Session *sn, Request *rq)
{
    char *method_str = pblock_findval(ACL_ATTR_METHOD, pb);
    char *attr = pblock_findval(ACL_ATTR_ATTRIBUTE, pb);
    char *funcStr = pblock_findval(ACL_ATTR_GETTERFN, pb);
    char *dbtype_str = pblock_findval(ACL_ATTR_DBTYPE, pb);
    char *position_str = pblock_findval(ACL_ATTR_POSITION, pb);
    ACLDbType_t dbtype = ACL_DBTYPE_ANY;
    ACLMethod_t method = ACL_METHOD_ANY;
    ACLAttrGetterFn_t func;
    char err[BIG_LINE];
    NSErr_t *errp = 0;
    int position = ACL_AT_END;

    if (method_str) {
	ACL_REG(ACL_MethodFind(errp, method_str, &method),
		"Method \"%s\" is not registered", method_str);
    }

    if (dbtype_str) {
	ACL_REG(ACL_DbTypeFind(errp, dbtype_str, &dbtype),
		"Database type \"%s\" is not registered", dbtype_str);
    }

    if (!attr || !*attr) {
	pblock_nvinsert("error", "Attribute name is missing", pb);
	return REQ_ABORTED;
    }

    if (!funcStr || !*funcStr) {
	pblock_nvinsert("error", "Attribute getter function name is missing", pb);
	return REQ_ABORTED;
    }

    if (!position_str) {
	if (!strcmp(position_str, "ACL_AT_FRONT")) position = ACL_AT_FRONT;
	else if (!strcmp(position_str, "ACL_AT_END")) position = ACL_AT_END;
	else if (!strcmp(position_str, "ACL_REPLACE_ALL")) position = ACL_REPLACE_ALL;
	else if (!strcmp(position_str, "ACL_REPLACE_MATCHING")) position = ACL_REPLACE_MATCHING;
	else {
	    sprintf(err, "Position attribute \"%s\" is not valid", position_str);
	    pblock_nvinsert("error", err, pb);
	    return REQ_ABORTED;
	}
    }

    func = (ACLAttrGetterFn_t)func_find((char *)funcStr);

    if (!func) {
	sprintf(err, "Could not map \"%s\" to a function", funcStr);
	pblock_nvinsert("error", err, pb);
	return REQ_ABORTED;
    }

    ACL_REG(ACL_AttrGetterRegister(errp, attr, func, method, dbtype, position,
                                   NULL),
	    "Failed to register attribute getter for %s",
	    attr);

    return REQ_PROCEED;
}

int register_method (pblock *pb, Session *sn, Request *rq)
{
    char *method = pblock_findval(ACL_ATTR_METHOD, pb);
    ACLMethod_t t;
    NSErr_t *errp = 0;

    ACL_REG(ACL_MethodRegister (errp, method, &t),
	    "Failed to register method \"%s\"", method);

    return REQ_PROCEED;
}

int set_default_method (pblock *pb, Session *sn, Request *rq)
{
    char *method = pblock_findval(ACL_ATTR_METHOD, pb);
    NSErr_t *errp = 0;
    ACLMethod_t	t;

    ACL_REG(ACL_MethodFind(errp, method, &t),
	    "Method \"%s\" is not registered", method);

    ACL_MethodSetDefault(errp, t);

    return REQ_PROCEED;
}

int set_default_database (pblock *pb, Session *sn, Request *rq)
{
    char *dbname = pblock_findval(ACL_ATTR_DBNAME, pb);
    NSErr_t *errp = 0;

    ACL_REG(ACL_DatabaseSetDefault(errp, dbname),
	    "Failed to set default method \"%s\"", dbname);

    return REQ_PROCEED;
}

int register_module (pblock *pb, Session *sn, Request *rq)
{
    char *module = pblock_findval(ACL_ATTR_MODULE, pb);
    char *funcStr = pblock_findval(ACL_ATTR_MODULEFUNC, pb);
    AclModuleInitFunc func;
    NSErr_t *errp = 0;

    if (!funcStr || !*funcStr) {
	ereport(LOG_SECURITY, "register_module: module init function is missing\n");
	return REQ_ABORTED;
    }

    func = (AclModuleInitFunc)func_find(funcStr);

    if (!func) {
	ereport(LOG_SECURITY, "register_module: could find module init function\n");
	return REQ_ABORTED;
    }

    ACL_REG(ACL_ModuleRegister (errp, module, func),
	    "Failed to register module \"%s\"", module);

    return REQ_PROCEED;
}

int register_database_type(pblock *pb, Session *sn, Request *rq)
{
    char *dbtype_str = pblock_findval(ACL_ATTR_DBTYPE, pb);
    char *parseFuncStr = pblock_findval(ACL_ATTR_PARSEFN, pb);
    ACLDbType_t dbtype = ACL_DBTYPE_INVALID;
    DbParseFn_t parseFunc;
    char err[BIG_LINE];
    NSErr_t *errp = 0;

    if (!dbtype_str || !*dbtype_str) {
	pblock_nvinsert("error", "dbtype is missing", pb);
	return REQ_ABORTED;
    }

    if (!parseFuncStr) {
	pblock_nvinsert("error", "parse function is missing", pb);
	return REQ_ABORTED;
    }

    parseFunc = (DbParseFn_t)func_find(parseFuncStr);

    if (!parseFunc) {
	sprintf(err, "Could map \"%s\" to a function", parseFunc);
	pblock_nvinsert("error", err, pb);
	return REQ_ABORTED;
    }

    ACL_REG(ACL_DbTypeRegister(errp, dbtype_str, parseFunc, &dbtype),
	    "Failed to register database type \"%s\"", dbtype_str);

    return REQ_PROCEED;
}

int register_database_name(pblock *pb, Session *sn, Request *rq)
{
    char *dbtype_str = pblock_findval(ACL_ATTR_DBTYPE, pb);
    char *dbname = pblock_findval(ACL_ATTR_DBNAME, pb);
    char *url = pblock_findval(ACL_ATTR_DATABASE_URL, pb);
    ACLDbType_t dbtype = ACL_DBTYPE_INVALID;
    NSErr_t *errp = 0;
    PList_t plist = PListCreate(NULL, ACL_ATTR_INDEX_MAX, NULL, NULL);

    if (!dbtype_str || !*dbtype_str) {
	pblock_nvinsert("error", "dbtype is missing", pb);
	return REQ_ABORTED;
    }

    if (!dbname || !*dbname) {
	pblock_nvinsert("error", "database name is missing", pb);
	return REQ_ABORTED;
    }

    if (!url || !*url) {
	pblock_nvinsert("error", "database url is missing", pb);
	return REQ_ABORTED;
    }

    ACL_REG(ACL_DbTypeFind(errp, dbtype_str, &dbtype),
	    "Database type \"%s\" is not registered", dbtype_str);

    ACL_REG(ACL_DatabaseRegister(errp, dbtype, dbname, url, plist),
	    "Failed to register database \"%s\"", dbname);

    return REQ_PROCEED;
}

static int init_dbtypes (NSErr_t *errp)
{
    return REQ_PROCEED;
}

static int init_methods (NSErr_t *errp)
{
    /* Register the basic method */
    ACL_REG(ACL_MethodRegister(errp, ACL_AUTHTYPE_BASIC, &ACL_MethodBasic),
	    "Failed to register the method \"%s\"", ACL_AUTHTYPE_BASIC);

    /* Register the ssl method */
    ACL_REG(ACL_MethodRegister(errp, ACL_AUTHTYPE_SSL, &ACL_MethodSSL),
	    "Failed to register the method \"%s\"", ACL_AUTHTYPE_SSL);

    return REQ_PROCEED;
}

static int init_getters (NSErr_t *errp)
{
    int pos = ACL_AT_END;

    /* Initialize getters for method=="any", dbtype == "ldap" */
    ACL_REG(ACL_AttrGetterRegister(errp, ACL_ATTR_USER_ISMEMBER,
				   get_user_ismember_ldap,
				   ACL_METHOD_ANY, ACL_DbTypeLdap, pos, NULL),
	    "Failed to register attr getter for \"%s\"", ACL_ATTR_RAW_USER);
    ACL_REG(ACL_AttrGetterRegister(errp, ACL_ATTR_USER_EXISTS,
				   get_user_exists_ldap,
				   ACL_METHOD_ANY, ACL_DbTypeLdap, pos, NULL),
	    "Failed to register attr getter for \"%s\"", ACL_ATTR_USER_EXISTS);

    /* Initialize getters for method=="basic", dbtype == "any" */
    ACL_REG(ACL_AttrGetterRegister(errp, ACL_ATTR_USER, get_auth_user_basic,
				   ACL_MethodBasic, ACL_DBTYPE_ANY, pos, NULL),
	    "Failed to register attr getter for \"%s\"", ACL_ATTR_USER);
    ACL_REG(ACL_AttrGetterRegister(errp, ACL_ATTR_RAW_USER, get_user_login_basic,
				   ACL_MethodBasic, ACL_DBTYPE_ANY, pos, NULL),
	    "Failed to register attr getter for \"%s\"", ACL_ATTR_RAW_USER);
    ACL_REG(ACL_AttrGetterRegister(errp, ACL_ATTR_RAW_PASSWORD, get_user_login_basic,
				   ACL_MethodBasic, ACL_DBTYPE_ANY, pos, NULL),
	    "Failed to register attr getter for \"%s\"", ACL_ATTR_RAW_PASSWORD);
    ACL_REG(ACL_AttrGetterRegister(errp, ACL_ATTR_AUTHORIZATION,
				   get_authorization_basic,
				   ACL_MethodBasic, ACL_DBTYPE_ANY, pos, NULL),
	    "Failed to register attr getter for \"%s\"", ACL_ATTR_AUTHORIZATION);

    /* Initialize getters for method=="basic", dbtype == "ldap" */
    ACL_REG(ACL_AttrGetterRegister(errp, ACL_ATTR_IS_VALID_PASSWORD,
				   get_is_valid_password_basic_ldap, 
				   ACL_MethodBasic, ACL_DbTypeLdap, pos, NULL),
	    "Failed to register attr getter for \"%s\"",
	    ACL_ATTR_IS_VALID_PASSWORD);

#if defined(CLIENT_AUTH)
    /* Initialize getters for method=="ssl", dbtype == "any" */
    ACL_REG(ACL_AttrGetterRegister(errp, ACL_ATTR_USER_CERT, get_user_cert_ssl,
				   ACL_MethodSSL, ACL_DBTYPE_ANY, pos, NULL),
	    "Failed to register attr getter for \"%s\"", ACL_ATTR_USER_CERT);
    /* Initialize getters for method=="ssl", dbtype == "ldap" */
    ACL_REG(ACL_AttrGetterRegister(errp, ACL_ATTR_USER, get_auth_user_ssl,
				   ACL_MethodSSL, ACL_DbTypeLdap, pos, NULL),
	    "Failed to register attr getter for \"%s\"", ACL_ATTR_USER);

#endif

    return REQ_PROCEED;
}

int init_acl_modules (NSErr_t *errp)
{
    if (init_dbtypes(errp) == REQ_PROCEED &&
	init_methods(errp) == REQ_PROCEED &&
	init_getters(errp) == REQ_PROCEED)
    {
	return 0;
    }
    else {
	return -1;
    }
}

