/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * logbuf.c: General functions to handle buffered log writes. Requires a 
 * threaded server.
 * 
 * Rob McCool
 */


#include "netsite.h"

#ifdef LOG_BUFFERING


#include "logbuf.h"
#include "base/systhr.h"

#include "frame/log.h"         /* log_error */
#include "base/util.h"         /* util_sprintf */

#ifdef XP_UNIX
#include <sys/uio.h>
#endif

/* Leave off by default for now. */
static int logbuf_size = 0, logbuf_flush = DEFAULT_LOGBUF_FLUSH;
static SYS_THREAD syncthread = NULL;

static LOGBUF_T *all_logs = NULL;


/* Assumes you've entered lb->free_crit */
struct logdata_s *logbuf_free_remove(LOGBUF_T *lb)
{
    struct logdata_s *ret;

    /* Take one off the free list */
    ret = lb->free_list;
    lb->free_list = ret->next;

    ret->next = NULL;
    return ret;
}



/* Assumes you're in ldflush->crit and ldflush->log->crit */
void logbuf_flush_add(struct logdata_s *ldflush)
{
    crit_enter(ldflush->log->flush_crit);
    if(ldflush->log->flush_list) {
        ldflush->log->flush_list_end->next = ldflush;
    }
    else {
        ldflush->log->flush_list = ldflush;
        ldflush->log->flush_list_end = ldflush;
    }
    ldflush->next = NULL;
    crit_exit(ldflush->log->flush_crit);
}


/*
   robm - This function runs every 60 seconds and flushes the data from all 
   of the log caches. Notice that it is created not during Init phase, but 
   when a function enters clf_record and notices that the thread is not
   present. There is also no mutex protecting this data.

   I chose to avoid having to enter and exit a global mutex on each CLF
   record by having this function check its thread id periodically. If it's
   not the one listed in the global, the thread exits. I believe this will
   pay the price of starting too many threads every so often, but will avoid
   global mutex contention.
 */
static void logbuf_sync(void *unused)
{
    while(1) {
        LOGBUF_T *bufp;
        struct logdata_s *datap, *dataq, *flushbegin, *flushend;
        struct logdata_thread_s *thrp;
        int flushlen;

        /* If someone went and created another thread, exit by returning. */
        /* Also exit if we've restarted, XXXrobm race here */
        if((syncthread != systhread_current()) || (!all_logs))
            return;

        systhread_sleep(logbuf_flush);
        for(bufp = all_logs; bufp; bufp = bufp->next) {
            /* Get full buffers from flush list */
            crit_enter(bufp->flush_crit);
            flushbegin = bufp->flush_list;
            flushend = bufp->flush_list_end;
            bufp->flush_list = NULL;
            bufp->flush_list_end = NULL;
            flushlen = bufp->flush_len;
            bufp->flush_len = 0;
            crit_exit(bufp->flush_crit);

            /* Get partial buffers from thread list */
            crit_enter(bufp->thread_crit);
            /* 
               Pull all partial buffers onto the end of our flush list.
             */
            for(thrp = bufp->thread_list; thrp; thrp = thrp->next) {
                /* Take away each thread's buffer, and let it go and get 
                   another one. */
                crit_enter(thrp->crit);
                datap = thrp->logdata;
                thrp->logdata = NULL;
                crit_exit(thrp->crit);

                if(datap) {
                    ++flushlen;
                    if(flushbegin) {
                        flushend->next = datap;
                        datap->next = NULL;
                        flushend = datap;
                    }
                    else {
                        flushbegin = datap;
                        flushend = datap;
                    }
                }
            }
            crit_exit(bufp->thread_crit);

            if(flushbegin) {
                /* Now flush them */
#ifdef XP_UNIX
                struct iovec *iov;
                int x;
                iov = (struct iovec *) MALLOC(flushlen * sizeof(struct iovec));
                for(x=0, datap = flushbegin; datap; ++x, datap = datap->next) {
                    iov[x].iov_base = (caddr_t) datap->data;
                    iov[x].iov_len = datap->pos;
                    datap->pos = 0;
                }
                writev(PR_FileDesc2NativeHandle(bufp->fd), iov, flushlen);
                FREE(iov);
#else
                for(datap = flushbegin; datap; datap = datap->next) {
                    system_fwrite_atomic(bufp->fd, datap->data, datap->pos);
                    datap->pos = 0;
                }
#endif
                /* Now put them all on free list */
                crit_enter(bufp->free_crit);
                flushend->next = bufp->free_list;
                bufp->free_list = flushbegin;
                crit_exit(bufp->free_crit);
            }
        }
    }
}


struct logdata_s *logbuf_alloc(LOGBUF_T *lb)
{
    struct logdata_s *ret;

    ret = (struct logdata_s *) PERM_MALLOC(sizeof(struct logdata_s));
    ret->data = (char *) PERM_MALLOC(logbuf_size);
    ret->pos = 0;
    ret->log = lb;
    ret->next = NULL;

    return ret;
}

struct logdata_s *logbuf_new_data(LOGBUF_T *lb)
{
    struct logdata_s *ret;

    ret = NULL;
    crit_enter(lb->free_crit);
    if(lb->free_list) {
        ret = logbuf_free_remove(lb);
    }
    /* Exit now to make sure we don't tie up the crit for long */
    crit_exit(lb->free_crit);

    if(!ret) {
        /* Or allocate a new one */
        ret = logbuf_alloc(lb);
    }
    return ret;
}


void logbuf_write(LOGBUF_T *lb, SYS_FILE fd, char *buf, int len)
{
    struct logdata_thread_s *threadp;
    struct logdata_s *datap;

    if(!logbuf_size) {
        system_fwrite_atomic(fd, buf, len);
        return;
    }

    /* WHY NO MUTEX? See above. */
    if(logbuf_flush && (syncthread == NULL)) {
        if(!(syncthread = systhread_start(1, 65536, logbuf_sync, NULL))) {
            log_error(LOG_FAILURE, "logbuf-alloc", NULL, NULL, 
                      "could not create new thread (%s)", system_errmsg());
            system_fwrite_atomic(fd, buf, len);
            return;
        }
    }

    /* See if we have a per-thread data struct yet */
    threadp = (struct logdata_thread_s *) systhread_getdata(lb->data_key);
    if(!threadp) {
        threadp = (struct logdata_thread_s *) 
            PERM_MALLOC(sizeof(struct logdata_thread_s));
        threadp->crit = crit_init();
        threadp->logdata = NULL;
        crit_enter(lb->thread_crit);
        threadp->next = lb->thread_list;
        lb->thread_list = threadp;
        crit_exit(lb->thread_crit);
        systhread_setdata(lb->data_key, (void *) threadp);
    }

    crit_enter(threadp->crit);
    datap = threadp->logdata;

    if(!datap) {
        datap = logbuf_new_data(lb);
        threadp->logdata = datap;
    }
    else {
        /* See the data will fit into what's left of the buffer */
        /* This assumes the line will fit in the empty buffer */
        if((datap->pos + len + 1) >= logbuf_size) {
            logbuf_flush_add(datap);
            datap = logbuf_new_data(lb);
            threadp->logdata = datap;
        }
    }
    strncpy(&datap->data[datap->pos], buf, len);
    datap->pos += len;
    crit_exit(threadp->crit);
}


struct logdata_s *logbuf_destroy_data(struct logdata_s *datap)
{
    LOGBUF_T *lb;
    struct logdata_s *ret;

    lb = datap->log;
    if(datap->pos)
        system_fwrite_atomic(lb->fd, datap->data, datap->pos);
    PERM_FREE(datap->data);
    ret = datap->next;
    PERM_FREE(datap);
    return ret;
}

struct logdata_thread_s *logbuf_destroy_thrdata(struct logdata_thread_s *thrp)
{
    struct logdata_thread_s *ret;
    struct logdata_s *datap = thrp->logdata;

    if(datap)
        (void) logbuf_destroy_data(datap);
    ret = thrp->next;
    crit_terminate(thrp->crit);
    PERM_FREE(thrp);
    return ret;
}


void logbuf_close(LOGBUF_T *lb)
{
    if(logbuf_size) {
        LOGBUF_T *lbp, *lbq;
        struct logdata_thread_s *thrp;
        struct logdata_s *datap;

        lbp = all_logs;
        while(lbp) {
            crit_enter(lbp->flush_crit);
            datap = lbp->flush_list; 
            while(datap)
                datap = logbuf_destroy_data(datap);
            crit_exit(lbp->flush_crit);
            crit_terminate(lbp->flush_crit);

            crit_enter(lbp->free_crit);
            datap = lbp->free_list;
            while(datap)
                datap = logbuf_destroy_data(datap);
            crit_exit(lbp->free_crit);
            crit_terminate(lbp->free_crit);

            crit_enter(lbp->thread_crit);
            thrp = lbp->thread_list;
            while(thrp)
                thrp = logbuf_destroy_thrdata(thrp);
            crit_exit(lbp->thread_crit);
            crit_terminate(lbp->thread_crit);

            /* fd is closed by the logging code */

            lbq = lbp->next;
            PERM_FREE(lbp);
            lbp = lbq;
        }
        all_logs = NULL;
    }
}




int logbuf_init(pblock *args)
{ 
    char err[MAGNUS_ERROR_LEN];
    char *t;

    logbuf_size = 0;
    logbuf_flush = DEFAULT_LOGBUF_FLUSH;
    if( (t = pblock_findval("buffer-size", args)) ) {
        logbuf_size = atoi(t);
        if(logbuf_size < 0) {
            util_snprintf(err, MAGNUS_ERROR_LEN, "buffer size of %d is invalid", logbuf_size);
            pblock_nvinsert("error", err, args);
            return REQ_ABORTED;
        }
        if(logbuf_size < (REQ_MAX_LINE * 2))
            logbuf_size = REQ_MAX_LINE * 2;
    }
    if( (t = pblock_findval("buffer-flush", args)) ) {
        logbuf_flush = atoi(t) * 1000;
        if(logbuf_flush < 0) {
            util_snprintf(err, MAGNUS_ERROR_LEN,"flush rate of %d seconds is invalid", 
                         logbuf_flush);
            pblock_nvinsert("error", err, args);
            return REQ_ABORTED;
        }
    }
    return REQ_PROCEED;
}


LOGBUF_T *logbuf_new(pblock *pb, SYS_FILE fd)
{
    LOGBUF_T *lb;

    if(!logbuf_size)
        return NULL;

    lb = (LOGBUF_T *) PERM_MALLOC(sizeof(struct logbuf_s));

    lb->fd = fd;
    lb->data_key = systhread_newkey();

    lb->free_crit = crit_init();
    lb->free_list = NULL;
    lb->thread_crit = crit_init();
    lb->thread_list = NULL;
    lb->flush_crit = crit_init();
    lb->flush_list = NULL;
    lb->flush_list_end = NULL;

    lb->next = all_logs;
    all_logs = lb;
    return lb;
}
#endif /* LOG_BUFFERING */
