/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * ntcgi.c: Handle Common Gateway Interface scripts
 *
 * CGI is an interface definition for external programs to the server which
 * can be executed to produce documents on the fly. Handy for forms etc.
 * 
 * Rob McCool
 */


#include "netsite.h"
#include "safs/cgi.h"
#include "base/pblock.h"
#include "base/session.h"
#include "frame/req.h"
#include "frame/log.h"


#include "base/util.h"      /* environment functions, can_exec */
#include "base/net.h"       /* SYS_NETFD */
#include "base/daemon.h"    /* child_fork */
#include "frame/protocol.h" /* protocol_status */
#include "frame/http.h"     /* http_hdrs2env */
#include "frame/conf.h"     /* server action globals */

#include "frame/httpact.h"
#include <frame/objset.h>

#include <base/buffer.h>
#include <base/daemon.h>
#include <base/file.h>

#include <nspr.h>
#include <private/pprio.h>
#ifdef NET_SSL
extern "C" {
#include <ssl.h>
#include <key.h>
#include <cert.h>
#include <pk11func.h>
}
#endif

#include "prio.h"

#define port conf_getglobals()->Vport

#define ENV_CGI
#define PIPE_BUFFERSIZE 8192

#ifdef MCC_ADMSERV
extern "C" char **admin_cgivars(Session *sn, Request *rq, DWORD ppid, SYS_FILE pwpipe, char **env);
#endif /* MCC_ADMSERV */

#define CGI_VERSION "CGI/1.1"
#define CGI_READ_TIMEOUT 300
#define DEFAULT_PATH "c:\\bin;c:\\DOS;c:\\windows"

static pblock *initenv = NULL;
static char **initenv_vars = NULL;

extern PRPollDesc *poll_desc;
extern int poll_index;


/* ----------------------------- _find_exe_path ----------------------------- */


/*
 * _find_exe_path takes a path and returns its file association if there is one.
 */

char *_find_exe_path(char *path)
{
    HINSTANCE result;
    char *association;

    if (path == NULL)
        return NULL;

    association = (char *)MALLOC(MAX_PATH);

    result = FindExecutable(path, NULL, association);
    if (result <= (HINSTANCE)32) {
        FREE(association);
        if (result == (HINSTANCE)31) {
            /* No association is present */
            return NULL;
        } else {
            /* some other error returned by the call. record it. */
             ereport(LOG_WARN, "send-cgi:"
                "failed to get associated file for (%s) Error %d ", path, result);
             return NULL;
        }
    } else {
        return association;
    }
}

/* ----------------------------- _create_argv ----------------------------- */


/*
 * _create_argv takes a program_name, a query_string (URL-encoded)
 * and possibly a shellcgi_path and converts it to command line parameters. 
 * Doesn't handle "foo bar" being one argument, etc. or $LOGNAME stuff.
 *
 * There are two special cases to fix here.
 * 
 * First, the spec states that if there is a "=" in the query string, then the
 * query string is NOT passed on the command line to the CGI program.  However,
 * for ISINDEX queries do want the query string passed on the command line- so
 * to differentiate the ISINDEX queries from the normal queries, ISINDEX queries
 * will send the "=" in its encoded format.
 *
 * The second function of this routine is to check for any character which 
 * is a special bourne-shell character is escaped with a preceding backslash.
 *
 * A side-effect of this function is to unescape/backslash the query string 
 * (args). XXX- Looks like there is a bug where the query string does not
 * unescaped if there was an = in the command line... We're too close to 
 * release to fix right now.
 *
 */

char *_create_argv(char *program_name, char *orig_query_string, char *shellcgi_path){
    register int x;
    int ChildCmdLineLength;
    char *w, *ChildCmdLine, c, *l, *t, *ChildCmdLinePointer;

    char *query_string = (orig_query_string) ? STRDUP(orig_query_string) : NULL;

	/*
     * First check if there is an equal character 
     */
	if (!query_string || (strchr(query_string,'='))) {
        /*
         * The returned command line will not have the query string in it.
         */
        if (shellcgi_path) {
			int length = strlen(program_name) + 3 + strlen(shellcgi_path);
            ChildCmdLine = (char *)MALLOC(length);
            util_snprintf(ChildCmdLine, length, "%s %s\0", program_name, 
                shellcgi_path);
        } else {
    		ChildCmdLine = (char *)MALLOC(strlen(program_name) + 2);
    		strcpy(ChildCmdLine,program_name);
    		strcat(ChildCmdLine, " \0");
        }
		if(query_string) FREE(query_string);
		return ChildCmdLine;
	}

	/* 
     * The returned command line will have the query string in it 
     */
    ChildCmdLineLength = strlen(query_string)*2 + 1 + strlen(program_name) + 2;
    if (shellcgi_path)
        ChildCmdLineLength += strlen(shellcgi_path);
    ChildCmdLine = (char *) MALLOC(ChildCmdLineLength);

	strcpy(ChildCmdLine,program_name);
	strcat(ChildCmdLine, " ");

    if (shellcgi_path) {
    	strcat(ChildCmdLine, shellcgi_path);
    	strcat(ChildCmdLine, " ");
    }
	ChildCmdLinePointer = ChildCmdLine + strlen(program_name) + 1;
    if (shellcgi_path)
        ChildCmdLinePointer += strlen(shellcgi_path) + 1;

    /*
     * unescape the query string
     */
    for(x=0;query_string[x];x++) {
        if(query_string[x] == '+') {
            query_string[x] = ' ';
        }
    }
    l = query_string;
    util_uri_unescape(l);
    x = 0; w = l;

    /* 
     * put a backslash before any special bourne-shell characters 
     */
    while(1) {
        if((!l[x]) || (l[x] == ' ')) {
            c = l[x];
            l[x] = '\0';
    		for(t = w; *t; ++t, ++ChildCmdLinePointer) {
        		if(strchr("&;`'\"|*?~<>^()[]{}$\\ #!", *t))
            		*ChildCmdLinePointer++ = '\\';
        		*ChildCmdLinePointer = *t;
		    }
    		*ChildCmdLinePointer++ = ' ';
            if(c)
                w = &l[x+1];
            else
                break;
        }
        ++x;
    }
	*ChildCmdLinePointer = '\0';
	if(query_string) FREE(query_string);
	return ChildCmdLine;
}



/* -------------------------- cgi_specific_vars --------------------------- */

#define MAX_COMMON_VARS 19
#define MAX_CGI_VARS (MAX_COMMON_VARS+7)
#define MAX_CGI_SSI_VARS (MAX_CGI_VARS+7)

char **
cgi_specific_vars(Session *sn, Request *rq, char *args, char **env)
{
    int x, y;
    register pblock *pb;
    char c, *t, *u;

    env = util_env_create(env, MAX_CGI_VARS, &x);

    pb = rq->reqpb;

    env[x++] = util_env_str("GATEWAY_INTERFACE", CGI_VERSION);

    env[x++] = util_env_str("SERVER_PROTOCOL", 
                pblock_findval("protocol", pb));

    env[x++] = util_env_str("REQUEST_METHOD",
                pblock_findval("method", pb));

    t = pblock_findval("uri", pb);

    /* Simulate CGI URIs by truncating path info */
    if( (u = pblock_findval("path-info", rq->vars)) ) {
        y = strlen(t) - strlen(u);
        c = t[y];
        t[y] = '\0';
        env[x++] = util_env_str("SCRIPT_NAME", t);
        t[y] = c;

        env[x++] = util_env_str("PATH_INFO", u);
        if((t = request_translate_uri(u, sn))) {
            env[x++] = util_env_str("PATH_TRANSLATED", t);
            FREE(t);
        }
    }
    else  {
        env[x++] = util_env_str("SCRIPT_NAME", t);
    }

    if(args) 
        env[x++] = util_env_str("QUERY_STRING", args);

    env[x] = NULL;
    return env;
}


/* --------------------------- cgi_ssi_vars ---------------------------- */

/* passing SSI variables as environment vars for SSI CGIs.  These also
 * apply to regular CGI's as well.  FIXME: This is duplicated code for both
 * Unix and NT. - mikep 
 */
char **cgi_ssi_vars(Session *sn, Request *rq, char *args, char **env)
{
    int x;
    register pblock *pb;
    char *t;
 
   env = util_env_create(env, MAX_CGI_SSI_VARS,  &x);

   pb = rq->reqpb;

   if (t = pblock_findval("DOCUMENT_URI", pb))	
     env[x++] = util_env_str("DOCUMENT_URI", t);

   if (t = pblock_findval("DOCUMENT_NAME",  pb))
     env[x++] = util_env_str("DOCUMENT_NAME", t);

   if (t = pblock_findval("DATE_LOCAL", pb))
     env[x++] = util_env_str("DATE_LOCAL", t);

   if (t = pblock_findval("DATE_GMT", pb))
     env[x++] = util_env_str("DATE_GMT", t);

   if (t = pblock_findval("LAST_MODIFIED",  pb))
     env[x++] = util_env_str("LAST_MODIFIED", t);

   if (t = pblock_findval("QUERY_STRING_UNESCAPED",  pb))
     env[x++] = util_env_str("QUERY_STRING_UNESCAPED", t);

   env[x] = NULL;
   return env;
}
 
/* --------------------------- cgi_common_vars ---------------------------- */

char **cgi_common_vars(Session *sn, Request *rq, char **env) 
{
    char *t, *ip, tp[15];
    int x;

    env = util_env_create(env,MAX_COMMON_VARS,&x);
    
    if(!(t = getenv("PATH")))
        t = DEFAULT_PATH;
    env[x++] = util_env_str("PATH", t);

    /* winsock needs this to be set */
    if(t = getenv("SystemRoot"))
        env[x++] = util_env_str("SystemRoot", t);

    if( (t = getenv("TZ")) )
        env[x++] = util_env_str("TZ", t);
    if( (t = getenv("LANG")) )
        env[x++] = util_env_str("LANG", t);

    env[x++] = util_env_str("SERVER_SOFTWARE", MAGNUS_VERSION_STRING);

    util_itoa(port, tp);
    env[x++] = util_env_str("SERVER_PORT", tp);

    env[x++] = util_env_str("SERVER_NAME", server_hostname);

    t = http_uri2url_dynamic("","",sn,rq);
    env[x++] = util_env_str("SERVER_URL", t);
    FREE(t);

    ip = pblock_findval("ip", sn->client);
    t = session_dns(sn);
    env[x++] = util_env_str("REMOTE_HOST", (t ? t : ip));
    env[x++] = util_env_str("REMOTE_ADDR", ip);

    if((t = pblock_findval("auth-user", rq->vars))) {
        env[x++] = util_env_str("REMOTE_USER", t);
        env[x++] = util_env_str("AUTH_TYPE", 
                                pblock_findval("auth-type", rq->vars));
    }

#ifdef CLIENT_AUTH
    if ((t = pblock_findval("auth-cert", rq->vars))) {
    /* Alan made up this environment variable name */
        env[x++] = util_env_str("CLIENT_CERT", t);
    }
#endif

#ifdef NET_SSL
#define NUM_RBYTES 32
    if(security_active)  {
        char *random_bytes = (char *) MALLOC(NUM_RBYTES+2);
	char *random_string = (char *) MALLOC(NUM_RBYTES*2+2);
        char *sslid = pblock_findval("ssl-id", sn->client);
	register int i;

        env[x++] = util_env_str("HTTPS", "ON");

        if (pblock_findval("keysize", sn->client) != NULL)
          env[x++] = util_env_str("HTTPS_KEYSIZE",
                                  pblock_findval("keysize", sn->client));

        if (pblock_findval("secret-keysize", sn->client) != NULL)
          env[x++] = util_env_str("HTTPS_SECRETKEYSIZE",
                                  pblock_findval("secret-keysize", sn->client));

        env[x++] = util_env_str("HTTPS_SESSIONID",
                                sslid?sslid:"");
        PK11_GenerateRandom((unsigned char *)random_bytes, NUM_RBYTES);
        for(i=0; i < NUM_RBYTES; i++)  {
            sprintf(&random_string[i*2], "%02x", 
		    (unsigned int) (random_bytes[i] & 0xff));
        }
        random_string[NUM_RBYTES*2]='\0';
        env[x++] = util_env_str("HTTPS_RANDOM", random_string);
        FREE(random_string);
        FREE(random_bytes);
    }  else  {
        env[x++] = util_env_str("HTTPS", "OFF");
    }
#endif

    env[x] = NULL;
    return env;
}

/* --------------------------- cgi_scan_headers --------------------------- */


#define BAD_CGI(s1, s2) \
    util_sprintf(err, s1, s2); \
    log_error(LOG_FAILURE, "cgi-parse-output", sn, rq, \
              "the CGI program %s did not produce a valid header (%s)", \
              pblock_findval("path", rq->vars), err); \
    return REQ_ABORTED;


int cgi_scan_headers(Session *sn, Request *rq, filebuf_t *nb)
{
    register int x,y;
    register char c;
    int nh, i, alen, asz;
	filebuf_t *buf = nb;
    pb_param *acc;
    char t[REQ_MAX_LINE], err[REQ_MAX_LINE];

    nh = 0;
    x = 0; y = -1;
    acc = NULL; alen = 0; asz = 0;
    while(1) {
        if(x > REQ_MAX_LINE) {
            BAD_CGI("CGI header line too long (max %d)", REQ_MAX_LINE);
        }
        if(nh > HTTP_MAX_HEADERS) {
            BAD_CGI("too many headers from CGI script (max %d)",
                        HTTP_MAX_HEADERS);
        }
		i = pipebuf_getc(buf);

        if(i == IO_ERROR) {
            BAD_CGI("read failed, error is %s", system_errmsg());
        }
        else if(i == IO_EOF) {
            BAD_CGI("program terminated without a valid CGI header"
            	" (check for core dump or other abnormal termination", "");
        }

        c = (char) i;
        switch(c) {
          case LF:
            if((!x) || ((x == 1) && (t[0] == CR)))
                return REQ_PROCEED;

            if(t[x-1] == CR)
                --x;
            t[x] = '\0';
            if(y == -1) {
                /* Ignore CRLF on a single line entry after a message header */
                return REQ_PROCEED;
            }
            while(t[y] && isspace(t[y])) ++y;

            pblock_nvinsert(t, &t[y], rq->srvhdrs);

            x = 0; y = -1;
            ++nh;
            break;
          case ':':
            if(y == -1) {
                y = x+1;
                c = '\0';
            }
          default:
            t[x++] = ((y == -1) && isupper(c) ? tolower(c) : c);
        }
    }
}


/* --------------------------- cgi_send_content --------------------------- */


/* 
 * This routine is not forgiving of pinheads: If content-length
 * or content-type is not right, it punts.
 */
#define TMP_BUF_SIZE	8192
int cgi_send_content(Session *sn, Request *rq, SYS_FILE s2c)
{
    char *cls, *cts;
    int cl, x;

	int result, netresult;
	char tmp_buf[TMP_BUF_SIZE];
	struct timeval timeout = {0, 0};

    if((request_header("content-length", &cls, sn, rq) == REQ_ABORTED) || 
       (request_header("content-type", &cts, sn, rq) == REQ_ABORTED))
        return REQ_ABORTED;

    if((!cls) || (!cts))
        return REQ_PROCEED;

    while(isspace(*cls)) ++cls;
    for(x = 0; cls[x] && (!isspace(cls[x])); ++x) {
        if(!isdigit(cls[x])) {
            log_error(LOG_WARN, "send-cgi", sn, rq, 
                      "client sent bad content-length (%s)", cls);
            return REQ_ABORTED;
        }
    }

    cl = atoi(cls);
    if(cl < 0) {
        log_error(LOG_WARN, "send-cgi", sn, rq, 
                  "client sent bad content-length (%s)", cls);
        return REQ_ABORTED;
    }
   clear_poll_desc(poll_desc);
   add_poll_desc(sn->inbuf->sd, &poll_desc);

    if(pipebuf_netbuf2pipe(sn->inbuf, s2c, cl) == IO_ERROR) {
		result = GetLastError();
		while(TRUE) {
         clear_poll_desc(poll_desc);
         add_poll_desc(sn->inbuf->sd, &poll_desc);

			/* Select on NT behaves differently from the UNIX
			 * version; NT will return that the socket is ready
			 * for reading even if it is closed.
			 */
/*
			netresult = PR_Select(1, &readfds, NULL, NULL, 0);
*/
         netresult = PR_Poll(poll_desc, poll_index, 0);

			if (netresult == SOCKET_ERROR) {
					result = WSAECONNRESET;
					break;
			} else if (exists_poll_desc(sn->inbuf->sd, poll_desc)) {
				if ( (netresult = net_read(sn->inbuf->sd, tmp_buf, 
					TMP_BUF_SIZE, sn->inbuf->rdtimeout)) <= 0) {
					result = WSAECONNRESET;
					break;
				} else {
					/* loop */;
				}
			} else {
				break;
			}
		}
		
		switch (result) {
				/* If the pipe closed, that means the cgi program terminated
				 * This could be normal; so continue trying to send back data.
				 */
			case ERROR_BROKEN_PIPE:
			case ERROR_BAD_PIPE:
			case ERROR_PIPE_BUSY:
			case ERROR_PIPE_LISTENING:
			case ERROR_PIPE_NOT_CONNECTED:
			case ERROR_NO_DATA:
				break;
			default:
	            log_error(LOG_FAILURE, "send-cgi", sn, rq, 
	                      "error occurred while sending content to script (%s)",
	                      sn->inbuf->errmsg);
	            return REQ_ABORTED;
		}
    }

	/* Kludge, kludge, kludge. This is a workaround to get around
	 * the problem of the browser sending two extra bytes
	 * "\r\n" as part of post data. These two bytes are not counted
	 * in content-length and if they remain unreceived in the socket
	 * buffer the server does a hard reset of the connection upon
	 * the closesocket call
	 */
/*
	result = PR_Select(1, &readfds, NULL, NULL, 0);
*/
   result = PR_Poll(poll_desc, poll_index, 0);

	if (result == SOCKET_ERROR) {
		ereport(LOG_INFORM, "CGI select failure (%s)",
			system_errmsg());
		return REQ_ABORTED;
	} else if (result){
		if ( (result = net_read(sn->inbuf->sd, tmp_buf, TMP_BUF_SIZE, 
				sn->inbuf->rdtimeout)) <= 0 ) {
			log_error(LOG_FAILURE, "send-cgi", sn, rq, 
				"error occurred while reading last data from client (%s)",
				system_errmsg());
			return REQ_ABORTED;
		}
	}
    return REQ_PROCEED;
}

#ifdef MCC_ADMSERV
/* MLM - send a username and password pair to a CGI.  This shouldn't really
 *       be cut and pasted but these cgi.cpp and ntcgi.cpp don't have an 
 *       common area. */
int adm_cgi_send_password(Session *sn, Request *rq, SYS_FILE fd)
{
    char *user, *pass, *auth, *userDN, *ans;
    char *siepwd, *sievalue;
    int r;
#ifdef FORTEZZA
    char *pin;
#endif /* FORTEZZA */

    /* Yuck.  Well, if there's no user, I guess we'll send a nonsense
     * string rather than tearing down the pipe. */
    user = pblock_findval("auth-user", rq->vars);
    pass = pblock_findval("auth-password", rq->vars);
    userDN = pblock_findval("userdn", rq->vars);
    r    = request_header("authorization", &auth, sn, rq);
    if( (r == REQ_ABORTED) )
        return REQ_ABORTED;
 
    if(!user)  {
        user = ADM_NO_VALUE_STRING;
    }
    if(!pass)  {
        pass = ADM_NO_VALUE_STRING;
    }
    if(!auth)  {
        auth = ADM_NO_VALUE_STRING;
    }
    if(!userDN)  {
        userDN = ADM_NO_VALUE_STRING;
    }

    siepwd  = admGetCachedSIEPWD();
    sievalue = (siepwd?siepwd:ADM_NO_VALUE_STRING);

#ifdef FORTEZZA

    pin = NULL;
    pin = FortezzaGetAuthPin();
    if(!pin)  {
	pin = ADM_NO_VALUE_STRING;
    }

    ans = (char *) MALLOC(strlen(user) + strlen(ADM_USER_STRING) +
                          strlen(pass) + strlen(ADM_PASS_STRING) + 
                          strlen(auth) + strlen(ADM_AUTH_STRING) +
			              strlen(pin) + strlen(ADM_PIN_STRING) +
                          strlen(userDN) + strlen(ADM_USERDN_STRING) +
                          strlen(sievalue) + strlen(ADM_SIEPWD_STRING) +
                          20);
    sprintf(ans, "%s%s\n%s%s\n%s%s\n%s%s\n%s%s\n%s%s\n",
        ADM_USER_STRING, user, 
        ADM_PASS_STRING, pass, 
        ADM_AUTH_STRING, auth,
        ADM_PIN_STRING, pin,
        ADM_USERDN_STRING, userDN,
        ADM_SIEPWD_STRING, sievalue,
        );
#else

    ans = (char *) MALLOC(strlen(user) + strlen(ADM_USER_STRING) +
                          strlen(pass) + strlen(ADM_PASS_STRING) + 
                          strlen(auth) + strlen(ADM_AUTH_STRING) +
                          strlen(userDN) + strlen(ADM_USERDN_STRING) +
                          strlen(sievalue) + strlen(ADM_SIEPWD_STRING) +
                          16);
    sprintf(ans, "%s%s\n%s%s\n%s%s\n%s%s\n%s%s\n",
        ADM_USER_STRING, user, 
        ADM_PASS_STRING, pass, 
        ADM_AUTH_STRING, auth,
        ADM_USERDN_STRING, userDN,
        ADM_SIEPWD_STRING, sievalue
        );
#endif /* FORTEZZA */

 
 
    /* strlen(ans)+1 to include null */
    r = system_pwrite(fd, ans, strlen(ans)+1);
    FREE(ans);

#ifdef FORTEZZA
    if(strcmp(pin, ADM_NO_VALUE_STRING) != 0)
        FortezzaDestroyAuthPin(pin);
#endif

    if (siepwd) free(siepwd);

    if(r==-1)  {
        if(errno != EPIPE)  {
            log_error(LOG_FAILURE, "send-cgi", sn, rq, 
                      "error occurred while sending username to script (%s)",
                      system_errmsg());
            return REQ_ABORTED;
        }
    }
    return REQ_PROCEED;
}
#endif

char *
_build_cwd(char *path)
{
   /* EXACTLY HOW HOSED IS THE FILE_PATHSEP CRAP ANYWAY??? XXXMB */
    char *new_path;
    char *ptr;

    if (!path)
        return NULL;

    if (*path == 0) {
	/* Make sure it can hold at least 2 chars -- for NT */
	new_path = (char *)MALLOC(2);
	*new_path = 0;
    }
    else {
	/* intentionally overallocate; we don't really care */
	new_path = (char *)STRDUP(path);
    }

    if (new_path) {
        if ( (ptr = strrchr(new_path, '/')) ) 
            *ptr = '\0';
        else if  ( (ptr = strrchr(new_path, '\\')) ) 
	    *ptr = '\0';
	else
	    *new_path = 0;

#ifdef XP_WIN32
	if (*new_path == 0) {
	    /* NT doesn't like the NULL path when used in CreateProcess */
	    *new_path = '\\';
	    *(new_path+1) = 0;
	}
#endif /* XP_WIN32 */
    }

    return new_path;
}

/* _convert_unix_to_nt_env()
 * Unix environments are stored as an array of character ptrs.  NT stores
 * the environment as one big long string; it is terminated by two 
 * NULL characters.  Each individual entry is delinieated with a single
 * NULL character.
 */
char *
_convert_unix_to_nt_env(char **env)
{
    int total_len = 0;
    char **env_ptr;
    char *newenv;
    char *newenv_ptr;
    char *tmp;

    for (env_ptr = env; *env_ptr; env_ptr++)
        total_len += (strlen(*env_ptr)+1);
    total_len++;  /* terminating NULL byte */

    if ( (newenv = (char *)MALLOC(sizeof(char) * total_len)) == NULL)
        return NULL;

    newenv_ptr = newenv;
    for (env_ptr = env; *env_ptr; env_ptr++) {
        for (tmp = *env_ptr; *tmp; tmp++) {
            *newenv_ptr = *tmp;
            newenv_ptr++;
        }
        *newenv_ptr = '\0';
        newenv_ptr++;
    }
    *newenv_ptr = '\0';
    
    return newenv;
}

/* _cgi_is_dangerous
 * Check for .BAT or .CMD scripts which might have bad args passed to them.
 * Returns 1 if the request looks "dangerous", 0 otherwise.
 */
_cgi_is_dangerous(char *path, char *cmd_line, pblock *pb)
{
	/* Fix .BAT/.CMD security hole
	 * Note- everything has been unescaped at this point 
	 * Note- path has all forward slashes at this point
	 */
	char *ext, *last_slash;

	if (!(ext = strrchr(path, '.')))
		return 0;

	/* Do a quick check for ".[bB]" or ".[cC]" */
	if ( ext[1] != 'b' && ext[1] != 'B' &&
		ext[1] != 'c' && ext[1] != 'C')
		return 0;

	last_slash = strrchr(path, '/');

	if ( ext > last_slash ) {
		ext++;
		/* A ".bat" pattern will only be allowed if the character following the
		 * ".bat" is alphanumeric or ".".  Paths which end with "." have already
		 * been weeded out in the clean-uri function.
		 */
		if ( (!strncasecmp(ext, "bat", 3) ||
			!strncasecmp(ext, "cmd", 3)) &&
			(!ext[3] || (!isalpha(ext[3]) && !isdigit(ext[3])) && ext[3] != '.') ) {

			/* ok, this is a .BAT script.  Disallow &, |, <, >.
			 * Allow these if 'dangerous' is enabled in the obj.conf.
			 */
			if (!pblock_findval("allow-dangerous", pb)) {
				if ( strchr(cmd_line, '&') ||
					strchr(cmd_line, '|') ||
					strchr(cmd_line, '>') ||
					strchr(cmd_line, '<') ) {
						return 1;
				}
			}
		}
	}
	return 0;
}

#define WAIT_INTERVAL	30000 	/* 30 secs */
#define TERMINATE_TRIES	5
void
wait_for_child_death(HANDLE pid, int wait_flag, int terminate_flag)
{
	int tries;
	int result = FALSE;
	unsigned long exit_code;

	/* Check to see if the process finished */
	if ( GetExitCodeProcess(pid, &exit_code) ) {
		if ( exit_code != STILL_ACTIVE ) 
			/* cool- the process is already dead */
			goto killed;
	} else 
		/* If GetExitCodeProcess() failed, pid is most likely invalid
		 * Don't call CloseHandle() again.
		 */
		return;

	/* First, wait for the damn thing to die- up to 30 seconds? */
	if ( wait_flag ) {
		if ((WaitForSingleObject(pid, WAIT_INTERVAL)) == WAIT_FAILED) {
			ereport(LOG_CATASTROPHE, "wait for CGI child failed (%s)",
				system_errmsg());
			goto killed;
		}
	}

	if ( terminate_flag ) {
		for (tries=TERMINATE_TRIES; tries; tries--) {
			if (TerminateProcess(pid, 1) == FALSE) {
				if ( GetExitCodeProcess(pid, &exit_code) ) {
					if ( exit_code != STILL_ACTIVE ) 
						/* cool- the process is already dead */
						goto killed;
				}

				if (WaitForSingleObject(pid, WAIT_INTERVAL) == WAIT_FAILED) 
					ereport(LOG_CATASTROPHE, "wait2 for CGI child failed (%s)", 
						system_errmsg());
			} else 
				goto killed;
		}
	}

	/* Drat; we never did kill the darn thing... */
	ereport(LOG_FAILURE, "cgi-send: TerminateProcess(pid %d) returned %s",
		pid, system_errmsg());
		
killed:
	if ( !CloseHandle(pid) ) {
		ereport(LOG_FAILURE, "CloseHandle(pid) failed for %d", pid);
	}

	return;
}

/*
 * Takes a path and request structures, and a pid_t to store the child pid
 * in. Returns -1 on error, and a file descriptor on success.
 * 
 * Note that NPH is no longer taken care of here. Data from the CGI
 * program is still buffered in case SSL is active.
 */

int cgi_start_exec(pblock *pb, Session *sn, Request *rq, HANDLE *pid, char *path, SYS_FILE& sysFile)
{
  char *argv0 = 0;
  char *args = 0;
  int  n, pathLength;;
  struct stat *finfo;
  STARTUPINFO siStartInfo;
  PROCESS_INFORMATION piProcInfo;
  PCHAR ChildCmdLine;
  HANDLE hChildStdinRd=NULL,hChildStdinWr=NULL, hChildStdoutWr=NULL,
    hChildStdinWrDup=NULL, hChildStdoutRd=NULL;
  SECURITY_ATTRIBUTES saAttr;
  SYS_FILE writeFile;
  struct linger option;
  char **env = NULL;
  char *dir = NULL;
  char *nt_env = NULL;
  char *shellcgi_path = NULL;
  char *oldPath = NULL;
#ifdef MCC_ADMSERV
  /* MLM - adding a new pipe for username and password information */
  DWORD ppid = GetCurrentProcessId();
  SYS_FILE hChildPWPipeRd = NULL;
  SYS_FILE hChildPWPipeWr = NULL;
#endif /* MCC_ADMSERV */

  option.l_onoff = 1;
  option.l_linger = 120; 

  ZeroMemory(&piProcInfo, sizeof(PROCESS_INFORMATION));
  *pid = NULL;

  saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
  saAttr.bInheritHandle = TRUE;
  saAttr.lpSecurityDescriptor = NULL;
                               
  if (request_loadheaders(sn, rq) == REQ_ABORTED)
    return REQ_ABORTED;
  args = pblock_findval("query", rq->reqpb);

  finfo = request_stat_path(path, rq);
  if (!finfo) {
    pathLength = strlen(path);
    if (!strcmp( (path + pathLength - 4), ".exe")) {
      /* try out path without .exe extension */
      oldPath = strdup(path);
      path[pathLength-4] = '\0';
      finfo = request_stat_path(path, rq);
    }
    /* path is not found */
    if (!finfo) {
      if (oldPath) {
	log_error(LOG_WARN, "send-cgi", sn, rq, 
		  "cannot find CGI program %s/%s (%s)",
		  oldPath, path, rq->staterr);
	free(oldPath);
      }
      else log_error(LOG_WARN, "send-cgi", sn, rq, 
		     "cannot find CGI program %s (%s)",
		     path, rq->staterr);
      protocol_status(sn, rq, PROTOCOL_NOT_FOUND, NULL);
      return REQ_ABORTED;
    }
  } 
    
  if(!S_ISREG(finfo->st_mode)) {
    log_error(LOG_WARN, "send-cgi", sn, rq, 
	      "cannot execute %s (is a directory)",
	      path);
    protocol_status(sn, rq, PROTOCOL_NOT_FOUND, NULL);
    return REQ_ABORTED;
  }
	
  if((argv0 = strrchr(path, FILE_PATHSEP)) != NULL) {
    argv0++;
  } 
  else if ((argv0 = strrchr(path, '/')) != NULL) {
    argv0++;
  } 
  else {
    argv0 = path;
  }

/* MLM - I moved the env to after the pipe creation step, so we could use
 *       the pipe information for the admin server. */

  shellcgi_path = pblock_findval("shellcgi-path", rq->vars);
  ChildCmdLine = _create_argv(argv0, args, shellcgi_path);

  if (_cgi_is_dangerous(path, ChildCmdLine, pb)) {
    log_error(LOG_SECURITY, "send-cgi", sn, rq, 
	      "rejecting request due to dangerous characters (&,|,<,>) requested for .bat script");
    protocol_status(sn, rq, PROTOCOL_NOT_FOUND, NULL);
    goto REQ_ABORT;
  }

  saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
  saAttr.bInheritHandle = TRUE;
  saAttr.lpSecurityDescriptor = NULL;

  //Create a pipe for the child's STDOUT
  if (!CreatePipe(&hChildStdoutRd, &hChildStdoutWr, &saAttr, 0)) {
    log_error(LOG_CATASTROPHE, "send-cgi", sn, rq, 
	      "could not create STDOUT pipe (%s)", system_errmsg());
    goto REQ_ABORT;
  }
	
  //Create a pipe for the child's STDIN
  if (!CreatePipe(&hChildStdinRd, &hChildStdinWr, &saAttr, 0)) {
    log_error(LOG_CATASTROPHE, "send-cgi", sn, rq, 
	      "could not create STDIN pipe (%s)", system_errmsg());
    goto REQ_ABORT;
  }
  // Duplicate the write handle to the pipe so that it is not inherited

  if (! DuplicateHandle(GetCurrentProcess(), hChildStdinWr, GetCurrentProcess(),
			&hChildStdinWrDup, 0, FALSE, DUPLICATE_SAME_ACCESS)) {
    log_error(LOG_CATASTROPHE, "send-cgi", sn, rq, 
	      "could not duplicate hChildStdinWr (%s)", system_errmsg());
    goto REQ_ABORT;
  }
  if ( !CloseHandle(hChildStdinWr) ) {
    ereport(LOG_FAILURE, "CloseHandle(hChildStdinWr) failed for %d", hChildStdinWr);
  }
  hChildStdinWr = NULL;

#ifdef MCC_ADMSERV
  /* MLM - Create a pipe for admin server password */
  saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
  saAttr.bInheritHandle = TRUE;
  saAttr.lpSecurityDescriptor = NULL;

  {
    HANDLE prd, pwr;
    if (!CreatePipe(&prd, &pwr, &saAttr, 0)) {
      log_error(LOG_CATASTROPHE, "send-cgi", sn, rq, 
                "could not create password pipe (%s)", system_errmsg());
      goto REQ_ABORT;
    }
    hChildPWPipeRd = PR_ImportFile((int32)prd);
    hChildPWPipeWr = PR_ImportFile((int32)pwr);
  }
#endif

/* MLM - moving env stuff to after stdin and stdout creation so we can set
 *       admin handle in environment correctly */
  env = http_hdrs2env(rq->headers);
  if (initenv_vars)
    env = util_env_copy(initenv_vars, env);
#ifdef MCC_ADMSERV
  /* MLM - add new pipe for password */
  env = admin_cgivars(sn, rq, ppid, hChildPWPipeRd, env);
#endif /* MCC_ADMSERV */
  env = cgi_common_vars(sn, rq, env);
  env = cgi_specific_vars(sn, rq, args, env);
  env = cgi_ssi_vars(sn, rq, args, env);
  if (initenv)
    env = pblock_pb2env(initenv, env);

  /* env now contains the UNIX style environment variables. 
   * convert them to a format which we can pass to create process 
   */
  nt_env = _convert_unix_to_nt_env(env);
  util_env_free(env);
  /* MLM - env move end */

  ZeroMemory(&siStartInfo, sizeof(STARTUPINFO));
  siStartInfo.cb = sizeof(STARTUPINFO);
  siStartInfo.dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
  siStartInfo.hStdInput = hChildStdinRd;
  siStartInfo.hStdOutput = siStartInfo.hStdError = hChildStdoutWr;
  siStartInfo.wShowWindow = SW_HIDE;


  /* Temporary fix for CreateProcess to work under Win95 -XXX aruna.
   * CreateProcess returns a FILE_NOT_FOUND error under Win95 if
   * forward slashes are used in the pathname 
   */

  CONVERT_TO_NATIVE_FS(path);
  dir = _build_cwd(path);

  if (!(CreateProcess(path, 
		      ChildCmdLine, 
		      NULL, 		/* process security attr */
		      NULL, 		/* thread security attr */
		      TRUE,		/* inherit handles */
		      0,		/* create flags */
		      (LPVOID)nt_env,	/* environment */
		      dir, 		/* directory name */
		      &siStartInfo,	/* startup info */
		      &piProcInfo))) {/* process info */
    log_error(LOG_FAILURE, "send-cgi", sn, rq, 
	      "could not send new process (%s)", system_errmsg());
    goto REQ_ABORT;
  }
  
  if ( !CloseHandle(hChildStdinRd) ) {
    ereport(LOG_FAILURE, "CloseHandle(hChildStdinRd) failed for %d", hChildStdinRd);
  }
  hChildStdinRd = NULL;
  if ( !CloseHandle(hChildStdoutWr) ) {
    ereport(LOG_FAILURE, "CloseHandle(hChildStdoutWr) failed for %d", 
	    hChildStdoutWr);
  }
  hChildStdoutWr = NULL;
  
  *pid = piProcInfo.hProcess;

  FREE(ChildCmdLine);	ChildCmdLine = NULL;
  if (!CloseHandle(piProcInfo.hThread) ) {
    ereport(LOG_FAILURE, "CloseHandle(piProcInfo.hThread) failed for %d",
	    piProcInfo.hThread);
  }
  piProcInfo.hThread = NULL;
  sysFile = PR_ImportFile((int32)hChildStdoutRd);
  writeFile = PR_ImportFile((int32)hChildStdinWrDup);

  n = cgi_send_content(sn, rq, writeFile);

  /* cgi_send_content() gives its best effort to read all trailing data;
   * so remove content-length so that the error_report code won't try to
   * read the data too
   */
  pblock_remove("content-length", rq->headers);

  if ( PR_Close(writeFile) == PR_FAILURE ) {
    ereport(LOG_FAILURE, "PR_Close(writeFile) failed for %d",
	    hChildStdinWrDup);
  } 
  hChildStdinWrDup = NULL;
  
  writeFile = SYS_ERROR_FD;

  if(n == REQ_ABORTED) {
    goto REQ_ABORT;
  }

#ifdef MCC_ADMSERV
  /* MLM - send auth data to CGI */
  n = adm_cgi_send_password(sn, rq, hChildPWPipeWr);
  if( PR_Close(hChildPWPipeRd) == PR_FAILURE)  {
    ereport(LOG_FAILURE, "PR_Close(hChildPWPipeRd) failed for %d",
	    hChildPWPipeRd);
  }
  hChildPWPipeRd = NULL;
  if( PR_Close(hChildPWPipeWr) == PR_FAILURE)  {
    ereport(LOG_FAILURE, "PR_Close(hChildPWPipeWr) failed for %d",
	    hChildPWPipeWr);
  }
  hChildPWPipeWr = NULL;
  if(n == REQ_ABORTED) {
    goto REQ_ABORT;
  }
#endif

  if (dir) 
    FREE(dir);
  if (nt_env) 
    FREE(nt_env);

  return REQ_PROCEED;

REQ_ABORT:
  if (hChildStdinRd) 
    if ( !CloseHandle(hChildStdinRd) ) {
      ereport(LOG_FAILURE, "CloseHandle(hChildStdinRd) failed for %d", 
	      hChildStdinRd);
    }
  if (hChildStdinWr) 
    if ( !CloseHandle(hChildStdinWr) ) {
      ereport(LOG_FAILURE, "CloseHandle(hChildStdinWr) failed for %d", 
	      hChildStdinWr);
    }
  if (hChildStdoutRd) 
    if ( !CloseHandle(hChildStdoutRd) ) {
      ereport(LOG_FAILURE, "CloseHandle(hChildStdoutRd) failed for %d", 
	      hChildStdoutRd);
    }
  if (hChildStdoutWr) 
    if ( !CloseHandle(hChildStdoutWr) ) {
      ereport(LOG_FAILURE, "CloseHandle(hChildStdoutWr) failed for %d", 
	      hChildStdoutWr);
    }
  if (hChildStdinWrDup) 
    if ( !CloseHandle(hChildStdinWrDup) ) {
      ereport(LOG_FAILURE, "CloseHandle(hChildStdinWrDup) failed for %d",
	      hChildStdinWrDup);
    } 
#ifdef MCC_ADMSERV
  /* MLM - clean up password pipe */
  if(hChildPWPipeRd)  {
    if ( PR_Close(hChildPWPipeRd) == PR_FAILURE) {
      ereport(LOG_FAILURE,"PR_Close(hChildPWPipeRd) failed for %d",
	      hChildPWPipeRd);
    }
  }
  if(hChildPWPipeWr)  {
    if ( PR_Close(hChildPWPipeWr) == PR_FAILURE) {
      ereport(LOG_FAILURE,"PR_Close(hChildPWPipeWr) failed for %d",
	      hChildPWPipeWr);
    }
  }
#endif
  if (ChildCmdLine) 
    FREE (ChildCmdLine);
  if (dir) 
    FREE(dir);
  if (nt_env)
    FREE(nt_env);
  if(piProcInfo.hProcess)	{
    wait_for_child_death(piProcInfo.hProcess, TRUE, TRUE);
    *pid = NULL;
  }
  if(piProcInfo.hThread)
    if ( !CloseHandle(piProcInfo.hThread) ) {
      ereport(LOG_FAILURE, "CloseHandle(piProcInfo.hThread) failed for %d",
	      piProcInfo.hThread);
    }
  return REQ_ABORTED;
}


/* --------------------------- cgi_parse_output --------------------------- */


/*
 * Returns -1 on error, 0 for proceed normally, and 1 for dump output and 
 * restart request.
 */

int cgi_parse_output(filebuf_t *buf, Session *sn, Request *rq)
{
    char *s, *l;
    if(cgi_scan_headers(sn, rq, buf) == REQ_ABORTED)
        return -1;
    l = pblock_findval("location", rq->srvhdrs);

    if((s = pblock_findval("status", rq->srvhdrs))) {
        if((strlen(s) < 3) || (!isdigit(s[0]) || (!isdigit(s[1])) || 
                                                           (!isdigit(s[2]))))
        {
            log_error(LOG_WARN, "send-cgi", sn, rq, 
                      "ignoring CGI status line %s (no 3-digit code)",s);
            s = NULL;
        }
        else pblock_nvinsert("status", s, rq->srvhdrs);
    }
    if(l && (!util_is_url(l)))
        return 1;

    if(!s) {
		if (l)
			pblock_nvinsert("url", l, rq->vars); 
        protocol_status(sn, rq, (l ? PROTOCOL_REDIRECT : PROTOCOL_OK), NULL);
	}

    return 0;
}


/* ------------------------------- cgi_send ------------------------------- */


/*
 * Execute a CGI script and return its output to the client. A Service 
 * class function.
 */
int cgi_send(pblock *pb, Session *sn, Request *rq)
{
    char *t, *path = pblock_findval("path", rq->vars), num[16];
    SYS_NETFD tfd;
	HANDLE pid = NULL;
	filebuf_t *buf;
	SYS_FILE cfd;
    pb_param *pp;
    int restart, ns;
	int linger = 1;

    if((pp = pblock_remove("content-type", rq->srvhdrs)))
        param_free(pp);

    // Need to pass a reference to cfd, not address of cfd
    // if( cgi_start_exec(pb, sn, rq, &pid, path, &cfd) == REQ_ABORTED) {
    if( cgi_start_exec(pb, sn, rq, &pid, path, cfd) == REQ_ABORTED) {
		ereport(LOG_FAILURE,"cgi_send:cgi_start_exec %s failed", path);
		if (pid) 
			wait_for_child_death(pid, TRUE, TRUE);
        return REQ_ABORTED;
	}

	cgiwatch_add_client(sn, rq, pid);

	buf = pipebuf_open(cfd, PIPE_BUFFERSIZE, 0);
	if (!buf) {
		log_error(LOG_CATASTROPHE, "send-cgi", sn, rq, 
                          "could not open pipe (%s)", system_errmsg());
		PR_Close(cfd);
		cgiwatch_remove_client(sn, rq, pid);
		wait_for_child_death(pid, TRUE, TRUE);
		return REQ_ABORTED;
	}

    t = strrchr(path, FILE_PATHSEP);
    /* XXXrobm FILE_PATHSEP is a forward slash. But at this point, the path 
       may have backslashes */
    if(!t)
        t = strrchr(path, '\\');
    if(t && (!strncmp(t+1, "nph-", 4))) {
        tfd = sn->csd;
        restart = 0;
    }
    else {
        switch(cgi_parse_output(buf, sn, rq)) {
          case -1:
			pipebuf_close(buf);
			cgiwatch_remove_client(sn, rq, pid);
			wait_for_child_death(pid, TRUE, TRUE);
            return REQ_ABORTED;
          case 0:
            if(protocol_start_response(sn, rq) == REQ_NOACTION)
                tfd = SYS_NET_ERRORFD;
            else
                tfd = sn->csd;
            restart = 0;
            break;
          case 1:
            tfd = SYS_NET_ERRORFD;
            restart = 1;
        }
    }
    if( (ns = pipebuf_buf2sd(buf, tfd, -1)) == IO_ERROR) {
		BOOLEAN WinsockError = FALSE;
		int Error = GetLastError();

		if (Error == 0) {
			Error = WSAGetLastError();
			if (Error) {
				WinsockError = TRUE;
			}
		}
		switch (Error) {
			case ERROR_BROKEN_PIPE:
			case ERROR_BAD_PIPE:
			case ERROR_PIPE_BUSY:
			case ERROR_PIPE_LISTENING:
			case ERROR_PIPE_NOT_CONNECTED:
			case ERROR_NO_DATA:
				break;
			default:
	            log_error(LOG_FAILURE, "send-cgi", sn, rq, 
	                      "error occurred while sending script output (%s)",
	                      buf->errmsg);
				pipebuf_close(buf);
				cgiwatch_remove_client(sn, rq, pid);
				wait_for_child_death(pid, TRUE, TRUE);
				if (WinsockError) {
					WSASetLastError(Error);
				}
	            return REQ_ABORTED;
        }
    }
    else {
        util_itoa(ns, num);
        pblock_nvinsert("content-length", num, rq->srvhdrs);
    }
	pipebuf_close(buf);
	cgiwatch_remove_client(sn, rq, pid);
	wait_for_child_death(pid, TRUE, FALSE);
	
    if(restart) {
        pb_param *l = pblock_remove("location", rq->srvhdrs);
        request_restart_internal(l->value, rq);
        param_free(l);
        return request_handle_processed(sn, rq);
    }
    return REQ_PROCEED;
}

/* ------------------------------- shellcgi_send ------------------------------- */

/*
 * Execute a Shell CGI script and return its output to the client. A Service 
 * class function.
 */
int shellcgi_send(pblock *pb, Session *sn, Request *rq)
{
    pb_param *ppath = pblock_find("path", rq->vars);
    pb_param *pargs = pblock_find("query", rq->reqpb);

    char *path = ppath->value;
    char *args = NULL;

    char *shell_exe_path = _find_exe_path(path);
    struct stat fi;

    if (shell_exe_path) {
        /* Check to see if the associated file actually exists
         * This check may have been covered by the FindExecutable call */

        if(system_stat(shell_exe_path, &fi) == -1) {
            protocol_status(sn, rq, PROTOCOL_NOT_FOUND, NULL);
            log_error(LOG_FAILURE, "shellcgi-send", sn, rq, 
                      "can't find %s for execution", shell_exe_path);
            return REQ_ABORTED;
        }

        if(pargs)
            args = pargs->value;
                    
        /* We are going to override the old path with the path for the
         * executable.  But, in order to avoid us changing the query variable
         * and in order to get the program arguments right, we'll need to
         * know what the shellcgi-path originally was.  Save it in 
         * shellcgi-path.
         */
        pblock_nvinsert("shellcgi-path", path, rq->vars);
        FREE(ppath->value);
        ppath->value = shell_exe_path;

        return(cgi_send(pb, sn, rq));
    } else {
 
        protocol_status(sn, rq, PROTOCOL_NOT_FOUND, NULL);
        log_error(LOG_FAILURE, "shellcgi-send", sn, rq, 
                  "can't find file association of %s for execution", path);
        return REQ_ABORTED;
    }
}

/* ------------------------------ cgi_query ------------------------------- */


int cgi_query(pblock *pb, Session *sn, Request *rq)
{
    pb_param *rqpath = pblock_find("path", rq->vars);
    char *path = pblock_findval("path", pb);

    FREE(rqpath->value);
    rqpath->value = STRDUP(path);

    if(!(pblock_findval("path-info", rq->vars)))
        pblock_nvinsert("path-info", pblock_findval("uri", rq->reqpb), 
                        rq->vars);

    return cgi_send(NULL, sn, rq);
}

void cgi_terminate(void *unused)
{
    if(initenv) {
        pblock_free(initenv);
        initenv = NULL;
    }

    if (initenv_vars) {
        util_env_free(initenv_vars);
        initenv_vars = NULL;
    }
}

/* cgi_init_generic()
 * Copies all environment variables known to the server to the 
 * initenv.  Why do this?  The reason is because the server used to
 * pass the environment variables by allowing the CGI program to
 * inherit the server's environment.  This has been changed, but
 * we don't want to break any user which may be using that environment,
 * so we still manually copy all server variables into the CGI 
 * variables.
 */
int cgi_init_generic()
{
    char **env;
    int env_count = 0;
    int index, pos;
    extern char **_environ;

    /* Count the number of env vars */

    if (_environ)
        for (env = _environ, env_count = 0; *env; env++, env_count++);

    if (env_count == 0)
        return 0;

    initenv_vars = util_env_create(NULL, env_count+1, &pos);

    if (_environ)
        for (index = 0; index < env_count; index++) 
            initenv_vars[index] = STRDUP(_environ[index]);
    initenv_vars[index] = NULL;

    return 0;
}

int cgi_init(pblock *pb, Session *sn, Request *rq)
{
    /* 
     * For backwards compatibility we provide the load-all option
     * for CGI on NT.  The NT server used to allow all env vars to
     * be inherited by the CGI program, now it does not.  If the user
     * specifies "load-all" on the cgi-init line, it will emulate
     * the old style and pass all server vars to the CGI program.
     * This is NT specific.
     */
    char *old_init = pblock_findval("load-server-vars", pb);

    if(!initenv)
        initenv = pblock_create(4);
    pblock_copy(pb, initenv);
    param_free(pblock_remove("fn", initenv));

    if (old_init) {
        pblock_remove("load-server-vars", initenv);
        (void)cgi_init_generic();
    }

#if 0
    magnus_atrestart(cgi_terminate, NULL);
#endif
    return TRUE;
}

