/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * ntlogbuf.c: General functions to handle buffered log writes. Requires a 
 * threaded server.
 * 
 * Mike Belshe
 */



#include "netsite.h"
#include "safs/ntlogbuf.h"
#include "base/ereport.h"
#include "base/systhr.h"
#include "frame/conf_api.h"
#include "nspr.h"
/* AIX/WINNT can't live with private/primpl.h, but others can't live without it */
/* Now it is GONE from nspr 19971028 release
#if (!defined(AIX) && !defined(XP_WIN_32))
#include "private/primpl.h"
#endif
*/


/* --- Globals --------------------------------------------------- */
int logbuf_tls = -1;
CRITICAL logbuf_list_lock = NULL;
log_file_buf_t *logbuf_list = NULL;

PRIntervalTime logFlushInterval = 0;
static PRThread	*_logbufthread;
#define GetConfigParm(name, default) \
    (conf_findGlobal(name)?atoi(conf_findGlobal(name)):default)
void _logbuf_flush_disk(LOGBUF_T *fb);

/* --- logbuf_init ----------------------------------------------- */
int
logbuf_init(pblock *pb)
{
    /* There is a race here on startup; but since we go through the
     * init functions in serial, this will work ok. 
     */
	if (logbuf_tls == -1) {
		logbuf_tls = systhread_newkey();
	    logbuf_list_lock = crit_init();
    }

	return 0;
}

static void
logbufthread_main(void *unused)
{
    PRLock *l;
    PRCondVar *cv;
    log_file_buf_t *ptr;
    log_thread_buf_t *tls_buf;

    if (logbuf_tls == -1)
	return;

    l = PR_NewLock();
    if (!l)
        return;
    cv = PR_NewCondVar(l);
    if (!l)
        return;

    PR_Lock(l);
    while(1) {
        PR_WaitCondVar(cv, logFlushInterval);

	crit_enter(logbuf_list_lock);
	ptr = logbuf_list;
	while(ptr) {
		/* flush all buffers */
		for(tls_buf = ptr->list; tls_buf; tls_buf = tls_buf->next) 
			logbuf_flush(tls_buf);

		_logbuf_flush_disk(ptr);	

		ptr = ptr->next;
	}
	crit_exit(logbuf_list_lock);
    }

    /* Not reached */
    PR_Unlock(l);
}

PR_IMPLEMENT(PRBool) logbufthread_init()
{
    /* Configure the log flush timeout */
    logFlushInterval = 
        GetConfigParm("LogFlushInterval", logFlushInterval);
    if (logFlushInterval == 0)
        logFlushInterval = DEFAULT_FLUSH_TIMEOUT;
    logFlushInterval = PR_SecondsToInterval(logFlushInterval);

/* PR_CreateThreadGCAble is gone for NSPR 21 (19971028)
#ifdef SOLARIS
    _logbufthread = PR_CreateThreadGCAble(PR_SYSTEM_THREAD,
#else
    _logbufthread = PR_CreateThread(PR_SYSTEM_THREAD,
#endif
                                      logbufthread_main,
                                      NULL,
                                      PR_PRIORITY_NORMAL,
                                      PR_GLOBAL_THREAD,
                                      PR_UNJOINABLE_THREAD,
                                      0);
*/
    _logbufthread = PR_CreateThread(PR_SYSTEM_THREAD,
				    logbufthread_main,
				    NULL,
				    PR_PRIORITY_NORMAL,
				    PR_GLOBAL_THREAD,
				    PR_UNJOINABLE_THREAD,
				    0);
    if (!_logbufthread)
        return PR_FALSE;

    return PR_TRUE;
}

/* --- logbuf_new ----------------------------------------------- */
LOGBUF_T *
logbuf_new(pblock *pb, SYS_FILE fd)
{
	LOGBUF_T *newentry;
	int index;
	int num_buffers;
	char *str_num_buffers;
	int buffer_size;
	char *str_buffer_size;
	int thread_buffer_size;
	char *str_thread_buffer_size;

	if ( (str_num_buffers = pblock_findval("num-buffers", pb)) != NULL) {
		num_buffers = atoi(str_num_buffers);
	}
	else
		num_buffers = DEFAULT_NUM_BUFFERS;

	if ( (str_buffer_size = pblock_findval("buffer-size", pb)) != NULL) {
		buffer_size = atoi(str_buffer_size);
	}
	else
		buffer_size = DEFAULT_LARGE_BUFFER_SIZE;
	
	if ( (str_thread_buffer_size = pblock_findval("thread-buffer-size", pb)) != NULL) {
		thread_buffer_size = atoi(str_thread_buffer_size);
	}
	else
		thread_buffer_size = DEFAULT_THREAD_BUFFER_SIZE;
	

	if ( (newentry = (LOGBUF_T*)PERM_MALLOC(sizeof(log_file_buf_t))) == NULL){
		ereport(LOG_FAILURE, "log-buf-create: out of memory");
		return NULL;
	}

	newentry->num_buffers = num_buffers;
	if ( (newentry->buf = (log_thread_buf_t **)PERM_MALLOC(sizeof(log_thread_buf_t *) * (newentry->num_buffers))) == NULL) {
		ereport(LOG_FAILURE, "log-buf-create: out of memory");
		return NULL;
	}

	newentry->tls_key = systhread_newkey();


	for (index=0; index < newentry->num_buffers; index++) {
		newentry->buf[index] = _log_thread_buf_create(NULL, buffer_size);
	}

	newentry->current_buffer = 0;
	newentry->fd = fd;
	newentry->thread_buf_size = thread_buffer_size;
	newentry->list = NULL;
	
	newentry->crit = crit_init();
	newentry->cv_crit = crit_init();
	newentry->cv = condvar_init(newentry->cv_crit);

	crit_enter(logbuf_list_lock);
	newentry->next = logbuf_list;
	logbuf_list = newentry;
	crit_exit(logbuf_list_lock);

	return newentry;
}

/* --- logbuf_destroy ------------------------------------------- */
void
logbuf_destroy(LOGBUF_T *buf)
{
	int index;
	log_file_buf_t *search;

	for (index= 0; index<buf->num_buffers; index++)
		_log_thread_buf_destroy(buf->buf[index]);

	crit_enter(logbuf_list_lock);

	if (logbuf_list == buf) {
		logbuf_list = buf->next;
	} else {
		for (search = logbuf_list; search && search->next != buf; 
			search = search->next);

		if (search->next == buf) 
			search->next = buf->next;
		else {
			ereport(LOG_WARN, "logbuf_destroy: log buffer list is corrupt");
		}
	}
	crit_exit(logbuf_list_lock);

	PERM_FREE(buf);

	return;
}

/* --- _log_thread_buf_create ------------------------------------ */
log_thread_buf_t *
_log_thread_buf_create(LOGBUF_T *buf, int size)
{
	log_thread_buf_t *newentry;

	if ( (newentry = (log_thread_buf_t *)PERM_MALLOC(sizeof(log_thread_buf_t))) == NULL) {
		ereport(LOG_FAILURE, "log-thread-buf-create: out of memory");
		return NULL;
	}

	if ( (newentry->data = (char *)PERM_MALLOC(sizeof(char) * size)) == NULL) {
		ereport(LOG_FAILURE, "log-thread-buf-create: out of memory");
		return NULL;
	}

	newentry->current_size = 0;
	newentry->max_size = size;
	newentry->tls_next = NULL;
	newentry->log_file_buf = buf;
	newentry->state = LB_STATE_IDLE;
	newentry->next = NULL;

	/* Add to global linked list of outstanding buffers for this log */
	if (buf) {
		crit_enter(buf->crit);
		newentry->next = buf->list;
		buf->list = newentry;
		crit_exit(buf->crit);
	}
	
	return newentry;
}

/* --- _log_thread_buf_destroy ----------------------------------- */
void
_log_thread_buf_destroy(log_thread_buf_t *buf)
{
	FREE(buf->data);
        buf->data = 0;
	FREE(buf);

	return;
}

/* --- _log_thread_link ------------------------------------------ */
/* Adds a buffer to this thread's TLS linked list for cleanup when
 * the thread goes away 
 */
void
_log_thread_link(log_thread_buf_t *buf)
{
	log_thread_buf_t *head;

	head = (log_thread_buf_t *)systhread_getdata(logbuf_tls);
	
	buf->tls_next = head;

	systhread_setdata(logbuf_tls, (void *)buf);

	return;
}


/* --- logbuf_write --------------------------------------------- */
int
logbuf_write(LOGBUF_T *buf, SYS_FILE unused, char *data, int len)
{
	log_thread_buf_t *tls_buf;

	/* Get the pointer to the TLS buffer */
	if ( (tls_buf = (log_thread_buf_t *)systhread_getdata(buf->tls_key)) == NULL) {
		/* This thread has not yet logged */
		if ( (tls_buf = _log_thread_buf_create(buf, buf->thread_buf_size)) == NULL){
			return system_fwrite_atomic(buf->fd, data, len);
		}
		systhread_setdata(buf->tls_key, tls_buf);
		_log_thread_link(tls_buf);
	}

	/* Check to see if we need to flush first */
	if ( len + tls_buf->current_size >= tls_buf->max_size) {
		logbuf_flush(tls_buf);
		tls_buf->current_size = 0;
	}

	crit_enter(buf->crit);

	memcpy(&(tls_buf->data[tls_buf->current_size]), data, len);
	tls_buf->current_size += len;

	crit_exit(buf->crit);

	return 0;
}

/* flush a buffer all the way to disk */
void
_logbuf_flush_disk(LOGBUF_T *fb)
{
	int flush_buffer;

	crit_enter(fb->crit);

	if (fb->buf[fb->current_buffer]->current_size == 0) {
		crit_exit(fb->crit);
		return;
	}

	fb->buf[fb->current_buffer]->state = LB_STATE_WRITING;

	flush_buffer = fb->current_buffer;
	fb->current_buffer++;
	if (fb->current_buffer >= fb->num_buffers)
		fb->current_buffer = 0;
	crit_exit(fb->crit);

	system_fwrite_atomic(fb->fd, fb->buf[flush_buffer]->data, 
		fb->buf[flush_buffer]->current_size);

	fb->buf[flush_buffer]->current_size = 0;

	crit_enter(fb->cv_crit);
	fb->buf[flush_buffer]->state = LB_STATE_IDLE;
	condvar_notify(fb->cv);
	crit_exit(fb->cv_crit);

	return;
}

/* --- logbuf_flush --------------------------------------------- */
void
logbuf_flush(log_thread_buf_t *buf)
{
	LOGBUF_T *fb = buf->log_file_buf;

	if (buf->current_size == 0)	/* nothing to flush! */
		return;

	crit_enter(fb->crit);

	while (fb->buf[fb->current_buffer]->state == LB_STATE_WRITING) {
		/* oh no!  all buffers are writing! */

		/*ereport(LOG_WARN, "logbuf_flush: buffers are filled to capacity and blocking threads; may want to increase buffer size");*/

		/* Note we still hold fb->crit and we are grabbing another lock.
		 * This is ok; both buffers are full and nobody can write anyway
		 */

		crit_enter(fb->cv_crit);
		/* The first state check was a quickie to avoid grabbing a lock;
		 * do the real check now that we have cv_crit 
		 */
		if ( fb->buf[fb->current_buffer]->state == LB_STATE_WRITING) 
			condvar_wait(fb->cv);

		crit_exit(fb->cv_crit);
	}

	/* Check for overflow */
	if ( fb->buf[fb->current_buffer]->current_size + buf->current_size >=
		fb->buf[fb->current_buffer]->max_size) {

		/* need to flush to disk */
		_logbuf_flush_disk(fb);

		/* now that we did the work of flushing; recursively see if we
		 * can't write our own buffer now.  No starvation possible because
		 * we hold the lock.
		 */
		logbuf_flush(buf);
		crit_exit(fb->crit);
		return;
	}
	
	memcpy(&(fb->buf[fb->current_buffer]->data[fb->buf[fb->current_buffer]->current_size]),
		buf->data, buf->current_size);
	fb->buf[fb->current_buffer]->current_size += buf->current_size;
	buf->current_size = 0;

	crit_exit(fb->crit);

	return ;
}

/* --- logbuf_flush_all ----------------------------------------- */
void
logbuf_flush_all()
{
	log_thread_buf_t *head;

	head = (log_thread_buf_t *)systhread_getdata(logbuf_tls);

	while(head) {
		logbuf_flush(head);						/* flush to secondary buffer */
		_logbuf_flush_disk(head->log_file_buf);	/* flush secondary to disk */

		head = head->tls_next;
	}

	return;
}

/* --- logbuf_cleanup_thread ------------------------------------ */
void
logbuf_cleanup_thread()
{

    /* XXXMB - WARNING WARNING WARNING- this routine does not work; 
     * it fails to remove the buffers from the log's lists; later calling
     * something like logbuf_cleanup_all() will fail!
     */
    log_thread_buf_t *ptr, *dead;

    ptr = (log_thread_buf_t *)systhread_getdata(logbuf_tls);
    systhread_setdata(logbuf_tls, NULL);

    while(ptr) {
        logbuf_flush(ptr);
        dead = ptr;
        ptr = dead->tls_next;
        _log_thread_buf_destroy(dead);
    }
}

/* --- logbuf_cleanup_all --------------------------------------- */
void
logbuf_cleanup_all()
{
	log_file_buf_t *ptr, *tmp;

    if (logbuf_tls == -1)
		return;

	crit_enter(logbuf_list_lock);
	ptr = logbuf_list;
	while(ptr) {
		tmp = ptr;
		ptr = ptr->next;
		logbuf_close(tmp);
	}
	crit_exit(logbuf_list_lock);
}

/* --- logbuf_close --------------------------------------------- */
void
logbuf_close(LOGBUF_T *buf)
{
	log_thread_buf_t *ptr;

	/* All threads must have stopped writing to the log at this point;
	 * otherwise armageddon will occur....
	 */

	/* flush all buffers */
	for(ptr = buf->list; ptr; ptr = ptr->next) 
		logbuf_flush(ptr);

	crit_enter(buf->crit);

	buf->buf[buf->current_buffer]->state = LB_STATE_WRITING;

	crit_exit(buf->crit);

	system_fwrite_atomic(buf->fd, buf->buf[buf->current_buffer]->data, 
		buf->buf[buf->current_buffer]->current_size);

	buf->buf[buf->current_buffer]->state = LB_STATE_IDLE;

	logbuf_destroy(buf);
}

/* --- logbuf_isempty ------------------------------------------- */
int
logbuf_isempty(LOGBUF_T *buf)
{
	log_thread_buf_t *tls_buf;

	tls_buf = (log_thread_buf_t *)systhread_getdata(buf->tls_key);

	return tls_buf->current_size == 0;
}

/* --- logbuf_allempty ------------------------------------------ */
int
logbuf_allempty()
{
	log_thread_buf_t *head;
	int isempty = 1;

    if (logbuf_tls == -1)
		return 1;

	head = (log_thread_buf_t *)systhread_getdata(logbuf_tls);

	while(head) {
		isempty &= (head->current_size == 0);
		head = head->tls_next;
	}

	return isempty;
}
