/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * ntrans.c: Functions to be used with the NameTrans directive.
 * 
 * Rob McCool
 */


#include "netsite.h"
#include "safs/ntrans.h"
#include "base/session.h"
#include "frame/req.h"
#include "frame/log.h"            /* log_error */
#include "base/util.h"            /* util_sprintf */

#include "base/shexp.h"
#include "frame/http.h"
#include "frame/conf.h"

#include "base/language.h"
#include "i18n.h"				  /* GetFileForLanguage */

#include <sys/stat.h>

#ifdef XP_UNIX
#include <pwd.h>

#define PATH_SIZE 128


/* -------------------------- ntrans_uhome_init --------------------------- */


#include "base/crit.h"

static pblock *pwpb = NULL;
static CRITICAL pwcrit = NULL;

extern "C" void ntrans_init_crits(void)
{
    pwcrit = crit_init();
}

int _passwd_scan(char *fn, char *user, char *l)
{
    SYS_FILE fd;
    filebuf_t *fb;
    char *t, *u, *dir, *uid;
    int eof, ln;

    if((fd = system_fopenRO(fn)) == SYS_ERROR_FD) {
        strcpy(l, system_errmsg());
        return -1;
    }
    if(!(fb = filebuf_open(fd, FILE_BUFFERSIZE))) {
        strcpy(l, system_errmsg());
        return -1;
    }
    for(eof = 0, ln = 0; (!eof); ) {
        if( (eof = util_getline(fb, ln, 256, l)) == -1)
            return -1;

        if(!(t = strchr(l, ':')))
            continue;
        *t++ = '\0';
        if(user && strcmp(user, l))
            continue;
        if(!(t = strchr(t, ':')))
            continue;
        ++t;
        if(!(u = strchr(t, ':')))
            continue;
        *u = '\0';
        uid = t;
        if(!(t = strchr(u+1, ':')))
            continue;
        if(!(t = strchr(t+1, ':')))
            continue;
        dir = ++t;
        if(!(u = strchr(t, ':')))
            continue;
        *u = '\0';
        if(user) {
            filebuf_close(fb);
            util_sprintf(l, "%s:%s", uid, dir); /* overlap should be safe */
            return 1;
        }
        else {
            char l2[256];
            util_sprintf(l2, "%s:%s", uid, dir);
            pblock_nvinsert(l, l2, pwpb);
        }
    }
    filebuf_close(fb);

    return 0;
}


void ntrans_uhome_terminate(void *data)
{
    pblock_free(pwpb);
    pwpb = NULL;
}


int ntrans_uhome_init(pblock *pb, Session *sn, Request *rq)
{
    char *fn = pblock_findval("pwfile", pb), l[256];

    if(!pwpb) {
        pwpb = pblock_create(27);
#if 0
        magnus_atrestart(ntrans_uhome_terminate, NULL);
#endif
    }
    if(fn) {
        if(_passwd_scan(fn, NULL, l)) {
            pblock_nvinsert("errors", l, pb);
            return REQ_ABORTED;
        }
        else
            return REQ_PROCEED;
    }
    else {
        struct passwd *pw;

        setpwent();
        while( (pw = getpwent()) ) {
            util_sprintf(l, "%d:%s", pw->pw_uid, pw->pw_dir);
            pblock_nvinsert(pw->pw_name, l, pwpb);
        }
        endpwent();
    }
    return REQ_PROCEED;
}


char *_name2dir(char *name, char *fn, char *uid)
{
    struct passwd *pw, pw_ret;
    char *t, l[256];
    int ret;
    char *homedir;

    if(pwpb) {
        if(!(t = pblock_findval(name, pwpb)))
            return NULL;
        strcpy(l, t);
        t = strchr(l, ':');
        *t++ = '\0';
        strcpy(uid, l);
        return STRDUP(t);
    }
    else if(fn) {
        crit_enter(pwcrit);
        ret = _passwd_scan(fn, name, l);
        crit_exit(pwcrit);
        switch(ret) {
        case -1:
            log_error(LOG_WARN, "unix-home", NULL, NULL, 
                      "can't open password file %s (%s)", fn, l);
        default: /* 0: no match */
            return NULL;
        case 1:
            t = strchr(l, ':');
            *t++ = '\0';
            strcpy(uid, l);
            return STRDUP(t);
        }
    }
    else {
	char pwline[DEF_PWBUF * 2];
        pw = util_getpwnam(name, &pw_ret, pwline, DEF_PWBUF * 2);
	if (!pw)
	    return NULL;
        util_itoa(pw->pw_uid, uid);
        return STRDUP(pw->pw_dir);
    }
}


/* --------------------------- ntrans_unix_home --------------------------- */


int ntrans_unix_home(pblock *pb, Session *sn, Request *rq)
{
    char *from = pblock_findval("from", pb);
    char *name = pblock_findval("name", pb);
    char *new_path = pblock_findval("subdir", pb);
    char *pwfile = pblock_findval("pwfile", pb);
    pb_param *path = pblock_find("ppath", rq->vars);
    char *t, *u, *dir, c;
    char uid[16];
    int fl;

    rq->directive_is_cacheable = 1;

    if((!from) || (!new_path)) {
        log_error(LOG_MISCONFIG, "unix-home", sn, rq,
                  "missing parameter (need from and subdir)");
        return REQ_ABORTED;
    }

    if(!path)
        return REQ_NOACTION;

    fl = strlen(from);
    if(strncmp(path->value, from, fl))
        return REQ_NOACTION;

    t = &path->value[fl];
    u = t;
    while(*u && (*u != '/'))
        ++u;

    c = *u;
    *u = '\0';
    if((!*t) || (!(dir = _name2dir(t, pwfile, uid)))) {
        log_error(LOG_WARN, "unix-home", sn, rq,
                  "could not find home directory for user %s", t);
        protocol_status(sn, rq, PROTOCOL_NOT_FOUND, NULL);
        return REQ_ABORTED;
    }
    pblock_nvinsert("user", t, rq->vars);
    pblock_nvinsert("uid", uid, rq->vars);
    *u = c;

    t = (char *) MALLOC(strlen(dir) + strlen(new_path) + strlen(u) + 2);

    util_sprintf(t, "%s/%s", dir, new_path);
    pblock_nvinsert("ntrans-base", t, rq->vars);
    strcat(t, u);

    FREE(dir);
    FREE(path->value);
    path->value = t;

    if(name)
        pblock_nvinsert("name", name, rq->vars);

    return REQ_PROCEED;
}

#endif /* XP_UNIX */

/* ---------------------------- ntrans_pfx2dir ---------------------------- */


int ntrans_pfx2dir(pblock *pb, Session *sn, Request *rq)
{
    register int x, y;
    char *from = pblock_findval("from", pb);
    char *dir = pblock_findval("dir", pb);
    pb_param *pp = pblock_find("ppath", rq->vars);
    char *t, *u;
    int fl;

    rq->directive_is_cacheable = 1;

    if((!from) || (!dir)) {
        log_error(LOG_MISCONFIG, "pfx2dir", sn, rq, 
                  "missing parameter (need from and dir)");
        return REQ_ABORTED;
    }

    if(!pp)
        return REQ_NOACTION;

    fl = strlen(from);
    u = pp->value;

    /* The second or clause is to prevent partials from shadowing.
       That is, if you have a cgi-bin directory and cgi-bin-sucks.html, 
       cgi-bin-sucks.html is not hidden by the mapping. Side effect is that 
       trailing slashes on from or dir causes problems.
    */
#ifdef XP_WIN32
    if((strncasecmp(from, u, fl)) || (u[fl] && (u[fl] != '/')))
#else
    if((strncmp(from, u, fl)) || (u[fl] && (u[fl] != '/')))
#endif
        return REQ_NOACTION;

    t = (char *) MALLOC(strlen(u) - fl + strlen(dir) + 1);

    for(x = 0; (t[x] = dir[x]); x++);
    for(y = fl; (t[x] = u[y]); x++, y++);

    FREE(u);

    /* translate name if accept-language info is there */
    if (conf_getglobals()->accept_language) {
        char *langpath;
        if (langpath = lang_acceptlang_file(t, rq->headers)) {
            FREE(t);
            t = langpath;
        }
    }

    pp->value = t;

    pblock_nvinsert("ntrans-base", dir, rq->vars);
    if( (t = pblock_findval("name", pb)) )
        pblock_nvinsert("name", t, rq->vars);

    return REQ_PROCEED;
}


/* ---------------------------- ntrans_docroot ---------------------------- */


int ntrans_docroot(pblock *pb, Session *sn, Request *rq)
{
    char *dr = pblock_findval("root", pb);
    char *address = pblock_findval("address", pb);
    char *port    = pblock_findval("port", pb);
    pb_param *path = pblock_find("ppath", rq->vars);
    char *t;
    char *language;  /* used for accept-language testing */

    rq->directive_is_cacheable = 1;

    if(!path)
        return REQ_NOACTION;

    if(!dr) {
        log_error(LOG_MISCONFIG, "document-root", sn, rq, 
                  "missing parameter (need root)");
        return REQ_ABORTED;
    }

    if(path->value[0] != '/') {
        protocol_status(sn, rq, PROTOCOL_NOT_FOUND, NULL);
        return REQ_ABORTED;
    }

    /* Check the IP address to see if it matches the one in address */
    /* XXXMB - this is really inefficient with all these string compares */
    if (address) {
        if (sn->inbuf->address) {
            if (strcmp(address, sn->inbuf->address)) {
                return REQ_NOACTION;
            }
        }
    }

    if (port) {
        /* XXXMB - what a shame to call atoi here */
        if (sn->local_addr.sin_port != atoi(port))
            return REQ_NOACTION;
    }

    t = (char *) MALLOC(strlen(path->value) + strlen(dr) + 1 + 10); /* 10: /$$LANGDIR  */
    util_sprintf(t, "%s%s", dr, path->value);
    FREE(path->value);

    /* translate name if accept-language info is there */
    if (conf_getglobals()->accept_language) {
        char *langpath;
        if (langpath = lang_acceptlang_file(t, rq->headers)) {
            FREE(t);
            t = langpath;
            /* set HTTP Vary tag, which notify client that this document is subject
               to change based on Accept-Language header from client */
            if (pblock_findval("Vary", rq->srvhdrs) == NULL)
                pblock_nvinsert("Vary", "Accept-language", rq->srvhdrs);

            /* disable cache if the file is fetched based 
             * on accept-language header
             */ 
            rq->directive_is_cacheable = 0;

        }
    }

    path->value = t;

    pblock_nvinsert("ntrans-base", dr, rq->vars);

    return REQ_PROCEED;
}


/* ------------------------------ user2name ------------------------------- */


int ntrans_user2name(pblock *pb, Session *sn, Request *rq)
{
    pb_param *user = pblock_find("user", rq->vars);
    pb_param *name = pblock_remove("name", rq->vars);

    rq->directive_is_cacheable = 1;

    if(name) {
        FREE(user->name);
        user->name = name->name;
        FREE(name->value);
        FREE(name);
    }
    else {
        FREE(user->name);
        user->name = STRDUP("name");
    }
    return REQ_PROCEED;
}


/* ------------------------------- redirect ------------------------------- */


int ntrans_redirect(pblock *pb, Session *sn, Request *rq)
{
    char *path = pblock_findval("ppath", rq->vars);
    char *pfx = pblock_findval("from", pb);
    char *url = pblock_findval("url", pb);
    char *upfx = pblock_findval("url-prefix", pb);
    char *escape = pblock_findval("escape", pb);
    pb_param *pp;
    int l;

    rq->directive_is_cacheable = 1;

    if((!pfx) || (!(url || upfx))) {
        log_error(LOG_MISCONFIG, "redirect", sn, rq, 
                  "missing parameter (need from, url)");
        return REQ_ABORTED;
    }
    l = strlen(pfx);

/*    if(strncmp(path, pfx, l) || (path[l] && (path[l] != '/')))
 *    Per bug 40890, some folks believe that the prefix must match a complete
 *    pathname component, while others believe that a simple pattern match
 *    is the right thing to do.  In the 1.x servers, a simple pattern match was
 *    done.  In 2.x, this was intentionally changed to a full component (per
 *    the code above.  In 3.x, this is being changed back due to customer
 *    complaint, and to match NCSA/Apache behavior.  It could allow users to
 *    specify redirect loops more easily (i.e. by mistake).
 */
#ifdef XP_WIN32
    if(strncasecmp(path, pfx, l))
#else
    if(strncmp(path, pfx, l))
#endif
        return REQ_NOACTION;

    protocol_status(sn, rq, PROTOCOL_REDIRECT, NULL);

    if(url)
        pblock_nvinsert("url", url, rq->vars);
    else {
        pp = param_create("url", NULL);
        pp->value = (char *) MALLOC(strlen(path) - l + strlen(upfx) + 1);
        util_sprintf(pp->value, "%s%s", upfx, path + l);
        pblock_pinsert(pp, rq->vars);
    }

    /* Check to see if the directive specifies that the URL to be
     * redirected should not be escaped. By default, URLs that are
     * redirected are always escaped. */

    if (escape)
        pblock_nvinsert("escape", escape, rq->vars);

    return REQ_ABORTED;
}


/* --------------------------- mozilla-redirect --------------------------- */


int ntrans_mozilla_redirect(pblock *pb, Session *sn, Request *rq)
{
    char *ppath = pblock_findval("ppath", rq->vars);
    char *from = pblock_findval("from", pb);
    char *url = pblock_findval("url", pb);
    char *ua;

    if((!from) || (!url)) {
        log_error(LOG_MISCONFIG, "mozilla-redirect", sn, rq,
                  "missing parameter (need from, url)");
        return REQ_ABORTED;
    }
    if(!WILDPAT_CMP(ppath, from)) {
        if(request_header("user-agent", &ua, sn, rq) == REQ_ABORTED)
            return REQ_ABORTED;

        if(util_is_mozilla(ua, "0", "96")) {
            protocol_status(sn, rq, PROTOCOL_REDIRECT, NULL);
            pblock_nvinsert("url", url, rq->vars);
            return REQ_ABORTED;            
        }
    }
    return REQ_NOACTION;
}


/* ------------------------------- homepage ------------------------------- */


int ntrans_homepage(pblock *pb, Session *sn, Request *rq)
{
    char *hp = pblock_findval("path", pb);
    pb_param *pp = pblock_find("ppath", rq->vars);

    rq->directive_is_cacheable = 1;

    if(strcmp(pp->value, "/"))
        return REQ_NOACTION;

    /* Check if this is an absolute path or a relative path */
#ifdef XP_WIN32
    if ((hp[0] == '/' || hp[0] == '\\') ||
        (hp[0] != '\0' && hp[1] == ':')) {
#else
    if(hp[0] == '/') {
#endif
        FREE(pp->value);
        pp->value = STRDUP(hp);
        return REQ_PROCEED;
    }
    pp->value = (char *) REALLOC(pp->value, 1 + strlen(hp) + 1);
    strcpy(&pp->value[1], hp);
    return REQ_NOACTION;
}


/* ----------------------------- assign-name ------------------------------ */


int ntrans_assign_name(pblock *pb, Session *sn, Request *rq)
{
    char *path = pblock_findval("ppath", rq->vars);
    char *from = pblock_findval("from", pb);
    char *name = pblock_findval("name", pb);
    char *stop = pblock_findval("stop", pb);

    rq->directive_is_cacheable = 1;

    if((!from) || (!name)) {
        log_error(LOG_MISCONFIG, "assign-name", sn, rq, 
                  "missing parameter (need from, name)");
        return REQ_ABORTED;
    }
    /* Don't override any names previously assigned */
    if(pblock_find("name", rq->vars))
        return REQ_NOACTION;
    if(!WILDPAT_CMP(path, from))  {
        pblock_nvinsert("name", name, rq->vars);
        if (stop)
            return REQ_PROCEED;
    }

    return REQ_NOACTION;
}

