/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * otype.c: ObjectType directive related functions
 * 
 * Rob McCool
 * 
 */


#include "netsite.h"       /* FREE, STRDUP */
#include "safs/otype.h"
#include "base/cinfo.h"     /* cinfo_find */
#include "frame/log.h"

#include "base/shexp.h"     /* shexp_cmp */
#include "frame/protocol.h" /* protocol_status */

#include "base/util.h"      /* is_mozilla, sprintf */

#include <sys/types.h>
#include <sys/stat.h>


/* --------------------------- otype_forcetype ---------------------------- */


#define OTYPE_ADD(s1, s2) \
  if((ps = pblock_findval(s1, pb))) \
      if(!(pp = pblock_find(s2, rq->srvhdrs))) \
          pblock_nvinsert(s2, ps, rq->srvhdrs);


int otype_forcetype(pblock *pb, Session *sn, Request *rq)
{
    char *ps;
    pb_param *pp;

    rq->directive_is_cacheable = 1;

    OTYPE_ADD("type", "content-type");
    OTYPE_ADD("enc",  "content-encoding");
    OTYPE_ADD("lang", "content-language");
    OTYPE_ADD("charset", "magnus-charset");

    return REQ_PROCEED;
}



/* --------------------------- otype_typebyexp ---------------------------- */


int otype_typebyexp(pblock *pb, Session *sn, Request *rq) {
    char *exp = pblock_findval("exp", pb);
    char *path = pblock_findval("path", rq->vars);

    rq->directive_is_cacheable = 1;

    if(!exp) {
        log_error(LOG_MISCONFIG, "type-by-exp", sn, rq, 
                  "parameter missing (exp)");
        return REQ_ABORTED;
    }

    if(!(WILDPAT_CMP(path, exp)))
        return otype_forcetype(pb, sn, rq);
    else 
        return REQ_NOACTION;
}


/* --------------------------- otype_exec_shtml --------------------------- */


/* Aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaack hack barrf */

int otype_shtmlhacks(pblock *pb, Session *sn, Request *rq)
{
    char *path = pblock_findval("path", rq->vars);
    char *do_exec = pblock_findval("exec-hack", pb);
    int l;

    /* This is cachable- we always do the same thing based SOLELY
     * on the URI...
     */
    rq->directive_is_cacheable = 1;

    l = strlen(path);
    if((!strcasecmp(&path[l-4],".htm")) || (!strcasecmp(&path[l-5],".html"))) {
#ifdef XP_UNIX
        if(do_exec) {
            struct stat *fi = request_stat_path(NULL, rq);

            if(fi && (!(fi->st_mode & S_IXUSR)))
                return REQ_NOACTION;
        }
#endif /* XP_UNIX */
        param_free(pblock_remove("content-type", rq->srvhdrs));
        pblock_nvinsert("content-type", "magnus-internal/parsed-html",
                        rq->srvhdrs);
        return REQ_PROCEED;
    }
    return REQ_NOACTION;
}


/* --------------------------- otype_find2pblock -------------------------- */


#define OTYPE_CIADD(c, s) \
  if(c) \
      if(!(pp = pblock_find(s, rq->srvhdrs))) \
          pblock_nvinsert(s, c, rq->srvhdrs); \


NSAPI_PUBLIC int otype_ext2type(pblock *param, Session *sn, Request *rq)
{
    cinfo *ci;
    pb_param *pp;
    char *path = pblock_findval("path", rq->vars);
    struct stat *finfo;

    rq->directive_is_cacheable = 1;

    /* Check for directory and set type appropriately */
    /* If not found, ignore error */
    if( (finfo = request_stat_path(path, rq)) ) {
        if(S_ISDIR(finfo->st_mode)) {
            if(!(pp = pblock_find("content-type", rq->srvhdrs)))
                pblock_nvinsert("content-type", "magnus-internal/directory", 
                                rq->srvhdrs);
            return REQ_PROCEED;
        }
    }
    ci = cinfo_find(path);
    if(ci) {
        OTYPE_CIADD(ci->type, "content-type");
        OTYPE_CIADD(ci->encoding, "content-encoding");
        OTYPE_CIADD(ci->language, "content-language");

        FREE(ci);
    }

    return REQ_PROCEED;
}


/* ------------------------------ otype_init ------------------------------ */


int otype_init(pblock *pb, Session *sn, Request *rq)
{
    char *smt = pblock_findval("mime-types", pb);
    char *lmt = pblock_findval("local-types", pb);
    char *err;

    cinfo_init();
#if 0
    magnus_atrestart((void (*)(void *))cinfo_terminate, NULL);
#endif

    if((err = cinfo_merge(smt))) {
        pblock_nvinsert("error", err, pb);
        FREE(err);
        return REQ_ABORTED;
    }

    if(lmt && (err = cinfo_merge(lmt))) {
        pblock_nvinsert("error", err, pb);
        FREE(err);
        return REQ_ABORTED;
    }
    return REQ_PROCEED;
}


/* --------------------------- otype_imgswitch ---------------------------- */


int otype_imgswitch(pblock *pb, Session *sn, Request *rq)
{
    char *ct, *ua, *pa, *npath, *t;
    struct stat fi;
    pb_param *pp;

    /* This routine might be cacheable.  Lets check the objtype so
     * far to see if we are talking about an image file.  If this path
     * is not an image, we can go ahead and cache.
     */
    rq->directive_is_cacheable = 1;

    if(!(ct = pblock_findval("content-type", rq->srvhdrs)))
        return REQ_NOACTION;
    if(strncasecmp(ct, "image/", 6))
        return REQ_NOACTION;

    /* This routine is still running, so it must be an image we are
     * dealing with.  This is not cacheable since we need the user-agent
     * info to determine which file we return.
     */
    rq->directive_is_cacheable=0;

    /* In the absence of a capabilities header, use user-agent */
    if(request_header("user-agent", &ua, sn, rq) == REQ_ABORTED)
        return REQ_ABORTED;
    /* We want to be nice to proxies */
    if(request_header("proxy-agent", &pa, sn, rq) == REQ_ABORTED)
        return REQ_ABORTED;

    if((!ua) || pa || strstr(ua, "roxy"))
        return REQ_NOACTION;

    /* Look for jpeg if we're talking to mozilla and image is .gif */
    if((!strncasecmp(ua, "mozilla", 7)) && (!strcasecmp(ct, "image/gif"))) {
        npath = STRDUP(pblock_findval("path", rq->vars));
        if(!(t = strstr(npath, ".gif")))
            return REQ_NOACTION;
        t[1] = 'j'; t[2] = 'p'; t[3] = 'g';
        if(stat(npath, &fi) == -1) {
            FREE(npath);
            return REQ_NOACTION;
        }
        pp = pblock_find("path", rq->vars);
        FREE(pp->value);
        pp->value = npath;

        /* don't check return; it should work. */
        request_stat_path(npath, rq);

        pp = pblock_find("content-type", rq->srvhdrs);
        FREE(pp->value);
        pp->value = STRDUP("image/jpeg");

        return REQ_PROCEED;
    }
    return REQ_NOACTION;
}


/* --------------------------- otype_htmlswitch --------------------------- */


/* Sigh. Another stupid pet trick that will get dropped on the floor */
int otype_htmlswitch(pblock *pb, Session *sn, Request *rq)
{
    char *ct, *ua, *pa, *npath, *t;
    struct stat fi;
    pb_param *pp;

    /* This routine might be cacheable.  Lets check the objtype so
     * far to see if we are talking about a text file.  If this path
     * is not a text, we can go ahead and cache.
     */
    rq->directive_is_cacheable = 1;

    if(!(ct = pblock_findval("content-type", rq->srvhdrs)))
        return REQ_NOACTION;
    if(strncasecmp(ct, "text/", 5))
        return REQ_NOACTION;

    /* This is still running, so it must be a text file we are
     * dealing with.  This is not cacheable since we need the user-agent
     * info to determine which file we return.
     */
    rq->directive_is_cacheable = 0;

    /* In the absence of a capabilities header, use user-agent */
    if(request_header("user-agent", &ua, sn, rq) == REQ_ABORTED)
        return REQ_ABORTED;
    /* We want to be nice to proxies */
    if(request_header("proxy-agent", &pa, sn, rq) == REQ_ABORTED)
        return REQ_ABORTED;

    if((!ua) || pa || strstr(ua, "roxy"))
        return REQ_NOACTION;

    /* Look for html3 if we're talking to mozilla and find HTML */
    if(util_is_mozilla(ua, "1", "1") && (!strcasecmp(ct, "text/html"))) {
        t = pblock_findval("path", rq->vars);
        npath = (char *) MALLOC(strlen(t) + 1 + 1);
        util_sprintf(npath, "%s3", t);
        if(stat(npath, &fi) == -1) {
            FREE(npath);
            return REQ_NOACTION;
        }
        pp = pblock_find("path", rq->vars);
        FREE(pp->value);
        pp->value = npath;

        /* don't check return; it should work. */
        request_stat_path(npath, rq);
        return REQ_PROCEED;
    }
    return REQ_NOACTION;
}
