/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * pcheck.c: PathCheck functions
 * 
 * Rob McCool
 */


#include "netsite.h"
/* removed for ns security integration
#include "sslio/sslio.h"
*/
extern "C" {
#include "ssl.h"
}
#include "safs/pcheck.h"
#include "base/shexp.h"
#include "base/session.h"
#include "frame/req.h"
#include "frame/log.h"
#include "frame/http.h"
#include "base/util.h"  /* util_uri_is_evil */
#include "base/language.h"

#include "base/pblock.h"
#include "frame/protocol.h"
#include "frame/conf.h"

#include <sys/stat.h>
#include <ctype.h>


/* --------------------------- pcheck_find_path --------------------------- */


/*
 * Takes a given path, and figures out if there is any path_info attached.
 * If no explicit path_info was provided by nametrans, and the file doesn't 
 * exist as specified, it tries to find it by groping through the filesystem.
 * 
 * This type of implicit path_info cannot be attached to directories. Such
 * a request will be flagged as not found.
 * 
 * pb unused.
 */

int pcheck_find_path(pblock *pb, Session *sn, Request *rq) 
{
    pb_param *pp = pblock_find("path", rq->vars);
    char *path;
    char *path_info = pblock_findval("path-info", rq->vars);
    struct stat *finfo;

    path = pp->value;

    rq->directive_is_cacheable = 1;

    if( (finfo = request_stat_path(path, rq)) )
        return REQ_NOACTION;

    if(path_info)
        goto bong;

    path_info = &path[strlen(path)];
    while(1) {
        /* Change all occurrences of '/' to FILE_PATHSEP for WIN32 */
        for(  ; path_info != path; --path_info)
            if(*path_info == FILE_PATHSEP)
                break;

        if(path_info == path)
            break;

        *path_info = '\0';
        if( (finfo = request_stat_path(path, rq))) {
            if(!(S_ISREG(finfo->st_mode))) {
                *path_info = FILE_PATHSEP;
                goto bong;
            }
            else {
                *path_info = FILE_PATHSEP;
                pblock_nvinsert("path-info", path_info, rq->vars);
                *path_info = '\0';
                return REQ_PROCEED;
            }
        }
        else *path_info-- = FILE_PATHSEP;
    }

  bong:
    /* This was changed to support virtual documents */
    return REQ_NOACTION;
}



/* ------------------------ pcheck_deny_existence ------------------------- */


int pcheck_deny_existence(pblock *pb, Session *sn, Request *rq)
{
    char *bong_file = pblock_findval("bong-file", pb);
    pb_param *path = pblock_find("path", rq->vars);
    char *check_path = pblock_findval("path", pb);

    rq->directive_is_cacheable = 1;

    if((!check_path) || (!WILDPAT_CMP(path->value, check_path))) {
        log_error(LOG_SECURITY, "deny-existence", sn, rq, 
                  "denying existence of %s", pblock_findval("path", rq->vars));
        if(bong_file) {
            FREE(path->value);
            path->value = STRDUP(bong_file);
            return REQ_PROCEED;
        }
        protocol_status(sn, rq, PROTOCOL_NOT_FOUND, NULL);
        return REQ_ABORTED;
    }
    else
        return REQ_NOACTION;
}


/* -------------------------- pcheck_find_index --------------------------- */


/* Helper: Take dir and append path */

char *_make_path(char *dir, char *file, char *t)
{
    char *p;

    for(p = t; (*p = *dir); ++p, ++dir);
    while( (*p = *file) )
        (++p, ++file);
    return t;
}

/*
 * This function will first determine if the given path is a directory.
 * If so, it will try to find each of the files in the index-names list
 * in order. If not, it will return REQ_NOACTION.
 * 
 * If the file is found, and there was no trailing slash, this function will
 * force the client to redirect.
 */

int pcheck_find_index(pblock *pb, Session *sn, Request *rq)
{
    pb_param *pp;
    char *path;
    char *flist, *fl = pblock_findval("index-names", pb);
    char *t, *u, *v;
    char *uri, *luri;
    struct stat *finfo;
    int lp;
	char *method;
    char *langpath;
    char *qs;

    rq->directive_is_cacheable = 1;

    if(!fl) {
        log_error(LOG_WARN, "find-index", sn, rq, "missing parameter (flist)");
        return REQ_ABORTED;
    }

    /* If the url contains query, do not add index.html to the end of it */
    if ( (qs = pblock_findval("query", rq->reqpb)) ) {
        /* THIS SUCKS- XXXMB; fixing bug 79435 */
        if (!strcmp(qs, "properties"))
            return REQ_NOACTION;
    }

	/* If method is GET,  append "index.html" to the uri */
    if(( method = pblock_findval( "method", rq->reqpb ))) 
	    if( strcmp( method, "GET" ) )
	        return REQ_NOACTION;

    pp = pblock_find("path", rq->vars);
    path = pp->value;
    lp = strlen(path);

    finfo = request_stat_path(path, rq);
    if((!finfo) || (!S_ISDIR(finfo->st_mode)))
        return REQ_NOACTION;


    if(path[lp - 1] != FILE_PATHSEP) {
        pb_param *url = param_create("url", NULL);
        url->value = protocol_uri2url_dynamic(pblock_findval("uri", rq->reqpb),
                                              "/", sn, rq);
        pblock_pinsert(url, rq->vars);
        protocol_status(sn, rq, PROTOCOL_REDIRECT, NULL);
        return REQ_ABORTED;
    }

    flist = STRDUP(fl);
    t = flist;
    u = flist;

    /* We overallocate v a bit but shouldn't be much */
    v = (char *) MALLOC(strlen(t) + lp + 1);

    while(*t) {
        while((*u) && (*u != ',')) ++u;
        if(u == t) {
            /* grumble grumble what a dope */
            t = ++u;
            continue;
        }
        if(*u)
            *u++ = '\0';

        /* additonal language lookup if accept-language feature is on */
        langpath = _make_path(path, t, v);
        if (conf_getglobals()->accept_language) {
            char *newpath;
            if (newpath = lang_acceptlang_file(v, rq->headers)) {
                langpath = newpath;
                /* disable cache if the file is fetched based 
                 * on accept-language header
                 */ 
                rq->directive_is_cacheable = 0;
            }
        }

        if((finfo = request_stat_path(langpath, rq))) {
            if(S_ISREG(finfo->st_mode)) {
                FREE(pp->value);
                pp->value = langpath;
                if (v != langpath)
                    FREE(v);
                FREE(flist);
		/* Create a new pblock entry called luri with the index name 
		 * This is needed by the ACL gathering code.
		 */
		uri = pblock_findval("uri", rq->reqpb);
		luri = (char *)MALLOC(strlen(t) + strlen(uri) + 1);
		strcpy(luri, uri);
		strcat(luri, t);
		pblock_nvinsert("luri", luri, rq->reqpb);
                return REQ_PROCEED;
            }
        }
        t = u;
    }
    FREE(v);
    FREE(flist);
    return REQ_NOACTION;
}


/* --------------------------- pcheck_uri_clean --------------------------- */


int pcheck_uri_clean(pblock *pb, Session *sn, Request *rq)
{ 
	
    char *t = pblock_findval("path", rq->vars);

    rq->directive_is_cacheable = 1;

    if(!t[0])
        return REQ_NOACTION;
#ifdef XP_UNIX
    if((t[0]) != '/') {
#else /* WIN32 */
    if((t[0] != '/') &&  (!(isalpha(t[0]) && (t[1] == ':')))) {
#endif /* XP_UNIX */

        /* chastise stupid client */
        log_error(LOG_WARN, "unix-uri-clean", sn, rq, "not full path: %s", t);
        return REQ_ABORTED;
    }

    if(util_uri_is_evil(t)) {
        protocol_status(sn, rq, PROTOCOL_NOT_FOUND, NULL);
        return REQ_ABORTED;
    }

    return REQ_PROCEED;
}


/* -------------------------- pcheck_ntcgicheck --------------------------- */


#ifdef XP_WIN32
int pcheck_ntcgicheck(pblock *pb, Session *sn, Request *rq)
{ 
    char *extension = pblock_findval("extension", pb);
    pb_param *pp = pblock_find("path", rq->vars);
    char *cgiPath, *newPath ;
	struct stat statBuffer;

    int pathLength;

    rq->directive_is_cacheable = 1;

    if((!extension)) {
        log_error(LOG_MISCONFIG, "ntcgicheck", sn, rq, 
                  "missing parameter (need extension)");
        return REQ_ABORTED;
    }

    if(!pp)
        return REQ_NOACTION;

    cgiPath = pp->value;
	pathLength = strlen(cgiPath);

	if (cgiPath[pathLength - 1] == '/') {
		// if the path is not that of a plain file, dont add to it...
		return REQ_NOACTION;
	} else {
		// check if the file is a directory...
		// I think that the file handle is cached.
		// If not, to do this differently will make the code ugly...

		// stat returns 0 if it succeeds, if it fails, let PathCheck
		// catch it after we append the extension...

		if (!(stat(cgiPath, &statBuffer))) {
			if ( _S_IFDIR & statBuffer.st_mode ) {
				return REQ_NOACTION;
			}
		}
	}

	// If the last three letters end in .cgi change to .exe
	// If the path does not have an extension, add the default extension...
	// The <= is there in case it has neither a '.' nor a '/'.

	if (strrchr(cgiPath, '.') <= strrchr(cgiPath,'/')) {
		pathLength = strlen(cgiPath) + strlen(extension) + 1;
		newPath = (PCHAR)MALLOC(pathLength);
		sprintf(newPath, "%s%s\0", cgiPath, extension);
		FREE( pp->value );
		pp -> value  =  newPath;
	} else if (!strcmp( (cgiPath + pathLength - 4), ".cgi")) {
		// We are removing .cgi and adding in the default extension...
		pathLength = strlen(cgiPath) - 4 + strlen(extension) + 1;
		newPath = (PCHAR)MALLOC(pathLength);
		strncpy( newPath, cgiPath, pathLength - 4);
		strcpy( newPath + pathLength - 5 , extension);
		newPath[pathLength - 1] = '\0';
		FREE(pp->value);
		pp -> value = newPath;
	}

    return REQ_PROCEED;
}
#endif /* XP_WIN32 */


/* ------------------------- pcheck_require_auth -------------------------- */


int pcheck_require_auth(pblock *pb, Session *sn, Request *rq)
{
    char *t, *u, *type;
    char buf[256];
    int lt;

    rq->directive_is_cacheable = 1;

    if( (t = pblock_findval("path", pb)) )
        if(WILDPAT_CMP(pblock_findval("path", rq->vars), t))
            return REQ_NOACTION;

    rq->directive_is_cacheable = 0;

    if(!(type = pblock_findval("auth-type", pb))) {
        log_error(LOG_MISCONFIG, "require-auth", sn, rq, 
                  "required parameter missing (auth-type)");
        return REQ_ABORTED;
    }
    if(!(t = pblock_findval("auth-type", rq->vars)))
        goto punt;

    if(strcasecmp(t, type))
        goto punt;

    if( (t = pblock_findval("auth-user", rq->vars)) ) {
        if( (u = pblock_findval("auth-user", pb))) {
            if(shexp_cmp(t, u))
                goto punt;
        }
    }

    if(!(u = pblock_findval("auth-group", pb)))
        return REQ_PROCEED;
    if(!(t = pblock_findval("auth-group", rq->vars)))
        goto punt;  /* group required but none given */

    if(!shexp_cmp(t, u))
        return REQ_PROCEED;

  punt:
    if(!(u = pblock_findval("realm", pb))) {
        log_error(LOG_MISCONFIG, "require-auth", sn, rq, 
                  "missing parameter (realm)");
        return REQ_ABORTED;
    }
    lt = strlen(type);

    if(lt > 256)
        type[255] = '\0';
    if(strlen(u) > (256 - lt))
        u[256 - lt] = '\0';

    util_sprintf(buf,"%s realm=\"%s\"", type, u);

    pblock_nvinsert("WWW-authenticate", buf, rq->srvhdrs);
    protocol_status(sn, rq, PROTOCOL_UNAUTHORIZED, NULL);

    return REQ_ABORTED;
}


/* -------------------------- pcheck_find_links --------------------------- */

#ifdef XP_UNIX
int pcheck_find_links(pblock *pb, Session *sn, Request *rq)
{
    char *path = pblock_findval("path", rq->vars);
    char *disable = pblock_findval("disable", pb);
    char c, *t, *dir = pblock_findval("dir", pb);
    struct stat fi;
    int soft, hard;
    register int l;
    uid_t uid = 0;

    rq->directive_is_cacheable = 1;

    if((!disable) || (!dir)) {
        log_error(LOG_MISCONFIG, "find-links", sn, rq, 
                  "missing parameter (need disable and dir)");
        return REQ_ABORTED;
    }
    for(soft = 0, hard = 0; *disable; ++disable) {
        if(*disable == 'h')
            hard = 1;
        else if(*disable == 's')
            soft = 1;
        else if(*disable == 'o') {
            uid = getuid();
            soft = 1;
        }
    }

    l = strlen(dir);
    if(l && (dir[l-1] == '/')) --l;

    t = path;

    if(l) {
        if(dir[0] == '/') {
            if(strncmp(path, dir, l))
                return REQ_NOACTION;
            t += l;
        }
        else {
            if(!(t = strstr(path, dir)))
                return REQ_NOACTION;
            t += l;
        }
    }

    while(1) {
        while(*t && (*t != '/')) ++t;
        c = *t;
        *t = '\0';
        if(lstat(path, &fi) == -1) {
            log_error(LOG_WARN, "find-links", sn, rq, 
                      "while servicing %s, lstat of %s failed (%s)",
                      pblock_findval("uri", rq->vars), path, system_errmsg());
            protocol_status(sn, rq, PROTOCOL_FORBIDDEN, NULL);
            return REQ_ABORTED;
        }
        if(soft && S_ISLNK(fi.st_mode)) {
                struct stat realfi;

                if(stat(path, &realfi) == -1) {
                    log_error(LOG_WARN, "find-links", sn, rq, 
                              "while servicing %s, stat of %s failed (%s)",
                              pblock_findval("uri", rq->vars), path, 
                              system_errmsg());
                    protocol_status(sn, rq, PROTOCOL_FORBIDDEN, NULL);
                    return REQ_ABORTED;
                }
                if(realfi.st_uid != uid)
                    goto bong;
        }
        if(hard && S_ISREG(fi.st_mode) && (fi.st_nlink != 1)) {
          bong:
            log_error(LOG_WARN, "find-links", sn, rq,
                      "will not follow link %s", path);
            protocol_status(sn, rq, PROTOCOL_FORBIDDEN, NULL);
            return REQ_ABORTED;
        }
        if(!(*t++ = c))
            break;
    }
    return REQ_PROCEED;
}
#endif /* XP_UNIX */

/* MLM - Check for vulnerable (pre-1.12) versions of the navigator.  
 *       Bong them if they try to access you.  Requested by customers.
 *       Was in 1.12, somehow got regressed out of 2.0.
 */
/* --------------------- pcheck_detect_vulnerable ------------------------ */
 
int pcheck_detect_vulnerable(pblock *pb, Session *sn, Request *rq)
{
    char *ua;
    char *whereto;
 
    if(request_header("user-agent", &ua, sn, rq) == REQ_ABORTED)
        return REQ_ABORTED;
 
    if((!ua) || strncasecmp(ua, "Mozilla/", 8))
        return REQ_PROCEED;
 
    /* Detect if it's a major rev <= 1. */
    /* If not, they're using 2.0 and are Happy.*/
    if(ua[8] > '1')
        return REQ_PROCEED;
 
    /* Bong 0.xx releases */
    else if((ua[8] < '1') || (ua[9] != '.'))
        goto deny_request;

    /* Minor version. Check to see if it's patch level "1.12" or "1.22" */
    if(ua[10] > '2')
        return REQ_PROCEED;
    if((!isdigit(ua[11])) || (ua[11] < '2'))
        goto deny_request;
    else
        return REQ_PROCEED;
 
deny_request:
    whereto = pblock_findval("dest", pb);
    if(!whereto)  {
        log_error(LOG_MISCONFIG, "detect_vulnerable", sn, rq,
                                 "missing from parameter (need 'dest')");
        return REQ_ABORTED;
    }
 
    if(util_is_url(whereto))   {
        protocol_status(sn, rq, PROTOCOL_REDIRECT, NULL);
        pblock_nvinsert("url", whereto, rq->vars);
        return REQ_ABORTED;
    }  else  {
        char *newloc = http_uri2url(whereto, "");
        char *uri = pblock_findval("uri", rq->reqpb);
 
        /* If they only gave a partial path, then allow them to get ONLY
         * that path.  (This means no graphics, but hey.)  If they're not
         * requesting the same path, redirect them. */
        if(!strcmp(uri, whereto))
            return REQ_PROCEED;
 
        protocol_status(sn, rq, PROTOCOL_REDIRECT, NULL);
        pblock_nvinsert("url", newloc, rq->vars);
        FREE(newloc);
        return REQ_ABORTED;
    }
}

/* Invalidate the current SSL session in the SSL session cache */
int pcheck_ssl_logout(pblock *pb, Session *sn, Request *rq)
{
    if (security_active) {
        SSL_InvalidateSession(sn->csd);
        return REQ_PROCEED;
    }

    return REQ_NOACTION;
}

int pcheck_set_cache_control(pblock *pb, Session *sn, Request *rq)
{
    char *cc_value = pblock_findval("control", pb);
    pb_param *pp = pblock_find("path", rq->vars);
	char cc_buf[100];


	if (rq->protv_num < 101)		/* Must be HTTP/1.1 or higher */
		return REQ_NOACTION;

    rq->directive_is_cacheable = 1;

    if(!cc_value) {
        log_error(LOG_MISCONFIG, "cache-control", sn, rq, 
                  "missing parameter (need a cache-control value)");
        return REQ_ABORTED;
    }


    if(!pp)
        return REQ_NOACTION;

	util_sprintf(cc_buf,"\"%s\"",cc_value);

    pblock_nvinsert("cache-control", cc_buf, rq->srvhdrs);

    return REQ_PROCEED;
}

