/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * service.c: Handle Service-related things (including directive and user-
 * callable functions
 * 
 * Rob McCool
 */

#include "netsite.h"

#include "safs/service.h"
#include "frame/log.h"      /* log_error */
#include "frame/protocol.h" /* protocol_start_response */
#include "frame/req.h"      /* request_handle_processed */
#include "base/pblock.h"    /* pblock_findval */
#include "base/file.h"      /* system_errmsg */
#include "base/util.h"      /* util_getline, util_is_url */
#include "frame/httpact.h"	/* servact_handle_processed */
#include "frame/http.h"
#ifdef ACCELERATOR_CACHE
#include <sys/types.h>
#ifdef XP_UNIX
#include <sys/mman.h>
#include <sys/uio.h>
#endif
#include "base/time_cache.h"
#include "frame/accel_file_cache.h"
#endif /* ACCELERATOR_CACHE */

#include <errno.h>

#define TRAILER_MAX_LEN 512
#define TRAILER_LASTMOD ":LASTMOD:"
#define TRAILER_LASTMOD_LEN 9


#define CM_PROPERTY_STR     "properties"
#define CM_WEBPUB_NS        "webpub-ui"
#define CM_SYS_PROP_NS      "wp-sys-prop"
#define CM_TOC_NS           "wp-toc"
#define CM_VER_INFO_NS      "wp-ver-info"
#define CM_LINK_INFO_NS     "wp-link-info"
#define CM_CUSTOM_FEILD_NS  "wp-custom-info"
#define CM_VER_DIFF_NS      "wp-ver_diff"
#define CM_START_VERSION_NS "wp-start-ver"
#define CM_STOP_VERSION_NS  "wp-stop-ver"
#define CM_VER_LINKS_NS     "wp-verify-link"
#define CM_UNCHECKOUT_NS    "wp-uncheckout"
#define CM_ADMIN_NS         "wp-admin"
#define CM_ADMIN_SYS_NS     "wp-admin-sys"
#define CM_USR_PROP_NS      "wp-usr-prop"



static int service_is_CM_req( char * query );


/* ------------------------- service_plain_range -------------------------- */


void _range_setvars(char *n1, char *n2, int maxlen,
                    int *rstart, int *rend, int *rlen)
{
    if(*n1) {
        *rstart = atoi(n1);
        if(*n2) {
            /* bytes=start-end */
            *rend = atoi(n2);
        }
        else {
            /* bytes=start- */
            *rend = maxlen - 1;
        }
    }
    else {
        /* bytes=-end */
        *rstart = maxlen - atoi(n2);
        *rend = maxlen - 1;
    }
    if(*rend > (maxlen - 1))
        *rend = maxlen - 1;
    if(*rstart > *rend)
        *rstart = *rend;

    *rlen = (*rend - *rstart) + 1;
}


int _range_send(SYS_FILE fd, SYS_NETFD sd, 
                int rstart, int rend, int rlen)
{
    int n, o, l = rlen;
    char buf[FILE_BUFFERSIZE];

    /* We have no guarantee where the fd is at this point */
    system_lseek(fd, rstart, SEEK_SET);

    for(o = 0; o != l; o += n) {
        if( (n = l - o) > sizeof(buf))
            n = sizeof(buf);
        if( ((n = system_fread(fd, buf, n)) == IO_ERROR) ||
           (net_write(sd, buf, n) == IO_ERROR) )
            return IO_ERROR;
    }
    return IO_OKAY;
}


int service_plain_range(pblock *param, Session *sn, Request *rq)
{
    char *rqrange = pblock_findval("range", rq->headers);
    char *rangebuf;
    char *range;
    char *ranges, sep[128], *path, *t, *ums, numstr[16], hdr[256];
    int rstart, rend, rlen, l, seplen = 0, domulti;
    struct stat *finfo;
    pb_param *doc_type;
    SYS_FILE fd;

    if (!rqrange)
	return REQ_NOACTION;
    range = rangebuf = STRDUP(rqrange);
    if((!range) || (strncmp(range, "bytes=", 6))) {
	FREE(rangebuf);
        return REQ_NOACTION;
    }
    range += 6;
    if(strchr(range, ';')) {
        /* According to draft... */
        protocol_status(sn, rq, PROTOCOL_NOT_IMPLEMENTED, NULL);
	FREE(rangebuf);
        return REQ_ABORTED;
    }
    if( (ranges = strchr(range, ',')) ) {
        domulti = 1;
        *ranges++ = '\0';
    }
    else
        domulti = 0;
    path = pblock_findval("path", rq->vars);
    if( (t = pblock_findval("path-info", rq->vars)) ) {
        protocol_status(sn, rq, PROTOCOL_NOT_FOUND, NULL);
        log_error(LOG_WARN, "send-file", sn, rq, 
                  "%s%s not found", path, t);
	FREE(rangebuf);
        return REQ_ABORTED;
    }
    if(!(finfo = request_stat_path(path, rq))) {
        protocol_status(sn, rq, (rtfile_notfound() ? PROTOCOL_NOT_FOUND : 
                                 PROTOCOL_FORBIDDEN), NULL);
        log_error(LOG_WARN, "send-file", sn, rq, "can't find %s (%s)",
                  path, rq->staterr);
	FREE(rangebuf);
        return REQ_ABORTED;
    }

    if(!(t = strchr(range, '-'))) {
	FREE(rangebuf);
        return REQ_NOACTION;
    }
    *t++ = '\0';

    /* Make sure IMS isn't present */
    param_free(pblock_remove("if-modified-since", rq->headers));
    protocol_set_finfo(sn, rq, finfo);

    /* XXXrobm this will require THREE gmtimes. Should this be in set_finfo */
    if( (ums = pblock_findval("unless-modified-since", rq->headers)) ) {
        struct tm *tms, tmss;
        tms = system_gmtime(&finfo->st_mtime, &tmss);
        /* Handle as normal request */
        if(!util_time_equal(tms, ums)) {
            param_free(pblock_remove("last-modified", rq->srvhdrs));
            param_free(pblock_remove("content-length", rq->srvhdrs));
	    FREE(rangebuf);
            return REQ_NOACTION;
        }
    }
    _range_setvars(range, t, finfo->st_size, &rstart, &rend, &rlen);
    if(!(pblock_find("status", rq->srvhdrs)))
        protocol_status(sn, rq, PROTOCOL_PARTIAL_CONTENT, NULL);

    /* Set content-length only if there's one range. */
    param_free(pblock_remove("content-length", rq->srvhdrs));
    if(!domulti) {
        util_itoa(rlen, numstr);
        pblock_nvinsert("content-length", numstr, rq->srvhdrs);

        util_sprintf(hdr, "bytes %d-%d/%d", rstart, rend, finfo->st_size);
        pblock_nvinsert("content-range", hdr, rq->srvhdrs);
        doc_type = NULL;
    }
    else {
        doc_type = pblock_remove("content-type", rq->srvhdrs);
        seplen = util_mime_separator(sep);
        util_sprintf(hdr, "multipart/x-byteranges; boundary=%s", &sep[4]);
        pblock_nvinsert("content-type", hdr, rq->srvhdrs);
        util_sprintf(&sep[seplen], "%c%c", CR, LF);
        seplen += 2;
    }
    if((fd = system_fopenRO(path)) == SYS_ERROR_FD) {
        protocol_status(sn, rq, (rtfile_notfound() ? PROTOCOL_NOT_FOUND : 
                                 PROTOCOL_FORBIDDEN), NULL);
        log_error(LOG_WARN, "send-file", sn, rq, "error opening %s (%s)", 
                  path, system_errmsg());
	FREE(rangebuf);
        return REQ_ABORTED;
    }
    if(protocol_start_response(sn,rq) != REQ_NOACTION) {
        while(1) {
            if(domulti) {
                net_write(sn->csd, sep, seplen);
                l = util_sprintf(hdr, "Content-type: %s\n"
                                 "Range: bytes %d-%d/%d\n\n",
                                 doc_type->value, rstart, rend, 
                                 finfo->st_size);
                net_write(sn->csd, hdr, l);
            }
            if(_range_send(fd, sn->csd, rstart, rend, rlen) == IO_ERROR) {
                system_fclose(fd);
                param_free(doc_type);
	        FREE(rangebuf);
                return REQ_EXIT;
            }
            if(!ranges)
                break;
            range = ranges;
            if( (ranges = strchr(ranges, ',')) )
                *ranges++ = '\0';
            if(!(t = strchr(range, '-')))
                break;
            *t++ = '\0';
            _range_setvars(range, t, finfo->st_size, &rstart, &rend, &rlen);
        }
    }
    if(domulti) {
        util_sprintf(&sep[seplen-2], "--%c%c", CR, LF);
        seplen += 2;
        net_write(sn->csd, sep, seplen);
    }
    system_fclose(fd);
    param_free(doc_type);
    FREE(rangebuf);
    return REQ_PROCEED;
}


/* -------------------------- service_plain_file -------------------------- */


int service_plain_file(pblock *param, Session *sn, Request *rq) 
{
    char *path, *t;
    SYS_FILE fd;
    filebuf_t *buf;
    struct stat *finfo;
    int ret;
#ifdef ACCELERATOR_CACHE
    accel_file_cache_entry_t *ac_entry = NULL;
#endif /* ACCELERATOR_CACHE */
 
    if( (ret = service_plain_range(param, sn, rq)) != REQ_NOACTION)
        return ret;
    path = pblock_findval("path", rq->vars);
    if( (t = pblock_findval("path-info", rq->vars)) ) {
        protocol_status(sn, rq, PROTOCOL_NOT_FOUND, NULL);
        log_error(LOG_WARN, "send-file", sn, rq, 
                  "%s%s not found", path, t);
        return REQ_ABORTED;
    }
    if(!(finfo = request_stat_path(path, rq))) {
        protocol_status(sn, rq, (rtfile_notfound() ? PROTOCOL_NOT_FOUND : 
                                 PROTOCOL_FORBIDDEN), NULL);
        log_error(LOG_WARN, "send-file", sn, rq, "can't find %s (%s)",
                  path, rq->staterr);
        return REQ_ABORTED;
    }

    if(protocol_set_finfo(sn, rq, finfo) == REQ_ABORTED)
        return REQ_ABORTED;

    if((fd = system_fopenRO(path)) == SYS_ERROR_FD) {
        protocol_status(sn, rq, (file_notfound() ? PROTOCOL_NOT_FOUND :
                                 PROTOCOL_FORBIDDEN), NULL);
        log_error(LOG_WARN, "send-file", sn, rq, "error opening %s (%s)",
                  path, system_errmsg());
        return REQ_ABORTED;
    }

    if(!(buf = filebuf_open_nostat(fd, FILE_BUFFERSIZE, finfo))) {
        system_fclose(fd);
        protocol_status(sn, rq, PROTOCOL_FORBIDDEN, NULL);
        log_error(LOG_WARN,"send-file", sn, rq,
                  "error opening buffer from %s (%s)", path,
              system_errmsg());
        return REQ_ABORTED;
    }

    pblock_nvinsert("accept-ranges", "bytes", rq->srvhdrs);
    if(!(pblock_find("status", rq->srvhdrs)))
        protocol_status(sn, rq, PROTOCOL_OK, NULL);

    if(protocol_start_response(sn,rq) != REQ_NOACTION) {
        if(filebuf_buf2sd(buf, sn->csd) == IO_ERROR) {
            filebuf_close(buf);
            return REQ_EXIT;
        }
    }
    filebuf_close(buf);

#ifdef ACCELERATOR_CACHE
    /* If we are cacheable, try to insert in cache now.  */
    if (rq->request_is_cacheable && rq->cached_headers && rq->status_num==PROTOCOL_OK) {
        char *uri = pblock_findval("uri", rq->reqpb);
        char *host = pblock_findval("host", rq->headers);
        char *local_ip = strlen(sn->inbuf->address)?sn->inbuf->address:NULL;
        int local_port = sn->local_addr.sin_port;

        if ( (ac_entry = accel_file_cache_lookup_uri(uri, 
            host, &sn->local_addr)) == NULL) {
            ac_entry = accel_file_cache_insert(
                uri, 
                host, 
                &sn->local_addr,
                path, 
                rq->cached_headers, 
                rq->cached_headers_len, 
                rq->os,
                NULL, 
                NULL, 
                NULL,
                NULL, /* user_data */
		rq->acllist 
            ); 
        }
        if (ac_entry) 
            accel_file_cache_use_decrement(ac_entry);
    }
#endif /* ACCELERATOR_CACHE */


    return REQ_PROCEED;
}

/* -------------------------- service_send_error -------------------------- */


int service_send_error(pblock *pb, Session *sn, Request *rq)
{
    char *new_path = pblock_findval("path", pb);
    pb_param *path = pblock_find("path", rq->vars);

    if(!new_path) {
        log_error(LOG_MISCONFIG, "send-error", sn, rq, 
                  "missing parameter (path)");
        return REQ_ABORTED;
    }
    if(path) {
        FREE(path->value);
        path->value = STRDUP(new_path);
    }
    else pblock_nvinsert("path", new_path, rq->vars);

    param_free(pblock_remove("content-type", rq->srvhdrs));
    /* If this was a CGI access, make sure path-info doesn't cause a 404 */
    param_free(pblock_remove("path-info", rq->vars));
    pblock_nvinsert("content-type", "text/html", rq->srvhdrs);

    return service_plain_file(NULL, sn, rq);
}


/* --------------------------- service_trailer ---------------------------- */


int service_trailer(pblock *pb, Session *sn, Request *rq)
{
    char *trail = pblock_findval("trailer", pb), *trailer = 0;
    char *timefmt = pblock_findval("timefmt", pb);
    char *path, timebuf[256], tbuf[TRAILER_MAX_LEN], *t, *u;
    struct stat *finfo;
    struct tm *tms, tmss;
    pb_param *pp;
    SYS_FILE fd;
    filebuf_t *buf;
    int l, cl;

    if(!trail) {
        log_error(LOG_MISCONFIG, "append-trailer", sn, rq, 
                  "missing parameter (need trailer)");
        goto punt;
    }
    trailer = STRDUP(trail);
    util_uri_unescape(trailer);
    path = pblock_findval("path", rq->vars);
    if( (t = pblock_findval("path-info", rq->vars)) ) {
        protocol_status(sn, rq, PROTOCOL_NOT_FOUND, NULL);
        log_error(LOG_WARN, "append-trailer", sn, rq, 
                  "%s%s not found", path, t);
        goto punt;
    }
    if((fd = system_fopenRO(path)) == SYS_ERROR_FD) {
        protocol_status(sn, rq, (file_notfound() ? PROTOCOL_NOT_FOUND : 
                                 PROTOCOL_FORBIDDEN), NULL);
        log_error(LOG_WARN, "append-trailer", sn, rq, "error opening %s (%s)",
                  path, system_errmsg());
        goto punt;
    }
    if(!(finfo = request_stat_path(path, rq))) {
        protocol_status(sn, rq, PROTOCOL_FORBIDDEN, NULL);
        log_error(LOG_WARN,"append-trailer", sn, rq, "can't fstat %s (%s)", 
                  path, rq->staterr);
        goto punt;
    }

    if(!(buf = filebuf_open_nostat(fd, FILE_BUFFERSIZE, finfo))) {
        protocol_status(sn, rq, PROTOCOL_FORBIDDEN, NULL);
        log_error(LOG_WARN, "append-trailer", sn, rq,
                  "error opening buffer from %s (%s)",path, system_errmsg());
        goto punt;
    }
    if(!(pblock_find("status", rq->srvhdrs)))
        protocol_status(sn, rq, PROTOCOL_OK, NULL);
    if(protocol_set_finfo(sn, rq, finfo) == REQ_ABORTED) {
        filebuf_close(buf);
        goto punt;
    }

    if(timefmt) {
        tms = system_localtime(&finfo->st_mtime, &tmss);
        (void)util_strftime(timebuf, timefmt, tms);
        l = strlen(timebuf);
        t = trailer;
        u = tbuf;

        while(*t) {
            if(!strncmp(t, TRAILER_LASTMOD, TRAILER_LASTMOD_LEN)) {
                strcpy(u, timebuf);
                t += TRAILER_LASTMOD_LEN;
                u += l;
            }
            else *u++ = *t++;

            if((u - tbuf) > (TRAILER_MAX_LEN - l))
                break;
        }
        *u = '\0';
        l = u - tbuf;
        t = tbuf;
    }
    else {
        t = trailer;
        l = strlen(t);
    }
    if( (pp = pblock_find("content-length", rq->srvhdrs)) ) {
        cl = l + atoi(pp->value);
        FREE(pp->value);
        util_itoa(cl, timebuf);
        pp->value = STRDUP(timebuf);
    }
    if(protocol_start_response(sn, rq) != REQ_NOACTION) {
        if(filebuf_buf2sd(buf, sn->csd) == IO_ERROR) {
            FREE(trailer);
            filebuf_close(buf);
            return REQ_EXIT;
        }
        net_write(sn->csd, t, l);
    }
    filebuf_close(buf);
    FREE(trailer);
    return REQ_PROCEED;
  punt:
    FREE(trailer);
    return REQ_ABORTED;
}


/* --------------------------- service_imagemap --------------------------- */


#define MAXLINE 500
#define MAXVERTS 100
#define X 0
#define Y 1

#define isname(c) (!isspace((c)))

#define IMAP_BADREQ \
"<TITLE>Imagemap Error</TITLE><H1>Imagemap Error</H1>Your client did not send any coordinates. Either your client doesn't support imagemap or the map file was not accessed as a map."


void imap_die(char *msg, Session *sn, Request *rq)
{
    int ml = strlen(msg);
    char buf[16];

    protocol_status(sn, rq, PROTOCOL_OK, NULL);
    util_itoa(ml, buf);
    pblock_nvinsert("content-length", buf, rq->srvhdrs);
    pblock_nvinsert("content-type", "text/html", rq->srvhdrs);
    protocol_start_response(sn, rq);
    net_write(sn->csd, msg, ml);
}

int imap_handle(char *url, Session *sn, Request *rq, char *path, 
                int ncsa_compat)
{
    char *uri, *fullurl, *t, c = 0, *fulluri;

    if(util_is_url(url)) {
        pblock_nvinsert("url", url, rq->vars);
        protocol_status(sn, rq, PROTOCOL_REDIRECT, NULL);
        return REQ_ABORTED;
    }  else  {
	if(!ncsa_compat)  {
            uri = pblock_findval("uri", rq->reqpb);

            if( (t = strrchr(uri, '/')) ) {
                c = *(++t);
                *t = '\0';
            }

	    fulluri = (char *) MALLOC(strlen(uri)+strlen(url)+2);
	    if(url[0] != '/')  
	        util_sprintf(fulluri, "%s%s", uri, url);
	    else 
		util_sprintf(fulluri, "%s", url);
	    util_uri_parse(fulluri);	         /* strip /../, /./, etc */

            fullurl = protocol_uri2url_dynamic(fulluri, "", sn, rq);

            if(t) 
                *t = c;
            pblock_nvinsert("url", fullurl, rq->vars);
            protocol_status(sn, rq, PROTOCOL_REDIRECT, NULL);

	    FREE(fulluri);
            FREE(fullurl);

            return REQ_ABORTED;
        }  else  {
            /* Emulate old httpd behavior. Bad? */
            request_restart_internal(url, rq);
            return REQ_RESTART;
        }
    }
}


int pointinpoly(double where[2], double target[MAXVERTS][2])
{
    int totalv = 0;
    int intersects = 0;
    double wherex = where[X];
    double wherey = where[Y];
    double xval, yval;
    double *pointer;
    double *end;

    while((target[totalv][X] != -1) && (totalv < MAXVERTS))
	totalv++;
    xval = target[totalv - 1][X]; 
    yval = target[totalv - 1][Y];  
    end = target[totalv];

    pointer = (double *) target+1;

    if ((yval >= wherey) != (*pointer >= wherey)) 
        if ((xval >= wherex) == (*(double *) target >= wherex)) 
	    intersects += (xval >= wherex);
        else 
            intersects += (xval - (yval - wherey) * 
                          (*(double *) target - xval) /
                          (*pointer - yval)) >= wherex;

    while(pointer < end)  {
        yval = *pointer;
        pointer += 2;

        if(yval >= wherey)  {
	   while((pointer < end) && (*pointer >= wherey))
		pointer+=2;
	   if (pointer >= end)
		break;
           if( (*(pointer-3) >= wherex) == (*(pointer-1) >= wherex) )
	        intersects += (*(pointer-3) >= wherex);
 	    else
		intersects += (*(pointer-3) - (*(pointer-2) - wherey) *
                              (*(pointer-1) - *(pointer-3)) /
                              (*pointer - *(pointer - 2))) >= wherex;

        }  else  {
	   while((pointer < end) && (*pointer < wherey))
		pointer+=2;
	   if (pointer >= end)
		break;
           if( (*(pointer-3) >= wherex) == (*(pointer-1) >= wherex) )
		intersects += (*(pointer-3) >= wherex);
 	   else
	        intersects += (*(pointer-3) - (*(pointer-2) - wherey) *
                              (*(pointer-1) - *(pointer-3)) /
                              (*pointer - *(pointer - 2))) >= wherex;
        }  
    }
    return (intersects & 1);
}


int service_imagemap(pblock *pb, Session *sn, Request *rq)
{
    char *coord = pblock_findval("query", rq->reqpb), *t;
    char *path = pblock_findval("path", rq->vars);
    int ncsa_compat = pblock_findval("ncsa-compat", pb) ? 1 : 0;
    double testpoint[2], pointarray[MAXVERTS][2], closepoint, len, ftmp;
    char input[MAXLINE], type[MAXLINE], url[MAXLINE], def[MAXLINE], num[16];
    filebuf_t *buf;
    SYS_FILE fd;
    int ln, i, j, k, eof;


	/* return NOACTION, if the request is for CM  (i.e. /x/y.img?properties ) */
	if ( coord && service_is_CM_req( coord ) )
	  return REQ_NOACTION;
	
    param_free(pblock_remove("content-type", rq->srvhdrs));

    if(!coord) {
        imap_die(IMAP_BADREQ, sn, rq);
        return REQ_PROCEED;
    }
    for(t = coord; *t != ','; ++t) {
        if((!*t) || (!isdigit(*t))) {
            imap_die(IMAP_BADREQ, sn, rq);
            return REQ_PROCEED;
        }
    }
    *t++ = '\0';
    testpoint[X] = atoi(coord);

    coord = t;
    while(*t) {
        if(!isdigit(*t)) {
            imap_die(IMAP_BADREQ, sn, rq);
            return REQ_PROCEED;
        }
        ++t;
    }
    testpoint[Y] = atoi(coord);

    if((fd = system_fopenRO(path)) == SYS_ERROR_FD) {
        protocol_status(sn, rq, 
                    file_notfound() ? PROTOCOL_NOT_FOUND : PROTOCOL_FORBIDDEN,
                    NULL);
        log_error(LOG_WARN, "imagemap", sn, rq, "error opening %s (%s)", path,
                  system_errmsg());
        return REQ_ABORTED;
    }
    if(!(buf = filebuf_open(fd, FILE_BUFFERSIZE))) {
        protocol_status(sn, rq, PROTOCOL_FORBIDDEN, NULL);
        log_error(LOG_WARN,"imagemap", sn, rq, 
                  "error opening buffer from %s (%s)", path, 
                  system_errmsg());
        return REQ_ABORTED;
    }
    ln = 1;
    def[0] = '\0';
    closepoint = -1.0;

    for(ln = 1, eof = 0; (!eof); ++ln) {
        if( (eof = util_getline(buf, ln, MAXLINE, input)) == -1) {
            filebuf_close(buf);
            log_error(LOG_WARN, "imagemap", sn, rq, 
                      "cannot read imagemap file %s (%s)", path, input);
            return REQ_ABORTED;
        }
        if((!input[0]) || (input[0] == '#'))
            continue;

        type[0] = '\0'; url[0] = '\0';

        for(i=0;isname(input[i]) && (input[i]);i++)
            type[i] = input[i];
        type[i] = '\0';

        while(isspace(input[i])) ++i;
        for(j=0;input[i] && isname(input[i]);++i,++j)
            url[j] = input[i];
        url[j] = '\0';

        if((closepoint < 0) && (!strcmp(type,"default"))) {
            strcpy(def,url);
            continue;
        }
        k=0;
        while (input[i]) {
            while (isspace(input[i]) || input[i] == ',')
                i++;
            j = 0;
            while (isdigit(input[i]))
                num[j++] = input[i++];
            num[j] = '\0';
            if (num[0] != '\0')
                pointarray[k][X] = (double) atoi(num);
            else
                break;
            while (isspace(input[i]) || input[i] == ',')
                i++;
            j = 0;
            while (isdigit(input[i]))
                num[j++] = input[i++];
            num[j] = '\0';
            if (num[0] != '\0')
                pointarray[k++][Y] = (double) atoi(num);
            else {
                filebuf_close(buf);
                imap_die("Missing y value.", sn, rq);
                return REQ_PROCEED;
            }
        }
        pointarray[k][X] = -1;
        if(!strcasecmp(type,"poly")) {
            if(pointinpoly(testpoint,pointarray)) {
                filebuf_close(buf);
                return imap_handle(url, sn, rq, path, ncsa_compat);
            }
        }
        else if(!strcasecmp(type,"circle")) {
            double radius1, radius2, t;

            t = pointarray[0][Y] - pointarray[1][Y];
            radius1 = t * t;
            t = pointarray[0][X] - pointarray[1][X];
            radius1 += t * t;

            t = (pointarray[0][Y] - testpoint[Y]);
            radius2 = t * t;
            t = (pointarray[0][X] - testpoint[X]);
            radius2 += t * t;

            if(radius2 <= radius1) {
                filebuf_close(buf);
                return imap_handle(url, sn, rq, path, ncsa_compat);
            }
        }
        else if(!strcasecmp(type,"rect")) {
            if((testpoint[X] >= pointarray[0][X]) && 
               (testpoint[X] <= pointarray[1][X]) &&
               (testpoint[Y] >= pointarray[0][Y]) && 
               (testpoint[Y] <= pointarray[1][Y]))
            {
                filebuf_close(buf);
                return imap_handle(url, sn, rq, path, ncsa_compat);
            }
        }
        else if(!strcasecmp(type, "point")) {
            ftmp = (testpoint[X] - pointarray[0][X]);
            len = ftmp*ftmp;
            ftmp = (testpoint[Y] - pointarray[0][Y]);
            len += ftmp*ftmp;
            if((closepoint < 0) || (len < closepoint)) {
                closepoint = len;
                strcpy(def, url);
            }
        }
    }
    filebuf_close(buf);
    if(def[0])
        return imap_handle(def, sn, rq, path, ncsa_compat);
#if 0
    imap_die("No default URL specified in map file. Sorry.", sn, rq);
#else
    protocol_status(sn, rq, PROTOCOL_NO_RESPONSE, NULL);
    protocol_start_response(sn, rq);
#endif
    return REQ_PROCEED;
}



/* returns true, if the query is for content manager.
 * XXXXdp, we need to do a better job here!
 */
static int service_is_CM_req( char * query )
{
  if ( !strcmp( query, CM_PROPERTY_STR ) ||  !strcmp( query, CM_SYS_PROP_NS ) ||
	   !strcmp( query, CM_TOC_NS )       ||  !strcmp( query, CM_VER_INFO_NS ) ||
	   !strcmp( query, CM_LINK_INFO_NS ) ||  !strcmp( query, CM_CUSTOM_FEILD_NS ) ||
	   !strcmp( query, CM_USR_PROP_NS )  )
	return 1;

  return 0;
}





/* ------------------------- service_disable_type ------------------------- */


int service_disable_type(pblock *pb, Session *sn, Request *rq)
{
    char *type = pblock_findval("content-type", rq->srvhdrs);

    log_error(LOG_WARN, "disable-type", sn, rq,
              "server not configured to allow the MIME type %s from this resource", type ? type : "(none)");
    return REQ_ABORTED;
}
/* ------------------------- service_keytoosmall -------------------------- */


#ifdef NET_SSL

#define CRYPT_HEADER "Insufficient encryption"

int service_keytoosmall(pblock *pb, Session *sn, Request *rq)
{
    char buf[MAGNUS_ERROR_LEN];
    int l;

    param_free(pblock_remove("content-type", rq->srvhdrs));
    pblock_nvinsert("content-type", "text/html", rq->srvhdrs);

    protocol_status(sn, rq, PROTOCOL_FORBIDDEN, NULL);
    protocol_start_response(sn, rq);

    l = util_snprintf(buf, MAGNUS_ERROR_LEN, "<title>%s</title><h1>%s</h1>\n", CRYPT_HEADER,
                     CRYPT_HEADER);
    l += util_snprintf(&buf[l], MAGNUS_ERROR_LEN - l, "This document requires a larger secret key size for encryption than your browser is capable of supporting.\n");

    net_write(sn->csd, buf, l);
    return REQ_PROCEED;
}
#endif

#include "httpdaemon/internalstats.h"

int
service_dumpstats(pblock *pb, Session *sn, Request *rq)
{
    PRUint32 numSockets;
    listenSocketInfo *listenSocketData;
    keepAliveInfo keepAliveData;
    cacheInfo cacheData;
    DNSCacheInfo dnsCacheData;
    PRBool listenSocketDataReady;
    PRBool keepAliveDataReady;
    PRBool cacheDataReady;
    PRBool dnsCacheDataReady;
#define INFOBUF_LEN	4096
    char infobuf[INFOBUF_LEN];
    int infobuflen;
    int index;
    char *refresh, *refresh_val = NULL;
    char *cookie;
    int cookiedate;
#define COOKIEMAX 128
    char newcookie[COOKIEMAX];

    /* See if client asked for automatic refresh in query string */
    if ((refresh = pblock_findval("query", rq->reqpb)) != NULL ) {
        if (!strncmp("refresh", refresh, 7)) {
             refresh_val = strchr(refresh, '=');
             if (refresh_val)
                 refresh_val++;
        }
    }

    param_free(pblock_remove("content-type", rq->srvhdrs));
    pblock_nvinsert("content-type", "text/plain", rq->srvhdrs);
    if (refresh_val)
        pblock_nvinsert("refresh", refresh_val, rq->srvhdrs);
    protocol_status(sn, rq, PROTOCOL_OK, NULL);
    protocol_start_response(sn,rq);

    listenSocketDataReady = GetListenSocketInfo(&numSockets, &listenSocketData);
    keepAliveDataReady = GetKeepAliveInfo(&keepAliveData);
    cacheDataReady = GetCacheInfo(&cacheData);
    dnsCacheDataReady = GetDNSCacheInfo(&dnsCacheData);

    /* Get listen socket info */
    if (listenSocketDataReady) {
        for(index=0; index<numSockets; index++ ) {
            infobuflen = util_snprintf(infobuf, INFOBUF_LEN, 
                   "ListenSocket #%d:\n"
                   "------------------\n"
                   "Address           https:\\\\%s:%d\n"
                   "ActiveThreads     %d\n"
                   "WaitingThreads    %d\n"
                   "BusyThreads       %d\n"
                   "IdleThreads       %d\n\n"
                   "Thread limits     %d/%d\n",
                index,
                listenSocketData[index].ipAddr?
                    listenSocketData[index].ipAddr:"INADDR_ANY",
                listenSocketData[index].port,
                listenSocketData[index].numActiveThreads,
                listenSocketData[index].numWaitingThreads,
                listenSocketData[index].numBusyThreads,
                listenSocketData[index].numIdleThreads,
                listenSocketData[index].numMinThreads,
                listenSocketData[index].numMaxThreads);
            net_write(sn->csd, infobuf, infobuflen);
        }
    }
    FREE(listenSocketData);

    /* Get keepalive info */
    if (keepAliveDataReady) {
        infobuflen = util_snprintf(infobuf, INFOBUF_LEN, 
               "\nKeepAliveInfo:\n"
               "------------------\n"
               "KeepAliveCount        %d/%d\n"
               "KeepAliveHits         %d\n"
               "KeepAliveFlushes      %d\n",
               keepAliveData.numKeepAlives,
               keepAliveData.maxKeepAlives,
               keepAliveData.numKeepAliveHits,
               keepAliveData.numKeepAliveFlushes);
        net_write(sn->csd, infobuf, infobuflen);
    }

    /* Get cache info */
    if (cacheDataReady) {
        if (cacheData.enabled) {
            PRUint32 totalhits =
                (cacheData.numCacheMisses+cacheData.numCacheHits);
            float hitratio;

            if (totalhits > 0.0)
                 hitratio = (float)cacheData.numCacheHits / (float)totalhits;
            else
                 hitratio = 0.0;
            infobuflen = util_snprintf(infobuf, INFOBUF_LEN, 
                   "\nCacheInfo:\n"
                   "------------------\n"
                   "enabled             yes\n"
                   "CacheEntries        %d/%d\n"
#ifdef MCC_DEBUG
                   "OpenCacheEntries    %d/%d\n"
#endif
                   "CacheSize(bytes)    %d/%d\n\n"
                   "Hit Ratio           %d/%d (%6.2f)\n"
#ifdef MCC_DEBUG
                   "numCacheInsertsOk   %d\n"
                   "numCacheInsertsFail %d\n"
                   "numCacheDeletes     %d\n\n"
#endif
                   "pollInterval        %d\n"
                   "maxFileSize         %d\n",
                   cacheData.numCacheEntries,
                   cacheData.maxCacheEntries,
#ifdef MCC_DEBUG
                   cacheData.numOpenCacheEntries,
                   cacheData.maxOpenCacheEntries,
#endif
                   cacheData.maxCurrentTotalCacheSize,
                   cacheData.maxTotalCacheSize,
                   cacheData.numCacheHits,
                   totalhits,
                   hitratio * 100.0,
#ifdef MCC_DEBUG
                   cacheData.numCacheInsertsOk,
                   cacheData.numCacheInsertsFail,
                   cacheData.numCacheDeletes,
#endif
                   cacheData.pollInterval,
                   cacheData.maxFileSize);
            net_write(sn->csd, infobuf, infobuflen);
        } else {
            infobuflen = util_snprintf(infobuf, INFOBUF_LEN, 
                   "\nServer cache disabled\n");
            net_write(sn->csd, infobuf, infobuflen);
        }
    }
    /* Get dns cache info */
    if (dnsCacheDataReady) {
        if (dnsCacheData.enabled) {
            PRUint32 totalhits =
                dnsCacheData.numCacheMisses+dnsCacheData.numCacheHits;
            float hitratio;

            if (totalhits > 0.0)
                 hitratio = (float)dnsCacheData.numCacheHits / (float)totalhits;
            else
                 hitratio = 0.0;
            infobuflen = util_snprintf(infobuf, INFOBUF_LEN, 
                   "\nDNSCacheInfo:\n"
                   "------------------\n"
                   "enabled             yes\n"
                   "CacheEntries        %d/%d\n\n"
                   "HitRatio            %d/%d (%6.2f)\n"
#ifdef MCC_DEBUG
                   "numCacheInsertsOk   %d\n"
                   "numCacheInsertsFail %d\n"
                   "numCacheDeletes     %d\n\n",
#else
                   ,
#endif
                   dnsCacheData.numCacheEntries,
                   dnsCacheData.maxCacheEntries,
                   dnsCacheData.numCacheHits,
                   totalhits,
#ifdef MCC_DEBUG
                   hitratio * 100.0,
                   dnsCacheData.numCacheInsertsOk,
                   dnsCacheData.numCacheInsertsFail,
                   dnsCacheData.numCacheDeletes);
#else
                   hitratio * 100.0 );
#endif
            net_write(sn->csd, infobuf, infobuflen);
        } else {
            infobuflen = util_snprintf(infobuf, INFOBUF_LEN, 
                   "\nServer DNS cache disabled\n");
            net_write(sn->csd, infobuf, infobuflen);
        }
    }

    return REQ_PROCEED;
}

