/** BEGIN COPYRIGHT BLOCK
 * Copyright (C) 2001 Sun Microsystems, Inc.  Used by permission.
 * Copyright (C) 2005 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/
/*
 * SubSystem loading functions.
 *
 * Romy Ramil (08/21/96)
 */

#include "base/subsystem.h"
#include "base/util.h"
#include "safs/dl.h"
#include "base/dll.h"
#include "oad_c.hh"


/* Global Subsytem List */
subsystem_list *subs_list=NULL;

/*
 * load_subsystem():  08/19/96  RR
 *
 * Extract the subsystem info from the configuration file and register it with the
 ( Object Activation Daemon (OAD)
 */
int reg_subsystem(pblock *pb, Session *sn, Request *rq)
{
    char *s_name = pblock_findval("sname", pb);		/* Subsystem name */
	char *s_obj  = pblock_findval("sobj", pb);		/* Object name */
	char *s_proc = pblock_findval("sproc", pb);		/* IN or OUT process */
    char *s_path = pblock_findval("spath", pb);		/* Subsystem path name */
    char *s_arg = pblock_findval("sargs", pb);		/* Optional command argument */
	char *s_env = pblock_findval("senv", pb);		/* Optional environment variable */
													/* to be passed */
    char c, err[MAGNUS_ERROR_LEN];
	int oargc = 0;

#ifdef _ORB_ON_

	/* Initialize ORB and BOA */
	CORBA::ORB_ptr orb = CORBA::ORB_init(oargc, NULL);
	CORBA::BOA_ptr boa = orb->BOA_init(oargc, NULL);

	CORBA::ReferenceData id;
	/* Optionally initialize id to desired value */

	CORBA::CreationImplDef impl_def(s_name, s_obj,
									id,
									s_path,
									(CORBA::StringSequence&)s_arg,		/* Can be used to pass a key(principal) for */
												/* security purposes */
									(CORBA::StringSequence&)s_env );
	Activation::OAD_var	oad;
#endif /*_ORB_ON */

	/*
	 * Load subsystem methods 
	 */
	if ( !init_subsystem(pb, sn, rq) ){
		return REQ_ABORTED;
	}

#ifdef _ORB_ON_
	if (strcmp(s_proc,"IN") == 0) {
		if (dl_init(pb,sn,rq) == REQ_ABORTED)
			return REQ_ABORTED;
		goto add_to_sublist;
	}

	if (!s_name || !s_path || !s_obj) { /* Required */
		pblock_nvinsert("error", "missing parameter (need sname and spath)",pb);
		return REQ_ABORTED;
	}
	try {
		oad = Activation::OAD::_bind();
	}
	catch (const CORBA::Exception&excep){
		util_sprintf(err, "Unable to bind to OAD for subsystem <%s>.",s_name);
		pblock_nvinsert("error",err, pb);
		return REQ_ABORTED;
	}

	try {
		oad->reg_implementation(&impl_def);
	}
	catch(Activation::DuplicateEntry&excep){}
	catch(const CORBA::Exception&excep){
        util_sprintf(err, "Unable to register with OAD for subsystem <%s>.",s_name);
        pblock_nvinsert("error",err, pb);
        return REQ_ABORTED;
    }
#endif _ORB_ON_

add_to_sublist:
	/*
	 * Add subsystem to master list.
	 */
	if ( !subs_list ) {
		subs_list = new subsystem_list;
		magnus_atrestart(unreg_subsystems, NULL);
	}
	if ( !subs_list->add_to_list(s_name, s_obj) )
        return REQ_ABORTED;
		
	return REQ_PROCEED;
}

NSAPI_PUBLIC void 
unreg_subsystems( void * param)
{
	int oargc = 0;
   /* Initialize ORB and BOA */
    CORBA::ORB_ptr orb = CORBA::ORB_init(oargc, NULL);
    CORBA::BOA_ptr boa = orb->BOA_init(oargc, NULL);
 
    Activation::OAD_var oad;
 
    try {
        oad = Activation::OAD::_bind();
    }
    catch (...){
    }

	try {
		for(int i = 0;i<subs_list->_scount(); i++) 
			oad->unreg_implementation( subs_list->_sname(i), subs_list->_oname(i));
	}
	catch(...){
	}

	delete subs_list;
}

NSAPI_PUBLIC int
init_subsystem( pblock * pb, Session *sn, Request *rq)
{
	char *s_lib  = pblock_findval("shlib", pb);		/* shared library */
	char *s_init = pblock_findval("func", pb);		/* Load function */
	char err[MAGNUS_ERROR_LEN];
    DLHANDLE dlp;
    FuncPtr tfn;

	if ( !s_lib || !s_init ) {
        pblock_nvinsert("error","missing parameter (need shlib and funcs)",pb);
		return 0;
	} 

    if(!(dlp = dll_open(s_lib))) {
    	char* dllerr = dll_error(dlp);
        util_sprintf(err, "dlopen of %s failed (%s)", s_lib, dllerr);
        pblock_nvinsert("error", err, pb);
		return 0;
	}

	tfn = (FuncPtr) dll_findsym(dlp, s_init);

	if ( !tfn ) {
 		util_sprintf(err, "dll_findsym for %s failed (%s)", s_init, dll_error(dlp));
      	pblock_nvinsert("error", err, pb);
      	FREE(s_init);
		return 0;
	}

	/* Execute subsystem loading function */
	(tfn)(pb, sn, rq);
	return 1;
}

NSAPI_PUBLIC void
load_subsystem ( sub_methods_t *sublist )
{
	while(sublist) {	
		if ( !sublist->fname )
			break;
		func_insert(sublist->fname, sublist->fptr);
		sublist++;
	}	
}
